! mod_combinedlibgrey.f90
! Grey Gordon 2010-current
! Soli Deo Gloria
! 
! This is a stripped down version of libgrey combined into one .f90 file. The only requirements are BLAS and LAPACK routines.
! 
! Proper attributions have been attempted. If I missed citing something or lack a license, please let me know.


subroutine normal_01_cdf_values ( n_data, x, fx )

!*****************************************************************************80
!
!! NORMAL_01_CDF_VALUES returns some values of the Normal 01 CDF.
!
!  Discussion:
!
!    In Mathematica, the function can be evaluated by:
!
!      Needs["Statistics`ContinuousDistributions`"]
!      dist = NormalDistribution [ 0, 1 ]
!      CDF [ dist, x ]
!
!  Licensing:
!
!    This code is distributed under the GNU LGPL license.
!
!  Modified:
!
!    28 August 2004
!
!  Author:
!
!    John Burkardt
!
!  Reference:
!
!    Milton Abramowitz and Irene Stegun,
!    Handbook of Mathematical Functions,
!    US Department of Commerce, 1964.
!
!    Stephen Wolfram,
!    The Mathematica Book,
!    Fourth Edition,
!    Wolfram Media / Cambridge University Press, 1999.
!
!  Parameters:
!
!    Input/output, integer ( kind = 4 ) N_DATA.  The user sets N_DATA to 0 
!    before the first call.  On each call, the routine increments N_DATA by 1, 
!    and returns the corresponding data; when there is no more data, the
!    output value of N_DATA will be 0 again.
!
!    Output, real ( kind = 8 ) X, the argument of the function.
!
!    Output, real ( kind = 8 ) FX, the value of the function.
!
  implicit none

  integer ( kind = 4 ), parameter :: n_max = 17

  real ( kind = 8 ) fx
  real ( kind = 8 ), save, dimension ( n_max ) :: fx_vec = (/ &
    0.5000000000000000D+00, &
    0.5398278372770290D+00, &
    0.5792597094391030D+00, &
    0.6179114221889526D+00, &
    0.6554217416103242D+00, &
    0.6914624612740131D+00, &
    0.7257468822499270D+00, &
    0.7580363477769270D+00, &
    0.7881446014166033D+00, &
    0.8159398746532405D+00, &
    0.8413447460685429D+00, &
    0.9331927987311419D+00, &
    0.9772498680518208D+00, &
    0.9937903346742239D+00, &
    0.9986501019683699D+00, &
    0.9997673709209645D+00, &
    0.9999683287581669D+00 /)
  integer ( kind = 4 ) n_data
  real ( kind = 8 ) x
  real ( kind = 8 ), save, dimension ( n_max ) :: x_vec = (/ &
    0.0000000000000000D+00, &
    0.1000000000000000D+00, &
    0.2000000000000000D+00, &
    0.3000000000000000D+00, &
    0.4000000000000000D+00, &
    0.5000000000000000D+00, &
    0.6000000000000000D+00, &
    0.7000000000000000D+00, &
    0.8000000000000000D+00, &
    0.9000000000000000D+00, &
    0.1000000000000000D+01, &
    0.1500000000000000D+01, &
    0.2000000000000000D+01, &
    0.2500000000000000D+01, &
    0.3000000000000000D+01, &
    0.3500000000000000D+01, &
    0.4000000000000000D+01 /)

  if ( n_data < 0 ) then
    n_data = 0
  end if

  n_data = n_data + 1

  if ( n_max < n_data ) then
    n_data = 0
    x = 0.0D+00
    fx = 0.0D+00
  else
    x = x_vec(n_data)
    fx = fx_vec(n_data)
  end if

  return
end
function r4_normal_01_cdf_inverse ( p )

!*****************************************************************************80
!
!! R4_NORMAL_01_CDF_INVERSE inverts the standard normal CDF.
!
!  Discussion:
!
!    The result is accurate to about 1 part in 10**7.
!
!  Licensing:
!
!    This code is distributed under the GNU LGPL license.
!
!  Modified:
!
!    27 December 2004
!
!  Author:
!
!    Original FORTRAN77 version by Michael Wichura.
!    FORTRAN90 version by John Burkardt.
!
!  Reference:
!
!    Michael Wichura,
!    The Percentage Points of the Normal Distribution,
!    Algorithm AS 241,
!    Applied Statistics,
!    Volume 37, Number 3, pages 477-484, 1988.
!
!  Parameters:
!
!    Input, real ( kind = 4 ) P, the value of the cumulative probability 
!    densitity function.  0 < P < 1.  If P is outside this range, an 
!    "infinite" value will be returned. 
!
!    Output, real ( kind = 4 ) R4_NORMAL_01_CDF_INVERSE, the normal deviate 
!    value with the property that the probability of a standard normal deviate 
!    being less than or equal to this value is P.
!
  implicit none

  real ( kind = 4 ), parameter, dimension ( 4 ) :: a = (/ &
    3.3871327179E+00, 50.434271938E+00, 159.29113202E+00, 59.109374720E+00 /)
  real ( kind = 4 ), parameter, dimension ( 4 ) :: b = (/ &
    1.0E+00, 17.895169469E+00, 78.757757664E+00, 67.187563600E+00 /)
  real ( kind = 4 ), parameter, dimension ( 4 ) :: c = (/ &
    1.4234372777E+00, 2.7568153900E+00, 1.3067284816E+00, 0.17023821103E+00 /)
  real ( kind = 4 ), parameter :: const1 = 0.180625E+00
  real ( kind = 4 ), parameter :: const2 = 1.6E+00
  real ( kind = 4 ), parameter, dimension ( 3 ) :: d = (/ &
    1.0E+00, 0.73700164250E+00, 0.12021132975E+00 /)
  real ( kind = 4 ), parameter, dimension ( 4 ) :: e = (/ &
    6.6579051150E+00, 3.0812263860E+00, 0.42868294337E+00, 0.017337203997E+00 /)
  real ( kind = 4 ), parameter, dimension ( 3 ) :: f = (/ &
    1.0E+00, 0.24197894225E+00, 0.012258202635E+00 /)
  real ( kind = 4 ) p
  real ( kind = 4 ) q
  real ( kind = 4 ) r
  real ( kind = 4 ) r4_normal_01_cdf_inverse
  real ( kind = 4 ) r4poly_value
  real ( kind = 4 ), parameter :: split1 = 0.425E+00
  real ( kind = 4 ), parameter :: split2 = 5.0E+00

  if ( p <= 0.0E+00 ) then
    r4_normal_01_cdf_inverse = - huge ( p )
    return
  end if

  if ( 1.0E+00 <= p ) then
    r4_normal_01_cdf_inverse = huge ( p )
    return
  end if

  q = p - 0.5E+00

  if ( abs ( q ) <= split1 ) then

    r = const1 - q * q
    r4_normal_01_cdf_inverse = q * r4poly_value ( 4, a, r ) &
                                 / r4poly_value ( 4, b, r )

  else

    if ( q < 0.0E+00 ) then
      r = p
    else
      r = 1.0E+00 - p
    end if

    if ( r <= 0.0E+00 ) then
      r4_normal_01_cdf_inverse = -1.0E+00
      stop
    end if

    r = sqrt ( -log ( r ) )

    if ( r <= split2 ) then

      r = r - const2
      r4_normal_01_cdf_inverse = r4poly_value ( 4, c, r ) &
                               / r4poly_value ( 3, d, r )

    else

      r = r - split2
      r4_normal_01_cdf_inverse = r4poly_value ( 4, e, r ) &
                               / r4poly_value ( 3, f, r )

    end if

    if ( q < 0.0E+00 ) then
      r4_normal_01_cdf_inverse = - r4_normal_01_cdf_inverse
    end if

  end if

  return
end
function r4poly_value ( n, a, x )

!*****************************************************************************80
!
!! R4POLY_VALUE evaluates an R4POLY.
!
!  Discussion:
!
!    For sanity's sake, the value of N indicates the NUMBER of 
!    coefficients, or more precisely, the ORDER of the polynomial,
!    rather than the DEGREE of the polynomial.  The two quantities
!    differ by 1, but cause a great deal of confusion.
!
!    Given N and A, the form of the polynomial is:
!
!      p(x) = a(1) + a(2) * x + ... + a(n-1) * x^(n-2) + a(n) * x^(n-1)
!
!  Licensing:
!
!    This code is distributed under the GNU LGPL license.
!
!  Modified:
!
!    13 August 2004
!
!  Author:
!
!    John Burkardt
!
!  Parameters:
!
!    Input, integer ( kind = 4 ) N, the order of the polynomial.
!
!    Input, real ( kind = 4 ) A(N), the coefficients of the polynomial.
!    A(1) is the constant term.
!
!    Input, real ( kind = 4 ) X, the point at which the polynomial is 
!    to be evaluated.
!
!    Output, real ( kind = 4 ) R4POLY_VALUE, the value of the polynomial at X.
!
  implicit none

  integer ( kind = 4 ) n

  real ( kind = 4 ) a(n)
  integer ( kind = 4 ) i
  real ( kind = 4 ) r4poly_value
  real ( kind = 4 ) x

  r4poly_value = 0.0E+00
  do i = n, 1, -1
    r4poly_value = r4poly_value * x + a(i)
  end do

  return
end
function r8_normal_01_cdf_inverse ( p )

!*****************************************************************************80
!
!! R8_NORMAL_01_CDF_INVERSE inverts the standard normal CDF.
!
!  Discussion:
!
!    The result is accurate to about 1 part in 10**16.
!
!  Licensing:
!
!    This code is distributed under the GNU LGPL license.
!
!  Modified:
!
!    27 December 2004
!
!  Author:
!
!    Original FORTRAN77 version by Michael Wichura.
!    FORTRAN90 version by John Burkardt.
!
!  Reference:
!
!    Michael Wichura,
!    The Percentage Points of the Normal Distribution,
!    Algorithm AS 241,
!    Applied Statistics,
!    Volume 37, Number 3, pages 477-484, 1988.
!
!  Parameters:
!
!    Input, real ( kind = 8 ) P, the value of the cumulative probability 
!    densitity function.  0 < P < 1.  If P is outside this range,
!    an "infinite" value will be returned.
!
!    Output, real ( kind = 8 ) D_NORMAL_01_CDF_INVERSE, the normal deviate 
!    value with the property that the probability of a standard normal 
!    deviate being less than or equal to the value is P.
!
  implicit none

  real ( kind = 8 ), parameter, dimension ( 8 ) :: a = (/ &
    3.3871328727963666080D+00, &
    1.3314166789178437745D+02, &
    1.9715909503065514427D+03, &
    1.3731693765509461125D+04, &
    4.5921953931549871457D+04, &
    6.7265770927008700853D+04, &
    3.3430575583588128105D+04, &
    2.5090809287301226727D+03 /)
  real ( kind = 8 ), parameter, dimension ( 8 ) :: b = (/ &
    1.0D+00, &
    4.2313330701600911252D+01, &
    6.8718700749205790830D+02, &
    5.3941960214247511077D+03, &
    2.1213794301586595867D+04, &
    3.9307895800092710610D+04, &
    2.8729085735721942674D+04, &
    5.2264952788528545610D+03 /)
  real   ( kind = 8 ), parameter, dimension ( 8 ) :: c = (/ &
    1.42343711074968357734D+00, &
    4.63033784615654529590D+00, &
    5.76949722146069140550D+00, &
    3.64784832476320460504D+00, &
    1.27045825245236838258D+00, &
    2.41780725177450611770D-01, &
    2.27238449892691845833D-02, &
    7.74545014278341407640D-04 /)
  real ( kind = 8 ), parameter :: const1 = 0.180625D+00
  real ( kind = 8 ), parameter :: const2 = 1.6D+00
  real ( kind = 8 ), parameter, dimension ( 8 ) :: d = (/ &
    1.0D+00, &
    2.05319162663775882187D+00, &
    1.67638483018380384940D+00, &
    6.89767334985100004550D-01, &
    1.48103976427480074590D-01, &
    1.51986665636164571966D-02, &
    5.47593808499534494600D-04, &
    1.05075007164441684324D-09 /)
  real ( kind = 8 ), parameter, dimension ( 8 ) :: e = (/ &
    6.65790464350110377720D+00, &
    5.46378491116411436990D+00, &
    1.78482653991729133580D+00, &
    2.96560571828504891230D-01, &
    2.65321895265761230930D-02, &
    1.24266094738807843860D-03, &
    2.71155556874348757815D-05, &
    2.01033439929228813265D-07 /)
  real ( kind = 8 ), parameter, dimension ( 8 ) :: f = (/ &
    1.0D+00, &
    5.99832206555887937690D-01, &
    1.36929880922735805310D-01, &
    1.48753612908506148525D-02, &
    7.86869131145613259100D-04, &
    1.84631831751005468180D-05, &
    1.42151175831644588870D-07, &
    2.04426310338993978564D-15 /)
  real ( kind = 8 ) p
  real ( kind = 8 ) q
  real ( kind = 8 ) r
  real ( kind = 8 ) r8_normal_01_cdf_inverse 
  real ( kind = 8 ) r8poly_value
  real ( kind = 8 ), parameter :: split1 = 0.425D+00
  real ( kind = 8 ), parameter :: split2 = 5.0D+00

  if ( p <= 0.0D+00 ) then
    r8_normal_01_cdf_inverse = - huge ( p )
    return
  end if

  if ( 1.0D+00 <= p ) then
    r8_normal_01_cdf_inverse = huge ( p )
    return
  end if

  q = p - 0.5D+00

  if ( abs ( q ) <= split1 ) then

    r = const1 - q * q
    r8_normal_01_cdf_inverse = q * r8poly_value ( 8, a, r ) &
                                 / r8poly_value ( 8, b, r )

  else

    if ( q < 0.0D+00 ) then
      r = p
    else
      r = 1.0D+00 - p
    end if

    if ( r <= 0.0D+00 ) then
      r8_normal_01_cdf_inverse = - 1.0D+00
      stop
    end if

    r = sqrt ( -log ( r ) )

    if ( r <= split2 ) then

      r = r - const2
      r8_normal_01_cdf_inverse = r8poly_value ( 8, c, r ) &
                               / r8poly_value ( 8, d, r )

    else

      r = r - split2
      r8_normal_01_cdf_inverse = r8poly_value ( 8, e, r ) &
                               / r8poly_value ( 8, f, r )
   
    end if

    if ( q < 0.0D+00 ) then
      r8_normal_01_cdf_inverse = - r8_normal_01_cdf_inverse
    end if

  end if

  return
end
function r8poly_value ( n, a, x )

!*****************************************************************************80
!
!! R8POLY_VALUE evaluates an R8POLY
!
!  Discussion:
!
!    For sanity's sake, the value of N indicates the NUMBER of 
!    coefficients, or more precisely, the ORDER of the polynomial,
!    rather than the DEGREE of the polynomial.  The two quantities
!    differ by 1, but cause a great deal of confusion.
!
!    Given N and A, the form of the polynomial is:
!
!      p(x) = a(1) + a(2) * x + ... + a(n-1) * x^(n-2) + a(n) * x^(n-1)
!
!  Licensing:
!
!    This code is distributed under the GNU LGPL license.
!
!  Modified:
!
!    13 August 2004
!
!  Author:
!
!    John Burkardt
!
!  Parameters:
!
!    Input, integer ( kind = 4 ) N, the order of the polynomial.
!
!    Input, real ( kind = 8 ) A(N), the coefficients of the polynomial.
!    A(1) is the constant term.
!
!    Input, real ( kind = 8 ) X, the point at which the polynomial is 
!    to be evaluated.
!
!    Output, real ( kind = 8 ) R8POLY_VALUE, the value of the polynomial at X.
!
  implicit none

  integer ( kind = 4 ) n

  real ( kind = 8 ) a(n)
  integer ( kind = 4 ) i
  real ( kind = 8 ) r8poly_value
  real ( kind = 8 ) x

  r8poly_value = 0.0D+00
  do i = n, 1, -1
    r8poly_value = r8poly_value * x + a(i)
  end do

  return
end
subroutine timestamp ( )

!*****************************************************************************80
!
!! TIMESTAMP prints the current YMDHMS date as a time stamp.
!
!  Example:
!
!    31 May 2001   9:45:54.872 AM
!
!  Licensing:
!
!    This code is distributed under the GNU LGPL license.
!
!  Modified:
!
!    06 August 2005
!
!  Author:
!
!    John Burkardt
!
!  Parameters:
!
!    None
!
  implicit none

  character ( len = 8 ) ampm
  integer ( kind = 4 ) d
  integer ( kind = 4 ) h
  integer ( kind = 4 ) m
  integer ( kind = 4 ) mm
  character ( len = 9 ), parameter, dimension(12) :: month = (/ &
    'January  ', 'February ', 'March    ', 'April    ', &
    'May      ', 'June     ', 'July     ', 'August   ', &
    'September', 'October  ', 'November ', 'December ' /)
  integer ( kind = 4 ) n
  integer ( kind = 4 ) s
  integer ( kind = 4 ) values(8)
  integer ( kind = 4 ) y

  call date_and_time ( values = values )

  y = values(1)
  m = values(2)
  d = values(3)
  h = values(5)
  n = values(6)
  s = values(7)
  mm = values(8)

  if ( h < 12 ) then
    ampm = 'AM'
  else if ( h == 12 ) then
    if ( n == 0 .and. s == 0 ) then
      ampm = 'Noon'
    else
      ampm = 'PM'
    end if
  else
    h = h - 12
    if ( h < 12 ) then
      ampm = 'PM'
    else if ( h == 12 ) then
      if ( n == 0 .and. s == 0 ) then
        ampm = 'Midnight'
      else
        ampm = 'AM'
      end if
    end if
  end if

  write ( *, '(i2,1x,a,1x,i4,2x,i2,a1,i2.2,a1,i2.2,a1,i3.3,1x,a)' ) &
    d, trim ( month(m) ), y, h, ':', n, ':', s, '.', mm, trim ( ampm )

  return
end
! ORDERPACK v2.0
! http://www.fortran-2000.com/rank/index.html
! To access any of these routines, one may "use" mod_orderpack


Module m_mrgrnk
Integer, Parameter :: kdp = selected_real_kind(15)
public :: mrgrnk
private :: kdp
private :: R_mrgrnk, I_mrgrnk, D_mrgrnk
interface mrgrnk
  module procedure D_mrgrnk, R_mrgrnk, I_mrgrnk
end interface mrgrnk
contains

Subroutine D_mrgrnk (XDONT, IRNGT)
! __________________________________________________________
!   MRGRNK = Merge-sort ranking of an array
!   For performance reasons, the first 2 passes are taken
!   out of the standard loop, and use dedicated coding.
! __________________________________________________________
! __________________________________________________________
      Real (kind=kdp), Dimension (:), Intent (In) :: XDONT
      Integer, Dimension (:), Intent (Out) :: IRNGT
! __________________________________________________________
      Real (kind=kdp) :: XVALA, XVALB
!
      Integer, Dimension (SIZE(IRNGT)) :: JWRKT
      Integer :: LMTNA, LMTNC, IRNG1, IRNG2
      Integer :: NVAL, IIND, IWRKD, IWRK, IWRKF, JINDA, IINDA, IINDB
!
      NVAL = Min (SIZE(XDONT), SIZE(IRNGT))
      Select Case (NVAL)
      Case (:0)
         Return
      Case (1)
         IRNGT (1) = 1
         Return
      Case Default
         Continue
      End Select
!
!  Fill-in the index array, creating ordered couples
!
      Do IIND = 2, NVAL, 2
         If (XDONT(IIND-1) <= XDONT(IIND)) Then
            IRNGT (IIND-1) = IIND - 1
            IRNGT (IIND) = IIND
         Else
            IRNGT (IIND-1) = IIND
            IRNGT (IIND) = IIND - 1
         End If
      End Do
      If (Modulo(NVAL, 2) /= 0) Then
         IRNGT (NVAL) = NVAL
      End If
!
!  We will now have ordered subsets A - B - A - B - ...
!  and merge A and B couples into     C   -   C   - ...
!
      LMTNA = 2
      LMTNC = 4
!
!  First iteration. The length of the ordered subsets goes from 2 to 4
!
      Do
         If (NVAL <= 2) Exit
!
!   Loop on merges of A and B into C
!
         Do IWRKD = 0, NVAL - 1, 4
            If ((IWRKD+4) > NVAL) Then
               If ((IWRKD+2) >= NVAL) Exit
!
!   1 2 3
!
               If (XDONT(IRNGT(IWRKD+2)) <= XDONT(IRNGT(IWRKD+3))) Exit
!
!   1 3 2
!
               If (XDONT(IRNGT(IWRKD+1)) <= XDONT(IRNGT(IWRKD+3))) Then
                  IRNG2 = IRNGT (IWRKD+2)
                  IRNGT (IWRKD+2) = IRNGT (IWRKD+3)
                  IRNGT (IWRKD+3) = IRNG2
!
!   3 1 2
!
               Else
                  IRNG1 = IRNGT (IWRKD+1)
                  IRNGT (IWRKD+1) = IRNGT (IWRKD+3)
                  IRNGT (IWRKD+3) = IRNGT (IWRKD+2)
                  IRNGT (IWRKD+2) = IRNG1
               End If
               Exit
            End If
!
!   1 2 3 4
!
            If (XDONT(IRNGT(IWRKD+2)) <= XDONT(IRNGT(IWRKD+3))) Cycle
!
!   1 3 x x
!
            If (XDONT(IRNGT(IWRKD+1)) <= XDONT(IRNGT(IWRKD+3))) Then
               IRNG2 = IRNGT (IWRKD+2)
               IRNGT (IWRKD+2) = IRNGT (IWRKD+3)
               If (XDONT(IRNG2) <= XDONT(IRNGT(IWRKD+4))) Then
!   1 3 2 4
                  IRNGT (IWRKD+3) = IRNG2
               Else
!   1 3 4 2
                  IRNGT (IWRKD+3) = IRNGT (IWRKD+4)
                  IRNGT (IWRKD+4) = IRNG2
               End If
!
!   3 x x x
!
            Else
               IRNG1 = IRNGT (IWRKD+1)
               IRNG2 = IRNGT (IWRKD+2)
               IRNGT (IWRKD+1) = IRNGT (IWRKD+3)
               If (XDONT(IRNG1) <= XDONT(IRNGT(IWRKD+4))) Then
                  IRNGT (IWRKD+2) = IRNG1
                  If (XDONT(IRNG2) <= XDONT(IRNGT(IWRKD+4))) Then
!   3 1 2 4
                     IRNGT (IWRKD+3) = IRNG2
                  Else
!   3 1 4 2
                     IRNGT (IWRKD+3) = IRNGT (IWRKD+4)
                     IRNGT (IWRKD+4) = IRNG2
                  End If
               Else
!   3 4 1 2
                  IRNGT (IWRKD+2) = IRNGT (IWRKD+4)
                  IRNGT (IWRKD+3) = IRNG1
                  IRNGT (IWRKD+4) = IRNG2
               End If
            End If
         End Do
!
!  The Cs become As and Bs
!
         LMTNA = 4
         Exit
      End Do
!
!  Iteration loop. Each time, the length of the ordered subsets
!  is doubled.
!
      Do
         If (LMTNA >= NVAL) Exit
         IWRKF = 0
         LMTNC = 2 * LMTNC
!
!   Loop on merges of A and B into C
!
         Do
            IWRK = IWRKF
            IWRKD = IWRKF + 1
            JINDA = IWRKF + LMTNA
            IWRKF = IWRKF + LMTNC
            If (IWRKF >= NVAL) Then
               If (JINDA >= NVAL) Exit
               IWRKF = NVAL
            End If
            IINDA = 1
            IINDB = JINDA + 1
!
!   Shortcut for the case when the max of A is smaller
!   than the min of B. This line may be activated when the
!   initial set is already close to sorted.
!
!          IF (XDONT(IRNGT(JINDA)) <= XDONT(IRNGT(IINDB))) CYCLE
!
!  One steps in the C subset, that we build in the final rank array
!
!  Make a copy of the rank array for the merge iteration
!
            JWRKT (1:LMTNA) = IRNGT (IWRKD:JINDA)
!
            XVALA = XDONT (JWRKT(IINDA))
            XVALB = XDONT (IRNGT(IINDB))
!
            Do
               IWRK = IWRK + 1
!
!  We still have unprocessed values in both A and B
!
               If (XVALA > XVALB) Then
                  IRNGT (IWRK) = IRNGT (IINDB)
                  IINDB = IINDB + 1
                  If (IINDB > IWRKF) Then
!  Only A still with unprocessed values
                     IRNGT (IWRK+1:IWRKF) = JWRKT (IINDA:LMTNA)
                     Exit
                  End If
                  XVALB = XDONT (IRNGT(IINDB))
               Else
                  IRNGT (IWRK) = JWRKT (IINDA)
                  IINDA = IINDA + 1
                  If (IINDA > LMTNA) Exit! Only B still with unprocessed values
                  XVALA = XDONT (JWRKT(IINDA))
               End If
!
            End Do
         End Do
!
!  The Cs become As and Bs
!
         LMTNA = 2 * LMTNA
      End Do
!
      Return
!
End Subroutine D_mrgrnk

Subroutine R_mrgrnk (XDONT, IRNGT)
! __________________________________________________________
!   MRGRNK = Merge-sort ranking of an array
!   For performance reasons, the first 2 passes are taken
!   out of the standard loop, and use dedicated coding.
! __________________________________________________________
! _________________________________________________________
      Real, Dimension (:), Intent (In) :: XDONT
      Integer, Dimension (:), Intent (Out) :: IRNGT
! __________________________________________________________
      Real :: XVALA, XVALB
!
      Integer, Dimension (SIZE(IRNGT)) :: JWRKT
      Integer :: LMTNA, LMTNC, IRNG1, IRNG2
      Integer :: NVAL, IIND, IWRKD, IWRK, IWRKF, JINDA, IINDA, IINDB
!
      NVAL = Min (SIZE(XDONT), SIZE(IRNGT))
      Select Case (NVAL)
      Case (:0)
         Return
      Case (1)
         IRNGT (1) = 1
         Return
      Case Default
         Continue
      End Select
!
!  Fill-in the index array, creating ordered couples
!
      Do IIND = 2, NVAL, 2
         If (XDONT(IIND-1) <= XDONT(IIND)) Then
            IRNGT (IIND-1) = IIND - 1
            IRNGT (IIND) = IIND
         Else
            IRNGT (IIND-1) = IIND
            IRNGT (IIND) = IIND - 1
         End If
      End Do
      If (Modulo(NVAL, 2) /= 0) Then
         IRNGT (NVAL) = NVAL
      End If
!
!  We will now have ordered subsets A - B - A - B - ...
!  and merge A and B couples into     C   -   C   - ...
!
      LMTNA = 2
      LMTNC = 4
!
!  First iteration. The length of the ordered subsets goes from 2 to 4
!
      Do
         If (NVAL <= 2) Exit
!
!   Loop on merges of A and B into C
!
         Do IWRKD = 0, NVAL - 1, 4
            If ((IWRKD+4) > NVAL) Then
               If ((IWRKD+2) >= NVAL) Exit
!
!   1 2 3
!
               If (XDONT(IRNGT(IWRKD+2)) <= XDONT(IRNGT(IWRKD+3))) Exit
!
!   1 3 2
!
               If (XDONT(IRNGT(IWRKD+1)) <= XDONT(IRNGT(IWRKD+3))) Then
                  IRNG2 = IRNGT (IWRKD+2)
                  IRNGT (IWRKD+2) = IRNGT (IWRKD+3)
                  IRNGT (IWRKD+3) = IRNG2
!
!   3 1 2
!
               Else
                  IRNG1 = IRNGT (IWRKD+1)
                  IRNGT (IWRKD+1) = IRNGT (IWRKD+3)
                  IRNGT (IWRKD+3) = IRNGT (IWRKD+2)
                  IRNGT (IWRKD+2) = IRNG1
               End If
               Exit
            End If
!
!   1 2 3 4
!
            If (XDONT(IRNGT(IWRKD+2)) <= XDONT(IRNGT(IWRKD+3))) Cycle
!
!   1 3 x x
!
            If (XDONT(IRNGT(IWRKD+1)) <= XDONT(IRNGT(IWRKD+3))) Then
               IRNG2 = IRNGT (IWRKD+2)
               IRNGT (IWRKD+2) = IRNGT (IWRKD+3)
               If (XDONT(IRNG2) <= XDONT(IRNGT(IWRKD+4))) Then
!   1 3 2 4
                  IRNGT (IWRKD+3) = IRNG2
               Else
!   1 3 4 2
                  IRNGT (IWRKD+3) = IRNGT (IWRKD+4)
                  IRNGT (IWRKD+4) = IRNG2
               End If
!
!   3 x x x
!
            Else
               IRNG1 = IRNGT (IWRKD+1)
               IRNG2 = IRNGT (IWRKD+2)
               IRNGT (IWRKD+1) = IRNGT (IWRKD+3)
               If (XDONT(IRNG1) <= XDONT(IRNGT(IWRKD+4))) Then
                  IRNGT (IWRKD+2) = IRNG1
                  If (XDONT(IRNG2) <= XDONT(IRNGT(IWRKD+4))) Then
!   3 1 2 4
                     IRNGT (IWRKD+3) = IRNG2
                  Else
!   3 1 4 2
                     IRNGT (IWRKD+3) = IRNGT (IWRKD+4)
                     IRNGT (IWRKD+4) = IRNG2
                  End If
               Else
!   3 4 1 2
                  IRNGT (IWRKD+2) = IRNGT (IWRKD+4)
                  IRNGT (IWRKD+3) = IRNG1
                  IRNGT (IWRKD+4) = IRNG2
               End If
            End If
         End Do
!
!  The Cs become As and Bs
!
         LMTNA = 4
         Exit
      End Do
!
!  Iteration loop. Each time, the length of the ordered subsets
!  is doubled.
!
      Do
         If (LMTNA >= NVAL) Exit
         IWRKF = 0
         LMTNC = 2 * LMTNC
!
!   Loop on merges of A and B into C
!
         Do
            IWRK = IWRKF
            IWRKD = IWRKF + 1
            JINDA = IWRKF + LMTNA
            IWRKF = IWRKF + LMTNC
            If (IWRKF >= NVAL) Then
               If (JINDA >= NVAL) Exit
               IWRKF = NVAL
            End If
            IINDA = 1
            IINDB = JINDA + 1
!
!   Shortcut for the case when the max of A is smaller
!   than the min of B. This line may be activated when the
!   initial set is already close to sorted.
!
!          IF (XDONT(IRNGT(JINDA)) <= XDONT(IRNGT(IINDB))) CYCLE
!
!  One steps in the C subset, that we build in the final rank array
!
!  Make a copy of the rank array for the merge iteration
!
            JWRKT (1:LMTNA) = IRNGT (IWRKD:JINDA)
!
            XVALA = XDONT (JWRKT(IINDA))
            XVALB = XDONT (IRNGT(IINDB))
!
            Do
               IWRK = IWRK + 1
!
!  We still have unprocessed values in both A and B
!
               If (XVALA > XVALB) Then
                  IRNGT (IWRK) = IRNGT (IINDB)
                  IINDB = IINDB + 1
                  If (IINDB > IWRKF) Then
!  Only A still with unprocessed values
                     IRNGT (IWRK+1:IWRKF) = JWRKT (IINDA:LMTNA)
                     Exit
                  End If
                  XVALB = XDONT (IRNGT(IINDB))
               Else
                  IRNGT (IWRK) = JWRKT (IINDA)
                  IINDA = IINDA + 1
                  If (IINDA > LMTNA) Exit! Only B still with unprocessed values
                  XVALA = XDONT (JWRKT(IINDA))
               End If
!
            End Do
         End Do
!
!  The Cs become As and Bs
!
         LMTNA = 2 * LMTNA
      End Do
!
      Return
!
End Subroutine R_mrgrnk
Subroutine I_mrgrnk (XDONT, IRNGT)
! __________________________________________________________
!   MRGRNK = Merge-sort ranking of an array
!   For performance reasons, the first 2 passes are taken
!   out of the standard loop, and use dedicated coding.
! __________________________________________________________
! __________________________________________________________
      Integer, Dimension (:), Intent (In)  :: XDONT
      Integer, Dimension (:), Intent (Out) :: IRNGT
! __________________________________________________________
      Integer :: XVALA, XVALB
!
      Integer, Dimension (SIZE(IRNGT)) :: JWRKT
      Integer :: LMTNA, LMTNC, IRNG1, IRNG2
      Integer :: NVAL, IIND, IWRKD, IWRK, IWRKF, JINDA, IINDA, IINDB
!
      NVAL = Min (SIZE(XDONT), SIZE(IRNGT))
      Select Case (NVAL)
      Case (:0)
         Return
      Case (1)
         IRNGT (1) = 1
         Return
      Case Default
         Continue
      End Select
!
!  Fill-in the index array, creating ordered couples
!
      Do IIND = 2, NVAL, 2
         If (XDONT(IIND-1) <= XDONT(IIND)) Then
            IRNGT (IIND-1) = IIND - 1
            IRNGT (IIND) = IIND
         Else
            IRNGT (IIND-1) = IIND
            IRNGT (IIND) = IIND - 1
         End If
      End Do
      If (Modulo(NVAL, 2) /= 0) Then
         IRNGT (NVAL) = NVAL
      End If
!
!  We will now have ordered subsets A - B - A - B - ...
!  and merge A and B couples into     C   -   C   - ...
!
      LMTNA = 2
      LMTNC = 4
!
!  First iteration. The length of the ordered subsets goes from 2 to 4
!
      Do
         If (NVAL <= 2) Exit
!
!   Loop on merges of A and B into C
!
         Do IWRKD = 0, NVAL - 1, 4
            If ((IWRKD+4) > NVAL) Then
               If ((IWRKD+2) >= NVAL) Exit
!
!   1 2 3
!
               If (XDONT(IRNGT(IWRKD+2)) <= XDONT(IRNGT(IWRKD+3))) Exit
!
!   1 3 2
!
               If (XDONT(IRNGT(IWRKD+1)) <= XDONT(IRNGT(IWRKD+3))) Then
                  IRNG2 = IRNGT (IWRKD+2)
                  IRNGT (IWRKD+2) = IRNGT (IWRKD+3)
                  IRNGT (IWRKD+3) = IRNG2
!
!   3 1 2
!
               Else
                  IRNG1 = IRNGT (IWRKD+1)
                  IRNGT (IWRKD+1) = IRNGT (IWRKD+3)
                  IRNGT (IWRKD+3) = IRNGT (IWRKD+2)
                  IRNGT (IWRKD+2) = IRNG1
               End If
               Exit
            End If
!
!   1 2 3 4
!
            If (XDONT(IRNGT(IWRKD+2)) <= XDONT(IRNGT(IWRKD+3))) Cycle
!
!   1 3 x x
!
            If (XDONT(IRNGT(IWRKD+1)) <= XDONT(IRNGT(IWRKD+3))) Then
               IRNG2 = IRNGT (IWRKD+2)
               IRNGT (IWRKD+2) = IRNGT (IWRKD+3)
               If (XDONT(IRNG2) <= XDONT(IRNGT(IWRKD+4))) Then
!   1 3 2 4
                  IRNGT (IWRKD+3) = IRNG2
               Else
!   1 3 4 2
                  IRNGT (IWRKD+3) = IRNGT (IWRKD+4)
                  IRNGT (IWRKD+4) = IRNG2
               End If
!
!   3 x x x
!
            Else
               IRNG1 = IRNGT (IWRKD+1)
               IRNG2 = IRNGT (IWRKD+2)
               IRNGT (IWRKD+1) = IRNGT (IWRKD+3)
               If (XDONT(IRNG1) <= XDONT(IRNGT(IWRKD+4))) Then
                  IRNGT (IWRKD+2) = IRNG1
                  If (XDONT(IRNG2) <= XDONT(IRNGT(IWRKD+4))) Then
!   3 1 2 4
                     IRNGT (IWRKD+3) = IRNG2
                  Else
!   3 1 4 2
                     IRNGT (IWRKD+3) = IRNGT (IWRKD+4)
                     IRNGT (IWRKD+4) = IRNG2
                  End If
               Else
!   3 4 1 2
                  IRNGT (IWRKD+2) = IRNGT (IWRKD+4)
                  IRNGT (IWRKD+3) = IRNG1
                  IRNGT (IWRKD+4) = IRNG2
               End If
            End If
         End Do
!
!  The Cs become As and Bs
!
         LMTNA = 4
         Exit
      End Do
!
!  Iteration loop. Each time, the length of the ordered subsets
!  is doubled.
!
      Do
         If (LMTNA >= NVAL) Exit
         IWRKF = 0
         LMTNC = 2 * LMTNC
!
!   Loop on merges of A and B into C
!
         Do
            IWRK = IWRKF
            IWRKD = IWRKF + 1
            JINDA = IWRKF + LMTNA
            IWRKF = IWRKF + LMTNC
            If (IWRKF >= NVAL) Then
               If (JINDA >= NVAL) Exit
               IWRKF = NVAL
            End If
            IINDA = 1
            IINDB = JINDA + 1
!
!   Shortcut for the case when the max of A is smaller
!   than the min of B. This line may be activated when the
!   initial set is already close to sorted.
!
!          IF (XDONT(IRNGT(JINDA)) <= XDONT(IRNGT(IINDB))) CYCLE
!
!  One steps in the C subset, that we build in the final rank array
!
!  Make a copy of the rank array for the merge iteration
!
            JWRKT (1:LMTNA) = IRNGT (IWRKD:JINDA)
!
            XVALA = XDONT (JWRKT(IINDA))
            XVALB = XDONT (IRNGT(IINDB))
!
            Do
               IWRK = IWRK + 1
!
!  We still have unprocessed values in both A and B
!
               If (XVALA > XVALB) Then
                  IRNGT (IWRK) = IRNGT (IINDB)
                  IINDB = IINDB + 1
                  If (IINDB > IWRKF) Then
!  Only A still with unprocessed values
                     IRNGT (IWRK+1:IWRKF) = JWRKT (IINDA:LMTNA)
                     Exit
                  End If
                  XVALB = XDONT (IRNGT(IINDB))
               Else
                  IRNGT (IWRK) = JWRKT (IINDA)
                  IINDA = IINDA + 1
                  If (IINDA > LMTNA) Exit! Only B still with unprocessed values
                  XVALA = XDONT (JWRKT(IINDA))
               End If
!
            End Do
         End Do
!
!  The Cs become As and Bs
!
         LMTNA = 2 * LMTNA
      End Do
!
      Return
!
End Subroutine I_mrgrnk
end module m_mrgrnk


Module m_ctrper
Use m_mrgrnk
Private
Integer, Parameter :: kdp = selected_real_kind(15)
public :: ctrper
private :: kdp
private :: R_ctrper, I_ctrper, D_ctrper
interface ctrper
  module procedure d_ctrper, r_ctrper, i_ctrper
end interface ctrper
contains

Subroutine D_ctrper (XDONT, PCLS)
!   Permute array XVALT randomly, but leaving elements close
!   to their initial locations (nearbyness is controled by PCLS).
! _________________________________________________________________
!   The routine takes the 1...size(XVALT) index array as real
!   values, takes a combination of these values and of random
!   values as a perturbation of the index array, and sorts the
!   initial set according to the ranks of these perturbated indices.
!   The relative proportion of initial order and random order
!   is 1-PCLS / PCLS, thus when PCLS = 0, there is no change in
!   the order whereas the new order is fully random when PCLS = 1.
!   Michel Olagnon - May 2000.
! _________________________________________________________________
! __________________________________________________________
      Real (kind=kdp), Dimension (:), Intent (InOut) :: XDONT
      Real, Intent (In) :: PCLS
! __________________________________________________________
!
      Real, Dimension (Size(XDONT)) :: XINDT
      Integer, Dimension (Size(XDONT)) :: JWRKT
      Real :: PWRK
      Integer :: I
!
      Call Random_number (XINDT(:))
      PWRK = Min (Max (0.0, PCLS), 1.0)
      XINDT = Real(Size(XDONT)) * XINDT
      XINDT = PWRK*XINDT + (1.0-PWRK)*(/ (Real(I), I=1,size(XDONT)) /)
      Call MRGRNK (XINDT, JWRKT)
      XDONT = XDONT (JWRKT)
!
End Subroutine D_ctrper

Subroutine R_ctrper (XDONT, PCLS)
!   Permute array XVALT randomly, but leaving elements close
!   to their initial locations (nearbyness is controled by PCLS).
! _________________________________________________________________
!   The routine takes the 1...size(XVALT) index array as real
!   values, takes a combination of these values and of random
!   values as a perturbation of the index array, and sorts the
!   initial set according to the ranks of these perturbated indices.
!   The relative proportion of initial order and random order
!   is 1-PCLS / PCLS, thus when PCLS = 0, there is no change in
!   the order whereas the new order is fully random when PCLS = 1.
!   Michel Olagnon - May 2000.
! _________________________________________________________________
! _________________________________________________________
      Real, Dimension (:), Intent (InOut) :: XDONT
      Real, Intent (In) :: PCLS
! __________________________________________________________
!
      Real, Dimension (Size(XDONT)) :: XINDT
      Integer, Dimension (Size(XDONT)) :: JWRKT
      Real :: PWRK
      Integer :: I
!
      Call Random_number (XINDT(:))
      PWRK = Min (Max (0.0, PCLS), 1.0)
      XINDT = Real(Size(XDONT)) * XINDT
      XINDT = PWRK*XINDT + (1.0-PWRK)*(/ (Real(I), I=1,size(XDONT)) /)
      Call MRGRNK (XINDT, JWRKT)
      XDONT = XDONT (JWRKT)
!
End Subroutine R_ctrper
Subroutine I_ctrper (XDONT, PCLS)
!   Permute array XVALT randomly, but leaving elements close
!   to their initial locations (nearbyness is controled by PCLS).
! _________________________________________________________________
!   The routine takes the 1...size(XVALT) index array as real
!   values, takes a combination of these values and of random
!   values as a perturbation of the index array, and sorts the
!   initial set according to the ranks of these perturbated indices.
!   The relative proportion of initial order and random order
!   is 1-PCLS / PCLS, thus when PCLS = 0, there is no change in
!   the order whereas the new order is fully random when PCLS = 1.
!   Michel Olagnon - May 2000.
! _________________________________________________________________
! __________________________________________________________
      Integer, Dimension (:), Intent (InOut)  :: XDONT
      Real, Intent (In) :: PCLS
! __________________________________________________________
!
      Real, Dimension (Size(XDONT)) :: XINDT
      Integer, Dimension (Size(XDONT)) :: JWRKT
      Real :: PWRK
      Integer :: I
!
      Call Random_number (XINDT(:))
      PWRK = Min (Max (0.0, PCLS), 1.0)
      XINDT = Real(Size(XDONT)) * XINDT
      XINDT = PWRK*XINDT + (1.0-PWRK)*(/ (Real(I), I=1,size(XDONT)) /)
      Call MRGRNK (XINDT, JWRKT)
      XDONT = XDONT (JWRKT)
!
End Subroutine I_ctrper
end module m_ctrper

Module m_fndnth
Integer, Parameter :: kdp = selected_real_kind(15)
public :: fndnth
private :: kdp
private :: R_fndnth, I_fndnth, D_fndnth
interface fndnth
  module procedure d_fndnth, r_fndnth, i_fndnth
end interface fndnth
contains

Function D_fndnth (XDONT, NORD) Result (FNDNTH)
!  Return NORDth value of XDONT, i.e fractile of order NORD/SIZE(XDONT).
! ______________________________________________________________________
!  This subroutine uses insertion sort, limiting insertion
!  to the first NORD values. It is faster when NORD is very small (2-5),
!  and it requires only a workarray of size NORD and type of XDONT,
!  but worst case behavior can happen fairly probably (initially inverse
!  sorted). In many cases, the refined quicksort method is faster.
!  Michel Olagnon - Aug. 2000
! __________________________________________________________
! __________________________________________________________
      Real (Kind=kdp), Dimension (:), Intent (In) :: XDONT
      Real (Kind=kdp) :: FNDNTH
      Integer, Intent (In) :: NORD
! __________________________________________________________
      Real (Kind=kdp), Dimension (NORD) :: XWRKT
      Real (Kind=kdp) :: XWRK, XWRK1
!
!
      Integer :: ICRS, IDCR, ILOW, NDON
!
      XWRKT (1) = XDONT (1)
      Do ICRS = 2, NORD
         XWRK = XDONT (ICRS)
         Do IDCR = ICRS - 1, 1, - 1
            If (XWRK >= XWRKT(IDCR)) Exit
            XWRKT (IDCR+1) = XWRKT (IDCR)
         End Do
         XWRKT (IDCR+1) = XWRK
      End Do
!
      NDON = SIZE (XDONT)
      XWRK1 = XWRKT (NORD)
      ILOW = 2*NORD - NDON
      Do ICRS = NORD + 1, NDON
         If (XDONT(ICRS) < XWRK1) Then
            XWRK = XDONT (ICRS)
            Do IDCR = NORD - 1, MAX (1, ILOW) , - 1
               If (XWRK >= XWRKT(IDCR)) Exit
               XWRKT (IDCR+1) = XWRKT (IDCR)
            End Do
            XWRKT (IDCR+1) = XWRK
            XWRK1 = XWRKT(NORD)
         End If
         ILOW = ILOW + 1
      End Do
      FNDNTH = XWRK1

!
End Function D_fndnth

Function R_fndnth (XDONT, NORD) Result (FNDNTH)
!  Return NORDth value of XDONT, i.e fractile of order NORD/SIZE(XDONT).
! ______________________________________________________________________
!  This subroutine uses insertion sort, limiting insertion
!  to the first NORD values. It is faster when NORD is very small (2-5),
!  and it requires only a workarray of size NORD and type of XDONT,
!  but worst case behavior can happen fairly probably (initially inverse
!  sorted). In many cases, the refined quicksort method is faster.
!  Michel Olagnon - Aug. 2000
! __________________________________________________________
! _________________________________________________________
      Real, Dimension (:), Intent (In) :: XDONT
      Real :: FNDNTH
      Integer, Intent (In) :: NORD
! __________________________________________________________
      Real, Dimension (NORD) :: XWRKT
      Real :: XWRK, XWRK1
!
!
      Integer :: ICRS, IDCR, ILOW, NDON
!
      XWRKT (1) = XDONT (1)
      Do ICRS = 2, NORD
         XWRK = XDONT (ICRS)
         Do IDCR = ICRS - 1, 1, - 1
            If (XWRK >= XWRKT(IDCR)) Exit
            XWRKT (IDCR+1) = XWRKT (IDCR)
         End Do
         XWRKT (IDCR+1) = XWRK
      End Do
!
      NDON = SIZE (XDONT)
      XWRK1 = XWRKT (NORD)
      ILOW = 2*NORD - NDON
      Do ICRS = NORD + 1, NDON
         If (XDONT(ICRS) < XWRK1) Then
            XWRK = XDONT (ICRS)
            Do IDCR = NORD - 1, MAX (1, ILOW) , - 1
               If (XWRK >= XWRKT(IDCR)) Exit
               XWRKT (IDCR+1) = XWRKT (IDCR)
            End Do
            XWRKT (IDCR+1) = XWRK
            XWRK1 = XWRKT(NORD)
         End If
         ILOW = ILOW + 1
      End Do
      FNDNTH = XWRK1

!
End Function R_fndnth
Function I_fndnth (XDONT, NORD) Result (FNDNTH)
!  Return NORDth value of XDONT, i.e fractile of order NORD/SIZE(XDONT).
! ______________________________________________________________________
!  This subroutine uses insertion sort, limiting insertion
!  to the first NORD values. It is faster when NORD is very small (2-5),
!  and it requires only a workarray of size NORD and type of XDONT,
!  but worst case behavior can happen fairly probably (initially inverse
!  sorted). In many cases, the refined quicksort method is faster.
!  Michel Olagnon - Aug. 2000
! __________________________________________________________
! __________________________________________________________
      Integer, Dimension (:), Intent (In) :: XDONT
      Integer :: fndnth
      Integer, Intent (In) :: NORD
! __________________________________________________________
      Integer, Dimension (NORD) :: XWRKT
      Integer :: XWRK, XWRK1
!
!
      Integer :: ICRS, IDCR, ILOW, NDON
!
      XWRKT (1) = XDONT (1)
      Do ICRS = 2, NORD
         XWRK = XDONT (ICRS)
         Do IDCR = ICRS - 1, 1, - 1
            If (XWRK >= XWRKT(IDCR)) Exit
            XWRKT (IDCR+1) = XWRKT (IDCR)
         End Do
         XWRKT (IDCR+1) = XWRK
      End Do
!
      NDON = SIZE (XDONT)
      XWRK1 = XWRKT (NORD)
      ILOW = 2*NORD - NDON
      Do ICRS = NORD + 1, NDON
         If (XDONT(ICRS) < XWRK1) Then
            XWRK = XDONT (ICRS)
            Do IDCR = NORD - 1, MAX (1, ILOW) , - 1
               If (XWRK >= XWRKT(IDCR)) Exit
               XWRKT (IDCR+1) = XWRKT (IDCR)
            End Do
            XWRKT (IDCR+1) = XWRK
            XWRK1 = XWRKT(NORD)
         End If
         ILOW = ILOW + 1
      End Do
      FNDNTH = XWRK1

!
End Function I_fndnth
end module m_fndnth

Module m_indmed
Integer, Parameter :: kdp = selected_real_kind(15)
public :: indmed
private :: kdp
private :: R_indmed, I_indmed, D_indmed
private :: r_med, i_med, d_med
Integer, Allocatable, Dimension(:), Private, Save :: IDONT
interface indmed
  module procedure d_indmed, r_indmed, i_indmed
end interface indmed
contains

Subroutine D_indmed (XDONT, INDM)
!  Returns index of median value of XDONT.
! __________________________________________________________
      Real (kind=kdp), Dimension (:), Intent (In) :: XDONT
      Integer, Intent (Out) :: INDM
! __________________________________________________________
      Integer :: IDON
!
      Allocate (IDONT (SIZE(XDONT)))
      Do IDON = 1, SIZE(XDONT)
         IDONT (IDON) = IDON
      End Do
!
      Call d_med (XDONT, IDONT, INDM)
!
      Deallocate (IDONT)
End Subroutine D_indmed
   Recursive Subroutine d_med (XDATT, IDATT, ires_med)
!  Finds the index of the median of XDONT using the recursive procedure
!  described in Knuth, The Art of Computer Programming,
!  vol. 3, 5.3.3 - This procedure is linear in time, and
!  does not require to be able to interpolate in the
!  set as the one used in INDNTH. It also has better worst
!  case behavior than INDNTH, but is about 30% slower in
!  average for random uniformly distributed values.
! __________________________________________________________
      Real (kind=kdp), Dimension (:), Intent (In) :: XDATT
      Integer, Dimension (:), Intent (In) :: IDATT
      Integer, Intent (Out):: ires_med
! __________________________________________________________
!
      Real (kind=kdp), Parameter :: XHUGE = HUGE (XDATT)
      Real (kind=kdp) :: XWRK, XWRK1, XMED7, XMAX, XMIN
!
      Integer, Dimension (7*(((Size (IDATT)+6)/7+6)/7)) :: ISTRT, IENDT, IMEDT
      Integer, Dimension (7*((Size(IDATT)+6)/7)) :: IWRKT
      Integer :: NTRI, NMED, NORD, NEQU, NLEQ, IMED, IDON, IDON1
      Integer :: IDEB, ITMP, IDCR, ICRS, ICRS1, ICRS2, IMAX, IMIN
      Integer :: IWRK, IWRK1, IMED1, IMED7, NDAT
!
      NDAT = Size (IDATT)
      NMED = (NDAT+1) / 2
      IWRKT = IDATT
!
!  If the number of values is small, then use insertion sort
!
     If (NDAT < 35) Then
!
!  Bring minimum to first location to save test in decreasing loop
!
         IDCR = NDAT
         If (XDATT (IWRKT (1)) < XDATT (IWRKT (IDCR))) Then
            IWRK = IWRKT (1)
         Else
            IWRK = IWRKT (IDCR)
            IWRKT (IDCR) = IWRKT (1)
         Endif
         XWRK = XDATT (IWRK)
         Do ITMP = 1, NDAT - 2
            IDCR = IDCR - 1
            IWRK1 = IWRKT (IDCR)
            XWRK1 = XDATT (IWRK1)
            If (XWRK1 < XWRK) Then
                IWRKT (IDCR) = IWRK
                XWRK = XWRK1
                IWRK = IWRK1
            Endif
         End Do
         IWRKT (1) = IWRK
!
! Sort the first half, until we have NMED sorted values
!
         Do ICRS = 3, NMED
            XWRK = XDATT (IWRKT (ICRS))
            IWRK = IWRKT (ICRS)
            IDCR = ICRS - 1
            Do
                  If (XWRK >= XDATT (IWRKT(IDCR))) Exit
                  IWRKT (IDCR+1) = IWRKT (IDCR)
                  IDCR = IDCR - 1
            End Do
            IWRKT (IDCR+1) = IWRK
         End Do
!
!  Insert any value less than the current median in the first half
!
         XWRK1 = XDATT (IWRKT (NMED))
         Do ICRS = NMED+1, NDAT
            XWRK = XDATT (IWRKT (ICRS))
            IWRK = IWRKT (ICRS)
            If (XWRK < XWRK1) Then
               IDCR = NMED - 1
               Do
                  If (XWRK >= XDATT (IWRKT(IDCR))) Exit
                  IWRKT (IDCR+1) = IWRKT (IDCR)
                  IDCR = IDCR - 1
               End Do
               IWRKT (IDCR+1) = IWRK
               XWRK1 = XDATT (IWRKT (NMED))
            End If
         End Do
         ires_med = IWRKT (NMED)
         Return
      End If
!
!  Make sorted subsets of 7 elements
!  This is done by a variant of insertion sort where a first
!  pass is used to bring the smallest element to the first position
!  decreasing disorder at the same time, so that we may remove
!  remove the loop test in the insertion loop.
!
      IMAX = 1
      IMIN = 1
      XMAX = XDATT (IWRKT(IMAX))
      XMIN = XDATT (IWRKT(IMIN))
      DO IDEB = 1, NDAT-6, 7
         IDCR = IDEB + 6
         If (XDATT (IWRKT(IDEB)) < XDATT (IWRKT(IDCR))) Then
            IWRK = IWRKT(IDEB)
         Else
            IWRK = IWRKT (IDCR)
            IWRKT (IDCR) = IWRKT(IDEB)
         Endif
         XWRK = XDATT (IWRK)
         Do ITMP = 1, 5
            IDCR = IDCR - 1
            IWRK1 = IWRKT (IDCR)
            XWRK1 = XDATT (IWRK1)
            If (XWRK1 < XWRK) Then
                IWRKT (IDCR) = IWRK
                IWRK = IWRK1
                XWRK = XWRK1
            Endif
         End Do
         IWRKT (IDEB) = IWRK
         If (XWRK < XMIN) Then
             IMIN = IWRK
             XMIN = XWRK
         End If
         Do ICRS = IDEB+1, IDEB+5
            IWRK = IWRKT (ICRS+1)
            XWRK = XDATT (IWRK)
            IDON = IWRKT(ICRS)
            If (XWRK < XDATT(IDON)) Then
               IWRKT (ICRS+1) = IDON
               IDCR = ICRS
               IWRK1 = IWRKT (IDCR-1)
               XWRK1 = XDATT (IWRK1)
               Do
                  If (XWRK >= XWRK1) Exit
                  IWRKT (IDCR) = IWRK1
                  IDCR = IDCR - 1
                  IWRK1 = IWRKT (IDCR-1)
                  XWRK1 = XDATT (IWRK1)
               End Do
               IWRKT (IDCR) = IWRK
            EndIf
         End Do
         If (XWRK > XMAX) Then
             IMAX = IWRK
             XMAX = XWRK
         End If
      End Do
!
!  Add-up alternatively MAX and MIN values to make the number of data
!  an exact multiple of 7.
!
      IDEB = 7 * (NDAT/7)
      NTRI = NDAT
      If (IDEB < NDAT) Then
!
         Do ICRS = IDEB+1, NDAT
            XWRK1 = XDATT (IWRKT (ICRS))
            IF (XWRK1 > XMAX) Then
               IMAX = IWRKT (ICRS)
               XMAX = XWRK1
            End If
            IF (XWRK1 < XMIN) Then
               IMIN = IWRKT (ICRS)
               XMIN = XWRK1
            End If
         End Do
         IWRK1 = IMAX
         Do ICRS = NDAT+1, IDEB+7
               IWRKT (ICRS) = IWRK1
               If (IWRK1 == IMAX) Then
                  IWRK1 = IMIN
               Else
                  NMED = NMED + 1
                  IWRK1 = IMAX
               End If
         End Do
!
         Do ICRS = IDEB+2, IDEB+7
            IWRK = IWRKT (ICRS)
            XWRK = XDATT (IWRK)
            Do IDCR = ICRS - 1, IDEB+1, - 1
               If (XWRK >= XDATT (IWRKT(IDCR))) Exit
               IWRKT (IDCR+1) = IWRKT (IDCR)
            End Do
            IWRKT (IDCR+1) = IWRK
         End Do
!
         NTRI = IDEB+7
      End If
!
!  Make the set of the indices of median values of each sorted subset
!
         IDON1 = 0
         Do IDON = 1, NTRI, 7
            IDON1 = IDON1 + 1
            IMEDT (IDON1) = IWRKT (IDON + 3)
         End Do
!
!  Find XMED7, the median of the medians
!
         Call d_med (XDATT, IMEDT(1:IDON1), IMED7)
         XMED7 = XDATT (IMED7)
!
!  Count how many values are not higher than (and how many equal to) XMED7
!  This number is at least 4 * 1/2 * (N/7) : 4 values in each of the
!  subsets where the median is lower than the median of medians. For similar
!  reasons, we also have at least 2N/7 values not lower than XMED7. At the
!  same time, we find in each subset the index of the last value < XMED7,
!  and that of the first > XMED7. These indices will be used to restrict the
!  search for the median as the Kth element in the subset (> or <) where
!  we know it to be.
!
         IDON1 = 1
         NLEQ = 0
         NEQU = 0
         Do IDON = 1, NTRI, 7
            IMED = IDON+3
            If (XDATT (IWRKT (IMED)) > XMED7) Then
                  IMED = IMED - 2
                  If (XDATT (IWRKT (IMED)) > XMED7) Then
                     IMED = IMED - 1
                  Else If (XDATT (IWRKT (IMED)) < XMED7) Then
                     IMED = IMED + 1
                  Endif
            Else If (XDATT (IWRKT (IMED)) < XMED7) Then
                  IMED = IMED + 2
                  If (XDATT (IWRKT (IMED)) > XMED7) Then
                     IMED = IMED - 1
                  Else If (XDATT (IWRKT (IMED)) < XMED7) Then
                     IMED = IMED + 1
                  Endif
            Endif
            If (XDATT (IWRKT (IMED)) > XMED7) Then
               NLEQ = NLEQ + IMED - IDON
               IENDT (IDON1) = IMED - 1
               ISTRT (IDON1) = IMED
            Else If (XDATT (IWRKT (IMED)) < XMED7) Then
               NLEQ = NLEQ + IMED - IDON + 1
               IENDT (IDON1) = IMED
               ISTRT (IDON1) = IMED + 1
            Else                    !       If (XDATT (IWRKT (IMED)) == XMED7)
               NLEQ = NLEQ + IMED - IDON + 1
               NEQU = NEQU + 1
               IENDT (IDON1) = IMED - 1
               Do IMED1 = IMED - 1, IDON, -1
                  If (XDATT (IWRKT (IMED1)) == XMED7) Then
                     NEQU = NEQU + 1
                     IENDT (IDON1) = IMED1 - 1
                  Else
                     Exit
                  End If
               End Do
               ISTRT (IDON1) = IMED + 1
               Do IMED1 = IMED + 1, IDON + 6
                  If (XDATT (IWRKT (IMED1)) == XMED7) Then
                     NEQU = NEQU + 1
                     NLEQ = NLEQ + 1
                     ISTRT (IDON1) = IMED1 + 1
                  Else
                     Exit
                  End If
               End Do
            Endif
            IDON1 = IDON1 + 1
         End Do
!
!  Carry out a partial insertion sort to find the Kth smallest of the
!  large values, or the Kth largest of the small values, according to
!  what is needed.
!
!
         If (NLEQ - NEQU + 1 <= NMED) Then
            If (NLEQ < NMED) Then   !      Not enough low values
                IWRK1 = IMAX
                XWRK1 = XDATT (IWRK1)
                NORD = NMED - NLEQ
                IDON1 = 0
                ICRS1 = 1
                ICRS2 = 0
                IDCR = 0
                Do IDON = 1, NTRI, 7
                   IDON1 = IDON1 + 1
                   If (ICRS2 < NORD) Then
                      Do ICRS = ISTRT (IDON1), IDON + 6
                         If (XDATT (IWRKT (ICRS)) < XWRK1) Then
                            IWRK = IWRKT (ICRS)
                            XWRK = XDATT (IWRK)
                            Do IDCR = ICRS1 - 1, 1, - 1
                               If (XWRK >= XDATT (IWRKT (IDCR))) Exit
                               IWRKT  (IDCR+1) = IWRKT (IDCR)
                            End Do
                            IWRKT (IDCR+1) = IWRK
                            IWRK1 = IWRKT (ICRS1)
                            XWRK1 = XDATT (IWRK1)
                         Else
                           If (ICRS2 < NORD) Then
                              IWRKT (ICRS1) = IWRKT (ICRS)
                              IWRK1 = IWRKT (ICRS1)
                              XWRK1 = XDATT (IWRK1)
                           Endif
                         End If
                         ICRS1 = MIN (NORD, ICRS1 + 1)
                         ICRS2 = MIN (NORD, ICRS2 + 1)
                      End Do
                   Else
                      Do ICRS = ISTRT (IDON1), IDON + 6
                         If (XDATT (IWRKT (ICRS)) >= XWRK1) Exit
                         IWRK = IWRKT (ICRS)
                         XWRK = XDATT (IWRK)
                         Do IDCR = ICRS1 - 1, 1, - 1
                               If (XWRK >= XDATT (IWRKT (IDCR))) Exit
                               IWRKT  (IDCR+1) = IWRKT (IDCR)
                         End Do
                         IWRKT (IDCR+1) = IWRK
                         IWRK1 = IWRKT (ICRS1)
                         XWRK1 = XDATT (IWRK1)
                      End Do
                   End If
                End Do
                ires_med = IWRK1
                Return
            Else
                ires_med = IMED7
                Return
            End If
         Else                       !      If (NLEQ > NMED)
!                                          Not enough high values
                XWRK1 = -XHUGE
                NORD = NLEQ - NEQU - NMED + 1
                IDON1 = 0
                ICRS1 = 1
                ICRS2 = 0
                Do IDON = 1, NTRI, 7
                   IDON1 = IDON1 + 1
                   If (ICRS2 < NORD) Then
!
                      Do ICRS = IDON, IENDT (IDON1)
                         If (XDATT(IWRKT (ICRS)) > XWRK1) Then
                            IWRK = IWRKT (ICRS)
                            XWRK = XDATT (IWRK)
                            IDCR = ICRS1 - 1
                            Do IDCR = ICRS1 - 1, 1, - 1
                               If (XWRK <= XDATT(IWRKT (IDCR))) Exit
                               IWRKT (IDCR+1) = IWRKT (IDCR)
                            End Do
                            IWRKT (IDCR+1) = IWRK
                            IWRK1 = IWRKT(ICRS1)
                            XWRK1 = XDATT(IWRK1)
                         Else
                            If (ICRS2 < NORD) Then
                               IWRKT (ICRS1) = IWRKT (ICRS)
                               IWRK1 = IWRKT(ICRS1)
                               XWRK1 = XDATT(IWRK1)
                            End If
                         End If
                         ICRS1 = MIN (NORD, ICRS1 + 1)
                         ICRS2 = MIN (NORD, ICRS2 + 1)
                      End Do
                   Else
                      Do ICRS = IENDT (IDON1), IDON, -1
                         If (XDATT(IWRKT (ICRS)) <= XWRK1) Exit
                         IWRK = IWRKT (ICRS)
                         XWRK = XDATT (IWRK)
                         IDCR = ICRS1 - 1
                         Do IDCR = ICRS1 - 1, 1, - 1
                               If (XWRK <= XDATT(IWRKT (IDCR))) Exit
                               IWRKT (IDCR+1) = IWRKT (IDCR)
                         End Do
                         IWRKT (IDCR+1) = IWRK
                         IWRK1 = IWRKT(ICRS1)
                         XWRK1 = XDATT(IWRK1)
                      End Do
                   Endif
                End Do
!
                ires_med = IWRK1
                Return
         End If
!
   END Subroutine d_med
!
Subroutine R_indmed (XDONT, INDM)
!  Returns index of median value of XDONT.
! __________________________________________________________
      Real, Dimension (:), Intent (In) :: XDONT
      Integer, Intent (Out) :: INDM
! __________________________________________________________
      Integer :: IDON
!
      Allocate (IDONT (SIZE(XDONT)))
      Do IDON = 1, SIZE(XDONT)
         IDONT (IDON) = IDON
      End Do
!
      Call r_med (XDONT, IDONT, INDM)
!
      Deallocate (IDONT)
End Subroutine R_indmed
   Recursive Subroutine r_med (XDATT, IDATT, ires_med)
!  Finds the index of the median of XDONT using the recursive procedure
!  described in Knuth, The Art of Computer Programming,
!  vol. 3, 5.3.3 - This procedure is linear in time, and
!  does not require to be able to interpolate in the
!  set as the one used in INDNTH. It also has better worst
!  case behavior than INDNTH, but is about 30% slower in
!  average for random uniformly distributed values.
! __________________________________________________________
      Real, Dimension (:), Intent (In) :: XDATT
      Integer, Dimension (:), Intent (In) :: IDATT
      Integer, Intent (Out) :: ires_med
! __________________________________________________________
!
      Real, Parameter :: XHUGE = HUGE (XDATT)
      Real :: XWRK, XWRK1, XMED7, XMAX, XMIN
!
      Integer, Dimension (7*(((Size (IDATT)+6)/7+6)/7)) :: ISTRT, IENDT, IMEDT
      Integer, Dimension (7*((Size(IDATT)+6)/7)) :: IWRKT
      Integer :: NTRI, NMED, NORD, NEQU, NLEQ, IMED, IDON, IDON1
      Integer :: IDEB, ITMP, IDCR, ICRS, ICRS1, ICRS2, IMAX, IMIN
      Integer :: IWRK, IWRK1, IMED1, IMED7, NDAT
!
      NDAT = Size (IDATT)
      NMED = (NDAT+1) / 2
      IWRKT = IDATT
!
!  If the number of values is small, then use insertion sort
!
     If (NDAT < 35) Then
!
!  Bring minimum to first location to save test in decreasing loop
!
         IDCR = NDAT
         If (XDATT (IWRKT (1)) < XDATT (IWRKT (IDCR))) Then
            IWRK = IWRKT (1)
         Else
            IWRK = IWRKT (IDCR)
            IWRKT (IDCR) = IWRKT (1)
         Endif
         XWRK = XDATT (IWRK)
         Do ITMP = 1, NDAT - 2
            IDCR = IDCR - 1
            IWRK1 = IWRKT (IDCR)
            XWRK1 = XDATT (IWRK1)
            If (XWRK1 < XWRK) Then
                IWRKT (IDCR) = IWRK
                XWRK = XWRK1
                IWRK = IWRK1
            Endif
         End Do
         IWRKT (1) = IWRK
!
! Sort the first half, until we have NMED sorted values
!
         Do ICRS = 3, NMED
            XWRK = XDATT (IWRKT (ICRS))
            IWRK = IWRKT (ICRS)
            IDCR = ICRS - 1
            Do
                  If (XWRK >= XDATT (IWRKT(IDCR))) Exit
                  IWRKT (IDCR+1) = IWRKT (IDCR)
                  IDCR = IDCR - 1
            End Do
            IWRKT (IDCR+1) = IWRK
         End Do
!
!  Insert any value less than the current median in the first half
!
         XWRK1 = XDATT (IWRKT (NMED))
         Do ICRS = NMED+1, NDAT
            XWRK = XDATT (IWRKT (ICRS))
            IWRK = IWRKT (ICRS)
            If (XWRK < XWRK1) Then
               IDCR = NMED - 1
               Do
                  If (XWRK >= XDATT (IWRKT(IDCR))) Exit
                  IWRKT (IDCR+1) = IWRKT (IDCR)
                  IDCR = IDCR - 1
               End Do
               IWRKT (IDCR+1) = IWRK
               XWRK1 = XDATT (IWRKT (NMED))
            End If
         End Do
         ires_med = IWRKT (NMED)
         Return
      End If
!
!  Make sorted subsets of 7 elements
!  This is done by a variant of insertion sort where a first
!  pass is used to bring the smallest element to the first position
!  decreasing disorder at the same time, so that we may remove
!  remove the loop test in the insertion loop.
!
      IMAX = 1
      IMIN = 1
      XMAX = XDATT (IWRKT(IMAX))
      XMIN = XDATT (IWRKT(IMIN))
      DO IDEB = 1, NDAT-6, 7
         IDCR = IDEB + 6
         If (XDATT (IWRKT(IDEB)) < XDATT (IWRKT(IDCR))) Then
            IWRK = IWRKT(IDEB)
         Else
            IWRK = IWRKT (IDCR)
            IWRKT (IDCR) = IWRKT(IDEB)
         Endif
         XWRK = XDATT (IWRK)
         Do ITMP = 1, 5
            IDCR = IDCR - 1
            IWRK1 = IWRKT (IDCR)
            XWRK1 = XDATT (IWRK1)
            If (XWRK1 < XWRK) Then
                IWRKT (IDCR) = IWRK
                IWRK = IWRK1
                XWRK = XWRK1
            Endif
         End Do
         IWRKT (IDEB) = IWRK
         If (XWRK < XMIN) Then
             IMIN = IWRK
             XMIN = XWRK
         End If
         Do ICRS = IDEB+1, IDEB+5
            IWRK = IWRKT (ICRS+1)
            XWRK = XDATT (IWRK)
            IDON = IWRKT(ICRS)
            If (XWRK < XDATT(IDON)) Then
               IWRKT (ICRS+1) = IDON
               IDCR = ICRS
               IWRK1 = IWRKT (IDCR-1)
               XWRK1 = XDATT (IWRK1)
               Do
                  If (XWRK >= XWRK1) Exit
                  IWRKT (IDCR) = IWRK1
                  IDCR = IDCR - 1
                  IWRK1 = IWRKT (IDCR-1)
                  XWRK1 = XDATT (IWRK1)
               End Do
               IWRKT (IDCR) = IWRK
            EndIf
         End Do
         If (XWRK > XMAX) Then
             IMAX = IWRK
             XMAX = XWRK
         End If
      End Do
!
!  Add-up alternatively MAX and MIN values to make the number of data
!  an exact multiple of 7.
!
      IDEB = 7 * (NDAT/7)
      NTRI = NDAT
      If (IDEB < NDAT) Then
!
         Do ICRS = IDEB+1, NDAT
            XWRK1 = XDATT (IWRKT (ICRS))
            IF (XWRK1 > XMAX) Then
               IMAX = IWRKT (ICRS)
               XMAX = XWRK1
            End If
            IF (XWRK1 < XMIN) Then
               IMIN = IWRKT (ICRS)
               XMIN = XWRK1
            End If
         End Do
         IWRK1 = IMAX
         Do ICRS = NDAT+1, IDEB+7
               IWRKT (ICRS) = IWRK1
               If (IWRK1 == IMAX) Then
                  IWRK1 = IMIN
               Else
                  NMED = NMED + 1
                  IWRK1 = IMAX
               End If
         End Do
!
         Do ICRS = IDEB+2, IDEB+7
            IWRK = IWRKT (ICRS)
            XWRK = XDATT (IWRK)
            Do IDCR = ICRS - 1, IDEB+1, - 1
               If (XWRK >= XDATT (IWRKT(IDCR))) Exit
               IWRKT (IDCR+1) = IWRKT (IDCR)
            End Do
            IWRKT (IDCR+1) = IWRK
         End Do
!
         NTRI = IDEB+7
      End If
!
!  Make the set of the indices of median values of each sorted subset
!
         IDON1 = 0
         Do IDON = 1, NTRI, 7
            IDON1 = IDON1 + 1
            IMEDT (IDON1) = IWRKT (IDON + 3)
         End Do
!
!  Find XMED7, the median of the medians
!
         Call r_med (XDATT, IMEDT(1:IDON1), IMED7)
         XMED7 = XDATT (IMED7)
!
!  Count how many values are not higher than (and how many equal to) XMED7
!  This number is at least 4 * 1/2 * (N/7) : 4 values in each of the
!  subsets where the median is lower than the median of medians. For similar
!  reasons, we also have at least 2N/7 values not lower than XMED7. At the
!  same time, we find in each subset the index of the last value < XMED7,
!  and that of the first > XMED7. These indices will be used to restrict the
!  search for the median as the Kth element in the subset (> or <) where
!  we know it to be.
!
         IDON1 = 1
         NLEQ = 0
         NEQU = 0
         Do IDON = 1, NTRI, 7
            IMED = IDON+3
            If (XDATT (IWRKT (IMED)) > XMED7) Then
                  IMED = IMED - 2
                  If (XDATT (IWRKT (IMED)) > XMED7) Then
                     IMED = IMED - 1
                  Else If (XDATT (IWRKT (IMED)) < XMED7) Then
                     IMED = IMED + 1
                  Endif
            Else If (XDATT (IWRKT (IMED)) < XMED7) Then
                  IMED = IMED + 2
                  If (XDATT (IWRKT (IMED)) > XMED7) Then
                     IMED = IMED - 1
                  Else If (XDATT (IWRKT (IMED)) < XMED7) Then
                     IMED = IMED + 1
                  Endif
            Endif
            If (XDATT (IWRKT (IMED)) > XMED7) Then
               NLEQ = NLEQ + IMED - IDON
               IENDT (IDON1) = IMED - 1
               ISTRT (IDON1) = IMED
            Else If (XDATT (IWRKT (IMED)) < XMED7) Then
               NLEQ = NLEQ + IMED - IDON + 1
               IENDT (IDON1) = IMED
               ISTRT (IDON1) = IMED + 1
            Else                    !       If (XDATT (IWRKT (IMED)) == XMED7)
               NLEQ = NLEQ + IMED - IDON + 1
               NEQU = NEQU + 1
               IENDT (IDON1) = IMED - 1
               Do IMED1 = IMED - 1, IDON, -1
                  If (XDATT (IWRKT (IMED1)) == XMED7) Then
                     NEQU = NEQU + 1
                     IENDT (IDON1) = IMED1 - 1
                  Else
                     Exit
                  End If
               End Do
               ISTRT (IDON1) = IMED + 1
               Do IMED1 = IMED + 1, IDON + 6
                  If (XDATT (IWRKT (IMED1)) == XMED7) Then
                     NEQU = NEQU + 1
                     NLEQ = NLEQ + 1
                     ISTRT (IDON1) = IMED1 + 1
                  Else
                     Exit
                  End If
               End Do
            Endif
            IDON1 = IDON1 + 1
         End Do
!
!  Carry out a partial insertion sort to find the Kth smallest of the
!  large values, or the Kth largest of the small values, according to
!  what is needed.
!
!
         If (NLEQ - NEQU + 1 <= NMED) Then
            If (NLEQ < NMED) Then   !      Not enough low values
                IWRK1 = IMAX
                XWRK1 = XDATT (IWRK1)
                NORD = NMED - NLEQ
                IDON1 = 0
                ICRS1 = 1
                ICRS2 = 0
                IDCR = 0
                Do IDON = 1, NTRI, 7
                   IDON1 = IDON1 + 1
                   If (ICRS2 < NORD) Then
                      Do ICRS = ISTRT (IDON1), IDON + 6
                         If (XDATT (IWRKT (ICRS)) < XWRK1) Then
                            IWRK = IWRKT (ICRS)
                            XWRK = XDATT (IWRK)
                            Do IDCR = ICRS1 - 1, 1, - 1
                               If (XWRK >= XDATT (IWRKT (IDCR))) Exit
                               IWRKT  (IDCR+1) = IWRKT (IDCR)
                            End Do
                            IWRKT (IDCR+1) = IWRK
                            IWRK1 = IWRKT (ICRS1)
                            XWRK1 = XDATT (IWRK1)
                         Else
                           If (ICRS2 < NORD) Then
                              IWRKT (ICRS1) = IWRKT (ICRS)
                              IWRK1 = IWRKT (ICRS1)
                              XWRK1 = XDATT (IWRK1)
                           Endif
                         End If
                         ICRS1 = MIN (NORD, ICRS1 + 1)
                         ICRS2 = MIN (NORD, ICRS2 + 1)
                      End Do
                   Else
                      Do ICRS = ISTRT (IDON1), IDON + 6
                         If (XDATT (IWRKT (ICRS)) >= XWRK1) Exit
                         IWRK = IWRKT (ICRS)
                         XWRK = XDATT (IWRK)
                         Do IDCR = ICRS1 - 1, 1, - 1
                               If (XWRK >= XDATT (IWRKT (IDCR))) Exit
                               IWRKT  (IDCR+1) = IWRKT (IDCR)
                         End Do
                         IWRKT (IDCR+1) = IWRK
                         IWRK1 = IWRKT (ICRS1)
                         XWRK1 = XDATT (IWRK1)
                      End Do
                   End If
                End Do
                ires_med = IWRK1
                Return
            Else
                ires_med = IMED7
                Return
            End If
         Else                       !      If (NLEQ > NMED)
!                                          Not enough high values
                XWRK1 = -XHUGE
                NORD = NLEQ - NEQU - NMED + 1
                IDON1 = 0
                ICRS1 = 1
                ICRS2 = 0
                Do IDON = 1, NTRI, 7
                   IDON1 = IDON1 + 1
                   If (ICRS2 < NORD) Then
!
                      Do ICRS = IDON, IENDT (IDON1)
                         If (XDATT(IWRKT (ICRS)) > XWRK1) Then
                            IWRK = IWRKT (ICRS)
                            XWRK = XDATT (IWRK)
                            IDCR = ICRS1 - 1
                            Do IDCR = ICRS1 - 1, 1, - 1
                               If (XWRK <= XDATT(IWRKT (IDCR))) Exit
                               IWRKT (IDCR+1) = IWRKT (IDCR)
                            End Do
                            IWRKT (IDCR+1) = IWRK
                            IWRK1 = IWRKT(ICRS1)
                            XWRK1 = XDATT(IWRK1)
                         Else
                            If (ICRS2 < NORD) Then
                               IWRKT (ICRS1) = IWRKT (ICRS)
                               IWRK1 = IWRKT(ICRS1)
                               XWRK1 = XDATT(IWRK1)
                            End If
                         End If
                         ICRS1 = MIN (NORD, ICRS1 + 1)
                         ICRS2 = MIN (NORD, ICRS2 + 1)
                      End Do
                   Else
                      Do ICRS = IENDT (IDON1), IDON, -1
                         If (XDATT(IWRKT (ICRS)) <= XWRK1) Exit
                         IWRK = IWRKT (ICRS)
                         XWRK = XDATT (IWRK)
                         IDCR = ICRS1 - 1
                         Do IDCR = ICRS1 - 1, 1, - 1
                               If (XWRK <= XDATT(IWRKT (IDCR))) Exit
                               IWRKT (IDCR+1) = IWRKT (IDCR)
                         End Do
                         IWRKT (IDCR+1) = IWRK
                         IWRK1 = IWRKT(ICRS1)
                         XWRK1 = XDATT(IWRK1)
                      End Do
                   Endif
                End Do
!
                ires_med = IWRK1
                Return
         End If
!
   END Subroutine r_med
Subroutine I_indmed (XDONT, INDM)
!  Returns index of median value of XDONT.
! __________________________________________________________
      Integer, Dimension (:), Intent (In) :: XDONT
      Integer, Intent (Out) :: INDM
! __________________________________________________________
      Integer :: IDON
!
      Allocate (IDONT (SIZE(XDONT)))
      Do IDON = 1, SIZE(XDONT)
         IDONT (IDON) = IDON
      End Do
!
      Call i_med (XDONT, IDONT, INDM)
!
      Deallocate (IDONT)
End Subroutine I_indmed
   Recursive Subroutine i_med (XDATT, IDATT, ires_med)
!  Finds the index of the median of XDONT using the recursive procedure
!  described in Knuth, The Art of Computer Programming,
!  vol. 3, 5.3.3 - This procedure is linear in time, and
!  does not require to be able to interpolate in the
!  set as the one used in INDNTH. It also has better worst
!  case behavior than INDNTH, but is about 30% slower in
!  average for random uniformly distributed values.
! __________________________________________________________
      Integer, Dimension (:), Intent (In) :: XDATT
      Integer, Dimension (:), Intent (In) :: IDATT
      Integer, Intent (Out) :: ires_med
! __________________________________________________________
!
      Integer, Parameter :: XHUGE = HUGE (XDATT)
      Integer :: XWRK, XWRK1, XMED7, XMAX, XMIN
!
      Integer, Dimension (7*(((Size (IDATT)+6)/7+6)/7)) :: ISTRT, IENDT, IMEDT
      Integer, Dimension (7*((Size(IDATT)+6)/7)) :: IWRKT
      Integer :: NTRI, NMED, NORD, NEQU, NLEQ, IMED, IDON, IDON1
      Integer :: IDEB, ITMP, IDCR, ICRS, ICRS1, ICRS2, IMAX, IMIN
      Integer :: IWRK, IWRK1, IMED1, IMED7, NDAT
!
      NDAT = Size (IDATT)
      NMED = (NDAT+1) / 2
      IWRKT = IDATT
!
!  If the number of values is small, then use insertion sort
!
     If (NDAT < 35) Then
!
!  Bring minimum to first location to save test in decreasing loop
!
         IDCR = NDAT
         If (XDATT (IWRKT (1)) < XDATT (IWRKT (IDCR))) Then
            IWRK = IWRKT (1)
         Else
            IWRK = IWRKT (IDCR)
            IWRKT (IDCR) = IWRKT (1)
         Endif
         XWRK = XDATT (IWRK)
         Do ITMP = 1, NDAT - 2
            IDCR = IDCR - 1
            IWRK1 = IWRKT (IDCR)
            XWRK1 = XDATT (IWRK1)
            If (XWRK1 < XWRK) Then
                IWRKT (IDCR) = IWRK
                XWRK = XWRK1
                IWRK = IWRK1
            Endif
         End Do
         IWRKT (1) = IWRK
!
! Sort the first half, until we have NMED sorted values
!
         Do ICRS = 3, NMED
            XWRK = XDATT (IWRKT (ICRS))
            IWRK = IWRKT (ICRS)
            IDCR = ICRS - 1
            Do
                  If (XWRK >= XDATT (IWRKT(IDCR))) Exit
                  IWRKT (IDCR+1) = IWRKT (IDCR)
                  IDCR = IDCR - 1
            End Do
            IWRKT (IDCR+1) = IWRK
         End Do
!
!  Insert any value less than the current median in the first half
!
         XWRK1 = XDATT (IWRKT (NMED))
         Do ICRS = NMED+1, NDAT
            XWRK = XDATT (IWRKT (ICRS))
            IWRK = IWRKT (ICRS)
            If (XWRK < XWRK1) Then
               IDCR = NMED - 1
               Do
                  If (XWRK >= XDATT (IWRKT(IDCR))) Exit
                  IWRKT (IDCR+1) = IWRKT (IDCR)
                  IDCR = IDCR - 1
               End Do
               IWRKT (IDCR+1) = IWRK
               XWRK1 = XDATT (IWRKT (NMED))
            End If
         End Do
         ires_med = IWRKT (NMED)
         Return
      End If
!
!  Make sorted subsets of 7 elements
!  This is done by a variant of insertion sort where a first
!  pass is used to bring the smallest element to the first position
!  decreasing disorder at the same time, so that we may remove
!  remove the loop test in the insertion loop.
!
      IMAX = 1
      IMIN = 1
      XMAX = XDATT (IWRKT(IMAX))
      XMIN = XDATT (IWRKT(IMIN))
      DO IDEB = 1, NDAT-6, 7
         IDCR = IDEB + 6
         If (XDATT (IWRKT(IDEB)) < XDATT (IWRKT(IDCR))) Then
            IWRK = IWRKT(IDEB)
         Else
            IWRK = IWRKT (IDCR)
            IWRKT (IDCR) = IWRKT(IDEB)
         Endif
         XWRK = XDATT (IWRK)
         Do ITMP = 1, 5
            IDCR = IDCR - 1
            IWRK1 = IWRKT (IDCR)
            XWRK1 = XDATT (IWRK1)
            If (XWRK1 < XWRK) Then
                IWRKT (IDCR) = IWRK
                IWRK = IWRK1
                XWRK = XWRK1
            Endif
         End Do
         IWRKT (IDEB) = IWRK
         If (XWRK < XMIN) Then
             IMIN = IWRK
             XMIN = XWRK
         End If
         Do ICRS = IDEB+1, IDEB+5
            IWRK = IWRKT (ICRS+1)
            XWRK = XDATT (IWRK)
            IDON = IWRKT(ICRS)
            If (XWRK < XDATT(IDON)) Then
               IWRKT (ICRS+1) = IDON
               IDCR = ICRS
               IWRK1 = IWRKT (IDCR-1)
               XWRK1 = XDATT (IWRK1)
               Do
                  If (XWRK >= XWRK1) Exit
                  IWRKT (IDCR) = IWRK1
                  IDCR = IDCR - 1
                  IWRK1 = IWRKT (IDCR-1)
                  XWRK1 = XDATT (IWRK1)
               End Do
               IWRKT (IDCR) = IWRK
            EndIf
         End Do
         If (XWRK > XMAX) Then
             IMAX = IWRK
             XMAX = XWRK
         End If
      End Do
!
!  Add-up alternatively MAX and MIN values to make the number of data
!  an exact multiple of 7.
!
      IDEB = 7 * (NDAT/7)
      NTRI = NDAT
      If (IDEB < NDAT) Then
!
         Do ICRS = IDEB+1, NDAT
            XWRK1 = XDATT (IWRKT (ICRS))
            IF (XWRK1 > XMAX) Then
               IMAX = IWRKT (ICRS)
               XMAX = XWRK1
            End If
            IF (XWRK1 < XMIN) Then
               IMIN = IWRKT (ICRS)
               XMIN = XWRK1
            End If
         End Do
         IWRK1 = IMAX
         Do ICRS = NDAT+1, IDEB+7
               IWRKT (ICRS) = IWRK1
               If (IWRK1 == IMAX) Then
                  IWRK1 = IMIN
               Else
                  NMED = NMED + 1
                  IWRK1 = IMAX
               End If
         End Do
!
         Do ICRS = IDEB+2, IDEB+7
            IWRK = IWRKT (ICRS)
            XWRK = XDATT (IWRK)
            Do IDCR = ICRS - 1, IDEB+1, - 1
               If (XWRK >= XDATT (IWRKT(IDCR))) Exit
               IWRKT (IDCR+1) = IWRKT (IDCR)
            End Do
            IWRKT (IDCR+1) = IWRK
         End Do
!
         NTRI = IDEB+7
      End If
!
!  Make the set of the indices of median values of each sorted subset
!
         IDON1 = 0
         Do IDON = 1, NTRI, 7
            IDON1 = IDON1 + 1
            IMEDT (IDON1) = IWRKT (IDON + 3)
         End Do
!
!  Find XMED7, the median of the medians
!
         Call i_med (XDATT, IMEDT(1:IDON1), IMED7)
         XMED7 = XDATT (IMED7)
!
!  Count how many values are not higher than (and how many equal to) XMED7
!  This number is at least 4 * 1/2 * (N/7) : 4 values in each of the
!  subsets where the median is lower than the median of medians. For similar
!  reasons, we also have at least 2N/7 values not lower than XMED7. At the
!  same time, we find in each subset the index of the last value < XMED7,
!  and that of the first > XMED7. These indices will be used to restrict the
!  search for the median as the Kth element in the subset (> or <) where
!  we know it to be.
!
         IDON1 = 1
         NLEQ = 0
         NEQU = 0
         Do IDON = 1, NTRI, 7
            IMED = IDON+3
            If (XDATT (IWRKT (IMED)) > XMED7) Then
                  IMED = IMED - 2
                  If (XDATT (IWRKT (IMED)) > XMED7) Then
                     IMED = IMED - 1
                  Else If (XDATT (IWRKT (IMED)) < XMED7) Then
                     IMED = IMED + 1
                  Endif
            Else If (XDATT (IWRKT (IMED)) < XMED7) Then
                  IMED = IMED + 2
                  If (XDATT (IWRKT (IMED)) > XMED7) Then
                     IMED = IMED - 1
                  Else If (XDATT (IWRKT (IMED)) < XMED7) Then
                     IMED = IMED + 1
                  Endif
            Endif
            If (XDATT (IWRKT (IMED)) > XMED7) Then
               NLEQ = NLEQ + IMED - IDON
               IENDT (IDON1) = IMED - 1
               ISTRT (IDON1) = IMED
            Else If (XDATT (IWRKT (IMED)) < XMED7) Then
               NLEQ = NLEQ + IMED - IDON + 1
               IENDT (IDON1) = IMED
               ISTRT (IDON1) = IMED + 1
            Else                    !       If (XDATT (IWRKT (IMED)) == XMED7)
               NLEQ = NLEQ + IMED - IDON + 1
               NEQU = NEQU + 1
               IENDT (IDON1) = IMED - 1
               Do IMED1 = IMED - 1, IDON, -1
                  If (XDATT (IWRKT (IMED1)) == XMED7) Then
                     NEQU = NEQU + 1
                     IENDT (IDON1) = IMED1 - 1
                  Else
                     Exit
                  End If
               End Do
               ISTRT (IDON1) = IMED + 1
               Do IMED1 = IMED + 1, IDON + 6
                  If (XDATT (IWRKT (IMED1)) == XMED7) Then
                     NEQU = NEQU + 1
                     NLEQ = NLEQ + 1
                     ISTRT (IDON1) = IMED1 + 1
                  Else
                     Exit
                  End If
               End Do
            Endif
            IDON1 = IDON1 + 1
         End Do
!
!  Carry out a partial insertion sort to find the Kth smallest of the
!  large values, or the Kth largest of the small values, according to
!  what is needed.
!
!
         If (NLEQ - NEQU + 1 <= NMED) Then
            If (NLEQ < NMED) Then   !      Not enough low values
                IWRK1 = IMAX
                XWRK1 = XDATT (IWRK1)
                NORD = NMED - NLEQ
                IDON1 = 0
                ICRS1 = 1
                ICRS2 = 0
                IDCR = 0
                Do IDON = 1, NTRI, 7
                   IDON1 = IDON1 + 1
                   If (ICRS2 < NORD) Then
                      Do ICRS = ISTRT (IDON1), IDON + 6
                         If (XDATT (IWRKT (ICRS)) < XWRK1) Then
                            IWRK = IWRKT (ICRS)
                            XWRK = XDATT (IWRK)
                            Do IDCR = ICRS1 - 1, 1, - 1
                               If (XWRK >= XDATT (IWRKT (IDCR))) Exit
                               IWRKT  (IDCR+1) = IWRKT (IDCR)
                            End Do
                            IWRKT (IDCR+1) = IWRK
                            IWRK1 = IWRKT (ICRS1)
                            XWRK1 = XDATT (IWRK1)
                         Else
                           If (ICRS2 < NORD) Then
                              IWRKT (ICRS1) = IWRKT (ICRS)
                              IWRK1 = IWRKT (ICRS1)
                              XWRK1 = XDATT (IWRK1)
                           Endif
                         End If
                         ICRS1 = MIN (NORD, ICRS1 + 1)
                         ICRS2 = MIN (NORD, ICRS2 + 1)
                      End Do
                   Else
                      Do ICRS = ISTRT (IDON1), IDON + 6
                         If (XDATT (IWRKT (ICRS)) >= XWRK1) Exit
                         IWRK = IWRKT (ICRS)
                         XWRK = XDATT (IWRK)
                         Do IDCR = ICRS1 - 1, 1, - 1
                               If (XWRK >= XDATT (IWRKT (IDCR))) Exit
                               IWRKT  (IDCR+1) = IWRKT (IDCR)
                         End Do
                         IWRKT (IDCR+1) = IWRK
                         IWRK1 = IWRKT (ICRS1)
                         XWRK1 = XDATT (IWRK1)
                      End Do
                   End If
                End Do
                ires_med = IWRK1
                Return
            Else
                ires_med = IMED7
                Return
            End If
         Else                       !      If (NLEQ > NMED)
!                                          Not enough high values
                XWRK1 = -XHUGE
                NORD = NLEQ - NEQU - NMED + 1
                IDON1 = 0
                ICRS1 = 1
                ICRS2 = 0
                Do IDON = 1, NTRI, 7
                   IDON1 = IDON1 + 1
                   If (ICRS2 < NORD) Then
!
                      Do ICRS = IDON, IENDT (IDON1)
                         If (XDATT(IWRKT (ICRS)) > XWRK1) Then
                            IWRK = IWRKT (ICRS)
                            XWRK = XDATT (IWRK)
                            IDCR = ICRS1 - 1
                            Do IDCR = ICRS1 - 1, 1, - 1
                               If (XWRK <= XDATT(IWRKT (IDCR))) Exit
                               IWRKT (IDCR+1) = IWRKT (IDCR)
                            End Do
                            IWRKT (IDCR+1) = IWRK
                            IWRK1 = IWRKT(ICRS1)
                            XWRK1 = XDATT(IWRK1)
                         Else
                            If (ICRS2 < NORD) Then
                               IWRKT (ICRS1) = IWRKT (ICRS)
                               IWRK1 = IWRKT(ICRS1)
                               XWRK1 = XDATT(IWRK1)
                            End If
                         End If
                         ICRS1 = MIN (NORD, ICRS1 + 1)
                         ICRS2 = MIN (NORD, ICRS2 + 1)
                      End Do
                   Else
                      Do ICRS = IENDT (IDON1), IDON, -1
                         If (XDATT(IWRKT (ICRS)) <= XWRK1) Exit
                         IWRK = IWRKT (ICRS)
                         XWRK = XDATT (IWRK)
                         IDCR = ICRS1 - 1
                         Do IDCR = ICRS1 - 1, 1, - 1
                               If (XWRK <= XDATT(IWRKT (IDCR))) Exit
                               IWRKT (IDCR+1) = IWRKT (IDCR)
                         End Do
                         IWRKT (IDCR+1) = IWRK
                         IWRK1 = IWRKT(ICRS1)
                         XWRK1 = XDATT(IWRK1)
                      End Do
                   Endif
                End Do
!
                ires_med = IWRK1
                Return
         End If
!
   END Subroutine i_med
end module m_indmed
Module m_indnth
Integer, Parameter :: kdp = selected_real_kind(15)
public :: indnth
private :: kdp
private :: R_indnth, I_indnth, D_indnth
interface indnth
  module procedure d_indnth, r_indnth, i_indnth
end interface indnth
contains

Function D_indnth (XDONT, NORD) Result (INDNTH)
!  Return NORDth value of XDONT, i.e fractile of order NORD/SIZE(XDONT).
! __________________________________________________________
!  This routine uses a pivoting strategy such as the one of
!  finding the median based on the quicksort algorithm, but
!  we skew the pivot choice to try to bring it to NORD as
!  fast as possible. It uses 2 temporary arrays, where it
!  stores the indices of the values smaller than the pivot
!  (ILOWT), and the indices of values larger than the pivot
!  that we might still need later on (IHIGT). It iterates
!  until it can bring the number of values in ILOWT to
!  exactly NORD, and then finds the maximum of this set.
!  Michel Olagnon - Aug. 2000
! __________________________________________________________
! __________________________________________________________
      Real (kind=kdp), Dimension (:), Intent (In) :: XDONT
      Integer :: INDNTH
      Integer, Intent (In) :: NORD
! __________________________________________________________
      Real (kind=kdp) :: XPIV, XPIV0, XWRK, XWRK1, XMIN, XMAX
!
      Integer, Dimension (NORD) :: IRNGT
      Integer, Dimension (SIZE(XDONT)) :: ILOWT, IHIGT
      Integer :: NDON, JHIG, JLOW, IHIG, IWRK, IWRK1, IWRK2, IWRK3
      Integer :: IMIL, IFIN, ICRS, IDCR, ILOW
      Integer :: JLM2, JLM1, JHM2, JHM1, INTH
!
      NDON = SIZE (XDONT)
      INTH = NORD
!
!    First loop is used to fill-in ILOWT, IHIGT at the same time
!
      If (NDON < 2) Then
         If (INTH == 1) INDNTH = 1
         Return
      End If
!
!  One chooses a pivot, best estimate possible to put fractile near
!  mid-point of the set of low values.
!
      If (XDONT(2) < XDONT(1)) Then
         ILOWT (1) = 2
         IHIGT (1) = 1
      Else
         ILOWT (1) = 1
         IHIGT (1) = 2
      End If
!
      If (NDON < 3) Then
         If (INTH == 1) INDNTH = ILOWT (1)
         If (INTH == 2) INDNTH = IHIGT (1)
         Return
      End If
!
      If (XDONT(3) < XDONT(IHIGT(1))) Then
         IHIGT (2) = IHIGT (1)
         If (XDONT(3) < XDONT(ILOWT(1))) Then
            IHIGT (1) = ILOWT (1)
            ILOWT (1) = 3
         Else
            IHIGT (1) = 3
         End If
      Else
         IHIGT (2) = 3
      End If
!
      If (NDON < 4) Then
         If (INTH == 1) INDNTH = ILOWT (1)
         If (INTH == 2) INDNTH = IHIGT (1)
         If (INTH == 3) INDNTH = IHIGT (2)
         Return
      End If
!
      If (XDONT(NDON) < XDONT(IHIGT(1))) Then
         IHIGT (3) = IHIGT (2)
         IHIGT (2) = IHIGT (1)
         If (XDONT(NDON) < XDONT(ILOWT(1))) Then
            IHIGT (1) = ILOWT (1)
            ILOWT (1) = NDON
         Else
            IHIGT (1) = NDON
         End If
      Else
         IHIGT (3) = NDON
      End If
!
      If (NDON < 5) Then
         If (INTH == 1) INDNTH = ILOWT (1)
         If (INTH == 2) INDNTH = IHIGT (1)
         If (INTH == 3) INDNTH = IHIGT (2)
         If (INTH == 4) INDNTH = IHIGT (3)
         Return
      End If
!

      JLOW = 1
      JHIG = 3
      XPIV = XDONT (ILOWT(1)) + REAL(2*INTH)/REAL(NDON+INTH) * &
                                   (XDONT(IHIGT(3))-XDONT(ILOWT(1)))
      If (XPIV >= XDONT(IHIGT(1))) Then
         XPIV = XDONT (ILOWT(1)) + REAL(2*INTH)/REAL(NDON+INTH) * &
                                      (XDONT(IHIGT(2))-XDONT(ILOWT(1)))
         If (XPIV >= XDONT(IHIGT(1))) &
             XPIV = XDONT (ILOWT(1)) + REAL (2*INTH) / REAL (NDON+INTH) * &
                                          (XDONT(IHIGT(1))-XDONT(ILOWT(1)))
      End If
      XPIV0 = XPIV
!
!  One puts values > pivot in the end and those <= pivot
!  at the beginning. This is split in 2 cases, so that
!  we can skip the loop test a number of times.
!  As we are also filling in the work arrays at the same time
!  we stop filling in the IHIGT array as soon as we have more
!  than enough values in ILOWT.
!
!
      If (XDONT(NDON) > XPIV) Then
         ICRS = 3
         Do
            ICRS = ICRS + 1
            If (XDONT(ICRS) > XPIV) Then
               If (ICRS >= NDON) Exit
               JHIG = JHIG + 1
               IHIGT (JHIG) = ICRS
            Else
               JLOW = JLOW + 1
               ILOWT (JLOW) = ICRS
               If (JLOW >= INTH) Exit
            End If
         End Do
!
!  One restricts further processing because it is no use
!  to store more high values
!
         If (ICRS < NDON-1) Then
            Do
               ICRS = ICRS + 1
               If (XDONT(ICRS) <= XPIV) Then
                  JLOW = JLOW + 1
                  ILOWT (JLOW) = ICRS
               Else If (ICRS >= NDON) Then
                  Exit
               End If
            End Do
         End If
!
!
      Else
!
!  Same as above, but this is not as easy to optimize, so the
!  DO-loop is kept
!
         Do ICRS = 4, NDON - 1
            If (XDONT(ICRS) > XPIV) Then
               JHIG = JHIG + 1
               IHIGT (JHIG) = ICRS
            Else
               JLOW = JLOW + 1
               ILOWT (JLOW) = ICRS
               If (JLOW >= INTH) Exit
            End If
         End Do
!
         If (ICRS < NDON-1) Then
            Do
               ICRS = ICRS + 1
               If (XDONT(ICRS) <= XPIV) Then
                  If (ICRS >= NDON) Exit
                  JLOW = JLOW + 1
                  ILOWT (JLOW) = ICRS
               End If
            End Do
         End If
      End If
!
      JLM2 = 0
      JLM1 = 0
      JHM2 = 0
      JHM1 = 0
      Do
         If (JLM2 == JLOW .And. JHM2 == JHIG) Then
!
!   We are oscillating. Perturbate by bringing JLOW closer by one
!   to INTH
!
             If (INTH > JLOW) Then
                XMIN = XDONT (IHIGT(1))
                IHIG = 1
                Do ICRS = 2, JHIG
                   If (XDONT(IHIGT(ICRS)) < XMIN) Then
                      XMIN = XDONT (IHIGT(ICRS))
                      IHIG = ICRS
                   End If
                End Do
!
                JLOW = JLOW + 1
                ILOWT (JLOW) = IHIGT (IHIG)
                IHIGT (IHIG) = IHIGT (JHIG)
                JHIG = JHIG - 1
             Else

                ILOW = ILOWT (1)
                XMAX = XDONT (ILOW)
                Do ICRS = 2, JLOW
                   If (XDONT(ILOWT(ICRS)) > XMAX) Then
                      IWRK = ILOWT (ICRS)
                      XMAX = XDONT (IWRK)
                      ILOWT (ICRS) = ILOW
                      ILOW = IWRK
                   End If
                End Do
                JLOW = JLOW - 1
             End If
         End If
         JLM2 = JLM1
         JLM1 = JLOW
         JHM2 = JHM1
         JHM1 = JHIG
!
!   We try to bring the number of values in the low values set
!   closer to INTH.
!
         Select Case (INTH-JLOW)
         Case (2:)
!
!   Not enough values in low part, at least 2 are missing
!
            INTH = INTH - JLOW
            JLOW = 0
            Select Case (JHIG)
!!!!!           CASE DEFAULT
!!!!!              write (unit=*,fmt=*) "Assertion failed"
!!!!!              STOP
!
!   We make a special case when we have so few values in
!   the high values set that it is bad performance to choose a pivot
!   and apply the general algorithm.
!
            Case (2)
               If (XDONT(IHIGT(1)) <= XDONT(IHIGT(2))) Then
                  JLOW = JLOW + 1
                  ILOWT (JLOW) = IHIGT (1)
                  JLOW = JLOW + 1
                  ILOWT (JLOW) = IHIGT (2)
               Else
                  JLOW = JLOW + 1
                  ILOWT (JLOW) = IHIGT (2)
                  JLOW = JLOW + 1
                  ILOWT (JLOW) = IHIGT (1)
               End If
               Exit
!
            Case (3)
!
!
               IWRK1 = IHIGT (1)
               IWRK2 = IHIGT (2)
               IWRK3 = IHIGT (3)
               If (XDONT(IWRK2) < XDONT(IWRK1)) Then
                  IHIGT (1) = IWRK2
                  IHIGT (2) = IWRK1
                  IWRK2 = IWRK1
               End If
               If (XDONT(IWRK2) > XDONT(IWRK3)) Then
                  IHIGT (3) = IWRK2
                  IHIGT (2) = IWRK3
                  IWRK2 = IWRK3
                  If (XDONT(IWRK2) < XDONT(IHIGT(1))) Then
                     IHIGT (2) = IHIGT (1)
                     IHIGT (1) = IWRK2
                  End If
               End If
               JHIG = 0
               Do ICRS = JLOW + 1, INTH
                  JHIG = JHIG + 1
                  ILOWT (ICRS) = IHIGT (JHIG)
               End Do
               JLOW = INTH
               Exit
!
            Case (4:)
!
!
               XPIV0 = XPIV
               IFIN = JHIG
!
!  One chooses a pivot from the 2 first values and the last one.
!  This should ensure sufficient renewal between iterations to
!  avoid worst case behavior effects.
!
               IWRK1 = IHIGT (1)
               IWRK2 = IHIGT (2)
               IWRK3 = IHIGT (IFIN)
               If (XDONT(IWRK2) < XDONT(IWRK1)) Then
                  IHIGT (1) = IWRK2
                  IHIGT (2) = IWRK1
                  IWRK2 = IWRK1
               End If
               If (XDONT(IWRK2) > XDONT(IWRK3)) Then
                  IHIGT (IFIN) = IWRK2
                  IHIGT (2) = IWRK3
                  IWRK2 = IWRK3
                  If (XDONT(IWRK2) < XDONT(IHIGT(1))) Then
                     IHIGT (2) = IHIGT (1)
                     IHIGT (1) = IWRK2
                  End If
               End If
!
               IWRK1 = IHIGT (1)
               JLOW = JLOW + 1
               ILOWT (JLOW) = IWRK1
               XPIV = XDONT (IWRK1) + 0.5 * (XDONT(IHIGT(IFIN))-XDONT(IWRK1))
!
!  One takes values <= pivot to ILOWT
!  Again, 2 parts, one where we take care of the remaining
!  high values because we might still need them, and the
!  other when we know that we will have more than enough
!  low values in the end.
!
               JHIG = 0
               Do ICRS = 2, IFIN
                  If (XDONT(IHIGT(ICRS)) <= XPIV) Then
                     JLOW = JLOW + 1
                     ILOWT (JLOW) = IHIGT (ICRS)
                     If (JLOW >= INTH) Exit
                  Else
                     JHIG = JHIG + 1
                     IHIGT (JHIG) = IHIGT (ICRS)
                  End If
               End Do
!
               Do ICRS = ICRS + 1, IFIN
                  If (XDONT(IHIGT(ICRS)) <= XPIV) Then
                     JLOW = JLOW + 1
                     ILOWT (JLOW) = IHIGT (ICRS)
                  End If
               End Do
            End Select
!
!
         Case (1)
!
!  Only 1 value is missing in low part
!
            XMIN = XDONT (IHIGT(1))
            IHIG = 1
            Do ICRS = 2, JHIG
               If (XDONT(IHIGT(ICRS)) < XMIN) Then
                  XMIN = XDONT (IHIGT(ICRS))
                  IHIG = ICRS
               End If
            End Do
!
            INDNTH = IHIGT (IHIG)
            Return
!
!
         Case (0)
!
!  Low part is exactly what we want
!
            Exit
!
!
         Case (-5:-1)
!
!  Only few values too many in low part
!
            IRNGT (1) = ILOWT (1)
            ILOW = 1 + INTH - JLOW
            Do ICRS = 2, INTH
               IWRK = ILOWT (ICRS)
               XWRK = XDONT (IWRK)
               Do IDCR = ICRS - 1, MAX (1, ILOW), - 1
                  If (XWRK < XDONT(IRNGT(IDCR))) Then
                     IRNGT (IDCR+1) = IRNGT (IDCR)
                  Else
                     Exit
                  End If
               End Do
               IRNGT (IDCR+1) = IWRK
               ILOW = ILOW + 1
            End Do
!
            XWRK1 = XDONT (IRNGT(INTH))
            ILOW = 2*INTH - JLOW
            Do ICRS = INTH + 1, JLOW
               If (XDONT(ILOWT (ICRS)) < XWRK1) Then
                  XWRK = XDONT (ILOWT (ICRS))
                  Do IDCR = INTH - 1, MAX (1, ILOW), - 1
                     If (XWRK >= XDONT(IRNGT(IDCR))) Exit
                     IRNGT (IDCR+1) = IRNGT (IDCR)
                  End Do
                  IRNGT (IDCR+1) = ILOWT (ICRS)
                  XWRK1 = XDONT (IRNGT(INTH))
               End If
               ILOW = ILOW + 1
            End Do
!
            INDNTH = IRNGT(INTH)
            Return
!
!
         Case (:-6)
!
! last case: too many values in low part
!

            IMIL = (JLOW+1) / 2
            IFIN = JLOW
!
!  One chooses a pivot from 1st, last, and middle values
!
            If (XDONT(ILOWT(IMIL)) < XDONT(ILOWT(1))) Then
               IWRK = ILOWT (1)
               ILOWT (1) = ILOWT (IMIL)
               ILOWT (IMIL) = IWRK
            End If
            If (XDONT(ILOWT(IMIL)) > XDONT(ILOWT(IFIN))) Then
               IWRK = ILOWT (IFIN)
               ILOWT (IFIN) = ILOWT (IMIL)
               ILOWT (IMIL) = IWRK
               If (XDONT(ILOWT(IMIL)) < XDONT(ILOWT(1))) Then
                  IWRK = ILOWT (1)
                  ILOWT (1) = ILOWT (IMIL)
                  ILOWT (IMIL) = IWRK
               End If
            End If
            If (IFIN <= 3) Exit
!
            XPIV = XDONT (ILOWT(1)) + REAL(INTH)/REAL(JLOW+INTH) * &
                                      (XDONT(ILOWT(IFIN))-XDONT(ILOWT(1)))

!
!  One takes values > XPIV to IHIGT
!
            JHIG = 0
            JLOW = 0
!
            If (XDONT(ILOWT(IFIN)) > XPIV) Then
               ICRS = 0
               Do
                  ICRS = ICRS + 1
                  If (XDONT(ILOWT(ICRS)) > XPIV) Then
                     JHIG = JHIG + 1
                     IHIGT (JHIG) = ILOWT (ICRS)
                     If (ICRS >= IFIN) Exit
                  Else
                     JLOW = JLOW + 1
                     ILOWT (JLOW) = ILOWT (ICRS)
                     If (JLOW >= INTH) Exit
                  End If
               End Do
!
               If (ICRS < IFIN) Then
                  Do
                     ICRS = ICRS + 1
                     If (XDONT(ILOWT(ICRS)) <= XPIV) Then
                        JLOW = JLOW + 1
                        ILOWT (JLOW) = ILOWT (ICRS)
                     Else
                        If (ICRS >= IFIN) Exit
                     End If
                  End Do
               End If
            Else
               Do ICRS = 1, IFIN
                  If (XDONT(ILOWT(ICRS)) > XPIV) Then
                     JHIG = JHIG + 1
                     IHIGT (JHIG) = ILOWT (ICRS)
                  Else
                     JLOW = JLOW + 1
                     ILOWT (JLOW) = ILOWT (ICRS)
                     If (JLOW >= INTH) Exit
                  End If
               End Do
!
               Do ICRS = ICRS + 1, IFIN
                  If (XDONT(ILOWT(ICRS)) <= XPIV) Then
                     JLOW = JLOW + 1
                     ILOWT (JLOW) = ILOWT (ICRS)
                  End If
               End Do
            End If
!
         End Select
!
      End Do
!
!  Now, we only need to find maximum of the 1:INTH set
!

      IWRK1 = ILOWT (1)
      XWRK1 =  XDONT (IWRK1)
      Do ICRS = 1+1, INTH
         IWRK = ILOWT (ICRS)
         XWRK = XDONT (IWRK)
         If (XWRK > XWRK1) Then
            XWRK1 = XWRK
            IWRK1 = IWRK
         End If
      End Do
      INDNTH = IWRK1
      Return
!
!
End Function D_indnth

Function R_indnth (XDONT, NORD) Result (INDNTH)
!  Return NORDth value of XDONT, i.e fractile of order NORD/SIZE(XDONT).
! __________________________________________________________
!  This routine uses a pivoting strategy such as the one of
!  finding the median based on the quicksort algorithm, but
!  we skew the pivot choice to try to bring it to NORD as
!  fast as possible. It uses 2 temporary arrays, where it
!  stores the indices of the values smaller than the pivot
!  (ILOWT), and the indices of values larger than the pivot
!  that we might still need later on (IHIGT). It iterates
!  until it can bring the number of values in ILOWT to
!  exactly NORD, and then finds the maximum of this set.
!  Michel Olagnon - Aug. 2000
! __________________________________________________________
! _________________________________________________________
      Real, Dimension (:), Intent (In) :: XDONT
      Integer :: INDNTH
      Integer, Intent (In) :: NORD
! __________________________________________________________
      Real :: XPIV, XPIV0, XWRK, XWRK1, XMIN, XMAX
!
      Integer, Dimension (NORD) :: IRNGT
      Integer, Dimension (SIZE(XDONT)) :: ILOWT, IHIGT
      Integer :: NDON, JHIG, JLOW, IHIG, IWRK, IWRK1, IWRK2, IWRK3
      Integer :: IMIL, IFIN, ICRS, IDCR, ILOW
      Integer :: JLM2, JLM1, JHM2, JHM1, INTH
!
      NDON = SIZE (XDONT)
      INTH = NORD
!
!    First loop is used to fill-in ILOWT, IHIGT at the same time
!
      If (NDON < 2) Then
         If (INTH == 1) INDNTH = 1
         Return
      End If
!
!  One chooses a pivot, best estimate possible to put fractile near
!  mid-point of the set of low values.
!
      If (XDONT(2) < XDONT(1)) Then
         ILOWT (1) = 2
         IHIGT (1) = 1
      Else
         ILOWT (1) = 1
         IHIGT (1) = 2
      End If
!
      If (NDON < 3) Then
         If (INTH == 1) INDNTH = ILOWT (1)
         If (INTH == 2) INDNTH = IHIGT (1)
         Return
      End If
!
      If (XDONT(3) < XDONT(IHIGT(1))) Then
         IHIGT (2) = IHIGT (1)
         If (XDONT(3) < XDONT(ILOWT(1))) Then
            IHIGT (1) = ILOWT (1)
            ILOWT (1) = 3
         Else
            IHIGT (1) = 3
         End If
      Else
         IHIGT (2) = 3
      End If
!
      If (NDON < 4) Then
         If (INTH == 1) INDNTH = ILOWT (1)
         If (INTH == 2) INDNTH = IHIGT (1)
         If (INTH == 3) INDNTH = IHIGT (2)
         Return
      End If
!
      If (XDONT(NDON) < XDONT(IHIGT(1))) Then
         IHIGT (3) = IHIGT (2)
         IHIGT (2) = IHIGT (1)
         If (XDONT(NDON) < XDONT(ILOWT(1))) Then
            IHIGT (1) = ILOWT (1)
            ILOWT (1) = NDON
         Else
            IHIGT (1) = NDON
         End If
      Else
         IHIGT (3) = NDON
      End If
!
      If (NDON < 5) Then
         If (INTH == 1) INDNTH = ILOWT (1)
         If (INTH == 2) INDNTH = IHIGT (1)
         If (INTH == 3) INDNTH = IHIGT (2)
         If (INTH == 4) INDNTH = IHIGT (3)
         Return
      End If
!

      JLOW = 1
      JHIG = 3
      XPIV = XDONT (ILOWT(1)) + REAL(2*INTH)/REAL(NDON+INTH) * &
                                   (XDONT(IHIGT(3))-XDONT(ILOWT(1)))
      If (XPIV >= XDONT(IHIGT(1))) Then
         XPIV = XDONT (ILOWT(1)) + REAL(2*INTH)/REAL(NDON+INTH) * &
                                      (XDONT(IHIGT(2))-XDONT(ILOWT(1)))
         If (XPIV >= XDONT(IHIGT(1))) &
             XPIV = XDONT (ILOWT(1)) + REAL (2*INTH) / REAL (NDON+INTH) * &
                                          (XDONT(IHIGT(1))-XDONT(ILOWT(1)))
      End If
      XPIV0 = XPIV
!
!  One puts values > pivot in the end and those <= pivot
!  at the beginning. This is split in 2 cases, so that
!  we can skip the loop test a number of times.
!  As we are also filling in the work arrays at the same time
!  we stop filling in the IHIGT array as soon as we have more
!  than enough values in ILOWT.
!
!
      If (XDONT(NDON) > XPIV) Then
         ICRS = 3
         Do
            ICRS = ICRS + 1
            If (XDONT(ICRS) > XPIV) Then
               If (ICRS >= NDON) Exit
               JHIG = JHIG + 1
               IHIGT (JHIG) = ICRS
            Else
               JLOW = JLOW + 1
               ILOWT (JLOW) = ICRS
               If (JLOW >= INTH) Exit
            End If
         End Do
!
!  One restricts further processing because it is no use
!  to store more high values
!
         If (ICRS < NDON-1) Then
            Do
               ICRS = ICRS + 1
               If (XDONT(ICRS) <= XPIV) Then
                  JLOW = JLOW + 1
                  ILOWT (JLOW) = ICRS
               Else If (ICRS >= NDON) Then
                  Exit
               End If
            End Do
         End If
!
!
      Else
!
!  Same as above, but this is not as easy to optimize, so the
!  DO-loop is kept
!
         Do ICRS = 4, NDON - 1
            If (XDONT(ICRS) > XPIV) Then
               JHIG = JHIG + 1
               IHIGT (JHIG) = ICRS
            Else
               JLOW = JLOW + 1
               ILOWT (JLOW) = ICRS
               If (JLOW >= INTH) Exit
            End If
         End Do
!
         If (ICRS < NDON-1) Then
            Do
               ICRS = ICRS + 1
               If (XDONT(ICRS) <= XPIV) Then
                  If (ICRS >= NDON) Exit
                  JLOW = JLOW + 1
                  ILOWT (JLOW) = ICRS
               End If
            End Do
         End If
      End If
!
      JLM2 = 0
      JLM1 = 0
      JHM2 = 0
      JHM1 = 0
      Do
         If (JLM2 == JLOW .And. JHM2 == JHIG) Then
!
!   We are oscillating. Perturbate by bringing JLOW closer by one
!   to INTH
!
             If (INTH > JLOW) Then
                XMIN = XDONT (IHIGT(1))
                IHIG = 1
                Do ICRS = 2, JHIG
                   If (XDONT(IHIGT(ICRS)) < XMIN) Then
                      XMIN = XDONT (IHIGT(ICRS))
                      IHIG = ICRS
                   End If
                End Do
!
                JLOW = JLOW + 1
                ILOWT (JLOW) = IHIGT (IHIG)
                IHIGT (IHIG) = IHIGT (JHIG)
                JHIG = JHIG - 1
             Else

                ILOW = ILOWT (1)
                XMAX = XDONT (ILOW)
                Do ICRS = 2, JLOW
                   If (XDONT(ILOWT(ICRS)) > XMAX) Then
                      IWRK = ILOWT (ICRS)
                      XMAX = XDONT (IWRK)
                      ILOWT (ICRS) = ILOW
                      ILOW = IWRK
                   End If
                End Do
                JLOW = JLOW - 1
             End If
         End If
         JLM2 = JLM1
         JLM1 = JLOW
         JHM2 = JHM1
         JHM1 = JHIG
!
!   We try to bring the number of values in the low values set
!   closer to INTH.
!
         Select Case (INTH-JLOW)
         Case (2:)
!
!   Not enough values in low part, at least 2 are missing
!
            INTH = INTH - JLOW
            JLOW = 0
            Select Case (JHIG)
!!!!!           CASE DEFAULT
!!!!!              write (unit=*,fmt=*) "Assertion failed"
!!!!!              STOP
!
!   We make a special case when we have so few values in
!   the high values set that it is bad performance to choose a pivot
!   and apply the general algorithm.
!
            Case (2)
               If (XDONT(IHIGT(1)) <= XDONT(IHIGT(2))) Then
                  JLOW = JLOW + 1
                  ILOWT (JLOW) = IHIGT (1)
                  JLOW = JLOW + 1
                  ILOWT (JLOW) = IHIGT (2)
               Else
                  JLOW = JLOW + 1
                  ILOWT (JLOW) = IHIGT (2)
                  JLOW = JLOW + 1
                  ILOWT (JLOW) = IHIGT (1)
               End If
               Exit
!
            Case (3)
!
!
               IWRK1 = IHIGT (1)
               IWRK2 = IHIGT (2)
               IWRK3 = IHIGT (3)
               If (XDONT(IWRK2) < XDONT(IWRK1)) Then
                  IHIGT (1) = IWRK2
                  IHIGT (2) = IWRK1
                  IWRK2 = IWRK1
               End If
               If (XDONT(IWRK2) > XDONT(IWRK3)) Then
                  IHIGT (3) = IWRK2
                  IHIGT (2) = IWRK3
                  IWRK2 = IWRK3
                  If (XDONT(IWRK2) < XDONT(IHIGT(1))) Then
                     IHIGT (2) = IHIGT (1)
                     IHIGT (1) = IWRK2
                  End If
               End If
               JHIG = 0
               Do ICRS = JLOW + 1, INTH
                  JHIG = JHIG + 1
                  ILOWT (ICRS) = IHIGT (JHIG)
               End Do
               JLOW = INTH
               Exit
!
            Case (4:)
!
!
               XPIV0 = XPIV
               IFIN = JHIG
!
!  One chooses a pivot from the 2 first values and the last one.
!  This should ensure sufficient renewal between iterations to
!  avoid worst case behavior effects.
!
               IWRK1 = IHIGT (1)
               IWRK2 = IHIGT (2)
               IWRK3 = IHIGT (IFIN)
               If (XDONT(IWRK2) < XDONT(IWRK1)) Then
                  IHIGT (1) = IWRK2
                  IHIGT (2) = IWRK1
                  IWRK2 = IWRK1
               End If
               If (XDONT(IWRK2) > XDONT(IWRK3)) Then
                  IHIGT (IFIN) = IWRK2
                  IHIGT (2) = IWRK3
                  IWRK2 = IWRK3
                  If (XDONT(IWRK2) < XDONT(IHIGT(1))) Then
                     IHIGT (2) = IHIGT (1)
                     IHIGT (1) = IWRK2
                  End If
               End If
!
               IWRK1 = IHIGT (1)
               JLOW = JLOW + 1
               ILOWT (JLOW) = IWRK1
               XPIV = XDONT (IWRK1) + 0.5 * (XDONT(IHIGT(IFIN))-XDONT(IWRK1))
!
!  One takes values <= pivot to ILOWT
!  Again, 2 parts, one where we take care of the remaining
!  high values because we might still need them, and the
!  other when we know that we will have more than enough
!  low values in the end.
!
               JHIG = 0
               Do ICRS = 2, IFIN
                  If (XDONT(IHIGT(ICRS)) <= XPIV) Then
                     JLOW = JLOW + 1
                     ILOWT (JLOW) = IHIGT (ICRS)
                     If (JLOW >= INTH) Exit
                  Else
                     JHIG = JHIG + 1
                     IHIGT (JHIG) = IHIGT (ICRS)
                  End If
               End Do
!
               Do ICRS = ICRS + 1, IFIN
                  If (XDONT(IHIGT(ICRS)) <= XPIV) Then
                     JLOW = JLOW + 1
                     ILOWT (JLOW) = IHIGT (ICRS)
                  End If
               End Do
            End Select
!
!
         Case (1)
!
!  Only 1 value is missing in low part
!
            XMIN = XDONT (IHIGT(1))
            IHIG = 1
            Do ICRS = 2, JHIG
               If (XDONT(IHIGT(ICRS)) < XMIN) Then
                  XMIN = XDONT (IHIGT(ICRS))
                  IHIG = ICRS
               End If
            End Do
!
            INDNTH = IHIGT (IHIG)
            Return
!
!
         Case (0)
!
!  Low part is exactly what we want
!
            Exit
!
!
         Case (-5:-1)
!
!  Only few values too many in low part
!
            IRNGT (1) = ILOWT (1)
            ILOW = 1 + INTH - JLOW
            Do ICRS = 2, INTH
               IWRK = ILOWT (ICRS)
               XWRK = XDONT (IWRK)
               Do IDCR = ICRS - 1, MAX (1, ILOW), - 1
                  If (XWRK < XDONT(IRNGT(IDCR))) Then
                     IRNGT (IDCR+1) = IRNGT (IDCR)
                  Else
                     Exit
                  End If
               End Do
               IRNGT (IDCR+1) = IWRK
               ILOW = ILOW + 1
            End Do
!
            XWRK1 = XDONT (IRNGT(INTH))
            ILOW = 2*INTH - JLOW
            Do ICRS = INTH + 1, JLOW
               If (XDONT(ILOWT (ICRS)) < XWRK1) Then
                  XWRK = XDONT (ILOWT (ICRS))
                  Do IDCR = INTH - 1, MAX (1, ILOW), - 1
                     If (XWRK >= XDONT(IRNGT(IDCR))) Exit
                     IRNGT (IDCR+1) = IRNGT (IDCR)
                  End Do
                  IRNGT (IDCR+1) = ILOWT (ICRS)
                  XWRK1 = XDONT (IRNGT(INTH))
               End If
               ILOW = ILOW + 1
            End Do
!
            INDNTH = IRNGT(INTH)
            Return
!
!
         Case (:-6)
!
! last case: too many values in low part
!

            IMIL = (JLOW+1) / 2
            IFIN = JLOW
!
!  One chooses a pivot from 1st, last, and middle values
!
            If (XDONT(ILOWT(IMIL)) < XDONT(ILOWT(1))) Then
               IWRK = ILOWT (1)
               ILOWT (1) = ILOWT (IMIL)
               ILOWT (IMIL) = IWRK
            End If
            If (XDONT(ILOWT(IMIL)) > XDONT(ILOWT(IFIN))) Then
               IWRK = ILOWT (IFIN)
               ILOWT (IFIN) = ILOWT (IMIL)
               ILOWT (IMIL) = IWRK
               If (XDONT(ILOWT(IMIL)) < XDONT(ILOWT(1))) Then
                  IWRK = ILOWT (1)
                  ILOWT (1) = ILOWT (IMIL)
                  ILOWT (IMIL) = IWRK
               End If
            End If
            If (IFIN <= 3) Exit
!
            XPIV = XDONT (ILOWT(1)) + REAL(INTH)/REAL(JLOW+INTH) * &
                                      (XDONT(ILOWT(IFIN))-XDONT(ILOWT(1)))

!
!  One takes values > XPIV to IHIGT
!
            JHIG = 0
            JLOW = 0
!
            If (XDONT(ILOWT(IFIN)) > XPIV) Then
               ICRS = 0
               Do
                  ICRS = ICRS + 1
                  If (XDONT(ILOWT(ICRS)) > XPIV) Then
                     JHIG = JHIG + 1
                     IHIGT (JHIG) = ILOWT (ICRS)
                     If (ICRS >= IFIN) Exit
                  Else
                     JLOW = JLOW + 1
                     ILOWT (JLOW) = ILOWT (ICRS)
                     If (JLOW >= INTH) Exit
                  End If
               End Do
!
               If (ICRS < IFIN) Then
                  Do
                     ICRS = ICRS + 1
                     If (XDONT(ILOWT(ICRS)) <= XPIV) Then
                        JLOW = JLOW + 1
                        ILOWT (JLOW) = ILOWT (ICRS)
                     Else
                        If (ICRS >= IFIN) Exit
                     End If
                  End Do
               End If
            Else
               Do ICRS = 1, IFIN
                  If (XDONT(ILOWT(ICRS)) > XPIV) Then
                     JHIG = JHIG + 1
                     IHIGT (JHIG) = ILOWT (ICRS)
                  Else
                     JLOW = JLOW + 1
                     ILOWT (JLOW) = ILOWT (ICRS)
                     If (JLOW >= INTH) Exit
                  End If
               End Do
!
               Do ICRS = ICRS + 1, IFIN
                  If (XDONT(ILOWT(ICRS)) <= XPIV) Then
                     JLOW = JLOW + 1
                     ILOWT (JLOW) = ILOWT (ICRS)
                  End If
               End Do
            End If
!
         End Select
!
      End Do
!
!  Now, we only need to find maximum of the 1:INTH set
!

      IWRK1 = ILOWT (1)
      XWRK1 =  XDONT (IWRK1)
      Do ICRS = 1+1, INTH
         IWRK = ILOWT (ICRS)
         XWRK = XDONT (IWRK)
         If (XWRK > XWRK1) Then
            XWRK1 = XWRK
            IWRK1 = IWRK
         End If
      End Do
      INDNTH = IWRK1
      Return
!
!
End Function R_indnth
Function I_indnth (XDONT, NORD) Result (INDNTH)
!  Return NORDth value of XDONT, i.e fractile of order NORD/SIZE(XDONT).
! __________________________________________________________
!  This routine uses a pivoting strategy such as the one of
!  finding the median based on the quicksort algorithm, but
!  we skew the pivot choice to try to bring it to NORD as
!  fast as possible. It uses 2 temporary arrays, where it
!  stores the indices of the values smaller than the pivot
!  (ILOWT), and the indices of values larger than the pivot
!  that we might still need later on (IHIGT). It iterates
!  until it can bring the number of values in ILOWT to
!  exactly NORD, and then finds the maximum of this set.
!  Michel Olagnon - Aug. 2000
! __________________________________________________________
! __________________________________________________________
      Integer, Dimension (:), Intent (In)  :: XDONT
      Integer :: INDNTH
      Integer, Intent (In) :: NORD
! __________________________________________________________
      Integer :: XPIV, XPIV0, XWRK, XWRK1, XMIN, XMAX
!
      Integer, Dimension (NORD) :: IRNGT
      Integer, Dimension (SIZE(XDONT)) :: ILOWT, IHIGT
      Integer :: NDON, JHIG, JLOW, IHIG, IWRK, IWRK1, IWRK2, IWRK3
      Integer :: IMIL, IFIN, ICRS, IDCR, ILOW
      Integer :: JLM2, JLM1, JHM2, JHM1, INTH
!
      NDON = SIZE (XDONT)
      INTH = NORD
!
!    First loop is used to fill-in ILOWT, IHIGT at the same time
!
      If (NDON < 2) Then
         If (INTH == 1) INDNTH = 1
         Return
      End If
!
!  One chooses a pivot, best estimate possible to put fractile near
!  mid-point of the set of low values.
!
      If (XDONT(2) < XDONT(1)) Then
         ILOWT (1) = 2
         IHIGT (1) = 1
      Else
         ILOWT (1) = 1
         IHIGT (1) = 2
      End If
!
      If (NDON < 3) Then
         If (INTH == 1) INDNTH = ILOWT (1)
         If (INTH == 2) INDNTH = IHIGT (1)
         Return
      End If
!
      If (XDONT(3) < XDONT(IHIGT(1))) Then
         IHIGT (2) = IHIGT (1)
         If (XDONT(3) < XDONT(ILOWT(1))) Then
            IHIGT (1) = ILOWT (1)
            ILOWT (1) = 3
         Else
            IHIGT (1) = 3
         End If
      Else
         IHIGT (2) = 3
      End If
!
      If (NDON < 4) Then
         If (INTH == 1) INDNTH = ILOWT (1)
         If (INTH == 2) INDNTH = IHIGT (1)
         If (INTH == 3) INDNTH = IHIGT (2)
         Return
      End If
!
      If (XDONT(NDON) < XDONT(IHIGT(1))) Then
         IHIGT (3) = IHIGT (2)
         IHIGT (2) = IHIGT (1)
         If (XDONT(NDON) < XDONT(ILOWT(1))) Then
            IHIGT (1) = ILOWT (1)
            ILOWT (1) = NDON
         Else
            IHIGT (1) = NDON
         End If
      Else
         IHIGT (3) = NDON
      End If
!
      If (NDON < 5) Then
         If (INTH == 1) INDNTH = ILOWT (1)
         If (INTH == 2) INDNTH = IHIGT (1)
         If (INTH == 3) INDNTH = IHIGT (2)
         If (INTH == 4) INDNTH = IHIGT (3)
         Return
      End If
!

      JLOW = 1
      JHIG = 3
      XPIV = XDONT (ILOWT(1)) + REAL(2*INTH)/REAL(NDON+INTH) * &
                                   (XDONT(IHIGT(3))-XDONT(ILOWT(1)))
      If (XPIV >= XDONT(IHIGT(1))) Then
         XPIV = XDONT (ILOWT(1)) + REAL(2*INTH)/REAL(NDON+INTH) * &
                                      (XDONT(IHIGT(2))-XDONT(ILOWT(1)))
         If (XPIV >= XDONT(IHIGT(1))) &
             XPIV = XDONT (ILOWT(1)) + REAL (2*INTH) / REAL (NDON+INTH) * &
                                          (XDONT(IHIGT(1))-XDONT(ILOWT(1)))
      End If
      XPIV0 = XPIV
!
!  One puts values > pivot in the end and those <= pivot
!  at the beginning. This is split in 2 cases, so that
!  we can skip the loop test a number of times.
!  As we are also filling in the work arrays at the same time
!  we stop filling in the IHIGT array as soon as we have more
!  than enough values in ILOWT.
!
!
      If (XDONT(NDON) > XPIV) Then
         ICRS = 3
         Do
            ICRS = ICRS + 1
            If (XDONT(ICRS) > XPIV) Then
               If (ICRS >= NDON) Exit
               JHIG = JHIG + 1
               IHIGT (JHIG) = ICRS
            Else
               JLOW = JLOW + 1
               ILOWT (JLOW) = ICRS
               If (JLOW >= INTH) Exit
            End If
         End Do
!
!  One restricts further processing because it is no use
!  to store more high values
!
         If (ICRS < NDON-1) Then
            Do
               ICRS = ICRS + 1
               If (XDONT(ICRS) <= XPIV) Then
                  JLOW = JLOW + 1
                  ILOWT (JLOW) = ICRS
               Else If (ICRS >= NDON) Then
                  Exit
               End If
            End Do
         End If
!
!
      Else
!
!  Same as above, but this is not as easy to optimize, so the
!  DO-loop is kept
!
         Do ICRS = 4, NDON - 1
            If (XDONT(ICRS) > XPIV) Then
               JHIG = JHIG + 1
               IHIGT (JHIG) = ICRS
            Else
               JLOW = JLOW + 1
               ILOWT (JLOW) = ICRS
               If (JLOW >= INTH) Exit
            End If
         End Do
!
         If (ICRS < NDON-1) Then
            Do
               ICRS = ICRS + 1
               If (XDONT(ICRS) <= XPIV) Then
                  If (ICRS >= NDON) Exit
                  JLOW = JLOW + 1
                  ILOWT (JLOW) = ICRS
               End If
            End Do
         End If
      End If
!
      JLM2 = 0
      JLM1 = 0
      JHM2 = 0
      JHM1 = 0
      Do
         If (JLM2 == JLOW .And. JHM2 == JHIG) Then
!
!   We are oscillating. Perturbate by bringing JLOW closer by one
!   to INTH
!
             If (INTH > JLOW) Then
                XMIN = XDONT (IHIGT(1))
                IHIG = 1
                Do ICRS = 2, JHIG
                   If (XDONT(IHIGT(ICRS)) < XMIN) Then
                      XMIN = XDONT (IHIGT(ICRS))
                      IHIG = ICRS
                   End If
                End Do
!
                JLOW = JLOW + 1
                ILOWT (JLOW) = IHIGT (IHIG)
                IHIGT (IHIG) = IHIGT (JHIG)
                JHIG = JHIG - 1
             Else

                ILOW = ILOWT (1)
                XMAX = XDONT (ILOW)
                Do ICRS = 2, JLOW
                   If (XDONT(ILOWT(ICRS)) > XMAX) Then
                      IWRK = ILOWT (ICRS)
                      XMAX = XDONT (IWRK)
                      ILOWT (ICRS) = ILOW
                      ILOW = IWRK
                   End If
                End Do
                JLOW = JLOW - 1
             End If
         End If
         JLM2 = JLM1
         JLM1 = JLOW
         JHM2 = JHM1
         JHM1 = JHIG
!
!   We try to bring the number of values in the low values set
!   closer to INTH.
!
         Select Case (INTH-JLOW)
         Case (2:)
!
!   Not enough values in low part, at least 2 are missing
!
            INTH = INTH - JLOW
            JLOW = 0
            Select Case (JHIG)
!!!!!           CASE DEFAULT
!!!!!              write (unit=*,fmt=*) "Assertion failed"
!!!!!              STOP
!
!   We make a special case when we have so few values in
!   the high values set that it is bad performance to choose a pivot
!   and apply the general algorithm.
!
            Case (2)
               If (XDONT(IHIGT(1)) <= XDONT(IHIGT(2))) Then
                  JLOW = JLOW + 1
                  ILOWT (JLOW) = IHIGT (1)
                  JLOW = JLOW + 1
                  ILOWT (JLOW) = IHIGT (2)
               Else
                  JLOW = JLOW + 1
                  ILOWT (JLOW) = IHIGT (2)
                  JLOW = JLOW + 1
                  ILOWT (JLOW) = IHIGT (1)
               End If
               Exit
!
            Case (3)
!
!
               IWRK1 = IHIGT (1)
               IWRK2 = IHIGT (2)
               IWRK3 = IHIGT (3)
               If (XDONT(IWRK2) < XDONT(IWRK1)) Then
                  IHIGT (1) = IWRK2
                  IHIGT (2) = IWRK1
                  IWRK2 = IWRK1
               End If
               If (XDONT(IWRK2) > XDONT(IWRK3)) Then
                  IHIGT (3) = IWRK2
                  IHIGT (2) = IWRK3
                  IWRK2 = IWRK3
                  If (XDONT(IWRK2) < XDONT(IHIGT(1))) Then
                     IHIGT (2) = IHIGT (1)
                     IHIGT (1) = IWRK2
                  End If
               End If
               JHIG = 0
               Do ICRS = JLOW + 1, INTH
                  JHIG = JHIG + 1
                  ILOWT (ICRS) = IHIGT (JHIG)
               End Do
               JLOW = INTH
               Exit
!
            Case (4:)
!
!
               XPIV0 = XPIV
               IFIN = JHIG
!
!  One chooses a pivot from the 2 first values and the last one.
!  This should ensure sufficient renewal between iterations to
!  avoid worst case behavior effects.
!
               IWRK1 = IHIGT (1)
               IWRK2 = IHIGT (2)
               IWRK3 = IHIGT (IFIN)
               If (XDONT(IWRK2) < XDONT(IWRK1)) Then
                  IHIGT (1) = IWRK2
                  IHIGT (2) = IWRK1
                  IWRK2 = IWRK1
               End If
               If (XDONT(IWRK2) > XDONT(IWRK3)) Then
                  IHIGT (IFIN) = IWRK2
                  IHIGT (2) = IWRK3
                  IWRK2 = IWRK3
                  If (XDONT(IWRK2) < XDONT(IHIGT(1))) Then
                     IHIGT (2) = IHIGT (1)
                     IHIGT (1) = IWRK2
                  End If
               End If
!
               IWRK1 = IHIGT (1)
               JLOW = JLOW + 1
               ILOWT (JLOW) = IWRK1
               XPIV = XDONT (IWRK1) + 0.5 * (XDONT(IHIGT(IFIN))-XDONT(IWRK1))
!
!  One takes values <= pivot to ILOWT
!  Again, 2 parts, one where we take care of the remaining
!  high values because we might still need them, and the
!  other when we know that we will have more than enough
!  low values in the end.
!
               JHIG = 0
               Do ICRS = 2, IFIN
                  If (XDONT(IHIGT(ICRS)) <= XPIV) Then
                     JLOW = JLOW + 1
                     ILOWT (JLOW) = IHIGT (ICRS)
                     If (JLOW >= INTH) Exit
                  Else
                     JHIG = JHIG + 1
                     IHIGT (JHIG) = IHIGT (ICRS)
                  End If
               End Do
!
               Do ICRS = ICRS + 1, IFIN
                  If (XDONT(IHIGT(ICRS)) <= XPIV) Then
                     JLOW = JLOW + 1
                     ILOWT (JLOW) = IHIGT (ICRS)
                  End If
               End Do
            End Select
!
!
         Case (1)
!
!  Only 1 value is missing in low part
!
            XMIN = XDONT (IHIGT(1))
            IHIG = 1
            Do ICRS = 2, JHIG
               If (XDONT(IHIGT(ICRS)) < XMIN) Then
                  XMIN = XDONT (IHIGT(ICRS))
                  IHIG = ICRS
               End If
            End Do
!
            INDNTH = IHIGT (IHIG)
            Return
!
!
         Case (0)
!
!  Low part is exactly what we want
!
            Exit
!
!
         Case (-5:-1)
!
!  Only few values too many in low part
!
            IRNGT (1) = ILOWT (1)
            ILOW = 1 + INTH - JLOW
            Do ICRS = 2, INTH
               IWRK = ILOWT (ICRS)
               XWRK = XDONT (IWRK)
               Do IDCR = ICRS - 1, MAX (1, ILOW), - 1
                  If (XWRK < XDONT(IRNGT(IDCR))) Then
                     IRNGT (IDCR+1) = IRNGT (IDCR)
                  Else
                     Exit
                  End If
               End Do
               IRNGT (IDCR+1) = IWRK
               ILOW = ILOW + 1
            End Do
!
            XWRK1 = XDONT (IRNGT(INTH))
            ILOW = 2*INTH - JLOW
            Do ICRS = INTH + 1, JLOW
               If (XDONT(ILOWT (ICRS)) < XWRK1) Then
                  XWRK = XDONT (ILOWT (ICRS))
                  Do IDCR = INTH - 1, MAX (1, ILOW), - 1
                     If (XWRK >= XDONT(IRNGT(IDCR))) Exit
                     IRNGT (IDCR+1) = IRNGT (IDCR)
                  End Do
                  IRNGT (IDCR+1) = ILOWT (ICRS)
                  XWRK1 = XDONT (IRNGT(INTH))
               End If
               ILOW = ILOW + 1
            End Do
!
            INDNTH = IRNGT(INTH)
            Return
!
!
         Case (:-6)
!
! last case: too many values in low part
!

            IMIL = (JLOW+1) / 2
            IFIN = JLOW
!
!  One chooses a pivot from 1st, last, and middle values
!
            If (XDONT(ILOWT(IMIL)) < XDONT(ILOWT(1))) Then
               IWRK = ILOWT (1)
               ILOWT (1) = ILOWT (IMIL)
               ILOWT (IMIL) = IWRK
            End If
            If (XDONT(ILOWT(IMIL)) > XDONT(ILOWT(IFIN))) Then
               IWRK = ILOWT (IFIN)
               ILOWT (IFIN) = ILOWT (IMIL)
               ILOWT (IMIL) = IWRK
               If (XDONT(ILOWT(IMIL)) < XDONT(ILOWT(1))) Then
                  IWRK = ILOWT (1)
                  ILOWT (1) = ILOWT (IMIL)
                  ILOWT (IMIL) = IWRK
               End If
            End If
            If (IFIN <= 3) Exit
!
            XPIV = XDONT (ILOWT(1)) + REAL(INTH)/REAL(JLOW+INTH) * &
                                      (XDONT(ILOWT(IFIN))-XDONT(ILOWT(1)))

!
!  One takes values > XPIV to IHIGT
!
            JHIG = 0
            JLOW = 0
!
            If (XDONT(ILOWT(IFIN)) > XPIV) Then
               ICRS = 0
               Do
                  ICRS = ICRS + 1
                  If (XDONT(ILOWT(ICRS)) > XPIV) Then
                     JHIG = JHIG + 1
                     IHIGT (JHIG) = ILOWT (ICRS)
                     If (ICRS >= IFIN) Exit
                  Else
                     JLOW = JLOW + 1
                     ILOWT (JLOW) = ILOWT (ICRS)
                     If (JLOW >= INTH) Exit
                  End If
               End Do
!
               If (ICRS < IFIN) Then
                  Do
                     ICRS = ICRS + 1
                     If (XDONT(ILOWT(ICRS)) <= XPIV) Then
                        JLOW = JLOW + 1
                        ILOWT (JLOW) = ILOWT (ICRS)
                     Else
                        If (ICRS >= IFIN) Exit
                     End If
                  End Do
               End If
            Else
               Do ICRS = 1, IFIN
                  If (XDONT(ILOWT(ICRS)) > XPIV) Then
                     JHIG = JHIG + 1
                     IHIGT (JHIG) = ILOWT (ICRS)
                  Else
                     JLOW = JLOW + 1
                     ILOWT (JLOW) = ILOWT (ICRS)
                     If (JLOW >= INTH) Exit
                  End If
               End Do
!
               Do ICRS = ICRS + 1, IFIN
                  If (XDONT(ILOWT(ICRS)) <= XPIV) Then
                     JLOW = JLOW + 1
                     ILOWT (JLOW) = ILOWT (ICRS)
                  End If
               End Do
            End If
!
         End Select
!
      End Do
!
!  Now, we only need to find maximum of the 1:INTH set
!

      IWRK1 = ILOWT (1)
      XWRK1 =  XDONT (IWRK1)
      Do ICRS = 1+1, INTH
         IWRK = ILOWT (ICRS)
         XWRK = XDONT (IWRK)
         If (XWRK > XWRK1) Then
            XWRK1 = XWRK
            IWRK1 = IWRK
         End If
      End Do
      INDNTH = IWRK1
      Return
!
!
End Function I_indnth
end module m_indnth

Module m_inspar
Integer, Parameter :: kdp = selected_real_kind(15)
public :: inspar
private :: kdp
private :: R_inspar, I_inspar, D_inspar
interface inspar
  module procedure d_inspar, r_inspar, i_inspar
end interface inspar
contains

Subroutine D_inspar (XDONT, NORD)
!  Sorts partially XDONT, bringing the NORD lowest values at the
!  begining of the array
! __________________________________________________________
!  This subroutine uses insertion sort, limiting insertion
!  to the first NORD values. It does not use any work array
!  and is faster when NORD is very small (2-5), but worst case
!  behavior can happen fairly probably (initially inverse sorted)
!  In many cases, the refined quicksort method is faster.
!  Michel Olagnon - Feb. 2000
! __________________________________________________________
! __________________________________________________________
      Real (kind=kdp), Dimension (:), Intent (InOut) :: XDONT
      Integer, Intent (In) :: NORD
! __________________________________________________________
      Real (kind=kdp) :: XWRK, XWRK1
!
      Integer :: ICRS, IDCR
!
      Do ICRS = 2, NORD
         XWRK = XDONT (ICRS)
         Do IDCR = ICRS - 1, 1, - 1
            If (XWRK >= XDONT(IDCR)) Exit
            XDONT (IDCR+1) = XDONT (IDCR)
         End Do
         XDONT (IDCR+1) = XWRK
      End Do
!
      XWRK1 = XDONT (NORD)
      Do ICRS = NORD + 1, SIZE (XDONT)
         If (XDONT(ICRS) < XWRK1) Then
            XWRK = XDONT (ICRS)
            XDONT (ICRS) = XWRK1
            Do IDCR = NORD - 1, 1, - 1
               If (XWRK >= XDONT(IDCR)) Exit
               XDONT (IDCR+1) = XDONT (IDCR)
            End Do
            XDONT (IDCR+1) = XWRK
            XWRK1 = XDONT (NORD)
         End If
      End Do
!
!
End Subroutine D_inspar

Subroutine R_inspar (XDONT, NORD)
!  Sorts partially XDONT, bringing the NORD lowest values at the
!  begining of the array
! __________________________________________________________
!  This subroutine uses insertion sort, limiting insertion
!  to the first NORD values. It does not use any work array
!  and is faster when NORD is very small (2-5), but worst case
!  behavior can happen fairly probably (initially inverse sorted)
!  In many cases, the refined quicksort method is faster.
!  Michel Olagnon - Feb. 2000
! __________________________________________________________
! _________________________________________________________
      Real, Dimension (:), Intent (InOut) :: XDONT
      Integer, Intent (In) :: NORD
! __________________________________________________________
      Real    :: XWRK, XWRK1
!
      Integer :: ICRS, IDCR
!
      Do ICRS = 2, NORD
         XWRK = XDONT (ICRS)
         Do IDCR = ICRS - 1, 1, - 1
            If (XWRK >= XDONT(IDCR)) Exit
            XDONT (IDCR+1) = XDONT (IDCR)
         End Do
         XDONT (IDCR+1) = XWRK
      End Do
!
      XWRK1 = XDONT (NORD)
      Do ICRS = NORD + 1, SIZE (XDONT)
         If (XDONT(ICRS) < XWRK1) Then
            XWRK = XDONT (ICRS)
            XDONT (ICRS) = XWRK1
            Do IDCR = NORD - 1, 1, - 1
               If (XWRK >= XDONT(IDCR)) Exit
               XDONT (IDCR+1) = XDONT (IDCR)
            End Do
            XDONT (IDCR+1) = XWRK
            XWRK1 = XDONT (NORD)
         End If
      End Do
!
!
End Subroutine R_inspar
Subroutine I_inspar (XDONT, NORD)
!  Sorts partially XDONT, bringing the NORD lowest values at the
!  begining of the array
! __________________________________________________________
!  This subroutine uses insertion sort, limiting insertion
!  to the first NORD values. It does not use any work array
!  and is faster when NORD is very small (2-5), but worst case
!  behavior can happen fairly probably (initially inverse sorted)
!  In many cases, the refined quicksort method is faster.
!  Michel Olagnon - Feb. 2000
! __________________________________________________________
! __________________________________________________________
      Integer, Dimension (:), Intent (InOut)  :: XDONT
      Integer, Intent (In) :: NORD
! __________________________________________________________
      Integer :: XWRK, XWRK1
!
      Integer :: ICRS, IDCR
!
      Do ICRS = 2, NORD
         XWRK = XDONT (ICRS)
         Do IDCR = ICRS - 1, 1, - 1
            If (XWRK >= XDONT(IDCR)) Exit
            XDONT (IDCR+1) = XDONT (IDCR)
         End Do
         XDONT (IDCR+1) = XWRK
      End Do
!
      XWRK1 = XDONT (NORD)
      Do ICRS = NORD + 1, SIZE (XDONT)
         If (XDONT(ICRS) < XWRK1) Then
            XWRK = XDONT (ICRS)
            XDONT (ICRS) = XWRK1
            Do IDCR = NORD - 1, 1, - 1
               If (XWRK >= XDONT(IDCR)) Exit
               XDONT (IDCR+1) = XDONT (IDCR)
            End Do
            XDONT (IDCR+1) = XWRK
            XWRK1 = XDONT (NORD)
         End If
      End Do
!
!
End Subroutine I_inspar
end module m_inspar
Module m_inssor
Integer, Parameter :: kdp = selected_real_kind(15)
public :: inssor
private :: kdp
private :: R_inssor, I_inssor, D_inssor
interface inssor
  module procedure d_inssor, r_inssor, i_inssor
end interface inssor
contains

Subroutine D_inssor (XDONT)
!  Sorts XDONT into increasing order (Insertion sort)
! __________________________________________________________
!  This subroutine uses insertion sort. It does not use any
!  work array and is faster when XDONT is of very small size
!  (< 20), or already almost sorted, but worst case behavior
!  can happen fairly probably (initially inverse sorted).
!  In many cases, the quicksort or merge sort method is faster.
!  Michel Olagnon - Apr. 2000
! __________________________________________________________
! __________________________________________________________
! __________________________________________________________
      Real (kind=kdp), Dimension (:), Intent (InOut) :: XDONT
! __________________________________________________________
      Real (Kind=kdp) :: XWRK, XMIN
!
! __________________________________________________________
!
      Integer :: ICRS, IDCR, NDON
!
      NDON = Size (XDONT)
!
! We first bring the minimum to the first location in the array.
! That way, we will have a "guard", and when looking for the
! right place to insert a value, no loop test is necessary.
!
      If (XDONT (1) < XDONT (NDON)) Then
          XMIN = XDONT (1)
      Else
          XMIN = XDONT (NDON)
          XDONT (NDON) = XDONT (1)
      Endif
      Do IDCR = NDON-1, 2, -1
         XWRK = XDONT(IDCR)
         IF (XWRK < XMIN) Then
            XDONT (IDCR) = XMIN
            XMIN = XWRK
         End If
      End Do
      XDONT (1) = XMIN
!
! The first value is now the minimum
! Loop over the array, and when a value is smaller than
! the previous one, loop down to insert it at its right place.
!
      Do ICRS = 3, NDON
         XWRK = XDONT (ICRS)
         IDCR = ICRS - 1
         If (XWRK < XDONT(IDCR)) Then
            XDONT (ICRS) = XDONT (IDCR)
            IDCR = IDCR - 1
            Do
               If (XWRK >= XDONT(IDCR)) Exit
               XDONT (IDCR+1) = XDONT (IDCR)
               IDCR = IDCR - 1
            End Do
            XDONT (IDCR+1) = XWRK
         End If
      End Do
!
      Return
!
End Subroutine D_inssor

Subroutine R_inssor (XDONT)
!  Sorts XDONT into increasing order (Insertion sort)
! __________________________________________________________
!  This subroutine uses insertion sort. It does not use any
!  work array and is faster when XDONT is of very small size
!  (< 20), or already almost sorted, but worst case behavior
!  can happen fairly probably (initially inverse sorted).
!  In many cases, the quicksort or merge sort method is faster.
!  Michel Olagnon - Apr. 2000
! __________________________________________________________
! __________________________________________________________
! _________________________________________________________
      Real, Dimension (:), Intent (InOut) :: XDONT
! __________________________________________________________
      Real :: XWRK, XMIN
!
! __________________________________________________________
!
      Integer :: ICRS, IDCR, NDON
!
      NDON = Size (XDONT)
!
! We first bring the minimum to the first location in the array.
! That way, we will have a "guard", and when looking for the
! right place to insert a value, no loop test is necessary.
!
      If (XDONT (1) < XDONT (NDON)) Then
          XMIN = XDONT (1)
      Else
          XMIN = XDONT (NDON)
          XDONT (NDON) = XDONT (1)
      Endif
      Do IDCR = NDON-1, 2, -1
         XWRK = XDONT(IDCR)
         IF (XWRK < XMIN) Then
            XDONT (IDCR) = XMIN
            XMIN = XWRK
         End If
      End Do
      XDONT (1) = XMIN
!
! The first value is now the minimum
! Loop over the array, and when a value is smaller than
! the previous one, loop down to insert it at its right place.
!
      Do ICRS = 3, NDON
         XWRK = XDONT (ICRS)
         IDCR = ICRS - 1
         If (XWRK < XDONT(IDCR)) Then
            XDONT (ICRS) = XDONT (IDCR)
            IDCR = IDCR - 1
            Do
               If (XWRK >= XDONT(IDCR)) Exit
               XDONT (IDCR+1) = XDONT (IDCR)
               IDCR = IDCR - 1
            End Do
            XDONT (IDCR+1) = XWRK
         End If
      End Do
!
      Return
!
End Subroutine R_inssor
Subroutine I_inssor (XDONT)
!  Sorts XDONT into increasing order (Insertion sort)
! __________________________________________________________
!  This subroutine uses insertion sort. It does not use any
!  work array and is faster when XDONT is of very small size
!  (< 20), or already almost sorted, but worst case behavior
!  can happen fairly probably (initially inverse sorted).
!  In many cases, the quicksort or merge sort method is faster.
!  Michel Olagnon - Apr. 2000
! __________________________________________________________
! __________________________________________________________
! __________________________________________________________
      Integer, Dimension (:), Intent (InOut)  :: XDONT
! __________________________________________________________
      Integer :: XWRK, XMIN
!
! __________________________________________________________
!
      Integer :: ICRS, IDCR, NDON
!
      NDON = Size (XDONT)
!
! We first bring the minimum to the first location in the array.
! That way, we will have a "guard", and when looking for the
! right place to insert a value, no loop test is necessary.
!
      If (XDONT (1) < XDONT (NDON)) Then
          XMIN = XDONT (1)
      Else
          XMIN = XDONT (NDON)
          XDONT (NDON) = XDONT (1)
      Endif
      Do IDCR = NDON-1, 2, -1
         XWRK = XDONT(IDCR)
         IF (XWRK < XMIN) Then
            XDONT (IDCR) = XMIN
            XMIN = XWRK
         End If
      End Do
      XDONT (1) = XMIN
!
! The first value is now the minimum
! Loop over the array, and when a value is smaller than
! the previous one, loop down to insert it at its right place.
!
      Do ICRS = 3, NDON
         XWRK = XDONT (ICRS)
         IDCR = ICRS - 1
         If (XWRK < XDONT(IDCR)) Then
            XDONT (ICRS) = XDONT (IDCR)
            IDCR = IDCR - 1
            Do
               If (XWRK >= XDONT(IDCR)) Exit
               XDONT (IDCR+1) = XDONT (IDCR)
               IDCR = IDCR - 1
            End Do
            XDONT (IDCR+1) = XWRK
         End If
      End Do
!
      Return
!
End Subroutine I_inssor
end module m_inssor
Module m_median
Integer, Parameter :: kdp = selected_real_kind(15)
public :: median
private :: kdp
private :: R_median, I_median, D_median
interface median
  module procedure d_median, r_median, i_median
end interface median
contains

Function D_median (XDONT) Result (median)
!  Return median value of XDONT
!  If even number of data, average of the two "medians".
! __________________________________________________________
!  This routine uses a pivoting strategy such as the one of
!  finding the median based on the quicksort algorithm, but
!  we skew the pivot choice to try to bring it to NORD as
!  fast as possible. It uses 2 temporary arrays, where it
!  stores the indices of the values smaller than the pivot
!  (ILOWT), and the indices of values larger than the pivot
!  that we might still need later on (IHIGT). It iterates
!  until it can bring the number of values in ILOWT to
!  exactly NORD, and then finds the maximum of this set.
!  Michel Olagnon - Aug. 2000
! __________________________________________________________
! __________________________________________________________
      Real (Kind=kdp), Dimension (:), Intent (In) :: XDONT
      Real (Kind=kdp) :: median
! __________________________________________________________
      Real (Kind=kdp), Dimension (SIZE(XDONT)) :: XLOWT, XHIGT
      Real (Kind=kdp) :: XPIV, XPIV0, XWRK, XWRK1, XWRK2, XWRK3, XMIN, XMAX
!!
      Logical :: IFODD
      Integer :: NDON, JHIG, JLOW, IHIG !, NORD
      Integer :: IMIL, IFIN, ICRS, IDCR !, ILOW
      Integer :: JLM2, JLM1, JHM2, JHM1, INTH
!
      NDON = SIZE (XDONT)
      INTH = NDON/2 + 1
      IFODD = (2*INTH == NDON + 1)
!
!    First loop is used to fill-in XLOWT, XHIGT at the same time
!
      If (NDON < 3) Then
         If (NDON > 0) median = 0.5 * (XDONT (1) + XDONT (NDON))
         Return
      End If
!
!  One chooses a pivot, best estimate possible to put fractile near
!  mid-point of the set of low values.
!
      If (XDONT(2) < XDONT(1)) Then
         XLOWT (1) = XDONT(2)
         XHIGT (1) = XDONT(1)
      Else
         XLOWT (1) = XDONT(1)
         XHIGT (1) = XDONT(2)
      End If
!
!
      If (XDONT(3) < XHIGT(1)) Then
         XHIGT (2) = XHIGT (1)
         If (XDONT(3) < XLOWT(1)) Then
            XHIGT (1) = XLOWT (1)
            XLOWT (1) = XDONT(3)
         Else
            XHIGT (1) = XDONT(3)
         End If
      Else
         XHIGT (2) = XDONT(3)
      End If
!
      If (NDON < 4) Then ! 3 values
         median = XHIGT (1)
         Return
      End If
!
      If (XDONT(NDON) < XHIGT(1)) Then
         XHIGT (3) = XHIGT (2)
         XHIGT (2) = XHIGT (1)
         If (XDONT(NDON) < XLOWT(1)) Then
            XHIGT (1) = XLOWT (1)
            XLOWT (1) = XDONT(NDON)
         Else
            XHIGT (1) = XDONT(NDON)
         End If
      Else
         If (XDONT(NDON) < XHIGT(2)) Then
            XHIGT (3) = XHIGT (2)
            XHIGT (2) = XDONT(NDON)
         Else
            XHIGT (3) = XDONT(NDON)
         End If
      End If
!
      If (NDON < 5) Then ! 4 values
         median = 0.5*(XHIGT (1) + XHIGT (2))
         Return
      End If
!
      JLOW = 1
      JHIG = 3
      XPIV = XLOWT(1) + 2.0 * (XHIGT(3)-XLOWT(1)) / 3.0
      If (XPIV >= XHIGT(1)) Then
         XPIV = XLOWT(1) + 2.0 * (XHIGT(2)-XLOWT(1)) / 3.0
         If (XPIV >= XHIGT(1)) XPIV = XLOWT(1) + 2.0 * (XHIGT(1)-XLOWT(1)) / 3.0
      End If
      XPIV0 = XPIV
!
!  One puts values > pivot in the end and those <= pivot
!  at the beginning. This is split in 2 cases, so that
!  we can skip the loop test a number of times.
!  As we are also filling in the work arrays at the same time
!  we stop filling in the XHIGT array as soon as we have more
!  than enough values in XLOWT.
!
!
      If (XDONT(NDON) > XPIV) Then
         ICRS = 3
         Do
            ICRS = ICRS + 1
            If (XDONT(ICRS) > XPIV) Then
               If (ICRS >= NDON) Exit
               JHIG = JHIG + 1
               XHIGT (JHIG) = XDONT(ICRS)
            Else
               JLOW = JLOW + 1
               XLOWT (JLOW) = XDONT(ICRS)
               If (JLOW >= INTH) Exit
            End If
         End Do
!
!  One restricts further processing because it is no use
!  to store more high values
!
         If (ICRS < NDON-1) Then
            Do
               ICRS = ICRS + 1
               If (XDONT(ICRS) <= XPIV) Then
                  JLOW = JLOW + 1
                  XLOWT (JLOW) = XDONT(ICRS)
               Else If (ICRS >= NDON) Then
                  Exit
               End If
            End Do
         End If
!
!
      Else
!
!  Same as above, but this is not as easy to optimize, so the
!  DO-loop is kept
!
         Do ICRS = 4, NDON - 1
            If (XDONT(ICRS) > XPIV) Then
               JHIG = JHIG + 1
               XHIGT (JHIG) = XDONT(ICRS)
            Else
               JLOW = JLOW + 1
               XLOWT (JLOW) = XDONT(ICRS)
               If (JLOW >= INTH) Exit
            End If
         End Do
!
         If (ICRS < NDON-1) Then
            Do
               ICRS = ICRS + 1
               If (XDONT(ICRS) <= XPIV) Then
                  If (ICRS >= NDON) Exit
                  JLOW = JLOW + 1
                  XLOWT (JLOW) = XDONT(ICRS)
               End If
            End Do
         End If
      End If
!
      JLM2 = 0
      JLM1 = 0
      JHM2 = 0
      JHM1 = 0
      Do
         If (JLM2 == JLOW .And. JHM2 == JHIG) Then
!
!   We are oscillating. Perturbate by bringing JLOW closer by one
!   to INTH
! 
             If (INTH > JLOW) Then
                XMIN = XHIGT(1)
                IHIG = 1
                Do ICRS = 2, JHIG
                   If (XHIGT(ICRS) < XMIN) Then
                      XMIN = XHIGT(ICRS)
                      IHIG = ICRS
                   End If
                End Do
!
                JLOW = JLOW + 1
                XLOWT (JLOW) = XHIGT (IHIG)
                XHIGT (IHIG) = XHIGT (JHIG)
                JHIG = JHIG - 1
             Else

                XMAX = XLOWT (JLOW)
                JLOW = JLOW - 1
                Do ICRS = 1, JLOW
                   If (XLOWT(ICRS) > XMAX) Then
                      XWRK = XMAX
                      XMAX = XLOWT(ICRS)
                      XLOWT (ICRS) = XWRK
                   End If
                End Do
             End If
         End If
         JLM2 = JLM1
         JLM1 = JLOW
         JHM2 = JHM1
         JHM1 = JHIG
!
!   We try to bring the number of values in the low values set
!   closer to INTH.
!
         Select Case (INTH-JLOW)
         Case (2:)
!
!   Not enough values in low part, at least 2 are missing
!
            INTH = INTH - JLOW
            JLOW = 0
            Select Case (JHIG)
!!!!!           CASE DEFAULT
!!!!!              write (unit=*,fmt=*) "Assertion failed"
!!!!!              STOP
!
!   We make a special case when we have so few values in
!   the high values set that it is bad performance to choose a pivot
!   and apply the general algorithm.
!
            Case (2)
               If (XHIGT(1) <= XHIGT(2)) Then
                  JLOW = JLOW + 1
                  XLOWT (JLOW) = XHIGT (1)
                  JLOW = JLOW + 1
                  XLOWT (JLOW) = XHIGT (2)
               Else
                  JLOW = JLOW + 1
                  XLOWT (JLOW) = XHIGT (2)
                  JLOW = JLOW + 1
                  XLOWT (JLOW) = XHIGT (1)
               End If
               Exit
!
            Case (3)
!
!
               XWRK1 = XHIGT (1)
               XWRK2 = XHIGT (2)
               XWRK3 = XHIGT (3)
               If (XWRK2 < XWRK1) Then
                  XHIGT (1) = XWRK2
                  XHIGT (2) = XWRK1
                  XWRK2 = XWRK1
               End If
               If (XWRK2 > XWRK3) Then
                  XHIGT (3) = XWRK2
                  XHIGT (2) = XWRK3
                  XWRK2 = XWRK3
                  If (XWRK2 < XHIGT(1)) Then
                     XHIGT (2) = XHIGT (1)
                     XHIGT (1) = XWRK2
                  End If
               End If
               JHIG = 0
               Do ICRS = JLOW + 1, INTH
                  JHIG = JHIG + 1
                  XLOWT (ICRS) = XHIGT (JHIG)
               End Do
               JLOW = INTH
               Exit
!
            Case (4:)
!
!
               XPIV0 = XPIV
               IFIN = JHIG
!
!  One chooses a pivot from the 2 first values and the last one.
!  This should ensure sufficient renewal between iterations to
!  avoid worst case behavior effects.
!
               XWRK1 = XHIGT (1)
               XWRK2 = XHIGT (2)
               XWRK3 = XHIGT (IFIN)
               If (XWRK2 < XWRK1) Then
                  XHIGT (1) = XWRK2
                  XHIGT (2) = XWRK1
                  XWRK2 = XWRK1
               End If
               If (XWRK2 > XWRK3) Then
                  XHIGT (IFIN) = XWRK2
                  XHIGT (2) = XWRK3
                  XWRK2 = XWRK3
                  If (XWRK2 < XHIGT(1)) Then
                     XHIGT (2) = XHIGT (1)
                     XHIGT (1) = XWRK2
                  End If
               End If
!
               XWRK1 = XHIGT (1)
               JLOW = JLOW + 1
               XLOWT (JLOW) = XWRK1
               XPIV = XWRK1 + 0.5 * (XHIGT(IFIN)-XWRK1)
!
!  One takes values <= pivot to XLOWT
!  Again, 2 parts, one where we take care of the remaining
!  high values because we might still need them, and the
!  other when we know that we will have more than enough
!  low values in the end.
!
               JHIG = 0
               Do ICRS = 2, IFIN
                  If (XHIGT(ICRS) <= XPIV) Then
                     JLOW = JLOW + 1
                     XLOWT (JLOW) = XHIGT (ICRS)
                     If (JLOW >= INTH) Exit
                  Else
                     JHIG = JHIG + 1
                     XHIGT (JHIG) = XHIGT (ICRS)
                  End If
               End Do
!
               Do ICRS = ICRS + 1, IFIN
                  If (XHIGT(ICRS) <= XPIV) Then
                     JLOW = JLOW + 1
                     XLOWT (JLOW) = XHIGT (ICRS)
                  End If
               End Do
            End Select
!
!
         Case (1)
!
!  Only 1 value is missing in low part
!
            XMIN = XHIGT(1)
            Do ICRS = 2, JHIG
               If (XHIGT(ICRS) < XMIN) Then
                  XMIN = XHIGT(ICRS)
               End If
            End Do
!
            JLOW = JLOW + 1
            XLOWT (JLOW) = XMIN
            Exit
!
!
         Case (0)
!
!  Low part is exactly what we want
!
            Exit
!
!
         Case (-5:-1)
!
!  Only few values too many in low part
!
            IF (IFODD) THEN
              JHIG = JLOW - INTH + 1 
            Else
              JHIG = JLOW - INTH + 2
            Endif
            XHIGT (1) = XLOWT (1)
            Do ICRS = 2, JHIG
               XWRK = XLOWT (ICRS)
               Do IDCR = ICRS - 1, 1, - 1
                  If (XWRK < XHIGT(IDCR)) Then
                     XHIGT (IDCR+1) = XHIGT (IDCR)
                  Else
                     Exit
                  End If
               End Do
               XHIGT (IDCR+1) = XWRK
            End Do
!
            Do ICRS = JHIG + 1, JLOW
               If (XLOWT (ICRS) > XHIGT(1)) Then 
                  XWRK = XLOWT (ICRS)
                  Do IDCR = 2, JHIG
                     If (XWRK >= XHIGT(IDCR)) Then
                        XHIGT (IDCR-1) = XHIGT (IDCR)
                     else
                        exit
                     endif
                  End Do
                  XHIGT (IDCR-1) = XWRK
               End If
            End Do
!
            IF (IFODD) THEN
              median = XHIGT(1)
            Else
              median = 0.5*(XHIGT(1)+XHIGT(2))
            Endif
            Return
!
!
         Case (:-6)
!
! last case: too many values in low part
!

            IMIL = (JLOW+1) / 2
            IFIN = JLOW
!
!  One chooses a pivot from 1st, last, and middle values
!
            If (XLOWT(IMIL) < XLOWT(1)) Then
               XWRK = XLOWT (1)
               XLOWT (1) = XLOWT (IMIL)
               XLOWT (IMIL) = XWRK
            End If
            If (XLOWT(IMIL) > XLOWT(IFIN)) Then
               XWRK = XLOWT (IFIN)
               XLOWT (IFIN) = XLOWT (IMIL)
               XLOWT (IMIL) = XWRK
               If (XLOWT(IMIL) < XLOWT(1)) Then
                  XWRK = XLOWT (1)
                  XLOWT (1) = XLOWT (IMIL)
                  XLOWT (IMIL) = XWRK
               End If
            End If
            If (IFIN <= 3) Exit
!
            XPIV = XLOWT(1) + REAL(INTH)/REAL(JLOW+INTH) * &
                              (XLOWT(IFIN)-XLOWT(1))

!
!  One takes values > XPIV to XHIGT
!
            JHIG = 0
            JLOW = 0
!
            If (XLOWT(IFIN) > XPIV) Then
               ICRS = 0
               Do
                  ICRS = ICRS + 1
                  If (XLOWT(ICRS) > XPIV) Then
                     JHIG = JHIG + 1
                     XHIGT (JHIG) = XLOWT (ICRS)
                     If (ICRS >= IFIN) Exit
                  Else
                     JLOW = JLOW + 1
                     XLOWT (JLOW) = XLOWT (ICRS)
                     If (JLOW >= INTH) Exit
                  End If
               End Do
!
               If (ICRS < IFIN) Then
                  Do
                     ICRS = ICRS + 1
                     If (XLOWT(ICRS) <= XPIV) Then
                        JLOW = JLOW + 1
                        XLOWT (JLOW) = XLOWT (ICRS)
                     Else
                        If (ICRS >= IFIN) Exit
                     End If
                  End Do
               End If
            Else
               Do ICRS = 1, IFIN
                  If (XLOWT(ICRS) > XPIV) Then
                     JHIG = JHIG + 1
                     XHIGT (JHIG) = XLOWT (ICRS)
                  Else
                     JLOW = JLOW + 1
                     XLOWT (JLOW) = XLOWT (ICRS)
                     If (JLOW >= INTH) Exit
                  End If
               End Do
!
               Do ICRS = ICRS + 1, IFIN
                  If (XLOWT(ICRS) <= XPIV) Then
                     JLOW = JLOW + 1
                     XLOWT (JLOW) = XLOWT (ICRS)
                  End If
               End Do
            End If
!
         End Select
!
      End Do
!
!  Now, we only need to find maximum of the 1:INTH set
!
      if (IFODD) then
        median = MAXVAL (XLOWT (1:INTH))
      else
        XWRK = MAX (XLOWT (1), XLOWT (2))
        XWRK1 = MIN (XLOWT (1), XLOWT (2))
        DO ICRS = 3, INTH
          IF (XLOWT (ICRS) > XWRK1) THEN
            IF (XLOWT (ICRS) > XWRK) THEN
              XWRK1 = XWRK
              XWRK  = XLOWT (ICRS)
            Else
              XWRK1 = XLOWT (ICRS)
            ENDIF
          ENDIF
        ENDDO
        median = 0.5*(XWRK+XWRK1)
      endif
      Return
!
End Function D_median

Function R_median (XDONT) Result (median)
!  Return median value of XDONT
! __________________________________________________________
!  This routine uses a pivoting strategy such as the one of
!  finding the median based on the quicksort algorithm, but
!  we skew the pivot choice to try to bring it to NORD as
!  fast as possible. It uses 2 temporary arrays, where it
!  stores the indices of the values smaller than the pivot
!  (ILOWT), and the indices of values larger than the pivot
!  that we might still need later on (IHIGT). It iterates
!  until it can bring the number of values in ILOWT to
!  exactly NORD, and then finds the maximum of this set.
!  Michel Olagnon - Aug. 2000
! __________________________________________________________
! _________________________________________________________
      Real, Dimension (:), Intent (In) :: XDONT
      Real :: median
! __________________________________________________________
      Real, Dimension (SIZE(XDONT)) :: XLOWT, XHIGT
      Real :: XPIV, XPIV0, XWRK, XWRK1, XWRK2, XWRK3, XMIN, XMAX
!!
      Logical :: IFODD
      Integer :: NDON, JHIG, JLOW, IHIG !, NORD
      Integer :: IMIL, IFIN, ICRS, IDCR !, ILOW
      Integer :: JLM2, JLM1, JHM2, JHM1, INTH
!
      NDON = SIZE (XDONT)
      INTH = NDON/2 + 1
      IFODD = (2*INTH == NDON + 1)
!
!    First loop is used to fill-in XLOWT, XHIGT at the same time
!
      If (NDON < 3) Then
         If (NDON > 0) median = 0.5 * (XDONT (1) + XDONT (NDON))
         Return
      End If
!
!  One chooses a pivot, best estimate possible to put fractile near
!  mid-point of the set of low values.
!
      If (XDONT(2) < XDONT(1)) Then
         XLOWT (1) = XDONT(2)
         XHIGT (1) = XDONT(1)
      Else
         XLOWT (1) = XDONT(1)
         XHIGT (1) = XDONT(2)
      End If
!
!
      If (XDONT(3) < XHIGT(1)) Then
         XHIGT (2) = XHIGT (1)
         If (XDONT(3) < XLOWT(1)) Then
            XHIGT (1) = XLOWT (1)
            XLOWT (1) = XDONT(3)
         Else
            XHIGT (1) = XDONT(3)
         End If
      Else
         XHIGT (2) = XDONT(3)
      End If
!
      If (NDON < 4) Then ! 3 values
         median = XHIGT (1)
         Return
      End If
!
      If (XDONT(NDON) < XHIGT(1)) Then
         XHIGT (3) = XHIGT (2)
         XHIGT (2) = XHIGT (1)
         If (XDONT(NDON) < XLOWT(1)) Then
            XHIGT (1) = XLOWT (1)
            XLOWT (1) = XDONT(NDON)
         Else
            XHIGT (1) = XDONT(NDON)
         End If
      Else
         If (XDONT(NDON) < XHIGT(2)) Then
            XHIGT (3) = XHIGT (2)
            XHIGT (2) = XDONT(NDON)
         Else
            XHIGT (3) = XDONT(NDON)
         End If
      End If
!
      If (NDON < 5) Then ! 4 values
         median = 0.5*(XHIGT (1) + XHIGT (2))
         Return
      End If
!
      JLOW = 1
      JHIG = 3
      XPIV = XLOWT(1) + 2.0 * (XHIGT(3)-XLOWT(1)) / 3.0
      If (XPIV >= XHIGT(1)) Then
         XPIV = XLOWT(1) + 2.0 * (XHIGT(2)-XLOWT(1)) / 3.0
         If (XPIV >= XHIGT(1)) XPIV = XLOWT(1) + 2.0 * (XHIGT(1)-XLOWT(1)) / 3.0
      End If
      XPIV0 = XPIV
!
!  One puts values > pivot in the end and those <= pivot
!  at the beginning. This is split in 2 cases, so that
!  we can skip the loop test a number of times.
!  As we are also filling in the work arrays at the same time
!  we stop filling in the XHIGT array as soon as we have more
!  than enough values in XLOWT.
!
!
      If (XDONT(NDON) > XPIV) Then
         ICRS = 3
         Do
            ICRS = ICRS + 1
            If (XDONT(ICRS) > XPIV) Then
               If (ICRS >= NDON) Exit
               JHIG = JHIG + 1
               XHIGT (JHIG) = XDONT(ICRS)
            Else
               JLOW = JLOW + 1
               XLOWT (JLOW) = XDONT(ICRS)
               If (JLOW >= INTH) Exit
            End If
         End Do
!
!  One restricts further processing because it is no use
!  to store more high values
!
         If (ICRS < NDON-1) Then
            Do
               ICRS = ICRS + 1
               If (XDONT(ICRS) <= XPIV) Then
                  JLOW = JLOW + 1
                  XLOWT (JLOW) = XDONT(ICRS)
               Else If (ICRS >= NDON) Then
                  Exit
               End If
            End Do
         End If
!
!
      Else
!
!  Same as above, but this is not as easy to optimize, so the
!  DO-loop is kept
!
         Do ICRS = 4, NDON - 1
            If (XDONT(ICRS) > XPIV) Then
               JHIG = JHIG + 1
               XHIGT (JHIG) = XDONT(ICRS)
            Else
               JLOW = JLOW + 1
               XLOWT (JLOW) = XDONT(ICRS)
               If (JLOW >= INTH) Exit
            End If
         End Do
!
         If (ICRS < NDON-1) Then
            Do
               ICRS = ICRS + 1
               If (XDONT(ICRS) <= XPIV) Then
                  If (ICRS >= NDON) Exit
                  JLOW = JLOW + 1
                  XLOWT (JLOW) = XDONT(ICRS)
               End If
            End Do
         End If
      End If
!
      JLM2 = 0
      JLM1 = 0
      JHM2 = 0
      JHM1 = 0
      Do
         If (JLM2 == JLOW .And. JHM2 == JHIG) Then
!
!   We are oscillating. Perturbate by bringing JLOW closer by one
!   to INTH
! 
             If (INTH > JLOW) Then
                XMIN = XHIGT(1)
                IHIG = 1
                Do ICRS = 2, JHIG
                   If (XHIGT(ICRS) < XMIN) Then
                      XMIN = XHIGT(ICRS)
                      IHIG = ICRS
                   End If
                End Do
!
                JLOW = JLOW + 1
                XLOWT (JLOW) = XHIGT (IHIG)
                XHIGT (IHIG) = XHIGT (JHIG)
                JHIG = JHIG - 1
             Else

                XMAX = XLOWT (JLOW)
                JLOW = JLOW - 1
                Do ICRS = 1, JLOW
                   If (XLOWT(ICRS) > XMAX) Then
                      XWRK = XMAX
                      XMAX = XLOWT(ICRS)
                      XLOWT (ICRS) = XWRK
                   End If
                End Do
             End If
         End If
         JLM2 = JLM1
         JLM1 = JLOW
         JHM2 = JHM1
         JHM1 = JHIG
!
!   We try to bring the number of values in the low values set
!   closer to INTH.
!
         Select Case (INTH-JLOW)
         Case (2:)
!
!   Not enough values in low part, at least 2 are missing
!
            INTH = INTH - JLOW
            JLOW = 0
            Select Case (JHIG)
!!!!!           CASE DEFAULT
!!!!!              write (unit=*,fmt=*) "Assertion failed"
!!!!!              STOP
!
!   We make a special case when we have so few values in
!   the high values set that it is bad performance to choose a pivot
!   and apply the general algorithm.
!
            Case (2)
               If (XHIGT(1) <= XHIGT(2)) Then
                  JLOW = JLOW + 1
                  XLOWT (JLOW) = XHIGT (1)
                  JLOW = JLOW + 1
                  XLOWT (JLOW) = XHIGT (2)
               Else
                  JLOW = JLOW + 1
                  XLOWT (JLOW) = XHIGT (2)
                  JLOW = JLOW + 1
                  XLOWT (JLOW) = XHIGT (1)
               End If
               Exit
!
            Case (3)
!
!
               XWRK1 = XHIGT (1)
               XWRK2 = XHIGT (2)
               XWRK3 = XHIGT (3)
               If (XWRK2 < XWRK1) Then
                  XHIGT (1) = XWRK2
                  XHIGT (2) = XWRK1
                  XWRK2 = XWRK1
               End If
               If (XWRK2 > XWRK3) Then
                  XHIGT (3) = XWRK2
                  XHIGT (2) = XWRK3
                  XWRK2 = XWRK3
                  If (XWRK2 < XHIGT(1)) Then
                     XHIGT (2) = XHIGT (1)
                     XHIGT (1) = XWRK2
                  End If
               End If
               JHIG = 0
               Do ICRS = JLOW + 1, INTH
                  JHIG = JHIG + 1
                  XLOWT (ICRS) = XHIGT (JHIG)
               End Do
               JLOW = INTH
               Exit
!
            Case (4:)
!
!
               XPIV0 = XPIV
               IFIN = JHIG
!
!  One chooses a pivot from the 2 first values and the last one.
!  This should ensure sufficient renewal between iterations to
!  avoid worst case behavior effects.
!
               XWRK1 = XHIGT (1)
               XWRK2 = XHIGT (2)
               XWRK3 = XHIGT (IFIN)
               If (XWRK2 < XWRK1) Then
                  XHIGT (1) = XWRK2
                  XHIGT (2) = XWRK1
                  XWRK2 = XWRK1
               End If
               If (XWRK2 > XWRK3) Then
                  XHIGT (IFIN) = XWRK2
                  XHIGT (2) = XWRK3
                  XWRK2 = XWRK3
                  If (XWRK2 < XHIGT(1)) Then
                     XHIGT (2) = XHIGT (1)
                     XHIGT (1) = XWRK2
                  End If
               End If
!
               XWRK1 = XHIGT (1)
               JLOW = JLOW + 1
               XLOWT (JLOW) = XWRK1
               XPIV = XWRK1 + 0.5 * (XHIGT(IFIN)-XWRK1)
!
!  One takes values <= pivot to XLOWT
!  Again, 2 parts, one where we take care of the remaining
!  high values because we might still need them, and the
!  other when we know that we will have more than enough
!  low values in the end.
!
               JHIG = 0
               Do ICRS = 2, IFIN
                  If (XHIGT(ICRS) <= XPIV) Then
                     JLOW = JLOW + 1
                     XLOWT (JLOW) = XHIGT (ICRS)
                     If (JLOW >= INTH) Exit
                  Else
                     JHIG = JHIG + 1
                     XHIGT (JHIG) = XHIGT (ICRS)
                  End If
               End Do
!
               Do ICRS = ICRS + 1, IFIN
                  If (XHIGT(ICRS) <= XPIV) Then
                     JLOW = JLOW + 1
                     XLOWT (JLOW) = XHIGT (ICRS)
                  End If
               End Do
            End Select
!
!
         Case (1)
!
!  Only 1 value is missing in low part
!
            XMIN = XHIGT(1)
            Do ICRS = 2, JHIG
               If (XHIGT(ICRS) < XMIN) Then
                  XMIN = XHIGT(ICRS)
               End If
            End Do
!
            JLOW = JLOW + 1
            XLOWT (JLOW) = XMIN
            Exit
!
!
         Case (0)
!
!  Low part is exactly what we want
!
            Exit
!
!
         Case (-5:-1)
!
!  Only few values too many in low part
!
            IF (IFODD) THEN
              JHIG = JLOW - INTH + 1 
            Else
              JHIG = JLOW - INTH + 2
            Endif
            XHIGT (1) = XLOWT (1)
            Do ICRS = 2, JHIG
               XWRK = XLOWT (ICRS)
               Do IDCR = ICRS - 1, 1, - 1
                  If (XWRK < XHIGT(IDCR)) Then
                     XHIGT (IDCR+1) = XHIGT (IDCR)
                  Else
                     Exit
                  End If
               End Do
               XHIGT (IDCR+1) = XWRK
            End Do
!
            Do ICRS = JHIG + 1, JLOW
               If (XLOWT (ICRS) > XHIGT(1)) Then 
                  XWRK = XLOWT (ICRS)
                  Do IDCR = 2, JHIG
                     If (XWRK >= XHIGT(IDCR)) Then
                        XHIGT (IDCR-1) = XHIGT (IDCR)
                     else
                        exit
                     endif
                  End Do
                  XHIGT (IDCR-1) = XWRK
               End If
            End Do
!
            IF (IFODD) THEN
              median = XHIGT(1)
            Else
              median = 0.5*(XHIGT(1)+XHIGT(2))
            Endif
            Return
!
!
         Case (:-6)
!
! last case: too many values in low part
!

            IMIL = (JLOW+1) / 2
            IFIN = JLOW
!
!  One chooses a pivot from 1st, last, and middle values
!
            If (XLOWT(IMIL) < XLOWT(1)) Then
               XWRK = XLOWT (1)
               XLOWT (1) = XLOWT (IMIL)
               XLOWT (IMIL) = XWRK
            End If
            If (XLOWT(IMIL) > XLOWT(IFIN)) Then
               XWRK = XLOWT (IFIN)
               XLOWT (IFIN) = XLOWT (IMIL)
               XLOWT (IMIL) = XWRK
               If (XLOWT(IMIL) < XLOWT(1)) Then
                  XWRK = XLOWT (1)
                  XLOWT (1) = XLOWT (IMIL)
                  XLOWT (IMIL) = XWRK
               End If
            End If
            If (IFIN <= 3) Exit
!
            XPIV = XLOWT(1) + REAL(INTH)/REAL(JLOW+INTH) * &
                              (XLOWT(IFIN)-XLOWT(1))

!
!  One takes values > XPIV to XHIGT
!
            JHIG = 0
            JLOW = 0
!
            If (XLOWT(IFIN) > XPIV) Then
               ICRS = 0
               Do
                  ICRS = ICRS + 1
                  If (XLOWT(ICRS) > XPIV) Then
                     JHIG = JHIG + 1
                     XHIGT (JHIG) = XLOWT (ICRS)
                     If (ICRS >= IFIN) Exit
                  Else
                     JLOW = JLOW + 1
                     XLOWT (JLOW) = XLOWT (ICRS)
                     If (JLOW >= INTH) Exit
                  End If
               End Do
!
               If (ICRS < IFIN) Then
                  Do
                     ICRS = ICRS + 1
                     If (XLOWT(ICRS) <= XPIV) Then
                        JLOW = JLOW + 1
                        XLOWT (JLOW) = XLOWT (ICRS)
                     Else
                        If (ICRS >= IFIN) Exit
                     End If
                  End Do
               End If
            Else
               Do ICRS = 1, IFIN
                  If (XLOWT(ICRS) > XPIV) Then
                     JHIG = JHIG + 1
                     XHIGT (JHIG) = XLOWT (ICRS)
                  Else
                     JLOW = JLOW + 1
                     XLOWT (JLOW) = XLOWT (ICRS)
                     If (JLOW >= INTH) Exit
                  End If
               End Do
!
               Do ICRS = ICRS + 1, IFIN
                  If (XLOWT(ICRS) <= XPIV) Then
                     JLOW = JLOW + 1
                     XLOWT (JLOW) = XLOWT (ICRS)
                  End If
               End Do
            End If
!
         End Select
!
      End Do
!
!  Now, we only need to find maximum of the 1:INTH set
!
      if (IFODD) then
        median = MAXVAL (XLOWT (1:INTH))
      else
        XWRK = MAX (XLOWT (1), XLOWT (2))
        XWRK1 = MIN (XLOWT (1), XLOWT (2))
        DO ICRS = 3, INTH
          IF (XLOWT (ICRS) > XWRK1) THEN
            IF (XLOWT (ICRS) > XWRK) THEN
              XWRK1 = XWRK
              XWRK  = XLOWT (ICRS)
            Else
              XWRK1 = XLOWT (ICRS)
            ENDIF
          ENDIF
        ENDDO
        median = 0.5*(XWRK+XWRK1)
      endif
      Return
!
End Function R_median
Function I_median (XDONT) Result (median)
!  Return median value of XDONT
! __________________________________________________________
!  This routine uses a pivoting strategy such as the one of
!  finding the median based on the quicksort algorithm, but
!  we skew the pivot choice to try to bring it to NORD as
!  fast as possible. It uses 2 temporary arrays, where it
!  stores the indices of the values smaller than the pivot
!  (ILOWT), and the indices of values larger than the pivot
!  that we might still need later on (IHIGT). It iterates
!  until it can bring the number of values in ILOWT to
!  exactly NORD, and then finds the maximum of this set.
!  Michel Olagnon - Aug. 2000
! __________________________________________________________
! __________________________________________________________
      Integer, Dimension (:), Intent (In) :: XDONT
      Integer :: median
! __________________________________________________________
      Integer, Dimension (SIZE(XDONT)) :: XLOWT, XHIGT
      Integer :: XPIV, XPIV0, XWRK, XWRK1, XWRK2, XWRK3, XMIN, XMAX
!!
      Logical :: IFODD
      Integer :: NDON, JHIG, JLOW, IHIG !, NORD
      Integer :: IMIL, IFIN, ICRS, IDCR !, ILOW
      Integer :: JLM2, JLM1, JHM2, JHM1, INTH
!
      NDON = SIZE (XDONT)
      INTH = NDON/2 + 1
      IFODD = (2*INTH == NDON + 1)
!
!    First loop is used to fill-in XLOWT, XHIGT at the same time
!
      If (NDON < 3) Then
         If (NDON > 0) median = 0.5 * (XDONT (1) + XDONT (NDON))
         Return
      End If
!
!  One chooses a pivot, best estimate possible to put fractile near
!  mid-point of the set of low values.
!
      If (XDONT(2) < XDONT(1)) Then
         XLOWT (1) = XDONT(2)
         XHIGT (1) = XDONT(1)
      Else
         XLOWT (1) = XDONT(1)
         XHIGT (1) = XDONT(2)
      End If
!
!
      If (XDONT(3) < XHIGT(1)) Then
         XHIGT (2) = XHIGT (1)
         If (XDONT(3) < XLOWT(1)) Then
            XHIGT (1) = XLOWT (1)
            XLOWT (1) = XDONT(3)
         Else
            XHIGT (1) = XDONT(3)
         End If
      Else
         XHIGT (2) = XDONT(3)
      End If
!
      If (NDON < 4) Then ! 3 values
         median = XHIGT (1)
         Return
      End If
!
      If (XDONT(NDON) < XHIGT(1)) Then
         XHIGT (3) = XHIGT (2)
         XHIGT (2) = XHIGT (1)
         If (XDONT(NDON) < XLOWT(1)) Then
            XHIGT (1) = XLOWT (1)
            XLOWT (1) = XDONT(NDON)
         Else
            XHIGT (1) = XDONT(NDON)
         End If
      Else
         If (XDONT(NDON) < XHIGT(2)) Then
            XHIGT (3) = XHIGT (2)
            XHIGT (2) = XDONT(NDON)
         Else
            XHIGT (3) = XDONT(NDON)
         End If
      End If
!
      If (NDON < 5) Then ! 4 values
         median = 0.5*(XHIGT (1) + XHIGT (2))
         Return
      End If
!
      JLOW = 1
      JHIG = 3
      XPIV = XLOWT(1) + 2.0 * (XHIGT(3)-XLOWT(1)) / 3.0
      If (XPIV >= XHIGT(1)) Then
         XPIV = XLOWT(1) + 2.0 * (XHIGT(2)-XLOWT(1)) / 3.0
         If (XPIV >= XHIGT(1)) XPIV = XLOWT(1) + 2.0 * (XHIGT(1)-XLOWT(1)) / 3.0
      End If
      XPIV0 = XPIV
!
!  One puts values > pivot in the end and those <= pivot
!  at the beginning. This is split in 2 cases, so that
!  we can skip the loop test a number of times.
!  As we are also filling in the work arrays at the same time
!  we stop filling in the XHIGT array as soon as we have more
!  than enough values in XLOWT.
!
!
      If (XDONT(NDON) > XPIV) Then
         ICRS = 3
         Do
            ICRS = ICRS + 1
            If (XDONT(ICRS) > XPIV) Then
               If (ICRS >= NDON) Exit
               JHIG = JHIG + 1
               XHIGT (JHIG) = XDONT(ICRS)
            Else
               JLOW = JLOW + 1
               XLOWT (JLOW) = XDONT(ICRS)
               If (JLOW >= INTH) Exit
            End If
         End Do
!
!  One restricts further processing because it is no use
!  to store more high values
!
         If (ICRS < NDON-1) Then
            Do
               ICRS = ICRS + 1
               If (XDONT(ICRS) <= XPIV) Then
                  JLOW = JLOW + 1
                  XLOWT (JLOW) = XDONT(ICRS)
               Else If (ICRS >= NDON) Then
                  Exit
               End If
            End Do
         End If
!
!
      Else
!
!  Same as above, but this is not as easy to optimize, so the
!  DO-loop is kept
!
         Do ICRS = 4, NDON - 1
            If (XDONT(ICRS) > XPIV) Then
               JHIG = JHIG + 1
               XHIGT (JHIG) = XDONT(ICRS)
            Else
               JLOW = JLOW + 1
               XLOWT (JLOW) = XDONT(ICRS)
               If (JLOW >= INTH) Exit
            End If
         End Do
!
         If (ICRS < NDON-1) Then
            Do
               ICRS = ICRS + 1
               If (XDONT(ICRS) <= XPIV) Then
                  If (ICRS >= NDON) Exit
                  JLOW = JLOW + 1
                  XLOWT (JLOW) = XDONT(ICRS)
               End If
            End Do
         End If
      End If
!
      JLM2 = 0
      JLM1 = 0
      JHM2 = 0
      JHM1 = 0
      Do
         If (JLM2 == JLOW .And. JHM2 == JHIG) Then
!
!   We are oscillating. Perturbate by bringing JLOW closer by one
!   to INTH
! 
             If (INTH > JLOW) Then
                XMIN = XHIGT(1)
                IHIG = 1
                Do ICRS = 2, JHIG
                   If (XHIGT(ICRS) < XMIN) Then
                      XMIN = XHIGT(ICRS)
                      IHIG = ICRS
                   End If
                End Do
!
                JLOW = JLOW + 1
                XLOWT (JLOW) = XHIGT (IHIG)
                XHIGT (IHIG) = XHIGT (JHIG)
                JHIG = JHIG - 1
             Else

                XMAX = XLOWT (JLOW)
                JLOW = JLOW - 1
                Do ICRS = 1, JLOW
                   If (XLOWT(ICRS) > XMAX) Then
                      XWRK = XMAX
                      XMAX = XLOWT(ICRS)
                      XLOWT (ICRS) = XWRK
                   End If
                End Do
             End If
         End If
         JLM2 = JLM1
         JLM1 = JLOW
         JHM2 = JHM1
         JHM1 = JHIG
!
!   We try to bring the number of values in the low values set
!   closer to INTH.
!
         Select Case (INTH-JLOW)
         Case (2:)
!
!   Not enough values in low part, at least 2 are missing
!
            INTH = INTH - JLOW
            JLOW = 0
            Select Case (JHIG)
!!!!!           CASE DEFAULT
!!!!!              write (unit=*,fmt=*) "Assertion failed"
!!!!!              STOP
!
!   We make a special case when we have so few values in
!   the high values set that it is bad performance to choose a pivot
!   and apply the general algorithm.
!
            Case (2)
               If (XHIGT(1) <= XHIGT(2)) Then
                  JLOW = JLOW + 1
                  XLOWT (JLOW) = XHIGT (1)
                  JLOW = JLOW + 1
                  XLOWT (JLOW) = XHIGT (2)
               Else
                  JLOW = JLOW + 1
                  XLOWT (JLOW) = XHIGT (2)
                  JLOW = JLOW + 1
                  XLOWT (JLOW) = XHIGT (1)
               End If
               Exit
!
            Case (3)
!
!
               XWRK1 = XHIGT (1)
               XWRK2 = XHIGT (2)
               XWRK3 = XHIGT (3)
               If (XWRK2 < XWRK1) Then
                  XHIGT (1) = XWRK2
                  XHIGT (2) = XWRK1
                  XWRK2 = XWRK1
               End If
               If (XWRK2 > XWRK3) Then
                  XHIGT (3) = XWRK2
                  XHIGT (2) = XWRK3
                  XWRK2 = XWRK3
                  If (XWRK2 < XHIGT(1)) Then
                     XHIGT (2) = XHIGT (1)
                     XHIGT (1) = XWRK2
                  End If
               End If
               JHIG = 0
               Do ICRS = JLOW + 1, INTH
                  JHIG = JHIG + 1
                  XLOWT (ICRS) = XHIGT (JHIG)
               End Do
               JLOW = INTH
               Exit
!
            Case (4:)
!
!
               XPIV0 = XPIV
               IFIN = JHIG
!
!  One chooses a pivot from the 2 first values and the last one.
!  This should ensure sufficient renewal between iterations to
!  avoid worst case behavior effects.
!
               XWRK1 = XHIGT (1)
               XWRK2 = XHIGT (2)
               XWRK3 = XHIGT (IFIN)
               If (XWRK2 < XWRK1) Then
                  XHIGT (1) = XWRK2
                  XHIGT (2) = XWRK1
                  XWRK2 = XWRK1
               End If
               If (XWRK2 > XWRK3) Then
                  XHIGT (IFIN) = XWRK2
                  XHIGT (2) = XWRK3
                  XWRK2 = XWRK3
                  If (XWRK2 < XHIGT(1)) Then
                     XHIGT (2) = XHIGT (1)
                     XHIGT (1) = XWRK2
                  End If
               End If
!
               XWRK1 = XHIGT (1)
               JLOW = JLOW + 1
               XLOWT (JLOW) = XWRK1
               XPIV = XWRK1 + 0.5 * (XHIGT(IFIN)-XWRK1)
!
!  One takes values <= pivot to XLOWT
!  Again, 2 parts, one where we take care of the remaining
!  high values because we might still need them, and the
!  other when we know that we will have more than enough
!  low values in the end.
!
               JHIG = 0
               Do ICRS = 2, IFIN
                  If (XHIGT(ICRS) <= XPIV) Then
                     JLOW = JLOW + 1
                     XLOWT (JLOW) = XHIGT (ICRS)
                     If (JLOW >= INTH) Exit
                  Else
                     JHIG = JHIG + 1
                     XHIGT (JHIG) = XHIGT (ICRS)
                  End If
               End Do
!
               Do ICRS = ICRS + 1, IFIN
                  If (XHIGT(ICRS) <= XPIV) Then
                     JLOW = JLOW + 1
                     XLOWT (JLOW) = XHIGT (ICRS)
                  End If
               End Do
            End Select
!
!
         Case (1)
!
!  Only 1 value is missing in low part
!
            XMIN = XHIGT(1)
            Do ICRS = 2, JHIG
               If (XHIGT(ICRS) < XMIN) Then
                  XMIN = XHIGT(ICRS)
               End If
            End Do
!
            JLOW = JLOW + 1
            XLOWT (JLOW) = XMIN
            Exit
!
!
         Case (0)
!
!  Low part is exactly what we want
!
            Exit
!
!
         Case (-5:-1)
!
!  Only few values too many in low part
!
            IF (IFODD) THEN
              JHIG = JLOW - INTH + 1 
            Else
              JHIG = JLOW - INTH + 2
            Endif
            XHIGT (1) = XLOWT (1)
            Do ICRS = 2, JHIG
               XWRK = XLOWT (ICRS)
               Do IDCR = ICRS - 1, 1, - 1
                  If (XWRK < XHIGT(IDCR)) Then
                     XHIGT (IDCR+1) = XHIGT (IDCR)
                  Else
                     Exit
                  End If
               End Do
               XHIGT (IDCR+1) = XWRK
            End Do
!
            Do ICRS = JHIG + 1, JLOW
               If (XLOWT (ICRS) > XHIGT(1)) Then 
                  XWRK = XLOWT (ICRS)
                  Do IDCR = 2, JHIG
                     If (XWRK >= XHIGT(IDCR)) Then
                        XHIGT (IDCR-1) = XHIGT (IDCR)
                     else
                        exit
                     endif
                  End Do
                  XHIGT (IDCR-1) = XWRK
               End If
            End Do
!
            IF (IFODD) THEN
              median = XHIGT(1)
            Else
              median = 0.5*(XHIGT(1)+XHIGT(2))
            Endif
            Return
!
!
         Case (:-6)
!
! last case: too many values in low part
!

            IMIL = (JLOW+1) / 2
            IFIN = JLOW
!
!  One chooses a pivot from 1st, last, and middle values
!
            If (XLOWT(IMIL) < XLOWT(1)) Then
               XWRK = XLOWT (1)
               XLOWT (1) = XLOWT (IMIL)
               XLOWT (IMIL) = XWRK
            End If
            If (XLOWT(IMIL) > XLOWT(IFIN)) Then
               XWRK = XLOWT (IFIN)
               XLOWT (IFIN) = XLOWT (IMIL)
               XLOWT (IMIL) = XWRK
               If (XLOWT(IMIL) < XLOWT(1)) Then
                  XWRK = XLOWT (1)
                  XLOWT (1) = XLOWT (IMIL)
                  XLOWT (IMIL) = XWRK
               End If
            End If
            If (IFIN <= 3) Exit
!
            XPIV = XLOWT(1) + REAL(INTH)/REAL(JLOW+INTH) * &
                              (XLOWT(IFIN)-XLOWT(1))

!
!  One takes values > XPIV to XHIGT
!
            JHIG = 0
            JLOW = 0
!
            If (XLOWT(IFIN) > XPIV) Then
               ICRS = 0
               Do
                  ICRS = ICRS + 1
                  If (XLOWT(ICRS) > XPIV) Then
                     JHIG = JHIG + 1
                     XHIGT (JHIG) = XLOWT (ICRS)
                     If (ICRS >= IFIN) Exit
                  Else
                     JLOW = JLOW + 1
                     XLOWT (JLOW) = XLOWT (ICRS)
                     If (JLOW >= INTH) Exit
                  End If
               End Do
!
               If (ICRS < IFIN) Then
                  Do
                     ICRS = ICRS + 1
                     If (XLOWT(ICRS) <= XPIV) Then
                        JLOW = JLOW + 1
                        XLOWT (JLOW) = XLOWT (ICRS)
                     Else
                        If (ICRS >= IFIN) Exit
                     End If
                  End Do
               End If
            Else
               Do ICRS = 1, IFIN
                  If (XLOWT(ICRS) > XPIV) Then
                     JHIG = JHIG + 1
                     XHIGT (JHIG) = XLOWT (ICRS)
                  Else
                     JLOW = JLOW + 1
                     XLOWT (JLOW) = XLOWT (ICRS)
                     If (JLOW >= INTH) Exit
                  End If
               End Do
!
               Do ICRS = ICRS + 1, IFIN
                  If (XLOWT(ICRS) <= XPIV) Then
                     JLOW = JLOW + 1
                     XLOWT (JLOW) = XLOWT (ICRS)
                  End If
               End Do
            End If
!
         End Select
!
      End Do
!
!  Now, we only need to find maximum of the 1:INTH set
!
      if (IFODD) then
        median = MAXVAL (XLOWT (1:INTH))
      else
        XWRK = MAX (XLOWT (1), XLOWT (2))
        XWRK1 = MIN (XLOWT (1), XLOWT (2))
        DO ICRS = 3, INTH
          IF (XLOWT (ICRS) > XWRK1) THEN
            IF (XLOWT (ICRS) > XWRK) THEN
              XWRK1 = XWRK
              XWRK  = XLOWT (ICRS)
            Else
              XWRK1 = XLOWT (ICRS)
            ENDIF
          ENDIF
        ENDDO
        median = 0.5*(XWRK+XWRK1)
      endif
      Return
!
End Function I_median
end module m_median

Module m_mrgref
Integer, Parameter :: kdp = selected_real_kind(15)
public :: mrgref
private :: kdp
private :: R_mrgref, I_mrgref, D_mrgref
interface mrgref
  module procedure d_mrgref, r_mrgref, i_mrgref
end interface mrgref
contains

Subroutine D_mrgref (XVALT, IRNGT)
!   Ranks array XVALT into index array IRNGT, using merge-sort
! __________________________________________________________
!   This version is not optimized for performance, and is thus
!   not as difficult to read as some other ones.
!   Michel Olagnon - April 2000
! __________________________________________________________
! __________________________________________________________
      Real (kind=kdp), Dimension (:), Intent (In) :: XVALT
      Integer, Dimension (:), Intent (Out) :: IRNGT
! __________________________________________________________
!
      Integer, Dimension (:), Allocatable :: JWRKT
      Integer :: LMTNA, LMTNC
      Integer :: NVAL, IIND, IWRKD, IWRK, IWRKF, JINDA, IINDA, IINDB
!
      NVAL = Min (SIZE(XVALT), SIZE(IRNGT))
      If (NVAL <= 0) Then
         Return
      End If
!
!  Fill-in the index array, creating ordered couples
!
      Do IIND = 2, NVAL, 2
         If (XVALT(IIND-1) <= XVALT(IIND)) Then
            IRNGT (IIND-1) = IIND - 1
            IRNGT (IIND) = IIND
         Else
            IRNGT (IIND-1) = IIND
            IRNGT (IIND) = IIND - 1
         End If
      End Do
      If (Modulo (NVAL, 2) /= 0) Then
         IRNGT (NVAL) = NVAL
      End If
!
!  We will now have ordered subsets A - B - A - B - ...
!  and merge A and B couples into     C   -   C   - ...
!
      Allocate (JWRKT(1:NVAL))
      LMTNC = 2
      LMTNA = 2
!
!  Iteration. Each time, the length of the ordered subsets
!  is doubled.
!
      Do
         If (LMTNA >= NVAL) Exit
         IWRKF = 0
         LMTNC = 2 * LMTNC
         IWRK = 0
!
!   Loop on merges of A and B into C
!
         Do
            IINDA = IWRKF
            IWRKD = IWRKF + 1
            IWRKF = IINDA + LMTNC
            JINDA = IINDA + LMTNA
            If (IWRKF >= NVAL) Then
               If (JINDA >= NVAL) Exit
               IWRKF = NVAL
            End If
            IINDB = JINDA
!
!   Shortcut for the case when the max of A is smaller
!   than the min of B (no need to do anything)
!
            If (XVALT(IRNGT(JINDA)) <= XVALT(IRNGT(JINDA+1))) Then
               IWRK = IWRKF
               Cycle
            End If
!
!  One steps in the C subset, that we create in the final rank array
!
            Do
               If (IWRK >= IWRKF) Then
!
!  Make a copy of the rank array for next iteration
!
                  IRNGT (IWRKD:IWRKF) = JWRKT (IWRKD:IWRKF)
                  Exit
               End If
!
               IWRK = IWRK + 1
!
!  We still have unprocessed values in both A and B
!
               If (IINDA < JINDA) Then
                  If (IINDB < IWRKF) Then
                     If (XVALT(IRNGT(IINDA+1)) > XVALT(IRNGT(IINDB+1))) &
                    & Then
                        IINDB = IINDB + 1
                        JWRKT (IWRK) = IRNGT (IINDB)
                     Else
                        IINDA = IINDA + 1
                        JWRKT (IWRK) = IRNGT (IINDA)
                     End If
                  Else
!
!  Only A still with unprocessed values
!
                     IINDA = IINDA + 1
                     JWRKT (IWRK) = IRNGT (IINDA)
                  End If
               Else
!
!  Only B still with unprocessed values
!
                  IRNGT (IWRKD:IINDB) = JWRKT (IWRKD:IINDB)
                  IWRK = IWRKF
                  Exit
               End If
!
            End Do
         End Do
!
!  The Cs become As and Bs
!
         LMTNA = 2 * LMTNA
      End Do
!
!  Clean up
!
      Deallocate (JWRKT)
      Return
!
End Subroutine D_mrgref

Subroutine R_mrgref (XVALT, IRNGT)
!   Ranks array XVALT into index array IRNGT, using merge-sort
! __________________________________________________________
!   This version is not optimized for performance, and is thus
!   not as difficult to read as some other ones.
!   Michel Olagnon - April 2000
! __________________________________________________________
! _________________________________________________________
      Real, Dimension (:), Intent (In) :: XVALT
      Integer, Dimension (:), Intent (Out) :: IRNGT
! __________________________________________________________
!
      Integer, Dimension (:), Allocatable :: JWRKT
      Integer :: LMTNA, LMTNC
      Integer :: NVAL, IIND, IWRKD, IWRK, IWRKF, JINDA, IINDA, IINDB
!
      NVAL = Min (SIZE(XVALT), SIZE(IRNGT))
      If (NVAL <= 0) Then
         Return
      End If
!
!  Fill-in the index array, creating ordered couples
!
      Do IIND = 2, NVAL, 2
         If (XVALT(IIND-1) <= XVALT(IIND)) Then
            IRNGT (IIND-1) = IIND - 1
            IRNGT (IIND) = IIND
         Else
            IRNGT (IIND-1) = IIND
            IRNGT (IIND) = IIND - 1
         End If
      End Do
      If (Modulo (NVAL, 2) /= 0) Then
         IRNGT (NVAL) = NVAL
      End If
!
!  We will now have ordered subsets A - B - A - B - ...
!  and merge A and B couples into     C   -   C   - ...
!
      Allocate (JWRKT(1:NVAL))
      LMTNC = 2
      LMTNA = 2
!
!  Iteration. Each time, the length of the ordered subsets
!  is doubled.
!
      Do
         If (LMTNA >= NVAL) Exit
         IWRKF = 0
         LMTNC = 2 * LMTNC
         IWRK = 0
!
!   Loop on merges of A and B into C
!
         Do
            IINDA = IWRKF
            IWRKD = IWRKF + 1
            IWRKF = IINDA + LMTNC
            JINDA = IINDA + LMTNA
            If (IWRKF >= NVAL) Then
               If (JINDA >= NVAL) Exit
               IWRKF = NVAL
            End If
            IINDB = JINDA
!
!   Shortcut for the case when the max of A is smaller
!   than the min of B (no need to do anything)
!
            If (XVALT(IRNGT(JINDA)) <= XVALT(IRNGT(JINDA+1))) Then
               IWRK = IWRKF
               Cycle
            End If
!
!  One steps in the C subset, that we create in the final rank array
!
            Do
               If (IWRK >= IWRKF) Then
!
!  Make a copy of the rank array for next iteration
!
                  IRNGT (IWRKD:IWRKF) = JWRKT (IWRKD:IWRKF)
                  Exit
               End If
!
               IWRK = IWRK + 1
!
!  We still have unprocessed values in both A and B
!
               If (IINDA < JINDA) Then
                  If (IINDB < IWRKF) Then
                     If (XVALT(IRNGT(IINDA+1)) > XVALT(IRNGT(IINDB+1))) &
                    & Then
                        IINDB = IINDB + 1
                        JWRKT (IWRK) = IRNGT (IINDB)
                     Else
                        IINDA = IINDA + 1
                        JWRKT (IWRK) = IRNGT (IINDA)
                     End If
                  Else
!
!  Only A still with unprocessed values
!
                     IINDA = IINDA + 1
                     JWRKT (IWRK) = IRNGT (IINDA)
                  End If
               Else
!
!  Only B still with unprocessed values
!
                  IRNGT (IWRKD:IINDB) = JWRKT (IWRKD:IINDB)
                  IWRK = IWRKF
                  Exit
               End If
!
            End Do
         End Do
!
!  The Cs become As and Bs
!
         LMTNA = 2 * LMTNA
      End Do
!
!  Clean up
!
      Deallocate (JWRKT)
      Return
!
End Subroutine R_mrgref
Subroutine I_mrgref (XVALT, IRNGT)
!   Ranks array XVALT into index array IRNGT, using merge-sort
! __________________________________________________________
!   This version is not optimized for performance, and is thus
!   not as difficult to read as some other ones.
!   Michel Olagnon - April 2000
! __________________________________________________________
! __________________________________________________________
      Integer, Dimension (:), Intent (In)  :: XVALT
      Integer, Dimension (:), Intent (Out) :: IRNGT
! __________________________________________________________
!
      Integer, Dimension (:), Allocatable :: JWRKT
      Integer :: LMTNA, LMTNC
      Integer :: NVAL, IIND, IWRKD, IWRK, IWRKF, JINDA, IINDA, IINDB
!
      NVAL = Min (SIZE(XVALT), SIZE(IRNGT))
      If (NVAL <= 0) Then
         Return
      End If
!
!  Fill-in the index array, creating ordered couples
!
      Do IIND = 2, NVAL, 2
         If (XVALT(IIND-1) <= XVALT(IIND)) Then
            IRNGT (IIND-1) = IIND - 1
            IRNGT (IIND) = IIND
         Else
            IRNGT (IIND-1) = IIND
            IRNGT (IIND) = IIND - 1
         End If
      End Do
      If (Modulo (NVAL, 2) /= 0) Then
         IRNGT (NVAL) = NVAL
      End If
!
!  We will now have ordered subsets A - B - A - B - ...
!  and merge A and B couples into     C   -   C   - ...
!
      Allocate (JWRKT(1:NVAL))
      LMTNC = 2
      LMTNA = 2
!
!  Iteration. Each time, the length of the ordered subsets
!  is doubled.
!
      Do
         If (LMTNA >= NVAL) Exit
         IWRKF = 0
         LMTNC = 2 * LMTNC
         IWRK = 0
!
!   Loop on merges of A and B into C
!
         Do
            IINDA = IWRKF
            IWRKD = IWRKF + 1
            IWRKF = IINDA + LMTNC
            JINDA = IINDA + LMTNA
            If (IWRKF >= NVAL) Then
               If (JINDA >= NVAL) Exit
               IWRKF = NVAL
            End If
            IINDB = JINDA
!
!   Shortcut for the case when the max of A is smaller
!   than the min of B (no need to do anything)
!
            If (XVALT(IRNGT(JINDA)) <= XVALT(IRNGT(JINDA+1))) Then
               IWRK = IWRKF
               Cycle
            End If
!
!  One steps in the C subset, that we create in the final rank array
!
            Do
               If (IWRK >= IWRKF) Then
!
!  Make a copy of the rank array for next iteration
!
                  IRNGT (IWRKD:IWRKF) = JWRKT (IWRKD:IWRKF)
                  Exit
               End If
!
               IWRK = IWRK + 1
!
!  We still have unprocessed values in both A and B
!
               If (IINDA < JINDA) Then
                  If (IINDB < IWRKF) Then
                     If (XVALT(IRNGT(IINDA+1)) > XVALT(IRNGT(IINDB+1))) &
                    & Then
                        IINDB = IINDB + 1
                        JWRKT (IWRK) = IRNGT (IINDB)
                     Else
                        IINDA = IINDA + 1
                        JWRKT (IWRK) = IRNGT (IINDA)
                     End If
                  Else
!
!  Only A still with unprocessed values
!
                     IINDA = IINDA + 1
                     JWRKT (IWRK) = IRNGT (IINDA)
                  End If
               Else
!
!  Only B still with unprocessed values
!
                  IRNGT (IWRKD:IINDB) = JWRKT (IWRKD:IINDB)
                  IWRK = IWRKF
                  Exit
               End If
!
            End Do
         End Do
!
!  The Cs become As and Bs
!
         LMTNA = 2 * LMTNA
      End Do
!
!  Clean up
!
      Deallocate (JWRKT)
      Return
!
End Subroutine I_mrgref
end module m_mrgref
Module m_uniinv
Integer, Parameter :: kdp = selected_real_kind(15)
public :: uniinv
private :: kdp
private :: R_uniinv, I_uniinv, D_uniinv
private :: R_nearless, I_nearless, D_nearless, nearless
interface uniinv
  module procedure d_uniinv, r_uniinv, i_uniinv
end interface uniinv
interface nearless
  module procedure D_nearless, R_nearless, I_nearless
end interface nearless
contains

Subroutine D_uniinv (XDONT, IGOEST)
! __________________________________________________________
!   UNIINV = Merge-sort inverse ranking of an array, with removal of
!   duplicate entries.
!   The routine is similar to pure merge-sort ranking, but on
!   the last pass, it sets indices in IGOEST to the rank
!   of the value in the ordered set with duplicates removed.
!   For performance reasons, the first 2 passes are taken
!   out of the standard loop, and use dedicated coding.
! __________________________________________________________
! __________________________________________________________
      Real (kind=kdp), Dimension (:), Intent (In) :: XDONT
      Integer, Dimension (:), Intent (Out) :: IGOEST
! __________________________________________________________
      Real (kind=kdp) :: XTST, XDONA, XDONB
!
! __________________________________________________________
      Integer, Dimension (SIZE(IGOEST)) :: JWRKT, IRNGT
      Integer :: LMTNA, LMTNC, IRNG, IRNG1, IRNG2, NUNI
      Integer :: NVAL, IIND, IWRKD, IWRK, IWRKF, JINDA, IINDA, IINDB
!
      NVAL = Min (SIZE(XDONT), SIZE(IGOEST))
!
      Select Case (NVAL)
      Case (:0)
         Return
      Case (1)
         IGOEST (1) = 1
         Return
      Case Default
         Continue
      End Select
!
!  Fill-in the index array, creating ordered couples
!
      Do IIND = 2, NVAL, 2
         If (XDONT(IIND-1) < XDONT(IIND)) Then
            IRNGT (IIND-1) = IIND - 1
            IRNGT (IIND) = IIND
         Else
            IRNGT (IIND-1) = IIND
            IRNGT (IIND) = IIND - 1
         End If
      End Do
      If (Modulo (NVAL, 2) /= 0) Then
         IRNGT (NVAL) = NVAL
      End If
!
!  We will now have ordered subsets A - B - A - B - ...
!  and merge A and B couples into     C   -   C   - ...
!
      LMTNA = 2
      LMTNC = 4
!
!  First iteration. The length of the ordered subsets goes from 2 to 4
!
      Do
         If (NVAL <= 4) Exit
!
!   Loop on merges of A and B into C
!
         Do IWRKD = 0, NVAL - 1, 4
            If ((IWRKD+4) > NVAL) Then
               If ((IWRKD+2) >= NVAL) Exit
!
!   1 2 3
!
               If (XDONT(IRNGT(IWRKD+2)) <= XDONT(IRNGT(IWRKD+3))) Exit
!
!   1 3 2
!
               If (XDONT(IRNGT(IWRKD+1)) <= XDONT(IRNGT(IWRKD+3))) Then
                  IRNG2 = IRNGT (IWRKD+2)
                  IRNGT (IWRKD+2) = IRNGT (IWRKD+3)
                  IRNGT (IWRKD+3) = IRNG2
!
!   3 1 2
!
               Else
                  IRNG1 = IRNGT (IWRKD+1)
                  IRNGT (IWRKD+1) = IRNGT (IWRKD+3)
                  IRNGT (IWRKD+3) = IRNGT (IWRKD+2)
                  IRNGT (IWRKD+2) = IRNG1
               End If
               Exit
            End If
!
!   1 2 3 4
!
            If (XDONT(IRNGT(IWRKD+2)) <= XDONT(IRNGT(IWRKD+3))) Cycle
!
!   1 3 x x
!
            If (XDONT(IRNGT(IWRKD+1)) <= XDONT(IRNGT(IWRKD+3))) Then
               IRNG2 = IRNGT (IWRKD+2)
               IRNGT (IWRKD+2) = IRNGT (IWRKD+3)
               If (XDONT(IRNG2) <= XDONT(IRNGT(IWRKD+4))) Then
!   1 3 2 4
                  IRNGT (IWRKD+3) = IRNG2
               Else
!   1 3 4 2
                  IRNGT (IWRKD+3) = IRNGT (IWRKD+4)
                  IRNGT (IWRKD+4) = IRNG2
               End If
!
!   3 x x x
!
            Else
               IRNG1 = IRNGT (IWRKD+1)
               IRNG2 = IRNGT (IWRKD+2)
               IRNGT (IWRKD+1) = IRNGT (IWRKD+3)
               If (XDONT(IRNG1) <= XDONT(IRNGT(IWRKD+4))) Then
                  IRNGT (IWRKD+2) = IRNG1
                  If (XDONT(IRNG2) <= XDONT(IRNGT(IWRKD+4))) Then
!   3 1 2 4
                     IRNGT (IWRKD+3) = IRNG2
                  Else
!   3 1 4 2
                     IRNGT (IWRKD+3) = IRNGT (IWRKD+4)
                     IRNGT (IWRKD+4) = IRNG2
                  End If
               Else
!   3 4 1 2
                  IRNGT (IWRKD+2) = IRNGT (IWRKD+4)
                  IRNGT (IWRKD+3) = IRNG1
                  IRNGT (IWRKD+4) = IRNG2
               End If
            End If
         End Do
!
!  The Cs become As and Bs
!
         LMTNA = 4
         Exit
      End Do
!
!  Iteration loop. Each time, the length of the ordered subsets
!  is doubled.
!
      Do
         If (2*LMTNA >= NVAL) Exit
         IWRKF = 0
         LMTNC = 2 * LMTNC
!
!   Loop on merges of A and B into C
!
         Do
            IWRK = IWRKF
            IWRKD = IWRKF + 1
            JINDA = IWRKF + LMTNA
            IWRKF = IWRKF + LMTNC
            If (IWRKF >= NVAL) Then
               If (JINDA >= NVAL) Exit
               IWRKF = NVAL
            End If
            IINDA = 1
            IINDB = JINDA + 1
!
!  One steps in the C subset, that we create in the final rank array
!
!  Make a copy of the rank array for the iteration
!
            JWRKT (1:LMTNA) = IRNGT (IWRKD:JINDA)
            XDONA = XDONT (JWRKT(IINDA))
            XDONB = XDONT (IRNGT(IINDB))
!
            Do
               IWRK = IWRK + 1
!
!  We still have unprocessed values in both A and B
!
               If (XDONA > XDONB) Then
                  IRNGT (IWRK) = IRNGT (IINDB)
                  IINDB = IINDB + 1
                  If (IINDB > IWRKF) Then
!  Only A still with unprocessed values
                     IRNGT (IWRK+1:IWRKF) = JWRKT (IINDA:LMTNA)
                     Exit
                  End If
                  XDONB = XDONT (IRNGT(IINDB))
               Else
                  IRNGT (IWRK) = JWRKT (IINDA)
                  IINDA = IINDA + 1
                  If (IINDA > LMTNA) Exit! Only B still with unprocessed values
                  XDONA = XDONT (JWRKT(IINDA))
               End If
!
            End Do
         End Do
!
!  The Cs become As and Bs
!
         LMTNA = 2 * LMTNA
      End Do
!
!   Last merge of A and B into C, with removal of duplicates.
!
      IINDA = 1
      IINDB = LMTNA + 1
      NUNI = 0
!
!  One steps in the C subset, that we create in the final rank array
!
      JWRKT (1:LMTNA) = IRNGT (1:LMTNA)
      If (IINDB <= NVAL) Then
        XTST = NEARLESS (Min(XDONT(JWRKT(1)), XDONT(IRNGT(IINDB))))
      Else
        XTST = NEARLESS (XDONT(JWRKT(1)))
      Endif
      Do IWRK = 1, NVAL
!
!  We still have unprocessed values in both A and B
!
         If (IINDA <= LMTNA) Then
            If (IINDB <= NVAL) Then
               If (XDONT(JWRKT(IINDA)) > XDONT(IRNGT(IINDB))) Then
                  IRNG = IRNGT (IINDB)
                  IINDB = IINDB + 1
               Else
                  IRNG = JWRKT (IINDA)
                  IINDA = IINDA + 1
               End If
            Else
!
!  Only A still with unprocessed values
!
               IRNG = JWRKT (IINDA)
               IINDA = IINDA + 1
            End If
         Else
!
!  Only B still with unprocessed values
!
            IRNG = IRNGT (IWRK)
         End If
         If (XDONT(IRNG) > XTST) Then
            XTST = XDONT (IRNG)
            NUNI = NUNI + 1
         End If
         IGOEST (IRNG) = NUNI
!
      End Do
!
      Return
!
End Subroutine D_uniinv

Subroutine R_uniinv (XDONT, IGOEST)
! __________________________________________________________
!   UNIINV = Merge-sort inverse ranking of an array, with removal of
!   duplicate entries.
!   The routine is similar to pure merge-sort ranking, but on
!   the last pass, it sets indices in IGOEST to the rank
!   of the value in the ordered set with duplicates removed.
!   For performance reasons, the first 2 passes are taken
!   out of the standard loop, and use dedicated coding.
! __________________________________________________________
! _________________________________________________________
      Real, Dimension (:), Intent (In) :: XDONT
      Integer, Dimension (:), Intent (Out) :: IGOEST
! __________________________________________________________
      Real    :: XTST, XDONA, XDONB
!
! __________________________________________________________
      Integer, Dimension (SIZE(IGOEST)) :: JWRKT, IRNGT
      Integer :: LMTNA, LMTNC, IRNG, IRNG1, IRNG2, NUNI
      Integer :: NVAL, IIND, IWRKD, IWRK, IWRKF, JINDA, IINDA, IINDB
!
      NVAL = Min (SIZE(XDONT), SIZE(IGOEST))
!
      Select Case (NVAL)
      Case (:0)
         Return
      Case (1)
         IGOEST (1) = 1
         Return
      Case Default
         Continue
      End Select
!
!  Fill-in the index array, creating ordered couples
!
      Do IIND = 2, NVAL, 2
         If (XDONT(IIND-1) < XDONT(IIND)) Then
            IRNGT (IIND-1) = IIND - 1
            IRNGT (IIND) = IIND
         Else
            IRNGT (IIND-1) = IIND
            IRNGT (IIND) = IIND - 1
         End If
      End Do
      If (Modulo (NVAL, 2) /= 0) Then
         IRNGT (NVAL) = NVAL
      End If
!
!  We will now have ordered subsets A - B - A - B - ...
!  and merge A and B couples into     C   -   C   - ...
!
      LMTNA = 2
      LMTNC = 4
!
!  First iteration. The length of the ordered subsets goes from 2 to 4
!
      Do
         If (NVAL <= 4) Exit
!
!   Loop on merges of A and B into C
!
         Do IWRKD = 0, NVAL - 1, 4
            If ((IWRKD+4) > NVAL) Then
               If ((IWRKD+2) >= NVAL) Exit
!
!   1 2 3
!
               If (XDONT(IRNGT(IWRKD+2)) <= XDONT(IRNGT(IWRKD+3))) Exit
!
!   1 3 2
!
               If (XDONT(IRNGT(IWRKD+1)) <= XDONT(IRNGT(IWRKD+3))) Then
                  IRNG2 = IRNGT (IWRKD+2)
                  IRNGT (IWRKD+2) = IRNGT (IWRKD+3)
                  IRNGT (IWRKD+3) = IRNG2
!
!   3 1 2
!
               Else
                  IRNG1 = IRNGT (IWRKD+1)
                  IRNGT (IWRKD+1) = IRNGT (IWRKD+3)
                  IRNGT (IWRKD+3) = IRNGT (IWRKD+2)
                  IRNGT (IWRKD+2) = IRNG1
               End If
               Exit
            End If
!
!   1 2 3 4
!
            If (XDONT(IRNGT(IWRKD+2)) <= XDONT(IRNGT(IWRKD+3))) Cycle
!
!   1 3 x x
!
            If (XDONT(IRNGT(IWRKD+1)) <= XDONT(IRNGT(IWRKD+3))) Then
               IRNG2 = IRNGT (IWRKD+2)
               IRNGT (IWRKD+2) = IRNGT (IWRKD+3)
               If (XDONT(IRNG2) <= XDONT(IRNGT(IWRKD+4))) Then
!   1 3 2 4
                  IRNGT (IWRKD+3) = IRNG2
               Else
!   1 3 4 2
                  IRNGT (IWRKD+3) = IRNGT (IWRKD+4)
                  IRNGT (IWRKD+4) = IRNG2
               End If
!
!   3 x x x
!
            Else
               IRNG1 = IRNGT (IWRKD+1)
               IRNG2 = IRNGT (IWRKD+2)
               IRNGT (IWRKD+1) = IRNGT (IWRKD+3)
               If (XDONT(IRNG1) <= XDONT(IRNGT(IWRKD+4))) Then
                  IRNGT (IWRKD+2) = IRNG1
                  If (XDONT(IRNG2) <= XDONT(IRNGT(IWRKD+4))) Then
!   3 1 2 4
                     IRNGT (IWRKD+3) = IRNG2
                  Else
!   3 1 4 2
                     IRNGT (IWRKD+3) = IRNGT (IWRKD+4)
                     IRNGT (IWRKD+4) = IRNG2
                  End If
               Else
!   3 4 1 2
                  IRNGT (IWRKD+2) = IRNGT (IWRKD+4)
                  IRNGT (IWRKD+3) = IRNG1
                  IRNGT (IWRKD+4) = IRNG2
               End If
            End If
         End Do
!
!  The Cs become As and Bs
!
         LMTNA = 4
         Exit
      End Do
!
!  Iteration loop. Each time, the length of the ordered subsets
!  is doubled.
!
      Do
         If (2*LMTNA >= NVAL) Exit
         IWRKF = 0
         LMTNC = 2 * LMTNC
!
!   Loop on merges of A and B into C
!
         Do
            IWRK = IWRKF
            IWRKD = IWRKF + 1
            JINDA = IWRKF + LMTNA
            IWRKF = IWRKF + LMTNC
            If (IWRKF >= NVAL) Then
               If (JINDA >= NVAL) Exit
               IWRKF = NVAL
            End If
            IINDA = 1
            IINDB = JINDA + 1
!
!  One steps in the C subset, that we create in the final rank array
!
!  Make a copy of the rank array for the iteration
!
            JWRKT (1:LMTNA) = IRNGT (IWRKD:JINDA)
            XDONA = XDONT (JWRKT(IINDA))
            XDONB = XDONT (IRNGT(IINDB))
!
            Do
               IWRK = IWRK + 1
!
!  We still have unprocessed values in both A and B
!
               If (XDONA > XDONB) Then
                  IRNGT (IWRK) = IRNGT (IINDB)
                  IINDB = IINDB + 1
                  If (IINDB > IWRKF) Then
!  Only A still with unprocessed values
                     IRNGT (IWRK+1:IWRKF) = JWRKT (IINDA:LMTNA)
                     Exit
                  End If
                  XDONB = XDONT (IRNGT(IINDB))
               Else
                  IRNGT (IWRK) = JWRKT (IINDA)
                  IINDA = IINDA + 1
                  If (IINDA > LMTNA) Exit! Only B still with unprocessed values
                  XDONA = XDONT (JWRKT(IINDA))
               End If
!
            End Do
         End Do
!
!  The Cs become As and Bs
!
         LMTNA = 2 * LMTNA
      End Do
!
!   Last merge of A and B into C, with removal of duplicates.
!
      IINDA = 1
      IINDB = LMTNA + 1
      NUNI = 0
!
!  One steps in the C subset, that we create in the final rank array
!
      JWRKT (1:LMTNA) = IRNGT (1:LMTNA)
      If (IINDB <= NVAL) Then
        XTST = NEARLESS (Min(XDONT(JWRKT(1)), XDONT(IRNGT(IINDB))))
      Else
        XTST = NEARLESS (XDONT(JWRKT(1)))
      Endif
      Do IWRK = 1, NVAL
!
!  We still have unprocessed values in both A and B
!
         If (IINDA <= LMTNA) Then
            If (IINDB <= NVAL) Then
               If (XDONT(JWRKT(IINDA)) > XDONT(IRNGT(IINDB))) Then
                  IRNG = IRNGT (IINDB)
                  IINDB = IINDB + 1
               Else
                  IRNG = JWRKT (IINDA)
                  IINDA = IINDA + 1
               End If
            Else
!
!  Only A still with unprocessed values
!
               IRNG = JWRKT (IINDA)
               IINDA = IINDA + 1
            End If
         Else
!
!  Only B still with unprocessed values
!
            IRNG = IRNGT (IWRK)
         End If
         If (XDONT(IRNG) > XTST) Then
            XTST = XDONT (IRNG)
            NUNI = NUNI + 1
         End If
         IGOEST (IRNG) = NUNI
!
      End Do
!
      Return
!
End Subroutine R_uniinv
Subroutine I_uniinv (XDONT, IGOEST)
! __________________________________________________________
!   UNIINV = Merge-sort inverse ranking of an array, with removal of
!   duplicate entries.
!   The routine is similar to pure merge-sort ranking, but on
!   the last pass, it sets indices in IGOEST to the rank
!   of the value in the ordered set with duplicates removed.
!   For performance reasons, the first 2 passes are taken
!   out of the standard loop, and use dedicated coding.
! __________________________________________________________
! __________________________________________________________
      Integer, Dimension (:), Intent (In)  :: XDONT
      Integer, Dimension (:), Intent (Out) :: IGOEST
! __________________________________________________________
      Integer :: XTST, XDONA, XDONB
!
! __________________________________________________________
      Integer, Dimension (SIZE(IGOEST)) :: JWRKT, IRNGT
      Integer :: LMTNA, LMTNC, IRNG, IRNG1, IRNG2, NUNI
      Integer :: NVAL, IIND, IWRKD, IWRK, IWRKF, JINDA, IINDA, IINDB
!
      NVAL = Min (SIZE(XDONT), SIZE(IGOEST))
!
      Select Case (NVAL)
      Case (:0)
         Return
      Case (1)
         IGOEST (1) = 1
         Return
      Case Default
         Continue
      End Select
!
!  Fill-in the index array, creating ordered couples
!
      Do IIND = 2, NVAL, 2
         If (XDONT(IIND-1) < XDONT(IIND)) Then
            IRNGT (IIND-1) = IIND - 1
            IRNGT (IIND) = IIND
         Else
            IRNGT (IIND-1) = IIND
            IRNGT (IIND) = IIND - 1
         End If
      End Do
      If (Modulo (NVAL, 2) /= 0) Then
         IRNGT (NVAL) = NVAL
      End If
!
!  We will now have ordered subsets A - B - A - B - ...
!  and merge A and B couples into     C   -   C   - ...
!
      LMTNA = 2
      LMTNC = 4
!
!  First iteration. The length of the ordered subsets goes from 2 to 4
!
      Do
         If (NVAL <= 4) Exit
!
!   Loop on merges of A and B into C
!
         Do IWRKD = 0, NVAL - 1, 4
            If ((IWRKD+4) > NVAL) Then
               If ((IWRKD+2) >= NVAL) Exit
!
!   1 2 3
!
               If (XDONT(IRNGT(IWRKD+2)) <= XDONT(IRNGT(IWRKD+3))) Exit
!
!   1 3 2
!
               If (XDONT(IRNGT(IWRKD+1)) <= XDONT(IRNGT(IWRKD+3))) Then
                  IRNG2 = IRNGT (IWRKD+2)
                  IRNGT (IWRKD+2) = IRNGT (IWRKD+3)
                  IRNGT (IWRKD+3) = IRNG2
!
!   3 1 2
!
               Else
                  IRNG1 = IRNGT (IWRKD+1)
                  IRNGT (IWRKD+1) = IRNGT (IWRKD+3)
                  IRNGT (IWRKD+3) = IRNGT (IWRKD+2)
                  IRNGT (IWRKD+2) = IRNG1
               End If
               Exit
            End If
!
!   1 2 3 4
!
            If (XDONT(IRNGT(IWRKD+2)) <= XDONT(IRNGT(IWRKD+3))) Cycle
!
!   1 3 x x
!
            If (XDONT(IRNGT(IWRKD+1)) <= XDONT(IRNGT(IWRKD+3))) Then
               IRNG2 = IRNGT (IWRKD+2)
               IRNGT (IWRKD+2) = IRNGT (IWRKD+3)
               If (XDONT(IRNG2) <= XDONT(IRNGT(IWRKD+4))) Then
!   1 3 2 4
                  IRNGT (IWRKD+3) = IRNG2
               Else
!   1 3 4 2
                  IRNGT (IWRKD+3) = IRNGT (IWRKD+4)
                  IRNGT (IWRKD+4) = IRNG2
               End If
!
!   3 x x x
!
            Else
               IRNG1 = IRNGT (IWRKD+1)
               IRNG2 = IRNGT (IWRKD+2)
               IRNGT (IWRKD+1) = IRNGT (IWRKD+3)
               If (XDONT(IRNG1) <= XDONT(IRNGT(IWRKD+4))) Then
                  IRNGT (IWRKD+2) = IRNG1
                  If (XDONT(IRNG2) <= XDONT(IRNGT(IWRKD+4))) Then
!   3 1 2 4
                     IRNGT (IWRKD+3) = IRNG2
                  Else
!   3 1 4 2
                     IRNGT (IWRKD+3) = IRNGT (IWRKD+4)
                     IRNGT (IWRKD+4) = IRNG2
                  End If
               Else
!   3 4 1 2
                  IRNGT (IWRKD+2) = IRNGT (IWRKD+4)
                  IRNGT (IWRKD+3) = IRNG1
                  IRNGT (IWRKD+4) = IRNG2
               End If
            End If
         End Do
!
!  The Cs become As and Bs
!
         LMTNA = 4
         Exit
      End Do
!
!  Iteration loop. Each time, the length of the ordered subsets
!  is doubled.
!
      Do
         If (2*LMTNA >= NVAL) Exit
         IWRKF = 0
         LMTNC = 2 * LMTNC
!
!   Loop on merges of A and B into C
!
         Do
            IWRK = IWRKF
            IWRKD = IWRKF + 1
            JINDA = IWRKF + LMTNA
            IWRKF = IWRKF + LMTNC
            If (IWRKF >= NVAL) Then
               If (JINDA >= NVAL) Exit
               IWRKF = NVAL
            End If
            IINDA = 1
            IINDB = JINDA + 1
!
!  One steps in the C subset, that we create in the final rank array
!
!  Make a copy of the rank array for the iteration
!
            JWRKT (1:LMTNA) = IRNGT (IWRKD:JINDA)
            XDONA = XDONT (JWRKT(IINDA))
            XDONB = XDONT (IRNGT(IINDB))
!
            Do
               IWRK = IWRK + 1
!
!  We still have unprocessed values in both A and B
!
               If (XDONA > XDONB) Then
                  IRNGT (IWRK) = IRNGT (IINDB)
                  IINDB = IINDB + 1
                  If (IINDB > IWRKF) Then
!  Only A still with unprocessed values
                     IRNGT (IWRK+1:IWRKF) = JWRKT (IINDA:LMTNA)
                     Exit
                  End If
                  XDONB = XDONT (IRNGT(IINDB))
               Else
                  IRNGT (IWRK) = JWRKT (IINDA)
                  IINDA = IINDA + 1
                  If (IINDA > LMTNA) Exit! Only B still with unprocessed values
                  XDONA = XDONT (JWRKT(IINDA))
               End If
!
            End Do
         End Do
!
!  The Cs become As and Bs
!
         LMTNA = 2 * LMTNA
      End Do
!
!   Last merge of A and B into C, with removal of duplicates.
!
      IINDA = 1
      IINDB = LMTNA + 1
      NUNI = 0
!
!  One steps in the C subset, that we create in the final rank array
!
      JWRKT (1:LMTNA) = IRNGT (1:LMTNA)
      If (IINDB <= NVAL) Then
        XTST = NEARLESS (Min(XDONT(JWRKT(1)), XDONT(IRNGT(IINDB))))
      Else
        XTST = NEARLESS (XDONT(JWRKT(1)))
      Endif
      Do IWRK = 1, NVAL
!
!  We still have unprocessed values in both A and B
!
         If (IINDA <= LMTNA) Then
            If (IINDB <= NVAL) Then
               If (XDONT(JWRKT(IINDA)) > XDONT(IRNGT(IINDB))) Then
                  IRNG = IRNGT (IINDB)
                  IINDB = IINDB + 1
               Else
                  IRNG = JWRKT (IINDA)
                  IINDA = IINDA + 1
               End If
            Else
!
!  Only A still with unprocessed values
!
               IRNG = JWRKT (IINDA)
               IINDA = IINDA + 1
            End If
         Else
!
!  Only B still with unprocessed values
!
            IRNG = IRNGT (IWRK)
         End If
         If (XDONT(IRNG) > XTST) Then
            XTST = XDONT (IRNG)
            NUNI = NUNI + 1
         End If
         IGOEST (IRNG) = NUNI
!
      End Do
!
      Return
!
End Subroutine I_uniinv

Function D_nearless (XVAL) result (D_nl)
!  Nearest value less than given value
! __________________________________________________________
      Real (kind=kdp), Intent (In) :: XVAL
      Real (kind=kdp) :: D_nl
! __________________________________________________________
      D_nl = nearest (XVAL, -1.0_kdp)
      return
!
End Function D_nearless
Function R_nearless (XVAL) result (R_nl)
!  Nearest value less than given value
! __________________________________________________________
      Real, Intent (In) :: XVAL
      Real :: R_nl
! __________________________________________________________
      R_nl = nearest (XVAL, -1.0)
      return
!
End Function R_nearless
Function I_nearless (XVAL) result (I_nl)
!  Nearest value less than given value
! __________________________________________________________
      Integer, Intent (In) :: XVAL
      Integer :: I_nl
! __________________________________________________________
      I_nl = XVAL - 1
      return
!
End Function I_nearless

end module m_uniinv

Module m_mulcnt
Use m_uniinv
Private
Integer, Parameter :: kdp = selected_real_kind(15)
public :: mulcnt
private :: kdp
private :: R_mulcnt, I_mulcnt, D_mulcnt
interface mulcnt
  module procedure d_mulcnt, r_mulcnt, i_mulcnt
end interface mulcnt
contains

Subroutine D_mulcnt (XDONT, IMULT)
!   MULCNT = Give for each array value its multiplicity
!            (number of times that it appears in the array)
! __________________________________________________________
!  Michel Olagnon - Mar. 2000
! __________________________________________________________
! __________________________________________________________
      Real (kind=kdp), Dimension (:), Intent (In) :: XDONT
      Integer, Dimension (:), Intent (Out) :: IMULT
! __________________________________________________________
!
      Integer, Dimension (Size(XDONT)) :: IWRKT
      Integer, Dimension (Size(XDONT)) :: ICNTT
      Integer :: ICRS
! __________________________________________________________
      Call UNIINV (XDONT, IWRKT)
      ICNTT = 0
      Do ICRS = 1, Size(XDONT)
            ICNTT(IWRKT(ICRS)) = ICNTT(IWRKT(ICRS)) + 1
      End Do
      Do ICRS = 1, Size(XDONT)
            IMULT(ICRS) = ICNTT(IWRKT(ICRS))
      End Do

!
End Subroutine D_mulcnt

Subroutine R_mulcnt (XDONT, IMULT)
!   MULCNT = Give for each array value its multiplicity
!            (number of times that it appears in the array)
! __________________________________________________________
!  Michel Olagnon - Mar. 2000
! __________________________________________________________
! _________________________________________________________
      Real, Dimension (:), Intent (In) :: XDONT
      Integer, Dimension (:), Intent (Out) :: IMULT
! __________________________________________________________
!
      Integer, Dimension (Size(XDONT)) :: IWRKT
      Integer, Dimension (Size(XDONT)) :: ICNTT
      Integer :: ICRS
! __________________________________________________________
      Call UNIINV (XDONT, IWRKT)
      ICNTT = 0
      Do ICRS = 1, Size(XDONT)
            ICNTT(IWRKT(ICRS)) = ICNTT(IWRKT(ICRS)) + 1
      End Do
      Do ICRS = 1, Size(XDONT)
            IMULT(ICRS) = ICNTT(IWRKT(ICRS))
      End Do

!
End Subroutine R_mulcnt
Subroutine I_mulcnt (XDONT, IMULT)
!   MULCNT = Give for each array value its multiplicity
!            (number of times that it appears in the array)
! __________________________________________________________
!  Michel Olagnon - Mar. 2000
! __________________________________________________________
! __________________________________________________________
      Integer, Dimension (:), Intent (In)  :: XDONT
      Integer, Dimension (:), Intent (Out) :: IMULT
! __________________________________________________________
!
      Integer, Dimension (Size(XDONT)) :: IWRKT
      Integer, Dimension (Size(XDONT)) :: ICNTT
      Integer :: ICRS
! __________________________________________________________
      Call UNIINV (XDONT, IWRKT)
      ICNTT = 0
      Do ICRS = 1, Size(XDONT)
            ICNTT(IWRKT(ICRS)) = ICNTT(IWRKT(ICRS)) + 1
      End Do
      Do ICRS = 1, Size(XDONT)
            IMULT(ICRS) = ICNTT(IWRKT(ICRS))
      End Do

!
End Subroutine I_mulcnt
end module m_mulcnt
Module m_rapknr
Integer, Parameter :: kdp = selected_real_kind(15)
public :: rapknr
private :: kdp
private :: R_rapknr, I_rapknr, D_rapknr
interface rapknr
  module procedure d_rapknr, r_rapknr, i_rapknr
end interface rapknr
contains

Subroutine D_rapknr (XDONT, IRNGT, NORD)
!  Ranks partially XDONT by IRNGT, up to order NORD, in decreasing order.
!  rapknr = (rnkpar backwards)
! __________________________________________________________
!  This routine uses a pivoting strategy such as the one of
!  finding the median based on the quicksort algorithm, but
!  we skew the pivot choice to try to bring it to NORD as
!  fast as possible. It uses 2 temporary arrays, where it
!  stores the indices of the values larger than the pivot
!  (IHIGT), and the indices of values smaller than the pivot
!  that we might still need later on (ILOWT). It iterates
!  until it can bring the number of values in IHIGT to
!  exactly NORD, and then uses an insertion sort to rank
!  this set, since it is supposedly small.
!  Michel Olagnon - Feb. 2011
! __________________________________________________________
! __________________________________________________________
      Real (kind=kdp), Dimension (:), Intent (In) :: XDONT
      Integer, Dimension (:), Intent (Out) :: IRNGT
      Integer, Intent (In) :: NORD
! __________________________________________________________
      Real (kind=kdp) :: XPIV, XPIV0, XWRK, XWRK1, XMIN, XMAX
!
      Integer, Dimension (SIZE(XDONT)) :: ILOWT, IHIGT
      Integer :: NDON, JHIG, JLOW, IHIG, IWRK, IWRK1, IWRK2, IWRK3
      Integer :: IDEB, JDEB, IMIL, IFIN, NWRK, ICRS, IDCR, ILOW
      Integer :: JLM2, JLM1, JHM2, JHM1
!
      NDON = SIZE (XDONT)
!
!    First loop is used to fill-in ILOWT, IHIGT at the same time
!
      If (NDON < 2) Then
         If (NORD >= 1) IRNGT (1) = 1
         Return
      End If
!
!  One chooses a pivot, best estimate possible to put fractile near
!  mid-point of the set of high values.
!
      If (XDONT(2) < XDONT(1)) Then
         ILOWT (1) = 2
         IHIGT (1) = 1
      Else
         ILOWT (1) = 1
         IHIGT (1) = 2
      End If
!
      If (NDON < 3) Then
         If (NORD >= 1) IRNGT (1) = IHIGT (1)
         If (NORD >= 2) IRNGT (2) = ILOWT (1)
         Return
      End If
! ---
      If (XDONT(3) > XDONT(ILOWT(1))) Then
         ILOWT (2) = ILOWT (1)
         If (XDONT(3) > XDONT(IHIGT(1))) Then
            ILOWT (1) = IHIGT (1)
            IHIGT (1) = 3
         Else
            ILOWT (1) = 3
         End If
      Else
         ILOWT (2) = 3
      End If
! ---
      If (NDON < 4) Then
         If (NORD >= 1) IRNGT (1) = IHIGT (1)
         If (NORD >= 2) IRNGT (2) = ILOWT (1)
         If (NORD >= 3) IRNGT (3) = ILOWT (2)
         Return
      End If
!
      If (XDONT(NDON) > XDONT(ILOWT(1))) Then
         ILOWT (3) = ILOWT (2)
         ILOWT (2) = ILOWT (1)
         If (XDONT(NDON) > XDONT(IHIGT(1))) Then
            ILOWT (1) = IHIGT (1)
            IHIGT (1) = NDON
         Else
            ILOWT (1) = NDON
         End If
      Else
         if (XDONT (NDON) > XDONT (ILOWT(2))) Then
            ILOWT (3) = ILOWT (2)
            ILOWT (2) = NDON
         else
            ILOWT (3) = NDON
         endif 
      End If
!
      If (NDON < 5) Then
         If (NORD >= 1) IRNGT (1) = IHIGT (1)
         If (NORD >= 2) IRNGT (2) = ILOWT (1)
         If (NORD >= 3) IRNGT (3) = ILOWT (2)
         If (NORD >= 4) IRNGT (4) = ILOWT (3)
         Return
      End If
! ---
      JDEB = 0
      IDEB = JDEB + 1
      JHIG = IDEB
      JLOW = 3
      XPIV = XDONT (IHIGT(IDEB)) + REAL(2*NORD)/REAL(NDON+NORD) * &
                                   (XDONT(ILOWT(3))-XDONT(IHIGT(IDEB)))
      If (XPIV >= XDONT(ILOWT(1))) Then
         XPIV = XDONT (IHIGT(IDEB)) + REAL(2*NORD)/REAL(NDON+NORD) * &
                                      (XDONT(ILOWT(2))-XDONT(IHIGT(IDEB)))
         If (XPIV >= XDONT(ILOWT(1))) &
             XPIV = XDONT (IHIGT(IDEB)) + REAL (2*NORD) / REAL (NDON+NORD) * &
                                          (XDONT(ILOWT(1))-XDONT(IHIGT(IDEB)))
      End If
      XPIV0 = XPIV
! ---
!  One puts values < pivot in the end and those >= pivot
!  at the beginning. This is split in 2 cases, so that
!  we can skip the loop test a number of times.
!  As we are also filling in the work arrays at the same time
!  we stop filling in the ILOWT array as soon as we have more
!  than enough values in IHIGT.
!
!
      If (XDONT(NDON) < XPIV) Then
         ICRS = 3
         Do
            ICRS = ICRS + 1
            If (XDONT(ICRS) < XPIV) Then
               If (ICRS >= NDON) Exit
               JLOW = JLOW + 1
               ILOWT (JLOW) = ICRS
            Else
               JHIG = JHIG + 1
               IHIGT (JHIG) = ICRS
               If (JHIG >= NORD) Exit
            End If
         End Do
!
!  One restricts further processing because it is no use
!  to store more low values
!
         If (ICRS < NDON-1) Then
            Do
               ICRS = ICRS + 1
               If (XDONT(ICRS) >= XPIV) Then
                  JHIG = JHIG + 1
                  IHIGT (JHIG) = ICRS
               Else If (ICRS >= NDON) Then
                  Exit
               End If
            End Do
         End If
!
! ---
      Else
!
!  Same as above, but this is not as easy to optimize, so the
!  DO-loop is kept
!
         Do ICRS = 4, NDON - 1
            If (XDONT(ICRS) < XPIV) Then
               JLOW = JLOW + 1
               ILOWT (JLOW) = ICRS
            Else
               JHIG = JHIG + 1
               IHIGT (JHIG) = ICRS
               If (JHIG >= NORD) Exit
            End If
         End Do
!
         If (ICRS < NDON-1) Then
            Do
               ICRS = ICRS + 1
               If (XDONT(ICRS) >= XPIV) Then
                  If (ICRS >= NDON) Exit
                  JHIG = JHIG + 1
                  IHIGT (JHIG) = ICRS
               End If
            End Do
         End If
      End If
! ---
      JLM2 = 0
      JLM1 = 0
      JHM2 = 0
      JHM1 = 0
      Do
         if (JHIG == NORD) Exit
         If (JHM2 == JHIG .And. JLM2 == JLOW) Then
!
!   We are oscillating. Perturbate by bringing JHIG closer by one
!   to NORD
!
           If (NORD > JHIG) Then
                XMAX = XDONT (ILOWT(1))
                ILOW = 1
                Do ICRS = 2, JLOW
                   If (XDONT(ILOWT(ICRS)) > XMAX) Then
                      XMAX = XDONT (ILOWT(ICRS))
                      ILOW = ICRS
                   End If
                End Do
!
                JHIG = JHIG + 1
                IHIGT (JHIG) = ILOWT (ILOW)
                ILOWT (ILOW) = ILOWT (JLOW)
                JLOW = JLOW - 1
             Else
                IHIG = IHIGT (JHIG)
                XMIN = XDONT (IHIG)
                Do ICRS = 1, JHIG
                   If (XDONT(IHIGT(ICRS)) < XMIN) Then
                      IWRK = IHIGT (ICRS)
                      XMIN = XDONT (IWRK)
                      IHIGT (ICRS) = IHIG
                      IHIG = IWRK
                   End If
                End Do
                JHIG = JHIG - 1
             End If
         End If
         JLM2 = JLM1
         JLM1 = JLOW
         JHM2 = JHM1
         JHM1 = JHIG
! ---
!   We try to bring the number of values in the high values set
!   closer to NORD.
!
        Select Case (NORD-JHIG)
         Case (2:)
!
!   Not enough values in low part, at least 2 are missing
!
            Select Case (JLOW)
!!!!!           CASE DEFAULT
!!!!!              write (*,*) "Assertion failed"
!!!!!              STOP
!
!   We make a special case when we have so few values in
!   the low values set that it is bad performance to choose a pivot
!   and apply the general algorithm.
!
            Case (2)
               If (XDONT(ILOWT(1)) >= XDONT(ILOWT(2))) Then
                  JHIG = JHIG + 1
                  IHIGT (JHIG) = ILOWT (1)
                  JHIG = JHIG + 1
                  IHIGT (JHIG) = ILOWT (2)
               Else
                  JHIG = JHIG + 1
                  IHIGT (JHIG) = ILOWT (2)
                  JHIG = JHIG + 1
                  IHIGT (JHIG) = ILOWT (1)
               End If
               Exit
! ---
            Case (3)
!
!
               IWRK1 = ILOWT (1)
               IWRK2 = ILOWT (2)
               IWRK3 = ILOWT (3)
               If (XDONT(IWRK2) > XDONT(IWRK1)) Then
                  ILOWT (1) = IWRK2
                  ILOWT (2) = IWRK1
                  IWRK2 = IWRK1
               End If
               If (XDONT(IWRK2) < XDONT(IWRK3)) Then
                  ILOWT (3) = IWRK2
                  ILOWT (2) = IWRK3
                  IWRK2 = IWRK3
                  If (XDONT(IWRK2) > XDONT(ILOWT(1))) Then
                     ILOWT (2) = ILOWT (1)
                     ILOWT (1) = IWRK2
                  End If
               End If
               JLOW = 0
               Do ICRS = JHIG + 1, NORD
                  JLOW = JLOW + 1
                  IHIGT (ICRS) = ILOWT (JLOW)
               End Do
               JHIG = NORD
               Exit
! ---
            Case (4:)
!
!
               XPIV0 = XPIV
               IFIN = JLOW
!
!  One chooses a pivot from the 2 first values and the last one.
!  This should ensure sufficient renewal between iterations to
!  avoid worst case behavior effects.
!
               IWRK1 = ILOWT (1)
               IWRK2 = ILOWT (2)
               IWRK3 = ILOWT (IFIN)
               If (XDONT(IWRK2) > XDONT(IWRK1)) Then
                  ILOWT (1) = IWRK2
                  ILOWT (2) = IWRK1
                  IWRK2 = IWRK1
               End If
               If (XDONT(IWRK2) < XDONT(IWRK3)) Then
                  ILOWT (IFIN) = IWRK2
                  ILOWT (2) = IWRK3
                  IWRK2 = IWRK3
                  If (XDONT(IWRK2) > XDONT(IHIGT(1))) Then
                     ILOWT (2) = ILOWT (1)
                     ILOWT (1) = IWRK2
                  End If
               End If
!
               JDEB = JHIG
               NWRK = NORD - JHIG
               IWRK1 = ILOWT (1)
               JHIG = JHIG + 1
               IHIGT (JHIG) = IWRK1
               XPIV = XDONT (IWRK1) + REAL (NWRK) / REAL (NORD+NWRK) * &
                                      (XDONT(ILOWT(IFIN))-XDONT(IWRK1))
!
!  One takes values >= pivot to IHIGT
!  Again, 2 parts, one where we take care of the remaining
!  low values because we might still need them, and the
!  other when we know that we will have more than enough
!  high values in the end.
! ---
               JLOW = 0
               Do ICRS = 2, IFIN
                  If (XDONT(ILOWT(ICRS)) >= XPIV) Then
                     JHIG = JHIG + 1
                     IHIGT (JHIG) = ILOWT (ICRS)
                     If (JHIG >= NORD) Exit
                  Else
                     JLOW = JLOW + 1
                     ILOWT (JLOW) = ILOWT (ICRS)
                  End If
               End Do
!
               Do ICRS = ICRS + 1, IFIN
                  If (XDONT(ILOWT(ICRS)) >= XPIV) Then
                     JHIG = JHIG + 1
                     IHIGT (JHIG) = ILOWT (ICRS)
                  End If
               End Do
           End Select
! ---
!
         Case (1)
!
!  Only 1 value is missing in high part
!
            XMAX = XDONT (ILOWT(1))
            ILOW = 1
            Do ICRS = 2, JLOW
               If (XDONT(ILOWT(ICRS)) > XMAX) Then
                  XMAX = XDONT (ILOWT(ICRS))
                  ILOW = ICRS
               End If
            End Do
!
            JHIG = JHIG + 1
            IHIGT (JHIG) = ILOWT (ILOW)
            Exit
!
!
         Case (0)
!
!  Low part is exactly what we want
!
            Exit
! ---
!
         Case (-5:-1)
!
!  Only few values too many in high part
!
            IRNGT (1) = IHIGT (1)
            Do ICRS = 2, NORD
               IWRK = IHIGT (ICRS)
               XWRK = XDONT (IWRK)
               Do IDCR = ICRS - 1, 1, - 1
                  If (XWRK > XDONT(IRNGT(IDCR))) Then
                     IRNGT (IDCR+1) = IRNGT (IDCR)
                  Else
                     Exit
                  End If
               End Do
               IRNGT (IDCR+1) = IWRK
            End Do
!
            XWRK1 = XDONT (IRNGT(NORD))
            Do ICRS = NORD + 1, JHIG
               If (XDONT(IHIGT (ICRS)) > XWRK1) Then
                  XWRK = XDONT (IHIGT (ICRS))
                  Do IDCR = NORD - 1, 1, - 1
                     If (XWRK <= XDONT(IRNGT(IDCR))) Exit
                     IRNGT (IDCR+1) = IRNGT (IDCR)
                  End Do
                  IRNGT (IDCR+1) = IHIGT (ICRS)
                  XWRK1 = XDONT (IRNGT(NORD))
               End If
            End Do
!
            Return
!
!
         Case (:-6)
!
! last case: too many values in high part
! ---
            IDEB = JDEB + 1
            IMIL = (JHIG+IDEB) / 2
            IFIN = JHIG
! ---
!  One chooses a pivot from 1st, last, and middle values
!
            If (XDONT(IHIGT(IMIL)) > XDONT(IHIGT(IDEB))) Then
               IWRK = IHIGT (IDEB)
               IHIGT (IDEB) = IHIGT (IMIL)
               IHIGT (IMIL) = IWRK
            End If
            If (XDONT(IHIGT(IMIL)) < XDONT(IHIGT(IFIN))) Then
               IWRK = IHIGT (IFIN)
               IHIGT (IFIN) = IHIGT (IMIL)
               IHIGT (IMIL) = IWRK
               If (XDONT(IHIGT(IMIL)) > XDONT(IHIGT(IDEB))) Then
                  IWRK = IHIGT (IDEB)
                  IHIGT (IDEB) = IHIGT (IMIL)
                  IHIGT (IMIL) = IWRK
               End If
            End If
            If (IFIN <= 3) Exit
! ---
            XPIV = XDONT (IHIGT(1)) + REAL(NORD)/REAL(JHIG+NORD) * &
                                      (XDONT(IHIGT(IFIN))-XDONT(IHIGT(1)))
            If (JDEB > 0) Then
               If (XPIV <= XPIV0) &
                   XPIV = XPIV0 + REAL(2*NORD-JDEB)/REAL (JHIG+NORD) * &
                                  (XDONT(IHIGT(IFIN))-XPIV0)
            Else
               IDEB = 1
            End If
!
!  One takes values < XPIV to ILOWT
!  However, we do not process the first values if we have been
!  through the case when we did not have enough high values
! ---
            JLOW = 0
            JHIG = JDEB
! ---
            If (XDONT(IHIGT(IFIN)) < XPIV) Then
               ICRS = JDEB
               Do
                 ICRS = ICRS + 1
                  If (XDONT(IHIGT(ICRS)) < XPIV) Then
                     JLOW = JLOW + 1
                     ILOWT (JLOW) = IHIGT (ICRS)
                     If (ICRS >= IFIN) Exit
                  Else
                     JHIG = JHIG + 1
                     IHIGT (JHIG) = IHIGT (ICRS)
                     If (JHIG >= NORD) Exit
                  End If
               End Do
! ---
               If (ICRS < IFIN) Then
                  Do
                     ICRS = ICRS + 1
                     If (XDONT(IHIGT(ICRS)) >= XPIV) Then
                        JHIG = JHIG + 1
                        IHIGT (JHIG) = IHIGT (ICRS)
                     Else
                        If (ICRS >= IFIN) Exit
                     End If
                  End Do
               End If
           Else
               Do ICRS = IDEB, IFIN
                  If (XDONT(IHIGT(ICRS)) < XPIV) Then
                     JLOW = JLOW + 1
                     ILOWT (JLOW) = IHIGT (ICRS)
                  Else
                     JHIG = JHIG + 1
                     IHIGT (JHIG) = IHIGT (ICRS)
                     If (JHIG >= NORD) Exit
                  End If
               End Do
!
               Do ICRS = ICRS + 1, IFIN
                  If (XDONT(IHIGT(ICRS)) >= XPIV) Then
                     JHIG = JHIG + 1
                     IHIGT (JHIG) = IHIGT (ICRS)
                  End If
               End Do
            End If
!
         End Select
!
      End Do
! ---
!  Now, we only need to complete ranking of the 1:NORD set
!  Assuming NORD is small, we use a simple insertion sort
!
      IRNGT (1) = IHIGT (1)
      Do ICRS = 2, NORD
         IWRK = IHIGT (ICRS)
         XWRK = XDONT (IWRK)
         Do IDCR = ICRS - 1, 1, - 1
            If (XWRK > XDONT(IRNGT(IDCR))) Then
               IRNGT (IDCR+1) = IRNGT (IDCR)
            Else
               Exit
            End If
         End Do
         IRNGT (IDCR+1) = IWRK
      End Do
     Return
!
!
End Subroutine D_rapknr

Subroutine R_rapknr (XDONT, IRNGT, NORD)
!  Ranks partially XDONT by IRNGT, up to order NORD, in decreasing order.
!  rapknr = (rnkpar backwards)
! __________________________________________________________
!  This routine uses a pivoting strategy such as the one of
!  finding the median based on the quicksort algorithm, but
!  we skew the pivot choice to try to bring it to NORD as
!  fast as possible. It uses 2 temporary arrays, where it
!  stores the indices of the values larger than the pivot
!  (IHIGT), and the indices of values smaller than the pivot
!  that we might still need later on (ILOWT). It iterates
!  until it can bring the number of values in IHIGT to
!  exactly NORD, and then uses an insertion sort to rank
!  this set, since it is supposedly small.
!  Michel Olagnon - Feb. 2011
! __________________________________________________________
! __________________________________________________________
! _________________________________________________________
      Real, Dimension (:), Intent (In) :: XDONT
      Integer, Dimension (:), Intent (Out) :: IRNGT
      Integer, Intent (In) :: NORD
! __________________________________________________________
      Real    :: XPIV, XPIV0, XWRK, XWRK1, XMIN, XMAX
!
      Integer, Dimension (SIZE(XDONT)) :: ILOWT, IHIGT
      Integer :: NDON, JHIG, JLOW, IHIG, IWRK, IWRK1, IWRK2, IWRK3
      Integer :: IDEB, JDEB, IMIL, IFIN, NWRK, ICRS, IDCR, ILOW
      Integer :: JLM2, JLM1, JHM2, JHM1
!
      NDON = SIZE (XDONT)
!
!    First loop is used to fill-in ILOWT, IHIGT at the same time
!
      If (NDON < 2) Then
         If (NORD >= 1) IRNGT (1) = 1
         Return
      End If
!
!  One chooses a pivot, best estimate possible to put fractile near
!  mid-point of the set of high values.
!
      If (XDONT(2) < XDONT(1)) Then
         ILOWT (1) = 2
         IHIGT (1) = 1
      Else
         ILOWT (1) = 1
         IHIGT (1) = 2
      End If
!
      If (NDON < 3) Then
         If (NORD >= 1) IRNGT (1) = IHIGT (1)
         If (NORD >= 2) IRNGT (2) = ILOWT (1)
         Return
      End If
! ---
      If (XDONT(3) > XDONT(ILOWT(1))) Then
         ILOWT (2) = ILOWT (1)
         If (XDONT(3) > XDONT(IHIGT(1))) Then
            ILOWT (1) = IHIGT (1)
            IHIGT (1) = 3
         Else
            ILOWT (1) = 3
         End If
      Else
         ILOWT (2) = 3
      End If
! ---
      If (NDON < 4) Then
         If (NORD >= 1) IRNGT (1) = IHIGT (1)
         If (NORD >= 2) IRNGT (2) = ILOWT (1)
         If (NORD >= 3) IRNGT (3) = ILOWT (2)
         Return
      End If
!
      If (XDONT(NDON) > XDONT(ILOWT(1))) Then
         ILOWT (3) = ILOWT (2)
         ILOWT (2) = ILOWT (1)
         If (XDONT(NDON) > XDONT(IHIGT(1))) Then
            ILOWT (1) = IHIGT (1)
            IHIGT (1) = NDON
         Else
            ILOWT (1) = NDON
         End If
      Else
         if (XDONT (NDON) > XDONT (ILOWT(2))) Then
            ILOWT (3) = ILOWT (2)
            ILOWT (2) = NDON
         else
            ILOWT (3) = NDON
         endif 
      End If
!
      If (NDON < 5) Then
         If (NORD >= 1) IRNGT (1) = IHIGT (1)
         If (NORD >= 2) IRNGT (2) = ILOWT (1)
         If (NORD >= 3) IRNGT (3) = ILOWT (2)
         If (NORD >= 4) IRNGT (4) = ILOWT (3)
         Return
      End If
! ---
      JDEB = 0
      IDEB = JDEB + 1
      JHIG = IDEB
      JLOW = 3
      XPIV = XDONT (IHIGT(IDEB)) + REAL(2*NORD)/REAL(NDON+NORD) * &
                                   (XDONT(ILOWT(3))-XDONT(IHIGT(IDEB)))
      If (XPIV >= XDONT(ILOWT(1))) Then
         XPIV = XDONT (IHIGT(IDEB)) + REAL(2*NORD)/REAL(NDON+NORD) * &
                                      (XDONT(ILOWT(2))-XDONT(IHIGT(IDEB)))
         If (XPIV >= XDONT(ILOWT(1))) &
             XPIV = XDONT (IHIGT(IDEB)) + REAL (2*NORD) / REAL (NDON+NORD) * &
                                          (XDONT(ILOWT(1))-XDONT(IHIGT(IDEB)))
      End If
      XPIV0 = XPIV
! ---
!  One puts values < pivot in the end and those >= pivot
!  at the beginning. This is split in 2 cases, so that
!  we can skip the loop test a number of times.
!  As we are also filling in the work arrays at the same time
!  we stop filling in the ILOWT array as soon as we have more
!  than enough values in IHIGT.
!
!
      If (XDONT(NDON) < XPIV) Then
         ICRS = 3
         Do
            ICRS = ICRS + 1
            If (XDONT(ICRS) < XPIV) Then
               If (ICRS >= NDON) Exit
               JLOW = JLOW + 1
               ILOWT (JLOW) = ICRS
            Else
               JHIG = JHIG + 1
               IHIGT (JHIG) = ICRS
               If (JHIG >= NORD) Exit
            End If
         End Do
!
!  One restricts further processing because it is no use
!  to store more low values
!
         If (ICRS < NDON-1) Then
            Do
               ICRS = ICRS + 1
               If (XDONT(ICRS) >= XPIV) Then
                  JHIG = JHIG + 1
                  IHIGT (JHIG) = ICRS
               Else If (ICRS >= NDON) Then
                  Exit
               End If
            End Do
         End If
!
! ---
      Else
!
!  Same as above, but this is not as easy to optimize, so the
!  DO-loop is kept
!
         Do ICRS = 4, NDON - 1
            If (XDONT(ICRS) < XPIV) Then
               JLOW = JLOW + 1
               ILOWT (JLOW) = ICRS
            Else
               JHIG = JHIG + 1
               IHIGT (JHIG) = ICRS
               If (JHIG >= NORD) Exit
            End If
         End Do
!
         If (ICRS < NDON-1) Then
            Do
               ICRS = ICRS + 1
               If (XDONT(ICRS) >= XPIV) Then
                  If (ICRS >= NDON) Exit
                  JHIG = JHIG + 1
                  IHIGT (JHIG) = ICRS
               End If
            End Do
         End If
      End If
! ---
      JLM2 = 0
      JLM1 = 0
      JHM2 = 0
      JHM1 = 0
      Do
         if (JHIG == NORD) Exit
         If (JHM2 == JHIG .And. JLM2 == JLOW) Then
!
!   We are oscillating. Perturbate by bringing JHIG closer by one
!   to NORD
!
           If (NORD > JHIG) Then
                XMAX = XDONT (ILOWT(1))
                ILOW = 1
                Do ICRS = 2, JLOW
                   If (XDONT(ILOWT(ICRS)) > XMAX) Then
                      XMAX = XDONT (ILOWT(ICRS))
                      ILOW = ICRS
                   End If
                End Do
!
                JHIG = JHIG + 1
                IHIGT (JHIG) = ILOWT (ILOW)
                ILOWT (ILOW) = ILOWT (JLOW)
                JLOW = JLOW - 1
             Else
                IHIG = IHIGT (JHIG)
                XMIN = XDONT (IHIG)
                Do ICRS = 1, JHIG
                   If (XDONT(IHIGT(ICRS)) < XMIN) Then
                      IWRK = IHIGT (ICRS)
                      XMIN = XDONT (IWRK)
                      IHIGT (ICRS) = IHIG
                      IHIG = IWRK
                   End If
                End Do
                JHIG = JHIG - 1
             End If
         End If
         JLM2 = JLM1
         JLM1 = JLOW
         JHM2 = JHM1
         JHM1 = JHIG
! ---
!   We try to bring the number of values in the high values set
!   closer to NORD.
!
        Select Case (NORD-JHIG)
         Case (2:)
!
!   Not enough values in low part, at least 2 are missing
!
            Select Case (JLOW)
!!!!!           CASE DEFAULT
!!!!!              write (*,*) "Assertion failed"
!!!!!              STOP
!
!   We make a special case when we have so few values in
!   the low values set that it is bad performance to choose a pivot
!   and apply the general algorithm.
!
            Case (2)
               If (XDONT(ILOWT(1)) >= XDONT(ILOWT(2))) Then
                  JHIG = JHIG + 1
                  IHIGT (JHIG) = ILOWT (1)
                  JHIG = JHIG + 1
                  IHIGT (JHIG) = ILOWT (2)
               Else
                  JHIG = JHIG + 1
                  IHIGT (JHIG) = ILOWT (2)
                  JHIG = JHIG + 1
                  IHIGT (JHIG) = ILOWT (1)
               End If
               Exit
! ---
            Case (3)
!
!
               IWRK1 = ILOWT (1)
               IWRK2 = ILOWT (2)
               IWRK3 = ILOWT (3)
               If (XDONT(IWRK2) > XDONT(IWRK1)) Then
                  ILOWT (1) = IWRK2
                  ILOWT (2) = IWRK1
                  IWRK2 = IWRK1
               End If
               If (XDONT(IWRK2) < XDONT(IWRK3)) Then
                  ILOWT (3) = IWRK2
                  ILOWT (2) = IWRK3
                  IWRK2 = IWRK3
                  If (XDONT(IWRK2) > XDONT(ILOWT(1))) Then
                     ILOWT (2) = ILOWT (1)
                     ILOWT (1) = IWRK2
                  End If
               End If
               JLOW = 0
               Do ICRS = JHIG + 1, NORD
                  JLOW = JLOW + 1
                  IHIGT (ICRS) = ILOWT (JLOW)
               End Do
               JHIG = NORD
               Exit
! ---
            Case (4:)
!
!
               XPIV0 = XPIV
               IFIN = JLOW
!
!  One chooses a pivot from the 2 first values and the last one.
!  This should ensure sufficient renewal between iterations to
!  avoid worst case behavior effects.
!
               IWRK1 = ILOWT (1)
               IWRK2 = ILOWT (2)
               IWRK3 = ILOWT (IFIN)
               If (XDONT(IWRK2) > XDONT(IWRK1)) Then
                  ILOWT (1) = IWRK2
                  ILOWT (2) = IWRK1
                  IWRK2 = IWRK1
               End If
               If (XDONT(IWRK2) < XDONT(IWRK3)) Then
                  ILOWT (IFIN) = IWRK2
                  ILOWT (2) = IWRK3
                  IWRK2 = IWRK3
                  If (XDONT(IWRK2) > XDONT(IHIGT(1))) Then
                     ILOWT (2) = ILOWT (1)
                     ILOWT (1) = IWRK2
                  End If
               End If
!
               JDEB = JHIG
               NWRK = NORD - JHIG
               IWRK1 = ILOWT (1)
               JHIG = JHIG + 1
               IHIGT (JHIG) = IWRK1
               XPIV = XDONT (IWRK1) + REAL (NWRK) / REAL (NORD+NWRK) * &
                                      (XDONT(ILOWT(IFIN))-XDONT(IWRK1))
!
!  One takes values >= pivot to IHIGT
!  Again, 2 parts, one where we take care of the remaining
!  low values because we might still need them, and the
!  other when we know that we will have more than enough
!  high values in the end.
! ---
               JLOW = 0
               Do ICRS = 2, IFIN
                  If (XDONT(ILOWT(ICRS)) >= XPIV) Then
                     JHIG = JHIG + 1
                     IHIGT (JHIG) = ILOWT (ICRS)
                     If (JHIG >= NORD) Exit
                  Else
                     JLOW = JLOW + 1
                     ILOWT (JLOW) = ILOWT (ICRS)
                  End If
               End Do
!
               Do ICRS = ICRS + 1, IFIN
                  If (XDONT(ILOWT(ICRS)) >= XPIV) Then
                     JHIG = JHIG + 1
                     IHIGT (JHIG) = ILOWT (ICRS)
                  End If
               End Do
           End Select
! ---
!
         Case (1)
!
!  Only 1 value is missing in high part
!
            XMAX = XDONT (ILOWT(1))
            ILOW = 1
            Do ICRS = 2, JLOW
               If (XDONT(ILOWT(ICRS)) > XMAX) Then
                  XMAX = XDONT (ILOWT(ICRS))
                  ILOW = ICRS
               End If
            End Do
!
            JHIG = JHIG + 1
            IHIGT (JHIG) = ILOWT (ILOW)
            Exit
!
!
         Case (0)
!
!  Low part is exactly what we want
!
            Exit
! ---
!
         Case (-5:-1)
!
!  Only few values too many in high part
!
            IRNGT (1) = IHIGT (1)
            Do ICRS = 2, NORD
               IWRK = IHIGT (ICRS)
               XWRK = XDONT (IWRK)
               Do IDCR = ICRS - 1, 1, - 1
                  If (XWRK > XDONT(IRNGT(IDCR))) Then
                     IRNGT (IDCR+1) = IRNGT (IDCR)
                  Else
                     Exit
                  End If
               End Do
               IRNGT (IDCR+1) = IWRK
            End Do
!
            XWRK1 = XDONT (IRNGT(NORD))
            Do ICRS = NORD + 1, JHIG
               If (XDONT(IHIGT (ICRS)) > XWRK1) Then
                  XWRK = XDONT (IHIGT (ICRS))
                  Do IDCR = NORD - 1, 1, - 1
                     If (XWRK <= XDONT(IRNGT(IDCR))) Exit
                     IRNGT (IDCR+1) = IRNGT (IDCR)
                  End Do
                  IRNGT (IDCR+1) = IHIGT (ICRS)
                  XWRK1 = XDONT (IRNGT(NORD))
               End If
            End Do
!
            Return
!
!
         Case (:-6)
!
! last case: too many values in high part
! ---
            IDEB = JDEB + 1
            IMIL = (JHIG+IDEB) / 2
            IFIN = JHIG
! ---
!  One chooses a pivot from 1st, last, and middle values
!
            If (XDONT(IHIGT(IMIL)) > XDONT(IHIGT(IDEB))) Then
               IWRK = IHIGT (IDEB)
               IHIGT (IDEB) = IHIGT (IMIL)
               IHIGT (IMIL) = IWRK
            End If
            If (XDONT(IHIGT(IMIL)) < XDONT(IHIGT(IFIN))) Then
               IWRK = IHIGT (IFIN)
               IHIGT (IFIN) = IHIGT (IMIL)
               IHIGT (IMIL) = IWRK
               If (XDONT(IHIGT(IMIL)) > XDONT(IHIGT(IDEB))) Then
                  IWRK = IHIGT (IDEB)
                  IHIGT (IDEB) = IHIGT (IMIL)
                  IHIGT (IMIL) = IWRK
               End If
            End If
            If (IFIN <= 3) Exit
! ---
            XPIV = XDONT (IHIGT(1)) + REAL(NORD)/REAL(JHIG+NORD) * &
                                      (XDONT(IHIGT(IFIN))-XDONT(IHIGT(1)))
            If (JDEB > 0) Then
               If (XPIV <= XPIV0) &
                   XPIV = XPIV0 + REAL(2*NORD-JDEB)/REAL (JHIG+NORD) * &
                                  (XDONT(IHIGT(IFIN))-XPIV0)
            Else
               IDEB = 1
            End If
!
!  One takes values < XPIV to ILOWT
!  However, we do not process the first values if we have been
!  through the case when we did not have enough high values
! ---
            JLOW = 0
            JHIG = JDEB
! ---
            If (XDONT(IHIGT(IFIN)) < XPIV) Then
               ICRS = JDEB
               Do
                 ICRS = ICRS + 1
                  If (XDONT(IHIGT(ICRS)) < XPIV) Then
                     JLOW = JLOW + 1
                     ILOWT (JLOW) = IHIGT (ICRS)
                     If (ICRS >= IFIN) Exit
                  Else
                     JHIG = JHIG + 1
                     IHIGT (JHIG) = IHIGT (ICRS)
                     If (JHIG >= NORD) Exit
                  End If
               End Do
! ---
               If (ICRS < IFIN) Then
                  Do
                     ICRS = ICRS + 1
                     If (XDONT(IHIGT(ICRS)) >= XPIV) Then
                        JHIG = JHIG + 1
                        IHIGT (JHIG) = IHIGT (ICRS)
                     Else
                        If (ICRS >= IFIN) Exit
                     End If
                  End Do
               End If
           Else
               Do ICRS = IDEB, IFIN
                  If (XDONT(IHIGT(ICRS)) < XPIV) Then
                     JLOW = JLOW + 1
                     ILOWT (JLOW) = IHIGT (ICRS)
                  Else
                     JHIG = JHIG + 1
                     IHIGT (JHIG) = IHIGT (ICRS)
                     If (JHIG >= NORD) Exit
                  End If
               End Do
!
               Do ICRS = ICRS + 1, IFIN
                  If (XDONT(IHIGT(ICRS)) >= XPIV) Then
                     JHIG = JHIG + 1
                     IHIGT (JHIG) = IHIGT (ICRS)
                  End If
               End Do
            End If
!
         End Select
!
      End Do
! ---
!  Now, we only need to complete ranking of the 1:NORD set
!  Assuming NORD is small, we use a simple insertion sort
!
      IRNGT (1) = IHIGT (1)
      Do ICRS = 2, NORD
         IWRK = IHIGT (ICRS)
         XWRK = XDONT (IWRK)
         Do IDCR = ICRS - 1, 1, - 1
            If (XWRK > XDONT(IRNGT(IDCR))) Then
               IRNGT (IDCR+1) = IRNGT (IDCR)
            Else
               Exit
            End If
         End Do
         IRNGT (IDCR+1) = IWRK
      End Do
     Return
!
!
End Subroutine R_rapknr
Subroutine I_rapknr (XDONT, IRNGT, NORD)
!  Ranks partially XDONT by IRNGT, up to order NORD, in decreasing order.
!  rapknr = (rnkpar backwards)
! __________________________________________________________
!  This routine uses a pivoting strategy such as the one of
!  finding the median based on the quicksort algorithm, but
!  we skew the pivot choice to try to bring it to NORD as
!  fast as possible. It uses 2 temporary arrays, where it
!  stores the indices of the values larger than the pivot
!  (IHIGT), and the indices of values smaller than the pivot
!  that we might still need later on (ILOWT). It iterates
!  until it can bring the number of values in IHIGT to
!  exactly NORD, and then uses an insertion sort to rank
!  this set, since it is supposedly small.
!  Michel Olagnon - Feb. 2011
! __________________________________________________________
! __________________________________________________________
! __________________________________________________________
      Integer, Dimension (:), Intent (In)  :: XDONT
      Integer, Dimension (:), Intent (Out) :: IRNGT
      Integer, Intent (In) :: NORD
! __________________________________________________________
      Integer :: XPIV, XPIV0, XWRK, XWRK1, XMIN, XMAX
!
      Integer, Dimension (SIZE(XDONT)) :: ILOWT, IHIGT
      Integer :: NDON, JHIG, JLOW, IHIG, IWRK, IWRK1, IWRK2, IWRK3
      Integer :: IDEB, JDEB, IMIL, IFIN, NWRK, ICRS, IDCR, ILOW
      Integer :: JLM2, JLM1, JHM2, JHM1
!
      NDON = SIZE (XDONT)
!
!    First loop is used to fill-in ILOWT, IHIGT at the same time
!
      If (NDON < 2) Then
         If (NORD >= 1) IRNGT (1) = 1
         Return
      End If
!
!  One chooses a pivot, best estimate possible to put fractile near
!  mid-point of the set of high values.
!
      If (XDONT(2) < XDONT(1)) Then
         ILOWT (1) = 2
         IHIGT (1) = 1
      Else
         ILOWT (1) = 1
         IHIGT (1) = 2
      End If
!
      If (NDON < 3) Then
         If (NORD >= 1) IRNGT (1) = IHIGT (1)
         If (NORD >= 2) IRNGT (2) = ILOWT (1)
         Return
      End If
! ---
      If (XDONT(3) > XDONT(ILOWT(1))) Then
         ILOWT (2) = ILOWT (1)
         If (XDONT(3) > XDONT(IHIGT(1))) Then
            ILOWT (1) = IHIGT (1)
            IHIGT (1) = 3
         Else
            ILOWT (1) = 3
         End If
      Else
         ILOWT (2) = 3
      End If
! ---
      If (NDON < 4) Then
         If (NORD >= 1) IRNGT (1) = IHIGT (1)
         If (NORD >= 2) IRNGT (2) = ILOWT (1)
         If (NORD >= 3) IRNGT (3) = ILOWT (2)
         Return
      End If
!
      If (XDONT(NDON) > XDONT(ILOWT(1))) Then
         ILOWT (3) = ILOWT (2)
         ILOWT (2) = ILOWT (1)
         If (XDONT(NDON) > XDONT(IHIGT(1))) Then
            ILOWT (1) = IHIGT (1)
            IHIGT (1) = NDON
         Else
            ILOWT (1) = NDON
         End If
      Else
         if (XDONT (NDON) > XDONT (ILOWT(2))) Then
            ILOWT (3) = ILOWT (2)
            ILOWT (2) = NDON
         else
            ILOWT (3) = NDON
         endif 
      End If
!
      If (NDON < 5) Then
         If (NORD >= 1) IRNGT (1) = IHIGT (1)
         If (NORD >= 2) IRNGT (2) = ILOWT (1)
         If (NORD >= 3) IRNGT (3) = ILOWT (2)
         If (NORD >= 4) IRNGT (4) = ILOWT (3)
         Return
      End If
! ---
      JDEB = 0
      IDEB = JDEB + 1
      JHIG = IDEB
      JLOW = 3
      XPIV = XDONT (IHIGT(IDEB)) + REAL(2*NORD)/REAL(NDON+NORD) * &
                                   (XDONT(ILOWT(3))-XDONT(IHIGT(IDEB)))
      If (XPIV >= XDONT(ILOWT(1))) Then
         XPIV = XDONT (IHIGT(IDEB)) + REAL(2*NORD)/REAL(NDON+NORD) * &
                                      (XDONT(ILOWT(2))-XDONT(IHIGT(IDEB)))
         If (XPIV >= XDONT(ILOWT(1))) &
             XPIV = XDONT (IHIGT(IDEB)) + REAL (2*NORD) / REAL (NDON+NORD) * &
                                          (XDONT(ILOWT(1))-XDONT(IHIGT(IDEB)))
      End If
      XPIV0 = XPIV
! ---
!  One puts values < pivot in the end and those >= pivot
!  at the beginning. This is split in 2 cases, so that
!  we can skip the loop test a number of times.
!  As we are also filling in the work arrays at the same time
!  we stop filling in the ILOWT array as soon as we have more
!  than enough values in IHIGT.
!
!
      If (XDONT(NDON) < XPIV) Then
         ICRS = 3
         Do
            ICRS = ICRS + 1
            If (XDONT(ICRS) < XPIV) Then
               If (ICRS >= NDON) Exit
               JLOW = JLOW + 1
               ILOWT (JLOW) = ICRS
            Else
               JHIG = JHIG + 1
               IHIGT (JHIG) = ICRS
               If (JHIG >= NORD) Exit
            End If
         End Do
!
!  One restricts further processing because it is no use
!  to store more low values
!
         If (ICRS < NDON-1) Then
            Do
               ICRS = ICRS + 1
               If (XDONT(ICRS) >= XPIV) Then
                  JHIG = JHIG + 1
                  IHIGT (JHIG) = ICRS
               Else If (ICRS >= NDON) Then
                  Exit
               End If
            End Do
         End If
!
! ---
      Else
!
!  Same as above, but this is not as easy to optimize, so the
!  DO-loop is kept
!
         Do ICRS = 4, NDON - 1
            If (XDONT(ICRS) < XPIV) Then
               JLOW = JLOW + 1
               ILOWT (JLOW) = ICRS
            Else
               JHIG = JHIG + 1
               IHIGT (JHIG) = ICRS
               If (JHIG >= NORD) Exit
            End If
         End Do
!
         If (ICRS < NDON-1) Then
            Do
               ICRS = ICRS + 1
               If (XDONT(ICRS) >= XPIV) Then
                  If (ICRS >= NDON) Exit
                  JHIG = JHIG + 1
                  IHIGT (JHIG) = ICRS
               End If
            End Do
         End If
      End If
! ---
      JLM2 = 0
      JLM1 = 0
      JHM2 = 0
      JHM1 = 0
      Do
         if (JHIG == NORD) Exit
         If (JHM2 == JHIG .And. JLM2 == JLOW) Then
!
!   We are oscillating. Perturbate by bringing JHIG closer by one
!   to NORD
!
           If (NORD > JHIG) Then
                XMAX = XDONT (ILOWT(1))
                ILOW = 1
                Do ICRS = 2, JLOW
                   If (XDONT(ILOWT(ICRS)) > XMAX) Then
                      XMAX = XDONT (ILOWT(ICRS))
                      ILOW = ICRS
                   End If
                End Do
!
                JHIG = JHIG + 1
                IHIGT (JHIG) = ILOWT (ILOW)
                ILOWT (ILOW) = ILOWT (JLOW)
                JLOW = JLOW - 1
             Else
                IHIG = IHIGT (JHIG)
                XMIN = XDONT (IHIG)
                Do ICRS = 1, JHIG
                   If (XDONT(IHIGT(ICRS)) < XMIN) Then
                      IWRK = IHIGT (ICRS)
                      XMIN = XDONT (IWRK)
                      IHIGT (ICRS) = IHIG
                      IHIG = IWRK
                   End If
                End Do
                JHIG = JHIG - 1
             End If
         End If
         JLM2 = JLM1
         JLM1 = JLOW
         JHM2 = JHM1
         JHM1 = JHIG
! ---
!   We try to bring the number of values in the high values set
!   closer to NORD.
!
        Select Case (NORD-JHIG)
         Case (2:)
!
!   Not enough values in low part, at least 2 are missing
!
            Select Case (JLOW)
!!!!!           CASE DEFAULT
!!!!!              write (*,*) "Assertion failed"
!!!!!              STOP
!
!   We make a special case when we have so few values in
!   the low values set that it is bad performance to choose a pivot
!   and apply the general algorithm.
!
            Case (2)
               If (XDONT(ILOWT(1)) >= XDONT(ILOWT(2))) Then
                  JHIG = JHIG + 1
                  IHIGT (JHIG) = ILOWT (1)
                  JHIG = JHIG + 1
                  IHIGT (JHIG) = ILOWT (2)
               Else
                  JHIG = JHIG + 1
                  IHIGT (JHIG) = ILOWT (2)
                  JHIG = JHIG + 1
                  IHIGT (JHIG) = ILOWT (1)
               End If
               Exit
! ---
            Case (3)
!
!
               IWRK1 = ILOWT (1)
               IWRK2 = ILOWT (2)
               IWRK3 = ILOWT (3)
               If (XDONT(IWRK2) > XDONT(IWRK1)) Then
                  ILOWT (1) = IWRK2
                  ILOWT (2) = IWRK1
                  IWRK2 = IWRK1
               End If
               If (XDONT(IWRK2) < XDONT(IWRK3)) Then
                  ILOWT (3) = IWRK2
                  ILOWT (2) = IWRK3
                  IWRK2 = IWRK3
                  If (XDONT(IWRK2) > XDONT(ILOWT(1))) Then
                     ILOWT (2) = ILOWT (1)
                     ILOWT (1) = IWRK2
                  End If
               End If
               JLOW = 0
               Do ICRS = JHIG + 1, NORD
                  JLOW = JLOW + 1
                  IHIGT (ICRS) = ILOWT (JLOW)
               End Do
               JHIG = NORD
               Exit
! ---
            Case (4:)
!
!
               XPIV0 = XPIV
               IFIN = JLOW
!
!  One chooses a pivot from the 2 first values and the last one.
!  This should ensure sufficient renewal between iterations to
!  avoid worst case behavior effects.
!
               IWRK1 = ILOWT (1)
               IWRK2 = ILOWT (2)
               IWRK3 = ILOWT (IFIN)
               If (XDONT(IWRK2) > XDONT(IWRK1)) Then
                  ILOWT (1) = IWRK2
                  ILOWT (2) = IWRK1
                  IWRK2 = IWRK1
               End If
               If (XDONT(IWRK2) < XDONT(IWRK3)) Then
                  ILOWT (IFIN) = IWRK2
                  ILOWT (2) = IWRK3
                  IWRK2 = IWRK3
                  If (XDONT(IWRK2) > XDONT(IHIGT(1))) Then
                     ILOWT (2) = ILOWT (1)
                     ILOWT (1) = IWRK2
                  End If
               End If
!
               JDEB = JHIG
               NWRK = NORD - JHIG
               IWRK1 = ILOWT (1)
               JHIG = JHIG + 1
               IHIGT (JHIG) = IWRK1
               XPIV = XDONT (IWRK1) + REAL (NWRK) / REAL (NORD+NWRK) * &
                                      (XDONT(ILOWT(IFIN))-XDONT(IWRK1))
!
!  One takes values >= pivot to IHIGT
!  Again, 2 parts, one where we take care of the remaining
!  low values because we might still need them, and the
!  other when we know that we will have more than enough
!  high values in the end.
! ---
               JLOW = 0
               Do ICRS = 2, IFIN
                  If (XDONT(ILOWT(ICRS)) >= XPIV) Then
                     JHIG = JHIG + 1
                     IHIGT (JHIG) = ILOWT (ICRS)
                     If (JHIG >= NORD) Exit
                  Else
                     JLOW = JLOW + 1
                     ILOWT (JLOW) = ILOWT (ICRS)
                  End If
               End Do
!
               Do ICRS = ICRS + 1, IFIN
                  If (XDONT(ILOWT(ICRS)) >= XPIV) Then
                     JHIG = JHIG + 1
                     IHIGT (JHIG) = ILOWT (ICRS)
                  End If
               End Do
           End Select
! ---
!
         Case (1)
!
!  Only 1 value is missing in high part
!
            XMAX = XDONT (ILOWT(1))
            ILOW = 1
            Do ICRS = 2, JLOW
               If (XDONT(ILOWT(ICRS)) > XMAX) Then
                  XMAX = XDONT (ILOWT(ICRS))
                  ILOW = ICRS
               End If
            End Do
!
            JHIG = JHIG + 1
            IHIGT (JHIG) = ILOWT (ILOW)
            Exit
!
!
         Case (0)
!
!  Low part is exactly what we want
!
            Exit
! ---
!
         Case (-5:-1)
!
!  Only few values too many in high part
!
            IRNGT (1) = IHIGT (1)
            Do ICRS = 2, NORD
               IWRK = IHIGT (ICRS)
               XWRK = XDONT (IWRK)
               Do IDCR = ICRS - 1, 1, - 1
                  If (XWRK > XDONT(IRNGT(IDCR))) Then
                     IRNGT (IDCR+1) = IRNGT (IDCR)
                  Else
                     Exit
                  End If
               End Do
               IRNGT (IDCR+1) = IWRK
            End Do
!
            XWRK1 = XDONT (IRNGT(NORD))
            Do ICRS = NORD + 1, JHIG
               If (XDONT(IHIGT (ICRS)) > XWRK1) Then
                  XWRK = XDONT (IHIGT (ICRS))
                  Do IDCR = NORD - 1, 1, - 1
                     If (XWRK <= XDONT(IRNGT(IDCR))) Exit
                     IRNGT (IDCR+1) = IRNGT (IDCR)
                  End Do
                  IRNGT (IDCR+1) = IHIGT (ICRS)
                  XWRK1 = XDONT (IRNGT(NORD))
               End If
            End Do
!
            Return
!
!
         Case (:-6)
!
! last case: too many values in high part
! ---
            IDEB = JDEB + 1
            IMIL = (JHIG+IDEB) / 2
            IFIN = JHIG
! ---
!  One chooses a pivot from 1st, last, and middle values
!
            If (XDONT(IHIGT(IMIL)) > XDONT(IHIGT(IDEB))) Then
               IWRK = IHIGT (IDEB)
               IHIGT (IDEB) = IHIGT (IMIL)
               IHIGT (IMIL) = IWRK
            End If
            If (XDONT(IHIGT(IMIL)) < XDONT(IHIGT(IFIN))) Then
               IWRK = IHIGT (IFIN)
               IHIGT (IFIN) = IHIGT (IMIL)
               IHIGT (IMIL) = IWRK
               If (XDONT(IHIGT(IMIL)) > XDONT(IHIGT(IDEB))) Then
                  IWRK = IHIGT (IDEB)
                  IHIGT (IDEB) = IHIGT (IMIL)
                  IHIGT (IMIL) = IWRK
               End If
            End If
            If (IFIN <= 3) Exit
! ---
            XPIV = XDONT (IHIGT(1)) + REAL(NORD)/REAL(JHIG+NORD) * &
                                      (XDONT(IHIGT(IFIN))-XDONT(IHIGT(1)))
            If (JDEB > 0) Then
               If (XPIV <= XPIV0) &
                   XPIV = XPIV0 + REAL(2*NORD-JDEB)/REAL (JHIG+NORD) * &
                                  (XDONT(IHIGT(IFIN))-XPIV0)
            Else
               IDEB = 1
            End If
!
!  One takes values < XPIV to ILOWT
!  However, we do not process the first values if we have been
!  through the case when we did not have enough high values
! ---
            JLOW = 0
            JHIG = JDEB
! ---
            If (XDONT(IHIGT(IFIN)) < XPIV) Then
               ICRS = JDEB
               Do
                 ICRS = ICRS + 1
                  If (XDONT(IHIGT(ICRS)) < XPIV) Then
                     JLOW = JLOW + 1
                     ILOWT (JLOW) = IHIGT (ICRS)
                     If (ICRS >= IFIN) Exit
                  Else
                     JHIG = JHIG + 1
                     IHIGT (JHIG) = IHIGT (ICRS)
                     If (JHIG >= NORD) Exit
                  End If
               End Do
! ---
               If (ICRS < IFIN) Then
                  Do
                     ICRS = ICRS + 1
                     If (XDONT(IHIGT(ICRS)) >= XPIV) Then
                        JHIG = JHIG + 1
                        IHIGT (JHIG) = IHIGT (ICRS)
                     Else
                        If (ICRS >= IFIN) Exit
                     End If
                  End Do
               End If
           Else
               Do ICRS = IDEB, IFIN
                  If (XDONT(IHIGT(ICRS)) < XPIV) Then
                     JLOW = JLOW + 1
                     ILOWT (JLOW) = IHIGT (ICRS)
                  Else
                     JHIG = JHIG + 1
                     IHIGT (JHIG) = IHIGT (ICRS)
                     If (JHIG >= NORD) Exit
                  End If
               End Do
!
               Do ICRS = ICRS + 1, IFIN
                  If (XDONT(IHIGT(ICRS)) >= XPIV) Then
                     JHIG = JHIG + 1
                     IHIGT (JHIG) = IHIGT (ICRS)
                  End If
               End Do
            End If
!
         End Select
!
      End Do
! ---
!  Now, we only need to complete ranking of the 1:NORD set
!  Assuming NORD is small, we use a simple insertion sort
!
      IRNGT (1) = IHIGT (1)
      Do ICRS = 2, NORD
         IWRK = IHIGT (ICRS)
         XWRK = XDONT (IWRK)
         Do IDCR = ICRS - 1, 1, - 1
            If (XWRK > XDONT(IRNGT(IDCR))) Then
               IRNGT (IDCR+1) = IRNGT (IDCR)
            Else
               Exit
            End If
         End Do
         IRNGT (IDCR+1) = IWRK
      End Do
     Return
!
!
End Subroutine I_rapknr
end module m_rapknr
Module m_refpar
Integer, Parameter :: kdp = selected_real_kind(15)
public :: refpar
private :: kdp
private :: R_refpar, I_refpar, D_refpar
interface refpar
  module procedure d_refpar, r_refpar, i_refpar
end interface refpar
contains

Subroutine D_refpar (XDONT, IRNGT, NORD)
!  Ranks partially XDONT by IRNGT, up to order NORD
! __________________________________________________________
!  This routine uses a pivoting strategy such as the one of
!  finding the median based on the quicksort algorithm. It uses
!  a temporary array, where it stores the partially ranked indices
!  of the values. It iterates until it can bring the number of
!  values lower than the pivot to exactly NORD, and then uses an
!  insertion sort to rank this set, since it is supposedly small.
!  Michel Olagnon - Feb. 2000
! __________________________________________________________
! __________________________________________________________
      Real (kind=kdp), Dimension (:), Intent (In) :: XDONT
      Integer, Dimension (:), Intent (Out) :: IRNGT
      Integer, Intent (In) :: NORD
! __________________________________________________________
      Real (kind=kdp) :: XPIV, XWRK
! __________________________________________________________
!
      Integer, Dimension (SIZE(XDONT)) :: IWRKT
      Integer :: NDON, ICRS, IDEB, IDCR, IFIN, IMIL, IWRK
!
      NDON = SIZE (XDONT)
!
      Do ICRS = 1, NDON
         IWRKT (ICRS) = ICRS
      End Do
      IDEB = 1
      IFIN = NDON
      Do
         If (IDEB >= IFIN) Exit
         IMIL = (IDEB+IFIN) / 2
!
!  One chooses a pivot, median of 1st, last, and middle values
!
         If (XDONT(IWRKT(IMIL)) < XDONT(IWRKT(IDEB))) Then
            IWRK = IWRKT (IDEB)
            IWRKT (IDEB) = IWRKT (IMIL)
            IWRKT (IMIL) = IWRK
         End If
         If (XDONT(IWRKT(IMIL)) > XDONT(IWRKT(IFIN))) Then
            IWRK = IWRKT (IFIN)
            IWRKT (IFIN) = IWRKT (IMIL)
            IWRKT (IMIL) = IWRK
            If (XDONT(IWRKT(IMIL)) < XDONT(IWRKT(IDEB))) Then
               IWRK = IWRKT (IDEB)
               IWRKT (IDEB) = IWRKT (IMIL)
               IWRKT (IMIL) = IWRK
            End If
         End If
         If ((IFIN-IDEB) < 3) Exit
         XPIV = XDONT (IWRKT(IMIL))
!
!  One exchanges values to put those > pivot in the end and
!  those <= pivot at the beginning
!
         ICRS = IDEB
         IDCR = IFIN
         ECH2: Do
            Do
               ICRS = ICRS + 1
               If (ICRS >= IDCR) Then
!
!  the first  >  pivot is IWRKT(IDCR)
!  the last   <= pivot is IWRKT(ICRS-1)
!  Note: If one arrives here on the first iteration, then
!        the pivot is the maximum of the set, the last value is equal
!        to it, and one can reduce by one the size of the set to process,
!        as if XDONT (IWRKT(IFIN)) > XPIV
!
                  Exit ECH2
!
               End If
               If (XDONT(IWRKT(ICRS)) > XPIV) Exit
            End Do
            Do
               If (XDONT(IWRKT(IDCR)) <= XPIV) Exit
               IDCR = IDCR - 1
               If (ICRS >= IDCR) Then
!
!  The last value < pivot is always IWRKT(ICRS-1)
!
                  Exit ECH2
               End If
            End Do
!
            IWRK = IWRKT (IDCR)
            IWRKT (IDCR) = IWRKT (ICRS)
            IWRKT (ICRS) = IWRK
         End Do ECH2
!
!  One restricts further processing to find the fractile value
!
         If (ICRS <= NORD) IDEB = ICRS
         If (ICRS > NORD) IFIN = ICRS - 1
      End Do
!
!  Now, we only need to complete ranking of the 1:NORD set
!  Assuming NORD is small, we use a simple insertion sort
!
      Do ICRS = 2, NORD
         IWRK = IWRKT (ICRS)
         XWRK = XDONT (IWRK)
         Do IDCR = ICRS - 1, 1, - 1
            If (XWRK <= XDONT(IWRKT(IDCR))) Then
               IWRKT (IDCR+1) = IWRKT (IDCR)
            Else
               Exit
            End If
         End Do
         IWRKT (IDCR+1) = IWRK
      End Do
      IRNGT (1:NORD) = IWRKT (1:NORD)
      Return
!
End Subroutine D_refpar

Subroutine R_refpar (XDONT, IRNGT, NORD)
!  Ranks partially XDONT by IRNGT, up to order NORD
! __________________________________________________________
!  This routine uses a pivoting strategy such as the one of
!  finding the median based on the quicksort algorithm. It uses
!  a temporary array, where it stores the partially ranked indices
!  of the values. It iterates until it can bring the number of
!  values lower than the pivot to exactly NORD, and then uses an
!  insertion sort to rank this set, since it is supposedly small.
!  Michel Olagnon - Feb. 2000
! __________________________________________________________
! _________________________________________________________
      Real, Dimension (:), Intent (In) :: XDONT
      Integer, Dimension (:), Intent (Out) :: IRNGT
      Integer, Intent (In) :: NORD
! __________________________________________________________
      Real :: XPIV, XWRK
! __________________________________________________________
!
      Integer, Dimension (SIZE(XDONT)) :: IWRKT
      Integer :: NDON, ICRS, IDEB, IDCR, IFIN, IMIL, IWRK
!
      NDON = SIZE (XDONT)
!
      Do ICRS = 1, NDON
         IWRKT (ICRS) = ICRS
      End Do
      IDEB = 1
      IFIN = NDON
      Do
         If (IDEB >= IFIN) Exit
         IMIL = (IDEB+IFIN) / 2
!
!  One chooses a pivot, median of 1st, last, and middle values
!
         If (XDONT(IWRKT(IMIL)) < XDONT(IWRKT(IDEB))) Then
            IWRK = IWRKT (IDEB)
            IWRKT (IDEB) = IWRKT (IMIL)
            IWRKT (IMIL) = IWRK
         End If
         If (XDONT(IWRKT(IMIL)) > XDONT(IWRKT(IFIN))) Then
            IWRK = IWRKT (IFIN)
            IWRKT (IFIN) = IWRKT (IMIL)
            IWRKT (IMIL) = IWRK
            If (XDONT(IWRKT(IMIL)) < XDONT(IWRKT(IDEB))) Then
               IWRK = IWRKT (IDEB)
               IWRKT (IDEB) = IWRKT (IMIL)
               IWRKT (IMIL) = IWRK
            End If
         End If
         If ((IFIN-IDEB) < 3) Exit
         XPIV = XDONT (IWRKT(IMIL))
!
!  One exchanges values to put those > pivot in the end and
!  those <= pivot at the beginning
!
         ICRS = IDEB
         IDCR = IFIN
         ECH2: Do
            Do
               ICRS = ICRS + 1
               If (ICRS >= IDCR) Then
!
!  the first  >  pivot is IWRKT(IDCR)
!  the last   <= pivot is IWRKT(ICRS-1)
!  Note: If one arrives here on the first iteration, then
!        the pivot is the maximum of the set, the last value is equal
!        to it, and one can reduce by one the size of the set to process,
!        as if XDONT (IWRKT(IFIN)) > XPIV
!
                  Exit ECH2
!
               End If
               If (XDONT(IWRKT(ICRS)) > XPIV) Exit
            End Do
            Do
               If (XDONT(IWRKT(IDCR)) <= XPIV) Exit
               IDCR = IDCR - 1
               If (ICRS >= IDCR) Then
!
!  The last value < pivot is always IWRKT(ICRS-1)
!
                  Exit ECH2
               End If
            End Do
!
            IWRK = IWRKT (IDCR)
            IWRKT (IDCR) = IWRKT (ICRS)
            IWRKT (ICRS) = IWRK
         End Do ECH2
!
!  One restricts further processing to find the fractile value
!
         If (ICRS <= NORD) IDEB = ICRS
         If (ICRS > NORD) IFIN = ICRS - 1
      End Do
!
!  Now, we only need to complete ranking of the 1:NORD set
!  Assuming NORD is small, we use a simple insertion sort
!
      Do ICRS = 2, NORD
         IWRK = IWRKT (ICRS)
         XWRK = XDONT (IWRK)
         Do IDCR = ICRS - 1, 1, - 1
            If (XWRK <= XDONT(IWRKT(IDCR))) Then
               IWRKT (IDCR+1) = IWRKT (IDCR)
            Else
               Exit
            End If
         End Do
         IWRKT (IDCR+1) = IWRK
      End Do
      IRNGT (1:NORD) = IWRKT (1:NORD)
      Return
!
End Subroutine R_refpar
Subroutine I_refpar (XDONT, IRNGT, NORD)
!  Ranks partially XDONT by IRNGT, up to order NORD
! __________________________________________________________
!  This routine uses a pivoting strategy such as the one of
!  finding the median based on the quicksort algorithm. It uses
!  a temporary array, where it stores the partially ranked indices
!  of the values. It iterates until it can bring the number of
!  values lower than the pivot to exactly NORD, and then uses an
!  insertion sort to rank this set, since it is supposedly small.
!  Michel Olagnon - Feb. 2000
! __________________________________________________________
! __________________________________________________________
      Integer, Dimension (:), Intent (In)  :: XDONT
      Integer, Dimension (:), Intent (Out) :: IRNGT
      Integer, Intent (In) :: NORD
! __________________________________________________________
      Integer :: XPIV, XWRK
!
      Integer, Dimension (SIZE(XDONT)) :: IWRKT
      Integer :: NDON, ICRS, IDEB, IDCR, IFIN, IMIL, IWRK
!
      NDON = SIZE (XDONT)
!
      Do ICRS = 1, NDON
         IWRKT (ICRS) = ICRS
      End Do
      IDEB = 1
      IFIN = NDON
      Do
         If (IDEB >= IFIN) Exit
         IMIL = (IDEB+IFIN) / 2
!
!  One chooses a pivot, median of 1st, last, and middle values
!
         If (XDONT(IWRKT(IMIL)) < XDONT(IWRKT(IDEB))) Then
            IWRK = IWRKT (IDEB)
            IWRKT (IDEB) = IWRKT (IMIL)
            IWRKT (IMIL) = IWRK
         End If
         If (XDONT(IWRKT(IMIL)) > XDONT(IWRKT(IFIN))) Then
            IWRK = IWRKT (IFIN)
            IWRKT (IFIN) = IWRKT (IMIL)
            IWRKT (IMIL) = IWRK
            If (XDONT(IWRKT(IMIL)) < XDONT(IWRKT(IDEB))) Then
               IWRK = IWRKT (IDEB)
               IWRKT (IDEB) = IWRKT (IMIL)
               IWRKT (IMIL) = IWRK
            End If
         End If
         If ((IFIN-IDEB) < 3) Exit
         XPIV = XDONT (IWRKT(IMIL))
!
!  One exchanges values to put those > pivot in the end and
!  those <= pivot at the beginning
!
         ICRS = IDEB
         IDCR = IFIN
         ECH2: Do
            Do
               ICRS = ICRS + 1
               If (ICRS >= IDCR) Then
!
!  the first  >  pivot is IWRKT(IDCR)
!  the last   <= pivot is IWRKT(ICRS-1)
!  Note: If one arrives here on the first iteration, then
!        the pivot is the maximum of the set, the last value is equal
!        to it, and one can reduce by one the size of the set to process,
!        as if XDONT (IWRKT(IFIN)) > XPIV
!
                  Exit ECH2
!
               End If
               If (XDONT(IWRKT(ICRS)) > XPIV) Exit
            End Do
            Do
               If (XDONT(IWRKT(IDCR)) <= XPIV) Exit
               IDCR = IDCR - 1
               If (ICRS >= IDCR) Then
!
!  The last value < pivot is always IWRKT(ICRS-1)
!
                  Exit ECH2
               End If
            End Do
!
            IWRK = IWRKT (IDCR)
            IWRKT (IDCR) = IWRKT (ICRS)
            IWRKT (ICRS) = IWRK
         End Do ECH2
!
!  One restricts further processing to find the fractile value
!
         If (ICRS <= NORD) IDEB = ICRS
         If (ICRS > NORD) IFIN = ICRS - 1
      End Do
!
!  Now, we only need to complete ranking of the 1:NORD set
!  Assuming NORD is small, we use a simple insertion sort
!
      Do ICRS = 2, NORD
         IWRK = IWRKT (ICRS)
         XWRK = XDONT (IWRK)
         Do IDCR = ICRS - 1, 1, - 1
            If (XWRK <= XDONT(IWRKT(IDCR))) Then
               IWRKT (IDCR+1) = IWRKT (IDCR)
            Else
               Exit
            End If
         End Do
         IWRKT (IDCR+1) = IWRK
      End Do
      IRNGT (1:NORD) = IWRKT (1:NORD)
      Return
!
End Subroutine I_refpar
end module m_refpar
Module m_refsor
Integer, Parameter :: kdp = selected_real_kind(15)
public :: refsor
private :: kdp
private :: R_refsor, I_refsor, D_refsor
private :: R_inssor, I_inssor, D_inssor
private :: R_subsor, I_subsor, D_subsor
interface refsor
  module procedure d_refsor, r_refsor, i_refsor
end interface refsor
contains

Subroutine D_refsor (XDONT)
!  Sorts XDONT into ascending order - Quicksort
! __________________________________________________________
!  Quicksort chooses a "pivot" in the set, and explores the
!  array from both ends, looking for a value > pivot with the
!  increasing index, for a value <= pivot with the decreasing
!  index, and swapping them when it has found one of each.
!  The array is then subdivided in 2 ([3]) subsets:
!  { values <= pivot} {pivot} {values > pivot}
!  One then call recursively the program to sort each subset.
!  When the size of the subarray is small enough, one uses an
!  insertion sort that is faster for very small sets.
!  Michel Olagnon - Apr. 2000
! __________________________________________________________
! __________________________________________________________
      Real (kind=kdp), Dimension (:), Intent (InOut) :: XDONT
! __________________________________________________________
!
!
      Call D_subsor (XDONT, 1, Size (XDONT))
      Call D_inssor (XDONT)
      Return
End Subroutine D_refsor
Recursive Subroutine D_subsor (XDONT, IDEB1, IFIN1)
!  Sorts XDONT from IDEB1 to IFIN1
! __________________________________________________________
      Real(kind=kdp), dimension (:), Intent (InOut) :: XDONT
      Integer, Intent (In) :: IDEB1, IFIN1
! __________________________________________________________
      Integer, Parameter :: NINS = 16 ! Max for insertion sort
      Integer :: ICRS, IDEB, IDCR, IFIN, IMIL
      Real(kind=kdp) :: XPIV, XWRK
!
      IDEB = IDEB1
      IFIN = IFIN1
!
!  If we don't have enough values to make it worth while, we leave
!  them unsorted, and the final insertion sort will take care of them
!
      If ((IFIN - IDEB) > NINS) Then
         IMIL = (IDEB+IFIN) / 2
!
!  One chooses a pivot, median of 1st, last, and middle values
!
         If (XDONT(IMIL) < XDONT(IDEB)) Then
            XWRK = XDONT (IDEB)
            XDONT (IDEB) = XDONT (IMIL)
            XDONT (IMIL) = XWRK
         End If
         If (XDONT(IMIL) > XDONT(IFIN)) Then
            XWRK = XDONT (IFIN)
            XDONT (IFIN) = XDONT (IMIL)
            XDONT (IMIL) = XWRK
            If (XDONT(IMIL) < XDONT(IDEB)) Then
               XWRK = XDONT (IDEB)
               XDONT (IDEB) = XDONT (IMIL)
               XDONT (IMIL) = XWRK
            End If
         End If
         XPIV = XDONT (IMIL)
!
!  One exchanges values to put those > pivot in the end and
!  those <= pivot at the beginning
!
         ICRS = IDEB
         IDCR = IFIN
         ECH2: Do
            Do
               ICRS = ICRS + 1
               If (ICRS >= IDCR) Then
!
!  the first  >  pivot is IDCR
!  the last   <= pivot is ICRS-1
!  Note: If one arrives here on the first iteration, then
!        the pivot is the maximum of the set, the last value is equal
!        to it, and one can reduce by one the size of the set to process,
!        as if XDONT (IFIN) > XPIV
!
                  Exit ECH2
!
               End If
               If (XDONT(ICRS) > XPIV) Exit
            End Do
            Do
               If (XDONT(IDCR) <= XPIV) Exit
               IDCR = IDCR - 1
               If (ICRS >= IDCR) Then
!
!  The last value < pivot is always ICRS-1
!
                  Exit ECH2
               End If
            End Do
!
            XWRK = XDONT (IDCR)
            XDONT (IDCR) = XDONT (ICRS)
            XDONT (ICRS) = XWRK
         End Do ECH2
!
!  One now sorts each of the two sub-intervals
!
         Call D_subsor (XDONT, IDEB1, ICRS-1)
         Call D_subsor (XDONT, IDCR, IFIN1)
      End If
      Return
   End Subroutine D_subsor
   Subroutine D_inssor (XDONT)
!  Sorts XDONT into increasing order (Insertion sort)
! __________________________________________________________
      Real(kind=kdp), dimension (:), Intent (InOut) :: XDONT
! __________________________________________________________
      Integer :: ICRS, IDCR
      Real(kind=kdp) :: XWRK
!
      Do ICRS = 2, Size (XDONT)
         XWRK = XDONT (ICRS)
         If (XWRK >= XDONT(ICRS-1)) Cycle
         XDONT (ICRS) = XDONT (ICRS-1)
         Do IDCR = ICRS - 2, 1, - 1
            If (XWRK >= XDONT(IDCR)) Exit
            XDONT (IDCR+1) = XDONT (IDCR)
         End Do
         XDONT (IDCR+1) = XWRK
      End Do
!
      Return
!
End Subroutine D_inssor
!
Subroutine R_refsor (XDONT)
!  Sorts XDONT into ascending order - Quicksort
! __________________________________________________________
!  Quicksort chooses a "pivot" in the set, and explores the
!  array from both ends, looking for a value > pivot with the
!  increasing index, for a value <= pivot with the decreasing
!  index, and swapping them when it has found one of each.
!  The array is then subdivided in 2 ([3]) subsets:
!  { values <= pivot} {pivot} {values > pivot}
!  One then call recursively the program to sort each subset.
!  When the size of the subarray is small enough, one uses an
!  insertion sort that is faster for very small sets.
!  Michel Olagnon - Apr. 2000
! __________________________________________________________
! _________________________________________________________
      Real, Dimension (:), Intent (InOut) :: XDONT
! __________________________________________________________
!
!
      Call R_subsor (XDONT, 1, Size (XDONT))
      Call R_inssor (XDONT)
      Return
End Subroutine R_refsor
Recursive Subroutine R_subsor (XDONT, IDEB1, IFIN1)
!  Sorts XDONT from IDEB1 to IFIN1
! __________________________________________________________
      Real, dimension (:), Intent (InOut) :: XDONT
      Integer, Intent (In) :: IDEB1, IFIN1
! __________________________________________________________
      Integer, Parameter :: NINS = 16 ! Max for insertion sort
      Integer :: ICRS, IDEB, IDCR, IFIN, IMIL
      Real :: XPIV, XWRK
!
      IDEB = IDEB1
      IFIN = IFIN1
!
!  If we don't have enough values to make it worth while, we leave
!  them unsorted, and the final insertion sort will take care of them
!
      If ((IFIN - IDEB) > NINS) Then
         IMIL = (IDEB+IFIN) / 2
!
!  One chooses a pivot, median of 1st, last, and middle values
!
         If (XDONT(IMIL) < XDONT(IDEB)) Then
            XWRK = XDONT (IDEB)
            XDONT (IDEB) = XDONT (IMIL)
            XDONT (IMIL) = XWRK
         End If
         If (XDONT(IMIL) > XDONT(IFIN)) Then
            XWRK = XDONT (IFIN)
            XDONT (IFIN) = XDONT (IMIL)
            XDONT (IMIL) = XWRK
            If (XDONT(IMIL) < XDONT(IDEB)) Then
               XWRK = XDONT (IDEB)
               XDONT (IDEB) = XDONT (IMIL)
               XDONT (IMIL) = XWRK
            End If
         End If
         XPIV = XDONT (IMIL)
!
!  One exchanges values to put those > pivot in the end and
!  those <= pivot at the beginning
!
         ICRS = IDEB
         IDCR = IFIN
         ECH2: Do
            Do
               ICRS = ICRS + 1
               If (ICRS >= IDCR) Then
!
!  the first  >  pivot is IDCR
!  the last   <= pivot is ICRS-1
!  Note: If one arrives here on the first iteration, then
!        the pivot is the maximum of the set, the last value is equal
!        to it, and one can reduce by one the size of the set to process,
!        as if XDONT (IFIN) > XPIV
!
                  Exit ECH2
!
               End If
               If (XDONT(ICRS) > XPIV) Exit
            End Do
            Do
               If (XDONT(IDCR) <= XPIV) Exit
               IDCR = IDCR - 1
               If (ICRS >= IDCR) Then
!
!  The last value < pivot is always ICRS-1
!
                  Exit ECH2
               End If
            End Do
!
            XWRK = XDONT (IDCR)
            XDONT (IDCR) = XDONT (ICRS)
            XDONT (ICRS) = XWRK
         End Do ECH2
!
!  One now sorts each of the two sub-intervals
!
         Call R_subsor (XDONT, IDEB1, ICRS-1)
         Call R_subsor (XDONT, IDCR, IFIN1)
      End If
      Return
   End Subroutine R_subsor
   Subroutine R_inssor (XDONT)
!  Sorts XDONT into increasing order (Insertion sort)
! __________________________________________________________
      Real, dimension (:), Intent (InOut) :: XDONT
! __________________________________________________________
      Integer :: ICRS, IDCR
      Real :: XWRK
!
      Do ICRS = 2, Size (XDONT)
         XWRK = XDONT (ICRS)
         If (XWRK >= XDONT(ICRS-1)) Cycle
         XDONT (ICRS) = XDONT (ICRS-1)
         Do IDCR = ICRS - 2, 1, - 1
            If (XWRK >= XDONT(IDCR)) Exit
            XDONT (IDCR+1) = XDONT (IDCR)
         End Do
         XDONT (IDCR+1) = XWRK
      End Do
!
      Return
!
End Subroutine R_inssor
!
Subroutine I_refsor (XDONT)
!  Sorts XDONT into ascending order - Quicksort
! __________________________________________________________
!  Quicksort chooses a "pivot" in the set, and explores the
!  array from both ends, looking for a value > pivot with the
!  increasing index, for a value <= pivot with the decreasing
!  index, and swapping them when it has found one of each.
!  The array is then subdivided in 2 ([3]) subsets:
!  { values <= pivot} {pivot} {values > pivot}
!  One then call recursively the program to sort each subset.
!  When the size of the subarray is small enough, one uses an
!  insertion sort that is faster for very small sets.
!  Michel Olagnon - Apr. 2000
! __________________________________________________________
! __________________________________________________________
      Integer, Dimension (:), Intent (InOut)  :: XDONT
! __________________________________________________________
!
!
      Call I_subsor (XDONT, 1, Size (XDONT))
      Call I_inssor (XDONT)
      Return
End Subroutine I_refsor
Recursive Subroutine I_subsor (XDONT, IDEB1, IFIN1)
!  Sorts XDONT from IDEB1 to IFIN1
! __________________________________________________________
      Integer, dimension (:), Intent (InOut) :: XDONT
      Integer, Intent (In) :: IDEB1, IFIN1
! __________________________________________________________
      Integer, Parameter :: NINS = 16 ! Max for insertion sort
      Integer :: ICRS, IDEB, IDCR, IFIN, IMIL
      Integer :: XPIV, XWRK
!
      IDEB = IDEB1
      IFIN = IFIN1
!
!  If we don't have enough values to make it worth while, we leave
!  them unsorted, and the final insertion sort will take care of them
!
      If ((IFIN - IDEB) > NINS) Then
         IMIL = (IDEB+IFIN) / 2
!
!  One chooses a pivot, median of 1st, last, and middle values
!
         If (XDONT(IMIL) < XDONT(IDEB)) Then
            XWRK = XDONT (IDEB)
            XDONT (IDEB) = XDONT (IMIL)
            XDONT (IMIL) = XWRK
         End If
         If (XDONT(IMIL) > XDONT(IFIN)) Then
            XWRK = XDONT (IFIN)
            XDONT (IFIN) = XDONT (IMIL)
            XDONT (IMIL) = XWRK
            If (XDONT(IMIL) < XDONT(IDEB)) Then
               XWRK = XDONT (IDEB)
               XDONT (IDEB) = XDONT (IMIL)
               XDONT (IMIL) = XWRK
            End If
         End If
         XPIV = XDONT (IMIL)
!
!  One exchanges values to put those > pivot in the end and
!  those <= pivot at the beginning
!
         ICRS = IDEB
         IDCR = IFIN
         ECH2: Do
            Do
               ICRS = ICRS + 1
               If (ICRS >= IDCR) Then
!
!  the first  >  pivot is IDCR
!  the last   <= pivot is ICRS-1
!  Note: If one arrives here on the first iteration, then
!        the pivot is the maximum of the set, the last value is equal
!        to it, and one can reduce by one the size of the set to process,
!        as if XDONT (IFIN) > XPIV
!
                  Exit ECH2
!
               End If
               If (XDONT(ICRS) > XPIV) Exit
            End Do
            Do
               If (XDONT(IDCR) <= XPIV) Exit
               IDCR = IDCR - 1
               If (ICRS >= IDCR) Then
!
!  The last value < pivot is always ICRS-1
!
                  Exit ECH2
               End If
            End Do
!
            XWRK = XDONT (IDCR)
            XDONT (IDCR) = XDONT (ICRS)
            XDONT (ICRS) = XWRK
         End Do ECH2
!
!  One now sorts each of the two sub-intervals
!
         Call I_subsor (XDONT, IDEB1, ICRS-1)
         Call I_subsor (XDONT, IDCR, IFIN1)
      End If
      Return
   End Subroutine I_subsor
   Subroutine I_inssor (XDONT)
!  Sorts XDONT into increasing order (Insertion sort)
! __________________________________________________________
      Integer, dimension (:), Intent (InOut) :: XDONT
! __________________________________________________________
      Integer :: ICRS, IDCR
      Integer :: XWRK
!
      Do ICRS = 2, Size (XDONT)
         XWRK = XDONT (ICRS)
         If (XWRK >= XDONT(ICRS-1)) Cycle
         XDONT (ICRS) = XDONT (ICRS-1)
         Do IDCR = ICRS - 2, 1, - 1
            If (XWRK >= XDONT(IDCR)) Exit
            XDONT (IDCR+1) = XDONT (IDCR)
         End Do
         XDONT (IDCR+1) = XWRK
      End Do
!
      Return
!
End Subroutine I_inssor
!
end module m_refsor
Module m_rinpar
Integer, Parameter :: kdp = selected_real_kind(15)
public :: rinpar
private :: kdp
private :: R_rinpar, I_rinpar, D_rinpar
interface rinpar
  module procedure d_rinpar, r_rinpar, i_rinpar
end interface rinpar
contains

Subroutine D_rinpar (XDONT, IRNGT, NORD)
!  Ranks partially XDONT by IRNGT, up to order NORD = size (IRNGT)
! __________________________________________________________
!  This subroutine uses insertion sort, limiting insertion
!  to the first NORD values. It does not use any work array
!  and is faster when NORD is very small (2-5), but worst case
!  behavior can happen fairly probably (initially inverse sorted)
!  In many cases, the refined quicksort method is faster.
!  Michel Olagnon - Feb. 2000
! __________________________________________________________
! __________________________________________________________
      Real (kind=kdp), Dimension (:), Intent (In) :: XDONT
      Integer, Dimension (:), Intent (Out) :: IRNGT
      Integer, Intent (In) :: NORD
! __________________________________________________________
      Real (kind=kdp) :: XWRK, XWRK1
!
      Integer :: ICRS, IDCR
!
      IRNGT (1) = 1
      Do ICRS = 2, NORD
         XWRK = XDONT (ICRS)
         Do IDCR = ICRS - 1, 1, - 1
            If (XWRK >= XDONT(IRNGT(IDCR))) Exit
            IRNGT (IDCR+1) = IRNGT (IDCR)
         End Do
         IRNGT (IDCR+1) = ICRS
      End Do
!
      XWRK1 = XDONT (IRNGT(NORD))
      Do ICRS = NORD + 1, SIZE (XDONT)
         If (XDONT(ICRS) < XWRK1) Then
            XWRK = XDONT (ICRS)
            Do IDCR = NORD - 1, 1, - 1
               If (XWRK >= XDONT(IRNGT(IDCR))) Exit
               IRNGT (IDCR+1) = IRNGT (IDCR)
            End Do
            IRNGT (IDCR+1) = ICRS
            XWRK1 = XDONT (IRNGT(NORD))
         End If
      End Do
!
!
End Subroutine D_rinpar

Subroutine R_rinpar (XDONT, IRNGT, NORD)
!  Ranks partially XDONT by IRNGT, up to order NORD = size (IRNGT)
! __________________________________________________________
!  This subroutine uses insertion sort, limiting insertion
!  to the first NORD values. It does not use any work array
!  and is faster when NORD is very small (2-5), but worst case
!  behavior can happen fairly probably (initially inverse sorted)
!  In many cases, the refined quicksort method is faster.
!  Michel Olagnon - Feb. 2000
! __________________________________________________________
! _________________________________________________________
      Real, Dimension (:), Intent (In) :: XDONT
      Integer, Dimension (:), Intent (Out) :: IRNGT
      Integer, Intent (In) :: NORD
! __________________________________________________________
      Real    :: XWRK, XWRK1
!
      Integer :: ICRS, IDCR
!
      IRNGT (1) = 1
      Do ICRS = 2, NORD
         XWRK = XDONT (ICRS)
         Do IDCR = ICRS - 1, 1, - 1
            If (XWRK >= XDONT(IRNGT(IDCR))) Exit
            IRNGT (IDCR+1) = IRNGT (IDCR)
         End Do
         IRNGT (IDCR+1) = ICRS
      End Do
!
      XWRK1 = XDONT (IRNGT(NORD))
      Do ICRS = NORD + 1, SIZE (XDONT)
         If (XDONT(ICRS) < XWRK1) Then
            XWRK = XDONT (ICRS)
            Do IDCR = NORD - 1, 1, - 1
               If (XWRK >= XDONT(IRNGT(IDCR))) Exit
               IRNGT (IDCR+1) = IRNGT (IDCR)
            End Do
            IRNGT (IDCR+1) = ICRS
            XWRK1 = XDONT (IRNGT(NORD))
         End If
      End Do
!
!
End Subroutine R_rinpar
Subroutine I_rinpar (XDONT, IRNGT, NORD)
!  Ranks partially XDONT by IRNGT, up to order NORD = size (IRNGT)
! __________________________________________________________
!  This subroutine uses insertion sort, limiting insertion
!  to the first NORD values. It does not use any work array
!  and is faster when NORD is very small (2-5), but worst case
!  behavior can happen fairly probably (initially inverse sorted)
!  In many cases, the refined quicksort method is faster.
!  Michel Olagnon - Feb. 2000
! __________________________________________________________
! __________________________________________________________
      Integer, Dimension (:), Intent (In)  :: XDONT
      Integer, Dimension (:), Intent (Out) :: IRNGT
      Integer, Intent (In) :: NORD
! __________________________________________________________
      Integer :: XWRK, XWRK1
!
      Integer :: ICRS, IDCR
!
      IRNGT (1) = 1
      Do ICRS = 2, NORD
         XWRK = XDONT (ICRS)
         Do IDCR = ICRS - 1, 1, - 1
            If (XWRK >= XDONT(IRNGT(IDCR))) Exit
            IRNGT (IDCR+1) = IRNGT (IDCR)
         End Do
         IRNGT (IDCR+1) = ICRS
      End Do
!
      XWRK1 = XDONT (IRNGT(NORD))
      Do ICRS = NORD + 1, SIZE (XDONT)
         If (XDONT(ICRS) < XWRK1) Then
            XWRK = XDONT (ICRS)
            Do IDCR = NORD - 1, 1, - 1
               If (XWRK >= XDONT(IRNGT(IDCR))) Exit
               IRNGT (IDCR+1) = IRNGT (IDCR)
            End Do
            IRNGT (IDCR+1) = ICRS
            XWRK1 = XDONT (IRNGT(NORD))
         End If
      End Do
!
!
End Subroutine I_rinpar
end module m_rinpar
Module m_rnkpar
Integer, Parameter :: kdp = selected_real_kind(15)
public :: rnkpar
private :: kdp
private :: R_rnkpar, I_rnkpar, D_rnkpar
interface rnkpar
  module procedure d_rnkpar, r_rnkpar, i_rnkpar
end interface rnkpar
contains

Subroutine D_rnkpar (XDONT, IRNGT, NORD)
!  Ranks partially XDONT by IRNGT, up to order NORD
! __________________________________________________________
!  This routine uses a pivoting strategy such as the one of
!  finding the median based on the quicksort algorithm, but
!  we skew the pivot choice to try to bring it to NORD as
!  fast as possible. It uses 2 temporary arrays, where it
!  stores the indices of the values smaller than the pivot
!  (ILOWT), and the indices of values larger than the pivot
!  that we might still need later on (IHIGT). It iterates
!  until it can bring the number of values in ILOWT to
!  exactly NORD, and then uses an insertion sort to rank
!  this set, since it is supposedly small.
!  Michel Olagnon - Feb. 2000
! __________________________________________________________
! __________________________________________________________
      Real (kind=kdp), Dimension (:), Intent (In) :: XDONT
      Integer, Dimension (:), Intent (Out) :: IRNGT
      Integer, Intent (In) :: NORD
! __________________________________________________________
      Real (kind=kdp) :: XPIV, XPIV0, XWRK, XWRK1, XMIN, XMAX
!
      Integer, Dimension (SIZE(XDONT)) :: ILOWT, IHIGT
      Integer :: NDON, JHIG, JLOW, IHIG, IWRK, IWRK1, IWRK2, IWRK3
      Integer :: IDEB, JDEB, IMIL, IFIN, NWRK, ICRS, IDCR, ILOW
      Integer :: JLM2, JLM1, JHM2, JHM1
!
      NDON = SIZE (XDONT)
!
!    First loop is used to fill-in ILOWT, IHIGT at the same time
!
      If (NDON < 2) Then
         If (NORD >= 1) IRNGT (1) = 1
         Return
      End If
!
!  One chooses a pivot, best estimate possible to put fractile near
!  mid-point of the set of low values.
!
      If (XDONT(2) < XDONT(1)) Then
         ILOWT (1) = 2
         IHIGT (1) = 1
      Else
         ILOWT (1) = 1
         IHIGT (1) = 2
      End If
!
      If (NDON < 3) Then
         If (NORD >= 1) IRNGT (1) = ILOWT (1)
         If (NORD >= 2) IRNGT (2) = IHIGT (1)
         Return
      End If
!
      If (XDONT(3) <= XDONT(IHIGT(1))) Then
         IHIGT (2) = IHIGT (1)
         If (XDONT(3) < XDONT(ILOWT(1))) Then
            IHIGT (1) = ILOWT (1)
            ILOWT (1) = 3
         Else
            IHIGT (1) = 3
         End If
      Else
         IHIGT (2) = 3
      End If
!
      If (NDON < 4) Then
         If (NORD >= 1) IRNGT (1) = ILOWT (1)
         If (NORD >= 2) IRNGT (2) = IHIGT (1)
         If (NORD >= 3) IRNGT (3) = IHIGT (2)
         Return
      End If
!
      If (XDONT(NDON) <= XDONT(IHIGT(1))) Then
         IHIGT (3) = IHIGT (2)
         IHIGT (2) = IHIGT (1)
         If (XDONT(NDON) < XDONT(ILOWT(1))) Then
            IHIGT (1) = ILOWT (1)
            ILOWT (1) = NDON
         Else
            IHIGT (1) = NDON
         End If
      Else
         if (XDONT (NDON) < XDONT (IHIGT(2))) Then
            IHIGT (3) = IHIGT (2)
            IHIGT (2) = NDON
         else
            IHIGT (3) = NDON
         endif 
      End If
!
      If (NDON < 5) Then
         If (NORD >= 1) IRNGT (1) = ILOWT (1)
         If (NORD >= 2) IRNGT (2) = IHIGT (1)
         If (NORD >= 3) IRNGT (3) = IHIGT (2)
         If (NORD >= 4) IRNGT (4) = IHIGT (3)
         Return
      End If
!
      JDEB = 0
      IDEB = JDEB + 1
      JLOW = IDEB
      JHIG = 3
      XPIV = XDONT (ILOWT(IDEB)) + REAL(2*NORD)/REAL(NDON+NORD) * &
                                   (XDONT(IHIGT(3))-XDONT(ILOWT(IDEB)))
      If (XPIV >= XDONT(IHIGT(1))) Then
         XPIV = XDONT (ILOWT(IDEB)) + REAL(2*NORD)/REAL(NDON+NORD) * &
                                      (XDONT(IHIGT(2))-XDONT(ILOWT(IDEB)))
         If (XPIV >= XDONT(IHIGT(1))) &
             XPIV = XDONT (ILOWT(IDEB)) + REAL (2*NORD) / REAL (NDON+NORD) * &
                                          (XDONT(IHIGT(1))-XDONT(ILOWT(IDEB)))
      End If
      XPIV0 = XPIV
!
!  One puts values > pivot in the end and those <= pivot
!  at the beginning. This is split in 2 cases, so that
!  we can skip the loop test a number of times.
!  As we are also filling in the work arrays at the same time
!  we stop filling in the IHIGT array as soon as we have more
!  than enough values in ILOWT.
!
!
      If (XDONT(NDON) > XPIV) Then
         ICRS = 3
         Do
            ICRS = ICRS + 1
            If (XDONT(ICRS) > XPIV) Then
               If (ICRS >= NDON) Exit
               JHIG = JHIG + 1
               IHIGT (JHIG) = ICRS
            Else
               JLOW = JLOW + 1
               ILOWT (JLOW) = ICRS
               If (JLOW >= NORD) Exit
            End If
         End Do
!
!  One restricts further processing because it is no use
!  to store more high values
!
         If (ICRS < NDON-1) Then
            Do
               ICRS = ICRS + 1
               If (XDONT(ICRS) <= XPIV) Then
                  JLOW = JLOW + 1
                  ILOWT (JLOW) = ICRS
               Else If (ICRS >= NDON) Then
                  Exit
               End If
            End Do
         End If
!
!
      Else
!
!  Same as above, but this is not as easy to optimize, so the
!  DO-loop is kept
!
         Do ICRS = 4, NDON - 1
            If (XDONT(ICRS) > XPIV) Then
               JHIG = JHIG + 1
               IHIGT (JHIG) = ICRS
            Else
               JLOW = JLOW + 1
               ILOWT (JLOW) = ICRS
               If (JLOW >= NORD) Exit
            End If
         End Do
!
         If (ICRS < NDON-1) Then
            Do
               ICRS = ICRS + 1
               If (XDONT(ICRS) <= XPIV) Then
                  If (ICRS >= NDON) Exit
                  JLOW = JLOW + 1
                  ILOWT (JLOW) = ICRS
               End If
            End Do
         End If
      End If
!
      JLM2 = 0
      JLM1 = 0
      JHM2 = 0
      JHM1 = 0
      Do
         if (JLOW == NORD) Exit
         If (JLM2 == JLOW .And. JHM2 == JHIG) Then
!
!   We are oscillating. Perturbate by bringing JLOW closer by one
!   to NORD
!
           If (NORD > JLOW) Then
                XMIN = XDONT (IHIGT(1))
                IHIG = 1
                Do ICRS = 2, JHIG
                   If (XDONT(IHIGT(ICRS)) < XMIN) Then
                      XMIN = XDONT (IHIGT(ICRS))
                      IHIG = ICRS
                   End If
                End Do
!
                JLOW = JLOW + 1
                ILOWT (JLOW) = IHIGT (IHIG)
                IHIGT (IHIG) = IHIGT (JHIG)
                JHIG = JHIG - 1
             Else
                ILOW = ILOWT (JLOW)
                XMAX = XDONT (ILOW)
                Do ICRS = 1, JLOW
                   If (XDONT(ILOWT(ICRS)) > XMAX) Then
                      IWRK = ILOWT (ICRS)
                      XMAX = XDONT (IWRK)
                      ILOWT (ICRS) = ILOW
                      ILOW = IWRK
                   End If
                End Do
                JLOW = JLOW - 1
             End If
         End If
         JLM2 = JLM1
         JLM1 = JLOW
         JHM2 = JHM1
         JHM1 = JHIG
!
!   We try to bring the number of values in the low values set
!   closer to NORD.
!
        Select Case (NORD-JLOW)
         Case (2:)
!
!   Not enough values in low part, at least 2 are missing
!
            Select Case (JHIG)
!!!!!           CASE DEFAULT
!!!!!              write (*,*) "Assertion failed"
!!!!!              STOP
!
!   We make a special case when we have so few values in
!   the high values set that it is bad performance to choose a pivot
!   and apply the general algorithm.
!
            Case (2)
               If (XDONT(IHIGT(1)) <= XDONT(IHIGT(2))) Then
                  JLOW = JLOW + 1
                  ILOWT (JLOW) = IHIGT (1)
                  JLOW = JLOW + 1
                  ILOWT (JLOW) = IHIGT (2)
               Else
                  JLOW = JLOW + 1
                  ILOWT (JLOW) = IHIGT (2)
                  JLOW = JLOW + 1
                  ILOWT (JLOW) = IHIGT (1)
               End If
               Exit
!
            Case (3)
!
!
               IWRK1 = IHIGT (1)
               IWRK2 = IHIGT (2)
               IWRK3 = IHIGT (3)
               If (XDONT(IWRK2) < XDONT(IWRK1)) Then
                  IHIGT (1) = IWRK2
                  IHIGT (2) = IWRK1
                  IWRK2 = IWRK1
               End If
               If (XDONT(IWRK2) > XDONT(IWRK3)) Then
                  IHIGT (3) = IWRK2
                  IHIGT (2) = IWRK3
                  IWRK2 = IWRK3
                  If (XDONT(IWRK2) < XDONT(IHIGT(1))) Then
                     IHIGT (2) = IHIGT (1)
                     IHIGT (1) = IWRK2
                  End If
               End If
               JHIG = 0
               Do ICRS = JLOW + 1, NORD
                  JHIG = JHIG + 1
                  ILOWT (ICRS) = IHIGT (JHIG)
               End Do
               JLOW = NORD
               Exit
!
            Case (4:)
!
!
               XPIV0 = XPIV
               IFIN = JHIG
!
!  One chooses a pivot from the 2 first values and the last one.
!  This should ensure sufficient renewal between iterations to
!  avoid worst case behavior effects.
!
               IWRK1 = IHIGT (1)
               IWRK2 = IHIGT (2)
               IWRK3 = IHIGT (IFIN)
               If (XDONT(IWRK2) < XDONT(IWRK1)) Then
                  IHIGT (1) = IWRK2
                  IHIGT (2) = IWRK1
                  IWRK2 = IWRK1
               End If
               If (XDONT(IWRK2) > XDONT(IWRK3)) Then
                  IHIGT (IFIN) = IWRK2
                  IHIGT (2) = IWRK3
                  IWRK2 = IWRK3
                  If (XDONT(IWRK2) < XDONT(IHIGT(1))) Then
                     IHIGT (2) = IHIGT (1)
                     IHIGT (1) = IWRK2
                  End If
               End If
!
               JDEB = JLOW
               NWRK = NORD - JLOW
               IWRK1 = IHIGT (1)
               JLOW = JLOW + 1
               ILOWT (JLOW) = IWRK1
               XPIV = XDONT (IWRK1) + REAL (NWRK) / REAL (NORD+NWRK) * &
                                      (XDONT(IHIGT(IFIN))-XDONT(IWRK1))
!
!  One takes values <= pivot to ILOWT
!  Again, 2 parts, one where we take care of the remaining
!  high values because we might still need them, and the
!  other when we know that we will have more than enough
!  low values in the end.
!
               JHIG = 0
               Do ICRS = 2, IFIN
                  If (XDONT(IHIGT(ICRS)) <= XPIV) Then
                     JLOW = JLOW + 1
                     ILOWT (JLOW) = IHIGT (ICRS)
                     If (JLOW >= NORD) Exit
                  Else
                     JHIG = JHIG + 1
                     IHIGT (JHIG) = IHIGT (ICRS)
                  End If
               End Do
!
               Do ICRS = ICRS + 1, IFIN
                  If (XDONT(IHIGT(ICRS)) <= XPIV) Then
                     JLOW = JLOW + 1
                     ILOWT (JLOW) = IHIGT (ICRS)
                  End If
               End Do
           End Select
!
!
         Case (1)
!
!  Only 1 value is missing in low part
!
            XMIN = XDONT (IHIGT(1))
            IHIG = 1
            Do ICRS = 2, JHIG
               If (XDONT(IHIGT(ICRS)) < XMIN) Then
                  XMIN = XDONT (IHIGT(ICRS))
                  IHIG = ICRS
               End If
            End Do
!
            JLOW = JLOW + 1
            ILOWT (JLOW) = IHIGT (IHIG)
            Exit
!
!
         Case (0)
!
!  Low part is exactly what we want
!
            Exit
!
!
         Case (-5:-1)
!
!  Only few values too many in low part
!
            IRNGT (1) = ILOWT (1)
            Do ICRS = 2, NORD
               IWRK = ILOWT (ICRS)
               XWRK = XDONT (IWRK)
               Do IDCR = ICRS - 1, 1, - 1
                  If (XWRK < XDONT(IRNGT(IDCR))) Then
                     IRNGT (IDCR+1) = IRNGT (IDCR)
                  Else
                     Exit
                  End If
               End Do
               IRNGT (IDCR+1) = IWRK
            End Do
!
            XWRK1 = XDONT (IRNGT(NORD))
            Do ICRS = NORD + 1, JLOW
               If (XDONT(ILOWT (ICRS)) < XWRK1) Then
                  XWRK = XDONT (ILOWT (ICRS))
                  Do IDCR = NORD - 1, 1, - 1
                     If (XWRK >= XDONT(IRNGT(IDCR))) Exit
                     IRNGT (IDCR+1) = IRNGT (IDCR)
                  End Do
                  IRNGT (IDCR+1) = ILOWT (ICRS)
                  XWRK1 = XDONT (IRNGT(NORD))
               End If
            End Do
!
            Return
!
!
         Case (:-6)
!
! last case: too many values in low part
!
            IDEB = JDEB + 1
            IMIL = (JLOW+IDEB) / 2
            IFIN = JLOW
!
!  One chooses a pivot from 1st, last, and middle values
!
            If (XDONT(ILOWT(IMIL)) < XDONT(ILOWT(IDEB))) Then
               IWRK = ILOWT (IDEB)
               ILOWT (IDEB) = ILOWT (IMIL)
               ILOWT (IMIL) = IWRK
            End If
            If (XDONT(ILOWT(IMIL)) > XDONT(ILOWT(IFIN))) Then
               IWRK = ILOWT (IFIN)
               ILOWT (IFIN) = ILOWT (IMIL)
               ILOWT (IMIL) = IWRK
               If (XDONT(ILOWT(IMIL)) < XDONT(ILOWT(IDEB))) Then
                  IWRK = ILOWT (IDEB)
                  ILOWT (IDEB) = ILOWT (IMIL)
                  ILOWT (IMIL) = IWRK
               End If
            End If
            If (IFIN <= 3) Exit
!
            XPIV = XDONT (ILOWT(1)) + REAL(NORD)/REAL(JLOW+NORD) * &
                                      (XDONT(ILOWT(IFIN))-XDONT(ILOWT(1)))
            If (JDEB > 0) Then
               If (XPIV <= XPIV0) &
                   XPIV = XPIV0 + REAL(2*NORD-JDEB)/REAL (JLOW+NORD) * &
                                  (XDONT(ILOWT(IFIN))-XPIV0)
            Else
               IDEB = 1
            End If
!
!  One takes values > XPIV to IHIGT
!  However, we do not process the first values if we have been
!  through the case when we did not have enough low values
!
            JHIG = 0
            JLOW = JDEB
!
            If (XDONT(ILOWT(IFIN)) > XPIV) Then
               ICRS = JDEB
               Do
                 ICRS = ICRS + 1
                  If (XDONT(ILOWT(ICRS)) > XPIV) Then
                     JHIG = JHIG + 1
                     IHIGT (JHIG) = ILOWT (ICRS)
                     If (ICRS >= IFIN) Exit
                  Else
                     JLOW = JLOW + 1
                     ILOWT (JLOW) = ILOWT (ICRS)
                     If (JLOW >= NORD) Exit
                  End If
               End Do
!
               If (ICRS < IFIN) Then
                  Do
                     ICRS = ICRS + 1
                     If (XDONT(ILOWT(ICRS)) <= XPIV) Then
                        JLOW = JLOW + 1
                        ILOWT (JLOW) = ILOWT (ICRS)
                     Else
                        If (ICRS >= IFIN) Exit
                     End If
                  End Do
               End If
           Else
               Do ICRS = IDEB, IFIN
                  If (XDONT(ILOWT(ICRS)) > XPIV) Then
                     JHIG = JHIG + 1
                     IHIGT (JHIG) = ILOWT (ICRS)
                  Else
                     JLOW = JLOW + 1
                     ILOWT (JLOW) = ILOWT (ICRS)
                     If (JLOW >= NORD) Exit
                  End If
               End Do
!
               Do ICRS = ICRS + 1, IFIN
                  If (XDONT(ILOWT(ICRS)) <= XPIV) Then
                     JLOW = JLOW + 1
                     ILOWT (JLOW) = ILOWT (ICRS)
                  End If
               End Do
            End If
!
         End Select
!
      End Do
!
!  Now, we only need to complete ranking of the 1:NORD set
!  Assuming NORD is small, we use a simple insertion sort
!
      IRNGT (1) = ILOWT (1)
      Do ICRS = 2, NORD
         IWRK = ILOWT (ICRS)
         XWRK = XDONT (IWRK)
         Do IDCR = ICRS - 1, 1, - 1
            If (XWRK < XDONT(IRNGT(IDCR))) Then
               IRNGT (IDCR+1) = IRNGT (IDCR)
            Else
               Exit
            End If
         End Do
         IRNGT (IDCR+1) = IWRK
      End Do
     Return
!
!
End Subroutine D_rnkpar

Subroutine R_rnkpar (XDONT, IRNGT, NORD)
!  Ranks partially XDONT by IRNGT, up to order NORD
! __________________________________________________________
!  This routine uses a pivoting strategy such as the one of
!  finding the median based on the quicksort algorithm, but
!  we skew the pivot choice to try to bring it to NORD as
!  fast as possible. It uses 2 temporary arrays, where it
!  stores the indices of the values smaller than the pivot
!  (ILOWT), and the indices of values larger than the pivot
!  that we might still need later on (IHIGT). It iterates
!  until it can bring the number of values in ILOWT to
!  exactly NORD, and then uses an insertion sort to rank
!  this set, since it is supposedly small.
!  Michel Olagnon - Feb. 2000
! __________________________________________________________
! _________________________________________________________
      Real, Dimension (:), Intent (In) :: XDONT
      Integer, Dimension (:), Intent (Out) :: IRNGT
      Integer, Intent (In) :: NORD
! __________________________________________________________
      Real    :: XPIV, XPIV0, XWRK, XWRK1, XMIN, XMAX
!
      Integer, Dimension (SIZE(XDONT)) :: ILOWT, IHIGT
      Integer :: NDON, JHIG, JLOW, IHIG, IWRK, IWRK1, IWRK2, IWRK3
      Integer :: IDEB, JDEB, IMIL, IFIN, NWRK, ICRS, IDCR, ILOW
      Integer :: JLM2, JLM1, JHM2, JHM1
!
      NDON = SIZE (XDONT)
!
!    First loop is used to fill-in ILOWT, IHIGT at the same time
!
      If (NDON < 2) Then
         If (NORD >= 1) IRNGT (1) = 1
         Return
      End If
!
!  One chooses a pivot, best estimate possible to put fractile near
!  mid-point of the set of low values.
!
      If (XDONT(2) < XDONT(1)) Then
         ILOWT (1) = 2
         IHIGT (1) = 1
      Else
         ILOWT (1) = 1
         IHIGT (1) = 2
      End If
!
      If (NDON < 3) Then
         If (NORD >= 1) IRNGT (1) = ILOWT (1)
         If (NORD >= 2) IRNGT (2) = IHIGT (1)
         Return
      End If
!
      If (XDONT(3) <= XDONT(IHIGT(1))) Then
         IHIGT (2) = IHIGT (1)
         If (XDONT(3) < XDONT(ILOWT(1))) Then
            IHIGT (1) = ILOWT (1)
            ILOWT (1) = 3
         Else
            IHIGT (1) = 3
         End If
      Else
         IHIGT (2) = 3
      End If
!
      If (NDON < 4) Then
         If (NORD >= 1) IRNGT (1) = ILOWT (1)
         If (NORD >= 2) IRNGT (2) = IHIGT (1)
         If (NORD >= 3) IRNGT (3) = IHIGT (2)
         Return
      End If
!
      If (XDONT(NDON) <= XDONT(IHIGT(1))) Then
         IHIGT (3) = IHIGT (2)
         IHIGT (2) = IHIGT (1)
         If (XDONT(NDON) < XDONT(ILOWT(1))) Then
            IHIGT (1) = ILOWT (1)
            ILOWT (1) = NDON
         Else
            IHIGT (1) = NDON
         End If
      Else
         if (XDONT (NDON) < XDONT (IHIGT(2))) Then
            IHIGT (3) = IHIGT (2)
            IHIGT (2) = NDON
         else
            IHIGT (3) = NDON
         endif 
      End If
!
      If (NDON < 5) Then
         If (NORD >= 1) IRNGT (1) = ILOWT (1)
         If (NORD >= 2) IRNGT (2) = IHIGT (1)
         If (NORD >= 3) IRNGT (3) = IHIGT (2)
         If (NORD >= 4) IRNGT (4) = IHIGT (3)
         Return
      End If
!
      JDEB = 0
      IDEB = JDEB + 1
      JLOW = IDEB
      JHIG = 3
      XPIV = XDONT (ILOWT(IDEB)) + REAL(2*NORD)/REAL(NDON+NORD) * &
                                   (XDONT(IHIGT(3))-XDONT(ILOWT(IDEB)))
      If (XPIV >= XDONT(IHIGT(1))) Then
         XPIV = XDONT (ILOWT(IDEB)) + REAL(2*NORD)/REAL(NDON+NORD) * &
                                      (XDONT(IHIGT(2))-XDONT(ILOWT(IDEB)))
         If (XPIV >= XDONT(IHIGT(1))) &
             XPIV = XDONT (ILOWT(IDEB)) + REAL (2*NORD) / REAL (NDON+NORD) * &
                                          (XDONT(IHIGT(1))-XDONT(ILOWT(IDEB)))
      End If
      XPIV0 = XPIV
!
!  One puts values > pivot in the end and those <= pivot
!  at the beginning. This is split in 2 cases, so that
!  we can skip the loop test a number of times.
!  As we are also filling in the work arrays at the same time
!  we stop filling in the IHIGT array as soon as we have more
!  than enough values in ILOWT.
!
!
      If (XDONT(NDON) > XPIV) Then
         ICRS = 3
         Do
            ICRS = ICRS + 1
            If (XDONT(ICRS) > XPIV) Then
               If (ICRS >= NDON) Exit
               JHIG = JHIG + 1
               IHIGT (JHIG) = ICRS
            Else
               JLOW = JLOW + 1
               ILOWT (JLOW) = ICRS
               If (JLOW >= NORD) Exit
            End If
         End Do
!
!  One restricts further processing because it is no use
!  to store more high values
!
         If (ICRS < NDON-1) Then
            Do
               ICRS = ICRS + 1
               If (XDONT(ICRS) <= XPIV) Then
                  JLOW = JLOW + 1
                  ILOWT (JLOW) = ICRS
               Else If (ICRS >= NDON) Then
                  Exit
               End If
            End Do
         End If
!
!
      Else
!
!  Same as above, but this is not as easy to optimize, so the
!  DO-loop is kept
!
         Do ICRS = 4, NDON - 1
            If (XDONT(ICRS) > XPIV) Then
               JHIG = JHIG + 1
               IHIGT (JHIG) = ICRS
            Else
               JLOW = JLOW + 1
               ILOWT (JLOW) = ICRS
               If (JLOW >= NORD) Exit
            End If
         End Do
!
         If (ICRS < NDON-1) Then
            Do
               ICRS = ICRS + 1
               If (XDONT(ICRS) <= XPIV) Then
                  If (ICRS >= NDON) Exit
                  JLOW = JLOW + 1
                  ILOWT (JLOW) = ICRS
               End If
            End Do
         End If
      End If
!
      JLM2 = 0
      JLM1 = 0
      JHM2 = 0
      JHM1 = 0
      Do
         if (JLOW == NORD) Exit
         If (JLM2 == JLOW .And. JHM2 == JHIG) Then
!
!   We are oscillating. Perturbate by bringing JLOW closer by one
!   to NORD
!
           If (NORD > JLOW) Then
                XMIN = XDONT (IHIGT(1))
                IHIG = 1
                Do ICRS = 2, JHIG
                   If (XDONT(IHIGT(ICRS)) < XMIN) Then
                      XMIN = XDONT (IHIGT(ICRS))
                      IHIG = ICRS
                   End If
                End Do
!
                JLOW = JLOW + 1
                ILOWT (JLOW) = IHIGT (IHIG)
                IHIGT (IHIG) = IHIGT (JHIG)
                JHIG = JHIG - 1
             Else
                ILOW = ILOWT (JLOW)
                XMAX = XDONT (ILOW)
                Do ICRS = 1, JLOW
                   If (XDONT(ILOWT(ICRS)) > XMAX) Then
                      IWRK = ILOWT (ICRS)
                      XMAX = XDONT (IWRK)
                      ILOWT (ICRS) = ILOW
                      ILOW = IWRK
                   End If
                End Do
                JLOW = JLOW - 1
             End If
         End If
         JLM2 = JLM1
         JLM1 = JLOW
         JHM2 = JHM1
         JHM1 = JHIG
!
!   We try to bring the number of values in the low values set
!   closer to NORD.
!
        Select Case (NORD-JLOW)
         Case (2:)
!
!   Not enough values in low part, at least 2 are missing
!
            Select Case (JHIG)
!!!!!           CASE DEFAULT
!!!!!              write (*,*) "Assertion failed"
!!!!!              STOP
!
!   We make a special case when we have so few values in
!   the high values set that it is bad performance to choose a pivot
!   and apply the general algorithm.
!
            Case (2)
               If (XDONT(IHIGT(1)) <= XDONT(IHIGT(2))) Then
                  JLOW = JLOW + 1
                  ILOWT (JLOW) = IHIGT (1)
                  JLOW = JLOW + 1
                  ILOWT (JLOW) = IHIGT (2)
               Else
                  JLOW = JLOW + 1
                  ILOWT (JLOW) = IHIGT (2)
                  JLOW = JLOW + 1
                  ILOWT (JLOW) = IHIGT (1)
               End If
               Exit
!
            Case (3)
!
!
               IWRK1 = IHIGT (1)
               IWRK2 = IHIGT (2)
               IWRK3 = IHIGT (3)
               If (XDONT(IWRK2) < XDONT(IWRK1)) Then
                  IHIGT (1) = IWRK2
                  IHIGT (2) = IWRK1
                  IWRK2 = IWRK1
               End If
               If (XDONT(IWRK2) > XDONT(IWRK3)) Then
                  IHIGT (3) = IWRK2
                  IHIGT (2) = IWRK3
                  IWRK2 = IWRK3
                  If (XDONT(IWRK2) < XDONT(IHIGT(1))) Then
                     IHIGT (2) = IHIGT (1)
                     IHIGT (1) = IWRK2
                  End If
               End If
               JHIG = 0
               Do ICRS = JLOW + 1, NORD
                  JHIG = JHIG + 1
                  ILOWT (ICRS) = IHIGT (JHIG)
               End Do
               JLOW = NORD
               Exit
!
            Case (4:)
!
!
               XPIV0 = XPIV
               IFIN = JHIG
!
!  One chooses a pivot from the 2 first values and the last one.
!  This should ensure sufficient renewal between iterations to
!  avoid worst case behavior effects.
!
               IWRK1 = IHIGT (1)
               IWRK2 = IHIGT (2)
               IWRK3 = IHIGT (IFIN)
               If (XDONT(IWRK2) < XDONT(IWRK1)) Then
                  IHIGT (1) = IWRK2
                  IHIGT (2) = IWRK1
                  IWRK2 = IWRK1
               End If
               If (XDONT(IWRK2) > XDONT(IWRK3)) Then
                  IHIGT (IFIN) = IWRK2
                  IHIGT (2) = IWRK3
                  IWRK2 = IWRK3
                  If (XDONT(IWRK2) < XDONT(IHIGT(1))) Then
                     IHIGT (2) = IHIGT (1)
                     IHIGT (1) = IWRK2
                  End If
               End If
!
               JDEB = JLOW
               NWRK = NORD - JLOW
               IWRK1 = IHIGT (1)
               JLOW = JLOW + 1
               ILOWT (JLOW) = IWRK1
               XPIV = XDONT (IWRK1) + REAL (NWRK) / REAL (NORD+NWRK) * &
                                      (XDONT(IHIGT(IFIN))-XDONT(IWRK1))
!
!  One takes values <= pivot to ILOWT
!  Again, 2 parts, one where we take care of the remaining
!  high values because we might still need them, and the
!  other when we know that we will have more than enough
!  low values in the end.
!
               JHIG = 0
               Do ICRS = 2, IFIN
                  If (XDONT(IHIGT(ICRS)) <= XPIV) Then
                     JLOW = JLOW + 1
                     ILOWT (JLOW) = IHIGT (ICRS)
                     If (JLOW >= NORD) Exit
                  Else
                     JHIG = JHIG + 1
                     IHIGT (JHIG) = IHIGT (ICRS)
                  End If
               End Do
!
               Do ICRS = ICRS + 1, IFIN
                  If (XDONT(IHIGT(ICRS)) <= XPIV) Then
                     JLOW = JLOW + 1
                     ILOWT (JLOW) = IHIGT (ICRS)
                  End If
               End Do
           End Select
!
!
         Case (1)
!
!  Only 1 value is missing in low part
!
            XMIN = XDONT (IHIGT(1))
            IHIG = 1
            Do ICRS = 2, JHIG
               If (XDONT(IHIGT(ICRS)) < XMIN) Then
                  XMIN = XDONT (IHIGT(ICRS))
                  IHIG = ICRS
               End If
            End Do
!
            JLOW = JLOW + 1
            ILOWT (JLOW) = IHIGT (IHIG)
            Exit
!
!
         Case (0)
!
!  Low part is exactly what we want
!
            Exit
!
!
         Case (-5:-1)
!
!  Only few values too many in low part
!
            IRNGT (1) = ILOWT (1)
            Do ICRS = 2, NORD
               IWRK = ILOWT (ICRS)
               XWRK = XDONT (IWRK)
               Do IDCR = ICRS - 1, 1, - 1
                  If (XWRK < XDONT(IRNGT(IDCR))) Then
                     IRNGT (IDCR+1) = IRNGT (IDCR)
                  Else
                     Exit
                  End If
               End Do
               IRNGT (IDCR+1) = IWRK
            End Do
!
            XWRK1 = XDONT (IRNGT(NORD))
            Do ICRS = NORD + 1, JLOW
               If (XDONT(ILOWT (ICRS)) < XWRK1) Then
                  XWRK = XDONT (ILOWT (ICRS))
                  Do IDCR = NORD - 1, 1, - 1
                     If (XWRK >= XDONT(IRNGT(IDCR))) Exit
                     IRNGT (IDCR+1) = IRNGT (IDCR)
                  End Do
                  IRNGT (IDCR+1) = ILOWT (ICRS)
                  XWRK1 = XDONT (IRNGT(NORD))
               End If
            End Do
!
            Return
!
!
         Case (:-6)
!
! last case: too many values in low part
!
            IDEB = JDEB + 1
            IMIL = (JLOW+IDEB) / 2
            IFIN = JLOW
!
!  One chooses a pivot from 1st, last, and middle values
!
            If (XDONT(ILOWT(IMIL)) < XDONT(ILOWT(IDEB))) Then
               IWRK = ILOWT (IDEB)
               ILOWT (IDEB) = ILOWT (IMIL)
               ILOWT (IMIL) = IWRK
            End If
            If (XDONT(ILOWT(IMIL)) > XDONT(ILOWT(IFIN))) Then
               IWRK = ILOWT (IFIN)
               ILOWT (IFIN) = ILOWT (IMIL)
               ILOWT (IMIL) = IWRK
               If (XDONT(ILOWT(IMIL)) < XDONT(ILOWT(IDEB))) Then
                  IWRK = ILOWT (IDEB)
                  ILOWT (IDEB) = ILOWT (IMIL)
                  ILOWT (IMIL) = IWRK
               End If
            End If
            If (IFIN <= 3) Exit
!
            XPIV = XDONT (ILOWT(1)) + REAL(NORD)/REAL(JLOW+NORD) * &
                                      (XDONT(ILOWT(IFIN))-XDONT(ILOWT(1)))
            If (JDEB > 0) Then
               If (XPIV <= XPIV0) &
                   XPIV = XPIV0 + REAL(2*NORD-JDEB)/REAL (JLOW+NORD) * &
                                  (XDONT(ILOWT(IFIN))-XPIV0)
            Else
               IDEB = 1
            End If
!
!  One takes values > XPIV to IHIGT
!  However, we do not process the first values if we have been
!  through the case when we did not have enough low values
!
            JHIG = 0
            JLOW = JDEB
!
            If (XDONT(ILOWT(IFIN)) > XPIV) Then
               ICRS = JDEB
               Do
                 ICRS = ICRS + 1
                  If (XDONT(ILOWT(ICRS)) > XPIV) Then
                     JHIG = JHIG + 1
                     IHIGT (JHIG) = ILOWT (ICRS)
                     If (ICRS >= IFIN) Exit
                  Else
                     JLOW = JLOW + 1
                     ILOWT (JLOW) = ILOWT (ICRS)
                     If (JLOW >= NORD) Exit
                  End If
               End Do
!
               If (ICRS < IFIN) Then
                  Do
                     ICRS = ICRS + 1
                     If (XDONT(ILOWT(ICRS)) <= XPIV) Then
                        JLOW = JLOW + 1
                        ILOWT (JLOW) = ILOWT (ICRS)
                     Else
                        If (ICRS >= IFIN) Exit
                     End If
                  End Do
               End If
           Else
               Do ICRS = IDEB, IFIN
                  If (XDONT(ILOWT(ICRS)) > XPIV) Then
                     JHIG = JHIG + 1
                     IHIGT (JHIG) = ILOWT (ICRS)
                  Else
                     JLOW = JLOW + 1
                     ILOWT (JLOW) = ILOWT (ICRS)
                     If (JLOW >= NORD) Exit
                  End If
               End Do
!
               Do ICRS = ICRS + 1, IFIN
                  If (XDONT(ILOWT(ICRS)) <= XPIV) Then
                     JLOW = JLOW + 1
                     ILOWT (JLOW) = ILOWT (ICRS)
                  End If
               End Do
            End If
!
         End Select
!
      End Do
!
!  Now, we only need to complete ranking of the 1:NORD set
!  Assuming NORD is small, we use a simple insertion sort
!
      IRNGT (1) = ILOWT (1)
      Do ICRS = 2, NORD
         IWRK = ILOWT (ICRS)
         XWRK = XDONT (IWRK)
         Do IDCR = ICRS - 1, 1, - 1
            If (XWRK < XDONT(IRNGT(IDCR))) Then
               IRNGT (IDCR+1) = IRNGT (IDCR)
            Else
               Exit
            End If
         End Do
         IRNGT (IDCR+1) = IWRK
      End Do
     Return
!
!
End Subroutine R_rnkpar
Subroutine I_rnkpar (XDONT, IRNGT, NORD)
!  Ranks partially XDONT by IRNGT, up to order NORD
! __________________________________________________________
!  This routine uses a pivoting strategy such as the one of
!  finding the median based on the quicksort algorithm, but
!  we skew the pivot choice to try to bring it to NORD as
!  fast as possible. It uses 2 temporary arrays, where it
!  stores the indices of the values smaller than the pivot
!  (ILOWT), and the indices of values larger than the pivot
!  that we might still need later on (IHIGT). It iterates
!  until it can bring the number of values in ILOWT to
!  exactly NORD, and then uses an insertion sort to rank
!  this set, since it is supposedly small.
!  Michel Olagnon - Feb. 2000
! __________________________________________________________
! __________________________________________________________
      Integer, Dimension (:), Intent (In)  :: XDONT
      Integer, Dimension (:), Intent (Out) :: IRNGT
      Integer, Intent (In) :: NORD
! __________________________________________________________
      Integer :: XPIV, XPIV0, XWRK, XWRK1, XMIN, XMAX
!
      Integer, Dimension (SIZE(XDONT)) :: ILOWT, IHIGT
      Integer :: NDON, JHIG, JLOW, IHIG, IWRK, IWRK1, IWRK2, IWRK3
      Integer :: IDEB, JDEB, IMIL, IFIN, NWRK, ICRS, IDCR, ILOW
      Integer :: JLM2, JLM1, JHM2, JHM1
!
      NDON = SIZE (XDONT)
!
!    First loop is used to fill-in ILOWT, IHIGT at the same time
!
      If (NDON < 2) Then
         If (NORD >= 1) IRNGT (1) = 1
         Return
      End If
!
!  One chooses a pivot, best estimate possible to put fractile near
!  mid-point of the set of low values.
!
      If (XDONT(2) < XDONT(1)) Then
         ILOWT (1) = 2
         IHIGT (1) = 1
      Else
         ILOWT (1) = 1
         IHIGT (1) = 2
      End If
!
      If (NDON < 3) Then
         If (NORD >= 1) IRNGT (1) = ILOWT (1)
         If (NORD >= 2) IRNGT (2) = IHIGT (1)
         Return
      End If
!
      If (XDONT(3) <= XDONT(IHIGT(1))) Then
         IHIGT (2) = IHIGT (1)
         If (XDONT(3) < XDONT(ILOWT(1))) Then
            IHIGT (1) = ILOWT (1)
            ILOWT (1) = 3
         Else
            IHIGT (1) = 3
         End If
      Else
         IHIGT (2) = 3
      End If
!
      If (NDON < 4) Then
         If (NORD >= 1) IRNGT (1) = ILOWT (1)
         If (NORD >= 2) IRNGT (2) = IHIGT (1)
         If (NORD >= 3) IRNGT (3) = IHIGT (2)
         Return
      End If
!
      If (XDONT(NDON) <= XDONT(IHIGT(1))) Then
         IHIGT (3) = IHIGT (2)
         IHIGT (2) = IHIGT (1)
         If (XDONT(NDON) < XDONT(ILOWT(1))) Then
            IHIGT (1) = ILOWT (1)
            ILOWT (1) = NDON
         Else
            IHIGT (1) = NDON
         End If
      Else
         if (XDONT (NDON) < XDONT (IHIGT(2))) Then
            IHIGT (3) = IHIGT (2)
            IHIGT (2) = NDON
         else
            IHIGT (3) = NDON
         endif 
      End If
!
      If (NDON < 5) Then
         If (NORD >= 1) IRNGT (1) = ILOWT (1)
         If (NORD >= 2) IRNGT (2) = IHIGT (1)
         If (NORD >= 3) IRNGT (3) = IHIGT (2)
         If (NORD >= 4) IRNGT (4) = IHIGT (3)
         Return
      End If
!
      JDEB = 0
      IDEB = JDEB + 1
      JLOW = IDEB
      JHIG = 3
      XPIV = XDONT (ILOWT(IDEB)) + REAL(2*NORD)/REAL(NDON+NORD) * &
                                   (XDONT(IHIGT(3))-XDONT(ILOWT(IDEB)))
      If (XPIV >= XDONT(IHIGT(1))) Then
         XPIV = XDONT (ILOWT(IDEB)) + REAL(2*NORD)/REAL(NDON+NORD) * &
                                      (XDONT(IHIGT(2))-XDONT(ILOWT(IDEB)))
         If (XPIV >= XDONT(IHIGT(1))) &
             XPIV = XDONT (ILOWT(IDEB)) + REAL (2*NORD) / REAL (NDON+NORD) * &
                                          (XDONT(IHIGT(1))-XDONT(ILOWT(IDEB)))
      End If
      XPIV0 = XPIV
!
!  One puts values > pivot in the end and those <= pivot
!  at the beginning. This is split in 2 cases, so that
!  we can skip the loop test a number of times.
!  As we are also filling in the work arrays at the same time
!  we stop filling in the IHIGT array as soon as we have more
!  than enough values in ILOWT.
!
!
      If (XDONT(NDON) > XPIV) Then
         ICRS = 3
         Do
            ICRS = ICRS + 1
            If (XDONT(ICRS) > XPIV) Then
               If (ICRS >= NDON) Exit
               JHIG = JHIG + 1
               IHIGT (JHIG) = ICRS
            Else
               JLOW = JLOW + 1
               ILOWT (JLOW) = ICRS
               If (JLOW >= NORD) Exit
            End If
         End Do
!
!  One restricts further processing because it is no use
!  to store more high values
!
         If (ICRS < NDON-1) Then
            Do
               ICRS = ICRS + 1
               If (XDONT(ICRS) <= XPIV) Then
                  JLOW = JLOW + 1
                  ILOWT (JLOW) = ICRS
               Else If (ICRS >= NDON) Then
                  Exit
               End If
            End Do
         End If
!
!
      Else
!
!  Same as above, but this is not as easy to optimize, so the
!  DO-loop is kept
!
         Do ICRS = 4, NDON - 1
            If (XDONT(ICRS) > XPIV) Then
               JHIG = JHIG + 1
               IHIGT (JHIG) = ICRS
            Else
               JLOW = JLOW + 1
               ILOWT (JLOW) = ICRS
               If (JLOW >= NORD) Exit
            End If
         End Do
!
         If (ICRS < NDON-1) Then
            Do
               ICRS = ICRS + 1
               If (XDONT(ICRS) <= XPIV) Then
                  If (ICRS >= NDON) Exit
                  JLOW = JLOW + 1
                  ILOWT (JLOW) = ICRS
               End If
            End Do
         End If
      End If
!
      JLM2 = 0
      JLM1 = 0
      JHM2 = 0
      JHM1 = 0
      Do
         if (JLOW == NORD) Exit
         If (JLM2 == JLOW .And. JHM2 == JHIG) Then
!
!   We are oscillating. Perturbate by bringing JLOW closer by one
!   to NORD
!
           If (NORD > JLOW) Then
                XMIN = XDONT (IHIGT(1))
                IHIG = 1
                Do ICRS = 2, JHIG
                   If (XDONT(IHIGT(ICRS)) < XMIN) Then
                      XMIN = XDONT (IHIGT(ICRS))
                      IHIG = ICRS
                   End If
                End Do
!
                JLOW = JLOW + 1
                ILOWT (JLOW) = IHIGT (IHIG)
                IHIGT (IHIG) = IHIGT (JHIG)
                JHIG = JHIG - 1
             Else
                ILOW = ILOWT (JLOW)
                XMAX = XDONT (ILOW)
                Do ICRS = 1, JLOW
                   If (XDONT(ILOWT(ICRS)) > XMAX) Then
                      IWRK = ILOWT (ICRS)
                      XMAX = XDONT (IWRK)
                      ILOWT (ICRS) = ILOW
                      ILOW = IWRK
                   End If
                End Do
                JLOW = JLOW - 1
             End If
         End If
         JLM2 = JLM1
         JLM1 = JLOW
         JHM2 = JHM1
         JHM1 = JHIG
!
!   We try to bring the number of values in the low values set
!   closer to NORD.
!
        Select Case (NORD-JLOW)
         Case (2:)
!
!   Not enough values in low part, at least 2 are missing
!
            Select Case (JHIG)
!!!!!           CASE DEFAULT
!!!!!              write (*,*) "Assertion failed"
!!!!!              STOP
!
!   We make a special case when we have so few values in
!   the high values set that it is bad performance to choose a pivot
!   and apply the general algorithm.
!
            Case (2)
               If (XDONT(IHIGT(1)) <= XDONT(IHIGT(2))) Then
                  JLOW = JLOW + 1
                  ILOWT (JLOW) = IHIGT (1)
                  JLOW = JLOW + 1
                  ILOWT (JLOW) = IHIGT (2)
               Else
                  JLOW = JLOW + 1
                  ILOWT (JLOW) = IHIGT (2)
                  JLOW = JLOW + 1
                  ILOWT (JLOW) = IHIGT (1)
               End If
               Exit
!
            Case (3)
!
!
               IWRK1 = IHIGT (1)
               IWRK2 = IHIGT (2)
               IWRK3 = IHIGT (3)
               If (XDONT(IWRK2) < XDONT(IWRK1)) Then
                  IHIGT (1) = IWRK2
                  IHIGT (2) = IWRK1
                  IWRK2 = IWRK1
               End If
               If (XDONT(IWRK2) > XDONT(IWRK3)) Then
                  IHIGT (3) = IWRK2
                  IHIGT (2) = IWRK3
                  IWRK2 = IWRK3
                  If (XDONT(IWRK2) < XDONT(IHIGT(1))) Then
                     IHIGT (2) = IHIGT (1)
                     IHIGT (1) = IWRK2
                  End If
               End If
               JHIG = 0
               Do ICRS = JLOW + 1, NORD
                  JHIG = JHIG + 1
                  ILOWT (ICRS) = IHIGT (JHIG)
               End Do
               JLOW = NORD
               Exit
!
            Case (4:)
!
!
               XPIV0 = XPIV
               IFIN = JHIG
!
!  One chooses a pivot from the 2 first values and the last one.
!  This should ensure sufficient renewal between iterations to
!  avoid worst case behavior effects.
!
               IWRK1 = IHIGT (1)
               IWRK2 = IHIGT (2)
               IWRK3 = IHIGT (IFIN)
               If (XDONT(IWRK2) < XDONT(IWRK1)) Then
                  IHIGT (1) = IWRK2
                  IHIGT (2) = IWRK1
                  IWRK2 = IWRK1
               End If
               If (XDONT(IWRK2) > XDONT(IWRK3)) Then
                  IHIGT (IFIN) = IWRK2
                  IHIGT (2) = IWRK3
                  IWRK2 = IWRK3
                  If (XDONT(IWRK2) < XDONT(IHIGT(1))) Then
                     IHIGT (2) = IHIGT (1)
                     IHIGT (1) = IWRK2
                  End If
               End If
!
               JDEB = JLOW
               NWRK = NORD - JLOW
               IWRK1 = IHIGT (1)
               JLOW = JLOW + 1
               ILOWT (JLOW) = IWRK1
               XPIV = XDONT (IWRK1) + REAL (NWRK) / REAL (NORD+NWRK) * &
                                      (XDONT(IHIGT(IFIN))-XDONT(IWRK1))
!
!  One takes values <= pivot to ILOWT
!  Again, 2 parts, one where we take care of the remaining
!  high values because we might still need them, and the
!  other when we know that we will have more than enough
!  low values in the end.
!
               JHIG = 0
               Do ICRS = 2, IFIN
                  If (XDONT(IHIGT(ICRS)) <= XPIV) Then
                     JLOW = JLOW + 1
                     ILOWT (JLOW) = IHIGT (ICRS)
                     If (JLOW >= NORD) Exit
                  Else
                     JHIG = JHIG + 1
                     IHIGT (JHIG) = IHIGT (ICRS)
                  End If
               End Do
!
               Do ICRS = ICRS + 1, IFIN
                  If (XDONT(IHIGT(ICRS)) <= XPIV) Then
                     JLOW = JLOW + 1
                     ILOWT (JLOW) = IHIGT (ICRS)
                  End If
               End Do
           End Select
!
!
         Case (1)
!
!  Only 1 value is missing in low part
!
            XMIN = XDONT (IHIGT(1))
            IHIG = 1
            Do ICRS = 2, JHIG
               If (XDONT(IHIGT(ICRS)) < XMIN) Then
                  XMIN = XDONT (IHIGT(ICRS))
                  IHIG = ICRS
               End If
            End Do
!
            JLOW = JLOW + 1
            ILOWT (JLOW) = IHIGT (IHIG)
            Exit
!
!
         Case (0)
!
!  Low part is exactly what we want
!
            Exit
!
!
         Case (-5:-1)
!
!  Only few values too many in low part
!
            IRNGT (1) = ILOWT (1)
            Do ICRS = 2, NORD
               IWRK = ILOWT (ICRS)
               XWRK = XDONT (IWRK)
               Do IDCR = ICRS - 1, 1, - 1
                  If (XWRK < XDONT(IRNGT(IDCR))) Then
                     IRNGT (IDCR+1) = IRNGT (IDCR)
                  Else
                     Exit
                  End If
               End Do
               IRNGT (IDCR+1) = IWRK
            End Do
!
            XWRK1 = XDONT (IRNGT(NORD))
            Do ICRS = NORD + 1, JLOW
               If (XDONT(ILOWT (ICRS)) < XWRK1) Then
                  XWRK = XDONT (ILOWT (ICRS))
                  Do IDCR = NORD - 1, 1, - 1
                     If (XWRK >= XDONT(IRNGT(IDCR))) Exit
                     IRNGT (IDCR+1) = IRNGT (IDCR)
                  End Do
                  IRNGT (IDCR+1) = ILOWT (ICRS)
                  XWRK1 = XDONT (IRNGT(NORD))
               End If
            End Do
!
            Return
!
!
         Case (:-6)
!
! last case: too many values in low part
!
            IDEB = JDEB + 1
            IMIL = (JLOW+IDEB) / 2
            IFIN = JLOW
!
!  One chooses a pivot from 1st, last, and middle values
!
            If (XDONT(ILOWT(IMIL)) < XDONT(ILOWT(IDEB))) Then
               IWRK = ILOWT (IDEB)
               ILOWT (IDEB) = ILOWT (IMIL)
               ILOWT (IMIL) = IWRK
            End If
            If (XDONT(ILOWT(IMIL)) > XDONT(ILOWT(IFIN))) Then
               IWRK = ILOWT (IFIN)
               ILOWT (IFIN) = ILOWT (IMIL)
               ILOWT (IMIL) = IWRK
               If (XDONT(ILOWT(IMIL)) < XDONT(ILOWT(IDEB))) Then
                  IWRK = ILOWT (IDEB)
                  ILOWT (IDEB) = ILOWT (IMIL)
                  ILOWT (IMIL) = IWRK
               End If
            End If
            If (IFIN <= 3) Exit
!
            XPIV = XDONT (ILOWT(1)) + REAL(NORD)/REAL(JLOW+NORD) * &
                                      (XDONT(ILOWT(IFIN))-XDONT(ILOWT(1)))
            If (JDEB > 0) Then
               If (XPIV <= XPIV0) &
                   XPIV = XPIV0 + REAL(2*NORD-JDEB)/REAL (JLOW+NORD) * &
                                  (XDONT(ILOWT(IFIN))-XPIV0)
            Else
               IDEB = 1
            End If
!
!  One takes values > XPIV to IHIGT
!  However, we do not process the first values if we have been
!  through the case when we did not have enough low values
!
            JHIG = 0
            JLOW = JDEB
!
            If (XDONT(ILOWT(IFIN)) > XPIV) Then
               ICRS = JDEB
               Do
                 ICRS = ICRS + 1
                  If (XDONT(ILOWT(ICRS)) > XPIV) Then
                     JHIG = JHIG + 1
                     IHIGT (JHIG) = ILOWT (ICRS)
                     If (ICRS >= IFIN) Exit
                  Else
                     JLOW = JLOW + 1
                     ILOWT (JLOW) = ILOWT (ICRS)
                     If (JLOW >= NORD) Exit
                  End If
               End Do
!
               If (ICRS < IFIN) Then
                  Do
                     ICRS = ICRS + 1
                     If (XDONT(ILOWT(ICRS)) <= XPIV) Then
                        JLOW = JLOW + 1
                        ILOWT (JLOW) = ILOWT (ICRS)
                     Else
                        If (ICRS >= IFIN) Exit
                     End If
                  End Do
               End If
           Else
               Do ICRS = IDEB, IFIN
                  If (XDONT(ILOWT(ICRS)) > XPIV) Then
                     JHIG = JHIG + 1
                     IHIGT (JHIG) = ILOWT (ICRS)
                  Else
                     JLOW = JLOW + 1
                     ILOWT (JLOW) = ILOWT (ICRS)
                     If (JLOW >= NORD) Exit
                  End If
               End Do
!
               Do ICRS = ICRS + 1, IFIN
                  If (XDONT(ILOWT(ICRS)) <= XPIV) Then
                     JLOW = JLOW + 1
                     ILOWT (JLOW) = ILOWT (ICRS)
                  End If
               End Do
            End If
!
         End Select
!
      End Do
!
!  Now, we only need to complete ranking of the 1:NORD set
!  Assuming NORD is small, we use a simple insertion sort
!
      IRNGT (1) = ILOWT (1)
      Do ICRS = 2, NORD
         IWRK = ILOWT (ICRS)
         XWRK = XDONT (IWRK)
         Do IDCR = ICRS - 1, 1, - 1
            If (XWRK < XDONT(IRNGT(IDCR))) Then
               IRNGT (IDCR+1) = IRNGT (IDCR)
            Else
               Exit
            End If
         End Do
         IRNGT (IDCR+1) = IWRK
      End Do
     Return
!
!
End Subroutine I_rnkpar
end module m_rnkpar
Module m_unipar
Integer, Parameter :: kdp = selected_real_kind(15)
public :: unipar
private :: kdp
private :: R_unipar, I_unipar, D_unipar
interface unipar
  module procedure d_unipar, r_unipar, i_unipar
end interface unipar
contains

Subroutine D_unipar (XDONT, IRNGT, NORD)
!  Ranks partially XDONT by IRNGT, up to order NORD at most,
!  removing duplicate entries
! __________________________________________________________
!  This routine uses a pivoting strategy such as the one of
!  finding the median based on the quicksort algorithm, but
!  we skew the pivot choice to try to bring it to NORD as
!  quickly as possible. It uses 2 temporary arrays, where it
!  stores the indices of the values smaller than the pivot
!  (ILOWT), and the indices of values larger than the pivot
!  that we might still need later on (IHIGT). It iterates
!  until it can bring the number of values in ILOWT to
!  exactly NORD, and then uses an insertion sort to rank
!  this set, since it is supposedly small. At all times, the
!  NORD first values in ILOWT correspond to distinct values
!  of the input array.
!  Michel Olagnon - Feb. 2000
! __________________________________________________________
! __________________________________________________________
      Real (kind=kdp), Dimension (:), Intent (In) :: XDONT
      Integer, Dimension (:), Intent (Out) :: IRNGT
      Integer, Intent (InOut) :: NORD
! __________________________________________________________
      Real (kind=kdp) :: XPIV, XWRK, XWRK1, XMIN, XMAX, XPIV0
!
      Integer, Dimension (SIZE(XDONT)) :: ILOWT, IHIGT
      Integer :: NDON, JHIG, JLOW, IHIG, IWRK, IWRK1, IWRK2, IWRK3
      Integer :: IDEB, JDEB, IMIL, IFIN, NWRK, ICRS, IDCR, ILOW
      Integer :: JLM2, JLM1, JHM2, JHM1
!
      NDON = SIZE (XDONT)
!
!    First loop is used to fill-in ILOWT, IHIGT at the same time
!
      If (NDON < 2) Then
         If (NORD >= 1) Then
            NORD = 1
            IRNGT (1) = 1
         End If
         Return
      End If
!
!  One chooses a pivot, best estimate possible to put fractile near
!  mid-point of the set of low values.
!
     Do ICRS = 2, NDON
        If (XDONT(ICRS) == XDONT(1)) Then
          Cycle
        Else If (XDONT(ICRS) < XDONT(1)) Then
           ILOWT (1) = ICRS
           IHIGT (1) = 1
        Else
           ILOWT (1) = 1
           IHIGT (1) = ICRS
        End If
        Exit
     End Do
!
      If (NDON <= ICRS) Then
         NORD = Min (NORD, 2)
         If (NORD >= 1) IRNGT (1) = ILOWT (1)
         If (NORD >= 2) IRNGT (2) = IHIGT (1)
         Return
      End If
!
      ICRS = ICRS + 1
      JHIG = 1
      If (XDONT(ICRS) < XDONT(IHIGT(1))) Then
         If (XDONT(ICRS) < XDONT(ILOWT(1))) Then
            JHIG = JHIG + 1
            IHIGT (JHIG) = IHIGT (1)
            IHIGT (1) = ILOWT (1)
            ILOWT (1) = ICRS
         Else If (XDONT(ICRS) > XDONT(ILOWT(1))) Then
            JHIG = JHIG + 1
            IHIGT (JHIG) = IHIGT (1)
            IHIGT (1) = ICRS
         End If
      ElseIf (XDONT(ICRS) > XDONT(IHIGT(1))) Then
         JHIG = JHIG + 1
         IHIGT (JHIG) = ICRS
      End If
!
      If (NDON <= ICRS) Then
         NORD = Min (NORD, JHIG+1)
         If (NORD >= 1) IRNGT (1) = ILOWT (1)
         If (NORD >= 2) IRNGT (2) = IHIGT (1)
         If (NORD >= 3) IRNGT (3) = IHIGT (2)
         Return
      End If
!
      If (XDONT(NDON) < XDONT(IHIGT(1))) Then
         If (XDONT(NDON) < XDONT(ILOWT(1))) Then
            Do IDCR = JHIG, 1, -1
              IHIGT (IDCR+1) = IHIGT (IDCR)
            End Do
            IHIGT (1) = ILOWT (1)
            ILOWT (1) = NDON
            JHIG = JHIG + 1
         ElseIf (XDONT(NDON) > XDONT(ILOWT(1))) Then
            Do IDCR = JHIG, 1, -1
              IHIGT (IDCR+1) = IHIGT (IDCR)
            End Do
            IHIGT (1) = NDON
            JHIG = JHIG + 1
         End If
      ElseIf (XDONT(NDON) > XDONT(IHIGT(1))) Then
         JHIG = JHIG + 1
         IHIGT (JHIG) = NDON
      End If
!
      If (NDON <= ICRS+1) Then
         NORD = Min (NORD, JHIG+1)
         If (NORD >= 1) IRNGT (1) = ILOWT (1)
         If (NORD >= 2) IRNGT (2) = IHIGT (1)
         If (NORD >= 3) IRNGT (3) = IHIGT (2)
         If (NORD >= 4) IRNGT (4) = IHIGT (3)
         Return
      End If
!
      JDEB = 0
      IDEB = JDEB + 1
      JLOW = IDEB
      XPIV = XDONT (ILOWT(IDEB)) + REAL(2*NORD)/REAL(NDON+NORD) * &
                                   (XDONT(IHIGT(3))-XDONT(ILOWT(IDEB)))
      If (XPIV >= XDONT(IHIGT(1))) Then
         XPIV = XDONT (ILOWT(IDEB)) + REAL(2*NORD)/REAL(NDON+NORD) * &
                                      (XDONT(IHIGT(2))-XDONT(ILOWT(IDEB)))
         If (XPIV >= XDONT(IHIGT(1))) &
             XPIV = XDONT (ILOWT(IDEB)) + REAL (2*NORD) / REAL (NDON+NORD) * &
                                          (XDONT(IHIGT(1))-XDONT(ILOWT(IDEB)))
      End If
      XPIV0 = XPIV
!
!  One puts values > pivot in the end and those <= pivot
!  at the beginning. This is split in 2 cases, so that
!  we can skip the loop test a number of times.
!  As we are also filling in the work arrays at the same time
!  we stop filling in the IHIGT array as soon as we have more
!  than enough values in ILOWT, i.e. one more than
!  strictly necessary so as to be able to come out of the
!  case where JLOWT would be NORD distinct values followed
!  by values that are exclusively duplicates of these.
!
!
      If (XDONT(NDON) > XPIV) Then
         lowloop1: Do
            ICRS = ICRS + 1
            If (XDONT(ICRS) > XPIV) Then
               If (ICRS >= NDON) Exit
               JHIG = JHIG + 1
               IHIGT (JHIG) = ICRS
            Else
               Do ILOW = 1, JLOW
                 If (XDONT(ICRS) == XDONT(ILOWT(ILOW))) Cycle lowloop1
               End Do
               JLOW = JLOW + 1
               ILOWT (JLOW) = ICRS
               If (JLOW >= NORD) Exit
            End If
         End Do lowloop1
!
!  One restricts further processing because it is no use
!  to store more high values
!
         If (ICRS < NDON-1) Then
            Do
               ICRS = ICRS + 1
               If (XDONT(ICRS) <= XPIV) Then
                  JLOW = JLOW + 1
                  ILOWT (JLOW) = ICRS
               Else If (ICRS >= NDON) Then
                  Exit
               End If
            End Do
         End If
!
!
      Else
!
!  Same as above, but this is not as easy to optimize, so the
!  DO-loop is kept
!
         lowloop2: Do ICRS = ICRS + 1, NDON - 1
            If (XDONT(ICRS) > XPIV) Then
               JHIG = JHIG + 1
               IHIGT (JHIG) = ICRS
            Else
               Do ILOW = 1, JLOW
                 If (XDONT(ICRS) == XDONT (ILOWT(ILOW))) Cycle lowloop2
               End Do
               JLOW = JLOW + 1
               ILOWT (JLOW) = ICRS
               If (JLOW >= NORD) Exit
            End If
         End Do lowloop2
!
         If (ICRS < NDON-1) Then
            Do
               ICRS = ICRS + 1
               If (XDONT(ICRS) <= XPIV) Then
                  If (ICRS >= NDON) Exit
                  JLOW = JLOW + 1
                  ILOWT (JLOW) = ICRS
               End If
            End Do
         End If
      End If
!
      JLM2 = 0
      JLM1 = 0
      JHM2 = 0
      JHM1 = 0
      Do
         if (JLOW == NORD) Exit
         If (JLM2 == JLOW .And. JHM2 == JHIG) Then
!
!   We are oscillating. Perturbate by bringing JLOW closer by one
!   to NORD
!
           If (NORD > JLOW) Then
                XMIN = XDONT (IHIGT(1))
                IHIG = 1
                Do ICRS = 2, JHIG
                   If (XDONT(IHIGT(ICRS)) < XMIN) Then
                      XMIN = XDONT (IHIGT(ICRS))
                      IHIG = ICRS
                   End If
                End Do
!
                JLOW = JLOW + 1
                ILOWT (JLOW) = IHIGT (IHIG)
                IHIG = 0
                Do ICRS = 1, JHIG
                   If (XDONT(IHIGT (ICRS)) /= XMIN) then
                      IHIG = IHIG + 1
                      IHIGT (IHIG ) = IHIGT (ICRS)
                   End If
                End Do
                JHIG = IHIG
             Else
                ILOW = ILOWT (JLOW)
                XMAX = XDONT (ILOW)
                Do ICRS = 1, JLOW
                   If (XDONT(ILOWT(ICRS)) > XMAX) Then
                      IWRK = ILOWT (ICRS)
                      XMAX = XDONT (IWRK)
                      ILOWT (ICRS) = ILOW
                      ILOW = IWRK
                   End If
                End Do
                JLOW = JLOW - 1
             End If
         End If
         JLM2 = JLM1
         JLM1 = JLOW
         JHM2 = JHM1
         JHM1 = JHIG
!
!   We try to bring the number of values in the low values set
!   closer to NORD. In order to make better pivot choices, we
!   decrease NORD if we already know that we don't have that
!   many distinct values as a whole.
!
         IF (JLOW+JHIG < NORD) NORD = JLOW+JHIG
         Select Case (NORD-JLOW)
! ______________________________
         Case (2:)
!
!   Not enough values in low part, at least 2 are missing
!
            Select Case (JHIG)
!
!   Not enough values in high part either (too many duplicates)
!
            Case (0)
               NORD = JLOW
!
            Case (1)
               JLOW = JLOW + 1
               ILOWT (JLOW) = IHIGT (1)
               NORD = JLOW
!
!   We make a special case when we have so few values in
!   the high values set that it is bad performance to choose a pivot
!   and apply the general algorithm.
!
            Case (2)
               If (XDONT(IHIGT(1)) <= XDONT(IHIGT(2))) Then
                  JLOW = JLOW + 1
                  ILOWT (JLOW) = IHIGT (1)
                  JLOW = JLOW + 1
                  ILOWT (JLOW) = IHIGT (2)
               ElseIf (XDONT(IHIGT(1)) == XDONT(IHIGT(2))) Then
                  JLOW = JLOW + 1
                  ILOWT (JLOW) = IHIGT (1)
                  NORD = JLOW
               Else
                  JLOW = JLOW + 1
                  ILOWT (JLOW) = IHIGT (2)
                  JLOW = JLOW + 1
                  ILOWT (JLOW) = IHIGT (1)
               End If
               Exit
!
            Case (3)
!
!
               IWRK1 = IHIGT (1)
               IWRK2 = IHIGT (2)
               IWRK3 = IHIGT (3)
               If (XDONT(IWRK2) < XDONT(IWRK1)) Then
                  IHIGT (1) = IWRK2
                  IHIGT (2) = IWRK1
                  IWRK2 = IWRK1
               End If
               If (XDONT(IWRK2) > XDONT(IWRK3)) Then
                  IHIGT (3) = IWRK2
                  IHIGT (2) = IWRK3
                  IWRK2 = IWRK3
                  If (XDONT(IWRK2) < XDONT(IHIGT(1))) Then
                     IHIGT (2) = IHIGT (1)
                     IHIGT (1) = IWRK2
                  End If
               End If
               JHIG = 1
               JLOW = JLOW + 1
               ILOWT (JLOW) = IHIGT (1)
               JHIG = JHIG + 1
               IF (XDONT(IHIGT(JHIG)) /= XDONT(ILOWT(JLOW))) Then
                 JLOW = JLOW + 1
                 ILOWT (JLOW) = IHIGT (JHIG)
               End If
               JHIG = JHIG + 1
               IF (XDONT(IHIGT(JHIG)) /= XDONT(ILOWT(JLOW))) Then
                 JLOW = JLOW + 1
                 ILOWT (JLOW) = IHIGT (JHIG)
               End If
               NORD = Min (JLOW, NORD)
               Exit
!
            Case (4:)
!
!
               XPIV0 = XPIV
               IFIN = JHIG
!
!  One chooses a pivot from the 2 first values and the last one.
!  This should ensure sufficient renewal between iterations to
!  avoid worst case behavior effects.
!
               IWRK1 = IHIGT (1)
               IWRK2 = IHIGT (2)
               IWRK3 = IHIGT (IFIN)
               If (XDONT(IWRK2) < XDONT(IWRK1)) Then
                  IHIGT (1) = IWRK2
                  IHIGT (2) = IWRK1
                  IWRK2 = IWRK1
               End If
               If (XDONT(IWRK2) > XDONT(IWRK3)) Then
                  IHIGT (IFIN) = IWRK2
                  IHIGT (2) = IWRK3
                  IWRK2 = IWRK3
                  If (XDONT(IWRK2) < XDONT(IHIGT(1))) Then
                     IHIGT (2) = IHIGT (1)
                     IHIGT (1) = IWRK2
                  End If
               End If
!
               JDEB = JLOW
               NWRK = NORD - JLOW
               IWRK1 = IHIGT (1)
               XPIV = XDONT (IWRK1) + REAL (NWRK) / REAL (NORD+NWRK) * &
                                      (XDONT(IHIGT(IFIN))-XDONT(IWRK1))
!
!  One takes values <= pivot to ILOWT
!  Again, 2 parts, one where we take care of the remaining
!  high values because we might still need them, and the
!  other when we know that we will have more than enough
!  low values in the end.
!
               JHIG = 0
               lowloop3: Do ICRS = 1, IFIN
                  If (XDONT(IHIGT(ICRS)) <= XPIV) Then
                     Do ILOW = 1, JLOW
                        If (XDONT(IHIGT(ICRS)) == XDONT (ILOWT(ILOW))) &
                            Cycle lowloop3
                     End Do
                     JLOW = JLOW + 1
                     ILOWT (JLOW) = IHIGT (ICRS)
                     If (JLOW > NORD) Exit
                  Else
                     JHIG = JHIG + 1
                     IHIGT (JHIG) = IHIGT (ICRS)
                  End If
               End Do lowloop3
!
               Do ICRS = ICRS + 1, IFIN
                  If (XDONT(IHIGT(ICRS)) <= XPIV) Then
                     JLOW = JLOW + 1
                     ILOWT (JLOW) = IHIGT (ICRS)
                  End If
               End Do
           End Select
!
! ______________________________
!
         Case (1)
!
!  Only 1 value is missing in low part
!
            XMIN = XDONT (IHIGT(1))
            IHIG = 1
            Do ICRS = 2, JHIG
               If (XDONT(IHIGT(ICRS)) < XMIN) Then
                  XMIN = XDONT (IHIGT(ICRS))
                  IHIG = ICRS
               End If
            End Do
!
            JLOW = JLOW + 1
            ILOWT (JLOW) = IHIGT (IHIG)
            Exit
!
! ______________________________
!
         Case (0)
!
!  Low part is exactly what we want
!
            Exit
!
! ______________________________
!
         Case (-5:-1)
!
!  Only few values too many in low part
!
            IRNGT (1) = ILOWT (1)
            Do ICRS = 2, NORD
               IWRK = ILOWT (ICRS)
               XWRK = XDONT (IWRK)
               Do IDCR = ICRS - 1, 1, - 1
                  If (XWRK < XDONT(IRNGT(IDCR))) Then
                     IRNGT (IDCR+1) = IRNGT (IDCR)
                  Else
                     Exit
                  End If
               End Do
               IRNGT (IDCR+1) = IWRK
            End Do
!
            XWRK1 = XDONT (IRNGT(NORD))
            insert1: Do ICRS = NORD + 1, JLOW
               If (XDONT(ILOWT (ICRS)) < XWRK1) Then
                  XWRK = XDONT (ILOWT (ICRS))
                  Do ILOW = 1, NORD - 1
                     If (XWRK <= XDONT(IRNGT(ILOW))) Then
                        If (XWRK == XDONT(IRNGT(ILOW))) Cycle insert1
                        Exit
                     End If
                  End Do
                  Do IDCR = NORD - 1, ILOW, - 1
                     IRNGT (IDCR+1) = IRNGT (IDCR)
                  End Do
                  IRNGT (IDCR+1) = ILOWT (ICRS)
                  XWRK1 = XDONT (IRNGT(NORD))
               End If
            End Do insert1
!
            Return
!
! ______________________________
!
         Case (:-6)
!
! last case: too many values in low part
!
            IDEB = JDEB + 1
            IMIL = MIN ((JLOW+IDEB) / 2, NORD)
            IFIN = MIN (JLOW, NORD+1)
!
!  One chooses a pivot from 1st, last, and middle values
!
            If (XDONT(ILOWT(IMIL)) < XDONT(ILOWT(IDEB))) Then
               IWRK = ILOWT (IDEB)
               ILOWT (IDEB) = ILOWT (IMIL)
               ILOWT (IMIL) = IWRK
            End If
            If (XDONT(ILOWT(IMIL)) > XDONT(ILOWT(IFIN))) Then
               IWRK = ILOWT (IFIN)
               ILOWT (IFIN) = ILOWT (IMIL)
               ILOWT (IMIL) = IWRK
               If (XDONT(ILOWT(IMIL)) < XDONT(ILOWT(IDEB))) Then
                  IWRK = ILOWT (IDEB)
                  ILOWT (IDEB) = ILOWT (IMIL)
                  ILOWT (IMIL) = IWRK
               End If
            End If
            If (IFIN <= 3) Exit
!
            XPIV = XDONT (ILOWT(IDEB)) + REAL(NORD)/REAL(JLOW+NORD) * &
                                      (XDONT(ILOWT(IFIN))-XDONT(ILOWT(1)))
            If (JDEB > 0) Then
               If (XPIV <= XPIV0) &
                   XPIV = XPIV0 + REAL(2*NORD-JDEB)/REAL (JLOW+NORD) * &
                                  (XDONT(ILOWT(IFIN))-XPIV0)
            Else
               IDEB = 1
            End If
!
!  One takes values > XPIV to IHIGT
!  However, we do not process the first values if we have been
!  through the case when we did not have enough low values
!
            JHIG = 0
            IFIN = JLOW
            JLOW = JDEB
!
            If (XDONT(ILOWT(IFIN)) > XPIV) Then
               ICRS = JDEB
              lowloop4: Do
                 ICRS = ICRS + 1
                  If (XDONT(ILOWT(ICRS)) > XPIV) Then
                     JHIG = JHIG + 1
                     IHIGT (JHIG) = ILOWT (ICRS)
                     If (ICRS >= IFIN) Exit
                  Else
                     XWRK1 = XDONT(ILOWT(ICRS))
                     Do ILOW = IDEB, JLOW
                        If (XWRK1 == XDONT(ILOWT(ILOW))) &
                            Cycle lowloop4
                     End Do
                     JLOW = JLOW + 1
                     ILOWT (JLOW) = ILOWT (ICRS)
                     If (JLOW >= NORD) Exit
                  End If
               End Do lowloop4
!
               If (ICRS < IFIN) Then
                  Do
                     ICRS = ICRS + 1
                     If (XDONT(ILOWT(ICRS)) <= XPIV) Then
                        JLOW = JLOW + 1
                        ILOWT (JLOW) = ILOWT (ICRS)
                     Else
                        If (ICRS >= IFIN) Exit
                     End If
                  End Do
               End If
           Else
              lowloop5: Do ICRS = IDEB, IFIN
                  If (XDONT(ILOWT(ICRS)) > XPIV) Then
                     JHIG = JHIG + 1
                     IHIGT (JHIG) = ILOWT (ICRS)
                  Else
                     XWRK1 = XDONT(ILOWT(ICRS))
                     Do ILOW = IDEB, JLOW
                        If (XWRK1 == XDONT(ILOWT(ILOW))) &
                            Cycle lowloop5
                     End Do
                     JLOW = JLOW + 1
                     ILOWT (JLOW) = ILOWT (ICRS)
                     If (JLOW >= NORD) Exit
                  End If
               End Do lowloop5
!
               Do ICRS = ICRS + 1, IFIN
                  If (XDONT(ILOWT(ICRS)) <= XPIV) Then
                     JLOW = JLOW + 1
                     ILOWT (JLOW) = ILOWT (ICRS)
                  End If
               End Do
            End If
!
         End Select
! ______________________________
!
      End Do
!
!  Now, we only need to complete ranking of the 1:NORD set
!  Assuming NORD is small, we use a simple insertion sort
!
      IRNGT (1) = ILOWT (1)
      Do ICRS = 2, NORD
         IWRK = ILOWT (ICRS)
         XWRK = XDONT (IWRK)
         Do IDCR = ICRS - 1, 1, - 1
            If (XWRK < XDONT(IRNGT(IDCR))) Then
               IRNGT (IDCR+1) = IRNGT (IDCR)
            Else
               Exit
            End If
         End Do
         IRNGT (IDCR+1) = IWRK
      End Do
     Return
!
!
End Subroutine D_unipar

Subroutine R_unipar (XDONT, IRNGT, NORD)
!  Ranks partially XDONT by IRNGT, up to order NORD at most,
!  removing duplicate entries
! __________________________________________________________
!  This routine uses a pivoting strategy such as the one of
!  finding the median based on the quicksort algorithm, but
!  we skew the pivot choice to try to bring it to NORD as
!  quickly as possible. It uses 2 temporary arrays, where it
!  stores the indices of the values smaller than the pivot
!  (ILOWT), and the indices of values larger than the pivot
!  that we might still need later on (IHIGT). It iterates
!  until it can bring the number of values in ILOWT to
!  exactly NORD, and then uses an insertion sort to rank
!  this set, since it is supposedly small. At all times, the
!  NORD first values in ILOWT correspond to distinct values
!  of the input array.
!  Michel Olagnon - Feb. 2000
! __________________________________________________________
! _________________________________________________________
      Real, Dimension (:), Intent (In) :: XDONT
      Integer, Dimension (:), Intent (Out) :: IRNGT
      Integer, Intent (InOut) :: NORD
! __________________________________________________________
      Real    :: XPIV, XWRK, XWRK1, XMIN, XMAX, XPIV0
!
      Integer, Dimension (SIZE(XDONT)) :: ILOWT, IHIGT
      Integer :: NDON, JHIG, JLOW, IHIG, IWRK, IWRK1, IWRK2, IWRK3
      Integer :: IDEB, JDEB, IMIL, IFIN, NWRK, ICRS, IDCR, ILOW
      Integer :: JLM2, JLM1, JHM2, JHM1
!
      NDON = SIZE (XDONT)
!
!    First loop is used to fill-in ILOWT, IHIGT at the same time
!
      If (NDON < 2) Then
         If (NORD >= 1) Then
            NORD = 1
            IRNGT (1) = 1
         End If
         Return
      End If
!
!  One chooses a pivot, best estimate possible to put fractile near
!  mid-point of the set of low values.
!
     Do ICRS = 2, NDON
        If (XDONT(ICRS) == XDONT(1)) Then
          Cycle
        Else If (XDONT(ICRS) < XDONT(1)) Then
           ILOWT (1) = ICRS
           IHIGT (1) = 1
        Else
           ILOWT (1) = 1
           IHIGT (1) = ICRS
        End If
        Exit
     End Do
!
      If (NDON <= ICRS) Then
         NORD = Min (NORD, 2)
         If (NORD >= 1) IRNGT (1) = ILOWT (1)
         If (NORD >= 2) IRNGT (2) = IHIGT (1)
         Return
      End If
!
      ICRS = ICRS + 1
      JHIG = 1
      If (XDONT(ICRS) < XDONT(IHIGT(1))) Then
         If (XDONT(ICRS) < XDONT(ILOWT(1))) Then
            JHIG = JHIG + 1
            IHIGT (JHIG) = IHIGT (1)
            IHIGT (1) = ILOWT (1)
            ILOWT (1) = ICRS
         Else If (XDONT(ICRS) > XDONT(ILOWT(1))) Then
            JHIG = JHIG + 1
            IHIGT (JHIG) = IHIGT (1)
            IHIGT (1) = ICRS
         End If
      ElseIf (XDONT(ICRS) > XDONT(IHIGT(1))) Then
         JHIG = JHIG + 1
         IHIGT (JHIG) = ICRS
      End If
!
      If (NDON <= ICRS) Then
         NORD = Min (NORD, JHIG+1)
         If (NORD >= 1) IRNGT (1) = ILOWT (1)
         If (NORD >= 2) IRNGT (2) = IHIGT (1)
         If (NORD >= 3) IRNGT (3) = IHIGT (2)
         Return
      End If
!
      If (XDONT(NDON) < XDONT(IHIGT(1))) Then
         If (XDONT(NDON) < XDONT(ILOWT(1))) Then
            Do IDCR = JHIG, 1, -1
              IHIGT (IDCR+1) = IHIGT (IDCR)
            End Do
            IHIGT (1) = ILOWT (1)
            ILOWT (1) = NDON
            JHIG = JHIG + 1
         ElseIf (XDONT(NDON) > XDONT(ILOWT(1))) Then
            Do IDCR = JHIG, 1, -1
              IHIGT (IDCR+1) = IHIGT (IDCR)
            End Do
            IHIGT (1) = NDON
            JHIG = JHIG + 1
         End If
      ElseIf (XDONT(NDON) > XDONT(IHIGT(1))) Then
         JHIG = JHIG + 1
         IHIGT (JHIG) = NDON
      End If
!
      If (NDON <= ICRS+1) Then
         NORD = Min (NORD, JHIG+1)
         If (NORD >= 1) IRNGT (1) = ILOWT (1)
         If (NORD >= 2) IRNGT (2) = IHIGT (1)
         If (NORD >= 3) IRNGT (3) = IHIGT (2)
         If (NORD >= 4) IRNGT (4) = IHIGT (3)
         Return
      End If
!
      JDEB = 0
      IDEB = JDEB + 1
      JLOW = IDEB
      XPIV = XDONT (ILOWT(IDEB)) + REAL(2*NORD)/REAL(NDON+NORD) * &
                                   (XDONT(IHIGT(3))-XDONT(ILOWT(IDEB)))
      If (XPIV >= XDONT(IHIGT(1))) Then
         XPIV = XDONT (ILOWT(IDEB)) + REAL(2*NORD)/REAL(NDON+NORD) * &
                                      (XDONT(IHIGT(2))-XDONT(ILOWT(IDEB)))
         If (XPIV >= XDONT(IHIGT(1))) &
             XPIV = XDONT (ILOWT(IDEB)) + REAL (2*NORD) / REAL (NDON+NORD) * &
                                          (XDONT(IHIGT(1))-XDONT(ILOWT(IDEB)))
      End If
      XPIV0 = XPIV
!
!  One puts values > pivot in the end and those <= pivot
!  at the beginning. This is split in 2 cases, so that
!  we can skip the loop test a number of times.
!  As we are also filling in the work arrays at the same time
!  we stop filling in the IHIGT array as soon as we have more
!  than enough values in ILOWT, i.e. one more than
!  strictly necessary so as to be able to come out of the
!  case where JLOWT would be NORD distinct values followed
!  by values that are exclusively duplicates of these.
!
!
      If (XDONT(NDON) > XPIV) Then
         lowloop1: Do
            ICRS = ICRS + 1
            If (XDONT(ICRS) > XPIV) Then
               If (ICRS >= NDON) Exit
               JHIG = JHIG + 1
               IHIGT (JHIG) = ICRS
            Else
               Do ILOW = 1, JLOW
                 If (XDONT(ICRS) == XDONT(ILOWT(ILOW))) Cycle lowloop1
               End Do
               JLOW = JLOW + 1
               ILOWT (JLOW) = ICRS
               If (JLOW >= NORD) Exit
            End If
         End Do lowloop1
!
!  One restricts further processing because it is no use
!  to store more high values
!
         If (ICRS < NDON-1) Then
            Do
               ICRS = ICRS + 1
               If (XDONT(ICRS) <= XPIV) Then
                  JLOW = JLOW + 1
                  ILOWT (JLOW) = ICRS
               Else If (ICRS >= NDON) Then
                  Exit
               End If
            End Do
         End If
!
!
      Else
!
!  Same as above, but this is not as easy to optimize, so the
!  DO-loop is kept
!
         lowloop2: Do ICRS = ICRS + 1, NDON - 1
            If (XDONT(ICRS) > XPIV) Then
               JHIG = JHIG + 1
               IHIGT (JHIG) = ICRS
            Else
               Do ILOW = 1, JLOW
                 If (XDONT(ICRS) == XDONT (ILOWT(ILOW))) Cycle lowloop2
               End Do
               JLOW = JLOW + 1
               ILOWT (JLOW) = ICRS
               If (JLOW >= NORD) Exit
            End If
         End Do lowloop2
!
         If (ICRS < NDON-1) Then
            Do
               ICRS = ICRS + 1
               If (XDONT(ICRS) <= XPIV) Then
                  If (ICRS >= NDON) Exit
                  JLOW = JLOW + 1
                  ILOWT (JLOW) = ICRS
               End If
            End Do
         End If
      End If
!
      JLM2 = 0
      JLM1 = 0
      JHM2 = 0
      JHM1 = 0
      Do
         if (JLOW == NORD) Exit
         If (JLM2 == JLOW .And. JHM2 == JHIG) Then
!
!   We are oscillating. Perturbate by bringing JLOW closer by one
!   to NORD
!
           If (NORD > JLOW) Then
                XMIN = XDONT (IHIGT(1))
                IHIG = 1
                Do ICRS = 2, JHIG
                   If (XDONT(IHIGT(ICRS)) < XMIN) Then
                      XMIN = XDONT (IHIGT(ICRS))
                      IHIG = ICRS
                   End If
                End Do
!
                JLOW = JLOW + 1
                ILOWT (JLOW) = IHIGT (IHIG)
                IHIG = 0
                Do ICRS = 1, JHIG
                   If (XDONT(IHIGT (ICRS)) /= XMIN) then
                      IHIG = IHIG + 1
                      IHIGT (IHIG ) = IHIGT (ICRS)
                   End If
                End Do
                JHIG = IHIG
             Else
                ILOW = ILOWT (JLOW)
                XMAX = XDONT (ILOW)
                Do ICRS = 1, JLOW
                   If (XDONT(ILOWT(ICRS)) > XMAX) Then
                      IWRK = ILOWT (ICRS)
                      XMAX = XDONT (IWRK)
                      ILOWT (ICRS) = ILOW
                      ILOW = IWRK
                   End If
                End Do
                JLOW = JLOW - 1
             End If
         End If
         JLM2 = JLM1
         JLM1 = JLOW
         JHM2 = JHM1
         JHM1 = JHIG
!
!   We try to bring the number of values in the low values set
!   closer to NORD. In order to make better pivot choices, we
!   decrease NORD if we already know that we don't have that
!   many distinct values as a whole.
!
         IF (JLOW+JHIG < NORD) NORD = JLOW+JHIG
         Select Case (NORD-JLOW)
! ______________________________
         Case (2:)
!
!   Not enough values in low part, at least 2 are missing
!
            Select Case (JHIG)
!
!   Not enough values in high part either (too many duplicates)
!
            Case (0)
               NORD = JLOW
!
            Case (1)
               JLOW = JLOW + 1
               ILOWT (JLOW) = IHIGT (1)
               NORD = JLOW
!
!   We make a special case when we have so few values in
!   the high values set that it is bad performance to choose a pivot
!   and apply the general algorithm.
!
            Case (2)
               If (XDONT(IHIGT(1)) <= XDONT(IHIGT(2))) Then
                  JLOW = JLOW + 1
                  ILOWT (JLOW) = IHIGT (1)
                  JLOW = JLOW + 1
                  ILOWT (JLOW) = IHIGT (2)
               ElseIf (XDONT(IHIGT(1)) == XDONT(IHIGT(2))) Then
                  JLOW = JLOW + 1
                  ILOWT (JLOW) = IHIGT (1)
                  NORD = JLOW
               Else
                  JLOW = JLOW + 1
                  ILOWT (JLOW) = IHIGT (2)
                  JLOW = JLOW + 1
                  ILOWT (JLOW) = IHIGT (1)
               End If
               Exit
!
            Case (3)
!
!
               IWRK1 = IHIGT (1)
               IWRK2 = IHIGT (2)
               IWRK3 = IHIGT (3)
               If (XDONT(IWRK2) < XDONT(IWRK1)) Then
                  IHIGT (1) = IWRK2
                  IHIGT (2) = IWRK1
                  IWRK2 = IWRK1
               End If
               If (XDONT(IWRK2) > XDONT(IWRK3)) Then
                  IHIGT (3) = IWRK2
                  IHIGT (2) = IWRK3
                  IWRK2 = IWRK3
                  If (XDONT(IWRK2) < XDONT(IHIGT(1))) Then
                     IHIGT (2) = IHIGT (1)
                     IHIGT (1) = IWRK2
                  End If
               End If
               JHIG = 1
               JLOW = JLOW + 1
               ILOWT (JLOW) = IHIGT (1)
               JHIG = JHIG + 1
               IF (XDONT(IHIGT(JHIG)) /= XDONT(ILOWT(JLOW))) Then
                 JLOW = JLOW + 1
                 ILOWT (JLOW) = IHIGT (JHIG)
               End If
               JHIG = JHIG + 1
               IF (XDONT(IHIGT(JHIG)) /= XDONT(ILOWT(JLOW))) Then
                 JLOW = JLOW + 1
                 ILOWT (JLOW) = IHIGT (JHIG)
               End If
               NORD = Min (JLOW, NORD)
               Exit
!
            Case (4:)
!
!
               XPIV0 = XPIV
               IFIN = JHIG
!
!  One chooses a pivot from the 2 first values and the last one.
!  This should ensure sufficient renewal between iterations to
!  avoid worst case behavior effects.
!
               IWRK1 = IHIGT (1)
               IWRK2 = IHIGT (2)
               IWRK3 = IHIGT (IFIN)
               If (XDONT(IWRK2) < XDONT(IWRK1)) Then
                  IHIGT (1) = IWRK2
                  IHIGT (2) = IWRK1
                  IWRK2 = IWRK1
               End If
               If (XDONT(IWRK2) > XDONT(IWRK3)) Then
                  IHIGT (IFIN) = IWRK2
                  IHIGT (2) = IWRK3
                  IWRK2 = IWRK3
                  If (XDONT(IWRK2) < XDONT(IHIGT(1))) Then
                     IHIGT (2) = IHIGT (1)
                     IHIGT (1) = IWRK2
                  End If
               End If
!
               JDEB = JLOW
               NWRK = NORD - JLOW
               IWRK1 = IHIGT (1)
               XPIV = XDONT (IWRK1) + REAL (NWRK) / REAL (NORD+NWRK) * &
                                      (XDONT(IHIGT(IFIN))-XDONT(IWRK1))
!
!  One takes values <= pivot to ILOWT
!  Again, 2 parts, one where we take care of the remaining
!  high values because we might still need them, and the
!  other when we know that we will have more than enough
!  low values in the end.
!
               JHIG = 0
               lowloop3: Do ICRS = 1, IFIN
                  If (XDONT(IHIGT(ICRS)) <= XPIV) Then
                     Do ILOW = 1, JLOW
                        If (XDONT(IHIGT(ICRS)) == XDONT (ILOWT(ILOW))) &
                            Cycle lowloop3
                     End Do
                     JLOW = JLOW + 1
                     ILOWT (JLOW) = IHIGT (ICRS)
                     If (JLOW > NORD) Exit
                  Else
                     JHIG = JHIG + 1
                     IHIGT (JHIG) = IHIGT (ICRS)
                  End If
               End Do lowloop3
!
               Do ICRS = ICRS + 1, IFIN
                  If (XDONT(IHIGT(ICRS)) <= XPIV) Then
                     JLOW = JLOW + 1
                     ILOWT (JLOW) = IHIGT (ICRS)
                  End If
               End Do
           End Select
!
! ______________________________
!
         Case (1)
!
!  Only 1 value is missing in low part
!
            XMIN = XDONT (IHIGT(1))
            IHIG = 1
            Do ICRS = 2, JHIG
               If (XDONT(IHIGT(ICRS)) < XMIN) Then
                  XMIN = XDONT (IHIGT(ICRS))
                  IHIG = ICRS
               End If
            End Do
!
            JLOW = JLOW + 1
            ILOWT (JLOW) = IHIGT (IHIG)
            Exit
!
! ______________________________
!
         Case (0)
!
!  Low part is exactly what we want
!
            Exit
!
! ______________________________
!
         Case (-5:-1)
!
!  Only few values too many in low part
!
            IRNGT (1) = ILOWT (1)
            Do ICRS = 2, NORD
               IWRK = ILOWT (ICRS)
               XWRK = XDONT (IWRK)
               Do IDCR = ICRS - 1, 1, - 1
                  If (XWRK < XDONT(IRNGT(IDCR))) Then
                     IRNGT (IDCR+1) = IRNGT (IDCR)
                  Else
                     Exit
                  End If
               End Do
               IRNGT (IDCR+1) = IWRK
            End Do
!
            XWRK1 = XDONT (IRNGT(NORD))
            insert1: Do ICRS = NORD + 1, JLOW
               If (XDONT(ILOWT (ICRS)) < XWRK1) Then
                  XWRK = XDONT (ILOWT (ICRS))
                  Do ILOW = 1, NORD - 1
                     If (XWRK <= XDONT(IRNGT(ILOW))) Then
                        If (XWRK == XDONT(IRNGT(ILOW))) Cycle insert1
                        Exit
                     End If
                  End Do
                  Do IDCR = NORD - 1, ILOW, - 1
                     IRNGT (IDCR+1) = IRNGT (IDCR)
                  End Do
                  IRNGT (IDCR+1) = ILOWT (ICRS)
                  XWRK1 = XDONT (IRNGT(NORD))
               End If
            End Do insert1
!
            Return
!
! ______________________________
!
         Case (:-6)
!
! last case: too many values in low part
!
            IDEB = JDEB + 1
            IMIL = MIN ((JLOW+IDEB) / 2, NORD)
            IFIN = MIN (JLOW, NORD+1)
!
!  One chooses a pivot from 1st, last, and middle values
!
            If (XDONT(ILOWT(IMIL)) < XDONT(ILOWT(IDEB))) Then
               IWRK = ILOWT (IDEB)
               ILOWT (IDEB) = ILOWT (IMIL)
               ILOWT (IMIL) = IWRK
            End If
            If (XDONT(ILOWT(IMIL)) > XDONT(ILOWT(IFIN))) Then
               IWRK = ILOWT (IFIN)
               ILOWT (IFIN) = ILOWT (IMIL)
               ILOWT (IMIL) = IWRK
               If (XDONT(ILOWT(IMIL)) < XDONT(ILOWT(IDEB))) Then
                  IWRK = ILOWT (IDEB)
                  ILOWT (IDEB) = ILOWT (IMIL)
                  ILOWT (IMIL) = IWRK
               End If
            End If
            If (IFIN <= 3) Exit
!
            XPIV = XDONT (ILOWT(IDEB)) + REAL(NORD)/REAL(JLOW+NORD) * &
                                      (XDONT(ILOWT(IFIN))-XDONT(ILOWT(1)))
            If (JDEB > 0) Then
               If (XPIV <= XPIV0) &
                   XPIV = XPIV0 + REAL(2*NORD-JDEB)/REAL (JLOW+NORD) * &
                                  (XDONT(ILOWT(IFIN))-XPIV0)
            Else
               IDEB = 1
            End If
!
!  One takes values > XPIV to IHIGT
!  However, we do not process the first values if we have been
!  through the case when we did not have enough low values
!
            JHIG = 0
            IFIN = JLOW
            JLOW = JDEB
!
            If (XDONT(ILOWT(IFIN)) > XPIV) Then
               ICRS = JDEB
              lowloop4: Do
                 ICRS = ICRS + 1
                  If (XDONT(ILOWT(ICRS)) > XPIV) Then
                     JHIG = JHIG + 1
                     IHIGT (JHIG) = ILOWT (ICRS)
                     If (ICRS >= IFIN) Exit
                  Else
                     XWRK1 = XDONT(ILOWT(ICRS))
                     Do ILOW = IDEB, JLOW
                        If (XWRK1 == XDONT(ILOWT(ILOW))) &
                            Cycle lowloop4
                     End Do
                     JLOW = JLOW + 1
                     ILOWT (JLOW) = ILOWT (ICRS)
                     If (JLOW >= NORD) Exit
                  End If
               End Do lowloop4
!
               If (ICRS < IFIN) Then
                  Do
                     ICRS = ICRS + 1
                     If (XDONT(ILOWT(ICRS)) <= XPIV) Then
                        JLOW = JLOW + 1
                        ILOWT (JLOW) = ILOWT (ICRS)
                     Else
                        If (ICRS >= IFIN) Exit
                     End If
                  End Do
               End If
           Else
              lowloop5: Do ICRS = IDEB, IFIN
                  If (XDONT(ILOWT(ICRS)) > XPIV) Then
                     JHIG = JHIG + 1
                     IHIGT (JHIG) = ILOWT (ICRS)
                  Else
                     XWRK1 = XDONT(ILOWT(ICRS))
                     Do ILOW = IDEB, JLOW
                        If (XWRK1 == XDONT(ILOWT(ILOW))) &
                            Cycle lowloop5
                     End Do
                     JLOW = JLOW + 1
                     ILOWT (JLOW) = ILOWT (ICRS)
                     If (JLOW >= NORD) Exit
                  End If
               End Do lowloop5
!
               Do ICRS = ICRS + 1, IFIN
                  If (XDONT(ILOWT(ICRS)) <= XPIV) Then
                     JLOW = JLOW + 1
                     ILOWT (JLOW) = ILOWT (ICRS)
                  End If
               End Do
            End If
!
         End Select
! ______________________________
!
      End Do
!
!  Now, we only need to complete ranking of the 1:NORD set
!  Assuming NORD is small, we use a simple insertion sort
!
      IRNGT (1) = ILOWT (1)
      Do ICRS = 2, NORD
         IWRK = ILOWT (ICRS)
         XWRK = XDONT (IWRK)
         Do IDCR = ICRS - 1, 1, - 1
            If (XWRK < XDONT(IRNGT(IDCR))) Then
               IRNGT (IDCR+1) = IRNGT (IDCR)
            Else
               Exit
            End If
         End Do
         IRNGT (IDCR+1) = IWRK
      End Do
     Return
!
!
End Subroutine R_unipar
Subroutine I_unipar (XDONT, IRNGT, NORD)
!  Ranks partially XDONT by IRNGT, up to order NORD at most,
!  removing duplicate entries
! __________________________________________________________
!  This routine uses a pivoting strategy such as the one of
!  finding the median based on the quicksort algorithm, but
!  we skew the pivot choice to try to bring it to NORD as
!  quickly as possible. It uses 2 temporary arrays, where it
!  stores the indices of the values smaller than the pivot
!  (ILOWT), and the indices of values larger than the pivot
!  that we might still need later on (IHIGT). It iterates
!  until it can bring the number of values in ILOWT to
!  exactly NORD, and then uses an insertion sort to rank
!  this set, since it is supposedly small. At all times, the
!  NORD first values in ILOWT correspond to distinct values
!  of the input array.
!  Michel Olagnon - Feb. 2000
! __________________________________________________________
! __________________________________________________________
      Integer, Dimension (:), Intent (In)  :: XDONT
      Integer, Dimension (:), Intent (Out) :: IRNGT
      Integer, Intent (InOut) :: NORD
! __________________________________________________________
      Integer :: XPIV, XWRK, XWRK1, XMIN, XMAX, XPIV0
!
      Integer, Dimension (SIZE(XDONT)) :: ILOWT, IHIGT
      Integer :: NDON, JHIG, JLOW, IHIG, IWRK, IWRK1, IWRK2, IWRK3
      Integer :: IDEB, JDEB, IMIL, IFIN, NWRK, ICRS, IDCR, ILOW
      Integer :: JLM2, JLM1, JHM2, JHM1
!
      NDON = SIZE (XDONT)
!
!    First loop is used to fill-in ILOWT, IHIGT at the same time
!
      If (NDON < 2) Then
         If (NORD >= 1) Then
            NORD = 1
            IRNGT (1) = 1
         End If
         Return
      End If
!
!  One chooses a pivot, best estimate possible to put fractile near
!  mid-point of the set of low values.
!
     Do ICRS = 2, NDON
        If (XDONT(ICRS) == XDONT(1)) Then
          Cycle
        Else If (XDONT(ICRS) < XDONT(1)) Then
           ILOWT (1) = ICRS
           IHIGT (1) = 1
        Else
           ILOWT (1) = 1
           IHIGT (1) = ICRS
        End If
        Exit
     End Do
!
      If (NDON <= ICRS) Then
         NORD = Min (NORD, 2)
         If (NORD >= 1) IRNGT (1) = ILOWT (1)
         If (NORD >= 2) IRNGT (2) = IHIGT (1)
         Return
      End If
!
      ICRS = ICRS + 1
      JHIG = 1
      If (XDONT(ICRS) < XDONT(IHIGT(1))) Then
         If (XDONT(ICRS) < XDONT(ILOWT(1))) Then
            JHIG = JHIG + 1
            IHIGT (JHIG) = IHIGT (1)
            IHIGT (1) = ILOWT (1)
            ILOWT (1) = ICRS
         Else If (XDONT(ICRS) > XDONT(ILOWT(1))) Then
            JHIG = JHIG + 1
            IHIGT (JHIG) = IHIGT (1)
            IHIGT (1) = ICRS
         End If
      ElseIf (XDONT(ICRS) > XDONT(IHIGT(1))) Then
         JHIG = JHIG + 1
         IHIGT (JHIG) = ICRS
      End If
!
      If (NDON <= ICRS) Then
         NORD = Min (NORD, JHIG+1)
         If (NORD >= 1) IRNGT (1) = ILOWT (1)
         If (NORD >= 2) IRNGT (2) = IHIGT (1)
         If (NORD >= 3) IRNGT (3) = IHIGT (2)
         Return
      End If
!
      If (XDONT(NDON) < XDONT(IHIGT(1))) Then
         If (XDONT(NDON) < XDONT(ILOWT(1))) Then
            Do IDCR = JHIG, 1, -1
              IHIGT (IDCR+1) = IHIGT (IDCR)
            End Do
            IHIGT (1) = ILOWT (1)
            ILOWT (1) = NDON
            JHIG = JHIG + 1
         ElseIf (XDONT(NDON) > XDONT(ILOWT(1))) Then
            Do IDCR = JHIG, 1, -1
              IHIGT (IDCR+1) = IHIGT (IDCR)
            End Do
            IHIGT (1) = NDON
            JHIG = JHIG + 1
         End If
      ElseIf (XDONT(NDON) > XDONT(IHIGT(1))) Then
         JHIG = JHIG + 1
         IHIGT (JHIG) = NDON
      End If
!
      If (NDON <= ICRS+1) Then
         NORD = Min (NORD, JHIG+1)
         If (NORD >= 1) IRNGT (1) = ILOWT (1)
         If (NORD >= 2) IRNGT (2) = IHIGT (1)
         If (NORD >= 3) IRNGT (3) = IHIGT (2)
         If (NORD >= 4) IRNGT (4) = IHIGT (3)
         Return
      End If
!
      JDEB = 0
      IDEB = JDEB + 1
      JLOW = IDEB
      XPIV = XDONT (ILOWT(IDEB)) + REAL(2*NORD)/REAL(NDON+NORD) * &
                                   (XDONT(IHIGT(3))-XDONT(ILOWT(IDEB)))
      If (XPIV >= XDONT(IHIGT(1))) Then
         XPIV = XDONT (ILOWT(IDEB)) + REAL(2*NORD)/REAL(NDON+NORD) * &
                                      (XDONT(IHIGT(2))-XDONT(ILOWT(IDEB)))
         If (XPIV >= XDONT(IHIGT(1))) &
             XPIV = XDONT (ILOWT(IDEB)) + REAL (2*NORD) / REAL (NDON+NORD) * &
                                          (XDONT(IHIGT(1))-XDONT(ILOWT(IDEB)))
      End If
      XPIV0 = XPIV
!
!  One puts values > pivot in the end and those <= pivot
!  at the beginning. This is split in 2 cases, so that
!  we can skip the loop test a number of times.
!  As we are also filling in the work arrays at the same time
!  we stop filling in the IHIGT array as soon as we have more
!  than enough values in ILOWT, i.e. one more than
!  strictly necessary so as to be able to come out of the
!  case where JLOWT would be NORD distinct values followed
!  by values that are exclusively duplicates of these.
!
!
      If (XDONT(NDON) > XPIV) Then
         lowloop1: Do
            ICRS = ICRS + 1
            If (XDONT(ICRS) > XPIV) Then
               If (ICRS >= NDON) Exit
               JHIG = JHIG + 1
               IHIGT (JHIG) = ICRS
            Else
               Do ILOW = 1, JLOW
                 If (XDONT(ICRS) == XDONT(ILOWT(ILOW))) Cycle lowloop1
               End Do
               JLOW = JLOW + 1
               ILOWT (JLOW) = ICRS
               If (JLOW >= NORD) Exit
            End If
         End Do lowloop1
!
!  One restricts further processing because it is no use
!  to store more high values
!
         If (ICRS < NDON-1) Then
            Do
               ICRS = ICRS + 1
               If (XDONT(ICRS) <= XPIV) Then
                  JLOW = JLOW + 1
                  ILOWT (JLOW) = ICRS
               Else If (ICRS >= NDON) Then
                  Exit
               End If
            End Do
         End If
!
!
      Else
!
!  Same as above, but this is not as easy to optimize, so the
!  DO-loop is kept
!
         lowloop2: Do ICRS = ICRS + 1, NDON - 1
            If (XDONT(ICRS) > XPIV) Then
               JHIG = JHIG + 1
               IHIGT (JHIG) = ICRS
            Else
               Do ILOW = 1, JLOW
                 If (XDONT(ICRS) == XDONT (ILOWT(ILOW))) Cycle lowloop2
               End Do
               JLOW = JLOW + 1
               ILOWT (JLOW) = ICRS
               If (JLOW >= NORD) Exit
            End If
         End Do lowloop2
!
         If (ICRS < NDON-1) Then
            Do
               ICRS = ICRS + 1
               If (XDONT(ICRS) <= XPIV) Then
                  If (ICRS >= NDON) Exit
                  JLOW = JLOW + 1
                  ILOWT (JLOW) = ICRS
               End If
            End Do
         End If
      End If
!
      JLM2 = 0
      JLM1 = 0
      JHM2 = 0
      JHM1 = 0
      Do
         if (JLOW == NORD) Exit
         If (JLM2 == JLOW .And. JHM2 == JHIG) Then
!
!   We are oscillating. Perturbate by bringing JLOW closer by one
!   to NORD
!
           If (NORD > JLOW) Then
                XMIN = XDONT (IHIGT(1))
                IHIG = 1
                Do ICRS = 2, JHIG
                   If (XDONT(IHIGT(ICRS)) < XMIN) Then
                      XMIN = XDONT (IHIGT(ICRS))
                      IHIG = ICRS
                   End If
                End Do
!
                JLOW = JLOW + 1
                ILOWT (JLOW) = IHIGT (IHIG)
                IHIG = 0
                Do ICRS = 1, JHIG
                   If (XDONT(IHIGT (ICRS)) /= XMIN) then
                      IHIG = IHIG + 1
                      IHIGT (IHIG ) = IHIGT (ICRS)
                   End If
                End Do
                JHIG = IHIG
             Else
                ILOW = ILOWT (JLOW)
                XMAX = XDONT (ILOW)
                Do ICRS = 1, JLOW
                   If (XDONT(ILOWT(ICRS)) > XMAX) Then
                      IWRK = ILOWT (ICRS)
                      XMAX = XDONT (IWRK)
                      ILOWT (ICRS) = ILOW
                      ILOW = IWRK
                   End If
                End Do
                JLOW = JLOW - 1
             End If
         End If
         JLM2 = JLM1
         JLM1 = JLOW
         JHM2 = JHM1
         JHM1 = JHIG
!
!   We try to bring the number of values in the low values set
!   closer to NORD. In order to make better pivot choices, we
!   decrease NORD if we already know that we don't have that
!   many distinct values as a whole.
!
         IF (JLOW+JHIG < NORD) NORD = JLOW+JHIG
         Select Case (NORD-JLOW)
! ______________________________
         Case (2:)
!
!   Not enough values in low part, at least 2 are missing
!
            Select Case (JHIG)
!
!   Not enough values in high part either (too many duplicates)
!
            Case (0)
               NORD = JLOW
!
            Case (1)
               JLOW = JLOW + 1
               ILOWT (JLOW) = IHIGT (1)
               NORD = JLOW
!
!   We make a special case when we have so few values in
!   the high values set that it is bad performance to choose a pivot
!   and apply the general algorithm.
!
            Case (2)
               If (XDONT(IHIGT(1)) <= XDONT(IHIGT(2))) Then
                  JLOW = JLOW + 1
                  ILOWT (JLOW) = IHIGT (1)
                  JLOW = JLOW + 1
                  ILOWT (JLOW) = IHIGT (2)
               ElseIf (XDONT(IHIGT(1)) == XDONT(IHIGT(2))) Then
                  JLOW = JLOW + 1
                  ILOWT (JLOW) = IHIGT (1)
                  NORD = JLOW
               Else
                  JLOW = JLOW + 1
                  ILOWT (JLOW) = IHIGT (2)
                  JLOW = JLOW + 1
                  ILOWT (JLOW) = IHIGT (1)
               End If
               Exit
!
            Case (3)
!
!
               IWRK1 = IHIGT (1)
               IWRK2 = IHIGT (2)
               IWRK3 = IHIGT (3)
               If (XDONT(IWRK2) < XDONT(IWRK1)) Then
                  IHIGT (1) = IWRK2
                  IHIGT (2) = IWRK1
                  IWRK2 = IWRK1
               End If
               If (XDONT(IWRK2) > XDONT(IWRK3)) Then
                  IHIGT (3) = IWRK2
                  IHIGT (2) = IWRK3
                  IWRK2 = IWRK3
                  If (XDONT(IWRK2) < XDONT(IHIGT(1))) Then
                     IHIGT (2) = IHIGT (1)
                     IHIGT (1) = IWRK2
                  End If
               End If
               JHIG = 1
               JLOW = JLOW + 1
               ILOWT (JLOW) = IHIGT (1)
               JHIG = JHIG + 1
               IF (XDONT(IHIGT(JHIG)) /= XDONT(ILOWT(JLOW))) Then
                 JLOW = JLOW + 1
                 ILOWT (JLOW) = IHIGT (JHIG)
               End If
               JHIG = JHIG + 1
               IF (XDONT(IHIGT(JHIG)) /= XDONT(ILOWT(JLOW))) Then
                 JLOW = JLOW + 1
                 ILOWT (JLOW) = IHIGT (JHIG)
               End If
               NORD = Min (JLOW, NORD)
               Exit
!
            Case (4:)
!
!
               XPIV0 = XPIV
               IFIN = JHIG
!
!  One chooses a pivot from the 2 first values and the last one.
!  This should ensure sufficient renewal between iterations to
!  avoid worst case behavior effects.
!
               IWRK1 = IHIGT (1)
               IWRK2 = IHIGT (2)
               IWRK3 = IHIGT (IFIN)
               If (XDONT(IWRK2) < XDONT(IWRK1)) Then
                  IHIGT (1) = IWRK2
                  IHIGT (2) = IWRK1
                  IWRK2 = IWRK1
               End If
               If (XDONT(IWRK2) > XDONT(IWRK3)) Then
                  IHIGT (IFIN) = IWRK2
                  IHIGT (2) = IWRK3
                  IWRK2 = IWRK3
                  If (XDONT(IWRK2) < XDONT(IHIGT(1))) Then
                     IHIGT (2) = IHIGT (1)
                     IHIGT (1) = IWRK2
                  End If
               End If
!
               JDEB = JLOW
               NWRK = NORD - JLOW
               IWRK1 = IHIGT (1)
               XPIV = XDONT (IWRK1) + REAL (NWRK) / REAL (NORD+NWRK) * &
                                      (XDONT(IHIGT(IFIN))-XDONT(IWRK1))
!
!  One takes values <= pivot to ILOWT
!  Again, 2 parts, one where we take care of the remaining
!  high values because we might still need them, and the
!  other when we know that we will have more than enough
!  low values in the end.
!
               JHIG = 0
               lowloop3: Do ICRS = 1, IFIN
                  If (XDONT(IHIGT(ICRS)) <= XPIV) Then
                     Do ILOW = 1, JLOW
                        If (XDONT(IHIGT(ICRS)) == XDONT (ILOWT(ILOW))) &
                            Cycle lowloop3
                     End Do
                     JLOW = JLOW + 1
                     ILOWT (JLOW) = IHIGT (ICRS)
                     If (JLOW > NORD) Exit
                  Else
                     JHIG = JHIG + 1
                     IHIGT (JHIG) = IHIGT (ICRS)
                  End If
               End Do lowloop3
!
               Do ICRS = ICRS + 1, IFIN
                  If (XDONT(IHIGT(ICRS)) <= XPIV) Then
                     JLOW = JLOW + 1
                     ILOWT (JLOW) = IHIGT (ICRS)
                  End If
               End Do
           End Select
!
! ______________________________
!
         Case (1)
!
!  Only 1 value is missing in low part
!
            XMIN = XDONT (IHIGT(1))
            IHIG = 1
            Do ICRS = 2, JHIG
               If (XDONT(IHIGT(ICRS)) < XMIN) Then
                  XMIN = XDONT (IHIGT(ICRS))
                  IHIG = ICRS
               End If
            End Do
!
            JLOW = JLOW + 1
            ILOWT (JLOW) = IHIGT (IHIG)
            Exit
!
! ______________________________
!
         Case (0)
!
!  Low part is exactly what we want
!
            Exit
!
! ______________________________
!
         Case (-5:-1)
!
!  Only few values too many in low part
!
            IRNGT (1) = ILOWT (1)
            Do ICRS = 2, NORD
               IWRK = ILOWT (ICRS)
               XWRK = XDONT (IWRK)
               Do IDCR = ICRS - 1, 1, - 1
                  If (XWRK < XDONT(IRNGT(IDCR))) Then
                     IRNGT (IDCR+1) = IRNGT (IDCR)
                  Else
                     Exit
                  End If
               End Do
               IRNGT (IDCR+1) = IWRK
            End Do
!
            XWRK1 = XDONT (IRNGT(NORD))
            insert1: Do ICRS = NORD + 1, JLOW
               If (XDONT(ILOWT (ICRS)) < XWRK1) Then
                  XWRK = XDONT (ILOWT (ICRS))
                  Do ILOW = 1, NORD - 1
                     If (XWRK <= XDONT(IRNGT(ILOW))) Then
                        If (XWRK == XDONT(IRNGT(ILOW))) Cycle insert1
                        Exit
                     End If
                  End Do
                  Do IDCR = NORD - 1, ILOW, - 1
                     IRNGT (IDCR+1) = IRNGT (IDCR)
                  End Do
                  IRNGT (IDCR+1) = ILOWT (ICRS)
                  XWRK1 = XDONT (IRNGT(NORD))
               End If
            End Do insert1
!
            Return
!
! ______________________________
!
         Case (:-6)
!
! last case: too many values in low part
!
            IDEB = JDEB + 1
            IMIL = MIN ((JLOW+IDEB) / 2, NORD)
            IFIN = MIN (JLOW, NORD+1)
!
!  One chooses a pivot from 1st, last, and middle values
!
            If (XDONT(ILOWT(IMIL)) < XDONT(ILOWT(IDEB))) Then
               IWRK = ILOWT (IDEB)
               ILOWT (IDEB) = ILOWT (IMIL)
               ILOWT (IMIL) = IWRK
            End If
            If (XDONT(ILOWT(IMIL)) > XDONT(ILOWT(IFIN))) Then
               IWRK = ILOWT (IFIN)
               ILOWT (IFIN) = ILOWT (IMIL)
               ILOWT (IMIL) = IWRK
               If (XDONT(ILOWT(IMIL)) < XDONT(ILOWT(IDEB))) Then
                  IWRK = ILOWT (IDEB)
                  ILOWT (IDEB) = ILOWT (IMIL)
                  ILOWT (IMIL) = IWRK
               End If
            End If
            If (IFIN <= 3) Exit
!
            XPIV = XDONT (ILOWT(IDEB)) + REAL(NORD)/REAL(JLOW+NORD) * &
                                      (XDONT(ILOWT(IFIN))-XDONT(ILOWT(1)))
            If (JDEB > 0) Then
               If (XPIV <= XPIV0) &
                   XPIV = XPIV0 + REAL(2*NORD-JDEB)/REAL (JLOW+NORD) * &
                                  (XDONT(ILOWT(IFIN))-XPIV0)
            Else
               IDEB = 1
            End If
!
!  One takes values > XPIV to IHIGT
!  However, we do not process the first values if we have been
!  through the case when we did not have enough low values
!
            JHIG = 0
            IFIN = JLOW
            JLOW = JDEB
!
            If (XDONT(ILOWT(IFIN)) > XPIV) Then
               ICRS = JDEB
              lowloop4: Do
                 ICRS = ICRS + 1
                  If (XDONT(ILOWT(ICRS)) > XPIV) Then
                     JHIG = JHIG + 1
                     IHIGT (JHIG) = ILOWT (ICRS)
                     If (ICRS >= IFIN) Exit
                  Else
                     XWRK1 = XDONT(ILOWT(ICRS))
                     Do ILOW = IDEB, JLOW
                        If (XWRK1 == XDONT(ILOWT(ILOW))) &
                            Cycle lowloop4
                     End Do
                     JLOW = JLOW + 1
                     ILOWT (JLOW) = ILOWT (ICRS)
                     If (JLOW >= NORD) Exit
                  End If
               End Do lowloop4
!
               If (ICRS < IFIN) Then
                  Do
                     ICRS = ICRS + 1
                     If (XDONT(ILOWT(ICRS)) <= XPIV) Then
                        JLOW = JLOW + 1
                        ILOWT (JLOW) = ILOWT (ICRS)
                     Else
                        If (ICRS >= IFIN) Exit
                     End If
                  End Do
               End If
           Else
              lowloop5: Do ICRS = IDEB, IFIN
                  If (XDONT(ILOWT(ICRS)) > XPIV) Then
                     JHIG = JHIG + 1
                     IHIGT (JHIG) = ILOWT (ICRS)
                  Else
                     XWRK1 = XDONT(ILOWT(ICRS))
                     Do ILOW = IDEB, JLOW
                        If (XWRK1 == XDONT(ILOWT(ILOW))) &
                            Cycle lowloop5
                     End Do
                     JLOW = JLOW + 1
                     ILOWT (JLOW) = ILOWT (ICRS)
                     If (JLOW >= NORD) Exit
                  End If
               End Do lowloop5
!
               Do ICRS = ICRS + 1, IFIN
                  If (XDONT(ILOWT(ICRS)) <= XPIV) Then
                     JLOW = JLOW + 1
                     ILOWT (JLOW) = ILOWT (ICRS)
                  End If
               End Do
            End If
!
         End Select
! ______________________________
!
      End Do
!
!  Now, we only need to complete ranking of the 1:NORD set
!  Assuming NORD is small, we use a simple insertion sort
!
      IRNGT (1) = ILOWT (1)
      Do ICRS = 2, NORD
         IWRK = ILOWT (ICRS)
         XWRK = XDONT (IWRK)
         Do IDCR = ICRS - 1, 1, - 1
            If (XWRK < XDONT(IRNGT(IDCR))) Then
               IRNGT (IDCR+1) = IRNGT (IDCR)
            Else
               Exit
            End If
         End Do
         IRNGT (IDCR+1) = IWRK
      End Do
     Return
!
!
End Subroutine I_unipar
end module m_unipar
Module m_unirnk
Integer, Parameter :: kdp = selected_real_kind(15)
public :: unirnk
private :: kdp
private :: R_unirnk, I_unirnk, D_unirnk
private :: R_nearless, I_nearless, D_nearless, nearless
interface unirnk
  module procedure D_unirnk, R_unirnk, I_unirnk
end interface unirnk
interface nearless
  module procedure D_nearless, R_nearless, I_nearless
end interface nearless

contains

Subroutine D_unirnk (XVALT, IRNGT, NUNI)
! __________________________________________________________
!   UNIRNK = Merge-sort ranking of an array, with removal of
!   duplicate entries.
!   The routine is similar to pure merge-sort ranking, but on
!   the last pass, it discards indices that correspond to
!   duplicate entries.
!   For performance reasons, the first 2 passes are taken
!   out of the standard loop, and use dedicated coding.
! __________________________________________________________
! __________________________________________________________
      Real (Kind=kdp), Dimension (:), Intent (In) :: XVALT
      Integer, Dimension (:), Intent (Out) :: IRNGT
      Integer, Intent (Out) :: NUNI
! __________________________________________________________
      Integer, Dimension (SIZE(IRNGT)) :: JWRKT
      Integer :: LMTNA, LMTNC, IRNG, IRNG1, IRNG2
      Integer :: NVAL, IIND, IWRKD, IWRK, IWRKF, JINDA, IINDA, IINDB
      Real (Kind=kdp) :: XTST, XVALA, XVALB
!
!
      NVAL = Min (SIZE(XVALT), SIZE(IRNGT))
      NUNI = NVAL
!
      Select Case (NVAL)
      Case (:0)
         Return
      Case (1)
         IRNGT (1) = 1
         Return
      Case Default
         Continue
      End Select
!
!  Fill-in the index array, creating ordered couples
!
      Do IIND = 2, NVAL, 2
         If (XVALT(IIND-1) < XVALT(IIND)) Then
            IRNGT (IIND-1) = IIND - 1
            IRNGT (IIND) = IIND
         Else
            IRNGT (IIND-1) = IIND
            IRNGT (IIND) = IIND - 1
         End If
      End Do
      If (Modulo(NVAL, 2) /= 0) Then
         IRNGT (NVAL) = NVAL
      End If
!
!  We will now have ordered subsets A - B - A - B - ...
!  and merge A and B couples into     C   -   C   - ...
!
      LMTNA = 2
      LMTNC = 4
!
!  First iteration. The length of the ordered subsets goes from 2 to 4
!
      Do
         If (NVAL <= 4) Exit
!
!   Loop on merges of A and B into C
!
         Do IWRKD = 0, NVAL - 1, 4
            If ((IWRKD+4) > NVAL) Then
               If ((IWRKD+2) >= NVAL) Exit
!
!   1 2 3
!
               If (XVALT(IRNGT(IWRKD+2)) <= XVALT(IRNGT(IWRKD+3))) Exit
!
!   1 3 2
!
               If (XVALT(IRNGT(IWRKD+1)) <= XVALT(IRNGT(IWRKD+3))) Then
                  IRNG2 = IRNGT (IWRKD+2)
                  IRNGT (IWRKD+2) = IRNGT (IWRKD+3)
                  IRNGT (IWRKD+3) = IRNG2
!
!   3 1 2
!
               Else
                  IRNG1 = IRNGT (IWRKD+1)
                  IRNGT (IWRKD+1) = IRNGT (IWRKD+3)
                  IRNGT (IWRKD+3) = IRNGT (IWRKD+2)
                  IRNGT (IWRKD+2) = IRNG1
               End If
               Exit
            End If
!
!   1 2 3 4
!
            If (XVALT(IRNGT(IWRKD+2)) <= XVALT(IRNGT(IWRKD+3))) Cycle
!
!   1 3 x x
!
            If (XVALT(IRNGT(IWRKD+1)) <= XVALT(IRNGT(IWRKD+3))) Then
               IRNG2 = IRNGT (IWRKD+2)
               IRNGT (IWRKD+2) = IRNGT (IWRKD+3)
               If (XVALT(IRNG2) <= XVALT(IRNGT(IWRKD+4))) Then
!   1 3 2 4
                  IRNGT (IWRKD+3) = IRNG2
               Else
!   1 3 4 2
                  IRNGT (IWRKD+3) = IRNGT (IWRKD+4)
                  IRNGT (IWRKD+4) = IRNG2
               End If
!
!   3 x x x
!
            Else
               IRNG1 = IRNGT (IWRKD+1)
               IRNG2 = IRNGT (IWRKD+2)
               IRNGT (IWRKD+1) = IRNGT (IWRKD+3)
               If (XVALT(IRNG1) <= XVALT(IRNGT(IWRKD+4))) Then
                  IRNGT (IWRKD+2) = IRNG1
                  If (XVALT(IRNG2) <= XVALT(IRNGT(IWRKD+4))) Then
!   3 1 2 4
                     IRNGT (IWRKD+3) = IRNG2
                  Else
!   3 1 4 2
                     IRNGT (IWRKD+3) = IRNGT (IWRKD+4)
                     IRNGT (IWRKD+4) = IRNG2
                  End If
               Else
!   3 4 1 2
                  IRNGT (IWRKD+2) = IRNGT (IWRKD+4)
                  IRNGT (IWRKD+3) = IRNG1
                  IRNGT (IWRKD+4) = IRNG2
               End If
            End If
         End Do
!
!  The Cs become As and Bs
!
         LMTNA = 4
         Exit
      End Do
!
!  Iteration loop. Each time, the length of the ordered subsets
!  is doubled.
!
      Do
         If (2*LMTNA >= NVAL) Exit
         IWRKF = 0
         LMTNC = 2 * LMTNC
!
!   Loop on merges of A and B into C
!
         Do
            IWRK = IWRKF
            IWRKD = IWRKF + 1
            JINDA = IWRKF + LMTNA
            IWRKF = IWRKF + LMTNC
            If (IWRKF >= NVAL) Then
               If (JINDA >= NVAL) Exit
               IWRKF = NVAL
            End If
            IINDA = 1
            IINDB = JINDA + 1
!
!  One steps in the C subset, that we create in the final rank array
!
!  Make a copy of the rank array for the iteration
!
            JWRKT (1:LMTNA) = IRNGT (IWRKD:JINDA)
            XVALA = XVALT (JWRKT(IINDA))
            XVALB = XVALT (IRNGT(IINDB))
!
            Do
               IWRK = IWRK + 1
!
!  We still have unprocessed values in both A and B
!
               If (XVALA > XVALB) Then
                  IRNGT (IWRK) = IRNGT (IINDB)
                  IINDB = IINDB + 1
                  If (IINDB > IWRKF) Then
!  Only A still with unprocessed values
                     IRNGT (IWRK+1:IWRKF) = JWRKT (IINDA:LMTNA)
                     Exit
                  End If
                  XVALB = XVALT (IRNGT(IINDB))
               Else
                  IRNGT (IWRK) = JWRKT (IINDA)
                  IINDA = IINDA + 1
                  If (IINDA > LMTNA) Exit! Only B still with unprocessed values
                  XVALA = XVALT (JWRKT(IINDA))
               End If
!
            End Do
         End Do
!
!  The Cs become As and Bs
!
         LMTNA = 2 * LMTNA
      End Do
!
!   Last merge of A and B into C, with removal of duplicates.
!
      IINDA = 1
      IINDB = LMTNA + 1
      NUNI = 0
!
!  One steps in the C subset, that we create in the final rank array
!
      JWRKT (1:LMTNA) = IRNGT (1:LMTNA)
      If (IINDB <= NVAL) Then
        XTST = NEARLESS (Min(XVALT(JWRKT(1)), XVALT(IRNGT(IINDB))))
      Else
        XTST = NEARLESS (XVALT(JWRKT(1)))
      Endif
      Do IWRK = 1, NVAL
!
!  We still have unprocessed values in both A and B
!
         If (IINDA <= LMTNA) Then
            If (IINDB <= NVAL) Then
               If (XVALT(JWRKT(IINDA)) > XVALT(IRNGT(IINDB))) Then
                  IRNG = IRNGT (IINDB)
                  IINDB = IINDB + 1
               Else
                  IRNG = JWRKT (IINDA)
                  IINDA = IINDA + 1
               End If
            Else
!
!  Only A still with unprocessed values
!
               IRNG = JWRKT (IINDA)
               IINDA = IINDA + 1
            End If
         Else
!
!  Only B still with unprocessed values
!
            IRNG = IRNGT (IWRK)
         End If
         If (XVALT(IRNG) > XTST) Then
            XTST = XVALT (IRNG)
            NUNI = NUNI + 1
            IRNGT (NUNI) = IRNG
         End If
!
      End Do
!
      Return
!
End Subroutine D_unirnk

Subroutine R_unirnk (XVALT, IRNGT, NUNI)
! __________________________________________________________
!   UNIRNK = Merge-sort ranking of an array, with removal of
!   duplicate entries.
!   The routine is similar to pure merge-sort ranking, but on
!   the last pass, it discards indices that correspond to
!   duplicate entries.
!   For performance reasons, the first 2 passes are taken
!   out of the standard loop, and use dedicated coding.
! __________________________________________________________
! __________________________________________________________
      Real, Dimension (:), Intent (In) :: XVALT
      Integer, Dimension (:), Intent (Out) :: IRNGT
      Integer, Intent (Out) :: NUNI
! __________________________________________________________
      Integer, Dimension (SIZE(IRNGT)) :: JWRKT
      Integer :: LMTNA, LMTNC, IRNG, IRNG1, IRNG2
      Integer :: NVAL, IIND, IWRKD, IWRK, IWRKF, JINDA, IINDA, IINDB
      Real :: XTST, XVALA, XVALB
!
!
      NVAL = Min (SIZE(XVALT), SIZE(IRNGT))
      NUNI = NVAL
!
      Select Case (NVAL)
      Case (:0)
         Return
      Case (1)
         IRNGT (1) = 1
         Return
      Case Default
         Continue
      End Select
!
!  Fill-in the index array, creating ordered couples
!
      Do IIND = 2, NVAL, 2
         If (XVALT(IIND-1) < XVALT(IIND)) Then
            IRNGT (IIND-1) = IIND - 1
            IRNGT (IIND) = IIND
         Else
            IRNGT (IIND-1) = IIND
            IRNGT (IIND) = IIND - 1
         End If
      End Do
      If (Modulo(NVAL, 2) /= 0) Then
         IRNGT (NVAL) = NVAL
      End If
!
!  We will now have ordered subsets A - B - A - B - ...
!  and merge A and B couples into     C   -   C   - ...
!
      LMTNA = 2
      LMTNC = 4
!
!  First iteration. The length of the ordered subsets goes from 2 to 4
!
      Do
         If (NVAL <= 4) Exit
!
!   Loop on merges of A and B into C
!
         Do IWRKD = 0, NVAL - 1, 4
            If ((IWRKD+4) > NVAL) Then
               If ((IWRKD+2) >= NVAL) Exit
!
!   1 2 3
!
               If (XVALT(IRNGT(IWRKD+2)) <= XVALT(IRNGT(IWRKD+3))) Exit
!
!   1 3 2
!
               If (XVALT(IRNGT(IWRKD+1)) <= XVALT(IRNGT(IWRKD+3))) Then
                  IRNG2 = IRNGT (IWRKD+2)
                  IRNGT (IWRKD+2) = IRNGT (IWRKD+3)
                  IRNGT (IWRKD+3) = IRNG2
!
!   3 1 2
!
               Else
                  IRNG1 = IRNGT (IWRKD+1)
                  IRNGT (IWRKD+1) = IRNGT (IWRKD+3)
                  IRNGT (IWRKD+3) = IRNGT (IWRKD+2)
                  IRNGT (IWRKD+2) = IRNG1
               End If
               Exit
            End If
!
!   1 2 3 4
!
            If (XVALT(IRNGT(IWRKD+2)) <= XVALT(IRNGT(IWRKD+3))) Cycle
!
!   1 3 x x
!
            If (XVALT(IRNGT(IWRKD+1)) <= XVALT(IRNGT(IWRKD+3))) Then
               IRNG2 = IRNGT (IWRKD+2)
               IRNGT (IWRKD+2) = IRNGT (IWRKD+3)
               If (XVALT(IRNG2) <= XVALT(IRNGT(IWRKD+4))) Then
!   1 3 2 4
                  IRNGT (IWRKD+3) = IRNG2
               Else
!   1 3 4 2
                  IRNGT (IWRKD+3) = IRNGT (IWRKD+4)
                  IRNGT (IWRKD+4) = IRNG2
               End If
!
!   3 x x x
!
            Else
               IRNG1 = IRNGT (IWRKD+1)
               IRNG2 = IRNGT (IWRKD+2)
               IRNGT (IWRKD+1) = IRNGT (IWRKD+3)
               If (XVALT(IRNG1) <= XVALT(IRNGT(IWRKD+4))) Then
                  IRNGT (IWRKD+2) = IRNG1
                  If (XVALT(IRNG2) <= XVALT(IRNGT(IWRKD+4))) Then
!   3 1 2 4
                     IRNGT (IWRKD+3) = IRNG2
                  Else
!   3 1 4 2
                     IRNGT (IWRKD+3) = IRNGT (IWRKD+4)
                     IRNGT (IWRKD+4) = IRNG2
                  End If
               Else
!   3 4 1 2
                  IRNGT (IWRKD+2) = IRNGT (IWRKD+4)
                  IRNGT (IWRKD+3) = IRNG1
                  IRNGT (IWRKD+4) = IRNG2
               End If
            End If
         End Do
!
!  The Cs become As and Bs
!
         LMTNA = 4
         Exit
      End Do
!
!  Iteration loop. Each time, the length of the ordered subsets
!  is doubled.
!
      Do
         If (2*LMTNA >= NVAL) Exit
         IWRKF = 0
         LMTNC = 2 * LMTNC
!
!   Loop on merges of A and B into C
!
         Do
            IWRK = IWRKF
            IWRKD = IWRKF + 1
            JINDA = IWRKF + LMTNA
            IWRKF = IWRKF + LMTNC
            If (IWRKF >= NVAL) Then
               If (JINDA >= NVAL) Exit
               IWRKF = NVAL
            End If
            IINDA = 1
            IINDB = JINDA + 1
!
!  One steps in the C subset, that we create in the final rank array
!
!  Make a copy of the rank array for the iteration
!
            JWRKT (1:LMTNA) = IRNGT (IWRKD:JINDA)
            XVALA = XVALT (JWRKT(IINDA))
            XVALB = XVALT (IRNGT(IINDB))
!
            Do
               IWRK = IWRK + 1
!
!  We still have unprocessed values in both A and B
!
               If (XVALA > XVALB) Then
                  IRNGT (IWRK) = IRNGT (IINDB)
                  IINDB = IINDB + 1
                  If (IINDB > IWRKF) Then
!  Only A still with unprocessed values
                     IRNGT (IWRK+1:IWRKF) = JWRKT (IINDA:LMTNA)
                     Exit
                  End If
                  XVALB = XVALT (IRNGT(IINDB))
               Else
                  IRNGT (IWRK) = JWRKT (IINDA)
                  IINDA = IINDA + 1
                  If (IINDA > LMTNA) Exit! Only B still with unprocessed values
                  XVALA = XVALT (JWRKT(IINDA))
               End If
!
            End Do
         End Do
!
!  The Cs become As and Bs
!
         LMTNA = 2 * LMTNA
      End Do
!
!   Last merge of A and B into C, with removal of duplicates.
!
      IINDA = 1
      IINDB = LMTNA + 1
      NUNI = 0
!
!  One steps in the C subset, that we create in the final rank array
!
      JWRKT (1:LMTNA) = IRNGT (1:LMTNA)
      If (IINDB <= NVAL) Then
        XTST = NEARLESS (Min(XVALT(JWRKT(1)), XVALT(IRNGT(IINDB))))
      Else
        XTST = NEARLESS (XVALT(JWRKT(1)))
      Endif
      Do IWRK = 1, NVAL
!
!  We still have unprocessed values in both A and B
!
         If (IINDA <= LMTNA) Then
            If (IINDB <= NVAL) Then
               If (XVALT(JWRKT(IINDA)) > XVALT(IRNGT(IINDB))) Then
                  IRNG = IRNGT (IINDB)
                  IINDB = IINDB + 1
               Else
                  IRNG = JWRKT (IINDA)
                  IINDA = IINDA + 1
               End If
            Else
!
!  Only A still with unprocessed values
!
               IRNG = JWRKT (IINDA)
               IINDA = IINDA + 1
            End If
         Else
!
!  Only B still with unprocessed values
!
            IRNG = IRNGT (IWRK)
         End If
         If (XVALT(IRNG) > XTST) Then
            XTST = XVALT (IRNG)
            NUNI = NUNI + 1
            IRNGT (NUNI) = IRNG
         End If
!
      End Do
!
      Return
!
End Subroutine R_unirnk
Subroutine I_unirnk (XVALT, IRNGT, NUNI)
! __________________________________________________________
!   UNIRNK = Merge-sort ranking of an array, with removal of
!   duplicate entries.
!   The routine is similar to pure merge-sort ranking, but on
!   the last pass, it discards indices that correspond to
!   duplicate entries.
!   For performance reasons, the first 2 passes are taken
!   out of the standard loop, and use dedicated coding.
! __________________________________________________________
! __________________________________________________________
      Integer, Dimension (:), Intent (In) :: XVALT
      Integer, Dimension (:), Intent (Out) :: IRNGT
      Integer, Intent (Out) :: NUNI
! __________________________________________________________
      Integer, Dimension (SIZE(IRNGT)) :: JWRKT
      Integer :: LMTNA, LMTNC, IRNG, IRNG1, IRNG2
      Integer :: NVAL, IIND, IWRKD, IWRK, IWRKF, JINDA, IINDA, IINDB
      Integer :: XTST, XVALA, XVALB
!
!
      NVAL = Min (SIZE(XVALT), SIZE(IRNGT))
      NUNI = NVAL
!
      Select Case (NVAL)
      Case (:0)
         Return
      Case (1)
         IRNGT (1) = 1
         Return
      Case Default
         Continue
      End Select
!
!  Fill-in the index array, creating ordered couples
!
      Do IIND = 2, NVAL, 2
         If (XVALT(IIND-1) < XVALT(IIND)) Then
            IRNGT (IIND-1) = IIND - 1
            IRNGT (IIND) = IIND
         Else
            IRNGT (IIND-1) = IIND
            IRNGT (IIND) = IIND - 1
         End If
      End Do
      If (Modulo(NVAL, 2) /= 0) Then
         IRNGT (NVAL) = NVAL
      End If
!
!  We will now have ordered subsets A - B - A - B - ...
!  and merge A and B couples into     C   -   C   - ...
!
      LMTNA = 2
      LMTNC = 4
!
!  First iteration. The length of the ordered subsets goes from 2 to 4
!
      Do
         If (NVAL <= 4) Exit
!
!   Loop on merges of A and B into C
!
         Do IWRKD = 0, NVAL - 1, 4
            If ((IWRKD+4) > NVAL) Then
               If ((IWRKD+2) >= NVAL) Exit
!
!   1 2 3
!
               If (XVALT(IRNGT(IWRKD+2)) <= XVALT(IRNGT(IWRKD+3))) Exit
!
!   1 3 2
!
               If (XVALT(IRNGT(IWRKD+1)) <= XVALT(IRNGT(IWRKD+3))) Then
                  IRNG2 = IRNGT (IWRKD+2)
                  IRNGT (IWRKD+2) = IRNGT (IWRKD+3)
                  IRNGT (IWRKD+3) = IRNG2
!
!   3 1 2
!
               Else
                  IRNG1 = IRNGT (IWRKD+1)
                  IRNGT (IWRKD+1) = IRNGT (IWRKD+3)
                  IRNGT (IWRKD+3) = IRNGT (IWRKD+2)
                  IRNGT (IWRKD+2) = IRNG1
               End If
               Exit
            End If
!
!   1 2 3 4
!
            If (XVALT(IRNGT(IWRKD+2)) <= XVALT(IRNGT(IWRKD+3))) Cycle
!
!   1 3 x x
!
            If (XVALT(IRNGT(IWRKD+1)) <= XVALT(IRNGT(IWRKD+3))) Then
               IRNG2 = IRNGT (IWRKD+2)
               IRNGT (IWRKD+2) = IRNGT (IWRKD+3)
               If (XVALT(IRNG2) <= XVALT(IRNGT(IWRKD+4))) Then
!   1 3 2 4
                  IRNGT (IWRKD+3) = IRNG2
               Else
!   1 3 4 2
                  IRNGT (IWRKD+3) = IRNGT (IWRKD+4)
                  IRNGT (IWRKD+4) = IRNG2
               End If
!
!   3 x x x
!
            Else
               IRNG1 = IRNGT (IWRKD+1)
               IRNG2 = IRNGT (IWRKD+2)
               IRNGT (IWRKD+1) = IRNGT (IWRKD+3)
               If (XVALT(IRNG1) <= XVALT(IRNGT(IWRKD+4))) Then
                  IRNGT (IWRKD+2) = IRNG1
                  If (XVALT(IRNG2) <= XVALT(IRNGT(IWRKD+4))) Then
!   3 1 2 4
                     IRNGT (IWRKD+3) = IRNG2
                  Else
!   3 1 4 2
                     IRNGT (IWRKD+3) = IRNGT (IWRKD+4)
                     IRNGT (IWRKD+4) = IRNG2
                  End If
               Else
!   3 4 1 2
                  IRNGT (IWRKD+2) = IRNGT (IWRKD+4)
                  IRNGT (IWRKD+3) = IRNG1
                  IRNGT (IWRKD+4) = IRNG2
               End If
            End If
         End Do
!
!  The Cs become As and Bs
!
         LMTNA = 4
         Exit
      End Do
!
!  Iteration loop. Each time, the length of the ordered subsets
!  is doubled.
!
      Do
         If (2*LMTNA >= NVAL) Exit
         IWRKF = 0
         LMTNC = 2 * LMTNC
!
!   Loop on merges of A and B into C
!
         Do
            IWRK = IWRKF
            IWRKD = IWRKF + 1
            JINDA = IWRKF + LMTNA
            IWRKF = IWRKF + LMTNC
            If (IWRKF >= NVAL) Then
               If (JINDA >= NVAL) Exit
               IWRKF = NVAL
            End If
            IINDA = 1
            IINDB = JINDA + 1
!
!  One steps in the C subset, that we create in the final rank array
!
!  Make a copy of the rank array for the iteration
!
            JWRKT (1:LMTNA) = IRNGT (IWRKD:JINDA)
            XVALA = XVALT (JWRKT(IINDA))
            XVALB = XVALT (IRNGT(IINDB))
!
            Do
               IWRK = IWRK + 1
!
!  We still have unprocessed values in both A and B
!
               If (XVALA > XVALB) Then
                  IRNGT (IWRK) = IRNGT (IINDB)
                  IINDB = IINDB + 1
                  If (IINDB > IWRKF) Then
!  Only A still with unprocessed values
                     IRNGT (IWRK+1:IWRKF) = JWRKT (IINDA:LMTNA)
                     Exit
                  End If
                  XVALB = XVALT (IRNGT(IINDB))
               Else
                  IRNGT (IWRK) = JWRKT (IINDA)
                  IINDA = IINDA + 1
                  If (IINDA > LMTNA) Exit! Only B still with unprocessed values
                  XVALA = XVALT (JWRKT(IINDA))
               End If
!
            End Do
         End Do
!
!  The Cs become As and Bs
!
         LMTNA = 2 * LMTNA
      End Do
!
!   Last merge of A and B into C, with removal of duplicates.
!
      IINDA = 1
      IINDB = LMTNA + 1
      NUNI = 0
!
!  One steps in the C subset, that we create in the final rank array
!
      JWRKT (1:LMTNA) = IRNGT (1:LMTNA)
      If (IINDB <= NVAL) Then
        XTST = NEARLESS (Min(XVALT(JWRKT(1)), XVALT(IRNGT(IINDB))))
      Else
        XTST = NEARLESS (XVALT(JWRKT(1)))
      Endif
      Do IWRK = 1, NVAL
!
!  We still have unprocessed values in both A and B
!
         If (IINDA <= LMTNA) Then
            If (IINDB <= NVAL) Then
               If (XVALT(JWRKT(IINDA)) > XVALT(IRNGT(IINDB))) Then
                  IRNG = IRNGT (IINDB)
                  IINDB = IINDB + 1
               Else
                  IRNG = JWRKT (IINDA)
                  IINDA = IINDA + 1
               End If
            Else
!
!  Only A still with unprocessed values
!
               IRNG = JWRKT (IINDA)
               IINDA = IINDA + 1
            End If
         Else
!
!  Only B still with unprocessed values
!
            IRNG = IRNGT (IWRK)
         End If
         If (XVALT(IRNG) > XTST) Then
            XTST = XVALT (IRNG)
            NUNI = NUNI + 1
            IRNGT (NUNI) = IRNG
         End If
!
      End Do
!
      Return
!
End Subroutine I_unirnk

Function D_nearless (XVAL) result (D_nl)
!  Nearest value less than given value
! __________________________________________________________
      Real (kind=kdp), Intent (In) :: XVAL
      Real (kind=kdp) :: D_nl
! __________________________________________________________
      D_nl = nearest (XVAL, -1.0_kdp)
      return
!
End Function D_nearless
Function R_nearless (XVAL) result (R_nl)
!  Nearest value less than given value
! __________________________________________________________
      Real, Intent (In) :: XVAL
      Real :: R_nl
! __________________________________________________________
      R_nl = nearest (XVAL, -1.0)
      return
!
End Function R_nearless
Function I_nearless (XVAL) result (I_nl)
!  Nearest value less than given value
! __________________________________________________________
      Integer, Intent (In) :: XVAL
      Integer :: I_nl
! __________________________________________________________
      I_nl = XVAL - 1
      return
!
End Function I_nearless

end module m_unirnk
Module m_unista
Use m_uniinv
Private
Integer, Parameter :: kdp = selected_real_kind(15)
public :: unista
private :: kdp
private :: R_unista, I_unista, D_unista
interface unista
  module procedure d_unista, r_unista, i_unista
end interface unista
contains

Subroutine D_unista (XDONT, NUNI)
!   UNISTA = (Stable unique) Removes duplicates from an array,
!            leaving unique entries in the order of their first
!            appearance in the initial set.
!  Michel Olagnon - Feb. 2000
! __________________________________________________________
! __________________________________________________________
      Real (kind=kdp), Dimension (:), Intent (InOut) :: XDONT
      Integer, Intent (Out) :: NUNI
! __________________________________________________________
!
      Integer, Dimension (Size(XDONT)) :: IWRKT
      Logical, Dimension (Size(XDONT)) :: IFMPTYT
      Integer :: ICRS
! __________________________________________________________
      Call UNIINV (XDONT, IWRKT)
      IFMPTYT = .True.
      NUNI = 0
      Do ICRS = 1, Size(XDONT)
         If (IFMPTYT(IWRKT(ICRS))) Then
            IFMPTYT(IWRKT(ICRS)) = .False.
            NUNI = NUNI + 1
            XDONT (NUNI) = XDONT (ICRS)
         End If
      End Do
      Return
!
End Subroutine D_unista

Subroutine R_unista (XDONT, NUNI)
!   UNISTA = (Stable unique) Removes duplicates from an array,
!            leaving unique entries in the order of their first
!            appearance in the initial set.
!  Michel Olagnon - Feb. 2000
! __________________________________________________________
! _________________________________________________________
      Real, Dimension (:), Intent (InOut) :: XDONT
      Integer, Intent (Out) :: NUNI
! __________________________________________________________
!
      Integer, Dimension (Size(XDONT)) :: IWRKT
      Logical, Dimension (Size(XDONT)) :: IFMPTYT
      Integer :: ICRS
! __________________________________________________________
      Call UNIINV (XDONT, IWRKT)
      IFMPTYT = .True.
      NUNI = 0
      Do ICRS = 1, Size(XDONT)
         If (IFMPTYT(IWRKT(ICRS))) Then
            IFMPTYT(IWRKT(ICRS)) = .False.
            NUNI = NUNI + 1
            XDONT (NUNI) = XDONT (ICRS)
         End If
      End Do
      Return
!
End Subroutine R_unista
Subroutine I_unista (XDONT, NUNI)
!   UNISTA = (Stable unique) Removes duplicates from an array,
!            leaving unique entries in the order of their first
!            appearance in the initial set.
!  Michel Olagnon - Feb. 2000
! __________________________________________________________
! __________________________________________________________
      Integer, Dimension (:), Intent (InOut)  :: XDONT
      Integer, Intent (Out) :: NUNI
! __________________________________________________________
!
      Integer, Dimension (Size(XDONT)) :: IWRKT
      Logical, Dimension (Size(XDONT)) :: IFMPTYT
      Integer :: ICRS
! __________________________________________________________
      Call UNIINV (XDONT, IWRKT)
      IFMPTYT = .True.
      NUNI = 0
      Do ICRS = 1, Size(XDONT)
         If (IFMPTYT(IWRKT(ICRS))) Then
            IFMPTYT(IWRKT(ICRS)) = .False.
            NUNI = NUNI + 1
            XDONT (NUNI) = XDONT (ICRS)
         End If
      End Do
      Return
!
End Subroutine I_unista
end module m_unista
Module m_valmed
Integer, Parameter :: kdp = selected_real_kind(15)
public :: valmed
private :: kdp
private :: R_valmed, I_valmed, D_valmed
interface valmed
  module procedure d_valmed, r_valmed, i_valmed
end interface valmed
contains

Recursive Function D_valmed (XDONT) Result (res_med)
!  Finds the median of XDONT using the recursive procedure
!  described in Knuth, The Art of Computer Programming,
!  vol. 3, 5.3.3 - This procedure is linear in time, and
!  does not require to be able to interpolate in the
!  set as the one used in INDNTH. It also has better worst
!  case behavior than INDNTH, but is about 30% slower in
!  average for random uniformly distributed values.
! __________________________________________________________
! __________________________________________________________
      Real (kind=kdp), Dimension (:), Intent (In) :: XDONT
      Real (kind=kdp) :: res_med
! __________________________________________________________
      Real (kind=kdp), Parameter :: XHUGE = HUGE (XDONT)
      Real (kind=kdp), Dimension (SIZE(XDONT)+6) :: XWRKT
      Real (kind=kdp) :: XWRK, XWRK1, XMED7
!
      Integer, Dimension ((SIZE(XDONT)+6)/7) :: ISTRT, IENDT, IMEDT
      Integer :: NDON, NTRI, NMED, NORD, NEQU, NLEQ, IMED, IDON, IDON1
      Integer :: IDEB, IWRK, IDCR, ICRS, ICRS1, ICRS2, IMED1
!
      NDON = SIZE (XDONT)
      NMED = (NDON+1) / 2
!      write(unit=*,fmt=*) NMED, NDON
!
!  If the number of values is small, then use insertion sort
!
      If (NDON < 35) Then
!
!  Bring minimum to first location to save test in decreasing loop
!
         IDCR = NDON
         If (XDONT (1) < XDONT (NDON)) Then
            XWRK = XDONT (1)
            XWRKT (IDCR) = XDONT (IDCR)
         Else
            XWRK = XDONT (IDCR)
            XWRKT (IDCR) = XDONT (1)
         Endif
         Do IWRK = 1, NDON - 2
            IDCR = IDCR - 1
            XWRK1 = XDONT (IDCR)
            If (XWRK1 < XWRK) Then
                XWRKT (IDCR) = XWRK
                XWRK = XWRK1
            Else
                XWRKT (IDCR) = XWRK1
            Endif
         End Do
         XWRKT (1) = XWRK
!
! Sort the first half, until we have NMED sorted values
!
         Do ICRS = 3, NMED
            XWRK = XWRKT (ICRS)
               IDCR = ICRS - 1
               Do
                  If (XWRK >= XWRKT(IDCR)) Exit
                  XWRKT (IDCR+1) = XWRKT (IDCR)
                  IDCR = IDCR - 1
               End Do
            XWRKT (IDCR+1) = XWRK
         End Do
!
!  Insert any value less than the current median in the first half
!
         Do ICRS = NMED+1, NDON
            XWRK = XWRKT (ICRS)
            If (XWRK < XWRKT (NMED)) Then
               IDCR = NMED - 1
               Do
                  If (XWRK >= XWRKT(IDCR)) Exit
                  XWRKT (IDCR+1) = XWRKT (IDCR)
                  IDCR = IDCR - 1
               End Do
               XWRKT (IDCR+1) = XWRK
            End If
         End Do
         res_med = XWRKT (NMED)
         Return
      End If
!
!  Make sorted subsets of 7 elements
!  This is done by a variant of insertion sort where a first
!  pass is used to bring the smallest element to the first position
!  decreasing disorder at the same time, so that we may remove
!  remove the loop test in the insertion loop.
!
      DO IDEB = 1, NDON-6, 7
         IDCR = IDEB + 6
         If (XDONT (IDEB) < XDONT (IDCR)) Then
            XWRK = XDONT (IDEB)
            XWRKT (IDCR) = XDONT (IDCR)
         Else
            XWRK = XDONT (IDCR)
            XWRKT (IDCR) = XDONT (IDEB)
         Endif
         Do IWRK = 1, 5
            IDCR = IDCR - 1
            XWRK1 = XDONT (IDCR)
            If (XWRK1 < XWRK) Then
                XWRKT (IDCR) = XWRK
                XWRK = XWRK1
            Else
                XWRKT (IDCR) = XWRK1
            Endif
         End Do
         XWRKT (IDEB) = XWRK
         Do ICRS = IDEB+2, IDEB+6
            XWRK = XWRKT (ICRS)
            If (XWRK < XWRKT(ICRS-1)) Then
               XWRKT (ICRS) = XWRKT (ICRS-1)
               IDCR = ICRS - 1
               XWRK1 = XWRKT (IDCR-1)
               Do
                  If (XWRK >= XWRK1) Exit
                  XWRKT (IDCR) = XWRK1
                  IDCR = IDCR - 1
                  XWRK1 = XWRKT (IDCR-1)
               End Do
               XWRKT (IDCR) = XWRK
            EndIf
         End Do
      End Do
!
!  Add-up alternatively + and - HUGE values to make the number of data
!  an exact multiple of 7.
!
      IDEB = 7 * (NDON/7)
      NTRI = NDON
      If (IDEB < NDON) Then
!
         XWRK1 = XHUGE
         Do ICRS = IDEB+1, IDEB+7
            If (ICRS <= NDON) Then
               XWRKT (ICRS) = XDONT (ICRS)
            Else
               If (XWRK1 /= XHUGE) NMED = NMED + 1
               XWRKT (ICRS) = XWRK1
               XWRK1 = - XWRK1
            Endif
         End Do
!
         Do ICRS = IDEB+2, IDEB+7
            XWRK = XWRKT (ICRS)
            Do IDCR = ICRS - 1, IDEB+1, - 1
               If (XWRK >= XWRKT(IDCR)) Exit
               XWRKT (IDCR+1) = XWRKT (IDCR)
            End Do
            XWRKT (IDCR+1) = XWRK
         End Do
!
         NTRI = IDEB+7
      End If
!
!  Make the set of the indices of median values of each sorted subset
!
         IDON1 = 0
         Do IDON = 1, NTRI, 7
            IDON1 = IDON1 + 1
            IMEDT (IDON1) = IDON + 3
         End Do
!
!  Find XMED7, the median of the medians
!
         XMED7 = D_valmed (XWRKT (IMEDT))
!
!  Count how many values are not higher than (and how many equal to) XMED7
!  This number is at least 4 * 1/2 * (N/7) : 4 values in each of the
!  subsets where the median is lower than the median of medians. For similar
!  reasons, we also have at least 2N/7 values not lower than XMED7. At the
!  same time, we find in each subset the index of the last value < XMED7,
!  and that of the first > XMED7. These indices will be used to restrict the
!  search for the median as the Kth element in the subset (> or <) where
!  we know it to be.
!
         IDON1 = 1
         NLEQ = 0
         NEQU = 0
         Do IDON = 1, NTRI, 7
            IMED = IDON+3
            If (XWRKT (IMED) > XMED7) Then
                  IMED = IMED - 2
                  If (XWRKT (IMED) > XMED7) Then
                     IMED = IMED - 1
                  Else If (XWRKT (IMED) < XMED7) Then
                     IMED = IMED + 1
                  Endif
            Else If (XWRKT (IMED) < XMED7) Then
                  IMED = IMED + 2
                  If (XWRKT (IMED) > XMED7) Then
                     IMED = IMED - 1
                  Else If (XWRKT (IMED) < XMED7) Then
                     IMED = IMED + 1
                  Endif
            Endif
            If (XWRKT (IMED) > XMED7) Then
               NLEQ = NLEQ + IMED - IDON
               IENDT (IDON1) = IMED - 1
               ISTRT (IDON1) = IMED
            Else If (XWRKT (IMED) < XMED7) Then
               NLEQ = NLEQ + IMED - IDON + 1
               IENDT (IDON1) = IMED
               ISTRT (IDON1) = IMED + 1
            Else                    !       If (XWRKT (IMED) == XMED7)
               NLEQ = NLEQ + IMED - IDON + 1
               NEQU = NEQU + 1
               IENDT (IDON1) = IMED - 1
               Do IMED1 = IMED - 1, IDON, -1
                  If (XWRKT (IMED1) == XMED7) Then
                     NEQU = NEQU + 1
                     IENDT (IDON1) = IMED1 - 1
                  Else
                     Exit
                  End If
               End Do
               ISTRT (IDON1) = IMED + 1
               Do IMED1 = IMED + 1, IDON + 6
                  If (XWRKT (IMED1) == XMED7) Then
                     NEQU = NEQU + 1
                     NLEQ = NLEQ + 1
                     ISTRT (IDON1) = IMED1 + 1
                  Else
                     Exit
                  End If
               End Do
            Endif
            IDON1 = IDON1 + 1
         End Do
!
!  Carry out a partial insertion sort to find the Kth smallest of the
!  large values, or the Kth largest of the small values, according to
!  what is needed.
!
        If (NLEQ - NEQU + 1 <= NMED) Then
            If (NLEQ < NMED) Then   !      Not enough low values
                XWRK1 = XHUGE
                NORD = NMED - NLEQ
                IDON1 = 0
                ICRS1 = 1
                ICRS2 = 0
                IDCR = 0
               Do IDON = 1, NTRI, 7
                   IDON1 = IDON1 + 1
                   If (ICRS2 < NORD) Then
                      Do ICRS = ISTRT (IDON1), IDON + 6
                         If (XWRKT(ICRS) < XWRK1) Then
                            XWRK = XWRKT (ICRS)
                            Do IDCR = ICRS1 - 1, 1, - 1
                               If (XWRK >= XWRKT(IDCR)) Exit
                               XWRKT (IDCR+1) = XWRKT (IDCR)
                            End Do
                            XWRKT (IDCR+1) = XWRK
                            XWRK1 = XWRKT(ICRS1)
                         Else
                           If (ICRS2 < NORD) Then
                              XWRKT (ICRS1) = XWRKT (ICRS)
                              XWRK1 = XWRKT(ICRS1)
                           Endif
                         End If
                         ICRS1 = MIN (NORD, ICRS1 + 1)
                         ICRS2 = MIN (NORD, ICRS2 + 1)
                      End Do
                   Else
                      Do ICRS = ISTRT (IDON1), IDON + 6
                         If (XWRKT(ICRS) >= XWRK1) Exit
                         XWRK = XWRKT (ICRS)
                         Do IDCR = ICRS1 - 1, 1, - 1
                               If (XWRK >= XWRKT(IDCR)) Exit
                               XWRKT (IDCR+1) = XWRKT (IDCR)
                         End Do
                         XWRKT (IDCR+1) = XWRK
                         XWRK1 = XWRKT(ICRS1)
                      End Do
                   End If
                End Do
                res_med = XWRK1
                Return
            Else
                res_med = XMED7
                Return
            End If
         Else                       !      If (NLEQ > NMED)
!                                          Not enough high values
                XWRK1 = -XHUGE
                NORD = NLEQ - NEQU - NMED + 1
                IDON1 = 0
                ICRS1 = 1
                ICRS2 = 0
                Do IDON = 1, NTRI, 7
                   IDON1 = IDON1 + 1
                   If (ICRS2 < NORD) Then
!
                      Do ICRS = IDON, IENDT (IDON1)
                         If (XWRKT(ICRS) > XWRK1) Then
                            XWRK = XWRKT (ICRS)
                            IDCR = ICRS1 - 1
                            Do IDCR = ICRS1 - 1, 1, - 1
                               If (XWRK <= XWRKT(IDCR)) Exit
                               XWRKT (IDCR+1) = XWRKT (IDCR)
                            End Do
                            XWRKT (IDCR+1) = XWRK
                            XWRK1 = XWRKT(ICRS1)
                         Else
                            If (ICRS2 < NORD) Then
                               XWRKT (ICRS1) = XWRKT (ICRS)
                               XWRK1 = XWRKT (ICRS1)
                            End If
                         End If
                         ICRS1 = MIN (NORD, ICRS1 + 1)
                         ICRS2 = MIN (NORD, ICRS2 + 1)
                      End Do
                   Else
                      Do ICRS = IENDT (IDON1), IDON, -1
                         If (XWRKT(ICRS) > XWRK1) Then
                            XWRK = XWRKT (ICRS)
                            IDCR = ICRS1 - 1
                            Do IDCR = ICRS1 - 1, 1, - 1
                               If (XWRK <= XWRKT(IDCR)) Exit
                               XWRKT (IDCR+1) = XWRKT (IDCR)
                            End Do
                            XWRKT (IDCR+1) = XWRK
                            XWRK1 = XWRKT(ICRS1)
                         Else
                            Exit
                         End If
                      End Do
                   Endif
                End Do
!
                res_med = XWRK1
                Return
         End If
!
End Function D_valmed

Recursive Function R_valmed (XDONT) Result (res_med)
!  Finds the median of XDONT using the recursive procedure
!  described in Knuth, The Art of Computer Programming,
!  vol. 3, 5.3.3 - This procedure is linear in time, and
!  does not require to be able to interpolate in the
!  set as the one used in INDNTH. It also has better worst
!  case behavior than INDNTH, but is about 30% slower in
!  average for random uniformly distributed values.
! __________________________________________________________
! _________________________________________________________
      Real, Dimension (:), Intent (In) :: XDONT
      Real :: res_med
! __________________________________________________________
      Real, Parameter :: XHUGE = HUGE (XDONT)
      Real, Dimension (SIZE(XDONT)+6) :: XWRKT
      Real :: XWRK, XWRK1, XMED7
!
      Integer, Dimension ((SIZE(XDONT)+6)/7) :: ISTRT, IENDT, IMEDT
      Integer :: NDON, NTRI, NMED, NORD, NEQU, NLEQ, IMED, IDON, IDON1
      Integer :: IDEB, IWRK, IDCR, ICRS, ICRS1, ICRS2, IMED1
!
      NDON = SIZE (XDONT)
      NMED = (NDON+1) / 2
!      write(unit=*,fmt=*) NMED, NDON
!
!  If the number of values is small, then use insertion sort
!
      If (NDON < 35) Then
!
!  Bring minimum to first location to save test in decreasing loop
!
         IDCR = NDON
         If (XDONT (1) < XDONT (NDON)) Then
            XWRK = XDONT (1)
            XWRKT (IDCR) = XDONT (IDCR)
         Else
            XWRK = XDONT (IDCR)
            XWRKT (IDCR) = XDONT (1)
         Endif
         Do IWRK = 1, NDON - 2
            IDCR = IDCR - 1
            XWRK1 = XDONT (IDCR)
            If (XWRK1 < XWRK) Then
                XWRKT (IDCR) = XWRK
                XWRK = XWRK1
            Else
                XWRKT (IDCR) = XWRK1
            Endif
         End Do
         XWRKT (1) = XWRK
!
! Sort the first half, until we have NMED sorted values
!
         Do ICRS = 3, NMED
            XWRK = XWRKT (ICRS)
               IDCR = ICRS - 1
               Do
                  If (XWRK >= XWRKT(IDCR)) Exit
                  XWRKT (IDCR+1) = XWRKT (IDCR)
                  IDCR = IDCR - 1
               End Do
            XWRKT (IDCR+1) = XWRK
         End Do
!
!  Insert any value less than the current median in the first half
!
         Do ICRS = NMED+1, NDON
            XWRK = XWRKT (ICRS)
            If (XWRK < XWRKT (NMED)) Then
               IDCR = NMED - 1
               Do
                  If (XWRK >= XWRKT(IDCR)) Exit
                  XWRKT (IDCR+1) = XWRKT (IDCR)
                  IDCR = IDCR - 1
               End Do
               XWRKT (IDCR+1) = XWRK
            End If
         End Do
         res_med = XWRKT (NMED)
         Return
      End If
!
!  Make sorted subsets of 7 elements
!  This is done by a variant of insertion sort where a first
!  pass is used to bring the smallest element to the first position
!  decreasing disorder at the same time, so that we may remove
!  remove the loop test in the insertion loop.
!
      DO IDEB = 1, NDON-6, 7
         IDCR = IDEB + 6
         If (XDONT (IDEB) < XDONT (IDCR)) Then
            XWRK = XDONT (IDEB)
            XWRKT (IDCR) = XDONT (IDCR)
         Else
            XWRK = XDONT (IDCR)
            XWRKT (IDCR) = XDONT (IDEB)
         Endif
         Do IWRK = 1, 5
            IDCR = IDCR - 1
            XWRK1 = XDONT (IDCR)
            If (XWRK1 < XWRK) Then
                XWRKT (IDCR) = XWRK
                XWRK = XWRK1
            Else
                XWRKT (IDCR) = XWRK1
            Endif
         End Do
         XWRKT (IDEB) = XWRK
         Do ICRS = IDEB+2, IDEB+6
            XWRK = XWRKT (ICRS)
            If (XWRK < XWRKT(ICRS-1)) Then
               XWRKT (ICRS) = XWRKT (ICRS-1)
               IDCR = ICRS - 1
               XWRK1 = XWRKT (IDCR-1)
               Do
                  If (XWRK >= XWRK1) Exit
                  XWRKT (IDCR) = XWRK1
                  IDCR = IDCR - 1
                  XWRK1 = XWRKT (IDCR-1)
               End Do
               XWRKT (IDCR) = XWRK
            EndIf
         End Do
      End Do
!
!  Add-up alternatively + and - HUGE values to make the number of data
!  an exact multiple of 7.
!
      IDEB = 7 * (NDON/7)
      NTRI = NDON
      If (IDEB < NDON) Then
!
         XWRK1 = XHUGE
         Do ICRS = IDEB+1, IDEB+7
            If (ICRS <= NDON) Then
               XWRKT (ICRS) = XDONT (ICRS)
            Else
               If (XWRK1 /= XHUGE) NMED = NMED + 1
               XWRKT (ICRS) = XWRK1
               XWRK1 = - XWRK1
            Endif
         End Do
!
         Do ICRS = IDEB+2, IDEB+7
            XWRK = XWRKT (ICRS)
            Do IDCR = ICRS - 1, IDEB+1, - 1
               If (XWRK >= XWRKT(IDCR)) Exit
               XWRKT (IDCR+1) = XWRKT (IDCR)
            End Do
            XWRKT (IDCR+1) = XWRK
         End Do
!
         NTRI = IDEB+7
      End If
!
!  Make the set of the indices of median values of each sorted subset
!
         IDON1 = 0
         Do IDON = 1, NTRI, 7
            IDON1 = IDON1 + 1
            IMEDT (IDON1) = IDON + 3
         End Do
!
!  Find XMED7, the median of the medians
!
         XMED7 = R_valmed (XWRKT (IMEDT))
!
!  Count how many values are not higher than (and how many equal to) XMED7
!  This number is at least 4 * 1/2 * (N/7) : 4 values in each of the
!  subsets where the median is lower than the median of medians. For similar
!  reasons, we also have at least 2N/7 values not lower than XMED7. At the
!  same time, we find in each subset the index of the last value < XMED7,
!  and that of the first > XMED7. These indices will be used to restrict the
!  search for the median as the Kth element in the subset (> or <) where
!  we know it to be.
!
         IDON1 = 1
         NLEQ = 0
         NEQU = 0
         Do IDON = 1, NTRI, 7
            IMED = IDON+3
            If (XWRKT (IMED) > XMED7) Then
                  IMED = IMED - 2
                  If (XWRKT (IMED) > XMED7) Then
                     IMED = IMED - 1
                  Else If (XWRKT (IMED) < XMED7) Then
                     IMED = IMED + 1
                  Endif
            Else If (XWRKT (IMED) < XMED7) Then
                  IMED = IMED + 2
                  If (XWRKT (IMED) > XMED7) Then
                     IMED = IMED - 1
                  Else If (XWRKT (IMED) < XMED7) Then
                     IMED = IMED + 1
                  Endif
            Endif
            If (XWRKT (IMED) > XMED7) Then
               NLEQ = NLEQ + IMED - IDON
               IENDT (IDON1) = IMED - 1
               ISTRT (IDON1) = IMED
            Else If (XWRKT (IMED) < XMED7) Then
               NLEQ = NLEQ + IMED - IDON + 1
               IENDT (IDON1) = IMED
               ISTRT (IDON1) = IMED + 1
            Else                    !       If (XWRKT (IMED) == XMED7)
               NLEQ = NLEQ + IMED - IDON + 1
               NEQU = NEQU + 1
               IENDT (IDON1) = IMED - 1
               Do IMED1 = IMED - 1, IDON, -1
                  If (XWRKT (IMED1) == XMED7) Then
                     NEQU = NEQU + 1
                     IENDT (IDON1) = IMED1 - 1
                  Else
                     Exit
                  End If
               End Do
               ISTRT (IDON1) = IMED + 1
               Do IMED1 = IMED + 1, IDON + 6
                  If (XWRKT (IMED1) == XMED7) Then
                     NEQU = NEQU + 1
                     NLEQ = NLEQ + 1
                     ISTRT (IDON1) = IMED1 + 1
                  Else
                     Exit
                  End If
               End Do
            Endif
            IDON1 = IDON1 + 1
         End Do
!
!  Carry out a partial insertion sort to find the Kth smallest of the
!  large values, or the Kth largest of the small values, according to
!  what is needed.
!
        If (NLEQ - NEQU + 1 <= NMED) Then
            If (NLEQ < NMED) Then   !      Not enough low values
                XWRK1 = XHUGE
                NORD = NMED - NLEQ
                IDON1 = 0
                ICRS1 = 1
                ICRS2 = 0
                IDCR = 0
               Do IDON = 1, NTRI, 7
                   IDON1 = IDON1 + 1
                   If (ICRS2 < NORD) Then
                      Do ICRS = ISTRT (IDON1), IDON + 6
                         If (XWRKT(ICRS) < XWRK1) Then
                            XWRK = XWRKT (ICRS)
                            Do IDCR = ICRS1 - 1, 1, - 1
                               If (XWRK >= XWRKT(IDCR)) Exit
                               XWRKT (IDCR+1) = XWRKT (IDCR)
                            End Do
                            XWRKT (IDCR+1) = XWRK
                            XWRK1 = XWRKT(ICRS1)
                         Else
                           If (ICRS2 < NORD) Then
                              XWRKT (ICRS1) = XWRKT (ICRS)
                              XWRK1 = XWRKT(ICRS1)
                           Endif
                         End If
                         ICRS1 = MIN (NORD, ICRS1 + 1)
                         ICRS2 = MIN (NORD, ICRS2 + 1)
                      End Do
                   Else
                      Do ICRS = ISTRT (IDON1), IDON + 6
                         If (XWRKT(ICRS) >= XWRK1) Exit
                         XWRK = XWRKT (ICRS)
                         Do IDCR = ICRS1 - 1, 1, - 1
                               If (XWRK >= XWRKT(IDCR)) Exit
                               XWRKT (IDCR+1) = XWRKT (IDCR)
                         End Do
                         XWRKT (IDCR+1) = XWRK
                         XWRK1 = XWRKT(ICRS1)
                      End Do
                   End If
                End Do
                res_med = XWRK1
                Return
            Else
                res_med = XMED7
                Return
            End If
         Else                       !      If (NLEQ > NMED)
!                                          Not enough high values
                XWRK1 = -XHUGE
                NORD = NLEQ - NEQU - NMED + 1
                IDON1 = 0
                ICRS1 = 1
                ICRS2 = 0
                Do IDON = 1, NTRI, 7
                   IDON1 = IDON1 + 1
                   If (ICRS2 < NORD) Then
!
                      Do ICRS = IDON, IENDT (IDON1)
                         If (XWRKT(ICRS) > XWRK1) Then
                            XWRK = XWRKT (ICRS)
                            IDCR = ICRS1 - 1
                            Do IDCR = ICRS1 - 1, 1, - 1
                               If (XWRK <= XWRKT(IDCR)) Exit
                               XWRKT (IDCR+1) = XWRKT (IDCR)
                            End Do
                            XWRKT (IDCR+1) = XWRK
                            XWRK1 = XWRKT(ICRS1)
                         Else
                            If (ICRS2 < NORD) Then
                               XWRKT (ICRS1) = XWRKT (ICRS)
                               XWRK1 = XWRKT (ICRS1)
                            End If
                         End If
                         ICRS1 = MIN (NORD, ICRS1 + 1)
                         ICRS2 = MIN (NORD, ICRS2 + 1)
                      End Do
                   Else
                      Do ICRS = IENDT (IDON1), IDON, -1
                         If (XWRKT(ICRS) > XWRK1) Then
                            XWRK = XWRKT (ICRS)
                            IDCR = ICRS1 - 1
                            Do IDCR = ICRS1 - 1, 1, - 1
                               If (XWRK <= XWRKT(IDCR)) Exit
                               XWRKT (IDCR+1) = XWRKT (IDCR)
                            End Do
                            XWRKT (IDCR+1) = XWRK
                            XWRK1 = XWRKT(ICRS1)
                         Else
                            Exit
                         End If
                      End Do
                   Endif
                End Do
!
                res_med = XWRK1
                Return
         End If
!
End Function R_valmed
Recursive Function I_valmed (XDONT) Result (res_med)
!  Finds the median of XDONT using the recursive procedure
!  described in Knuth, The Art of Computer Programming,
!  vol. 3, 5.3.3 - This procedure is linear in time, and
!  does not require to be able to interpolate in the
!  set as the one used in INDNTH. It also has better worst
!  case behavior than INDNTH, but is about 30% slower in
!  average for random uniformly distributed values.
! __________________________________________________________
! __________________________________________________________
      Integer, Dimension (:), Intent (In)  :: XDONT
      Integer :: res_med
! __________________________________________________________
      Integer, Parameter :: XHUGE = HUGE (XDONT)
      Integer, Dimension (SIZE(XDONT)+6) :: XWRKT
      Integer :: XWRK, XWRK1, XMED7
!
      Integer, Dimension ((SIZE(XDONT)+6)/7) :: ISTRT, IENDT, IMEDT
      Integer :: NDON, NTRI, NMED, NORD, NEQU, NLEQ, IMED, IDON, IDON1
      Integer :: IDEB, IWRK, IDCR, ICRS, ICRS1, ICRS2, IMED1
!
      NDON = SIZE (XDONT)
      NMED = (NDON+1) / 2
!      write(unit=*,fmt=*) NMED, NDON
!
!  If the number of values is small, then use insertion sort
!
      If (NDON < 35) Then
!
!  Bring minimum to first location to save test in decreasing loop
!
         IDCR = NDON
         If (XDONT (1) < XDONT (NDON)) Then
            XWRK = XDONT (1)
            XWRKT (IDCR) = XDONT (IDCR)
         Else
            XWRK = XDONT (IDCR)
            XWRKT (IDCR) = XDONT (1)
         Endif
         Do IWRK = 1, NDON - 2
            IDCR = IDCR - 1
            XWRK1 = XDONT (IDCR)
            If (XWRK1 < XWRK) Then
                XWRKT (IDCR) = XWRK
                XWRK = XWRK1
            Else
                XWRKT (IDCR) = XWRK1
            Endif
         End Do
         XWRKT (1) = XWRK
!
! Sort the first half, until we have NMED sorted values
!
         Do ICRS = 3, NMED
            XWRK = XWRKT (ICRS)
               IDCR = ICRS - 1
               Do
                  If (XWRK >= XWRKT(IDCR)) Exit
                  XWRKT (IDCR+1) = XWRKT (IDCR)
                  IDCR = IDCR - 1
               End Do
            XWRKT (IDCR+1) = XWRK
         End Do
!
!  Insert any value less than the current median in the first half
!
         Do ICRS = NMED+1, NDON
            XWRK = XWRKT (ICRS)
            If (XWRK < XWRKT (NMED)) Then
               IDCR = NMED - 1
               Do
                  If (XWRK >= XWRKT(IDCR)) Exit
                  XWRKT (IDCR+1) = XWRKT (IDCR)
                  IDCR = IDCR - 1
               End Do
               XWRKT (IDCR+1) = XWRK
            End If
         End Do
         res_med = XWRKT (NMED)
         Return
      End If
!
!  Make sorted subsets of 7 elements
!  This is done by a variant of insertion sort where a first
!  pass is used to bring the smallest element to the first position
!  decreasing disorder at the same time, so that we may remove
!  remove the loop test in the insertion loop.
!
      DO IDEB = 1, NDON-6, 7
         IDCR = IDEB + 6
         If (XDONT (IDEB) < XDONT (IDCR)) Then
            XWRK = XDONT (IDEB)
            XWRKT (IDCR) = XDONT (IDCR)
         Else
            XWRK = XDONT (IDCR)
            XWRKT (IDCR) = XDONT (IDEB)
         Endif
         Do IWRK = 1, 5
            IDCR = IDCR - 1
            XWRK1 = XDONT (IDCR)
            If (XWRK1 < XWRK) Then
                XWRKT (IDCR) = XWRK
                XWRK = XWRK1
            Else
                XWRKT (IDCR) = XWRK1
            Endif
         End Do
         XWRKT (IDEB) = XWRK
         Do ICRS = IDEB+2, IDEB+6
            XWRK = XWRKT (ICRS)
            If (XWRK < XWRKT(ICRS-1)) Then
               XWRKT (ICRS) = XWRKT (ICRS-1)
               IDCR = ICRS - 1
               XWRK1 = XWRKT (IDCR-1)
               Do
                  If (XWRK >= XWRK1) Exit
                  XWRKT (IDCR) = XWRK1
                  IDCR = IDCR - 1
                  XWRK1 = XWRKT (IDCR-1)
               End Do
               XWRKT (IDCR) = XWRK
            EndIf
         End Do
      End Do
!
!  Add-up alternatively + and - HUGE values to make the number of data
!  an exact multiple of 7.
!
      IDEB = 7 * (NDON/7)
      NTRI = NDON
      If (IDEB < NDON) Then
!
         XWRK1 = XHUGE
         Do ICRS = IDEB+1, IDEB+7
            If (ICRS <= NDON) Then
               XWRKT (ICRS) = XDONT (ICRS)
            Else
               If (XWRK1 /= XHUGE) NMED = NMED + 1
               XWRKT (ICRS) = XWRK1
               XWRK1 = - XWRK1
            Endif
         End Do
!
         Do ICRS = IDEB+2, IDEB+7
            XWRK = XWRKT (ICRS)
            Do IDCR = ICRS - 1, IDEB+1, - 1
               If (XWRK >= XWRKT(IDCR)) Exit
               XWRKT (IDCR+1) = XWRKT (IDCR)
            End Do
            XWRKT (IDCR+1) = XWRK
         End Do
!
         NTRI = IDEB+7
      End If
!
!  Make the set of the indices of median values of each sorted subset
!
         IDON1 = 0
         Do IDON = 1, NTRI, 7
            IDON1 = IDON1 + 1
            IMEDT (IDON1) = IDON + 3
         End Do
!
!  Find XMED7, the median of the medians
!
         XMED7 = I_valmed (XWRKT (IMEDT))
!
!  Count how many values are not higher than (and how many equal to) XMED7
!  This number is at least 4 * 1/2 * (N/7) : 4 values in each of the
!  subsets where the median is lower than the median of medians. For similar
!  reasons, we also have at least 2N/7 values not lower than XMED7. At the
!  same time, we find in each subset the index of the last value < XMED7,
!  and that of the first > XMED7. These indices will be used to restrict the
!  search for the median as the Kth element in the subset (> or <) where
!  we know it to be.
!
         IDON1 = 1
         NLEQ = 0
         NEQU = 0
         Do IDON = 1, NTRI, 7
            IMED = IDON+3
            If (XWRKT (IMED) > XMED7) Then
                  IMED = IMED - 2
                  If (XWRKT (IMED) > XMED7) Then
                     IMED = IMED - 1
                  Else If (XWRKT (IMED) < XMED7) Then
                     IMED = IMED + 1
                  Endif
            Else If (XWRKT (IMED) < XMED7) Then
                  IMED = IMED + 2
                  If (XWRKT (IMED) > XMED7) Then
                     IMED = IMED - 1
                  Else If (XWRKT (IMED) < XMED7) Then
                     IMED = IMED + 1
                  Endif
            Endif
            If (XWRKT (IMED) > XMED7) Then
               NLEQ = NLEQ + IMED - IDON
               IENDT (IDON1) = IMED - 1
               ISTRT (IDON1) = IMED
            Else If (XWRKT (IMED) < XMED7) Then
               NLEQ = NLEQ + IMED - IDON + 1
               IENDT (IDON1) = IMED
               ISTRT (IDON1) = IMED + 1
            Else                    !       If (XWRKT (IMED) == XMED7)
               NLEQ = NLEQ + IMED - IDON + 1
               NEQU = NEQU + 1
               IENDT (IDON1) = IMED - 1
               Do IMED1 = IMED - 1, IDON, -1
                  If (XWRKT (IMED1) == XMED7) Then
                     NEQU = NEQU + 1
                     IENDT (IDON1) = IMED1 - 1
                  Else
                     Exit
                  End If
               End Do
               ISTRT (IDON1) = IMED + 1
               Do IMED1 = IMED + 1, IDON + 6
                  If (XWRKT (IMED1) == XMED7) Then
                     NEQU = NEQU + 1
                     NLEQ = NLEQ + 1
                     ISTRT (IDON1) = IMED1 + 1
                  Else
                     Exit
                  End If
               End Do
            Endif
            IDON1 = IDON1 + 1
         End Do
!
!  Carry out a partial insertion sort to find the Kth smallest of the
!  large values, or the Kth largest of the small values, according to
!  what is needed.
!
        If (NLEQ - NEQU + 1 <= NMED) Then
            If (NLEQ < NMED) Then   !      Not enough low values
                XWRK1 = XHUGE
                NORD = NMED - NLEQ
                IDON1 = 0
                ICRS1 = 1
                ICRS2 = 0
                IDCR = 0
               Do IDON = 1, NTRI, 7
                   IDON1 = IDON1 + 1
                   If (ICRS2 < NORD) Then
                      Do ICRS = ISTRT (IDON1), IDON + 6
                         If (XWRKT(ICRS) < XWRK1) Then
                            XWRK = XWRKT (ICRS)
                            Do IDCR = ICRS1 - 1, 1, - 1
                               If (XWRK >= XWRKT(IDCR)) Exit
                               XWRKT (IDCR+1) = XWRKT (IDCR)
                            End Do
                            XWRKT (IDCR+1) = XWRK
                            XWRK1 = XWRKT(ICRS1)
                         Else
                           If (ICRS2 < NORD) Then
                              XWRKT (ICRS1) = XWRKT (ICRS)
                              XWRK1 = XWRKT(ICRS1)
                           Endif
                         End If
                         ICRS1 = MIN (NORD, ICRS1 + 1)
                         ICRS2 = MIN (NORD, ICRS2 + 1)
                      End Do
                   Else
                      Do ICRS = ISTRT (IDON1), IDON + 6
                         If (XWRKT(ICRS) >= XWRK1) Exit
                         XWRK = XWRKT (ICRS)
                         Do IDCR = ICRS1 - 1, 1, - 1
                               If (XWRK >= XWRKT(IDCR)) Exit
                               XWRKT (IDCR+1) = XWRKT (IDCR)
                         End Do
                         XWRKT (IDCR+1) = XWRK
                         XWRK1 = XWRKT(ICRS1)
                      End Do
                   End If
                End Do
                res_med = XWRK1
                Return
            Else
                res_med = XMED7
                Return
            End If
         Else                       !      If (NLEQ > NMED)
!                                          Not enough high values
                XWRK1 = -XHUGE
                NORD = NLEQ - NEQU - NMED + 1
                IDON1 = 0
                ICRS1 = 1
                ICRS2 = 0
                Do IDON = 1, NTRI, 7
                   IDON1 = IDON1 + 1
                   If (ICRS2 < NORD) Then
!
                      Do ICRS = IDON, IENDT (IDON1)
                         If (XWRKT(ICRS) > XWRK1) Then
                            XWRK = XWRKT (ICRS)
                            IDCR = ICRS1 - 1
                            Do IDCR = ICRS1 - 1, 1, - 1
                               If (XWRK <= XWRKT(IDCR)) Exit
                               XWRKT (IDCR+1) = XWRKT (IDCR)
                            End Do
                            XWRKT (IDCR+1) = XWRK
                            XWRK1 = XWRKT(ICRS1)
                         Else
                            If (ICRS2 < NORD) Then
                               XWRKT (ICRS1) = XWRKT (ICRS)
                               XWRK1 = XWRKT (ICRS1)
                            End If
                         End If
                         ICRS1 = MIN (NORD, ICRS1 + 1)
                         ICRS2 = MIN (NORD, ICRS2 + 1)
                      End Do
                   Else
                      Do ICRS = IENDT (IDON1), IDON, -1
                         If (XWRKT(ICRS) > XWRK1) Then
                            XWRK = XWRKT (ICRS)
                            IDCR = ICRS1 - 1
                            Do IDCR = ICRS1 - 1, 1, - 1
                               If (XWRK <= XWRKT(IDCR)) Exit
                               XWRKT (IDCR+1) = XWRKT (IDCR)
                            End Do
                            XWRKT (IDCR+1) = XWRK
                            XWRK1 = XWRKT(ICRS1)
                         Else
                            Exit
                         End If
                      End Do
                   Endif
                End Do
!
                res_med = XWRK1
                Return
         End If
!
End Function I_valmed
end module m_valmed
Module m_valnth
Integer, Parameter :: kdp = selected_real_kind(15)
public :: valnth
private :: kdp
private :: R_valnth, I_valnth, D_valnth
interface valnth
  module procedure d_valnth, r_valnth, i_valnth
end interface valnth
contains

Function D_valnth (XDONT, NORD) Result (valnth)
!  Return NORDth value of XDONT, i.e fractile of order NORD/SIZE(XDONT).
! __________________________________________________________
!  This routine uses a pivoting strategy such as the one of
!  finding the median based on the quicksort algorithm, but
!  we skew the pivot choice to try to bring it to NORD as
!  fast as possible. It uses 2 temporary arrays, where it
!  stores the indices of the values smaller than the pivot
!  (ILOWT), and the indices of values larger than the pivot
!  that we might still need later on (IHIGT). It iterates
!  until it can bring the number of values in ILOWT to
!  exactly NORD, and then finds the maximum of this set.
!  Michel Olagnon - Aug. 2000
! __________________________________________________________
! __________________________________________________________
      Real (Kind=kdp), Dimension (:), Intent (In) :: XDONT
      Real (Kind=kdp) :: valnth
      Integer, Intent (In) :: NORD
! __________________________________________________________
      Real (Kind=kdp), Dimension (SIZE(XDONT)) :: XLOWT, XHIGT
      Real (Kind=kdp) :: XPIV, XPIV0, XWRK, XWRK1, XWRK2, XWRK3, XMIN, XMAX
!
      Integer :: NDON, JHIG, JLOW, IHIG
      Integer :: IMIL, IFIN, ICRS, IDCR, ILOW
      Integer :: JLM2, JLM1, JHM2, JHM1, INTH
!
      NDON = SIZE (XDONT)
      INTH = MAX (MIN (NORD, NDON), 1)
!
!    First loop is used to fill-in XLOWT, XHIGT at the same time
!
      If (NDON < 2) Then
         If (INTH == 1) VALNTH = XDONT (1)
         Return
      End If
!
!  One chooses a pivot, best estimate possible to put fractile near
!  mid-point of the set of low values.
!
      If (XDONT(2) < XDONT(1)) Then
         XLOWT (1) = XDONT(2)
         XHIGT (1) = XDONT(1)
      Else
         XLOWT (1) = XDONT(1)
         XHIGT (1) = XDONT(2)
      End If
!
      If (NDON < 3) Then
         If (INTH == 1) VALNTH = XLOWT (1)
         If (INTH == 2) VALNTH = XHIGT (1)
         Return
      End If
!
      If (XDONT(3) < XHIGT(1)) Then
         XHIGT (2) = XHIGT (1)
         If (XDONT(3) < XLOWT(1)) Then
            XHIGT (1) = XLOWT (1)
            XLOWT (1) = XDONT(3)
         Else
            XHIGT (1) = XDONT(3)
         End If
      Else
         XHIGT (2) = XDONT(3)
      End If
!
      If (NDON < 4) Then
         If (INTH == 1) Then
             VALNTH = XLOWT (1)
         Else
             VALNTH = XHIGT (INTH - 1)
         End If
         Return
      End If
!
      If (XDONT(NDON) < XHIGT(1)) Then
         XHIGT (3) = XHIGT (2)
         XHIGT (2) = XHIGT (1)
         If (XDONT(NDON) < XLOWT(1)) Then
            XHIGT (1) = XLOWT (1)
            XLOWT (1) = XDONT(NDON)
         Else
            XHIGT (1) = XDONT(NDON)
         End If
      Else
         XHIGT (3) = XDONT(NDON)
      End If
!
      If (NDON < 5) Then
         If (INTH == 1) Then
             VALNTH = XLOWT (1)
         Else
             VALNTH = XHIGT (INTH - 1)
         End If
         Return
      End If
!

      JLOW = 1
      JHIG = 3
      XPIV = XLOWT(1) + REAL(2*INTH)/REAL(NDON+INTH) * (XHIGT(3)-XLOWT(1))
      If (XPIV >= XHIGT(1)) Then
         XPIV = XLOWT(1) + REAL(2*INTH)/REAL(NDON+INTH) * &
                           (XHIGT(2)-XLOWT(1))
         If (XPIV >= XHIGT(1)) &
             XPIV = XLOWT(1) + REAL (2*INTH) / REAL (NDON+INTH) * &
                               (XHIGT(1)-XLOWT(1))
      End If
      XPIV0 = XPIV
!
!  One puts values > pivot in the end and those <= pivot
!  at the beginning. This is split in 2 cases, so that
!  we can skip the loop test a number of times.
!  As we are also filling in the work arrays at the same time
!  we stop filling in the XHIGT array as soon as we have more
!  than enough values in XLOWT.
!
!
      If (XDONT(NDON) > XPIV) Then
         ICRS = 3
         Do
            ICRS = ICRS + 1
            If (XDONT(ICRS) > XPIV) Then
               If (ICRS >= NDON) Exit
               JHIG = JHIG + 1
               XHIGT (JHIG) = XDONT(ICRS)
            Else
               JLOW = JLOW + 1
               XLOWT (JLOW) = XDONT(ICRS)
               If (JLOW >= INTH) Exit
            End If
         End Do
!
!  One restricts further processing because it is no use
!  to store more high values
!
         If (ICRS < NDON-1) Then
            Do
               ICRS = ICRS + 1
               If (XDONT(ICRS) <= XPIV) Then
                  JLOW = JLOW + 1
                  XLOWT (JLOW) = XDONT(ICRS)
               Else If (ICRS >= NDON) Then
                  Exit
               End If
            End Do
         End If
!
!
      Else
!
!  Same as above, but this is not as easy to optimize, so the
!  DO-loop is kept
!
         Do ICRS = 4, NDON - 1
            If (XDONT(ICRS) > XPIV) Then
               JHIG = JHIG + 1
               XHIGT (JHIG) = XDONT(ICRS)
            Else
               JLOW = JLOW + 1
               XLOWT (JLOW) = XDONT(ICRS)
               If (JLOW >= INTH) Exit
            End If
         End Do
!
         If (ICRS < NDON-1) Then
            Do
               ICRS = ICRS + 1
               If (XDONT(ICRS) <= XPIV) Then
                  If (ICRS >= NDON) Exit
                  JLOW = JLOW + 1
                  XLOWT (JLOW) = XDONT(ICRS)
               End If
            End Do
         End If
      End If
!
      JLM2 = 0
      JLM1 = 0
      JHM2 = 0
      JHM1 = 0
      Do
         If (JLM2 == JLOW .And. JHM2 == JHIG) Then
!
!   We are oscillating. Perturbate by bringing JLOW closer by one
!   to INTH
!
             If (INTH > JLOW) Then
                XMIN = XHIGT(1)
                IHIG = 1
                Do ICRS = 2, JHIG
                   If (XHIGT(ICRS) < XMIN) Then
                      XMIN = XHIGT(ICRS)
                      IHIG = ICRS
                   End If
                End Do
!
                JLOW = JLOW + 1
                XLOWT (JLOW) = XHIGT (IHIG)
                XHIGT (IHIG) = XHIGT (JHIG)
                JHIG = JHIG - 1
             Else

                XMAX = XLOWT (JLOW)
                JLOW = JLOW - 1
                Do ICRS = 1, JLOW
                   If (XLOWT(ICRS) > XMAX) Then
                      XWRK = XMAX
                      XMAX = XLOWT(ICRS)
                      XLOWT (ICRS) = XWRK
                   End If
                End Do
             End If
         End If
         JLM2 = JLM1
         JLM1 = JLOW
         JHM2 = JHM1
         JHM1 = JHIG
!
!   We try to bring the number of values in the low values set
!   closer to INTH.
!
         Select Case (INTH-JLOW)
         Case (2:)
!
!   Not enough values in low part, at least 2 are missing
!
            INTH = INTH - JLOW
            JLOW = 0
            Select Case (JHIG)
!!!!!           CASE DEFAULT
!!!!!              write (unit=*,fmt=*) "Assertion failed"
!!!!!              STOP
!
!   We make a special case when we have so few values in
!   the high values set that it is bad performance to choose a pivot
!   and apply the general algorithm.
!
            Case (2)
               If (XHIGT(1) <= XHIGT(2)) Then
                  JLOW = JLOW + 1
                  XLOWT (JLOW) = XHIGT (1)
                  JLOW = JLOW + 1
                  XLOWT (JLOW) = XHIGT (2)
               Else
                  JLOW = JLOW + 1
                  XLOWT (JLOW) = XHIGT (2)
                  JLOW = JLOW + 1
                  XLOWT (JLOW) = XHIGT (1)
               End If
               Exit
!
            Case (3)
!
!
               XWRK1 = XHIGT (1)
               XWRK2 = XHIGT (2)
               XWRK3 = XHIGT (3)
               If (XWRK2 < XWRK1) Then
                  XHIGT (1) = XWRK2
                  XHIGT (2) = XWRK1
                  XWRK2 = XWRK1
               End If
               If (XWRK2 > XWRK3) Then
                  XHIGT (3) = XWRK2
                  XHIGT (2) = XWRK3
                  XWRK2 = XWRK3
                  If (XWRK2 < XHIGT(1)) Then
                     XHIGT (2) = XHIGT (1)
                     XHIGT (1) = XWRK2
                  End If
               End If
               JHIG = 0
               Do ICRS = JLOW + 1, INTH
                  JHIG = JHIG + 1
                  XLOWT (ICRS) = XHIGT (JHIG)
               End Do
               JLOW = INTH
               Exit
!
            Case (4:)
!
!
               XPIV0 = XPIV
               IFIN = JHIG
!
!  One chooses a pivot from the 2 first values and the last one.
!  This should ensure sufficient renewal between iterations to
!  avoid worst case behavior effects.
!
               XWRK1 = XHIGT (1)
               XWRK2 = XHIGT (2)
               XWRK3 = XHIGT (IFIN)
               If (XWRK2 < XWRK1) Then
                  XHIGT (1) = XWRK2
                  XHIGT (2) = XWRK1
                  XWRK2 = XWRK1
               End If
               If (XWRK2 > XWRK3) Then
                  XHIGT (IFIN) = XWRK2
                  XHIGT (2) = XWRK3
                  XWRK2 = XWRK3
                  If (XWRK2 < XHIGT(1)) Then
                     XHIGT (2) = XHIGT (1)
                     XHIGT (1) = XWRK2
                  End If
               End If
!
               XWRK1 = XHIGT (1)
               JLOW = JLOW + 1
               XLOWT (JLOW) = XWRK1
               XPIV = XWRK1 + 0.5 * (XHIGT(IFIN)-XWRK1)
!
!  One takes values <= pivot to XLOWT
!  Again, 2 parts, one where we take care of the remaining
!  high values because we might still need them, and the
!  other when we know that we will have more than enough
!  low values in the end.
!
               JHIG = 0
               Do ICRS = 2, IFIN
                  If (XHIGT(ICRS) <= XPIV) Then
                     JLOW = JLOW + 1
                     XLOWT (JLOW) = XHIGT (ICRS)
                     If (JLOW >= INTH) Exit
                  Else
                     JHIG = JHIG + 1
                     XHIGT (JHIG) = XHIGT (ICRS)
                  End If
               End Do
!
               Do ICRS = ICRS + 1, IFIN
                  If (XHIGT(ICRS) <= XPIV) Then
                     JLOW = JLOW + 1
                     XLOWT (JLOW) = XHIGT (ICRS)
                  End If
               End Do
            End Select
!
!
         Case (1)
!
!  Only 1 value is missing in low part
!
            XMIN = XHIGT(1)
            IHIG = 1
            Do ICRS = 2, JHIG
               If (XHIGT(ICRS) < XMIN) Then
                  XMIN = XHIGT(ICRS)
                  IHIG = ICRS
               End If
            End Do
!
            VALNTH = XHIGT (IHIG)
            Return
!
!
         Case (0)
!
!  Low part is exactly what we want
!
            Exit
!
!
         Case (-5:-1)
!
!  Only few values too many in low part
!
            XHIGT (1) = XLOWT (1)
            ILOW = 1 + INTH - JLOW
            Do ICRS = 2, INTH
               XWRK = XLOWT (ICRS)
               Do IDCR = ICRS - 1, MAX (1, ILOW), - 1
                  If (XWRK < XHIGT(IDCR)) Then
                     XHIGT (IDCR+1) = XHIGT (IDCR)
                  Else
                     Exit
                  End If
               End Do
               XHIGT (IDCR+1) = XWRK
               ILOW = ILOW + 1
            End Do
!
            XWRK1 = XHIGT(INTH)
            ILOW = 2*INTH - JLOW
            Do ICRS = INTH + 1, JLOW
               If (XLOWT (ICRS) < XWRK1) Then
                  XWRK = XLOWT (ICRS)
                  Do IDCR = INTH - 1, MAX (1, ILOW), - 1
                     If (XWRK >= XHIGT(IDCR)) Exit
                     XHIGT (IDCR+1) = XHIGT (IDCR)
                  End Do
                  XHIGT (IDCR+1) = XLOWT (ICRS)
                  XWRK1 = XHIGT(INTH)
               End If
               ILOW = ILOW + 1
            End Do
!
            VALNTH = XHIGT(INTH)
            Return
!
!
         Case (:-6)
!
! last case: too many values in low part
!

            IMIL = (JLOW+1) / 2
            IFIN = JLOW
!
!  One chooses a pivot from 1st, last, and middle values
!
            If (XLOWT(IMIL) < XLOWT(1)) Then
               XWRK = XLOWT (1)
               XLOWT (1) = XLOWT (IMIL)
               XLOWT (IMIL) = XWRK
            End If
            If (XLOWT(IMIL) > XLOWT(IFIN)) Then
               XWRK = XLOWT (IFIN)
               XLOWT (IFIN) = XLOWT (IMIL)
               XLOWT (IMIL) = XWRK
               If (XLOWT(IMIL) < XLOWT(1)) Then
                  XWRK = XLOWT (1)
                  XLOWT (1) = XLOWT (IMIL)
                  XLOWT (IMIL) = XWRK
               End If
            End If
            If (IFIN <= 3) Exit
!
            XPIV = XLOWT(1) + REAL(INTH)/REAL(JLOW+INTH) * &
                              (XLOWT(IFIN)-XLOWT(1))

!
!  One takes values > XPIV to XHIGT
!
            JHIG = 0
            JLOW = 0
!
            If (XLOWT(IFIN) > XPIV) Then
               ICRS = 0
               Do
                  ICRS = ICRS + 1
                  If (XLOWT(ICRS) > XPIV) Then
                     JHIG = JHIG + 1
                     XHIGT (JHIG) = XLOWT (ICRS)
                     If (ICRS >= IFIN) Exit
                  Else
                     JLOW = JLOW + 1
                     XLOWT (JLOW) = XLOWT (ICRS)
                     If (JLOW >= INTH) Exit
                  End If
               End Do
!
               If (ICRS < IFIN) Then
                  Do
                     ICRS = ICRS + 1
                     If (XLOWT(ICRS) <= XPIV) Then
                        JLOW = JLOW + 1
                        XLOWT (JLOW) = XLOWT (ICRS)
                     Else
                        If (ICRS >= IFIN) Exit
                     End If
                  End Do
               End If
            Else
               Do ICRS = 1, IFIN
                  If (XLOWT(ICRS) > XPIV) Then
                     JHIG = JHIG + 1
                     XHIGT (JHIG) = XLOWT (ICRS)
                  Else
                     JLOW = JLOW + 1
                     XLOWT (JLOW) = XLOWT (ICRS)
                     If (JLOW >= INTH) Exit
                  End If
               End Do
!
               Do ICRS = ICRS + 1, IFIN
                  If (XLOWT(ICRS) <= XPIV) Then
                     JLOW = JLOW + 1
                     XLOWT (JLOW) = XLOWT (ICRS)
                  End If
               End Do
            End If
!
         End Select
!
      End Do
!
!  Now, we only need to find maximum of the 1:INTH set
!
      VALNTH = MAXVAL (XLOWT (1:INTH))
      Return
!
!
End Function D_valnth

Function R_valnth (XDONT, NORD) Result (valnth)
!  Return NORDth value of XDONT, i.e fractile of order NORD/SIZE(XDONT).
! __________________________________________________________
!  This routine uses a pivoting strategy such as the one of
!  finding the median based on the quicksort algorithm, but
!  we skew the pivot choice to try to bring it to NORD as
!  fast as possible. It uses 2 temporary arrays, where it
!  stores the indices of the values smaller than the pivot
!  (ILOWT), and the indices of values larger than the pivot
!  that we might still need later on (IHIGT). It iterates
!  until it can bring the number of values in ILOWT to
!  exactly NORD, and then finds the maximum of this set.
!  Michel Olagnon - Aug. 2000
! __________________________________________________________
! _________________________________________________________
      Real, Dimension (:), Intent (In) :: XDONT
      Real :: valnth
      Integer, Intent (In) :: NORD
! __________________________________________________________
      Real, Dimension (SIZE(XDONT)) :: XLOWT, XHIGT
      Real :: XPIV, XPIV0, XWRK, XWRK1, XWRK2, XWRK3, XMIN, XMAX
!
      Integer :: NDON, JHIG, JLOW, IHIG
      Integer :: IMIL, IFIN, ICRS, IDCR, ILOW
      Integer :: JLM2, JLM1, JHM2, JHM1, INTH
!
      NDON = SIZE (XDONT)
      INTH = MAX (MIN (NORD, NDON), 1)
!
!    First loop is used to fill-in XLOWT, XHIGT at the same time
!
      If (NDON < 2) Then
         If (INTH == 1) VALNTH = XDONT (1)
         Return
      End If
!
!  One chooses a pivot, best estimate possible to put fractile near
!  mid-point of the set of low values.
!
      If (XDONT(2) < XDONT(1)) Then
         XLOWT (1) = XDONT(2)
         XHIGT (1) = XDONT(1)
      Else
         XLOWT (1) = XDONT(1)
         XHIGT (1) = XDONT(2)
      End If
!
      If (NDON < 3) Then
         If (INTH == 1) VALNTH = XLOWT (1)
         If (INTH == 2) VALNTH = XHIGT (1)
         Return
      End If
!
      If (XDONT(3) < XHIGT(1)) Then
         XHIGT (2) = XHIGT (1)
         If (XDONT(3) < XLOWT(1)) Then
            XHIGT (1) = XLOWT (1)
            XLOWT (1) = XDONT(3)
         Else
            XHIGT (1) = XDONT(3)
         End If
      Else
         XHIGT (2) = XDONT(3)
      End If
!
      If (NDON < 4) Then
         If (INTH == 1) Then
             VALNTH = XLOWT (1)
         Else
             VALNTH = XHIGT (INTH - 1)
         End If
         Return
      End If
!
      If (XDONT(NDON) < XHIGT(1)) Then
         XHIGT (3) = XHIGT (2)
         XHIGT (2) = XHIGT (1)
         If (XDONT(NDON) < XLOWT(1)) Then
            XHIGT (1) = XLOWT (1)
            XLOWT (1) = XDONT(NDON)
         Else
            XHIGT (1) = XDONT(NDON)
         End If
      Else
         XHIGT (3) = XDONT(NDON)
      End If
!
      If (NDON < 5) Then
         If (INTH == 1) Then
             VALNTH = XLOWT (1)
         Else
             VALNTH = XHIGT (INTH - 1)
         End If
         Return
      End If
!

      JLOW = 1
      JHIG = 3
      XPIV = XLOWT(1) + REAL(2*INTH)/REAL(NDON+INTH) * (XHIGT(3)-XLOWT(1))
      If (XPIV >= XHIGT(1)) Then
         XPIV = XLOWT(1) + REAL(2*INTH)/REAL(NDON+INTH) * &
                           (XHIGT(2)-XLOWT(1))
         If (XPIV >= XHIGT(1)) &
             XPIV = XLOWT(1) + REAL (2*INTH) / REAL (NDON+INTH) * &
                               (XHIGT(1)-XLOWT(1))
      End If
      XPIV0 = XPIV
!
!  One puts values > pivot in the end and those <= pivot
!  at the beginning. This is split in 2 cases, so that
!  we can skip the loop test a number of times.
!  As we are also filling in the work arrays at the same time
!  we stop filling in the XHIGT array as soon as we have more
!  than enough values in XLOWT.
!
!
      If (XDONT(NDON) > XPIV) Then
         ICRS = 3
         Do
            ICRS = ICRS + 1
            If (XDONT(ICRS) > XPIV) Then
               If (ICRS >= NDON) Exit
               JHIG = JHIG + 1
               XHIGT (JHIG) = XDONT(ICRS)
            Else
               JLOW = JLOW + 1
               XLOWT (JLOW) = XDONT(ICRS)
               If (JLOW >= INTH) Exit
            End If
         End Do
!
!  One restricts further processing because it is no use
!  to store more high values
!
         If (ICRS < NDON-1) Then
            Do
               ICRS = ICRS + 1
               If (XDONT(ICRS) <= XPIV) Then
                  JLOW = JLOW + 1
                  XLOWT (JLOW) = XDONT(ICRS)
               Else If (ICRS >= NDON) Then
                  Exit
               End If
            End Do
         End If
!
!
      Else
!
!  Same as above, but this is not as easy to optimize, so the
!  DO-loop is kept
!
         Do ICRS = 4, NDON - 1
            If (XDONT(ICRS) > XPIV) Then
               JHIG = JHIG + 1
               XHIGT (JHIG) = XDONT(ICRS)
            Else
               JLOW = JLOW + 1
               XLOWT (JLOW) = XDONT(ICRS)
               If (JLOW >= INTH) Exit
            End If
         End Do
!
         If (ICRS < NDON-1) Then
            Do
               ICRS = ICRS + 1
               If (XDONT(ICRS) <= XPIV) Then
                  If (ICRS >= NDON) Exit
                  JLOW = JLOW + 1
                  XLOWT (JLOW) = XDONT(ICRS)
               End If
            End Do
         End If
      End If
!
      JLM2 = 0
      JLM1 = 0
      JHM2 = 0
      JHM1 = 0
      Do
         If (JLM2 == JLOW .And. JHM2 == JHIG) Then
!
!   We are oscillating. Perturbate by bringing JLOW closer by one
!   to INTH
!
             If (INTH > JLOW) Then
                XMIN = XHIGT(1)
                IHIG = 1
                Do ICRS = 2, JHIG
                   If (XHIGT(ICRS) < XMIN) Then
                      XMIN = XHIGT(ICRS)
                      IHIG = ICRS
                   End If
                End Do
!
                JLOW = JLOW + 1
                XLOWT (JLOW) = XHIGT (IHIG)
                XHIGT (IHIG) = XHIGT (JHIG)
                JHIG = JHIG - 1
             Else

                XMAX = XLOWT (JLOW)
                JLOW = JLOW - 1
                Do ICRS = 1, JLOW
                   If (XLOWT(ICRS) > XMAX) Then
                      XWRK = XMAX
                      XMAX = XLOWT(ICRS)
                      XLOWT (ICRS) = XWRK
                   End If
                End Do
             End If
         End If
         JLM2 = JLM1
         JLM1 = JLOW
         JHM2 = JHM1
         JHM1 = JHIG
!
!   We try to bring the number of values in the low values set
!   closer to INTH.
!
         Select Case (INTH-JLOW)
         Case (2:)
!
!   Not enough values in low part, at least 2 are missing
!
            INTH = INTH - JLOW
            JLOW = 0
            Select Case (JHIG)
!!!!!           CASE DEFAULT
!!!!!              write (unit=*,fmt=*) "Assertion failed"
!!!!!              STOP
!
!   We make a special case when we have so few values in
!   the high values set that it is bad performance to choose a pivot
!   and apply the general algorithm.
!
            Case (2)
               If (XHIGT(1) <= XHIGT(2)) Then
                  JLOW = JLOW + 1
                  XLOWT (JLOW) = XHIGT (1)
                  JLOW = JLOW + 1
                  XLOWT (JLOW) = XHIGT (2)
               Else
                  JLOW = JLOW + 1
                  XLOWT (JLOW) = XHIGT (2)
                  JLOW = JLOW + 1
                  XLOWT (JLOW) = XHIGT (1)
               End If
               Exit
!
            Case (3)
!
!
               XWRK1 = XHIGT (1)
               XWRK2 = XHIGT (2)
               XWRK3 = XHIGT (3)
               If (XWRK2 < XWRK1) Then
                  XHIGT (1) = XWRK2
                  XHIGT (2) = XWRK1
                  XWRK2 = XWRK1
               End If
               If (XWRK2 > XWRK3) Then
                  XHIGT (3) = XWRK2
                  XHIGT (2) = XWRK3
                  XWRK2 = XWRK3
                  If (XWRK2 < XHIGT(1)) Then
                     XHIGT (2) = XHIGT (1)
                     XHIGT (1) = XWRK2
                  End If
               End If
               JHIG = 0
               Do ICRS = JLOW + 1, INTH
                  JHIG = JHIG + 1
                  XLOWT (ICRS) = XHIGT (JHIG)
               End Do
               JLOW = INTH
               Exit
!
            Case (4:)
!
!
               XPIV0 = XPIV
               IFIN = JHIG
!
!  One chooses a pivot from the 2 first values and the last one.
!  This should ensure sufficient renewal between iterations to
!  avoid worst case behavior effects.
!
               XWRK1 = XHIGT (1)
               XWRK2 = XHIGT (2)
               XWRK3 = XHIGT (IFIN)
               If (XWRK2 < XWRK1) Then
                  XHIGT (1) = XWRK2
                  XHIGT (2) = XWRK1
                  XWRK2 = XWRK1
               End If
               If (XWRK2 > XWRK3) Then
                  XHIGT (IFIN) = XWRK2
                  XHIGT (2) = XWRK3
                  XWRK2 = XWRK3
                  If (XWRK2 < XHIGT(1)) Then
                     XHIGT (2) = XHIGT (1)
                     XHIGT (1) = XWRK2
                  End If
               End If
!
               XWRK1 = XHIGT (1)
               JLOW = JLOW + 1
               XLOWT (JLOW) = XWRK1
               XPIV = XWRK1 + 0.5 * (XHIGT(IFIN)-XWRK1)
!
!  One takes values <= pivot to XLOWT
!  Again, 2 parts, one where we take care of the remaining
!  high values because we might still need them, and the
!  other when we know that we will have more than enough
!  low values in the end.
!
               JHIG = 0
               Do ICRS = 2, IFIN
                  If (XHIGT(ICRS) <= XPIV) Then
                     JLOW = JLOW + 1
                     XLOWT (JLOW) = XHIGT (ICRS)
                     If (JLOW >= INTH) Exit
                  Else
                     JHIG = JHIG + 1
                     XHIGT (JHIG) = XHIGT (ICRS)
                  End If
               End Do
!
               Do ICRS = ICRS + 1, IFIN
                  If (XHIGT(ICRS) <= XPIV) Then
                     JLOW = JLOW + 1
                     XLOWT (JLOW) = XHIGT (ICRS)
                  End If
               End Do
            End Select
!
!
         Case (1)
!
!  Only 1 value is missing in low part
!
            XMIN = XHIGT(1)
            IHIG = 1
            Do ICRS = 2, JHIG
               If (XHIGT(ICRS) < XMIN) Then
                  XMIN = XHIGT(ICRS)
                  IHIG = ICRS
               End If
            End Do
!
            VALNTH = XHIGT (IHIG)
            Return
!
!
         Case (0)
!
!  Low part is exactly what we want
!
            Exit
!
!
         Case (-5:-1)
!
!  Only few values too many in low part
!
            XHIGT (1) = XLOWT (1)
            ILOW = 1 + INTH - JLOW
            Do ICRS = 2, INTH
               XWRK = XLOWT (ICRS)
               Do IDCR = ICRS - 1, MAX (1, ILOW), - 1
                  If (XWRK < XHIGT(IDCR)) Then
                     XHIGT (IDCR+1) = XHIGT (IDCR)
                  Else
                     Exit
                  End If
               End Do
               XHIGT (IDCR+1) = XWRK
               ILOW = ILOW + 1
            End Do
!
            XWRK1 = XHIGT(INTH)
            ILOW = 2*INTH - JLOW
            Do ICRS = INTH + 1, JLOW
               If (XLOWT (ICRS) < XWRK1) Then
                  XWRK = XLOWT (ICRS)
                  Do IDCR = INTH - 1, MAX (1, ILOW), - 1
                     If (XWRK >= XHIGT(IDCR)) Exit
                     XHIGT (IDCR+1) = XHIGT (IDCR)
                  End Do
                  XHIGT (IDCR+1) = XLOWT (ICRS)
                  XWRK1 = XHIGT(INTH)
               End If
               ILOW = ILOW + 1
            End Do
!
            VALNTH = XHIGT(INTH)
            Return
!
!
         Case (:-6)
!
! last case: too many values in low part
!

            IMIL = (JLOW+1) / 2
            IFIN = JLOW
!
!  One chooses a pivot from 1st, last, and middle values
!
            If (XLOWT(IMIL) < XLOWT(1)) Then
               XWRK = XLOWT (1)
               XLOWT (1) = XLOWT (IMIL)
               XLOWT (IMIL) = XWRK
            End If
            If (XLOWT(IMIL) > XLOWT(IFIN)) Then
               XWRK = XLOWT (IFIN)
               XLOWT (IFIN) = XLOWT (IMIL)
               XLOWT (IMIL) = XWRK
               If (XLOWT(IMIL) < XLOWT(1)) Then
                  XWRK = XLOWT (1)
                  XLOWT (1) = XLOWT (IMIL)
                  XLOWT (IMIL) = XWRK
               End If
            End If
            If (IFIN <= 3) Exit
!
            XPIV = XLOWT(1) + REAL(INTH)/REAL(JLOW+INTH) * &
                              (XLOWT(IFIN)-XLOWT(1))

!
!  One takes values > XPIV to XHIGT
!
            JHIG = 0
            JLOW = 0
!
            If (XLOWT(IFIN) > XPIV) Then
               ICRS = 0
               Do
                  ICRS = ICRS + 1
                  If (XLOWT(ICRS) > XPIV) Then
                     JHIG = JHIG + 1
                     XHIGT (JHIG) = XLOWT (ICRS)
                     If (ICRS >= IFIN) Exit
                  Else
                     JLOW = JLOW + 1
                     XLOWT (JLOW) = XLOWT (ICRS)
                     If (JLOW >= INTH) Exit
                  End If
               End Do
!
               If (ICRS < IFIN) Then
                  Do
                     ICRS = ICRS + 1
                     If (XLOWT(ICRS) <= XPIV) Then
                        JLOW = JLOW + 1
                        XLOWT (JLOW) = XLOWT (ICRS)
                     Else
                        If (ICRS >= IFIN) Exit
                     End If
                  End Do
               End If
            Else
               Do ICRS = 1, IFIN
                  If (XLOWT(ICRS) > XPIV) Then
                     JHIG = JHIG + 1
                     XHIGT (JHIG) = XLOWT (ICRS)
                  Else
                     JLOW = JLOW + 1
                     XLOWT (JLOW) = XLOWT (ICRS)
                     If (JLOW >= INTH) Exit
                  End If
               End Do
!
               Do ICRS = ICRS + 1, IFIN
                  If (XLOWT(ICRS) <= XPIV) Then
                     JLOW = JLOW + 1
                     XLOWT (JLOW) = XLOWT (ICRS)
                  End If
               End Do
            End If
!
         End Select
!
      End Do
!
!  Now, we only need to find maximum of the 1:INTH set
!
      VALNTH = MAXVAL (XLOWT (1:INTH))
      Return
!
!
End Function R_valnth
Function I_valnth (XDONT, NORD) Result (valnth)
!  Return NORDth value of XDONT, i.e fractile of order NORD/SIZE(XDONT).
! __________________________________________________________
!  This routine uses a pivoting strategy such as the one of
!  finding the median based on the quicksort algorithm, but
!  we skew the pivot choice to try to bring it to NORD as
!  fast as possible. It uses 2 temporary arrays, where it
!  stores the indices of the values smaller than the pivot
!  (ILOWT), and the indices of values larger than the pivot
!  that we might still need later on (IHIGT). It iterates
!  until it can bring the number of values in ILOWT to
!  exactly NORD, and then finds the maximum of this set.
!  Michel Olagnon - Aug. 2000
! __________________________________________________________
! __________________________________________________________
      Integer, Dimension (:), Intent (In) :: XDONT
      Integer :: valnth
      Integer, Intent (In) :: NORD
! __________________________________________________________
      Integer, Dimension (SIZE(XDONT)) :: XLOWT, XHIGT
      Integer :: XPIV, XPIV0, XWRK, XWRK1, XWRK2, XWRK3, XMIN, XMAX
!
      Integer :: NDON, JHIG, JLOW, IHIG
      Integer :: IMIL, IFIN, ICRS, IDCR, ILOW
      Integer :: JLM2, JLM1, JHM2, JHM1, INTH
!
      NDON = SIZE (XDONT)
      INTH = MAX (MIN (NORD, NDON), 1)
!
!    First loop is used to fill-in XLOWT, XHIGT at the same time
!
      If (NDON < 2) Then
         If (INTH == 1) VALNTH = XDONT (1)
         Return
      End If
!
!  One chooses a pivot, best estimate possible to put fractile near
!  mid-point of the set of low values.
!
      If (XDONT(2) < XDONT(1)) Then
         XLOWT (1) = XDONT(2)
         XHIGT (1) = XDONT(1)
      Else
         XLOWT (1) = XDONT(1)
         XHIGT (1) = XDONT(2)
      End If
!
      If (NDON < 3) Then
         If (INTH == 1) VALNTH = XLOWT (1)
         If (INTH == 2) VALNTH = XHIGT (1)
         Return
      End If
!
      If (XDONT(3) < XHIGT(1)) Then
         XHIGT (2) = XHIGT (1)
         If (XDONT(3) < XLOWT(1)) Then
            XHIGT (1) = XLOWT (1)
            XLOWT (1) = XDONT(3)
         Else
            XHIGT (1) = XDONT(3)
         End If
      Else
         XHIGT (2) = XDONT(3)
      End If
!
      If (NDON < 4) Then
         If (INTH == 1) Then
             VALNTH = XLOWT (1)
         Else
             VALNTH = XHIGT (INTH - 1)
         End If
         Return
      End If
!
      If (XDONT(NDON) < XHIGT(1)) Then
         XHIGT (3) = XHIGT (2)
         XHIGT (2) = XHIGT (1)
         If (XDONT(NDON) < XLOWT(1)) Then
            XHIGT (1) = XLOWT (1)
            XLOWT (1) = XDONT(NDON)
         Else
            XHIGT (1) = XDONT(NDON)
         End If
      Else
         XHIGT (3) = XDONT(NDON)
      End If
!
      If (NDON < 5) Then
         If (INTH == 1) Then
             VALNTH = XLOWT (1)
         Else
             VALNTH = XHIGT (INTH - 1)
         End If
         Return
      End If
!

      JLOW = 1
      JHIG = 3
      XPIV = XLOWT(1) + REAL(2*INTH)/REAL(NDON+INTH) * (XHIGT(3)-XLOWT(1))
      If (XPIV >= XHIGT(1)) Then
         XPIV = XLOWT(1) + REAL(2*INTH)/REAL(NDON+INTH) * &
                           (XHIGT(2)-XLOWT(1))
         If (XPIV >= XHIGT(1)) &
             XPIV = XLOWT(1) + REAL (2*INTH) / REAL (NDON+INTH) * &
                               (XHIGT(1)-XLOWT(1))
      End If
      XPIV0 = XPIV
!
!  One puts values > pivot in the end and those <= pivot
!  at the beginning. This is split in 2 cases, so that
!  we can skip the loop test a number of times.
!  As we are also filling in the work arrays at the same time
!  we stop filling in the XHIGT array as soon as we have more
!  than enough values in XLOWT.
!
!
      If (XDONT(NDON) > XPIV) Then
         ICRS = 3
         Do
            ICRS = ICRS + 1
            If (XDONT(ICRS) > XPIV) Then
               If (ICRS >= NDON) Exit
               JHIG = JHIG + 1
               XHIGT (JHIG) = XDONT(ICRS)
            Else
               JLOW = JLOW + 1
               XLOWT (JLOW) = XDONT(ICRS)
               If (JLOW >= INTH) Exit
            End If
         End Do
!
!  One restricts further processing because it is no use
!  to store more high values
!
         If (ICRS < NDON-1) Then
            Do
               ICRS = ICRS + 1
               If (XDONT(ICRS) <= XPIV) Then
                  JLOW = JLOW + 1
                  XLOWT (JLOW) = XDONT(ICRS)
               Else If (ICRS >= NDON) Then
                  Exit
               End If
            End Do
         End If
!
!
      Else
!
!  Same as above, but this is not as easy to optimize, so the
!  DO-loop is kept
!
         Do ICRS = 4, NDON - 1
            If (XDONT(ICRS) > XPIV) Then
               JHIG = JHIG + 1
               XHIGT (JHIG) = XDONT(ICRS)
            Else
               JLOW = JLOW + 1
               XLOWT (JLOW) = XDONT(ICRS)
               If (JLOW >= INTH) Exit
            End If
         End Do
!
         If (ICRS < NDON-1) Then
            Do
               ICRS = ICRS + 1
               If (XDONT(ICRS) <= XPIV) Then
                  If (ICRS >= NDON) Exit
                  JLOW = JLOW + 1
                  XLOWT (JLOW) = XDONT(ICRS)
               End If
            End Do
         End If
      End If
!
      JLM2 = 0
      JLM1 = 0
      JHM2 = 0
      JHM1 = 0
      Do
         If (JLM2 == JLOW .And. JHM2 == JHIG) Then
!
!   We are oscillating. Perturbate by bringing JLOW closer by one
!   to INTH
!
             If (INTH > JLOW) Then
                XMIN = XHIGT(1)
                IHIG = 1
                Do ICRS = 2, JHIG
                   If (XHIGT(ICRS) < XMIN) Then
                      XMIN = XHIGT(ICRS)
                      IHIG = ICRS
                   End If
                End Do
!
                JLOW = JLOW + 1
                XLOWT (JLOW) = XHIGT (IHIG)
                XHIGT (IHIG) = XHIGT (JHIG)
                JHIG = JHIG - 1
             Else

                XMAX = XLOWT (JLOW)
                JLOW = JLOW - 1
                Do ICRS = 1, JLOW
                   If (XLOWT(ICRS) > XMAX) Then
                      XWRK = XMAX
                      XMAX = XLOWT(ICRS)
                      XLOWT (ICRS) = XWRK
                   End If
                End Do
             End If
         End If
         JLM2 = JLM1
         JLM1 = JLOW
         JHM2 = JHM1
         JHM1 = JHIG
!
!   We try to bring the number of values in the low values set
!   closer to INTH.
!
         Select Case (INTH-JLOW)
         Case (2:)
!
!   Not enough values in low part, at least 2 are missing
!
            INTH = INTH - JLOW
            JLOW = 0
            Select Case (JHIG)
!!!!!           CASE DEFAULT
!!!!!              write (unit=*,fmt=*) "Assertion failed"
!!!!!              STOP
!
!   We make a special case when we have so few values in
!   the high values set that it is bad performance to choose a pivot
!   and apply the general algorithm.
!
            Case (2)
               If (XHIGT(1) <= XHIGT(2)) Then
                  JLOW = JLOW + 1
                  XLOWT (JLOW) = XHIGT (1)
                  JLOW = JLOW + 1
                  XLOWT (JLOW) = XHIGT (2)
               Else
                  JLOW = JLOW + 1
                  XLOWT (JLOW) = XHIGT (2)
                  JLOW = JLOW + 1
                  XLOWT (JLOW) = XHIGT (1)
               End If
               Exit
!
            Case (3)
!
!
               XWRK1 = XHIGT (1)
               XWRK2 = XHIGT (2)
               XWRK3 = XHIGT (3)
               If (XWRK2 < XWRK1) Then
                  XHIGT (1) = XWRK2
                  XHIGT (2) = XWRK1
                  XWRK2 = XWRK1
               End If
               If (XWRK2 > XWRK3) Then
                  XHIGT (3) = XWRK2
                  XHIGT (2) = XWRK3
                  XWRK2 = XWRK3
                  If (XWRK2 < XHIGT(1)) Then
                     XHIGT (2) = XHIGT (1)
                     XHIGT (1) = XWRK2
                  End If
               End If
               JHIG = 0
               Do ICRS = JLOW + 1, INTH
                  JHIG = JHIG + 1
                  XLOWT (ICRS) = XHIGT (JHIG)
               End Do
               JLOW = INTH
               Exit
!
            Case (4:)
!
!
               XPIV0 = XPIV
               IFIN = JHIG
!
!  One chooses a pivot from the 2 first values and the last one.
!  This should ensure sufficient renewal between iterations to
!  avoid worst case behavior effects.
!
               XWRK1 = XHIGT (1)
               XWRK2 = XHIGT (2)
               XWRK3 = XHIGT (IFIN)
               If (XWRK2 < XWRK1) Then
                  XHIGT (1) = XWRK2
                  XHIGT (2) = XWRK1
                  XWRK2 = XWRK1
               End If
               If (XWRK2 > XWRK3) Then
                  XHIGT (IFIN) = XWRK2
                  XHIGT (2) = XWRK3
                  XWRK2 = XWRK3
                  If (XWRK2 < XHIGT(1)) Then
                     XHIGT (2) = XHIGT (1)
                     XHIGT (1) = XWRK2
                  End If
               End If
!
               XWRK1 = XHIGT (1)
               JLOW = JLOW + 1
               XLOWT (JLOW) = XWRK1
               XPIV = XWRK1 + 0.5 * (XHIGT(IFIN)-XWRK1)
!
!  One takes values <= pivot to XLOWT
!  Again, 2 parts, one where we take care of the remaining
!  high values because we might still need them, and the
!  other when we know that we will have more than enough
!  low values in the end.
!
               JHIG = 0
               Do ICRS = 2, IFIN
                  If (XHIGT(ICRS) <= XPIV) Then
                     JLOW = JLOW + 1
                     XLOWT (JLOW) = XHIGT (ICRS)
                     If (JLOW >= INTH) Exit
                  Else
                     JHIG = JHIG + 1
                     XHIGT (JHIG) = XHIGT (ICRS)
                  End If
               End Do
!
               Do ICRS = ICRS + 1, IFIN
                  If (XHIGT(ICRS) <= XPIV) Then
                     JLOW = JLOW + 1
                     XLOWT (JLOW) = XHIGT (ICRS)
                  End If
               End Do
            End Select
!
!
         Case (1)
!
!  Only 1 value is missing in low part
!
            XMIN = XHIGT(1)
            IHIG = 1
            Do ICRS = 2, JHIG
               If (XHIGT(ICRS) < XMIN) Then
                  XMIN = XHIGT(ICRS)
                  IHIG = ICRS
               End If
            End Do
!
            VALNTH = XHIGT (IHIG)
            Return
!
!
         Case (0)
!
!  Low part is exactly what we want
!
            Exit
!
!
         Case (-5:-1)
!
!  Only few values too many in low part
!
            XHIGT (1) = XLOWT (1)
            ILOW = 1 + INTH - JLOW
            Do ICRS = 2, INTH
               XWRK = XLOWT (ICRS)
               Do IDCR = ICRS - 1, MAX (1, ILOW), - 1
                  If (XWRK < XHIGT(IDCR)) Then
                     XHIGT (IDCR+1) = XHIGT (IDCR)
                  Else
                     Exit
                  End If
               End Do
               XHIGT (IDCR+1) = XWRK
               ILOW = ILOW + 1
            End Do
!
            XWRK1 = XHIGT(INTH)
            ILOW = 2*INTH - JLOW
            Do ICRS = INTH + 1, JLOW
               If (XLOWT (ICRS) < XWRK1) Then
                  XWRK = XLOWT (ICRS)
                  Do IDCR = INTH - 1, MAX (1, ILOW), - 1
                     If (XWRK >= XHIGT(IDCR)) Exit
                     XHIGT (IDCR+1) = XHIGT (IDCR)
                  End Do
                  XHIGT (IDCR+1) = XLOWT (ICRS)
                  XWRK1 = XHIGT(INTH)
               End If
               ILOW = ILOW + 1
            End Do
!
            VALNTH = XHIGT(INTH)
            Return
!
!
         Case (:-6)
!
! last case: too many values in low part
!

            IMIL = (JLOW+1) / 2
            IFIN = JLOW
!
!  One chooses a pivot from 1st, last, and middle values
!
            If (XLOWT(IMIL) < XLOWT(1)) Then
               XWRK = XLOWT (1)
               XLOWT (1) = XLOWT (IMIL)
               XLOWT (IMIL) = XWRK
            End If
            If (XLOWT(IMIL) > XLOWT(IFIN)) Then
               XWRK = XLOWT (IFIN)
               XLOWT (IFIN) = XLOWT (IMIL)
               XLOWT (IMIL) = XWRK
               If (XLOWT(IMIL) < XLOWT(1)) Then
                  XWRK = XLOWT (1)
                  XLOWT (1) = XLOWT (IMIL)
                  XLOWT (IMIL) = XWRK
               End If
            End If
            If (IFIN <= 3) Exit
!
            XPIV = XLOWT(1) + REAL(INTH)/REAL(JLOW+INTH) * &
                              (XLOWT(IFIN)-XLOWT(1))

!
!  One takes values > XPIV to XHIGT
!
            JHIG = 0
            JLOW = 0
!
            If (XLOWT(IFIN) > XPIV) Then
               ICRS = 0
               Do
                  ICRS = ICRS + 1
                  If (XLOWT(ICRS) > XPIV) Then
                     JHIG = JHIG + 1
                     XHIGT (JHIG) = XLOWT (ICRS)
                     If (ICRS >= IFIN) Exit
                  Else
                     JLOW = JLOW + 1
                     XLOWT (JLOW) = XLOWT (ICRS)
                     If (JLOW >= INTH) Exit
                  End If
               End Do
!
               If (ICRS < IFIN) Then
                  Do
                     ICRS = ICRS + 1
                     If (XLOWT(ICRS) <= XPIV) Then
                        JLOW = JLOW + 1
                        XLOWT (JLOW) = XLOWT (ICRS)
                     Else
                        If (ICRS >= IFIN) Exit
                     End If
                  End Do
               End If
            Else
               Do ICRS = 1, IFIN
                  If (XLOWT(ICRS) > XPIV) Then
                     JHIG = JHIG + 1
                     XHIGT (JHIG) = XLOWT (ICRS)
                  Else
                     JLOW = JLOW + 1
                     XLOWT (JLOW) = XLOWT (ICRS)
                     If (JLOW >= INTH) Exit
                  End If
               End Do
!
               Do ICRS = ICRS + 1, IFIN
                  If (XLOWT(ICRS) <= XPIV) Then
                     JLOW = JLOW + 1
                     XLOWT (JLOW) = XLOWT (ICRS)
                  End If
               End Do
            End If
!
         End Select
!
      End Do
!
!  Now, we only need to find maximum of the 1:INTH set
!
      VALNTH = MAXVAL (XLOWT (1:INTH))
      Return
!
!
End Function I_valnth
end module m_valnth




module mod_orderpack
    use m_mrgrnk
    use m_ctrper
    use m_fndnth
    use m_indmed
    use m_indnth
    use m_inspar
    use m_inssor
    use m_median
    use m_mrgref
    use m_mulcnt
    use m_rapknr
    use m_refpar
    use m_refsor
    use m_rinpar
    use m_rnkpar
    use m_uniinv
    use m_unipar
    use m_unirnk
    use m_unista
    use m_valmed
    use m_valnth
    implicit none
end module

module mod_types_private
    implicit none
    include 'mpif.h'
end module 
module mod_types
    use mod_types_private, only: MPI_REAL,MPI_DOUBLE_PRECISION
    implicit none

    ! Single precision
    ! integer, parameter :: rt = 4 
    ! integer, parameter :: mympi_rt = MPI_REAL 

    ! Double precision
    integer, parameter :: rt = 8
    integer, parameter :: mympi_rt = MPI_DOUBLE_PRECISION 

    private 
    public :: rt, mympi_rt
end module 
! A collection of core routines used by many of the libgrey routines. These are designed to be basic, i.e., they have no dependencies.

! Routines
module mod_core
    use mod_types
    implicit none

    interface disp
        module procedure disp_scal, disp_vec, disp_mat, disp_3d, disp_vec_i, disp_mat_i, &
                         disp_scal_title, disp_vec_title, disp_mat_title, disp_3d_title, &
                         disp_vec_title_i, disp_mat_title_i, disp_mat_title_l, disp_mat_l
    end interface disp

    integer(8), private :: datetime_start(8)


contains

! Define a function continue to prevent the use of 
integer function continue()
    implicit none
    print*,'Don''t use continue, use cycle. Continue it doesn''t do what you think!'
    continue = -1
end function

! Produces a square n by n identity matrix
function eye(n) result(mat)
    implicit none
    integer, intent(in) :: n
    real(rt), dimension(n,n) :: mat    
    integer :: i

    mat = 0d0
    do i = 1,n
        mat(i,i) = 1d0
    end do

end function eye

subroutine disp_mat_title_l(title,X)
    implicit none
    character(len=*), intent(in) :: title
    logical, dimension(1:,1:), intent(in) :: X    
    write(*,'(A)') title
    call disp_mat_l(X)
end subroutine 
subroutine disp_mat_l(X)
    implicit none
    logical, dimension(1:,1:), intent(in) :: X    
    integer :: i
    character(len=5) :: dim2
    write(dim2,'(i5)') size(X,2)
    do i = 1,size(X,1)
        write(*,'('//dim2//'L2)') X(i,:)
    end do
end subroutine disp_mat_l

subroutine disp_mat_title_i(title,X)
    implicit none
    character(len=*), intent(in) :: title
    integer, dimension(1:,1:), intent(in) :: X    
    write(*,'(A)') title
    call disp_mat_i(X)
end subroutine 
subroutine disp_mat_i(X)
    implicit none
    integer, dimension(1:,1:), intent(in) :: X    
    integer :: i
    character(len=5) :: dim2
    write(dim2,'(i5)') size(X,2)
    do i = 1,size(X,1)
        write(*,'('//dim2//'i8)') X(i,:)
    end do
end subroutine disp_mat_i

subroutine disp_vec_title_i(title,X)
    implicit none
    character(len=*), intent(in) :: title
    integer, dimension(1:), intent(in) :: X    
    write(*,'(A)') title
    call disp_vec_i(X)
end subroutine
subroutine disp_vec_i(X)
    implicit none
    integer, dimension(1:), intent(in) :: X    
    integer :: i
    do i = 1,size(X,1)
        write(*,'(i8)') X(i)
    end do
end subroutine disp_vec_i


subroutine disp_vec(X,uid)
    implicit none
    real(rt), dimension(:), intent(in) :: X    
    integer, intent(in), optional :: uid
    if (     present(uid)) call disp_core(X,size(X,1),1,1,'',uid)
    if (.not.present(uid)) call disp_core(X,size(X,1),1,1,'',6)
end subroutine 
subroutine disp_mat(X,uid)
    implicit none
    real(rt), dimension(:,:), intent(in) :: X    
    integer, intent(in), optional :: uid
    if (     present(uid)) call disp_core(X,size(X,1),size(X,2),1,'',uid)
    if (.not.present(uid)) call disp_core(X,size(X,1),size(X,2),1,'',6)
end subroutine 
subroutine disp_3d(X,uid)
    implicit none
    real(rt), dimension(:,:,:), intent(in) :: X    
    integer, intent(in), optional :: uid
    if (     present(uid)) call disp_core(X,size(X,1),size(X,2),size(X,3),'',uid)
    if (.not.present(uid)) call disp_core(X,size(X,1),size(X,2),size(X,3),'',6)
end subroutine 
subroutine disp_vec_title(title,X,uid)
    implicit none
    real(rt), dimension(:), intent(in) :: X    
    character(len=*), intent(in) :: title
    integer, intent(in), optional :: uid
    if (     present(uid)) call disp_core(X,size(X,1),1,1,title,uid)
    if (.not.present(uid)) call disp_core(X,size(X,1),1,1,title,6)
end subroutine 
subroutine disp_mat_title(title,X,uid)
    implicit none
    real(rt), dimension(:,:), intent(in) :: X    
    character(len=*), intent(in) :: title
    integer, intent(in), optional :: uid
    if (     present(uid)) call disp_core(X,size(X,1),size(X,2),1,title,uid)
    if (.not.present(uid)) call disp_core(X,size(X,1),size(X,2),1,title,6)
end subroutine 
subroutine disp_3d_title(title,X,uid)
    implicit none
    real(rt), dimension(:,:,:), intent(in) :: X    
    character(len=*), intent(in) :: title
    integer, intent(in), optional :: uid
    if (     present(uid)) call disp_core(X,size(X,1),size(X,2),size(X,3),title,uid)
    if (.not.present(uid)) call disp_core(X,size(X,1),size(X,2),size(X,3),title,6)
end subroutine 

subroutine disp_core(X,n1,n2,n3,title,uid)
    implicit none
    integer, intent(in) :: n1,n2,n3,uid
    character(len=*), intent(in) :: title
    real(rt), intent(in) :: X(n1,n2,*)
    ! local
    integer :: i1,i2,i3

    ! If no title or uid is present, then use title='' and uid=6
    do i3 = 1,n3
        if (n3>1) then
            write(uid,'(A,i4,A)') title//'(:,:,',i3,')'
        else
            if (title/='') write(uid,'(A)') title
        end if
        do i1 = 1,n1
            do i2 = 1,n2
                if (abs(X(i1,i2,i3))<1d5) then
                    write(uid,'(f16.5)',advance='no') X(i1,i2,i3)
                else
                    write(uid,'(es16.7e3)',advance='no') X(i1,i2,i3)
                end if
            end do
            write(uid,*)
        end do
    end do

end subroutine



subroutine disp_scal_title(title,X)
    implicit none
    character(len=*), intent(in) :: title
    real(rt), intent(in) :: X    
    write(*,'(A)') title
    call disp_scal(X)
end subroutine 
subroutine disp_scal(X)
    implicit none
    real(rt), intent(in) :: X    
    write(*,'(f12.5)') X
end subroutine disp_scal

! Acts just like the colon in Matlab
pure function colon(a,b)
    implicit none
    integer, intent(in) :: a,b
    integer, dimension(b-a+1) :: colon
    integer :: i
    do i = a,b
        colon(i-a+1) = i
    end do
end function colon

! Use binary search to find the lowerbound on a bracketing interval for 
! Given sorted ascending array, find lowerbound on bracketing interval. Note that
! ilo will always be in 1,...,size(x)-1.
! 
! This is supposed to work much like the C++ standard library where the interface comes from
subroutine bsearch(ilo,xhat,x)
    implicit none
    real(rt), dimension(:), intent(in) :: x
    real(rt), intent(in) :: xhat
    integer, intent(out) :: ilo
    !local
!     integer :: ihi,mid

    call bsearch_core(ilo,xhat,x,1,size(x))
        
!     ilo = 1
!     ihi = size(x)
!     mid = (ihi+ilo)/2 ! mid will have ilo<=mid<ihi as long as ihi>ilo. If ihi=ilo, then mid=ilo, so the 
!                       ! while loop is not entered and ilo=1 is returned 
! 
!     do while (mid>ilo) ! mid=ilo means ihi=ilo+1.
! 
!         ! Here, we must have ihi>ilo and ilo<=mid<ihi
! 
!         ! Bisect. 
!         ! If xhat<x(1), then xhat>=x(mid) will never occur and ilo will never be modified.
!         ! If xhat>=x(nx), then xhat>=x(mid) will always occur (note mid<ihi). Consequently, 
!         !      ilo = mid occurs until mid==ilo, which only occures if ihi==ilo+1.
!         if (xhat>=x(mid)) then
!             ilo = mid
!         else
!             ihi = mid
!         end if
! 
!         mid = (ihi+ilo)/2 
! 
!     end do
! 
!     ! Check the result as a precaution.  Make sure x in fact lies between.
!     ! If this has a severe negative performance impact, I should get rid of this
!     ! if (xhat<x(1) .and. ilo/=1) STOP 'bsearch: wrong case 1'
!     ! if (xhat>x(size(x)) .and. ilo/=size(x)-1) STOP 'bsearch: wrong case 2'
!     ! if (size(x)==1 .and. ilo/=1) STOP 'bsearch: wrong case 3'
!     ! if (size(x)>=2 .and. xhat>=x(1) .and. xhat<=x(size(x)) .and. (xhat<x(ilo) .or. xhat>x(ilo+1))) STOP 'bsearch: wrong case 4'

end subroutine 

subroutine bsearch_core(ilo,xhat,x,search_lb,search_ub)
    implicit none
    real(rt), intent(in) :: x(*),xhat
    integer, intent(in) :: search_lb,search_ub
    integer, intent(out) :: ilo
    !local
    integer :: ihi,mid
        
    ilo = search_lb
    ihi = search_ub
    mid = (ihi+ilo)/2 ! mid will have ilo<=mid<ihi as long as ihi>ilo. If ihi=ilo, then mid=ilo, so the 
                      ! while loop is not entered and ilo=1 is returned 
    
    do while (mid>ilo) ! mid=ilo means ihi=ilo+1.

        ! Here, we must have ihi>ilo and ilo<=mid<ihi

        ! Bisect. 
        ! If xhat<x(1), then xhat>=x(mid) will never occur and ilo will never be modified.
        ! If xhat>=x(nx), then xhat>=x(mid) will always occur (note mid<ihi). Consequently, 
        !      ilo = mid occurs until mid==ilo, which only occures if ihi==ilo+1.
        if (xhat>=x(mid)) then
            ilo = mid
        else
            ihi = mid
        end if

        mid = (ihi+ilo)/2 

    end do

end subroutine 

! Given sorted (ascending) lists xhat and x, finds ilo as if call bsearch(ilo(i),xhat(i),x) was called, but does so faster
subroutine bsearch_many(ilo,xhat,x)
    implicit none
    integer, intent(out) :: ilo(:)
    real(rt), intent(in) :: xhat(:),x(:)
    ! local
    integer :: n,nx

    n = size(ilo) 
    nx = size(x)

    if (n>=2) then
        call bsearch_core(ilo(1),xhat(1),x,1,nx)
        call bsearch_core(ilo(n),xhat(n),x,ilo(1),nx)

        if (n==2) return
        call bsearch_many_core(ilo,xhat,x,1,n)

    elseif (n==1) then
        call bsearch(ilo(1),xhat(1),x)
    elseif (n<=0) then
        stop 'ERROR: bsearch_many: size(ilo)<=0'
    end if


end subroutine

! Assumes ilo(lb) and ilo(ub) are known, exploits monotonicity
recursive subroutine bsearch_many_core(ilo,xhat,x,lb,ub)
    implicit none
    integer, intent(inout) :: ilo(*)
    real(rt), intent(in) :: xhat(*),x(*)
    integer, intent(in) :: lb,ub
    ! local
    integer :: m

    m = (lb+ub)/2

    call bsearch_core(ilo(m),xhat(m),x,ilo(lb),ilo(ub)+1) ! need to add 1 to the upper bound because ilo always returns the lower part of the bracketing interval

    if (m>lb+1) call bsearch_many_core(ilo,xhat,x,lb,m)
    if (ub>m+1) call bsearch_many_core(ilo,xhat,x,m,ub)


end subroutine 


! linspace(a,b,n) (of course) constructs a grid of n linearly spaced points between
! a and b. If n==1, then the grid is just "a", consistent with Matlab. If n<=0, the routine
! stops with an error.
function linspace(x1,x2,n) result(grid)
    implicit none
    real(rt), intent(in) :: x1,x2
    integer, intent(in) :: n
    real(rt), dimension(1:n) :: grid
    !local
    integer :: i 

    if (n>1) then
        grid(1) = x1
        do i=2,n-1
            grid(i) = x1 + (x2-x1)*(real(i,rt)-1._rt)/(real(n,rt)-1._rt)
        end do   
        grid(n) = x2
    elseif (n==1) then
        grid(1) = x1
    else
        STOP 'linspace: ERROR: grid has size <= 0'
    end if
end function linspace

! Tic and toc for timing routines.
! Because of the odd behavior of the system clock, there is a known bug here that is 
! more likely to be experienced for long run times.
subroutine tic()
    implicit none
    ! call cpu_time(time_start)
    ! call system_clock(count=count_start,count_rate=countrate_start)
    call date_and_time(values=datetime_start)
end subroutine tic
subroutine toc(time_out)
    implicit none
    real(rt), intent(out), optional :: time_out
    integer(8) :: datetime_end(8), ta(8)
    real(rt) :: minutes,seconds
    logical :: addmonth

    call date_and_time(values=datetime_end)

    ! Starting at the start time, go by minutes and increment
    ta = datetime_start
    minutes = 0d0
    do while (minutes<=144000) ! go for up to 100 days (place the max in case there is an error below)
        if (all(ta([1,2,3,5,6])==datetime_end([1,2,3,5,6]))) exit
        minutes = minutes + 1._rt

        ta(6) = ta(6) + 1 ! add a minute
        if (ta(6)>59) then
            ta(6) = 0
            ta(5) = ta(5) + 1 ! add an hour
            if (ta(5)>23) then
                ta(5) = 0 
                ta(3) = ta(3) + 1 ! add a day
                select case (ta(2))
                    case (2)
                        addmonth = ta(3)>28
                    case (4,6,9,11)
                        addmonth = ta(3)>30
                    case default
                        addmonth = ta(3)>31
                end select
                if (addmonth) then
                    ta(3) = 1 ! reset day to 1
                    ta(2) = ta(2) + 1 ! add a month
                    if (ta(2)>12) then
                        ta(1) = ta(1) + 1 ! add a year
                    end if
                end if
            end if
        end if
    end do

    ! Now have the measure down to minutes, get seconds and milliseconds
    seconds = datetime_end(7) - datetime_start(7) + (datetime_end(8) - datetime_start(8))/1000d0

    if (present(time_out)) then
        time_out = minutes + seconds/60d0
    else
        write(*,'(A,f14.6,A,f14.6,A,f14.6,A,f14.6,A)') 'Elapsed time: ',&
            (minutes+seconds/60d0)/60d0,' (hr) or ',&
            minutes+seconds/60d0,' (min) or ',&
            minutes*60d0 + seconds,' (sec) or ',&
            minutes*60d3 + seconds*1d3,' (ms)'
    end if

    
end subroutine toc


end module mod_core

module mod_core_test
    use mod_core
    implicit none
contains

    subroutine test_bsearch_many()
        use mod_types
        implicit none
        integer, allocatable :: ilo(:),ilo2(:)
        real(rt), allocatable :: xhat(:),x(:)
        real(rt) :: tmp
        integer, parameter :: n = 1000, nx = 570
        real(rt), parameter :: xlb = -3._rt, xub = 4._rt
        real(rt), parameter :: xhatlb = -6._rt, xhatub = 5._rt
        integer :: i,iter

        allocate(x(nx))
        allocate(xhat(n))
        allocate(ilo(n))
        allocate(ilo2(n))

        ! Check that the position agrees with bsearch
        do iter = 1,1000
            call randomgrids()
            call bsearch_many(ilo,xhat,x)
            do i = 1,n
                call bsearch(ilo2(i),xhat(i),x)
            end do
            if (any(ilo/=ilo2)) stop 'test_bsearch_many: failed test 1'
        end do

        print*,'test_bsearch_many: passed'

    contains

        subroutine randomgrids()
            implicit none

            ! Create increasing random lists
            xhat(1) = 0._rt
            do i = 2,n
                call random_number(tmp)
                xhat(i) = xhat(i-1) + tmp
            end do
            xhat = xhat*(xhatub-xhatlb) + xhatlb
            
            x(1) = 0._rt
            do i = 2,nx
                call random_number(tmp)
                x(i) = x(i-1) + tmp
            end do
            x = x*(xub-xlb) + xlb
        end subroutine 

    end subroutine

end module mod_core_test
! Collection of utility programs
! -> isNaN and isInvalidNumber
! -> sleep
! -> getcmdarg_? : gets a requested command argument (the type must be specified by "?", d for real(rt), i for integer, s for string)
! -> *_delim_arg : process a delimited string
! -> conv_i_to_s : convert integer to string
module mod_util
    use mod_types
    implicit none
    private :: getcmdarg_core

    type timer 
        real(8) :: refpt
        logical :: ispaused = .false.
        real(8) :: total = 0d0
    contains
        procedure, pass :: stop => timer_stop
        procedure, pass :: start => timer_start
        procedure, pass :: elapsed => timer_elapsed
    end type

    type type_seed
        integer, allocatable :: saved_seed(:)
    contains
        procedure, pass :: save => seed_save
        procedure, pass :: revert => seed_revert
        procedure, pass :: simpleset => seed_simpleset
        procedure, pass :: randomize => seed_randomize
    end type

contains
    
    subroutine assert(tf,msg,val,ival)
        implicit none
        logical, intent(in) :: tf
        character(len=*), intent(in) :: msg
        real(rt), intent(in), optional :: val
        logical, intent(in), optional :: ival
        if (.not. tf) then
            if (present(val)) then
                write(0,*) 'ERROR: '//msg,val
                stop
            elseif (present(ival)) then
                write(0,*) 'ERROR: '//msg,ival
                stop 
            else
                write(0,*) 'ERROR: '//msg
                stop
            end if
        end if
    end subroutine

    ! https://gcc.gnu.org/onlinedocs/gfortran/RANDOM_005fSEED.html with slight modifications
    subroutine seed_randomize(this)
        use iso_fortran_env, only: int64
        implicit none
        class(type_seed), intent(in) :: this
        ! local
        integer, allocatable :: s(:)
        integer :: i, n, un, istat, dt(8) !, pid
        integer(int64) :: t

        call random_seed(size = n)
        allocate(s(n))
        ! First try if the OS provides a random number generator
        open(newunit=un, file="/dev/urandom", access="stream", &
             form="unformatted", action="read", status="old", iostat=istat)
        if (istat == 0) then
           read(un) s
           close(un)
        else
           ! Fallback to XOR:ing the current time and pid. The PID is
           ! useful in case one launches multiple instances of the same
           ! program in parallel.
           call system_clock(t)
           if (t == 0) then
              call date_and_time(values=dt)
              t = (dt(1) - 1970) * 365_int64 * 24 * 60 * 60 * 1000 &
                   + dt(2) * 31_int64 * 24 * 60 * 60 * 1000 &
                   + dt(3) * 24_int64 * 60 * 60 * 1000 &
                   + dt(5) * 60 * 60 * 1000 &
                   + dt(6) * 60 * 1000 + dt(7) * 1000 &
                   + dt(8)
           end if
!            pid = getpid()
!            t = ieor(t, int(pid, kind(t)))
           do i = 1, n
              s(i) = lcg(t)
           end do
        end if
        call random_seed(put=s)

    contains
        ! This simple PRNG might not be good enough for real work, but is
        ! sufficient for seeding a better PRNG.
        function lcg(s)
          integer :: lcg
          integer(int64) :: s
          if (s == 0) then
             s = 104729
          else
             s = mod(s, 4294967296_int64)
          end if
          s = mod(s * 279470273_int64, 4294967291_int64)
          lcg = int(mod(s, int(huge(0), int64)), kind(0))
        end function lcg
    end subroutine 


    ! Saves seed
    subroutine seed_save(this)
        implicit none
        class(type_seed), intent(inout) :: this
        integer :: lenseed

        if (allocated(this%saved_seed)) deallocate(this%saved_seed)

        call random_seed(size=lenseed)

        allocate(this%saved_seed(lenseed)) 

        call random_seed(get=this%saved_seed)

    end subroutine 

    ! Reverts to a saved seet
    subroutine seed_revert(this)
        implicit none
        class(type_seed), intent(in) :: this
        if (allocated(this%saved_seed)) then
            call random_seed(put=this%saved_seed)
        else
            write(0,*) 'WARNING: seed_revert: revert was called, but no seed had been saved, doing nothing.'
        end if
    end subroutine 

    ! Sets the seed based just on one integer value
    subroutine seed_simpleset(this,i)
        implicit none
        class(type_seed), intent(in) :: this
        integer, intent(in) :: i
        ! local
        integer :: iseed,lenseed
        integer, allocatable :: seed(:)
        call random_seed(size=lenseed)
        allocate(seed(lenseed))
        seed(1) = i     
        do iseed = 2,lenseed
            seed(iseed) = mod(seed(iseed-1)*17+7919,(huge(seed(1))-10000)/20) ! 7919 is prime
        end do
        call random_seed(put=seed)
    end subroutine 
    

    !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    ! Compute the time
!     subroutine timer_init(this)
!         use omp_lib
!         implicit none
!         class(timer) :: this
!         this%refpt = omp_get_wtime()
!         this%ispaused = .false.
!         this%total = 0d0
!     end subroutine 
    subroutine timer_stop(this)
        use omp_lib
        implicit none
        class(timer) :: this
        this%total = this%total + (omp_get_wtime()-this%refpt)
        this%ispaused = .true.
    end subroutine
    subroutine timer_start(this)
        use omp_lib
        implicit none
        class(timer) :: this
        this%refpt = omp_get_wtime()
        this%ispaused = .false.
    end subroutine
    real(8) function timer_elapsed(this)
        use omp_lib
        implicit none
        class(timer) :: this

        if (this%ispaused) then
            timer_elapsed = this%total 
        else
            timer_elapsed = this%total + (omp_get_wtime()-this%refpt)
        end if

    end function

    subroutine test_util()
        use ieee_arithmetic
        implicit none
        real(rt) :: x
        type(type_seed) :: randstate
        real(rt) :: t,t2,t3,t4
!         character(len=1000) :: str

        ! http://www.ibm.com/support/knowledgecenter/SSGH4D_14.1.0/com.ibm.xlf141.aix.doc/language_ref/isnan.html
        ! NOTE: fftrap must be disabled or taking sqrt(x) will crash.
        x = -1._rt
        if (ieee_support_datatype(x)) then
            if (ieee_support_sqrt(x)) then 
                if (ieee_support_nan(x)) then
                    call assert(ieee_is_nan(sqrt(x)),'ERROR: test_util: ieee_is_nan does not correctly identify sqrt(-1) as being nan')
                end if
            end if
        end if
        call assert(isNaN(sqrt(x)),'ERROR: test_util: ieee_is_nan does not correctly identify sqrt(-1) as being nan')

!         if (.not. isInvalidNumber(log(x))) STOP 'ERROR: isInvalidNumber failed test 1'
        if (isInvalidNumber(3e0_rt)) STOP 'ERROR: isInvalidNumber failed test 2'

        ! str = printx([-tiny(0d0),-huge(0d0)])
        ! print*,'printx',trim(str)

        call randstate%randomize()
        call randstate%save()
        call random_number(t)

        call randstate%simpleset(120480)
        call random_number(t2)
        call assert(t2/=t,'t2 should not equal t')

        call randstate%simpleset(120480)
        call random_number(t3)
        call assert(t2==t3,'t2 must equal t3')

        call randstate%revert()
        call random_number(t4)
        call assert(t==t4,'t must equal t4')

        print*,'test_util: passed'

    end subroutine 

    ! NOTE: this may not work depending on compiler options. E.g., if -ffast-math is enabled, then IEEE arithmetic rules
    ! are ignored.
!     function isNaN(x) result(tf)
!         use ieee_arithmetic, only: ieee_is_nan ! F2003 intrinsic
!         implicit none
!         real(rt) :: x
!         logical :: tf
! 
!         ! NaN is not equal to anything, not even itself
!         tf = ieee_is_nan(x)
! 
!     end function 
    elemental function isNaN(x) result(tf)
        implicit none
        real(rt), intent(in) :: x 
        logical :: tf

        ! NaN is not equal to anything, not even itself
        tf = x/=x

    end function 

    ! This works even with ffast math
    elemental function isInvalidNumber(x) result(tf)
        implicit none
        real(rt), intent(in) :: x
        logical :: tf

        ! This should evaluate to false if x is a normal number (not infinite, not NaN)
        tf = .not. ( x>=-huge(0d0) .and. x<=huge(0d0))

    end function 


    ! http://stackoverflow.com/questions/6931846/sleep-in-fortran
    subroutine sleep(sec)
        use, intrinsic :: iso_c_binding, only: c_int
        implicit none
        integer, intent(in) :: sec
        integer (c_int) :: sec_c
        integer (c_int) :: err

        interface
            function callcsleep(seconds)  bind ( C, name="sleep" )
                use, intrinsic :: iso_c_binding, only: c_int
                import
                implicit none
                integer (c_int) :: callcsleep
                integer (c_int), value :: seconds
            end function 
        end interface

        sec_c = int(sec,c_int)

        err = callcsleep(sec_c)

    end subroutine

    !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    ! DISPLAY A DOUBLE PRECISION VECTOR IN A NICE FORMAT
    function printx(x) result(strout)
        implicit none
        integer, parameter :: stride = 17
        real(rt), intent(in) :: x(:)
        character(len=1+stride*size(x)) :: strout
        ! local
        character(len=1+stride*size(x)) :: str
        integer :: ix

        if (1+stride*size(x)>len(str)) then
            str = 'size(x) is too large, cannot display'
            return 
        end if

        str = '' ! blank it out
        write(str(1:1),'(A)') '['
        do ix = 1,size(x)-1
            write(str(2+stride*(ix-1):1+stride*ix),'(ES16.8E3,A)') x(ix),','
        end do
        write(str(2+stride*(ix-1):1+stride*ix),'(ES16.8E3,A)') x(size(x)),']'

        strout = str
        do ix = 1,1+stride*size(x)
            if (str(ix:ix)=="E") strout(ix:ix)="D"
        end do


    end function


    !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    ! PARSING
    ! Given a delimited string, such as "xtoler=1d-6,printmod=1" (in this case, the delimiter could be ","), gives the ith
    ! argument (1 would be "xtoler=1d-6" and 2 would be "printmod=1" in this example). There is also a routine to determine
    ! how many arguments there are.
    function count_delim_arg(str,del) result(n)
        implicit none
        character(len=*) :: str
        character(len=1) :: del
        integer :: n
        ! local
        integer :: i,nloc

        nloc = 0
        do i = 1,len(str)
            if (str(i:i)==del) nloc = nloc + 1
        end do
        n = nloc + 1

    end function 
    function get_delim_arg(str,del,i) result(arg)
        implicit none
        character(len=*) :: str
        character(len=1) :: del
        integer, intent(in) :: i
        character(len=len(str)) :: arg
        ! local
        integer :: j,n,ta,tb

        if (i>count_delim_arg(str,del)) then
            arg = 'error_getting_arg'
            return
        end if

        n = 0
        tb = 1 
        do j = 1,len(str)
            if (str(j:j)==del .or. j==len(str)) then
                ta = tb
                tb = j
                n = n + 1
                if (n==i) exit
            end if
        end do

        ! Here, should have str(ta:tb) almost as the argument. 
        ! However, str(ta) and str(tb) may or may not be delimiters.
        if (ta>1) then 
            ta = ta + 1
        end if
        if (tb<len(str)) then
            tb = tb - 1
        end if

        ! Output the string
        arg = '' ! put blank, or else there can be garbage
        arg(1:tb-ta+1) = str(ta:tb)

    end function 
    subroutine test_delim_arg()
        implicit none
        character(len=200) :: str,chk
        character(len=20) :: a,b,c
        integer :: n 

        a = "argone"
        b = "argtwo"
        c = "argthree"
        
        str = trim(a)//";"//trim(b)//";"//trim(c)

        
        n = count_delim_arg(str,";")

        if (n/=3) stop 'test_delim_arg: failed test 1'

        chk = get_delim_arg(str,";",1)
        if (trim(chk)/=trim(a)) then
            print*,'trim(a)  :',trim(a),'len_trim(a)',len_trim(a)
            print*,'trim(chk):',trim(chk),'len_trim(chk)',len_trim(chk)
            print*,'trim(chk)==trim(a)',trim(chk)==trim(a)
            print*,'trim(chk)/=trim(a)',trim(chk)/=trim(a)
            print*,'chk(1:1)==a(1:1)',chk(1:1)==a(1:1)
            print*,'chk(2:2)==a(2:2)',chk(2:2)==a(2:2)
            print*,'chk(3:3)==a(3:3)',chk(3:3)==a(3:3)
            print*,'chk(4:4)==a(4:4)',chk(4:4)==a(4:4)
            print*,'chk(5:5)==a(5:5)',chk(5:5)==a(5:5)
            print*,'chk(6:6)==a(6:6)',chk(6:6)==a(6:6)
            print*,'chk(7:7)==a(7:7)',chk(7:7)==a(7:7)
            stop 'test_delim_arg: failed test 2a'
        end if
        chk = get_delim_arg(str,";",2)
        if (trim(chk)/=trim(b)) stop 'test_delim_arg: failed test 2b'
        chk = get_delim_arg(str,";",3)
        if (trim(chk)/=trim(c)) stop 'test_delim_arg: failed test 2c'

        print*,'test_delim_arg: passed'

    end subroutine




    !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    ! GET COMMAND LINE ARGUMENTS
    function getcmdarg_d(i) result(arg)
        implicit none
        integer, intent(in) :: i
        real(rt) :: arg
        ! local
        real(rt) :: xd
        integer :: xi
        character(1024) :: xs

        call getcmdarg_core(i,xd,xi,xs,'d')
        arg = xd
        
    end function 
    function getcmdarg_i(i) result(arg)
        implicit none
        integer, intent(in) :: i
        integer :: arg
        ! local
        real(rt) :: xd
        integer :: xi
        character(1024) :: xs

        call getcmdarg_core(i,xd,xi,xs,'i')
        arg = xi
        
    end function 
    function getcmdarg_s(i) result(arg)
        implicit none
        integer, intent(in) :: i
        character(len=1024) :: arg
        ! local
        real(rt) :: xd
        integer :: xi
        character(1024) :: xs

        call getcmdarg_core(i,xd,xi,xs,'s')
        arg = adjustl(trim(xs))
        
    end function 

    subroutine getcmdarg_core(i,xd,xi,xs,type_expected)
        implicit none
        integer, intent(in) :: i ! which command argument is desired
        real(rt), intent(inout) :: xd ! output (if real)
        integer, intent(inout) :: xi ! output (if integer)
        character(*), intent(inout) :: xs ! output (if string)
        character(1), intent(in) :: type_expected ! d for double, i for integer, s for string
        ! local
        character(len=1024) :: str

        if (i>command_argument_count()) then
            xd = -1d0
            xi = -1
            xs = "-1"
            write(0,'(A,i5,A,i5,A)') 'WARNING: getcmdarg: arg ',i,' couldn''t be obtained, only ',&
                                               command_argument_count(),' arguments present'
            return
        end if

        ! Get the argument as a string
        call get_command_argument(i,str)

        ! Convert it to the specified type
        select case (type_expected)
            case ('d')
                xd = -1d0
                ! read(str,'(d30.16)',err=999) xd
                read(str,*,err=999) xd
            case ('i')
                xi = -1
                read(str,'(i30)',err=999) xi
            case ('s')
                xs = trim(adjustl(str))
        end select

        return

        ! If caught an error, warn about it
999     write(0,*) 'ERROR: mod_util: getcmdarg_core: error while reading command argument ',i,' argument is ',trim(adjustl(str))

    end subroutine 
    !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    function conv_s_to_i(s) result(i)
        implicit none
        character(*), intent(in) :: s
        integer :: i
        ! local
        integer :: itmp
        i = -1
        read(s,*,err=999) itmp
        i = itmp
        return

999     write(0,*) 'ERROR: could not convert '//s//' to an integer'
    end function 

    function conv_i_to_s(i) result(s)
        implicit none
        integer, intent(in) :: i
        character(len=11) :: s ! largest integer is 2147483647 (10 digits)
        ! local
        integer :: j,k
        character(len=1) :: t

        s = ' '

        j = abs(i)

        if (j==0) then
            s = '0'
            return
        end if

        k = 0
        do while (j>0)
            k = k+1
            write(s(k:k),'(i1)') mod(j,10)
            j = j/10
        end do

        ! Now reverse it
        do j = 1,k/2
            t = s(j:j)
            s(j:j) = s(k-j+1:k-j+1)
            s(k-j+1:k-j+1) = t
        end do

        ! If negative, shift right and add -
        if (i<0) then
            do j = k,1,-1
                s(j+1:j+1) = s(j:j)
            end do
            s(1:1) = '-'
        end if

    end function


end module

! program main
!     use mod_cmdarg
!     implicit none
!     character(1024) :: str
! 
!     call get_command(str)
!     print*,repeat('*',100)
!     print*,'Full command line: ',trim(adjustl(str))
!     print*,repeat('*',100)
!     print*,'command argument #0 (name of program):    ', trim(getcmdarg_s(0))
!     print*,'command argument #1 (expecting double):   ', getcmdarg_d(1)
!     print*,'command argument #2 (expecting integer):  ', getcmdarg_i(2)
!     print*,'command argument #3 (expecting string):   ', trim(getcmdarg_s(3))
!     print*,'command argument #4 (only 3 args, error): ', trim(getcmdarg_s(4))
!     print*,repeat('*',100)
! 
! end program
module mod_sort
    use mod_types
    implicit none
    integer, parameter, private :: qsort_nthresh = 100 ! When a list is <= this threshold, insertion sort is called. MUST be >= 3

    ! I found using ifort that a high number like 60 or 100 was close to optimal.
    ! I usually have nearly sorted data, it's better to err on the high side.
    ! I also found middle pivot was better slightly, but will keep median as it is more robust.
    
    interface msort
        module procedure msort_i,msort_r
    end interface msort
    interface qsort
        module procedure qsort_i,qsort_r
    end interface qsort
    interface isort
        module procedure isort_i,isort_r
    end interface isort

    ! qsort is based on algorithms I found in several places online, most citing textbooks
    ! isort is based on wiki I believe
    ! msort I think I wrote completely on my own (which is why it's so bad!)

contains

    ! tests whether x is sorted according to optional argument mode ("ascend" or "descend").
    ! By default, checks for ascended sort.
    function issorted(x,mode) result(sorted)
        implicit none
        logical :: sorted
        real(rt), dimension(1:) :: x
        character(len=*), intent(in) :: mode    !either 'ascend' or 'descend'
        !local
        integer :: i

        if (size(x)==1) then
            sorted = .TRUE. 
            return
        end if

        if (mode=='ascend') then
            do i = 2,size(x)
                if (x(i)<x(i-1)) then
                    sorted = .FALSE.
                    return
                end if
            end do
        elseif (mode=='descend') then
            do i = 2,size(x)
                if (x(i)>x(i-1)) then
                    sorted = .FALSE.
                    return
                end if
            end do
        else
            write(*,*) 'issorted: error: sorting mode not found'
            STOP 'issorted: error: sorting mode not found'
        end if

        sorted = .TRUE.

    end function issorted

    ! Insertion sort
    pure subroutine isort_i(X,I)
        implicit none
        integer, dimension(1:), intent(inout) :: X
        integer, dimension(1:size(X)), intent(out) :: I
        !local
        real(rt), dimension(1:size(X)) :: Xreal

        Xreal = real(X,rt) ! Convert to double prec
        call isort_r(Xreal,I) ! Sort the double prec
        X = nint(Xreal) ! Nearest integer to Xreal

    end subroutine isort_i 
    pure subroutine isort_r(X,I) 
        implicit none
        real(rt), dimension(1:), intent(inout) :: X ! Input/Output array
        integer, dimension(1:size(X)), intent(out) :: I ! I s.t. X(I) is sorted
        !local
        integer :: j

        do j = 1,size(X)
            I(j) = j
        end do

        call isort_r_core(X,I)

    end subroutine isort_r
    pure subroutine isort_r_core(X,I) 
        implicit none
        real(rt), dimension(1:), intent(inout) :: X ! Input/Output array
        integer, dimension(1:), intent(inout) :: I ! I s.t. X(I) is sorted
        !local
        real(rt) :: val
        integer :: k,j,n,ival
        
        n = size(x)

        ! Loop over items to be sorted 
        do k = 2,n
            ! 1 to k-1 is sorted. Now place X(i) in the sorted list, shifting
            ! the sorted elements to the right as necessary.
            val = X(k)
            ival = I(k)
            ! Going backwards from j=k-1 to 1, if X(k)<X(j), then we know X(k) is to the 
            ! left of X(j) somewhere. Consequently, X(j) should be shifted to the right.
            ! If not, then X(k)>=X(j), and so X(k) should be put in the j+1th position
            do j = k-1,1,-1
                if (val<X(j)) then
                    X(j+1) = X(j) ! Shift forward
                    I(j+1) = I(j) ! Ditto
                else
                    ! Exit here
                    ! When we exit, j retains it's last value
                    exit
                end if
            end do
            ! The j value here is either what it was when "exit" was executed or 0 if the loop completed
            ! Consequently, we always want to replace the j+1 th value.
            X(j+1) = val ! Put
            I(j+1) = ival
        end do

    end subroutine isort_r_core

    ! Quick sort
    ! Sorts X in place and also returns I such that X=X(I) with RHS X input.
    subroutine qsort_i(X,I)
        implicit none
        integer, dimension(1:), intent(inout) :: X
        integer, dimension(1:size(X)), intent(out) :: I
        !local
        real(rt), dimension(1:size(X)) :: Xreal
print*,'WARNING: I AM NOT SURE WHETHER THIS VERSION OF QUICKSORT IS STABLE!'

        Xreal = real(X,rt) ! Convert to double prec
        call qsort_r(Xreal,I) ! Sort the double prec
        X = nint(Xreal) ! Nearest integer to Xreal
        
    end subroutine qsort_i
    subroutine qsort_r(X,I)
        implicit none
        real(rt), dimension(1:), intent(inout) :: X ! Input/Output array
        integer, dimension(1:size(X)), intent(out) :: I ! I s.t. X(I) is sorted
        ! local
        integer :: l,n ! # of elts
print*,'WARNING: I AM NOT SURE WHETHER THIS VERSION OF QUICKSORT IS STABLE!'
        
        n = size(X)
        
        ! Test whether should go directly to insertion sort
        if (n<=qsort_nthresh) then
            call isort_r(X,I)
            return
        end if 

        ! Set the index array
        do l = 1,n
            I(l) = l
        end do

        ! Call the recursion 
        call qsort_r_core(X,I,1,n)

        ! I have seen it suggested that one should not do the many isorts in the core, 
        ! but one large one at the end since isort works well on almost sorted data.
        ! I do not pursue that here.
        ! call isort_core(I,X)

    end subroutine qsort_r

    ! The core routine for qsort
    pure recursive subroutine qsort_r_core(X,I,left,right)
        implicit none
        real(rt), dimension(1:), intent(inout) :: X ! Input/Output array
        integer, dimension(1:), intent(inout) :: I ! I s.t. X(I) is sorted
        integer, intent(in) :: left,right
        !local
        integer :: p,l

        ! If size is small, use isort
        if (right-left+1<=qsort_nthresh) then
            call isort_r_core(X(left:right),I(left:right))
            return
        end if

        ! Middle
!         p = (left + right)/2

        ! Median of three
        l = (left + right)/2 ! l used for temporary storage
        if (X(l)>=X(left)) then
            ! l >= left
            if (X(l)<=X(right)) then
                ! l <= right
                p = l
            else
                ! l>=left and l>right
                if (X(left)<=X(right)) then
                    ! left<=right
                    p = right
                else
                    p = left
                end if
            end if
        else
            ! l < left
            if (X(l)>=X(right)) then
                ! l >= right
                p = l
            else
                ! l <= right
                if (X(left)<=X(right)) then
                    ! left<=right and l<left
                    p = left
                else
                    ! left > right and l<=right
                    p = right
                end if
            end if
        end if

        ! Partition the list
        call qsort_r_partition(l,X,I,left,right,p)

        ! Now, recursively sort the sublists.
        if (left<l-1) call qsort_r_core(X,I,left,l-1)
        if (l<right) call qsort_r_core(X,I,l,right)
        
    end subroutine qsort_r_core

    ! Given the array X and bounds left,right, and a pivot p, 
    ! rearranges X(left:right) so that X(left:l)<=X(p), X(l+1:right)>=X(p) 
    ! where l is integer output. 
    pure subroutine qsort_r_partition(l,X,I,left,right,p)
        implicit none
        real(rt), dimension(1:), intent(inout) :: X
        integer, dimension(1:), intent(inout) :: I
        integer, intent(in) :: left,right,p
        integer, intent(out) :: l
        ! local
        real(rt) :: pval, tmp
        integer :: r, tmpi

        ! Get the pivot value
        pval = X(p)

        ! 
        l = left
        r = right
        do while (l<=r) 

            ! Find the first l greater or equal to the current l such that X(l)>=pval is found
            do while(X(l)<pval) ! Exits when X(l)>=pval. Since pval lies in left,right, this is not an infinite loop
                l = l+1
            end do
            ! Find the first r less than or equal to the current r such that X(r)<=pval is found
            do while (X(r)>pval) ! Exits when X(r)<=pval
                r = r-1
            end do

            ! Here, X(l)>=pval and X(r)<=pval and l<r, so swap these elements to guarantee
            if (l<=r) then
                tmp = X(l)
                X(l) = X(r)
                X(r) = tmp
                tmpi = I(l)
                I(l) = I(r)
                I(r) = tmpi
                ! At this point, left:l satisfies X(left:l)<=pval and X(r:right)>=pval. 

                ! Increment l and r
                l = l+1
                r = r-1
            end if

        end do
        
        ! Here, since l>r (we exited the loop) and X(left:l-1)<=pval and X(l:right)>=pval, we 
        ! have partitioned X 
        
    end subroutine qsort_r_partition 



    ! Performs mergesort for real numbers
    subroutine msort_r(A,I)
        use mod_orderpack
        implicit none
        real(rt), dimension(1:), intent(inout) :: A !list to be sorted
        integer, dimension(1:size(A)), intent(out) :: I !index of sorting s.t. to sort B according to A use B=B(I)

        ! Call the ORDERPACK 2.0 routine for a stable mergesort
        call mrgrnk(A,I)
        A = A(I)

    end subroutine msort_r

!     !Returns sorted vector A and indices I sorted according to A
!     ! I am embarassed by this routine. I should go back and fix it to use
!     ! pointers rather than straight up copying a million times with automatic
!     ! local arrays.
!     recursive subroutine msort_recursion_r(A,I)
!         implicit none
!         real(rt), dimension(1:), intent(inout) :: A
!         integer, dimension(1:), intent(inout) :: I
!         !local
!         real(rt), dimension(1:size(A)) :: Acopy
!         integer, dimension(1:size(I)) :: Icopy
!         integer :: na,j,a1,a2
!         na = size(a)
! 
!         !If na is 0 or 1, the list is sorted, so return
!         if (na<=1) then        
!             return
!         end if
! 
!         !If na>1 then divide the list in half
!         call msort_recursion_r(A(1:na/2),I(1:na/2))
!         call msort_recursion_r(A(na/2+1:na),I(na/2+1:na))
! 
!         !Merge these two sorted lists
!         a1 = 1
!         a2 = na/2+1
!         do j=1,na      
!             if (a1<=na/2 .AND. a2<=na) then
!                 if (A(a1)<=A(a2)) then
!                     Acopy(j) = A(a1)
!                     Icopy(j) = I(a1)
!                     a1 = a1+1
!                 else
!                     Acopy(j) = A(a2)
!                     Icopy(j) = I(a2)
!                     a2 = a2+1
!                 end if
!                 !if (j==na) exit
!             elseif (a1>na/2) then !first list has been exhausted, fill remaining elements with 2nd list
!                 Acopy(j:na) = A(a2:na)
!                 Icopy(j:na) = I(a2:na)
!                 exit
!             elseif (a2>na) then !second list has been exhausted, fill remaining elements with 1st list
!                 Acopy(j:na) = A(a1:na/2)
!                 Icopy(j:na) = I(a1:na/2)
!                 exit
!             end if
!         end do
! 
!         A = Acopy
!         I = Icopy
! 
!     end subroutine msort_recursion_r

    subroutine msort_i(A,I)
        use mod_orderpack
        implicit none
        integer, dimension(1:), intent(inout) :: A !list to be sorted
        integer, dimension(1:size(A)), intent(out) :: I !index of sorting s.t. to sort B according to A use B=B(I)

        ! Call the ORDERPACK 2.0 routine for a stable mergesort
        call mrgrnk(A,I)
        A = A(I)

    end subroutine msort_i

!     !Returns sorted vector A and indices I sorted according to A
!     recursive subroutine msort_recursion_i(A,I)
!         implicit none
!         integer, dimension(1:), intent(inout) :: A
!         integer, dimension(1:), intent(inout) :: I
!         !local
!         integer, dimension(1:size(A)) :: Acopy
!         integer, dimension(1:size(I)) :: Icopy
!         integer :: na,j,a1,a2
!         na = size(a)
! 
!         !If na is 0 or 1, the list is sorted, so return
!         if (na<=1) then        
!             return
!         end if
! 
!         !If na>1 then divide the list in half
!         call msort_recursion_i(A(1:na/2),I(1:na/2))
!         call msort_recursion_i(A(na/2+1:na),I(na/2+1:na))
! 
!         !Merge these two sorted lists
!         a1 = 1
!         a2 = na/2+1
!         do j=1,na      
!             if (a1<=na/2 .AND. a2<=na) then
!                 if (A(a1)<=A(a2)) then
!                     Acopy(j) = A(a1)
!                     Icopy(j) = I(a1)
!                     a1 = a1+1
!                 else
!                     Acopy(j) = A(a2)
!                     Icopy(j) = I(a2)
!                     a2 = a2+1
!                 end if
!                 !if (j==na) exit
!             elseif (a1>na/2) then !first list has been exhausted, fill remaining elements with 2nd list
!                 Acopy(j:na) = A(a2:na)
!                 Icopy(j:na) = I(a2:na)
!                 exit
!             elseif (a2>na) then !second list has been exhausted, fill remaining elements with 1st list
!                 Acopy(j:na) = A(a1:na/2)
!                 Icopy(j:na) = I(a1:na/2)
!                 exit
!             end if
!         end do
! 
!         A = Acopy
!         I = Icopy
! 
!     end subroutine msort_recursion_i

    subroutine test_sortmethod(sortMethod,n)
        implicit none
        interface 
            subroutine sortMethod(X,I)
                use mod_types
                implicit none
                real(rt), dimension(1:), intent(inout) :: X
                integer, dimension(1:size(X)), intent(out) :: I
            end subroutine sortMethod
        end interface
        integer, intent(in) :: n
        real(rt), allocatable :: X(:),X0(:)
        integer, allocatable :: I(:) ! Output sorted ints
        logical :: passed
        real :: ta, tb
        ! Test data

        allocate(X(n),X0(n),I(n))

        call cpu_time(ta)

        ! Random input
        call random_number(X)
        X0 = X; call sortMethod(X,I); passed = issorted(X,'ascend') .and. all(X0(I)==X)
        if (.not. passed) STOP 'failed random data' 

        ! Sorted data
        X0 = X; call sortMethod(X,I); passed = issorted(X,'ascend') .and. all(X0(I)==X)
        if (.not. passed) STOP 'failed sorted data' 

        ! Descendingly sorted data
        X = -X;
        X0 = X; call sortMethod(X,I); passed = issorted(X,'ascend') .and. all(X0(I)==X)
        if (.not. passed) STOP 'failed descendingly sorted data' 

        ! Random, but not distinct
        ! gcc4.9.0 has a internal compiler error here, so I disabled this test
!         call random_number(X)
!         X = floor(X*1000d0) 
!         X0 = X; call sortMethod(X,I); passed = issorted(X,'ascend') .and. all(X0(I)==X)
!         if (.not. passed) STOP 'failed random but not distinct pts' 

        call cpu_time(tb)

        print*,'sorting time',tb-ta

        deallocate(X,X0,I)

    end subroutine test_sortmethod
    subroutine test_sort
        implicit none
        call test_sortmethod(isort_r,10); print*,'test_sort: isort_r passed'
        call test_sortmethod(qsort_r,1000); print*,'test_sort: qsort_r passed'
        call test_sortmethod(msort_r,1000); print*,'test_sort: msort_r passed'
    end subroutine test_sort

end module mod_sort
module mod_interp
    use mod_types
    use mod_core, only: bsearch
    implicit none

    interface sub_s2interp
        module procedure sub_s2interp_elt, sub_s2interp_vec, sub_s2interp_elt_loc, sub_s2interp_vec_loc, sub_s2interp_elt_loc_mass
    end interface
    interface sub_s2interp_precomp
        module procedure sub_s2interp_elt_loc_getTheta
    end interface

    integer, parameter, private :: search_cutoff = 20 ! If number of elements is less than this, lsearch is used, o/w bsearch. For random data, this should be close to 25. 

contains

    ! "search" attempts to provide the best of lsearch and bsearch. For small arrays, calls lsearch. For large, bsearch.
    subroutine search(ilo,xhat,x)
        implicit none
        real(rt), dimension(1:), intent(in) :: x
        real(rt), intent(in) :: xhat
        integer, intent(out) :: ilo
        
        if (size(x)<=search_cutoff) then
            call lsearch(ilo,xhat,x)
        else
            call bsearch(ilo,xhat,x)
        end if

!         if (xhat<x(1) .and. ilo/=1) STOP 'search: wrong case 1'
!         if (xhat>x(size(x)) .and. ilo/=size(x)-1) STOP 'search: wrong case 2'
!         if (size(x)==1 .and. ilo/=1) STOP 'search: wrong case 3'
!         if (size(x)>=2 .and. xhat>=x(1) .and. xhat<=x(size(x)) .and. (xhat<x(ilo) .or. xhat>x(ilo+1))) STOP 'search: wrong case 4'

    end subroutine

    ! Interpolation/Extrapolation search
    ! Tries to find ilo by assuming x is linear in a given interval
    ! If in fact x is linear, this is a very fast method. But worst case performance is O(n)
    ! Finds an i such that xhat \in [x(i),x(i+1)] if such an interval exists.
    subroutine isearch(ilo,xhat,x)
        implicit none
        real(rt), dimension(1:), intent(in) :: x
        real(rt), intent(in) :: xhat
        integer, intent(out) :: ilo

        if (size(x)==1) then
            ilo = 1
            return
        end if

        if (xhat<x(1)) then
            ilo = 1
        elseif (xhat>=x(size(x)-1)) then
            ilo = size(x)-1
        else 
            
            ! Take a guess at what it should be based on linear spacing
            ilo = ceiling( 1d-14 + (size(x)-1)*(xhat-x(1))/(x(size(x))-x(1)) ) ! adding a small contant helps ensure that if xhat is on the x grid, the ilo is correctly found

            ! If this fails, call bsearch
            if (ilo<1 .or. ilo>=size(x)) then
                call bsearch(ilo,xhat,x)
                return
            end if
            if (xhat<x(ilo) .or. xhat>x(ilo+1)) then
                call bsearch(ilo,xhat,x)
                return
            end if


!             if (ilo>=size(x)) then
!                 ilo = size(x)-1
!                 if (xhat<x(ilo)) then
!                     ! Error: switch to bsearch
!     !                 write(*,'(A)',advance='no') '1'
!                     call bsearch(ilo,xhat,x)
!                 end if
!             elseif (ilo<1) then
!                 ilo = 1
!                 if (xhat>x(2)) then
!                     ! Error: switch to bsearch
!     !                 write(*,'(A)',advance='no') '2'
!                     call bsearch(ilo,xhat,x)
!                 end if
!             else
!                 ! Check that x = max(min(x,x(n)),x(1)) has x \in [x(i),x(i+1)]
!     !             write(*,'(A)',advance='no') 'g'
!                 if (xhat<x(ilo) .or. xhat>x(ilo+1)) then
!                     ! Error: switch to bsearch
!     !                 write(*,'(A)',advance='no') '3'
!                     call bsearch(ilo,xhat,x)
!                 end if
!             end if
        end if
        
    end subroutine isearch

!     subroutine isearch(ilo,xhat,x)
!         implicit none
!         real(rt), dimension(1:), intent(in) :: x
!         real(rt), intent(in) :: xhat
!         integer, intent(out) :: ilo
!         ! local
!         integer :: itry,ihi
!         
!         ilo = 1
!         ihi = size(x)
! 
!         do while (ihi-ilo>1) ! Tests that there is a point between ilo and ihi (if not, return ilo!)
! 
!             ! If x(i) is linear in i between [ilo,ihi], then x(i) = x(ilo) + (x(ihi)-x(ilo))/(ihi-ilo)*(i-ilo)
!             ! So, xhat = x(i) occurs for i equal to 
!             ! xhat = x(ilo) + (x(ihi)-x(ilo))/(ihi-ilo)*(i-ilo)
!             ! (xhat-x(ilo))/(x(ihi)-x(ilo))*(ihi-ilo)  = i-ilo
!             ! i = ilo + (ihi-ilo)*(xhat-x(ilo))/(x(ihi)-x(ilo))
!             ! The last terms with x values is in [0,1] IF x is sorted. 
! 
!             itry = floor(ilo + (ihi-ilo)*(xhat-x(ilo))/(x(ihi)-x(ilo)))
!             
!             ! The following two if statemetns may only be necessary on the first pass 
!             ! (to check whether xhat<x(1) or xhat>x(n)), but since using mixed integer and double arithmetic,
!             ! probably safer to leave them.
!             
!             ! If itry is less than or equal to ilo, and x is sorted, this means xhat<x(ilo) so return ilo
!             ! Note: if itry==ilo, this can mean xhat>=x(ilo) and so must continue search
!             if (itry<ilo) then
!                 return 
!             end if
! 
!             ! If itry is greater than or equal to ihi, this means xhat>=x(ihi) so return ilo=ihi-1
!             if (itry>=ihi) then
!                 ilo = ihi-1
!                 return
!             end if
! 
!             ! Now itry is in [ilo,...,ihi-1] so following x checks are safe
!             
!             ! Construct the new bracket and test whether have found true ilo
!             if (x(itry)>xhat) then
!                 ! Here, ihi should be replaced. Note: already tested itry>=ihi and it was false
!                 ihi = itry 
!             else
!                 ! HERE: x(itry)<=xhat
!                 ! Since user is expecting near-linear case, test whether have succeeded in finding interval
!                 if (x(itry+1)>=xhat) then
!                     ! HERE: xhat \in [x(itry),x(itry+1)]
!                     ilo = itry
!                     return 
!                 end if
! 
!                 ! Here, ilo should be replaced. 
!                 ilo = max(itry,ilo+1) 
! 
!             end if
! 
!         end do
! 
!     end subroutine isearch

    ! This routine requires "step" as argument. The user should ensure 
    ! step = real(size(x)-1),8)/(x(size(x))-x(1)). 
    subroutine isearch_with_step(ilo,xhat,x,step)
        implicit none
        real(rt), dimension(1:), intent(in) :: x
        real(rt), intent(in) :: xhat,step
        integer, intent(out) :: ilo
        ! local
        integer :: itry,ihi
        
        ! Check whether exactly works
        ilo = floor(1e0_rt + step*(xhat-x(1)))
        if (ilo>=1 .and. ilo<=size(x)-1) then
            if (xhat>=x(ilo) .and. xhat<=x(ilo+1)) return            
        end if  

        ! If it didn't work, just call isearch for simplicity
        call isearch(ilo,xhat,x)

    end subroutine isearch_with_step
    
    ! Use sequential/linear search to find the lowerbound on a bracketing interval
    ! This has far worse theoretical performance than binary search O(n) compared to 
    ! O(log2(n)), but it is much easier for the compiler to understand and use effectively.
    ! Consequently for small arrays, it is faster.
    subroutine lsearch(ilo,xhat,x)
        implicit none
        real(rt), dimension(1:), intent(in) :: x
        real(rt), intent(in) :: xhat
        integer, intent(out) :: ilo
        ! local
!         integer :: ihi

!    do ihi = 2,size(x)
!        ! If here, then size(x)-1>=1 <=> size(x)>=2 (if size(x)==1, we just return ilo = 1)
! 
!        ! Three cases to be concerned with
!        ! (1) xhat<x(1)
!        ! (2) xhat>=x(n)
!        ! (3) xhat>=x(1) .and. xhat<x(n)
! 
!        ! Test whether xhat<x(ilo). 
!        ! In case (1), this will always evaluate to true. So we should exit immediately returning ilo=ilo-1.
!        ! In case (2), this will never evaluate to true. 
!        ! In case (3), this will evaluate to true when xhat<x(ilo) and there was no j<ilo such that xhat<x(j).
!        !   Consequently, we have xhat>=x(j) for all j<ilo, and in particular, xhat>=x(ilo-1). This means, xhat \in [x(ilo-1),x(ilo))
!        !   So, return ilo-1, just as we did in case (1).
!        if (xhat<x(ihi)) then
!            ilo = ihi-1
!            return
!        end if
!    end do
!    ! If still here, we have case (2) or the size of x is 1. So, we return ilo = n-1 if size>1 and 1 o/w
!    ilo = max(size(x)-1,1)
        

        do ilo = 1,size(x)-1
            ! If here, then size(x)-1>=1 <=> size(x)>=2 (if size(x)==1, we just return ilo = 1)

            ! Three cases to be concerned with
            ! (1) xhat<x(1)
            ! (2) xhat>=x(n)
            ! (3) xhat>=x(1) .and. xhat<x(n)

            ! Test whether xhat<x(ilo+1). 
            ! In case (1), this will always evaluate to true. So we should exit immediately returning ilo=1
            ! In case (2), this will never evaluate to true. 
            ! In case (3), this will evaluate to true when xhat<x(ilo+1) and there was no j<ilo+1 such that xhat<x(j).
            !   Consequently, we have xhat>=x(j) for all j<ilo+1, and in particular, xhat>=x(ilo). This means,
            !   xhat \in [x(ilo),x(ilo+1)). So, return ilo, just as we did in case (1).
            if (xhat<x(ilo+1)) return
        end do
        ! If still here, we have case (2) or the size of x is 1. So, we return ilo = n-1 if size>1 and 1 o/w
        ilo = max(size(x)-1,1)

    end subroutine

    subroutine bsearch_old(ilo,xhat,x)
        implicit none
        real(rt), dimension(1:), intent(in) :: x
        real(rt), intent(in) :: xhat
        integer, intent(out) :: ilo
        !local
        integer :: ihi,mid

        if (size(x)<=1) then
            ilo = 1
            STOP 'bsearch: size(x) is only 1.  should be at least 2'
        elseif (xhat<=x(1)) then
            ilo = 1
            return
        elseif (xhat>=x(size(x))) then
            ilo = size(x)-1
            return
        else

            ilo = 1
            ihi = size(x)
            mid = (ihi+ilo)/2
            
            do while (mid>ilo)
            
                if (xhat>x(mid)) then
                    ilo = mid
                elseif (xhat<x(mid)) then
                    ihi = mid
                else
                    ilo = mid
                    return
                end if
                mid = (ihi+ilo)/2

            end do

        end if

        ! Check the result as a precaution.  Make sure x in fact lies between.
        ! If this has a severe negative performance impact, I should get rid of this
        if (xhat<x(ilo) .or. xhat>x(ilo+1)) then
            write(*,*) 'bsearch: error, bracketing interval not found.  is x sorted?'
            write(*,*) 'bsearch: xhat x(ilo) x(ilo+1) ilo ihi', xhat, x(ilo),x(ilo+1), ilo, ihi
        end if 

    end subroutine 

    ! Computes the simplical 2-dimensional interpolant 
    !
    ! Input :: X,Y,V,xi,yi,isEqSpaced 
    ! Output :: Vi
    ! 
    ! X and Y are the grids V is defined on. X and Y must both be strictly increasing.
    ! V is a matrix (size(X),size(Y)) s.t. V(i,j) = f(X(i),X(j))
    ! xi and yi are two vectors of the same length giving desired values
    !   (xi(i),yi(i)) where the interpolant should be computed
    ! Vi is the interpolated value, i.e. V(i) = fhat(xi(i),yi(i))
    !
    ! space determines what spacing is assumed for X and Y according to
    !   0: X and Y have no spacing requirements (binary search is used)
    !   1: X and Y are both equally spaced (inversion of linspacing formula used)
    !   2: The positions of xi,yi in X,Y are given explicitly by vectors xiloc,yiloc.
    !      If this option is selected, optional arguments xiloc and yiloc must be passed. 
    !      Further, to work properly, must have xi(i) \in [X(xiloc(i)),X(xiloc(i+1))] and 
    !      similarly for y
    !   3+: TBD
    subroutine sub_s2interp_elt(Vi,X,Y,V,xi,yi,space)! ,doDebug) 
        implicit none
        real(rt), dimension(1:), intent(in) :: X,Y
        real(rt), dimension(1:,1:), intent(in) :: V
        real(rt), intent(in) :: xi,yi
        real(rt), intent(out) :: Vi ! Interpolated V values
        integer, intent(in) :: space
        !logical, intent(in), optional :: doDebug
        ! local
        real(rt) :: Vicopy(1)

        ! It would be more efficient to copy routine in and make adjustments for scalar, 
        ! but I don't pursue this now
        call sub_s2interp_vec(Vicopy,X,Y,V,(/xi/),(/yi/),space)
        Vi = Vicopy(1)

    end subroutine sub_s2interp_elt
    subroutine sub_s2interp_vec(Vi,X,Y,V,xi,yi,space)
        implicit none
        real(rt), dimension(1:), intent(in) :: X,Y
        real(rt), dimension(1:,1:), intent(in) :: V
        real(rt), dimension(1:), intent(in) :: xi,yi
        real(rt), dimension(1:size(xi)), intent(out) :: Vi ! Interpolated V values
        integer, intent(in) :: space
        ! logical, intent(in), optional :: doDebug
        !local
        integer :: nX,nY,i,xiloc,yiloc
        real(rt) :: stepX,stepY,thetaxi,thetayi

        nX = size(X)
        nY = size(Y)

        if (size(V,1)/=nX .or. size(V,2)/=nY) STOP 'sub_s2interp: ERROR: V must have size nX,nY'
        if (size(xi)/=size(yi)) STOP 'sub_s2interp: ERROR: xi and yi must have the same length'

        if (space==1) then
            stepX = (X(nX)-X(1))/real(nX-1,rt)
            stepY = (Y(nY)-Y(1))/real(nY-1,rt)
        end if

        ! Iterate over the interpolation points
        do i = 1,size(xi)

            ! Get the grid locations for each 
            if (space==1) then
                xiloc = eqSpacedLoc(X,nX,stepX,xi(i))
                yiloc = eqSpacedLoc(Y,nY,stepY,yi(i))
            else
                call search(xiloc,xi(i),X)
                call search(yiloc,yi(i),Y)
            end if

            ! Convert to the unit rectangle via a linear change of variables 
            thetaxi = (xi(i)-X(xiloc))/(X(xiloc+1)-X(xiloc))
            thetayi = (yi(i)-Y(yiloc))/(Y(yiloc+1)-Y(yiloc))

            ! thetaxi+thetaya {>=,<=} 1 determines the simplex so we can do the interpolation
            if (thetaxi+thetayi<=1e0_rt) then
                Vi(i) = V(xiloc,yiloc) + &
                        (V(xiloc,yiloc+1)-V(xiloc,yiloc))*thetayi + &
                        (V(xiloc+1,yiloc)-V(xiloc,yiloc))*thetaxi
            else
                Vi(i) = V(xiloc+1,yiloc+1) -  &
                        (V(xiloc+1,yiloc+1)-V(xiloc+1,yiloc))*(1e0_rt-thetayi) - &
                        (V(xiloc+1,yiloc+1)-V(xiloc,yiloc+1))*(1e0_rt-thetaxi)
            end if

        end do

    end subroutine sub_s2interp_vec
    subroutine sub_s2interp_elt_loc(Vi,X,Y,V,xi,yi,space,xiloc,yiloc)
        implicit none
        real(rt), dimension(1:), intent(in) :: X,Y
        real(rt), dimension(1:,1:), intent(in) :: V
        real(rt), intent(in) :: xi,yi
        real(rt), intent(out) :: Vi ! Interpolated V values
        integer, intent(in) :: space
        integer, intent(in) :: xiloc,yiloc
        ! local
        real(rt) :: thetaxi,thetayi

        if (size(V,1)/=size(X) .or. size(V,2)/=size(Y)) STOP 'sub_s2interp: ERROR: V must have size nX,nY'
        if (space/=3) STOP 'sub_s2interp_vec_loc: ERROR: space should be 3 in this scenario'

        ! Convert to the unit rectangle via a linear change of variables 
        thetaxi = (xi-X(xiloc))/(X(xiloc+1)-X(xiloc))
        thetayi = (yi-Y(yiloc))/(Y(yiloc+1)-Y(yiloc))

        ! thetaxi+thetaya {>=,<=} 1 determines the simplex so we can do the interpolation
        if (thetaxi+thetayi<=1e0_rt) then
            Vi = V(xiloc,yiloc) + &
                    (V(xiloc,yiloc+1)-V(xiloc,yiloc))*thetayi + &
                    (V(xiloc+1,yiloc)-V(xiloc,yiloc))*thetaxi
        else
            Vi = V(xiloc+1,yiloc+1) -  &
                    (V(xiloc+1,yiloc+1)-V(xiloc+1,yiloc))*(1e0_rt-thetayi) - &
                    (V(xiloc+1,yiloc+1)-V(xiloc,yiloc+1))*(1e0_rt-thetaxi)
        end if

    end subroutine sub_s2interp_elt_loc
    subroutine sub_s2interp_vec_loc(Vi,X,Y,V,xi,yi,space,xiloc,yiloc)
        implicit none
        real(rt), dimension(1:), intent(in) :: X,Y
        real(rt), dimension(1:,1:), intent(in) :: V
        real(rt), dimension(1:), intent(in) :: xi,yi
        integer, dimension(1:), intent(in) :: xiloc,yiloc ! Positions of xi and yi in grid
        real(rt), dimension(1:size(xi)), intent(out) :: Vi ! Interpolated V values
        integer, intent(in) :: space
        !local
        integer :: i
        real(rt) :: thetaxi,thetayi

        if (size(V,1)/=size(X) .or. size(V,2)/=size(Y)) &
            STOP 'sub_s2interp: ERROR: V must have size nX,nY'
        if (size(xi)/=size(yi)) &
            STOP 'sub_s2interp: ERROR: xi and yi must have the same length'
        if (size(xiloc)/=size(yiloc) .or. size(xiloc)/=size(xi)) &
            STOP 'sub_s2interp: ERROR: xi,yi,xiloc,yiloc must have the same length'
        if (space/=3) STOP 'sub_s2interp_vec_loc: ERROR: space should be 3 in this scenario'

        ! Iterate over the interpolation points
        do i = 1,size(xi)

            ! Convert to the unit rectangle via a linear change of variables 
            thetaxi = (xi(i)-X(xiloc(i)))/(X(xiloc(i)+1)-X(xiloc(i)))
            thetayi = (yi(i)-Y(yiloc(i)))/(Y(yiloc(i)+1)-Y(yiloc(i)))

            ! thetaxi+thetaya {>=,<=} 1 determines the simplex so we can do the interpolation
            if (thetaxi+thetayi<=1e0_rt) then
                Vi(i) = V(xiloc(i),yiloc(i)) + &
                        (V(xiloc(i),yiloc(i)+1)-V(xiloc(i),yiloc(i)))*thetayi + &
                        (V(xiloc(i)+1,yiloc(i))-V(xiloc(i),yiloc(i)))*thetaxi
            else
                Vi(i) = V(xiloc(i)+1,yiloc(i)+1) -  &
                        (V(xiloc(i)+1,yiloc(i)+1)-V(xiloc(i)+1,yiloc(i)))*(1e0_rt-thetayi) - &
                        (V(xiloc(i)+1,yiloc(i)+1)-V(xiloc(i),yiloc(i)+1))*(1e0_rt-thetaxi)
            end if

        end do

    end subroutine sub_s2interp_vec_loc






    ! This routine is the ultimate in terms of performance because just evaluates the formula
    ! without computing any weights. This gets rid of 2 divisions and 4 subtractions since the
    ! thetaxi and thetayi are passed as arguments. The thetas are of the form
    ! thetaxi = (xi-X(xiloc))/(X(xiloc+1)-X(xiloc)) ...

    subroutine sub_s2interp_elt_loc_getTheta(thetaxi,thetayi,X,Y,xi,yi,xiloc,yiloc)
        implicit none
!         !DEC$ ATTRIBUTES INLINE :: sub_s2interp_elt_loc_getTheta 
        real(rt), dimension(1:), intent(in) :: X,Y
        real(rt), intent(in) :: xi,yi
        integer, intent(in) :: xiloc,yiloc
        real(rt), intent(out) :: thetaxi,thetayi

        ! Computes the theta
        thetaxi = (xi-X(xiloc))/(X(xiloc+1)-X(xiloc))
        thetayi = (yi-Y(yiloc))/(Y(yiloc+1)-Y(yiloc))

    end subroutine sub_s2interp_elt_loc_getTheta 
    subroutine sub_s2interp_elt_loc_mass(Vi,X,Y,V,xi,yi,space,xiloc,yiloc,thetaxi,thetayi)
        implicit none
!         !DEC$ ATTRIBUTES INLINE :: sub_s2interp_elt_loc_mass
        real(rt), dimension(1:), intent(in) :: X,Y
        real(rt), dimension(1:,1:), intent(in) :: V
        real(rt), intent(in) :: xi,yi
        real(rt), intent(out) :: Vi ! Interpolated V values
        integer, intent(in) :: space
        integer, intent(in) :: xiloc,yiloc
        real(rt), intent(in) :: thetaxi,thetayi

        ! thetaxi+thetaya {>=,<=} 1 determines the simplex so we can do the interpolation
        if (thetaxi+thetayi<=1e0_rt) then
            Vi = V(xiloc,yiloc) + &
                    (V(xiloc,yiloc+1)-V(xiloc,yiloc))*thetayi + &
                    (V(xiloc+1,yiloc)-V(xiloc,yiloc))*thetaxi
        else
            Vi = V(xiloc+1,yiloc+1) -  &
                    (V(xiloc+1,yiloc+1)-V(xiloc+1,yiloc))*(1e0_rt-thetayi) - &
                    (V(xiloc+1,yiloc+1)-V(xiloc,yiloc+1))*(1e0_rt-thetaxi)
        end if

    end subroutine sub_s2interp_elt_loc_mass

!     !DEC$ ATTRIBUTES INLINE :: eqSpacedLoc
    function eqSpacedLoc(X,nX,stepX,xval) result(xiloc)
        implicit none
        real(rt), dimension(1:), intent(in) :: X
        real(rt), intent(in) :: xval,stepX
        integer, intent(in) :: nX 
        integer :: xiloc

        ! 6 operations
        xiloc = max(min(floor(1.e0_rt+(xval-X(1))/stepX),nX-1),1)    

    end function eqSpacedLoc


    subroutine test_s2interp()
        implicit none
        real(rt), allocatable :: X(:),Y(:),V(:,:),Xi(:),Yi(:),Vi(:)
        real(rt) :: h
        integer :: i,j

        allocate(X(20),Y(30))
        allocate(V(size(X),size(Y)))
        allocate(Xi(size(X)*size(Y)),Yi(size(X)*size(Y)),Vi(size(X)*size(Y)))

        X(1) = -1e0_rt
        X(size(X)) = 15e0_rt
        h = (X(size(X))-X(1))/real(size(X)-1,rt)
        do i = 2,size(X)-1
            X(i) = h + X(i-1)
        end do

        Y(1) = 2e0_rt
        Y(size(Y)) = 3e0_rt
        h = (Y(size(Y))-Y(1))/real(size(Y)-1,rt)
        do i = 2,size(Y)-1
            Y(i) = h + Y(i-1)
        end do
        
        do j = 1,size(Y)
            do i = 1,size(X)
                V(i,j) = f(x(i),y(j))
            end do
        end do

        ! First check that it actually interpolates
        do i = 1,size(Y)
            Xi(1:size(X)) = X
            Yi(1:size(X)) = Y(i)
            call sub_s2interp(Vi(1:size(X)),X,Y,V,Xi(1:size(X)),Yi(1:size(X)),1) 
            if (maxval(abs(Vi(1:size(X))-f(Xi(1:size(X)),Yi(1:size(X)))))>sqrt(epsilon(0._rt))) &
                STOP 'sub_s2interp failed interp check'
        end do

        ! Now check error at random points in the state space
        call random_number(Xi)
        call random_number(Yi)
        Xi = X(1) + Xi*(X(size(X))-X(1))
        Yi = Y(1) + Yi*(Y(size(Y))-Y(1))

        call sub_s2interp(Vi,X,Y,V,Xi,Yi,1) 

!         print*,'Min error is ',minval(abs(Vi-f(Xi,Yi)))
!         print*,'Avg error is ',sum(abs(Vi-f(Xi,Yi)))/real(size(Vi),8)
!         print*,'Max error is ',maxval(abs(Vi-f(Xi,Yi)))

        if (maxval(abs(Vi-f(Xi,Yi)))>sqrt(epsilon(0._rt))) then
            STOP 'sub_s2interp failed random check'
        end if


        ! do i = 1,size(Xi)
        !     write(*,'(A,4f6.2)')'x,y,Vi,f',Xi(i),Yi(i),Vi(i),f(Xi(i),Yi(i))
        ! end do


!         call usedir('~/Desktop/data/')
! 
!         call myprint('X',X)
!         call myprint('Y',Y)
!         call myprint('V',V)
!         call myprint('Xi',Xi)
!         call myprint('Yi',Yi)
!         call myprint('Vi',Vi)

        print*,'test_s2interp: passed'


    end subroutine test_s2interp

    elemental function f(x,y)
        real(rt), intent(in) :: x,y
        real(rt) :: f
        ! f = x*y + 3e0_rt*x - 2e0_rt*y + 5e0_rt
        f = 3e0_rt*x - 2e0_rt*y + 5e0_rt
        ! f = x + y
        ! f = sqrt(x**2 + y**2) 
        ! f = log(.01+abs(x))*y + y**2 - .01e0_rt*x**3 
        
    end function

    subroutine test_search
        use mod_core
        implicit none
        real(rt) :: X(1000)
        integer :: i
        integer, parameter :: ntrials = 10000
        real(rt), allocatable :: xhat(:)
        integer, allocatable :: j(:)

        allocate(xhat(ntrials),j(ntrials))
        call random_number(xhat)
        
        ! Linearly spaced data
        X = linspace(.02e0_rt,.98e0_rt,size(X)) 

        ! Random data
        do i = 1,size(X)
            call random_number(X(i))
            if (i>1) X(i) = X(i) + X(i-1)
        end do
        X = (X-X(1))/(X(size(X))-X(1)) ! [0,1]
        X = .02e0_rt + .96e0_rt*X ! [0.02,.98]


        ! Test interpolation search
        do i = 1,ntrials
            call isearch(j(i),xhat(i),X)
        end do
        call check('isearch')
        print*,'test_search: isearch passed'

        do i = 1,ntrials
            call bsearch(j(i),xhat(i),X)
        end do
        call check('bsearch')
        print*,'test_search: bsearch passed'

        do i = 1,ntrials
            call lsearch(j(i),xhat(i),X)
        end do
        call check('lsearch')
        print*,'test_search: lsearch passed'

        do i = 1,ntrials
            call search(j(i),xhat(i),X)
        end do
        call check('search')
        print*,'test_search: lsearch passed'

    contains

        subroutine check(msg)
            implicit none
            character(len=*), intent(in) :: msg
            
            do i = 1,ntrials
                if (xhat(i)<X(1)) then 
                    if (j(i)/=1) then
                        print*,msg,': j(i) should equal one but is ',j(i),' xhat(i) ',xhat(i),' X(1) ',X(1)
                        STOP 'test_search: error 1'
                    end if
                elseif (xhat(i)>X(size(X))) Then
                    if (j(i)/=size(X)-1) then
                        print*,msg
                        STOP 'test_search: error 2'
                    end if
                else
                    if (xhat(i)<X(j(i)) .or. xhat(i)>=X(j(i)+1)) then
                        print*,msg
                        STOP 'test_search: error 3'
                    end if
                end if
            end do
        end subroutine

    end subroutine

end module mod_interp
! A module for plotting from Fortran using gnuplot

! subroutine clear()
! subroutine scratchdir(s)
! subroutine checkinit()
! subroutine legend(s1,s2,s3,s4,s5,s6,s7,s8,s9,s10,s11,s12)
! subroutine axis(x)
! subroutine xlabel(s)
! subroutine ylabel(s)
! subroutine title(s)

! subroutine triplet(x,y,s)
! * Inputs
! --- x,y are the data
! --- s is a string giving the style. One of "lines", "linespoints", "points", "dots"

! subroutine saveplot(fname)
! subroutine drawnow()
! subroutine test_plot()
module mod_plot
    use mod_types
    implicit none

    logical :: disable_plotting = .false.

    character(len=1000) :: dir = '/tmp/'

    ! Defaults
    character(len=7) :: default_colors(1000) 
    integer :: n_default_colors = 1

    ! Triplet data type
    type xys
        real(rt), allocatable :: x(:), y(:)
        integer :: nx,ncol
        character(len=4) :: s ! style
    end type

    ! Plot type
    type type_plotopts
        type(xys) :: p(10) ! stores the plot data
        integer :: npairs = 0
        logical :: haslegend = .false.
        character(len=1000), dimension(12) :: legendlist ! hardcoded max legend limit is 10
        integer :: nlegendlist = 0 
        character(len=1000) :: legendlocation = ''
        integer :: persist 
        real(rt) :: xub,xlb,yub,ylb
        logical :: hasxub=.false.,hasxlb=.false.,hasyub=.false.,hasylb=.false.
        logical :: dosaveplot = .false.
        character(len=1000) :: savefname= ''
    end type

    type(type_plotopts), dimension(100), target :: plot_list ! list of all plots
    type(type_plotopts), pointer :: cf => null() ! current figure
    
    logical :: isinit = .false.

    interface triplet
        module procedure triplet_vec, triplet_mat
    end interface

    private 
    public :: xlabel,ylabel,title,legend,triplet,drawnow,axis,scratchdir,clear,test_plot,initplot,add_gnuplot_cmd,disable_plotting,gnuplot

    interface title_renamed
        module procedure title
    end interface 

    interface gnuplot
        module procedure gnuplot_mat, gnuplot_vec
    end interface

contains

    subroutine gnuplot_vec(x,y,title,saveas)
        implicit none
        real(8), intent(in) :: x(:)
        real(8), intent(in) :: y(:)
        character(*), intent(in), optional :: title,saveas

        write(0,*) 'WARNING: passing optional arguments, this may segfault'
        call gnuplot_mat(x,reshape(y,[size(y),1]),title,saveas)

    end subroutine 

    subroutine gnuplot_mat(x,y,title,saveas)
        implicit none
        real(8), intent(in) :: x(:)
        real(8), intent(in) :: y(:,:)
        character(*), intent(in), optional :: title,saveas
        ! local
        integer :: uid

        call clear()

        if (present(saveas)) then
            uid = initplot(saveas)
        else
            uid = initplot('/tmp/gnuplot_core.eps')
        end if

        if (present(title)) then
            call title_renamed(uid,title)
        end if

        call triplet(x,y,' ')

        call drawnow(uid)
        

    end subroutine 

    ! Creates "handle" (really just a unit number) for a particular plot
    ! * Inputs
    ! --- fname : file name for eps file
    ! 
    function initplot(fname) result(uid)
        implicit none
        character(len=*), intent(in) :: fname
        integer :: uid
        ! local
        integer :: uidtmp

        call checkinit()

        open(newunit=uidtmp,file=trim(dir)//"mod_plot_tmp.cfg",status='replace',err=999)
        uid = uidtmp

        ! Write the default style file 

        write(uid,'(3A)') 'set output "',trim(fname),'"'
        write(uid,'(A)') 'set terminal postscript eps colour enhanced "Helvetica" 14' 
        write(uid,'(A)') 'set linetype 1 lc rgb "#4169E1" linewidth 4 ' ! royal blue
        write(uid,'(A)') 'set linetype 2 lc rgb "#FF7F50" linewidth 4 dashtype "-." # coral' ! coral
        write(uid,'(A)') 'set linetype 3 lc rgb "#3CB371" linewidth 4 dashtype "._. ._."' ! medium sea-green
        write(uid,'(A)') 'set linetype 4 lc rgb "#DC143C" linewidth 4 ' ! crimson
        write(uid,'(A)') 'set linetype cycle 4'
        write(uid,'(A)') 'set border linewidth 1'
        write(uid,'(A)') 'set key off' ! by default, no legend, but if legend is called, turned on
        write(uid,'(A)') 'set style data linespoints' ! set line as default style 

        return

999     write(0,*) 'WARNING: could not open ',trim(dir)//'mod_plot_tmp.cfg, plot will not be created'
        uid = 0

    end function 

    subroutine add_gnuplot_cmd(uid,s)
        implicit none
        integer, intent(in) :: uid
        character(len=*), intent(in) :: s
        write(uid,'(A)') s
    end subroutine

    ! legend
    ! Creates a legend with names and location specified 
    !
    ! * Inputs
    ! --- names : comma separated list of data names
    ! --- <location> : gnuplot style location, e.g. "inside bottom right"
    subroutine legend(uid,names,location)
        use mod_util
        implicit none
        integer, intent(in) :: uid
        character(len=*), intent(in) :: names
        character(len=*), intent(in), optional :: location
        ! local
        integer :: n, i

        if (present(location)) then
            write(uid,'(A)') 'set key '//location
        else
            write(uid,'(A)') 'set key on'
        end if

        cf%haslegend = .true.

        cf%nlegendlist = count_delim_arg(names,',')

        do i = 1,cf%nlegendlist 
            cf%legendlist(i) = get_delim_arg(names,',',i)
        end do

    end subroutine

    subroutine xlabel(uid,s)
        implicit none
        integer, intent(in) :: uid
        character(len=*), intent(in) :: s
        write(uid,'(A)') 'set xlabel "'//s//'"'
    end subroutine
    subroutine ylabel(uid,s)
        implicit none
        integer, intent(in) :: uid
        character(len=*), intent(in) :: s
        write(uid,'(A)') 'set ylabel "'//s//'"'
    end subroutine
    subroutine title(uid,s)
        implicit none
        integer, intent(in) :: uid
        character(len=*), intent(in) :: s
        write(uid,'(A)') 'set title "'//s//'"'
    end subroutine
    
    ! frees the specified figure our, by default, current figure
    subroutine clear()
        implicit none
        ! local
        integer :: i
        type(type_plotopts) :: defaulttype

        call checkinit()
        
        ! I don't think it is necessarily, but to be careful, manually free x and y data 
        do i = 1,cf%npairs
            if (allocated(cf%p(i)%x)) deallocate(cf%p(i)%x)
            if (allocated(cf%p(i)%y)) deallocate(cf%p(i)%y)
!             cf%p(i)%nx = -1
!             cf%p(i)%ncol = -1
!             cf%p(i)%s = ''

        end do
!         cf%npairs = 0

        ! Use this to reset to whatever the default specification is
        cf = defaulttype

    end subroutine

    subroutine write_default_style(uid)
        implicit none
        integer, intent(in) :: uid

        write(uid,'(A)') '# Temporary gnuplot configuration file created by grey''s plot routine'
        write(uid,'(A)') 'set terminal postscript eps colour enhanced "Helvetica" 16' 
        write(uid,'(A)') 'set linetype 1 lc rgb "#4169E1" linewidth 4 ' ! royal blue
        write(uid,'(A)') 'set linetype 2 lc rgb "#FF7F50" linewidth 4 dashtype "-." # coral' ! coral
        write(uid,'(A)') 'set linetype 3 lc rgb "#DC143C" linewidth 4 ' ! crimson
        write(uid,'(A)') 'set linetype 4 lc rgb "#3CB371" linewidth 4 dashtype "._. ._."' ! medium sea-green
        write(uid,'(A)') 'set linetype cycle 4'
        write(uid,'(A)') 'set border linewidth 1'

    end subroutine

    
    subroutine scratchdir(s)
        implicit none
        character(len=*), intent(in) :: s
        dir = s
    end subroutine 

    subroutine checkinit()
        implicit none
        if (.not. isinit) then
            cf => plot_list(1)
            isinit = .true.
        end if
    end subroutine

!     subroutine legend(s1,s2,s3,s4,s5,s6,s7,s8,s9,s10,s11,s12)
!         implicit none
!         character(len=*), intent(in), optional :: s1,s2,s3,s4,s5,s6,s7,s8,s9,s10,s11,s12
!         ! local
!         integer :: narg
! 
!         cf%haslegend = .true.
! 
!         if (present(s1))  then; cf%legendlist(1) = s1; narg = 1; end if
!         if (present(s2))  then; cf%legendlist(2) = s2; narg = 2; end if
!         if (present(s3))  then; cf%legendlist(3) = s3; narg = 3; end if
!         if (present(s4))  then; cf%legendlist(4) = s4; narg = 4; end if
!         if (present(s5))  then; cf%legendlist(5) = s5; narg = 5; end if
!         if (present(s6))  then; cf%legendlist(6) = s6; narg = 6; end if
!         if (present(s7))  then; cf%legendlist(7) = s7; narg = 7; end if
!         if (present(s8))  then; cf%legendlist(8) = s8; narg = 8; end if
!         if (present(s9))  then; cf%legendlist(9) = s9; narg = 9; end if
!         if (present(s10)) then; cf%legendlist(10) = s10; narg = 10; end if
!         if (present(s11)) then; cf%legendlist(11) = s11; narg = 11; end if
!         if (present(s12)) then; cf%legendlist(12) = s12; narg = 12; end if
! 
!         cf%nlegendlist = narg
! 
!         if (trim(cf%legendlist(narg-1))=='location') then
!             cf%legendlocation = trim(cf%legendlist(narg))
!             cf%nlegendlist = narg - 2
!         end if
!         
!     end subroutine

    subroutine axis(x)
        implicit none
        real(rt), intent(in) :: x(4)
        call checkinit()
        
        if (x(1)/=-huge(0e0_rt)) then; cf%xlb = x(1); cf%hasxlb = .true.; end if
        if (x(2)/=huge(0e0_rt)) then; cf%xub = x(2); cf%hasxub = .true.; end if
        if (x(3)/=-huge(0e0_rt)) then; cf%ylb = x(3); cf%hasylb = .true.; end if
        if (x(4)/=huge(0e0_rt)) then; cf%yub = x(4); cf%hasyub = .true.; end if

    end subroutine
    subroutine triplet_vec(x,y,s)
        implicit none
        real(rt), intent(in) :: x(:),y(:)
        character(len=*), intent(in) :: s
        call triplet_core(x,y,s,1)
    end subroutine
    subroutine triplet_mat(x,y,s)
        implicit none
        real(rt), intent(in) :: x(:),y(:,:)
        character(len=*), intent(in) :: s
        call triplet_core(x,y,s,size(y,2))
    end subroutine
    subroutine triplet_core(x,y,s,ncol)
        implicit none
        real(rt), intent(in) :: x(:),y(*)
        integer, intent(in) :: ncol
        character(len=*), intent(in) :: s

        call checkinit()

        cf%npairs = cf%npairs + 1

        associate (p => cf%p(cf%npairs)) 

            p%s = s

            p%nx = size(x)
            allocate(p%x(p%nx))
            p%x = x

            p%ncol = ncol

            allocate(p%y(p%nx*p%ncol))
            p%y(1:p%nx*p%ncol) = y(1:p%nx*p%ncol)

        end associate

    end subroutine

!     subroutine saveplot(fname)
!         implicit none
!         character(len=*), intent(in) :: fname
! 
!         cf%dosaveplot = .true.
!         cf%savefname = fname
!         call drawnow()
!         cf%dosaveplot = .false.
!         cf%savefname = ''
! 
!     end subroutine 

    subroutine drawnow(uid)
        implicit none
        integer, intent(in) :: uid
        integer :: i, unitno !,j,ni,nj
        character(len=3), parameter :: newline = ', \'
        character(len=3) :: tmpchar
        character(len=100) :: configfile, datafile 
        integer :: ierr,ierr2
        integer :: irow,nrow,icol,xaxiscol,ilegend
        character(len=*), parameter :: doublefmt = '1PE18.9E3'
        character(len=100) :: persistcmd
        ! FOR PGI, need below line
!         include 'lib3f.h' 

        if (disable_plotting) then
            close(uid)
            return
        end if

!         !$OMP SINGLE

        ! As an extra precaution, I try to make sure the threads write to unique files
        datafile = trim(dir)//'mod_plot_tmp'//'.dat'
        configfile = trim(dir)//'mod_plot_tmp'//'.cfg'

        ! Write raw data to a temporary file

        ! To handle uneven data like triplet(#2,#2) triplet(#3,#3), need to figure out the max length, (#3 in this example) and write all the rows at one time
        ! Data:
        ! 0 1.2 0 1.3
        ! 1 2   .5 1.6
        !       2 2
        ! Then, when displaying, select 1st col and 2nd col as first set of data and 3rd col and 4th col as second set of data
        nrow = -1
        do i = 1,cf%npairs
            nrow = max(nrow,cf%p(i)%nx)
        end do

        open(newunit=unitno,file=trim(datafile),status='unknown',err=999)
        do irow = 1,nrow
            do i = 1,cf%npairs
                if (irow<=cf%p(i)%nx) then
                    
                    write(unitno,'('//doublefmt//',A)',advance='no') cf%p(i)%x(irow),' '
                    do icol = 1,cf%p(i)%ncol
                        write(unitno,'('//doublefmt//',A)',advance='no') cf%p(i)%y(irow + (icol-1)*cf%p(i)%nx),' '
                    end do
                else 
                    ! write(unitno,'(A)',advance='no') repeat('? ',1+cf%p(i)%ncol)
                    write(unitno,'('//doublefmt//',A)',advance='no') cf%p(i)%x(cf%p(i)%nx),' '
                    do icol = 1,cf%p(i)%ncol
                        write(unitno,'(A19)',advance='no') '? '
                    end do
                end if
            end do
            write(unitno,*)
        end do
        close(unitno)

        ! Write the configuration file
        


        ! Write config to a temporary file
        unitno = uid
        ! open(newunit=unitno,file=trim(configfile),status='unknown')

        ! write(unitno,*) 'set term x11'
        ! write(unitno,*) 'set term epslatex'


        ! if (cf%xlabel/='') write(unitno,*) 'set xlabel "'//trim(cf%xlabel)//'"'
        ! if (cf%ylabel/='') write(unitno,*) 'set ylabel "'//trim(cf%ylabel)//'"'
        ! if (cf%title/='') write(unitno,*) 'set title "'//trim(cf%title)//'"'


        write(unitno,*) 'set datafile missing "?" ' ! set line as default style 

        ! Legend entries
        if (.not. cf%haslegend) write(unitno,*) 'set key off'
        if (cf%haslegend) then
                if (trim(cf%legendlocation)=='northwest') then; write(unitno,*) 'set key on top left'; 
            elseif (trim(cf%legendlocation)=='northeast') then; write(unitno,*) 'set key on top right'; 
            else
                ! do nothing
            end if

        end if

        ! Autoscaling is a bit tricky, so determine appropriate rnages

        ! Set xrange
        if (cf%hasylb .or. cf%hasyub .or. cf%hasxub .or. cf%hasxlb) then
            write(unitno,'(A)',advance='no') 'set xr ['
            if (cf%hasxlb) then
                write(unitno,'('//doublefmt//')',advance='no') cf%xlb
            else
                cf%xlb = huge(0e0_rt)
                do i = 1,cf%npairs 
                    cf%xlb = min(cf%xlb,minval(cf%p(i)%x))
                end do
                write(unitno,'('//doublefmt//')',advance='no') cf%xlb
            end if
            write(unitno,'(A)',advance='no') ':'
            if (cf%hasxub) then
                write(unitno,'('//doublefmt//')',advance='no') cf%xub
            else
                cf%xub = -huge(0e0_rt)
                do i = 1,cf%npairs 
                    cf%xub = max(cf%xub,maxval(cf%p(i)%x))
                end do
                write(unitno,'('//doublefmt//')',advance='no') cf%xub
            end if
            write(unitno,'(A)',advance='yes') ']'
        end if
        ! Set yrange
        if (cf%hasylb .or. cf%hasyub .or. cf%hasxub .or. cf%hasxlb) then
            write(unitno,'(A)',advance='no') 'set yr ['
            if (cf%hasylb) then
                write(unitno,'('//doublefmt//')',advance='no') cf%ylb
            else
                cf%ylb = huge(0e0_rt)
                do i = 1,cf%npairs 
                    cf%ylb = min(cf%ylb,minval(cf%p(i)%x))
                end do
                write(unitno,'('//doublefmt//')',advance='no') cf%ylb
            end if
            write(unitno,'(A)',advance='no') ':'
            if (cf%hasyub) then
                write(unitno,'('//doublefmt//')',advance='no') cf%yub
            else
                cf%yub = -huge(0e0_rt)
                do i = 1,cf%npairs 
                    cf%yub = max(cf%yub,maxval(cf%p(i)%y))
                end do
                write(unitno,'('//doublefmt//')',advance='no') cf%yub
            end if
            write(unitno,'(A)',advance='yes') ']'
        end if


    !     ! Check inputs
    !     ni = size(x)
    !     nj = size(y,2)
    !     if (size(y,1)/=ni) STOP 'sub_plot_x_y2: x and y dims don''t agree'
    ! 
    !     ! Set defaults and replace them if optional arg present
    !     do j = 1,nJ
    !         write(jAsString,'(i0)') j
    !         dls(j) = 'data '//jAsString
    !     end do
    !     ms = 'line'
    !     cls = ''
    ! 
    !     ! Write data to a temporary file
    !     open(unitno,file=trim(datafile),status='unknown')
    !     write(unitno,*) '# Temporary data file created by grey''s plot routine' 
    !     do i = 1,ni
    !         write(unitno,'(1PE24.15E3)',advance='no') x(i)
    !         do j = 1,nj
    !             write(unitno,'(1PE24.15E3)',advance='no') y(i,j)
    !         end do
    !         write(unitno,*) 
    !     end do
    !     close(unitno)

        write(unitno,'(A)',advance='no') 'plot' ! this is the first part of the command

        xaxiscol = 1
        ilegend = 1
! print*,'cf%npairs',cf%npairs
        do i = 1,cf%npairs

! print*,'cf%p(i)%ncol',cf%p(i)%ncol
            do icol = 1,cf%p(i)%ncol
                
                ! Determine whether new line is needed
                if (i<cf%npairs .or. icol<cf%p(i)%ncol) then
                    tmpchar = newline
                else
                    tmpchar = '   '
                end if

                ! Output the command. NOTE: must use advance no here to prevent ifort from breaking the line
                if (cf%haslegend .and. ilegend<=cf%nlegendlist) then
                    write(unitno,'(A)',advance='no') ' "'//trim(datafile)//'" using '//trim(i_to_s(xaxiscol))//':' &
                            //trim(i_to_s(xaxiscol+icol))//' title "'//trim(cf%legendlist(ilegend))&
                            //'"'//trim(tmpchar) 
                elseif (cf%haslegend .and. ilegend>cf%nlegendlist) then
                    write(unitno,'(A)',advance='no') ' "'//trim(datafile)//'" using '//trim(i_to_s(xaxiscol))//':' &
                            //trim(i_to_s(xaxiscol+icol))//' title "'//' '//'" '//trim(tmpchar) ! Title in latex is more like legend in matlab
                else 
                    write(unitno,'(A)',advance='no') ' "'//trim(datafile)//'" using '//trim(i_to_s(xaxiscol))//':' &
                            //trim(i_to_s(xaxiscol+icol))//trim(tmpchar) ! Title in latex is more like legend in matlab
                end if
                ! Make a new line
                write(unitno,*)

                ilegend = ilegend + 1
            end do

            xaxiscol = xaxiscol + 1 + cf%p(i)%ncol

        end do

!         if (cf%dosaveplot) then 
! !             write(unitno,*) 'set terminal postscript
! !             write(unitno,*) 'set terminal postscript eps size 3.5,2.62 enhanced color font ''Helvetica,20'' linewidth 2 '
! !             write(unitno,*) 'set terminal postscript eps enhanced color font ''Helvetica,20'' linewidth 2 '
!             write(unitno,*) 'set terminal postscript eps enhanced color font ''Helvetica,14'' linewidth 2 '
!             write(unitno,*) 'set output "'//trim(cf%savefname)//'.eps"'
!             write(unitno,*) 'replot'
! 
!             ! write(unitno,*) 'set terminal epslatex size 3.5,2.62 color colortext'
!             ! write(unitno,*) 'set output "'//trim(cf%savefname)//'.tex"'
!             ! write(unitno,*) 'replot'
!         end if

        close(unitno)

!         if (cf%dosaveplot) then
!             persistcmd = ''
!         else
!             persistcmd = '-persist'
!         end if

        ! Now use a nonstandard fortran routine (appearing in ifort and gfortran) to call gnuplot
        ! NOTE: if gnuplot is not available or an error is received, just report it and the program
        ! will continue.
        ! ierr = system('gnuplot -persist '//trim(configfile)) ! for PGI
        call execute_command_line('gnuplot -persist '//trim(configfile)//' 2> /dev/null',exitstat=ierr,cmdstat=ierr2)
!         call execute_command_line('gnuplot '//trim(persistcmd)//' '//trim(configfile)),exitstat=ierr)
!         if (ierr/=0) print*,'sub_plot_x_y2: WARNING: error in calling gnuplot'

!         !$OMP END SINGLE
        
        return

999     write(0,*) 'WARNING: could not open ',trim(datafile),', plot will not be created'

    end subroutine drawnow

    ! Converts an integer into a string
    function i_to_s(i) result(s)
        implicit none
        integer, intent(in) :: i 
        character(len=100) :: s
        ! local
        character(len=100) :: slocal

        write(slocal,'(i100)') i
        s = adjustl(slocal)

    end function 

    subroutine test_plot()
        implicit none
        integer :: i

        i = initplot('/tmp/deleteme.eps') 
        call title(i,'test plot')
        call xlabel(i,'x axis')
        call ylabel(i,'y axis')
        call legend(i,'first triplet,second triplet,third triplet','inside bottom right')

        call triplet([0e0_rt,1e0_rt],reshape([.1e0_rt,2e0_rt,.1e0_rt,1e0_rt],[2,2]),'r-')
        call triplet([0e0_rt,.5e0_rt,1e0_rt],[.1e0_rt,2.4e0_rt,2e0_rt],'b-.')
    !     call axis([.3e0_rt,1.1e0_rt,-huge(0e0_rt),huge(0e0_rt)])

        call drawnow(i)

        print*,'Saved plot to /tmp/deleteme.eps'
!         call saveplot('/tmp/savedplot')

    end subroutine

end module

! 
! program testplot
!     use mod_plot
!     implicit none
! 
! 
! end program
! 
!     ! Returns rgb value 
!     ! Hex Code #RRGGBB
!     ! Values taken from http://www.rapidtables.com/web/color/RGB_Color.htm
!     function rgb(color) 
!         implicit none
!         character(7) :: rgb
!         character(30), intent(in) :: color
!         character(30,2), dimension(139) :: colors
!         integer :: i
! 
!         i = 1
!         colors(i,1) = "maroon      ";                       colors(i,2) = "#800000"; i = i + 1
!         colors(i,1) = "dark red        ";                   colors(i,2) = "#8B0000"; i = i + 1
!         colors(i,1) = "brown       ";                       colors(i,2) = "#A52A2A"; i = i + 1
!         colors(i,1) = "firebrick       ";                   colors(i,2) = "#B22222"; i = i + 1
!         colors(i,1) = "crimson     ";                       colors(i,2) = "#DC143C"; i = i + 1
!         colors(i,1) = "red     ";                           colors(i,2) = "#FF0000"; i = i + 1
!         colors(i,1) = "tomato      ";                       colors(i,2) = "#FF6347"; i = i + 1
!         colors(i,1) = "coral   ";                           colors(i,2) = "#FF7F50"; i = i + 1
!         colors(i,1) = "indian red  ";                       colors(i,2) = "#CD5C5C"; i = i + 1
!         colors(i,1) = "light coral     ";                   colors(i,2) = "#F08080"; i = i + 1
!         colors(i,1) = "dark salmon         ";               colors(i,2) = "#E9967A"; i = i + 1
!         colors(i,1) = "salmon  ";                           colors(i,2) = "#FA8072"; i = i + 1
!         colors(i,1) = "light salmon        ";               colors(i,2) = "#FFA07A"; i = i + 1
!         colors(i,1) = "orange red      ";                   colors(i,2) = "#FF4500"; i = i + 1
!         colors(i,1) = "dark orange         ";               colors(i,2) = "#FF8C00"; i = i + 1
!         colors(i,1) = "orange  ";                           colors(i,2) = "#FFA500"; i = i + 1
!         colors(i,1) = "gold        ";                       colors(i,2) = "#FFD700"; i = i + 1
!         colors(i,1) = "dark golden rod         ";           colors(i,2) = "#B8860B"; i = i + 1
!         colors(i,1) = "golden rod  ";                       colors(i,2) = "#DAA520"; i = i + 1
!         colors(i,1) = "pale golden rod         ";           colors(i,2) = "#EEE8AA"; i = i + 1
!         colors(i,1) = "dark khaki  ";                       colors(i,2) = "#BDB76B"; i = i + 1
!         colors(i,1) = "khaki   ";                           colors(i,2) = "#F0E68C"; i = i + 1
!         colors(i,1) = "olive       ";                       colors(i,2) = "#808000"; i = i + 1
!         colors(i,1) = "yellow  ";                           colors(i,2) = "#FFFF00"; i = i + 1
!         colors(i,1) = "yellow green        ";               colors(i,2) = "#9ACD32"; i = i + 1
!         colors(i,1) = "dark olive green        ";           colors(i,2) = "#556B2F"; i = i + 1
!         colors(i,1) = "olive drab  ";                       colors(i,2) = "#6B8E23"; i = i + 1
!         colors(i,1) = "lawn green      ";                   colors(i,2) = "#7CFC00"; i = i + 1
!         colors(i,1) = "chart reuse         ";               colors(i,2) = "#7FFF00"; i = i + 1
!         colors(i,1) = "green yellow    ";                   colors(i,2) = "#ADFF2F"; i = i + 1
!         colors(i,1) = "dark green  ";                       colors(i,2) = "#006400"; i = i + 1
!         colors(i,1) = "green   ";                           colors(i,2) = "#008000"; i = i + 1
!         colors(i,1) = "forest green        ";               colors(i,2) = "#228B22"; i = i + 1
!         colors(i,1) = "lime    ";                           colors(i,2) = "#00FF00"; i = i + 1
!         colors(i,1) = "lime green  ";                       colors(i,2) = "#32CD32"; i = i + 1
!         colors(i,1) = "light green     ";                   colors(i,2) = "#90EE90"; i = i + 1
!         colors(i,1) = "pale green  ";                       colors(i,2) = "#98FB98"; i = i + 1
!         colors(i,1) = "dark sea green  ";                   colors(i,2) = "#8FBC8F"; i = i + 1
!         colors(i,1) = "medium spring green         ";       colors(i,2) = "#00FA9A"; i = i + 1
!         colors(i,1) = "spring green    ";                   colors(i,2) = "#00FF7F"; i = i + 1
!         colors(i,1) = "sea green   ";                       colors(i,2) = "#2E8B57"; i = i + 1
!         colors(i,1) = "medium aqua marine      ";           colors(i,2) = "#66CDAA"; i = i + 1
!         colors(i,1) = "medium sea green    ";               colors(i,2) = "#3CB371"; i = i + 1
!         colors(i,1) = "light sea green         ";           colors(i,2) = "#20B2AA"; i = i + 1
!         colors(i,1) = "dark slate gray     ";               colors(i,2) = "#2F4F4F"; i = i + 1
!         colors(i,1) = "teal    ";                           colors(i,2) = "#008080"; i = i + 1
!         colors(i,1) = "dark cyan   ";                       colors(i,2) = "#008B8B"; i = i + 1
!         colors(i,1) = "aqua    ";                           colors(i,2) = "#00FFFF"; i = i + 1
!         colors(i,1) = "cyan        ";                       colors(i,2) = "#00FFFF"; i = i + 1
!         colors(i,1) = "light cyan      ";                   colors(i,2) = "#E0FFFF"; i = i + 1
!         colors(i,1) = "dark turquoise      ";               colors(i,2) = "#00CED1"; i = i + 1
!         colors(i,1) = "turquoise       ";                   colors(i,2) = "#40E0D0"; i = i + 1
!         colors(i,1) = "medium turquoise    ";               colors(i,2) = "#48D1CC"; i = i + 1
!         colors(i,1) = "pale turquoise  ";                   colors(i,2) = "#AFEEEE"; i = i + 1
!         colors(i,1) = "aqua marine         ";               colors(i,2) = "#7FFFD4"; i = i + 1
!         colors(i,1) = "powder blue     ";                   colors(i,2) = "#B0E0E6"; i = i + 1
!         colors(i,1) = "cadet blue  ";                       colors(i,2) = "#5F9EA0"; i = i + 1
!         colors(i,1) = "steel blue      ";                   colors(i,2) = "#4682B4"; i = i + 1
!         colors(i,1) = "corn flower blue    ";               colors(i,2) = "#6495ED"; i = i + 1
!         colors(i,1) = "deep sky blue   ";                   colors(i,2) = "#00BFFF"; i = i + 1
!         colors(i,1) = "dodger blue         ";               colors(i,2) = "#1E90FF"; i = i + 1
!         colors(i,1) = "light blue      ";                   colors(i,2) = "#ADD8E6"; i = i + 1
!         colors(i,1) = "sky blue    ";                       colors(i,2) = "#87CEEB"; i = i + 1
!         colors(i,1) = "light sky blue  ";                   colors(i,2) = "#87CEFA"; i = i + 1
!         colors(i,1) = "midnight blue       ";               colors(i,2) = "#191970"; i = i + 1
!         colors(i,1) = "navy    ";                           colors(i,2) = "#000080"; i = i + 1
!         colors(i,1) = "dark blue   ";                       colors(i,2) = "#00008B"; i = i + 1
!         colors(i,1) = "medium blue     ";                   colors(i,2) = "#0000CD"; i = i + 1
!         colors(i,1) = "blue        ";                       colors(i,2) = "#0000FF"; i = i + 1
!         colors(i,1) = "royal blue      ";                   colors(i,2) = "#4169E1"; i = i + 1
!         colors(i,1) = "blue violet         ";               colors(i,2) = "#8A2BE2"; i = i + 1
!         colors(i,1) = "indigo  ";                           colors(i,2) = "#4B0082"; i = i + 1
!         colors(i,1) = "dark slate blue     ";               colors(i,2) = "#483D8B"; i = i + 1
!         colors(i,1) = "slate blue      ";                   colors(i,2) = "#6A5ACD"; i = i + 1
!         colors(i,1) = "medium slate blue   ";               colors(i,2) = "#7B68EE"; i = i + 1
!         colors(i,1) = "medium purple   ";                   colors(i,2) = "#9370DB"; i = i + 1
!         colors(i,1) = "dark magenta        ";               colors(i,2) = "#8B008B"; i = i + 1
!         colors(i,1) = "dark violet     ";                   colors(i,2) = "#9400D3"; i = i + 1
!         colors(i,1) = "dark orchid         ";               colors(i,2) = "#9932CC"; i = i + 1
!         colors(i,1) = "medium orchid   ";                   colors(i,2) = "#BA55D3"; i = i + 1
!         colors(i,1) = "purple      ";                       colors(i,2) = "#800080"; i = i + 1
!         colors(i,1) = "thistle         ";                   colors(i,2) = "#D8BFD8"; i = i + 1
!         colors(i,1) = "plum        ";                       colors(i,2) = "#DDA0DD"; i = i + 1
!         colors(i,1) = "violet  ";                           colors(i,2) = "#EE82EE"; i = i + 1
!         colors(i,1) = "magenta             ";               colors(i,2) = "#FF00FF"; i = i + 1
!         colors(i,1) = "orchid  ";                           colors(i,2) = "#DA70D6"; i = i + 1
!         colors(i,1) = "medium violet red   ";               colors(i,2) = "#C71585"; i = i + 1
!         colors(i,1) = "pale violet red         ";           colors(i,2) = "#DB7093"; i = i + 1
!         colors(i,1) = "deep pink   ";                       colors(i,2) = "#FF1493"; i = i + 1
!         colors(i,1) = "hot pink        ";                   colors(i,2) = "#FF69B4"; i = i + 1
!         colors(i,1) = "light pink  ";                       colors(i,2) = "#FFB6C1"; i = i + 1
!         colors(i,1) = "pink    ";                           colors(i,2) = "#FFC0CB"; i = i + 1
!         colors(i,1) = "antique white       ";               colors(i,2) = "#FAEBD7"; i = i + 1
!         colors(i,1) = "beige   ";                           colors(i,2) = "#F5F5DC"; i = i + 1
!         colors(i,1) = "bisque      ";                       colors(i,2) = "#FFE4C4"; i = i + 1
!         colors(i,1) = "blanched almond         ";           colors(i,2) = "#FFEBCD"; i = i + 1
!         colors(i,1) = "wheat       ";                       colors(i,2) = "#F5DEB3"; i = i + 1
!         colors(i,1) = "corn silk       ";                   colors(i,2) = "#FFF8DC"; i = i + 1
!         colors(i,1) = "lemon chiffon       ";               colors(i,2) = "#FFFACD"; i = i + 1
!         colors(i,1) = "light golden rod yellow         ";   colors(i,2) = "#FAFAD2"; i = i + 1
!         colors(i,1) = "light yellow        ";               colors(i,2) = "#FFFFE0"; i = i + 1
!         colors(i,1) = "saddle brown    ";                   colors(i,2) = "#8B4513"; i = i + 1
!         colors(i,1) = "sienna      ";                       colors(i,2) = "#A0522D"; i = i + 1
!         colors(i,1) = "chocolate       ";                   colors(i,2) = "#D2691E"; i = i + 1
!         colors(i,1) = "peru        ";                       colors(i,2) = "#CD853F"; i = i + 1
!         colors(i,1) = "sandy brown     ";                   colors(i,2) = "#F4A460"; i = i + 1
!         colors(i,1) = "burly wood  ";                       colors(i,2) = "#DEB887"; i = i + 1
!         colors(i,1) = "tan     ";                           colors(i,2) = "#D2B48C"; i = i + 1
!         colors(i,1) = "rosy brown  ";                       colors(i,2) = "#BC8F8F"; i = i + 1
!         colors(i,1) = "moccasin        ";                   colors(i,2) = "#FFE4B5"; i = i + 1
!         colors(i,1) = "navajo white        ";               colors(i,2) = "#FFDEAD"; i = i + 1
!         colors(i,1) = "peach puff      ";                   colors(i,2) = "#FFDAB9"; i = i + 1
!         colors(i,1) = "misty rose  ";                       colors(i,2) = "#FFE4E1"; i = i + 1
!         colors(i,1) = "lavender blush  ";                   colors(i,2) = "#FFF0F5"; i = i + 1
!         colors(i,1) = "linen       ";                       colors(i,2) = "#FAF0E6"; i = i + 1
!         colors(i,1) = "old lace        ";                   colors(i,2) = "#FDF5E6"; i = i + 1
!         colors(i,1) = "papaya whip         ";               colors(i,2) = "#FFEFD5"; i = i + 1
!         colors(i,1) = "sea shell       ";                   colors(i,2) = "#FFF5EE"; i = i + 1
!         colors(i,1) = "mint cream  ";                       colors(i,2) = "#F5FFFA"; i = i + 1
!         colors(i,1) = "slate gray      ";                   colors(i,2) = "#708090"; i = i + 1
!         colors(i,1) = "light slate gray    ";               colors(i,2) = "#778899"; i = i + 1
!         colors(i,1) = "light steel blue        ";           colors(i,2) = "#B0C4DE"; i = i + 1
!         colors(i,1) = "lavender    ";                       colors(i,2) = "#E6E6FA"; i = i + 1
!         colors(i,1) = "floral white    ";                   colors(i,2) = "#FFFAF0"; i = i + 1
!         colors(i,1) = "alice blue  ";                       colors(i,2) = "#F0F8FF"; i = i + 1
!         colors(i,1) = "ghost white     ";                   colors(i,2) = "#F8F8FF"; i = i + 1
!         colors(i,1) = "honeydew    ";                       colors(i,2) = "#F0FFF0"; i = i + 1
!         colors(i,1) = "ivory   ";                           colors(i,2) = "#FFFFF0"; i = i + 1
!         colors(i,1) = "azure       ";                       colors(i,2) = "#F0FFFF"; i = i + 1
!         colors(i,1) = "snow    ";                           colors(i,2) = "#FFFAFA"; i = i + 1
!         colors(i,1) = "black       ";                       colors(i,2) = "#000000"; i = i + 1
!         colors(i,1) = "dim gray ";                          colors(i,2) = "#696969"; i = i + 1
!         colors(i,1) = "gray                ";               colors(i,2) = "#808080"; i = i + 1
!         colors(i,1) = "dark gray               ";           colors(i,2) = "#A9A9A9"; i = i + 1
!         colors(i,1) = "silver      ";                       colors(i,2) = "#C0C0C0"; i = i + 1
!         colors(i,1) = "light gray                      ";   colors(i,2) = "#D3D3D3"; i = i + 1
!         colors(i,1) = "gainsboro   ";                       colors(i,2) = "#DCDCDC"; i = i + 1
!         colors(i,1) = "white smoke     ";                   colors(i,2) = "#F5F5F5"; i = i + 1
!         colors(i,1) = "white       ";                       colors(i,2) = "#FFFFFF"; i = i + 1
! 
!         ! Default to black
!         rgb = "#000000"
!     
!         ! Search for colors
!         do i = 1,size(colors)
!             if (colors(i,1)==color) then
!                 rgb = trim(colors(i,2))
!                 return
!             else 
!                 write(0,*) 'WARNING: rgb: color not found'
!             end if
!         end do
! 
!     end function

! A triplet specification consists of three things: 
! color (can be any name listed inside rgb)
! linestyle (one one lines, linespoints, lin

! b     blue          .     point              -     solid
! g     green         o     circle             :     dotted
! r     red           x     x-mark             -.    dashdot 
! c     cyan          +     plus               --    dashed   
! m     magenta       *     star             (none)  no line
! y     yellow        s     square
! k     black         d     diamond
! w     white         v     triangle (down)
!                     ^     triangle (up)
!                     <     triangle (left)
!                     >     triangle (right)
!                     p     pentagram
!                     h     hexagram
! 
!     
! 
!     function tripletspec(color,mrk,lne) result(cmdout)
!         implicit none
!         character(len=*), optional, intent(in) :: col
!         character(len=*), optional, intent(in) :: mrk,lne
! 
!         character(len=50) :: cmd,cmdout
!         character(len=1) :: c,marker
!         character(len=2) :: style
!         integer :: i
! 
! 
!         marker = ''
!         do i = 1,4
!             select case (t%s(i:i))
!                 case ('b','g','r','c','m','y','k','w')
!                     c = t%s(i:i)
!                 case ('-',':') 
!                     if (t%s(i+1:i+1)=='.') then
!                         style = '-.'
!                     else
!                         style(1:1) = t%s(i:i)
!                         style(2:2) = ' '
!                     end if
!                 case default 
!             end select
!         end do
! 
! 
!         cmd = ''
!         select case (style)
!             case ("- ")
!                 cmd = trim(cmd)//" with lines "
!             case (": ")
!                 cmd = trim(cmd)//" with dots "
!             case ("-.")
!                 cmd = trim(cmd)//" with linespoints "
!             case ("  ")
!                 cmd = trim(cmd)//" with points "
!             case default
!         end select
!         select case (c)
!             case ('b')
!                 cmd = trim(cmd)//' lt rgb "blue" '
!             case ('g')
!                 cmd = trim(cmd)//' lt rgb "green" '
!             case ('r')
!                 cmd = trim(cmd)//' lt rgb "red" '
!             case ('c')
!                 cmd = trim(cmd)//' lt rgb "cyan" '
!             case ('m')
!                 cmd = trim(cmd)//' lt rgb "magenta" '
!             case ('y')
!                 cmd = trim(cmd)//' lt rgb "yellow" '
!             case ('k')
!                 cmd = trim(cmd)//' lt rgb "black" '
!             case ('w')
!                 cmd = trim(cmd)//' lt rgb "white" '
!             case default
!         end select
!         select case (marker)
!             ! Currently not implemented
!             case default
!         end select
!             
! !         print*,'cmd:',cmd
!         cmdout = trim(cmd)
! 
!     end function 
! mod_matlab: 
! A collection of Matlab style routines in Fortran.  Uses BLAS and LAPACK libraries. 
! These routines have not been extensively tested, and some have not really been
! tested at all.  Use with caution!
!
! The routine interfaces are intentionally meant to resemble their matlab counterpart.
! As such, I feel that I should mention that I have explicitly copied some lines of code,
! in particular the interfaces (e.g. function eye(n) result(I)) from the Matlab documentation.
! Here is a basic citation:
! Matlab Help Files.  Mathworks, Cambridge MA. 2002.
! 
! Additionally, many of the matrix algorithms (like Chol) were constructed either
! following Wikipedia or the Intel MKL documentation. 
!
! Various other routines are borrowed from various collections on the internet (e.g.
! the hp filter routine is modified from a Matlab exchange code). I have done
! my best to make acknowledgements where they are do. If you find code that has 
! not been properly acknowledged, please let me know and I'll rectify the situation 
! immediately.
!
! Author: Grey Gordon 2010-2013
! Soli Deo Gloria
!
module mod_matlab

    use mod_core ! Some core routines, like tic,toc,linspace,...
    use mod_sort, only: issorted, &
                        defsort=>msort ! Here I am naming msort as the default sort routine 
                                       ! Leave this as the default because (1) it comes from
                                       ! ORDERPACK 2.0 and so is probably better tested 
                                       ! (2) it is quick and (3) mergesort is stable 
    use mod_plot
    implicit none

    private :: defsort ! Make the routine private

    ! Overload 
    interface dot
        real(8) function ddot(n,dx,incx,dy,incy)
            integer :: incx,incy,n
            real(8) :: dx(*),dy(*)
        end function 
        real(4) function sdot(n,dx,incx,dy,incy)
            integer :: incx,incy,n
            real(4) :: dx(*),dy(*)
        end function 
    end interface

    interface gesvd
        subroutine dgesvd(jobu,jobvt,m,n,a,lda,s,u,ldu,vt,ldvt,work,lwork,info)
            character :: jobu, jobvt
            integer :: info, lda, ldu, ldvt, lwork, m, n
            real(8) :: a(lda,*), s(*), u(ldu,*), vt(ldvt,*), work(*)
        end subroutine
        subroutine sgesvd(jobu,jobvt,m,n,a,lda,s,u,ldu,vt,ldvt,work,lwork,info)
            character :: jobu, jobvt
            integer :: info, lda, ldu, ldvt, lwork, m, n
            real(4) :: a(lda,*), s(*), u(ldu,*), vt(ldvt,*), work(*)
        end subroutine
    end interface

    interface gesv
        subroutine dgesv( n, nrhs, a, lda, ipiv, b, ldb, info )
            integer            info, lda, ldb, n, nrhs
            integer            ipiv( * )
            double precision   a( lda, * ), b( ldb, * )
        end subroutine
        subroutine sgesv( n, nrhs, a, lda, ipiv, b, ldb, info )
            integer            info, lda, ldb, n, nrhs
            integer            ipiv( * )
            real a( lda, * ), b( ldb, * )
        end subroutine
    end interface

    interface gemm
        subroutine dgemm(transa,transb,m,n,k,alpha,a,lda,b,ldb,beta,c,ldc)
            double precision alpha,beta
            integer k,lda,ldb,ldc,m,n
            character transa,transb
            double precision a(lda,*),b(ldb,*),c(ldc,*)
        end subroutine 
        subroutine sgemm(transa,transb,m,n,k,alpha,a,lda,b,ldb,beta,c,ldc)
            real(4) alpha,beta
            integer k,lda,ldb,ldc,m,n
            character transa,transb
            real(4) a(lda,*),b(ldb,*),c(ldc,*)
        end subroutine
    end interface

    interface gemv
          subroutine dgemv(trans,m,n,alpha,a,lda,x,incx,beta,y,incy)
                double precision alpha,beta
                integer incx,incy,lda,m,n
                character trans
                double precision a(lda,*),x(*),y(*)
        end subroutine 
          subroutine sgemv(trans,m,n,alpha,a,lda,x,incx,beta,y,incy)
                real(4) alpha,beta
                integer incx,incy,lda,m,n
                character trans
                real(4) a(lda,*),x(*),y(*)
        end subroutine 
    end interface

    interface gbtrf
        subroutine dgbtrf( m, n, kl, ku, ab, ldab, ipiv, info )
            integer            info, kl, ku, ldab, m, n
            integer            ipiv( * )
            double precision   ab( ldab, * )
        end subroutine
        subroutine sgbtrf( m, n, kl, ku, ab, ldab, ipiv, info )
            integer            info, kl, ku, ldab, m, n
            integer            ipiv( * )
            real               ab( ldab, * )
        end subroutine
    end interface

    interface gbtrs
        subroutine dgbtrs( trans, n, kl, ku, nrhs, ab, ldab, ipiv, b, ldb, info )
            character          trans
            integer            info, kl, ku, ldab, ldb, n, nrhs
            integer            ipiv( * )
            double precision   ab( ldab, * ), b( ldb, * )
        end subroutine
        subroutine sgbtrs( trans, n, kl, ku, nrhs, ab, ldab, ipiv, b, ldb, info )
            character          trans
            integer            info, kl, ku, ldab, ldb, n, nrhs
            integer            ipiv( * )
            real               ab( ldab, * ), b( ldb, * )
        end subroutine
    end interface

    interface hseqr
        subroutine dhseqr( job, compz, n, ilo, ihi, h, ldh, wr, wi, z, ldz, work, lwork, info )
        integer            ihi, ilo, info, ldh, ldz, lwork, n
        character          compz, job
        double precision   h( ldh, * ), wi( * ), work( * ), wr( * ), z( ldz, * )
        end subroutine
        subroutine shseqr( job, compz, n, ilo, ihi, h, ldh, wr, wi, z, ldz, work, lwork, info )
        integer            ihi, ilo, info, ldh, ldz, lwork, n
        character          compz, job
        real               h( ldh, * ), wi( * ), work( * ), wr( * ), z( ldz, * )
        end subroutine
    end interface

    
    interface gttrf
        subroutine dgttrf( n, dl, d, du, du2, ipiv, info )
            integer            info, n
            integer            ipiv( * )
            double precision   d( * ), dl( * ), du( * ), du2( * )
        end subroutine
        subroutine sgttrf( n, dl, d, du, du2, ipiv, info )
            integer            info, n
            integer            ipiv( * )
            real               d( * ), dl( * ), du( * ), du2( * )
        end subroutine
    end interface

    interface gttrs
        subroutine dgttrs( trans, n, nrhs, dl, d, du, du2, ipiv, b, ldb, info )
            character          trans
            integer            info, ldb, n, nrhs
            integer            ipiv( * )
            double precision   b( ldb, * ), d( * ), dl( * ), du( * ), du2( * )
        end subroutine 
        subroutine sgttrs( trans, n, nrhs, dl, d, du, du2, ipiv, b, ldb, info )
            character          trans
            integer            info, ldb, n, nrhs
            integer            ipiv( * )
            real               b( ldb, * ), d( * ), dl( * ), du( * ), du2( * )
        end subroutine 
    end interface


    ! Overload procedures
    interface corr
        module procedure corr_mat, corr_vecs, corr_mats, corr_matvec, corr_vecmat
    end interface corr
    interface cumsum
        module procedure cumsum_vec, cumsum_vec_int, cumsum_mat
    end interface cumsum
    interface cumprod
        module procedure cumprod_r, cumprod_i
    end interface cumprod
    interface diag
        module procedure diag_mat, diag_vec, diag_mat_k
    end interface diag
    interface diff
        module procedure diff_vec, diff_mat
    end interface diff
    interface find1
        module procedure find1_xs
    end interface find1
!     interface find
!         module procedure find_x, find_xk, find_xks
!     end interface find
    interface finda
        module procedure find_x_alloc
    end interface
    interface hist
        module procedure hist_x, hist_xn
    end interface hist
    interface horzcat
        module procedure horzcat2, horzcat3, horzcat4, horzcat2mat
    end interface horzcat
    interface interp1
        module procedure interp1_3dim, interp1_mat, interp1_vec, interp1_scal, interp1_scal_mat, interp1_scal_3d
    end interface interp1
    interface interp1q
        module procedure interp1q_scal,interp1q_vec,interp1q_scal_ymat,interp1q_vec_ymat
    end interface interp1q
    interface interp1qf ! even faster version of interp1q, but requires xi be sorted
        module procedure interp1q_scal,interp1q_scal_ymat,interp1qf_vec,interp1qf_vec_ymat ! note: when the xi are scalrs, should be equivalent to interp1q
    end interface interp1qf
    interface interp2q
        module procedure interp2q_scal_ymat,interp2q_vec_ymat,interp2q_vec_ymat2,interp2q_scal
    end interface interp2q
    interface mean
        module procedure mean_vec, mean_mat
    end interface mean   
    interface meshgrid
        module procedure meshgrid_int2d,meshgrid_int3d,meshgrid_double2d,meshgrid_double3d
    end interface meshgrid
    interface mldivide
        module procedure mldivide_full, mldivide_band, mldivide_full_mat, mldivide_tridiag
    end interface
    interface mtimes
        module procedure mtimes_matmat, mtimes_matvec
    end interface 
    interface mydot
        module procedure mydot_4d
    end interface 
    interface ndgrid 
        module procedure ndgrid_double2d
    end interface ndgrid
    interface num2str
        module procedure num2str_double, num2str_int
    end interface
    interface ones
        module procedure ones_sqmat, ones_mat, ones_vec2
    end interface ones
    interface plot
        module procedure plot_y_fine,plot_xy_fine,plot_y2_fine,plot_xy2_fine
    end interface plot
    interface plot_coarse
        module procedure plot_y_coarse,plot_xy_coarse
    end interface plot_coarse
    interface randU
        module procedure scal_rand,vec_rand,mat_rand
    end interface randU
    interface randN
        module procedure scal_randN,vec_randN, mat_randN
    end interface randN
    interface repmat
        module procedure repmat_mat_2, repmat_scal_2, repmat_scal_1
    end interface repmat
    interface sort
        module procedure fn_sort_vec
    end interface sort   
    interface sort2
        module procedure sub_sort_vec_i, sub_sort_vec, sub_sort_mat
    end interface sort2
    interface squeeze
        module procedure squeeze_vec, squeeze_mat, squeeze_vec_i
    end interface squeeze
    interface sortrows
        module procedure sortrows_real, sortrows_int
    end interface
    interface std
        module procedure std_vec, std_mat
    end interface std
    interface supnorm
        module procedure supnorm_4d
    end interface supnorm
    interface true
        module procedure true_sqmat, true_mat
    end interface true
    interface unique
        module procedure unique_real, unique_real_vec, unique_int, unique_real_vec_noinds, unique_int_vec_noinds
    end interface unique
    interface vec
        module procedure vec_mat, vec_3d, vec_4d, vec_mat_i, vec_3d_i, vec_4d_i
    end interface vec
    interface vertcat
        module procedure vertcat2mat
    end interface vertcat

    logical, private :: randomSeedInitialized = .FALSE.
    type stream
        integer :: m
        integer, allocatable :: seeds(:)
    end type

    ! Parameter constants
    real(rt), parameter :: pi = 3.14159265358979323846264338327950288419716939937510582097494459e0_rt
    real(rt), parameter :: sqrt_of_2 = sqrt(2e0_rt)
    real(rt), parameter :: sqrt_of_2pi = sqrt(2e0_rt*pi)

    type type_interp2q_work 
        integer :: ilo1,ilo2
        real(rt) :: wa,wb,wc,wd
    end type

contains

! Solves a tridiagonal system A*X = B for X where A is tridiagonal
function mldivide_tridiag(Al,Ad,Au,B) result(X)
    implicit none
    real(rt), intent(in) :: Al(:),Ad(:),Au(:),B(:)
    real(rt) :: X(size(B))

    if (size(Ad)/=size(Al)+1) stop 'ERROR: mldivide_tridiag: input: wrong size'
    if (size(Ad)/=size(Au)+1) stop 'ERROR: mldivide_tridiag: input: wrong size'
    if (size(Ad)/=size(B)) stop 'ERROR: mldivide_tridiag: input: wrong size'

    call mldivide_tridiag_core(X,'N',size(Ad),Al,Ad,Au,size(B),1,B)

end function
subroutine mldivide_tridiag_core(X,tA,n,Al,Ad,Au,rowB,colB,B)
    implicit none
    character(len=1), intent(in) :: tA
    integer, intent(in) :: n,rowB,colB
    real(rt), intent(in) :: Al(:),Ad(:),Au(:),B(rowB,*)
    real(rt), intent(out) :: X(rowB,*)
    ! local
    real(rt), allocatable :: Au2(:)
    integer, allocatable :: ipiv(:)
    integer :: info

    allocate(Au2(n-2))
    allocate(ipiv(n))

    ! LU factorization
    call gttrf(n,Al,Ad,Au,Au2,ipiv,info)

    if (info/=0) write(0,*) 'WARNING: mldivid_tridiag: gttrf had an error'

    ! Solution
    call gttrs(tA,n,colB,Al,Ad,Au,Au2,ipiv,B,rowB,info)
    if (info/=0) write(0,*) 'WARNING: mldivid_tridiag: gttrs had an error'

    ! Output
    X(:,1:colB) = B(:,1:colB)


end subroutine



! From http://fortranwiki.org/fortran/show/random_seed with slight modifications
subroutine init_random_seed(shift)
    implicit none
    integer, intent(in), optional :: shift
    integer :: i, n, clock
    integer, dimension(:), allocatable :: seed

    call random_seed(size = n)
    allocate(seed(n))

    call system_clock(count=clock)

    if (present(shift)) then
        seed = clock + 37 * (/ (i - 1, i = 1, n) /) + 67*shift
    else
        seed = clock + 37 * (/ (i - 1, i = 1, n) /)
    end if
    call random_seed(put = seed)

    deallocate(seed)
end subroutine

! Computes the supnorm of two vectors v0,v1 (maximum absolute value between them)
function supnorm_4d(v0,v1) result(supnorm)
    implicit none
    real(rt), intent(in) :: v0(:,:,:,:),v1(:,:,:,:)
    real(rt) :: supnorm
    supnorm = supnorm_core(v0,v1,size(v0))
end function 
function supnorm_core(v0,v1,n) result(supnorm)
    implicit none
    integer, intent(in) :: n
    real(rt) :: v0(*),v1(*)
    real(rt) :: supnorm
    ! local
    real(rt) :: m
    integer :: i

    m = abs(v0(1) - v1(1))
    do i = 2,n
        m = max(m,abs(v0(i) - v1(i)))
    end do
    supnorm = m

end function 

! BLAS ddot call
function mydot_4d(x,y) result(myddot)
    implicit none
    real(rt), dimension(:,:,:,:), intent(in) :: x,y
    real(rt) :: myddot
    real(rt) :: ddot
    external :: ddot
    myddot = ddot(size(x),x,1,y,1)
end function 


! Code is based of the matlab ind2sub.m
! Given an integer vector siz that gives the shape of
! some array and the linear index i, convert the linear
! index into an i1,i2,... such that 
!  a(i) = a(i1,i2,...)
function ind2sub(siz,i) result(ivec)
    integer, intent(in) :: siz(:)
    integer :: i
    integer :: ivec(size(siz))
    ! local
    integer :: d
    integer :: j,k
    integer :: cumprodivec(size(siz))

    cumprodivec(1) = 1
    cumprodivec(2:size(siz)) = cumprod(siz(1:size(siz)-1))

    j = i
    do d = size(siz),1,-1
        k = mod(j-1,cumprodivec(d)) + 1
        ivec(d) = (j-k)/cumprodivec(d) + 1
        j = k
    end do

end function

! Same as Matlab's vertcat. 
function vertcat2mat(x1,x2) result(x)
    implicit none
    real(rt), dimension(1:,1:), intent(in) :: x1,x2
    real(rt) :: x(size(x1,1)+size(x2,1),size(x1,2))

    if (size(x1,2)/=size(x2,2)) STOP 'vertcat: ERROR: x1,x2 dimensions don''t conform'

    x(1:size(x1,1),:) = x1
    x(size(x1,1)+1:size(x1,1)+size(x2,1),:) = x2

end function 

! Same as Matlab's horzcat. 
function horzcat2mat(x1,x2) result(x)
    implicit none
    real(rt), dimension(1:,1:), intent(in) :: x1,x2
    real(rt) :: x(size(x1,1),size(x1,2)+size(x2,2))

    if (size(x1,1)/=size(x2,1)) STOP 'horzcat: ERROR: x1,x2 dimensions don''t conform'

    x(:,1:size(x1,2)) = x1
    x(:,size(x1,2)+1:size(x1,2)+size(x2,2)) = x2

end function 
function horzcat2(x1,x2) result(x)
    implicit none
    real(rt), dimension(1:), intent(in) :: x1,x2
    real(rt) :: x(size(x1),2)

    if (size(x1)/=size(x2)) STOP 'horzcat: ERROR: x1,x2 dimensions don''t conform'

    x(:,1) = x1
    x(:,2) = x2

end function 
function horzcat3(x1,x2,x3) result(x)
    implicit none
    real(rt), dimension(1:), intent(in) :: x1,x2,x3
    real(rt) :: x(size(x1),3)

    if (size(x1)/=size(x2)) STOP 'horzcat: ERROR: x1,x2 dimensions don''t conform'
    if (size(x1)/=size(x3)) STOP 'horzcat: ERROR: x1,x3 dimensions don''t conform'

    x(:,1) = x1
    x(:,2) = x2
    x(:,3) = x3

end function 
function horzcat4(x1,x2,x3,x4) result(x)
    implicit none
    real(rt), dimension(1:), intent(in) :: x1,x2,x3,x4
    real(rt) :: x(size(x1),4)

    if (size(x1)/=size(x2)) STOP 'horzcat: ERROR: x1,x2 dimensions don''t conform'
    if (size(x1)/=size(x3)) STOP 'horzcat: ERROR: x1,x3 dimensions don''t conform'
    if (size(x1)/=size(x4)) STOP 'horzcat: ERROR: x1,x3 dimensions don''t conform'

    x(:,1) = x1
    x(:,2) = x2
    x(:,3) = x3
    x(:,4) = x4

end function 

! Given a matrix X that has a singleton dimension, will convert to a vector. O/w error is thrown.
function mat2vec(X) result(Y)
    implicit none
    real(rt), intent(in) :: X(1:,1:)
    real(rt) :: Y(size(X))

    if (size(X,1)==1) then
        Y = X(1,:)
    elseif (size(X,2)==1) then
        Y = X(:,1)
    else
        STOP 'mat2vec: ERROR: X must have a singleton dim'
    end if

end function

! Given a vector X, reshapes it into a matrix w size (nX,1) 
function vec2mat(X) result(Y)
    implicit none
    real(rt), intent(in) :: X(1:)
    real(rt) :: Y(1:size(X),1)
    
    Y = reshape(X,(/size(X),1/))

end function vec2mat
! Given a length one vector X, return a scalar
function vec2scal(X) result(Y)
    implicit none
    real(rt), intent(in) :: X(1)
    real(rt) :: Y
    Y = X(1)
end function

! Similar to repmat in Matlab, will take X and tile it (a,b) times. 
! Can only be called using a matrix X (if need to use a vector, first use
! vec2mat).
!
! Note: could overload this by doing repmat(X,a,b,c,d,...)
function repmat_mat_2(X,a,b) result(R)
    implicit none
    real(rt), intent(in) :: X(1:,1:)
    integer, intent(in) :: a,b
    real(rt) :: R(1:size(X,1)*a,1:size(X,2)*b)
    ! local
    integer :: i,j,nx1,nx2

    nx1 = size(X,1)
    nx2 = size(X,2)

    ! Tile
    do j = 1,nx2*b
        do i = 1,nx1*a
            R(i,j) = X(1+mod(i-1,nx1),1+mod(j-1,nx2))
        end do
    end do

end function repmat_mat_2
function repmat_scal_2(X,a,b) result(R)
    implicit none
    real(rt), intent(in) :: X
    integer, intent(in) :: a,b
    real(rt) :: R(a,b)

    R = x

end function repmat_scal_2
function repmat_scal_1(X,a) result(R)
    implicit none
    real(rt), intent(in) :: X
    integer, intent(in) :: a
    real(rt) :: R(a)

    R = x

end function repmat_scal_1

! HP Filters the series y using smoothing parameter w
! Returns the deviation "dev"
! w = 100 is common for annual
! w = 1600 is common for quarterly
! This code is based off Wilmer Henao's HP filter code for matlab 
! Copyright (c) 2003, Wilmer Henao
! All rights reserved.
! 
! Redistribution and use in source and binary forms, with or without 
! modification, are permitted provided that the following conditions are 
! met:
! 
!     * Redistributions of source code must retain the above copyright 
!       notice, this list of conditions and the following disclaimer.
!     * Redistributions in binary form must reproduce the above copyright 
!       notice, this list of conditions and the following disclaimer in 
!       the documentation and/or other materials provided with the distribution
!             
!       THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
!       AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
!       IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
!       ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
!       LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
!       CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
!       SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
!       INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
!       CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
!       ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
!       POSSIBILITY OF SUCH DAMAGE.
function hpfilter(y,w) result(dev) 
    implicit none
    real(rt), dimension(:), intent(in) :: y
    real(rt), intent(in) :: w !
    real(rt), dimension(size(y)) :: dev
    !local
    integer :: m,j
    real(rt), allocatable :: B(:,:), d(:,:)

    allocate(B(size(y),size(y)),d(size(y),3))
    
    ! Ensure that there is something to filter by checking the stdev is not zero
    if (std(y)<10.0_rt*epsilon(0._rt)) then
        print*,'hpfilter: WARNING: input series had little or no variability: stdev is ',&
                std(y),' max is ',maxval(y),' min is ',minval(y),' returning zero deviation.'
        dev = 0._rt
        return
    end if



    m = size(y)
    
    d = spread((/w, -4.e0_rt*w, (6.e0_rt*w+1.e0_rt)/2.e0_rt /),1,m)
    d(1,2) = -2.e0_rt*w;            d(m-1,2) = -2.e0_rt*w
    d(1,3) = (1.e0_rt+w)/2.e0_rt;      d(m,3) = (1.e0_rt+w)/2.e0_rt
    d(2,3) = (5.e0_rt*w+1.e0_rt)/2.e0_rt; d(m-1,3) = (5.e0_rt*w+1.e0_rt)/2.e0_rt

    ! Produce the banded matrix
!     B = 0.e0_rt
!     do j = 1,m
!         B(j,j) = d(j,3)
!         if (j<m) B(j+1,j) = d(j,2)
!         if (j<m-1) B(j+2,j) = d(j,1)        
!     end do
!     B = B + transpose(B)

    do j = 1,m
        B(j,j) = 2e0_rt*d(j,3)
        if (j<m) then
            B(j+1,j) = d(j,2)
            B(j,j+1) = d(j,2)
        end if
        if (j<m-1) then
            B(j+2,j) = d(j,1)        
            B(j,j+2) = d(j,1)        
        end if
    end do

    !!! NOTE: it is quite slow to call mldivide here.
    !!! Because B is a banded sparse matrix, we could use the routine
    !!! dgbtrf followed by dgbtrs to compute the result.

    !!! DO NOT!!! USE THIS COMMENTED LINE OF CODE W/O FIRST SETTING B=0 ON OFF DIAGONAL ELEMENTS
!     dev = y - mldivide(B,y)

    !!! Do use this
    dev = y - mldivide_band(B,y,2,2)

    deallocate(B,d)

end function hpfilter

!  [auto]=autocorr(x,F,mu)
!  Computes the autocorrelation of some discrete process with grid x,
!  transition matrix F, and initial distribution mu.
function autocorr(x,F,mu) result(auto)
    implicit none
    real(rt), intent(in) :: x(1:),F(1:,1:),mu(1:)
    real(rt) :: auto
    !local
    real(rt) :: Ex,Exp,sigx,sigxp,accum
    real(rt), allocatable :: mup(:)
    integer :: i

    allocate(mup(size(mu)))

    ! Check that mu is a distribution
    if (abs(sum(mu)-1.e0_rt)>1d-15) STOP 'autocorr: mu is not a distribution, it doesnt sum to 1'
    if (any(mu<0.e0_rt)) STOP 'autocorr: mu is not a distribution, it has negative elements'
    ! Check that F is a transition matrix
    if (any(abs(sum(F,2)-1.e0_rt)>1d-15)) STOP 'F is not a transition matrix'
    ! Check that x,F, and mu have the correct dimensions
    if (size(x)/=size(mu)) STOP 'autocorr: x and mu sizes don''t agree'
    if (size(x)/=size(F,1)) STOP 'autocorr: x and F1 sizes don''t agree'
    if (size(x)/=size(F,2)) STOP 'autocorr: x and F2 sizes don''t agree'

    ! Compute mean today
    Ex = dot_product(x,mu)
    
    ! Compute stdev today
    sigx = sqrt(dot_product(mu,(x-Ex)**2))
    
    ! Compute dsn tomorrow
    mup = matmul(mu,F)

    ! Compute mean tomorrow
    Exp = dot_product(x,mup)
    
    ! Compute stdev tomorrow
    sigxp = sqrt(dot_product(mup,(x-Exp)**2))

    ! Compute autocorrelation
    accum = 0.e0_rt
    do i = 1,size(x)
        accum = accum + mu(i)*(x(i)-Ex)*(dot_product(F(i,:),(x-Exp)))
    end do
    auto = accum/(sigx*sigxp)

    deallocate(mup)

end function autocorr

! Computes the Cholesky decomposition of A into an upper triangular matrix U satisfying U'*U=A
! I believe this algorithm is based on the descriptions of how to compute the Cholesky decomp
! in the Intel MKL help files. 
function chol(Ain) result(U)
    implicit none
    real(rt), intent(in) :: Ain(1:,1:)
    real(rt) :: U(size(Ain,1),size(Ain,2))
    !local
    character(len=1) :: uplo
    real(rt), allocatable :: A(:,:) 
    integer :: m,n,info, i,j

    allocate(A(size(Ain,1),size(Ain,2)))

    m = size(A,1)
    n = size(A,2)
    if (m/=n) STOP 'chol: matrix must be square'

    uplo = 'U' ! This is the Matlab convention
    
    A = Ain

    call dpotrf( uplo, n, A, n, info )
    
    select case (info)
        case (:-1)
            STOP 'chol: in call to dpotrf, the ith parameter had an illegal value'
        case (0)
            ! Success
        case (1:)
            STOP 'chol: A is not positive definite and so the factorization could not be completed'
    end select

    U = 0.e0_rt
    do i = 1,m
        do j=i,m
            U(i,j) = A(i,j)
        end do
    end do

    deallocate(A)

end function chol


! Compares two strings. Returns equal in the Matlab style way rather than Fortran (which allows unequal
! length strings to be equal by padding the right with spaces).
function strcmp(s1,s2) result(tf)
    implicit none
    character(len=*), intent(in) :: s1,s2
    logical :: tf

    if (len(s1)/=len(s2)) then
        tf = .false.
    else
        tf = s1==s2
    end if

end function strcmp

! Computes the determinant of a matrix A by  
function det(Ain) result(d)
    implicit none
    real(rt), dimension(1:,1:), intent(in) :: Ain
    real(rt) :: d
    !local
    real(rt), allocatable :: A(:,:)
    integer, allocatable :: ipiv(:) 
    integer :: nRowChanges
    integer :: m,n,info
    
    allocate(A(size(Ain,1),size(Ain,2)))
    allocate(ipiv(size(Ain,1)))
    
    m = size(A,1)
    n = size(A,2)
    if (m/=n) STOP 'det: matrix must be square'

    ! Copy over since the dgetrf overwrites A
    A = Ain

    call dgetrf(m,n,A,m,ipiv,info)

    ! Check for an error from dgetrf
    select case (info)
        case (:-1)
            STOP 'det: in call to dgetrf, the ith parameter had an illegal value'
        case (0)

        case (1:)
            write(*,*) 'WARNING: det: in LU factorization, u_ii had zero value.  This may result in problems'
    end select

    ! Count the number of row changes (if even, then the determinant of the Perm matrix is 1, o/w it is -1)
    nRowChanges = count(ipiv /= colon(1,size(Ain,1)))
    
    ! Calculate determinant
    if (mod(nRowChanges,2) == 0) then !if number of row changes was even
        d = product(diag(A))
    else
        d = -product(diag(A))
    end if

    ! Clean up
    deallocate(A,ipiv)
    

end function det

! Sort a vector of indices returning only the unique values
subroutine unique_int_vec_noinds(B,A)
    use mod_orderpack, only: unirnk
    implicit none
    integer, intent(in) :: A(:)
    integer, allocatable, intent(out) :: B(:)
    ! local
    integer, allocatable :: tmp(:)
    integer :: nuni ! number of unique

    allocate(tmp(size(A)))
    call unirnk(A,tmp,nuni)
    allocate(B(nuni))
    B = A(tmp(1:nuni))

end subroutine 

subroutine unique_real_vec_noinds(B,A)
    implicit none
    real(rt), intent(in) :: A(:)
    real(rt), allocatable, intent(out) :: B(:)
    ! local
    integer, allocatable :: redo(:),undo(:)

    allocate(undo(size(A)))
    call unique_real_vec(B,redo,undo,A)

end subroutine 
! Given A, finds the unique elements and places them in B. 
! Also, computes redo and undo integer vectors s.t. A(redo)==B
! and B(undo)==A.
subroutine unique_real_vec(B,redo,undo,A)
    implicit none
    real(rt), intent(in) :: A(1:)
    real(rt), allocatable, intent(out) :: B(:)
    integer, allocatable, intent(out) :: redo(:)
    integer, intent(out) :: undo(size(A))
    !local
    real(rt), allocatable :: Asort(:)
    integer, allocatable :: IsortA(:), IrecoverA(:), D(:)
    integer :: nunique

    allocate(Asort(size(A)),IsortA(size(A)),IrecoverA(size(A)),D(size(A)))

    !First sort
    call sort2(Asort,IsortA,A)
    IrecoverA(IsortA) = colon(1,size(A))
    
    !Get indicator vector of unique locations
    D(1) = 1
    D(2:size(A)) = merge(1,0,Asort(1:size(A)-1)/=Asort(2:size(A)))

    !Then allocate the right amount
    nunique = sum(D)
    allocate(B(nunique),redo(nunique))
    
    !Then copy over the unique values
    B = pack(Asort,D==1)
    redo = pack(IsortA,D==1)

    undo = cumsum(D)
    undo(IsortA) = undo

    deallocate(Asort,IsortA,IrecoverA,D)

end subroutine unique_real_vec

! Same as above, but with A a matrix. Currently only does unique rows.
subroutine unique_real(B,redo,undo,A,byRows)
    implicit none
    real(rt), intent(in) :: A(1:,1:)
    real(rt), allocatable, intent(out) :: B(:,:)
    integer, intent(out) :: undo(size(A,1))
    integer, allocatable, intent(out) :: redo(:)
    character(len=*), intent(in), optional :: byRows
    !local
    real(rt), allocatable :: Asort(:,:)
    integer, allocatable :: rank(:), undorank(:), keepRows(:) 
    integer :: i, nKeep

    allocate(Asort(size(A,1),size(A,2)),rank(size(A,1)),undorank(size(A,1)),keepRows(size(A,1)))

    if (present(byRows)) then
        call uniqueByRows()
    else
        STOP 'havent programmed this'
!         call uniqueVals()
    end if

    deallocate(Asort,rank,undorank,keepRows)

contains

    subroutine uniqueByRows()
        implicit none

        ! Sort A by rows
        call sortrows(Asort,rank,A)
        undorank(rank) = colon(1,size(A,1)) !Undoes the sorting

        ! Compare each row with the next row
        nKeep = 1
        keepRows(1) = 1
        undo(1) = nKeep
        do i = 2,size(Asort,1)
            if (any(Asort(i-1,:)/=Asort(i,:))) then
                !if i is unique, keep it
                nKeep = nKeep + 1
                keepRows(nKeep) = i
                undo(i) = nKeep
            else
                !if the row is not unique, find the last row that was and use its index
                undo(i) = nKeep
            end if
        end do
        undo=undo(undorank)        

        ! Get B and redo
        allocate(B(nKeep,size(A,2)),redo(nKeep))
        redo = rank(keepRows(1:nKeep))
        B = Asort(keepRows(1:nKeep),:)

    end subroutine uniqueByRows

end subroutine unique_real
subroutine unique_int(B,redo,undo,A,byRows)
    implicit none
    integer, intent(in) :: A(1:,1:)
    integer, allocatable, intent(out) :: B(:,:)
    integer, intent(out) :: undo(size(A,1))
    integer, allocatable, intent(out) :: redo(:)
    character(len=*), intent(in), optional :: byRows
    !local
    integer, allocatable :: Asort(:,:), rank(:), undorank(:), keepRows(:) 
    integer :: i, nkeep

    allocate(Asort(size(A,1),size(A,2)),rank(size(A,1)),undorank(size(A,1)),keepRows(size(A,1)))

    if (present(byRows)) then
        call uniqueByRows()
    else
        STOP 'ERROR: unique_int: havent programmed this'
!         call uniqueVals()
    end if

    deallocate(Asort,rank,undorank,keepRows)

contains

    subroutine uniqueByRows()
        implicit none

        ! Sort A by rows
        call sortrows(Asort,rank,A)
        undorank(rank) = colon(1,size(A,1)) !Undoes the sorting

        ! Compare each row with the next row
        nKeep = 1
        keepRows(1) = 1
        undo(1) = nKeep
        do i = 2,size(Asort,1)
            if (any(Asort(i-1,:)/=Asort(i,:))) then
                !if i is unique, keep it
                nKeep = nKeep+1
                keepRows(nKeep) = i
                undo(i) = nKeep
            else
                !if the row is not unique, find the last row that was and use its index
                undo(i) = nKeep
            end if
        end do
        undo=undo(undorank)        

        ! Get B and redo
        allocate(B(nKeep,size(A,2)),redo(nKeep))
        redo = rank(keepRows(1:nKeep))
        B = Asort(keepRows(1:nKeep),:)

    end subroutine

end subroutine unique_int

! Sorts A by rows
subroutine sortrows_real(B,redo,A)
    use mod_orderpack, only: mrgrnk
    implicit none
    real(rt), intent(in) :: A(1:,1:)
    real(rt), intent(out) :: B(size(A,1),size(A,2))
    integer, intent(out) :: redo(size(A,1))
    ! local
    integer, allocatable :: tmp_redo(:)
    integer :: icol,ncol,nrow

    ! Sort according to columns from right to left. I believe this 
    ! ensures that when there are equal elements, they are ranked according to
    ! the column to their right as long as the sort is table

    ncol = size(A,2)
    nrow = size(A,1)
    
    redo = colon(1,nrow)

    allocate(tmp_redo(nrow))

    do icol = ncol,1,-1
        
        ! Redo gives the current assortment necessary to redo all changes
        B(:,icol) = A(redo,icol) ! Copy over to prevent a temporary copy being made by the compiler
        call mrgrnk(B(:,icol),tmp_redo) ! Gets the ranking but does not change B which is changed at the end
        redo = redo(tmp_redo)

    end do

    deallocate(tmp_redo)

    ! At the last pass, have to redo all 
    B = A(redo,:)


end subroutine sortrows_real

!Sorts A by rows (keeping each row as a unit)
! Also returns an integer vector redo s.t. B=A(redo,:)
subroutine sortrows_int(B,redo,A)
    use mod_orderpack, only: mrgrnk
    implicit none
    integer, intent(in) :: A(1:,1:)
    integer, intent(out) :: B(size(A,1),size(A,2))
    integer, intent(out) :: redo(size(A,1))
    ! local
    integer, allocatable :: tmp_redo(:)
    integer :: icol,ncol,nrow
    
    ! Sort according to columns from right to left. I believe this 
    ! ensures that when there are equal elements, they are ranked according to
    ! the column to their right as long as the sort is table

    ncol = size(A,2)
    nrow = size(A,1)
    
    redo = colon(1,nrow)

    allocate(tmp_redo(nrow))

    do icol = ncol,1,-1
        
        ! Redo gives the current assortment necessary to redo all changes
        B(:,icol) = A(redo,icol) ! Copy over to prevent a temporary copy being made by the compiler
        call mrgrnk(B(:,icol),tmp_redo) ! Gets the ranking but does not change B which is changed at the end
        redo = redo(tmp_redo)

    end do

    deallocate(tmp_redo)

    ! At the last pass, have to redo all 
    B = A(redo,:)


end subroutine sortrows_int

! Computes n choose k
function choose(n,k) result(val)
    implicit none
    integer, intent(in) :: n,k
    integer :: val

    if (k>n .or. k<0) STOP 'choose: k is unacceptable'

    val = pure_choose(n,k)

end function choose
elemental function pure_choose(n,k) result(val)
    implicit none
    integer, intent(in) :: n,k
    integer :: val
    
    if (k>=n-k) then
        val = product(colon(k+1,n))/pure_factorial(n-k)
    else
        val = product(colon(n-k+1,n))/pure_factorial(k)        
    end if

end function

! Computes the factorial of n
function factorial(n) result(val)
    implicit none
    integer, intent(in) :: n
    integer :: val

    if (n<0) then
        write(*,*) 'factorial: n value',n
        STOP 'factorial: n is less than zero!'
    end if

    val = pure_factorial(n) 

end function factorial
elemental function pure_factorial(n) result(val)
    implicit none
    integer, intent(in) :: n
    integer :: val
    !local
    integer :: ind,tmp

    tmp = 1
    do ind = 2,n
        tmp = tmp*ind
    end do
    val = tmp

end function 

! Converts a number to a string
function num2str_double(A,format) result(str)
    implicit none
    real(rt), intent(in) :: A
    character(len=*), intent(in) :: format
    character(len=50) :: str
    write(str,format) A
end function
function num2str_int(A,format) result(str)
    implicit none
    integer, intent(in) :: A
    character(len=*), intent(in) :: format
    character(len=50) :: str
    write(str,format) A
end function

! Very basic histogram plotting to std out
subroutine hist_x(x)
    implicit none
    real(rt), dimension(1:), intent(in) :: x
    call hist_xn(x,40)
end subroutine
subroutine hist_xn(x,n)
    implicit none
    real(rt), dimension(1:), intent(in) :: x
    integer, intent(in) :: n
    integer, parameter :: max_xpixels = 40,ypixels=20
    character(len=min(n,max_xpixels)) :: horizontal_line
    integer :: i, y_ind, x_ind, xpixels
    integer :: ylocmax
    real(rt), allocatable :: x_axis(:), y_axis(:) 
    logical, allocatable :: xy_filled(:,:) 
    integer, allocatable :: x_count(:) 
    
    allocate(x_axis(min(n,max_xpixels)),y_axis(1:ypixels),xy_filled(min(n,max_xpixels),ypixels),x_count(min(n,max_xpixels)-1))

    xpixels = min(n,max_xpixels)

    !Generate xaxis/bins
    x_axis = linspace(minval(x),maxval(x),xpixels)

    !Create a count for each bin
    ! This is a very slow method, but speed is not essential here yet. 
    x_count(1) = count(x<=x_axis(1))
    do i = 2,xpixels-2
        x_count(i) = count(x_axis(i)<=x .and. x<x_axis(i+1))
    end do
    x_count(xpixels-1) = count(x>=x_axis(xpixels))

    !Generate yaxis
    y_axis = linspace(0.e0_rt,real(maxval(x_count),rt),ypixels)

    !Fill in the plot
    xy_filled = .FALSE.
    do i = 1,size(x_count)
        ylocmax = int(real(x_count(i),8)/y_axis(ypixels)*real(ypixels,8))
        xy_filled(i,1:ylocmax) = .TRUE.
    end do

    !Draw plot to screen
    do y_ind = ypixels,1,-1
        do x_ind = 1,xpixels
            if (xy_filled(x_ind,y_ind)) then
                horizontal_line(x_ind:x_ind) = '.'
            else 
                horizontal_line(x_ind:x_ind) = ' '
            end if
        end do
        write(*,'(f10.5,A,A)') y_axis(y_ind),'|',horizontal_line
    end do
    do x_ind = 1,xpixels
        horizontal_line(x_ind:x_ind) = '-'
    end do
    write(*,*) '          ',horizontal_line
    write(*,*) 'x_range: ',x_axis(1),',',x_axis(xpixels)

    deallocate(x_axis,y_axis,xy_filled,x_count)

end subroutine

!Very basic plotting to std out
!Only have a limited number of pixels to work with    
subroutine plot_y_coarse(y)
    implicit none
    real(rt), dimension(1:), intent(in) :: y
    !local
    real(rt), dimension(1:size(y)) :: x
    x = real(colon(1,size(y)),rt)
    call plot_xy_coarse(x,y)
end subroutine plot_y_coarse
subroutine plot_xy_coarse(x,y,xpixels,ypixels,xlb,xub,ylb,yub,uid)
    implicit none
    real(rt), dimension(1:), intent(in) :: x,y
    real(rt), intent(in), optional :: xlb,xub,ylb,yub
    integer, intent(in), optional :: xpixels,ypixels,uid
    !local
    integer :: xp, yp, unitno
    real(rt), allocatable :: x_axis(:),y_axis(:)
    real(rt) :: xpt,ypt
    logical, allocatable :: xy_filled(:,:)
    logical :: lx,ly
    integer :: i, y_ind, x_ind
    character(len=10000) :: horizontal_line

    unitno = 6 ! stdout
    xp = 40
    yp = 20

    if (present(xpixels)) xp = xpixels 
    if (present(ypixels)) yp = ypixels 
    if (present(uid)) unitno = uid

    allocate(x_axis(xp))
    allocate(y_axis(yp))
    allocate(xy_filled(xp,yp))

    !Check input
    if (size(x)/=size(y)) STOP 'plot: wrong size'       

    !Generate axes
    x_axis(1)  = minval(x); if (present(xlb)) x_axis(1) = xlb
    x_axis(xp) = maxval(x); if (present(xub)) x_axis(xp) = xub
    y_axis(1)  = minval(y); if (present(ylb)) y_axis(1) = ylb
    y_axis(yp) = maxval(y); if (present(yub)) y_axis(yp) = yub

    x_axis = linspace(x_axis(1),x_axis(xp),xp)
    y_axis = linspace(y_axis(1),y_axis(yp),yp)

    !Fill plot
    xy_filled = .false.
    do i = 1,size(x)
        xpt = x(i)
        ypt = y(i)

        call bsearch(x_ind,x(i),x_axis)
        call bsearch(y_ind,y(i),y_axis)

        lx = .true.
        ly = .true.
        if (x(i)<x_axis(1)) lx = .false.
        if (y(i)<y_axis(1)) ly = .false.
        if (x(i)>x_axis(xp)) lx = .false.
        if (y(i)>y_axis(yp)) ly = .false.

        xy_filled(x_ind,y_ind) = lx .and. ly
                    
    end do

    !Draw plot to screen
    do y_ind = yp,1,-1
        horizontal_line = ' '
        do x_ind = 1,xp
            if (xy_filled(x_ind,y_ind)) horizontal_line(x_ind:x_ind) = 'o'
        end do
        write(unitno,'(f10.5,A,A)') y_axis(y_ind),'|',horizontal_line(1:xp)
    end do
    do x_ind = 1,xp
        horizontal_line(x_ind:x_ind) = '-'
    end do
    write(unitno,'(2A)') '          +',horizontal_line(1:xp)

    write(unitno,'(A11)',advance='no') ''
    do x_ind = 1,xp/10-1 ! divides the xp space into 10 because the space taken below is 10
        write(unitno,'(f10.5)',advance='no') x_axis(x_ind*10 - 10 + 1)
    end do
    write(unitno,'(f10.5)') x_axis(xp)

end subroutine plot_xy_coarse




subroutine plot_y_fine(y,title_,xlabel_,ylabel_)
    implicit none
    real(rt), intent(in) :: y(:)
    character(len=*), intent(in), optional :: title_,xlabel_,ylabel_

!     call clear()
!     if (present(title_)) call title(title_)
!     if (present(xlabel_)) call xlabel(xlabel_)
!     if (present(ylabel_)) call ylabel(ylabel_)
! 
!     call triplet(real(colon(1,size(y)),rt),y,'')
!     call drawnow()

end subroutine 
subroutine plot_y2_fine(y,title_,xlabel_,ylabel_)
    implicit none
    real(rt), intent(in) :: y(:,:)
    character(len=*), intent(in), optional :: title_,xlabel_,ylabel_

!     call clear()
!     if (present(title_)) call title(title_)
!     if (present(xlabel_)) call xlabel(xlabel_)
!     if (present(ylabel_)) call ylabel(ylabel_)
!    
!     call triplet(real(colon(1,size(y,1)),rt),y,'')
!     call drawnow()

end subroutine plot_y2_fine


subroutine plot_xy_fine(x,y,title_,xlabel_,ylabel_,fname_)
    implicit none
    real(rt), intent(in) :: x(:), y(:)
    character(len=*), intent(in), optional :: title_,xlabel_,ylabel_,fname_
    ! local
    integer :: i

    call clear()

    if (present(fname_)) then
        i = initplot(trim(fname_)) 
    else
        i = initplot('/tmp/mod_matlab_plot.eps')
    end if

    if (present(title_)) call title(i,title_)
    if (present(xlabel_)) call xlabel(i,xlabel_)
    if (present(ylabel_)) call ylabel(i,ylabel_)

    call triplet(x,y,'')
    call drawnow(i)

    call clear()

    if (present(fname_)) then
        ! if (.not. disable_plotting) print*,'Plot written to ',trim(fname_)
    else
        call execute_command_line('open /tmp/mod_matlab_plot.eps',wait=.true.)
    end if


end subroutine plot_xy_fine
subroutine plot_xy2_fine(x,y,title_,xlabel_,ylabel_,fname_)
    implicit none
    real(rt), intent(in) :: x(:), y(:,:)
    character(len=*), intent(in), optional :: title_,xlabel_,ylabel_,fname_
    ! local
    integer :: i

    call clear()

    if (present(fname_)) then
        i = initplot(trim(fname_)) 
    else
        i = initplot('/tmp/mod_matlab_plot.eps')
    end if

    if (present(title_)) call title(i,title_)
    if (present(xlabel_)) call xlabel(i,xlabel_)
    if (present(ylabel_)) call ylabel(i,ylabel_)

    call triplet(x,y,'')
    call drawnow(i)

    call clear()

    if (present(fname_)) then
        ! if (.not. disable_plotting) print*,'Plot written to ',trim(fname_)
    else
        call execute_command_line('open /tmp/mod_matlab_plot.eps',wait=.true.)
    end if


end subroutine plot_xy2_fine


subroutine ezplot(f,a,b)
    implicit none
    interface
        function f(x) result(y)
            use mod_types
            real(rt), intent(in) :: x
            real(rt) :: y
        end function
    end interface
    real(rt), intent(in) :: a,b
    ! local
    real(rt), allocatable :: x(:),y(:)
    integer :: i

    allocate(x(1000))
    allocate(y(size(x)))
    x = linspace(a,b,size(x))

    do i = 1,size(x)
        y(i) = f(x(i))
    end do

    call plot(x,y,'ezplot')

end subroutine 

! subroutine plot_y_fine(y,title)
!     implicit none
!     real(rt), intent(in) :: y(1:)
!     character(len=*), intent(in), optional :: title
! 
!     if (present(title)) call plot_xy_fine(real(colon(1,size(y)),8),y,title)
!     if (.not. present(title)) call plot_xy_fine(real(colon(1,size(y)),8),y)
! 
! end subroutine
! subroutine plot_xy_fine(x,y,title,xlabel,ylabel,extra)
!     implicit none
!     real(rt), intent(in) :: x(:), y(:)
!     character(len=*), intent(in), optional :: title,xlabel,ylabel,extra
!     ! local
!     character(len=1000) :: t,xl,yl,ex
!     logical :: useDefArg(4)
! 
!     useDefArg(1) = present(title)
!     useDefArg(2) = present(xlabel)
!     useDefArg(3) = present(ylabel)
!     useDefArg(4) = present(extra)
! 
!     t=' '; if (present(title)) t=title
!     xl=' '; if (present(xlabel)) xl=xlabel 
!     yl=' '; if (present(ylabel)) yl=ylabel
!     ex=' '; if (present(extra)) ex=extra
! 
!     call plot_xy2_fine(x,reshape(y,[size(y),1]),title,xlabel,ylabel,extra,useDefArg)
! 
! end subroutine plot_xy_fine
! subroutine plot_y2_fine(y,title,xlabel,ylabel,extra)
!     implicit none
!     real(rt), intent(in) :: y(:,:)
!     character(len=*), intent(in), optional :: title,xlabel,ylabel,extra
!     ! local
!     real(rt), allocatable :: x(:)
!     character(len=1000) :: t,xl,yl,ex
!     logical :: useDefArg(4)
! 
!     useDefArg(1) = present(title)
!     useDefArg(2) = present(xlabel)
!     useDefArg(3) = present(ylabel)
!     useDefArg(4) = present(extra)
! 
!     t=' '; if (present(title)) t=title
!     xl=' '; if (present(xlabel)) xl=xlabel 
!     yl=' '; if (present(ylabel)) yl=ylabel
!     ex=' '; if (present(extra)) ex=extra
! 
!     allocate(x(size(y,1)))
!     x = colon(1,size(y,1))
! 
!     call plot_xy2_fine(x,y,title,xlabel,ylabel,extra,useDefArg)
! 
! end subroutine plot_y2_fine 
! 
! ! plot_x_y2_fine plots x,y data using a (not compliant Fortran standard) system call to gnuplot.
! ! Many options are available and for the most part self-descriptive. Temporary files are saved 
! ! to /tmp/ (which of course is *nix) standard.
! !
! ! Arguments are mostly obvious. The unobvious ones are 
! ! extra 
! ! -> allows the user to pass any commands to gnuplot before the data is plotted.
! ! -> ex: "set style data scatter" will produce a scatter plot instead of the default line plot.
! ! 
! ! useDefArg 
! ! -> should not be called by the user (internal mechanism)
! ! 
! subroutine plot_xy2_fine(x,y,title,xlabel,ylabel,extra,useDefArg)
!     !$ use omp_lib ! conditional compilation, will only be examined by compiler if openmp is enabled
! 
!     implicit none
!     real(rt), intent(in) :: x(:), y(:,:)
!     character(len=*), intent(in), optional :: title,xlabel,ylabel,extra
!     logical, intent(in), optional :: useDefArg(4) ! Logical specifying which Default arguments
!     !local
!     integer :: i,j,ni,nj, unitno
!     logical :: isPresent_xlabel, isPresent_ylabel, isPresent_title, isPresent_extra
!     character(len=3), parameter :: newline = ', \'
!     character(len=3) :: tmpchar
!     character(len=5) :: jp1AsString,jAsString,tmpChar5
!     character(len=1000) :: dls(size(y,2)), ms(size(y,2)),cls(size(y,2)) ! data labels and methods
!     character(len=100) :: configfile, datafile 
!     integer(4) :: ierr
! 
! !     interface 
! !         function system(str) result(ierr)
! !             implicit none
! !             integer(4) :: ierr
! !             character(*) :: str
! !         end function 
! !     end interface
! 
!     ! Must ensure that only one thread executes this at a time otherwise they may modify 
!     ! the files at the same time. W/o the critical, this routine is not thread safe. Even 
!     ! with the critical, it is only thead safe if the system call to gnuplot is thread safe.
!     ! 
!     ! NOTE: I must not understand critical sections well enough, because I nested two and that
!     ! cause the program to hang. B/c of that, I am not using the critical section
! !         !$OMP CRITICAL
! 
!     ! As an extra precaution, I try to make sure the threads write to unique files
!     write(tmpChar5,'(i0)') 1
!     !$ write(tmpChar5,'(i0)') omp_get_thread_num() ! only compiled if openmp is enabled
!     datafile = '/tmp/greysdeleteme'//trim(tmpChar5)//'.dat'
!     configfile = '/tmp/greysdeleteme'//trim(tmpChar5)//'.cfg'
!     unitno = 726
!     !$ unitno = 726+omp_get_thread_num()
!     
! !         datafile = '/tmp/greysdeleteme.dat'
! !         configfile = '/tmp/greysdeleteme.cfg'
! !         unitno = 726
! 
!     if (present(useDefArg)) then
!         isPresent_title = useDefArg(1)
!         isPresent_xlabel = useDefArg(2)
!         isPresent_ylabel = useDefArg(3)
!         isPresent_extra = useDefArg(4)
!     else
!         isPresent_title = present(title) 
!         isPresent_xlabel = present(xlabel)
!         isPresent_ylabel = present(ylabel)
!         isPresent_extra = present(extra)
!     end if
! 
!     ! Check inputs
!     ni = size(x)
!     nj = size(y,2)
!     if (size(y,1)/=ni) STOP 'sub_plot_x_y2: x and y dims don''t agree'
! 
!     ! Set defaults and replace them if optional arg present
!     do j = 1,nJ
!         write(jAsString,'(i0)') j
!         dls(j) = 'data '//jAsString
!     end do
!     ms = 'line'
!     cls = ''
! 
!     ! Write data to a temporary file
!     open(unitno,file=trim(datafile),status='unknown')
!     write(unitno,*) '# Temporary data file created by grey''s plot routine' 
!     do i = 1,ni
!         write(unitno,'(1PE24.15E3)',advance='no') x(i)
!         do j = 1,nj
!             write(unitno,'(1PE24.15E3)',advance='no') y(i,j)
!         end do
!         write(unitno,*) 
!     end do
!     close(unitno)
! 
!     ! Write config to a temporary file
!     open(unitno,file=trim(configfile),status='unknown')
!     write(unitno,*) '# Temporary gnuplot configuration file created by grey''s plot routine'
!     write(unitno,*) 'set style data line' ! set line as default style 
!     if (isPresent_xlabel) write(unitno,*) 'set xlabel "'//trim(xlabel)//'"'
!     if (isPresent_ylabel) write(unitno,*) 'set ylabel "'//trim(ylabel)//'"'
!     if (isPresent_title) write(unitno,*) 'set title "'//trim(title)//'"'
!     if (isPresent_extra) write(unitno,*) trim(extra)//' ' ! extra pre options
! 
!     write(unitno,'(A)',advance='no') 'plot' ! this is the first part of the command
!     do j = 1,nj
!         
!         ! Create j+1 as a string
!         write(jp1AsString,'(i0)') j+1
! 
!         ! Determine whether new line is needed
!         if (j<nj) then
!             tmpchar = newline
!         else
!             tmpchar = '   '
!         end if
! 
!         ! Output the command
!         ! NOTE: must use advance no here to prevent ifort from breaking the line
! !         write(unitno,'(A)',advance='no') ' "'//trim(datafile)//'" using 1:' &
! !                 //trim(jp1AsString)//' title "'//trim(dls(j)) // &
! !               '" with '//trim(ms(j))//' '//trim(cls(j))//' '//trim(tmpchar)
!         write(unitno,'(A)',advance='no') ' "'//trim(datafile)//'" using 1:' &
!                 //trim(jp1AsString)//' title "'//trim(dls(j))//'" '//trim(tmpchar)
!         ! Make a new line
!         write(unitno,*)
! 
!     end do
! 
! !     if (isPresent_savefile) then 
! !         write(unitno,*) 'set terminal postscript'
! !         write(unitno,*) 'set output "'//trim(savefile)//'"'
! !         write(unitno,*) 'replot'
! !         write(unitno,*) 'set term x11' ! Not strictly necessary
! !     end if
! 
!     close(unitno)
! 
!     ! Now use a nonstandard fortran routine (appearing in ifort and gfortran) to call gnuplot
!     ! NOTE: if gnuplot is not available or an error is received, just report it and the program
!     ! will continue.
! !     ierr = system('gnuplot -persist '//trim(configfile)) ! was using until 5/14/14
!     call execute_command_line('gnuplot -persist '//trim(configfile),exitstat=ierr)
!     if (ierr/=0) print*,'sub_plot_x_y2: WARNING: error in calling gnuplot'
! 
! !         !$OMP END CRITICAL
! 
! end subroutine plot_xy2_fine

! Cumulatively sums the vector x
function cumsum_vec(x_in) result(x_out)
    implicit none
    real(rt), dimension(1:), intent(in) :: x_in
    real(rt), dimension(1:size(x_in)) :: x_out
    integer :: i
    real(rt) :: accum

    accum = 0e0_rt
    do i = 1,size(x_in)
        accum = accum + x_in(i)
        x_out(i) = accum
    end do

end function cumsum_vec
function cumsum_vec_int(x_in) result(x_out)
    implicit none
    integer, dimension(1:), intent(in) :: x_in
    integer, dimension(1:size(x_in)) :: x_out
    integer :: i,accum

    accum = 0
    do i = 1,size(x_in)
        accum = accum + x_in(i)
        x_out(i) = accum
    end do

end function cumsum_vec_int

! Cumulatively sums the rows of x
function cumsum_mat(x_in) result(x_out)
    implicit none
    real(rt), dimension(1:,1:), intent(in) :: x_in
    real(rt), dimension(1:size(x_in,1),1:size(x_in,2)) :: x_out  ! illegal to reference x_out
    ! local
    integer :: i,d 
    real(rt) :: accum

    do d = 1,size(x_in,2)

        accum = 0e0_rt
        do i = 1,size(x_in,1)
            accum = accum + x_in(i,d)
            x_out(i,d) = accum
        end do

    end do

end function cumsum_mat


subroutine sub_cumsum(x_out,x_in)
    implicit none
    real(rt), dimension(:), intent(in) :: x_in
    real(rt), dimension(1:size(x_in)), intent(inout) :: x_out
    integer :: i

    x_out(1)=x_in(1)
    do i=2,size(x_in)
        x_out(i) = x_in(i) + x_out(i-1)
    end do

end subroutine sub_cumsum
subroutine sub_cumsum_mat(x_out,x_in,DIM)
    implicit none
    real(rt), dimension(:,:), intent(in) :: x_in
    real(rt), dimension(1:size(x_in,1),1:size(x_in,2)), intent(inout) :: x_out
    integer, intent(in) :: DIM 
    integer :: col_ind, row_ind

    if (DIM==1) then
        do col_ind = 1,size(x_in,2)
            call sub_cumsum(x_out(:,col_ind),x_in(:,col_ind))
        end do
    elseif (DIM==2) then
        do row_ind = 1,size(x_in,1)
            call sub_cumsum(x_out(row_ind,:),x_in(row_ind,:))
        end do
    else
        STOP 'error in cumsum'
    end if

end subroutine sub_cumsum_mat

subroutine ndgrid_double2d(xx,yy,x,y)
    implicit none
    real(rt), intent(in) :: x(:), y(:)
    real(rt), intent(out) :: xx(:,:), yy(:,:)
    ! local
    integer :: ix, iy
    if (any(shape(xx)/=[size(x),size(y)])) write(0,*) 'ERROR: ndgrid: size xx does not conform with x,y'
    if (any(shape(yy)/=[size(x),size(y)])) write(0,*) 'ERROR: ndgrid: size yy does not conform with x,y'
    do ix = 1,size(x)
        do iy = 1,size(y)
            xx(ix,iy) = x(ix)
            yy(ix,iy) = y(iy)
        end do
    end do
end subroutine ndgrid_double2d

! Uses the same convention as the Matlab routines. ndgrid might be more natural.
subroutine meshgrid_int2d(xx,yy,x,y)
    implicit none
    integer, intent(in) :: x(1:), y(1:)
    integer, intent(out) :: xx(size(y),size(x)), yy(size(y),size(x))
    xx = spread(x,1,size(y))
    yy = spread(y,2,size(x))
end subroutine meshgrid_int2d
subroutine meshgrid_int3d(xxx,yyy,zzz,x,y,z)
    implicit none
    integer, intent(in) :: x(1:), y(1:), z(1:)
    integer, intent(out) :: xxx(size(y),size(x),size(z)), yyy(size(y),size(x),size(z)), zzz(size(y),size(x),size(z))
    xxx = spread(spread(x,1,size(y)),3,size(z))
    yyy = spread(spread(y,2,size(x)),3,size(z))
    zzz = spread(spread(z,1,size(y)),2,size(x))
end subroutine meshgrid_int3d
subroutine meshgrid_double2d(xx,yy,x,y)
    implicit none
    real(rt), intent(in) :: x(1:), y(1:)
    real(rt), intent(out) :: xx(size(y),size(x)), yy(size(y),size(x))
    xx = spread(x,1,size(y))
    yy = spread(y,2,size(x))
end subroutine meshgrid_double2d
subroutine meshgrid_double3d(xxx,yyy,zzz,x,y,z)
    implicit none
    real(rt), intent(in) :: x(1:), y(1:), z(1:)
    real(rt), intent(out) :: xxx(size(y),size(x),size(z)), yyy(size(y),size(x),size(z)), zzz(size(y),size(x),size(z))
    xxx = spread(spread(x,1,size(y)),3,size(z))
    yyy = spread(spread(y,2,size(x)),3,size(z))
    zzz = spread(spread(z,1,size(y)),2,size(x))
end subroutine meshgrid_double3d

! As in the matlab routine, fix rounds x to the nearest integer between x and 0.
! In fortran, this is just the "int" conversion routine.
elemental function fix(x)
    implicit none
    real(rt), intent(in) :: x
    integer :: fix

    fix = int(x)

!     fix = floor(x)
!     if (x<0.e0_rt) fix = fix + 1
end function fix

! Uniform random number draw between 0 and 1.
function scal_rand() result(btwn01)
    implicit none
    real(rt) :: btwn01
    call random_number(btwn01)
end function scal_rand
function vec_rand(n) result(btwn01)
    implicit none
    integer, intent(in) :: n
    real(rt), dimension(1:n) :: btwn01
    call random_number(btwn01)
end function vec_rand
function mat_rand(n1,n2) result(btwn01)
    implicit none
    integer, intent(in) :: n1,n2
    real(rt), dimension(1:n1,1:n2) :: btwn01
    call random_number(btwn01)
end function mat_rand

!Generates random normal variables using the Box-Muller method as described in Wikipedia
function scal_randn() result(RV)
    implicit none
    real(rt) :: RV
    !local
    real(rt) :: tmp(1)
    tmp = randn(1)
    RV = tmp(1)
end function 
function vec_randN(n) result(RVs)
    implicit none
    integer, intent(in) :: n
    real(rt), dimension(1:n) :: RVs
    !local
    real(rt) :: U(2),twopiU1,sqrtneg2logU2
    integer :: i
    
    do i = 1,n,2
        ! Fill U with two U[0,1]
        call random_number(U)
        do while (U(2)==0._rt)
            call random_number(U)
        end do

        
        ! Precompute
        twopiU1 = 2.e0_rt*pi*U(1)
        sqrtneg2logU2 = sqrt(-2.e0_rt*log(U(2)))
        
        ! Fill the random variables
        RVs(i) = cos(twopiU1)*sqrtneg2logU2
        if (i+1<=n) RVs(i+1) = sin(twopiU1)*sqrtneg2logU2

    end do

end function vec_randN
function mat_randN(n1,n2) result(RVs)
    implicit none
    integer, intent(in) :: n1,n2
    real(rt), dimension(n1,n2) :: RVs
    RVs = reshape(vec_randN(n1*n2),(/n1,n2/))
end function

! reset(.) set the seed of Fortran's random number generator.
! To reset to a default stream, do the following:   call reset(getDefaultStream())
subroutine reset(streamIn)
    implicit none
    type(stream) :: streamIn
    integer :: m
    call random_seed(size=m)
    if (m/=streamIn%m) STOP 'reset: stream size and seed size do not agree'

    call random_seed(put=streamIn%seeds)
end subroutine reset
! Sets a default stream in a Fortran standard compliant way.
function getDefaultStream() result(streamOut)
    implicit none
    type(stream) :: streamOut
    integer :: m

    if (.not. randomSeedInitialized) then
        call random_seed()
        randomSeedInitialized = .true.
    end if

    call random_seed(size=m)
    streamOut%m = m

    if (allocated(streamOut%seeds)) deallocate(streamOut%seeds)
    allocate(streamOut%seeds(m))

    streamOut%seeds = 5

end function getDefaultStream


!Vectorize (equivalent to matlab colon)
function vec_mat(X) result(vec)
    implicit none
    real(rt), dimension(1:,1:) :: X
    real(rt), dimension(1:size(X)) :: vec
    vec = reshape(X,(/size(X)/))
end function vec_mat
function vec_3d(X) result(vec)
    implicit none
    real(rt), dimension(1:,1:,1:) :: X
    real(rt), dimension(1:size(X)) :: vec
    vec = reshape(X,(/size(X)/))
end function vec_3d
function vec_4d(X) result(vec)
    implicit none
    real(rt), dimension(1:,1:,1:,1:) :: X
    real(rt), dimension(1:size(X)) :: vec
    vec = reshape(X,(/size(X)/))
end function vec_4d
function vec_mat_i(X) result(vec)
    implicit none
    integer, dimension(1:,1:) :: X
    integer, dimension(1:size(X)) :: vec
    vec = reshape(X,(/size(X)/))
end function vec_mat_i
function vec_3d_i(X) result(vec)
    implicit none
    integer, dimension(1:,1:,1:) :: X
    integer, dimension(1:size(X)) :: vec
    vec = reshape(X,(/size(X)/))
end function vec_3d_i
function vec_4d_i(X) result(vec)
    implicit none
    integer, dimension(1:,1:,1:,1:) :: X
    integer, dimension(1:size(X)) :: vec
    vec = reshape(X,(/size(X)/))
end function vec_4d_i

! Wraps blas dgemm with a simplified interface. 
! Has the options to transpose and can handle nD arrays. The penalty to nD arrays is
! two shape arguments must be passed.
! C = alpha*op(A)*op(B) + beta*C
subroutine mtimes_matmat(C,A,B,tA,tB,alpha,beta)
    implicit none
    real(rt), intent(out) :: C(:,:)
    real(rt), intent(in) :: A(:,:),B(:,:)
    character(len=1), intent(in), optional :: tA,tB
    real(rt), intent(in), optional :: alpha, beta
    ! local
    character(len=1) :: tAselect,tBselect
    real(rt) :: alphaselect,betaselect

    ! Set the most reasonable defaults and override if necessary
    tAselect = 'N'; if (present(tA)) tAselect = tA
    tBselect = 'N'; if (present(tB)) tBselect = tB
    alphaselect = 1e0_rt; if (present(alpha)) alphaselect = alpha
    betaselect = 0e0_rt; if (present(beta)) betaselect = beta
    
    call mygemm(C,A,B,size(A,1),size(A,2),size(B,1),size(B,2),tAselect,tBselect,alphaselect,betaselect)

end subroutine 

! C = alpha*op(A)*B + beta*C
subroutine mtimes_matvec(C,A,B,tA,alpha,beta)
    implicit none
    real(rt), intent(out) :: C(:)
    real(rt), intent(in) :: A(:,:),B(:)
    character(len=1), intent(in), optional :: tA
    real(rt), intent(in), optional :: alpha, beta
    ! local
    character(len=1) :: tAselect
    real(rt) :: alphaselect,betaselect

    ! Set the most reasonable defaults and override if necessary
    tAselect = 'N'; if (present(tA)) tAselect = tA
    alphaselect = 1e0_rt; if (present(alpha)) alphaselect = alpha
    betaselect = 0e0_rt; if (present(beta)) betaselect = beta
    
    call mygemv(C,A,B,size(A,1),size(A,2),tAselect,alphaselect,betaselect)

end subroutine 



subroutine mygemm(C,A,B,sA1,sA2,sB1,sB2,tA,tB,alpha,beta)
    implicit none
    integer, intent(in) :: sA1,sA2,sB1,sB2
    real(rt), intent(out) :: C(1,*) ! make a 2d array, should not matter, but needed to match the exact interface of gemm
    real(rt), intent(in) :: A(sA1,*),B(sB1,*)
    character(len=1), intent(in), optional :: tA,tB
    real(rt), intent(in), optional :: alpha, beta
    ! local
    character(len=1) :: tAselect,tBselect
    real(rt) :: alphaselect,betaselect
    integer :: m,n,k

    ! Set the most reasonable defaults and override if necessary
    tAselect = 'N'; if (present(tA)) tAselect = tA
    tBselect = 'N'; if (present(tB)) tBselect = tB
    alphaselect = 1e0_rt; if (present(alpha)) alphaselect = alpha
    betaselect = 0e0_rt; if (present(beta)) betaselect = beta

    ! Get the dimensions accounting for transposes
    m = merge(sA1,sA2,tAselect=='N') ! m is # rows of C and opA 
    n = merge(sB2,sB1,tBselect=='N') ! n is # columns of C and opB
    k = merge(sB1,sB2,tBselect=='N') ! k is # columns of opA and rows of opB
    
    ! Check for incompatible dimensions
    if (merge(sA2,sA1,tAselect=='N')/=k) STOP 'mygemm: dimensions are not compatible'

    ! Call BLAS
    call gemm(tAselect,tBselect,m,n,k,alphaselect,A,sA1,B,sB1,betaselect,C,m)

end subroutine 

! Wraps blas gemv with a simplified interface. 
! Has the options to transpose and can handle nD arrays. The penalty to nD arrays is
! two shape arguments must be passed.
! C = alpha*op(A)*B + beta*C
subroutine mygemv(C,A,B,sA1,sA2,tA,alpha,beta)
    implicit none
    integer, intent(in) :: sA1,sA2
    real(rt), intent(out) :: C(*)
    real(rt), intent(in) :: A(sA1,*),B(*)
    character(len=1), intent(in), optional :: tA
    real(rt), intent(in), optional :: alpha, beta
    ! local
    character(len=1) :: tAselect
    real(rt) :: alphaselect,betaselect
    integer :: m,n

    ! Set the most reasonable defaults and override if necessary
    tAselect = 'N'; if (present(tA)) tAselect = tA
    alphaselect = 1e0_rt; if (present(alpha)) alphaselect = alpha
    betaselect = 0e0_rt; if (present(beta)) betaselect = beta

    ! Get the dimensions accounting for transposes
    m = sA1
    n = sA2

    ! Call BLAS
    call gemv(tAselect,m,n,alphaselect,A,sA1,B,1,betaselect,C,1)

end subroutine 



! Construct a true matrix
function true_sqmat(a) result(true)
    integer, intent(in) :: a
    logical :: true(a,a)
    true = .TRUE.
end function true_sqmat
function true_mat(a,b) result(true)
    integer, intent(in) :: a,b
    logical :: true(a,b)
    true = .TRUE.
end function true_mat

! logspace(a,b,n) places n points from 10^a to 10^b according to linear spacing in log10. 
function logspace(d1,d2,n) result(grid)
    implicit none
    real(rt), intent(in) :: d1,d2
    integer, intent(in) :: n
    real(rt), dimension(1:n) :: grid

    grid = 10e0_rt**(linspace(d1,d2,n))

end function logspace

! Produces a matrix filled with ones according to the Matlab convention
function ones_sqmat(dim) result(mat)
    implicit none
    integer, intent(in) :: dim
    real(rt), dimension(1:dim,1:dim) :: mat    
    mat = 1.e0_rt
end function ones_sqmat
function ones_mat(dim1,dim2) result(mat)
    implicit none
    integer, intent(in) :: dim1,dim2
    real(rt), dimension(1:dim1,1:dim2) :: mat    
    mat = 1.e0_rt
end function ones_mat
function ones_vec2(dims) result(mat)
    implicit none
    integer, dimension(1:2),intent(in) :: dims
    real(rt), dimension(1:dims(1)) :: mat
    if (dims(2)/=1) STOP 'ones_vec2: use ones_mat2 for a matrix.  This only handles vectors'
    mat = 1.e0_rt
end function ones_vec2

! Computes the OLS regression of X on Y to obtain Y = X*B + eps with eps orthogonal
function regress(Y,X) result(B)
    implicit none
    real(rt), dimension(1:), intent(in) :: Y
    real(rt), dimension(1:,1:), intent(in) :: X
    real(rt), dimension(1:size(X,2)) :: B
    !local
    real(rt), allocatable :: xpx(:,:), xpy(:) 
    integer :: ny,nb

    ny = size(Y)
    nb = size(X,2)

    ! Check inputs
    if (ny/=size(X,1)) STOP 'regress: matrix dimensions dont agree: y/=size(x,1)'

    allocate(xpx(nb,nb))
    allocate(xpy(nb))

    ! Calc X'X and X'Y
    call mygemm(Xpx,X,X,ny,nb,ny,nb,tA='T')
    call mygemv(XpY,X,Y,ny,nb,tA='T')

    ! Check the condition of X'X to check for possibly multicollinearity
    if (cond(XpX)>1d10) then 
        write(*,*) 'regress: warning: possible multicollinearity in regressors'
        write(*,*) 'regress: X''X has condition number', cond(XpX)
    end if

    ! Solve (X'X)B=(X'Y) (i.e. B = (X'X)^-1(X'Y))
    B = mldivide(XpX,XpY)

end function regress

! Computes the condition number of the matrix X
function cond(X) result(k)
    
    implicit none
    real(rt), dimension(1:,1:), intent(in) :: X
    real(rt) :: k
    !local
    real(rt), dimension(1:size(X,1)) :: e 
    real(rt) :: mine

    ! Get the eigenvalues of X
    e = eig(X)

    ! 
    mine = minval(abs(e))
    if (mine/=0.e0_rt) then
        k = maxval(abs(e))/mine
    else 
        k = huge(0.e0_rt)
    endif
end function cond

! Computes the eigenvalues of a matrix.
! Follows the algorithm laid out in the Intel MKL help files.
function eig(A_in) result(e)
    implicit none
    real(rt), dimension(1:,1:), intent(in) :: A_in
    real(rt), dimension(1:size(A_in,1)) :: e
    !local
    real(rt), dimension(1:size(A_in,1),1:size(A_in,2)) :: A
    real(rt), dimension(1:size(A_in,1),1:size(A_in,2)) :: H,Z
    real(rt), dimension(1:size(A_in,1)) :: wr,wi
    integer, parameter :: blocksize = 64
    real(rt), dimension(1:size(A_in,1)*blocksize) :: work
    real(rt), dimension(1:size(A_in,1)-1) :: tau
    integer :: n, ilo, ihi, lda, lwork, info, ldh, ldz
    character(len=1) :: job,compz
    !Ensure the A is square
    if (size(A_in,1)/=size(A_in,2)) STOP 'A not square'

    !Compute the hessenberg form of the matrix A=Q*H*Q^{H}
    A = A_in
    n = size(A_in,1)
    ilo = 1
    ihi = n
    lda = n
    lwork = size(work)
    call dgehrd(n, ilo, ihi, a, lda, tau, work, lwork, info) !note A is overwritten with details of H and Q
    
    !Compute the eigenvalues of the hessenberg matrix
    job = 'E'
    compz = 'N'
    H = A
    ldh = size(H,1)    
    !Z is not referenced and so need not be set    
    ldz = size(Z,1)
    !Can use the same work arrays as before
    call hseqr(job, compz, n, ilo, ihi, h, ldh, wr, wi, z, ldz, work, lwork, info)
    !wr and wi contain the real and imaginary parts of the eigenvalues unless info>0

    select case (info)
        case (0)
            !Executed successfully
            !Right only the real part of the eigenvalues
            e = wr
            if (any(abs(wi)>1d-14)) then
                write(*,*) 'eig: warning: some eigenvalues are complex!'
                write(*,*) 'eig:          only writing the real parts'
            end if                
        case (:-1)
            write(*,*) 'eig: the ', info, ' parameter has an illegal value'
            STOP
            !The info-th parameter had an illegal value
        case (1:)
            write(*,*) 'eig: not all eigenvalues found'
            STOP 
            !Elements 1,2,...,ilo-1 and i+1,i+2,...,n of wr and wi contain the real and imag parts of 
            !those eigenvalues which have been successfully found.        
    end select

end function eig

! Computes the inverse of a matrix. 
! Currently, only works for 2x2 matrices. Otherwise, one should generally use mldivide.
function inv(X) result(Xinv)
    
    implicit none
    real(rt), dimension(1:,1:), intent(in) :: X
    real(rt), dimension(1:size(X,1),1:size(X,2)) :: Xinv
    !local
    real(rt) :: determ, a,b,c,d
    if (any(shape(X)/=(/2,2/))) STOP 'inv: only built for 2x2 matrices'
    a = X(1,1)
    b = X(1,2)
    c = X(2,1)
    d = X(2,2)
    determ = a*d-b*c
    Xinv = reshape((/d,-c,-b,a/),(/2,2/))/determ
end function inv


!Removes one singleton dimension from A
function squeeze_vec_i(A) result(B)
    implicit none
    integer, dimension(1:) :: A
    integer :: B

    if (size(A)>1) then
        STOP 'squeeze_vec: A is not a length 1 vector!'
    end if

    B = A(1)

end function squeeze_vec_i
function squeeze_vec(A) result(B)
    implicit none
    real(rt), dimension(1:) :: A
    real(rt) :: B

    if (size(A)>1) then
        STOP 'squeeze_vec: A is not a length 1 vector!'
    end if

    B = A(1)

end function squeeze_vec
function squeeze_mat(A) result(B)
    implicit none
    real(rt), dimension(1:,1:) :: A
    real(rt), dimension(1:size(A)) :: B

    if (.NOT.(any(shape(A)==1))) then
        write(*,*) 'squeeze_mat: No singleton dimension to remove'
        STOP
!     elseif (count(shape(A)>1)) then
!         write(*,*) 'squeeze_mat: More than one singleton dimension to remove'
!         STOP
    end if   

    if (size(A,1)==1) then
        B = A(1,:)
    elseif (size(A,2)==1) then
        B = A(:,1)
    else
        STOP 'squeeze_mat: case not found'
    end if
end function squeeze_mat

! Computes the mean of X
function mean_vec(X,mask) result(mean)
    implicit none
    logical, dimension(:), intent(in), optional :: mask
    real(rt), dimension(:), intent(in) :: X
    real(rt) :: mean
    ! local
    integer :: n

    if (present(mask)) then
        n = count(mask)
        if (n>0) then
            mean = sum(X,mask=mask)/real(n,rt)
        else
            mean = -1234.0_rt
        end if
    else
        mean = sum(X)/real(size(X),rt)
    end if
end function mean_vec
function mean_mat(X) result(mean)
    
    implicit none
    real(rt), dimension(1:,1:), intent(in) :: X
    real(rt), dimension(1:size(X,2)) :: mean
    mean = sum(X,1)/real(size(X,1),rt)
end function mean_mat

! X = mldivide(A,B)
! mldivide computes an "X" s.t. AX = B if possible. 
! I am not sure what happens when this is not possible.
function mldivide_full_mat(A,b) result(x)
    implicit none
    real(rt), dimension(1:,1:), intent(in) :: A
    real(rt), dimension(1:,1:), intent(in) :: B
    real(rt), dimension(1:size(A,2),1:size(b,2)) :: x
    !local
    real(rt), dimension(:,:), allocatable :: A_local, b_local
    integer, dimension(:), allocatable :: ipiv
    integer :: info

    allocate(A_local(size(A,1),size(A,2)), b_local(size(b,1),size(b,2)), ipiv(size(A,1)))
    
    !Check inputs
    if (size(A,1)/=size(A,2)) then
        write(*,*) shape(A)
        STOP 'mldivide: A not square.  Cant handle this currently.'
    end if

    ! Copy A and b over because these are overwritten
    A_local = A
    b_local = b
    call gesv(size(A_local,1), size(b_local,2), A_local, size(A_local,1), ipiv, b_local, size(b_local,1), info )
    x = b_local
    
    ! Check for failure
    select case (info)
        case(0)
        case (:-1)
            write(*,*) 'mldivide: the ', -info, ' parameter had an illegal value'
        case(1:)
            write(*,*) 'mldivide: U(i,i) for i=', info, ' is exactly zero.  The factorization ', &
                    'has been completed, but the factor U is exactly singular, so the solution ', &
                    'could not be computed.'
    end select

    deallocate(A_local, b_local, ipiv)
    
end function mldivide_full_mat

! x = mldivide(A,b)
! mldivide computes an "x" s.t. Ax = b if possible. 
! I am not sure what happens when this is not possible.
function mldivide_full(A,b) result(x)
    implicit none
    real(rt), dimension(1:,1:), intent(in) :: A
    real(rt), dimension(1:), intent(in) :: b
    real(rt), dimension(1:size(A,2)) :: x
    !local
    real(rt), dimension(:,:), allocatable :: A_local, b_local
    integer, dimension(:), allocatable :: ipiv
    integer :: info

    allocate(A_local(size(A,1),size(A,2)), b_local(size(b),1), ipiv(size(A,1)))
    
    !Check inputs
    if (size(b)/=size(A,1)) then
        write(*,*) size(b), size(A,1)
        STOP 'mldivide: b wrong size'
    end if
    if (size(A,1)/=size(A,2)) then
        write(*,*) shape(A)
        STOP 'mldivide: A not square.  Cant handle this currently.'
    end if

    ! Copy A and b over because these are overwritten
    A_local = A
    b_local(:,1) = b
    call gesv(size(A_local,1), 1, A_local, size(A_local,1), ipiv, b_local, size(b_local,1), info )
    x = b_local(:,1)
    
    ! Check for failure
    select case (info)
        case(0)
        case (:-1)
            write(*,*) 'mldivide: the ', -info, ' parameter had an illegal value'
        case(1:)
            write(*,*) 'mldivide: U(i,i) for i=', info, ' is exactly zero.  The factorization ', &
                    'has been completed, but the factor U is exactly singular, so the solution ', &
                    'could not be computed.'
    end select

    deallocate(A_local, b_local, ipiv)
    
end function mldivide_full

! mldivide_band
! Solves for x in Ax=b when A is a banded matrix with kl subdiagonals, ku superdiagonals, 
! with the one main diagonal (kl>=0,ku>=0). For user convenience, I am not economizing
! on storage here, so A is a full square matrix (the banded matrix is created internally)
!
function mldivide_band(A,b,kl,ku) result(x)
    implicit none
    real(rt), dimension(1:,1:), intent(in) :: A
    real(rt), dimension(1:), intent(in) :: b
    real(rt), dimension(1:size(A,2)) :: x
    integer, intent(in) :: kl,ku
    !local
    real(rt), dimension(:,:), allocatable :: Ab, btmp ! banded A, tmp b
    integer, dimension(:), allocatable :: ipiv
    integer :: m,n,ldab,ldb,info,nrhs, i,j

    !Check inputs
    if (size(b)/=size(A,1)) then
        write(*,*) size(b), size(A,1)
        STOP 'mldivide_band: b wrong size'
    end if
    if (size(A,1)/=size(A,2)) then
        write(*,*) shape(A)
        STOP 'mldivide_band: A not square.  Cant handle this currently.'
    end if
    if (ku<0 .or. kl<0) then
        STOP 'mldivide_band: ku and kl must be nonnegative'
    end if

    ! Begin
    m = size(A,1) ! Also is size(A,2)
    n = size(A,2) ! Also is size(A,1)
    ldab = 2*kl + ku + 1 ! The extra kl is used for storage
    ldb = n ! Rows in b
    nrhs = 1 ! Columns of b

    allocate(Ab(ldab,n), btmp(n,nrhs), ipiv(n))

    ! Convert A into banded form
    do j = 1,n
        do i = max(1,j-ku),min(m,j+kl) 
            ab(kl + ku + 1 + i - j, j) = A(i,j) ! storage scheme in lapack 
        end do
    end do

    ! Copy over b as it will be overwritten
    btmp(:,1) = b

    ! Compute the lu factorization
    call gbtrf( m, n, kl, ku, ab, ldab, ipiv, info )

    ! Check for failure of lu factorization
    select case (info)
        case(0)
        case (:-1)
            write(*,*) 'mldivide_band: in LU factorizing, the ', -info, ' parameter had an illegal value'
            stop
        case(1:)
            write(*,*) 'mldivide_band: in LU factorizing, U(i,i) for i=', info, ' is exactly zero.  The factorization ', &
                    'has been completed, but the factor U is exactly singular, so the solution ', &
                    'could not be computed. Setting result to 0 and returning. Results will likely be wrong.'
            x = 0d0
            return 
    end select

    ! Compute the solution using the factorized A
    call gbtrs( 'N', n, kl, ku, nrhs, ab, ldab, ipiv, btmp, ldb, info )

    ! Copy out the solution
    x = btmp(:,1)
    
    ! Check for failure
    select case (info)
        case(0)
        case (:-1)
            write(*,*) 'mldivide_band: in solving, the ', -info, ' parameter had an illegal value'
    end select

    deallocate(Ab, btmp, ipiv)
    
end function mldivide_band

! normcdf(x,my,sigma) evaluates the normal cdf.
! x is pt to be eval
! mu is mean
! sigma is stdev
elemental function normcdf(x,mu,sigma) result(F)
    implicit none
    real(rt), intent(in) :: x,mu,sigma
    real(rt) :: F
    F=.5e0_rt*(1.e0_rt+erf((x-mu)/(sigma*sqrt_of_2)))
end function normcdf

! Standard normal cdf
elemental function stdnormcdf(x) result(F)
    implicit none
    real(rt), intent(in) :: x
    real(rt) :: F
    F=.5e0_rt*(1e0_rt+erf(x/sqrt_of_2))
end function 

! Normal density
elemental function normpdf(x,mu,sigma) result(f)
    implicit none
    real(rt), intent(in) :: x,mu,sigma
    real(rt) :: f
    !local

    f = exp(-(x-mu)**2/(2e0_rt*sigma**2))/(sigma*sqrt_of_2pi)

end function 

! Inverse of standard normal cdf
function stdnormcdfinv(F) result(x)
    use mod_types
    implicit none
    real(rt), intent(in) :: F
    real(rt) :: x
    interface 
        function r8_normal_01_cdf_inverse(F)
            use mod_types
            implicit none
            real(rt) :: F
            real(rt) :: r8_normal_01_cdf_inverse
        end function
    end interface

    x = r8_normal_01_cdf_inverse(F)

end function 
function normcdfinv(F,mu,sigma) result(x)
    use mod_types
    implicit none
    real(rt), intent(in) :: F,mu,sigma
    real(rt) :: x
    interface 
        function r8_normal_01_cdf_inverse(F)
            use mod_types
            implicit none
            real(rt) :: F
            real(rt) :: r8_normal_01_cdf_inverse
        end function
    end interface

    x = mu + sigma*r8_normal_01_cdf_inverse(F)

end function 

! Standard normal density
elemental function stdnormpdf(x) result(f)
    implicit none
    real(rt), intent(in) :: x
    real(rt) :: f
    !local

    f = exp(-x**2/2e0_rt)/sqrt_of_2pi

end function 



! A driver for the dgesvd LAPACK routine. 
! svd(U,S,V,A) produces matrices such that A = U*S*V' where U and V are
! "unitary" matrices and S is diagonal.
subroutine svd(U,S,V,A_in)
    implicit none
    real(rt), dimension(1:,1:), intent(in) :: A_in
    real(rt), dimension(1:size(A_in,1),1:size(A_in,2)), intent(out) :: S    
    real(rt), dimension(1:size(A_in,1),1:size(A_in,1)), intent(out) :: U
    real(rt), dimension(1:size(A_in,2),1:size(A_in,2)), intent(out) :: V
    !local
    real(rt), dimension(1:size(A_in,1),1:size(A_in,2)) :: A
    real(rt), dimension(1:size(A_in,2),1:size(A_in,2)) :: Vt
    real(rt), dimension(1:min(size(A_in,1),size(A_in,2))) :: s_values
    integer :: info
    character(len=1) :: jobu, jobvt
    integer :: m,n,lda,lwork,ldu,ldvt
    real(rt), dimension(:), allocatable :: work

    !Fortran 77 interface (Lapack)
    jobu = 'A'
    jobvt = 'A'
    m = size(A_in,1)
    n = size(A_in,2)
    lda = max(1,m)
    ldu = max(1,m)
    ldvt = max(1,n)
    lwork = 64 * max(3*min(m, n)+max(m, n), 5*min(m,n)) 
    
    allocate(work(1:max(1,lwork)))
    
    A = A_in
    call gesvd(jobu, jobvt, m, n, A, lda, s_values, U, ldu, Vt, ldvt, work, lwork, info) !A = U*S*Vt
    S = diag(s_values)
    V = transpose(Vt)
    
    deallocate(work)

    select case (info)
        case(0)
        case (:-1)
            write(*,*) 'svd: in reduction to bidiagonal, the ', -info, ' parameter had an illegal value'
        case (1:)
            write(*,*) 'svd: WARNING: ', info, ' of the superdiagonals failed to converge.'
    end select    
   
end subroutine svd


! x = nullspace(A,option)
! Computes the nullspace of A (i.e. x s.t. Ax = 0) using singular value decomposition.
! I have an optional argument "option" that normalizes x so it sums to one. Obviously,
! A had better not have full rank (<=> x=0) if this is selected!
! 
function nullspace(A,option) result(x_out)
    implicit none
    real(rt), dimension(1:,1:), intent(in) :: A
    real(rt), dimension(1:size(A,2)) :: x_out
    character(len=*), intent(in), optional :: option
    !local
    real(rt), dimension(1:size(A,2)) :: x
    real(rt), dimension(1:size(A,1),1:size(A,2)) :: S
    real(rt), dimension(1:size(A,1),1:size(A,1)) :: U
    real(rt), dimension(1:size(A,2),1:size(A,2)) :: V    
    real(rt) :: zero_toler    
    integer :: j
    logical, dimension(1:size(A,2)) :: is_singularcol
    
    call svd(U,S,V,A) !A = U*S*V'
       
    ! zero_toler = max(size(A,1),size(A,2))*maxval(S)*epsilon(0.e0_rt)
    ! The above was recommended, but I find a slightly higher tolerance is better
    zero_toler = 10e0_rt*max(size(A,1),size(A,2))*maxval(S)*epsilon(0.e0_rt)

    if (count(diag(S)<zero_toler)/=1) then
        print*,'null: nullity of A does not equal one.  '//&
               'This is okay, ... but not for transition matrices which this function was designed for.  '//&
               'Slight modifications are necessary if this is not what you want.'
        print*,'diag(S) should have exactly one element less than zero_toler'
        print*,'zero_toler is', zero_toler
        print*,'diag(S) is ', diag(S)
        print*,' '
        call disp('Matrix given:',A)
        call disp('If did null(eye - transpose(pi)), then pi was',transpose(eye(size(A,1))-A))
        call disp('The summation of pi across rows is',sum(transpose(eye(size(A,1))-A),dim=2))
        print*,'The above''s difference from 1 is ',1e0_rt-sum(transpose(eye(size(A,1))-A),dim=2)
        print*,' ' 
        print*,'If you believe you are using a transition matrix, it is likely that '
        print*,'1) the rows don''t sum to one'
        print*,'2) there are multiple ergodic sets instead of a unique one (indeterminacy)'
        print*,'3) the rows almost sum to one but not precisely'
        STOP
    end if

    is_singularcol = diag(S)<zero_toler
    do j=1,size(V,2)
        if (is_singularcol(j)) x = V(:,j)
    end do

    if (present(option)) then
        if (option=='normalize') then
            x = x/sum(x)
        else 
            write(*,*) 'did you mean normalize? may experience unexpected results'
        end if
    end if

    x_out = x
    
end function nullspace

! Finds all the indices of all the true elements of X.
subroutine find_x_alloc(I,x)
    implicit none
    integer, intent(out), allocatable :: I(:)
    logical, intent(in) :: x(1:)

    allocate(I(count(x)))

    I = pack(colon(1,size(x)),x)

end subroutine find_x_alloc

pure function find_x(X) result(I)
    implicit none
    logical, dimension(1:), intent(in) :: X
    integer, dimension(1:count(X)) :: I
    !local
    integer, dimension(1:size(X)) :: index_vec
    index_vec = colon(1,size(X))
    I = pack(index_vec,X)
end function find_x
! Returns a length k vector. Returns the indices of the first k
! true elements of X. If k true elements of X don't exist, a fatal
! error is called.
function find_xk(X,k) result(I)
    implicit none
    logical, dimension(1:), intent(in) :: X
    integer, intent(in) :: k
    integer, dimension(1:k) :: I !min(count(X),k)
    !local
    integer :: iind,xind
    
    iind = 1
    do xind = 1,size(X)
        if (X(xind)) then
            I(iind) = xind
            if (iind>=k) return
            iind = iind+1            
        end if
    end do
    STOP 'find_xk: ERROR: k true values did not exist'
    I(iind:k) = -1

end function find_xk
! Same as above, but either returns the first k elements or the last k depending on the 
! value of string \in {"first","last"}
function find_xks(X,k,string) result(I)
    implicit none
    logical, dimension(1:), intent(in) :: X
    integer, intent(in) :: k
    integer, dimension(1:k) :: I
    character(len=*), intent(in) :: string
    !local
    integer :: iind,xind

    if (string=='last') then
        iind = 1
        do xind = size(X),1,-1
            if (X(xind)) then
                I(iind) = xind
                if (iind>=k) return
                iind = iind+1            
            end if
        end do
        STOP 'find_xks: WARNING: k true values did not exist'
        I(iind:k) = -1
    elseif (string=='first') then
        iind = 1
        do xind = 1,size(X)
            if (X(xind)) then
                I(iind) = xind
                if (iind>=k) return
                iind = iind+1            
            end if
        end do
        STOP 'find_xks: WARNING: k true values did not exist'
        I(iind:k) = -1
    else 
        STOP 'find_xks: case not found'
    end if

end function find_xks

function find1_xs(X,string) result(I)
    implicit none
    logical, dimension(1:), intent(in) :: X
    integer :: I
    character(len=*), intent(in) :: string
    ! local
    integer :: Ivec(1)

    Ivec = find_xks(X,1,string)
    I = Ivec(1)

end function

! Computes the sample standard deviation estimate of X. The estimate is consistent
! in that it is normalized by sqrt(N-1).
function std_mat(X) result(Y)
    implicit none
    real(rt), dimension(1:,1:), intent(in) :: X
    real(rt), dimension(1:size(X,2)) :: Y
    !local
    integer :: N, j
    
    !Check dimensions
    if (size(X,1)<=1) STOP 'std: X must have at least N>1 rows'
    
    N = size(X,1)
    do j = 1,size(X,2)
        Y(j) = std_vec(X(:,j))
    end do
end function std_mat
function std_vec(x,mask) result(y)
    implicit none
    real(rt), dimension(:), intent(in) :: x
    logical, dimension(:), intent(in), optional :: mask
    real(rt) :: y

    if (present(mask)) then
        y = std_core(x,size(x),mask)
    else
        y = std_core(x,size(x))
    end if

end function std_vec

! std_core
function std_core(x,n,mask) result(s)
    implicit none
    real(rt), intent(in) :: x(*)
    integer, intent(in) :: n
    logical, intent(in), optional :: mask(*)
    real(rt) :: s
    ! local
    real(rt) :: Ex,Ex2
    integer :: i,m

    Ex = 0e0_rt
    Ex2 = 0e0_rt
    m = 0
    if (present(mask)) then
        do i = 1,n
            if (mask(i)) then
                Ex = Ex + x(i)
                Ex2 = Ex2 + x(i)**2
                m = m + 1
            end if
        end do
    else
        m = n
        do i = 1,n
            Ex = Ex + x(i)
            Ex2 = Ex2 + x(i)**2
        end do
    end if

    Ex = Ex/m

    s = sqrt(max(0._rt,Ex2/m - Ex**2)) ! in case the std is actually 0, bound at zero so don't take sqrt of -eps

end function 



! corr_core
function corr_core(x,y,n,mask) result(c)
    implicit none
    real(rt), intent(in) :: x(*),y(*)
    integer, intent(in) :: n
    logical, intent(in), optional :: mask(*)
    real(rt) :: c
    ! local
    real(rt) :: Ex,Ey,Ex2,Ey2,Exy,sdx,sdy
    integer :: i,m

    Ex = 0e0_rt
    Ex2 = 0e0_rt
    Ey = 0e0_rt
    Ey2 = 0e0_rt
    Exy = 0e0_rt
    m = 0
    if (present(mask)) then
        do i = 1,n
            if (mask(i)) then
                Ex = Ex + x(i)
                Ey = Ey + y(i)
                Ex2 = Ex2 + x(i)**2
                Ey2 = Ey2 + y(i)**2
                Exy = Exy + x(i)*y(i)
                m = m + 1
            end if
        end do
    else
        m = n
        do i = 1,n
            Ex = Ex + x(i)
            Ey = Ey + y(i)
            Ex2 = Ex2 + x(i)**2
            Ey2 = Ey2 + y(i)**2
            Exy = Exy + x(i)*y(i)
        end do
    end if

    sdx = sqrt(max(0._rt,m*Ex2 - Ex**2)) ! in case numerical error induces = -epsilon, bound at 0
    sdy = sqrt(max(0._rt,m*Ey2 - Ey**2))

    if (sdx>0e0_rt .and. sdy>0e0_rt) then
        c = (m*Exy-Ex*Ey)/sdx/sdy
    else
        c = -sqrt(huge(0._rt))
    end if

end function 

! corr(X,Y) with X,Y vectors, then corr is the unbiased small sample
! correlation estimate.
! See the matlab documentation for the complicated behavior of corr.
function corr_vecs(X,Y) result(rho)
    implicit none
    real(rt), dimension(:), intent(in) :: X, Y
    real(rt) :: rho

    rho = corr_core(x,y,size(x))

!     !local
!     integer :: N
!     
!     !Check dimensions
!     if (size(X,1)/=size(Y,1)) STOP 'corr_vecs: X must have at least N>1 rows'    
!     N = size(X,1)
!     
!     rho = sum((X-mean(X))*(Y-mean(Y)))/(real(N-1,rt)*std(X)*std(Y))
    
end function corr_vecs
function corr_mat(X) result(rho)
    implicit none
    real(rt), dimension(1:,1:), intent(in) :: X
    real(rt), dimension(1:size(X,2),1:size(X,2)) :: rho
    !local
    integer :: N, j1,j2
    
    !Check dimensions
    if (size(X,1)<=1) STOP 'corr: X must have at least N>1 rows'    
    N = size(X,1)
    
    do j1 = 1,size(X,2)
        do j2 = 1,size(X,2)
            rho(j1,j2) = corr_vecs(X(:,j1),X(:,j2))
        end do
    end do
    
end function corr_mat
function corr_mats(X1,X2) result(rho)
    
    implicit none
    real(rt), dimension(1:,1:), intent(in) :: X1,X2
    real(rt), dimension(1:size(X1,2),1:size(X2,2)) :: rho
    !local
    integer :: N, j1,j2
    
    !Check dimensions
    if (size(X1,1)<=1) STOP 'corr: X must have at least N>1 rows'    
    if (size(X1,1)/=size(X2,1)) STOP 'corr: X1 and X2 must have the same number of rows'    

    N = size(X1,1)
    
    do j1 = 1,size(X1,2)
        do j2 = 1,size(X2,2)
            rho(j1,j2) = corr_vecs(X1(:,j1),X2(:,j2))
        end do
    end do
    
end function corr_mats
function corr_matvec(X1,X2) result(rho)
    implicit none
    real(rt), dimension(1:,1:), intent(in) :: X1
    real(rt), dimension(1:), intent(in) :: X2
    real(rt), dimension(1:size(X1,2)) :: rho
    rho = (/corr_mats(X1,reshape(X2,(/size(X2),1/)))/)
end function corr_matvec
function corr_vecmat(X1,X2) result(rho)
    
    implicit none
    real(rt), dimension(1:), intent(in) :: X1
    real(rt), dimension(1:,1:), intent(in) :: X2
    real(rt), dimension(1:size(X2,2)) :: rho
    rho = (/corr_mats(reshape(X1,(/size(X1),1/)),X2)/)
end function corr_vecmat

! diag(X) for X a matrix produces the diagonal elements of X
! diag(X) for X a vector produces a diagonal matrix with X along the main 
! diagonal and 0 elsewhere.
function diag_mat(X) result(V)
    implicit none
    real(rt), dimension(1:,1:), intent(in) :: X
    real(rt), dimension(1:min(size(X,1),size(X,2))) :: V
    !local
    integer :: i
    
    if (size(X,1)/=size(X,2)) write(*,*) 'diag_mat: WARNING: X is not square.'
    
    do i = 1,min(size(X,1),size(X,2))
        V(i) = X(i,i)
    end do

end function diag_mat
function diag_mat_k(X,k) result(V)
    implicit none
    real(rt), dimension(1:,1:), intent(in) :: X
    integer, intent(in) :: k
    real(rt), dimension(1:size(X,1)-abs(k)) :: V
    !local
    integer :: i
    
    if (size(X,1)/=size(X,2)) STOP 'diag_mat_k: ERROR: X is not square. This routine needs to be changed to handle that case.'
    if (k>size(X,1)) STOP 'diag_mat_k: ERROR: k exceeds dim of X'
    
    if (k>=0) then
        do i = 1,size(X,1)-abs(k)
            V(i) = X(i,i+k)
        end do
    else
        do i = 1,size(X,1)-abs(k)
            V(i) = X(i-k,i)
        end do
    end if

end function diag_mat_k
pure function diag_vec(V) result(X)
    
    implicit none
    real(rt), dimension(1:), intent(in) :: V
    real(rt), dimension(1:size(V),1:size(V)) :: X
    !local
    integer :: i      
    X = 0.e0_rt
    forall (i = 1:size(V))
        X(i,i) = V(i)
    end forall
end function diag_vec

! Differences x. If x is a matrix, then this is done for each column.
function diff_vec(x) result(d)
    
    implicit none
    real(rt), dimension(1:), intent(in) :: x
    real(rt), dimension(1:size(x,1)-1) :: d
    !local
    integer :: n
    
    n = size(x,1)    
    d = x(2:n)-x(1:n-1)        
    
end function diff_vec

function diff_mat(x) result(d)
    
    implicit none
    real(rt), dimension(1:,1:), intent(in) :: x
    real(rt), dimension(1:size(x,1)-1,1:size(x,2)) :: d
    !local
    integer :: n
    
    n = size(x,1)    
    d = x(2:n,:)-x(1:n-1,:)
    
end function diff_mat

! Sorts x into y. Returns i s.t. x(i) = y. If dim is selected, then 
! sorts along that dimension. If mode is selected, then either sorts
! ascending or descending according to "ascend" / "descend"
subroutine sub_sort_vec_i(y,i,x,dim,mode)
    implicit none
    integer, dimension(1:), intent(in) :: x
    integer, dimension(1:size(x)), intent(inout) :: y
    integer, dimension(1:size(x)), intent(out) :: i
    integer, intent(in), optional :: dim
    character(len=*), intent(in), optional :: mode
    !local
    ! integer, dimension(1:size(x)) :: ylocal
    character(len=20) :: mode_eff

    ! Just here so I don't get bothered about unused variables
    if (present(dim)) then
    end if
        
    mode_eff = 'ascend'
    if (present(mode)) mode_eff = mode

    if (mode_eff=='ascend') then
        y = x
        call defsort(y,i)
    elseif (mode_eff=='descend') then
        y = -x
        call defsort(y,i)
        y = -y
    else
        write(*,*) 'sort: mode not found. stopping'
        STOP
    end if
   
end subroutine sub_sort_vec_i
subroutine sub_sort_vec(y,i,x,dim,mode)
    use mod_sort
    implicit none
    real(rt), dimension(1:), intent(in) :: x
    real(rt), dimension(1:size(x)), intent(inout) :: y
    integer, dimension(1:size(x)), intent(out) :: i
    integer, intent(in), optional :: dim
    character(len=*), intent(in), optional :: mode
    !local
    character(len=20) :: mode_eff

    ! Just here so I don't get bothered about unused variables
    if (present(dim)) then
    end if
        
    mode_eff = 'ascend'
    if (present(mode)) mode_eff = mode

    if (mode_eff=='ascend') then
        y = x
        call defsort(y,i)
    elseif (mode_eff=='descend') then
        y = -x
        call defsort(y,i)
        y = -y
    else
        write(*,*) 'sort: mode not found. stopping'
        STOP
    end if
   
end subroutine sub_sort_vec
subroutine sub_sort_mat(y,i,x,dim,mode)
    implicit none
    real(rt), dimension(1:,1:), intent(in) :: x
    real(rt), dimension(1:size(x,1),1:size(x,2)), intent(inout) :: y
    integer, dimension(1:size(x,1),1:size(x,2)), intent(out) :: i
    integer, intent(in), optional :: dim
    character(len=*), intent(in), optional :: mode
    !local
    integer :: ind    
        
    if (present(dim)) then        
        if (dim==2) then
            !sort each row
            do ind = 1,size(x,1)
                if (present(mode)) then
                    call sub_sort_vec(y(ind,:),i(ind,:),x(ind,:),1,mode)
                else
                    call sub_sort_vec(y(ind,:),i(ind,:),x(ind,:))
                end if            
            end do
            return        
        end if
    end if

    !sort each column
    do ind = 1,size(x,2)
        if (present(mode)) then
            call sub_sort_vec(y(:,ind),i(:,ind),x(:,ind),1,mode)
        else
            call sub_sort_vec(y(:,ind),i(:,ind),x(:,ind),1)
        end if
    end do
    
end subroutine sub_sort_mat
function fn_sort_vec(x,dim,mode) result(y)
    
    implicit none
    real(rt), dimension(1:), intent(in) :: x
    real(rt), dimension(1:size(x)) :: y
    integer, intent(in), optional :: dim
    character(len=*), intent(in), optional :: mode
    !local
    integer, dimension(1:size(x)):: i
    integer :: mycase
    
    mycase = merge(1,0,present(dim))+merge(2,0,present(mode))
       
    select case (mycase)
        case (0)
            call sub_sort_vec(y,i,x)
        case (1)
            call sub_sort_vec(y,i,x,dim=dim)
        case (2) 
            call sub_sort_vec(y,i,x,mode=mode)
        case (3)
            call sub_sort_vec(y,i,x,dim,mode)
    end select
    
end function fn_sort_vec


! interp1qs (s for special)
! Linearly interpolate along dimension d of some arbitrarily-dimensioned array y
! To handle arbitrary dimensions, the shape of y must be passed as an argument, as
! well as the desired dimension for interpolation.
! The shape of yi is inferred from the shape of y and the length of xi.
subroutine interp1qs(yi,x,y,xi,d,shapey)
    implicit none
    real(rt), intent(in) :: x(:),xi(:)
    real(rt), intent(in) :: y(*)
    real(rt), intent(out) :: yi(*)
    integer, intent(in) :: d,shapey(:)
    ! 
    integer :: nx,nxi,nlead,nlag

    if (shapey(d)/=size(x)) STOP 'ERROR: interp1qs: shape of y, d, and x do not agree'  

    nx = size(x)
    nxi = size(xi)

    ! To do the interpolation, it is most convenient to reshape y and yi into thre
    ! The first index is dimensions 1,...,d-1 stacked
    ! The second index is dimension d (the interpolation dimension)
    ! The third index is dimensions d+1,...,n stacked
    if (d>1) then
        nlead = product(shapey(1:d-1))
    else
        nlead = 1
    end if

    if (d<size(shapey)) then
        nlag = product(shapey(d+1:size(shapey)))
    else
        nlag = 1
    end if

    ! Call the core interpolation routine
    call interp1qs_core(yi,y,nx,nxi,nlead,nlag,x,xi)
    

end subroutine

    subroutine interp1qs_core(yi,y,nx,nxi,nlead,nlag,x,xi)
        implicit none
        real(rt), intent(in) :: x(:),xi(*)
        integer, intent(in) :: nx,nxi,nlead,nlag
        real(rt), intent(in) :: y(nlead,nx,*)
        real(rt), intent(out) :: yi(nlead,nxi,*)
        integer :: ixi,ilo
        real(rt) :: mlo

        ! When passed y and yi, can access naturally using a three dimensional array
        do ixi = 1,nxi

            ! Get the index ilo s.t. xi(ixi) \in [x(ilo),x(ilo+1)]
            call bsearch(ilo,xi(ixi),x)

            ! Get the corresponding weight
            mlo = (x(ilo+1)-xi(ixi))/(x(ilo+1)-x(ilo))

            ! Do the interpolation step
            yi(1:nlead,ixi,1:nlag) =      mlo *y(1:nlead,ilo  ,1:nlag) + &
                                     (1e0_rt-mlo)*y(1:nlead,ilo+1,1:nlag)

        end do

    end subroutine

! Get interp1q location and weights
subroutine interp1qlocweight(ilo,mlo,xi,x)
    implicit none
    real(rt), intent(in) :: xi,x(:)
    integer, intent(out) :: ilo
    real(rt), intent(out) :: mlo

    if (size(x)<=1) then
        write(0,*) 'WARNING: interp1qlocweight: size(x)<=1, returning ilo==1'
        ilo = 1
        return
    end if
    
    call bsearch(ilo,xi,x)
    mlo = (x(ilo+1)-xi)/(x(ilo+1)-x(ilo))

end subroutine
subroutine noexinterp1qlocweight(ilo,mlo,xi,x)
    implicit none
    real(rt), intent(in) :: xi,x(:)
    integer, intent(out) :: ilo
    real(rt), intent(out) :: mlo

    if (size(x)<=1) then
        write(0,*) 'WARNING: noexinterp1qlocweight: size(x)<=1, returning ilo==1'
        ilo = 1
        return
    end if
    
    call bsearch(ilo,xi,x)
    mlo = (x(ilo+1)-xi)/(x(ilo+1)-x(ilo))
    mlo = min(max(0e0_rt,mlo),1e0_rt) 

end subroutine


! interp1q
! Constructs and evaluates the linear interpolant of (x,y), evaluating at xi.
! Assumes x is sorted AND xi are sorted
function interp1q_scal(x,y,xi) result(yi)
    use mod_interp, only: search
    implicit none
    real(rt), dimension(:), intent(in) :: x,y
    real(rt), intent(in) :: xi
    real(rt) :: yi
    ! local 
    integer :: ilo 

    if (size(x)==size(y) .and. size(x)>1) then
        ! Find the location of xi in the x grid assuming it is sorted
        call search(ilo,xi,x)

        ! Do the interpolation
        yi = y(ilo) + (y(ilo+1)-y(ilo))*(xi-x(ilo))/(x(ilo+1)-x(ilo))
    elseif (size(x)==size(y) .and. size(x)==1) then
        yi = y(1) 
        return
    else
        write(0,*) 'interp1_scal: input: sizes wrong: #x ',size(x),' #y',size(y)
        stop 
    end if

end function interp1q_scal
function interp1q_vec(x,y,xi) result(yi)
    use mod_interp, only: search
    implicit none
    real(rt), dimension(:), intent(in) :: x
    real(rt), dimension(:), intent(in) :: y
    real(rt), dimension(:), intent(in) :: xi
    real(rt), dimension(size(xi)):: yi
    ! local 
    integer, allocatable :: ilo(:)
    real(rt) :: mlo
    integer :: i

    allocate(ilo(size(xi)))

    call bsearch_many(ilo,xi,x)

    do i = 1,size(xi)
        mlo = (x(ilo(i)+1)-xi(i))/(x(ilo(i)+1)-x(ilo(i)))
        yi(i) = y(ilo(i))*mlo + y(ilo(i)+1)*(1._rt-mlo)
    end do

!     do i = 1,size(xi)
!         yi(i) = interp1q_scal(x,y,xi(i))
!     end do

end function interp1q_vec
function interp1q_scal_ymat(x,y,xi) result(yi)
    use mod_interp, only: search
    implicit none
    real(rt), dimension(:), intent(in) :: x
    real(rt), dimension(:,:), intent(in) :: y
    real(rt), intent(in) :: xi
    real(rt), dimension(size(y,2)) :: yi
    ! local 
    integer :: nx, ny
    integer :: ilo 

    nx = size(x)
    ny = size(y,1)
    if (nx/=ny) STOP 'interp1_scal: dims dont agree'
    if (nx==1) then
        yi = y(1,:) 
        return
    end if

    ! Find the location of xi in the x grid assuming it is sorted
    call search(ilo,xi,x)

    ! Do the interpolation
    yi(:) = y(ilo,:) + (y(ilo+1,:)-y(ilo,:))*(xi-x(ilo))/(x(ilo+1)-x(ilo))

end function interp1q_scal_ymat

! Additional argument, transpose, determines whether to interpolate y along the first dimension (default) or the second
function interp1q_vec_ymat(x,y,xi) result(yi)
    use mod_interp, only: search
    implicit none
    real(rt), dimension(:), intent(in) :: x
    real(rt), dimension(:,:), intent(in) :: y
    real(rt), dimension(:), intent(in) :: xi
    real(rt), dimension(size(xi),size(y,2)):: yi
    ! local 
    integer, allocatable :: ilo(:)
    real(rt) :: mlo
    integer :: i

!     do i = 1,size(xi)
!         yi(i,:) = interp1q_scal_ymat(x,y,xi(i))
!     end do

    allocate(ilo(size(xi)))

    call bsearch_many(ilo,xi,x)

    do i = 1,size(xi)
        mlo = (x(ilo(i)+1)-xi(i))/(x(ilo(i)+1)-x(ilo(i)))
        yi(i,:) = y(ilo(i),:)*mlo + y(ilo(i)+1,:)*(1._rt-mlo)
    end do


end function interp1q_vec_ymat

! Additional argument says whether to interpolate along first (the default) or second dimension 
subroutine sub_interp1q(yi,x,y,xi,dim)
    use mod_util, only: assert
    use mod_interp, only: search
    implicit none
    real(rt), dimension(:,:), intent(out) :: yi
    real(rt), dimension(:), intent(in) :: x
    real(rt), dimension(:,:), intent(in) :: y
    real(rt), dimension(:), intent(in) :: xi
    integer, intent(in), optional :: dim
    ! local 
    integer, allocatable :: ilo(:)
    real(rt) :: mlo
    integer :: i,d
    logical :: normal 

    d = 1
    if (present(dim)) d = dim

!     normal = .true.
!     if (present(t)) normal = .not. (t=='T' .or. t=='t')

!     do i = 1,size(xi)
!         yi(i,:) = interp1q_scal_ymat(x,y,xi(i))
!     end do

    allocate(ilo(size(xi)))

    call bsearch_many(ilo,xi,x)

    if (d==1) then
        ! Check sizes
        call assert(size(yi,1)==size(xi),'sub_interp1q: size(yi,1) must equal size(xi)')
        call assert(size(yi,2)==size(y,2),'sub_interp1q: size(yi,2) must equal size(y,2)')


        do i = 1,size(xi)
            mlo = (x(ilo(i)+1)-xi(i))/(x(ilo(i)+1)-x(ilo(i)))
            yi(i,:) = y(ilo(i),:)*mlo + y(ilo(i)+1,:)*(1._rt-mlo)
        end do
    else
        ! Check sizes
        call assert(size(yi,2)==size(xi),'sub_interp1q: size(yi,2) must equal size(xi)')
        call assert(size(yi,1)==size(y,1),'sub_interp1q: size(yi,1) must equal size(y,1)')

        do i = 1,size(xi)
            mlo = (x(ilo(i)+1)-xi(i))/(x(ilo(i)+1)-x(ilo(i)))
            yi(:,i) = y(:,ilo(i))*mlo + y(:,ilo(i)+1)*(1._rt-mlo)
        end do
    end if


end subroutine

! This fast interpolation routine assumes that x and xi are sorted.
function interp1qf_vec(x,y,xi) result(yi)
    implicit none
    real(rt), intent(in) :: x(:),y(:),xi(:)
    real(rt) :: yi(size(xi))
    call interp1qf_core(yi,x,y,xi,1)
end function
function interp1qf_vec_ymat(x,y,xi) result(yi)
    implicit none
    real(rt), intent(in) :: x(:),y(:,:),xi(:)
    real(rt) :: yi(size(xi),size(y,2))
    call interp1qf_core(yi,x,y,xi,size(y,2))
end function
subroutine interp1qf_core(yi,x,y,xi,ncol)
    use mod_interp, only: search
    implicit none
    real(rt), intent(in) :: x(:),y(size(x),*),xi(:)
    integer, intent(in) :: ncol
    real(rt), dimension(size(xi),*), intent(out):: yi
    ! local 
    real(rt) :: mlo
    integer :: i,ilo,m,j

    m = size(x)

    ! Limit the upper bound of the search range by resetting m
    i = size(xi)
    call search(ilo,xi(i),x)
    mlo = (x(ilo+1)-xi(i))/(x(ilo+1)-x(ilo))
    yi(i,1:ncol) = y(ilo,1:ncol)*mlo + y(ilo+1,1:ncol)*(1e0_rt-mlo)
    m = ilo+1 

    ! The lower bound is repeatedly updated below
    ilo = 1

    do i = 1,size(xi)-1

        ! Find the location of xi in the x grid assuming x and xi are sorted
        j = ilo
        call search(ilo,xi(i),x(j:m)) 
        ilo = ilo - 1 + j

        ! print*,'xi(i)',xi(i),'ilo',ilo,'[x,x]',x(ilo:ilo+1)

        ! Do the interpolation
        mlo = (x(ilo+1)-xi(i))/(x(ilo+1)-x(ilo))
        yi(i,1:ncol) = y(ilo,1:ncol)*mlo + y(ilo+1,1:ncol)*(1e0_rt-mlo)

    end do

end subroutine interp1qf_core


! Fast two dimensional interpolation that is similar conceptually
! to interp1q
function interp2q_scal(x1,x2,y,xi1,xi2,workout,workin) result(yi)
    use mod_interp, only: search
    implicit none
    real(rt), dimension(:), intent(in) :: x1,x2
    real(rt), dimension(:,:), intent(in) :: y
    real(rt), intent(in) :: xi1,xi2
    real(rt) :: yi
    type(type_interp2q_work), optional, intent(out) :: workout
    type(type_interp2q_work), optional, intent(out) :: workin
    ! local 
    integer :: ilo1,ilo2
    real(rt) :: mlo1,mlo2

    if (size(x1)/=size(y,1)) STOP 'ERROR: interp2q_scal_ymat: sizes do not match'
    if (size(x2)/=size(y,2)) STOP 'ERROR: interp2q_scal_ymat: sizes do not match'

    if (size(x1)<=1 .or. size(x2)<=1) then
        STOP 'ERROR: x1 or x2 has size <= 1'
    end if

    ! Find the location of xi in the x grid assuming it is sorted
    if (.not. present(workin)) then

        call search(ilo1,xi1,x1)
        call search(ilo2,xi2,x2)

        ! Do the interpolation
        mlo1 = (x1(ilo1+1)-xi1)/(x1(ilo1+1)-x1(ilo1))
        mlo2 = (x2(ilo2+1)-xi2)/(x2(ilo2+1)-x2(ilo2))

        if (present(workout)) then
            workout%ilo1 = ilo1
            workout%ilo2 = ilo2
            workout%wa = mlo1*mlo2
            workout%wb = (1e0_rt-mlo1)*mlo2
            workout%wc = mlo1*(1e0_rt-mlo2)
            workout%wd = (1e0_rt-mlo1)*(1e0_rt-mlo2)

            yi = y(ilo1  ,ilo2  )*workout%wa + &
                 y(ilo1+1,ilo2  )*workout%wb + &
                 y(ilo1  ,ilo2+1)*workout%wc + &
                 y(ilo1+1,ilo2+1)*workout%wd 
        else

            yi = y(ilo1  ,ilo2  )*(    mlo1)*(    mlo2) + &
                 y(ilo1+1,ilo2  )*(1e0_rt-mlo1)*(    mlo2) + &
                 y(ilo1  ,ilo2+1)*(    mlo1)*(1e0_rt-mlo2) + &
                 y(ilo1+1,ilo2+1)*(1e0_rt-mlo1)*(1e0_rt-mlo2) 

        end if

    else

        yi = y(workin%ilo1  ,workin%ilo2  )*workin%wa + &
             y(workin%ilo1+1,workin%ilo2  )*workin%wb + &
             y(workin%ilo1  ,workin%ilo2+1)*workin%wc + &
             y(workin%ilo1+1,workin%ilo2+1)*workin%wd 

    end if

end function interp2q_scal
function interp2q_scal_ymat(x1,x2,y,xi1,xi2) result(yi)
    use mod_interp, only: search
    implicit none
    real(rt), dimension(:), intent(in) :: x1,x2
    real(rt), dimension(:,:,:), intent(in) :: y
    real(rt), intent(in) :: xi1,xi2
    real(rt), dimension(size(y,3)) :: yi
    ! local 
    integer :: ilo1,ilo2
    real(rt) :: mlo1,mlo2

    if (size(x1)/=size(y,1)) STOP 'ERROR: interp2q_scal_ymat: sizes do not match'
    if (size(x2)/=size(y,2)) STOP 'ERROR: interp2q_scal_ymat: sizes do not match'

    if (size(x1)<=1 .or. size(x2)<=1) then
        STOP 'ERROR: x1 or x2 has size <= 1'
    end if

    ! Find the location of xi in the x grid assuming it is sorted
    call search(ilo1,xi1,x1)
    call search(ilo2,xi2,x2)

    ! Do the interpolation
    mlo1 = (x1(ilo1+1)-xi1)/(x1(ilo1+1)-x1(ilo1))
    mlo2 = (x2(ilo2+1)-xi2)/(x2(ilo2+1)-x2(ilo2))

    yi(:) = y(ilo1  ,ilo2  ,:)*(    mlo1)*(    mlo2) + &
            y(ilo1+1,ilo2  ,:)*(1e0_rt-mlo1)*(    mlo2) + &
            y(ilo1  ,ilo2+1,:)*(    mlo1)*(1e0_rt-mlo2) + &
            y(ilo1+1,ilo2+1,:)*(1e0_rt-mlo1)*(1e0_rt-mlo2) 

end function 
function interp2q_vec_ymat(x1,x2,y,xi1,xi2) result(yi)
    use mod_interp, only: search
    implicit none
    real(rt), dimension(:), intent(in) :: x1,x2,xi1,xi2
    real(rt), dimension(:,:,:), intent(in) :: y
    real(rt), dimension(size(xi1),size(xi2),size(y,3)):: yi
    ! local 
    integer :: i1,i2

    if (size(x1)/=size(y,1)) STOP 'ERROR: interp2q_vec_ymat: sizes do not match'
    if (size(x2)/=size(y,2)) STOP 'ERROR: interp2q_vec_ymat: sizes do not match'

    do i2 = 1,size(xi2)
        do i1 = 1,size(xi1)
            yi(i1,i2,:) = interp2q_scal_ymat(x1,x2,y,xi1(i1),xi2(i2))
        end do
    end do

end function interp2q_vec_ymat
function interp2q_vec_ymat2(x1,x2,y,xi1,xi2) result(yi)
    use mod_interp, only: search
    implicit none
    real(rt), dimension(:), intent(in) :: x1,x2,xi1,xi2
    real(rt), dimension(:,:), intent(in) :: y
    real(rt), dimension(size(xi1),size(xi2)):: yi
    ! local 
    integer :: i1,i2

    if (size(x1)/=size(y,1)) STOP 'ERROR: interp2q_vec_ymat: sizes do not match'
    if (size(x2)/=size(y,2)) STOP 'ERROR: interp2q_vec_ymat: sizes do not match'

    do i2 = 1,size(xi2)
        do i1 = 1,size(xi1)
            yi(i1,i2) = interp2q_scal(x1,x2,y,xi1(i1),xi2(i2))
        end do
    end do

end function 

! interp1
! Constructs and evaluates the linear interpolant of (x,y), evaluating at xi.
function interp1_scal_3d(x,y,xi) result(yi)
    implicit none
    real(rt), dimension(:), intent(in) :: x
    real(rt), dimension(:,:,:), intent(in) :: y
    real(rt), intent(in) :: xi
    real(rt), dimension(size(y,2),size(y,3)) :: yi
    ! local

    call interp1_scal_core(yi,x,y,xi,size(x),size(y,2)*size(y,3))

end function 
function interp1_scal_mat(x,y,xi) result(yi)
    implicit none
    real(rt), dimension(:), intent(in) :: x
    real(rt), dimension(:,:), intent(in) :: y
    real(rt), intent(in) :: xi
    real(rt), dimension(size(y,2)) :: yi
    ! local

    call interp1_scal_core(yi,x,y,xi,size(x),size(y,2))

end function 
function interp1_scal(x,y,xi) result(yi)
    implicit none
    real(rt), dimension(1:), intent(in) :: x
    real(rt), dimension(1:), intent(in) :: y
    real(rt), intent(in) :: xi
    real(rt) :: yi
    ! local
    real(rt) :: yitmp(1)

    call interp1_scal_core(yitmp,x,y,xi,size(x),1)
    yi = yitmp(1)

end function interp1_scal
subroutine interp1_scal_core(yi,x,y,xi,nx,ncol)
    implicit none
    integer, intent(in) :: nx,ncol
    real(rt), dimension(:), intent(in) :: x
    real(rt), dimension(nx,*), intent(in) :: y
    real(rt), intent(in) :: xi
    real(rt), dimension(*):: yi
    !local for new algorithm   
    integer :: i, ilo, ihi, ilo_old, ihi_old

    !Special algorithm for xi==1
    if (nx/=size(x)) STOP 'ERROR: interp1_scal_core: x doesnt have nx elements'
    if (nx==1) then
        yi(1:ncol) = y(1,1:ncol) 
        return
    end if

    ! Find the closest indices below and above xi. That is, find i,j s.t. x(i)<=xi<=x(j)
    ilo = -1
    ihi = -1
    ilo_old = -1
    ihi_old = -1
    do i = 1,nx
        if (x(i)<xi) then
            ! Determine whether x(i) is greater than current x(ilo). If it is, replace ilo with i
            if (ilo/=-1) then
                ! We keep the order x(ilo_old)<=x(ilo)<xi whenever ilo_old/=-1/=ilo
                if (x(i)>x(ilo)) then
                    ilo_old = ilo
                    ilo = i
                elseif (ilo_old==-1) then
                    ilo_old = i
                elseif (x(i)>x(ilo_old)) then
                    ilo_old = i
                end if
            else 
                ilo = i
            end if
        elseif (x(i)>xi) then
            ! Determine whether x(i) is less than current x(ihi). If it is, replace ihi with i
            if (ihi/=-1) then
                ! We keep the order x(ihi_old)>=x(ihi)>xi whenever ihi_old/=-1/=ihi
                if (x(i)<x(ihi)) then
                    ihi_old = ihi
                    ihi = i
                elseif (ihi_old==-1) then
                    ihi_old = i
                elseif (x(i)<x(ihi_old)) then
                    ihi_old = i
                end if
            else
                ihi = i
            end if
        else
            ! Exact match, so return value
            yi(1:ncol) = y(i,1:ncol)
            return
        end if
    end do

    ! If still here, have found bracketing interval.
    ! If ilo/=-1 and ihi/=-1, then have x(ilo)<xi<x(ihi) and x(ilo) largest and x(ihi) smallest that satisfy.
    ! If ilo==-1 and ihi/=-1, then xi_in is less than any value of x. Consequently, ihi and ihi_old are the bracketing interval.
    ! If ilo/=-1 and ihi==-1, then xi_in is greater than any value of x. Consequently, ilo and ilo_old are the bracketing interval.
    ! The case of ilo==-1 and ihi==-1 is impossible, because either ilo or ihi is replaced at each step or the function returns
    if (ilo/=-1 .and. ihi/=-1) then
    elseif (ilo==-1 .and. ihi/=-1) then
        if (ihi_old==-1) then
            print*,'xi',xi
            print*,'x',x
            STOP 'interp1_scal: ERROR: not expecting ihi_old to be -1 here'
        end if
        ! ihi and ihi_old bracket it
        ilo = ihi_old
    elseif (ilo/=-1 .and. ihi==-1) then
        if (ilo_old==-1) then
            print*,'xi',xi
            print*,'x',x
            STOP 'interp1_scal: ERROR: not expecting ilo_old to be -1 here'
        end if
        ihi = ilo_old
    else
        print*,'ilo',ilo
        print*,'ihi',ihi
        print*,'xi',xi
        print*,'x',x
        STOP 'interp1_scal: ERROR: not expecting this case'
    end if

    ! Do the linear interpolation.
!     print*,'x',x
!     print*,'xi',xi
!     print*,'x(ilo)',x(ilo)
!     print*,'x(ihi)',x(ihi)
!     print*,'ilo',ilo
!     print*,'ihi',ihi

    yi(1:ncol) = y(ilo,1:ncol) + (y(ihi,1:ncol)-y(ilo,1:ncol))*(xi-x(ilo))/(x(ihi)-x(ilo))

end subroutine interp1_scal_core

function interp1_vec(x_in,y_in,xi_in) result(yi_out)
    implicit none
    real(rt), dimension(:), intent(in) :: x_in
    real(rt), dimension(:), intent(in), target :: y_in
    real(rt), dimension(:), intent(in) :: xi_in
    real(rt), dimension(size(xi_in)), target :: yi_out
    !local
!     real(rt), dimension(1:size(y_in),1:1) :: y
!     real(rt), dimension(1:size(xi_in),1:1) :: yi
!     real(rt), pointer :: y_in_ptr(:,:)

    ! Instead of doing the explict copying here, I should use pointers to "reshape" the values
!     y(:,1) = y_in    
!     yi = interp1_mat(x_in,y,xi_in)
!     yi_out = yi(:,1)


!     ! Create a pointer to y_in to reshape it implicitly, then call interp1_mat routine
!     y_in_ptr(1:size(y_in),1:1) => y_in
!     yi_out = reshape(interp1_mat(x_in,y_in_ptr,xi_in),(/size(xi_in)/))

    call interp1_core(yi_out,x_in,y_in,xi_in,size(x_in),size(xi_in),1)

end function interp1_vec

function interp1_3dim(x_in,y_in,xi_in) result(yi_out)
    use mod_sort, only: issorted
    implicit none
    real(rt), dimension(:), intent(in) :: x_in
    real(rt), dimension(:,:,:), intent(in) :: y_in
    real(rt), dimension(:), intent(in) :: xi_in
    real(rt), dimension(size(xi_in),size(y_in,2),size(y_in,3)) :: yi_out
    !local
!     integer :: ind
! 
!     ! Instead of doing the explict copying here, I should use pointers to "reshape" the values
!     do ind = 1,size(y_in,3)
!         yi_out(:,:,ind) = interp1_mat(x_in,y_in(:,:,ind),xi_in)
!     end do

    call interp1_core(yi_out,x_in,y_in,xi_in,size(x_in),size(xi_in),size(y_in,2)*size(y_in,3))

end function interp1_3dim

function interp1_mat(x_in,y_in,xi_in) result(yi_out)    
    use mod_sort, only: issorted
    implicit none
    real(rt), dimension(1:), intent(in) :: x_in
    real(rt), dimension(1:,1:), intent(in) :: y_in
    real(rt), dimension(1:), intent(in) :: xi_in
    real(rt), dimension(1:size(xi_in),1:size(y_in,2)) :: yi_out

    call interp1_core(yi_out,x_in,y_in,xi_in,size(x_in),size(xi_in),size(y_in,2)) 

!     !local
!     real(rt), allocatable :: x(:),y(:,:),xi(:),yi(:,:),theta(:)
!     integer, allocatable :: Isortx(:),Isortxi(:),Irecoverxi(:),xLoInds(:)
!     integer :: nx,nxi,col_ind,ncol,xi_ind,ny
!     logical :: issorted_ascend_x,issorted_descend_x,issorted_ascend_xi,issorted_descend_xi
! 
!     allocate(xLoInds(1:size(xi_in)))
!     allocate(theta(1:size(xi_in)))
!     allocate(x(1:size(x_in)))
!     allocate(y(1:size(y_in,1),1:size(y_in,2)))
!     allocate(xi(1:size(xi_in)))
!     allocate(yi(1:size(xi_in),1:size(y_in,2))) 
!     allocate(Isortx(1:size(x_in)))
!     allocate(Isortxi(1:size(xi_in)),Irecoverxi(1:size(xi_in)))
! 
!     nx = size(x_in)
!     ny = size(y_in,1)
!     ncol = size(y_in,2)
!     nxi = size(xi_in)
!  
!     if (nx/=ny) STOP 'interp1: x and y do not conform'
! 
!     x = x_in
!     y = y_in
!     xi = xi_in
! 
!     if (any(isNaN(xi))) then
!         write(*,*) 'xi'
!         call disp(xi)
!         STOP 'interp1: xi contains NaNs'
!     end if
! 
!     !Check whether data is sorted either in ascending or descending order
!     issorted_ascend_x = issorted(x,'ascend')
!     issorted_descend_x = .FALSE.
!     if (.NOT. issorted_ascend_x) issorted_descend_x = issorted(x,'descend')
! 
!     issorted_ascend_xi = issorted(xi,'ascend')
!     issorted_descend_xi = .FALSE.
!     if (.NOT. issorted_ascend_xi) issorted_descend_xi = issorted(xi,'descend')
! 
!     !Sort ascending if data is not sorted
!     if (issorted_ascend_x) then
!         !Do nothing
!     elseif (issorted_descend_x) then
!         !Reverse the order or x and y
!         x = reverse(x)
!         y = reverse_mat(y)
!     else
!         call sort2(x,Isortx,x)
!         do col_ind = 1,ncol
!             y(:,col_ind) = y(Isortx,col_ind)
!         end do
!     end if
!     if (issorted_ascend_xi) then
!         !Do nothing
!     elseif (issorted_descend_xi) then
!         xi = reverse(xi)
!     else
!         call sort2(xi,Isortxi,xi)
!         Irecoverxi(Isortxi) = 1 .colon. nxi
!     end if
! 
!     !NEW ALGORITHM: FIRST FINDS INDICES FOR TWO POINTS USED IN INTERPOLATING, SECOND DOES THE CALCULATION
! 
!     !Find bracketing intervals for all xi
!     xi_ind = 1
!     call bsearch(xLoInds(xi_ind),xi(xi_ind),x)
!     if (xLoInds(xi_ind)+1>nx) STOP 'interp1_mat: error'
! 
!     do xi_ind = 2,nxi
!         call bsearch(xLoInds(xi_ind),xi(xi_ind),x(xLoInds(xi_ind-1):nx)) 
!               !Since xi is ascending, we need only check the upper interval
!         xLoInds(xi_ind) = xLoInds(xi_ind) + xLoInds(xi_ind-1) - 1
!         if (xLoInds(xi_ind)+1>nx) STOP 'interp1_mat: error'    
!     end do
! 
!     ! Perform the interpolation
!     theta = (xi-x(xLoInds))/(x(xLoInds+1)-x(xLoInds))
!     do col_ind = 1,ncol
!         yi(:,col_ind) = y(xLoInds,col_ind) + theta*(y(xLoInds+1,col_ind)-y(xLoInds,col_ind))
!     end do
! 
!     !Unsort the data
!     if (issorted_ascend_xi) then
!         !Do nothing
!     elseif (issorted_descend_xi) then
!         yi = reverse_mat(yi)
!     else
!         do col_ind = 1,ncol
!             yi(:,col_ind) = yi(Irecoverxi,col_ind)
!         end do
!     end if
! 
!     !Write to result
!     yi_out = yi
! 
!     ! Clean up
!     deallocate(x,y,xi,yi,Isortx,Isortxi,Irecoverxi,xLoInds,theta)

end function interp1_mat

subroutine interp1_core(yi_out,x_in,y_in,xi_in,nx,nxi,ncol) 
    use mod_sort, only: issorted
    implicit none
    integer, intent(in) :: nx,nxi,ncol
    real(rt), dimension(:), intent(in) :: x_in
    real(rt), dimension(:), intent(in) :: xi_in
    real(rt), dimension(nx,*), intent(in) :: y_in
    real(rt), dimension(nxi,*) :: yi_out
    !local
    real(rt), allocatable :: x(:),y(:,:),xi(:),yi(:,:),theta(:)
    integer, allocatable :: Isortx(:),Isortxi(:),Irecoverxi(:),xLoInds(:)
    integer :: col_ind,xi_ind,ny 
    logical :: issorted_ascend_x,issorted_descend_x,issorted_ascend_xi,issorted_descend_xi

    ny = nx

    allocate(xLoInds(nxi))
    allocate(theta(nxi))
    allocate(x(nx))
    allocate(y(nx,ncol))
    allocate(xi(nxi))
    allocate(yi(nxi,ncol)) 
    allocate(Isortx(nx))
    allocate(Isortxi(nxi),Irecoverxi(nxi))
 
    if (nx/=size(x)) STOP 'interp1: x and nx do not agree'
    if (nxi/=size(xi)) STOP 'interp1: x and nx do not agree'

    ! Cover over inputs
    x = x_in
    y = y_in(:,1:ncol)
    xi = xi_in

!     if (any(isNaN(xi))) then
!         call disp('xi',xi)
!         STOP 'interp1: xi contains NaNs'
!     end if

    ! Check whether data is sorted either in ascending or descending order
    issorted_ascend_x = issorted(x,'ascend')
    issorted_descend_x = .FALSE.
    if (.NOT. issorted_ascend_x) issorted_descend_x = issorted(x,'descend')

    issorted_ascend_xi = issorted(xi,'ascend')
    issorted_descend_xi = .FALSE.
    if (.NOT. issorted_ascend_xi) issorted_descend_xi = issorted(xi,'descend')

    ! Sort ascending if data is not sorted
    if (issorted_ascend_x) then
        !Do nothing
    elseif (issorted_descend_x) then
        !Reverse the order or x and y
        x = reverse(x)
        y = reverse_mat(y)
    else
        call sort2(x,Isortx,x)
        do col_ind = 1,ncol
            y(:,col_ind) = y(Isortx,col_ind)
        end do
    end if
    if (issorted_ascend_xi) then
        !Do nothing
    elseif (issorted_descend_xi) then
        xi = reverse(xi)
    else
        call sort2(xi,Isortxi,xi)
        Irecoverxi(Isortxi) = colon(1,nxi)
    end if

    ! NEW ALGORITHM: FIRST FINDS INDICES FOR TWO POINTS USED IN INTERPOLATING, SECOND DOES THE CALCULATION

    ! Find bracketing intervals for all xi
    xi_ind = 1
    call bsearch(xLoInds(xi_ind),xi(xi_ind),x)
    if (xLoInds(xi_ind)+1>nx) STOP 'interp1_mat: error'

    do xi_ind = 2,nxi
        call bsearch(xLoInds(xi_ind),xi(xi_ind),x(xLoInds(xi_ind-1):nx)) !Since xi is ascending, 
                                                                         ! we need only check the upper interval
        xLoInds(xi_ind) = xLoInds(xi_ind) + xLoInds(xi_ind-1) - 1
        if (xLoInds(xi_ind)+1>nx) STOP 'interp1_mat: error'    
    end do

    ! Perform the interpolation
    ! NOTE: this could be expressed as matrix multiplication of yi = y + matmul(theta,...)
    theta = (xi-x(xLoInds))/(x(xLoInds+1)-x(xLoInds))
    do col_ind = 1,ncol
        yi(:,col_ind) = y(xLoInds,col_ind) + theta*(y(xLoInds+1,col_ind)-y(xLoInds,col_ind))
    end do

    ! Unsort the data
    if (issorted_ascend_xi) then
        !Do nothing
    elseif (issorted_descend_xi) then
        yi = reverse_mat(yi)
    else
        do col_ind = 1,ncol
            yi(:,col_ind) = yi(Irecoverxi,col_ind)
        end do
    end if

    ! Write to result
    yi_out(:,1:ncol) = yi

    ! Clean up
    deallocate(x,y,xi,yi,Isortx,Isortxi,Irecoverxi,xLoInds,theta)

end subroutine interp1_core

! Changes the order of the elements in x
function reverse(x) result(xr)
    
    implicit none
    real(rt), dimension(1:), intent(in) :: x
    real(rt), dimension(1:size(x)) :: xr
    !local
    integer :: i,nx
    nx = size(x)
    do i = 1,nx
        xr(i) = x(nx-i+1)
    end do
end function reverse
! Changes the order of the rows in y
function reverse_mat(y) result(yr)
    implicit none
    real(rt), dimension(1:,1:), intent(in) :: y
    real(rt), dimension(1:size(y,1),1:size(y,2)) :: yr
    !local
    integer :: i,ny
    ny = size(y,1)
    do i = 1,ny
        yr(i,:) = y(ny-i+1,:)
    end do
end function reverse_mat

! Compute the cumulative product of the vector x
function cumprod_r(x) result(y)
    implicit none
    real(rt), intent(in) :: x(1:)
    real(rt) :: y(size(x))
    !local
    real(rt) :: tmp
    integer :: i


    y(1) = x(1)
    tmp = x(1)
    do i = 2,size(x)
        tmp = tmp*x(i)
        y(i) = tmp
    end do

end function cumprod_r

function cumprod_i(x) result(y)
    implicit none
    integer, intent(in) :: x(1:)
    integer  :: y(size(x))
    !local
    integer :: tmp
    integer :: i

    y(1) = x(1)
    tmp = x(1)
    do i = 2,size(x)
        tmp = tmp*x(i)
        y(i) = tmp
    end do

end function cumprod_i

end module mod_matlab


!!! TEST ROUTINES FOR MOD_MATLAB
module mod_matlab_test
    use mod_matlab
    implicit none
contains

    !!! A

    !!! B
    !!! C
    subroutine test_colon
        implicit none
        integer :: x(6)

        x = colon(-3,2)
        if (any(x/=(/-3,-2,-1,0,1,2/))) STOP 'test_colon: failed test 1'
        
        print*,'test_colon: passed'
    end subroutine test_colon
    subroutine test_cumsum
        implicit none
        real(rt) :: A(5),cumA(5),cumAtrue(5)
        real(rt), parameter :: approx0 = epsilon(0.e0_rt)

        A = (/100e0_rt,-3.e0_rt,17e0_rt,-1.e0_rt, 2e0_rt/)
        cumAtrue = reshape((/100e0_rt,97e0_rt,114e0_rt,113e0_rt,115e0_rt/),(/5/))
        cumA = cumsum(A)
        
        if (maxval(abs(cumA-cumAtrue))>approx0) STOP 'test_cumsum: cumsum failed test 1'

        ! Matlab code
!         A = [100,-3,17,-1,2]';
!         formatMatFortran(cumsum(A))

        print*,'test_cumsum: passed'

    end subroutine test_cumsum
    !!! D
    subroutine test_disp
        implicit none
        real(8) :: x(7)

        x = 10d0**[-10d0,-1d0,0d0,1d0,4d0,5d0,10d0]

        call disp('x',x)
        
        print*,'test_disp: passed'
    end subroutine 
    !!! E
    subroutine test_eye
        implicit none
        real(rt) :: A(1,1), B(2,2)

        A = eye(1)
        B = eye(2)

        if (A(1,1)/=1.e0_rt) STOP 'test_eye: eye failed test 1'        
        if (abs(B(1,1)-1e0_rt)>1d-15 .or. abs(B(2,2)-1e0_rt)>1d-15 &
           .or. abs(B(1,2))>1d-15 .or. abs(B(2,1))>1d-15) STOP 'test_eye: eye failed test 2'
        print*,'B',B
        if (any(B/=reshape((/1e0_rt,0e0_rt,0e0_rt,1e0_rt/),(/2,2/)))) STOP 'test_eye: eye failed test 3'

        print*,'test_eye: passed'
    end subroutine test_eye
    !!! F
    subroutine test_find
        implicit none
        real(rt), allocatable :: x(:)
        integer, allocatable :: I(:),J(:),K(:)
!         integer :: J(2), K(1)

        allocate(x(100))

        x = linspace(-6.1e0_rt,5.1e0_rt,100)
        call finda(I,x>4e0_rt .and. x<4.7e0_rt)
        if (size(I)/=6) stop 'test_find: failed test 1'
        if (any(I/=(/91,92,93,94,95,96/))) stop 'test_find: failed test 2'
!         call finda(J,x>4e0_rt .and. x<4.7e0_rt,2,'first')
!         if (size(J)/=2) stop 'test_find: failed test 3'
!         if (any(J/=(/91,92/))) STOP 'test_find: failed test 4'
!         call finda(K,x>4e0_rt .and. x<4.7e0_rt,1,'last')
!         if (any(K/=(/96/))) STOP 'test_find: failed test 4'

        !Matlab Code
!         x = linspace(-6e0_rt,5e0_rt);
!         tmp = x>4e0_rt & x<4.7e0_rt;
!         I = find(tmp);        
!         formatMatFortran(I,1)
!         I = find(tmp,2,'first');
!         formatMatFortran(I,1)
!         I = find(tmp,1,'last');
!         formatMatFortran(I,1)

        print*,'test_find: passed'
    end subroutine test_find
    !!! G
    !!! H
    !!! I
    subroutine test_interp1
        implicit none
        real(rt) :: x(5),y(5,2), xi(8), yi(8,2), yitrue(8,2)
        integer :: i,j
        real(rt), parameter :: approx0 = epsilon(0.e0_rt)*1d3
        real(rt) :: tmp
        x = (/1.e0_rt, -1.1e0_rt, 2.2e0_rt, 5.e0_rt,-3.e0_rt/)
        y(:,1) = .5e0_rt*x + 1e0_rt
        y(:,2) = x**3
        xi = (/-6.e0_rt,1.1e0_rt,-3.e0_rt,-1.1e0_rt,-.5e0_rt,3.33e0_rt,10e0_rt,0e0_rt/) !Contains some interpolation, 
                                !some extrapolation, and some exact values in arbitrary order
        yitrue = reshape((/-2e0_rt,1.55e0_rt,-0.5e0_rt,0.45e0_rt,0.75e0_rt,2.665e0_rt,6e0_rt,1e0_rt,-67.53e0_rt,&
                    1.804e0_rt,-27e0_rt,-1.331e0_rt,-0.665e0_rt,56.7972e0_rt,329.2e0_rt,-0.11e0_rt/),(/8,2/))

        ! Test interp1(vec,mat,vec)
        yi = interp1(x,y,xi)
        if (maxval(abs(yi-yitrue))>approx0) STOP 'test_interp1: interp1 failed test 1'

        ! Test interp1(vec,vec,pt)
        yi = -sqrt(huge(0._rt))
        do j = 1,size(yi,2)
            do i = 1,size(yi,1)
                yi(i,j) = interp1(x,y(:,j),xi(i))
                if (abs(yi(i,j)-yitrue(i,j))>approx0) then
                    print*,'y',y(:,j)
                    print*,'yi',yi(i,j)
                    print*,'yitrue',yitrue(i,j)
                    STOP 'test_interp1: interp1 failed test 2'
                end if
            end do
        end do

        

        ! Test interp1(vec,vec,vec)
        yi = -huge(0e0_rt)
        do j = 1,size(yi,2)
            yi(:,j) = interp1(x,y(:,j),xi)
        end do        
        if (maxval(abs(yi-yitrue))>approx0) STOP 'test_interp1: interp1 failed test 3'

        ! Matlab code
!         x = [1,-1.1,2.2,5,-3]';
!         xi = [-6.e0_rt,1.1e0_rt,-3.e0_rt,-1.1e0_rt,-.5e0_rt,3.33e0_rt,10e0_rt,0e0_rt]';
!         y(:,1) = .5*x + 1;
!         y(:,2) = x.^3;
!         yitrue = interp1(x,y,xi,'linear','extrap');
!         formatMatFortran(yitrue)


        tmp = interp1(&
        (/ 0.24576e0_rt, 0.27585e0_rt, 0.30594e0_rt, 0.33604e0_rt, 0.36613e0_rt, &
        0.39622e0_rt, 0.42631e0_rt, 0.45641e0_rt, 0.4865e0_rt, 0.51659e0_rt, &
        0.54668e0_rt, 0.57678e0_rt, 0.60687e0_rt, 0.63696e0_rt, 0.66705e0_rt, &
        0.69715e0_rt, 0.72724e0_rt, 0.75733e0_rt, 0.78743e0_rt, 0.81752e0_rt, &
        0.84761e0_rt, 0.8777e0_rt, 0.9078e0_rt, 0.93789e0_rt, 0.96798e0_rt, &
        0.99807e0_rt, 1.02817e0_rt, 1.05826e0_rt, 1.08835e0_rt, 1.11845e0_rt, &
        1.14854e0_rt, 1.17863e0_rt, 1.20872e0_rt, 1.23882e0_rt, 1.26891e0_rt, &
        1.299e0_rt, 1.32909e0_rt, 1.35919e0_rt, 1.38928e0_rt, 1.41937e0_rt, &
        1.44947e0_rt, 1.47956e0_rt, 1.50965e0_rt, 1.53974e0_rt, 1.56984e0_rt, &
        1.59993e0_rt, 1.63002e0_rt, 1.66011e0_rt, 1.69021e0_rt, 1.7203e0_rt/), &
        real(colon(1,50),rt),0.245757053346832e0_rt)

        if (abs(tmp-0.999902072011698e0_rt)>1e-6_rt) then
            print*,'tmp',tmp
            STOP 'test_interp1: interp1 failed test 4' 
        end if

        print*,'test_interp1: passed'

    end subroutine test_interp1
    subroutine test_interp1q()
        call test_interp1q_opt(1)
    end subroutine
    subroutine test_interp1qf()
        call test_interp1q_opt(2)
    end subroutine
    subroutine test_interp1q_opt(opt)
        implicit none
        integer, intent(in) :: opt
        real(rt) :: x(7), y(7), xi(6), yi(6),yitrue(6)
        integer :: i

        x = (/ -12.1e0_rt, -0.1e0_rt, 0e0_rt, 0.2e0_rt, 4e0_rt, 9e0_rt, 100e0_rt/)
        y = x**3 - 10e0_rt*x
        xi = (/ -13e0_rt, -0.1e0_rt, -0.05e0_rt, 0e0_rt, 5e0_rt, 99.6e0_rt/)
        yitrue = (/ -1774.428e0_rt, 0.999e0_rt, 0.4995e0_rt, 0e0_rt, 147e0_rt, 9.946115999999999e+05_rt/)

        ! Test the vec routine
        if (opt==1) then
            yi = interp1q(x,y,xi)
        else
            yi = interp1qf(x,y,xi)
            print*,'yi',yi,'yitrue',yitrue
        end if
        if (any(abs(yi-yitrue)>100e0_rt*epsilon(0e0_rt)*abs(yitrue))) STOP 'test_interp1q: failed test 1'

        ! Test the scal routine
        do i = 1,size(xi)
            if (opt==1) then
                yi(i) = interp1q(x,y,xi(i))
            else
                yi(i) = interp1qf(x,y,xi(i))
            end if
        end do
        if (any(abs(yi-yitrue)>100e0_rt*epsilon(0e0_rt)*abs(yitrue))) STOP 'test_interp1q: failed test 2'
        
        ! Test random data against interp1
        call random_seed

        call random_number(x)
        x = x + .00000001e0_rt
        x = cumsum(x) ! x is now random, but increasing
        call random_number(y)
        y = 10e0_rt*y ! y is random 

        call random_number(xi)
        xi = x(1) + (x(size(x))-x(1))*xi 
        xi(1) = minval(x) - .01e0_rt
        xi(2) = maxval(x) + .01e0_rt

        if (opt==1) then
            yi = interp1q(x,y,xi)
            yitrue = interp1(x,y,xi)
            if (any(abs(yi-yitrue)>100e0_rt*epsilon(0e0_rt)*abs(yitrue))) &
                STOP 'test_interp1q: failed test 3: interp1 and interp1q don''t agree'
        else
            ! Can't do unsorted xi
        end if

        print*, 'test_interp1q: passed'

    end subroutine 
    subroutine test_interp1qs()
        implicit none
        real(rt) :: x(7), y(7), xi(6), yi(6),yitrue(6), y2(20,7,5), yi2(20,6,5), err(20,6,5)
        integer :: i,j

        x = (/ -12.1e0_rt, -0.1e0_rt, 0e0_rt, 0.2e0_rt, 4e0_rt, 9e0_rt, 100e0_rt/)
        y = x**3 - 10e0_rt*x
        xi = (/ -13e0_rt, -0.1e0_rt, -0.05e0_rt, 0e0_rt, 5e0_rt, 99.6e0_rt/)
        yitrue = (/ -1774.428e0_rt, 0.999e0_rt, 0.4995e0_rt, 0e0_rt, 147e0_rt, 9.946115999999999e+05_rt/)

        ! Test the routine for just vectors
        call interp1qs(yi,x,y,xi,1,shape(y))
        if (any(abs(yi-yitrue)>100e0_rt*epsilon(0e0_rt)*abs(yitrue))) STOP 'test_interp1qs: failed test 1'

!         print*,'yi',yi
!         print*,'yitrue',yitrue

        ! Test the routine by interpolating in the second dimension.
        ! Have the function just be the i*(x**3 - 10e0_rt*x) + j
        do j = 1,size(y2,3)
            do i = 1,size(y2,1)
                y2(i,:,j) = y*real(i,rt) + real(j,rt)
            end do
        end do

        call interp1qs(yi2,x,y2,xi,2,shape(y2))

        do j = 1,size(y2,3)
            do i = 1,size(y2,1)
                err(i,:,j) =  yi2(i,:,j) - real(i,rt)*yitrue - real(j,rt)
                if (any(abs(err(i,:,j))>100e0_rt*epsilon(0e0_rt)*abs(real(i,rt)*yitrue+real(j,rt)))) STOP 'test_interp1qs: failed test 2'
            end do
        end do


        ! Test the scal routine
!         do i = 1,size(xi)
!             yi(i) = interp1q(x,y,xi(i))
!         end do
!         if (any(abs(yi-yitrue)>100e0_rt*epsilon(0e0_rt)*abs(yitrue))) STOP 'test_interp1q: failed test 2'
!         
!         ! Test random data against interp1
!         call random_seed
! 
!         call random_number(x)
!         x = x + .00000001e0_rt
!         x = cumsum(x) ! x is now random, but increasing
!         call random_number(y)
!         y = 10e0_rt*y ! y is random 
! 
!         call random_number(xi)
!         xi = x(1) + (x(size(x))-x(1))*xi 
!         xi(1) = minval(x) - .01e0_rt
!         xi(2) = maxval(x) + .01e0_rt
! 
!         yi = interp1q(x,y,xi)
!         yitrue = interp1(x,y,xi)
!         if (any(abs(yi-yitrue)>100e0_rt*epsilon(0e0_rt)*abs(yitrue))) &
!             STOP 'test_interp1q: failed test 3: interp1 and interp1q don''t agree'
! 
        print*, 'test_interp1qs: passed'

    end subroutine 
    !!! J
    !!! K
    !!! L
    subroutine test_linspace
        implicit none
        real(rt) :: a,b
        integer :: n
        real(rt), parameter :: approx0 = epsilon(0.e0_rt)
        a = -3e0_rt
        b = 4e0_rt
        n = 5
        if (maxval(abs((/-3e0_rt,-1.25e0_rt,0.5e0_rt,2.25e0_rt,4e0_rt/) - linspace(a,b,n)))>approx0) &
            STOP 'test_linspace: linspace failed test 1'

        print*,'test_linspace: passed'

    end subroutine test_linspace
    !!! M
    subroutine test_mtimes
        implicit none
        real(rt) :: A(2,3),B(3,4),C(2,4), Ctrue(2,4), D(4), E(3), Etrue(3), F(2), G(4), Gtrue(4)
        real(rt), parameter :: approx0 = sqrt(epsilon(0.e0_rt))

        A = reshape((/2e0_rt,-6e0_rt,5e0_rt,12e0_rt,1.1e0_rt,0e0_rt/),(/2,3/))
        B = reshape((/-0.01e0_rt,6e0_rt,100e0_rt,1e0_rt,5e0_rt,-1e0_rt,3.1e0_rt,4e0_rt,-100e0_rt,9e0_rt,3e0_rt,0e0_rt/),(/3,4/))
        Ctrue = reshape((/139.98e0_rt,72.06e0_rt,25.9e0_rt,54e0_rt,-83.8e0_rt,29.4e0_rt,33e0_rt,-18e0_rt/),(/2,4/))
        D = (/-1.2e0_rt,-1.4e0_rt,6e0_rt,9e0_rt/)
        Etrue = [98.212e0_rt,36.8e0_rt,-718.6e0_rt]
        F = (/-2e0_rt,-1e0_rt/)
        Gtrue = (/-352.02e0_rt,-105.8e0_rt,138.2e0_rt,-48e0_rt/)

        call mtimes(C,A,B)
        call mtimes(E,B,D)
        call mtimes(G,transpose(C),F) !G=F*C = (C'*F')' with F and G vectors so no transpose required

        if (maxval(abs(C-Ctrue))>approx0*maxval(abs(Ctrue))) STOP 'test_mtimes: mtimes failed test 1'
        if (maxval(abs(E-Etrue))>approx0*maxval(abs(Etrue))) then
            print*,'E',E
            print*,'Etrue',Etrue
            STOP 'test_mtimes: mtimes failed test 2'
        end if
        if (maxval(abs(G-Gtrue))>approx0*maxval(abs(Gtrue))) STOP 'test_mtimes: mtimes failed test 3'

        !Matlab code
!         A = [2 5 1.1; -6 12 0];
!         B = [-.01 1 3.1 9; 6 5 4 3; 100 -1 -100 0];
!         C = A*B;
!         formatMatFortran(C)
!         D = [-1.2;-1.4;6;9];
!         E = B*D;
!         formatMatFortran(E)
!         F = [-2,-1];
!         G = F*C
!         formatMatFortran(G)

        print*,'test_mtimes: passed'

    end subroutine test_mtimes
    subroutine test_mygemm
        use omp_lib
        implicit none
        real(rt), allocatable :: X(:,:),Y(:,:),Z(:,:),Zcheck(:,:),timer(:,:)
        integer :: i,n,l,m


        allocate(timer(10,3))

        do i = 1,10
            n = 2**i
            allocate(X(n,n),Y(n,n),Z(n,n),Zcheck(n,n))

            do m = 1,n
                do l = 1,n
                    X(l,m) = real(l,rt)/n + 3e0_rt*real(m,rt)/n
                    Y(l,m) = real(m,rt)/n + 3e0_rt*real(l,rt)/n
                end do
            end do

            timer(i,1) = real(omp_get_wtime(),rt)

            Zcheck = matmul(X,Y)

            timer(i,2) = real(omp_get_wtime(),rt)

            call mygemm(Z,X,Y,n,n,n,n,'N','N')

            timer(i,3) = real(omp_get_wtime(),rt)

            if (maxval(abs(Z-Zcheck))>maxval(abs(Zcheck))*sqrt(epsilon(0._rt))) then
                print*,'max|Z-Zcheck|=',maxval(abs(Z-Zcheck))
                print*,'tolerance=',maxval(abs(Zcheck))*sqrt(epsilon(0._rt))
                stop 'test_mygemm: failed'
            end if

            write(*,'(A,i8,3(A,f16.8))') 'test_mygemm: n ',n,' mygemm time ',timer(i,3)-timer(i,2),&
                                        ' matmul time ',timer(i,2)-timer(i,1),' mygemm / matmul : ',&
                                        (timer(i,3)-timer(i,2))/max(1e-6_rt,timer(i,2)-timer(i,1))

            deallocate(X,Y,Z,Zcheck)
        end do

        print*,'test_mygemm: passed'


    end subroutine
    !!! N
    subroutine test_null
        implicit none
        real(rt) :: A(3,3),dsn(3) 

        A = reshape((/0.95e0_rt,0.2e0_rt,0.001e0_rt,0.05e0_rt,0.7e0_rt,0.7e0_rt,0e0_rt,0.1e0_rt,0.299e0_rt/),(/3,3/))

        dsn = nullspace(eye(3)-transpose(A))        

        if (any(abs(dsn-(/0.969604e0_rt,0.242228e0_rt,0.0345547e0_rt/))>max(1e-5_rt,sqrt(epsilon(0._rt))))) then
            print*,'dsn',dsn
            print*,'test_null: warning at test 1'
        end if

        dsn = dsn/sum(dsn)
        if (any(abs(dsn-(/0.777932e0_rt,0.194344e0_rt,0.0277239e0_rt/))>max(1e-5_rt,sqrt(epsilon(0._rt))))) then
            print*,'dsn',dsn
            STOP'test_null: failed test 2'
        end if

        !Matlab code
!         A = [.95 .05 0; .2 .7 .1; .001 .7 .299];
!         formatMatFortran(A)
!         dsn = null(eye(3) - A');
!         formatMatFortran(dsn)
!         dsn = dsn/sum(dsn);
!         formatMatFortran(dsn)

        print*,'test_null: passed'

    end subroutine test_null
    !!! O
    !!! P
    !!! Q
    !!! R
    subroutine test_repmat
        implicit none
        real(rt) :: A(4),B(2,3)
        real(rt) :: true1(8,3),true2(2,12),true3(2,6)

        A = (/1e0_rt,2e0_rt,3e0_rt,4e0_rt/)
        B = reshape((/1e0_rt, 7e0_rt,         2e0_rt,         8e0_rt,         3e0_rt,   9e0_rt/),(/2,3/))

        true1 = reshape((/1e0_rt,2e0_rt,3e0_rt,4e0_rt,1e0_rt,2e0_rt,3e0_rt,4e0_rt,1e0_rt,2e0_rt,3e0_rt,4e0_rt,1e0_rt,&
                  2e0_rt,3e0_rt,4e0_rt,1e0_rt,2e0_rt,3e0_rt,4e0_rt,1e0_rt,2e0_rt,3e0_rt,4e0_rt/),(/8,3/))
        true2 = reshape(&
                (/         1e0_rt,         1e0_rt,         2e0_rt,         2e0_rt,         3e0_rt, &
                           3e0_rt,         4e0_rt,         4e0_rt,         1e0_rt,         1e0_rt, &
                           2e0_rt,         2e0_rt,         3e0_rt,         3e0_rt,         4e0_rt, &
                           4e0_rt,         1e0_rt,         1e0_rt,         2e0_rt,         2e0_rt, &
                           3e0_rt,         3e0_rt,         4e0_rt,         4e0_rt/)&
                           ,(/2,12/))
        true3 = reshape((/ 1e0_rt, 7e0_rt, 2e0_rt, 8e0_rt, 3e0_rt, &
                           9e0_rt, 1e0_rt, 7e0_rt, 2e0_rt, 8e0_rt, &
                           3e0_rt, 9e0_rt/)&
                        ,(/2,6/))

        if (any(repmat(vec2mat(A),2,3)/=true1)) STOP 'test_repmat: failed 1'
        if (any(repmat(transpose(vec2mat(A)),2,3)/=true2)) STOP 'test_repmat: failed 2'
        if (any(repmat(B,1,2)/=true3)) STOP 'test_repmat: failed 3'

        print*,'test_repmat: passed'


    end subroutine
    !!! S
    subroutine test_sort
        implicit none
        real(rt) :: A(11),sortedAtrue(11),sortedA(11)
        real(rt), parameter :: approx0 = epsilon(0.e0_rt)
        A = (/20.6046e0_rt,-46.8167e0_rt,-22.3077e0_rt,-45.3829e0_rt,-40.2868e0_rt,32.3458e0_rt,19.4829e0_rt,&
             -18.2901e0_rt,45.0222e0_rt,-46.5554e0_rt,-45.3829e0_rt/) !random but with one duplicate

        !Test ascending sort
        sortedAtrue = (/-46.8167e0_rt,-46.5554e0_rt,-45.3829e0_rt,-45.3829e0_rt,-40.2868e0_rt,-22.3077e0_rt,&
                        -18.2901e0_rt,19.4829e0_rt,20.6046e0_rt,32.3458e0_rt,45.0222e0_rt/)
        sortedA = sort(A)
        if (maxval(abs(sortedA-sortedAtrue))>approx0) STOP 'test_sort: sort failed test 1'

        ! Test descending sort
        sortedA = 0.e0_rt
        sortedAtrue = (/45.0222e0_rt,32.3458e0_rt,20.6046e0_rt,19.4829e0_rt,-18.2901e0_rt,-22.3077e0_rt,-40.2868e0_rt,&
                        -45.3829e0_rt,-45.3829e0_rt,-46.5554e0_rt,-46.8167e0_rt/)
        sortedA = sort(A,mode='descend')
        if (maxval(abs(sortedA-sortedAtrue))>approx0) STOP 'test_sort: sort failed test 2'        

        !Matlab code
!         A = [20.6046  -46.8167  -22.3077  -45.3829  -40.2868   32.3458   19.4829  -18.2901   45.0222  -46.5554  -45.3829]';
!         formatMatFortran(sort(A))
!         formatMatFortran(sort(A,'descend'))

        print*,'test_sort: passed'

    end subroutine test_sort
    subroutine test_sort2
        implicit none
        real(rt) :: A(11),sortedAtrue(11),sortedA(11)
        integer :: B(11),sortedBtrue(11),sortedB(11)
        real(rt) :: C(2,3),sortedCtrue(2,3),sortedC(2,3)
        real(rt) :: D(2,3),sortedDtrue(2,3),sortedD(2,3)
        integer :: I(11),J(2,3),k
        real(rt), parameter :: approx0 = epsilon(0.e0_rt)

        A = (/20.6046e0_rt,-46.8167e0_rt,-22.3077e0_rt,-45.3829e0_rt,-40.2868e0_rt,32.3458e0_rt,&
              19.4829e0_rt,-18.2901e0_rt,45.0222e0_rt,-46.5554e0_rt,-45.3829e0_rt/) !random but with one duplicate
        B = (/45,32,20,19,-18,-22,-40,-45,-45,-46,-46/)
    
        !Test ascending sort
        sortedAtrue = (/-46.8167e0_rt,-46.5554e0_rt,-45.3829e0_rt,-45.3829e0_rt,-40.2868e0_rt,&
                        -22.3077e0_rt,-18.2901e0_rt,19.4829e0_rt,20.6046e0_rt,32.3458e0_rt,45.0222e0_rt/)
        call sort2(sortedA,I,A)
        if (maxval(abs(sortedA-sortedAtrue))>approx0) STOP 'test_sort2: sort2 failed test 1'
        if (any(A(I)/=sortedA)) STOP 'test_sort2: sort2 failed test 2'

        ! Test descending sort
        sortedA = 0.e0_rt
        sortedAtrue = (/45.0222e0_rt,32.3458e0_rt,20.6046e0_rt,19.4829e0_rt,-18.2901e0_rt,-22.3077e0_rt,&
                        -40.2868e0_rt,-45.3829e0_rt,-45.3829e0_rt,-46.5554e0_rt,-46.8167e0_rt/)
        call sort2(sortedA,I,A,mode='descend')
        if (maxval(abs(sortedA-sortedAtrue))>approx0) STOP 'test_sort2: sort2 failed test 3'        
        if (any(A(I)/=sortedA)) STOP 'test_sort2: sort2 failed test 4'

        !Test ascending sort with integers
        sortedBtrue = (/-46,-46,-45,-45,-40,-22,-18,19,20,32,45/)
        call sort2(sortedB,I,B)
        if (maxval(abs(sortedB-sortedBtrue))>approx0) STOP 'test_sort2: sort2 failed test 5'
        if (any(B(I)/=sortedB)) STOP 'test_sort2: sort2 failed test 6'

        !Test descending sort with integers
        sortedBtrue = (/45,32,20,19,-18,-22,-40,-45,-45,-46,-46/)
        call sort2(sortedB,I,B,mode='descend')
        if (maxval(abs(sortedB-sortedBtrue))>approx0) STOP 'test_sort2: sort2 failed test 7'
        if (any(B(I)/=sortedB)) STOP 'test_sort2: sort2 failed test 8'
        
        !Test ascending sort with matrix dim=1
        C = reshape((/3e0_rt,0e0_rt,7e0_rt,4e0_rt,5e0_rt,2e0_rt/),(/2,3/))
        sortedCtrue = reshape((/0e0_rt,3e0_rt,4e0_rt,7e0_rt,2e0_rt,5e0_rt/),(/2,3/)) 
        call sort2(sortedC,J,C,dim=1)
        if (maxval(abs(sortedC-sortedCtrue))>approx0) STOP 'test_sort2: sort2 failed test 9'
        do k = 1,size(C,2)
            if (any(C(J(:,k),k)/=sortedC(:,k))) STOP 'test_sort2: sort2 failed test 10'
        end do

        !Test ascending sort with matrix dim=2
        C = reshape((/3e0_rt,0e0_rt,7e0_rt,4e0_rt,5e0_rt,2e0_rt/),(/2,3/))
        sortedCtrue = reshape((/3e0_rt,0e0_rt,5e0_rt,2e0_rt,7e0_rt,4e0_rt/),(/2,3/))
        call sort2(sortedC,J,C,dim=2)
        if (maxval(abs(sortedC-sortedCtrue))>approx0) STOP 'test_sort2: sort2 failed test 11'
        do k = 1,size(C,1)
            if (any(C(k,J(k,:))/=sortedC(k,:))) STOP 'test_sort2: sort2 failed test 12'
        end do

        !Test ascending sort with matrix dim=1
        D = reshape((/3,0,7,4,5,2/),(/2,3/))
        sortedDtrue = reshape((/0,3,4,7,2,5/),(/2,3/)) 
        call sort2(sortedD,J,D,dim=1)
        if (maxval(abs(sortedD-sortedDtrue))>approx0) STOP 'test_sort2: sort2 failed test 13'
        do k = 1,size(D,2)
            if (any(D(J(:,k),k)/=sortedD(:,k))) STOP 'test_sort2: sort2 failed test 14'
        end do

        !Test ascending sort with matrix dim=2
        D = reshape((/3,0,7,4,5,2/),(/2,3/))
        sortedDtrue = reshape((/3,0,5,2,7,4/),(/2,3/))
        call sort2(sortedD,J,D,dim=2)
        if (maxval(abs(sortedD-sortedDtrue))>approx0) STOP 'test_sort2: sort2 failed test 15'
        do k = 1,size(D,1)
            if (any(D(k,J(k,:))/=sortedD(k,:))) STOP 'test_sort2: sort2 failed test 16'
        end do

        !Matlab code
!         A = [20.6046  -46.8167  -22.3077  -45.3829  -40.2868   32.3458   19.4829  -18.2901   45.0222  -46.5554  -45.3829]';
!         formatMatFortran(sort(A))
!         formatMatFortran(sort(A,'descend'))
!         B = [20  -46  -22  -45  -40 32 19 -18 45 -46 -45]';
!         formatMatFortran(sort(B),1)
!         formatMatFortran(sort(B,'descend'),1)
!         C = [3 7 5; 0 4 2];
!         formatMatFortran(sort(C,1))
!         formatMatFortran(sort(C,2))

        print*,'test_sort2: passed'
        
    end subroutine test_sort2

    subroutine test_sortrows
        implicit none
        integer :: XI(6,7),sortedXI(6,7),truesortedXI(6,7),redo(6),XIorig(6,7)
        real(rt) :: rXI(6,7),rsortedXI(6,7),rtruesortedXI(6,7),rXIorig(6,7)

        XI = reshape([    95,  95,  95,  95,  76,  76,  45,   7,   7,   7,  61,  79,  92,  73,  73,  40,&
                          93,  91,  41,  89,   5,  35,  81,   0,  13,  20,  19,  60,  27,  19,   1,  74,&
                          44,  93,  46,  41,  84,  52,  20,  67,  83,   1],[6,7])
        XIorig = XI
        truesortedXI = reshape([  76,  76,  95,  95,  95,  95,  61,  79,   7,   7,   7,  45,  93,  91,  40,  73,&
                                  73,  92,  81,   0,  35,   5,  89,  41,  27,  19,  60,  19,  20,  13,  46,  41,&
                                  93,  44,  74,   1,  83,   1,  67,  20,  52,  84],[6,7])
        rXI = real(XI,rt)
        rXIorig = rXI
        rtruesortedXI = real(truesortedXI,rt)

        call sortrows(sortedXI,redo,XI)

        if (any(sortedXI/=truesortedXI)) STOP 'test_sortrows: failed sorting integers'
        if (any(XIorig(redo,:)/=truesortedXI)) STOP 'test_sortrows: redo does not work for integers'
        
        call sortrows(rsortedXI,redo,rXI)

        if (any(rsortedXI/=rtruesortedXI)) STOP 'test_sortrows: failed sorting reals'
        if (any(rXIorig(redo,:)/=rtruesortedXI)) STOP 'test_sortrows: redo does not work for reals'

        print*,'test_sortrows: passed'

    end subroutine test_sortrows
    !!! T
    !!! U
    subroutine test_unique
        implicit none
        integer :: A(10) = [-4,-4,5,0,1,5,-2,3,3,5]
        integer :: Btrue(6) = [-4,-2,0,1,3,5]
        integer, allocatable :: B(:)

        call unique(B,A)

        if (any(B/=Btrue)) STOP 'test_unique: failed unique integer test'

        print*,'test_unique: passed'


    end subroutine 
    !!! V
    !!! W
    !!! X
    !!! Y
    !!! Z

end module mod_matlab_test



module mod_gini
    use mod_types
    implicit none
    integer, parameter :: nLorenzPts = 10001

contains 

! Computes an estimate of the gini coefficient via a monte carlo sampling of size k
function fn_gini_approx(k,dsn,grid,n) result(gini)
    use mod_matlab
    implicit none
    real(rt) :: gini
    integer, intent(in) :: k,n
    real(rt), intent(in) :: dsn(*),grid(*) ! size n
    ! local
    real(rt), dimension(:), allocatable :: cdsn,rg
    integer, dimension(:), allocatable :: redo
    real(rt) :: x,SN,SD
    integer :: i,j

    ! Sort the grid
    allocate(cdsn(0:n))
    allocate(rg(k),redo(k))

    cdsn(0) = 0._rt
    do i = 1,n
        cdsn(i) = cdsn(i-1) + dsn(i)
    end do
!     cdsn(0) = 0._rt
!     cdsn(1:n) = cumsum(dsn(1:n))
    if (abs(cdsn(n)-1d0)>sqrt(epsilon(0._rt))) then
        write(0,*) 'ERROR: fn_gini_approx: dsn doesn''t sum to one but to ',cdsn(n),' trying to recover'
    end if
    cdsn(n) = 1._rt

    ! Random sample
    do i = 1,k

        call random_number(x)
        call bsearch(j,x,cdsn)
        rg(i) = grid(j)
        
    end do

    ! Sort rg
    rg = sort(rg)

    ! Compute the gini
    SN = 0._rt
    SD = 0._rt
    do i = 1,k
        SN = SN + i*rg(i)
        SD = SD + rg(i)
    end do

    gini = (2*SN)/(k*SD) - (k+1)/k

end function


! Returns the requested lorenz points, as well as the gini
subroutine sub_lorenzpts(gini,lorenzpts,dsn,grid,n)
    use mod_sort, only: qsort
    use mod_matlab
    implicit none
    real(rt), intent(inout) :: gini, lorenzpts(:)
    real(rt), intent(in) :: dsn(*),grid(*)
    integer, intent(in) :: n
    ! local
    real(rt), dimension(:), allocatable :: Lorenz,Lorenz_grid,sdsn,sgrid
    integer, dimension(:), allocatable :: redo
    integer :: i

    ! Sort the grid
    allocate(sgrid(n),sdsn(n),redo(n))

    sgrid = grid(1:n)
    call qsort(sgrid,redo)
    sdsn = dsn(redo)

    allocate(Lorenz(nLorenzPts),Lorenz_grid(nLorenzPts))

    ! Get the Lorenz curve
    call sub_gini(gini,Lorenz,Lorenz_grid,sdsn,sgrid)

    ! Interpolate 
    if (size(lorenzpts)>0) then
        lorenzpts = interp1(Lorenz_grid,Lorenz,lorenzpts)
    end if

end subroutine 

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! y_in is the value and fy_in is the associated 
! density function.  e.g., a grid on wealth and 
! the invariant distribution of wealth
function fn_gini(y_in,fy_in) result(G)
    use mod_matlab
    implicit none
    real(rt), dimension(:), intent(in) :: y_in,fy_in
    real(rt), dimension(1:size(y_in)):: y,fy
    integer, dimension(1:size(y_in)) :: srt_ind
    real(rt), dimension(1:size(y_in)+1) :: S
    real(rt), dimension(1:size(y_in)) :: Siminus1_Si
    real(rt) :: G
    integer :: n, i
    
    n=size(y)
    y=y_in(1:n)
    fy=fy_in(1:n)
    
    do i=1,n
        srt_ind(i) = i
    end do
    
    call sort2(y,srt_ind,y)
    fy=fy(srt_ind)
    
    S(1)=0.e0_rt
    call sub_cumsum(S(2:n+1),y*fy)
    
    Siminus1_Si=S(2:n+1)+S(1:n)
    
    G=1-dot_product(fy,Siminus1_Si)/S(n+1)

end function fn_gini

subroutine sub_gini(gini,Lorenz,Lorenz_grid,dsn,grid)
    use mod_matlab, only: interp1, linspace, cumsum, issorted, unique
    implicit none
    real(rt), intent(out) :: gini
    real(rt), dimension(nLorenzPts), intent(out)  :: Lorenz,Lorenz_grid
    real(rt), dimension(1:), intent(in) :: dsn,grid
    !local
    real(rt), dimension(:), allocatable :: pct_of_stuff, pct_of_pop, pctpop, pctstuff
    integer, dimension(:), allocatable :: undo, redo

    allocate(pct_of_stuff(size(dsn)+2),pct_of_pop(size(dsn)+2),undo(size(dsn)+2))

    !Check inputs
    if (size(dsn)/=size(grid)) STOP 'sub_gini: dsn and grid size don''t agree'
    if (.not. issorted(grid,'ascend')) STOP 'sub_gini: grid is not sorted'
    
    Lorenz_grid = linspace(0.e0_rt,1.e0_rt,nLorenzPts)
    
    pct_of_stuff(1) = 0.e0_rt
    pct_of_stuff(2:size(dsn)+1) = cumsum(grid*dsn)/dot_product(grid,dsn)
    pct_of_stuff(size(dsn)+2) = 1.e0_rt
    
    pct_of_pop(1) = 0.e0_rt
    pct_of_pop(2:size(dsn)+1) = cumsum(dsn)/sum(dsn)
    pct_of_pop(size(dsn)+2) = 1.e0_rt


!     call unique(pctpop,redo,undo,pct_of_pop)
!     allocate(pctstuff(size(pctpop)))
!     pctstuff = pct_of_stuff(redo)
! 
!     Lorenz = interp1(pctpop,pctstuff,Lorenz_grid)
!     gini = sum(Lorenz_grid-Lorenz)/(sum(Lorenz_grid-Lorenz)+sum(Lorenz))  

    Lorenz = interp1(pct_of_pop,pct_of_stuff,Lorenz_grid)
    gini = sum(Lorenz_grid-Lorenz)/(sum(Lorenz_grid-Lorenz)+sum(Lorenz))  

end subroutine sub_gini


end module mod_gini
module mod_spline
    use mod_types
    implicit none

    interface ppval
        module procedure ppval_scal,ppval_vec
    end interface
!     interface ppval2
!         module procedure :: ppval2_scal,ppval2_vec
!     end interface
!     interface pchip
!         module procedure :: pchip_getpp,pchip_interp
!     end interface
    interface schumaker_interp
        module procedure schumaker_interp_vec, schumaker_interp_scal
    end interface

    ! One-dimensional spline
    type spline
        integer :: n = -1 ! Number of knots 
        integer :: d = -1 ! d-1 is the order of the polynomial in each interval
        real(rt), allocatable :: k(:) ! Knots 
        real(rt), allocatable :: co(:,:) ! Coefficients 
        character(len=20) :: method ! Method of spline
    !!! I removed this final procedure on 4/5/13 since gfortran doesn't support it yet
!     contains
!         final :: freeSpline
    end type

    ! Two dimensional spline
    type spline2 
        integer :: n1 = -1, n2=-1 ! Number of knots 
        integer :: d = -1 ! d-1 is the order of polynomial in each interval
        real(rt), allocatable :: k1(:),k2(:) ! Knots 
        real(rt), allocatable :: co1(:,:,:) ! Coefficients for each 2nd dim knot
!         real(rt), allocatable :: co2(:,:) ! Coefficients: these are not used currently
        character(len=20) :: method ! Method of spline (for now, assume same each dimension)
    end type

!     interface vardim_eval
!         module procedure :: vardim_eval_vec, vardim_eval_scal
!     end interface

    type, public :: vardimspline
        real(rt), allocatable :: xdata(:)
        real(rt), allocatable :: knots(:),taustar(:),f_taustar(:) ! basically work arrays
        real(rt) :: lb,ub
        integer :: p,nx
    contains 
        procedure, pass :: free => vardim_free
        procedure, pass :: config => vardim_config
        procedure, pass :: neval => vardim_getnumberofevalpts
        procedure, pass :: getieval => vardim_getithevalpt
        procedure, pass :: setieval => vardim_setithevalpt
        procedure, pass :: plot => vardim_plot
        procedure, pass :: test => vardim_test
        procedure, pass :: eval => vardim_eval_vec
        procedure, pass :: eval1 => vardim_eval_scal
        procedure, pass :: disp => vardim_disp
    end type

    ! type, public :: cubicspline
    type :: cubicspline
        real(rt), allocatable :: x(:),coef(:,:),a(:),b(:)
    contains 
        procedure, pass :: construct => cubic_construct
        procedure, pass :: eval_q => cubic_eval_quick
        procedure, pass :: eval_s => cubic_eval_stable
        procedure, pass :: plot => cubic_plot
    end type

    type :: type_pchip
        real(rt), allocatable :: x(:),y(:),slopes(:)
    contains 
        procedure, pass :: construct => pchip_construct
        procedure, pass :: construct_nomono => pchip_construct_nomono
        procedure, pass :: eval => pchip_eval_scal
        procedure, pass :: eval_w_i => pchip_eval_w_i_scal
        procedure, pass :: deriv => pchip_deriv_scal
        procedure, pass :: plot => pchip_plot
!         procedure, pass :: eval_q => cubic_eval_quick
!         procedure, pass :: eval_s => cubic_eval_stable
    end type

    private :: norm

contains

    subroutine pchip_plot(this,derivorder) 
        use mod_matlab
        implicit none
        class(type_pchip), intent(in) :: this
        real(rt), allocatable :: x(:),y(:),yp(:),ypp(:)
        integer, optional :: derivorder
        integer :: i,n

        n = 1000
        allocate(x(n),y(n),yp(n),ypp(n))

        x = linspace(this%x(1),this%x(size(this%x)),n)
        do i = 1,n
            ! y(i) = this%eval(x(i))
            call this%deriv(x(i),y(i),yp(i),ypp(i))
        end do

        if (present(derivorder)) then
            select case (derivorder)
                case (1)
                    call plot(x,yp,'Derivatives of pchip')
                    return
                case (2)
                    call plot(x,ypp,'Second derivatives of pchip')
                    return
                case (0) 
                case default
                    write(0,*) 'WARNING: pchip: bad deriv order, plot will nto be made'
            end select
        end if

        ! Default
        call plot(x,y,'Values of pchip')

    end subroutine 

    subroutine pchip_deriv_scal(this,x,y,yp,ypp)
        use mod_matlab
        implicit none
        class(type_pchip), intent(in) :: this
        real(rt), intent(in) :: x
        real(rt), intent(out), optional :: y,yp,ypp
        ! local
        real(rt) :: h,t,h00,h10,h01,h11,d00,d10,d01,d11,s00,s01,s10,s11
        integer :: i

        call bsearch(i,x,this%x)

        h = (this%x(i+1)-this%x(i))
        t = (x-this%x(i))/h

        h00 = (1._rt+2._rt*t)*(1._rt-t)**2
        d00 = -2*(1._rt+2._rt*t)*(1._rt-t) + 2*(1._rt-t)**2 ! derivative of above wrt t
        s00 = -2*(2._rt)*(1._rt-t) + 2*(1._rt+2._rt*t) - 4*(1._rt-t) ! 2nd derivative of above wrt t

        h10 = t*(1._rt-t)**2
        d10 = (1-t)**2 - 2*t*(1._rt-t) ! derivative of above wrt t
        s10 = -2*(1-t) - 2*(1-2*t) ! 2nd derivative of above wrt t

        h01 = t**2*(3._rt-2._rt*t)
        d01 = 2*t*(3._rt-2._rt*t) - 2*t**2 ! derivative of above wrt t
        s01 = 2*(3._rt-2._rt*t) -4*t - 4*t

        h11 = t**2*(t-1._rt)
        d11 = t**2 + 2*t*(t-1) ! derivative of above wrt t
        s11 = 2*t + 2*(2*t-1) ! 2nd derivative of above wrt t

        ! (f(t(x)))' : f'*t' = f'/h
        ! 2nd        : (f'*t')' = f''*t' + f'*t'' = f''/h + 0

        ! Convert to derivative wrt x
        d00 = d00/h
        d10 = d10/h
        d01 = d01/h
        d11 = d11/h

        s00 = s00/h
        s10 = s10/h
        s01 = s01/h
        s11 = s11/h

        if (present(y)) y = this%y(i)*h00 + h*this%slopes(i)*h10 + this%y(i+1)*h01 + h*this%slopes(i+1)*h11
        if (present(yp)) yp = this%y(i)*d00 + h*this%slopes(i)*d10 + this%y(i+1)*d01 + h*this%slopes(i+1)*d11
        if (present(ypp)) ypp = this%y(i)*s00 + h*this%slopes(i)*s10 + this%y(i+1)*s01 + h*this%slopes(i+1)*s11
        
    end subroutine
    
    ! User provides the index, avoiding a call to bsearch
    function pchip_eval_w_i_scal(this,x,i) result(y)
        use mod_matlab
        implicit none
        class(type_pchip), intent(in) :: this
        real(rt), intent(in) :: x
        integer, intent(in) :: i
        real(rt) :: y
        ! local
        real(rt) :: h,t,h00,h10,h01,h11

        ! Check that the i is correct
        if (i>=size(this%x)) stop 'WARNING: pchip_eval_w_i: i is greater than size(x)'
        if (x<this%x(i) .or. x>this%x(i+1)) write(0,*) 'WARNING: pchip_eval_w_i: i provided did not bracket x'

        h = (this%x(i+1)-this%x(i))
        t = (x-this%x(i))/h

        h00 = (1._rt+2._rt*t)*(1._rt-t)**2
        h10 = t*(1._rt-t)**2
        h01 = t**2*(3._rt-2._rt*t)
        h11 = t**2*(t-1._rt)

        y = this%y(i)*h00 + h*this%slopes(i)*h10 + this%y(i+1)*h01 + h*this%slopes(i+1)*h11
        
    end function

    function pchip_eval_scal(this,x) result(y)
        use mod_matlab
        implicit none
        class(type_pchip), intent(in) :: this
        real(rt), intent(in) :: x
        real(rt) :: y
        ! local
        real(rt) :: h,t,h00,h10,h01,h11
        integer :: i

        call bsearch(i,x,this%x)

        h = (this%x(i+1)-this%x(i))
        t = (x-this%x(i))/h

        h00 = (1._rt+2._rt*t)*(1._rt-t)**2
        h10 = t*(1._rt-t)**2
        h01 = t**2*(3._rt-2._rt*t)
        h11 = t**2*(t-1._rt)

        y = this%y(i)*h00 + h*this%slopes(i)*h10 + this%y(i+1)*h01 + h*this%slopes(i+1)*h11
        
    end function

    subroutine pchip_construct_nomono(this,x,y)
        use mod_util
        use mod_matlab
        implicit none
        class(type_pchip), intent(out) :: this
        real(rt), intent(in) :: x(:),y(:)
        ! local
        real(rt), dimension(:), allocatable :: delta, m
        integer :: k,n,sig

        if (allocated(this%x)) deallocate(this%x)
        if (allocated(this%y)) deallocate(this%y)
        if (allocated(this%slopes)) deallocate(this%slopes)

        n = size(x)
        call assert(size(y)==n,'phip_construct')
        
        allocate(this%x(n))
        allocate(this%y(n))
        allocate(this%slopes(n))
        allocate(delta(0:n))
        allocate(m(n))

        this%x = x
        this%y = y

        delta(1:n-1) = (y(2:n)-y(1:n-1))/(x(2:n)-x(1:n-1))
        delta(0) = delta(1)
        delta(n) = delta(n-1)

        m(2:n-1) = .5d0*(delta(1:n-2) + delta(2:n-1))
        m(1) = delta(1)
        m(n) = delta(n-1)

        this%slopes = m

    end subroutine 
    
    subroutine pchip_construct(this,x,y)
        use mod_util
        use mod_matlab
        implicit none
        class(type_pchip), intent(out) :: this
        real(rt), intent(in) :: x(:),y(:)
        ! local
        real(rt), dimension(:), allocatable :: delta, m 
        integer :: k,n,sig

        if (allocated(this%x)) deallocate(this%x)
        if (allocated(this%y)) deallocate(this%y)
        if (allocated(this%slopes)) deallocate(this%slopes)

        n = size(x)
        call assert(size(y)==n,'phip_construct')
        
        allocate(this%x(n))
        allocate(this%y(n))
        allocate(this%slopes(n))
        allocate(delta(0:n))
        allocate(m(n))

        this%x = x
        this%y = y

        delta(1:n-1) = (y(2:n)-y(1:n-1))/(x(2:n)-x(1:n-1))
        delta(0) = delta(1)
        delta(n) = delta(n-1)

        m(2:n-1) = .5d0*(delta(1:n-2) + delta(2:n-1))
        m(1) = delta(1)
        m(n) = delta(n-1)

        ! See page 476, eqn 4.2 of randall l. dougherty, alan edelman, and james m. hyman
        do k = 1,n
            if (delta(k-1)*delta(k)>0._rt) then
                sig = merge(1,-1,delta(k)>0._rt)
            else
                sig = 0
            end if
            select case (sig)
                case (1)
                    m(k) = min(max(0._rt,m(k)),3._rt*min(abs(delta(k-1)),abs(delta(k))))
                case (0)
                    m(k) = 0._rt
                case (-1)
                    m(k) = max(min(0._rt,m(k)),-3._rt*min(abs(delta(k-1)),abs(delta(k))))
            end select
        end do

        this%slopes = m


        return

!         where (delta/=0._rt) 
!             alpha = m(1:n-1)/delta
!             beta = m(2:n)/delta
!         end where

        ! Note beta = F.alpha*F.deta

!         do k = 2,n-1
!             if (delta(i-1)*delta(i)>0._rt) then
!                 
!             m = max(
!         end do
! 
!         do k = 1,n-1
!             if (delta(k)==0._rt .or. delta(k)==0._rt) then
!                 m(k) = 0._rt
!                 m(k+1) = 0._rt
!             elseif (alpha(k)<0._rt) then
!                 m(k) = 0._rt
! !                m(k+1) = 0._rt
!             elseif (k>1 .and. beta(max(k,1))<0._rt) then
!                 m(k) = 0._rt
!                 m(k+1) = 0._rt
!             else
!                 norm = norm2([alpha(k),beta(k)])
!                 if (norm>3._rt) then
!                     tau = 3._rt/norm
!                     m(k) = tau*alpha(k)*delta(k)
!                     m(k+1) = tau*beta(k)*delta(k)
!                 else
!                     m(k) = alpha(k)*delta(k)
!                     m(k+1) = beta(k)*delta(k)
!                 end if
!             end if
!         end do
! 
! 
!         this%slopes = m
! 
!         return
! 
!         i = 1
!         d_n = -huge(0._rt)
! 
!         do i = 1,n
! 
!             d_o = d_n
!             
!             if (i==1) then
!                 d_n = (y(i+1)-y(i))/(x(i+1)-x(i))
!                 d_o = d_n
!             elseif (i==n) then
!                 d_n = d_o
!             else
!                 d_n = (y(i+1)-y(i))/(x(i+1)-x(i))
!             end if
! 
!             if (d_n/=0._rt .and. d_o/=0._rt) then
!                 this%slopes(i) = .5d0*(d_o+d_n) ! average of finite difference slopes
!             else
!                 this%slopes(i) = 0d0 ! must be zero to preserve monotonicity
!             end if
!             
!             if (i<n) then
!                 a = this%slopes(i)/d_o
!                 b = this%slopes(i+1)/d_o
!                 if (a<0._rt .or. b<0._rt) then
!                     ! input data not monotone
!                     this%slopes(i) = 0d0
!                 else
!                     if (a**2+b**2<9) then
!                         this%slopes(i) = 
!                     else
!                         this%slopes(i) = 
! 
!                     end if
!                 end if
!             end if
!         end do
! 
!         do i = 1,n-1
!             a = this%slopes(i)
!             b = this%slopes(i+1)
!         end do
! 
! 
!         do i = 1,n-1
!             if (
! 
!             m_o = m_n
!             m_n = 
!         end do
! 

    end subroutine 
    

    
    subroutine vardim_free(this)
        implicit none
        class(vardimspline), intent(inout) :: this
        
        if (allocated(this%knots)) then
            deallocate(this%xdata)
            deallocate(this%knots)
            deallocate(this%taustar)
            deallocate(this%f_taustar)
            this%p = -1
            this%nx = -1
            this%lb = -1e0_rt
            this%ub = -1e0_rt
        end if

    end subroutine 

    subroutine vardim_config(this,order,xdata)
        implicit none
        class(vardimspline), intent(inout) :: this
        integer, intent(in) :: order
        real(rt), intent(in) :: xdata(:)
        ! local
        integer :: p,nx,j

        if (allocated(this%xdata)) call this%free()

        this%p = order
        this%nx = size(xdata)

        p = this%p
        nx = this%nx

        allocate(this%xdata(nx))
        this%xdata = xdata

        this%lb = xdata(1)
        this%ub = xdata(nx)

        allocate(this%knots(2*p + nx))
        this%knots(1:p) = xdata(1)
        this%knots(p+1:p+nx) = xdata
        this%knots(p+nx+1:2*p+nx) = xdata(nx)

        allocate(this%taustar(nx+p-1))
        do j = 1,nx+p-1
            this%taustar(j) = sum(this%knots(j+1:j+p))/p
        end do

        allocate(this%f_taustar(nx+p-1))

    end subroutine 
    function vardim_getnumberofevalpts(this) result(n)
        implicit none
        class(vardimspline), intent(in) :: this
        integer :: n

        n = size(this%taustar)

    end function 
    function vardim_getithevalpt(this,i) result(x)
        implicit none
        class(vardimspline), intent(in) :: this
        integer, intent(in) :: i
        real(rt) :: x

        x = this%taustar(i)

    end function
    subroutine vardim_setithevalpt(this,i,fx) 
        implicit none
        class(vardimspline), intent(inout) :: this
        integer, intent(in) :: i
        real(rt), intent(in) :: fx

        this%f_taustar(i) = fx


    end subroutine
    function vardim_eval_vec(this,x) result(y) 
        implicit none
        class(vardimspline), intent(in) :: this
        real(rt), intent(in) :: x(:)
        real(rt) :: y(size(x))
        ! local
        integer :: ix
        ! local
!         real(rt), allocatable :: work(:)
!         integer :: j
! 
!         allocate(work(size(x)))
! 
!         work = 0e0_rt
! 
!         do j = 1,this%nx + this%p - 1
!             work = work + this%f_taustar(j)*bspline(x,j,this%p+1,this%knots)
!         end do
! 
!         y = work
        
        do ix = 1,size(x)
            y(ix) = vardim_eval_scal(this,x(ix))
        end do

    end function 
    function vardim_eval_scal(this,x) result(y) 
        implicit none
        class(vardimspline), intent(in) :: this
        real(rt), intent(in) :: x
        real(rt) :: y
        ! local
        real(rt) :: B(this%p+1)
        integer :: ilo,ihi

        call bsplinefast(B,ilo,ihi,x,this%p+1,this%knots) 
        y = dot_product(this%f_taustar(ilo:ihi),B)

    end function 
        
    subroutine vardim_disp(this)
        implicit none
        class(vardimspline), intent(in) :: this

        print*,'*vardim ************************************************************************'
        if (allocated(this%xdata)) then
            print*,'*vardim xdata',this%xdata
        else
            print*,'*vardim xdata not allocated'
        end if
        if (allocated(this%knots)) then
            print*,'*vardim knots',this%knots
        else
            print*,'*vardim knots not allocated'
        end if
        if (allocated(this%taustar)) then
            print*,'*vardim taustar',this%taustar
        else
            print*,'*vardim taustar not allocated'
        end if
        if (allocated(this%f_taustar)) then
            print*,'*vardim f_taustar',this%f_taustar
        else
            print*,'*vardim f_taustar not allocated'
        end if
        print*,'*vardim p',this%p
        print*,'*vardim nx',this%nx
        print*,'*vardim ************************************************************************'

    end subroutine

    subroutine vardim_plot(this,xopt) 
        use mod_matlab
        implicit none
        class(vardimspline), intent(in) :: this
        real(rt), intent(in), optional :: xopt(:)
        ! local
        real(rt), allocatable :: x(:),y(:)
        integer :: nx

        nx = 1000
        if (present(xopt)) nx = size(xopt)

        allocate(x(nx),y(nx))

        x = linspace(this%lb,this%ub,nx)
        if (present(xopt)) x = xopt

        y = this%eval(x)

!         call this%disp()
!         print*,'x',x
!         print*,'y',y

        call plot(x,y,'vardimspline')

    end subroutine

    subroutine vardim_test(this,fcn,xdata,order) 
        use mod_matlab, only: linspace,plot
        implicit none
        class(vardimspline), intent(inout) :: this
        real(rt), intent(in) :: xdata(:)
        integer, intent(in) :: order
        interface 
            function fcn(x) 
                use mod_types
                implicit none
                real(rt), intent(in) :: x
                real(rt) :: fcn
            end function
        end interface
        ! local
        integer :: i
        real(rt) :: tmp
!         real(rt), allocatable :: x(:),y(:)

        call this%config(order,xdata)

        do i = 1,this%neval()
            tmp = fcn(this%getieval(i)) ! evaluate the function at the ith requested point
            call this%setieval(i,tmp) ! set the ith point
        end do

        call this%disp()

        call this%plot()

!         allocate(x(1000),y(1000))
!         
!         x = linspace(xdata(1),xdata(size(xdata)),1000)
!         y = this%eval(x)
! 
!         call plot(x,y,'test vardimspline')
! 

    end subroutine

    ! Uses a variation diminishing sple to interpolate
!     function interp_vardim(X,Y,XX,order) result(YY)
!         implicit none
!         real(8), intent(in) :: X(:),Y(:),XX(:)
!         integer, intent(in) :: order
!         real(8) :: YY(size(XX))
!         type(vardimspline) :: v
! 
!         call v%config(order,X)
! 
!     end function 

    subroutine cubic_plot(this,x)
        use mod_matlab
        implicit none
        class(cubicspline), intent(in) :: this
        real(rt), intent(in), optional :: x(:)
        ! local
        real(rt), allocatable :: xx(:),yy(:)
        integer :: i,n

        if (present(x)) then
            n = size(x)
            allocate(xx(n))
            xx = x
        else
            n = 100
            allocate(xx(n))
            xx = linspace(this%x(1),this%x(size(this%x)),n)
        end if

        allocate(yy(n))

        do i = 1,n
            yy(i) = this%eval_q(xx(i))
        end do

        call plot(xx,yy,'cubic spline')

    end subroutine

    function cubic_eval_quick(this,x) result(y)
        use mod_matlab
        implicit none
        class(cubicspline), intent(in) :: this
        real(rt), intent(in) :: x
        real(rt) :: y
        ! local
        integer :: i
        real(rt) :: t

        ! Check whether extrapolation
        if (x<this%x(1) .or. x>this%x(size(this%x))) then
            ! Linear
            y = interp1q(this%x,this%coef(:,1),x)
            return
        end if

        ! Get position in grid
        call bsearch(i,x,this%x)

        ! Evaluate spline using Horner's method
        t = (x - this%x(i))/(this%x(i+1)-this%x(i))
        y = this%coef(i,1) + t*(this%coef(i,2) + t*(this%coef(i,3) + t*this%coef(i,4)))

    end function 

    function cubic_eval_stable(this,x) result(y)
        use mod_matlab
        implicit none
        class(cubicspline), intent(in) :: this
        real(rt), intent(in) :: x
        real(rt) :: y
        ! local
        integer :: i
        real(rt) :: t

        ! Check whether extrapolation
        if (x<this%x(1) .or. x>this%x(size(this%x))) then
            ! Linear
            y = interp1q(this%x,this%coef(:,1),x)
            return
        end if


        ! Get position in grid
        call bsearch(i,x,this%x)

        ! Evaluate spline using Horner's method
        t = (x - this%x(i))/(this%x(i+1)-this%x(i))

        ! https://en.wikipedia.org/wiki/Spline_interpolation
        ! Better behaved numerically because all are convex combinations
        y = (1._rt-t)*this%coef(i,1) + t*this%coef(i+1,1) + t*(1._rt-t)*((1._rt-t)*this%a(i) + t*this%b(i))

    end function 
    
    subroutine cubic_construct(this,x,y)
        use mod_matlab
        implicit none
        class(cubicspline), intent(out) :: this
        real(rt), intent(in) :: x(0:),y(0:) ! note for convenience, am indexing from 0 
        ! local
        real(rt), allocatable :: Al(:),Ad(:),Au(:),D(:),dY(:)
        integer :: i,n

        n = size(x)-1

        if (allocated(this%x)) deallocate(this%x)
        if (allocated(this%coef)) deallocate(this%coef)
        if (allocated(this%a)) deallocate(this%a)
        if (allocated(this%b)) deallocate(this%b)

        if (size(y)/=size(x)) stop 'ERROR: cubic_construct: inputs wrongly sized'
        
        allocate(this%x(n+1))
        this%x = x


        allocate(Al(0:n-1),Ad(0:n),Au(0:n-1),D(0:n),Dy(0:n)) ! note lower bounds

! stop 'about to calc coefs'

        Au(0) = 1d0/(x(1)-x(0))
        Ad(0) = 2d0/(x(1)-x(0))
        Dy(0) = 3d0*(y(1)-y(0))/(x(1)-x(0))**2
        do i = 1,n-1
            Au(i) = 1d0/(x(i+1)-x(i))
            Ad(i) = 2d0*(1d0/(x(i+1)-x(i)) + 1d0/(x(i)-x(i-1)))
            Dy(i) = 3d0*((y(i+1)-y(i))/(x(i+1)-x(i))**2 + (y(i)-y(i-1))/(x(i)-x(i-1))**2)
        end do
        Ad(n) = 2d0/(x(n)-x(n-1))
        Dy(n) = 3d0*(y(n)-y(n-1))/(x(n)-x(n-1))**2

        Al = Au
        
        ! Solves the tridiagonal system here 
        ! http://mathworld.wolfram.com/CubicSpline.html
        D(0:n) = mldivide(Al,Ad,Au,Dy) 

        allocate(this%coef(n+1,4)) ! the last row is empty except for the first column, which has y(size(y))
        allocate(this%a(n))
        allocate(this%b(n))

        do i = 0,n-1
            ! this%coef(1+i,1) = y(i)
            ! this%coef(1+i,2) = D(i)
            ! this%coef(1+i,3) = 3d0*(y(i+1)-y(i)) - 2d0*D(i) - D(i+1)
            ! this%coef(1+i,4) = 2d0*(y(i)-y(i+1)) + D(i) + D(i+1)
            this%a(1+i) = D(i)*(x(i+1)-x(i)) - (y(i+1)-y(i))
            this%b(1+i) = -D(i+1)*(x(i+1)-x(i)) + (y(i+1)-y(i))
! print*,'a',a,'b',b
            this%coef(1+i,1) = y(i)
            this%coef(1+i,2) = y(i+1)-y(i) + this%a(1+i)
            this%coef(1+i,3) = this%b(1+i) - 2d0*this%a(1+i)
            this%coef(1+i,4) = this%a(1+i) - this%b(1+i)
            ! this%coef(1+i,1) = y(i)
            !  this%coef(1+i,2) = D(i)*(x(i+1)-x(i))
            ! this%coef(1+i,3) = 3d0*(y(i+1)-y(i)) + (x(i+1)-x(i))*(-2d0*D(i)-D(i+1))
            ! this%coef(1+i,4) = (x(i+1)-x(i))*(D(i+1)-D(i))
            ! this%coef(1+i,4) = 2d0*(y(i)-y(i+1)) + (x(i+1)-x(i))*(D(i)+D(i+1))
        end do
        this%coef(n+1,1) = y(n)

! call disp('this%coef',this%coef)
!          print*,'this%coef',this%coef
! print*,'g'

 !        stop 'checking coefs'



    end subroutine

    ! For a given value of x and knots t, finds the i such that x is in 
    ! [t(i),t(i+1)) and produces the non-zero B-splines from i-k+1,...,i or order k
    ! at x. These are the only non-zero basis functions for this x.
    ! 
    ! The routine, for convenience, returns ilo:=i-k+1 and ihi:=i which
    ! implicitly give the non-zero B-splines in the vector B(ilo:ihi). B itself
    ! may be much larger. 
    !
    ! The B-splines are given recursively by 
    ! 
    !  B_{i,1}(x) =  1[ x in [x_i,x_{i+1})] 
    !  B_{i,k}(x) = \frac{x-x_i}{x_{i+k-1}-x_i} B_{i,k-1}(x) + \frac{x_{i+k} - x}{x_{i+k} - x_{i+1}} B_{i+1,k-1}(x).
    ! 
    ! To use the 
    subroutine bsplinefast(Bout,ilo,ihi,x,k,t) 
        use mod_matlab
        implicit none
        real(rt), intent(in) :: x,t(:)
        integer, intent(in) :: k
        integer, intent(out) :: ilo,ihi
        real(rt), intent(out), target :: Bout(k)
        ! local
        real(rt), pointer :: Bnew(:), B(:)
        integer :: i,j,p

        call bsearch(i,x,t)

        if (.not. (x>=t(i) .and. x<t(i+1))) then
            Bout = 0e0_rt
            return
        end if

        if (k==1) then
            Bout = 1e0_rt
            return
        end if

        do p = 2,k
            
            B(i-p+2:i) => Bout(k-(p-2):k)
            Bnew(i-p+1:i) => Bout(k-(p-1):k)

            if (p==2) then
                B(i) = 1e0_rt
            end if

            ! Keep order
            do j = i-p+1,i 
!             do j = i,i-p+1,-1
                ! Compute B_{j,p} using B_{j,p-1} and B_{j+1,p-1}
                if (j==i) then
                    ! Combine B_{j,p-1}~=0 with the zero term B_{j+1,p-1}
                    Bnew(j) = (x-t(j))/(t(j+p-1)-t(j))*B(j)
                elseif (j==i-p+1) then
                    ! Combine B_{j+1,p-1}~=0 with the zero term B_{j,p-1}
                    Bnew(j) = (1-(x-t(j+1))/(t(j+1+p-1)-t(j+1)))*B(j+1)
                else
                    ! Combine B_{j,p-1}~=0 with B_{j+1,p-1}~=0
                    Bnew(j) = (x-t(j))/(t(j+p-1)-t(j))*B(j) + &
                              (1e0_rt-(x-t(j+1))/(t(j+1+p-1)-t(j+1)))*B(j+1)
                end if
            end do

        end do

        ilo = i-k+1
        ihi = i


    end subroutine 


    ! Evaluates the ith B-spline of a given order k at data x given knots t.
    ! The B-splines are given recursively by 
    ! 
    !  B_{i,1}(x) =  1[ x in [x_i,x_{i+1})] 
    !  B_{i,k}(x) = \frac{x-x_i}{x_{i+k-1}-x_i} B_{i,k-1}(x) + \frac{x_{i+k} - x}{x_{i+k} - x_{i+1}} B_{i+1,k-1}(x).
    ! 
    ! x should be sorted, as I plan to exploit this in the future.
    !
    recursive function bspline(x,i,k,t) result(B)
        implicit none
        real(rt), intent(in) :: x(:),t(:)
        integer, intent(in) :: i,k
        real(rt) :: B(size(x))
        ! local
        integer :: n,ix,nx

        if (k<=0) STOP 'ERROR: bspline: bad input value for k. Must be >= 1'

        n = size(t) 
        nx = size(x)
        
        if (i < 1 .or. i > n-k) then
            B = 0e0_rt
            return
        end if

        if (k == 1) then

            do ix = 1,nx
                if (x(ix)>=t(i) .and. x(ix)<t(i+1)) then
                    B(ix) = 1e0_rt
                else
                    B(ix) = 0e0_rt
                end if
            end do

        else
            
            if (t(i+k-1)==t(i) .and. t(i+k)==t(i+1)) then
                B = 0e0_rt
            elseif (t(i+k-1)==t(i)) then
                B = bspline(x,i+1,k-1,t)*(t(i+k)-x)/(t(i+k)-t(i+1))
            elseif (t(i+k)==t(i+1)) then
                B = bspline(x,i,k-1,t)*(x - t(i))/(t(i+k-1)-t(i)) 
            else
                B = bspline(x,i,k-1,t)*(x - t(i))/(t(i+k-1)-t(i)) + &
                    bspline(x,i+1,k-1,t)*(t(i+k)-x)/(t(i+k)-t(i+1))
            end if

        end if

    end function


!     ! Constructs the coefficients for tensor product pchip routine.
!     ! Note: this precomputes the coeficients (which are slopes) for only one dimension. 
!     subroutine pchip2(pp,X1,X2,Y)
!         implicit none
!         real(rt), dimension(1:), intent(in) :: X1,X2
!         real(rt), dimension(1:,1:), intent(in) :: Y
!         type(spline2), intent(out) :: pp
!         ! real(rt), dimension(size(X1)-1,4,size(X2)), intent(out) :: coef
!         ! local
!         integer :: i
!         type(spline) :: pp_tmp
! 
!         pp%n1 = size(X1)
!         pp%n2 = size(X2)
!         pp%method = 'pchip'
!         pp%d = 4
! 
!         ! Check sizes
!         if (size(Y,1)/=pp%n1) STOP 'size(Y,1) doesn''t agree with size(X1)'
!         if (size(Y,2)/=pp%n2) STOP 'size(Y,2) doesn''t agree with size(X2)'
! 
!         ! Precompute the coefficients for 1 dimension
!         allocate(pp%k1(pp%n1),pp%co1(pp%n1-1,4,pp%n2),pp%k2(pp%n2))
!         pp%k1 = x1
!         pp%k2 = x2
!         do i = 1,pp%n2
!             call pchip(pp_tmp,X1,Y(:,i))
!             pp%co1(:,:,i)= pp_tmp%co(:,:)
!         end do
! 
!     end subroutine pchip2 
! 
!     ! Evaluates a 2-dimensional piecewise polynomial
!     subroutine ppval2_scal(y,x1,x2,pp)
!         use mod_interp, only: search
!         implicit none
!         type(spline2), intent(in) :: pp 
!         real(rt), intent(in) :: x1,x2
!         real(rt), intent(out) :: y
!         !local
!         real(rt) :: del
!         real(rt), allocatable :: yint1(:)
!         integer :: i2,ilo1,id,d !,ilo2
!         type(spline) :: pp_tmp
! 
! !         if (checkSpline(pp)/=0) STOP 'ppval: ERROR: input spline has an error.'     
!         
!         d = pp%d
! 
!         allocate(yint1(pp%n2))
!             
!         ! Find the location of x1 and x2 in their grids
!         call search(ilo1,x1,pp%k1)
! !         call search(ilo2,x2,pp%k2)
! 
!         ! A spline has been computed for y(.,x2) for each x2 on the grid.
!         ! So evaluate the spline at (xi1,x2) for each xi1, x2, to get 
!         ! yhat(xi1,x2) for each xi1 and x2 on the grid.
! 
!         ! Evaluate the polynomial using Horner's method
!         del = x1 - pp%k1(ilo1)
!         do i2 = 1,pp%n2
!             yint1(i2) = pp%co1(ilo1,d-1,i2) + pp%co1(ilo1,d,i2)*del
!             do id = d-2,1,-1
!                 yint1(i2) = pp%co1(ilo1,id,i2) + yint1(i2)*del
!             end do
!         end do
! 
!         ! Now have yhat(xi1,x2). Construct one last spline for (.,yhat(xi1,.)) and 
!         ! evaluate it. 
!         if (pp%method=='pchip') then
!             ! pchip is mostly a local method, so this could be made more efficient
!             ! Would be better to construct a "pchipint" that would be of use also
!             ! in the one dimensional case, constructing just the local info.
!             call pchip(pp_tmp,pp%k2,yint1)
!             call ppval(y,x2,pp_tmp)
!         else
!             STOP 'ppval2: ERROR: currently, only pchip is an option'
!         end if
! 
!     end subroutine ppval2_scal

!     ! Evaluates a 2-dimensional piecewise polynomial
!     ! This is a "vectorized" version where x1 and x2 are vectors, the pp
!     ! is evaluted at the tensor product of x1 and x2. There is computational savings
!     ! in the tensor approach because, for #x1=#x2=n, there are only n spline 
!     ! constructions rather than n**2 evaluations.
!     subroutine ppval2_vec(y,x1,x2,pp)
!         use mod_interp, only: search
!         implicit none
!         type(spline2), intent(in) :: pp 
!         real(rt), dimension(1:), intent(in) :: x1,x2
!         real(rt), dimension(1:size(x1),1:size(x2)), intent(out) :: y
!         !local
!         real(rt) :: del
!         real(rt), allocatable :: yint1(:)
!         integer :: i2,ilo1,id,d, nx1,nx2, i1 
!         type(spline) :: pp_tmp
! 
! !         if (checkSpline(pp)/=0) STOP 'ppval: ERROR: input spline has an error.'     
!         
!         nx1 = size(x1)
!         nx2 = size(x2)
! 
!         d = pp%d
! 
!         allocate(yint1(pp%n2))
!             
!         ! Loop over the x1 values
!         do i1 = 1,nx1
! 
!             ! Find the location of x1 in the 1st dim grid
!             call search(ilo1,x1(i1),pp%k1)
! 
!             ! A spline has been computed for y(.,x2) for each x2 on the grid.
!             ! So evaluate the spline at (xi1,x2) for each xi1, x2, to get 
!             ! yhat(xi1,x2) for each xi1 and x2 on the grid.
! 
!             ! Evaluate the polynomial using Horner's method
!             del = x1(i1) - pp%k1(ilo1)
!             do i2 = 1,pp%n2
!                 yint1(i2) = pp%co1(ilo1,d-1,i2) + pp%co1(ilo1,d,i2)*del
!                 do id = d-2,1,-1
!                     yint1(i2) = pp%co1(ilo1,id,i2) + yint1(i2)*del
!                 end do
!             end do
! 
!             ! Now have yhat(xi1,x2). Construct one last spline for (.,yhat(xi1,.)) and 
!             ! evaluate it. 
!             if (pp%method=='pchip') then
!                 ! pchip is mostly a local method, so this could be made more efficient
!                 ! Would be better to construct a "pchipint" that would be of use also
!                 ! in the one dimensional case, constructing just the local info.
!                 call pchip(pp_tmp,pp%k2,yint1)
!                 call ppval_vec(y(i1,:),x2,pp_tmp) ! This is the vectorized version
!             else
!                 STOP 'ppval2: ERROR: currently, only pchip is an option'
!             end if
! 
!         end do
! 
!     end subroutine ppval2_vec


    ! freeSpline
    ! Free the spline memory
    subroutine freeSpline(pp) 
        implicit none
        type(spline) :: pp ! this deallocates the information
        if (allocated(pp%k)) deallocate(pp%k)
        if (allocated(pp%co)) deallocate(pp%co)
        pp%n = -1 ! this initializes to the default state 
        pp%d = -1 ! this initializes to the default state  
        pp%method = '' ! clears the method
    end subroutine

    ! checkSpline
    ! 
    !  0 == allocated and dimensions agree
    ! -1 == not allocated
    ! -2 == allocated but dimensions don't agree
    function checkSpline(pp) result(flag)
        implicit none
        type(spline), intent(in) :: pp
        integer :: flag

        if (allocated(pp%k) .and. allocated(pp%co)) then
            if (pp%n==size(pp%k) .and. pp%n-1==size(pp%co,1) &
                .and. pp%d==size(pp%co,2) .and. pp%d>0 .and. pp%n>1) then
                flag = 0
                return
            else
                flag = -2
                return
            end if
        else
            flag = -1
            return
        end if

    end function

    ! Given a piecewise polynomial pp, evaluates to recover values y.
    ! y(x) = c(i,1) + c(i,2)(x-xi) + c(i,3)(x-xi)**2 + ... if x \in [ki,ki+1]
    subroutine ppval_vec(y,x,pp) 
        use mod_interp, only: search
        implicit none
        type(spline), intent(in) :: pp 
        real(rt), intent(in) :: x(1:)
        real(rt), intent(out) :: y(1:size(x))
        !local
        real(rt) :: del
        integer :: ilo,ix,id,nx,d

        if (checkSpline(pp)/=0) STOP 'ppval: ERROR: input spline has an error.'     
        
        d = pp%d
        nx = size(x)
        
        ! Loop over each value of x
        do ix = 1,nx
            
            ! Find the location of x in grid k
            call search(ilo,x(ix),pp%k)
            del = x(ix)-pp%k(ilo)

            ! Evaluate the polynomial using Horner's method
            y(ix) = pp%co(ilo,d-1) + pp%co(ilo,d)*del
            do id = d-2,1,-1
                y(ix) = pp%co(ilo,id) + y(ix)*del
            end do

        end do

    end subroutine ppval_vec

    ! Given input x,k,c where k is knots, c is coefficients, and x is data,
    ! evaluate the spline defined by k,co at x to recover values y.
    ! y(x) = c(i,1) + c(i,2)(x-xi) + c(i,3)(x-xi)**2 + ... if x \in [ki,ki+1]
    subroutine ppval_scal(y,x,pp,ilo_in,ilo_out) 
        use mod_interp, only: search
        implicit none
        type(spline), intent(in) :: pp 
        real(rt), intent(in) :: x
        real(rt), intent(out) :: y
        integer, intent(in), optional :: ilo_in
        integer, intent(out), optional :: ilo_out
        !local
        real(rt) :: del
        integer :: ilo,id,d

        if (checkSpline(pp)/=0) STOP 'ppval: ERROR: input spline has an error.'     
        
        d = pp%d
            
        ! Find the location of x in grid k
        if (present(ilo_in)) then
            ilo = ilo_in
            ! Verify that x is in ilo,ilo+1
            if (x<pp%k(ilo) .or. x>pp%k(ilo+1)) then
                STOP 'ppval: ERROR: x is not in x(ilo,ilo+1)'
            end if
        else
            call search(ilo,x,pp%k)
        end if
        if (present(ilo_out)) then
            ilo_out = ilo
        end if

        ! Evaluate the polynomial using Horner's method
        del = x - pp%k(ilo)
        y = pp%co(ilo,d-1) + pp%co(ilo,d)*del
        do id = d-2,1,-1
            y = pp%co(ilo,id) + y*del
        end do

    end subroutine ppval_scal

    ! Given input x,k,c where k is knots, c is coefficients, and x is data,
    ! evaluate the spline defined by k,co at x to recover values y.
    ! y(x) = c(i,1) + c(i,2)(x-xi) + c(i,3)(x-xi)**2 + ... if x \in [ki,ki+1]
    subroutine ppderiv(yp,ypp,x,pp) 
        use mod_interp, only: search
        implicit none
        real(rt), intent(in) :: x(1:)
        type(spline), intent(in) :: pp
        real(rt), intent(out) :: yp(1:size(x)),ypp(1:size(x))
        !local
        real(rt) :: del
        integer :: ilo,ix,id,nx,d

        if (checkSpline(pp)/=0) STOP 'ppval: ERROR: input spline has an error.'     
        
        d = pp%d
        nx = size(x)
        
        ! Loop over each value of x
        do ix = 1,nx
            
            ! Find the location of x in grid k
            call search(ilo,x(ix),pp%k)
            del = x(ix)-pp%k(ilo)

            ! Evaluate the first derivative using Horner's method
            yp(ix) = real(d-1,rt)*pp%co(ilo,d)
            do id = d-1,2,-1
                yp(ix) = real(id-1,rt)*pp%co(ilo,id) + yp(ix)*del
            end do

            ! Evaluate the second derivative using Horner's method
            ypp(ix) = real(d-2,rt)*real(d-1,rt)*pp%co(ilo,d) ! This evaluates to zero if d is 1 or 2
            do id = d-1,3,-1
                ypp(ix) = real(id-2,rt)*real(id-1,rt)*pp%co(ilo,id) + ypp(ix)*del
            end do

        end do

    end subroutine ppderiv

    function schumaker_interp_vec(x,y,xp) result(yp)
        implicit none
        real(rt), dimension(:), intent(in) :: x,y,xp
        real(rt), dimension(size(xp)) :: yp
        type(spline) :: pp

        call schumaker(pp,x,y)
        call ppval(yp,xp,pp)

    end function
    function schumaker_interp_scal(x,y,xp) result(yp)
        implicit none
        real(rt), dimension(:), intent(in) :: x,y
        real(rt) :: xp
        real(rt) :: yp
        type(spline) :: pp

        call schumaker(pp,x,y)
        call ppval(yp,xp,pp)

    end function

    ! Produce a Schumaker shape-preserving spline 
    ! Input data is either Lagrange data x,y or Hermite data x,y,s where s is slope
    subroutine schumaker(pp,x,y,s)
        use mod_core
        implicit none
        real(rt), dimension(1:), intent(in) :: x,y
        real(rt), dimension(1:), intent(in), optional, target :: s
        type(spline), intent(out) :: pp
        !local
        real(rt), dimension(:), allocatable :: yp,dx,dy,m,L,Lm,dyp
        logical, dimension(:), allocatable :: specialCase
        real(rt) :: xsi,alpha,beta,ypbar,yplo_minus_m,yphi_minus_m,theta
        real(rt), parameter :: eps = epsilon(0e0_rt), sqrteps = sqrt(eps)
        integer :: i,n,j,nk

        pp%method = 'schumaker'
        pp%d = 3
        
        n = size(x)

        ! Check inputs
        if (size(y)/=n) STOP 'error: y and x do not have the same size'
!         if (any(isNaN(y))) then
!             print*,'y',y
!             STOP 'y is NaN'
!         end if

        ! Allocate space for local variables
        allocate(yp(n),dx(n-1),dy(n-1),m(n-1),L(n-1),Lm(n-1),dyp(n-1),specialCase(n-1))

        dx = x(2:n)-x(1:n-1)
        dy = y(2:n)-y(1:n-1)
        m = dy/dx

        ! Construct or copy the spline 
        if (present(s)) then
            ! Copy by assigning pointer
            if (size(s)/=n) STOP 'error: s and x do not have the same size'
            yp = s
        else

            ! Construct the slopes
            ! L = sqrt(dx**2 + dy**2)
            do i = 1,n-1
                L(i) = norm((/dx(i),dy(i)/))
            end do

            Lm = L*m
            do i = 2,n-1
                if ((m(i-1)>0e0_rt .and. m(i)>0e0_rt) .or. (m(i-1)<0e0_rt .and. m(i)<0e0_rt)) then
                    yp(i) = (Lm(i-1)+Lm(i))/(L(i-1)+L(i))
                else
                    yp(i) = 0e0_rt
                end if
            !    print*,(Lm(i-1)+Lm(i))/(L(i-1)+L(i))
            end do
            yp(1) = (3e0_rt*m(1) - yp(2))/2e0_rt
            yp(n) = (3e0_rt*m(n-1) - yp(n-1))/2e0_rt

        end if

        dyp = yp(2:n)-yp(1:n-1)
        specialCase = abs((yp(1:n-1)+yp(2:n))/2e0_rt - m)<4e0_rt*eps

        ! Get the number of knots - 1
        nk = 2*(n-1) - count(specialCase)

        pp%n = nk + 1
        
        ! Allocate space for output arguments
        allocate(pp%k(nk+1),pp%co(nk,3))

        ! Loop over intervals and compute coeffs for each
        j = 1
        do i = 1,n-1
        
            ! Check whether the special case where a quadratic polynomial works is satisfied
!             if (specialCase(i)) then
            if (specialCase(i)) then
                ! If so, store the coeffs and knots
                pp%co(j,1) = y(i)
                pp%co(j,2) = yp(i)
                pp%co(j,3) = dyp(i)/dx(i)/2e0_rt
                pp%k(j) = x(i)
                j = j+1

            else

                yplo_minus_m = yp(i)-m(i)
                yphi_minus_m = yp(i+1)-m(i)

                ! Note yplo_minus_m = yphi_minus_

                ! Test whether we can ensure xsi will be in (x(i),x(i+1))
                ! For this, ened |y'_j-m| \in (0, (|y'_{i+1}-y'_i|)^-1) for j = i,i+1
                

                ! Otherwise in the typical case where a quad poly doesn't work, insert a knot
                ! Three cases which will give the knot
                if (abs(yp(i+1)-yp(i))<epsilon(0e0_rt) .or. &
                    (yplo_minus_m<epsilon(0e0_rt) .and. yphi_minus_m<epsilon(0e0_rt)) .or. &
                    (yplo_minus_m>-epsilon(0e0_rt) .and. yphi_minus_m>-epsilon(0e0_rt))) then
!                 if ((yplo_minus_m<1d2*sqrteps .and. yphi_minus_m<=1d2*sqrteps) .or. (yplo_minus_m>=-1d2*sqrteps .and. yphi_minus_m>=-1d2*sqrteps)) then
                    ! Cannot shape preserve
                    xsi = (x(i)+x(i+1))/2e0_rt
                elseif (abs(yplo_minus_m)>abs(yphi_minus_m)) then
!                     xsibar = x(i) + 2e0_rt*dx(i)*yphi_minus_m(i)/dyp(i) 
!                     xsi = (x(i) + xsibar)/2e0_rt
!                     xsi = x(i) + dx(i)*yphi_minus_m/dyp(i) 

                    ! theta*dx will be the step size. we need to ensure this is not too close to 0 or 1
                    theta = yphi_minus_m/dyp(i)  ! < 0
                    theta = min(max(theta,sqrteps),1e0_rt-sqrteps)
                    xsi = x(i) + dx(i)*theta
                else 
!                     xsibar = x(i+1) + 2e0_rt*dx(i)*yplo_minus_m(i)/dyp(i)
!                     xsi = (x(i+1) + xsibar)/2e0_rt
!                     xsi = x(i+1) + dx(i)*yplo_minus_m/dyp(i)
                    theta = yphi_minus_m/dyp(i)  ! < 0
                    theta = min(max(theta,sqrteps),1e0_rt-sqrteps) 
                    theta = theta-1e0_rt
                    xsi = x(i+1) + dx(i)*theta
                end if

if (xsi<=x(i) .or. xsi>=x(i+1)) then
    print*,'spec case criterion',abs( (yp(i)+yp(i+1))/2e0_rt - m(i) )
    print*,'xsi',xsi
    print*,'x(i:i+1)',x(i:i+1)
    print*,'dyp(i)',dyp(i)
    print*,'dx(i)',dx(i)
    print*,'yphi_minus_m',yphi_minus_m
    print*,'yplo_minus_m',yplo_minus_m
    STOP 'xsi is not between x(i) and x(i+1)'
end if

                ! With the knot, there is a unique quadratic polynomial
                alpha = xsi - x(i)
                beta = x(i+1) - xsi
                ypbar = 2e0_rt*m(i) - (alpha*yp(i) + beta*yp(i+1))/dx(i)
                pp%co(j,1) = y(i)
                pp%co(j,2) = yp(i)

                pp%co(j,3) = (ypbar - yp(i))/(2e0_rt*alpha)
                pp%k(j) = x(i)
                pp%co(j+1,1) = pp%co(j,1) + alpha*(pp%co(j,2) + alpha*pp%co(j,3))
                pp%co(j+1,2) = ypbar
                pp%co(j+1,3) = (yp(i+1)-ypbar)/(2e0_rt*beta)
                pp%k(j+1) = xsi

!                 if (any(pp%co(j+1,:)>huge(0e0_rt) .or. pp%co(j,:)>huge(0e0_rt)) &
!                     .or. any(isNaN(pp%co(j+1,:)) .or. isNaN(pp%co(j,:))) &
!                     .or. max(pp%co(j+1,3),pp%co(j,3))>.5e0_rt*huge(0e0_rt)) then
                if (any(pp%co(j+1,:)>huge(0e0_rt) .or. pp%co(j,:)>huge(0e0_rt)) &
                    .or. max(pp%co(j+1,3),pp%co(j,3))>.5e0_rt*huge(0e0_rt)) then
                    print*,'xsi',xsi
                    print*,'x(i)',x(i),'x(i+1)',x(i+1)
                    print*,'y(i)',y(i),'y(i+1)',y(i+1)
                    print*,'dx(i)',dx(i)
                    print*,'dy(i)',dy(i)
                    print*,'m(i)',m(i)
                    print*,'L(i)',L(i)
                    print*,'c(j,1)',pp%co(j,1),'c(j+1,1)',pp%co(j+1,1)
                    print*,'c(j,2)',pp%co(j,2),'c(j+1,2)',pp%co(j+1,2)
                    print*,'c(j,3)',pp%co(j,3),'c(j+1,3)',pp%co(j+1,3)
                    print*,'yp(i)',yp(i),'yp(i+1)',yp(i+1)
                    print*,'yplo_minus_m',yplo_minus_m,'yphi_minus_m',yphi_minus_m
                    print*,'dx',dx(i)
                    print*,'ypbar',ypbar
                    print*,'alpha',alpha
                    print*,'beta',beta
                    print*,'c(j,1)+alpha(c(j,2)+alpha*c(j,3))',pp%co(j,1) + alpha*(pp%co(j,2) + alpha*pp%co(j,3))
                    print*,'c(j,1),alpha*c(j,2),alpha**2*c(j,3)',pp%co(j,1) , alpha*pp%co(j,2) , alpha**2*pp%co(j,3)
                    print*,'c(j,1)+alpha*c(j,2),alpha**2*c(j,3)',pp%co(j,1) + alpha*pp%co(j,2) , alpha**2*pp%co(j,3)
                    print*,'c(j,1),alpha*c(j,2)+alpha**2*c(j,3)',pp%co(j,1) , alpha*pp%co(j,2) + alpha**2*pp%co(j,3)
                    print*,'c(j,1)+alpha*c(j,2)+alpha**2*c(j,3)',pp%co(j,1) + alpha*pp%co(j,2) + alpha**2*pp%co(j,3)
                    print*,'j+1',j+1,'size(c,1)',size(pp%co,1)
                    STOP 'c is inf'
                end if

                j = j + 2

            end if
        end do

        ! Finalize k by adding the last point
        pp%k(j) = x(n)
       
        ! Deallocate space for local variables
        deallocate(dx,dy,m,L,Lm,dyp,specialCase)

        ! Deallocate space for yp if it was allocated
        if (.not. present(s)) deallocate(yp)

    end subroutine schumaker

    ! Computes the euclidean norm of X
    function norm(X) result(l2)
        implicit none
        real(rt), dimension(1:), intent(in) :: X
        real(rt) :: l2
        
        ! Call the minpack version of enorm
        l2 = sqrt(dot_product(X,X))
!         l2 = enorm(size(X),X)

!     contains
! 
!     DOUBLE PRECISION FUNCTION ENORM(N,X)
!     IMPLICIT NONE
!     INTEGER N
!     DOUBLE PRECISION X(N)
!     !     **********
!     !
!     !     function enorm
!     !
!     !     given an n-vector x, this function calculates the
!     !     euclidean norm of x.
!     !
!     !     the euclidean norm is computed by accumulating the sum of
!     !     squares in three different sums. the sums of squares for the
!     !     small and large components are scaled so that no overflows
!     !     occur. non-destructive underflows are permitted. underflows
!     !     and overflows do not occur in the computation of the unscaled
!     !     sum of squares for the intermediate components.
!     !     the definitions of small, intermediate and large components
!     !     depend on two constants, rdwarf and rgiant. the main
!     !     restrictions on these constants are that rdwarf**2 not
!     !     underflow and rgiant**2 not overflow. the constants
!     !     given here are suitable for every known computer.
!     !
!     !     the function statement is
!     !
!     !       double precision function enorm(n,x)
!     !
!     !     where
!     !
!     !       n is a positive integer input variable.
!     !
!     !       x is an input array of length n.
!     !
!     !     subprograms called
!     !
!     !       fortran-supplied ... dabs,dsqrt
!     !
!     !     argonne national laboratory. minpack project. march 1980.
!     !     burton s. garbow, kenneth e. hillstrom, jorge j. more
!     !
!     !     **********
!     INTEGER i
!     DOUBLE PRECISION agiant , floatn , one , rdwarf , rgiant , s1 , s2 , s3 , xabs , x1max , x3max , zero
!     DATA one , zero , rdwarf , rgiant/1.0D0 , 0.0D0 , 3.834D-20 ,  1.304D19/
!     s1 = zero
!     s2 = zero
!     s3 = zero
!     x1max = zero
!     x3max = zero
!     floatn = N
!     agiant = rgiant/floatn
!     DO i = 1 , N
!         xabs = DABS(X(i))
!         IF ( xabs.GT.rdwarf .AND. xabs.LT.agiant ) THEN
!         !
!         !           sum for intermediate components.
!         !
!             s2 = s2 + xabs**2
!         ELSEIF ( xabs.LE.rdwarf ) THEN
!         !
!         !              sum for small components.
!         !
!             IF ( xabs.LE.x3max ) THEN
!                 IF ( xabs.NE.zero ) s3 = s3 + (xabs/x3max)**2
!             ELSE
!                 s3 = one + s3*(x3max/xabs)**2
!                 x3max = xabs
!             ENDIF
!         !
!         !              sum for large components.
!         !
!         ELSEIF ( xabs.LE.x1max ) THEN
!             s1 = s1 + (xabs/x1max)**2
!         ELSE
!             s1 = one + s1*(x1max/xabs)**2
!             x1max = xabs
!         ENDIF
!     ENDDO
!     !
!     !     calculation of norm.
!     !
!     IF ( s1.NE.zero ) THEN
!         ENORM = x1max*DSQRT(s1+(s2/x1max)/x1max)
!     ELSEIF ( s2.EQ.zero ) THEN
!         ENORM = x3max*DSQRT(s3)
!     ELSE
!         IF ( s2.GE.x3max ) ENORM = DSQRT(s2*(one+(x3max/s2)*(x3max*s3)))
!         IF ( s2.LT.x3max ) ENORM = DSQRT(x3max*((s2/x3max)+(x3max*s3)))
!     ENDIF
!     !
!     !     last card of function enorm.
!     !
!     END function
! 

    end function 


!     ! Piecewise cubic hermite interpolating polynomial
!     ! Preserves montonicity 
!     subroutine pchip_interp(Yi,X,Y,Xi)
!         implicit none
!         real(rt), dimension(1:), intent(in) :: X,Y,Xi
!         real(rt), dimension(1:size(Xi)), intent(out) :: Yi
!         ! local
!         type(spline) :: pp 
! 
!         ! Get pp form of pchip
!         call pchip_getpp(pp,X,Y)
! 
!         ! Eval at Xi
!         call ppval(Yi,Xi,pp)
! 
!     end subroutine 
! 
!     subroutine pchip_getpp(pp,X,Y)
!         implicit none
!         type(spline), intent(out) :: pp ! intent(out) should deallocate all the components
!         real(rt), dimension(1:), intent(in) :: X,Y
!         ! local
!         real(rt), allocatable :: s(:)
!         real(rt) :: h
!         integer :: n,ierr,i
! 
!         n = size(X)
! 
!         pp%method = 'pchip'
!         pp%d = 4
! 
!         if (size(Y)/=n) STOP 'pchip: ERROR: X and Y dims don''t agree'
! 
!         pp%n = n
!         allocate(pp%k(n),pp%co(n-1,4))
!         allocate(s(n))
! 
!         ! Knots are just the X values
!         pp%k = X
! 
!         ! Coefficients: 
!         pp%co(:,1) = Y(1:n-1)
!         call dpchim(n,X,Y,s,1,ierr) ! Gets the slopes at each end point
!         pp%co(:,2) = s(1:n-1)
!         do i = 1,n-1
!             h = x(i+1) - x(i)
!             if (h<=0e0_rt) STOP 'pchip: ERROR: X is not increasing' 
!             pp%co(i,3) = - (s(i+1) + 2e0_rt*pp%co(i,2))/h + 3e0_rt*(Y(i+1)-pp%co(i,1))/(h**2)
!             pp%co(i,4) = (s(i+1) - 2e0_rt*pp%co(i,3)*h - pp%co(i,2))/(3e0_rt*h**2)
!         end do
! 
!         deallocate(s)
!     end subroutine

end module mod_spline

module mod_spline_test 
    use mod_types
    use mod_spline
    implicit none

contains

    real(rt) function f1(x)
        implicit none
        real(rt), intent(in) :: x
        f1 = sin(x)
    end function 
    real(rt) function f2(x)
        implicit none
        real(rt), intent(in) :: x
        f2 = sqrt(x)
    end function 

    subroutine test_cubic()
        use mod_spline
        use mod_matlab
        implicit none
        type(cubicspline) :: a
        real(rt), allocatable :: x(:),y(:)
        real(rt) :: lb,ub
        integer :: i

        lb = 1e-6_rt
        ub = 10d0

        allocate(x(250))
        call random_number(x)
        x = sort(x)
        x = lb + (ub-lb)*x 
        x(1) = lb
        x(size(x)) = ub


        call plot(x,sin(x),'Linear interp')

!         call a%construct([-1._rt,0._rt,3._rt],[.5_rt,0._rt,3._rt])
        call a%construct(x,sin(x))

! stop 'here'
        deallocate(x)
        allocate(x(1000),y(1000))
        x = linspace(lb,ub*1.1_rt,1000)

        call a%plot(x)

        do i = 1,size(x)
            y(i) = a%eval_s(x(i))
        end do

        call plot(x,y,'Cubic Spline, Stable Evaluation')


    end subroutine

    subroutine test_vardim()
        use mod_spline
        implicit none
        type(vardimspline) :: a

        call a%test(f1,[-2e0_rt,-1.5e0_rt,-1e0_rt,-.5e0_rt,1e0_rt,3e0_rt],1)
        call a%test(f1,[-2e0_rt,-1.5e0_rt,-1e0_rt,-.5e0_rt,1e0_rt,3e0_rt],2)
        call a%test(f1,[-2e0_rt,-1.5e0_rt,-1e0_rt,-.5e0_rt,1e0_rt,3e0_rt],3)
        call a%test(f1,[-2e0_rt,-1.5e0_rt,-1e0_rt,-.5e0_rt,1e0_rt,3e0_rt],4)
        call a%test(f2,[0e0_rt,.5e0_rt,1e0_rt,1.5e0_rt,2e0_rt,3e0_rt],1)
        call a%test(f2,[0e0_rt,.5e0_rt,1e0_rt,1.5e0_rt,2e0_rt,3e0_rt],2)
        call a%test(f2,[0e0_rt,.5e0_rt,1e0_rt,1.5e0_rt,2e0_rt,3e0_rt],3)
        call a%test(f2,[0e0_rt,.5e0_rt,1e0_rt,1.5e0_rt,2e0_rt,3e0_rt],4)

    end subroutine

    subroutine test_pchip()
        use mod_util
        use mod_matlab
        implicit none
        ! local
        type(type_pchip) :: p
        real(rt), allocatable :: x(:),y(:),yp(:),ypp(:),y2(:)
        integer :: i,n

        n = 20
        allocate(x(n),y(n),y2(n),yp(n),ypp(n))

        ! Check that it gets the linear function exactly right
        x = linspace(-1._rt,1._rt,n)
        y = 1._rt + x ! + .5_rt*x**2

        call p%construct(x,y)

        do i = 1,n
            call p%deriv(x(i),y2(i),yp(i),ypp(i))
            call assert(abs(y2(i)-y(i))<1d-6,'test_pchip: failed evaluation test')
            call assert(abs(-yp(i)+1._rt)<1d-6,'test_pchip: failed first derivative test')
            call assert(abs(-ypp(i))<1d-6,'test_pchip: failed second derivative test')
        end do

        ! Also check a less trivial function
        x = linspace(-10._rt,100._rt,n)
        y = max(10._rt,x,.02_rt*x**2)

        call p%construct_nomono(x,y)

        call p%plot(derivorder=2)

        print*,'test_pchip: passed'

    end subroutine 

!     subroutine test_spline()
!         use mod_matlab, only: plot,tic,toc,cumsum,linspace
!         use mod_spline
!         implicit none
!         type(spline) :: pp
!         type(spline2) :: pp2
!         real(rt) :: X(6),Y(6),X2(8),V(6,8),Vcheck(6,8)
!         real(rt) :: Xi(100),Yi(100),Ypi(100),Yppi(100),Xi2(100),Vi(100),Vitrue(100)
!         real(rt), allocatable :: tmp2(:,:)
!         integer :: i,j
! 
!         ! Test the interpolation properties of pchip
!         call random_number(Xi) 
!         Xi = cumsum(Xi+1d-6) ! xi must be strictly increasing
!         call random_number(Yi)
!         Yi = 1000e0_rt*Yi + log(Xi)
! 
!         ! Test the interpolation properties of pchip
!         call pchip(pp,Xi,Yi)
!         call ppval(Ypi,Xi,pp)
!         if (maxval(abs(Yi-Ypi))>1d3*epsilon(0e0_rt)) then
!             print*,'maxval(abs(Yi-Ypi))',maxval(abs(Yi-Ypi))
!             STOP 'test_spline: failed test 1'
!         end if
! 
!         ! Same for schumaker
!         call schumaker(pp,Xi,Yi)
!         call ppval(Ypi,Xi,pp)
!         if (maxval(abs(Yi-Ypi))>1d3*epsilon(0e0_rt)) STOP 'test_spline: failed test 2'
! 
!         ! Now test that linear functions can be recovered
!         Yi = -100e0_rt + .05e0_rt*Xi
!         call pchip(pp,Xi,Yi)
!         call ppval(Ypi,Xi,pp)
!         if (maxval(abs(Yi-Ypi))>epsilon(0e0_rt)) STOP 'test_spline: failed test 3'
! 
!         call schumaker(pp,Xi,Yi)
!         call ppval(Ypi,Xi,pp)
!         if (maxval(abs(Yi-Ypi))>epsilon(0e0_rt)) STOP 'test_spline: failed test 4'
! 
!         ! Now test that quadratic functions can be recovered
!         Yi = -100e0_rt + .05e0_rt*Xi - 6e0_rt*Xi**2
!         call pchip(pp,Xi,Yi)
!         call ppval(Ypi,Xi,pp)
!         if (maxval(abs(Yi-Ypi))>epsilon(0e0_rt)) STOP 'test_spline: failed test 5'
! 
!         call schumaker(pp,Xi,Yi)
!         call ppval(Ypi,Xi,pp)
!         if (maxval(abs(Yi-Ypi))>epsilon(0e0_rt)) STOP 'test_spline: failed test 6'
! 
!         ! Now test additional
! 
!         X = (/0e0_rt,.1e0_rt,.5e0_rt,1e0_rt,2e0_rt,10e0_rt/)
!         X2 =(/.0001e0_rt,1e0_rt,2e0_rt,3e0_rt,4e0_rt,7e0_rt,15d0,20d0/)
!         Y = sqrt(X)
! 
! 
!         call pchip(pp,X,Y)
! 
!         Xi = linspace(X(1),X(size(X)),size(Xi))
!         Xi2 = linspace(X2(1),X2(size(X2)),size(Xi))
! 
!         ! Test one dimensional
! !         call sub_pchip_eval(Yi,Xi,co,X,Y)
! !         call plot(Xi,Yi)
! 
! !         call sub_pchip_deriv(Ypi,Yi,Xi,co,X,Y)
! !         call plot(Xi,Yi)
! !         call plot(Xi,Ypi)
! 
! !         call myprint('X',X)
! !         call myprint('Y',Y)
! !         call myprint('Yi',Yi)
! !         call sub_pchip_int(yi,xi,X,Y)
! 
!         call pchip(pp,x,y)
!         call ppval(yi,xi,pp)
! ! call plot(x,y)        
! ! call plot(xi,yi,title='pchip')
!         call ppderiv(ypi,yppi,xi,pp)
! ! call plot(xi,ypi,title='pchip, deriv')
! ! call plot(xi,yppi,title='pchip, second deriv')
!         call schumaker(pp,x,y)
!         call ppval(yi,xi,pp)
! ! call plot(xi,yi,title='schumaker')
!         call ppderiv(ypi,yppi,xi,pp)
! ! call plot(xi,ypi,title='schumaker, deriv')
! ! call plot(xi,yppi,title='schumaker, second deriv')
! ! STOP 'plotting'
! 
!         ! Test two dimensional
!         do j = 1,8
!             do i = 1,6
!                 V(i,j) = sqrt(X(i)) - 10e0_rt*sqrt(X2(j))  - X(i)*X2(j)
!             end do
!         end do
!         Vitrue = sqrt(Xi) - 10e0_rt*sqrt(Xi2) - Xi*Xi2
! 
! 
!         call pchip2(pp2,X,X2,V)
! 
!         ! Check the vectorized version
!         call ppval2(Vcheck,X,X2,pp2)
! 
!         if (maxval(abs(Vcheck-V))>1000e0_rt*epsilon(0e0_rt)) then
!             print*,'maxval(abs(Vcheck-V))',maxval(abs(Vcheck-V))
!             STOP 'test_pchip: ERROR: failed test: ppval2 doesn''t interpolate'
!         end if
! 
!         do i = 1,size(xi2)
!             call ppval2(Vi(i),Xi(i),Xi2(i),pp2)
!         end do
! 
!         ! Can compare this with the matlab "test_pchip2.m" file. Good on 1/6/2013
! !         do i = 1,size(xi2)
! !             print*, Vi(i),Vitrue(i)
! !         end do
! !     STOP 'here'
! 
! 
!         if (maxval(abs(Vi-Vitrue)/abs(Vitrue))>.46e0_rt) then
!             print*,'Max relative error Vi,Vitrue',maxval(abs(Vi-Vitrue)/abs(Vitrue))
!             STOP 'test_pchip: ERROR: failed test1: spline error larger than expected'
!         end if
!         
! ! call plot(Vi)
! ! call plot(Vitrue)
! 
! !         STOP 'checking ppval2'
! 
! 
! ! call tic
! ! do i = 1,10000
! !         call sub_pchip2_coef(co2,X,X2,V)
! ! end do
! ! call toc
! ! do i = 1,10000
! !         call sub_pchip2_eval(Vi,xi,xi2,co2,X,X2,V)
! ! end do
! ! call toc
! !end do 
! !         call myprint('V',V)
! !         call myprint('co2',co2)
! !         call myprint('X2',X2)
! !         call myprint('Xi2',xi2)
! !         call myprint('Xi',xi)
! !         call myprint('Vi',Vi)
! !         call myprint('Vitrue',Vitrue)
! 
!         ! Test extreme values, 1 dim routine
!         call pchip(pp,[-2e0_rt,-1e0_rt,0e0_rt,1e0_rt,2e0_rt],[-1d150,-1e0_rt,0d0,1d0,1d150])
!         Xi = linspace(-2e0_rt,2e0_rt,size(Yi))
!         call ppval(Yi,Xi,pp)
!         call plot(Xi,Yi)
! 
!         call pchip2(pp2,[-2e0_rt,-1e0_rt,0e0_rt],[-2e0_rt,-1e0_rt,0e0_rt],reshape(&
!         (/ -2e0_rt, -10e0_rt, -1d200, &
!            -1e0_rt,  -5e0_rt, -200e0_rt, &
!             0e0_rt,  -1e0_rt,  -50e0_rt/)&
!            ,(/3,3/)))
!         Xi = linspace(-2e0_rt,0e0_rt,size(Xi))
!         Yi = linspace(-2e0_rt,0e0_rt,size(Yi))
!         allocate(tmp2(size(Xi),size(Yi)))
!         call ppval2(tmp2,Xi,Yi,pp2)
!         print*,'maxval(tmp2)',maxval(tmp2)
!         
!         call plot(Xi,tmp2)
! 
!         print*,'Yi:',Yi
!         STOP 'testing spline extreme'
! 
!     end subroutine
end module 
module mod_min1dim
    use mod_types
    implicit none

    interface sub_gridsearch1dim
        module procedure sub_gridsearch1dim_nogrid, sub_gridsearch1dim_grid
    end interface

    type type_brentrc
        integer :: flag
        integer :: istat ! should be set to 0 by user in the first call
        real(rt) :: a,b,x,fx
        real(rt) :: xtoler = 1e-6_rt
        integer :: printmod = -1
        real(rt) :: xsave,fxsave,a_in,b_in,t,u,v,w,fu,fv,fw,t2,r,tol,q,p,m,d,e ! should not be set by user
        integer :: iter,step ! should not be set by user
        character(len=10) :: msg ! should not be set by user
    end type
        
contains

! Computes the first and second derivative.
! Take as input the function f, the point where derivative is desired x, and the function 
! value at x. 
! Returns the first and second derivatives at x, fpx and fppx.
! The second derivative is computed at almost zero cost, but could be made optional
subroutine numderiv1dim(fx,f,x,fpx,fppx)
    implicit none
    real(rt), intent(in) :: x,fx
    interface
        function f(x)
            use mod_types
            implicit none
            real(rt), intent(in) :: x
            real(rt) :: f
        end function f
    end interface
    real(rt), intent(out) :: fpx,fppx
    !local
    real(rt) :: h,fxh,fxmh

    ! Compute the step size
    h = max(abs(x),1e0_rt)*(epsilon(1.e0_rt))**(1e0_rt/3e0_rt);

    ! Compute the function values at +/-h
    fxh = f(x+h)
    fxmh = f(x-h)

    ! Compute the derivatives
    fpx = (fxh-fxmh)/(2e0_rt*h)
    fppx = ((fxmh + fxh) - 2e0_rt*fx)/(h**2)

end subroutine numderiv1dim

! Grid search in 1 dimension
!
! flag  meaning
!    0  successfully converged
!   -1  max iterations reached
!
subroutine sub_gridsearch1dim_nogrid(x,f,a,b,n,printmod)
    use mod_matlab, only: linspace 
    implicit none
    real(rt), intent(inout) :: x ! Final solution
    interface
        function f(x)
            use mod_types
            implicit none
            real(rt), intent(in) :: x
            real(rt) :: f
        end function f
    end interface
    real(rt), intent(in) :: a,b
    integer, intent(in) :: n,printmod
    ! local
    real(rt), allocatable :: grid(:)

    allocate(grid(n))

    grid = linspace(a,b,n)

    call sub_gridsearch1dim_grid(x,f,grid,printmod)

    deallocate(grid)

end subroutine 
subroutine sub_gridsearch1dim_grid(x,f,grid,printmod)
    use mod_sort
    implicit none
    real(rt), intent(inout) :: x ! Final solution
    interface
        function f(x)
            use mod_types
            implicit none
            real(rt), intent(in) :: x
            real(rt) :: f
        end function f
    end interface
    real(rt), intent(in) :: grid(1:)
    integer, intent(in) :: printmod
    ! local
    integer :: i,n
    real(rt) :: xbest,fbest,tmp

    n = size(grid)

    if (printmod>0) write(*,'(A)') 'GS1:              x          f(x)         xbest      f(xbest)'
    fbest = huge(0e0_rt)
    do i = 1,n
        tmp = f(grid(i))
        if (tmp<fbest) then
            xbest = grid(i)
            fbest = tmp
        end if
        if (printmod>0 .and. mod(i,printmod)==0) write(*,'(A,4e14.6)') 'GS1: ',grid(i),tmp,xbest,fbest
    end do
    x = xbest
    if (printmod>0) write(*,'(A)') 'GS1: finished'

end subroutine 


! Simplex routine in 1 dimension
!
! flag  meaning
!    0  successfully converged
!   -1  max iterations reached
!
! As of Mar 5, 2014, this routine is stable in that it only moves away
! from xstar if a new point is strictly better. 
subroutine sub_simplex1dim(xstar,f,e,xtoler,maxiter,printmod,flag)
    use mod_sort
    implicit none
    real(rt), intent(inout) :: xstar ! Initial guess and final solution
    interface
        function f(x)
            use mod_types
            implicit none
            real(rt), intent(in) :: x
            real(rt) :: f
        end function f
    end interface
    real(rt), intent(in) :: e
    real(rt), intent(in) :: xtoler
    integer, intent(in) :: maxiter,printmod
    integer, intent(out) :: flag
    ! local
    real(rt) :: s(2), fs(2), xr, fxr, xe, fxe
    integer :: rank(2), iter

    if (printmod>0) write(*,'(A)') 'S1:              a             b            fa            fb'

    ! Get the initial simplex
    s = (/xstar,xstar+e/)
    fs = (/f(xstar),f(xstar+e)/)


    do iter = 1,maxiter

        ! Sort the points
!         call isort(fs,rank)
!         s = s(rank)

        ! Sort
        if (fs(2)<fs(1)) then
            xr = s(1) ! Uses xr for a temp
            fxr = fs(1) ! Uses fxr for a temp
            s(1) = s(2) ! Swap
            fs(1) = fs(2)
            s(2) = xr
            fs(2) = fxr
        end if
        
        ! Check for convergence
        if (abs(s(1)-s(2))<xtoler) then
            xstar = s(1)
            flag = 0
            if (printmod>0) write(*,'(A)') 'S1: converged successfully, simplex size<xtoler'
            return
        end if

        ! If desired, compute progress
        if (printmod>0 .and. mod(iter,printmod)==0) write(*,'(A,4e14.6)') 'S1: ',s(1),s(2),fs(1),fs(2)

        ! Now get rid of the worst point s(2)

        ! First, note the center of gravity is s(1)

        ! Now, compute the reflection point
        xr = 2e0_rt*s(1) - s(2)
        fxr = f(xr)
        
        if (fxr<fs(1)) then 

            ! If reflection is better than current best point, be greedy---try to expand.
            xe = 3e0_rt*s(1) - 2e0_rt*s(2)
            fxe = f(xe)

            if (fxe<fs(1)) then
                ! Expand
                s(2) = xe
                fs(2) = fxe
            else
                ! Reflect
                s(2) = xr
                fs(2) = fxr
            end if
        else
            ! Otherwise, contract (which here is the same as contract all)
            s(2) = 1.5e0_rt*s(1) - .5e0_rt*s(2)
            fs(2) = f(s(2))
        end if

    end do

    ! Return best if haven't converged yet
    xstar = merge(s(1),s(2),fs(1)<fs(2)) 
    flag = -1
    if (printmod>0) write(*,'(A)') 'S1: failed to converge, max iterations reached'


end subroutine sub_simplex1dim

! Newton's method in 1 dim
!
! Output flag returns info about the routine
! flag     meaning
!    2     converged successfully to a maximum, fpp<0
!    1     converged succesffuly to a minimum, fpp>0
!  >=0     converged successfully as measured by abs(s)<xtoler*(1e0_rt+abs(xstar))
!   -1     failed to converge, maxiter reached 
!   -2     failed to converge, bad function value encountered (NaN)
!   -3     failed to converge, implausibly large derivative value encountered
!   -4     failed to converge, second derivative of 0 encountered
subroutine sub_newton1dim(xstar,f,xtoler,maxiter,printmod,flag)
    implicit none
    real(rt), intent(inout) :: xstar ! Initial guess and final solution
    interface
        function f(x)
            use mod_types
            implicit none
            real(rt), intent(in) :: x
            real(rt) :: f
        end function f
    end interface
    real(rt), intent(in) :: xtoler
    integer, intent(in) :: maxiter,printmod
    integer, intent(out) :: flag
    ! local
    real(rt) :: s,fx,fpx,fppx
    integer :: iter

    if (printmod>0) write(*,'(A)') 'N1:              x            fx           fpx          fppx '

    ! Use Newton's method 
    do iter = 1,maxiter
        
        ! Compute the function value at x
        fx = f(xstar)

        ! Compute the derivatives at x
        call numderiv1dim(fx,f,xstar,fpx,fppx)

        ! Check for errors
!         if (isnan(fx) .or. isnan(fpx) .or. isnan(fppx)) then 
!             if (printmod>0) write(*,'(A)') 'N1: failed to converge, NaN encountered'
!             flag = -2
!             return
        if (abs(fpx)>1d50 .or. abs(fppx)>1d50) then
            if (printmod>0) write(*,'(A)') 'N1: failed to converge, huge derivative encountered' 
            flag = -3
            return
        elseif (abs(fppx)<1d-10) then
            if (printmod>0) write(*,'(A)') 'N1: failed to converge, second derivative is approx 0'
            flag = -4
            return
        end if

        ! Compute the Newton step
        s = fpx/fppx

        ! If desired, compute progress
        if (printmod>0 .and. mod(iter,printmod)==0) write(*,'(A,4e14.6)') 'N1: ',xstar,fx,fpx,fppx

        ! Check for convergence
        if (abs(s)<(1e0_rt+abs(xstar))*xtoler) then
            if (fppx<0e0_rt) then
                if (printmod>0) write(*,'(A)') 'N1: converged successfully to a max, step<toler'
                flag = 2
            else
                if (printmod>0) write(*,'(A)') 'N1: converged successfully to a min, step<toler'
                flag = 1
            end if
            return
        end if

        ! Otherwise, update the step
        xstar = xstar - s

    end do

    flag = -1
    if (printmod>0) write(*,'(A)') 'N1: failed to converge, maxiter reached'

end subroutine sub_newton1dim


! Reverse call back version of brent: no function needs to be passed
! 
! On the first call, user should set s%istat=0 and s%a and s%b, as well as s%xtoler and s%printmod if so desired
! 
! After the first call, the user should use the information in s%flag to determine what needs to be done
! s%flag | meaning
!     -1 | An error occurred
!      0 | Call the routine again without changing anything
!      1 | Store the value of f(s%x) in s%fx and call the routine again
!      2 | The solution is stored in s%x and the routine is finished
! 
subroutine sub_brent_min_rc(s)
    implicit none
    type (type_brentrc), intent(inout) :: s
    ! local
    real(rt), parameter :: phi=(1e0_rt+sqrt(5.e0_rt))/2.e0_rt, c = (3e0_rt-sqrt(5e0_rt))/2e0_rt, eps = epsilon(0e0_rt)

! print*,'s%istat',s%istat
    select case (s%istat)
        case (0)

            ! Initialization
            s%a_in = s%a
            s%b_in = s%b
            s%t = max(sqrt(epsilon(0e0_rt)),s%xtoler) ! Relative stopping criterion
            s%x = s%a + c*(s%b-s%a)
            s%w = s%x
            s%v = s%x

            ! Request f(s%x)
            s%flag = 1
            s%istat = 1

        case (1)

            ! Now have s%fx = f(s%x)
            s%fw = s%fx
            s%fv = s%fx
            s%e = 0e0_rt

            ! Go to the main loop with iter = 1 
            s%flag = 0 ! don't request function value
            s%istat = 2 ! main loop
            s%iter = 1 ! main loop iter = 1

        case (2) 

            if (s%iter==1) then
                s%msg = ' initial'
            else
                s%msg = merge(' parabolic',' golden   ',s%step==1)
            end if

            if (s%printmod>0 .and. mod(s%iter,s%printmod)==0) &
                write(*,'(A,i5,4f15.9,A)') 'B1: i,a,b,x,fx',s%iter,s%a,s%b,s%x,s%fx,s%msg

            s%m = .5e0_rt*(s%a+s%b)
            s%tol = eps*abs(s%x) + s%t
            s%t2 = 2e0_rt*s%tol

            ! Check stopping criterion
            if (.not. (abs(s%x-s%m)>s%t2 - .5e0_rt*(s%b-s%a))) then
                s%flag = 0
                s%istat = 100 ! exit main loop
                return 
            end if
                
            if (abs(s%e)>s%tol) then

                ! Fit parabola
                s%r = (s%x-s%w)*(s%fx-s%fv)
                s%q = (s%x-s%v)*(s%fx-s%fw)
                s%p = (s%x-s%v)*s%q - (s%x-s%w)*s%r 
                s%q = 2e0_rt*(s%q-s%r)

                if (s%q>0e0_rt) then
                    s%p = -s%p
                else
                    s%q = -s%q
                end if

                s%r = s%e
                s%e = s%d
            
            else 
                
                ! Don't fit parabola
                s%r = 0e0_rt
                s%q = 0e0_rt
                s%p = 0e0_rt

            end if

            ! if (abs(p)<abs(.5e0_rt*q*r) .and. p<q*(a-x) .and. p<q*(b-x)) then
            ! NOTE: THE ALGOL PROCEDURE IN BRENT HAS A BUG AND WOULD BE TRANSLATED
            ! INTO THE ABOVE LINE OF CODE. THE BUG IS NOT PRESENT IN THE FORTRAN 
            ! CODE WHICH NR HAS CORRECT.
            if (abs(s%p)<abs(.5e0_rt*s%q*s%r) .and. s%p>s%q*(s%a-s%x) .and. s%p<s%q*(s%b-s%x)) then

                ! Parabolic interpolation step
                s%d = s%p/s%q
                s%u = s%x + s%d

                ! f must not be evaluated too close to a or b
                if (s%u-s%a<s%t2 .or. s%b-s%u<s%t2) then
                    s%d = merge(s%tol,-s%tol,s%x<s%m)
                end if

                s%step = 1

            else

                ! Golden section step
                s%e = merge(s%b,s%a,s%x<s%m) - s%x
                s%d = c*s%e

                s%step = 0

            end if

            ! f must not be evaluated too close to x
            s%u = s%x + merge(s%d,merge(s%tol,-s%tol,s%d>0e0_rt),abs(s%d)>=s%tol)

            ! Request the value of f(u)
            s%xsave = s%x 
            s%fxsave = s%fx 
            s%x = s%u

            s%flag = 1 ! request
            s%istat = 3 ! go to second part of main loop

        case (3) 
            
            ! Store fu and restore the value of x
            s%fu = s%fx
            s%x = s%xsave
            s%fx = s%fxsave

            ! Update a,b,v,w, and x
            if (s%fu<=s%fx) then

                ! Replace upper or lower bound
                if (s%u<s%x) then
                    s%b = s%x
                else
                    s%a = s%x
                end if

                ! Save previous values
                s%v = s%w
                s%fv = s%fw
                s%w = s%x
                s%fw = s%fx
                s%x = s%u
                s%fx = s%fu

            else
                
                ! Replace upper or lower bound
                if (s%u<s%x) then
                    s%a = s%u
                else
                    s%b = s%u
                end if

                ! Save previous values
                if (s%fu <= s%fw .or. s%w == s%x) then
                    s%v = s%w 
                    s%fv = s%fw
                    s%w = s%u
                    s%fw = s%fu
                else if (s%fu <= s%fv .or. s%v == s%x .or. s%v == s%w) then
                    s%v = s%u
                    s%fv = s%fu
                end if
            end if

            ! Go to top of main loop 
            s%flag = 0 ! don't request a function value
            s%istat = 2 ! go to second part of main loop

            s%iter = s%iter + 1
            
            ! Check whether exceed max iteration
            if (s%iter>1000) then
                s%flag = -1 ! error
                s%istat = 999 ! error
            end if

        case (100) 
            
            ! After the main loop has been exited, here

            ! MODIFICATION: if close to boundary, compute the value at the boundary and take max
            s%tol = eps*abs(s%x) + s%t
            s%t2 = s%tol*2e0_rt
            if (abs(s%x-s%a_in)<=s%t2) then

                ! Request value of f(a) 
                s%xsave = s%x
                s%fxsave = s%fx
                s%x = s%a_in

                s%flag = 1 ! request the function evaluation
                s%istat = 101 ! go to case 1 after main loop
                return 

            elseif (abs(s%x-s%b_in)<=s%t2) then

                ! Request value of f(b) 
                s%xsave = s%x
                s%fxsave = s%fx
                s%x = s%b_in

                s%flag = 1 ! request the function evaluation
                s%istat = 102 ! go to case 1 after main loop
                return 

            else

                ! Not at boundary, so nothing further needs to be done: exit
                s%flag = 2 ! success
                s%istat = 1000 

            end if

        case (101)

                s%fu = s%fx
                s%x = s%xsave ! restore 
                s%fx = s%fxsave

                if (s%fu<s%fx) then
                    s%x = s%a_in
                end if

                s%msg = ' corner'
                s%iter = s%iter+1
                if (s%printmod>0 .and. mod(s%iter,s%printmod)==0) &
                    write(*,'(A,i5,4f15.9,A)') 'B1: i,a,b,x,fx',s%iter,s%a,s%b,s%x,s%fx,s%msg
                ! Exit
                s%flag = 2 ! success
                s%istat = 1000 

        case (102)

                s%fu = s%fx
                s%x = s%xsave ! restore 
                s%fx = s%fxsave

                if (s%fu<s%fx) then
                    s%x = s%b_in
                end if

                s%msg = ' corner'
                s%iter = s%iter+1
                if (s%printmod>0 .and. mod(s%iter,s%printmod)==0) &
                    write(*,'(A,i5,4f15.9,A)') 'B1: i,a,b,x,fx',s%iter,s%a,s%b,s%x,s%fx,s%msg

                ! Exit
                s%flag = 2 ! success
                s%istat = 1000

        case (999) 
            ! Error occurred
            STOP 'ERROR: sub_brent_min_rc: an error occurred; either reinitialize brent or don''t call'
        case (1000) 
            STOP 'ERROR: sub_brent_min_rc: a min was found successfully on a previous call; either reinitialize or don''t call'
    end select

end subroutine sub_brent_min_rc

! sub_brent_min
! This is the local minimization procedure described in Brent "Algorithms for 
! Minimization without Derivatives." The comments from the ALGOL procedure are
! the following (found on p.79):
!   If the function f is defined on the interval (a,b), then localmin finds an 
! approximation x to the point at which f attains its minimum (or the 
! appropriate limit point), and returns the value of f at x. t and eps define 
! a tolerance tol = eps|x| + t, and f is never evaluated at two points closer 
! together than tol. If f is delta-unimodal for some delta<tol, then x 
! approximates the global minimum of f with an error less than 3*tol. If f is 
! not delta-unimodal on (a,b), then x may approximate a local, but not global, 
! minimum. eps should be no smaller than 2*macheps, and preferably not much 
! less than sqrt (macheps), where macheps is the relative machine precision. t 
! should be [strictly] positive. For further details, see SEction 2.
!   The method used is a combination of golden section search and successive 
! parabolic interpolation. Convergence is never much slower than for a 
! Fibonacci search. If f has a continuous second derivative which is positive 
! at the minimum (not at a or b) then, ignoring rounding errors, convergence 
! is superlinear, and usually the order is at least 1.3247.
!
! [Grey:] Modifications: the original routine never evaluates f at the endpoints
! a and b. Consequently, a corner solution of a or b wil never be found exactly.
! I just add one check at the end if it looks like there may be a corner solution
! and return the corner solution if there is one.
! 
! I also consulted NR, but only after writing the code when it was producing golden
! steps only. I saw the code differed from the Algol procedure in Brent but not the
! Fortran procedure. Brent's Algol procedure had a bug! But once corrected, the 
! code works as expected. 
subroutine sub_brent_min(x,f,a_in,b_in,xtoler_in,printmod_in)
    implicit none
    real(rt), intent(out) :: x ! Final argmin
    interface
        function f(x)
            use mod_types
            implicit none
            real(rt), intent(in) :: x
            real(rt) :: f
        end function f
    end interface
    real(rt), intent(in) :: a_in,b_in
    ! optional arguments
    real(rt), intent(in), optional :: xtoler_in ! Relative stopping criterion for bracket size
    integer, intent(in), optional :: printmod_in ! Printing control. If <0, then no information is displayed
    ! local
    type(type_brentrc) :: s

    s%a = a_in
    s%b = b_in
    s%istat = 0
    if (present(xtoler_in)) s%xtoler = xtoler_in
    if (present(printmod_in)) s%printmod = printmod_in

    do 
        call sub_brent_min_rc(s)
        select case (s%flag)
            case (-1) 
                write(*,*) 'WARNING: sub_brent_min: an error occurred'
                exit
            case (0) 
                ! do nothing, call brent again
            case (1) 
                ! compute f(x) 
                s%fx = f(s%x)
            case (2)
                ! completed successfully, exit
                exit
            case default
                STOP 'sub_brent_min: case not found'
        end select
    end do

    ! Return the solution
    x = s%x

end subroutine 
subroutine sub_brent_min_old(x,f,a_in,b_in,xtoler_in,printmod_in)
    implicit none
    real(rt), intent(out) :: x ! Final argmin
    interface
        function f(x)
            use mod_types
            implicit none
            real(rt), intent(in) :: x
            real(rt) :: f
        end function f
    end interface
    real(rt), intent(in) :: a_in,b_in
    ! optional arguments
    real(rt), intent(in), optional :: xtoler_in ! Relative stopping criterion for bracket size
    integer, intent(in), optional :: printmod_in ! Printing control. If <0, then no information is displayed
    ! local
    real(rt) :: a,b,t,u,v,w,fu,fv,fw,fx,t2,r,tol,q,p,m,d,e
    integer :: iter,printmod,step
    character(len=10) :: msg
    real(rt), parameter :: phi=(1e0_rt+sqrt(5.e0_rt))/2.e0_rt, c = (3e0_rt-sqrt(5e0_rt))/2e0_rt, eps = epsilon(0e0_rt)

    ! Set defaults and replace if optional argument is present
    t = sqrt(epsilon(0e0_rt)); if (present(xtoler_in)) t = xtoler_in ! Relative stopping criterion
    printmod = 1; if (present(printmod_in)) printmod = printmod_in ! Printing control

    if (t<=0e0_rt) then
        STOP 'ERROR: sub_brent_min: xtoler_in must be > 0'
    end if

    ! Initialization
    a = a_in
    b = b_in

    x = a + c*(b-a)
    w = x
    v = x

    fx = f(x)
    fw = fx
    fv = fx

    e = 0e0_rt

    ! Main loop
    do iter = 1,10000

        if (iter==1) then
            msg = ' initial'
        else
            msg = merge(' parabolic',' golden   ',step==1)
        end if

        if (printmod>0 .and. mod(iter,printmod)==0) &
            write(*,'(A,i5,4f15.9,A)') 'B1: i,a,b,x,fx',iter,a,b,x,fx,msg

        m = .5e0_rt*(a+b)
        tol = eps*abs(x) + t
        t2 = 2e0_rt*tol

        ! Check stopping criterion
        if (.not. (abs(x-m)>t2 - .5e0_rt*(b-a))) exit

        if (abs(e)>tol) then

            ! Fit parabola
            r = (x-w)*(fx-fv)
            q = (x-v)*(fx-fw)
            p = (x-v)*q - (x-w)*r 
            q = 2e0_rt*(q-r)

            if (q>0e0_rt) then
                p = -p
            else
                q = -q
            end if

            r = e
            e = d
        
        else 
            
            ! Don't fit parabola
            r = 0e0_rt
            q = 0e0_rt
            p = 0e0_rt

        end if

        ! if (abs(p)<abs(.5e0_rt*q*r) .and. p<q*(a-x) .and. p<q*(b-x)) then
        ! NOTE: THE ALGOL PROCEDURE IN BRENT HAS A BUG AND WOULD BE TRANSLATED
        ! INTO THE ABOVE LINE OF CODE. THE BUG IS NOT PRESENT IN THE FORTRAN 
        ! CODE WHICH NR HAS CORRECT.
        if (abs(p)<abs(.5e0_rt*q*r) .and. p>q*(a-x) .and. p<q*(b-x)) then

            ! Parabolic interpolation step
            d = p/q
            u = x + d

            ! f must not be evaluated too close to a or b
            if (u-a<t2 .or. b-u<t2) then
                d = merge(tol,-tol,x<m)
            end if

            step = 1

        else

            ! Golden section step
            e = merge(b,a,x<m) - x
            d = c*e

            step = 0

        end if

        ! f must not be evaluated too close to x

        u = x + merge(d,merge(tol,-tol,d>0e0_rt),abs(d)>=tol)
        fu = f(u)

        ! Update a,b,v,w, and x
        if (fu<=fx) then

            ! Replace upper or lower bound
            if (u<x) then
                b = x
            else
                a = x
            end if

            ! Save previous values
            v = w
            fv = fw
            w = x
            fw = fx
            x = u
            fx = fu

        else
            
            ! Replace upper or lower bound
            if (u<x) then
                a = u
            else
                b = u
            end if

            ! Save previous values
            if (fu <= fw .or. w == x) then
                v = w 
                fv = fw
                w = u
                fw = fu
            else if (fu <= fv .or. v == x .or. v == w) then
                v = u
                fv = fu
            end if
        end if

    end do

    ! MODIFICATION: if close to boundary, compute the value at the boundary and take max
    tol = eps*abs(x) + t
    t2 = tol*2e0_rt
    if (abs(x-a_in)<=t2) then
        fu = f(a_in) 
        if (fu<fx) then
            x = a_in
        end if

        msg = ' corner'
        iter = iter+1
        if (printmod>0 .and. mod(iter,printmod)==0) &
            write(*,'(A,i5,4f15.9,A)') 'B1: i,a,b,x,fx',iter,a,b,x,fx,msg

    elseif (abs(x-b_in)<=t2) then
        fu = f(b_in) 
        if (fu<fx) then
            x = b_in
        end if

        msg = ' corner'
        iter = iter+1
        if (printmod>0 .and. mod(iter,printmod)==0) &
            write(*,'(A,i5,4f15.9,A)') 'B1: i,a,b,x,fx',iter,a,b,x,fx,msg
    end if

end subroutine sub_brent_min_old

! Golden section search to minimize a function
! The algorithm is from Wikipedia 
! xtoler_in :: the algorithm stops when |c-a|<xtoler*(1+abs(b)) where [a,c] brackets a min
! 
subroutine sub_golden(xstar,f,xtoler_in,step_in,printmod_in)
    implicit none
    real(rt), intent(inout) :: xstar ! Initial guess and final solution
    interface
        function f(x)
            use mod_types
            implicit none
            real(rt), intent(in) :: x
            real(rt) :: f
        end function f
    end interface
    ! optional arguments
    real(rt), intent(in), optional :: xtoler_in ! Relative stopping criterion for bracket size
    real(rt), intent(in), optional :: step_in ! First step size when trying to bracket 
    integer, intent(in), optional :: printmod_in ! Printing control. If <0, then no information is displayed
    ! local
    real(rt) :: a,b,c,fa,fb,fc,step,xtoler
    integer :: iter,printmod
    real(rt), parameter :: phi=(1.e0_rt+sqrt(5.e0_rt))/2.e0_rt

    ! Set defaults and replace if optional argument is present
    xtoler = sqrt(epsilon(0.e0_rt)); if (present(xtoler_in)) xtoler = xtoler_in ! Relative stopping criterion
    step = 1e-6_rt*(1.e0_rt+abs(xstar)); if (present(step_in)) step = step_in ! Initial step size to search
    printmod = -1; if (present(printmod_in)) printmod = printmod_in ! Printing control

    ! Try to find a bracket centered around xstar consistent with golden section search
    b = xstar
    fb = f(b)

    ! Bracket a minimum with section lengths having the golden ratio
    if (printmod>0) print*,'G1: Finding a bracket'
    do iter = 1,100

        ! Choose right interval so it is of size step
        c = b + phi*step

        ! Choose left interval so that it is of size phi*step where phi is the golden ratio
        a = b - step

        ! Evaluate f at a and c
        fa = f(a)
        fc = f(c)

        if (printmod>0 .and. mod(iter,printmod)==0) then
            write(*,'(A,6f9.3)') 'G1: a,b,c,fa,fb,fc',a,b,c,fa,fb,fc
        end if

        ! If f(a)>=f(b) and f(b)<=f(c), have found a bracket
        if (fa>=fb .and. fb<=fc) then
            if (printmod>0) write(*,'(A)') 'G1: Bracket found. Beginning min search.'
            exit
        else
            step = step*10.e0_rt
        end if

        if (iter==100) then
            write(*,'(A)') 'sub_golden: WARNING: couldn''t bracket a min, returning init guess'
            xstar = b
            return
        end if
    end do

    ! Call the core routine now that the first triple has been established
    call sub_golden_core(a,b,c,fa,fb,fc,f,xtoler,printmod)
    
    ! Return the b: this is better than (a+c)/2 because if c or a may not be feasible but b may be.
    xstar = b 

end subroutine sub_golden

! a<b<c bracket a minimum with f(a)>=f(b), f(b)<=f(c)
subroutine sub_golden_core(a,b,c,fa,fb,fc,f,xtoler,printmod)
    implicit none
    real(rt), intent(inout) :: a,b,c,fa,fb,fc
    interface
        function f(x)
            use mod_types
            implicit none
            real(rt), intent(in) :: x
            real(rt) :: f
        end function f
    end interface
    real(rt), intent(in) :: xtoler ! Relative stopping criterion
    integer, intent(in) :: printmod ! Print control
    real(rt), parameter :: phi=(1.e0_rt+sqrt(5.e0_rt))/2.e0_rt
    ! local
    real(rt) :: d,fd
    integer :: di
    integer :: iter

    do iter = 1,10000

        ! Check stopping criterion
        if (c-a<xtoler*(1.e0_rt+abs(b))) then
            if (printmod>0) write(*,'(A)') 'G1: finished'
            return
        end if

        ! Compute the new point
        if (c-b>b-a) then
            d = b + (2.e0_rt-phi)*(c-b)
        else
            d = b - (2.e0_rt-phi)*(b-a)
        end if

        ! Compute the function value
        fd = f(d)

        ! Display info if desired
        if (printmod>0 .and. mod(iter,printmod)==0) then
            write(*,'(A,6f9.3)') 'G1: a,b,c,fa,fb,fc',a,b,c,fa,fb,fc
        end if

        ! Compute the new bracket depending on the value of fd 
        if (fd<fb) then
            if (c-b>b-a) then
                a = b;  fa = fb
                b = d;  fb = fd
            else
                c = b;  fc = fb
                b = d;  fb = fd
            end if
        else
            if (c-b>b-a) then
                c = d;  fc = fd
            else
                a = d;  fa = fd
            end if
        end if

    end do

end subroutine sub_golden_core 

end module mod_min1dim


module mod_min1dim_test
    use mod_min1dim
    implicit none 
contains

! Test functions
! Defined on R
! minimum at x = 5e0_rt
function uniqMinQuad(x) result(y)
    real(rt), intent(in) :: x
    real(rt) :: y
    y = (x-5e0_rt)**2 + 2e0_rt
end function 
! Defined on R++, maximum at x = 1
function uniqMaxConc(x) result(y)
    real(rt), intent(in) :: x
    real(rt) :: y
    y = log(x) - x 
end function 
! Defined on R++, minimum at x = 1
function uniqMinConc(x) result(y)
    real(rt), intent(in) :: x
    real(rt) :: y
    y = -(log(x) - x )
end function 

subroutine test_brent_min
    implicit none
    real(rt) :: x

    call sub_brent_min(x,uniqMinConc,0e0_rt,100e0_rt,printmod_in=1)
    if (abs(x-1e0_rt)>4.0_rt*sqrt(epsilon(0._rt))) STOP 'ERROR: test_brent_min: failed test 1' 
    call sub_brent_min(x,uniqMinQuad,-100e0_rt,100e0_rt,printmod_in=1)
    if (abs(x-5e0_rt)>4.0_rt*sqrt(epsilon(0._rt))) STOP 'ERROR: test_brent_min: failed test 2' 

    print*,'test_brent_min: passed'
    

end subroutine

subroutine test_simplex1dim
    implicit none
    real(rt) :: x
    integer :: flag

    x = -10e0_rt
    call sub_simplex1dim(x,uniqMinQuad,100e0_rt,1e-6_rt,10000,-1,flag)
    
    if (flag<0) STOP 'ERROR: test_simplex1dim: failed to converge' 

end subroutine
subroutine test_newton1dim
    implicit none
    real(rt) :: x
    integer :: flag

    x = -10e0_rt
    call sub_newton1dim(x,uniqMinQuad,sqrt(epsilon(0._rt)),10000,-1,flag)
    if (flag<0) STOP 'ERROR: test_newton1dim: failed to converge' 
    x = .3e0_rt
    call sub_newton1dim(x,uniqMaxConc,sqrt(epsilon(0._rt)),10000,-1,flag)
    if (flag<0) STOP 'ERROR: test_newton1dim: failed to converge' 

end subroutine

end module 
! A module that attempts to simplify the process of using mpi w/o a significant
! computational cost (which of course would defeat the point!)
!
! Most routines, if available, are listed as myXXX where XXX is an mpi command.
! E.g., mympi_allgatherv calls MPI_Allgatherv.
! There are some exceptions, such as mympi_stop which calls MPI_Abort after
! displaying an error message.
module mod_mympi
    use mod_types
    implicit none

    include 'mpif.h'

    logical, private :: mympi_initialized = .false.
    integer, private :: nproc,iproc


    interface mympi_allgatherv
        module procedure mympi_allgatherv_vec, mympi_allgatherv_mat,  mympi_allgatherv_3d,&
                         mympi_allgatherv_4d, mympi_allgatherv_5d, mympi_allgatherv_6d,&
                         mympi_allgatherv_ivec,mympi_allgatherv_imat, mympi_allgatherv_i3d, mympi_allgatherv_i4d, &
                         mympi_allgatherv_l1d, mympi_allgatherv_l2d,mympi_allgatherv_l4d, mympi_allgatherv_l5d
    end interface
    interface mympi_allgatherv_r4
        module procedure mympi_allgatherv_r4vec,mympi_allgatherv_r4mat,mympi_allgatherv_r43d,mympi_allgatherv_r44d
    end interface

    interface mympi_allreduce
        module procedure mympi_allreduce_1d,mympi_allreduce_2d,mympi_allreduce_3d, &
                         mympi_allreduce_i0d,mympi_allreduce_i1d,mympi_allreduce_i2d
    end interface

    interface mympi_bcast
        module procedure mympi_bcast_vec, mympi_bcast_mat, mympi_bcast_3d, mympi_bcast_4d, mympi_bcast_5d, mympi_bcast_6d, &
                         mympi_bcast_ivec, mympi_bcast_imat, mympi_bcast_i3d, mympi_bcast_i4d, mympi_bcast_i5d, mympi_bcast_i6d, &
                         mympi_bcast_logical_vec, mympi_bcast_logical_mat, mympi_bcast_logical_3d, mympi_bcast_logical_4d, &
                         mympi_bcast_logical_5d, mympi_bcast_logical_6d
    end interface

    interface mympi_ibcast
        module procedure mympi_ibcast_vec, &
                         mympi_ibcast_ivec
    end interface

    interface mympi_irecv
        module procedure mympi_irecv_1d, mympi_irecv_0d, mympi_irecv_i1d
    end interface

    interface mympi_recv
        module procedure mympi_recv_2d, mympi_recv_1d, mympi_recv_0d, & 
                         mympi_recv_i2d, mympi_recv_i1d, mympi_recv_i0d
    end interface

    interface mympi_send
        module procedure mympi_send_0d, mympi_send_1d, mympi_send_2d, &
                         mympi_send_i0d, mympi_send_i1d, mympi_send_i2d
    end interface

    private :: mympi_bcast_core
    private :: mympi_allgatherv_core
    private :: mympi_allgatherv_core_logical
    private :: mympi_allgatherv_icore
    private :: checkinit
    
contains

    ! Performs a reduction operation and shares the result with all nodes
    ! [ MPI_MAX] maximum
    ! [ MPI_MIN] minimum
    ! [ MPI_SUM] sum
    ! [ MPI_PROD] product
    ! [ MPI_LAND] logical and
    ! [ MPI_BAND] bit-wise and
    ! [ MPI_LOR] logical or
    ! [ MPI_BOR] bit-wise or
    ! [ MPI_LXOR] logical xor
    ! [ MPI_BXOR] bit-wise xor
    ! [ MPI_MAXLOC] max value and location
    ! [ MPI_MINLOC] min value and location 
    subroutine mympi_allreduce_3d(B,A,operation)
        implicit none
        real(rt), intent(in) :: A(:,:,:)
        real(rt), intent(out) :: B(:,:,:)
        integer, intent(in) :: operation

        if (size(A)/=size(B)) STOP 'ERROR: mympi_allreduce: input and output do not have the same size'

        call mympi_allreduce_core_d(B,A,size(A),operation)

    end subroutine 
    subroutine mympi_allreduce_2d(B,A,operation)
        implicit none
        real(rt), intent(in) :: A(:,:)
        real(rt), intent(out) :: B(:,:)
        integer, intent(in) :: operation

        if (size(A)/=size(B)) STOP 'ERROR: mympi_allreduce: input and output do not have the same size'

        call mympi_allreduce_core_d(B,A,size(A),operation)

    end subroutine 
    subroutine mympi_allreduce_1d(B,A,operation)
        implicit none
        real(rt), intent(in) :: A(:)
        real(rt), intent(out) :: B(:)
        integer, intent(in) :: operation

        if (size(A)/=size(B)) STOP 'ERROR: mympi_allreduce: input and output do not have the same size'

        call mympi_allreduce_core_d(B,A,size(A),operation)

    end subroutine 
    subroutine mympi_allreduce_core_d(B,A,nelts,operation)
        use mod_matlab
        implicit none
        real(rt), intent(in) :: A(*)
        real(rt), intent(out) :: B(*)
        integer, intent(in) :: nelts,operation
        ! local
        integer :: ierr

        ! Call the mpi routine
        call MPI_Allreduce(A(1),& ! first element assigned to iproc
                           B(1),& ! where to put the reduction  
                           nelts,& ! total numbber of elements assigned to iproc --- may be zero
                           mympi_rt,& ! data type of send elements
                           operation,& ! type of operation
                           MPI_COMM_WORLD,&
                           ierr) 

    end subroutine
    
    subroutine mympi_allreduce_i2d(B,A,operation)
        implicit none
        integer, intent(in) :: A(:,:)
        integer, intent(out) :: B(:,:)
        integer, intent(in) :: operation

        if (size(A)/=size(B)) STOP 'ERROR: mympi_allreduce: input and output do not have the same size'

        call mympi_allreduce_core_i(B,A,size(A),operation)

    end subroutine 
    subroutine mympi_allreduce_i1d(B,A,operation)
        implicit none
        integer, intent(in) :: A(:)
        integer, intent(out) :: B(:)
        integer, intent(in) :: operation

        if (size(A)/=size(B)) STOP 'ERROR: mympi_allreduce: input and output do not have the same size'

        call mympi_allreduce_core_i(B,A,size(A),operation)

    end subroutine 
    subroutine mympi_allreduce_i0d(B,A,operation)
        implicit none
        integer, intent(in) :: A
        integer, intent(out) :: B
        integer, intent(in) :: operation
        integer :: ierr

        call MPI_Allreduce(A,& ! first element assigned to iproc
                           B,& ! where to put the reduction  
                           1,& ! total number of elements assigned to iproc --- may be zero
                           MPI_INTEGER,& ! data type of send elements
                           operation,& ! type of operation
                           MPI_COMM_WORLD,&
                           ierr) 

    end subroutine 
    subroutine mympi_allreduce_core_i(B,A,nelts,operation)
        use mod_matlab
        implicit none
        integer, intent(in) :: A(*)
        integer, intent(out) :: B(*)
        integer, intent(in) :: nelts,operation
        ! local
        integer :: ierr

        ! Call the mpi routine
        call MPI_Allreduce(A(1),& ! first element assigned to iproc
                           B(1),& ! where to put the reduction  
                           nelts,& ! total numbber of elements assigned to iproc --- may be zero
                           MPI_INTEGER,& ! data type of send elements
                           operation,& ! type of operation
                           MPI_COMM_WORLD,&
                           ierr) 

    end subroutine

function mympi_getprocessorname() result(nomen)
    implicit none
    character(len=MPI_MAX_PROCESSOR_NAME) :: nomen
    ! local
    integer :: ierr,lenout
    character(len=MPI_MAX_PROCESSOR_NAME) :: tmp

    
    call MPI_GET_PROCESSOR_NAME(tmp,lenout,ierr)

    nomen = ''
    nomen(1:lenout) = tmp(1:lenout) 

end function 

function get_ierr_msg(ierr) result(msg)
    implicit none
    integer, intent(in) :: ierr
    character(len=50) :: msg

    select case (ierr)
        case (MPI_SUCCESS)
            msg='Successful return code'
        case (MPI_ERR_BUFFER)
            msg='Invalid buffer pointer'
        case (MPI_ERR_COUNT)
            msg='Invalid count argument'
        case (MPI_ERR_TYPE)
            msg='Invalid datatype argument'
        case (MPI_ERR_TAG)
            msg='Invalid tag argument'
        case (MPI_ERR_COMM)
            msg='Invalid communicator'
        case (MPI_ERR_RANK)
            msg='Invalid rank'
        case (MPI_ERR_ROOT)
            msg='Invalid root'
        case (MPI_ERR_GROUP)
            msg='Null group passed to function'
        case (MPI_ERR_OP)
            msg='Invalid operation'
        case (MPI_ERR_TOPOLOGY)
            msg='Invalid topology'
        case (MPI_ERR_DIMS)
            msg='Illegal dimension argument'
        case (MPI_ERR_ARG)
            msg='Invalid argument'
        case (MPI_ERR_UNKNOWN)
            msg='Unknown error'
        case (MPI_ERR_TRUNCATE)
            msg='message truncated on receive'
        case (MPI_ERR_OTHER)
            msg='Other error; use Error_string'
        case (MPI_ERR_INTERN)
            msg='internal error code'
        case (MPI_ERR_IN_STATUS)
            msg='Look in status for error value'
        case (MPI_ERR_PENDING)
            msg='Pending request'
        case (MPI_ERR_REQUEST)
            msg='illegal mpi_request handle'
        case (MPI_ERR_LASTCODE)
            msg='Last error code -- always at end'
    end select

end function

subroutine checkinit()
    implicit none
    if (.not. mympi_initialized) STOP 'ERROR: mpi has not been initialized'
end subroutine 

integer function mympi_iproc()
    implicit none
    call checkinit()
    mympi_iproc = iproc
end function 
integer function mympi_nproc()
    implicit none
    call checkinit()
    mympi_nproc = nproc
end function 

subroutine mympi_barrier()
    implicit none
    integer :: ierr

    call checkinit()
    call MPI_Barrier(MPI_COMM_WORLD,ierr)

end subroutine 

! Aborts all mpi routines and also displays a message
subroutine mympi_abort()
    implicit none
    call mympi_stop('')
end subroutine 
subroutine mympi_stop(msg)
    implicit none
    character(len=*), intent(in) :: msg
    ! local
    integer :: errorcode
    integer :: ierr

    call checkinit()

    ! Display the error message
    write(0,*) msg

    ! Pass error code to environment
    errorcode = 1 
    
    ! Terminate all processes
    call MPI_Abort(MPI_COMM_WORLD,errorcode,ierr)

end subroutine 

! Non blocking test for a message
! function mympi_iprobe(source,tag) result(tf)
!     implicit none
!     integer, intent(in) :: source
!     integer, intent(in), optional :: tag 
!     logical :: tf
!     ! local
!     integer :: ierr, status(MPI_STATUS_SIZE)
!     logical :: flag
! 
!     if (present(tag)) then
!         call MPI_Iprobe(source, tag, MPI_COMM_WORLD, flag, status, ierr)
!     else
!         call MPI_Iprobe(source, MPI_ANY_TAG, MPI_COMM_WORLD, flag, status, ierr)
!     end if
! 
!     tf = flag
! 
! end function 

! Divides n elts as evenly as possible into ranges [a,b]
! If b<a, then b is set to -1 and a to 0. This is convenient as b-a+1 
! then returns the number of elements for each node regardless of whether 
! the range is 0 or not. Also, since b<a, "do i = a,b" is not executed. 
! NOTE: if used to assign trailing dimensions of an array, 
! as it defines contiguous array section this is very useful
subroutine mympi_assign_range(a,b,n)
    implicit none
    integer, intent(in) :: n
    integer, intent(out) :: a,b
    
    call checkinit()
    call mympi_assign_range_core(a,b,iproc,n)

end subroutine 
! Helper functions
integer function mympi_a(n)
    implicit none
    integer, intent(in) :: n
    ! local
    integer :: b
    call checkinit()
    call mympi_assign_range(mympi_a,b,n)
end function mympi_a
integer function mympi_b(n)
    implicit none
    integer, intent(in) :: n
    ! local
    integer :: a
    call checkinit()
    call mympi_assign_range(a,mympi_b,n)
end function mympi_b
! Gives the combined list
subroutine mympi_assign_range_list(a,b,n)
    implicit none
    integer, intent(in) :: n
    integer, intent(out) :: a(nproc),b(nproc)
    ! local
    integer :: ip

    call checkinit()

    do ip = 0,nproc-1
        call mympi_assign_range_core(a(ip+1),b(ip+1),ip,n)
    end do
end subroutine 

! The core routine for dividing elements
! ip = 0,...,nproc-1
subroutine mympi_assign_range_core(a,b,ip,n)
    implicit none
    integer, intent(in) :: ip,n
    integer, intent(out) :: a,b
    ! local
    integer :: rem,jproc,times

    call checkinit()
    if (n<nproc) then
        ! If here, can assign one element to each node. 
        ! Hence, ip gets the ip+1 element if ip+1<=n
        if (ip<n) then
            a = ip+1
            b = ip+1
        else
            a = 0
            b = -1
        end if
    else
        ! If here, need to assign at least n/nproc to each element and the remainder
        ! across the first elements

        rem = mod(n,nproc)
        times = n/nproc ! >=1

        if (n/=times*nproc + rem) call mympi_stop('ERROR: mympi_assign_range: arithmetic not as expected')

        b = 0
        do jproc = 0,ip
            if (jproc<rem) then
                ! Gets times+1 elements
                a = 1 + b
                b = times + 1 + b
            else
                ! Gets times elements
                a = 1 + b
                b = times + b
            end if
        end do

    end if

end subroutine


subroutine mympi_bcast_logical_6d(A,root)
    implicit none
    logical, intent(inout) :: A(:,:,:,:,:,:)
    integer, intent(in) :: root
    call mympi_bcast_logical_core(A,size(A),root)
end subroutine
subroutine mympi_bcast_logical_5d(A,root)
    implicit none
    logical, intent(inout) :: A(:,:,:,:,:)
    integer, intent(in) :: root
    call mympi_bcast_logical_core(A,size(A),root)
end subroutine
subroutine mympi_bcast_logical_4d(A,root)
    implicit none
    logical, intent(inout) :: A(:,:,:,:)
    integer, intent(in) :: root
    call mympi_bcast_logical_core(A,size(A),root)
end subroutine
subroutine mympi_bcast_logical_3d(A,root)
    implicit none
    logical, intent(inout) :: A(:,:,:)
    integer, intent(in) :: root
    call mympi_bcast_logical_core(A,size(A),root)
end subroutine
subroutine mympi_bcast_logical_mat(A,root)
    implicit none
    logical, intent(inout) :: A(:,:)
    integer, intent(in) :: root
    call mympi_bcast_logical_core(A,size(A),root)
end subroutine
subroutine mympi_bcast_logical_vec(A,root)
    implicit none
    logical, intent(inout) :: A(:)
    integer, intent(in) :: root
    call mympi_bcast_logical_core(A,size(A),root)
end subroutine

subroutine mympi_bcast_logical_core(A,n,root)
    implicit none
    integer, intent(in) :: n, root
    logical, intent(inout) :: A(*)
    ! local
    integer :: ierr

    call checkinit()

    ! Call the mpi routine
    call MPI_Bcast(A(1),n,MPI_LOGICAL,root,MPI_COMM_WORLD,ierr)

end subroutine

subroutine mympi_ibcast_ivec(request,A,root)
    implicit none
    integer, intent(out) :: request
    integer, intent(inout) :: A(:)
    integer, intent(in) :: root

    call mympi_ibcast_icore(request,A,size(A),root)

end subroutine
subroutine mympi_ibcast_icore(request,A,n,root)
    implicit none
    integer, intent(out) :: request
    integer, intent(inout) :: A(*)
    integer, intent(in) :: n,root
    ! local
    integer :: ierr

    call checkinit()

    ! Call the mpi routine
    call MPI_Ibcast(A(1:n),n,MPI_INTEGER,root,MPI_COMM_WORLD,request,ierr)

    if (ierr/=MPI_SUCCESS) then
        call mympi_stop('error calling mympi_ibcast_icore')
    end if

end subroutine

subroutine mympi_ibcast_vec(request,A,root)
    implicit none
    integer, intent(out) :: request
    real(rt), intent(inout) :: A(:)
    integer, intent(in) :: root
    call mympi_ibcast_core(request,A,size(A),root)
end subroutine
subroutine mympi_ibcast_core(request,A,n,root)
    implicit none
    integer, intent(out) :: request
    integer, intent(in) :: n,root
    real(rt), intent(inout) :: A(*)
    ! local
    integer :: ierr

    call checkinit()

    ! Call the mpi routine
    call MPI_Ibcast(A(1),n,mympi_rt,root,MPI_COMM_WORLD,request,ierr)

end subroutine

subroutine mympi_bcast_6d(A,root)
    implicit none
    real(rt), intent(inout) :: A(:,:,:,:,:,:)
    integer, intent(in) :: root
    call mympi_bcast_core(A,size(A),root)
end subroutine
subroutine mympi_bcast_5d(A,root)
    implicit none
    real(rt), intent(inout) :: A(:,:,:,:,:)
    integer, intent(in) :: root
    call mympi_bcast_core(A,size(A),root)
end subroutine
subroutine mympi_bcast_4d(A,root)
    implicit none
    real(rt), intent(inout) :: A(:,:,:,:)
    integer, intent(in) :: root
    call mympi_bcast_core(A,size(A),root)
end subroutine
subroutine mympi_bcast_3d(A,root)
    implicit none
    real(rt), intent(inout) :: A(:,:,:)
    integer, intent(in) :: root
    call mympi_bcast_core(A,size(A),root)
end subroutine
subroutine mympi_bcast_mat(A,root)
    implicit none
    real(rt), intent(inout) :: A(:,:)
    integer, intent(in) :: root
    call mympi_bcast_core(A,size(A),root)
end subroutine
subroutine mympi_bcast_vec(A,root)
    implicit none
    real(rt), intent(inout) :: A(:)
    integer, intent(in) :: root
    call mympi_bcast_core(A,size(A),root)
end subroutine

subroutine mympi_bcast_core(A,n,root)
    implicit none
    integer, intent(in) :: n,root
    real(rt), intent(inout) :: A(*)
    ! local
    integer :: ierr

    call checkinit()

    ! Call the mpi routine
    call MPI_Bcast(A(1),n,mympi_rt,root,MPI_COMM_WORLD,ierr)

end subroutine

subroutine mympi_bcast_i6d(A,root)
    implicit none
    integer, intent(inout) :: A(:,:,:,:,:,:)
    integer, intent(in) :: root
    call mympi_bcast_icore(A,size(A),root)
end subroutine
subroutine mympi_bcast_i5d(A,root)
    implicit none
    integer, intent(inout) :: A(:,:,:,:,:)
    integer, intent(in) :: root
    call mympi_bcast_icore(A,size(A),root)
end subroutine
subroutine mympi_bcast_i4d(A,root)
    implicit none
    integer, intent(inout) :: A(:,:,:,:)
    integer, intent(in) :: root
    call mympi_bcast_icore(A,size(A),root)
end subroutine
subroutine mympi_bcast_i3d(A,root)
    implicit none
    integer, intent(inout) :: A(:,:,:)
    integer, intent(in) :: root
    call mympi_bcast_icore(A,size(A),root)
end subroutine
subroutine mympi_bcast_imat(A,root)
    implicit none
    integer, intent(inout) :: A(:,:)
    integer, intent(in) :: root
    call mympi_bcast_icore(A,size(A),root)
end subroutine
subroutine mympi_bcast_ivec(A,root)
    implicit none
    integer, intent(inout) :: A(:)
    integer, intent(in) :: root
    call mympi_bcast_icore(A,size(A),root)
end subroutine

subroutine mympi_bcast_icore(A,n,root)
    implicit none
    integer, intent(in) :: n,root
    integer, intent(inout) :: A(*)
    ! local
    integer :: ierr

    call checkinit()

    ! Call the mpi routine
    call MPI_Bcast(A(1),n,MPI_INTEGER,root,MPI_COMM_WORLD,ierr)

end subroutine

function mympi_test(request) result(isComplete)
    implicit none
    integer, intent(in) :: request
    logical :: isComplete
    ! local
    logical :: flag
    integer :: ierr ! , status(MPI_STATUS_SIZE)

    call MPI_Test(request,flag,MPI_STATUS_IGNORE,ierr)

    isComplete = flag

end function

subroutine mympi_irecv_1d(request,A,source,tag)
    implicit none
    integer, intent(out) :: request
    real(rt), intent(out) :: A(:)
    integer, intent(in) :: source
    integer, intent(in), optional :: tag

    if (present(tag)) then
        call mympi_irecv_core(request,A,size(A),source,tag)
    else
        ! call mympi_irecv_core(request,A,size(A),source,MPI_ANY_TAG)
        call mympi_irecv_core(request,A,size(A),source,0)
    end if

end subroutine
subroutine mympi_irecv_0d(request,A,source,tag)
    implicit none
    integer, intent(out) :: request
    real(rt), intent(out) :: A
    integer, intent(in) :: source
    integer, intent(in), optional :: tag
    ! local
    real(rt) :: tmp(1)

    if (present(tag)) then
        call mympi_irecv_core(request,tmp,1,source,tag)
    else
        ! call mympi_irecv_core(request,tmp,1,source,MPI_ANY_TAG)
        call mympi_irecv_core(request,tmp,1,source,0)
    end if

    A = tmp(1)

end subroutine


subroutine mympi_irecv_core(request,A,n,source,tag)
    implicit none
    integer, intent(out) :: request
    real(rt), intent(out) :: A(*)
    integer, intent(in) :: n,source,tag
    ! local
    integer :: ierr

    call MPI_Irecv(A(1), n, mympi_rt, source, tag, &
                  MPI_COMM_WORLD, request, ierr)

    if (ierr/=MPI_SUCCESS) then
        print*,'ERROR: mympi_irecv: ',trim(get_ierr_msg(ierr))
        call MPI_Abort(MPI_COMM_WORLD,1,ierr)
    end if

end subroutine 

subroutine mympi_irecv_i1d(request,A,source,tag)
    implicit none
    integer, intent(out) :: request
    integer, intent(out) :: A(:)
    integer, intent(in) :: source
    integer, intent(in), optional :: tag

    if (present(tag)) then
        call mympi_irecv_icore(request,A,size(A),source,tag)
    else
        ! call mympi_irecv_icore(request,A,size(A),source,MPI_ANY_TAG)
        call mympi_irecv_icore(request,A,size(A),source,0)
    end if

end subroutine

subroutine mympi_irecv_icore(request,A,n,source,tag)
    implicit none
    integer, intent(out) :: request
    integer, intent(out) :: A(*)
    integer, intent(in) :: n,source,tag
    ! local
    integer :: ierr

    call MPI_Irecv(A(1), n, MPI_INTEGER, source, tag, &
                  MPI_COMM_WORLD, request, ierr)

    if (ierr/=MPI_SUCCESS) then
        print*,'ERROR: mympi_irecv: ',trim(get_ierr_msg(ierr))
        call MPI_Abort(MPI_COMM_WORLD,1,ierr)
    end if

end subroutine 

subroutine mympi_waitall(array_of_requests)
    implicit none
    integer, intent(in) :: array_of_requests(:)
    ! local
    integer :: ierr

    call MPI_Waitall(size(array_of_requests),array_of_requests(1),MPI_STATUSES_IGNORE,ierr)

    if (ierr/=MPI_SUCCESS) then
        print*,'ERROR: mympi_waitall: ',trim(get_ierr_msg(ierr))
        call MPI_Abort(MPI_COMM_WORLD,1,ierr)
    end if
end subroutine

subroutine mympi_wait(rqst)
    implicit none
    integer, intent(in) :: rqst
    ! local
    integer :: ierr

    call MPI_Wait(rqst,MPI_STATUS_IGNORE,ierr)

    if (ierr/=MPI_SUCCESS) then
        print*,'ERROR: mympi_wait: ',trim(get_ierr_msg(ierr))
        call MPI_Abort(MPI_COMM_WORLD,1,ierr)
    end if
end subroutine



subroutine mympi_recv_2d(A,source,tag)
    implicit none
    real(rt), intent(out) :: A(:,:)
    integer, intent(in) :: source
    integer, intent(in), optional :: tag

    if (present(tag)) then
        call mympi_recv_core(A,size(A),source,tag)
    else
        ! print*,'about to call recv_core'
        ! call mympi_recv_core(A,size(A),source,MPI_ANY_TAG)
        call mympi_recv_core(A,size(A),source,0)
    end if

end subroutine
subroutine mympi_recv_1d(A,source,tag)
    implicit none
    real(rt), intent(out) :: A(:)
    integer, intent(in) :: source
    integer, intent(in), optional :: tag

    if (present(tag)) then
        call mympi_recv_core(A,size(A),source,tag)
    else
        ! print*,'about to call recv_core'
        ! call mympi_recv_core(A,size(A),source,MPI_ANY_TAG)
        call mympi_recv_core(A,size(A),source,0)
    end if

end subroutine
subroutine mympi_recv_0d(A,source,tag)
    implicit none
    real(rt), intent(out) :: A
    integer, intent(in) :: source
    integer, intent(in), optional :: tag
    ! local
    real(rt) :: tmp(1)

    if (present(tag)) then
        call mympi_recv_core(tmp,1,source,tag)
    else
        ! call mympi_recv_core(tmp,1,source,MPI_ANY_TAG)
        call mympi_recv_core(tmp,1,source,0)
    end if

    A = tmp(1)

end subroutine


subroutine mympi_recv_core(A,n,source,tag)
    implicit none
    real(rt), intent(out) :: A(*)
    integer, intent(in) :: n,source,tag
    ! local
    integer :: ierr !, status(MPI_STATUS_SIZE)

    ! print*,'about to call recv'
    ! print*,'source: ',source
    ! print*,'tag: ',tag,'MPI_ANY_TAG',MPI_ANY_TAG
    ! print*,'MPI_STATUS_IGNORE: ',MPI_STATUS_IGNORE
    ! print*,'n',n
    ! print*,'A(1)',A(1)
    ! print*,'A(2)',A(2)
    ! print*,'A(3)',A(3)
    ! print*,'MPI_COMM_WORLD',MPI_COMM_WORLD
    call MPI_Recv(A(1), n, mympi_rt, source, tag, &
                  MPI_COMM_WORLD, MPI_STATUS_IGNORE, ierr)

    if (ierr/=MPI_SUCCESS) then
        print*,'ERROR: mympi_recv: ',trim(get_ierr_msg(ierr))
        call MPI_Abort(MPI_COMM_WORLD,1,ierr)
    end if

end subroutine 

subroutine mympi_recv_i0d(A,source,tag)
    implicit none
    integer, intent(out) :: A
    integer, intent(in) :: source
    integer, intent(in), optional :: tag
    ! local
    integer :: tmp(1)

    if (present(tag)) then
        call mympi_recv_icore(tmp,1,source,tag)
    else
        ! call mympi_recv_icore(tmp,1,source,MPI_ANY_TAG)
        call mympi_recv_icore(tmp,1,source,0)
    end if

    A = tmp(1)

end subroutine
subroutine mympi_recv_i2d(A,source,tag)
    implicit none
    integer, intent(out) :: A(:,:)
    integer, intent(in) :: source
    integer, intent(in), optional :: tag

    if (present(tag)) then
        call mympi_recv_icore(A,size(A),source,tag)
    else
        ! call mympi_recv_icore(A,size(A),source,MPI_ANY_TAG)
        call mympi_recv_icore(A,size(A),source,0)
    end if

end subroutine
subroutine mympi_recv_i1d(A,source,tag)
    implicit none
    integer, intent(out) :: A(:)
    integer, intent(in) :: source
    integer, intent(in), optional :: tag

    if (present(tag)) then
        call mympi_recv_icore(A,size(A),source,tag)
    else
        ! call mympi_recv_icore(A,size(A),source,MPI_ANY_TAG)
        call mympi_recv_icore(A,size(A),source,0)
    end if

end subroutine

subroutine mympi_recv_icore(A,n,source,tag)
    implicit none
    integer, intent(out) :: A(*)
    integer, intent(in) :: n,source,tag
    ! local
    integer :: ierr !, status(MPI_STATUS_SIZE)

    call MPI_Recv(A(1), n, MPI_INTEGER, source, tag, &
                  MPI_COMM_WORLD, MPI_STATUS_IGNORE, ierr)

    if (ierr/=MPI_SUCCESS) then
        print*,'ERROR: mympi_recv: ',trim(get_ierr_msg(ierr))
        call MPI_Abort(MPI_COMM_WORLD,1,ierr)
    end if

end subroutine 


subroutine mympi_send_2d(A,dest,tag)
    implicit none
    real(rt), intent(in) :: A(:,:)
    integer, intent(in) :: dest
    integer, intent(in), optional :: tag

    if (present(tag)) then
        call mympi_send_core(A,size(A),dest,tag)
    else
        ! NOTE: MPI_ANY_TAG may only be used in "RECV"s
        ! print*,'about to call mympi_send_core'
        call mympi_send_core(A,size(A),dest,0)
    end if

end subroutine
subroutine mympi_send_1d(A,dest,tag)
    implicit none
    real(rt), intent(in) :: A(:)
    integer, intent(in) :: dest
    integer, intent(in), optional :: tag

    if (present(tag)) then
        call mympi_send_core(A,size(A),dest,tag)
    else
        ! NOTE: MPI_ANY_TAG may only be used in "RECV"s
        ! print*,'about to call mympi_send_core'
        call mympi_send_core(A,size(A),dest,0)
    end if

end subroutine
subroutine mympi_send_0d(A,dest,tag)
    implicit none
    real(rt), intent(in) :: A
    integer, intent(in) :: dest
    integer, intent(in), optional :: tag
    ! local
    real(rt) :: tmp(1)

    tmp(1) = A

    if (present(tag)) then
        call mympi_send_core(tmp,1,dest,tag)
    else
        ! NOTE: MPI_ANY_TAG may only be used in "RECV"s
        call mympi_send_core(tmp,1,dest,0)
    end if

end subroutine

subroutine mympi_send_core(A,n,dest,tag)
    implicit none
    real(rt), intent(in) :: A(*)
    integer, intent(in) :: n,dest,tag
    ! local
    integer :: ierr

    ! print*,'about to send'
    call MPI_send(A(1), n, mympi_rt, dest, tag, &
                   MPI_COMM_WORLD, ierr)

    if (ierr/=MPI_SUCCESS) then
        print*,'ERROR: mympi_send: ',trim(get_ierr_msg(ierr))
        call MPI_Abort(MPI_COMM_WORLD,1,ierr)
    end if

end subroutine 

subroutine mympi_send_i2d(A,dest,tag)
    implicit none
    integer, intent(in) :: A(:,:)
    integer, intent(in) :: dest
    integer, intent(in), optional :: tag

    if (present(tag)) then
        call mympi_send_icore(A,size(A),dest,tag)
    else
        ! NOTE: MPI_ANY_TAG may only be used in "RECV"s
        call mympi_send_icore(A,size(A),dest,0)
    end if

end subroutine
subroutine mympi_send_i1d(A,dest,tag)
    implicit none
    integer, intent(in) :: A(:)
    integer, intent(in) :: dest
    integer, intent(in), optional :: tag

    if (present(tag)) then
        call mympi_send_icore(A,size(A),dest,tag)
    else
        ! NOTE: MPI_ANY_TAG may only be used in "RECV"s
        call mympi_send_icore(A,size(A),dest,0)
    end if

end subroutine
subroutine mympi_send_i0d(A,dest,tag)
    implicit none
    integer, intent(in) :: A
    integer, intent(in) :: dest
    integer, intent(in), optional :: tag
    ! local
    integer :: tmp(1)

    tmp(1) = A

    if (present(tag)) then
        call mympi_send_icore(tmp,1,dest,tag)
    else
        ! NOTE: MPI_ANY_TAG may only be used in "RECV"s
        call mympi_send_icore(tmp,1,dest,0)
    end if

end subroutine


subroutine mympi_send_icore(A,n,dest,tag)
    implicit none
    integer, intent(in) :: A(*)
    integer, intent(in) :: n,dest,tag
    ! local
    integer :: ierr

    call MPI_send(A(1), n, MPI_INTEGER, dest, tag, &
                   MPI_COMM_WORLD, ierr)

    if (ierr/=MPI_SUCCESS) then
        print*,'ERROR: mympi_send: ',trim(get_ierr_msg(ierr))
        call MPI_Abort(MPI_COMM_WORLD,1,ierr)
    end if

end subroutine 



subroutine mympi_allgatherv_l5d(A)
    use mod_matlab
    implicit none
    logical, intent(inout) :: A(:,:,:,:,:)
    call mympi_allgatherv_core_logical(A,size(A,1)*size(A,2)*size(A,3)*size(A,4),size(A,5))
end subroutine 
subroutine mympi_allgatherv_l4d(A)
    use mod_matlab
    implicit none
    logical, intent(inout) :: A(:,:,:,:)
    call mympi_allgatherv_core_logical(A,size(A,1)*size(A,2)*size(A,3),size(A,4))
end subroutine 
subroutine mympi_allgatherv_l2d(A)
    use mod_matlab
    implicit none
    logical, intent(inout) :: A(:,:)
    call mympi_allgatherv_core_logical(A,size(A,1),size(A,2))
end subroutine 
subroutine mympi_allgatherv_l1d(A)
    use mod_matlab
    implicit none
    logical, intent(inout) :: A(:)
    call mympi_allgatherv_core_logical(A,1,size(A,1))
end subroutine 

! This gathers data from all tasks and delivers the combined data to all tasks.
! NOTE: it assumes the work was distributed across the trailing dimension according
!       to mympi_assign_range_list.
subroutine mympi_allgatherv_6d(A)
    use mod_matlab
    implicit none
    real(rt), intent(inout) :: A(:,:,:,:,:,:)
    call mympi_allgatherv_core(A,size(A,1)*size(A,2)*size(A,3)*size(A,4)*size(A,5),size(A,6))
end subroutine 
subroutine mympi_allgatherv_5d(A)
    use mod_matlab
    implicit none
    real(rt), intent(inout) :: A(:,:,:,:,:)
    call mympi_allgatherv_core(A,size(A,1)*size(A,2)*size(A,3)*size(A,4),size(A,5))
end subroutine 
subroutine mympi_allgatherv_4d(A)
    use mod_matlab
    implicit none
    real(rt), intent(inout) :: A(:,:,:,:)
    call mympi_allgatherv_core(A,size(A,1)*size(A,2)*size(A,3),size(A,4))
end subroutine 
subroutine mympi_allgatherv_3d(A)
    use mod_matlab
    implicit none
    real(rt), intent(inout) :: A(:,:,:)
    call mympi_allgatherv_core(A,size(A,1)*size(A,2),size(A,3))
end subroutine 
subroutine mympi_allgatherv_mat(A)
    use mod_matlab
    implicit none
    real(rt), intent(inout) :: A(:,:)
    call mympi_allgatherv_core(A,size(A,1),size(A,2))
end subroutine 
subroutine mympi_allgatherv_vec(A)
    use mod_matlab
    implicit none
    real(rt), intent(inout) :: A(:)
    call mympi_allgatherv_core(A,1,size(A,1))
end subroutine 
subroutine mympi_allgatherv_core(A,nrow,ncol)
    use mod_matlab
    implicit none
    integer, intent(in) :: nrow,ncol
    real(rt), intent(inout) :: A(nrow,*)
    ! local
    real(rt), allocatable :: B(:,:)
    integer :: lo(nproc),hi(nproc),nelts(nproc),offset(nproc)
    integer :: ierr,lo_iproc,icol

    call checkinit()

    ! Get how the list was distributed
    call mympi_assign_range_list(lo,hi,ncol)

    ! Number of elements (columns)
    nelts = hi - lo + 1

    ! Offset
    offset(1) = 0
    offset(2:nproc) = cumsum(nrow*nelts(1:nproc-1))

    ! Check whether number of elements to be sent is > 0 
    if (nelts(iproc+1)>0) then
        ! If so, then first element of A to be sent occurs at lo_iproc
        lo_iproc = lo(iproc+1)
    else
        ! Otherwise, need to specify one knowing that no values are sent 
        lo_iproc = 1
    end if

    allocate(B(nrow,ncol))

    ! Call the mpi routine
    call MPI_Allgatherv(A(1,lo_iproc),& ! first element assigned to iproc
                        nrow*nelts(iproc+1),& ! total number of elements assigned to iproc --- may be zero
                        mympi_rt,& ! data type of send elements
                        B(1,1),& ! first element of receive 
                        nrow*nelts,& ! number of elements to be received by each process
                        offset,& ! Entry iproc specifies the displacement (relative to recvbuf) at which to place data from iproc
                        mympi_rt,& ! data type of receive elements
                        MPI_COMM_WORLD,&
                        ierr) 

    ! Now all has been received in B, so copy out to A (this is a restriction of MPI)
    do icol = 1,ncol
        A(1:nrow,icol) = B(:,icol)
    end do
    deallocate(B)

end subroutine

! Same as mympi_allgatherv_core but with logical
subroutine mympi_allgatherv_core_logical(A,nrow,ncol)
    use mod_matlab
    implicit none
    integer, intent(in) :: nrow,ncol
    logical, intent(inout) :: A(nrow,*)
    ! local
    logical, allocatable :: B(:,:)
    integer :: lo(nproc),hi(nproc),nelts(nproc),offset(nproc)
    integer :: ierr,lo_iproc,icol

    call checkinit()


    ! Get how the list was distributed
    call mympi_assign_range_list(lo,hi,ncol)

    ! Number of elements (columns)
    nelts = hi - lo + 1

    ! Offset
    offset(1) = 0
    offset(2:nproc) = cumsum(nrow*nelts(1:nproc-1))

    ! Check whether number of elements to be sent is > 0 
    if (nelts(iproc+1)>0) then
        ! If so, then first element of A to be sent occurs at lo_iproc
        lo_iproc = lo(iproc+1)
    else
        ! Otherwise, need to specify one knowing that no values are sent 
        lo_iproc = 1
    end if

    allocate(B(nrow,ncol))

    ! Call the mpi routine
    call MPI_Allgatherv(A(1,lo_iproc),& ! first element assigned to iproc
                        nrow*nelts(iproc+1),& ! total number of elements assigned to iproc --- may be zero
                        MPI_LOGICAL,& ! data type of send elements
                        B(1,1),& ! first element of receive 
                        nrow*nelts,& ! number of elements to be received by each process
                        offset,& ! Entry iproc specifies the displacement (relative to recvbuf) at which to place data from iproc
                        MPI_LOGICAL,& ! data type of receive elements
                        MPI_COMM_WORLD,&
                        ierr) 

    ! Now all has been received in B, so copy out to A (this is a restriction of MPI)
    do icol = 1,ncol
        A(1:nrow,icol) = B(:,icol)
    end do
    deallocate(B)


end subroutine

! Single precision version of the above
subroutine mympi_allgatherv_r44d(A)
    use mod_matlab
    implicit none
    real(4), intent(inout) :: A(:,:,:,:)
    call mympi_allgatherv_r4core(A,size(A,1)*size(A,2)*size(A,3),size(A,4))
end subroutine 
subroutine mympi_allgatherv_r43d(A)
    use mod_matlab
    implicit none
    real(4), intent(inout) :: A(:,:,:)
    call mympi_allgatherv_r4core(A,size(A,1)*size(A,2),size(A,3))
end subroutine 
subroutine mympi_allgatherv_r4mat(A)
    use mod_matlab
    implicit none
    real(4), intent(inout) :: A(:,:)
    call mympi_allgatherv_r4core(A,size(A,1),size(A,2))
end subroutine 
subroutine mympi_allgatherv_r4vec(A)
    use mod_matlab
    implicit none
    real(4), intent(inout) :: A(:)
    call mympi_allgatherv_r4core(A,1,size(A))
end subroutine 
subroutine mympi_allgatherv_r4core(A,nrow,ncol)
    use mod_matlab
    implicit none
    integer, intent(in) :: nrow,ncol
    real(4), intent(inout) :: A(nrow,*)
    ! local
    real(4), allocatable :: B(:,:)
    integer :: lo(nproc),hi(nproc),nelts(nproc),offset(nproc)
    integer :: ierr,lo_iproc,icol

    call checkinit()


    ! Get how the list was distributed
    call mympi_assign_range_list(lo,hi,ncol)

    ! Number of elements (columns)
    nelts = hi - lo + 1

    ! Offset
    offset(1) = 0
    offset(2:nproc) = cumsum(nrow*nelts(1:nproc-1))

    ! Check whether number of elements to be sent is > 0 
    if (nelts(iproc+1)>0) then
        ! If so, then first element of A to be sent occurs at lo_iproc
        lo_iproc = lo(iproc+1)
    else
        ! Otherwise, need to specify one knowing that no values are sent 
        lo_iproc = 1
    end if

    allocate(B(nrow,ncol))

    ! Call the mpi routine
    call MPI_Allgatherv(A(1,lo_iproc),& ! first element assigned to iproc
                        nrow*nelts(iproc+1),& ! total number of elements assigned to iproc --- may be zero
                        MPI_REAL4,& ! data type of send elements
                        B(1,1),& ! first element of receive 
                        nrow*nelts,& ! number of elements to be received by each process
                        offset,& ! Entry iproc specifies the displacement (relative to recvbuf) at which to place data from iproc
                        MPI_REAL4,& ! data type of receive elements
                        MPI_COMM_WORLD,&
                        ierr) 

    ! Now all has been received in B, so copy out to A (this is a restriction of MPI)
    do icol = 1,ncol
        A(1:nrow,icol) = B(:,icol)
    end do
    deallocate(B)


end subroutine


! Integer versions of the above
subroutine mympi_allgatherv_i4d(A)
    use mod_matlab
    implicit none
    integer, intent(inout) :: A(:,:,:,:)
    call mympi_allgatherv_icore(A,size(A,1)*size(A,2)*size(A,3),size(A,4))
end subroutine 
subroutine mympi_allgatherv_i3d(A)
    use mod_matlab
    implicit none
    integer, intent(inout) :: A(:,:,:)
    call mympi_allgatherv_icore(A,size(A,1)*size(A,2),size(A,3))
end subroutine 
subroutine mympi_allgatherv_imat(A)
    use mod_matlab
    implicit none
    integer, intent(inout) :: A(:,:)
    call mympi_allgatherv_icore(A,size(A,1),size(A,2))
end subroutine 
subroutine mympi_allgatherv_ivec(A)
    use mod_matlab
    implicit none
    integer, intent(inout) :: A(:)
    call mympi_allgatherv_icore(A,1,size(A))
end subroutine 
subroutine mympi_allgatherv_icore(A,nrow,ncol)
    use mod_matlab
    implicit none
    integer, intent(in) :: nrow,ncol
    integer, intent(inout) :: A(nrow,*)
    ! local
    integer, allocatable :: B(:,:)
    integer :: lo(nproc),hi(nproc),nelts(nproc),offset(nproc)
    integer :: ierr,lo_iproc,icol

    call checkinit()


    ! Get how the list was distributed
    call mympi_assign_range_list(lo,hi,ncol)

    ! Number of elements (columns)
    nelts = hi - lo + 1

    ! Offset
    offset(1) = 0
    offset(2:nproc) = cumsum(nrow*nelts(1:nproc-1))

    ! Check whether number of elements to be sent is > 0 
    if (nelts(iproc+1)>0) then
        ! If so, then first element of A to be sent occurs at lo_iproc
        lo_iproc = lo(iproc+1)
    else
        ! Otherwise, need to specify one knowing that no values are sent 
        lo_iproc = 1
    end if

    allocate(B(nrow,ncol))

    ! Call the mpi routine
    call MPI_Allgatherv(A(1,lo_iproc),& ! first element assigned to iproc
                        nrow*nelts(iproc+1),& ! total number of elements assigned to iproc --- may be zero
                        MPI_INTEGER,& ! data type of send elements
                        B(1,1),& ! first element of receive 
                        nrow*nelts,& ! number of elements to be received by each process
                        offset,& ! Entry iproc specifies the displacement (relative to recvbuf) at which to place data from iproc
                        MPI_INTEGER,& ! data type of receive elements
                        MPI_COMM_WORLD,&
                        ierr) 

    ! Now all has been received in B, so copy out to A (this is a restriction of MPI)
    do icol = 1,ncol
        A(1:nrow,icol) = B(:,icol)
    end do
    deallocate(B)



end subroutine

subroutine mympi_init()
    implicit none
    integer :: ierr,flag

    ! Test whether intialized already, and, if not, initialize
    call MPI_initialized(flag,ierr)
    if (flag==0) then 
        call MPI_init(ierr)

        ! Get the number of processors
        call MPI_Comm_size(MPI_COMM_WORLD,nproc,ierr)

        ! Get the processor id
        call MPI_Comm_rank(MPI_COMM_WORLD,iproc,ierr)

        ! Initialize mpi
        mympi_initialized = .true.

    end if

end subroutine
subroutine mympi_finalize()
    implicit none
    integer :: ierr

    call checkinit()
    call MPI_Finalize(ierr)

end subroutine 

subroutine mympi_hello_world()
    implicit none
    character(len=MPI_MAX_PROCESSOR_NAME) :: procname
    integer :: lenprocname,ierr

    call checkinit()

    ! Get the processor name
    call MPI_Get_processor_name(procname,lenprocname,ierr)

    ! Display
    write(*,'(3A,i0,A,i0,A)') 'Hello from ',trim(procname),'. I am process ',iproc,' of ',nproc,'.'

end subroutine 
    
    

end module mod_mympi

module mod_mympi_test
    use mod_mympi
    implicit none
contains
    
    ! mympi_init and mympi_finalize should be called before and after this routine
    subroutine test_mympi
        implicit none
        call test_mympi_core(1,7)
        call test_mympi_core(2,7)
        call test_mympi_core(3,7)
        call test_mympi_core(4,7)
        call test_mympi_core(5,7)
        call test_mympi_core(11,7)
        call test_mympi_core(24,7)
        call test_mympi_core(31,7)
        call test_mympi_core(1017,7)
        if (mympi_iproc()==0) print*,'test_mympi: passed'
    end subroutine
    subroutine test_mympi_core(ncol,ndim1)
        implicit none
        integer, intent(in) :: ncol, ndim1 
        real(rt), allocatable :: x_d_1d(:),x_d_2d(:,:),x_d_3d(:,:,:)
        real(4), allocatable :: x_f_1d(:),x_f_2d(:,:)
        integer, allocatable :: x_i_1d(:),x_i_2d(:,:)
        logical, allocatable :: x_l_1d(:),x_l_2d(:,:)
        real(rt), allocatable :: y_d_1d(:),y_d_2d(:,:),y_d_3d(:,:,:)
        real(4), allocatable :: y_f_1d(:),y_f_2d(:,:)
        integer, allocatable :: y_i_1d(:),y_i_2d(:,:)
        logical, allocatable :: y_l_1d(:),y_l_2d(:,:)
        integer, allocatable :: lo(:),hi(:),nelts(:),rqstlst(:)
        integer :: a,b,itmp,nproc,icol,cnt,itmp2
        real(rt) :: tmp

        nproc = mympi_nproc()
        allocate(lo(nproc),hi(nproc),nelts(nproc))
        allocate(rqstlst(ncol))

        ! Make space
        allocate(x_d_1d(ncol),x_d_2d(ndim1,ncol),x_d_3d(ndim1,ncol,3)) 
        allocate(x_f_1d(ncol),x_f_2d(ndim1,ncol)) 
        allocate(x_i_1d(ncol),x_i_2d(ndim1,ncol)) 
        allocate(x_l_1d(ncol),x_l_2d(ndim1,ncol)) 
        allocate(y_d_1d(ncol),y_d_2d(ndim1,ncol),y_d_3d(ndim1,ncol,3)) 
        allocate(y_f_1d(ncol),y_f_2d(ndim1,ncol)) 
        allocate(y_i_1d(ncol),y_i_2d(ndim1,ncol)) 
        allocate(y_l_1d(ncol),y_l_2d(ndim1,ncol)) 
        
        ! Hello world
!         call mympi_hello_world()

        ! Check list assignments
        call mympi_assign_range_list(lo,hi,ncol)

        ! Check list assignments
        nelts = hi - lo + 1
        if (any(nelts<0)) call mympi_stop('ERROR: List assignment has negative number of elements')
        if (maxval(nelts)>minval(nelts)+1) call mympi_stop('ERROR: List assignment was uneven')

        call mympi_assign_range(a,b,ncol)
        if (mympi_a(ncol)/=a) call mympi_stop('ERROR: in mympi_a')
        if (mympi_b(ncol)/=b) call mympi_stop('ERROR: in mympi_b')

        !!! Check that ever element was assigned to exactly one rank
        do icol = 1,ncol
            cnt = 0
            do itmp = 1,nproc
                if (icol>=lo(itmp) .and. icol<=hi(itmp)) cnt = cnt + 1
            end do
            if (cnt/=1) call mympi_stop('ERROR: failed to assign complete range or assigned multiply')
        end do

        ! Test barrier
        call mympi_barrier() ! prettier i/o

        ! Test mympi_allgatherv
        x_d_1d = -1e0_rt; x_d_2d = -1e0_rt
        x_f_1d = -1e0; x_f_2d = -1e0
        x_i_1d = -1  ; x_i_2d = -1
        x_l_1d = -1>0; x_l_2d = -1>0

        !!! Have each rank write its rank to x
        do icol = mympi_a(ncol),mympi_b(ncol)
            x_d_1d(  icol) = real(mympi_iproc(),rt)
            x_d_2d(:,icol) = real(mympi_iproc(),rt)
            x_f_1d(  icol) = real(mympi_iproc(),4)
            x_f_2d(:,icol) = real(mympi_iproc(),4)
            x_i_1d(  icol) = mympi_iproc()  
            x_i_2d(:,icol) = mympi_iproc()
            x_l_1d(  icol) = .true. 
            x_l_2d(:,icol) = .true.
        end do

        !!! Gather
        call mympi_allgatherv(x_d_1d)
        call mympi_allgatherv(x_d_2d)
        call mympi_allgatherv_r4(x_f_1d)
        call mympi_allgatherv_r4(x_f_2d)
        call mympi_allgatherv(x_i_1d)
        call mympi_allgatherv(x_i_2d)
        call mympi_allgatherv(x_l_1d)
        call mympi_allgatherv(x_l_2d)

        !!! Test that it is as it should be
        do icol = 1,ncol
            ! Search for who this column was assigned to
            do itmp = 1,nproc
                if (icol>=lo(itmp) .and. icol<=hi(itmp)) exit
            end do
            if (itmp<=nproc) then ! found an assignment
                itmp = itmp - 1
!                 print*,'icol',icol,'x_d_1d(icol)',x_d_1d(icol),'itmp',itmp
                if (   (x_d_1d(  icol) /= real(itmp,rt))) call mympi_stop('ERROR: failed test mympi_allgatherv d_1d')
                if (any(x_d_2d(:,icol) /= real(itmp,rt))) call mympi_stop('ERROR: failed test mympi_allgatherv d_2d')
                if (   (x_f_1d(  icol) /= real(itmp,4))) call mympi_stop('ERROR: failed test mympi_allgatherv f_1d')
                if (any(x_f_2d(:,icol) /= real(itmp,4))) call mympi_stop('ERROR: failed test mympi_allgatherv f_2d')
                if (   (x_i_1d(  icol) /= itmp))         call mympi_stop('ERROR: failed test mympi_allgatherv i_1d')
                if (any(x_i_2d(:,icol) /= itmp))         call mympi_stop('ERROR: failed test mympi_allgatherv i_2d')
                if (   (x_l_1d(  icol) .neqv. .true. ))      call mympi_stop('ERROR: failed test mympi_allgatherv l_1d')
                if (any(x_l_2d(:,icol) .neqv. .true.))       call mympi_stop('ERROR: failed test mympi_allgatherv l_2d')
            end if
        end do

        !!! Wait and display pass
        call mympi_barrier()

        ! Test reduction

        !!! Have each rank write its rank to x
        x_d_1d = real(mympi_iproc(),rt)
        x_d_2d = real(mympi_iproc(),rt)
        x_d_3d = real(mympi_iproc(),rt)

        !!! Reduce
        call random_number(y_d_1d)
        call random_number(y_d_2d)
        call random_number(y_d_3d)

        call mympi_allreduce(y_d_1d,x_d_1d,MPI_SUM)
        call mympi_allreduce(y_d_2d,x_d_2d,MPI_SUM)
        call mympi_allreduce(y_d_3d,x_d_3d,MPI_SUM)

        
        cnt = 0
        do itmp2 = 0,mympi_nproc()-1
            cnt = cnt + itmp2 + 2

        end do

        itmp = -1000000000
        call mympi_allreduce(itmp,2+mympi_iproc(),MPI_SUM)
        if (itmp/=cnt) call mympi_stop('ERROR: failed test mympi_allreduce i_0d')

        call mympi_barrier()
        stop 'checking'

        !!! Test that it is as it should be
        tmp = 0e0_rt
        do itmp = 0,nproc-1
            tmp = tmp + real(itmp,rt)
        end do
        x_d_1d = tmp
        x_d_2d = tmp
        x_d_3d = tmp

        if (any(y_d_1d/=x_d_1d)) call mympi_stop('ERROR: failed test mympi_allreduce d_1d')
        if (any(y_d_2d/=x_d_2d)) call mympi_stop('ERROR: failed test mympi_allreduce d_2d')
        if (any(y_d_3d/=x_d_3d)) call mympi_stop('ERROR: failed test mympi_allreduce d_3d')

        ! Check mympi_ibcast
        x_i_2d = 0
        if (mympi_iproc()==0) then
            do icol = 1,ncol
                x_i_2d(:,icol) = icol
            end do
        end if
        do icol = 1,ncol
            call mympi_ibcast(rqstlst(icol),x_i_2d(:,icol),0)
        end do
        call mympi_waitall(rqstlst) 
        do icol = 1,ncol
            if (any(x_i_2d(:,icol)/=icol)) call mympi_stop('ERROR: failed test mympi_ibcast')
        end do


    end subroutine

end module mod_mympi_test
module mod_root1dim
    use mod_types
    implicit none
    real(rt), parameter, private :: xtoler_def = 1e-6_rt, toler_def = 1e-6_rt
    integer, parameter, private :: printmod_def = -1, maxiter_def = 100
contains

! Tests whether the two functions values bracket a root -- i.e., have different signs
function bracketsRoot(fa,fb) result(tf)
    logical :: tf
    real(rt) :: fa,fb

    ! tf = sign(fa,fb)/=fa
    tf = sign(fa,fb)/=fa .or. fa==0e0_rt .or. fb==0e0_rt 

end function bracketsRoot

! Attempts to find n roots within a specific interval, [a,b], by checking n subintervals
! [a_i,a_{i+1}], i = 1,n where a_1:=a and a_{n+1}:=b. One must specify which intervals
! are to be checked by giving the n+1 points {a_i} points. Within each sub interval, 
! sub_brent is called with the toler, xtoler, maxiter, and printmod values desired.
subroutine sub_findallroots(roots,nroots,f,knots,toler_in,maxiter_in,xtoler_in,printmod_in)
    use mod_matlab
    implicit none    
    real(rt), intent(out) :: roots(:)
    integer, intent(out) :: nroots
    interface
        function f(x)
            use mod_types
            implicit none
            real(rt), intent(in) :: x
            real(rt) :: f
        end function f
    end interface
    real(rt), intent(in) :: knots(:)
    real(rt), intent(in), optional :: toler_in,xtoler_in
    integer, intent(in), optional :: maxiter_in,printmod_in
    ! local
    integer :: maxiter, printmod
    real(rt) :: toler, xtoler
    real(rt) :: fi, fip1
    integer :: i,n,iroots

    toler = 0e0_rt; if (present(toler_in)) toler = toler_in ! Same custom toler used in brent
    xtoler = xtoler_def; if (present(xtoler_in)) xtoler = xtoler_in
    maxiter = maxiter_def; if (present(maxiter_in)) maxiter = maxiter_in    
    printmod = printmod_def; if (present(printmod_in)) printmod = printmod_in

    if (size(roots)<size(knots)-1) STOP 'ERROR: sub_findallroots: roots vector not big enough'
    if (size(knots)<2) STOP 'ERROR: sub_findallroots: knots vector must have length of at least two'

    n = size(knots)-1 ! number of intervals


    fi = f(knots(1))
    iroots = 0

    do i = 1,n

        fip1 = f(knots(i+1))

        if (printmod>=1) then
            write(*,*) 'FA: Checking Interval ',i,' of ',n,'. Root found: ',bracketsRoot(fi,fip1)
        end if

        if (bracketsRoot(fi,fip1)) then
            iroots = iroots + 1
            call sub_brent(roots(iroots),f,knots(i),knots(i+1),&
                toler_in=toler,xtoler_in=xtoler,maxiter_in=maxiter,printmod_in=printmod)
        ! else, no root!
        end if

        fi = fip1

    end do

    nroots = iroots
    
    


end subroutine sub_findallroots

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! sub_nsect_omp
!
! Uses "n"-section to find a root f(x) = 0 in a bracket [a,b]
! At each step, a function evaluation is assigned to each thread. So, e.g.
! two threads will evaluated (a+b)/3 and 2*(a+b)/3. 
!
! Optional arguments are the evaluation of f at a AND or b.  
!
subroutine sub_nsect_omp(x,f,a_in,b_in,xtoler_in,printmod_in)
    implicit none
    real(rt), intent(out) :: x
    interface
        function f(x)
            use mod_types
            implicit none
            real(rt), intent(in) :: x
            real(rt) :: f
        end function f
    end interface
    real(rt), intent(in) :: a_in,b_in
    real(rt), intent(in), optional :: xtoler_in
    integer, intent(in), optional :: printmod_in
    !local
    real(rt), allocatable :: fx(:)
    real(rt) :: a,b,fa,fb,xtoler,last
    integer :: printmod,iter,i,nthreads
    real(rt), parameter :: eps = sqrt(epsilon(0e0_rt))

    ! Get initial bracket
    a = a_in
    b = b_in

    ! If a>b, switch them
    if (a>b) then
        last = a
        a = b
        b = last
        write(0,*) 'WARNING: a>b, switching'
    end if
    
    ! Get or set parameter values
    printmod = printmod_def; if (present(printmod_in)) printmod = printmod_in
    xtoler = xtoler_def; if (present(xtoler_in)) xtoler = xtoler_in


    ! Evaluate at end points and also at the 
    nthreads = 0
    !$OMP PARALLEL REDUCTION(+:nthreads)
    nthreads = nthreads + 1
    !$OMP END PARALLEL

    ! If running in serial, call sub_bisect instead
    if (nthreads==1) then
        write(0,*) 'WARNING: sub_nsect_omp being called with only one thread, calling sub_bisect instead.'
        call sub_bisect(x,f,a,b,xtoler_in=xtoler,printmod_in=printmod)
        return
    end if

    ! Now nthreads>=2

    allocate(fx(0:nthreads+1))

    !$OMP PARALLEL DO DEFAULT(SHARED)
    do i = 1,nthreads
        fx(i) = f(a + (i-1)*(b-a)/(nthreads-1))
    end do
    !$OMP END PARALLEL DO

    ! Replace [a,b] with the tightest interval possible (if multiple roots, select one)
    last = b-a
    do i = 1,nthreads-1
        if (bracketsRoot(fx(i),fx(i+1))) then
            b = a +    i *last/(nthreads-1)
            a = a + (i-1)*last/(nthreads-1)
            fa = fx(i)
            fb = fx(i+1)
            exit
        end if
    end do

    ! Check for an error
    if (b<a) then
        write(0,*) 'WARNING: nsect: may have found multiple roots: b<a: a',a,' b ',b
        x = merge(a,b,abs(fa)<abs(fb))
        return
    end if
    
    ! Check for bracketing
    if (.not. bracketsRoot(fa,fb)) then
        if (printmod>=-1) write(0,'(A)') 'nsect: WARNING: zero not bracketed, returning best endpoint'
        x = merge(a,b,abs(fa)<abs(fb))
        return
    end if

    ! Output progress
    iter = 0
    if (printmod==1) write(*,'(A,i4,A,e13.5,A,e13.5,A,e13.5,A,e13.5)') 'nsect_omp: iter ',iter,' a',a,' b',b,' fa',fa,' fb',fb

    ! Main loop
    do iter = 1,1000 ! should be at least halving at each step, so this should be fine log2(huge(0e0_rt)) is roughly 1000
        
        ! Save current precision
        last = b-a

        ! Check convergence criterion
        if (last<=eps*abs(b)+xtoler) then 
            x = merge(a,b,abs(fa)<abs(fb))
            exit
        end if
    
        ! N-section evaluation
        !$OMP PARALLEL DO DEFAULT(SHARED)
        do i = 1,nthreads
            fx(i) = f(a + i*(b-a)/(nthreads+1))
        end do
        !$OMP END PARALLEL DO

        ! Get new bracket as tight as possible, hopefully [x(i),x(i+1)] but don't theoretical guarantee
        !  [x(i),x]
        fx(0) = fa
        fx(nthreads+1) = fb
        do i = 0,nthreads
            if (bracketsRoot(fx(i),fx(i+1))) then
                b = a + (i+1)*last/(nthreads+1)
                a = a +    i *last/(nthreads+1)
                fa = fx(i)
                fb = fx(i+1)
                exit
            end if
        end do

        ! Output progress
        if (printmod>0 .and. mod(iter,printmod)==0) &
            write(*,"(A,i4,A,e13.5,A,e13.5,A,e13.5,A,e13.5,A,e13.5)") &
            'nsect_omp: iter ',iter,' a',a,' b',b,' fa',fa,' fb',fb,' ratio',(b-a)/last

    end do

contains
    
    logical function samesign(c,d)
        real(rt), intent(in) :: c,d
        samesign = (c<=0e0_rt .and. d<=0e0_rt) .or. (c>=0e0_rt .and. d>=0d0)
    end function 

end subroutine sub_nsect_omp

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! sub_nsect_mpi
!
! Uses "n"-section to find a root f(x) = 0 in a bracket [a,b]
! At each step, a function evaluation is assigned to each thread. So, e.g.
! two threads will evaluated (a+b)/3 and 2*(a+b)/3. 
!
! Optional arguments are the evaluation of f at a AND or b.  
!
subroutine sub_nsect_mpi(x,f,a_in,b_in,xtoler_in,printmod_in)
    use mod_mympi
    implicit none
    real(rt), intent(out) :: x
    interface
        function f(x)
            use mod_types
            implicit none
            real(rt), intent(in) :: x
            real(rt) :: f
        end function f
    end interface
    real(rt), intent(in) :: a_in,b_in
    real(rt), intent(in), optional :: xtoler_in
    integer, intent(in), optional :: printmod_in
    !local
    real(rt), allocatable :: fx(:)
    real(rt) :: a,b,fa,fb,xtoler,last
    integer :: printmod,iter,i,nproc
    real(rt), parameter :: eps = sqrt(epsilon(0e0_rt))

    ! Get initial bracket
    a = a_in
    b = b_in

    ! If a>b, switch them
    if (a>b) then
        last = a
        a = b
        b = last
        write(0,*) 'WARNING: a>b, switching'
    end if
    
    ! Get or set parameter values
    printmod = printmod_def; if (present(printmod_in)) printmod = printmod_in
    xtoler = xtoler_def; if (present(xtoler_in)) xtoler = xtoler_in

    ! Get number of ranks
    nproc = mympi_nproc()

    ! If running in serial, call sub_bisect instead
    if (nproc==1) then
        write(0,*) 'WARNING: sub_nsect_mpi is being called with only one rank, calling sub_bisect instead.'
        call sub_bisect(x,f,a,b,xtoler_in=xtoler,printmod_in=printmod)
        return
    end if

    ! If not master, don't display info 
    if (mympi_iproc()/=0) printmod = -1

    ! Now nproc>=2

    allocate(fx(0:nproc+1))

    ! Do the initial check
    i = mympi_iproc()+1
    fx(i) = f(a + (i-1)*(b-a)/(nproc-1))

    ! Collect from all the other nodes
    call mympi_allgatherv(fx(1:nproc))

    ! Replace [a,b] with the tightest interval possible (if multiple roots, select one)
    last = b-a
    do i = 1,nproc-1
        if (bracketsRoot(fx(i),fx(i+1))) then
            b = a +    i *last/(nproc-1)
            a = a + (i-1)*last/(nproc-1)
            fa = fx(i)
            fb = fx(i+1)
            exit
        end if
    end do

    ! Check for an error
    if (b<a) then
        write(0,*) 'WARNING: nsect: may have found multiple roots: b<a: a',a,' b ',b
        x = merge(a,b,abs(fa)<abs(fb))
        call mympi_barrier()
        return
    end if
    
    ! Check for bracketing
    if (.not. bracketsRoot(fa,fb)) then
        if (printmod>=-1) write(0,'(A)') 'nsect: WARNING: zero not bracketed, returning best endpoint'
        x = merge(a,b,abs(fa)<abs(fb))
        call mympi_barrier()
        return
    end if

    ! Output progress
    iter = 0
    if (printmod==1) &
        write(*,'(A,i4,A,e13.5,A,e13.5,A,e13.5,A,e13.5)') 'nsect_mpi: iter ',iter,' a',a,' b',b,' fa',fa,' fb',fb

    ! Main loop
    do iter = 1,1000 ! should be at least halving at each step, so this should be fine log2(huge(0e0_rt)) is roughly 1000
        
        ! Save current precision
        last = b-a

        ! Check convergence criterion
        if (last<=eps*abs(b)+xtoler) then 
            x = merge(a,b,abs(fa)<abs(fb))
            exit
        end if
    
        ! N-section evaluation
        i = mympi_iproc()+1
        fx(i) = f(a + i*(b-a)/(nproc+1))

        ! Collect
        call mympi_allgatherv(fx(1:nproc))

        ! Get new bracket as tight as possible, hopefully [x(i),x(i+1)] but don't theoretical guarantee
        !  [x(i),x]
        fx(0) = fa
        fx(nproc+1) = fb
        do i = 0,nproc
            if (bracketsRoot(fx(i),fx(i+1))) then
                b = a + (i+1)*last/(nproc+1)
                a = a +    i *last/(nproc+1)
                fa = fx(i)
                fb = fx(i+1)
                exit
            end if
        end do

        ! Output progress
        if (printmod>0 .and. mod(iter,printmod)==0) &
            write(*,"(A,i4,A,e13.5,A,e13.5,A,e13.5,A,e13.5,A,e13.5)") &
            'nsect_mpi: iter ',iter,' a',a,' b',b,' fa',fa,' fb',fb,' ratio',(b-a)/last

    end do

    ! Wait for all to finish
    call mympi_barrier()

contains
    
    logical function samesign(c,d)
        real(rt), intent(in) :: c,d
        samesign = (c<=0e0_rt .and. d<=0e0_rt) .or. (c>=0e0_rt .and. d>=0d0)
    end function 

end subroutine sub_nsect_mpi



!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! sub_bisect
!
! Uses bisection to find a root f(x) = 0 in a bracket [a,b]
! f should be defined as in the below interface block as just f(x)
!
! Optional arguments are the evaluation of f at a AND or b.  Alternatively,
! just the correct sign of fa AND fb can be passed.
!
subroutine sub_bisect(x_out,f,a_in,b_in,toler_in,maxiter_in,fa_in,fb_in,xtoler_in,printmod_in)
    implicit none
    real(rt), intent(out) :: x_out
    interface
        function f(x)
            use mod_types
            implicit none
            real(rt), intent(in) :: x
            real(rt) :: f
        end function f
    end interface
    real(rt), intent(in) :: a_in,b_in
    real(rt), intent(in), optional :: toler_in, fa_in,fb_in, xtoler_in
    integer, intent(in), optional :: maxiter_in, printmod_in
    !local
    real(rt) :: a,b,x,fa,fb,fx,res,toler,xtoler,t
    integer :: maxiter,printmod, iter
    
    ! Get initial bracket
    a = a_in
    b = b_in
    
    ! Get or set parameter values
    printmod = printmod_def; if (present(printmod_in)) printmod = printmod_in
    xtoler = xtoler_def; if (present(xtoler_in)) xtoler = xtoler_in
    toler = toler_def; if (present(toler_in)) toler = toler_in
    maxiter = maxiter_def; if (present(maxiter_in)) maxiter = maxiter_in

    
    ! Get or set bracket function values
    if (present(fa_in)) then 
        fa = fa_in
    else
        fa = f(a)
    end if
    if (present(fb_in)) then 
        fb = fb_in
    else
        fb = f(b)
    end if

    ! Check whether a zero is at one bracket 
    if (abs(fa)<toler) then
        if (printmod>0) write(*,'(A)') 'bisect: Converged successfully: |f(a)|<toler'
        x_out = a
        return
    elseif (abs(fb)<toler) then
        if (printmod>0) write(*,'(A)') 'bisect: Converged successfully: |f(b)|<toler'
        x_out = b
        return
    end if
    
    ! Check for bracketing
    if (.not. bracketsRoot(fa,fb)) then
        if (printmod>=-1) write(*,'(A)') 'bisect: WARNING: zero not bracketed, returning best endpoint'
        if (abs(fa)<abs(fb)) then
            x_out = a
        else
            x_out = b
        end if
        return
    end if
    
    ! Main loop
    do iter = 1,maxiter
        
        ! Output progress
        if (printmod>0 .and. mod(iter,printmod)==0) &
            write(*,"(A,i4,A,e13.5,A,e13.5,A,e13.5,A,e13.5)") &
            'bisect: iter ',iter,' a',a,' b',b,' fa',fa,' fb',fb

        ! Check convergence criterion
        t = 2e0_rt*epsilon(0._rt)*(a+b) 

        if (abs(b-a)<=max(xtoler,t)) then 

            if (printmod>0) write(*,*) 'bisect: Converged: |b-a|<xtoler. Returning midpoint.'

            x_out = (a+b)/2e0_rt
            exit
        end if
    
        ! Midpoint evaluation
        x = (a+b)/2e0_rt
        fx = f(x)

        
        ! Check other convergence criterion
        if (abs(fx)<toler) then
            if (printmod>0) then 
                write(*,'(A,e13.5,A,e13.5)') 'bisect: Converged: |f(x)|<toler. x ', x,' fx',fx
            end if
            x_out = x
            return
        end if

        ! Replace appropriate endpoint 
        if (bracketsRoot(fa,fx)) then
            b = x
            fb = fx
        else
            a = x
            fa = fx
        end if

        if (iter==maxiter) then
            if (printmod>0) write(*,*) 'bisect: Failed to converge: max iterations reached'
            x_out = merge(a,b,abs(fa)<abs(fb))
        end if

    end do

end subroutine sub_bisect

! Ridder's method
! There are not very good notes out there on this method. 
! subroutine sub_ridders(x_out,f,a_in,b_in,toler_in,maxiter_in,fa_in,fb_in,xtoler_in,printmod_in)
!     implicit none
!     real(rt), intent(out) :: x_out
!     interface
!         function f(x)
!             use mod_types
!             implicit none
!             real(rt), intent(in) :: x
!             real(rt) :: f
!         end function f
!     end interface
!     real(rt), intent(in) :: a_in,b_in
!     real(rt), intent(in), optional :: toler_in,fa_in,fb_in,xtoler_in
!     integer, intent(in), optional :: maxiter_in, printmod_in
!     !local
!     real(rt) :: toler, xtoler
!     integer :: maxiter, printmod, iter, minind(1)
!     real(rt) :: a, b, fa, fb, m, fm, r, fr
!     character(len=6) :: step
!     logical :: brack(3)
!     real(rt) :: dist(3)
!     
! 
!     ! Set and replace default values
!     toler = toler_def; if (present(toler_in)) toler = toler_in
!     xtoler = xtoler_def; if (present(xtoler_in)) xtoler = xtoler_in
!     maxiter = maxiter_def; if (present(maxiter_in)) maxiter = maxiter_in
!     printmod = printmod_def; if (present(printmod_in)) printmod = printmod_in
! 
!     ! Copy initial bracket
!     a = a_in
!     b = b_in
!     
!     ! Evaluate at brackets
!     if (present(fa_in)) then
!         fa = fa_in
!     else
!         fa = f(a)
!     end if
!     if (present(fb_in)) then
!         fb = fb_in
!     else
!         fb = f(b)
!     end if
! 
!     ! Check that a<b. If not, switch them
!     if (a>b) then
!         ! Use m for temp storage of b
!         m = b 
!         fm = fb
!         ! Replace b with a
!         b = a
!         fb = fa
!         ! Replace a with b
!         a = m
!         fa = fm
!     end if
! 
!     ! Check for covergence
!     if (abs(fa)<toler) then
!        x_out = a
!        return
!     elseif (abs(fb)<toler) then
!         x_out = b
!         return
!     end if
! 
!     ! If not bracketed, return error
!     if (.not. bracketsRoot(fa,fb)) then
!         x_out = merge(a,b,abs(fa)<abs(fb))
!         print*,'sub_ridders: WARNING: root is not bracketed, returning best'
!         return
!     end if
! 
!     ! Main loop
!     step = 'Init'
!     do iter = 1,maxiter
! 
!         ! Report progress
!         if (printmod>0 .and. mod(iter,printmod)==0) &
!             write(*,'(A,i3,A,e13.5,A,e13.5,A,e13.5,A,e13.5,A,A)') &
!             'ridders: iter ',iter,' a',a, ' b',b,' fa',fa, ' fb',fb, ' step ', trim(step) 
!         
!         ! Check stopping criterion
!         if (abs(b-a)<xtoler) then
!             x_out = (b+a)/2e0_rt
!             if (printmod>0) write(*,'(A)') 'ridders: Failed to converge: |b-a|<xtoler'
!             return
!         end if
! 
!         ! Midpoint
!         m = (b+a)/2e0_rt
!         fm = f(m)
!         if (abs(fm)<toler) then
!             x_out = m
!             if (printmod>0) write(*,'(A,e13.5,A,e13.5)') 'ridders: Converged: |f(m)|<toler. m',m,' fm',fm 
!             return
!         end if
! 
!         ! Exponential point (ridders pt)
!         if (fa>0e0_rt) then
!             r = m + (m-a)*fm/sqrt(fm**2-fa*fb)
!         else
!             r = m - (m-a)*fm/sqrt(fm**2-fa*fb)
!         end if
!         fr = f(r)
!         if (abs(fr)<toler) then
!             x_out = r
!             if (printmod>0) write(*,'(A,e13.5,A,e13.5)') 'ridders: Converged: |f(r)|<toler. r ', r, ' fr',fr 
!             return
!         end if
! 
!         ! Now among possible bracketing intervals, find the one with least length. 
!         ! There are 5 possible brackets. But I believe after extensive testing that 
!         ! [a,m] and [m,b] are always dominated. So, I am just not even considering these.
!         ! However, if I'm wrong this could cause problems with slow convergence (it will still
!         ! always convergence, but may do so slowly). 
!         brack(1) = bracketsRoot(fr,fm)
!         brack(2) = bracketsRoot(fa,fr)
!         brack(3) = bracketsRoot(fr,fb)
!         
!         dist(1) = abs(r-m)
!         dist(2) = abs(a-r)
!         dist(3) = abs(r-b)
!         
!         ! Get location of smallest bracket 
!         minind = minloc(dist,mask=brack)
!         
!         ! Replace bracket
!         select case (minind(1))
!             case (1)
!                 ! [r,m] best
!                 if (r<m) then
!                     a = r
!                     fa = fr
!                     b = m
!                     fb = fm
!                 else
!                     a = m
!                     fa = fm
!                     b = r
!                     fb = fr
!                 end if
!                 step = 'Best'
!             case (2)
!                 ! [a,r] best
!                 b = r
!                 fb = fr
!                 step = 'Good'
!             case (3)
!                 ! [r,b] best
!                 a = r
!                 fa = fr
!                 step = 'Good'
!         end select
! 
!     end do
! 
! end subroutine sub_ridders

! Written using the FORTRAN program zero in "Algorithms for Minimization 
! without Derivatives" by Brent. 
! 
! Description: 
!       Returns a zero x of the function f in the given interval [a,b], to within a tolerance 6macheps|x| + 2t, 
! where macheps is the relative machine precision and t is a positive tolerance. The procedure assumes that 
! f(a) and f(b) have different signs.
subroutine sub_brent(x,f,a_in,b_in,toler_in,maxiter_in,fa_in,fb_in,xtoler_in,printmod_in)
    use mod_matlab
    implicit none    
    real(rt), intent(out) :: x
    interface
        function f(x)
            use mod_types
            implicit none
            real(rt), intent(in) :: x
            real(rt) :: f
        end function f
    end interface
    real(rt), intent(in) :: a_in,b_in
    real(rt), intent(in), optional :: toler_in,fa_in,fb_in,xtoler_in
    integer, intent(in), optional :: maxiter_in, printmod_in
    ! local
    real(rt), parameter :: machep = epsilon(0e0_rt)
    real(rt) :: a,b,c,fa,fb,fc,toler,xtoler,e,d,m,p,q,tol,t,r,s
    integer :: maxiter,printmod,iter
    character(len=6) :: step

    ! Set of get parameters
    toler = 0e0_rt; if (present(toler_in)) toler = toler_in ! Better to use custom toler here
    xtoler = xtoler_def; if (present(xtoler_in)) xtoler = xtoler_in
    maxiter = maxiter_def; if (present(maxiter_in)) maxiter = maxiter_in    
    printmod = printmod_def; if (present(printmod_in)) printmod = printmod_in

    ! Set the user chosen tolerance t to xtoler
    if (xtoler<0e0_rt) then
        if (printmod>=-1) print*,'WARNING: xtoler must be positive. Resetting xtoler.'
        xtoler = 0e0_rt
    end if
    t = xtoler
    
    ! Get initial bracket
    a=a_in
    b=b_in
    if (present(fa_in)) then
        fa = fa_in
    else
        fa = f(a)
    end if

    if (printmod==1) write(*,"(2(A,e13.5))") 'brent: init       a', a, ' fa', fa

    if (present(fb_in)) then
        fb = fb_in
    else
        fb = f(b)
    end if

    if (printmod==1) write(*,"(2(A,e13.5))") 'brent: init       b', b, ' fb', fb

    ! Test whether root is bracketed
    if (.not. bracketsRoot(fa,fb)) then
        if (abs(fa)<abs(fb)) then
            if (printmod>=-1) write(*,*) 'brent: WARNING: root is not bracketed, returning best endpoint a'
            x = a
        else
            if (printmod>=-1) write(*,*) 'brent: WARNING: root is not bracketed, returning best endpoint b'
            x = b
        end if
        return
    end if

    step = 'init'

    ! At any point in time, b is the best guess of the root, a is the previous value of b, 
    ! and the root is bracketed by b and c.
    do iter = 1,maxiter

        if (iter==1 .or. (fb>0e0_rt .and. fc>0e0_rt) .or. (fb<=0e0_rt .and. fc<=0d0)) then
            c = a
            fc = fa
            e = b - a
            d = e
        end if

        ! If c is strictly better than b, swap b and c so b is the best guess. 
        if (abs(fc)<abs(fb)) then
            a = b
            b = c
            c  = a
            fa = fb
            fb = fc
            fc = fa
        end if

        ! Set the tolerance. Note: brent is very careful with these things, so don't deviate from this.
        tol = 2e0_rt*machep*abs(b) + t

        ! Determine what half the length of the bracket [b,c] is
        m = .5e0_rt*(c-b)

        ! If taking a bisection step would move the guess of the root less than tol, then return b the best guess.
        if ((abs(m)<=tol) .or. (fb==0e0_rt)) then
            x = b
            return
        end if

        ! Display info
        if (printmod>0 .and. mod(iter,printmod)==0) &
            write(*,"(A,i5,A,e13.5,A,e13.5,A,A)") &
            'brent: iter ',iter,' b',b,' fb',fb,' ',step

        ! If still here, then check whether need to do bisection or can do interpolation
        if ((abs(e)>=tol) .and. (abs(fa)>abs(fb))) then
            s = fb/fa
            if (a/=c) then
                ! Inverse quadratic interpolation
                q = fa/fc
                r = fb/fc
                p = s*(2e0_rt*m*q*(q-r) - (b-a)*(r-1e0_rt))
                q = (q-1e0_rt)*(r-1e0_rt)*(s-1e0_rt)
                
                step = 'quad'
            else
                ! Linear interpolation
                p = 2e0_rt*m*s
                q = 1e0_rt-s

                step = 'linear'
            end if

            ! Ensure p is positive
            if (p<=0e0_rt) then
                p = -p
            else
                q = -q
            end if

            s = e
            e = d
            if ((2e0_rt*p>=3e0_rt*m*q-abs(tol*q)) .or. &
                (p>=abs(.5e0_rt*s*q))) then
                ! Interpolation step failed to produce good step, bisect instead
                e = m
                d = m ! m is half the distance between b and c
                step = 'bisect'
            else
                !  Do interpolation step (either quadratic or linear)
                d = p/q
            end if
        else

            ! Do bisection step
            e = m 
            d = m
            
        end if

        ! Get new points. 
        !! Replace a (the old b) with b.
        a = b
        fa = fb

        !!! Increment b by d if that is greater than the tolerance. O/w, increment by tol.
        if (abs(d)<=tol) then
            ! m is .5*(c-b) with the bracket either [b,c] or [c,b]. 
            if (m > 0e0_rt) then
                ! If m>0e0_rt, then bracket is [b,c] so move towards c by tol
                b = b + tol
            else
                ! If m<=0e0_rt, then bracket is [c,b] so move towards c by tol
                b = b - tol
            end if
        else
            b = b + d
        end if

        !!! Evaluate at the new point
        fb = f(b)

        ! Check my custom tolerance 
        if (abs(fb)<toler) then
            x = b
            return
        end if
            
    end do

end subroutine sub_brent

!Adapted from Numerical Recipes
! 
! Because of the license, I am able to use this ON ONE COMPUTER!!! And that is only if I bought the book.
! So I did order the book, and I used this for one project (the garnish one). I can not use it on any
! other projects (since I'll be working on different computers). Use bisection or program my own routine
! for the next one --- this one is far too close to the original.
! 
subroutine sub_brent_old(x_out,f,a_in,b_in,toler_in,maxiter_in,fa_in,fb_in,xtoler_in,printmod_in,c_in,fc_in)
    use mod_matlab
    implicit none    
    real(rt), intent(out) :: x_out
    interface
        function f(x)
            use mod_types
            implicit none
            real(rt), intent(in) :: x
            real(rt) :: f
        end function f
    end interface
    real(rt), intent(in) :: a_in,b_in
    real(rt), intent(in), optional :: c_in,toler_in,fa_in,fb_in,fc_in,xtoler_in
    integer, intent(in), optional :: maxiter_in, printmod_in
    !local
    real(rt), parameter :: eps = epsilon(0.e0_rt) 
    integer :: maxiter
    real(rt) :: a,b,c,d,e,fa,fb,fc,p,q,r,s,tol1,xm,zbrent
    real(rt) :: toler,xtoler 
    integer :: iter,printmod,hist_iter

    ! Set of get parameters
    toler = toler_def; if (present(toler_in)) toler = toler_in
    xtoler = xtoler_def; if (present(xtoler_in)) xtoler = xtoler_in
    maxiter = maxiter_def; if (present(maxiter_in)) maxiter = maxiter_in    
    printmod = printmod_def; if (present(printmod_in)) printmod = printmod_in
    
    ! Get initial bracket
    a=a_in
    b=b_in

    if (.not. present(fa_in)) then
        fa = f(a)
    else
        fa = fa_in
    end if
    if (.not. present(fb_in)) then
        fb = f(b)
    else
        fb = fb_in
    end if

    if (.not. bracketsRoot(fa,fb)) then
        if (abs(fa)<abs(fb)) then
            write(*,*) 'brent: WARNING: root is not bracketed, returning best endpoint lb'
        else
            write(*,*) 'brent: WARNING: root is not bracketed, returning best endpoint ub'
        end if
        x_out = merge(a,b,abs(fa)<abs(fb))         
        return
    end if

    c = b
    fc = fb
    
    !main solver
    do iter=1,maxiter

        if((fb>0.e0_rt .AND. fc>0.e0_rt).or.(fb<0.e0_rt .AND. fc<0.d0)) then
            c=a !Rename a, b, c AND adjust bounding interval d.
            fc=fa
            d=b-a
            e=d
        endif
        if(abs(fc)<abs(fb)) then
            a=b
            b=c
            c=a
            fa=fb
            fb=fc
            fc=fa
        endif
        
        if (printmod>0 .and. mod(iter,printmod)==0) write(*,"(A,i5,A,e13.5,A,e13.5)") 'brent: iter ',iter,' b',b,' fb',fb
        
        tol1=2.e0_rt*EPS*abs(b)+0.5e0_rt*toler !Convergence check.
        xm=.5e0_rt*(c-b)

        !ORIGINAL CODE HAD THIS TERMINATION CRITERION
!         if(abs(xm)<=tol1 .or. fb==0.e0_rt) then
!             zbrent=b
!             exit !return
!         endif

        !MY CODE HAS THE TERM CRITERION
        if (abs(fb)<toler) then
            if (printmod>0) write(*,'(A)') 'brent: Converged successfully: |f(x)|<toler'
            x_out = b
            return
        elseif (abs(xm)<=tol1) then
            if (printmod>0) write(*,'(A)') 'brent: Failed to converge: change(x)<xtoler'
            x_out = b
            return
        end if

!         !MY CODE HAS THIS TERMINATION CRITERION
!         if (abs(fb)<toler .OR. abs(c-b)<xtoler) then
!             zbrent = b
!             exit
!         end if

        if(abs(e)>=tol1 .AND. abs(fa)>abs(fb)) then
            s=fb/fa !Attempt inverse quadratic interpolation.
            if(a==c) then
                p=2.e0_rt*xm*s
                q=1.e0_rt-s
            else
                q=fa/fc
                r=fb/fc
                p=s*(2.e0_rt*xm*q*(q-r)-(b-a)*(r-1.e0_rt))
                q=(q-1.e0_rt)*(r-1.e0_rt)*(s-1.e0_rt)
            endif
            if(p>0.e0_rt) q=-q !Check whether in bounds.
            p=abs(p)
            if(2.e0_rt*p < min(3.e0_rt*xm*q-abs(tol1*q),abs(e*q))) then
                e=d !Accept interpolation.
                d=p/q
            else
                d=xm !Interpolation failed, use bisection.
                e=d
            endif
        else !Bounds decreasing too slowly, use bisection.
            d=xm
            e=d
        endif
        a=b !Move last best guess to a.
        fa=fb
        if(abs(d) > tol1) then !Evaluate new trial root.
            b=b+d
        else
            b=b+sign(tol1,xm)
        endif
        fb=f(b)
    enddo

    if (printmod>0) write(*,*) 'brent: Failed to converge: max iters reached'
    x_out = b
    
end subroutine sub_brent_old

!OLD BISECTION ROUTINE
! subroutine sub_bisect(x_out,f,a_in,b_in,toler_in,maxiter_in,fa_in,fb_in,xtoler_in,printmod_in)
!     implicit none
!     real(rt), intent(out) :: x_out
!     interface
!         function f(x)
!         implicit none
!         real(rt), intent(in) :: x
!         real(rt) :: f
!         end function f
!     end interface
!     real(rt), intent(in) :: a_in,b_in
!     real(rt), intent(in), optional :: toler_in, fa_in,fb_in, xtoler_in
!     integer, intent(in), optional :: maxiter_in, printmod_in
!     !local
!     real(rt) :: a,b,x,fa,fb,fx,res,toler,xtoler, bestx,bestf
!     logical :: fapos, faneg, case1,case2,case3,case4,case5
!     integer :: maxiter,printmod, iter
!     
!     a = a_in
!     b = b_in
!     
!     !default values
!     printmod = 1
!     xtoler = 1.d-9
!     toler = 1.d-6
!     maxiter = 1000
!     if (present(printmod_in)) printmod = printmod_in
!     if (present(xtoler_in)) xtoler = xtoler_in
!     if (present(toler_in)) toler = toler_in
!     if (present(maxiter_in)) maxiter = maxiter_in
!     !check input values
!     if (xtoler<epsilon(0.e0_rt)*10.e0_rt) then
!         write(*,*) 'sub_bisect: warning: xtoler is set beyond machine precision.  increasing. '
!         xtoler = epsilon(0.e0_rt)*10.e0_rt
!     end if
!     
!     if (present(fa_in)) then 
!         fa = fa_in
!     else
!         fa = f(a)
!     end if
!     if (present(fb_in)) then 
!         fb = fb_in
!     else
!         fb = f(b)
!     end if
!     
!     if ((fa>0.e0_rt .AND. fb>0.e0_rt) .OR. (fa<0.e0_rt .AND. fb<0.d0)) then
!         write(*,*) 'sub_bisect: zero not bracketed:', fa, fb
!         x_out = merge(a,b,abs(fa)<abs(fb))
!         return
!     end if
!     
!     !Main routine
!     bestx = merge(a,b,abs(fa)<abs(fb))
!     bestf = merge(fa,fb,abs(fa)<abs(fb))
!     fapos = fa>=0.e0_rt
!     faneg = fa<=0.e0_rt
!     res = 1d10
!     if (printmod/=-1) write(*,"('|~~~~~~~~~~~~~~Beginning Bisection Algorithm~~~~~~~~~~~~~~~~~~|')")
!     if (printmod/=-1) write(*,"('| iter,            a,            b,            x,           fx|')")
!     iter = 1
!     do while(res>toler)
!     
!         x = .5e0_rt*a + .5e0_rt*b    
!         if (abs(x-a)<=xtoler) then 
!             write(*,*) 'sub_bisect: Warning: xtoler exceeded'
!             exit
!         end if
!         fx = f(x)
!     
!         bestx = merge(bestx,x,abs(bestf)<abs(fx))
!         bestf = merge(bestf,fx,abs(bestf)<abs(fx))
!         
!         if (printmod/=-1) write(*,"('|',i5,4e14.5,'|')") iter,a,b,x,fx
!     
!         case1 = fx<0.e0_rt .AND. faneg
!         case2 = fx>0.e0_rt .AND. faneg
!         case3 = fx<0.e0_rt .AND. fapos
!         case4 = fx>0.e0_rt .AND. fapos
!         case5 = fx==0.e0_rt
!     
!         if (case1) then
!             fa = fx
!             a = x
!         elseif (case2) then
!             fb = fx
!             b = x
!         elseif (case3) then
!             fb = fx
!             b = x
!         elseif (case4) then
!             fa = fx
!             a = x
!         elseif (case5) then
!             exit
!         end if
!          
!         res = abs(fx)
!         if (res<=toler) then 
!             if (printmod/=-1) write(*,"('|',i5,4e14.5,'|')") 0, 0.e0_rt, 0.e0_rt, x, fx
!             if (printmod/=-1) write(*,"('|~~~~~~~~~~~~~~Converged Successfully~~~~~~~~~~~~~~~~~~~~~~~~~|')")
!         end if
!     
!         iter = iter+1
!     
!         if (iter>maxiter) then
!             write(*,*) 'Max iterations exceeded'
!             exit
!         end if
!     
!     end do
!     
!     !Return the best found.  
!     x_out = bestx
!     !If fx /= bestf, want to call the routine with the best x in case the function has side effects
!     if (fx/=bestf) fx = f(bestx)
!     
!     if (printmod/=-1) write(*,"('|~~~~~~~~~~~~~~End Bisection Algorithm~~~~~~~~~~~~~~~~~~~~~~~~|')")
! 
! end subroutine sub_bisect

! Uses secant method with the bisection method as a backup
! subroutine sub_secant(x,f,a_in,b_in,toler_in,maxiter_in,fa_in,fb_in,xtoler_in,printmod_in)
!     use mod_matlab
!     implicit none   
!     interface    
!         function f(x)
!             real(rt), intent(in) :: x
!             real(rt) :: f
!         end function f
!     end interface
!     real(rt), intent(inout) :: x
!     real(rt), intent(in), optional :: a_in,b_in
!     real(rt), intent(in), optional :: toler_in,fa_in,fb_in,xtoler_in
!     integer, intent(in), optional :: maxiter_in, printmod_in
!     !local
!     real(rt) :: e, fx, fxpe, dfx, change_x, change_x_old, s, x_old, fx_old
!     real(rt), dimension(1:1000) :: xhist,fhist
!     real(rt) :: toler, xtoler
!     integer :: maxiter, iter, flag, feval, hist_iter, loc(1), printmod
!     
! 
!     !default values
!     printmod = 1
!     xtoler = 1.d-12
!     toler = 1.d-6
!     maxiter = 200
!     if (present(printmod_in)) printmod = printmod_in
!     if (present(xtoler_in)) xtoler = xtoler_in
!     if (present(toler_in)) toler = toler_in
!     if (present(maxiter_in)) maxiter = maxiter_in
!     
!     flag = 1
!     feval = 1
! 
!     change_x_old = huge(0.e0_rt)
!     change_x = huge(0.e0_rt)
!     hist_iter = 0
!     if (printmod>=1) write(*,*) 'sub_secant______________________________________________'
!     do iter = 1,maxiter
! 
!         !Compute fx at guess
!         fx = f(x)
! 
!         !Tracking progress
!         feval = feval + 1
!         hist_iter = hist_iter+1
!         xhist(hist_iter) = x
!         fhist(hist_iter) = fx        
!         if (hist_iter>size(xhist)) STOP 'increase secant history length'
! 
!         if (printmod>=1) write(*,'(A,i4,A,e12.5,A,e12.5)') ' iter ',iter,'  x: ',x,'  fx: ',fx
! 
!         !Check for goodness of fit
!         if (abs(fx)<toler) then
!             flag = 0
!             exit
!         end if
! 
!         !Check if have cycling
!         if (abs(change_x)>abs(change_x_old)) then
!             if ((change_x<0.e0_rt .AND. change_x_old<0.e0_rt) .or. (change_x>0.e0_rt .AND. change_x_old>0.d0)) then
!                 if (printmod>=1) write(*,*) 'It appears you are accelerating away from a root.  Switching to sub_bisect.'
!             else
!                 if (printmod>=1) write(*,*) 'It appears you are cycling around a root.  Switching to sub_bisect.'
!             end if
!             call bisect()
!             hist_iter = hist_iter+1
!             xhist(hist_iter) = x
!             fhist(hist_iter) = 0.e0_rt !This esnures x is selected below
!             exit
!         end if
!         change_x_old = change_x
! 
!         !Approximate derivative if first step
!         if (iter==1) then
!             e = .0001e0_rt !Step size        
!             fxpe = f(x+e)
!             dfx = (fxpe-fx)/e        
!             s = -fx/dfx
!             !Tracking progress
!             feval = feval + 1
!             hist_iter = hist_iter+1
!             xhist(hist_iter) = x+e
!             fhist(hist_iter) = fxpe        
!             if (hist_iter==size(xhist)) STOP 'increase secant history length'
!         else
!             s = -fx*(x-x_old)/(fx-fx_old)
!         end if
! 
!         !Set x' = x + s
!         x_old = x
!         fx_old = fx
!         change_x = s
!         x = x_old + s
! 
!     end do
! 
!     loc = minloc(abs(fhist(1:hist_iter)))
!     x = xhist(loc(1))
!     fx = fhist(loc(1))
! 
!     select case (flag)
!         case (0)
!             if (printmod>=1) write(*,*) 'Converged successfully, fx: ', fx
!         case default
!             if (printmod>=1) write(*,*) 'Failed to converge, fx: ', fx
!      end select
!     if (printmod>=1) write(*,*) '________________________________________________________'
! 
!     if (printmod>=1) call plot(xhist(1:hist_iter),fhist(1:hist_iter))
! 
! contains
!     !Locate bounds a,b if they exist and call sub_bisect
!     subroutine bisect()
!         implicit none
!         real(rt) :: a,b,fa,fb
!         logical :: call_bisect
! 
!         !Get bounds from xhist if possible
!         if (any(fhist(1:hist_iter)<0.e0_rt) .and. any(fhist(1:hist_iter)>0.e0_rt)) then
!             call_bisect = .TRUE.
! 
!             loc = maxloc(fhist(1:hist_iter),fhist(1:hist_iter)<0.e0_rt)
!             a = xhist(loc(1))
!             fa = fhist(loc(1))
!             loc = minloc(fhist(1:hist_iter),fhist(1:hist_iter)>0.e0_rt)
!             b = xhist(loc(1))
!             fb = fhist(loc(1))
! 
!         !If that failed, use bounds if present
!         elseif (present(a_in) .AND. present(b_in)) then
!             call_bisect = .TRUE.
! 
!             a = a_in
!             b = b_in
!             if (present(fa_in) .AND. present(fb_in)) then
!                 fa = fa_in
!                 fb = fb_in
!             else 
!                 fa = f(a)
!                 fb = f(b)
!             end if
!         
!         !If neither present alert the programmer and do nothing
!         else
!         
!             write(*,*) 'sub_secant_1dim: WARNING: bracket not provided and not found endogenously.  Trying large bounds.'
!             call_bisect = .TRUE.
!             a = -1d10 
!             b = 1d10
! 
!         end if
! 
!         if (call_bisect) then
!             call sub_bisect(x,f,a,b,toler,maxiter,fa,fb,xtoler,printmod)
!         end if
! 
!     end subroutine bisect
! 
! end subroutine sub_secant


end module mod_root1dim

! Contains test routines for mod_root1dim
module mod_root1dim_test
    use mod_root1dim
    implicit none
contains
    
    ! One dimensional test functions
    real(rt) function f0(x); real(rt), intent(in) :: x; f0 = sin(x) + cos(x); end function 
    real(rt) function f1(x); real(rt), intent(in) :: x; f1 = sqrt(abs(x-5e0_rt)) - 1e0_rt; end function
    real(rt) function f2(x); real(rt), intent(in) :: x; f2 = -10e0_rt + 2e0_rt*x + x**2 ; end function
    real(rt) function f3(x); real(rt), intent(in) :: x; f3 = exp(max(min(x,5.0_rt),-5.0_rt)) - 2e0_rt ; end function
    real(rt) function f4(x); real(rt), intent(in) :: x; f4 = sin(x) ; end function 
    ! Main routine
    subroutine test_root1dim() 
        implicit none
        logical :: tf(5)
        real(rt) :: a,b
        real(rt) :: xstar(5)
        
        a = -10; b = 20
        tf = passed(f0)
        if (.not. all(tf)) then
            write(0,*) 'Test failed:',tf,'x vals',xstar
            STOP 'test_root1dim: f0 test failed'
        end if
        a = -6; b = 6
        tf = passed(f1)
        if (.not. all(tf)) then
            write(0,*) 'Test failed:',tf,'x vals',xstar
            STOP 'test_root1dim: f1 test failed'
        end if
        a = -6; b = 2
        tf = passed(f2)
        if (.not. all(tf)) then
            write(0,*) 'Test failed:',tf,'x vals',xstar
            STOP 'test_root1dim: f2 test failed' 
        end if
        a = -100; b = 100
        tf = passed(f3)
        if (.not. all(tf)) then
            write(0,*) 'Test failed:',tf,'x vals',xstar
            STOP 'test_root1dim: f3 test failed' 
        end if
        a = -100; b = 100
        tf = passed(f4)
        if (.not. all(tf)) then
            write(0,*) 'Test failed:',tf,'x vals',xstar
            STOP 'test_root1dim: f4 test failed' 
        end if

        print*, 'test_root1dim: passed'

    contains 
        function passed(f) result(tfout)
            implicit none
            logical :: tf(5), tfout(5)
            interface
                function f(x)
                    use mod_types
                    implicit none
                    real(rt), intent(in) :: x
                    real(rt) :: f
                end function f
            end interface
            
            tf = .true.
            
            call sub_bisect(   xstar(1),f,a,b,xtoler_in=2e0_rt*epsilon(0e0_rt),printmod_in=1) 
            call sub_brent(    xstar(2),f,a,b,xtoler_in=2e0_rt*epsilon(0e0_rt),printmod_in=1) 
            ! call sub_ridders(  xstar(3),f,a,b,xtoler_in=2e0_rt*epsilon(0e0_rt),printmod_in=1)
            call sub_nsect_omp(xstar(4),f,a,b,xtoler_in=2e0_rt*epsilon(0e0_rt),printmod_in=1)
            ! call sub_nsect_mpi(xstar(5),f,a,b,xtoler_in=2e0_rt*epsilon(0e0_rt),printmod_in=1)
            
            if (abs(f(xstar(1)))>1e-2_rt) tf(1) = .false.
            if (abs(f(xstar(2)))>1e-2_rt) tf(2) = .false.
            ! if (abs(f(xstar(3)))>1e-2_rt) tf(3) = .false.
            if (abs(f(xstar(4)))>1e-2_rt) tf(4) = .false.
            ! if (abs(f(xstar(5)))>1e-2_rt) tf(5) = .false.

!             print*,'xstar',xstar
!             print*,'f(xstar)',f(xstar(1)),f(xstar(2)),f(xstar(3))
!             print*,'tf',tf

            tfout = tf

        end function passed
        
    end subroutine test_root1dim
    
    subroutine test_findallroots()
        use mod_matlab, only: linspace,pi
        implicit none
        real(rt) :: roots(1000)
        integer :: nroots,iroots

        call sub_findallroots(roots,nroots,f4,linspace(-5e0_rt,5e0_rt,1000))

        ! since f4 is just the sin function, the roots are any multiples of pi.
        ! with a range of -5e0_rt,5e0_rt, there are three roots (-pi,0,pi)
    
        if (nroots/=3e0_rt) STOP 'test_findallroots: test failed'
        if (any(abs(roots(1:3)-[-pi,0e0_rt,pi])>1d-7)) STOP 'test_findallroots: test 2 failed'

        print*, 'test_findallroots: passed'

!         do iroots = 1,nroots
!             print*,'roots(iroots)',roots(iroots)
!             print*,'f4(roots(iroots))',f4(roots(iroots))
!         end do

    end subroutine

end module mod_root1dim_test
module mod_quad
    use mod_types
    implicit none
    private :: en_r2_11_1, en_r2_11_1_size
contains


! Integrates 
! \int^b_a f(x) dx 
! using Gauss-Legendre quadrature adapted from [-1,1] with weighting function 1
subroutine legendrequad(x,w,a,b)
    use mod_matlab
    implicit none
    real(rt), intent(out) :: x(:),w(:)
    real(rt), intent(in) :: a,b
    real(rt) :: tmp
    integer :: i,order,uid
    logical :: found
    
    found = .false.

    open(newunit=uid,file='/tmp/legendrequadnodes.dat',status='old',action='read',err=999)
    inquire(uid,recl=i)
    if (i==size(x)) then
        read(uid,*) x
        found = .true.
    else
        found = .false.
    end if
    close(uid)

    ! If still here, check  they are increasing, and they are distinct
    if (any(x(1:size(x)-1)>=x(2:size(x)))) found = .false.

    ! Check that they are actually roots
    if (found) then
        !$OMP PARALLEL DO DEFAULT(SHARED) PRIVATE(i,tmp) SCHEDULE(STATIC,50) 
        do i = 1,order
            call legendre_eval(x(i),order,P=tmp)
            if (abs(tmp)>sqrt(epsilon(0._rt))) found = .false.
        end do
        !$OMP END PARALLEL DO 
    end if

    ! Get roots of legendre polynomials if couldn't read from file
999 if (.not. found) call legendre_roots(x)

    order = size(x)

    ! Get the weights
    !$OMP PARALLEL DO DEFAULT(SHARED) PRIVATE(i) SCHEDULE(STATIC,50) 
    do i = 1,order
        call legendre_eval(x(i),order,dP=w(i))
        ! Now w(i) stores P'_{n+1}. Use equation 12 here http://mathworld.wolfram.com/Legendre-GaussQuadrature.html (7/20/16)
        w(i) = 2d0/((1d0-x(i)**2)*w(i)**2)
    end do
    !$OMP END PARALLEL DO 

    ! Try to save to file 
    if (.not. found) then
        open(newunit=uid,file='/tmp/legendrequadnodes.dat',status='replace',action='write',err=998)
        write(uid,*,err=998) x
        close(uid)
    end if

    ! Now adapt to [a,b] via a linear change of variables
998 x = a+.5d0*(b-a)*(x+1d0)
    w = .5d0*(b-a)*w

end subroutine

! Computes the n roots of H_n, the Legendre polynomials
! P_0 = 1
! P_1 = x
!
! with the recurrence relations http://mathworld.wolfram.com/LegendrePolynomial.html (retrieved 7/20/16)
! P_n = (2*(n-1) + 1) * P_{n-1} - (n-1) * P_{n-2})/n
! P'_n = (-n * x * P_n + n * P_{n-1}) / (1d0-x**2)
! 
subroutine legendre_roots(x)
    implicit none
    real(rt), intent(inout) :: x(:)
    ! local
    integer :: order,i,n
    real(rt), allocatable :: xlast(:)

    n = size(x)

    allocate(xlast(n))

    do order = 1,n

        if (order==1) then
            x(1:order) = 0d0 
        else

            !$OMP PARALLEL DO DEFAULT(SHARED) PRIVATE(i) SCHEDULE(STATIC,20)
            do i = 1,order/2
                ! Look for a hermite root between the two previous ones
                if (i==1) then
                    call legendre_bracket(x(i),order,-1d0,xlast(1))
                elseif (i==order) then
                    call legendre_bracket(x(i),order,xlast(i-1),1d0)
                else
                    call legendre_bracket(x(i),order,xlast(i-1),xlast(i))
                end if
            end do
            !$OMP END PARALLEL DO

            if (mod(order,2)==1) then
                ! Odd number
                x((order+1)/2) = 0d0
                do i = (order+1)/2+1,order
                    x(i) = -x(order-i+1)
                end do
            else
                ! Even number
                do i = order/2+1,order
                    x(i) = -x(order-i+1)
                end do
            end if

        end if

        xlast = x

    end do

!     print*,'x',x

end subroutine

! Uses bracketing method to find root of P_n in a,b
subroutine legendre_bracket(x,n,a,b)
    use mod_root1dim
    implicit none
    real(rt), intent(inout) :: x
    real(rt), intent(in) :: a,b
    integer, intent(in) :: n
    ! local
    real(rt) :: s,tmp,P,dP,tmpP
    integer :: iter

    if (n==0) then
        write(0,*) 'WARNING: legendre_bracket: there is no root for P_0'
        return
    end if

    if (n==1) then
        x = 0d0
    else

        ! Try Newton
        x = .5d0*(a+b)
        do iter = 1,100
            call legendre_eval(x,n,P,dP)
            s = P/dP
            x = x - s
            if (abs(s)<1e2_rt*epsilon(0._rt)) exit
        end do

        if (abs(s)<1e2_rt*epsilon(0._rt) .and. x>a .and. x<b) then
!             print*,'Newton succeeded!'
            return
        end if

!         print*,'Newton failed :('

        ! If still here, Newton didn't work, so use Brent
        call sub_brent(x,res,a,b,printmod_in=-1)

        ! Try to polish off with Newton
        tmp = x
        call legendre_eval(x,n,tmpP)

        do iter = 1,10
            call legendre_eval(x,n,P,dP)
            s = P/dP
            x = x - s
        end do

        call legendre_eval(x,n,P)

        if (x<a .or. x>b .or. abs(tmpP)<abs(P)) then
            ! Newton failed to improve, so revert
            x = tmp
        end if

    end if

contains 

    real(rt) function res(x)
        implicit none
        real(rt), intent(in) :: x
        real(rt) :: P

        call legendre_eval(x,n,P)
        res = P
    end function 

end subroutine 

! Gives the P_n and P'_n at a given x
subroutine legendre_eval(x,n,P,dP)
    implicit none
    real(rt), intent(in) :: x
    integer, intent(in) :: n
    real(rt), intent(out), optional :: P,dP
    ! local
    real(8) :: Hzero,Hone,Htwo
    integer :: i

    ! H_0 = 1
    ! H_1 = x

    if (n==0) then
        if (present(P)) P = 1d0
        if (present(dP)) dP = 0d0
        return
    elseif (n==1) then
        if (present(P)) P = x
        if (present(dP)) dP = 1d0
        return
    end if

    Htwo = 1d0
    Hone = x
    do i = 2,n

        ! P_n = (2*(n-1) + 1) * x * P_{n-1} - (n-1) * P_{n-2})/n
        ! P_n = (2*n- 1) * x * P_{n-1} - (n-1) * P_{n-2})/n
        Hzero = ( (2*i-1)*x*Hone - (i-1)*Htwo)/i

        if (i<n) then
            Htwo = Hone
            Hone = Hzero
        end if
    end do

    ! Output the results
    if (present(P)) P = Hzero

    ! P'_n = (-n * x * P_n + n * P_{n-1}) / (1d0-x**2)
    if (present(dP)) then
        if (abs(x)/=1d0) then
            dP = (-n * x * Hzero + n*Hone )/(1._rt-x**2)
        else
            dP = (-n * x * Hzero + n*Hone )*1e10_rt
        end if
    end if


end subroutine 

! Computes the n roots of H_n, the hermite polynomials
! H_0 = 1
! H_1 = 2*x
! H_2 = 4*x**2 - 2
!
! with the recurrence relations http://mathworld.wolfram.com/HermitePolynomial.html (retrieved 7/20/16)
! H_n = 2 * x * H_{n-1} - 2 * (n-1) * H_{n-2}
! H'_n = 2 * n * H_{n-1}
! 
! 
! For Newton's method, need x' = x - H_n/H'_n. This is 
! x' = x - (x/n - (n-1)/n * H_{n-2}/H_{n-1})
! So, only need Hermite values rather than the derivatives
subroutine hermite_roots(x)
    implicit none
    real(rt), intent(inout) :: x(:)
    ! local
    integer :: order,i,n
    real(rt), allocatable :: xlast(:)

    n = size(x)

    allocate(xlast(n))

    do order = 1,n

        if (order==1) then
            x(1:order) = 0d0 
        elseif (order==2) then
            x(1:order) = .5d0*[-sqrt(2d0),sqrt(2d0)]
        elseif (order==3) then
            x(1:order) = .5d0*[-sqrt(6d0),0d0,sqrt(6d0)]
        else
            do i = 1,order
                ! Look for a hermite root between the two previous ones
                if (i==1) then
                    ! x(i) = (1d0+1d-3)*xlast(1) ! will be negative
                    ! call hermite_newton(x(i),order)
                    call hermite_bracket(x(i),order,-1d3,xlast(1))
                elseif (i==order) then
                    ! x(i) = (1d0+1d-3)*xlast(i-1)
                    ! call hermite_newton(x(i),order)
                    call hermite_bracket(x(i),order,xlast(i-1),1d3)
                else
                    ! x(i) = .5d0*(xlast(i-1)+xlast(i))
                    ! call hermite_newton(x(i),order)
                    call hermite_bracket(x(i),order,xlast(i-1),xlast(i))
                end if
            end do

        end if

        xlast = x

    end do

!     print*,'x',x

end subroutine

! Uses bracketing method to find root of H_n in a,b
subroutine hermite_bracket(x,n,a,b)
    use mod_root1dim
    implicit none
    real(rt), intent(inout) :: x
    real(rt), intent(in) :: a,b
    integer, intent(in) :: n
    ! local
    real(rt) :: s,H,Hp
    integer :: iter

    if (n==0) then
        write(0,*) 'WARNING: hermite_newton: there is no root for H_0'
        return
    end if

    if (n==1) then
        x = 0d0
    else
        call sub_brent(x,res,a,b,printmod_in=-1)
    end if

contains 

    real(rt) function res(x)
        implicit none
        real(rt), intent(in) :: x
        real(rt) :: H

        call hermite_eval(x,n,H)
        res = H
    end function 

end subroutine 


! Uses Newton's method to find a root of the nth polynomial starting from x
subroutine hermite_newton(x,n)
    implicit none
    real(rt), intent(inout) :: x
    integer, intent(in) :: n
    ! local
    real(rt) :: s,H,Hp
    integer :: iter

    if (n==0) then
        write(0,*) 'WARNING: hermite_newton: there is no root for H_0'
        return
    end if

    if (n==1) then
        x = 0d0
    else
        do iter = 1,1000

            call hermite_eval(x,n,H,Hp)
            print*,'n',n,'x,H,Hp',x,H,Hp
            s = H/Hp
            if (abs(s)<1d-10) return
            x = x - s
        end do
        write(0,*) 'WARNING: hermite_newton: failed to find root for H_n, step size is',s
    end if

end subroutine 

! Gives the H_n and H'_n at a given x
subroutine hermite_eval(x,n,H,Hp)
    implicit none
    real(rt), intent(in) :: x
    integer, intent(in) :: n
    real(rt), intent(out), optional :: H,Hp
    ! local
    real(8) :: Hzero,Hone,Htwo
    integer :: i

    ! H_0 = 1
    ! H_1 = 2*x
    ! H_2 = 4*x**2 - 2

    if (n==0) then
        if (present(H)) H = 1d0
        if (present(Hp)) Hp = 0d0
        return
    elseif (n==1) then
        if (present(H)) H = 2d0*x
        if (present(Hp)) Hp = 2d0
        return
    elseif (n==2) then
        if (present(H)) H = 4d0*x**2 - 2d0
        if (present(Hp)) Hp = 8d0*x
        return
    end if

    Htwo = 2d0*x
    Hone = 4d0*x**2 - 2d0
    do i = 3,n
        ! H_n = 2 * x * H_{n-1} - 2 * (n-1) * H_{n-2}
        Hzero = 2d0 * x * Hone - 2d0 * (i-1) * Htwo

        if (i<n) then
            Htwo = Hone
            Hone = Hzero
        end if
    end do

    ! Output the results
    if (present(H)) H = Hzero
    if (present(Hp)) Hp = 2d0 * n * Hone


end subroutine 

! Integrates 
! \int^b_a f(x) dx 
! using Chebyshev quadrature adapted from [-1,1] with weighting function 1/sqrt(1-x**2).
subroutine chebyquad(x,w,a,b)
    use mod_matlab
    implicit none
    real(rt), intent(out) :: x(:),w(:)
    real(rt), intent(in) :: a,b
    ! local
    real(rt), allocatable :: xsort(:)
    integer, allocatable :: Isort(:)
    real(rt) :: tmp
    integer :: n,i

    n = size(x)

    if (size(w)/=n) STOP 'ERROR: chebyquad: x,w don''t conform'

    do i = 1,n
        tmp = (2e0_rt*real(i,rt)-1e0_rt)/2e0_rt/real(n,rt)*pi
        tmp = cos(tmp)
        x(i) = a + (b-a)*(tmp+1e0_rt)/2e0_rt
        w(i) = (b-a)/2e0_rt*pi/real(n,rt)*sqrt(1e0_rt-tmp**2)
    end do

    ! Sort to make sure they are ascending
    allocate(xsort(n),Isort(n))
    call sort2(xsort,Isort,x)
    x = xsort
    w = w(Isort)

!     if (abs(sum(w)-(b-a))>1d-6*abs(b-a)) then
!         print*,'w',w
!         print*,'x',x
!         stop 'checking chebyquad'
!         print*,'b',b
!         print*,'a',a
!         print*,'b-a',b-a
!         print*,'sum(w)',sum(w)
!         STOP 'ERROR: chebyquad: weights do not sum to b-a'
!     end if


end subroutine

! gauss-legendre here has
!  \int^1_{-1} f(x) dx 
! but I use a change of variables to handle
!  \int^b_a f(x) dx
! n-point rules exactly integrate a 2n-1 degree polynomial wrt a weighting function
! n is inferred from the size of x and w
subroutine legendre(x,w,a,b)
    implicit none
    real(rt), intent(out) :: x(:),w(:)
    real(rt), intent(in) :: a,b
    ! local

    integer :: n

    n = size(x)

    if (size(w)/=n) STOP 'ERROR: x,w don''t conform'

    ! Get the nodes and weights for [-1,1] 
    select case (n)
        case (1)
            x = 0e0_rt
            w = 2e0_rt
        case (2)
            x = [-sqrt(1e0_rt/3e0_rt),sqrt(1e0_rt/3e0_rt)]
            w = 1e0_rt
        case (3)
            x(1) = -sqrt(3e0_rt/5e0_rt)
            x(2) = 0e0_rt
            x(3) = sqrt(3e0_rt/5e0_rt)
            w(1) = 5e0_rt/9e0_rt
            w(2) = 8e0_rt/9e0_rt
            w(3) = 5e0_rt/9e0_rt
        case (5) 

            x(1) = -0.90618e0_rt
            x(2) = -0.538469e0_rt
            x(3) = 0e0_rt 
            x(4) = -x(2)
            x(5) = -x(1)

            w(1) = 0.236927e0_rt
            w(2) = 0.478629e0_rt
            w(3) = 0.568889e0_rt
            w(4) = w(2)
            w(5) = w(1)

        case (11)

            ! http://keisan.casio.com/exec/system/1329114617
            x(1)  = -0.978228658146056992804e0_rt;  w(1)  = 0.0556685671161736664828e0_rt
            x(2)  = -0.8870625997680952990752e0_rt; w(2)  = 0.1255803694649046246347e0_rt
            x(3)  = -0.730152005574049324093e0_rt;  w(3)  = 0.1862902109277342514261e0_rt
            x(4)  = -0.519096129206811815926e0_rt;  w(4)  = 0.2331937645919904799185e0_rt
            x(5)  = -0.2695431559523449723315e0_rt; w(5)  = 0.262804544510246662181e0_rt
            x(6)  = 0e0_rt                        ; w(6)  = 0.2729250867779006307145e0_rt
            x(7)  = 0.2695431559523449723315e0_rt;  w(7)  = 0.262804544510246662181e0_rt
            x(8)  = 0.5190961292068118159257e0_rt;  w(8)  = 0.233193764591990479919e0_rt
            x(9)  = 0.7301520055740493240934e0_rt;  w(9)  = 0.186290210927734251425e0_rt
            x(10) = 0.887062599768095299075e0_rt;   w(10) = 0.125580369464904624635e0_rt
            x(11) = 0.9782286581460569928039e0_rt;  w(11) = 0.055668567116173666483e0_rt

        case (21)

            x(1 ) = -0.99375217062038950026e0_rt; w(1 ) = 0.016017228257774333324e0_rt
            x(2 ) = -0.967226838566306294317e0_rt; w(2 ) = 0.0369537897708524938e0_rt
            x(3 ) = -0.92009933415040082879e0_rt; w(3 ) = 0.0571344254268572082836e0_rt
            x(4 ) = -0.853363364583317283647e0_rt; w(4 ) = 0.076100113628379302017e0_rt
            x(5 ) = -0.768439963475677908616e0_rt; w(5 ) = 0.0934444234560338615533e0_rt
            x(6 ) = -0.667138804197412319306e0_rt; w(6 ) = 0.108797299167148377663e0_rt
            x(7 ) = -0.551618835887219807059e0_rt; w(7 ) = 0.1218314160537285341954e0_rt
            x(8 ) = -0.4243421202074387835737e0_rt; w(8 ) = 0.1322689386333374617811e0_rt
            x(9 ) = -0.288021316802401096601e0_rt; w(9 ) = 0.1398873947910731547221e0_rt
            x(10) = -0.145561854160895090937e0_rt; w(10) = 0.1445244039899700590638e0_rt
            x(11) = 0e0_rt;                        w(11) = 0.146081133649690427192e0_rt
            x(12) = 0.145561854160895090937e0_rt; w(12) = 0.144524403989970059064e0_rt
            x(13) = 0.2880213168024010966008e0_rt; w(13) = 0.139887394791073154722e0_rt
            x(14) = 0.4243421202074387835737e0_rt; w(14) = 0.132268938633337461781e0_rt
            x(15) = 0.551618835887219807059e0_rt; w(15) = 0.1218314160537285341956e0_rt
            x(16) = 0.667138804197412319306e0_rt; w(16) = 0.108797299167148377663e0_rt
            x(17) = 0.768439963475677908616e0_rt; w(17) = 0.0934444234560338615534e0_rt
            x(18) = 0.853363364583317283647e0_rt; w(18) = 0.0761001136283793020169e0_rt
            x(19) = 0.92009933415040082879e0_rt; w(19) = 0.057134425426857208284e0_rt
            x(20) = 0.967226838566306294317e0_rt; w(20) = 0.0369537897708524938e0_rt
            x(21) = 0.99375217062038950026e0_rt; w(21) = 0.016017228257774333324e0_rt

        case default
            STOP 'ERROR: legendre: formula for requested n not programmed yet'
    end select

    ! Adapt to [a,b]
    w = (b-a)/2e0_rt*w
    x = (b-a)/2e0_rt*x + (a+b)/2e0_rt

    ! Note: the weights should definitely sum to b-a because even with 1 quadrature point, \int 1 di can be integrated exactly.
    if (abs(sum(w)-(b-a))>4.0_rt*epsilon(0.0_rt)*abs(b-a)) then
        print*,'w',w
        print*,'x',x
        print*,'b',b
        print*,'a',a
        print*,'b-a',b-a
        print*,'sum(w)',sum(w)
        STOP 'ERROR: legendre: weights do not sum to b-a'
    end if
end subroutine


! Trapezoidal rule for a predetermined x grid
! This specifies the quadrature weights for the trapezoidal rule
! at the nodes x. Note: if y = f(x) is piecewise linear, then
! this integrates the function exactly.
function trapzquad_fixedx(x) result(w)
    implicit none
    real(rt), intent(in) :: x(1:)
    real(rt) :: w(size(x))
    ! local
    integer :: n,i

    n = size(x)

    w(1) = .5e0_rt*h(1)
    w(n) = .5e0_rt*h(n-1)
    do i = 2,n-1
        w(i) = .5e0_rt*(h(i-1)+h(i))
    end do

contains
    real(rt) function h(i)
        integer, intent(in) :: i
        h = x(i+1)-x(i)
    end function
end function 

! Simpson's rule 
! NOTE: n must be odd
subroutine sub_simpsonquad(x,w,a,b)
    use mod_matlab
    implicit none
    real(rt), intent(out) :: x(:),w(:)
    real(rt), intent(in) :: a,b
    ! local
    real(rt) :: h
    integer :: i,n

    n = size(x)

    if (n/=size(w)) STOP 'ERROR: simpson: size of x and w do not agree'
    if (mod(n,2)==0) STOP 'ERROR: simpson: n must be odd'
    if (n<3) STOP 'ERROR: simpson: n must be >= 3'

    x = linspace(a,b,n)

    ! Step size from x(i) to x(i+1)
    h = (b-a)/real(n-1,rt) 

    ! End point weights
    w(1) = 1e0_rt
    w(n) = 1e0_rt

    ! Even interior values
    do i = 2,n-1,2
        w(i) = 4e0_rt
    end do

    ! Odd interior values
    do i = 3,n-2,2
        w(i) = 2e0_rt
    end do

    ! Scale by h/3
    w = w*h/3e0_rt

end subroutine sub_simpsonquad

! Integrate the piecewise linear function f(x) over domain [a,b] wrt the Lebesgue measure
! That is, \int^b_a f(x) dx. It is assumed the knots x are strictly increasing. 
! If a or b is outside the domain of x, linear extrapolation is used
function intPiecewiseLinear(xin,yin,a,b) result(qout)
    use mod_matlab
    implicit none
    real(rt), intent(in) :: a,b,xin(1:),yin(1:)
    real(rt) :: qout
    ! local
    real(rt), allocatable :: x(:),y(:)
    real(rt) :: q,m,xlo,xhi
    integer :: nx,ialo,ibhi,i

    nx = size(xin) 
    if (nx/=size(yin)) STOP 'x and y must have the same size'
    if (nx<2) STOP 'x must have at least length 2'
    if (b<a) STOP 'b must be >= a'

    allocate(x(nx),y(nx))

    x = xin
    y = yin

    ! Since linearly extrapolating, can just extend the range of x to a and b if a<x(1) or b>x(nx)
    if (a<x(1)) then
        ! Replace (x(1),y(1)) with (a,f(a))
        m = (y(2)-y(1))/(x(2)-x(1))       
        y(1) = y(1) + m*(a-x(1))
        x(1) = a
    end if
    if (b>x(nx)) then
        ! Replace (x(nx),y(nx)) with (b,f(b))
        m = (y(nx)-y(nx-1))/(x(nx)-x(nx-1))       
        y(nx) = y(nx-1) + m*(b-x(nx-1))
        x(nx) = b
    end if
    
    ! Now, everything is interior with x(1)<=a<=b<=x(n)
    call bsearch(ialo,a,x)
    call bsearch(ibhi,b,x)
    if (b>x(ibhi)) ibhi = ibhi + 1 ! Note: bsearch always returns lower bracket. We want higher.

    !!! Check that conditions hold
    if (a<x(1)) STOP 'ERROR: intPiecewiseLinear: should have x(1)<=a'
    if (b<a) STOP 'ERROR: intPiecewiseLinear: should have a<=b'
    if (x(nx)<b) STOP 'ERROR: intPiecewiseLinear: should have b<=x(nx)'
    if (.not. (a>=x(ialo) .and. a<=x(ialo+1))) STOP 'ERROR: intPiecewiseLinear: error in getting ialo'
    if (.not. (b>=x(ibhi-1) .and. b<=x(ibhi))) STOP 'ERROR: intPiecewiseLinear: error in getting ibhi'

    ! Integrate looping through the intervals
    q = 0e0_rt
    do i = ialo,ibhi-1
        ! In range [max(a,x(i)),min(b,x(i+1))], integrate
        xlo = max(a,x(i))
        xhi = min(b,x(i+1))

        m = (y(i+1)-y(i))/(x(i+1)-x(i))

        q = q + y(i)*(xhi-xlo) + m*(xhi-x(i))**2/2e0_rt - m*(xlo-x(i))**2/2e0_rt

    end do

    qout = q

end function

! Provides quadrature nodes (x) and weights (w) that can be used to 
! integrate a function f of dim d with respect to a multivariate normal
! distribution N(mean,varcov) using monomial rules taken from Stroud.
! There are several methods available
! 
! Method name      Description
! 'deg3B'          A degree 3 rule using 2*d spread out equally weighted points 
! ! 'deg5'           A degree 5 rule using 2d^2+1 pts at various distances
subroutine myquad(x,w,d,mean,varcov,method)
    use mod_matlab, only: chol,det,pi,strcmp
    implicit none
    real(rt), allocatable, intent(out) :: x(:,:),w(:)
    integer, intent(in) :: d
    real(rt), intent(in) :: mean(d),varcov(d,d)
    character(len=*), intent(in) :: method
    !local
    real(rt) :: Omega(d,d), absdetOmega

    ! Construct the cholesky decomposition of varcov
    Omega = transpose(chol(varcov))
    absdetOmega = abs(det(Omega))

    ! Get the weights for standard form
    if (strcmp(method,'deg3B')) then
        call int_deg3ruleB(x,w,d)
!     elseif (strcmp(method,'deg5')) then
!         call int_deg5rule(x,w,d)
    else
        STOP 'myquad: Method not found.  Remember it''s case sensitive.'
    end if

    ! Convert to non-standard form
    x = sqrt(2.e0_rt)*x*transpose(Omega) + transpose(spread(mean,2,size(x,1)))
    w = w*(2.e0_rt**(real(d,rt)/2.e0_rt))*absdetOmega*1.e0_rt/sqrt(abs(det(varcov)))*(2.e0_rt*pi)**(-real(d,rt)/2.e0_rt)

end subroutine myquad

subroutine int_deg3ruleB(x,w,d)
    use mod_matlab, only: pi, diag
    implicit none
    real(rt), intent(out), allocatable :: x(:,:),w(:)
    integer, intent(in) :: d
    !local
    real(rt) :: r,V

    allocate(x(2*d,d),w(2*d))

    r = sqrt(real(d,rt)/2.e0_rt);    
    x(1:d,:) = diag(spread(r,1,d))
    x(d+1:2*d,:) = diag(spread(-r,1,d))

    V = pi**(real(d,rt)/2.e0_rt)
!     E = V/(2*d)*sum(f);
    w = V/(2.e0_rt*real(d,rt))
end subroutine int_deg3ruleB


subroutine test_quad_deg5rule
    use mod_matlab, only: randn, mean
    implicit none
    integer, parameter :: dint = 3
    real(rt) :: x(2*dint**2+1,dint) !
    real(rt) :: w(2*dint**2+1) !
    
    call sub_quad_deg5rule(x,w,dint)

    ! integrate sum(2*(.9*x-3)**5+1,2) using formula
    write(*,*) 'E according to quad rule', dot_product(sum(2.e0_rt*(.9e0_rt*x-3.e0_rt)**5+1.e0_rt,dim=2),w)

    write(*,*) 'E according to MC       ', mean(sum(2.e0_rt*(.9e0_rt*randn(100000,dint)-3.e0_rt)**5+1.e0_rt,dim=2))

end subroutine test_quad_deg5rule

! Integrate \int_{R^d} f(x) (2\pi)^{-d/2} e^{-(1/2) \sum^d_{i=1} x^2_i dx}
! Which is the integral of f wrt a multivariate unit normal distribution. 
!
! Based on integrating \int_{R^d} f(x) e^{-\sum^d_{i=1} x^2_i dx} but uses
! a change of variables to get it into a more useful form.
! Uses a degree 5 rule as in Judd p.275.
! 
! x are the quadrature points  2*d**2+1 by d
! w are the quadrature weights  2*d**2+1
! d is the dimension
!
subroutine sub_quad_deg5rule(x,w,dint)
    use mod_matlab, only: pi, diag
    implicit none
    integer, intent(in) :: dint
    real(rt), intent(inout) :: x(1:,1:) !(2*dint**2+1,dint)
    real(rt), intent(inout) :: w(1:) !(2*dint**2+1)
    !local
    integer :: N, ind, i, j
    real(rt) :: d, V, A, B, C !, E
    real(rt) :: r,s, tmpPP(dint),tmpPM(dint),tmpMP(dint),tmpMM(dint)

    ! Check for correct sizes
    if (size(x,1)/=2*dint**2+1) STOP 'sub_quad_deg5rule: x has wrong size1'
    if (size(x,2)/=dint) STOP 'sub_quad_deg5rule: x has wrong size2'
    if (size(w)/=2*dint**2+1) STOP 'sub_quad_deg5rule: w has wrong size'

    ! Get quadrature points
    N = 2*dint**2+1 !Number of points used in formula
    d = real(dint,rt)
    r = sqrt(1.e0_rt+.5e0_rt*d)
    s = sqrt(.5e0_rt+.25e0_rt*d)
    
    !x = 0            
    x(1,:) = 0.e0_rt
    ! x= r*ei
    x(2:2+dint-1,:) = diag(spread(r,1,dint))
    ! x= -r*ei
    x(2+dint:2+dint+dint-1,:) = diag(spread(-r,1,dint))
    ! x = +-s*ei +-s*ej
    ind = 2+dint+dint
    do i = 1,dint-1
        do j = i+1,dint
            tmpPP = 0.e0_rt
            tmpPP(i) = s
            tmpPP(j) = s
            tmpPM = 0.e0_rt
            tmpPM(i) = s
            tmpPM(j) = -s
            tmpMP = 0.e0_rt
            tmpMP(i) = -s
            tmpMP(j) = s
            tmpMM = 0.e0_rt
            tmpMM(i) = -s
            tmpMM(j) = -s

            x(ind,:) = tmpPP
            x(ind+1,:) = tmpPM
            x(ind+2,:) = tmpMP
            x(ind+3,:) = tmpMM
            ind = ind+4
        end do
    end do

    ! Change of variables to integrate f wrt unit normal
    x = sqrt(2.e0_rt)*x

    ! Get quadrature weights
    N = 2*dint**2+1 !Number of points used in formula
    d = real(dint,rt)

    V = pi**(d/2.e0_rt)
    A = 2.e0_rt/(d + 2.e0_rt)*V
    B = (4.e0_rt-d)/(2.e0_rt*(d+2.e0_rt)**2)*V
    C = V/(d+2.e0_rt)**2

    w(1) = A
    w(2:2+dint-1) = B
    w(2+dint:2+dint+dint-1) = B
    w(2+dint+dint:2*dint**2+1) = C

    ! Use change of variables formula
    w = 1.e0_rt/(sqrt(pi)**dint)*w  

end subroutine sub_quad_deg5rule

!Integrate \int_{R^d} f(x) e^{-\sum^d_{i=1} x^2_i dx}
!using a degree 5 rule as in Judd p.275.
!Proceed in two steps.  In the first step, you passed values of x for which
!you must evaluate f.  Then passing the values back, the integration is calculated.
!
function int_deg5rule_step1(dint) result(x)
    use mod_matlab
    implicit none
    integer, intent(in) :: dint
    real(rt) :: x(2*dint**2+1,dint)
    !local
    integer :: N, ind, i, j
    real(rt) :: d,r,s, tmpPP(dint),tmpPM(dint),tmpMP(dint),tmpMM(dint)
    !Values of x that need to be evaluated    
    !Variable definitions using monomial approach due to Stroud (1971) and found in Judd
    !integer, parameter :: dint = 4 !underscore J is for Judd

    N = 2*dint**2+1 !Number of points used in formula
    d = real(dint,rt)
    r = sqrt(1.e0_rt+.5e0_rt*d)
    s = sqrt(.5e0_rt+.25e0_rt*d)
    
    !x = 0            
    x(1,:) = 0.e0_rt
    ! x= r*ei
    x(2:2+dint-1,:) = diag(spread(r,1,dint))
    ! x= -r*ei
    x(2+dint:2+dint+dint-1,:) = diag(spread(-r,1,dint))

    ! x = +-s*ei +-s*ej
    ind = 2+dint+dint
    do i = 1,dint-1
        do j = i+1,dint
            tmpPP = 0.e0_rt
            tmpPP(i) = s
            tmpPP(j) = s
            tmpPM = 0.e0_rt
            tmpPM(i) = s
            tmpPM(j) = -s
            tmpMP = 0.e0_rt
            tmpMP(i) = -s
            tmpMP(j) = s
            tmpMM = 0.e0_rt
            tmpMM(i) = -s
            tmpMM(j) = -s

            x(ind,:) = tmpPP
            x(ind+1,:) = tmpPM
            x(ind+2,:) = tmpMP
            x(ind+3,:) = tmpMM
            ind = ind+4
        end do
    end do

end function

! See help for int_deg5rule_step1
function int_deg5rule_step2(dint,fx) result(E)
    use mod_matlab
    implicit none
    integer, intent(in) :: dint
    real(rt), dimension(1:), intent(in) :: fx
    real(rt) :: E
    !local
    integer :: N
    real(rt) :: d, V, A, B, C

    N = 2*dint**2+1 !Number of points used in formula
    d = real(dint,rt)

    V = pi**(d/2e0_rt)
    A = 2e0_rt/(d + 2e0_rt)*V
    B = (4e0_rt-d)/(2e0_rt*(d+2e0_rt)**2)*V
    C = V/(d+2e0_rt)**2

    E = (A*fx(1)+ B*sum(fx(2:2+dint-1)+fx(2+dint:2+dint+dint-1)) + C*sum(fx(2+dint+dint:size(fx,1))))

end function


! Degree 11 rule due to Stroud
function deg11rule_pts(d) result(x)
    implicit none
    integer, intent(in) :: d
    real(rt), allocatable :: x(:,:)
    !local
    real(rt), allocatable :: w(:)
    real(rt), allocatable :: xlocal(:,:)
    integer :: o
    if (d<3 .or. d>5) STOP 'deg11rule can only handle a dimension of 3,4, or 5'

!     N    O
!    __  ___
!     3  151
!     4  417
!     5  983
    if (d==3) o = 151
    if (d==4) o = 417
    if (d==5) o = 983

    allocate(xlocal(o,d),x(o,d),w(o))

    call en_r2_11_1(d,1,o,xlocal,w)
    
    ! Modify the points so they can be thought of as innovations to a normal distribution
    x = sqrt(2.e0_rt)*xlocal

end function deg11rule_pts


! THE TWO ROUTINES ARE ADAPTED TO FORTRAN FROM JON BURKARDT'S WEBSITE:
! http://people.sc.fsu.edu/~jburkardt/m_src/stroud/en_r2_11_1.m
subroutine en_r2_11_1 ( n, option, o, x, w )
!*****************************************************************************80
!
!! EN_R2_11_1 implements the Stroud rule 11.1 for region EN_R2.
!
!  Discussion:
!
!    The rule has order 
!
!      O = ( 4 * N^5 - 20 * N^4 + 140 * N^3 - 130 * N^2 + 96 * N + 15 ) / 15.
!
!    The rule has precision P = 11.
!
!    EN_R2 is the entire N-dimensional space with weight function
!
!      w(x) = product ( 1 <= i <= n ) ( exp ( - x(i)^2 ) 
!
!    There are two versions of each rule, chosen by setting the
!    OPTION variable to 1 or 2.
!
!    The rule as tabulated by Stenger is available for N = 2 through 20.
!    This function accepts N = 3 through 5.
!
!     N    O
!    __  ___
!     3  151
!     4  417
!     5  983
!
!  Licensing:
!
!    This code is distributed under the GNU LGPL license.
!
!  Modified:
!
!    23 January 2010
!
!  Author:
!
!    John Burkardt
!
!  Reference:
!
!    Arthur Stroud,
!    Approximate Calculation of Multiple Integrals,
!    Prentice Hall, 1971,
!    ISBN: 0130438936,
!    LC: QA311.S85.
!
!  Parameters:
!
!    Input, integer ( kind = 4 ) N, the spatial dimension.
!    3 <= N <= 5.
!
!    Input, integer ( kind = 4 ) OPTION, chooses rule option 1 or 2.
!
!    Input, integer ( kind = 4 ) O, the order.
!
!    Output, real ( kind = 8 ) X(N,O), the abscissas.
!
!    Output, real ( kind = 8 ) W(O), the weights.
!
  implicit none

  integer ( kind = 4 ) n
  integer ( kind = 4 ) o

  real    ( kind = rt ) b0
  real    ( kind = rt ) b1
  real    ( kind = rt ) b2
  real    ( kind = rt ) b3
  real    ( kind = rt ) b4
  real    ( kind = rt ) b5
  real    ( kind = rt ) b6
  real    ( kind = rt ) b7
  real    ( kind = rt ) b8
  real    ( kind = rt ) b9
  real    ( kind = rt ) b10
  real    ( kind = rt ) b11
  real    ( kind = rt ) b12
  real    ( kind = rt ) b13
  real    ( kind = rt ) b14
  real    ( kind = rt ) b15
  integer ( kind = 4 ) i
  integer ( kind = 4 ) i1
  integer ( kind = 4 ) i2
  integer ( kind = 4 ) i3
  integer ( kind = 4 ) i4
  integer ( kind = 4 ) i5
  integer ( kind = 4 ) j
  integer ( kind = 4 ) k
  integer ( kind = 4 ) l
  integer ( kind = 4 ) m
!   logical              more
  integer ( kind = 4 ) option
  real    ( kind = rt ), parameter :: pi = 3.141592653589793e+00_rt
  real    ( kind = rt ) u
  real    ( kind = rt ) v
  real    ( kind = rt ) volume
  real    ( kind = rt ) w2
  real    ( kind = rt ) w(o)
  real    ( kind = rt ) x(n,o)

  if ( n < 3 .or. 5 < n ) then
    write ( *, '(a)' ) ' '
    write ( *, '(a)' ) 'EN_R2_11_1 - Fatal error!'
    write ( *, '(a)' ) '  3 <= N <= 5 required.'
    stop
  end if

  if ( option < 1 .or. 2 < option ) then
    write ( *, '(a)' ) ' '
    write ( *, '(a)' ) 'EN_R2_11_1 - Fatal error!'
    write ( *, '(a)' ) '  1 <= OPTION <= 2 required.'
    stop
  end if

  volume = sqrt ( pi ** n )

  if ( n == 3 .and. option == 1 ) then
    u =     0.235060497367449e+01_rt
    v =     0.436077411927617e+00_rt
    w2 =    0.133584907401370e+01_rt
    b0 =  - 0.881591029957858e+01_rt
    b1 =  - 0.751996143360650e-01_rt
    b2 =    0.621743189471515e+01_rt
    b3 =    0.241426451456494e+00_rt
    b4 =  - 0.120709739276065e-02_rt
    b5 =  - 0.427751221210138e+01_rt
    b6 =    0.550169924840163e-01_rt
    b7 =    0.237084999634707e-01_rt
    b8 =  - 0.169791992887741e-02_rt
    b9 =  - 0.252266276123350e-04_rt
    b10 =   0.326777873717691e+01_rt
    b11 =   0.968469949206802e-02_rt
    b12 =   0.789754514877422e-03_rt
    b13 =   0.000000000000000e+00_rt
    b14 =   0.000000000000000e+00_rt
    b15 =   0.000000000000000e+00_rt
  else if ( n == 3 .and. option == 2 ) then
    u =     0.235060497367449e+01_rt
    v =     0.133584907401370e+01_rt
    w2 =    0.436077411927617e+00_rt
    b0 =  - 0.141214037032900e+02_rt
    b1 =  - 0.803730274707282e-01_rt
    b2 =    0.235546545595906e+00_rt
    b3 =    0.888123191556611e+01_rt
    b4 =    0.142467131155533e-03_rt
    b5 =    0.582993124006494e-01_rt
    b6 =  - 0.561099173155661e+01_rt
    b7 =  - 0.204028691521686e-02_rt
    b8 =    0.252880089932256e-01_rt
    b9 =  - 0.814378678627283e-04_rt
    b10 =   0.804353953375146e-02_rt
    b11 =   0.393451849690453e+01_rt
    b12 =   0.171183493169724e-03_rt
    b13 =   0.000000000000000e+00_rt
    b14 =   0.000000000000000e+00_rt
    b15 =   0.000000000000000e+00_rt
  else if ( n == 4 .and. option == 1 ) then
    u =     0.235060497367449e+01_rt
    v =     0.436077411927617e+00_rt
    w2 =    0.133584907401370e+01_rt
    b0 =    0.241502736147339e+03_rt
    b1 =  - 0.196095938531478e+00_rt
    b2 =  - 0.128675737999280e+03_rt
    b3 =    0.307568784278696e+00_rt
    b4 =  - 0.480908422319460e-02_rt
    b5 =    0.698087019367085e+02_rt
    b6 =    0.631837143743771e-01_rt
    b7 =    0.392226151971179e-01_rt
    b8 =  - 0.300948471646799e-02_rt
    b9 =  - 0.650235306755170e-04_rt
    b10 = - 0.386951974646715e+02_rt
    b11 =   0.171656829095787e-01_rt
    b12 =   0.139980343116450e-02_rt
    b13 =   0.101552487093372e-04_rt
    b14 =   0.222435922356439e+02_rt
    b15 =   0.000000000000000e+00_rt
  else if ( n == 4 .and. option == 2 ) then
    u =     0.235060497367449e+01_rt
    v =     0.133584907401370e+01_rt
    w2 =    0.436077411927617e+00_rt
    b0 =  - 0.151944464736584e+03_rt
    b1 =  - 0.223498438689039e+00_rt
    b2 =    0.243574919068010e+00_rt
    b3 =    0.634373877008693e+02_rt
    b4 =  - 0.782065187814018e-04_rt
    b5 =    0.911833754536616e-01_rt
    b6 =  - 0.238927288245914e+02_rt
    b7 =  - 0.422314408318853e-02_rt
    b8 =    0.448218289217760e-01_rt
    b9 =  - 0.138053374667391e-03_rt
    b10 =   0.607473265800655e-02_rt
    b11 =   0.697375246129742e+01_rt
    b12 =   0.303414841680135e-03_rt
    b13 = - 0.314574391771792e-05_rt
    b14 =   0.409103498175100e-02_rt
    b15 =   0.000000000000000e+00_rt
  else if ( n == 5 .and. option == 1 ) then
    u =     0.235060497367449e+01_rt
    v =     0.436077411927617e+00_rt
    w2 =    0.133584907401370e+01_rt
    b0 =    0.255885269311763e+04_rt
    b1 =  - 0.439598677491526e+00_rt
    b2 =  - 0.106541406144610e+04_rt
    b3 =    0.453540909054264e+00_rt
    b4 =  - 0.132100905623778e-01_rt
    b5 =    0.418606568954203e+03_rt
    b6 =    0.511394563043680e-01_rt
    b7 =    0.645581013845604e-01_rt
    b8 =  - 0.533417277494500e-02_rt
    b9 =  - 0.137981626254496e-03_rt
    b10 = - 0.147436933189884e+03_rt
    b11 =   0.304253807765057e-01_rt
    b12 =   0.248108698207828e-02_rt
    b13 =   0.113652094546015e-04_rt
    b14 =   0.394257407160391e+02_rt
    b15 =   0.331725011358320e-05_rt
  else if ( n == 5 .and. option == 2 ) then
    u =     0.235060497367449e+01_rt
    v =     0.133584907401370e+01_rt
    w2 =    0.436077411927617e+00_rt
    b0 =  - 0.761305347548192e+03_rt
    b1 =  - 0.536360805019297e+00_rt
    b2 =    0.110669832078736e+00_rt
    b3 =    0.246421088923968e+03_rt
    b4 =  - 0.773649327968607e-03_rt
    b5 =    0.169088641205970e+00_rt
    b6 =  - 0.670700680243651e+02_rt
    b7 =  - 0.856090560229205e-02_rt
    b8 =    0.794446232770302e-01_rt
    b9 =  - 0.220272863263544e-03_rt
    b10 = - 0.373515812228225e-02_rt
    b11 =   0.123606544052884e+02_rt
    b12 =   0.537788804557843e-03_rt
    b13 = - 0.122101861480881e-04_rt
    b14 =   0.725117070759373e-02_rt
    b15 =   0.331725011358320e-05_rt
  end if

  x(1:n,1:o) = 0.0D+00

  k = 0
!
!  1 point.
!
  k = k + 1
! x(1:n,k) = 0.0D+00
  w(k) = b0
!
!  2 * N points.
!
  do i = 1, n
    k = k + 1
    x(i,k) = - u
    w(k) = b1
    k = k + 1
    x(i,k) = + u
    w(k) = b1
  end do
!
!  2 * N points.
!
  do i = 1, n
    k = k + 1
    x(i,k) = - v
    w(k) = b2
    k = k + 1
    x(i,k) = + v
    w(k) = b2
  end do
!
!  2 * N points.
!
  do i = 1, n
    k = k + 1
    x(i,k) = - w2
    w(k) = b3
    k = k + 1
    x(i,k) = + w2
    w(k) = b3
  end do
!
!  4 * ( N * ( N - 1 ) / 2 ) points.
!
  do i = 1, n - 1
    do j = i + 1, n
      k = k + 1
      x(i,k) = - u
      x(j,k) = - u
      w(k) = b4
      k = k + 1
      x(i,k) = - u
      x(j,k) = + u
      w(k) = b4
      k = k + 1
      x(i,k) = + u
      x(j,k) = - u
      w(k) = b4
      k = k + 1
      x(i,k) = + u
      x(j,k) = + u
      w(k) = b4
    end do
  end do
!
!  4 * ( N * ( N - 1 ) / 2 ) points.
!
  do i = 1, n - 1
    do j = i + 1, n
      k = k + 1
      x(i,k) = - v
      x(j,k) = - v
      w(k) = b5
      k = k + 1
      x(i,k) = - v
      x(j,k) = + v
      w(k) = b5
      k = k + 1
      x(i,k) = + v
      x(j,k) = - v
      w(k) = b5
      k = k + 1
      x(i,k) = + v
      x(j,k) = + v
      w(k) = b5
    end do
  end do
!
!  4 * ( N * ( N - 1 ) / 2 ) points.
!
  do i = 1, n - 1
    do j = i + 1, n
      k = k + 1
      x(i,k) = - w2
      x(j,k) = - w2
      w(k) = b6
      k = k + 1
      x(i,k) = - w2
      x(j,k) = + w2
      w(k) = b6
      k = k + 1
      x(i,k) = + w2
      x(j,k) = - w2
      w(k) = b6
      k = k + 1
      x(i,k) = + w2
      x(j,k) = + w2
      w(k) = b6
    end do
  end do
!
!  4 * ( N * ( N - 1 ) ) points.
!
  do i = 1, n - 1
    do j = i + 1, n
      k = k + 1
      x(i,k) = - u
      x(j,k) = - v
      w(k) = b7
      k = k + 1
      x(i,k) = - u
      x(j,k) = + v
      w(k) = b7
      k = k + 1
      x(i,k) = + u
      x(j,k) = - v
      w(k) = b7
      k = k + 1
      x(i,k) = + u
      x(j,k) = + v
      w(k) = b7
      k = k + 1
      x(i,k) = - v
      x(j,k) = - u
      w(k) = b7
      k = k + 1
      x(i,k) = - v
      x(j,k) = + u
      w(k) = b7
      k = k + 1
      x(i,k) = + v
      x(j,k) = - u
      w(k) = b7
      k = k + 1
      x(i,k) = + v
      x(j,k) = + u
      w(k) = b7
    end do
  end do
!
!  4 * ( N * ( N - 1 ) ) points.
!
  do i = 1, n - 1
    do j = i + 1, n
      k = k + 1
      x(i,k) = - u
      x(j,k) = - w2
      w(k) = b8
      k = k + 1
      x(i,k) = - u
      x(j,k) = + w2
      w(k) = b8
      k = k + 1
      x(i,k) = + u
      x(j,k) = - w2
      w(k) = b8
      k = k + 1
      x(i,k) = + u
      x(j,k) = + w2
      w(k) = b8
      k = k + 1
      x(i,k) = - w2
      x(j,k) = - u
      w(k) = b8
      k = k + 1
      x(i,k) = - w2
      x(j,k) = + u
      w(k) = b8
      k = k + 1
      x(i,k) = + w2
      x(j,k) = - u
      w(k) = b8
      k = k + 1
      x(i,k) = + w2
      x(j,k) = + u
      w(k) = b8
    end do
  end do
!
!  8 * ( N * ( N - 1 ) * ( N - 2 ) / 6 ) points.
!
  do i = 1, n - 2
    do j = i + 1, n - 1
      do l = j + 1, n
        k = k + 1
        x(i,k) = - u
        x(j,k) = - u
        x(l,k) = - u
        w(k) = b9
        k = k + 1
        x(i,k) = - u
        x(j,k) = - u
        x(l,k) = + u
        w(k) = b9
        k = k + 1
        x(i,k) = - u
        x(j,k) = + u
        x(l,k) = - u
        w(k) = b9
        k = k + 1
        x(i,k) = - u
        x(j,k) = + u
        x(l,k) = + u
        w(k) = b9
        k = k + 1
        x(i,k) = + u
        x(j,k) = - u
        x(l,k) = - u
        w(k) = b9
        k = k + 1
        x(i,k) = + u
        x(j,k) = - u
        x(l,k) = + u
        w(k) = b9
        k = k + 1
        x(i,k) = + u
        x(j,k) = + u
        x(l,k) = - u
        w(k) = b9
        k = k + 1
        x(i,k) = + u
        x(j,k) = + u
        x(l,k) = + u
        w(k) = b9
      end do
    end do
  end do
!
!  8 * ( N * ( N - 1 ) * ( N - 2 ) / 6 ) points.
!
  do i = 1, n - 2
    do j = i + 1, n - 1
      do l = j + 1, n
        k = k + 1
        x(i,k) = - v
        x(j,k) = - v
        x(l,k) = - v
        w(k) = b10
        k = k + 1
        x(i,k) = - v
        x(j,k) = - v
        x(l,k) = + v
        w(k) = b10
        k = k + 1
        x(i,k) = - v
        x(j,k) = + v
        x(l,k) = - v
        w(k) = b10
        k = k + 1
        x(i,k) = - v
        x(j,k) = + v
        x(l,k) = + v
        w(k) = b10
        k = k + 1
        x(i,k) = + v
        x(j,k) = - v
        x(l,k) = - v
        w(k) = b10
        k = k + 1
        x(i,k) = + v
        x(j,k) = - v
        x(l,k) = + v
        w(k) = b10
        k = k + 1
        x(i,k) = + v
        x(j,k) = + v
        x(l,k) = - v
        w(k) = b10
        k = k + 1
        x(i,k) = + v
        x(j,k) = + v
        x(l,k) = + v
        w(k) = b10
      end do
    end do
  end do
!
!  8 * ( N * ( N - 1 ) * ( N - 2 ) / 6 ) points.
!
  do i = 1, n - 2
    do j = i + 1, n - 1
      do l = j + 1, n
        k = k + 1
        x(i,k) = - w2
        x(j,k) = - w2
        x(l,k) = - w2
        w(k) = b11
        k = k + 1
        x(i,k) = - w2
        x(j,k) = - w2
        x(l,k) = + w2
        w(k) = b11
        k = k + 1
        x(i,k) = - w2
        x(j,k) = + w2
        x(l,k) = - w2
        w(k) = b11
        k = k + 1
        x(i,k) = - w2
        x(j,k) = + w2
        x(l,k) = + w2
        w(k) = b11
        k = k + 1
        x(i,k) = + w2
        x(j,k) = - w2
        x(l,k) = - w2
        w(k) = b11
        k = k + 1
        x(i,k) = + w2
        x(j,k) = - w2
        x(l,k) = + w2
        w(k) = b11
        k = k + 1
        x(i,k) = + w2
        x(j,k) = + w2
        x(l,k) = - w2
        w(k) = b11
        k = k + 1
        x(i,k) = + w2
        x(j,k) = + w2
        x(l,k) = + w2
        w(k) = b11
      end do
    end do
  end do
!
!  8 * ( N * ( N - 1 ) * ( N - 2 ) / 2 ) points.
!
  do i = 1, n - 2
    do j = i + 1, n - 1
      do l = j + 1, n
        k = k + 1
        x(i,k) = - u
        x(j,k) = - u
        x(l,k) = - v
        w(k) = b12
        k = k + 1
        x(i,k) = - u
        x(j,k) = - u
        x(l,k) = + v
        w(k) = b12
        k = k + 1
        x(i,k) = - u
        x(j,k) = + u
        x(l,k) = - v
        w(k) = b12
        k = k + 1
        x(i,k) = - u
        x(j,k) = + u
        x(l,k) = + v
        w(k) = b12
        k = k + 1
        x(i,k) = + u
        x(j,k) = - u
        x(l,k) = - v
        w(k) = b12
        k = k + 1
        x(i,k) = + u
        x(j,k) = - u
        x(l,k) = + v
        w(k) = b12
        k = k + 1
        x(i,k) = + u
        x(j,k) = + u
        x(l,k) = - v
        w(k) = b12
        k = k + 1
        x(i,k) = + u
        x(j,k) = + u
        x(l,k) = + v
        w(k) = b12
        k = k + 1
        x(i,k) = - u
        x(j,k) = - v
        x(l,k) = - u
        w(k) = b12
        k = k + 1
        x(i,k) = - u
        x(j,k) = - v
        x(l,k) = + u
        w(k) = b12
        k = k + 1
        x(i,k) = - u
        x(j,k) = + v
        x(l,k) = - u
        w(k) = b12
        k = k + 1
        x(i,k) = - u
        x(j,k) = + v
        x(l,k) = + u
        w(k) = b12
        k = k + 1
        x(i,k) = + u
        x(j,k) = - v
        x(l,k) = - u
        w(k) = b12
        k = k + 1
        x(i,k) = + u
        x(j,k) = - v
        x(l,k) = + u
        w(k) = b12
        k = k + 1
        x(i,k) = + u
        x(j,k) = + v
        x(l,k) = - u
        w(k) = b12
        k = k + 1
        x(i,k) = + u
        x(j,k) = + v
        x(l,k) = + u
        w(k) = b12
        k = k + 1
        x(i,k) = - v
        x(j,k) = - u
        x(l,k) = - u
        w(k) = b12
        k = k + 1
        x(i,k) = - v
        x(j,k) = - u
        x(l,k) = + u
        w(k) = b12
        k = k + 1
        x(i,k) = - v
        x(j,k) = + u
        x(l,k) = - u
        w(k) = b12
        k = k + 1
        x(i,k) = - v
        x(j,k) = + u
        x(l,k) = + u
        w(k) = b12
        k = k + 1
        x(i,k) = + v
        x(j,k) = - u
        x(l,k) = - u
        w(k) = b12
        k = k + 1
        x(i,k) = + v
        x(j,k) = - u
        x(l,k) = + u
        w(k) = b12
        k = k + 1
        x(i,k) = + v
        x(j,k) = + u
        x(l,k) = - u
        w(k) = b12
        k = k + 1
        x(i,k) = + v
        x(j,k) = + u
        x(l,k) = + u
        w(k) = b12
      end do
    end do
  end do
!
!  16 * ( N * ( N - 1 ) * ( N - 2 ) * ( N - 3 ) / 24 ) points.
!
  do i = 1, n - 3
    do j = i + 1, n - 2
      do l = j + 1, n - 1
        do m = l + 1, n
          k = k + 1
          x(i,k) = - u
          x(j,k) = - u
          x(l,k) = - u
          x(m,k) = - u
          w(k) = b13
          k = k + 1
          x(i,k) = - u
          x(j,k) = - u
          x(l,k) = - u
          x(m,k) = + u
          w(k) = b13
          k = k + 1
          x(i,k) = - u
          x(j,k) = - u
          x(l,k) = + u
          x(m,k) = - u
          w(k) = b13
          k = k + 1
          x(i,k) = - u
          x(j,k) = - u
          x(l,k) = + u
          x(m,k) = + u
          w(k) = b13
          k = k + 1
          x(i,k) = - u
          x(j,k) = + u
          x(l,k) = - u
          x(m,k) = - u
          w(k) = b13
          k = k + 1
          x(i,k) = - u
          x(j,k) = + u
          x(l,k) = - u
          x(m,k) = + u
          w(k) = b13
          k = k + 1
          x(i,k) = - u
          x(j,k) = + u
          x(l,k) = + u
          x(m,k) = - u
          w(k) = b13
          k = k + 1
          x(i,k) = - u
          x(j,k) = + u
          x(l,k) = + u
          x(m,k) = + u
          w(k) = b13
          k = k + 1
          x(i,k) = + u
          x(j,k) = - u
          x(l,k) = - u
          x(m,k) = - u
          w(k) = b13
          k = k + 1
          x(i,k) = + u
          x(j,k) = - u
          x(l,k) = - u
          x(m,k) = + u
          w(k) = b13
          k = k + 1
          x(i,k) = + u
          x(j,k) = - u
          x(l,k) = + u
          x(m,k) = - u
          w(k) = b13
          k = k + 1
          x(i,k) = + u
          x(j,k) = - u
          x(l,k) = + u
          x(m,k) = + u
          w(k) = b13
          k = k + 1
          x(i,k) = + u
          x(j,k) = + u
          x(l,k) = - u
          x(m,k) = - u
          w(k) = b13
          k = k + 1
          x(i,k) = + u
          x(j,k) = + u
          x(l,k) = - u
          x(m,k) = + u
          w(k) = b13
          k = k + 1
          x(i,k) = + u
          x(j,k) = + u
          x(l,k) = + u
          x(m,k) = - u
          w(k) = b13
          k = k + 1
          x(i,k) = + u
          x(j,k) = + u
          x(l,k) = + u
          x(m,k) = + u
          w(k) = b13
        end do
      end do
    end do
  end do
!
!  16 * ( N * ( N - 1 ) * ( N - 2 ) * ( N - 3 ) / 24 ) points.
!
  do i = 1, n - 3
    do j = i + 1, n - 2
      do l = j + 1, n - 1
        do m = l + 1, n
          k = k + 1
          x(i,k) = - v
          x(j,k) = - v
          x(l,k) = - v
          x(m,k) = - v
          w(k) = b14
          k = k + 1
          x(i,k) = - v
          x(j,k) = - v
          x(l,k) = - v
          x(m,k) = + v
          w(k) = b14
          k = k + 1
          x(i,k) = - v
          x(j,k) = - v
          x(l,k) = + v
          x(m,k) = - v
          w(k) = b14
          k = k + 1
          x(i,k) = - v
          x(j,k) = - v
          x(l,k) = + v
          x(m,k) = + v
          w(k) = b14
          k = k + 1
          x(i,k) = - v
          x(j,k) = + v
          x(l,k) = - v
          x(m,k) = - v
          w(k) = b14
          k = k + 1
          x(i,k) = - v
          x(j,k) = + v
          x(l,k) = - v
          x(m,k) = + v
          w(k) = b14
          k = k + 1
          x(i,k) = - v
          x(j,k) = + v
          x(l,k) = + v
          x(m,k) = - v
          w(k) = b14
          k = k + 1
          x(i,k) = - v
          x(j,k) = + v
          x(l,k) = + v
          x(m,k) = + v
          w(k) = b14
          k = k + 1
          x(i,k) = + v
          x(j,k) = - v
          x(l,k) = - v
          x(m,k) = - v
          w(k) = b14
          k = k + 1
          x(i,k) = + v
          x(j,k) = - v
          x(l,k) = - v
          x(m,k) = + v
          w(k) = b14
          k = k + 1
          x(i,k) = + v
          x(j,k) = - v
          x(l,k) = + v
          x(m,k) = - v
          w(k) = b14
          k = k + 1
          x(i,k) = + v
          x(j,k) = - v
          x(l,k) = + v
          x(m,k) = + v
          w(k) = b14
          k = k + 1
          x(i,k) = + v
          x(j,k) = + v
          x(l,k) = - v
          x(m,k) = - v
          w(k) = b14
          k = k + 1
          x(i,k) = + v
          x(j,k) = + v
          x(l,k) = - v
          x(m,k) = + v
          w(k) = b14
          k = k + 1
          x(i,k) = + v
          x(j,k) = + v
          x(l,k) = + v
          x(m,k) = - v
          w(k) = b14
          k = k + 1
          x(i,k) = + v
          x(j,k) = + v
          x(l,k) = + v
          x(m,k) = + v
          w(k) = b14
        end do
      end do
    end do
  end do
!
!  All quintuples UUUUU with 32 sign combinations.
!
  do i1 = 1, n - 4
    do i2 = i1 + 1, n - 3
      do i3 = i2 + 1, n - 2
        do i4 = i3 + 1, n - 1
          do i5 = i4 + 1, n
            k = k + 1
            x(i1,k) = - u
            x(i2,k) = - u
            x(i3,k) = - u
            x(i4,k) = - u
            x(i5,k) = - u
            w(k) = b15
            k = k + 1
            x(i1,k) = - u
            x(i2,k) = - u
            x(i3,k) = - u
            x(i4,k) = - u
            x(i5,k) = + u
            w(k) = b15
            k = k + 1
            x(i1,k) = - u
            x(i2,k) = - u
            x(i3,k) = - u
            x(i4,k) = + u
            x(i5,k) = - u
            w(k) = b15
            k = k + 1
            x(i1,k) = - u
            x(i2,k) = - u
            x(i3,k) = - u
            x(i4,k) = + u
            x(i5,k) = + u
            w(k) = b15
            k = k + 1
            x(i1,k) = - u
            x(i2,k) = - u
            x(i3,k) = + u
            x(i4,k) = - u
            x(i5,k) = - u
            w(k) = b15
            k = k + 1
            x(i1,k) = - u
            x(i2,k) = - u
            x(i3,k) = + u
            x(i4,k) = - u
            x(i5,k) = + u
            w(k) = b15
            k = k + 1
            x(i1,k) = - u
            x(i2,k) = - u
            x(i3,k) = + u
            x(i4,k) = + u
            x(i5,k) = - u
            w(k) = b15
            k = k + 1
            x(i1,k) = - u
            x(i2,k) = - u
            x(i3,k) = + u
            x(i4,k) = + u
            x(i5,k) = + u
            w(k) = b15
            k = k + 1
            x(i1,k) = - u
            x(i2,k) = + u
            x(i3,k) = - u
            x(i4,k) = - u
            x(i5,k) = - u
            w(k) = b15
            k = k + 1
            x(i1,k) = - u
            x(i2,k) = + u
            x(i3,k) = - u
            x(i4,k) = - u
            x(i5,k) = + u
            w(k) = b15
            k = k + 1
            x(i1,k) = - u
            x(i2,k) = + u
            x(i3,k) = - u
            x(i4,k) = + u
            x(i5,k) = - u
            w(k) = b15
            k = k + 1
            x(i1,k) = - u
            x(i2,k) = + u
            x(i3,k) = - u
            x(i4,k) = + u
            x(i5,k) = + u
            w(k) = b15
            k = k + 1
            x(i1,k) = - u
            x(i2,k) = + u
            x(i3,k) = + u
            x(i4,k) = - u
            x(i5,k) = - u
            w(k) = b15
            k = k + 1
            x(i1,k) = - u
            x(i2,k) = + u
            x(i3,k) = + u
            x(i4,k) = - u
            x(i5,k) = + u
            w(k) = b15
            k = k + 1
            x(i1,k) = - u
            x(i2,k) = + u
            x(i3,k) = + u
            x(i4,k) = + u
            x(i5,k) = - u
            w(k) = b15
            k = k + 1
            x(i1,k) = - u
            x(i2,k) = + u
            x(i3,k) = + u
            x(i4,k) = + u
            x(i5,k) = + u
            w(k) = b15
            k = k + 1
            x(i1,k) = + u
            x(i2,k) = - u
            x(i3,k) = - u
            x(i4,k) = - u
            x(i5,k) = - u
            w(k) = b15
            k = k + 1
            x(i1,k) = + u
            x(i2,k) = - u
            x(i3,k) = - u
            x(i4,k) = - u
            x(i5,k) = + u
            w(k) = b15
            k = k + 1
            x(i1,k) = + u
            x(i2,k) = - u
            x(i3,k) = - u
            x(i4,k) = + u
            x(i5,k) = - u
            w(k) = b15
            k = k + 1
            x(i1,k) = + u
            x(i2,k) = - u
            x(i3,k) = - u
            x(i4,k) = + u
            x(i5,k) = + u
            w(k) = b15
            k = k + 1
            x(i1,k) = + u
            x(i2,k) = - u
            x(i3,k) = + u
            x(i4,k) = - u
            x(i5,k) = - u
            w(k) = b15
            k = k + 1
            x(i1,k) = + u
            x(i2,k) = - u
            x(i3,k) = + u
            x(i4,k) = - u
            x(i5,k) = + u
            w(k) = b15
            k = k + 1
            x(i1,k) = + u
            x(i2,k) = - u
            x(i3,k) = + u
            x(i4,k) = + u
            x(i5,k) = - u
            w(k) = b15
            k = k + 1
            x(i1,k) = + u
            x(i2,k) = - u
            x(i3,k) = + u
            x(i4,k) = + u
            x(i5,k) = + u
            w(k) = b15
            k = k + 1
            x(i1,k) = + u
            x(i2,k) = + u
            x(i3,k) = - u
            x(i4,k) = - u
            x(i5,k) = - u
            w(k) = b15
            k = k + 1
            x(i1,k) = + u
            x(i2,k) = + u
            x(i3,k) = - u
            x(i4,k) = - u
            x(i5,k) = + u
            w(k) = b15
            k = k + 1
            x(i1,k) = + u
            x(i2,k) = + u
            x(i3,k) = - u
            x(i4,k) = + u
            x(i5,k) = - u
            w(k) = b15
            k = k + 1
            x(i1,k) = + u
            x(i2,k) = + u
            x(i3,k) = - u
            x(i4,k) = + u
            x(i5,k) = + u
            w(k) = b15
            k = k + 1
            x(i1,k) = + u
            x(i2,k) = + u
            x(i3,k) = + u
            x(i4,k) = - u
            x(i5,k) = - u
            w(k) = b15
            k = k + 1
            x(i1,k) = + u
            x(i2,k) = + u
            x(i3,k) = + u
            x(i4,k) = - u
            x(i5,k) = + u
            w(k) = b15
            k = k + 1
            x(i1,k) = + u
            x(i2,k) = + u
            x(i3,k) = + u
            x(i4,k) = + u
            x(i5,k) = - u
            w(k) = b15
            k = k + 1
            x(i1,k) = + u
            x(i2,k) = + u
            x(i3,k) = + u
            x(i4,k) = + u
            x(i5,k) = + u
            w(k) = b15
          end do
        end do
      end do
    end do
  end do

  return
end subroutine

subroutine en_r2_11_1_size ( n, option, o )

!*****************************************************************************80
!
!! EN_R2_11_1_SIZE sizes the Stroud rule 11.1 for region EN_R2.
!
!  Discussion:
!
!    The rule has order 
!
!      O = ( 4 * N^5 - 20 * N^4 + 140 * N^3 - 130 * N^2 + 96 * N + 15 ) / 15.
!
!    The rule has precision P = 11.
!
!    EN_R2 is the entire N-dimensional space with weight function
!
!      w(x) = product ( 1 <= i <= n ) ( exp ( - x(i)^2 ) 
!
!    There are two versions of each rule, chosen by setting the
!    OPTION variable to 1 or 2.
!
!    The rule as tabulated by Stenger is available for N = 2 through 20.
!    This function accepts N = 3 through 5.
!
!     N    O
!    __  ___
!     3  151
!     4  417
!     5  983
!
!  Licensing:
!
!    This code is distributed under the GNU LGPL license.
!
!  Modified:
!
!    27 January 2010
!
!  Author:
!
!    John Burkardt
!
!  Reference:
!
!    Arthur Stroud,
!    Approximate Calculation of Multiple Integrals,
!    Prentice Hall, 1971,
!    ISBN: 0130438936,
!    LC: QA311.S85.
!
!  Parameters:
!
!    Input, integer ( kind = 4 ) N, the spatial dimension.
!    3 <= N <= 5.
!
!    Input, integer ( kind = 4 ) OPTION, chooses rule option 1 or 2.
!
!    Output, integer ( kind = 4 ) O, the order.
!
  implicit none

  integer ( kind = 4 ) n
  integer ( kind = 4 ) o
  integer ( kind = 4 ) option

  if ( n < 3 .or. 5 < n ) then
    write ( *, '(a)' ) ' '
    write ( *, '(a)' ) 'EN_R2_11_1_SIZE - Fatal error!'
    write ( *, '(a)' ) '  3 <= N <= 5 required.'
    stop
  end if

  if ( option < 1 .or. 2 < option ) then
    write ( *, '(a)' ) ' '
    write ( *, '(a)' ) 'EN_R2_11_1_SIZE - Fatal error!'
    write ( *, '(a)' ) '  1 <= OPTION <= 2 required.'
    stop
  end if

  o = ( 4 * n ** 5 - 20 * n ** 4 + 140 * n ** 3 - 130 * n ** 2 &
    + 96 * n + 15 ) / 15


  return
end subroutine

end module mod_quad


module mod_quad_test
    use mod_quad
contains

    subroutine test_gausslegendre()
        use mod_matlab
        implicit none
        real(rt) :: x(5),xtrue(5),w(5),wtrue(5),tmp
        real(rt), allocatable :: roots(:),P(:)
        integer :: i,order
        integer, parameter :: maxorder = 1001

        allocate(roots(maxorder),P(maxorder))
        do order = 1,maxorder

            ! Do randomly because this is costly
            if (randu()>.01d0) cycle 

            call legendre_roots(roots(1:order))
            ! Check that P is zero at the roots
            do i = 1,order
                call legendre_eval(roots(i),order,P(i))
            end do

            print*,'order : ',order,' max abs P ',maxval(abs(P(1:order))),'min distance between roots',minval(roots(2:order)-roots(1:order-1))
            if (maxval(abs(P(1:order)))>sqrt(epsilon(0._rt))) stop 'test_gausslegendre: failed test 0'
        end do


        ! x = [-1d0,-.5d0,0d0,.5d0,1d0]
        ! do i = 1,size(x)
        !     call legendre_eval(x(i),3,xtrue(i))
        ! end do
        ! call plot(x,xtrue,'4th order Legendre Polynomial')
        ! stop

        call legendrequad(x,w,-1d0,1d0)

        print*,'x',x
        print*,'w',w

        xtrue(1:3) = [-.90618d0,-.538469d0,0d0]
        xtrue(4) = -xtrue(2)
        xtrue(5) = -xtrue(1)

        wtrue(1:3) = [0.236927d0,0.478629d0,0.568889d0]
        wtrue(4) = wtrue(2)
        wtrue(5) = wtrue(1)

        if (maxval(abs(x-xtrue))>1d-4) stop 'test_gausslegendre: failed test 1'
        if (maxval(abs(w-wtrue))>1d-4) stop 'test_gausslegendre: failed test 2'

        ! Integrate a 4th degree polynomial from -5 to 10, so should be exact : int 5*x**4 dx = x**5 | 10,-5 = 10**5 + 5**5 
        call legendrequad(x,w,-5d0,10d0)
        x = 5d0*x**4
        tmp = 10**5+5**5
        ! print*,'dot_product(x,w)',dot_product(x,w),'true',10**5+5**5
        if (abs(dot_product(x,w)-tmp)>1e-6*tmp) stop 'test_gausslegendre: failed test 3'

        print*,'test_gausslegendre: passed'

    end subroutine 

    subroutine test_gausshermite()
        implicit none
        real(rt) :: x(5),xtrue(5)

        call hermite_roots(x)

        xtrue(1:3) = [-2.02018d0,-.958572d0,0d0]
        xtrue(4) = -xtrue(2)
        xtrue(5) = -xtrue(1)

        if (maxval(abs(x-xtrue))>1d-4) stop 'test_gausshemite: failed'

    end subroutine 

    subroutine test_sub_simpsonquad()
        implicit none
        real(rt) :: x(11),w(11),y(11)
        real(rt) :: lo,hi, Eapprox,E
        
        lo = 2e0_rt
        hi = 5e0_rt
        call sub_simpsonquad(x,w,lo,hi)

        ! Integrate a quadratic function (should be exact)
        ! y = \int^h_l a + bx + cx**2 
        !   = ax + bx**2/2 + c x**3/3 |^h_l
        y = -.5e0_rt - 5e0_rt*x + 3e0_rt*x**2
        Eapprox = dot_product(w,y)
        E = -.5e0_rt*hi - 5e0_rt*hi**2/2e0_rt + 3e0_rt*hi**3/3e0_rt
        E = E - (-.5e0_rt*lo - 5e0_rt*lo**2/2e0_rt + 3e0_rt*lo**3/3e0_rt)

        if (abs(E-Eapprox)>abs(E)*sqrt(epsilon(0.0_rt))) STOP 'ERROR: test_sub_simpsonquad: not exact'

        print*,'test_sub_simpsonquad: passed'

    end subroutine

    subroutine test_intPiecewiseLinear()
        implicit none
        real(rt) :: x(4),y(4),a,b,q,qtrue,q2
        integer :: i

        x = [-3e0_rt,0e0_rt,10e0_rt,20e0_rt]
        y = -sqrt(2e0_rt)*x + 1e0_rt  ! Linear function so should integrate exactly

        a = x(1)
        b = x(size(x))
        qtrue = (b-a) - sqrt(2e0_rt)/2e0_rt*(b**2-a**2)
        q2 = dot_product(y,trapzquad_fixedx(x))
        if (abs(qtrue-q2)>abs(qtrue)*sqrt(epsilon(0._rt))) then
            print*,'qtrue',qtrue
            print*,'q2',q2
            print*,'sum(trapzquad_fixedx(x))',sum(trapzquad_fixedx(x))
            print*,'|qtrue-q2|=',abs(qtrue-q2)
            print*,'tolerance=',abs(qtrue)*sqrt(epsilon(0._rt))
            STOP 'qtrue and q2 are far apart'
        end if

        do i = 1,10000

            call random_number(a)
            call random_number(b)

            ! Get a in range [-25e0_rt,25e0_rt]
            a = a*50e0_rt - 25e0_rt
            ! Get b-a in range [0e0_rt,30e0_rt]
            b = a + b*30e0_rt

            ! Integrate
            q = intPiecewiseLinear(x,y,a,b)
            qtrue = (b-a) - sqrt(2e0_rt)/2e0_rt*(b**2-a**2)

            if (abs(qtrue-q)>abs(qtrue)*sqrt(epsilon(0.0_rt))) then
                print*,'i',i,'a',a,'b',b
                print*,'qtrue',qtrue
                print*,'q',q
                print*,'|qtrue-q|=',abs(qtrue-q)
                print*,'tolerance=',abs(qtrue)*sqrt(epsilon(0._rt))
                STOP 'qtrue and q are far apart'
            end if


        end do
        print*,'test_intPiecewiseLinear: passed'

    end subroutine
end module mod_quad_test 
! This module contains several ways to discretize an AR1 process
! -- sub_rouwenhorst
! -- sub_tauchen
! Additional, it provides a routine that will sample from an AR1 process.
! -- sub_drawAR1
module mod_discreteAR1
    use mod_types
    implicit none
contains

! Discretizes the following process
! x_{i,t} = (1-rho(S_t))*mu(S_t) + rho(S_t)*x_{i,t-1} + nu(S_t) epsilon_{i,t}
! where epsilon_{i,t} is N(0,1) and 
! Prob(S_{t}|S_{t-1}) is specified by the diagonals
! Note the timing might not be what you expect
subroutine sub_regime_switch_2state_tauchen(z,Fzz,mu,rho,nu,Pdiag,nR)
    use mod_util
    use mod_matlab
    implicit none
    real(rt), intent(out) :: z(:),Fzz(:,:)
    real(rt), intent(in) :: mu(:), rho(:), nu(:), Pdiag(:) ! length 2
    integer, intent(in) :: nR(:) ! length 2
    ! local
    real(rt) :: P(2,2),invar(2),mid,mid2,tmp,cdnalmean,cdnalstd
    real(rt), allocatable :: s(:),ss(:)
    integer, allocatable :: R(:),Rlast(:)
    integer :: nZ,iter,iz,izp,ireg,iregp,i,ilo,ihi
    type(type_seed) :: randseed

    nZ = size(z)

    call assert(size(nR)==2,'size(nR)/=2')
    call assert(sum(nR)==nZ,'sum(nR)/=nZ')
    call assert(nZ==size(Fzz,1),'size(Fzz,1)/=nZ')
    call assert(nZ==size(Fzz,2),'size(Fzz,2)/=nZ')
    call assert(size(mu)==2,'size(mu)/=2')
    call assert(size(rho)==2,'size(rho)/=2')
    call assert(size(nu)==2,'size(nu)/=2')
    call assert(size(Pdiag)==2,'size(Pdiag)/=2')

    P(1,1) = Pdiag(1)
    P(2,2) = Pdiag(2)
    P(1,2) = 1d0-P(1,1)
    P(2,1) = 1d0-P(2,2)
    
    ! Compute the invariant distribution of the regime siwtching
    invar = nullspace(eye(2)-transpose(P))
    invar = invar / sum(invar)

    ! Save the current random seed and set a new one
    call randseed%save()
    call randseed%simpleset(123-16+59)

    ! Simulate to get an idea of the variance and where to place points
    allocate(s(min(100000,max(10000,nZ*1000))))
    allocate(R(size(s)))
    allocate(Rlast(size(s)))
    R(1:nint(invar(1)*size(s))) = 1
    R(nint(invar(1)*size(s))+1:size(s)) = 2
    s = mu(R)

    do iter = 1,1000
        Rlast = R
        do i = 1,size(s)

            call random_number(tmp)
            R(i) = merge(1,2,tmp<P(R(i),1))
            s(i) = (1d0-rho(R(i)))*mu(R(i)) + rho(R(i))*s(i) + nu(R(i))*randn()

            ! R and s are now the time t value of the state
        end do
    end do
    
    do ireg = 1,2

        allocate(ss(count(R==ireg)))

        ! Pack just the prod values for that regime
        ss = pack(s,R==ireg)

        ! Now, sort the ss and place into bins
        ss = sort(ss)

        do i = 1,nR(ireg)

            ilo = 1+(i-1)*size(ss)/nR(ireg)
            ihi = i*size(ss)/nR(ireg)

            ! Get the conditional mean within the bin
            z(i+(ireg-1)*nR(1)) = sum(ss(ilo:ihi))/(ihi-ilo+1) 
            
        end do

        deallocate(ss)

    end do

    ! Now, given grid points, use Tauchen's procedure
    do iz = 1,nZ

        ireg = merge(1,2,iz<=nR(1))

        do izp = 1,nZ

            iregp = merge(1,2,izp<=nR(1))

            cdnalmean = (1d0-rho(iregp))*mu(iregp) + rho(iregp)*z(iz)
            cdnalstd = nu(iregp)

            if (izp==1 .or. izp==nR(1)+1) then

                ! Left boundary
                mid = .5d0*(z(izp)+z(izp+1))
                Fzz(iz,izp) = P(ireg,iregp)*normcdf(mid,cdnalmean,cdnalstd)

            elseif (izp==nR(1) .or. izp==nR(1)+nR(2)) then

                ! Right boundary
                mid = .5d0*(z(izp-1)+z(izp)) 
                Fzz(iz,izp) = P(ireg,iregp)*(1d0-normcdf(mid,cdnalmean,cdnalstd))

            else

                ! Interior
                mid = .5d0*(z(izp-1)+z(izp))
                mid2 = .5d0*(z(izp)+z(izp+1))
                Fzz(iz,izp) = P(ireg,iregp)*(normcdf(mid2,cdnalmean,cdnalstd)-normcdf(mid,cdnalmean,cdnalstd))

            end if
            
        end do
    end do

    ! Revert to whatever the seed was
    call randseed%revert()
    
end subroutine

subroutine drawMarkovLifeCycle(ix,Fxpcxt,Fx0)
    implicit none
    real(rt), intent(in) :: Fxpcxt(:,:,:), Fx0(:)
    integer, intent(out) :: ix(:)
    ! local
    integer :: nx, it, nt

    nx = size(Fxpcxt,1)
    nt = size(ix)
    if (size(Fxpcxt,2)/=nx) stop 'ERROR: drawMarkovLifeCycle: input wrong dim'
    if (size(Fxpcxt,3)/=nt .and. size(Fxpcxt,3)/=nt-1) stop 'ERROR: drawMarkovLifeCycle: input wrong dim'
    if (size(Fx0)/=nx) stop 'ERROR: drawMarkovLifeCycle: input wrong dim'

    ! Get initial draw
    ix(1) = drawIid(Fx0,0) 

    ! Simulate the rest
    do it = 1,nt-1
        ix(it+1) = drawIid(Fxpcxt(:,ix(it),iT),0)
    end do
    

end subroutine

! Draw from the markov chain given current state i
! There are two options in this routine
! option == 0, slow but easy for user, F is the mdf
! option == 1, fast but involved for user, F is the cumulative summation transpose of the markov chain
subroutine draw_mdf_iid(idraws,F,nF)
    use mod_matlab, only: cumsum,bsearch
    implicit none
    integer, intent(in) :: nF
    real(rt), intent(in) :: F(*)
    integer, intent(out) :: idraws(:)
    ! local
    integer :: i,ilo
    real(rt) :: sumF,u01
    real(rt), allocatable :: CDF(:)

    sumF = sum(F(1:nF))
    if (abs(sumF-1d0)>1d-10) write(0,*) 'WARNING: draw_mdf_iid: input distribution does not sum to 1'
    
    ! Get the CDF
    allocate(CDF(1:nF+1))
    CDF(2:nF+1) = cumsum(F(1:nF))/sumF
    CDF(1) = 0d0

    do i = 1,size(idraws)
        call random_number(u01)
        call bsearch(ilo,u01,CDF) 
        ! ilo is the index of the lower point bracketing u01. So u01 in [CDF(ilo),CDF(ilo+1)]. Since CDF(1) = 0, this is what we want.
        idraws(i) = ilo
    end do
    
end subroutine


! Draw from the markov chain given current state i
! There are two options in this routine
! option == 0, slow but easy for user, F is the mdf
! option == 1, fast but involved for user, F is the cumulative summation transpose of the markov chain
! If u01 is present, then it is used as a random variable (o/w, one is drawn)
function drawIid(F,option,u01) result(izp)
    use mod_matlab, only: cumsum
    implicit none
    real(rt), dimension(1:), intent(in) :: F 
    integer, intent(in) :: option
    real(rt), optional, intent(in) :: u01
    integer :: izp
    ! local
    real(rt) :: randval
    real(rt), allocatable :: Ftmp(:)
    
    if (present(u01)) then
        randval = u01
    else
        call random_number(randval)
    end if

    if (option==0) then
        ! Slow: temporary created
        allocate(Ftmp(size(F)))
        Ftmp = cumsum(F)
        do izp = 1,size(F)
            if (randval<=Ftmp(izp)) return
        end do


    else
        ! Fast
        do izp = 1,size(F)
            if (randval<=F(izp)) return
        end do

    end if
    
end function drawIid

! Same as drawIid(F,iz) but takes randval as input
function drawIidcore(F,randval) result(izp_out)
    use mod_matlab, only: cumsum,bsearch
    implicit none
    real(rt), dimension(:), intent(in) :: F 
    integer :: izp_out
    real(rt), intent(in) :: randval 
    ! local
    integer :: izp
    
    ! Find the izp such that randval \in (F(izp-1,iZ),F(izp,iZ)]
    ! If randval<F(1), return 1
    ! If randval>F(n), then issue an error
    ! If randval\in (F(1),F(2)], return 2 
    ! If randval\in (F(i),F(i+1)], return i+1
    ! Note: bsearch finds the i such that randval \in (F(i),F(i+1)), so need to add 1
    if (randval<=F(1)) then
        izp = 1
    elseif (randval>=F(size(F,1))) then
        izp = size(F,1)
        print*,'WARNING: drawAR1: F must not sum to 1, simulation will be wrong'
    else
        call bsearch(izp,randval,F)
        izp = izp + 1
    end if

    izp_out = izp

end function

! drawAR1core(F,iz,randval) result(izp)
! Same as drawAR1(F,iz,option=1) but takes randval as input
function drawAR1core(F,iz,randval) result(izp_out)
    use mod_matlab, only: cumsum,bsearch
    implicit none
    real(rt), dimension(1:,1:), intent(in) :: F ! Either the transition matrix or the cumulatively summed transpose 
    integer, intent(in) :: iz !, option
    integer :: izp_out
    real(rt), intent(in) :: randval 
    ! local
    integer :: izp
    
    ! Find the izp such that randval \in (F(izp-1,iZ),F(izp,iZ)]
    ! If randval<F(1), return 1
    ! If randval>F(n), then issue an error
    ! If randval\in (F(1),F(2)], return 2 
    ! If randval\in (F(i),F(i+1)], return i+1
    ! Note: bsearch finds the i such that randval \in (F(i),F(i+1)), so need to add 1
    if (randval<=F(1,iZ)) then
        izp = 1
    elseif (randval>=F(size(F,1),iZ)) then
        izp = size(F,1)
        print*,'WARNING: drawIidcore: F must not sum to 1, simulation will be wrong'
    else
        call bsearch(izp,randval,F(:,iZ))
        izp = izp + 1
    end if

    izp_out = izp

end function
! drawAR1(F,iz,option) result(izp)
! 
! Given iz, will generate izp according to a discrete markov chain.
! If option==0, the Markov chain is taken to be F.
! If option==1, the Markov chain is given implicitly and F is treated as the cumulatively
!               summed transpose of it (with the convention that the markov matrix has 
!               pi(i,j) with i being today and j being tomorrow. 
! 
function drawAR1(F,iz,option) result(izp_out)
    use mod_matlab, only: cumsum,bsearch
    implicit none
    real(rt), dimension(1:,1:), intent(in) :: F ! Either the transition matrix or the cumulatively summed transpose 
    integer, intent(in) :: iz, option
    integer :: izp_out
    ! local
!     integer :: izp
    real(rt) :: randval
!     real(rt), allocatable :: Ftmp(:)
    
    call random_number(randval)

    if (option==0) then
        izp_out = drawAR1core(cumsum(transpose(F)),iz,randval) 
    elseif (option==1) then
        izp_out = drawAR1core(F,iz,randval)
    end if
    
end function 

! Construct a Markov chain of specified length. Normally, F
! is assumed to be F_ij = F(z'=j|z=i). However, if opt is 
! present and equal to 1, then F is assumed to be the cumsum 
! of the transpose of the Markov matrix.
subroutine drawMarkov(iseq,F,i1,opt)
    use mod_matlab, only: cumsum
    implicit none
    integer, intent(inout) :: iseq(1:)
    real(rt), dimension(1:,1:), intent(in) :: F ! The transition matrix
    integer, intent(in) :: i1
    integer, intent(in), optional :: opt
    ! local
    integer :: t,nseq
    real(rt), allocatable :: Fcumtrans(:,:)
    logical :: isspecial

    nseq = size(iseq)

    if (present(opt)) then
        isspecial = opt==1
    else
        isspecial = .false.
    end if

    if (.not. isspecial) then
        allocate(Fcumtrans(size(F,1),size(F,2)))

        Fcumtrans = transpose(F)
        Fcumtrans = cumsum(Fcumtrans)

        iseq(1) = i1
        do t = 2,nseq
            iseq(t) = drawAR1(Fcumtrans,iseq(t-1),1)
        end do

        deallocate(Fcumtrans)
    else

        iseq(1) = i1
        do t = 2,nseq
            iseq(t) = drawAR1(F,iseq(t-1),1)
        end do

    end if

end subroutine drawMarkov

! Adda cooper algorithm. 
! Based on Martin Floden's 2005 code
! %[Z,PI] = addacooper(n,mu,rho,sigma)
! % Approximate n-state AR(1) process following Tauchen (1986) and Tauchen & Hussey (1991). 
! % See Adda & Cooper (2003) pp 57-.
! %
! % Z(t+1) = mu*(1-rho) + rho*Z(t) + eps(t+1)
! %
! % where  std(eps) = sigma
! %
! % Martin Floden, 2005
! 
subroutine sub_addacooper_grid(Z,mu,rho,sigma)
    use mod_matlab
    ! use mod_quad
    implicit none
    real(rt), intent(out) :: Z(:)
    real(rt), intent(in) :: mu,rho,sigma
    ! local
    integer :: i,n
    real(rt) :: sigmaUNC
    real(rt), allocatable :: E(:)

    n = size(Z)

    if (n==1) then
        Z = mu
        return
    elseif (n<=0) then
        STOP 'ERROR: sub_addacooper: n<=0'
    end if

    sigmaUNC = sigma/sqrt(1e0_rt-rho**2)

    allocate(E(n+1))
    E = 0e0_rt
    Z = 0e0_rt

    E(1)   = mu - 8e0_rt*sigmaUNC ! 10 unconditional standard deviations below mean
    E(n+1) = mu + 8e0_rt*sigmaUNC
    do i = 2,n
        E(i) = sigmaUNC*normcdfinv(real(i-1,rt)/real(n,rt),0e0_rt,1e0_rt) + mu
    end do 

    do i = 1,n
        Z(i) = n*sigmaUNC*(stdnormpdf((E(i)-mu)/sigmaUNC) - stdnormpdf((E(i+1)-mu)/sigmaUNC)) + mu
    end do

end subroutine
subroutine sub_addacooper(Z,Fzz,mu,rho,sigma)
    use mod_matlab
    use mod_quad
    implicit none
    real(rt), intent(out) :: Z(:),Fzz(:,:)
    real(rt), intent(in) :: mu,rho,sigma
    ! local

    integer :: i,j,n
    real(rt) :: sigmaUNC,E1,E2
    real(rt), allocatable :: E(:),MFPI(:,:),xquad(:),wquad(:)
    integer, parameter :: nquad = 1000

    allocate(xquad(nquad),wquad(nquad))


    n = size(Z)
    if (size(Fzz,1)/=n) STOP 'ERROR: sub_addacooper size mismatch'
    if (size(Fzz,2)/=n) STOP 'ERROR: sub_addacooper size mismatch'


    if (n==1) then
        Z = mu
        Fzz = 1e0_rt
        return
    elseif (n<=0) then
        STOP 'ERROR: sub_addacooper: n<=0'
    end if

    sigmaUNC = sigma/sqrt(1e0_rt-rho**2);

    allocate(E(n+1))
    allocate(MFPI(n,n))
    E = 0e0_rt
    Z = 0e0_rt
    Fzz = 0e0_rt
    MFPI = 0e0_rt

    E(1)   = mu - 8e0_rt*sigmaUNC ! 10 unconditional standard deviations below mean
    E(n+1) = mu + 8e0_rt*sigmaUNC
    do i = 2,n
        E(i) = sigmaUNC*normcdfinv(real(i-1,rt)/real(n,rt),0e0_rt,1e0_rt) + mu
    end do 

    do i = 1,n
        Z(i) = n*sigmaUNC*(stdnormpdf((E(i)-mu)/sigmaUNC) - stdnormpdf((E(i+1)-mu)/sigmaUNC)) + mu
    end do

    do i = 1,n
        do j = 1,n
            E1 = E(j)
            E2 = E(j+1)
            ! PI(i,j) = quadl(th_fcn,E(i),E(i+1),1e-10)
            call chebyquad(xquad,wquad,E(i),E(i+1))
            Fzz(i,j) = dot_product(th_fcn(xquad),wquad)
!             if (any(isnan(xquad))) STOP 'ERROR'
!             if (any(isnan(wquad))) STOP 'ERROR'
! 
!             if (isnan(Fzz(i,j))) STOP 'ERROR'
!             print*,'Fzz(i,j)',Fzz(i,j)
            MFPI(i,j) = stdnormcdf((E(j+1)-mu*(1e0_rt-rho)-rho*Z(i))/sigma) - stdnormcdf((E(j)-mu*(1e0_rt-rho)-rho*Z(i))/sigma)
        end do
    end do

    do i = 1,n
        Fzz(i,:) = Fzz(i,:) / sum(Fzz(i,:))
        MFPI(i,:) = MFPI(i,:) / sum(MFPI(i,:))
    end do

contains

    function th_fcn(u) result(f)
        implicit none
        real(rt), intent(in) :: u(:)
        real(rt) :: f(size(u))

        f = real(n,rt)/sqrt(2e0_rt*pi*sigmaUNC**2) * (exp(-(u-mu)**2 / (2e0_rt*sigmaUNC**2)) * &
            (stdnormcdf((E2-mu*(1e0_rt-rho)-rho*u)/sigma) - stdnormcdf((E1-mu*(1e0_rt-rho)-rho*u)/sigma)))

    end function

end subroutine
        


!Rouwenhorst method as presented in Kopecky and Suen RED 2010
! Based on Karen Kopecky's Code
subroutine sub_rouwenhorst(z,P,rho,sig,n)
    use mod_matlab
    implicit none
    real(rt), intent(in) :: rho,sig
    integer, intent(in) :: n
    real(rt), intent(out) :: z(n),P(n,n)
    !local
    real(rt) :: mu_eps, q, nu
    real(rt), allocatable :: P0(:,:),P1(:,:)
    integer :: i

    if (n<=0) then
        STOP 'sub_rouwenhorst: n must be > 0'
    elseif (n==1) then
        z = 0e0_rt
        P = 1e0_rt
        return
    end if
    
    mu_eps = 0.e0_rt
    q = (rho+1.e0_rt)/2.e0_rt
    nu = ((real(n,rt)-1.e0_rt)/(1.e0_rt-rho**2))**(.5e0_rt) * sig
    
    allocate(P0(2,2))
    P0(1,:) = (/q, 1.e0_rt-q/)
    P0(2,:) = (/1.e0_rt-q, q/)
    
    do i = 2,n-1
        allocate(P1(i+1,i+1))
        P1 = 0.e0_rt
        !        
        P1(1:i,1:i) = q*P0 + P1(1:i,1:i)
        !
        P1(1:i,2:i+1) = (1.e0_rt-q)*P0 + P1(1:i,2:i+1)
        !
        P1(2:i+1,1:i) = (1.e0_rt-q)*P0 + P1(2:i+1,1:i)
        !
        P1(2:i+1,2:i+1) = q*P0 + P1(2:i+1,2:i+1)
        !
        P1(2:i,:) = P1(2:i,:)/2.e0_rt
        deallocate(P0)
        allocate(P0(i+1,i+1))
        P0 = P1
        deallocate(P1)
    end do
    
    P = P0
    deallocate(P0)
    
    z = linspace(mu_eps/(1.e0_rt-rho) - nu, mu_eps/(1.e0_rt-rho) + nu,n)
    

end subroutine sub_rouwenhorst

! subtauchen
! Generates Markov chain approximating an AR(1) function of the form
! y=ar1*y_{-1}+e,  e~N(0,sd_eps**2)
! ar1  is AR(1) coeff
! sd_eps is SD of normally distributed shock
! cover gives cover*SD coverage, e.g. 3 = 99% coverage, 2 = 95%
!
! Parts of this code may be motivated by Floden
subroutine sub_tauchen(y,transition,ar1,sd_eps,n,cover)
    use mod_matlab, only: linspace
    implicit none
    !Input/Output
    real(rt), intent(in)                      :: ar1        !ar(1) coeff
    real(rt), intent(in)                      :: sd_eps     !SD of normally distributed shocks
    real(rt), intent(in)                      :: cover      !number of SDs to cover (3=>99%)  
    integer,  intent(in)                      :: n          !no of tauchen points
    real(rt), dimension(:), intent(out)     :: y          !shock values
    real(rt), dimension(:,:), intent(out) :: transition !transition matrix
    ! local
    real(rt) :: sd_y            !Unconditional stdev of y

    if (n==1) then
        y = 0e0_rt
        transition = 1e0_rt
        return
    end if

    !Define discrete states of markov chain
    sd_y = sd_eps/sqrt(1-ar1**2)
    y = linspace(-cover*sd_y,cover*sd_y,n)

    ! Call tauchen with predetermined grid
    call sub_tauchen_fixedgrid(transition,y,ar1,sd_eps)

end subroutine sub_tauchen

! Iid shock version of tauchen, just call sub_tauchen and takes a row of the trans
subroutine sub_tauchen_iid(y,prob,sd_eps,n,cover)
    implicit none
    !Input/Output
    real(rt), intent(in)                  :: sd_eps     !SD of normally distributed shocks
    real(rt), intent(in)                  :: cover      !number of SDs to cover (3=>99%)  
    integer,  intent(in)                 :: n          !no of tauchen points
    real(rt), dimension(1:n), intent(out) :: y          !shock values
    real(rt), dimension(1:n), intent(out) :: prob !transition matrix
    ! local
    real(rt), dimension(:,:), allocatable :: transition

    allocate(transition(n,n))
    call sub_tauchen(y,transition,0e0_rt,sd_eps,n,cover)
    prob = transition(1,:)
    deallocate(transition)
    
end subroutine

subroutine sub_tauchen_iid_fixedgrid(prob,y,sd_eps)
    implicit none
    real(rt), intent(in) :: sd_eps
    real(rt), dimension(:), intent(in) :: y
    real(rt), dimension(size(y)), intent(out) :: prob !transition matrix
    ! local 
    real(rt), dimension(:,:), allocatable :: transition
    integer :: n

    n = size(y)

    allocate(transition(n,n))
    call sub_tauchen_fixedgrid(transition,y,0e0_rt,sd_eps) ! no autocorrelation
    prob = transition(1,:)
    deallocate(transition)

end subroutine 

subroutine sub_tauchen_iid_fixedgrid_with_mean(prob,y,sd_eps,mean)
    implicit none
    real(rt), intent(in) :: sd_eps,mean
    real(rt), dimension(:), intent(in) :: y
    real(rt), dimension(size(y)), intent(out) :: prob !transition matrix
    ! local 
    real(rt), dimension(:,:), allocatable :: transition
    integer :: n

    n = size(y)

    allocate(transition(n,n))
    call sub_tauchen_fixedgrid_with_mean(transition,y,0e0_rt,sd_eps,mean) ! no autocorrelation
    prob = transition(1,:)
    deallocate(transition)

end subroutine 

subroutine sub_tauchen_fixedgrid(transition,y,ar1,sd_eps)
    implicit none
    !Input/Output
    real(rt), intent(in)                      :: ar1        !ar(1) coeff
    real(rt), intent(in)                      :: sd_eps     !SD of normally distributed shocks
    real(rt), dimension(1:), intent(in)      :: y          !shock values
    real(rt), dimension(1:size(y),1:size(y)), intent(out) :: transition !transition matrix

    call sub_tauchen_fixedgrid_with_mean(transition,y,ar1,sd_eps,0e0_rt)

end subroutine sub_tauchen_fixedgrid

! Approximates 
! y = mean + ar1*y_{-1} + sd_eps*eps, eps~N(0,1)
! using tauchen on the fixed grid y.  
! NOTE: mean is not the unconditional mean of y, but the conditional mean of y|(y_{-1}=0)
! 
subroutine sub_tauchen_fixedgrid_with_mean(transition,y,ar1,sd_eps,mean)
    implicit none
    !Input/Output
    real(rt), intent(in) :: ar1, sd_eps, mean
    real(rt), dimension(1:), intent(in) :: y !shock values
    real(rt), dimension(1:size(y),1:size(y)), intent(out) :: transition !transition matrix
    !Variable Declaration
    integer :: n ! size of y
    integer :: y_ind            !shock
    integer :: yp_ind           !shock next period
    real(rt) :: mean_cond_on_y  !mean of y' cdnal on y
    real(rt) :: normcdfhigh     !cdf at ub of interval
    real(rt) :: normcdflow      !cdf at lb of interval
    real(rt), dimension(1:size(y)) :: betweenpts_high !interval ub
    real(rt), dimension(1:size(y)) :: betweenpts_low  !interval lb

    n = size(y)

    if (n==1) then
        transition = 1e0_rt
        return
    end if

    !Define intervals for tauchen points
    do y_ind = 1, n-1
        betweenpts_high(y_ind)=y(y_ind)+(y(y_ind+1) - y(y_ind))/2e0_rt        
    enddo

    do y_ind = 2, n
        betweenpts_low(y_ind)=y(y_ind)-(y(y_ind) - y(y_ind-1))/2e0_rt        
    enddo

    !Calculate transition probabilities
    do y_ind = 1,n
        mean_cond_on_y = mean + ar1*y(y_ind) 
        
        normcdfhigh=.5e0_rt*(1.0e0_rt+erf((betweenpts_high(1)-mean_cond_on_y)/(sd_eps*sqrt(2e0_rt))))
        normcdflow=0.e0_rt
        transition(y_ind, 1) = normcdfhigh-normcdflow
        
        normcdfhigh=1.e0_rt
        normcdflow=.5e0_rt*(1.0e0_rt+erf((betweenpts_low(n)-mean_cond_on_y)/(sd_eps*sqrt(2e0_rt))))
        transition(y_ind, n) = normcdfhigh-normcdflow
        
        do yp_ind = 2,n-1
            normcdfhigh=.5e0_rt*(1.0e0_rt+erf((betweenpts_high(yp_ind)-mean_cond_on_y)/(sd_eps*sqrt(2e0_rt))))
            normcdflow=.5e0_rt*(1.0e0_rt+erf((betweenpts_low(yp_ind)-mean_cond_on_y)/(sd_eps*sqrt(2e0_rt))))
            transition(y_ind,yp_ind) = (normcdfhigh - normcdflow)            
        enddo
    enddo

end subroutine sub_tauchen_fixedgrid_with_mean

end module mod_discreteAR1

module mod_discreteAR1_test
    use mod_discreteAR1
    implicit none
contains
    subroutine test_sub_regime_switch_2state_tauchen()
        use mod_matlab
        implicit none
        real(8) :: mu(2),rho(2),nu(2),Pdiag(2)
        integer :: nR(2)
        real(8), allocatable :: z(:),Fzz(:,:),invarz(:)
        integer, parameter :: nZ = 100

        allocate(z(nZ),Fzz(nZ,nZ),invarz(nZ))

        rho = [.4409d0,.7648d0]
        nu = [.001d0,.024d0]
        Pdiag = [.952d0,.635d0]
        mu = [.0497d0,-.0093d0]/(1d0-rho)
        nR = [nZ/2,nZ-nZ/2]

        call sub_regime_switch_2state_tauchen(z,Fzz,mu,rho,nu,Pdiag,nR)

        invarz = nullspace(eye(nZ)-transpose(Fzz))
        invarz = invarz/sum(invarz)

        call disp('z',z)
        call disp('Fzz',Fzz)
        call disp('invarz',invarz)

        print*,'mean(z)',dot_product(invarz,z)
        print*,'var(z)',dot_product(invarz,z**2)-dot_product(invarz,z)**2
        print*,'std(z)',sqrt(dot_product(invarz,z**2)-dot_product(invarz,z)**2)

    end subroutine
end module

! A module for maximizing the Bellman equation one way or another
!
! 1) one-dimensional maximization routine 
! 2) two-dimensional maximization routine
module mod_valueopt
    use mod_types
    use mod_spline
    implicit none

    real(rt), parameter, private :: deathv = -sqrt(huge(0._rt)), deathc = -1e0_rt, deathap = -1e0_rt
    integer, parameter, private :: deathapi = -1 ! must be <= 0

    private :: sub_bellman_continuous_core

    interface sub_bellman_discrete_choose2
        module procedure sub_bellman_discrete_choose2_1d
    end interface sub_bellman_discrete_choose2

    interface sub_bellman_interp_choose2
        module procedure sub_bellman_interp_choose2_1d
    end interface sub_bellman_interp_choose2

    interface sub_bellman_discrete
        module procedure sub_bellman_discrete_1d, &
                         sub_bellman_discrete_2d, &
                         sub_bellman_discrete_3d, &
                         sub_bellman_discrete_4d
    end interface sub_bellman_discrete

    interface sub_bellman_continuous
        module procedure sub_bellman_continuous_2d, &
                         sub_bellman_continuous_3d, &
                         sub_bellman_continuous_4d
    end interface sub_bellman_continuous

    interface sub_bellman_interp
        module procedure sub_bellman_interp_2d, &
                         sub_bellman_interp_3d, &
                         sub_bellman_interp_4d
    end interface sub_bellman_interp


    interface checkShape
        module procedure checkShapeReal,checkShapeInt,&
                         checkShapeReal2,checkShapeInt2,&
                         checkShapeReal3,checkShapeInt3,&
                         checkShapeReal4,checkShapeInt4
    end interface


    abstract interface
        function PUcustom_template(c) result(PU)
            use mod_types
            implicit none
            real(rt), intent(in) :: c
            real(rt) :: PU
        end function
    end interface

    procedure(PUcustom_template), pointer, private :: PUcustom 

    ! Don't point to null, weird language issue: http://software.intel.com/en-us/forums/topic/329011
    type(spline), private, save :: negV_pp
    real(rt), dimension(:), pointer, private :: negV_x, negV_W, negV_gAp
    real(rt), private :: negV_y, negV_crra
    logical, private :: negV_isSchumaker

    !$OMP THREADPRIVATE(negV_pp,negV_x,negV_W,negV_gAp,negV_y,negV_crra,negV_isSchumaker)

contains

    ! Solves V(i) = max_{1<=i'<=nip} u(i,i') + W(i') for 1<=i<=ni where u and W are functions taking 
    ! integer arguments. The resulting policy function is called g(i). ni is inferred from the size of V, but
    ! nip must be passed as an argument. 
    !
    ! The function u should return u(i,i') as well as a flag saying whether the choice is feasible (0) or not.
    ! The function W should return W(i') as well as a flag saying whether the choice is feasible or not.
    ! Both functions should take assumed-size arrays realargs and intargs. These arrays are never accessed by
    ! bellman and may be size zero. The point of these arrays is to pass additional arguments to u and W in a 
    ! convenient way. Note that realargs and intargs may be any rank or size when bellman is called.
    !
    ! The routine uses speedups, based on what the user selects via opt.
    ! 
    ! Monotonicity:
    ! opt(1) | meaning
    ! 2      | binary monotonicity
    !
    ! Concavity:
    ! opt(2) | meaning
    ! 0      | none
    ! 
    subroutine bellman(V,g,u,W,nip,opt,realargs,intargs)
        implicit none
        real(rt), intent(out) :: V(:)
        integer, intent(out) :: g(:)
        integer, intent(in) :: nip
        integer, intent(in) :: opt(:)
        real(rt), intent(inout) :: realargs(*)
        integer, intent(inout) :: intargs(*)
        interface
            subroutine u(pu,i,ip,flag,realargs,intargs)
                use mod_types
                implicit none
                real(rt), intent(out) :: pu
                integer, intent(in) :: i,ip
                integer, intent(out) :: flag
                real(rt), intent(inout) :: realargs(*)
                integer, intent(inout) :: intargs(*)
            end subroutine
            subroutine W(cv,ip,flag,realargs,intargs)
                use mod_types
                implicit none
                real(rt), intent(out) :: cv
                integer, intent(in) :: ip
                integer, intent(out) :: flag
                real(rt), intent(inout) :: realargs(*)
                integer, intent(inout) :: intargs(*)
            end subroutine
        end interface
        ! local
        integer :: ni

        if (size(opt)<2) STOP 'ERROR: bellman: opt must be at least length 2'
        if (any(opt(1:2)/=[2,0])) STOP 'ERROR: bellman: requested optimization not programmed' 

        ni = size(V)

        ! Compute the optimum at the endpoints
        call brute(1,1,nip) ! computes V(1),g(1)
        call brute(ni,g(1),nip) ! computes V(ni),g(ni) (assumes mono since lowerbound is g(1)

!         print*,1,V(1),g(1)
!         print*,ni,V(ni),g(ni)

        ! Uses binary monotonicity
        call binarymono(1,ni)

    contains

        ! Computes V(i) and g(i) for all i between ia and ib assuming g(ia) and g(ib) are known
        recursive subroutine binarymono(ia,ib)
            implicit none
            integer, intent(in) :: ia,ib
            ! local
            real(rt) :: putmp,cvtmp
            integer :: i,ip,im,flag


            if (ib<=ia+1) then
                ! Since g(ia) and g(ib) are assumed known, desired values are known
                return
            elseif (g(ia)==g(ib)) then

                ! Now know the entire policy, so fill out V and g
                ip = g(ia)

                call W(cvtmp,ip,flag,realargs,intargs)
                do i = ia+1,ib-1

                    call u(putmp,i,ip,flag,realargs,intargs)
                    V(i) = putmp + cvtmp
                    g(i) = ip

                end do
                
            else

                ! Evaluate the midpoint
                im = (ia+ib)/2
                call brute(im,g(ia),g(ib)) ! computes V(im),g(im)
!                 print*,im,V(im),g(im)

                ! Divide and conquer
                call binarymono(ia,im)
                call binarymono(im,ib)

            end if

        end subroutine

        ! Computes V(i),g(i) by doing a brute force search from ipa to ipb
        subroutine brute(i,ipa,ipb)
            implicit none
            integer, intent(in) :: i,ipa,ipb
            ! local
            real(rt) :: Vtmp,putmp,cvtmp
            integer :: ip,flag

            

            V(i) = -huge(0e0_rt)
            g(i) = 1
            do ip = ipa,ipb
                call u(putmp,i,ip,flag,realargs,intargs)
                call W(cvtmp,ip,flag,realargs,intargs)
                Vtmp = putmp + cvtmp
                if (Vtmp>V(i)) then
                    V(i) = Vtmp
                    g(i) = ip
                end if
            end do

        end subroutine

    end subroutine

    subroutine setPUcustom(PU)
        implicit none
        procedure(PUcustom_template) :: PU

        PUcustom => PU

    end subroutine

    ! Check that a function W(a',e) defined as the linear interpolant of 
    ! W at knots gAp is concave for each e. To check that it is globally 
    ! concave may computationally intensive because for each interval [a_i,a_j]
    ! all intermediate points k s.t. i<k<j must be checked. Instead, this routine
    ! just checks a random point and asks whether it lies above all the
    ! convex combinations. This can still be slow, but is much quicker than 
    ! checking all the combinations.
    function randCheckLinInterpConcave(W,gAP,lenAp,lenE) result(isconcave)
        use mod_matlab
        implicit none
        logical :: isconcave
        integer, intent(in) :: lenAp,lenE
        real(rt), intent(in) :: W(lenAp,*)
        real(rt), intent(in) :: gAP(lenAp)
        ! local
        real(rt) :: theta, aprand, Wrand
        integer :: iE,i,j, ilo

        ! Get the random value
        call random_number(aprand)
        aprand = gAP(1) + aprand*(gAP(lenAP)-gAP(1))

        call bsearch(ilo,aprand,gAP)

        ! Check it
        do iE = 1,lenE

            Wrand = interp1(gAP,W(:,iE),aprand)

            do i = 1,ilo
                do j = ilo+1,lenAp
                    ! The convex combination of X(i) and X(j) must lie below X(k)
                    theta = (aprand-gAp(i))/(gAp(j)-gAp(i))
                    if (Wrand<(1e0_rt-theta)*W(i,iE) + theta*W(j,iE) - 2e0_rt*abs(Wrand)*epsilon(0e0_rt)) then
                        ! not concave
                        print*,'randCheckLinInterpConcave: W is not concave'
                        print*,'aprand,Wrand: ',aprand,Wrand
                        print*,'gAP(i),W(i): ',gAP(i),W(i,iE)
                        print*,'gAP(j),W(j): ',gAP(j),W(j,iE)
                        print*,'Convex combination: ',(1e0_rt-theta)*W(i,iE) + theta*W(j,iE)
                        isconcave = .false.                            
                        return
                    end if
                end do
            end do
        end do
        isconcave = .true.

    end function 

    ! Check that a function W(a',e) defined as the linear interpolant of 
    ! W at knots gAp is concave for each e. To check that it is globally 
    ! concave may computationally intensive because for each interval [a_i,a_j]
    ! all intermediate points k s.t. i<k<j must be checked
    function checkLinInterpConcave(W,gAP,lenAp,lenE) result(isconcave)
        implicit none
        logical :: isconcave
        integer, intent(in) :: lenAp,lenE
        real(rt), intent(in) :: W(lenAp,*)
        real(rt), intent(in) :: gAP(lenAp)
        ! local
        real(rt) :: theta
        integer :: iE,i,j,k

        do iE = 1,lenE
            do i = 1,lenAp
                do j = lenAp,i+1,-1
                    do k = i+1,j-1
                        ! The convex combination of X(i) and X(j) must lie below X(k)
                        theta = (gAp(k)-gAp(i))/(gAp(j)-gAp(i))
                        if (W(k,iE)<(1e0_rt-theta)*W(i,iE) + theta*W(j,iE)) then
                            ! not concave
                            isconcave = .false.                            
                            return
                        end if
                    end do
                end do
            end do
        end do
        isconcave = .true.

    end function 

!     ! Both sub_bellman_discrete and sub_bellman_continuous solve
!     ! V(a,e) = max_{a'} u(y(a,e) - x(a',e)) + W(a',e)
!     ! Here, y(a,e) is income and x(a',e) could be q(a',e)a' as in CCNR
!     ! or just a' as in a RBC model. Options may be specified to control
!     ! various speedups and optimization options. 
!     ! 
!     ! Note: code is written to handle any problem with two or more dimensions
!     ! in the sense that any dimensions, other than the first, are treated as 
!     ! a single dimension. 
!     !
!     ! The code is threaded.
!     ! 
!     subroutine sub_policyiter_continuous(V,c,ap,W,y,x,gAP,lenA,lenE,lenAp,crra,opt,xopt)
!         use mod_matlab, only: bsearch, disp, interp1q
!         implicit none
!         integer, intent(in) :: lenA,lenE,lenAp
!         real(rt), dimension(lenA,*), intent(out) :: V
!         real(rt), dimension(lenA,*), intent(in) :: c,ap
!         real(rt), dimension(lenA,*), intent(in) :: y
!         real(rt), dimension(lenAp,*), intent(in) :: W,x
!         real(rt), dimension(lenAp), intent(in) :: gAP ! grid for AP
!         real(rt), intent(in) :: crra
!         integer, intent(in) :: opt(1:)
!         real(rt), intent(in) :: xopt(1:)
!         ! local
!         real(rt) :: tmp,step,work(3)
!         integer :: iA,iE,iwork(1)
! 
!         step = real(lenAp-1,8)/(gAp(lenAp)-gAp(1))
! 
!         STOP ' I need to update policiter_continuous if I want to use it '
! 
!         !$OMP PARALLEL DO DEFAULT(SHARED) PRIVATE(iA,iE,tmp,work,iwork)
!         do iE = 1,lenE
!             do iA = 1,lenA
!                 call getH(V(iA,iE),tmp,ap(iA,iE),iA,iE,W,y,x,gAP,step,crra,lenA,lenAp,work,iwork,opt)
!             end do
!         end do
!         !$OMP END PARALLEL DO
! 
!     end subroutine sub_policyiter_continuous

    ! Continuous case: linear interpolation

    ! Model options, governed by opt(1)
    ! 0         ERROR         
    ! 1         x(a',e) is increasing in a'
    !
    ! Speedup options, governed by opt(2)
    ! 0         ERROR
    ! 1         ERROR
    ! 2         Monotonicity and concavity assumption
    ! 
    ! Speedup options, governed by opt(3): NOTE: this is not currently used
    ! 0         Arbitrarily spaced gAP grid
    ! 1         Linearly spaced gAP grid
    ! 
    ! Interpolation options, governed by opt(4)
    ! 0         Linear interpolation
    ! 1         Schumaker shape-preserving spline interpolation
    ! 
    ! Tolerance, governed by xopt(1)
    ! xopt(1)   Termination critera, stop when bracket length<xopt(1)
    ! 
    subroutine sub_bellman_continuous_4d(V,c,ap,W,y,x,gAP,crra,opt,xopt)
        real(rt), dimension(:,:,:,:), intent(out) :: V,c
        real(rt), dimension(:,:,:,:), intent(out) :: ap
        real(rt), dimension(:,:,:,:), intent(in) :: y,W,x
        real(rt), dimension(:), intent(in) :: gAP ! grid for AP
        real(rt), intent(in) :: crra
        integer, intent(in) :: opt(1:)
        real(rt), intent(in) :: xopt(1:)
        ! local
        integer :: lenA,lenE1,lenE2,lenE3,lenAp

        lenA = size(V,1)
        lenE1 = size(V,2)
        lenE2 = size(V,3)
        lenE3 = size(V,4)
        lenAp = size(W,1)

        call checkShape(V,lenA,lenE1,lenE2,lenE3)
        call checkShape(c,lenA,lenE1,lenE2,lenE3)
        call checkShape(y,lenA,lenE1,lenE2,lenE3)
        call checkShape(x,lenAp,lenE1,lenE2,lenE3)
        call checkShape(W,lenAp,lenE1,lenE2,lenE3)
        call checkShape(ap,lenA,lenE1,lenE2,lenE3)
        
        call sub_bellman_continuous_core(V,c,ap,W,y,x,gAP,lenA,lenE1*lenE2*lenE3,lenAp,crra,opt,xopt)

    end subroutine 

    subroutine sub_bellman_continuous_3d(V,c,ap,W,y,x,gAP,crra,opt,xopt)
        real(rt), dimension(:,:,:), intent(out) :: V,c
        real(rt), dimension(:,:,:), intent(out) :: ap
        real(rt), dimension(:,:,:), intent(in) :: y,W,x
        real(rt), dimension(:), intent(in) :: gAP ! grid for AP
        real(rt), intent(in) :: crra
        integer, intent(in) :: opt(1:)
        real(rt), intent(in) :: xopt(1:)
        ! local
        integer :: lenA,lenE1,lenE2,lenAp

        lenA = size(V,1)
        lenE1 = size(V,2)
        lenE2 = size(V,3)
        lenAp = size(W,1)

        call checkShape(V,lenA,lenE1,lenE2)
        call checkShape(c,lenA,lenE1,lenE2)
        call checkShape(y,lenA,lenE1,lenE2)
        call checkShape(x,lenAp,lenE1,lenE2)
        call checkShape(W,lenAp,lenE1,lenE2)
        call checkShape(ap,lenA,lenE1,lenE2)
        
        call sub_bellman_continuous_core(V,c,ap,W,y,x,gAP,lenA,lenE1*lenE2,lenAp,crra,opt,xopt)

    end subroutine 

    subroutine sub_bellman_continuous_2d(V,c,ap,W,y,x,gAP,crra,opt,xopt)
        real(rt), dimension(:,:), intent(out) :: V,c
        real(rt), dimension(:,:), intent(out) :: ap
        real(rt), dimension(:,:), intent(in) :: y,W,x
        real(rt), dimension(:), intent(in) :: gAP ! grid for AP
        real(rt), intent(in) :: crra
        integer, intent(in) :: opt(1:)
        real(rt), intent(in) :: xopt(1:)
        ! local
        integer :: lenA,lenE,lenAp

        lenA = size(V,1)
        lenE = size(V,2)
        lenAp = size(W,1)

        call checkShape(V,lenA,lenE)
        call checkShape(c,lenA,lenE)
        call checkShape(y,lenA,lenE)
        call checkShape(x,lenAp,lenE)
        call checkShape(W,lenAp,lenE)
        call checkShape(ap,lenA,lenE)
        
        call sub_bellman_continuous_core(V,c,ap,W,y,x,gAP,lenA,lenE,lenAp,crra,opt,xopt)

    end subroutine 

    subroutine sub_bellman_continuous_core(V,c,ap,W,y,x,gAP,lenA,lenE,lenAp,crra,opt,xopt)
        use mod_matlab, only: bsearch, disp, interp1q
        use mod_interp, only: isearch
        use mod_spline
        use mod_min1dim
        implicit none
        integer, intent(in) :: lenA,lenE,lenAp
        real(rt), dimension(lenA,*), intent(out) :: V,c
        real(rt), dimension(lenA,*), intent(out) :: ap
        real(rt), dimension(lenA,*), intent(in) :: y
        real(rt), dimension(lenAp,*), intent(in), target :: W,x
        real(rt), dimension(lenAp), intent(in), target :: gAP ! grid for AP
        real(rt), intent(in) :: crra
        integer, intent(in) :: opt(1:)
        real(rt), intent(in) :: xopt(1:)
        ! local 
        integer, allocatable :: api(:,:)
        integer :: iA,iE,iAp_lb,iAp_ub,iAp
        real(rt) :: apmin,apmax
        logical :: isSchumaker


        if (size(xopt)<1) STOP 'sub_bellman_continuous: xopt must have length at least 1'
        if (xopt(1)<=2e0_rt*epsilon(0e0_rt)) STOP 'sub_bellman_continuous: xopt is set to small, should be >=eps'
        if (size(opt)<4) STOP 'sub_bellman_continuous: opt must have length at least 4'
        if (opt(2)/=2) STOP 'sub_bellman_continuous: concavity must be assumed'
        if (opt(1)/=1) STOP 'sub_bellman_continuous: x must be increasing in a'

        isSchumaker = opt(4)==1

        allocate(api(lenA,lenE))

        ! First, find the best value on the grid. Then, since objective function is concave,
        ! the optimum is bracketed in an interval + or - 1 index from the optimum. Since 
        ! concavity is assumed and it is much cheaper to check values on the grid than off
        ! the grid (since no interpolation is involved), this is extremely quick.
        call sub_bellman_discrete_core(V,c,api,W,y,x,lenA,lenE,lenAp,crra,[opt(1:2),1])

        ! api now has the optimum. 


!!! I THOUGHT THE OMP CODE SHOULD WORK, BUT SUPNORM IS SLIGHTLY DIFFERENT. AVOID UNTIL 
!!! HAVE TIME TO PROPERLY TEST IT. 
        !$OMP PARALLEL DO DEFAULT(SHARED) PRIVATE(iA,iE,iAp_lb,iAp_ub,iAp,apmin,apmax)
        do iE = 1,lenE

            ! Compute a spline of the continuation value if doing schumaker
            if (isSchumaker) then
                call schumaker(negV_pp,gAp,W(:,iE))
            end if

            ! Let negV know what isSchumaker is
            negV_isSchumaker = isSchumaker

            do iA = 1,lenA

                ! Note: need to determine feasible region, o/w the flat region created
                ! by deathv and result in multiple local minima.

                ! First, need to find a bracket that is feasible at both end points
                ! Since x is increasing in a', check that iAp=1 is feasible. O/w, nothing is.
                if (y(iA,iE)<x(1,iE)) then
                    ! Consumption is always negative
                    V(iA,iE) = deathv
                    c(iA,iE) = deathc
                    ap(iA,iE) = deathap
                    print*,'WARNING: no feasible choice found'
                    cycle
                else
                    apmin = gAP(1)
                end if

                ! Now, have to find a' such that c(a') = y-x(a') = 0 
                ! Begin by iterating from the top down and stopping if c>0. Since x is increasing
                ! in a', then c will be positive from gAP(1) to a'
                !!! Note: need to find iAp such that c(iAp)>=0e0_rt and c(iAp+1)<=0e0_rt if one exists. Equivalently,
                !!! find iAp such that y>=x(iAp) and y<=x(iAp+1). The routine isearch will find such an iAp, 
                !!! First though, check whether y>x(nAp) in which case all iAp are feasible.
                if (y(iA,iE)>x(lenAp,iE)) then
                    iAp = lenAp
                ! elseif (y(iA,iE)<x(1,iE)) then ! NOTE: this case has already been checked above
                else
                    call isearch(iAp,y(iA,iE),x(:,iE))
                end if
                
                if (iAp==lenAp) then
                    ! Entire choice set is feasible
                    apmax = gAP(lenAp)
                else
                    ! c > 0 at iAp but is <= at iAp+1
                    ! Now, need to find exact end point [aphi,gAp(iAp+1)]
                    ! y(iA,iE) = xval AND
                    ! xval = x(iAp,iE) + (x(iAp+1,iE)-x(iAp,iE))/(gAp(iAp+1)-gAp(iAp))*(ap-gAp(iAp))
                    ! <=> (y(iA,iE)-x(iAp,iE)) = (x(iAp+1,iE)-x(iAp,iE))/(gAp(iAp+1)-gAp(iAp))*(ap-gAp(iAp))
                    ! <=> (y(iA,iE)-x(iAp,iE))/(x(iAp+1,iE)-x(iAp,iE))*(gAp(iAp+1)-gAp(iAp)) = (ap-gAp(iAp))
                    ! <=> gAp(iAp) + (y(iA,iE)-x(iAp,iE))/(x(iAp+1,iE)-x(iAp,iE))*(gAp(iAp+1)-gAp(iAp)) = ap
                    apmax = gAp(iAp) + (y(iA,iE)-x(iAp,iE))/(x(iAp+1,iE)-x(iAp,iE))*(gAp(iAp+1)-gAp(iAp))
                end if

                ! Find the optimum using brent's method within the interval [lb,ub]
                iAp_lb = max(api(iA,iE)-1,1)
                iAp_ub = min(api(iA,iE)+1,lenAp)


                negV_crra = crra ! just a scalar, copy
                negV_y = y(iA,iE) ! "
                negV_gAp => gAp(iAp_lb:iAp_ub) ! variable length vectors, avoid copy
                negV_W => W(iAp_lb:iAp_ub,iE) ! "
                negV_x => x(iAp_lb:iAp_ub,iE) ! "

!                 print*,'negV_y',negV_y
!                 print*,'negV_W',negV_W
!                 print*,'negV_x',negV_x
!                 print*,'negV_gAp',negV_gAp
! 
!                 call negV_core(V(iA,iE),c(iA,iE),.5e0_rt)
!                 print*,'V',V(iA,iE)
!                 print*,'c',c(iA,iE)

                ! Use the minimization routine by brent. NOTE: only searching feasible region [apmin,apmax]
                apmin = max(gAp(iAp_lb),apmin)
                apmax = min(gAp(iAp_ub),apmax)
                call sub_brent_min(ap(iA,iE),negV,apmin,apmax,printmod_in=-1,xtoler_in=1e-8_rt)

                ! Evaluate last time at the optimum to get value and consumption policy
                call negV_core(V(iA,iE),c(iA,iE),ap(iA,iE))

            end do
        end do
        !$OMP END PARALLEL DO

!         STOP 'checking sub_brent_min'

    end subroutine sub_bellman_continuous_core

    ! Compute the negative of u(c)+W for ap 
    function negV(ap)
        implicit none
        real(rt), intent(in) :: ap
        real(rt) :: negV
        ! local
        real(rt) :: c,V

        call negV_core(V,c,ap)
        negV = -V

    end function 
    subroutine negV_core(V,c,ap) 
        use mod_interp, only: bsearch
        use mod_spline
        implicit none           
        real(rt), intent(in) :: ap
        real(rt), intent(out) :: V,c
        ! local
        integer :: ilo 
        real(rt) :: whi,xval,Wval
            
        ! Find the location of ap in the grid
        call bsearch(ilo,ap,negV_gAp)

        ! Interpolate to get x(a',e) and W(a',e)
        whi = (ap-negV_gAP(ilo))/(negV_gAP(ilo+1)-negV_gAP(ilo))
        xval = negV_x(ilo) + whi*(negV_x(ilo+1)-negV_x(ilo))

        if (negV_isSchumaker) then
            ! Evaluate the spline
            call ppval(Wval,ap,negV_pp)
        else
            Wval = negV_W(ilo) + whi*(negV_W(ilo+1)-negV_W(ilo))
        end if
! 
! print*,'ilo',ilo
! print*,'xval',xval
! print*,'Wval',Wval

        ! Evaluate consumption and utility
        c = negV_y - xval
! print*,'c',c
        if (c>0e0_rt) then
            if (negV_crra==2e0_rt) then
                V = -1e0_rt/c + Wval
            elseif (negV_crra==1e0_rt) then
                V = log(c) + Wval
            else
                V = c**(1e0_rt-negV_crra)/(1e0_rt-negV_crra) + Wval
            end if
        else
            V = deathv
            c = deathc
        end if
! print*,'V',V

    end subroutine  

!     subroutine getH(H,c,ap,iA,iE,W,y,x,gAP,step,crra,lenA,lenAp,work,iwork,opt) 
!         use mod_interp, only: bsearch,isearch_with_step
!         implicit none           
!         real(rt), intent(out) :: H,c
!         real(rt), intent(in) :: ap,step,crra
!         integer, intent(in) :: iA,iE
!         integer, intent(in) :: lenA,lenAp
!         real(rt), dimension(lenA,*), intent(in) :: y
!         real(rt), dimension(lenAp,*), intent(in) :: W,x
!         real(rt), dimension(lenAp), intent(in) :: gAP 
!         real(rt), intent(out) :: work(3)
!         integer, intent(out) :: iwork(1)
!         integer, intent(in) :: opt(1:)
! 
! ! integer :: itmp
!         associate (whi=>work(1), xval=>work(2), Wval=>work(3), ilo=>iwork(1))
! 
!         ! Get position of ap in gAP
!         if (opt(3)==0) then
!             ! Arbitrarily spaced grid
!             call bsearch(ilo,ap,gAP)
!         else
!             ! Linearly spaced grid
!             ! if gAP is linearly spaced, will find it in one iteration, 
!             ! but is robust to non-linear spacing.
!             call isearch_with_step(ilo,ap,gAP,step) 
!         end if
! 
!         ! Interpolate to get x(a',e) and W(a',e)
!         whi = (ap-gAP(ilo))/(gAP(ilo+1)-gAP(ilo))
!         xval = x(ilo,iE) + whi*(x(ilo+1,iE)-x(ilo,iE))
!         Wval = W(ilo,iE) + whi*(W(ilo+1,iE)-W(ilo,iE))
! 
!         ! Evaluate consumption and utility
!         c = y(iA,iE) - xval
!         if (c>0e0_rt) then
!             if (crra==2e0_rt) then
!                 H = -1e0_rt/c + Wval
!             elseif (crra==1e0_rt) then
!                 H = log(c) + Wval
!             else
!                 H = c**(1e0_rt-crra)/(1e0_rt-crra) + Wval
!             end if
!         else
!             H = deathv
!             c = deathc
!         end if
! 
!         end associate
! 
!     end subroutine  


    ! Grid search with interpolation
    ! 
    ! Identical to sub_bellman_discrete except that it first interpolates W and x
    ! from gAP onto gAPcheck and optimizes over the gAPfine grid. 
    ! Interpolation can be done via either linear interpolation or Schumaker shape-preserving
    ! splines. 
    ! 
    ! Interpolation options, governed by opt(1)
    ! 0         Linear interpolation
    ! XXX       Schumaker spline interpolation for W, linear for X (not implemented yet)
    ! XXX       Schumaker for both W and X (not implemented yet)
    !    
    ! Model options, governed by opt(2)
    ! 0         x is general form 
    ! 1         x is increasing in a' for each e (speedup available)
    !
    ! Speedup options, governed by opt(3)
    ! 0         Brute force grid search
    ! 1         Monotonicity assumption a'(a,e) monotone in a
    ! 2         Monotonicity and concavity assumption
    !
    ! Parallel options, governed by opt(4)
    ! 0         No parallelization
    ! 1         Parallelized over the e values using openmp
    ! 
    ! NOTE: gAP is assumed to be increasing so fast interpolation can be done
    ! NOTE: the concavity assumption is exploited using binary search (Heer and Maussner)
    ! NOTE: if concavity is assumed, then x must be increasing in a'
    subroutine sub_bellman_interp_4d(V,c,ap,W,y,x,gAP,gAPfine,crra,opt)
        implicit none
        real(rt), dimension(:,:,:,:), intent(out) :: V,c,ap
        real(rt), dimension(:,:,:,:), intent(in) :: y,W,x
        real(rt), dimension(:), intent(in) :: gAP,gAPfine
        real(rt), intent(in) :: crra
        integer, intent(in) :: opt(:)
        ! local
        integer :: lenA,lenE1,lenE2,lenE3,lenAp,lenApfine

        lenA = size(V,1)
        lenE1 = size(V,2)
        lenE2 = size(V,3)
        lenE3 = size(V,4)
        lenAp = size(W,1)
        lenApfine = size(gAPfine)

        call checkShape(V,lenA,lenE1,lenE2,lenE3)
        call checkShape(c,lenA,lenE1,lenE2,lenE3)
        call checkShape(y,lenA,lenE1,lenE2,lenE3)
        call checkShape(x,lenAp,lenE1,lenE2,lenE3)
        call checkShape(W,lenAp,lenE1,lenE2,lenE3)
        call checkShape(ap,lenA,lenE1,lenE2,lenE3)
        call checkShape(gAP,lenAp)
        call checkShape(opt,4)
        
        call sub_bellman_interp_core(V,c,ap,W,y,x,gAP,gAPfine,lenA,lenE1*lenE2*lenE3,lenAp,crra,opt)

    end subroutine 
    subroutine sub_bellman_interp_3d(V,c,ap,W,y,x,gAP,gAPfine,crra,opt)
        implicit none
        real(rt), dimension(:,:,:), intent(out) :: V,c,ap
        real(rt), dimension(:,:,:), intent(in) :: y,W,x
        real(rt), dimension(:), intent(in) :: gAP,gAPfine
        real(rt), intent(in) :: crra
        integer, intent(in) :: opt(:)
        ! local
        integer :: lenA,lenE1,lenE2,lenAp,lenApfine

        lenA = size(V,1)
        lenE1 = size(V,2)
        lenE2 = size(V,3)
        lenAp = size(W,1)
        lenApfine = size(gAPfine)

        call checkShape(V,lenA,lenE1,lenE2)
        call checkShape(c,lenA,lenE1,lenE2)
        call checkShape(y,lenA,lenE1,lenE2)
        call checkShape(x,lenAp,lenE1,lenE2)
        call checkShape(W,lenAp,lenE1,lenE2)
        call checkShape(ap,lenA,lenE1,lenE2)
        call checkShape(gAP,lenAp)
        call checkShape(opt,4)
        
        call sub_bellman_interp_core(V,c,ap,W,y,x,gAP,gAPfine,lenA,lenE1*lenE2,lenAp,crra,opt)

    end subroutine 
    subroutine sub_bellman_interp_2d(V,c,ap,W,y,x,gAP,gAPfine,crra,opt)
        implicit none
        real(rt), dimension(:,:), intent(out) :: V,c,ap
        real(rt), dimension(:,:), intent(in) :: y,W,x
        real(rt), dimension(:), intent(in) :: gAP,gAPfine
        real(rt), intent(in) :: crra
        integer, intent(in) :: opt(:)
        ! local
        integer :: lenA,lenE1,lenAp,lenApfine

        lenA = size(V,1)
        lenE1 = size(V,2)
        lenAp = size(W,1)
        lenApfine = size(gAPfine)

        call checkShape(V,lenA,lenE1)
        call checkShape(c,lenA,lenE1)
        call checkShape(y,lenA,lenE1)
        call checkShape(x,lenAp,lenE1)
        call checkShape(W,lenAp,lenE1)
        call checkShape(ap,lenA,lenE1)
        call checkShape(gAP,lenAp)
        call checkShape(opt,4)
        
        call sub_bellman_interp_core(V,c,ap,W,y,x,gAP,gAPfine,lenA,lenE1,lenAp,crra,opt)

    end subroutine 
    subroutine sub_bellman_interp_1d(V,c,ap,W,y,x,gAP,gAPfine,crra,opt)
        implicit none
        real(rt), dimension(:), intent(out) :: V,c,ap
        real(rt), dimension(:), intent(in) :: y,W,x
        real(rt), dimension(:), intent(in) :: gAP,gAPfine
        real(rt), intent(in) :: crra
        integer, intent(in) :: opt(:)
        ! local
        integer :: lenA,lenAp,lenApfine

        lenA = size(V,1)
        lenAp = size(W,1)
        lenApfine = size(gAPfine)

        call checkShape(V,lenA)
        call checkShape(c,lenA)
        call checkShape(y,lenA)
        call checkShape(x,lenAp)
        call checkShape(W,lenAp)
        call checkShape(ap,lenA)
        call checkShape(gAP,lenAp)
        call checkShape(opt,4)
        
        call sub_bellman_interp_core(V,c,ap,W,y,x,gAP,gAPfine,lenA,1,lenAp,crra,opt)

    end subroutine 

    subroutine sub_bellman_interp_core(V,c,ap,W,y,x,gAP,gAPfine,lenA,lenE,lenAp,crra,opt)
        use mod_matlab, only: bsearch, disp, interp1q
        use mod_interp, only: isearch
        use mod_spline
        use mod_min1dim
        implicit none
        integer, intent(in) :: lenA,lenE,lenAp
        real(rt), dimension(lenA,*), intent(out) :: V,c
        real(rt), dimension(lenA,*), intent(out) :: ap
        real(rt), dimension(lenA,*), intent(in) :: y
        real(rt), dimension(lenAp,*), intent(in) :: W,x
        real(rt), dimension(lenAp), intent(in) :: gAP ! grid for AP
        real(rt), dimension(:), intent(in) :: gAPfine ! interpolation grid
        real(rt), intent(in) :: crra
        integer, intent(in) :: opt(:)
        ! local 
        real(rt), allocatable :: Wfine(:,:), xfine(:,:)
        integer, allocatable :: api(:,:)
        integer :: lenAPfine,iE

        if (size(opt)<4) STOP 'sub_bellman_interp_core: opt must have length at least 5'

        lenAPfine = size(gAPfine)

        allocate(api(lenA,lenE))
        allocate(Wfine(lenAPfine,lenE))
        allocate(xfine(lenAPfine,lenE))

        Wfine = interp1q(gAP,W(:,1:lenE),gAPfine)
        xfine = interp1q(gAP,x(:,1:lenE),gAPfine)

        ! Call bellman discrete to do the actual optimization
        call sub_bellman_discrete_core(V,c,api,Wfine,y,xfine,lenA,lenE,lenAPfine,crra,opt(2:4))

        ! Recover the a' values
        do iE = 1,lenE
            where (c(:,iE)>0e0_rt) 
                ap(:,iE) = gAPfine(api(:,iE))
            elsewhere
                ap(:,iE) = deathap
            end where
        end do

    end subroutine sub_bellman_interp_core

    ! Discrete case: no interpolation
    !    
    ! Model options, governed by opt(1)
    ! 0         x is general form 
    ! 1         x is increasing in a' for each e (speedup available)
    !
    ! Speedup options, governed by opt(2)
    ! 0         Brute force grid search
    ! 1         Monotonicity assumption a'(a,e) monotone in a
    ! 2         Monotonicity and concavity assumption
    ! 
    ! Parallel options, governed by opt(3)
    ! 0         No parallelization
    ! 1         Parallelized over the e values using openmp
    !
    ! NOTE: the concavity assumption is exploited using binary search (Heer and Maussner)
    ! NOTE: if concavity is assumed, then x must be increasing in a'
    subroutine sub_bellman_discrete_4d(V,c,api,W,y,x,crra,opt)
        implicit none
        real(rt), dimension(:,:,:,:), intent(out) :: V,c
        integer, dimension(:,:,:,:), intent(out) :: api
        real(rt), dimension(:,:,:,:), intent(in) :: y,W,x
        real(rt), intent(in) :: crra
        integer, intent(in) :: opt(:)
        ! local
        integer :: lenA,lenE1,lenE2,lenE3,lenAp

        lenA = size(V,1)
        lenE1 = size(V,2)
        lenE2 = size(V,3)
        lenE3 = size(V,4)
        lenAp = size(W,1)

        call checkShape(V,lenA,lenE1,lenE2,lenE3)
        call checkShape(c,lenA,lenE1,lenE2,lenE3)
        call checkShape(y,lenA,lenE1,lenE2,lenE3)
        call checkShape(x,lenAp,lenE1,lenE2,lenE3)
        call checkShape(W,lenAp,lenE1,lenE2,lenE3)
        call checkShape(api,lenA,lenE1,lenE2,lenE3)
        
        call sub_bellman_discrete_core(V,c,api,W,y,x,lenA,lenE1*lenE2*lenE3,lenAp,crra,opt)

    end subroutine 
    subroutine sub_bellman_discrete_3d(V,c,api,W,y,x,crra,opt)
        implicit none
        real(rt), dimension(:,:,:), intent(out) :: V,c
        integer, dimension(:,:,:), intent(out) :: api
        real(rt), dimension(:,:,:), intent(in) :: y,W,x
        real(rt), intent(in) :: crra
        integer, intent(in) :: opt(:)
        ! local
        integer :: lenA,lenE1,lenE2,lenAp

        lenA = size(V,1)
        lenE1 = size(V,2)
        lenE2 = size(V,3)
        lenAp = size(W,1)

        call checkShape(V,lenA,lenE1,lenE2)
        call checkShape(c,lenA,lenE1,lenE2)
        call checkShape(y,lenA,lenE1,lenE2)
        call checkShape(x,lenAp,lenE1,lenE2)
        call checkShape(W,lenAp,lenE1,lenE2)
        call checkShape(api,lenA,lenE1,lenE2)
        
        call sub_bellman_discrete_core(V,c,api,W,y,x,lenA,lenE1*lenE2,lenAp,crra,opt)

    end subroutine 
    subroutine sub_bellman_discrete_2d(V,c,api,W,y,x,crra,opt)
        implicit none
        real(rt), dimension(:,:), intent(out) :: V,c
        integer, dimension(:,:), intent(out) :: api
        real(rt), dimension(:,:), intent(in) :: y,W,x
        real(rt), intent(in) :: crra
        integer, intent(in) :: opt(1:)
        ! local
        integer :: lenA,lenE,lenAp

        lenA = size(V,1)
        lenE = size(V,2)
        lenAp = size(W,1)

        call checkShape(V,lenA,lenE)
        call checkShape(c,lenA,lenE)
        call checkShape(y,lenA,lenE)
        call checkShape(x,lenAp,lenE)
        call checkShape(W,lenAp,lenE)
        call checkShape(api,lenA,lenE)

        call sub_bellman_discrete_core(V,c,api,W,y,x,lenA,lenE,lenAp,crra,opt)

    end subroutine 
    subroutine sub_bellman_discrete_1d(V,c,api,W,y,x,crra,opt)
        implicit none
        real(rt), dimension(:), intent(out) :: V,c
        integer, dimension(:), intent(out) :: api
        real(rt), dimension(:), intent(in) :: y,W,x
        real(rt), intent(in) :: crra
        integer, intent(in) :: opt(1:)
        ! local
        integer :: lenA,lenAp

        lenA = size(V,1)
        lenAp = size(W,1)

        call checkShape(V,lenA)
        call checkShape(c,lenA)
        call checkShape(y,lenA)
        call checkShape(x,lenAp)
        call checkShape(W,lenAp)
        call checkShape(api,lenA)

        call sub_bellman_discrete_core(V,c,api,W,y,x,lenA,1,lenAp,crra,opt)

    end subroutine 

!         subroutine checkShape(x,i,j)
!             implicit none
!             class(*), intent(in) :: x(:,:)
!             integer, intent(in) :: i,j
!             select type (a=>x)
!                 type is (real)
!                     if (any(shape(a)/=[i+1,j])) STOP 'ERROR: sub_bellman_discrete: incompatible dimensions'
!                 type is (integer)
!                     if (any(shape(a)/=[i+1,j])) STOP 'ERROR: sub_bellman_discrete: incompatible dimensions'
!             end select
!         end subroutine
    subroutine checkShapeReal4(x,i,j,k,l)
        implicit none
        real(rt), intent(in) :: x(:,:,:,:)
        integer, intent(in) :: i,j,k,l
        if (any(shape(x)/=[i,j,k,l])) STOP 'ERROR: bellman: incompatible dimensions'
    end subroutine
    subroutine checkShapeInt4(x,i,j,k,l)
        implicit none
        integer, intent(in) :: x(:,:,:,:)
        integer, intent(in) :: i,j,k,l
        if (any(shape(x)/=[i,j,k,l])) STOP 'ERROR: bellman: incompatible dimensions'
    end subroutine

    subroutine checkShapeReal3(x,i,j,k)
        implicit none
        real(rt), intent(in) :: x(:,:,:)
        integer, intent(in) :: i,j,k
        if (any(shape(x)/=[i,j,k])) STOP 'ERROR: bellman: incompatible dimensions'
    end subroutine
    subroutine checkShapeInt3(x,i,j,k)
        implicit none
        integer, intent(in) :: x(:,:,:)
        integer, intent(in) :: i,j,k
        if (any(shape(x)/=[i,j,k])) STOP 'ERROR: bellman: incompatible dimensions'
    end subroutine

    subroutine checkShapeReal2(x,i,j)
        implicit none
        real(rt), intent(in) :: x(:,:)
        integer, intent(in) :: i,j
        if (any(shape(x)/=[i,j])) STOP 'ERROR: bellman: incompatible dimensions'
    end subroutine
    subroutine checkShapeInt2(x,i,j)
        implicit none
        integer, intent(in) :: x(:,:)
        integer, intent(in) :: i,j
        if (any(shape(x)/=[i,j])) STOP 'ERROR: bellman: incompatible dimensions'
    end subroutine

    subroutine checkShapeReal(x,i)
        implicit none
        real(rt), intent(in) :: x(:)
        integer, intent(in) :: i
        if (any(shape(x)/=[i])) STOP 'ERROR: bellman: incompatible dimensions'
    end subroutine
    subroutine checkShapeInt(x,i)
        implicit none
        integer, intent(in) :: x(:)
        integer, intent(in) :: i
        if (any(shape(x)/=[i])) STOP 'ERROR: bellman: incompatible dimensions'
    end subroutine

    subroutine sub_bellman_discrete_core(V,c,api,W,y,x,lenA,lenE,lenAp,crra,opt)
        implicit none
        integer, intent(in) :: lenA,lenE,lenAp
        real(rt), dimension(lenA,*), intent(out) :: V,c
        integer, dimension(lenA,*), intent(out) :: api
        real(rt), dimension(lenA,*), intent(in) :: y
        real(rt), dimension(lenAp,*), intent(in) :: W,x
        real(rt), intent(in) :: crra
        integer, intent(in) :: opt(1:)
        ! local
        logical :: isConcave, isMono, xIsIncr, doOpenMP
        integer :: iE,iA,lastiap,lb,ub

        ! Check inputs
        if (size(opt)<3) then
            STOP 'Opt length must be at least three'
        end if

        ! Unpack the option values
        xIsIncr = opt(1)==1 ! whether x(a',e) is increasing in a'
        isMono = opt(2)>=1 ! whether a'(a,e) is mono in a
        isConcave = opt(2)==2 ! whether u+W is concave in a' for all e

        doOpenMP = opt(3)==1 ! use openmp

        if (isConcave .and. .not. xIsIncr) STOP 'ERROR: sub_bellman_discrete: x must be increasing if concave'

        !$OMP PARALLEL DO DEFAULT(SHARED) PRIVATE(iE,iA,lastiap,lb,ub) IF(doOpenMP)
        do iE = 1,lenE

            lastiap = 1

            do iA = 1,lenA

                ! Set defaults
                V(iA,iE) = deathv
                api(iA,iE) = deathapi
                c(iA,iE) = deathc

                if (.not. isConcave) then 

                    ! If not concave, then perform a grid search
                    call gridSearch(lastiap,lenAp,iA,iE)

                else

                    ! If concave and monotone, perform a binary search
                    lb = lastiap
                    ub = lenAp
                    do while (ub>lb)
                        call updateLbUb(lb,ub,iA,iE)
                    end do

                    ! Update the value and policy functions
                    if (lb/=ub) STOP 'ERROR: sub_bellman_discrete: logic did not work' 
                    call gridSearch(lb,ub,iA,iE) ! only checks one value

                end if

                ! If monotonicity assumption, then restrict search space 
                if (isMono) then
                    lastiap = max(api(iA,iE),1)
                end if

            end do
        end do
        !$OMP END PARALLEL DO

    contains

        ! Given a bound on the search space, lb and ub, replaces either lb or ub 
        ! with a midpoint such that [lb,ub] is cut roughly in half. It does this assuming
        ! concavity and checking the midpoint(s).
        subroutine updateLbUb(lb,ub,iA,iE)
            implicit none
            integer, intent(inout) :: lb,ub
            integer, intent(in) :: iA,iE
            ! local
            integer :: mid
            real(rt) :: clo,chi,vlo,vhi

            mid = floor(.5e0_rt*(lb+ub))

            ! Evaluate utility at midpoints
            call evalH(vlo,clo,mid,iA,iE)
            call evalH(vhi,chi,mid+1,iA,iE)

            if (clo>0e0_rt .and. chi>0e0_rt) then
                ! If both are feasible choices (typical case) then compare utility
                if (vlo>vhi) then
                    ub = mid
                else
                    lb = mid+1
                end if
            else
                ! If either is negative, then since x is increasing in iAp, know the optimum must be in 1,mid
                ub = mid
            end if
            
        end subroutine 

        ! Given bounds on search space, lb and ub, finds the index of a policy 
        ! resulting in V higher than the current V(iA,iE). If no such policy
        ! exists, V(iA,iE),c(iA,iE),api(iA,iE) are unchanged.
        subroutine gridSearch(lb,ub,iA,iE)
            implicit none
            integer, intent(in) :: lb,ub,iA,iE
            ! local
            integer :: iAp
            real(rt) :: cval,vval

            ! Search for the optimal policy
            do iAp = lb,ub

                ! Evaluate utility at a specific iAp
                call evalH(vval,cval,iAp,iA,iE)

                ! Check whether consumption is negative
                if (cval>0e0_rt) then
                    if (vval>V(iA,iE)) then
                        V(iA,iE) = vval
                        c(iA,iE) = cval
                        api(iA,iE) = iAp
                    end if
                elseif (xIsIncr) then
                    ! Since x is increasing in a', any subsequent values will result in lower c
                    exit
                end if

            end do

        end subroutine gridSearch

        ! Compute consumption and utility conditional on iA,iE for a given iAp. 
        ! If cval <= 0e0_rt, vval is not assigned a value.
        subroutine evalH(vval,cval,iAp,iA,iE)
            implicit none
            integer, intent(in) :: iAp,iA,iE
            real(rt), intent(out) :: cval,vval


            cval = y(iA,iE) - x(iAp,iE)  

            if (cval>0e0_rt) then

                ! Compute utility 
                if (crra==2e0_rt) then
                    vval = -1e0_rt/cval + W(iAp,iE)
                elseif (crra==1e0_rt) then
                    vval = log(cval) + W(iAp,iE)
                elseif (crra>0e0_rt) then
                    vval = cval**(1e0_rt-crra)/(1e0_rt-crra) + W(iAp,iE)
                else
                    vval = PUcustom(cval) + W(iAp,iE)
                end if

!                 ! Compute utility 
!                 if (crra==2e0_rt) then
!                     vval = -1e0_rt/(cval+1d-6) + W(iAp,iE)
!                 else
!                     STOP 'adding constant to utility'
!                 end if
            end if

        end subroutine

    end subroutine sub_bellman_discrete_core


    subroutine sub_bellman_discrete_choose2_1d(V,c,api,bpi,W,y,x,crra,opt)
        implicit none
        real(rt), dimension(:), intent(out) :: V,c
        integer, dimension(:), intent(out) :: api,bpi
        real(rt), dimension(:), intent(in) :: y
        real(rt), dimension(:,:), intent(in) :: W,x
        real(rt), intent(in) :: crra
        integer, intent(in) :: opt(1:)
        ! local
        integer :: lenA,lenAp,lenBp

        lenA = size(V,1)
        lenAp = size(W,1)
        lenBp = size(W,2)

        call checkShape(V,lenA)
        call checkShape(c,lenA)
        call checkShape(y,lenA)
        call checkShape(x,lenAp,lenBp)
        call checkShape(W,lenAp,lenBp)
        call checkShape(api,lenA)
        call checkShape(bpi,lenA)

        call sub_bellman_discrete_choose2_core(V,c,api,bpi,W,y,x,lenA,1,lenAp,lenBp,crra,opt)

    end subroutine 

    ! 2-dimensional optimization
    ! 
    ! Solves 
    !   V(a,e) = max_{a',b'} u(y(a,e)-x(a',b',e)) + W(a',b',e) for 
    ! a' and b' on grids (whose values need not be known).
    ! 
    ! Model options, governed by opt(1)
    ! 0         x is general form 
    ! 1         x is increasing in a' for each e (speedup available)
    ! 2 [same as 0] x is increasing in b' for each e
    ! 3 [same as 1] x is increasing in a' and b' for each e
    !
    ! Speedup options, governed by opt(2)
    ! 0         Brute force grid search
    ! 1         Monotonicity assumption a'(a,e) monotone in a
    ! 2         Monotonicity assumption b'(a,e) monotone in a
    ! 3         Monotonicity assumption a'(a,e) and b'(a,e) monotone in a
    ! XXX       Concavity Not programmed yet
    ! 
    subroutine sub_bellman_discrete_choose2_core(V,c,api,bpi,W,y,x,lenA,lenE,lenAp,lenBp,crra,opt)
        implicit none
        integer, intent(in) :: lenA,lenE,lenAp,lenBp
        real(rt), dimension(lenA,*), intent(out) :: V,c
        integer, dimension(lenA,*), intent(out) :: api,bpi
        real(rt), dimension(lenA,*), intent(in) :: y
        real(rt), dimension(lenAp,lenBp,*), intent(in) :: W,x
        real(rt), intent(in) :: crra
        integer, intent(in) :: opt(1:)
        ! local
        logical :: isMonoA, isMonoB, xIsIncrA, xIsIncrB ! isConcave
        integer :: iE,iA,lastiap,lastibp !,lb,ub

        ! Check inputs
        if (size(opt)<2) then
            STOP 'Opt length must be at least two'
        end if

        ! Unpack the option values
        xIsIncrA = opt(1)==1 .or. opt(1)==3 ! whether x(a',b',e) is increasing in a'
        xIsIncrB = opt(1)==2 .or. opt(1)==3 ! whether x(a',b',e) is increasing in b'
        isMonoA = opt(2)==1 .or. opt(2)==3 ! whether a'(a,e) is mono in a
        isMonoB = opt(2)==2 .or. opt(2)==3 ! whether b'(a,e) is mono in a

!         !$OMP PARALLEL DO DEFAULT(SHARED) PRIVATE(iE,iA,lastiap,lastibp)
    !!! CRAY compiler threw an error here saying xisincra is used before defined with openmp
        do iE = 1,lenE

            lastiap = 1
            lastibp = 1

            do iA = 1,lenA

                ! Set defaults
                V(iA,iE) = deathv
                api(iA,iE) = deathapi
                bpi(iA,iE) = deathapi
                c(iA,iE) = deathc

                ! If not concave, then perform a grid search
                call gridSearch(lastiap,lenAp,lastibp,lenBp,iA,iE)

                ! If monotonicity assumption, then restrict search space 
                if (isMonoA) lastiap = max(api(iA,iE),1)
                if (isMonoB) lastibp = max(bpi(iA,iE),1)

            end do
        end do
!         !$OMP END PARALLEL DO

    contains

        ! Given bounds on search space, lb and ub, finds the index of a policy 
        ! resulting in V higher than the current V(iA,iE). If no such policy
        ! exists, V(iA,iE),c(iA,iE),api(iA,iE),bpi(ia,iE) are unchanged.
        subroutine gridSearch(a_lb,a_ub,b_lb,b_ub,iA,iE)
            implicit none
            integer, intent(in) :: iA,iE
            integer, intent(in) :: a_lb,a_ub,b_lb,b_ub
            real(rt) :: cval,vval
            integer :: iAp,iBp

            ! Search for the optimal policy
            do iBp = b_lb,b_ub
                do iAp = a_lb,a_ub

                    ! Evaluate utility at a specific iAp
                    call evalH(vval,cval,iAp,iBp,iA,iE)

                    ! Check whether consumption is negative
                    if (cval>0e0_rt) then
                        if (vval>V(iA,iE)) then
                            V(iA,iE) = vval
                            c(iA,iE) = cval
                            api(iA,iE) = iAp
                            bpi(iA,iE) = iBp
                        end if
                    elseif (xIsIncrA) then
                        ! Since x is increasing in a', any subsequent values will result in lower c
                        exit
                    end if

                end do
            end do

        end subroutine gridSearch

        ! Compute consumption and utility conditional on iA,iE for a given iAp,iBp. 
        ! If cval <= 0e0_rt, vval is not assigned a value.
        subroutine evalH(vval,cval,iAp,iBp,iA,iE)
            implicit none
            integer, intent(in) :: iAp,iBp,iA,iE
            real(rt), intent(out) :: cval,vval

            cval = y(iA,iE) - x(iAp,iBp,iE)  

            if (cval>0e0_rt) then

                ! Compute utility 
                if (crra==2e0_rt) then
                    vval = -1e0_rt/cval + W(iAp,iBp,iE)
                elseif (crra==1e0_rt) then
                    vval = log(cval) + W(iAp,iBp,iE)
                elseif (crra>0e0_rt) then
                    vval = cval**(1e0_rt-crra)/(1e0_rt-crra) + W(iAp,iBp,iE)
                else
                    vval = PUcustom(cval) + W(iAp,iBp,iE)
                end if

            else

                vval = deathv

            end if

        end subroutine

    end subroutine sub_bellman_discrete_choose2_core

    subroutine sub_bellman_interp_choose2_1d(V,c,ap,bp,W,y,x,gAP,gAPfine,gBP,gBPfine,crra,opt)
        implicit none
        real(rt), dimension(:), intent(out) :: V,c,ap,bp
        real(rt), dimension(:), intent(in) :: y,gAP,gAPfine,gBP,gBPfine
        real(rt), dimension(:,:), intent(in) :: W,x
        real(rt), intent(in) :: crra
        integer, intent(in) :: opt(1:)
        ! local
        integer :: lenA,lenAp,lenBp

        lenA = size(V,1)
        lenAp = size(W,1)
        lenBp = size(W,2)

        call checkShape(V,lenA)
        call checkShape(c,lenA)
        call checkShape(y,lenA)
        call checkShape(x,lenAp,lenBp)
        call checkShape(W,lenAp,lenBp)
        call checkShape(ap,lenA)
        call checkShape(bp,lenA)
        call checkShape(gAP,lenAp)
        call checkShape(gBP,lenBp)

        call sub_bellman_interp_choose2_core(V,c,ap,bp,W,y,x,gAP,gAPfine,gBP,gBPfine,lenA,1,lenAp,lenBp,crra,opt)

    end subroutine 

    subroutine sub_bellman_interp_choose2_core(V,c,ap,bp,W,y,x,gAP,gAPfine,gBP,gBPfine,lenA,lenE,lenAp,lenBp,crra,opt)
        use mod_matlab, only: bsearch, disp, interp2q
        use mod_interp, only: isearch
        use mod_spline
        use mod_min1dim
        implicit none
        integer, intent(in) :: lenA,lenE,lenAp,lenBp
        real(rt), dimension(lenA,*), intent(out) :: V,c
        real(rt), dimension(lenA,*), intent(out) :: ap,bp
        real(rt), dimension(lenA,*), intent(in) :: y
        real(rt), dimension(lenAp,lenBp,*), intent(in) :: W,x
        real(rt), dimension(lenAp), intent(in) :: gAP
        real(rt), dimension(lenBp), intent(in) :: gBP 
        real(rt), dimension(:), intent(in) :: gAPfine, gBPfine ! interpolation grids
        real(rt), intent(in) :: crra
        integer, intent(in) :: opt(1:)
        ! local 
        real(rt), allocatable :: Wfine(:,:,:), xfine(:,:,:)
        integer, allocatable :: api(:,:),bpi(:,:)
        integer :: lenAPfine,lenBPfine,iE

        if (size(opt)<3) STOP 'sub_bellman_interp_core: opt must have length >= 3'

        lenAPfine = size(gAPfine)
        lenBPfine = size(gBPfine)

        allocate(api(lenA,lenE))
        allocate(bpi(lenA,lenE))
        allocate(Wfine(lenAPfine,lenBPfine,lenE))
        allocate(xfine(lenAPfine,lenBPfine,lenE))

        ! NOTE: it would probably be a lot faster to not use interp2 but
        ! rather manually do the interpolation and not call discrete_choose2. 
        ! For now, I don't pursue this for time's sake.
        Wfine = interp2q(gAP,gBP,W(:,:,1:lenE),gAPfine,gBPfine)
        xfine = interp2q(gAP,gBP,x(:,:,1:lenE),gAPfine,gBPfine)

        ! Call bellman discrete to do the actual optimization
        call sub_bellman_discrete_choose2_core(V,c,api,bpi,Wfine,y,xfine,lenA,lenE,lenAPfine,lenBPfine,crra,opt(2:3))

        ! Recover the a' values
        do iE = 1,lenE
            where (c(:,iE)>0e0_rt) 
                ap(:,iE) = gAPfine(api(:,iE))
                bp(:,iE) = gBPfine(bpi(:,iE))
            elsewhere
                ap(:,iE) = deathap
                bp(:,iE) = deathap
            end where
        end do

    end subroutine sub_bellman_interp_choose2_core

end module mod_valueopt


module mod_valueopt_test
    use mod_valueopt
    implicit none
    integer, parameter :: nK = 20, nZ = 3, maxiter = 3000 ! 1500
    ! integer, parameter :: nK = 2048, nZ = 64, maxiter = 3000 ! 1500
    real(rt), parameter :: alpha = .36e0_rt, beta = .99e0_rt, crra = 2e0_rt, deprec = .025e0_rt
    real(rt) :: Fzz(nZ,nZ),Z(nZ),K(nK),sup,lastsup,time(0:12),xopt(1)
    integer :: opt(5),iter,iZ,option
    real(rt), dimension(:,:), allocatable :: V,c,W,y,x,V0,kp,suphistory
    integer, dimension(:,:), allocatable :: kpi
contains

    subroutine u(pu,iK,iKp,flag,realargs,intargs)
        implicit none
        real(rt), intent(out) :: pu
        integer, intent(in) :: iK,iKp
        integer, intent(out) :: flag
        real(rt), intent(inout) :: realargs(*)
        integer, intent(inout) :: intargs(*)
        ! local
        real(rt) :: ctmp
        integer :: iZ

        iZ = intargs(1)

        ! Compute period utility
        ! ctmp = Z(iZ)*K(iK)**alpha + (1e0_rt-deprec)*K(iK) - K(iKp)
        ctmp = Y(iK,iZ) - K(iKp)
        if (ctmp>0e0_rt) then
            pu = -1e0_rt/ctmp
            flag = 0
        else
            pu = -sqrt(huge(0._rt))
            flag = 1
        end if

    end subroutine
    subroutine contutil(cv,iKp,flag,realargs,intargs)
        implicit none
        real(rt), intent(out) :: cv
        integer, intent(in) :: iKp
        integer, intent(out) :: flag
        real(rt), intent(inout) :: realargs(*)
        integer, intent(inout) :: intargs(*)
        ! local
        integer :: iZ

        iZ = intargs(1)

        ! Compute period utility
        cv = W(iKp,iZ)

        flag = 0

    end subroutine



    ! test_valueopt 
    ! Computes the solution to 
    ! V(k,z) = max u(c) + beta*E(V(k',z'))
    ! with c = -k' + z*k**alpha + (1e0_rt-deprec)*k
    subroutine test_valueopt
        use mod_discreteAR1
        use mod_matlab
        implicit none
        real(rt) :: realargs(1)
        integer :: intargs(1)

        allocate(V(nK,nZ))
        allocate(c(nK,nZ))
        allocate(W(nK,nZ))
        allocate(y(nK,nZ))
        allocate(x(nK,nZ))
        allocate(V0(nK,nZ))
        allocate(kp(nK,nZ))
        allocate(kpi(nK,nZ))
        allocate(suphistory(maxiter,0:12))

        ! Discretize the AR1
        call sub_rouwenhorst(Z,Fzz,.95e0_rt,.007e0_rt,nZ)
        Z = exp(Z)

!         call disp('Z',Z)
!         call disp('Fzz',Fzz)

        ! Get capital grid
        K = linspace(.7e0_rt,1.3e0_rt,nK)*((1e0_rt/beta+deprec-1e0_rt)/alpha)**(1e0_rt/(alpha-1e0_rt))

!         call disp('K',K)

        ! Compute y and x
        do iZ = 1,nZ
            y(:,iZ) = Z(iZ)*K**alpha + (1e0_rt-deprec)*K
            x(:,iZ) = K ! expenditure
        end do

        do option = 0,12

            ! Set opt according to which cycle we are doing
            select case (option)
                case (0)
                    print*,'new bellman: binary monotonicity'
               case (1)
                    print*,'Checking brute force w/ OpenMP'
                    opt(1:3) = [0,0,1]
                case (2)
                    print*,'Checking monotonicity w/ OpenMP'
                    opt(1:3) = [0,1,1]
                case (3)
                    print*,'Checking brute force w increasing x w/ OpenMP'
                    opt(1:3) = [1,0,1]
                case (4)
                    print*,'Checking monotonicity w increasing x w/ OpenMP'
                    opt(1:3) = [1,1,1]
                case (5)
                    print*,'Checking concavity w increasing x w/ OpenMP'
                    opt(1:3) = [1,2,1]
                case (6)
                    print*,'Checking brute force w/o OpenMP'
                    opt(1:3) = [0,0,0]
                case (7)
                    print*,'Checking monotonicity w/o OpenMP'
                    opt(1:3) = [0,1,0]
                case (8)
                    print*,'Checking brute force w increasing x w/o OpenMP'
                    opt(1:3) = [1,0,0]
                case (9)
                    print*,'Checking monotonicity w increasing x w/o OpenMP'
                    opt(1:3) = [1,1,0]
                case (10)
                    print*,'Checking concavity w increasing x w/o OpenMP'
                    opt(1:3) = [1,2,0]
                case (11)
                    print*,'Continuous: checking lininterp'
                    opt(1:4) = [1,2,1,0]
                    xopt(1) = sqrt(epsilon(0e0_rt))
                case (12)
                    print*,'Continuous: checking Schumaker'
                    opt(1:4) = [1,2,1,1]
                    xopt(1) = sqrt(epsilon(0e0_rt))
            end select

            suphistory(:,option) = -1e0_rt

            call tic()

            ! VFI 
            do iZ = 1,nZ
                V0(:,iZ) = 0e0_rt !log(K) ! - log(K(nK))! NOTE: guess must be strictly concave! 
            end do
            ! V0 = -1d-20
            do iter = 1,10000

                W = beta*matmul(V0,transpose(Fzz))

                if (option==0) then
                    !$OMP PARALLEL DO PRIVATE(iZ,intargs) DEFAULT(SHARED)
                    do iZ = 1,nZ
                        intargs(1) = iZ                        
                        call bellman(V(:,iZ),kpi(:,iZ),u,contutil,nK,[2,0],realargs,intargs)
                    end do
                    !$OMP END PARALLEL DO
                    ! call plot(V)
                    ! call plot(real(kpi,8))
                    ! if (iter==3) STOP 'check'
                elseif (option<=10) then
                    call sub_bellman_discrete(V,c,kpi,W,y,x,crra,opt)
                else
                    call sub_bellman_continuous(V,c,kp,W,y,x,K,crra,opt,xopt)
                end if

                ! Compute the supnorm
                if (iter>1) lastsup = sup

                sup = maxval(abs(V-V0)) 

                if (iter>1) then
                    if (sup/lastsup>beta+1e-3_rt*beta) then
                        print*,'option',option
                        print*,'iter',iter
                        print*,'sup',sup
                        print*,'lastsup',lastsup
                        print*,'sup/lastsup',sup/lastsup
                        print*,'beta',beta
                        print*,'ERROR: supnorm did not fall by beta'
                    end if
                end if
                suphistory(iter,option) = sup

                if (mod(iter,500)==0) print*,'iter',iter,'supnorm',sup

                if (sup<1e-3_rt*maxval(abs(V))) exit

                ! Update V
                V0 = V

            end do

            ! Check that the supnorm history agrees 
            if (option>0 .and. option<=10) then
                ! supnorm history should be completely the same in the discrete case
                if (any(suphistory(:,option)/=suphistory(:,option-1))) then 
                    STOP 'ERROR: supnorm history does not agree'
                end if
            end if

! call plot(K,V,'V')
! call plot(K,real(kpi,8),'kpi')
! 
            call toc(time(option))

!         STOP 'only checking valueopt option 0'

        end do

!         call plot(K,c,'c')
!         call plot(suphistory,'supnorm')

!         call disp('supnorm',suphistory)

        print*,'max diff in supnorm history from continuous and discrete'
        print*,maxval(abs(suphistory(:,1)-suphistory(:,11))),maxval(abs(suphistory(:,1)-suphistory(:,12)))
        print*,'max diff in supnorm history between two continuous methods'
        print*,maxval(abs(suphistory(:,11)-suphistory(:,12)))

        call disp('option times',time)

        print*,'test_valueopt: passed'
!         STOP 'only checking valueopt'

    end subroutine
end module 
! mod_csv
! 
! A module for reading and writing fortran arrays to csv files.
! The csv files can then be opened in Excel or imported into Matlab 
! via the corresponding matlab code.
!
! The default directory for the csv files to be placed is './results/'.
! If you need a different directory, use "usedir([dir])". You can determine the current
! directory via "getdir()".
! 
! To save an array A with filename "A.csv", call csvwrite('A',A).
! To read an array A with filename "A.csv", call csvread('A',A).
!
! Currently, very little error checking is done, so this may crash if used improperly. 
! E.g., in reading in an array, no attempt is made to ensure that the array in the csv file
! has the same shape or properties as the one being requested. 
! 
! Grey Gordon, 2014
! Soli Deo Gloria
module mod_csv
    implicit none

    character(len=255) :: dir = '.'
    character(len=255) :: scratchdir ! temporary directory that can be reverted to

    logical :: rwnotdisabled = .true.

    interface csvwrite
        module procedure w0d,w1d,w2d,w3d,w4d,w5d,w6d,&
                         w0i,w1i,w2i,w3i,w4i,w5i,w6i,&
                         w0l,w1l,w2l,w3l,w4l,w5l,w6l,&
                         w0f,w1f,w2f,w3f,w4f,w5f,w6f
    end interface
    interface csvread
        module procedure r0d,r1d,r2d,r3d,r4d,r5d,r6d,&
                         r0i,r1i,r2i,r3i,r4i,r5i,r6i,&
                         r0l,r1l,r2l,r3l,r4l,r5l,r6l,&
                         r0f,r1f,r2f,r3f,r4f,r5f,r6f
    end interface

    character(len=*), parameter :: fmt_d = '(e25.17e3)'
    character(len=*), parameter :: fmt_f = '(f25.17)'
    character(len=*), parameter :: fmt_i = '(i12)'
    character(len=*), parameter :: fmt_l = '(i1)'

    private 
    public :: csvwrite,csvread,usedir,getdir,test_csv,disablereadwrite,rdcore,enablereadwrite,rwnotdisabled

contains

    subroutine disablereadwrite()
        implicit none
        rwnotdisabled = .false.
    end subroutine 
    subroutine enablereadwrite()
        implicit none
        rwnotdisabled = .true.
    end subroutine 

    ! Setting or getting the directory
    function getdir() result(dirout)
        implicit none
        character(len=255) :: dirout
        dirout = dir
    end function

    subroutine usedir(dirin)
        implicit none
        character(len=*), intent(in) :: dirin
        scratchdir = dir
        dir = dirin

        ! Check that the directory actually exists
        call checkdir()
    end subroutine

    ! Tests whether the directory exists
    subroutine checkdir(var)
        implicit none
        integer :: unitno
        character(len=*), intent(in), optional :: var

        if (rwnotdisabled) then 

            ! Test whether the directory exists
            open(newunit=unitno,file=trim(dir)//'deleteme.txt',status='replace',err=1234)
            close (unitno)
            return

            ! If doesn't exist, end gracefully
1234        &
            if (present(var)) write(*,*) 'While trying to access file ',var,' found that '
            if (present(var)) write(0,*) 'While trying to access file ',var,' found that '
            write(*,*) 'data directory, '//trim(dir)//' does not exist or could not write there!'
            write(0,*) 'data directory, '//trim(dir)//' does not exist or could not write there!'
            STOP 

        end if

    end subroutine

    !!! READING ROUTINES
    subroutine r0d(str,var)
        real(8), intent(out) :: var
        character(len=*) :: str
        real(8) :: tmp(1)
        call rdcore(str,tmp,[1])
        var = tmp(1)
    end subroutine

    subroutine r1d(str,var); real(8), intent(out) :: var(:); character(len=*) :: str; call rdcore(str,var,shape(var)); end subroutine
    subroutine r2d(str,var); real(8), intent(out) :: var(:,:); character(len=*) :: str; call rdcore(str,var,shape(var)); end subroutine
    subroutine r3d(str,var); real(8), intent(out) :: var(:,:,:); character(len=*) :: str; call rdcore(str,var,shape(var)); end subroutine
    subroutine r4d(str,var); real(8), intent(out) :: var(:,:,:,:); character(len=*) :: str; call rdcore(str,var,shape(var)); end subroutine
    subroutine r5d(str,var); real(8), intent(out) :: var(:,:,:,:,:); character(len=*) :: str; call rdcore(str,var,shape(var)); end subroutine
    subroutine r6d(str,var); real(8), intent(out) :: var(:,:,:,:,:,:); character(len=*) :: str; call rdcore(str,var,shape(var)); end subroutine

    subroutine r0f(str,var); real(4), intent(out) :: var; character(len=*) :: str; real(4) :: tmp(1); call rfcore(str,tmp,[1]); var = tmp(1); end subroutine
    subroutine r1f(str,var); real(4), intent(out) :: var(:); character(len=*) :: str; call rfcore(str,var,shape(var)); end subroutine
    subroutine r2f(str,var); real(4), intent(out) :: var(:,:); character(len=*) :: str; call rfcore(str,var,shape(var)); end subroutine
    subroutine r3f(str,var); real(4), intent(out) :: var(:,:,:); character(len=*) :: str; call rfcore(str,var,shape(var)); end subroutine
    subroutine r4f(str,var); real(4), intent(out) :: var(:,:,:,:); character(len=*) :: str; call rfcore(str,var,shape(var)); end subroutine
    subroutine r5f(str,var); real(4), intent(out) :: var(:,:,:,:,:); character(len=*) :: str; call rfcore(str,var,shape(var)); end subroutine
    subroutine r6f(str,var); real(4), intent(out) :: var(:,:,:,:,:,:); character(len=*) :: str; call rfcore(str,var,shape(var)); end subroutine

    subroutine r0i(str,var); integer, intent(out) :: var; character(len=*) :: str; integer :: tmp(1); call ricore(str,tmp,[1]); var = tmp(1); end subroutine
    subroutine r1i(str,var); integer, intent(out) :: var(:); character(len=*) :: str; call ricore(str,var,shape(var)); end subroutine
    subroutine r2i(str,var); integer, intent(out) :: var(:,:); character(len=*) :: str; call ricore(str,var,shape(var)); end subroutine
    subroutine r3i(str,var); integer, intent(out) :: var(:,:,:); character(len=*) :: str; call ricore(str,var,shape(var)); end subroutine
    subroutine r4i(str,var); integer, intent(out) :: var(:,:,:,:); character(len=*) :: str; call ricore(str,var,shape(var)); end subroutine
    subroutine r5i(str,var); integer, intent(out) :: var(:,:,:,:,:); character(len=*) :: str; call ricore(str,var,shape(var)); end subroutine
    subroutine r6i(str,var); integer, intent(out) :: var(:,:,:,:,:,:); character(len=*) :: str; call ricore(str,var,shape(var)); end subroutine

    subroutine r0l(str,var); logical, intent(out) :: var; character(len=*) :: str; logical :: tmp(1); call rlcore(str,tmp,[1]); var = tmp(1); end subroutine
    subroutine r1l(str,var); logical, intent(out) :: var(:); character(len=*) :: str; call rlcore(str,var,shape(var)); end subroutine
    subroutine r2l(str,var); logical, intent(out) :: var(:,:); character(len=*) :: str; call rlcore(str,var,shape(var)); end subroutine
    subroutine r3l(str,var); logical, intent(out) :: var(:,:,:); character(len=*) :: str; call rlcore(str,var,shape(var)); end subroutine
    subroutine r4l(str,var); logical, intent(out) :: var(:,:,:,:); character(len=*) :: str; call rlcore(str,var,shape(var)); end subroutine
    subroutine r5l(str,var); logical, intent(out) :: var(:,:,:,:,:); character(len=*) :: str; call rlcore(str,var,shape(var)); end subroutine
    subroutine r6l(str,var); logical, intent(out) :: var(:,:,:,:,:,:); character(len=*) :: str; call rlcore(str,var,shape(var)); end subroutine


    !!! WRITING ROUTINES
    subroutine w0d(s,v);real(8),intent(in)::v;character(*)::s;call wdcore(s,[v],shape(v)); end subroutine
    subroutine w1d(s,v);real(8),intent(in)::v(:);character(*)::s;call wdcore(s,v,shape(v)); end subroutine
    subroutine w2d(s,v);real(8),intent(in)::v(:,:);character(*)::s;call wdcore(s,v,shape(v)); end subroutine
    subroutine w3d(s,v);real(8),intent(in)::v(:,:,:);character(*)::s;call wdcore(s,v,shape(v)); end subroutine
    subroutine w4d(s,v);real(8),intent(in)::v(:,:,:,:);character(*)::s;call wdcore(s,v,shape(v)); end subroutine
    subroutine w5d(s,v);real(8),intent(in)::v(:,:,:,:,:);character(*)::s;call wdcore(s,v,shape(v)); end subroutine
    subroutine w6d(s,v);real(8),intent(in)::v(:,:,:,:,:,:);character(*)::s;call wdcore(s,v,shape(v)); end subroutine
    subroutine w0f(s,v);real(4),intent(in)::v;character(*)::s;call wfcore(s,[v],shape(v)); end subroutine
    subroutine w1f(s,v);real(4),intent(in)::v(:);character(*)::s;call wfcore(s,v,shape(v)); end subroutine
    subroutine w2f(s,v);real(4),intent(in)::v(:,:);character(*)::s;call wfcore(s,v,shape(v)); end subroutine
    subroutine w3f(s,v);real(4),intent(in)::v(:,:,:);character(*)::s;call wfcore(s,v,shape(v)); end subroutine
    subroutine w4f(s,v);real(4),intent(in)::v(:,:,:,:);character(*)::s;call wfcore(s,v,shape(v)); end subroutine
    subroutine w5f(s,v);real(4),intent(in)::v(:,:,:,:,:);character(*)::s;call wfcore(s,v,shape(v)); end subroutine
    subroutine w6f(s,v);real(4),intent(in)::v(:,:,:,:,:,:); character(*) :: s; call wfcore(s,v,shape(v)); end subroutine

    subroutine w0i(str,var)
        integer, intent(in) :: var
        character(len=*) :: str
        integer :: tmp(1); tmp(1) = var
        call wicore(str,tmp,shape(var))
    end subroutine

    subroutine w1i(str,var)
        integer, intent(in) :: var(:)
        character(len=*) :: str
        call wicore(str,var,shape(var))
    end subroutine

    subroutine w2i(str,var)
        integer, intent(in) :: var(:,:)
        character(len=*) :: str
        call wicore(str,var,shape(var))
    end subroutine

    subroutine w3i(str,var)
        integer, intent(in) :: var(:,:,:)
        character(len=*) :: str
        call wicore(str,var,shape(var))
    end subroutine

    subroutine w4i(str,var)
        integer, intent(in) :: var(:,:,:,:)
        character(len=*) :: str
        call wicore(str,var,shape(var))
    end subroutine

    subroutine w5i(str,var)
        integer, intent(in) :: var(:,:,:,:,:)
        character(len=*) :: str
        call wicore(str,var,shape(var))
    end subroutine

    subroutine w6i(str,var)
        integer, intent(in) :: var(:,:,:,:,:,:)
        character(len=*) :: str
        call wicore(str,var,shape(var))
    end subroutine

    subroutine w0l(str,var)
        logical, intent(in) :: var
        character(len=*) :: str
        logical :: tmp(1); tmp(1) = var
        call wlcore(str,tmp,shape(var))
    end subroutine

    subroutine w1l(str,var)
        logical, intent(in) :: var(:)
        character(len=*) :: str
        call wlcore(str,var,shape(var))
    end subroutine

    subroutine w2l(str,var)
        logical, intent(in) :: var(:,:)
        character(len=*) :: str
        call wlcore(str,var,shape(var))
    end subroutine

    subroutine w3l(str,var)
        logical, intent(in) :: var(:,:,:)
        character(len=*) :: str
        call wlcore(str,var,shape(var))
    end subroutine

    subroutine w4l(str,var)
        logical, intent(in) :: var(:,:,:,:)
        character(len=*) :: str
        call wlcore(str,var,shape(var))
    end subroutine

    subroutine w5l(str,var)
        logical, intent(in) :: var(:,:,:,:,:)
        character(len=*) :: str
        call wlcore(str,var,shape(var))
    end subroutine

    subroutine w6l(str,var)
        logical, intent(in) :: var(:,:,:,:,:,:)
        character(len=*) :: str
        call wlcore(str,var,shape(var))
    end subroutine

    !!! CORE ROUTINES FOR READING AND WRITING
    subroutine makeheader(unitno,vartype,str,dims)
        implicit none
        character(len=*), intent(in) :: str
        integer, intent(in) :: unitno,vartype,dims(:)
        integer :: i

        if (.not. rwnotdisabled) return

        ! header info
        
        ! row 1: variable name | additional, reserved
        ! row 2: variable type | integer specifying type: 0 <=> real(8); 1 <=> integer; 2 <=> logical | variable format (string) | additional, reserved
        ! row 3: variable rank | additional, reserved
        ! row 4: variable dimensions
        ! row 5: reserved
        ! row 6: reserved
        ! row 7: reserved
        ! row 8: reserved
        ! row 9: reserved

        write(unitno,'(A,A,A)') 'Variable name:',',',trim(str) ! row 1
        select case (vartype)
            case (0)
                write(unitno,'(A,A,A,A,i3,A,A)') 'Variable type:',',','real(8)',',',0,',',fmt_d ! row 2
            case (1) 
                write(unitno,'(A,A,A,A,i3,A,A)') 'Variable type:',',','integer',',',1,',',fmt_i ! row 2
            case (2) 
                write(unitno,'(A,A,A,A,i3,A,A)') 'Variable type:',',','logical',',',2,',',fmt_l ! row 2
            case (3)
                write(unitno,'(A,A,A,A,i3,A,A)') 'Variable type:',',','real(4)',',',0,',',fmt_f ! row 2
        end select
        write(unitno,'(A,A,i3)') 'Variable rank:',',',size(dims)
        write(unitno,'(A,A)',advance='no') 'Variable dims:',','
        do i = 1,size(dims)
            write(unitno,'(i10)',advance='no') dims(i)
            if (i<size(dims)) write(unitno,'(A)',advance='no') ','
        end do
        if (size(dims)==0) then
            write(unitno,'(i10)',advance='no') 1
        end if
        write(unitno,'(A)') ''  ! row 4 (dimensions)
        write(unitno,'(A)') ''  ! row 5
        write(unitno,'(A)') ''  ! row 6
        write(unitno,'(A)') ''  ! row 7
        write(unitno,'(A)') ''  ! row 8
        write(unitno,'(A)') ''  ! row 9

    end subroutine

    subroutine getninjnk(ni,nj,nk,dims)
        implicit none
        integer, intent(in) :: dims(:)
        integer, intent(out) :: ni,nj,nk

        nk = 1
        nj = 1
        ni = 1
        if (size(dims)>=3) nk = product(dims(3:size(dims)))
        if (size(dims)>=2) nj = dims(2)
        if (size(dims)>=1) ni = dims(1)
        
    end subroutine 
    
    subroutine wdcore(str,var,dims)
        implicit none
        character(len=*), intent(in) :: str
        real(8), intent(in) :: var(*)
        integer, intent(in) :: dims(:)
        integer :: unitno,i,ni,j,nj,k,nk

        if (.not. rwnotdisabled) return

        ! Check that the directory exists
        call checkdir() 

        ! Open the file
        open(newunit=unitno,file=trim(dir)//str//'.csv',action='write',status='replace')

        ! create the header
        call makeheader(unitno,0,str,dims)

        ! data (begins at row 10)
        call getninjnk(ni,nj,nk,dims)

        do k = 1,nk
            do i = 1,ni
                do j = 1,nj
                    write(unitno,fmt_d,advance='no') var(i + ((j-1) + (k-1)*nj)*ni)
                    if (j<nj) write(unitno,'(A1)',advance='no')','
                end do
                write(unitno,'(A)') ''
            end do
        end do

        close(unitno)

    end subroutine

    subroutine wfcore(str,var,dims)
        implicit none
        character(len=*), intent(in) :: str
        real(4), intent(in) :: var(*)
        integer, intent(in) :: dims(:)
        integer :: unitno,i,ni,j,nj,k,nk

        if (.not. rwnotdisabled) return

        ! Check that the directory exists
        call checkdir() 

        ! Open the file
        open(newunit=unitno,file=trim(dir)//str//'.csv',action='write',status='replace')

        ! create the header
        call makeheader(unitno,3,str,dims)

        ! data (begins at row 10)
        call getninjnk(ni,nj,nk,dims)

        do k = 1,nk
            do i = 1,ni
                do j = 1,nj
                    write(unitno,fmt_f,advance='no') var(i + ((j-1) + (k-1)*nj)*ni)
                    if (j<nj) write(unitno,'(A1)',advance='no')','
                end do
                write(unitno,'(A)') ''
            end do
        end do

        close(unitno)

    end subroutine

    subroutine wicore(str,var,dims)
        implicit none
        character(len=*), intent(in) :: str
        integer, intent(in) :: var(*)
        integer, intent(in) :: dims(:)
        integer :: unitno,i,ni,j,nj,k,nk

        if (.not. rwnotdisabled) return

        ! Check that the directory exists
        call checkdir() 

        ! Open the file
        open(newunit=unitno,file=trim(dir)//str//'.csv',action='write',status='replace')

        ! create the header
        call makeheader(unitno,1,str,dims)

        ! data (begins at row 10)
        call getninjnk(ni,nj,nk,dims)

        do k = 1,nk
            do i = 1,ni
                do j = 1,nj
                    write(unitno,fmt_i,advance='no') var(i + ((j-1) + (k-1)*nj)*ni)
                    if (j<nj) write(unitno,'(A1)',advance='no')','
                end do
                write(unitno,'(A)') ''
            end do
        end do

        close(unitno)

    end subroutine

    subroutine wlcore(str,var,dims)
        implicit none
        character(len=*), intent(in) :: str
        logical, intent(in) :: var(*)
        integer, intent(in) :: dims(:)
        integer :: unitno,i,ni,j,nj,k,nk

        if (.not. rwnotdisabled) return

        ! Check that the directory exists
        call checkdir() 

        ! Open the file
        open(newunit=unitno,file=trim(dir)//str//'.csv',action='write',status='replace')

        ! create the header
        call makeheader(unitno,2,str,dims)

        ! data (begins at row 10)
        call getninjnk(ni,nj,nk,dims)

        do k = 1,nk
            do i = 1,ni
                do j = 1,nj
                    write(unitno,fmt_l,advance='no') merge(1,0,var(i + ((j-1) + (k-1)*nj)*ni))
                    if (j<nj) write(unitno,'(A1)',advance='no')','
                end do
                write(unitno,'(A)') ''
            end do
        end do

        close(unitno)

    end subroutine

    subroutine processheader(readfmt,unitno)
        implicit none
        character(len=25), intent(out) :: readfmt
        integer, intent(in) :: unitno
        ! local
        character(len=255) :: tmp
        integer :: i,cnt,a,b


        read(unitno,*)
        ! Go to second row, third entry to string format
        read(unitno,'(A)') tmp ! read in the entire second row
        ! Process the second row, finding the fourth entry which gives the format
        cnt = 0
        b = len(tmp) ! in case there are only four fields (and so only three commas), set b to lentmp
        do i = 1,len(tmp)
            if (tmp(i:i)==',') then
                cnt = cnt + 1
                if (cnt==3) a = i+1
                if (cnt==4) then
                    b = i-1
                    exit
                end if
            end if
        end do
        readfmt = tmp(a:b)

        do i = 3,9
            read(unitno,*)
        end do

    end subroutine
    
    subroutine rdcore(str,var,dims,err)
        implicit none
        character(len=*), intent(in) :: str
        real(8), intent(out) :: var(*)
        integer, intent(in) :: dims(:)
        integer, intent(out), optional :: err
        character(len=25) :: readfmt
        integer :: unitno,i,ni,j,nj,k,nk

        if (.not. rwnotdisabled) return

        ! Check that the directory exists
        call checkdir() 

        ! Open the file
        open(newunit=unitno,file=trim(dir)//str//'.csv',action='read',status='old',err=999)

        ! process the header 
        call processheader(readfmt,unitno)

        ! data (begins at row 10)
        call getninjnk(ni,nj,nk,dims)

        do k = 1,nk
            do i = 1,ni
                do j = 1,nj
                    read(unitno,trim(readfmt),advance='no') var(i + ((j-1) + (k-1)*nj)*ni)
                    if (j<nj) read(unitno,'(1X)',advance='no')
                end do
                read(unitno,*) 
            end do
        end do

        close(unitno)

        if (present(err)) err = 0

        return

        ! If there was an error reading, say so here
999     write(0,*) 'ERROR: csvread: failed to read ',trim(dir)//str//'.csv'
        if (present(err)) err = 1

    end subroutine

    subroutine rfcore(str,var,dims)
        implicit none
        character(len=*), intent(in) :: str
        real(4), intent(out) :: var(*)
        integer, intent(in) :: dims(:)
        character(len=25) :: readfmt
        integer :: unitno,i,ni,j,nj,k,nk

        if (.not. rwnotdisabled) return

        ! Check that the directory exists
        call checkdir() 

        ! Open the file
        open(newunit=unitno,file=trim(dir)//str//'.csv',action='read',status='old',err=999)

        ! process the header 
        call processheader(readfmt,unitno)

        ! data (begins at row 10)
        call getninjnk(ni,nj,nk,dims)

        do k = 1,nk
            do i = 1,ni
                do j = 1,nj
                    read(unitno,trim(readfmt),advance='no') var(i + ((j-1) + (k-1)*nj)*ni)
                    if (j<nj) read(unitno,'(1X)',advance='no')
                end do
                read(unitno,*) 
            end do
        end do

        close(unitno)

        return

        ! If there was an error reading, say so here
999     write(0,*) 'ERROR: csvread: failed to read ',trim(dir)//str//'.csv'

    end subroutine


    subroutine ricore(str,var,dims)
        implicit none
        character(len=*), intent(in) :: str
        integer, intent(out) :: var(*)
        integer, intent(in) :: dims(:)
        character(len=25) :: readfmt
        integer :: unitno,i,ni,j,nj,k,nk

        if (.not. rwnotdisabled) return

        ! Check that the directory exists
        call checkdir() 

        ! Open the file
        open(newunit=unitno,file=trim(dir)//str//'.csv',action='read',status='old',err=999)

        ! process the header 
        call processheader(readfmt,unitno)

        ! data (begins at row 10)
        call getninjnk(ni,nj,nk,dims)

        do k = 1,nk
            do i = 1,ni
                do j = 1,nj
                    read(unitno,trim(readfmt),advance='no') var(i + ((j-1) + (k-1)*nj)*ni)
                    if (j<nj) read(unitno,'(1X)',advance='no')
                end do
                read(unitno,*) 
            end do
        end do

        close(unitno)

        return

        ! If there was an error reading, say so here
999     write(0,*) 'ERROR: csvread: failed to read ',trim(dir)//str//'.csv'

    end subroutine

    subroutine rlcore(str,var,dims)
        implicit none
        character(len=*), intent(in) :: str
        logical, intent(out) :: var(*)
        integer, intent(in) :: dims(:)
        character(len=25) :: readfmt
        integer :: unitno,i,ni,j,nj,k,nk,tmp

        if (.not. rwnotdisabled) return

        ! Check that the directory exists
        call checkdir() 

        ! Open the file
        open(newunit=unitno,file=trim(dir)//str//'.csv',action='read',status='old',err=999)

        ! process the header 
        call processheader(readfmt,unitno)

        ! data (begins at row 10)
        call getninjnk(ni,nj,nk,dims)

        do k = 1,nk
            do i = 1,ni
                do j = 1,nj
                    read(unitno,trim(readfmt),advance='no') tmp
                    var(i + ((j-1) + (k-1)*nj)*ni) = tmp==1 ! recover logical value from {0,1}
                    if (j<nj) read(unitno,'(1X)',advance='no')
                end do
                read(unitno,*) 
            end do
        end do

        close(unitno)

        return

        ! If there was an error reading, say so here
999     write(0,*) 'ERROR: csvread: failed to read ',trim(dir)//str//'.csv'

    end subroutine

    subroutine test_csv()
        implicit none
        real(8) :: d0,d0_
        real(8), allocatable :: d1(:),d2(:,:),d3(:,:,:),d4(:,:,:,:)
        real(8), allocatable :: d1_(:),d2_(:,:),d3_(:,:,:),d4_(:,:,:,:)
        integer :: i0,i0_
        integer, allocatable :: i1(:),i2(:,:),i3(:,:,:),i4(:,:,:,:)
        integer, allocatable :: i1_(:),i2_(:,:),i3_(:,:,:),i4_(:,:,:,:)
        integer, parameter :: n1=7,n2=8,n3=9,n4=10
        allocate(d1(n1))
        allocate(d2(n1,n2))
        allocate(d3(n1,n2,n3))
        allocate(d4(n1,n2,n3,n4))
        allocate(i1(n1))
        allocate(i2(n1,n2))
        allocate(i3(n1,n2,n3))
        allocate(i4(n1,n2,n3,n4))
        allocate(d1_(n1))
        allocate(d2_(n1,n2))
        allocate(d3_(n1,n2,n3))
        allocate(d4_(n1,n2,n3,n4))
        allocate(i1_(n1))
        allocate(i2_(n1,n2))
        allocate(i3_(n1,n2,n3))
        allocate(i4_(n1,n2,n3,n4))

        call random_number(d0)
        call random_number(d1)
        call random_number(d2)
        call random_number(d3)
        call random_number(d4)
        i0 = nint((2d0*d0-1d0)*huge(1))
        i1 = nint((2d0*d1-1d0)*huge(1))
        i2 = nint((2d0*d2-1d0)*huge(1))
        i3 = nint((2d0*d3-1d0)*huge(1))
        i4 = nint((2d0*d4-1d0)*huge(1))
        d0 = (2d0*d0-1d0)*huge(0d0)
        d1 = (2d0*d1-1d0)*huge(0d0)
        d2 = (2d0*d2-1d0)*huge(0d0)
        d3 = (2d0*d3-1d0)*huge(0d0)
        d4 = (2d0*d4-1d0)*huge(0d0)


        call usedir('/tmp/')
        call csvwrite('d0',d0)
        call csvwrite('d1',d1)
        call csvwrite('d2',d2)
        call csvwrite('d3',d3)
        call csvwrite('d4',d4)
        call csvwrite('i0',i0)
        call csvwrite('i1',i1)
        call csvwrite('i2',i2)
        call csvwrite('i3',i3)
        call csvwrite('i4',i4)

        d0_ = d0
        d1_ = d1
        d2_ = d2
        d3_ = d3
        d4_ = d4
        i0_ = i0
        i1_ = i1
        i2_ = i2
        i3_ = i3
        i4_ = i4

        d0 = 0d0
        d1 = 0d0
        d2 = 0d0
        d3 = 0d0
        d4 = 0d0
        i0 = 0
        i1 = 0
        i2 = 0
        i3 = 0
        i4 = 0

        call csvread('d0',d0)
        call csvread('d1',d1)
        call csvread('d2',d2)
        call csvread('d3',d3)
        call csvread('d4',d4)
        call csvread('i0',i0)
        call csvread('i1',i1)
        call csvread('i2',i2)
        call csvread('i3',i3)
        call csvread('i4',i4)

        if (d0 /= d0_) STOP 'test_csv: failed test d0'
        if (any(d1 /= d1_)) STOP 'test_csv: failed test d1'
        if (any(d2 /= d2_)) STOP 'test_csv: failed test d2'
        if (any(d3 /= d3_)) STOP 'test_csv: failed test d3'
        if (any(d4 /= d4_)) STOP 'test_csv: failed test d4'
        if (i0 /= i0_) STOP 'test_csv: failed test i0'
        if (any(i1 /= i1_)) STOP 'test_csv: failed test i1'
        if (any(i2 /= i2_)) STOP 'test_csv: failed test i2'
        if (any(i3 /= i3_)) STOP 'test_csv: failed test i3'
        if (any(i4 /= i4_)) STOP 'test_csv: failed test i4'

        print*,'test_csv: passed'

    end subroutine 

end module mod_csv

! program main
!     use mod_csv
!     implicit none
!     real(8) :: A(7,5,2)
!     integer :: I(3,2),M(3,2,4)
!     logical :: L(1,2,3)
! 
!     A = 0d0
!     A(1,1,1) = 1d0
!     A(2,1,1) = 2d0
!     A(3,1,1) = huge(0d0)
!     A(4,1,1) = tiny(0d0)
!     A(1,2,1) = 3d0
!     A(6,3,2) = 4d0
! 
!     I = 0d0
!     I(1,1) = 1
!     I(2,1) = 2
!     I(3,1) = huge(0)
!     I(1,2) = -huge(0)
! 
!     L = .true.
!     L(1,2,1) = .false.
!     L(1,1,3) = .false.
! 
! 
!     call csvwrite('A0',A(1,1,1))
!     call csvwrite('A1',A(:,1,1))
!     call csvwrite('A2',A(:,:,1))
!     call csvwrite('A3',A)
! 
!     call csvwrite('I0',I(1,1))
!     call csvwrite('I1',I(:,1))
!     call csvwrite('I2',I(:,:))
! 
!     call csvwrite('L3',L)
! 
!     A = -1d0
!     call csvread('A1',A(:,1,1))
!     call csvwrite('A1check',A(:,1,1))
!     call csvwrite('A2check',A(:,:,1))
! 
!     call csvread('frommatlab',M)
! 
!     print*,'M',M
! 
! end program
module mod_deriv
    use mod_types
    implicit none
contains
    ! Computes the jacobian of a function f:R^n->R^n via forward differencing
    ! This is done in a robust way using (many more function evaluations)
    ! by computing more 3 jacobians at various step sizes and then taking the
    ! average of them. 
    subroutine sub_jacobian_robust(x,f,fx,jac,eps_in)
        implicit none
        real(rt), dimension(1:), intent(in) :: x
        real(rt), dimension(1:), intent(in) :: fx ! f evaluated at x
        interface    
            function f(x) result(y)
                use mod_types
                implicit none
                real(rt), dimension(1:), intent(in) :: x
                real(rt), dimension(1:size(x)) :: y
            end function f
        end interface
        real(rt), dimension(size(x),size(x)), intent(out) :: jac
        real(rt), optional, intent(in) :: eps_in
        ! local
        real(rt), allocatable :: jactmp(:,:)
        real(rt) :: eps

        eps = 1e-6_rt; if (present(eps_in)) eps = eps_in

        allocate(jactmp(size(x),size(x)))

        call sub_jacobian(x,f,fx,jactmp,eps)
        jac = jactmp/2e0_rt
        call sub_jacobian(x,f,fx,jactmp,eps*100e0_rt)
        jac = jac + jactmp/2e0_rt

        deallocate(jactmp)

    end subroutine


    ! Computes the jacobian of a function f:R^n->R^n via forward differencing
    ! The logic behind the tolerances and flow is from MINPACK's fdjac1
    ! "eps_in" should be roughly the amount of error in the calculation of f
    !
    ! The jacobian is of the form
    ! [df_1/dx_1 df_1/dx_2 ... df_1/dx_n]
    ! [df_2/dx_1 df_2/dx_2 ... df_2/dx_n]
    !  ...
    ! [df_n/dx_1 df_n/dx_2 ... df_n/dx_n]
    subroutine sub_jacobian(x,f,fx,jac,eps_in)
        implicit none
        real(rt), dimension(1:), intent(in) :: x
        real(rt), dimension(1:), intent(in) :: fx ! f evaluated at x
        interface    
            function f(x) result(y)
                use mod_types
                implicit none
                real(rt), dimension(1:), intent(in) :: x
                real(rt), dimension(1:size(x)) :: y
            end function f
        end interface
        real(rt), dimension(size(x),size(x)), intent(out) :: jac
        real(rt), optional, intent(in) :: eps_in
        ! local
        real(rt), allocatable :: wa(:), xlocal(:)
        real(rt) :: temp, h, eps
        integer :: n, i, j

        n = size(x)

        allocate(wa(n),xlocal(n))

        xlocal = x

        eps = sqrt(epsilon(0._rt)); if (present(eps_in)) eps = eps_in

        ! Now this is basically fdjac1 code
        eps = sqrt(max(eps,epsilon(0._rt)))
        do j = 1,n
            temp = xlocal(j)
            h = eps*abs(temp)
            if (h==0e0_rt) h = eps
            xlocal(j) = temp + h
            wa = f(xlocal)
            xlocal(j) = temp
            do i = 1,n
                jac(i,j) = (wa(i) - fx(i))/h
            end do
        end do

        deallocate(wa,xlocal)

    end subroutine


    ! sub_numGradHess
    ! 
    ! Compute the gradient and optionally the hessian of a function through finite differencing.
    ! There is a default step size used to try to obtain a good approximation. It is fairly small.
    ! However, for non or *almost* differentiable problems, this step size can be scaled up to make
    ! the steps larger. A "scaling" of 1 is the default, but the step size increases linearly in 
    ! scaling.
    !
    subroutine sub_numGradHess(x0,f,scaling,grad,hess,f0)
        implicit none
        real(rt), intent(in) :: x0(1:), scaling(1:)
        interface 
            function f(x) 
                use mod_types
                implicit none
                real(rt), dimension(1:), intent(in) :: x
                real(rt) :: f
            end function f
        end interface
        real(rt), intent(out) :: grad(size(x0))
        real(rt), intent(out), optional :: f0, hess(size(x0),size(x0))
        ! local        
        real(rt), parameter :: eps12 = epsilon(1._rt)**.5e0_rt
        real(rt), parameter :: eps13 = epsilon(1._rt)**.33333333333333e0_rt
        real(rt), parameter :: eps14 = epsilon(1._rt)**.25e0_rt
        real(rt) :: f0val
        real(rt), dimension(size(x0)) :: h,ei,ej, fxph
        integer :: i,j,n

        ! Get dim
        n = size(x0)

        h = scaling * max(abs(x0),1e0_rt) * eps12 ! * 2e0_rt*sqrt(epsilon(0e0_rt))/10e0_rt! eps14

        f0val = f(x0)
        if (present(f0)) then
            f0 = f0val
        end if

        ! Compute the gradient
        do i = 1,n
            ei = 0e0_rt
            ei(i) = h(i)
            fxph(i) = f(x0+ei)
        end do
        grad = (fxph-f0val)/h

        ! Compute the hessian

        ! Choose the new step size for the second derivative calculation. I have not done the error
        ! analysis to find the optimal, but this is the analagous one from the equal step size
        h = scaling * max(abs(x0),1e0_rt) * eps14

        if (present(hess)) then

            ! According to my calculations based on Taylor expansions,
            ! the formula for a cross partial d2f/didj using the 5 points f(x), f(x+-ei+-ej)
            ! puts 0 weight on f(x), 1/(4/(hi*hj)) where hi and hj are the step sizes. 
            do i = 1,n
                ei = 0e0_rt
                ei(i) = h(i)
                ! Handle the "special" i==j case (b/c can use f0val) separately
                hess(i,i) = (  f(x0+2e0_rt*ei) + f(x0-2e0_rt*ei) - 2e0_rt*f0val  )/(4e0_rt*h(i)**2)
                do j = i+1,n
                    ej = 0e0_rt
                    ej(j) = h(j)
                    hess(i,j) = ( f(x0+ei+ej) + f(x0-ei-ej) - f(x0+ei-ej) - f(x0-ei+ej) ) / (4e0_rt*h(i)*h(j)) 
                end do
            end do

            ! Fill out the symmetric hessian
            do i = 1,n
                do j = i+1,n
                    hess(j,i) = hess(i,j)
                end do
            end do
            
        end if

    end subroutine sub_numGradHess

! 
!     subroutine sub_numGradHess(x0,f,scaling,grad,hess,f0)
!         implicit none
!         real(rt), intent(in) :: x0(1:), scaling
!         interface 
!             function f(x) 
!                 implicit none
!                 real(rt), dimension(1:), intent(in) :: x
!                 real(rt) :: f
!             end function f
!         end interface
!         real(rt), intent(out) :: grad(size(x0))
!         real(rt), intent(out), optional :: f0, hess(size(x0),size(x0))
!         ! local        
!         real(rt), parameter :: eps13 = epsilon(1e0_rt)**.33333333333333e0_rt
!         real(rt), parameter :: eps14 = epsilon(1e0_rt)**.25e0_rt
!         real(rt) :: f0val
!         real(rt), dimension(size(x0)) :: h,ei,ej
!         integer :: i,j,n
! 
!         ! Get dim
!         n = size(x0)
! 
!         h = scaling * max(abs(x0),1e0_rt) * eps13
! 
!         ! Compute the gradient
!         do i = 1,n
!             ei = 0e0_rt
!             ei(i) = h(i)
!             grad(i) = (f(x0+ei)-f(x0-ei))/(2e0_rt*h(i))
!         end do
! 
!         ! Optionally compute the value at f0 and the hessian
!         if (present(f0)) then
!             f0 = f(x0)
!             f0val = f0
!         elseif (present(hess)) then
!             f0val = f(x0)
!         end if
! 
!         ! Choose the new step size for the second derivative calculation. I have not done the error
!         ! analysis to find the optimal, but this is the analagous one from the equal step size
!         h = scaling * max(abs(x0),1e0_rt) * eps14
! 
!         if (present(hess)) then
! 
!             ! According to my calculations based on Taylor expansions,
!             ! the formula for a cross partial d2f/didj using the 5 points f(x), f(x+-ei+-ej)
!             ! puts 0 weight on f(x), 1/(4/(hi*hj)) where hi and hj are the step sizes. 
!             do i = 1,n
!                 ei = 0e0_rt
!                 ei(i) = h(i)
!                 ! Handle the "special" i==j case (b/c can use f0val) separately
!                 hess(i,i) = (  f(x0+2e0_rt*ei) + f(x0-2e0_rt*ei) - 2e0_rt*f0val  )/(4e0_rt*h(i)**2)
!                 do j = i+1,n
!                     ej = 0e0_rt
!                     ej(j) = h(j)
!                     hess(i,j) = ( f(x0+ei+ej) + f(x0-ei-ej) - f(x0+ei-ej) - f(x0-ei+ej) ) / (4e0_rt*h(i)*h(j)) 
!                 end do
!             end do
! 
!             ! Fill out the symmetric hessian
!             do i = 1,n
!                 do j = i+1,n
!                     hess(j,i) = hess(i,j)
!                 end do
!             end do
!             
!         end if
! 
!     end subroutine sub_numGradHess


end module mod_deriv
! Author: Grey Gordon
! Year: 2011
!
! A Fortran implementation of the matlab Smolyak routines on my website.
! 
! All rights reserved. Please do not redistribute without permission.
!
! Soli Deo Gloria
!
! step1: get collocation points and compute many things
! step2: get coefficients of polynomial
! step3: evaluate polynomial at points x
! 
! Optionally, step3 can be broken into two parts
! step3A: get a work array of size that depends on x
! step3B: use the work array to evaluate the polynomial
module mod_smolyak
    use mod_types

    private
    public :: smolyakFcn, smolyakapprox_step1, smolyakapprox_step2, smolyakapprox_step3, &
              smolyakapprox_step3A, smolyakApprox_step3B, test_smolyak, equals, &
              smolyakGetWorkRows, smolyakGetWorkCols


    real(rt), parameter, private :: pi = 3.141592653589793238462643383279e0_rt
    
    type smolyakFcn
        integer :: d, mu, q, leni, napprox, work_array_cols
        integer, allocatable :: ibar(:)
        integer, allocatable :: i(:,:), j(:,:), constVec(:), redo(:), undo(:)
!         integer, allocatable :: j_eq_2(:,:), j_eq_3(:,:), j_eq_4(:,:), j_eq_5(:,:), j_gt_5(:,:)
        integer, allocatable :: Lb(:), Ub(:) 
        real(rt), allocatable :: a(:), b(:), Tovercjprod(:,:), cjprod(:), const(:)
        real(rt), allocatable :: constTimesTheta(:,:) ! Polynomial coefficients
        integer :: laststep = -1
    end type smolyakFcn

    interface assignment (=)
        module procedure equals
    end interface

    interface smolyakapprox_step2
        module procedure smolyakapprox_step2_vec, smolyakapprox_step2_mat 
    end interface smolyakapprox_step2
    interface smolyakapprox_step3
        module procedure smolyakapprox_step3_vec, smolyakapprox_step3_mat, &
                         smolyakapprox_step3_vec_evalpt, smolyakapprox_step3_mat_evalpt
    end interface smolyakapprox_step3
    interface smolyakapprox_step3A
        module procedure smolyakapprox_step3A_vec, smolyakapprox_step3A_mat
    end interface smolyakapprox_step3A
!     interface smolyakapprox_step4
!         module procedure smolyakapprox_step4_core
!     end interface smolyakapprox_step4

contains

! Provides the size
function smolyakGetWorkRows(s,nx) result(nrows)
    implicit none
    type(smolyakFcn), intent(in) :: s
    integer, intent(in) :: nx
    integer :: nrows
    nrows = nx
end function 
function smolyakGetWorkCols(s,nx) result(ncols)
    implicit none
    type(smolyakFcn), intent(in) :: s
    integer, intent(in) :: nx
    integer :: ncols
    ncols = s%work_array_cols
end function 

subroutine equals(b,a)
    implicit none
    type(smolyakFcn), intent(out) :: b
    type(smolyakFcn), intent(in) :: a
    integer :: statmsg

    if (a%laststep<=0) then
        print*,'WARNING: COPYING AN UNINITIALIZED SMOLYAKFCN DOES NOTHING'
        return
    end if
    
    ! Copy over scalars
    b%laststep = a%laststep
    b%d = a%d
    b%mu = a%mu
    b%q = a%q
    b%leni = a%leni
    b%work_array_cols = a%work_array_cols

    ! Allocate space
    allocate(b%ibar(size(a%ibar)),STAT=statmsg)
    allocate(b%i(size(a%i,1),size(a%i,2)),STAT=statmsg)
    allocate(b%j(size(a%j,1),size(a%j,2)),STAT=statmsg)
!     allocate(b%j_eq_2(size(a%j_eq_2,1),size(a%j_eq_2,2)),STAT=statmsg)
!     allocate(b%j_eq_3(size(a%j_eq_3,1),size(a%j_eq_3,2)),STAT=statmsg)
!     allocate(b%j_eq_4(size(a%j_eq_4,1),size(a%j_eq_4,2)),STAT=statmsg)
!     allocate(b%j_eq_5(size(a%j_eq_5,1),size(a%j_eq_5,2)),STAT=statmsg)
!     allocate(b%j_gt_5(size(a%j_gt_5,1),size(a%j_gt_5,2)),STAT=statmsg)
    allocate(b%constVec(size(a%constVec)),STAT=statmsg)
    allocate(b%redo(size(a%redo)),STAT=statmsg)
    allocate(b%undo(size(a%undo)),STAT=statmsg)
    allocate(b%Lb(size(a%Lb)),STAT=statmsg)
    allocate(b%Ub(size(a%Ub)),STAT=statmsg)
    allocate(b%Tovercjprod(size(a%Tovercjprod,1),size(a%Tovercjprod,2)),STAT=statmsg)
    allocate(b%cjprod(size(a%cjprod)),STAT=statmsg)
    allocate(b%const(size(a%const)),STAT=statmsg)
    allocate(b%a(size(a%a)),STAT=statmsg)
    allocate(b%b(size(a%b)),STAT=statmsg)

    ! Copy over allocatable arrays that are always allocated
    b%ibar = a%ibar
    b%i = a%i
    b%j = a%j
!     b%j_eq_2 = a%j_eq_2
!     b%j_eq_3 = a%j_eq_3
!     b%j_eq_4 = a%j_eq_4
!     b%j_eq_5 = a%j_eq_5
!     b%j_gt_5 = a%j_gt_5
    b%constVec = a%constVec
    b%redo = a%redo
    b%undo = a%undo
    b%Lb = a%Lb
    b%Ub = a%Ub    
    b%Tovercjprod = a%Tovercjprod
    b%cjprod = a%cjprod
    b%const = a%const
    b%a = a%a
    b%b = a%b

    if (allocated(a%constTimesTheta)) then
        allocate(b%constTimesTheta(size(a%constTimesTheta,1),size(a%constTimesTheta,2)))
        b%constTimesTheta = a%constTimesTheta
    end if

!     if (allocated(a%Tprod)) then
!         allocate(b%Tprod(size(a%Tprod,1),size(a%Tprod,2)),STAT=statmsg)
!         b%Tprod = a%Tprod
!     end if

end subroutine

! subroutine uniqueRowsInt(y,i,j,x) 
!     implicit none
!     integer, intent(in) :: x(1:,1:)
!     integer, allocatable, intent(out) :: y(:,:)
!     integer, allocatable, intent(out) :: i(:) !Redo
!     integer, intent(out) :: j(size(x,1)) !Undo
!     !local
!     integer :: tmp(size(x,1),size(x,2))
!     integer :: tmpLoc(size(x,1))
!     integer :: ind, nUnique, ind2
!     
!     !Find which rows are unique (current method is brute force)
!     !In future, should iteratively sort by columns
!     nUnique = 1
!     tmp(1,:) = x(1,:)
!     tmpLoc(1) = 1
!     do ind = 2,size(x,1)        
!         if (all(sum(abs(tmp(1:nUnique,:)-spread(x(ind,:),1,nUnique)),2)/=0)) then
!             nUnique = nUnique + 1
!             tmp(nUnique,:) = x(ind,:)
!             tmpLoc(nUnique) = ind
!         end if
!     end do
!     allocate(y(nUnique,size(x,2)), i(nUnique))
! !    write(*,*) nUnique
!     y = tmp(1:nUnique,:)    
!     i = tmpLoc(1:nUnique)
! 
!     !Find indices that will undo this operation -- again this seems terribly inefficient 
!     j = -1
!     do ind = 1,size(x,1)        
!         do ind2 = 1,nUnique
!             if (sum(abs(y(ind2,:)-x(ind,:)))==0) then
!                 j(ind) = ind2
!             end if
!         end do
!     end do
! 
! end subroutine
! subroutine uniqueRowsDouble(y,i,j,x) 
!     implicit none
!     real(rt), intent(in) :: x(1:,1:)
!     real(rt), allocatable, intent(out) :: y(:,:)
!     integer, allocatable, intent(out) :: i(:) !Redo
!     integer, intent(out) :: j(size(x,1)) !Undo
!     !local
!     real(rt) :: tmp(size(x,1),size(x,2))
!     integer :: tmpLoc(size(x,1))
!     integer :: ind, nUnique, ind2
!     
!     !Find which rows are unique (current method is brute force)
!     !In future, should iteratively sort by columns
!     nUnique = 1
!     tmp(1,:) = x(1,:)
!     tmpLoc(1) = 1
!     do ind = 2,size(x,1)
!         if (all(sum(abs(tmp(1:nUnique,:)-spread(x(ind,:),1,nUnique)),2)>1d-13)) then
!             nUnique = nUnique + 1
!             tmp(nUnique,:) = x(ind,:)
!             tmpLoc(nUnique) = ind
!         end if
!     end do
!     allocate(y(nUnique,size(x,2)), i(nUnique))
!     y = tmp(1:nUnique,:)    
!     i = tmpLoc(1:nUnique)
! 
!     !Find indices that will undo this operation -- again this seems terribly inefficient 
!     j = -1
!     do ind = 1,size(x,1)        
!         do ind2 = 1,nUnique
!             if (sum(abs(y(ind2,:)-x(ind,:)))<=1d-13) then
!                 j(ind) = ind2
!             end if
!         end do
!     end do
!    
! end subroutine

! This function is by Andreas Klimke, adapted to Fortran
elemental function spdimcc(n,d) result(sp)
    implicit none
    integer, intent(in) :: n,d
    integer :: sp
    ! % SPDIMCC   Compute number of sparse grid points, Clenshaw-Curtis grid
    ! %    SP = SPDIMCC(N,D) Computes the number of sparse grid points
    ! %      of the Clenshaw-Gurtis grid of dimension D and level N.
    ! %    (Internal function)

    ! % The function uses the formulas for N <= 7 given in A. Schreiber:
    ! % Die Methode von Smoliak bei der multivariaten Interpolation,
    ! % (2000), p.35/36. 
    !     
    !     % Author : Andreas Klimke, Universitaet Stuttgart
    !     % Version: 1.0
    !     % Date   : August 5, 2003

    !     % ------------------------------------------------------------
    !     % Sparse Grid Interpolation Toolbox
    !     % Copyright (c) 2006 W. Andreas Klimke, Universitaet Stuttgart 
    !     % Copyright (c) 2007-2008 W. A. Klimke. All Rights Reserved.
    !     % See LICENSE.txt for license. 
    !     % email: klimkeas@ians.uni-stuttgart.de
    !     % web  : http://www.ians.uni-stuttgart.de/spinterp
    !     % ------------------------------------------------------------

    select case (n)
        case (0)
            sp = 1
        case (1)
            sp = 2*d + 1
        case (2)
            sp = 2*d**2 + 2*d + 1
        case (3)
            sp = nint((4*d**3 + 6*d**2 + 14*d)/3e0_rt) + 1
        case (4)
            sp = nint((2*d**4 + 4*d**3 + 22*d**2 + 20*d)/3e0_rt) + 1
        case (5)
            sp = nint((4*d**5 + 10*d**4 + 100*d**3 + 170*d**2 + 196*d)/15e0_rt) + 1
        case (6)
            sp = nint((4*d**6 + 12*d**5 + 190*d**4 + 480*d**3 + 1246*d**2 + 948*d)/45e0_rt) + 1
        case (7)
            sp = nint((8*d**7 + 28*d**6 + 644*d**5 + 2170*d**4 + 9632*d**3 + 15442*d**2 + 12396*d)/315e0_rt) + 1
        case default 
            sp = -1 ! should crash the program
    !         STOP 'dimensionality case if too high, have not setup the routine for this'
    !         sp = nint((8*d^7 + 28*d^6 + 644*d^5 + 2170*d^4 + 9632*d^3 + 15442*d^2 + 12396*d)/315) + 1
    !         do m = 8:n
    !             seq = spgetseq(m,d);
    !             do k = 1:size(seq,1)
    !                 temp = 1;
    !                 do l = 1:d
    !                     lval = seq(k,l);
    !                     if (lval == 0) then 
    !                     elseif (lval < 3) then
    !                         temp = temp * 2;
    !                     else
    !                         temp = temp * 2^double(lval-1);
    !                     endif 
    !                 end do
    !                 sp = sp + temp;
    !             end do
    !         end do
    end select

end function 

! smolyakEnumerate2
! Given d and some ibar, finds all multi-indices i>=1 s.t. |i|=ibar.
! Just a wrapper for spgetseq
subroutine smolyakEnumerate2(i,d,ibar)
    use mod_matlab, only: pure_choose
    implicit none
    integer, intent(in) :: d,ibar
    integer, intent(out) :: i(pure_choose(ibar-1,d-1),d)

    if (ibar<d) STOP 'ERROR: smolyakEnumerate2: no such combination exists'

    call spgetseq(i,ibar-d,d)
    i = i+1

end subroutine smolyakEnumerate2

! An adaptation of spgetseq by A. Klimke
subroutine spgetseq(seq,n,d)
    use mod_matlab, only: pure_choose, disp
    implicit none
    integer, intent(in) :: n,d
    integer :: seq(pure_choose(n+d-1,d-1),d)
    integer :: nlevels,sum,max,k,l,m,temp
    ! % SPGETSEQ  Compute the sets of indices 
    ! %    (internal function)
    ! 
    ! % Author : Andreas Klimke, Universitaet Stuttgart
    ! % Version: 1.1
    ! % Date   : November 18, 2007
    ! 
    ! % Change log:
    ! % V1.0   : August 5, 2003
    ! %          Initial release
    ! % V1.1   : January 24, 2006
    ! %          Changed algorithm to operate on a uint8 rather than
    ! %          a double array.
    ! % V1.2   : November 18, 2007
    ! %          Added new grid type : Gauss-Patterson
    ! 
    ! % ------------------------------------------------------------
    ! % Sparse Grid Interpolation Toolbox
    ! % Copyright (c) 2006 W. Andreas Klimke, Universitaet Stuttgart 
    ! % Copyright (c) 2007-2008 W. A. Klimke. All Rights Reserved.
    ! % See LICENSE.txt for license. 
    ! % email: klimkeas@ians.uni-stuttgart.de
    ! % web  : http://www.ians.uni-stuttgart.de/spinterp
    ! % ------------------------------------------------------------
    ! 	
    ! % Examples:
    ! % For dimension 2, level n = 3, the sequence is
    ! %     3     0
    ! %     2     1
    ! %     1     2
    ! %     0     3
    ! %
    ! % For dimension 3, level n = 3, the sequence is
    ! %     3     0     0
    ! %     2     1     0
    ! %     1     2     0
    ! %     0     3     0
    ! %     2     0     1
    ! %     1     1     1
    ! %     0     2     1
    ! %     1     0     2
    ! %     0     1     2
    ! %     0     0     3

    nlevels = pure_choose(n+d-1,d-1)

    seq = 0

    seq(1,1) = n 
    max = n 
    do k = 2,nlevels
        if (seq(k-1,1) > 0) then
            seq(k,1) = seq(k-1,1) - 1 
            do l = 2,d
                if (seq(k-1,l) < max) then
                    seq(k,l) = seq(k-1,l) + 1 
                    do m = l+1,d
                        seq(k,m) = seq(k-1,m) 
                    end do
                    exit
                end if
            end do
        else
            sum = 0
            do l = 2,d
                if (seq(k-1,l) < max) then
                    seq(k,l) = seq(k-1,l) + 1 
                    sum = sum + seq(k,l) 
                    do m = l+1,d
                        seq(k,m) = seq(k-1,m) 
                        sum = sum + seq(k,m) 
                    end do
                    exit
                else
                    temp = 0 
                    do m = l+2,d
                        temp = temp + seq(k-1,m) 
                    end do
                    max = n-temp 
                    seq(k,l) = 0 
                end if 
            end do
            seq(k,1) = n - sum 
            max = n - sum 
        end if
    end do

end subroutine spgetseq



!smolyakapprox_step1 smolyak grid construction 
!   [x s] = smolyakapprox_step1(d,mu,a,b) constructs the smolyak
!   grid points for a hypercube in r^d defined by bounds a,b with level of 
!   approximation mu.  it outputs two elementts, x and s.  x is
!   the collocation points that your function should be evaluated.
!   s is a structure containing information used by
!   smolyakapprox_step2 and smolyakapprox_step3.
!   copyright grey gordon
!   date: 1/18/11
!   last revised: 7/22/13
!function [x s] = smolyakapprox_step1(d,mu,a,b)
subroutine smolyakapprox_step1(x,s,d,mu,a,b)
    use mod_matlab
    implicit none
    integer, intent(in) :: d,mu
    real(rt), intent(in) :: a(1:),b(1:)
    real(rt), intent(out), allocatable :: x(:,:) !Note b/c of intent(out) this is deallocated upon entry
    type(smolyakFcn), intent(out) :: s
    !local
    integer, allocatable :: k(:,:)
    real(rt), allocatable :: tmpMat(:,:),z(:,:)
    integer :: q, ibar, tmpi, tmpn, iind, dind, lind, jind
    real(rt) :: tmpProd, tmpSum

    ! Check inputs 
    if (size(a)/=size(b) .or. size(a)/=d) then
        STOP 'ERROR: smolyakapprox_step1: a and b must have length d'
    end if
    if (any(a>=b)) then
        STOP 'ERROR: smolyakapprox_step1: a must be strictly less than b'
    end if
    if (mu<1) STOP 'ERROR: smolyakapprox_step1: mu should be >= 1'
    if (d<1) STOP 'ERROR: smolyakapprox_step1: d should be >= 1' 

    s%laststep = 1

    q = max(d,mu+1)

    s%q = q
    s%d = d
    s%mu = mu
    
    allocate(s%a(d),s%b(d))
    s%a = a
    s%b = b    

    ! Construct necessary coefficients using fx values.
    ! first, enumerate all necessary theta
    ! -for all i satisfying q<=ibar<=d+mu, 
    ! -need { l | l(1) is in 1...m(i1), l(2) is in 1...m(i1)}
    
    !enumerate all such unique i, then all such m(i), then all such l
    ! determine all i that fit the criterion q<=|i|<=d+mu
    !!! Determine the number of such i
    tmpi = 0
    do ibar = q,d+mu
        tmpi = tmpi + choose(ibar-1,d-1)
    end do

    !!! Allocate space and get all such i
    allocate(s%i(tmpi,d))
    tmpi = 0
    do ibar = q,d+mu 
        tmpn = choose(ibar-1,d-1)
        call smolyakEnumerate2(s%i(tmpi+1:tmpi+tmpn,:),d,ibar)
        tmpi = tmpi + tmpn
    end do
    
    s%leni = size(s%i,1)
    
    allocate(k(s%leni,d))
    allocate(s%ibar(s%leni),s%Lb(s%leni),s%Ub(s%leni))

    s%ibar = sum(s%i,2)
    k = smolyakM(s%i)

    ! Lb(i) and Ub(i) give the starting and ending range for 
    ! all the cartesian products of G(m(i1))x...xG(m(id))
    tmpi = 0
    do iind = 1,s%leni        
        s%Lb(iind) = tmpi + 1
        tmpi = tmpi + product(k(iind,:))
        s%Ub(iind) = tmpi
    end do    

    allocate(x(s%Ub(s%leni),d),z(s%Ub(s%leni),d),s%j(s%Ub(s%leni),d))
    allocate(s%undo(s%Ub(s%leni)))

    ! Compute the cartesian products of the points
    !!! First, get their indices
    do iind = 1,s%leni
        call cartProdInt(s%j(s%Lb(iind):s%Ub(iind),:),k(iind,:))
    end do

    !!! Second, get their numeric values
    allocate(tmpMat(maxval(k),maxval(k))) ! Not sure the bounds on this are correct
    do tmpi = 1,maxval(k)
        call smolyakG(tmpMat(1:tmpi,tmpi),tmpi)
    end do
    do iind = 1,s%leni
        do dind = 1,d
            do tmpi = s%Lb(iind),s%Ub(iind) 
                z(tmpi,dind) = tmpMat(s%j(tmpi,dind),k(iind,dind))
            end do
        end do
    end do

    !!! Third, get the unique values (b/c of their nested nature, only return unique values)
    call unique(x,s%redo,s%undo,z,'rows')

    !!! Last, scale to [a,b], this gives the Sparse grid points
    do dind = 1,d
        x(:,dind) = (x(:,dind)+1.e0_rt)*(b(dind)-a(dind))/2.e0_rt + a(dind)
    end do


    !!! PRECOMPUTE SOME THINGS FOR SUBSEQUENT ROUTINES

    !To ease computation, we compute the values of t which are evaluated at roots in formula (25)
    allocate(s%Tovercjprod(s%Ub(s%leni),maxval(s%Ub-s%Lb)+1))
    allocate(s%cjprod(size(s%undo)))
    allocate(s%const(s%leni))

    ! For each l/j, there is a unique cjprod combination c_j1*c_j2*...*c_jd where c_1=c_kd=2 and c_i=1 o/w
    s%cjprod = -100.e0_rt
    do iind = 1,s%leni
        do lind = s%Lb(iind),s%Ub(iind)
            s%cjprod(lind) = 1.e0_rt
            do dind = 1,d
                ! Dimensions with only one dimensions are "dropped" according to Krueger and Kubler
                if (k(iind,dind)>1) then
                    ! The product is 2**(l==1 or l==k)
                    if ((s%j(lind,dind)==1) .or. (s%j(lind,dind)==k(iind,dind)))  then
                        s%cjprod(lind) = s%cjprod(lind)*2.e0_rt
                    end if
                end if
            end do
        end do
    end do

    ! To construct the tensor product coefficients theta_l, we need T_l1(zeta_j1)*...T_ld(zeta_j2) for each l,j combination    
    ! That is why T is quite large. However, we don't actually need to save T, but only T / cjprod
    do iind = 1,s%leni
        do lind = s%Lb(iind),s%Ub(iind)
            do jind = s%Lb(iind),s%Ub(iind)
                tmpProd = 1.e0_rt
                do dind = 1,d
                    if (k(iind,dind)>1) then
                        if (mod(s%j(lind,dind),2)==1) then
                            tmpProd = tmpProd*cos(pi*real((s%j(lind,dind)-1)*(s%j(jind,dind)-1),rt)/real(k(iind,dind)-1,rt))
                        else
                            tmpProd = -tmpProd*cos(pi*real((s%j(lind,dind)-1)*(s%j(jind,dind)-1),rt)/real(k(iind,dind)-1,rt))
                        end if
                    end if
                end do
                s%Tovercjprod(jind,lind-s%Lb(iind)+1) = tmpProd
            end do
        end do        
    end do
    do tmpi = 1,size(s%Tovercjprod,2)
        s%Tovercjprod(:,tmpi) = s%Tovercjprod(:,tmpi)/s%cjprod(:)
    end do

    ! For each i, there is a unique constant 2^dhat/((k_1-1)*...(k_dhat-1))/(c_l1*...*c_ldhat) 
    ! The first part, before "cjprod" (c_l1*...*c_ldhat), is "const".
    ! where dhat is the number of dimensions that has k=m(i)>1 and 1,...,dhat is a reordering of
    ! dimensions 1,...,d that drops dimensions with k=m(i)=1. 
    do iind = 1,s%leni
        tmpProd = 1e0_rt ! (k_1-1)*...*(k_dhat-1)
        tmpSum = 0e0_rt ! dhat
        do dind = 1,d
            if (k(iind,dind)>1) then
                tmpSum = tmpSum + 1._rt
                tmpProd = tmpProd*real(k(iind,dind)-1,rt)
            end if
        end do
        s%const(iind) = (2e0_rt**tmpSum)/tmpProd
    end do

    ! Store the constant values associated with the Smolyak hierarchy
    ! These are (-1)^(d+mu-|i|)*choose(d-1,d+\mu-|i|)
    allocate(s%constVec(d+mu+1-q))
    do ibar = q,d+mu
        s%constVec(ibar-q+1) = (-1)**(d+mu-ibar)*choose(d-1,d+mu-ibar)
    end do

    ! Set sizes for work array
    s%work_array_cols = size(s%undo)

    ! Precompute, for each dind, the indices such that s.j(:,dind)==x or >x. This speeds computation
    ! in step3.
!     allocate(s%j_eq_2(count(s%j(:,1)==2),d))
!     allocate(s%j_eq_3(count(s%j(:,1)==3),d))
!     allocate(s%j_eq_4(count(s%j(:,1)==4),d)) ! may be zero sized
!     allocate(s%j_eq_5(count(s%j(:,1)==5),d)) ! may be zero sized
!     allocate(s%j_gt_5(count(s%j(:,1)>5),d)) ! may be zero sized
!     do dind = 1,d
!         s%j_eq_2(:,dind) = find(s%j(:,dind)==2)
!         s%j_eq_3(:,dind) = find(s%j(:,dind)==3)
!         if (size(s%j_eq_4,1)>0) s%j_eq_4(:,dind) = find(s%j(:,dind)==4)
!         if (size(s%j_eq_5,1)>0) s%j_eq_5(:,dind) = find(s%j(:,dind)==5)
!         if (size(s%j_gt_5,1)>0) s%j_gt_5(:,dind) = find(s%j(:,dind)>5)
!     end do

    ! What we need going forward is the terms
    ! phi(l,i) := (-1)^(d+mu-|i|)*choose(d-1,d+\mu-|i|)*2^dhat/((k_1-1)*...(k_dhat-1))/(c_l1*...*c_ldhat) for each i
    ! The smolyak formula can be written as 
    ! \sum_{q<=|i|<=d+mu} \sum_l\in L(i) phi(l,i) 

end subroutine smolyakapprox_step1


! This routine takes as given the function values from the collocation points and constructs a polynomial approximation.
! Several sets of function values at collocation points are passed (organized so that each column contains a set of 
! collocation point values) and so several polynomial approximation are constructed, one for each column.
subroutine smolyakapprox_step2_vec(fx,s,tF)
    implicit none
    real(rt), dimension(1:), intent(in) :: fx
    type(smolyakFcn), intent(inout) :: s
    character(len=1), intent(in), optional :: tF

    if (present(tF)) print*,'WARNING: smolyakapprox_step2: when only approximating one function, tF does nothing'

    if (size(fx,1)/=size(s%redo)) STOP 'function values at collocation points must be along first dimension'
    s%napprox = 1

    call smolyakapprox_step2_core(fx,s,size(s%redo),'N') ! Since one dimensional , 

end subroutine smolyakapprox_step2_vec

subroutine smolyakapprox_step2_mat(fx,s,tF)
    implicit none
    real(rt), dimension(1:,1:), intent(in) :: fx
    type(smolyakFcn), intent(inout) :: s
    character(len=1), intent(in), optional :: tF
    ! local
    character(len=1) :: tFselect

    tFselect = 'N'; if (present(tF)) tFselect = tF

    if (tFselect=='N') then
        if (size(fx,1)/=size(s%redo)) STOP 'function values at colloc points must be along first dimension'
        s%napprox = size(fx,2)
    else
        if (size(fx,2)/=size(s%redo)) STOP 'function values at colloc points must be along second dimension'
        s%napprox = size(fx,1)
    end if

    call smolyakapprox_step2_core(fx,s,size(s%redo),tFselect)

end subroutine smolyakapprox_step2_mat

! This is the core routine. Normally, t='N' and then the approximated functions are the columns of fx. 
! However, if t='T', then the approximated functions are the rows of fx.
subroutine smolyakapprox_step2_core(fx,s,ncolloc,tFselect)
    use mod_matlab
    implicit none
    integer, intent(in) :: ncolloc
    type(smolyakFcn), intent(inout) :: s
    real(rt), dimension(*), intent(in), target :: fx 
    character(len=1), intent(in) :: tFselect
    ! local
    integer :: iind,lind,tmpi !,iapprox
    real(rt) :: tmp
    real(rt), pointer :: fxmat(:,:)
    real(rt), allocatable :: funpackedtrans(:,:) ! ,funpacked(:,:)

    if (allocated(s%constTimesTheta)) deallocate(s%constTimesTheta)
    allocate(s%constTimesTheta(s%napprox,size(s%undo)))
!     allocate(funpacked(size(s%undo),s%napprox))
    allocate(funpackedtrans(s%napprox,size(s%undo)))

    if (s%laststep<=0) STOP 'smolyakapprox_step3: step1 should be called before step2' 
    s%laststep = 2

    if (tFselect=='N') then 

        ! For convenience, make a fxmat out of fx
        fxmat(1:ncolloc,1:s%napprox) => fx(1:ncolloc*s%napprox)

        ! Unpack the function values to the sparse grid that has redundant values
    !     funpacked = fxmat(s%undo,:)
    !     funpackedtrans = transpose(funpacked)
        do tmpi = 1,size(s%undo)
            funpackedtrans(:,tmpi) = fxmat(s%undo(tmpi),:)
        end do
!         funpackedtrans = fxmat(s%undo,:)
    
    else
        
        fxmat(1:s%napprox,1:ncolloc) => fx(1:ncolloc*s%napprox)
        funpackedtrans = fxmat(:,s%undo)

    end if

!     print*,'size funpacked trans',shape(funpackedtrans)

    !Construct the polynomial coefficients theta_l and multiply by the constant terms (-1)^(d+mu-|i|)*choose(d-1,d+\mu-|i|)
!     do iapprox = 1,s%napprox
!         do iind = 1,s%leni
!             do lind = s%Lb(iind),s%Ub(iind)
!                 s%constTimesTheta(iapprox,lind) = (s%constVec(s%ibar(iind)-s%q+1)*s%const(iind)/s%cjprod(lind))*   &
!                                        dot_product(funpacked(s%Lb(iind):s%Ub(iind),iapprox),&
!                                                    s%Tovercjprod(s%Lb(iind):s%Ub(iind),lind-s%Lb(iind)+1))
!                                                  ! s%T(s%Lb(iind):s%Ub(iind),lind-s%Lb(iind)+1)/s%cjprod(s%Lb(iind):s%Ub(iind))) 
!             end do
!         end do
!     end do
    do iind = 1,s%leni
        associate (a=>s%Lb(iind),b=>s%Ub(iind))
        do lind = a,b
            tmp = s%constVec(s%ibar(iind)-s%q+1)*s%const(iind)/s%cjprod(lind) ! just a scalar
!             s%constTimesTheta(:,lind) = tmp*matmul(s%Tovercjprod(a:b,lind-a+1),funpacked(a:b,:))
!             s%constTimesTheta(:,lind) = tmp*matmul(s%Tovercjprod(a:b,lind-a+1),transpose(funpackedtrans(:,a:b)))
!             s%constTimesTheta(:,lind) = tmp*matmul(s%Tovercjprod(a:b,lind-a+1),transpose(funpackedtrans(:,a:b)))
            call mygemm(s%constTimesTheta(:,lind),s%Tovercjprod(a:b,lind-a+1),funpackedtrans(:,a:b),&
                         1,b-a+1,s%napprox,b-a+1,tB='T',alpha=tmp)
        end do
        end associate
    end do

end subroutine smolyakapprox_step2_core

! smolyakapprox_step3
! 
! Evaluates the approximations made in step2.  This particular routine applies if multiple approximations were made.
! If s was created with multiple functions, then it will be evaluated with them. 
! x should be #x1 by d  where #x1 is up to the user
! feval should be #x1 by napprox where napprox is the number of functions that were to be approximated
subroutine smolyakapprox_step3_vec_evalpt(feval,s,x) 
    implicit none
    type(smolyakFcn), intent(in) :: s
    real(rt), dimension(1:), intent(in) :: x
    real(rt), intent(out) :: feval
    ! local
    real(rt) :: tmp(1)

    if (size(x)/=s%d) STOP 'smolyakapprox_step3: x has the wrong size. Since it is a vector, it should have length s%d'

    call smolyakapprox_step3_core(tmp,s,x,1,1)
    feval = tmp(1)

end subroutine smolyakapprox_step3_vec_evalpt

subroutine smolyakapprox_step3_vec(feval,s,x) 
    implicit none
    type(smolyakFcn), intent(in) :: s
    real(rt), dimension(1:,1:), intent(in) :: x
    real(rt), dimension(1:), intent(inout) :: feval

    if (size(feval)/=size(x,1)) STOP 'smolyakapprox_step3: feval dim 2 is incorrect.  Wrong # of x points being evaluated'
    if (size(x,2)/=s%d) STOP 'smolyakapprox_step3: x dim 2 is incorrect, it should be s%d'

    call smolyakapprox_step3_core(feval,s,x,size(x,1),1)

end subroutine smolyakapprox_step3_vec

subroutine smolyakapprox_step3_mat_evalpt(feval,s,x) 
    implicit none
    type(smolyakFcn), intent(in) :: s
    real(rt), dimension(1:), intent(in) :: x
    real(rt), dimension(1:), intent(inout) :: feval

    if (size(feval)/=s%napprox) STOP 'smolyakapprox_step3: size of feval is incorrect.  Inconsistent # of fcns being approximated'
    if (size(x)/=s%d) STOP 'smolyakapprox_step3: x size is incorrect, it should be s%d'

    call smolyakapprox_step3_core(feval,s,x,1,size(feval))

end subroutine smolyakapprox_step3_mat_evalpt

subroutine smolyakapprox_step3_mat(feval,s,x) 
    implicit none
    type(smolyakFcn), intent(in) :: s
    real(rt), dimension(1:,1:), intent(in) :: x
    real(rt), dimension(1:,1:), intent(inout) :: feval

    if (size(feval,1)/=size(x,1)) STOP 'smolyakapprox_step3: feval dim 2 is incorrect.  Wrong # of x points being evaluated'
    if (size(feval,2)/=s%napprox) STOP 'smolyakapprox_step3: feval dim 1 is incorrect.  Inconsistent # of fcns being approximated'
    if (size(x,2)/=s%d) STOP 'smolyakapprox_step3: x dim 2 is incorrect, it should be s%d'

    call smolyakapprox_step3_core(feval,s,x,size(x,1),size(feval,2))

end subroutine smolyakapprox_step3_mat


! step3A: just get Tprod
subroutine smolyakapprox_step3A_vec(work,s,x) 
    implicit none
    real(rt), intent(in) :: x(1:)
    type(smolyakFcn), intent(in) :: s
    real(rt), intent(out) :: work(1,s%work_array_cols)

    if (size(x)/=s%d) STOP 'smolyakapprox_step3A: x is wrongly size'

    call smolyakapprox_step3A_core(work,s,x,1)

end subroutine smolyakapprox_step3A_vec
subroutine smolyakapprox_step3A_mat(work,s,x) 
    implicit none
    real(rt), intent(in) :: x(1:,1:)
    type(smolyakFcn), intent(in) :: s
    real(rt), intent(out) :: work(size(x,1),s%work_array_cols)

    if (size(x,2)/=s%d) STOP 'smolyakapprox_step3A: x is wrongly size'

    call smolyakapprox_step3A_core(work,s,x,size(x,1))

end subroutine smolyakapprox_step3A_mat

! NOTE: this routine only access components of s that are defined in 
! step1. Consequently, it can be called before step2 is called.
subroutine smolyakapprox_step3A_core(Tprod,s,xin,nx) 
    implicit none
    integer, intent(in) :: nx
    real(rt), dimension(*), intent(in), target :: xin
    type(smolyakFcn), intent(in) :: s
    real(rt), intent(out) :: Tprod(nx,s%work_array_cols)
    !
    real(rt), pointer :: x(:,:)
    real(rt), allocatable :: twoz(:),T(:,:)
    integer :: dind, oind, iind, lind
    
    ! Create a matrix x out of the assumed size array xin
    x(1:nx,1:s%d) => xin(1:nx*s%d)

    allocate(twoz(nx))
    allocate(T(nx,2:smolyakM(s%mu+1))) ! Note indexing in 2nd dim begins at 2

    ! The smolyak approximation can be written as 
    ! \sum_{q\le |i| \le d+mu} (-1)^{d+mu-|i|}choose(d-1,d+\mu-|i|)p^i(x)
    ! p^i(x) itself is composed of a summation across l-values (in the notation of MKK)
    ! of theta_l*T_l1(x_1)*T_l2(x_2)*...*T_ld(x_d). So, get the product across T and save
    ! as Tprod for each l.
    ! So, the summation can be written as 
    ! \sum_{q\le |i| \le d+mu} [(-1)^{d+mu-|i|}choose(d-1,d+\mu-|i|)theta_l ]*
    !                          [T_l1(x_1)*T_l2(x_2)*...*T_ld(x_d)]
    ! The first bracketed object has already been computed and saved as constTimesTheta(iapprox,l)
    ! The second has not been computed since it depends on the (transformed) x values
    ! 
    Tprod = 1._rt
    do dind = 1,s%d

        !Convert the x points to z via linear transformation
        twoz = 4._rt/(s%b(dind)-s%a(dind))*(x(:,dind)-s%a(dind)) - 2._rt ! two*z 
        
        ! Get the values of the Chebyshev basis at 
!         T(:,1) = 1e0_rt ! the first index is never accessed
        T(:,2) = .5_rt*twoz
        T(:,3) = twoz*T(:,2) - 1._rt ! don't access T(:,1)==1
        ! q<=ibar<=d+mu.  So ibar=d+mu is largest.  
        !Largest any order Cheby can be is the largest any component of m(i) can be.  
        !This occurs when i = 1 except for one component. 
        ! => icomp = d+mu-(d-1) = mu+1 => i = mu+1 => m(i) = m(mu+1);
        do oind = 4,smolyakM(s%mu+1) 
            T(:,oind) = twoz*T(:,oind-1) - T(:,oind-2)
        end do
        
!         do lind = 1,size(s%j,1)
!             if (s%j(lind,dind)>1) then
!                 Tprod(:,lind) = Tprod(:,lind)*T(:,s%j(lind,dind))
!             end if
!         end do

        do iind = 1,s%leni
            ! Test whether i>1. If not, then j = 1,...,M(i) has M(i)==1 so j<=1 implying T(:,j)==1
            if (s%i(iind,dind)>1) then
                do lind = s%Lb(iind),s%Ub(iind)
                    ! Again, test whether j>1 since can save possibly many multiplications if j==1
                    if (s%j(lind,dind)>1) then
                        Tprod(:,lind) = Tprod(:,lind)*T(:,s%j(lind,dind))
                    end if
                end do
            end if 
        end do

    end do

end subroutine smolyakapprox_step3A_core

! I've extended this to allow for a tranposed feval where, instead of having 
! nx # of rows and napprox # of columns it is switched. 
subroutine smolyakapprox_step3B(feval,s,work,tF) 
    implicit none
    type(smolyakFcn), intent(in) :: s
    real(rt), dimension(:,:), intent(inout) :: feval ! so can handle either vector or matrix
    real(rt), intent(in) :: work(1:,1:)
    character(len=1), intent(in), optional :: tF
    ! local
    character(len=1) :: tFselect

    tFselect = 'N'; if (present(tF)) tFselect = tF

    if (tFselect=='N') then
        if (size(feval,1)/=size(work,1)) STOP 'smolyakapprox_step3B: feval and work do not agree'
        if (size(feval,2)/=size(s%constTimesTheta,1)) STOP 'smolyakapprox_step3B: feval does not have rows equal to napprox'
    else
        if (size(feval,1)/=size(s%constTimesTheta,1)) STOP 'smolyakapprox_step3B: feval does not have rows equal to napprox'
        if (size(feval,2)/=size(work,1)) STOP 'smolyakapprox_step3B: feval and work do not agree'
    end if

    call smolyakapprox_step3B_core(feval,s,work,tFselect)

end subroutine smolyakapprox_step3B

subroutine smolyakapprox_step3B_core(feval,s,work,tFselect) 
    use mod_matlab
    implicit none
    type(smolyakFcn), intent(in) :: s
    real(rt), dimension(*), intent(inout) :: feval ! so can handle either vector or matrix
    real(rt), intent(in) :: work(1:,1:)
    character(len=1), intent(in) :: tFselect
    !local
    integer :: nx,napprox

    nx = size(work,1)
    napprox = size(s%constTimesTheta,1)

    if (size(work,2)/=s%work_array_cols) STOP 'smolyakapprox_step3B: work array is wrongly sized'

    !!! The rows of Tprod (work) represent different x points so nrows is size(x,1)
    !!! The columns of Tprod give T_l1(x_1)*T_l2(x_2)*....*T_ld(x_d) for a given l=(l1,l2,...,ld)
    !!! The columns of s%constTimesTheta give [(-1)^{d+mu-|i|}choose(d-1,d+\mu-|i|)theta_l ] for a given l
    !!! The rows of s%constTimesTheta represent different functions that have been approximated so # rows is s%napprox
    !!! Feval should have size(x,1) rows and s%napprox columns

!     ! SUBROUTINE DGEMM(TRANSA, TRANSB, M, N, K, ALPHA, A, LDA, B, LDB, BETA, C, LDC)
!     ! C := alpha*op(A)*op(B) + beta*C,
!     ! NOTE: M is rows of C, N is columns of C, K is # columns of B / # rows of A
!     ! Want: feval_{nx,napprox} = Tprod_{nx,size(s%undo)}*(s%constTimesTheta_{napprox,size(s%undo)}))'
!     ! If feval is tranposed, then 
!     ! Want: feval_{napprox,nx} = s%constTimesTheta_{napprox,size(s%undo)}*Tprod'
    if (tFselect=='N') then 
!         call dgemm('N', 'T', nx, napprox, size(s%undo), 1e0_rt, work, nx, s%constTimesTheta, napprox, 0e0_rt, feval, nx)
        call mygemm(feval,work,s%constTimesTheta,nx,size(s%undo),napprox,size(s%undo),tA='N',tB='T')
    else
        call mygemm(feval,s%constTimesTheta,work,napprox,size(s%undo),nx,size(s%undo),tA='N',tB='T')
    end if

end subroutine smolyakapprox_step3B_core


subroutine smolyakapprox_step3_core(feval,s,xin,nx,napprox) 
    implicit none
    integer, intent(in) :: nx,napprox
    type(smolyakFcn), intent(in) :: s
    real(rt), dimension(*), intent(in), target :: xin
    real(rt), dimension(*), intent(inout) :: feval ! so can handle either vector or matrix
    !local
    real(rt), allocatable :: Tprod(:,:) 

    if (smolyakM(s%mu+1)<3) then
        STOP 'smolyakapprox_step3: I was not expecting this case, change T(:,3)'
    end if

    ! Get Tprod
    allocate(Tprod(nx,s%work_array_cols))
    call smolyakapprox_step3A_core(Tprod,s,xin,nx) 

    ! Evaluate the approximation
    call smolyakapprox_step3B_core(feval,s,Tprod,'N') 

!     ! Create a matrix x out of the assumed size array xin
!     x(1:nx,1:s%d) => xin(1:nx*s%d)
!     !!! NOTE: no need to do the same thing with feval since only used by DGEMM (F77 routine)
! 
!     
!     allocate(z(nx))
!     allocate(T(nx,smolyakM(s%mu+1)))
! 
!     if (allocated(Tprod)) deallocate(Tprod)
!     allocate(Tprod(nx,size(s%undo)))
! 
!     ! The smolyak approximation can be written as 
!     ! \sum_{q\le |i| \le d+mu} (-1)^{d+mu-|i|}choose(d-1,d+\mu-|i|)p^i(x)
!     ! p^i(x) itself is composed of a summation across l-values (in the notation of MKK)
!     ! of theta_l*T_l1(x_1)*T_l2(x_2)*...*T_ld(x_d). So, get the product across T and save
!     ! as Tprod for each l.
!     ! So, the summation can be written as 
!     ! \sum_{q\le |i| \le d+mu} [(-1)^{d+mu-|i|}choose(d-1,d+\mu-|i|)theta_l ]*
!     !                          [T_l1(x_1)*T_l2(x_2)*...*T_ld(x_d)]
!     ! The first bracketed object has already been computed and saved as constTimesTheta(iapprox,l)
!     ! The second has not been computed since it depends on the (transformed) x values
!     ! 
!     Tprod = 1e0_rt
!     do dind = 1,s%d
! 
!         !Convert the x points to z via linear transformation
!         z = 2e0_rt/(s%b(dind)-s%a(dind))*(x(:,dind)-s%a(dind)) - 1e0_rt
!         
!         ! Get the values of the Chebyshev basis at 
!         T(:,1) = 1e0_rt
!         T(:,2) = z
!         T(:,3) = 2e0_rt*z**2 - 1e0_rt ! instead of 2d0*z*z - 1 which has three multiplications, do only two
!         ! q<=ibar<=d+mu.  So ibar=d+mu is largest.  
!         !Largest any order Cheby can be is the largest any component of m(i) can be.  
!         !This occurs when i = 1 except for one component. 
!         ! => icomp = d+mu-(d-1) = mu+1 => i = mu+1 => m(i) = m(mu+1);
!         do oind = 4,smolyakM(s%mu+1) 
!             T(:,oind) = 2e0_rt*z*T(:,oind-1) - T(:,oind-2)
!         end do
!         
! !         do lind = 1,size(s%j,1)
! !             if (s%j(lind,dind)>1) then
! !                 Tprod(:,lind) = Tprod(:,lind)*T(:,s%j(lind,dind))
! !             end if
! !         end do
! 
!         do iind = 1,s%leni
!             ! Test whether i>1. If not, then j = 1,...,M(i) has M(i)==1 so j<=1 implying T(:,j)==1
!             if (s%i(iind,dind)>1) then
!                 do lind = s%Lb(iind),s%Ub(iind)
!                     ! Again, test whether j>1 since can save possibly many multiplications if j==1
!                     if (s%j(lind,dind)>1) then
!                         Tprod(:,lind) = Tprod(:,lind)*T(:,s%j(lind,dind))
!                     end if
!                 end do
!             end if 
!         end do
! 
!     end do

    ! NOTE: Could easily make the above parallel by distributing the ix values

    ! Now, sum the values. For greater accuracy, can sum positive values into one accumulation and negative into another
    ! NOTE: I checked this and it turns out to be less precise

    ! This should be much faster
    !!! The rows of Tprod represent different x points so nrows is size(x,1)
    !!! The columns of Tprod give T_l1(x_1)*T_l2(x_2)*....*T_ld(x_d) for a given l=(l1,l2,...,ld)
    !!! The rows of s%constTimesTheta give [(-1)^{d+mu-|i|}choose(d-1,d+\mu-|i|)theta_l ] for a given l
    !!! The columns of s%constTimesTheta represent different functions that have been approximated so # rows is s%napprox
    !!! Feval should have size(x,1) rows and s%napprox columns

!     ! SUBROUTINE DGEMM(TRANSA, TRANSB, M, N, K, ALPHA, A, LDA, B, LDB, BETA, C, LDC)
!     ! C := alpha*op(A)*op(B) + beta*C,
!     ! NOTE: M is rows of C, N is columns of C, K is # columns of B / # rows of A
!     ! Want: feval_{nx,napprox} = Tprod_{nx,size(s%undo)}*s%constTimesTheta_{size(s%undo),napprox})
!     call dgemm('N', 'T', nx, napprox, size(s%undo), 1e0_rt, Tprod, nx, s%constTimesTheta, napprox, 0e0_rt, feval, nx)

end subroutine smolyakapprox_step3_core


! Given an integer vector x>=1, returns all the different vectors
! y s.t. 1<=y<=x elt by elt. This gives the cartesian product of the 
! vectors (1:x(1)) cross (1:x(2)) cross ... (1:x(end)) 
! x will be treated as a row vector
! y is returned as a matrix with columns equal to the length of x and rows
! equal to (prod(x)) (which gives all the enumerations).
subroutine cartProdInt(y,x)
    use mod_matlab
    implicit none
    integer, intent(in) :: x(1:)
    integer, intent(out) :: y(1:product(x),1:size(x))
    ! local
    integer :: nr,nx,ix
    integer :: base(size(x)),xrev(size(x)),i(size(y,1)),modvals(size(y,1),size(y,2))

    ! The technique here is to convert the 1:prod(x) base 10 numbers into 
    ! a custom base system given by x. 
    ! 
    ! Given x = [x(1)...x(n)], 
    ! we want for any i in 1:prod(x) a unique d = [d(1)...d(n)] s.t.
    ! i = d(1)*(x(2)*x(3)...*x(n)) + d(2)*(x(3)*...*x(n)) + ... +
    !    d(n-1)*x(n) + d(n)
    ! e.g., say
    ! x = [3,1,2];
    ! we convert i=0 to d=[0,0,0], 
    !            i=1 to d=[0,0,1], 
    !            i=2 to d=[1,0,0],
    !            i=3 to d=[1,0,1], ...
    ! This will give all enumerates once we add 1 to each d.
    ! 
    ! We see mod(i,x(n)) = d(n), mod(i,x(n-1)*x(n)) = d(n-1)*x(n) + d(n),
    ! ... and so we can find all the d values recursively.


    nr = product(x)
    nx = size(x)

    do ix = 1,nx
        xrev(ix) = x(nx-ix+1)
    end do

    base = cumprod(xrev)        

    i = colon(0,nr-1)

    !!! NOTE: mod in MATLAB is the same thing as modulo in Fortran (NOT mod)

    ! Convert from base10 to new base system
    modvals(:,nx) = modulo(i,base(1))
    do ix = nx-1,1,-1
        modvals(:,ix) = modulo(i,base(nx-ix+1))
    end do

    y(:,nx) = modvals(:,nx)
    do ix = nx-1,1,-1
        y(:,ix) = (modvals(:,ix) - modvals(:,ix+1))/base(nx-ix)
    end do

    ! Convert to one-based indexing
    y = y + 1

end subroutine cartProdInt

! Given an integer n, construct the set of n extrema of the Chebyshev 
! polynomials zeta_j = -cos(pi*(j-1)/(n-1)), j = 1,n
subroutine smolyakG(set,n)
    implicit none
    integer, intent(in) :: n
    real(rt), intent(out) :: set(n)
    !local
    integer :: ind
    
    if (n<1) then
        STOP 'smolyakG: n must be >=1'
    elseif (n==1) then
        set = 0e0_rt
    else       
        set(1) = -1e0_rt
        do ind = 2,n-1
            set(ind) = -cos(pi*(real(ind,rt)-1.e0_rt)/(real(n,rt)-1.e0_rt))
        end do
        set(n) = 1e0_rt
        if (mod(ind,2)==1) then
            set((ind+1)/2) = 0e0_rt
        end if 
    end if

end subroutine smolyakG

! Determines number of points in the set of smolyakG(set,i) I think
elemental function smolyakM(i) result(o)
    implicit none
    integer, intent(in) :: i
    integer :: o
    if (i==1) then
        o = 1
    elseif (i>1) then
        o = 2**(i-1) + 1
    elseif (i<-1) then
        o = -100000000
    end if
end function smolyakM

subroutine test_smolyak
    use mod_matlab, only: randU, disp, tic, toc
    implicit none
    integer, parameter :: d=200,mu=1
    integer, parameter :: nrand = 100
    real(rt), allocatable :: x(:,:)
    type(smolyakFcn) :: s,s_mult
    real(rt) :: a(d),b(d)
    real(rt) :: frand(nrand), frand_mult(nrand,3), xrand(nrand,d)
    real(rt), allocatable :: apa(:,:), work(:,:)
    
    a = 0.e0_rt
    b = 1.e0_rt

    write(*,*) 'Smolyak'    
    call smolyakapprox_step1(x,s,d,mu,a,b)

    print*, 'Copying over to s_mult'
    s_mult = s

!     write(*,*) 's%i'
!     call disp(s%i)
!     
!     write(*,*) 's%ibar'
!     call disp(s%ibar)
! 
!     write(*,*) 's%j'
!     call disp(s%j)
!         
!     write(*,*) 's%cjprod'
!     call disp(s%cjprod)
! 
!     write(*,*) 's%const'
!     call disp(s%const)
!     
!     write(*,*) 's%constVec'
!     call disp(s%constVec)
! 
!     write(*,*) 'x (collocation points)'
!     call disp(x)
! 
!     write(*,*) 's%redo'
!     call disp(s%redo)
! 
!     write(*,*) 's%undo'
!     call disp(s%undo)       

print*,'made past step1'

    ! Check the step2 routine
    call smolyakapprox_step2(f1(x),s)

    allocate(apa(size(x,1),2))
    apa(:,1) = f1(x)
    apa(:,2) = f2(x)
    call smolyakapprox_step2(apa,s_mult)

print*,'made past step2'

    ! Check the step3 routines
    xrand = randU(100,d)
    call smolyakapprox_step3(frand,s,xrand)
    write(*,*) 'These values should be zero to numerical precision'
    write(*,*) maxval(f1(xrand) - frand),minval(f1(xrand) - frand)

    call smolyakapprox_step3(frand_mult(:,1:2),s_mult,xrand)
    write(*,*) 'These values should be zero to numerical precision'
    write(*,*) maxval(f1(xrand) - frand_mult(:,1)),minval(f1(xrand) - frand_mult(:,1))
    write(*,*) maxval(f2(xrand) - frand_mult(:,2)),minval(f2(xrand) - frand_mult(:,2))

    ! Check the step3A and step3B routines
    !!! Change x without changing f
    xrand = randU(100,d)
    allocate(work(smolyakGetWorkRows(s,size(xrand,1)),smolyakGetWorkCols(s,size(xrand,1))))
    call smolyakapprox_step3A(work,s_mult,xrand)
    call smolyakapprox_step3B(frand_mult(:,1:2),s_mult,work)
    write(*,*) 'These values should be zero to numerical precision'
    write(*,*) maxval(f1(xrand) - frand_mult(:,1)),minval(f1(xrand) - frand_mult(:,1))
    write(*,*) maxval(f2(xrand) - frand_mult(:,2)),minval(f2(xrand) - frand_mult(:,2))

    !!! Change f without changing x
    apa(:,1) = f2(x)
    apa(:,2) = f1(x)
    call smolyakapprox_step2(apa,s_mult) 
    call smolyakapprox_step3B(frand_mult(:,1:2),s_mult,work)

    write(*,*) 'These values should be zero to numerical precision'
    write(*,*) maxval(f2(xrand) - frand_mult(:,1)),minval(f2(xrand) - frand_mult(:,1))
    write(*,*) maxval(f1(xrand) - frand_mult(:,2)),minval(f1(xrand) - frand_mult(:,2))

    ! Check the step4 routine
!     allocate(apa(size(x,1),3))
!     apa(:,1) = f2(x)
!     apa(:,2) = f3(x)
!     apa(:,3) = f1(x)
!     call smolyakapprox_step2(apa,s_mult)
!     call smolyakapprox_step4_core(frand_mult(:,1:3),s_mult)
! 
!     write(*,*) 'These values should be zero to numerical precision'
!     write(*,*) maxval(f2(xrand) - frand_mult(:,1)),minval(f2(xrand) - frand_mult(:,1))
!     write(*,*) maxval(f3(xrand) - frand_mult(:,2)),minval(f3(xrand) - frand_mult(:,2))
!     write(*,*) maxval(f1(xrand) - frand_mult(:,3)),minval(f1(xrand) - frand_mult(:,3))
! 



!     write(*,*) 'Speed testing step2 by running 10000 times'
!     call tic()
!     do ind = 1,10000
!         call smolyakapprox_step2(sum(x**2,dim=2),s)
!     end do
!     call toc()
! ! pause 'smolyak pause'
!     write(*,*) 'Speed testing step3 by running 10000 times'
!     call tic()
!     do ind = 1,10000
!         call smolyakapprox_step3(frand,s,xrand)
!         xrand = rand(100,d)
!     end do
!     call toc()
! 
!     STOP 'test_smolyak: here'

contains
    function f1(x) 
        real(rt) :: x(1:,1:)
        real(rt) :: f1(size(x,1))
        f1 = sum(x**2,dim=2)
    end function
    function f2(x) 
        real(rt) :: x(1:,1:)
        real(rt) :: f2(size(x,1))
        f2 = sum(12.e0_rt*x**2-3.e0_rt*x,dim=2)
    end function
    function f3(x) 
        real(rt) :: x(1:,1:)
        real(rt) :: f3(size(x,1))
        f3 = sum(-5e0_rt*x**2-x+.2e0_rt,dim=2)
    end function
    
    

end subroutine test_smolyak

end module mod_smolyak
! Grey Gordon, 2012
! 
! For God so loved the world, that He gave His only begotten Son, 
! that whoever believes in Him shall not perish, but have eternal life.
!
module mod_minNdim
    use mod_types
    implicit none

    ! Options for minNdim 
    type minNdim_opts
        ! Commonly used ones
        integer :: printmod = 1, maxiter = 10000
        real(rt) :: xabstoler = 1e-6_rt, xreltoler = 1e-6_rt
        real(rt) :: yabstoler = 1e-10_rt, yreltoler = 1e-10_rt
        integer :: dbglvl = 0
        ! Others (maybe only a few algorithms use them)
        ! For these, -1 means there is no default specified
        real(rt) :: fcnerr = 0._rt  ! used by MINPACK routines for computing approximate jacobian
        integer :: algorithm = 1
        integer :: poolsize = -1
        integer :: sleeptime = -1 
        integer :: displaybestx = 1 ! 1==means true (the default)
        real(rt) :: mutationprob = .001_rt
        integer :: reproducible = 0 ! random default
    end type 

    private :: slave

contains

    subroutine disp_opts(o)
        implicit none
        type(minNdim_opts), intent(in) :: o
        write(*,*) 'printmod',o%printmod
        write(*,*) 'maxiter',o%maxiter
        write(*,*) 'xabstoler',o%xabstoler
        write(*,*) 'xreltoler',o%xreltoler
        write(*,*) 'yabstoler',o%yabstoler
        write(*,*) 'yreltoler',o%yreltoler
        write(*,*) 'dbglvl',o%dbglvl
        write(*,*) 'poolsize',o%poolsize
        write(*,*) 'sleeptime',o%sleeptime
        write(*,*) 'displaybestx',o%displaybestx
        write(*,*) 'mutationprob',o%mutationprob
        write(*,*) 'reproducible',o%reproducible
        write(*,*) 'algorithm',o%algorithm
        write(*,*) 'fcnerr',o%fcnerr
    end subroutine

    ! Here, options are specified via a string such as "xtoler=1d6,printmod=10000"
    function optimset(str) result(opts)
        use mod_util
        implicit none
        character(len=*), intent(in) :: str
        type(minNdim_opts) :: opts
        ! local
        character(len=len(str)) :: s,arg,val
        type(minNdim_opts) :: def
        integer :: nargs,i

        nargs = count_delim_arg(str,",")

        do i = 1,nargs

            ! Get the substring
            s = ''
            s = trim(get_delim_arg(str,",",i))

            ! Process to get what is left and right of the equal sign
            arg = ''
            val = ''
            arg = get_delim_arg(s,"=",1)
            val = get_delim_arg(s,"=",2)

            if     (trim(arg)=="xabstoler")     then; read(val,'(d30.16)',err=999) opts%xabstoler 
            elseif (trim(arg)=="xreltoler")     then; read(val,'(d30.16)',err=999) opts%xreltoler 
            elseif (trim(arg)=="yabstoler")     then; read(val,'(d30.16)',err=999) opts%yabstoler 
            elseif (trim(arg)=="yreltoler")     then; read(val,'(d30.16)',err=999) opts%yreltoler 
            elseif (trim(arg)=="mutationprob")  then; read(val,'(d30.16)',err=999) opts%mutationprob
            elseif (trim(arg)=="fcnerr")        then; read(val,'(d30.16)',err=999) opts%fcnerr
            elseif (trim(arg)=="printmod")      then; read(val,'(i25)',   err=999) opts%printmod
            elseif (trim(arg)=="maxiter")       then; read(val,'(i25)',   err=999) opts%maxiter
            elseif (trim(arg)=="poolsize")      then; read(val,'(i25)',   err=999) opts%poolsize
            elseif (trim(arg)=="sleeptime")     then; read(val,'(i25)',   err=999) opts%sleeptime
            elseif (trim(arg)=="displaybestx")  then; read(val,'(i25)',   err=999) opts%displaybestx
            elseif (trim(arg)=="dbglvl")        then; read(val,'(i25)',   err=999) opts%dbglvl
            elseif (trim(arg)=="reproducible")  then; read(val,'(i25)',   err=999) opts%reproducible
            elseif (trim(arg)=="algorithm")     then; read(val,'(i25)',   err=999) opts%algorithm
            else
                write(0,*) 'WARNING: optimset: did not find option corresponding to argument ',&
                           trim(arg),'  with value ',trim(val),' and overall input ',trim(str)
            end if

        end do

        return

999     write(0,*) 'ERROR: optimset: error encountered processing ',trim(str),', with arg ',trim(arg),' and val ',trim(val),' falling back to default arguments'
        opts = def
    

    end function

    ! Levenberg-Marquardt with a finite-difference Jacobian
    ! 
    ! options specific to LM
    ! -> fcnerr : determines the step-size in computing the gradient. stepsize is (roughly?) 10**(fcnerr/2-8). So fcnerr=16 is 100%, fcnerr=14 is 10%, fcnerr=12 is 1%, .... If too small, will just use machine epsilon
    ! 
    ! Interface to the MINPACK routine lmdif
    subroutine sub_levenbergmarquardt(xstar,f,m,opts)
        implicit none
        interface 
            function f(x,m) 
                use mod_types
                implicit none
                real(rt), dimension(:), intent(in) :: x
                integer, intent(in) :: m
                real(rt) :: f(m)
            end function
        end interface
        real(rt), intent(in) :: xstar(:)
        integer, intent(in) :: m
        type(minNdim_opts), optional :: opts
        ! local
        type(minNdim_opts) :: o
        !!! NOTE: keep as double precision because required by MINPACK
        real(8), allocatable :: fvec(:),wa(:)
        integer, allocatable :: iwa(:)
        real(8) :: epsfcn,factor,ftol,gtol,xtol
        integer :: n,info,lwa,maxfev,mode,nfev,nprint,neval,iter

        if (present(opts)) o = opts

        neval = 0
        iter = 0
        n = size(xstar)
        lwa = m*n+5*n+m

        allocate(fvec(m))
        allocate(wa(lwa))
        allocate(iwa(n))

        info = 0
        if (n .le. 0 .or. m .lt. n .or. lwa .lt. m*n + 5*n + m) then
            write(0,*) 'WARNING: bad input param'
            return
        end if

        maxfev = o%maxiter*(n + 1)
        ftol = o%yreltoler
        xtol = o%xreltoler
        gtol = 0.d0
        epsfcn = o%fcnerr ! used to construct appropriate step-length (higher means error, and uses bigger step). if < epsilon(0), uses epsilon(0)
        mode = 1 ! internally scaled
        nprint = max(0,o%printmod)
        factor = 100d0

        call lmdif(sub_f,m,n,xstar,fvec,ftol,xtol,gtol,maxfev,epsfcn,wa(1),&
                   mode,factor,nprint,info,nfev,wa(m+5*n+1),m,iwa,& 
                   wa(n+1),wa(2*n+1),wa(3*n+1),wa(4*n+1),wa(5*n+1))

        ! call lmdif1(sub_f,m,n,xstar,fvec,tol,info,iwa,wa,lwa)

        select case (info)
            case(0); write(*,'(A)') 'LM: improper input parameters.'
            case(1); write(*,'(A)') 'LM: both actual and predicted relative reductions in the sum of squares are at most ftol.'
            case(2); write(*,'(A)') 'LM: relative error between two consecutive iterates is at most xtol.'
            case(3); write(*,'(A)') 'LM: conditions for info = 1 and info = 2 both hold.'
            case(4); write(*,'(A)') 'LM: the cosine of the angle between fvec and any column of the jacobian is at most gtol in absolute value.'
            case(5); write(*,'(A)') 'LM: number of calls to fcn has reached or exceeded maxfev.'
            case(6); write(*,'(A)') 'LM: ftol is too small. no further reduction in the sum of squares is possible.'
            case(7); write(*,'(A)') 'LM: xtol is too small. no further improvement in the approximate solution x is possible.'
            case(8); write(*,'(A)') 'LM: gtol is too small. fvec is orthogonal to the columns of the jacobian to machine precision.'
        end select

    contains 

        subroutine sub_f(m,n,x,fvec,iflag)
            use mod_util, only: printx
            implicit none
            integer :: m,n,iflag
            real(8) :: x(n),fvec(m)

            if (iflag==0) then
                write(*,'(6A,2(A,i5))')'LM: x = ',printx(x),' f(x) = ',printx(fvec),' SSRS = ',printx([norm2(fvec)]),' iter ',iter,' fevals ',neval
                iter = iter + nprint
            else
                fvec = f(x,m)
                neval = neval + 1
            end if

        end subroutine 

    end subroutine 

    ! Optimizes only along a given dimension
    subroutine sub_opt_along_dim(xstar,f,a,b,dim,opts,ngridsearch)
        use mod_min1dim
        implicit none
        real(rt), intent(inout) :: xstar(:)
        real(rt), intent(in) :: a,b
        integer, intent(in) :: dim
        type(minNdim_opts), optional :: opts
        integer, intent(in), optional :: ngridsearch
        interface 
            function f(x) 
                use mod_types
                implicit none
                real(rt), dimension(:), intent(in) :: x
                real(rt) :: f
            end function
        end interface
        ! local
        integer :: nx, ngs, flag
        real(rt), allocatable :: x(:)
        real(rt) :: x1
        type(minNdim_opts) :: o

        ngs = 100
        if (present(ngridsearch)) ngs = ngridsearch

        if (present(opts)) o = opts


        nx = size(xstar)

        if (dim<0 .or. dim>nx) stop 'ERROR: sub_opt_along_dim: invalid dim'

        allocate(x(nx))

        x = xstar

        print*,'sub_opt_along_dim: initial guess: ',xstar(dim)

        x1 = xstar(dim) 
        call sub_gridsearch1dim_nogrid(x1,g,a,b,ngs,printmod=o%printmod)
        call sub_simplex1dim(x1,g,e=.75d0*(b-a)/(ngs-1),xtoler=o%xabstoler,maxiter=o%maxiter,printmod=o%printmod,flag=flag) 
        ! Note: don't use brent as current setup doesn't use initial information for x1

        xstar(dim) = x1

        print*,'sub_opt_along_dim: final guess: ',xstar(dim)

    contains

        function g(x1) 
            use mod_types
            implicit none
            real(rt), intent(in) :: x1
            real(rt) :: g

            x(dim) = x1
            g = f(x)

        end function 


    end subroutine

    ! Controlled random search --- serial version
    ! 
    ! CRS attempts to solve min f(x) s.t. x in [a,b] (where all are in Rn)
    ! It generates N random uniform points.
    ! It takes the worst point and tries to get rid of it by reflecting it
    ! (like in Nelder-Mead) through a simplex of points. This simplex is chosen 
    ! by randomly selecting n+1 points (whose function values are known) with 
    ! replacement, giving a total of n+1 points. 
    !
    ! The centroid of the simplex is then used for reflecting the worst point. 
    ! If there reflected point improves on the worst point (and is in a,b), it is
    ! accepted. O/w, the process repeats.
    ! 
    ! This is modified for parallel purposes by attempting to simultaneously get 
    ! rid of several of the worst points simultaneously. The parallel design here,
    ! in contrast to CRS2LM, is more advantageous when the functions time to completion
    ! is not uniform across x values. The algorithm itself is more global.
    !
    ! The optional argument sleeptime is useful when there are only a few ranks. 
    ! It specifies how long the master should sleep if no free workers are found. This
    ! keeps processing cycles free for those ranks actually doing work. 
    subroutine sub_CRS(xstar,f,a,b,opts)
        use mod_util
        use mod_matlab
        implicit none
        real(rt), intent(inout) :: xstar(:)
        real(rt), intent(in) :: a(:),b(:)
        type(minNdim_opts), optional :: opts
        interface 
            function f(x) 
                use mod_types
                implicit none
                real(rt), dimension(:), intent(in) :: x
                real(rt) :: f
            end function
        end interface
        ! local other
        type(type_seed) :: randstate
        real(rt), allocatable :: X(:,:),y(:) 
        real(rt), allocatable :: R(:),g(:),smx(:,:),tmpy(:)
        integer, allocatable :: tmpI(:)
        real(rt) :: tmp, yR
        integer :: nx, N, M, ix, iter, i, ism, itmp(1)
        logical :: dostop
        type(minNdim_opts) :: o

        if (present(opts)) o = opts
    

        nx = size(xstar)
        if (size(a)/=nx) stop 'ERROR: sub_CRS: size of a and x do not agree'
        if (size(b)/=nx) stop 'ERROR: sub_CRS: size of b and x do not agree'
        if (any(a>=b)) stop 'ERROR: sub_CRS: a>=b at least once'

        ! Check that input x is in a,b for all i
        if (any(xstar<a) .or. any(xstar>b)) stop 'ERROR: sub_CRS: x is not in a,b for all dimensions'

        ! M is the number to simultaneously try to get rid of
        M = 1

        ! N is number of trial points (the pool of points)
        N = max(5*M, ((100+nx)/M+1)*M, nx+2+M)
        if (o%poolsize/=-1) then
            N = max(N,o%poolsize)
        end if

        if (N-M<=nx+1) stop 'ERROR: sub_CRS: too few N points set relative to M'

        ! Generate N random points uniformly from [a,b]
        allocate(X(nx,N))
        allocate(y(N))
        allocate(R(nx))
        allocate(g(nx))
        allocate(smx(nx,nx+1))                    
        allocate(tmpy(N),tmpI(N))

        ! Generate initial pool of points
        call randstate%save()
        call randstate%randomize()
        X(:,1) = xstar
        do i = 2,N
            do ix = 1,nx
                call random_number(tmp)
                X(ix,i) = tmp*a(ix) + (1e0_rt-tmp)*b(ix)
            end do
        end do
        call randstate%revert()

        ! Evaluate at each of the N points in parallel
        if (o%printmod>0 .and. o%printmod==1) write(*,'(A,i4,A)') 'CRS: Solving for initial pool of ',N,' points'
        do i = 1,N

            if (o%printmod>=1 .and. mod(i,o%printmod)==0 .and. o%displaybestx==1) &
                write(*,'(A,i4,A,i4,A,A)') 'CRS: solving for point ',i,' of ',N,' x = ',printx(X(:,i))
            if (o%printmod>=1 .and. mod(i,o%printmod)==0 .and. o%displaybestx==0) &
                write(*,'(A,i4,A,i4)')     'CRS: solving for point ',i,' of ',N

            ! Evaluate the point
            y(i) = f(X(:,i))

        end do

        call randstate%save()
        call randstate%randomize()
        call init_random_seed(N) ! in case f(.) has reset the seed, set a new one
        call randstate%revert()

        if (o%dbglvl>1) print*,'done initial point distribution, proceeding to main loop'

        do iter = 1,o%maxiter

            ! Sort the current points (ascending)
            call sort2(tmpy,tmpI,y) 
            X = X(:,tmpI) ! sort X
            y = tmpy ! sort y

            ! Display progress
            if (o%printmod>0 .and. mod(iter,o%printmod)==0) then
                if (o%displaybestx==1) &
                    write(*,'(A,i8,2(A,e14.6),2A)') 'CRS: iter',iter,'  best ',y(1),'  worst ',y(N),'  xbest = ',printx(X(:,1))
                if (o%displaybestx==0) &
                    write(*,'(A,i8,2(A,e14.6))')    'CRS: iter',iter,'  best ',y(1),'  worst ',y(N)
            end if
 
            ! In case exit either through maxiter being reached or satisfying exit criteria, save xstar
            xstar = X(:,1)

            ! Check stopping criteria
            dostop = abs(y(1)-y(N))<o%yabstoler .or. sqrt(sum((x(:,1)-x(:,N))**2))<o%xabstoler

            if (dostop .or. iter==o%maxiter) then

                if (dostop) then
                    if (o%printmod==1) write(*,'(A)') 'CRS: convergence criteria satisfied'
                else
                    if (o%printmod==1) write(*,'(A)') 'CRS: max iterations reached'
                end if

                return

            end if

            ! Form a random simplex excluding current worst point 
            do ism = 1,nx+1
                
                ! CRS with local mutation where simplex always includes best point
                if (ism==1) then
                    smx(:,ism) = x(:,1) 
                else
                    smx(:,ism) = x(:,randi(1,N-1))
                end if

            end do

            ! Compute the centroid
            g = sum(smx,2)/(nx+1)

            ! Reflect the worst point
            R = 2e0_rt*g - X(:,N) ! pure reflection

            ! If goes through boundary, then bound at boundary
            R = max(R,a)
            R = min(R,b)

            ! To make more local, can 50% of the time combine with local search which is convex combination with best point
            call randstate%save()
            call randstate%randomize()
            call random_number(tmp)
            if (tmp<=.5e0_rt) then
                do ix = 1,nx
                    call random_number(tmp)
                    R(ix) = tmp*x(ix,1) + (1e0_rt-tmp)*R(ix) ! convex combination, so in [a,b]
                end do
            end if
            call randstate%revert()
            
            ! Compute the reflected point
            yR = f(R)

            ! Replace the worst point if found a better one
            if (yR<maxval(y)) then
                itmp = maxloc(y) 
                y(itmp(1)) = yR
                x(:,itmp(1)) = R
            end if

        end do

    end subroutine

    subroutine check_guess_in_range(flagout,x,a,b)
        implicit none
        real(rt), intent(in) :: x(:),a(:),b(:)
        integer, intent(out) :: flagout
        ! local
        integer :: i,n,flag

        n = size(x)
        if (size(a)/=n) stop 'ERROR: check_guess_in_range: a and x do not have same length'
        if (size(b)/=n) stop 'ERROR: check_guess_in_range: b and x do not have same length'

        flag = 0
        do i = 1,n 
        if (x(i)<a(i)) then; write(0,*) 'WARNING: check_guess_in_range: x<a for element',i; flag = merge(-1,-999,flag==0); end if
        if (x(i)>b(i)) then; write(0,*) 'WARNING: check_guess_in_range: x>b for element',i; flag = merge(-2,-999,flag==0); end if
        if (b(i)<a(i)) then; write(0,*) 'WARNING: check_guess_in_range: b<a for element',i; flag = merge(-3,-999,flag==0); end if
        end do
        flagout = flag

    end subroutine 

    ! Controlled random search --- MPI version
    ! 
    ! CRS attempts to solve min f(x) s.t. x in [a,b] (where all are in Rn)
    ! It generates N random uniform points.
    ! It takes the worst point and tries to get rid of it by reflecting it
    ! (like in Nelder-Mead) through a simplex of points. This simplex is chosen 
    ! by randomly selecting n+1 points (whose function values are known) with 
    ! replacement, giving a total of n+1 points. 
    !
    ! The centroid of the simplex is then used for reflecting the worst point. 
    ! If there reflected point improves on the worst point (and is in a,b), it is
    ! accepted. O/w, the process repeats.
    ! 
    ! This is modified for parallel purposes by attempting to simultaneously get 
    ! rid of several of the worst points simultaneously. The parallel design here,
    ! in contrast to CRS2LM, is more advantageous when the functions time to completion
    ! is not uniform across x values. The algorithm itself is more global.
    !
    ! The optional argument sleeptime is useful when there are only a few ranks. 
    ! It specifies how long the master should sleep if no free workers are found. This
    ! keeps processing cycles free for those ranks actually doing work. 
    !
    ! If e_in is present, then random draws will not be taken uniformly from [a,b], but rather
    ! from [xstar-e,xstar+e] (component-wise). Combined with a small poolsize (which can be set 
    ! to 0 which will then be overridden), this makes the routine effectively like a parallel 
    ! version of Nelder-Mead.
    subroutine sub_parCRS_MPI(xstar,f,a,b,e_in,opts)
        use mod_util
        use mod_mympi
        use mod_matlab
        implicit none
        real(rt), intent(inout) :: xstar(:)
        real(rt), intent(in) :: a(:),b(:)
        real(rt), intent(in), optional :: e_in(:)
        type(minNdim_opts), intent(in), optional :: opts
        interface 
            function f(x) 
                use mod_types
                implicit none
                real(rt), dimension(:), intent(in) :: x
                real(rt) :: f
            end function
        end interface
        ! local other
        type(type_seed) :: randstate
        real(rt), allocatable, target :: X(:,:),y(:),slavexlist(:,:),slaveylist(:) ! NOTE: these have target attribute and will be pointed at
        logical, allocatable :: finishedtask(:),slaveeverassignedlist(:)
        real(rt), allocatable :: R(:),g(:),smx(:,:),tmpy(:)
        integer, allocatable :: tmpI(:), fevaldsn(:),requestlist(:)
        real(rt) :: tmp
        integer :: nx, N, M, ix, iter, i, ism, who, itmp(1), flag, minPoolSize, effPoolSize, effIter
        type(minNdim_opts) :: o

        if (present(opts)) o = opts

        if (mympi_nproc()<=1) then
            write(0,*) 'WARNING: parCRS MPI version was called with only one rank, calling serial version instead'
            call sub_CRS(xstar,f,a,b,o)
            return
        end if

        nx = size(xstar)
        if (size(a)/=nx) stop 'ERROR: sub_parCRS2LM: size of a and x do not agree'
        if (size(b)/=nx) stop 'ERROR: sub_parCRS2LM: size of b and x do not agree'
        if (any(a>=b)) stop 'ERROR: sub_parCRS2LM: a>=b at least once'

        ! Check that input x is in a,b for all i
        call check_guess_in_range(flag,xstar,a,b)
        if (flag/=0) stop 'ERROR: sub_parCRS2LM: x is not in a,b or b<a for some dimension'

        ! M is the number to simultaneously try to get rid of
        M = mympi_nproc()-1

        ! N is number of trial points (the pool of points)
        ! N = max(2*M, ((100+nx)/M+1)*M, nx+2+M)
        N = max(((100+nx)/M+1)*M, nx+2+M)
        ! N = ((100+nx)/M+1)*M ! roughly evenly divide 100+nx points across the M ranks
        if (o%poolsize/=-1) N = max(nx+2+M,o%poolsize)

        minPoolSize = max(nx+1,N-mympi_nproc(),M)

        if (N-M<=nx+1) call mympi_stop('ERROR: sub_parCRS2LM: too few N points set relative to M')

        ! If not master, go to the "slave" routine and wait for a x to evaluate
        if (mympi_iproc()/=0) then
            call slave(f,nx)
            call mympi_bcast(xstar,0) ! receive the best point from 
            call mympi_barrier()
            return
        end if

        ! Generate N random points uniformly from [a,b]
        allocate(finishedtask(mympi_nproc()-1))
        allocate(X(nx,N))
        allocate(y(N))
        allocate(R(nx))
        allocate(g(nx))
        allocate(smx(nx,nx+1))                    
        allocate(tmpy(N),tmpI(N))
        allocate(fevaldsn(mympi_nproc()-1))
        allocate(requestlist(mympi_nproc()-1))
        allocate(slavexlist(nx,mympi_nproc()-1))
        allocate(slaveylist(mympi_nproc()-1))
        allocate(slaveeverassignedlist(mympi_nproc()-1))

        fevaldsn = 0

        finishedtask = .true.
        slaveeverassignedlist = .false.

        ! Evaluate at each of the N points in parallel
        effPoolSize = 0
        write(*,'(A,i8,A,i8)') 'PARCRS: Solving for initial pool of ',N,' points; min pool size is ',minPoolSize
        do i = 1,N

            ! Wait for a free worker
            if (o%dbglvl>0) print*,'MASTER: WAITING FOR FREE WORKER'
            who = wait_for_free_worker()

            ! If they had a previously assigned point, use its value
            if (slaveeverassignedlist(who)) then
                ! Add to the pool
                effPoolSize = effPoolSize + 1
                X(:,effPoolsize) = slavexlist(:,who)
                Y(effPoolsize) = slaveylist(who)

                if (isnan(y(effPoolSize))) then
                    write(0,*) 'WARNING: sub_parCRS_MPI: function returned a NaN value for x :',printx(X(:,effPoolSize)),' replacing with huge(0d0)'
                    y(effPoolSize) = huge(0d0)
                end if

                if (o%printmod>0 .and. mod(effPoolSize,o%printmod)==0) then
                    itmp = minloc(Y(1:effPoolSize)) 
                    write(*,'(A,i8,2(A,ES14.6),2A)') 'PARCRS: pool ',effPoolSize,' trial ',y(effPoolSize),' best ',y(itmp(1)),'  xbest = ',printx(X(:,itmp(1)))
                end if
            end if

            ! Generate a random point and assign it
            call randstate%save()
            call randstate%randomize()
            slaveeverassignedlist(who) = .true.
            if (i==1) then
                slavexlist(:,who) = xstar
            else
                do ix = 1,nx
                    if (.not. present(e_in)) then
                        call random_number(tmp)
                        slavexlist(ix,who) = tmp*a(ix) + (1e0_rt-tmp)*b(ix)
                    else
                        call random_number(tmp)
                        tmp = 2._rt*tmp-1
                        slavexlist(ix,who) = min(b(ix),max(a(ix),xstar(ix)+tmp*e_in(ix)))
                    end if
                end do
            end if
            call randstate%revert()
            call assign_xy_to_worker(slavexlist(:,who),slaveylist(who),who)

            ! For statistical purposes, collect who was assigned that feval
            fevaldsn(who) = fevaldsn(who) + 1

        end do

        if (o%dbglvl>1) print*,'waiting for everyone to finish'

        ! Wait for everyone to finish
        ! do while (any(.not. finishedtask))
        do while (effPoolsize<minPoolSize) 

            do who = 1,mympi_nproc()-1
                if (.not. finishedtask(who)) then
                    finishedtask(who) = mympi_test(requestlist(who)) ! requestlist is not valid after mympi_test returns true
                    if (finishedtask(who)) then
                        ! Add to the pool
                        effPoolSize = effPoolSize + 1
                        X(:,effPoolsize) = slavexlist(:,who)
                        Y(effPoolsize) = slaveylist(who)
                        if (isnan(y(effPoolSize))) then
                            write(0,*) 'WARNING: sub_parCRS_MPI: function returned a NaN value for x :',printx(X(:,effPoolSize)),' replacing with huge(0d0)'
                            y(effPoolSize) = huge(0d0)
                        end if
                    end if
                end if
            end do

        end do

        write(*,'(A,i8,A)') 'PARCRS: Achieved pool size of ',effPoolSize,' proceeding to main loop, pool size will be increased in main loop'

        if (o%dbglvl>1) print*,'done initial point distribution, proceeding to main loop'

        effIter = 0

        do iter = 1,o%maxiter

            ! Wait for a slave. If there were previously assigned a task, then record its value.
            who = wait_for_free_worker()
!             print*,'iter',iter,'who',who

            ! If they had a previously assigned point, use its value
            if (slaveeverassignedlist(who)) then

                if (isnan(slaveylist(who))) then
                    write(0,*) 'WARNING: sub_parCRS_MPI: function returned a NaN value for x :',printx(slavexlist(:,who)),' replacing with huge(0d0)'
                    slaveylist(who) = huge(0d0)
                end if
                
                if (effPoolSize<N) then
                    ! If not at desired pool size yet, just add it to the pile
                    effPoolSize = effPoolSize + 1
                    X(:,effPoolsize) = slavexlist(:,who)
                    Y(effPoolsize) = slaveylist(who)

                else

                    ! Otherwise, replace the worst point if found a better one
                    itmp = maxloc(y(1:effPoolSize)) 
                    if (slaveylist(who)<y(itmp(1))) then
                        X(:,itmp(1)) = slavexlist(:,who)
                        Y(itmp(1)) = slaveylist(who)
                    end if

                end if

            end if

            effIter = merge(effIter+1,effIter,effPoolSize==N)

            ! After this point, "who" will be assigned an x
            slaveeverassignedlist(who) = .true.

            ! Sort the current points (ascending)
            call sort2(tmpy(1:effPoolSize),tmpI(1:effPoolSize),y(1:effPoolSize)) 
            X(:,1:effPoolSize) = X(:,tmpI(1:effPoolSize)) ! sort X
            y(1:effPoolSize) = tmpy(1:effPoolSize) ! sort y

            ! Display progress
            if (iter==1 .or. (o%printmod>0 .and. mod(iter,o%printmod)==0)) then
                
                if (effPoolSize/=N .or. effIter==0) then
                    write(*,'(A,i8,2(A,ES14.6),2A)') 'PARCRS: pool ',effPoolSize,' worst ',y(effPoolSize),' best ',y(1),'  xbest = ',printx(X(:,1))
!                     write(*,'(A,*(i4))') 'fevaldsn',fevaldsn
                else
                    write(*,'(A,i8,2(A,ES14.6),2A)') 'PARCRS: iter ',effIter,' worst ',y(effPoolSize),' best ',y(1),'  xbest = ',printx(X(:,1))
                end if

            end if
 
            ! Check stopping criteria
            if (effPoolSize==N .and. abs(y(1)-y(effPoolSize))<o%yabstoler .or. iter==o%maxiter) then

                ! In case exit either through maxiter being reached or satisfying exit criteria, save xstar
                xstar = X(:,1)

                if (abs(y(1)-y(effPoolSize))<o%yabstoler) then
                    write(*,'(A)') 'PARCRS: convergence criteria satisfied'
                else
                    write(*,'(A)') 'PARCRS: max iterations reached'
                end if

                write(*,'(*(A,i5))') 'PARCRS: rank eval. count; min: ',minval(fevaldsn),' max: ',maxval(fevaldsn),' avg: ',sum(fevaldsn)/size(fevaldsn)

                ! Shutdown slave
                call shutdown_slaves()

                ! Distribute xbest
                call mympi_bcast(xstar,0)

                ! Wait for everyone to finish 
                call mympi_barrier()

                return

            end if

            ! Form a random simplex excluding current worst point 
            call randstate%save()
            call randstate%randomize()
            do ism = 1,nx+1
                
                ! CRS with local mutation where simplex always includes best point
                if (ism==1) then
                    smx(:,ism) = x(:,1) 
                else
                    smx(:,ism) = x(:,randi(1,effPoolSize-1))
                end if

            end do

            ! Compute the centroid
            g = sum(smx,2)/(nx+1)

            ! Reflect the worst point
            R = 2._rt*g - X(:,effPoolSize) ! pure reflection

            ! If goes through boundary, then bound at boundary
            R = min(max(R,a),b)

            ! To make more local, can 50% of the time combine with local search which is convex combination with best point
            call random_number(tmp)
            if (tmp<=.5e0_rt) then
                do ix = 1,nx
                    call random_number(tmp)
                    R(ix) = tmp*x(ix,1) + (1e0_rt-tmp)*R(ix) ! convex combination, so in [a,b]
                end do
            end if
            call randstate%revert()
            
            ! Request that they solve for x and store it in slavey
            slavexlist(:,who) = R
            call assign_xy_to_worker(slavexlist(:,who),slaveylist(who),who)

            fevaldsn(who) = fevaldsn(who) + 1

        end do

    contains

        ! This routine sends the value of x to a slave "who", who will then store the value in y. 
        subroutine assign_xy_to_worker(x,y,who)
            use iso_c_binding, only: c_loc,c_ptr,c_f_pointer ! GETS THE ACTUALY ADDRESS OF Y WHICH WILL PERSIST AFTER CALL TO ASSIGN ENDS
            implicit none
            real(rt), intent(in), target :: x(:)
            real(rt), intent(out), target :: y
            integer, intent(in) :: who ! who you want assigned
            ! local
            type(c_ptr) :: yptr_c ! pointer in C
            real(rt), pointer :: yptr_f ! pointer in Fortran 
            integer :: flag,ierr

            ! Assign it
            flag = 1
            finishedtask(who) = .false.

            call mympi_send(flag,who) ! blocking send (tell them to proceed to execute)

            call mympi_send(x,who) ! assign an x value (wait for it to complete)

            ! In order to be able to do this assignment where the memory location associated with y will be stored even after 
            ! the subroutine is exited, the safest/most reliable way I have found is to send MPI the actual memory address.
            ! This can be accomplished by getting the memory address of y using the F2003 intrinsic "c_loc" and assigning that
            ! to a C pointer. Then, using c_f_pointer, copy that to a Fortran pointer. Then pass that Fortran poitner to MPI. 
            !
            ! This may be a hack (it seems like completely legal Fortran though), but it simplifies bookkeeping substantially.
            ! I.e., it becomes "set it and forget it."

            yptr_c = c_loc(y) ! get the actual address of y using "c_loc" and have yptr_c point to that
            call c_f_pointer(yptr_c,yptr_f) ! get a fortran pointer to that same address
            call MPI_Irecv(yptr_f, 1, MPI_DOUBLE_PRECISION, who, who, MPI_COMM_WORLD, requestlist(who), ierr)

        end subroutine 

        ! a routine to shut down all slaves
        subroutine shutdown_slaves() 
            implicit none
            ! local
            integer :: i
            logical :: ft

            ! Cycle through and shutdown
            do i = 1,mympi_nproc()-1

                ! Wait for rank to finish if not finished task
                if (.not. finishedtask(i)) then

                    ! Infinite loop, wait for them to finish 
                    do 

                        ! Check whether i has finished
                        ft = mympi_test(requestlist(i))
                        if (ft) then
                            finishedtask(i) = .true. ! show that they are now free
                            exit
                        end if

                        ! If sleeptime, then sleep
                        if (o%sleeptime>=1) then

                            call sleep(o%sleeptime)

                        else

                            ! Do some busy work so as to not thrash the system
                            call short_sleep()

                        end if

                    end do

                end if
                    
                ! Shut down, sending flag = -1 and xstar
                call mympi_send(-1,i) 

            end do

        end subroutine


        ! a routine returning the index of a free slave
        ! if no worker is available, who is -1
        function wait_for_free_worker() result(who)
            implicit none
            integer :: who
            ! local
            integer :: i
            logical :: ft

            do 

                ! Infinite loop
                do i = 1,mympi_nproc()-1

                    if (finishedtask(i)) then

                        who = i
                        return  

                    else

                        ! Check whether i has finished
                        if (o%dbglvl>1) print*,'MASTER: TESTING MESSAGE'
                        if (o%dbglvl>2) print*,'MASTER: finishedtask',finishedtask
                        if (o%dbglvl>2) print*,'MASTER: REQUESTLIST',REQUESTLIST
                        if (o%dbglvl>2) print*,'MASTER: i: ',i

                        ft = mympi_test(requestlist(i))
                        if (o%dbglvl>1) print*,'MASTER: FINISHED TEST'
                        if (ft) then

                            finishedtask(i) = .true. ! show that they are now free


                            who = i
                            if (o%dbglvl>1) print*,'MASTER: RECEIVED MESSAGE'
                            return
                        end if

                    end if

                end do

                ! If sleeptime, then sleep
                if (o%sleeptime>=1) then

                    call sleep(o%sleeptime)

                else

                    ! Do some busy work so as to not thrash the system
                    call short_sleep()

                end if

            end do

        end function


    end subroutine

    ! Do something according to the value of flag(this_image())[1]
    !  1 := evaluate slavex and store in slavey
    ! -1 := exit the slave routine
    subroutine slave(f,nx)
        use mod_mympi
        implicit none
        integer, intent(in) :: nx
        interface 
            function f(x) 
                use mod_types
                implicit none
                real(rt), dimension(:), intent(in) :: x
                real(rt) :: f
            end function
        end interface
        ! local
        integer :: flag
        real(rt) :: slavey
        real(rt), allocatable :: slavex(:)

        allocate(slavex(nx))

        do 

            ! Wait for a flag 
            call mympi_recv(flag,0)

            ! If flag==-1, then shutdown
            if (flag==-1) then
                return
            end if

            ! If still here, wait for an x
            call mympi_recv(slavex,0)

            ! If flag==1, execute f(x) and store on master.
            if (flag==1) then

                ! Compute the result
                slavey = f(slavex)

                ! Send the result to master tagged by mympi_iproc()
                call mympi_send(slavey,0,mympi_iproc()) 

            end if

        end do

    end subroutine

    ! Generate a random integer in a,...,b
    function randi(a,b) result(i)
        implicit none
        integer, intent(in) :: a,b
        integer :: i
        real(rt) :: tmp

        ! Generate a U[a,b+1-epsilon]
        ! Then floor to get a in 
        call random_number(tmp)
        tmp = tmp*(b+1-a) + a ! U[a,b+1]
        tmp = min(tmp,b+.99999999e0_rt)
        i = floor(tmp)

        if (floor(tmp)<a .or. floor(tmp)>b) STOP 'ERROR: randi: failed to generate i in a,...,b'

    end function


!     ! Noisy Levenberg-Marquardt with MPI
!     !
!     ! solve min_{x \in R^{nx}} \sum^{nf}_{j=1} f^j(x)
!     !
!     ! This is a routine I made up (pdf is in noisyLM folder).
!     ! Compute a linear projection of the function at hand via OLS and random points to obtain an 
!     ! approximate Jacobian. Then this Jacobian is used in the usual Levenberg-Marquardt procedure.
!     ! 
!     subroutine sub_parnoisyLM(x,f,nf,stdev,ndraws,opts)
!         use mod_mympi
!         use mod_matlab
!         implicit none
!         integer, intent(in) :: nf
!         real(rt), dimension(:), intent(inout) :: x
!         interface 
!             ! f is vector valued taking x (a length nx vector) to a length nf vector.
!             function f(x,nf) 
!                 implicit none
!                 integer, intent(in) :: nf
!                 real(rt), dimension(:), intent(in) :: x
!                 real(rt), dimension(nf) :: f
!             end function
!         end interface
!         real(rt), dimension(1:size(x)), intent(in), optional :: stdev
!         integer, intent(in), optional :: ndraws
!         type(minNdim_opts), optional :: opts
!         !local
!         type(minNdim_opts) :: o
!         real(rt), allocatable :: stdev_(:), xtrial(:), e_(:,:), Ymattranspose(:,:), Ymat(:,:), Xmat(:,:), J(:,:), fx(:), gL(:), s(:), eyex(:,:), ftrial(:), xtrials(:,:), ftrials(:,:)
!         integer, allocatable :: request(:)
!         logical, allocatable :: isFree(:)
!         integer :: iter, nx, itmp(1), nL, iL, idraw, ndraws_, iproc
!         logical :: haveRootDoIter1
!         real(rt) :: gradSSRSnorm,stepnorm,lambda,SSRSx,tmpSSRS,lambdabest
! 
!         if (mympi_nproc()==1) then
!             write(*,*) 'WARNING: If only using one node, call sub_noisyLM instead'
!             write(*,*) 'WARNING: Calling it for you with default parameters'
!             call sub_noisyLM(x,f,nf)
!             return
!         end if
! 
!         nx = size(x)
!         allocate(fx(nf))
! 
!     if (mympi_iproc()==0) then
! 
!         if (present(opts)) o = opts
! 
!         ndraws_ = 100 ! 10*nx
!         if (present(ndraws)) ndraws_ = ndraws
! 
!         if (ndraws_<mympi_nproc()-1) then
!             ! Increase ndraws so don't waste computation
!             write(*,*) 'WARNING: Increasing ndraws to not waste nodes'
!             ndraws_ = mympi_nproc()-1 
!         end if
!             
!         ! Set nL so that node nodes are wasted and that at least 12 points are evaluated
!         nL = max(12,mympi_nproc()-1)
! 
!         allocate(s(nx))
!         allocate(gL(nL))
!         allocate(stdev_(nx))
!         allocate(xtrial(nx))
!         allocate(ftrial(nf))
!         allocate(xtrials(nx,nL))
!         allocate(ftrials(nf,nL))
!         allocate(e_(nx,ndraws_))
!         allocate(Ymat(ndraws_,nf))
!         allocate(Ymattranspose(nf,ndraws_))
!         allocate(Xmat(ndraws_,nx))
!         allocate(eyex(nx,nx))
!         allocate(isFree(1:mympi_nproc()-1))
!         allocate(request(1:mympi_nproc()-1))
! 
!         isFree = .true.
!         eyex = eye(nx)
! 
!         if (nL==12) then
!             gL(1:12) = [0e0_rt,1d-6,1d-5,1d-4,1d-3,1d-2,1d-1,1e0_rt,1d1,1d2,1d3,1d4]
!         else
!             gL(1) = 0e0_rt
!             gL(2:nL) = logspace(-6e0_rt,4e0_rt,nL-1)
!         end if
! 
!         ! Give each process a random seed based on the clock
!         call init_random_seed(mympi_iproc())
! 
!         !Set default values
!         stdev_ = 1d-3
! 
!         !Overwrite if applicable
!         if (present(stdev)) stdev_ = stdev
! 
!         ! Determine whether the root node should compute f on the first iteration
!         ! This is done purely to avoid wasting resources because the first iteration requires 
!         ! ndraws_+1 function evaluations whereas every subsequent iteration requires ndraws_.
!         haveRootDoIter1 = ndraws_ == mympi_nproc()-1
! 
!         ! Main loop
!         do iter = 1,o%maxiter
! 
!             ! Compute draws
!             do idraw = 1,ndraws_
!                 e_(:,idraw) = stdev_*randn(nx)
!             end do
! 
!             ! Compute function values
!             do idraw = 1,merge(ndraws_+1,ndraws_,iter==1)
! 
!                 if (idraw==ndraws_+1) then
!                     ! Special case: first iteration, need to compute f of initial guess
!                     xtrial = x
!                 else
!                     ! Get xtrial
!                     xtrial = x + e_(:,idraw)
!                 end if
! 
!                 ! Compute the trial point
!                 if (idraw==ndraws_+1 .and. haveRootDoIter1) then
! 
!                     fx = f(xtrial,nf)
! 
!                 else
! 
!                     ! If there are no free nodes, wait for one to be free
!     ! write(*,'(A)',advance='no') '|'
!                     do while (.not. any(isFree))
! 
!                         ! Test each node to see if free
!                         do iproc = 1,mympi_nproc()-1
!                             
!                             ! Check whether finished
!                             isFree(iproc) = mympi_test(request(iproc))
! 
!                         end do
!                         
!                     end do
! 
!                     ! Use the free node
!                     if (any(isFree)) then
! 
!                         ! If there is a free node, get it
!                         iproc = find1(isFree,'first')
!                     
!                     else
!                         
!                         call mympi_stop('ERROR: could not find free node')
!                         
!                     end if
! 
!                     ! Send free node xtrial
!                     call mympi_send(xtrial,iproc)
! 
!                     ! Request the value of f(xtrial,nf) in Ymat(idraw,:)
!                     if (idraw==ndraws_+1) then
!                         ! print*,mympi_iproc(),': non-blocking receive of fx from ',iproc
!                         call mympi_irecv(request(iproc),fx,iproc)
!                     else
!                         ! print*,mympi_iproc(),': non-blocking receive of Ymattran from ',iproc
!                         call mympi_irecv(request(iproc),Ymattranspose(:,idraw),iproc)
!                     end if
! 
!                     ! Mark as not free 
!                     isFree(iproc) = .false.
! 
!                 end if
! 
!             end do
! 
!             ! Force all requests to complete before continuing
!             ! print*,mympi_iproc(),': waiting for all processes to complete'
!             call mympi_waitall(pack(request,.not. isFree))
!             isFree = .true.
! 
!             ! If first iteration, save SSRSx
!             if (iter==1) then
!                 SSRSx = dot_product(fx,fx)
!             end if
! 
!             ! Evaluate f at the trial points
!             Ymat = transpose(Ymattranspose)
!             Xmat = transpose(e_)
!             J = transpose(mldivide(matmul(transpose(Xmat),Xmat),matmul(transpose(Xmat),Ymat)))
! 
!             gradSSRSnorm = sqrt(dot_product(matmul(transpose(J),fx),matmul(transpose(J),fx)))
! 
!             ! For lambda on the line search, compute x+s and f(x+s)
!             do iL = 1,nL
! 
!                 lambda = gL(iL)
! 
!                 ! s = - mldivide(matmul(transpose(J),J) + lambda*eyex, matmul(transpose(J),fx))
!                 s = - mldivide(matmul(transpose(J),J) + lambda*diag(diag(matmul(transpose(J),J))), matmul(transpose(J),fx))
! 
!                 xtrials(:,iL) = x + s
! 
!                 ! If there are no free nodes, wait for one to be free
!                 do while (.not. any(isFree))
! 
!                     ! Test each node to see if free
!                     do iproc = 1,mympi_nproc()-1
!                         
!                         ! Check whether finished
!                         isFree(iproc) = mympi_test(request(iproc))
! 
!                     end do
!                     
!                 end do
! 
!                 ! If there is a free node, get it
!                 iproc = find1(isFree,'first')
! 
!                 ! Send the free node xtrial
!                 call mympi_send(xtrials(:,iL),iproc)
! 
!                 ! Request the value of f(xtrial,nf) in ftrials(:,iL)
!                 call mympi_irecv(request(iproc),ftrials(:,iL),iproc)
! 
!                 ! Mark as not free 
!                 isFree(iproc) = .false.
! 
!                 if (iL==nL) then
! 
!                     ! Wait for everyone to finish
!                     call mympi_waitall(pack(request,.not. isFree))
!                     isFree = .true.
! 
!                 end if
! 
!             end do 
! 
!             itmp = minloc(sum(ftrials**2,dim=1))
!             tmpSSRS = sum(ftrials(:,itmp(1))**2)
! 
!             if (tmpSSRS<SSRSx) then
! 
!                 ! Best line search is better than current best
!                 lambdabest = gL(itmp(1))
!                 s = xtrials(:,itmp(1)) - x
!                 stepnorm = sqrt(dot_product(s,s))
!                 x = xtrials(:,itmp(1))
!                 fx = ftrials(:,itmp(1))
!                 SSRSx = tmpSSRS
! 
!             else
! 
!                 ! Keep old value
!                 lambdabest = -1e0_rt
!                 stepnorm = 0e0_rt
! 
!             end if
! 
!             if (o%printmod>0 .and. mod(iter,o%printmod)==0) write(*,'(A,i8,4(A,d12.5))') ' NLM: iter',iter, ' step norm ',stepnorm,' SSRS ',SSRSx,' SSRS grad norm', gradSSRSnorm,' lambda ', lambdabest
! 
!             ! Check convergence criteria
!             if (stepnorm<o%xtoler .and.  gradSSRSnorm<o%xtoler) exit
!         
!         end do
! 
!         ! Tell every node to stop by sending them -huge(0e0_rt)
!         do iproc = 1,mympi_nproc()-1
!             xtrial = -huge(0e0_rt)
!             call mympi_send(xtrial,iproc)
!         end do
! 
!     else
!         
!         ! If not the root node, enter an infinite loop where you wait to receive an x value. Compute f(x) when you do.
!         do 
! 
!             ! Receive from root
!             ! print*,mympi_iproc(),': receiving x from root'
!             call mympi_recv(x,0) 
! 
!             ! Test whether x is less than -1d100, signaling it is time to stop the loop
!             if (all(x<-1d100)) exit
! 
!             ! Evaluate f
!             fx = f(x,nf)
! 
!             ! Send the value back to root
!             ! print*,mympi_iproc(),': sending fx to root'
!             call mympi_send(fx,0)
! 
!         end do
! 
! 
!     end if
! 
!         ! Send x from root to every node
!         call mympi_bcast(x,0)
! 
!     end subroutine 
! 
!     
!     ! Noisy Levenberg-Marquardt
!     !
!     ! solve min_{x \in R^{nx}} \sum^{nf}_{j=1} f^j(x)
!     !
!     ! This is a routine I made up (pdf is in noisyLM folder).
!     ! Compute a linear projection of the function at hand via OLS and random points to obtain an 
!     ! approximate Jacobian. Then this Jacobian is used in the usual Levenberg-Marquardt procedure.
!     ! 
!     subroutine sub_noisyLM(x,f,nf,stdev,ndraws,opts)
!         use mod_mympi
!         use mod_matlab
!         implicit none
!         integer, intent(in) :: nf
!         real(rt), dimension(:), intent(inout) :: x
!         interface 
!             ! f is vector valued taking x (a length nx vector) to a length nf vector.
!             function f(x,nf) 
!                 implicit none
!                 integer, intent(in) :: nf
!                 real(rt), dimension(:), intent(in) :: x
!                 real(rt), dimension(nf) :: f
!             end function
!         end interface
!         real(rt), dimension(1:size(x)), intent(in), optional :: stdev
!         integer, intent(in), optional :: ndraws
!         type(minNdim_opts), optional :: opts
!         !local
!         type(minNdim_opts) :: o
!         real(rt), allocatable :: stdev_(:), xtrial(:), e_(:,:), Ymat(:,:), Xmat(:,:), J(:,:), fx(:), gL(:), s(:), eyex(:,:), ftrial(:), fxbest(:), xbest(:)
!         integer :: iter, nx, nL, iL, idraw, ndraws_
!         real(rt) :: gradSSRSnorm,stepnorm,lambda,SSRSx,tmpSSRS,lambdabest,SSRSbest
! 
!         if (present(opts)) o = opts
! 
!         nx = size(x)
! 
!         ndraws_ = 100 
!         if (present(ndraws)) ndraws_ = ndraws
! 
!         nL = 12
! 
!         allocate(s(nx))
!         allocate(xbest(nx))
!         allocate(fx(nf))
!         allocate(fxbest(nf))
!         allocate(gL(nL))
!         allocate(stdev_(nx))
!         allocate(xtrial(nx))
!         allocate(ftrial(nf))
!         allocate(e_(nx,ndraws_))
!         allocate(Ymat(ndraws_,nf))
!         allocate(Xmat(ndraws_,nx))
!         allocate(eyex(nx,nx))
! 
!         eyex = eye(nx)
! 
!         if (nL==12) then
!             gL(1:12) = [0e0_rt,1d-6,1d-5,1d-4,1d-3,1d-2,1d-1,1e0_rt,1d1,1d2,1d3,1d4]
!         else
!             gL(1) = 0e0_rt
!             gL(2:nL) = logspace(-6e0_rt,4e0_rt,nL-1)
!         end if
! 
!         ! Give each process a random seed based on the clock
!         call init_random_seed(mympi_iproc())
! 
!         !Set default values
!         stdev_ = 1d-3
! 
!         !Overwrite if applicable
!         if (present(stdev)) stdev_ = stdev
! 
! 
!         ! Initialize
!         fx = f(x,nf) 
!         SSRSx = dot_product(fx,fx)
! 
!         ! Main loop
!         do iter = 1,o%maxiter
! 
!             do idraw = 1,ndraws_
! 
!                 ! Compute draws
!                 e_(:,idraw) = stdev_*randn(nx)
! 
!                 xtrial = x + e_(:,idraw)
! 
!                 ! Evaluate f at the trial points
!                 Ymat(idraw,:) = f(xtrial,nf)
! 
!             end do
! 
!             Xmat = transpose(e_)
!             J = transpose(mldivide(matmul(transpose(Xmat),Xmat),matmul(transpose(Xmat),Ymat)))
! 
!             gradSSRSnorm = sqrt(dot_product(matmul(transpose(J),fx),matmul(transpose(J),fx)))
! 
!             ! Do a line search
!             ! Initialize values to current
!             xbest = x
!             fxbest = fx
!             SSRSbest = SSRSx
!             lambdabest = -1e0_rt ! keeping current value 
!             stepnorm = 0e0_rt
!             do iL = 1,nL
!                 lambda = gL(iL)
! 
!                 ! s = - mldivide(matmul(transpose(J),J) + lambda*eyex, matmul(transpose(J),fx))
!                 s = - mldivide(matmul(transpose(J),J) + lambda*diag(diag(matmul(transpose(J),J))), matmul(transpose(J),fx))
! 
!                 xtrial = x + s
!                 ftrial = f(xtrial,nf)
!                 tmpSSRS = dot_product(ftrial,ftrial)
! 
!                 if (tmpSSRS<SSRSbest) then
!                     xbest = xtrial
!                     fxbest = ftrial
!                     SSRSbest = tmpSSRS
!                     lambdabest = lambda
!                     stepnorm = sqrt(dot_product(s,s))
!                 end if
! 
!             end do
! 
!             x = xbest
!             fx = fxbest
!             SSRSx = SSRSbest
! 
!             if (o%printmod>0 .and. mod(iter,o%printmod)==0) write(*,'(A,i8,4(A,d12.5))') ' NLM: iter',iter, ' step norm ',stepnorm,' SSRS ',SSRSx,' SSRS grad norm', gradSSRSnorm,' lambda ', lambdabest
! 
!             ! Check convergence criteria
!             if (stepnorm<o%xtoler .and.  gradSSRSnorm<o%xtoler) exit
!         
!         end do
! 
!     end subroutine 


    ! MPI implementation of random search
    ! Beginning with an initial point x, draws N proposals x+eps where eps is N(0,factor^2*e^2)
    ! where N is the number of MPI ranks, e is given by the user, and factor is internally set.
    ! The factor begins at 1e0_rt. If during an iteration an improvement is found, then factor = 
    ! factor*expand. O/w, factor = factor*shrink. The routine ends when factor*e<xtoler*abs(x) uniformly.
    ! 
    ! NOTE: this sets the random seed to distinct values for each node based on the clock. 
    subroutine sub_parrandom(x,f,e_in,maxiter_in,printmod_in,xtoler_in,shrink_in,expand_in)
        use mod_mympi
        use mod_matlab
        implicit none
        real(rt), dimension(1:), intent(inout) :: x
        interface 
            function f(x) 
                use mod_types
                implicit none
                real(rt), dimension(1:), intent(in) :: x
                real(rt) :: f
            end function f
        end interface
        real(rt), dimension(1:size(x)), intent(in), optional :: e_in
        integer, intent(in), optional :: maxiter_in,printmod_in
        real(rt), intent(in), optional :: xtoler_in, shrink_in, expand_in
        !local
        real(rt), allocatable :: e(:), ftrials(:), xtrial(:) !, xtmp(:)
        real(rt) :: xtoler, fx, expand, shrink, factor
        integer :: maxiter, printmod, iter, nproc, nx, itmp(1) !, iproc

        nx = size(x)

        ! Give each process a random seed based on the clock
        call init_random_seed(mympi_iproc())

        !Set default values
        allocate(e(nx))
        e = max(.1e0_rt*abs(x),10e0_rt*epsilon(0e0_rt)) !the size of neighborhood about each point which is a 
                                   !hypersphere stretched to edge length e in each dimension
        maxiter = 100000
        printmod = 1
        xtoler = 1e-6_rt
        shrink =  .9e0_rt 
        expand = 1.1e0_rt

        !Overwrite if applicable
        if (present(maxiter_in)) maxiter = maxiter_in
        if (present(printmod_in)) printmod = printmod_in
        if (present(e_in)) e = e_in
        if (present(xtoler_in)) xtoler = xtoler_in
        if (present(shrink_in)) shrink = shrink_in
        if (present(expand_in)) expand = expand_in

        ! Get MPI info
        nproc = mympi_nproc()
        allocate(ftrials(nproc))
        allocate(xtrial(nx))
        ! allocate(xtmp(nx))

        ! Initialize
        fx = huge(0e0_rt) ! set large so replace for sure
        factor = 1e0_rt

        ! Main loop
        do iter = 1,maxiter

            ! Compute a trial value
            xtrial = x + randn(nx)*e*factor
!             do iproc = 1,nproc
!                 if (iproc - 1 == mympi_iproc()) then
!                     xtrial = x + randn(nx)*e*factor
!                 else
!                     ! So that xtrial is definitely different across all nodes, call many randn calls
!                     xtmp = randn(nx)
!                 end if
!             end do

            ! Special case: first iteration, master node, compute value for x
            if (iter==1 .and. mympi_iproc()==0) then
                xtrial = x
            end if

            ! Compute the function value, but only for your rank
            ftrials(mympi_iproc()+1) = f(xtrial)

            ! Gather the function values only
            call mympi_allgatherv(ftrials)

            ! Select the best, check whether better than the current by a significant margin.
            if (minval(ftrials)<fx - (10e0_rt*epsilon(0e0_rt))*abs(fx)) then
                itmp = minloc(ftrials)
                fx = ftrials(itmp(1))
                
                ! If you are the best rank, send xtrial to everyone else. O/w, recieve.
                call mympi_bcast(xtrial,itmp(1)-1)
                
                ! Now all ranks have the copy of the best x
                x = xtrial
                factor = factor*expand
            else
                factor = factor*shrink
            end if

            if (mympi_iproc()==0 .and. printmod>0 .and. mod(iter,printmod)==0) &
                write(*,'(A,i10,A,d12.5,A,d12.5)') ' PR: iter',iter, ' f',fx,' factor', factor

            ! Check stopping criteria
            if (all(factor*e<xtoler*abs(x))) exit
        
        end do

        if (mympi_iproc()==0 .and. printmod>0 .and. printmod==1) then
            write(*,*) 'PR: final fx', fx, 'final x',x
        end if

    end subroutine 

    ! This code is a random-start simplex. Random draws are taken from [a,b] for each
    ! node. Each node then optimizes starting from that point. The controls specified
    ! by e_in and opt are passed along to the simplex. 
    !
    ! This is a ridiculously parallel minization. The optimization never stops, so it 
    ! is up to the user to keep track of the best (which is displayed to stdout after
    ! each iteration).
    ! 
    subroutine sub_parmin_MPI(x,f,a,b,opt,e_in)
        use mod_util
        use mod_matlab
        use mod_mympi
        implicit none
        real(rt), intent(inout) :: x(:)
        real(rt), intent(in) :: a(:), b(:)
        type(minNdim_opts), intent(in), optional :: opt
        real(rt), intent(in), optional :: e_in(:)
        interface 
            function f(x) 
                use mod_types
                implicit none
                real(rt), dimension(:), intent(in) :: x
                real(rt) :: f
            end function f
        end interface
        ! local
        real(rt), allocatable :: xtmp(:)
        type(minNdim_opts) :: o
        type(type_seed) :: randstate
        integer :: flag,i,iter

        allocate(xtmp(size(x)))

        ! Set default parameters and overwrite if optional argument is present
        if (present(opt)) o = opt

        call check_guess_in_range(flag,x,a,b)
        if (flag/=0) write(0,*) 'WARNING: x not in a,b for all dimensions'
        if (any(b<a)) stop 'ERROR: b<a in at least one dimension'

    

        iter = 0 

        ! BEGIN INFINITE LOOP
        ! Never exit (this way all ranks keep working forever) 
        do 

            iter = iter + 1
            
            call randstate%save()
            call randstate%randomize()

            ! Generate initial points (here, burn through points to ensure each node gets a different x)
            do i = 0,mympi_nproc()-1

                ! Generate a point in [a,b]
                call random_number(xtmp)
                xtmp = a + (b-a)*xtmp

                ! Use as starting point if i is for your node
                if (i==mympi_iproc()) x = xtmp

            end do 

            call randstate%revert()

            write(*,*) 'PARMIN: iter: ',iter,' rank ',mympi_iproc(),' xbeforeNM = ',printx(x)
            ! Call the minimization routine in parallel
            if (present(e_in)) then
                call sub_simplex(x,f,e_in=e_in,maxiter_in=o%maxiter,printmod_in=o%printmod,xtoler_in=o%xabstoler)
            else
                xtmp = .1e0_rt*(b-a)
                call sub_simplex(x,f,e_in=xtmp,maxiter_in=o%maxiter,printmod_in=o%printmod,xtoler_in=o%xabstoler)
            end if

            write(*,*) 'PARMIN: iter: ',iter,' rank ',mympi_iproc(),' xafterNM = ',printx(x)

        end do

    end subroutine 

    
    ! Same as the above, but with an OpenMP implementation
    ! NOTE:
    ! If e_in is present, the results will not be reproducible (even calling random_seed directly before the call)
    ! because the draws are from xtrial+-e_in.
    ! However, if e_in is not present, then the results will be reproducible since all the trials are taken randomly
    ! from a,b and the same trials are taken each time. So, instaed of using e_in, would could call with a=x-e,b=x+e
    ! repeatedly and have something similar, but reproducible.
    !
    ! If seed is present, then it is used to set the seed for random_number upon entry and is replaced with the
    ! state of the seed after sub_parmin_OMP has completed.
    ! 
    ! If randunitid is present, the random numbers are read from the file associated with randunitid
    ! 
    ! If randvals is present, then the random numbers are taken from randvals (as if they were read in from file). 
    ! Hence, randvals must be sufficient for the computation. The variable i_randvals is required in this case
    ! and tells the routine where to begin reading randvals. It is modified on exit to return the last location.
    !
    ! THIS CODE HAS BEEN WORKED EXTENSIVELY TO PRODUCE EXACTLY THE SAME RESULTS THREADED AS NON-THREADED (PROVIDED
    ! THE FUNCTION F DOES NOT HAVE ANY OMP ISSUES). THE ORDER IN WHICH ITERATIONS ARE EXECUTED AND THE TRIALS
    ! ARE EXACTLY THE SAME AS IN THE SERIAL CODE.
    subroutine sub_parmin_OMP(x,f,a,b,e_in,opt,seed,randunitid,randvals,i_randvals)
        use mod_matlab
        use omp_lib
        use mod_util
        implicit none
        real(rt), intent(inout) :: x(:)
        real(rt), intent(in) :: a(:),b(:)
        real(rt), optional, intent(in) :: e_in(:)
        type(minNdim_opts), intent(in), optional :: opt
        integer, intent(inout), optional :: seed(:)
        integer, intent(in), optional :: randunitid
        real(rt), intent(in), optional :: randvals(:)
        integer, intent(inout), optional :: i_randvals
        interface 
            function f(x) 
                use mod_types
                implicit none
                real(rt), dimension(:), intent(in) :: x
                real(rt) :: f
            end function f
        end interface
        ! local
        integer :: i,iter,tid,priviter,nthreads
        integer :: whosturn1,whosturn2
        integer :: whosturn1priv,whosturn2priv
        type(minNdim_opts) :: o
        real(rt), allocatable :: xtrial(:)
        real(rt) :: fx,fxtrial

        if (present(opt)) o = opt

        if (o%poolsize==-1) then
            write(0,*) 'WARNING: sub_parmin_OMP: poolsize should be specified. Setting to 1000.'
            o%poolsize = 1000
        end if

        if (any(b<a)) stop 'ERROR: sub_parmin: b is less than a'
        if (any(x<a .or. x>b)) stop 'ERROR: sub_parmin: x is not in [a,b]'

        iter = 0

        fx = f(x)

        if (present(seed)) then
            call random_seed(size=i)
            if (size(seed)/=i) stop 'ERROR: seed is not the correct length' 
            call random_seed(put=seed)
        end if

!         call dbgmsg(1,'about to enter parallel section')
    
        ! For strict reproducibility, need an ordered construct to handle cases of indifference. 

        !$OMP PARALLEL DEFAULT(NONE) SHARED(x,fx,o,a,b,whosturn1,whosturn2,randunitid,iter,nthreads,e_in,randvals,i_randvals) &
        !$OMP PRIVATE(xtrial,fxtrial,i,tid,priviter,whosturn1priv,whosturn2priv) 

        if (allocated(xtrial)) deallocate(xtrial)
        allocate(xtrial(size(x)))

        ! Get thread info
        tid = omp_get_thread_num()
        nthreads = omp_get_num_threads()

        ! Let thread 0 begin
        whosturn1 = 0
        whosturn2 = 0

        !$OMP BARRIER

!         print*,'tid',tid,'nthreads',nthreads,'whosturn1',whosturn1,'whosturn2',whosturn2

        ! Set counters to 0
        priviter = 0

        do i = 1,o%poolsize

            if (tid/=mod(i-1,nthreads)) cycle
    
            ! Wait for your turn to begin executiong
            do 
                !$OMP FLUSH (whosturn1)
                ! !$OMP ATOMIC READ
                whosturn1priv = whosturn1
                ! !$OMP END ATOMIC
                if (tid==mod(whosturn1priv,nthreads)) exit
                call spin(priviter)
            end do
            
            !$OMP CRITICAL (TURN1)

            ! Generate a new trial point
            if (present(randvals)) then
                xtrial = randvals(i_randvals:i_randvals+size(xtrial)-1)
                i_randvals = i_randvals + size(xtrial)
            elseif (present(randunitid)) then
                read(randunitid,*) xtrial
            else
                call random_number(xtrial)
            end if

            ! Generate a random point in [a,b]
            if (present(e_in)) then
                ! Draw uniformly about current best +/- e
                xtrial = x + (2e0_rt*xtrial - 1e0_rt)*e_in
                xtrial = max(a,min(b,xtrial)) ! bound
            else
                ! Draw uniformly from bounds
                xtrial = a + xtrial*(b-a)
            end if

            ! To ensure descent, make sure that do simplex from x
            if (tid==0 .and. priviter==0) xtrial = x 

            priviter = priviter + 1

            ! Free the next worker
            whosturn1 = whosturn1 + 1

            !$OMP END CRITICAL (TURN1)
            
            ! Call the minimization routine in parallel
            call sub_simplex(xtrial,f,a_in=a,b_in=b,e_in=(b-a)*.5e0_rt,xtoler_in=o%xabstoler,maxiter_in=o%maxiter,printmod_in=-1)

            ! Get value of best point found, xtrial
            fxtrial = f(xtrial)

            ! Wait for your turn to compare
            do
                !$OMP FLUSH (whosturn2)
                ! !$OMP ATOMIC READ
                whosturn2priv = whosturn2
                ! !$OMP END ATOMIC
                if (tid==mod(whosturn2priv,nthreads)) exit ! now its your turn
                call spin(priviter)
            end do

            !$OMP CRITICAL (TURN2)
            iter = iter + 1

            if (mod(iter-1,nthreads)/=tid) then
                write(0,*) 'WARNING: sub_parmin_OMP: execution of iterations was not ordered properly, there is a bug!'
            end if

            ! Compare against current best
            if (fxtrial<fx) then
                x = xtrial
                fx = fxtrial
            end if
            if (o%printmod>0 .and. mod(iter,o%printmod)==0) then
                if (o%displaybestx==1) then
                    write(*,'(A,i4,2(A,d16.8),2A)') 'PARMINOMP: iter ',iter,' best f ',fx,' trial f ',fxtrial,' best x ',printx(x)
                else
                    write(*,'(A,i4,2(A,d16.8))')    'PARMINOMP: iter ',iter,' best f ',fx,' trial f ',fxtrial
                end if
            end if

            ! Free the next worker
            whosturn2 = whosturn2 + 1

            !$OMP END CRITICAL (TURN2)

        end do

        ! don't exit the parallel region til everyone is finished (I think this is not necessary)
        !$OMP BARRIER 

        deallocate(xtrial)

        !$OMP END PARALLEL 

        ! Output the seed if requested
        if (present(seed)) then
            call random_seed(get=seed)
        end if

!     contains
!         subroutine dbgmsg(priority,msg)
!             implicit none
!             character(len=*), intent(in) :: msg
!             integer, intent(in) :: priority
!             if (priority>=o%dbglvl)  print*,'DEBUG: ',priority,msg
!         end subroutine 

    end subroutine 

    subroutine spin(i)
        implicit none
        integer :: i,j
        real(rt) :: tmp

        tmp = .12e0_rt + i

        do j = 0,1000
            tmp = cos(sin(tmp))
            if (tmp==2e0_rt) then
                write(0,*) 'This should not happen'
                exit
            end if
        end do
    end subroutine 

    ! Same as _OMP but with all openmp directives removed.
    subroutine sub_parmin_NOPAR(x,f,a,b,e_in,opt,seed,randunitid)
        use mod_matlab
        use omp_lib
        use mod_util
        implicit none
        real(rt), intent(inout) :: x(:)
        real(rt), intent(in) :: a(:),b(:)
        real(rt), optional, intent(in) :: e_in(:)
        type(minNdim_opts), intent(in), optional :: opt
        integer, intent(inout), optional :: seed(:)
        integer, intent(in), optional :: randunitid
        interface 
            function f(x) 
                use mod_types
                implicit none
                real(rt), dimension(:), intent(in) :: x
                real(rt) :: f
            end function f
        end interface
        ! local
        integer :: i,iter,tid,priviter
        type(minNdim_opts) :: o
        real(rt), allocatable :: xtrial(:)
        real(rt) :: fx,fxtrial

        allocate(xtrial(size(x)))

        if (present(opt)) o = opt

        if (o%poolsize==-1) then
            write(0,*) 'WARNING: sub_parmin_OMP: poolsize should be specified. Setting to 1000.'
            o%poolsize = 1000
        end if

        if (any(b<a)) stop 'ERROR: sub_parmin: b is less than a'
        if (any(x<a .or. x>b)) stop 'ERROR: sub_parmin: x is not in [a,b]'

        iter = 0
        priviter = 0

        fx = f(x)

        if (present(seed)) then
            call random_seed(size=i)
            if (size(seed)/=i) stop 'ERROR: seed is not the correct length' 
            call random_seed(put=seed)
        end if
    
        ! For strict reproducibility, need an ordered construct to handle cases of indifference. 

        do i = 1,o%poolsize

            tid = omp_get_thread_num()

            if (present(randunitid)) then
                read(randunitid,*) xtrial
            else
                call random_number(xtrial)
            end if

            ! Generate a random point in [a,b]
            if (present(e_in)) then
                ! Draw uniformly about current best +/- e
                xtrial = x + (2e0_rt*xtrial - 1e0_rt)*e_in
                xtrial = max(a,min(b,xtrial)) ! bound
            else
                ! Draw uniformly from bounds
                xtrial = a + xtrial*(b-a)
            end if

            ! To ensure descent, make sure that do simplex from x
            if (tid==0 .and. priviter==0) then
                xtrial = x 
                priviter = 1
            end if
            
            ! Call the minimization routine in parallel
            call sub_simplex(xtrial,f,a_in=a,b_in=b,e_in=(b-a)*.5e0_rt,xtoler_in=o%xabstoler,maxiter_in=o%maxiter,printmod_in=-1)

            ! Call two more times to compare against initial x. This way, the routine never does worse than the input guess.
            fxtrial = f(xtrial)

            if (fxtrial<fx) then
                x = xtrial
                fx = fxtrial
            end if
            iter = iter + 1
            if (o%printmod>0 .and. mod(iter,o%printmod)==0) then
                if (o%displaybestx==1) then
                    write(*,'(A,i4,2(A,d16.8),2A)') 'PARMINOMP: iter ',iter,' best f ',fx,' trial f ',fxtrial,' best x ',printx(x)
                else
                    write(*,'(A,i4,2(A,d16.8))')    'PARMINOMP: iter ',iter,' best f ',fx,' trial f ',fxtrial
                end if
            end if

        end do

        ! Output the seed if requested
        if (present(seed)) then
            call random_seed(get=seed)
        end if

    end subroutine 

    ! Maximizes using a genetic algorithm
    ! Note that this routine is different from the rest because
    ! (1) the search space is integers
    ! (2) it maximizes (as this is more natural in thinking about fitness)
    !
    ! By default, searches in binary.
    ! 
    ! The stopping criterion is measured in terms of the standard deviations of the fitness
    ! in the population. When this falls below toler, execution is terminated and the 
    ! best point is returned.
    ! 
    ! The crossover method can be selected through the cross option
    ! cross             meaning
    ! 0                 single crossover point
    ! 1 (default)       double crossover point
    ! ... others not yet implemented
    subroutine sub_ga(x,f,npop_in,nelite_in,pmute_in,cross_in,toler_in,maxiter_in,printmod_in,&
                        maxfevals_in,openmp_in,mpi_in)
        use mod_matlab
        implicit none
        integer, dimension(:), intent(inout) :: x
        interface 
            function f(x) 
                use mod_types
                implicit none
                integer, dimension(:), intent(in) :: x
                real(rt) :: f
            end function f
        end interface
        real(rt), intent(in), optional :: toler_in,pmute_in
        integer, intent(in), optional :: maxiter_in,printmod_in,maxfevals_in
        integer, intent(in), optional :: npop_in,nelite_in,cross_in
        logical, intent(in), optional :: openmp_in,mpi_in
        !local
        !controls
        integer :: maxiter,printmod,maxfevals,cross
        integer :: npop ! number of members in the population
        integer :: nelite ! number of elite members to carry over unchanged 
        logical :: openmp,mpi

        real(rt) :: pmute ! pmute/nx is the probability of a particular gene mutating
        real(rt), parameter :: pcross = .65e0_rt ! probability of two parents crossing
        
        !tools
        integer, dimension(:,:), allocatable :: pop ! population members
        integer, dimension(:,:), allocatable :: parents ! new population members
        integer, dimension(:,:), allocatable :: elite ! elite member
        real(rt), dimension(:), allocatable :: fit ! fitness of population members
        real(rt), dimension(:), allocatable :: probdraw ! used for drawing survivors
        real(rt), dimension(:), allocatable :: fit_elite
        real(rt) :: tmp,toler
        integer :: ipop,ix,nx,iter,itmp,itmp2,itmp3,fevals,ivec1(1)

        nx = size(x)        

        !Set defaults
        maxiter = 1000
        npop = 32
        nelite = 1
        pmute = 1e0_rt
        cross = 1
        printmod = 1
        toler = 1e-9_rt
        openmp = .false.
        mpi = .false.
        
        ! Overwrite if present
        if (present(maxiter_in)) maxiter = maxiter_in
        if (present(maxfevals_in)) maxfevals = maxfevals_in
        if (present(printmod_in)) printmod = printmod_in
        if (present(toler_in)) toler = toler_in 
        if (present(openmp_in)) openmp = openmp_in 
        if (present(mpi_in)) mpi = mpi_in  
        if (present(npop_in)) npop = npop_in
        if (present(cross_in)) cross = cross_in
        if (present(nelite_in)) nelite = nelite_in
        if (present(pmute_in)) pmute = pmute_in

        ! Set other defaults
        if (.not. present(maxfevals_in)) maxfevals = 1000*npop

        ! Check inputs
        if (mod(npop,2)/=0) STOP 'npop must be even'
        if (nelite>1) STOP 'nelite must less than or equal 1 or need to reprogram'

        ! Allocate space
        allocate(pop(nx,npop),parents(nx,npop),fit(npop),probdraw(npop))
        if (nelite>0) allocate(elite(nx,nelite),fit_elite(nelite))

        if (printmod>=0) write(*,'(A)') 'GA:________________________________________________________________________'            
        if (printmod>=0) write(*,'(A)') 'GA: Beginning Genetic Algorithm'            
        if (printmod>=0) write(*,'(A)') 'GA:  iter          best         worst           avg          stdev   fevals'

        !Create initial population by random draws except for input chromosone x
        pop(:,1) = x ! 
        do ipop = 2,npop
            do ix = 1,nx
                call random_number(tmp)
                pop(ix,ipop) = merge(1,0,tmp>.5e0_rt) ! binary
            end do
        end do

        fevals = 0
        if (nelite>0) fit_elite = -huge(0e0_rt)
        
        if (printmod==0) write(*,*) 'GA: Evaluating initial population: '

        ! Iterate
        do iter = 1,maxiter

            ! Evaluate the fitness of the population
            if (.not. openmp .and. .not. mpi) then
                do ipop = 1,npop
                    fit(ipop) = f(pop(:,ipop))
                end do
            elseif (openmp) then
                !$OMP PARALLEL DO DEFAULT(SHARED) PRIVATE(ipop)
                do ipop = 1,npop
                    fit(ipop) = f(pop(:,ipop))
                end do
                !$OMP END PARALLEL DO
            elseif (mpi) then
                STOP 'have not programmed mpi yet'
            end if
            fevals = fevals + npop


            ! Check that fitness is all positive
            if (any(fit<0e0_rt)) STOP 'ERROR: GA: fitness is negative but should be positive'

            ! Drop the worst performing completely and add back in elite
            if (nelite==1) then
                if (maxval(fit)>fit_elite(1)) then
                    ivec1 = maxloc(fit)
                    elite(:,1) = pop(:,ivec1(1))
                    fit_elite(1) = fit(ivec1(1))
                else
                    ivec1 = minloc(fit)
                    pop(:,ivec1(1)) = elite(:,1)
                    fit(ivec1(1)) = fit_elite(1)
                end if
            end if

            ! Display some info
            if (printmod==0) call disp('GA: pop fitness: ',fit)
            if (printmod==0 .or. (printmod>0 .and. mod(iter,printmod)==0)) &
                write(*,'(A,i6,A,e12.5,A,e12.5,A,e12.5,A,e12.5,A,i6)') &
                'GA:',iter,'  ', maxval(fit),'  ', minval(fit),'  ', mean(fit),'   ',std(fit),'   ',fevals

            ! Check stopping criteria
            if (iter==maxiter .or. fevals>=maxfevals .or. std(fit)<=toler) then
                print*,'iter',iter,'maxiter',maxiter
                print*,'fevals',fevals,'maxfevals',maxfevals
                print*,'std(fit)',std(fit),'toler',toler
                print*,'fit',fit
                exit
            end if

            ! Generate a new population 
            probdraw = cumsum(fit)/sum(fit)

            ! Roulette selection
            !!! Draw parents according to cumsum(fit)/sum(fit)
            do ipop = 1,npop
                call random_number(tmp)
                itmp = find1(tmp<=probdraw,'first')
                parents(:,ipop) = pop(:,ipop)
            end do

            ! Crossover the parents
            select case (cross)
                case (0)
                    !!! One-point crossover
                    do ipop = 1,npop-1,2

                        call random_number(tmp)
                        if (tmp>pcross) then 

                            ! Don't cross
                            pop(:,ipop) = parents(:,ipop)
                            pop(:,ipop+1) = parents(:,ipop+1)

                        else

                            !!! Draw one crossover 
                            call random_number(tmp)
                            itmp = nint(1e0_rt + tmp*(nx-1)) ! scale to 1,nx and round to nearest int

                            ! Now have a crossover point, so do it
                            pop(1:itmp,ipop) = parents(1:itmp,ipop)
                            pop(itmp+1:nx,ipop) = parents(itmp+1:nx,ipop+1)

                            pop(1:itmp,ipop+1) = parents(1:itmp,ipop+1)
                            pop(itmp+1:nx,ipop+1) = parents(itmp+1:nx,ipop)

                        end if

                    end do
                case (1)
                    !!! Two-point crossover
                    do ipop = 1,npop-1,2

                        call random_number(tmp)
                        if (tmp>pcross) then 

                            ! Don't cross
                            pop(:,ipop) = parents(:,ipop)
                            pop(:,ipop+1) = parents(:,ipop+1)

                        else

                            !!! Two-point crossover. Draw two crossover points
                            call random_number(tmp)
                            itmp = nint(1e0_rt + tmp*(nx-1)) ! scale to 1,nx and round to nearest int
                            call random_number(tmp)
                            itmp2 = nint(1e0_rt + tmp*(nx-1)) ! ditto

                            itmp3 = itmp
                            itmp = min(itmp3,itmp2)
                            itmp2 = max(itmp3,itmp2)

                            ! Now have two points for cross over. They may be equal.
                            pop(1:itmp,ipop) = parents(1:itmp,ipop)
                            pop(itmp+1:itmp2,ipop) = parents(itmp+1:itmp2,ipop+1)
                            pop(itmp2+1:nx,ipop) = parents(itmp2+1:nx,ipop)

                            pop(1:itmp,ipop+1) = parents(1:itmp,ipop+1)
                            pop(itmp+1:itmp2,ipop+1) = parents(itmp+1:itmp2,ipop)
                            pop(itmp2+1:nx,ipop+1) = parents(itmp2+1:nx,ipop+1)

                        end if

                    end do

!                 case (2)
!                     !!! WARNING: this is not actually uniform crossover
!                     !!! Uniform crossover
!                     do ipop = 1,npop-1,2
!                         do ix = 1,nx
!                             !!! W.p. 1/2, swap the gene
!                             call random_number(tmp)
! 
!                             if (tmp>.5e0_rt) then
!                                 pop(ix,ipop) = parents(ix,ipop)
!                                 pop(ix,ipop+1) = parents(ix,ipop+1)
!                             else
!                                 pop(ix,ipop) = parents(ix,ipop+1)
!                                 pop(ix,ipop+1) = parents(ix,ipop)
!                             end if
! 
!                         end do
!                     end do
                case default
                    STOP 'ERROR: crossover case not found'
            end select

            ! Mutate the population
            do ipop = 1,npop
                do ix = 1,nx
                    call random_number(tmp)
                    if (tmp<pmute) pop(ix,ipop) = merge(1,0,pop(ix,ipop)==0) ! mutate
                end do
            end do

        end do

        ! Return the best point and exit
        ivec1 = maxloc(fit)
        x = pop(:,ivec1(1))

!         if (printmod>0 .AND. printmod==1) then
!             write(*,*) 'GA:Final minimum x: '
!             call disp(x)            
!             write(*,*) 'GA:Final minimum f: '
!             write(*,*) 'GA:',minval(f_simplex)
!         end if
        if (printmod>=0) write(*,'(A)') 'GA:________________________________________________________________________'           

    end subroutine sub_ga


!     ! sub_sequentialmin
!     ! Cycles through each dimension and optimizes it before moving onto the next dimension
!     ! Good for problems where one-dimension may be more important than others or the problem
!     ! is mostly independent. This is definitely *not* guaranteed to find a global min or even
!     ! a local min because doing so may require changing x in more than one direction. 
!     ! However, it can still be useful!
!     subroutine sub_sequentialmin(x,f,e_in,shrink_in,toler_in,xtoler_in,maxiter_in,printmod_in)
!         use mod_min1dim
!         use mod_matlab
!         implicit none
!         real(rt), dimension(1:), intent(inout) :: x
!         interface 
!             function f(x) 
!                 implicit none
!                 real(rt), dimension(1:), intent(in) :: x
!                 real(rt) :: f
!             end function f
!         end interface
!         real(rt), dimension(1:size(x)), intent(in), optional :: e_in
!         real(rt), intent(in), optional :: toler_in, xtoler_in
!         integer, intent(in), optional :: maxiter_in,printmod_in
!         real(rt), intent(in), optional :: shrink_in
!         !local
!         real(rt) :: toler, xtoler, shrink, lastf, enorm
!         integer :: maxiter,printmod
!         real(rt), dimension(1:size(x)) :: e
!         integer :: n,i,iter,flag
!         
!         n = size(x)        
! 
!         !Set defaults
!         maxiter = 1000
!         e = max(.1e0_rt*abs(x),.01e0_rt) !inital step size
!         shrink = .5e0_rt
!         xtoler = 1d-12
!         toler = -huge(0.e0_rt) !stop iterations if f(x)<toler for some x (note that f may be negative)
!         printmod = 1
! 
!         !Overwrite if present
!         if (present(e_in)) e = e_in
!         if (present(shrink_in)) shrink = shrink_in
!         if (present(xtoler_in)) xtoler = xtoler_in
!         if (present(toler_in)) toler = toler_in
!         if (present(maxiter_in)) maxiter = maxiter_in
!         if (present(printmod_in)) printmod = printmod_in
! 
!         if (printmod>0) then
!             write(*,'(A)') 'SM:__________________________________________________________________'            
!             write(*,'(A)') 'SM:Beginning Sequential Minimization Method  '        
!             write(*,'(A)') 'SM:    iter   dim   f(x)   enorm                                  '
!         end if 
! 
!         ! Do the search
!         do iter = 1,maxiter
! 
!             enorm = sqrt(sum(e**2))
!            
!             ! Exit if xtoler exceeded 
!             if (enorm < xtoler) exit 
! 
!             ! Search in each direction
!             do i = 1,n
!                 
!                 ! 
!                 call sub_simplex1dim(x(i),fonedim,e(i),1d-6,10,-1,flag)
! 
!                 ! Update status 
!                 if (printmod>0 .and. (iter==1 .or. mod(iter,printmod)==0)) &
!                     write(*,'(A,2i5,2f13.7)') 'SM:   ',iter,i,lastf,enorm
! 
!             end do
!             
!             ! So it will eventually converge, shrink e
!             e = shrink*e
! 
!         end do
! 
!         if (printmod>0) then
!             write(*,'(A)',advance='no') 'SM: Final x:'
!             write(*,'(99d16.6)',advance='no') x
!             write(*,*)
!             write(*,'(A)',advance='no') 'SM: Final e:'
!             write(*,'(99d16.6)',advance='no') e
!             write(*,*)
!             write(*,'(A)') 'SM:__________________________________________________________________'            
!         end if
! 
!     contains
!         
!         function fonedim(xi)
!             implicit none
!             real(rt), intent(in) :: xi
!             real(rt) :: fonedim
! 
!             x(i) = xi
!             lastf = f(x) ! Used for stats
!             fonedim = lastf
! 
!         end function 
! 
!     end subroutine sub_sequentialmin


    ! Stochastic Gradient Descent
    ! 
    ! This minimizes a function that can only be measured noisily.
    ! That is, min_x f(x) where f(x) = E(g(x,Z)) and z is a measurement error.
    ! Is essential a noisy version of gradient descent.
    !
    subroutine sub_stochGradDescent(x,f,e_in,a_in,b_in,lambda_in,scaling_in,xtoler_in,maxiter_in,printmod_in)
        use mod_deriv
        use mod_matlab
        implicit none
        real(rt), dimension(1:), intent(inout) :: x
        interface 
            function f(x) 
                use mod_types
                implicit none
                real(rt), dimension(1:), intent(in) :: x
                real(rt) :: f
            end function f
        end interface
        real(rt), dimension(1:size(x)), intent(in), optional :: e_in,a_in,b_in
        real(rt), intent(in), optional :: xtoler_in, scaling_in, lambda_in
        integer, intent(in), optional :: maxiter_in,printmod_in
        !local
        real(rt) :: xtoler, f0
        integer :: maxiter,printmod
        real(rt), dimension(1:size(x)) :: e, a, b, grad, scaling_init, scaling
        real(rt) :: snorm, lambda, lambda_init, tmp
        integer :: n,iter,fevals
        
        n = size(x)        

        ! Set defaults and overwrite if present
        e = max(.1e0_rt*abs(x),.01e0_rt); if (present(e_in)) e = e_in
        a = -huge(0e0_rt);             if (present(a_in)) a = a_in
        b = huge(0e0_rt);              if (present(b_in)) b = b_in
        maxiter = 1000;             if (present(maxiter_in)) maxiter = maxiter_in
        xtoler = 1e-6_rt;              if (present(xtoler_in)) xtoler = xtoler_in
        printmod = 1;               if (present(printmod_in)) printmod = printmod_in
        lambda_init = 1e0_rt;          if (present(lambda_in)) lambda_init = lambda_in
        scaling_init = 1e0_rt;         if (present(scaling_in)) scaling_init = scaling_in

        ! Init
        fevals = 0

        if (printmod>0) then
            write(*,'(A)') 'SGD:__________________________________________________________________'            
            write(*,'(A)') 'SGD:Beginning Stochastic Gradient Descent Method       '            
            write(*,'(A)') 'SGD:    iter  fevals         f(x)        snorm'
        end if 

    
        do iter = 1,maxiter
            
            ! Compute the gradient estimate
            scaling = scaling_init*real(iter,rt)**(-1e0_rt/3e0_rt)
            call sub_numGradHess(x,f,scaling,grad,f0=f0)
            fevals = fevals + n + 1
            ! Compute the new point: use line search with a maximum step of lambda
            tmp = lambda_init/real(iter,rt)
            call sub_lineSearch(x,lambda,-grad,f,f0,grad,lambda_init_in=tmp,maxiter_in=10)
            ! x = x - lambda*grad

            ! Compute the norm of the step size
            snorm = lambda*sqrt(sum(grad**2))

            ! Update status 
            if (printmod>0 .and. (iter==1 .or. mod(iter,printmod)==0)) &
                write(*,'(A,i5,i8,2f13.7)') 'SGD:   ',iter, fevals, f0, snorm

            ! Exit if converged
            if (snorm<xtoler) exit

        end do 

        if (printmod>0) then
            write(*,'(A)',advance='no') 'SGD: Final x:'
            write(*,'(99d16.6)',advance='no') x
            write(*,*)
            write(*,'(A)') 'SGD:__________________________________________________________________'            
        end if

        
    end subroutine 


!     ! Smolyak search
!     ! 
!     ! Smolyak search computes a smolyak collocation projection of the function f on a hypercube
!     ! defined by x +- e. Subsequently, the projection (which is an infinitely differentiable
!     ! function) is optimized using derivative based methods. After the projection is optimized,
!     ! the hypercube is recentered around the new x. The projection is applied again on the 
!     ! hypercube x +- e. If little improvement is made, the hypercube is shrunk: e = shrink*e and
!     ! the process begins again (terminated when the L2 norm of e falls below xtoler).
!     ! 
!     ! The philosophy behind this is that if the underlying function is "roughly smooth" --- 
!     ! i.e. almost differentiable but maybe not because of noise --- that we can get a close
!     ! approximation by the smolyak function and then optimize that. The smolyak projection (verse
!     ! a tensor product approx) is used because it is suitable for high dimension problems.
!     ! 
!     ! This is totally heuristic --- although I believe it could be formalized. I believe it is
!     ! my creation, but it is a fairly obvious.
!     !
!     ! The user may choose e, shrink, and also "mu", which is the degree of the smolyak 
!     ! approximation used. mu should generally be at least 2, because mu==1 has no crossterms
!     ! in the approximation. Also, for higher order approximations, there may be a lot of local
!     ! minima. Generally, the default value mu==2 is recommended. a,b, if present, are bounds
!     ! on the problem that are enforced in both in the smolyak approximation construction and 
!     ! the inner maximization (which is the simplex method currently).
!     ! 
!     subroutine sub_smolyaksearch(x,f,e_in,a_in,b_in,shrink_in,mu_in,xtoler_in,maxiter_in,printmod_in)
!         use mod_matlab
!         use mod_smolyak
!         implicit none
!         real(rt), dimension(1:), intent(inout) :: x
!         interface 
!             function f(x) 
!                 implicit none
!                 real(rt), dimension(1:), intent(in) :: x
!                 real(rt) :: f
!             end function f
!         end interface
!         real(rt), dimension(1:size(x)), intent(in), optional :: e_in,a_in,b_in
!         real(rt), intent(in), optional :: xtoler_in
!         integer, intent(in), optional :: maxiter_in,printmod_in
!         real(rt), intent(in), optional :: shrink_in
!         integer, intent(in), optional :: mu_in
!         !local
!         type(smolyakFcn) :: sApprox ! Smolyak approximation
!         real(rt) :: xtoler
!         integer :: maxiter,printmod
!         real(rt), dimension(1:size(x)) :: e, xoldbest, xbest, a, b, hcubelb, hcubeub
!         real(rt) :: shrink, enorm, snorm, fbest, foldbest
!         integer :: n,i,iter,fevals,mu,inner, tmpi1(1)
!         real(rt), allocatable :: coll(:,:),fcoll(:)
!         
!         n = size(x)        
! 
!         ! Set defaults and overwrite if present
!         mu = 2;                     if (present(mu_in)) mu = mu_in
!         e = max(.1e0_rt*abs(x),.01e0_rt); if (present(e_in)) e = e_in
!         a = -huge(0e0_rt);             if (present(a_in)) a = a_in
!         b = huge(0e0_rt);              if (present(b_in)) b = b_in
!         shrink = .5e0_rt;              if (present(shrink_in)) shrink = shrink_in
!         maxiter = 1000;             if (present(maxiter_in)) maxiter = maxiter_in
!         xtoler = 1d-6;              if (present(xtoler_in)) xtoler = xtoler_in
!         printmod = 1;               if (present(printmod_in)) printmod = printmod_in
! 
!         ! Check inputs and try to recover when appropriate
!         if (any(e<0e0_rt)) then
!             print*,'WARNING: e has some negative, replacing'
!             e = abs(e)
!         end if
!         if (any(e==0e0_rt)) STOP 'SS: ERROR: e is equal to 0'
! 
!         ! Init
!         fevals = 0
! 
!         if (printmod>0) then
!             write(*,'(A)') 'SS:__________________________________________________________________'            
!             write(*,'(A)') 'SS:Beginning Smolyak Projection Search Method       '            
!             write(*,'(A)') 'SS:    iter  fevals    avg.fcoll        fbest        enorm        snorm'
!         end if 
! 
!         xoldbest = x
!         foldbest = f(x) ! inefficient
!         fevals = fevals + 1
! 
!         ! Do the smolyak search
!         OUTER: do iter = 1,maxiter
! 
!             ! Exit if xtoler exceeded 
!             enorm = sqrt(sum(e**2))
!             if (enorm < xtoler) exit 
! 
!             ! Inner loop: updating the hypercube center successively 
!             do inner = 1,20
! 
!                 ! Get the smolyak grid
!                 hcubelb = max(xoldbest-e,a)
!                 hcubeub = min(xoldbest+e,b)
!                 call smolyakapprox_step1(coll,sApprox,n,mu,hcubelb,hcubeub)
! 
!                 ! Evaluate the function at each collocation point
!                 if (allocated(fcoll)) deallocate(fcoll)
!                 allocate(fcoll(size(coll,1)))
!                 do i = 1,size(fcoll)
!                     fcoll(i) = f(coll(i,:))
!                     fevals = fevals + 1
!                 end do
! 
!                 ! Construct the smolyak approximation
!                 call smolyakapprox_step2(fcoll,sApprox)
! 
!                 ! Minimize the approximation within the hypercube 
!                 !!! Start from every collocation point and find the best of the approx
!                 do i = 1,size(fcoll)
!                     x = coll(i,:)                 
!                     call sub_simplex(x,fapprox,xtoler_in=1d-11,printmod_in=-1)
!                     if (i==1 .or. fapprox(x)<fbest) then
!                         xbest = x
!                         fbest = fapprox(x)
!                     end if
!                 end do
! 
!                 ! This is inefficient to do, but for info sake and ease, do it here 
!                 ! (could save this value since will be a collocation point next time
!                 fbest = f(xbest)
!                 fevals = fevals + 1
! 
!                 ! To ensure we always descend according to the *real* function values rather
!                 ! than the *approximate* function values, we can do this: 
!                 ! If this new "xbest" is worse than one of the collocation points we checked,
!                 ! replace it. 
!                 ! Note that with this, smolyak is at worst a grid search in several important
!                 ! dimensions.
!                 tmpi1 = minloc(fcoll)
!                 if (fcoll(tmpi1(1))<fbest) then
!                     xbest = coll(tmpi1(1),:) 
!                     fbest = fcoll(tmpi1(1))
!                 end if
! 
!                 snorm = sqrt(sum((xbest-xoldbest)**2))
! 
!                 xoldbest = xbest
!                 foldbest = fbest
! 
!                 ! Update status 
!                 if (printmod>0 .and. (iter==1 .or. mod(iter,printmod)==0)) &
!                     write(*,'(A,i5,i8,4f13.7)') 'SS:   ',iter, fevals, mean(fcoll), fbest, enorm, snorm
! 
!                 ! Exit inner loop if best x has not changed much
!                 ! if (xnorm<maxval(abs(xbest))*(1e0_rt + 1d-3)) exit
!                 if (snorm<1d2*xtoler) exit
! 
!             end do 
!                 
!             ! Shrink the hypercube dimensions
!             e = shrink*e
! 
!         end do OUTER
! 
!         if (printmod>0) then
!             write(*,'(A)',advance='no') 'SS: Final x:'
!             write(*,'(99d16.6)',advance='no') x
!             write(*,*)
!             write(*,'(A)',advance='no') 'SS: Final e:'
!             write(*,'(99d16.6)',advance='no') e
!             write(*,*)
!             write(*,'(A)') 'SS:__________________________________________________________________'            
!         end if
! 
!     contains 
!         
!         function fapprox(x)
!             implicit none
!             real(rt), intent(in) :: x(1:)
!             real(rt) :: fapprox
!             ! local
!             real(rt) :: fvec(1),xmat(1,size(x))
! 
!             if (any(x<hcubelb) .or. any(x>hcubeub)) then
!                 fapprox = 1d100
!                 return
!             end if
! 
!             xmat(1,:) = x
! 
!             ! Evaluate the smolyak approximation at x
!             call smolyakapprox_step3(fvec,sApprox,xmat)
!             fapprox = fvec(1)
! 
!         end function
!     
!     end subroutine sub_smolyaksearch


!     ! Compass search
!     ! 
!     ! This is a comparison method that searches in each coordinate direction successfully,
!     ! shrinking the step size only when no improvement is found. It can be awfully slow,
!     ! but it is also reliable --- more so than the simplex algorithm. In my testing on the 
!     ! rosenbrock function, SA got stuck far from the global optimium while CS was in the 
!     ! right area (but without a precise estimate).
!     ! 
!     ! The step sizes in each direction are initially given by "e". At each step, the new e
!     ! is e*shrink. The routine ceases when the L2 norm of e falls below xtoler or maxiter
!     ! is reached.
!     ! 
!     subroutine sub_compass(x,f,e_in,shrink_in,toler_in,xtoler_in,maxiter_in,printmod_in,usecache_in)
!         use mod_matlab
!         implicit none
!         real(rt), dimension(1:), intent(inout) :: x
!         procedure(minN_f_template) :: f
!         procedure(minN_f_template_w_cache), pointer :: g
!         type(minN_f_cache) :: cache
!         real(rt), dimension(1:size(x)), intent(in), optional :: e_in
!         real(rt), intent(in), optional :: toler_in, xtoler_in
!         integer, intent(in), optional :: maxiter_in,printmod_in
!         real(rt), intent(in), optional :: shrink_in
!         logical, intent(in), optional :: usecache_in
!         !local
!         real(rt) :: toler, xtoler
!         integer :: maxiter,printmod
!         real(rt), dimension(1:size(x)) :: e,xbest,xtrial
!         real(rt) :: shrink,fbest,ftrial,avgftrial,enorm, fpos, fneg
!         integer :: n,i,iter,nimprove,ntrial,fevals
!         logical :: usecache
!         
!         n = size(x)        
! 
!         !Set defaults
!         maxiter = 1000
!         e = max(.1e0_rt*abs(x),.01e0_rt) !inital step size
!         shrink = .5e0_rt
!         xtoler = 1d-12
!         toler = -huge(0.e0_rt) !stop iterations if f(x)<toler for some x (note that f may be negative)
!         printmod = 1
!         usecache = .true.
! 
!         !Overwrite if present
!         if (present(e_in)) e = e_in
!         if (present(shrink_in)) shrink = shrink_in
!         if (present(xtoler_in)) xtoler = xtoler_in
!         if (present(toler_in)) toler = toler_in
!         if (present(maxiter_in)) maxiter = maxiter_in
!         if (present(printmod_in)) printmod = printmod_in
!         if (present(usecache_in)) usecache = usecache_in
! 
!         ! Point g to either use f with a cache or without one
!         if (usecache) then
!             ! Create a cache
!             call init_cache(cache,size(x),1 + 2*n*maxiter)
!             g => f_with_cache
!         else
!             ! Proceed with a cache
!             g => f_without_cache 
!         end if
! 
!         ! Init
!         fevals = 0
! 
!         if (printmod>0) then
!             write(*,'(A)') 'CS:__________________________________________________________________'            
!             write(*,'(A)') 'CS:Beginning Compass Search Method       '            
!             write(*,'(A)') 'CS:    iter  fevals #better   avg.ftrial        fbest        xsize'
!         end if 
! 
!         ! Compute the function value at the initial point
!         xbest = x
!         fbest = g(x,f,cache)
! 
!         ! Do the compass search
!         do iter = 1,maxiter
! 
!             ! Search in each direction
!             nimprove = 0
!             avgftrial = 0e0_rt
!             ntrial = 0
!             do i = 1,n
! 
!                 ! Try positive direction
!                 xtrial = xbest
!                 xtrial(i) = xtrial(i) + e(i)
!                 fpos = g(xtrial,f,cache)
!                 avgftrial = avgftrial + fpos
!                 ntrial = ntrial + 1
! 
!                 ! Try negative direction
!                 xtrial = xbest
!                 xtrial(i) = xtrial(i) - e(i)
!                 fneg = g(xtrial,f,cache)
!                 avgftrial = avgftrial + fneg
!                 ntrial = ntrial + 1
!                 
!                 ! Get the best of the two directions
!                 if (fpos<fneg) then
!                     xtrial = xbest
!                     xtrial(i) = xtrial(i) + e(i)
!                     ftrial = fpos
!                 else
!                     xtrial = xbest
!                     xtrial(i) = xtrial(i) - e(i)
!                     ftrial = fneg
!                 end if
! 
!                 ! Store if better 
!                 if (ftrial<fbest) then
!                     xbest = xtrial
!                     fbest = ftrial
!                     nimprove = nimprove+1
! !                 else
! !                     ! Shrink that direction
! !                     e(i) = shrink*e(i)
!                 end if    
! 
!             end do
! 
!             ! If didn't find a better, shrink
!             if (nimprove==0) then
!                 e = shrink*e
!             end if
! 
!             ! Compute L2 norm of compass size e
!             enorm = sqrt(sum(e**2))
! 
!             ! Update status 
!             if (printmod>0 .and. (iter==1 .or. mod(iter,printmod)==0 .or. enorm<xtoler)) &
!               write(*,'(A,i5,2i8,3f13.7)') 'CS:   ',iter, fevals, nimprove, avgftrial/real(ntrial,8), fbest, enorm
!            
!             ! Exit if xtoler exceeded 
!             if (enorm < xtoler) exit 
! 
!         end do
! 
!         x = xbest
! 
!         if (printmod>0) then
!             write(*,'(A)',advance='no') 'CS: Final x:'
!             write(*,'(99d16.6)',advance='no') x
!             write(*,*)
!             write(*,'(A)',advance='no') 'CS: Final e:'
!             write(*,'(99d16.6)',advance='no') e
!             write(*,*)
!             write(*,'(A)') 'CS:__________________________________________________________________'            
!         end if
! 
!         if (usecache) then
!             ! Free cache
!             call free_cache(cache)
!         end if
! 
!     end subroutine sub_compass


    ! sub_lineSearch
    ! 
    ! This is a utility routine that, given a function, initial point (x), and direction (s),
    ! minimizes the function along the direction.
    !
    subroutine sub_lineSearch(x,lambda,s,f,fx,gradx,lambda_init_in,toler_in,xtoler_in,maxiter_in,printmod_in)
        use mod_min1dim
        use mod_matlab
        implicit none
        real(rt), dimension(1:), intent(inout) :: x
        real(rt), dimension(1:), intent(in) :: s, gradx ! direction and gradient at initial point x
        real(rt), intent(in) :: fx ! f at initial point x
        real(rt), intent(out) :: lambda ! step size chosen
        interface 
            function f(x) 
                use mod_types
                implicit none
                real(rt), dimension(1:), intent(in) :: x
                real(rt) :: f
            end function f
        end interface
        real(rt), intent(in), optional :: toler_in, xtoler_in, lambda_init_in
        integer, intent(in), optional :: maxiter_in,printmod_in
        !local
        integer :: maxiter,printmod,iter,n
        real(rt) :: toler, xtoler, gp0, g0, flambda, glambda, bestf, bestlambda, lambda_init
        real(rt), parameter :: sig = 2e0_rt, eps = 1e-4_rt

        n = size(x)        

        ! Set defaults and overwrite if present
        maxiter = 40; if (present(maxiter_in)) maxiter = maxiter_in
        xtoler = 1e-6_rt;  if (present(xtoler_in)) xtoler = xtoler_in
        toler = 1e-6_rt;   if (present(toler_in)) toler = toler_in
        printmod = -1;   if (present(printmod_in)) printmod = printmod_in
        lambda_init = 1e0_rt;  if (present(lambda_init_in)) lambda_init = lambda_init_in 

        ! We can use an inexact line search, the Armijo rule, to try to get a good value quickly
        ! Get the total derivative of f at x. This is the derivative of g(lambda) at lambda = 0  
        g0 = fx
        gp0 = dot_product(gradx,s)

        ! Begin with lambda = 1e0_rt, and creep it down until find an acceptable
        lambda = lambda_init 
        bestf = huge(0e0_rt)
        bestlambda = 1e0_rt
        do iter = 1,maxiter
            flambda = f(x+lambda*s)
            glambda = g0 + lambda*eps*gp0
            if (flambda<=glambda) then
                x = x+lambda*s
                return
            else
                lambda = lambda/sig
            end if
            if (flambda<bestf) then
                bestlambda = lambda
            end if
        end do
        
        lambda = bestlambda
        x = x + lambda*s

    end subroutine sub_lineSearch

    ! NOTE: THIS HAS NOT BEEN TESTED MUCH
    ! Broyden-Fletcher-Goldfarb-Shanno (BFGS) method
    ! 
    ! This uses line search with a Hessian estimate.
    ! 
    subroutine sub_BFGS(x,f,scaling_in,toler_in,xtoler_in,maxiter_in,printmod_in,hess_out)
        use mod_deriv
        use mod_matlab
        implicit none
        real(rt), dimension(1:), intent(inout) :: x
        interface 
            function f(x) 
                use mod_types
                implicit none
                real(rt), dimension(1:), intent(in) :: x
                real(rt) :: f
            end function f
        end interface
        real(rt), intent(in), optional :: scaling_in(1:size(x)),toler_in, xtoler_in
        integer, intent(in), optional :: maxiter_in,printmod_in
        real(rt), intent(out), optional :: hess_out(size(x),size(x))
        !local
        integer :: maxiter,printmod,iter,n
        real(rt) :: toler, xtoler, f0, snorm, gnorm, lambda
        real(rt), allocatable :: x0(:),s(:), hess(:,:), scaling(:), grad(:), grae0_rt(:), y(:)
        
        n = size(x)        
        allocate(x0(n),s(n),grad(n),hess(n,n),grae0_rt(n),y(n),scaling(n))

        !Set defaults and overwrite if present
        maxiter = 1000; if (present(maxiter_in)) maxiter = maxiter_in
        xtoler = 1e-6_rt;  if (present(xtoler_in)) xtoler = xtoler_in
        toler = 1e-6_rt;   if (present(toler_in)) toler = toler_in
        printmod = 1;   if (present(printmod_in)) printmod = printmod_in
        scaling = 1e0_rt;  if (present(scaling_in)) scaling = scaling_in

        if (printmod>0) then
            write(*,'(A)') 'BFGS:__________________________________________________________'            
            write(*,'(A)') 'BFGS: Beginning Broyden-Fletcher-Goldfarb-Shanno Method        '          
            write(*,'(A)') 'BFGS: Computing initial function value and gradient            '         
        end if

        ! Set the initial hessian to do steepest descent
        hess = eye(N)

        ! Compute the initial gradient
        x0 = x
        call sub_numGradHess(x0,f,scaling,grae0_rt,f0=f0)
        gnorm = sum(grae0_rt)**2
        
        if (printmod>0) then
            write(*,'(A)') 'BFGS:  iter          f0       snorm       gnorm      lambda    '
        end if

        ! Do the gradient descent method
        do iter = 1,maxiter
         
            ! Compute the direction using the hessian estimate and the gradient
            s = -mldivide(hess,grae0_rt)

            ! Compute the step via line search and the new x
            x = x0
            call sub_lineSearch(x,lambda,s,f,f0,grae0_rt) ! x overwritten
            s = x-x0 ! Now s is the step (not the direction)
            snorm = sum(s**2)

            ! Display output
            if (printmod>0 .and. mod(iter,printmod)==0) &
                write(*,'(A,i6,5e12.5)') 'BFGS:',iter,f0,snorm,gnorm,lambda

            ! Check for convergence
            if (snorm<xtoler*(1e0_rt+sum(x0**2))) then
                if (printmod>0) then
                    if (gnorm<toler*(1e0_rt+abs(f0))) then
                        write(*,'(A)') 'BFGS: Converged to critical point: gradient is nearly 0'
                    else
                        write(*,'(A)') 'BFGS: Converged to non optimal point: gradient is not nearly 0, but step is small'
                    end if
                end if
                exit
            end if

            ! Update the gradient and hessian
            !!! Compute the new gradient
            call sub_numGradHess(x,f,scaling,grad,f0=f0)
            gnorm = sum(grad**2)
            !!! Update the hessian
            y = grad-grae0_rt
            if (abs(dot_product(y,s))>1d-10 .and. sum(y**2)/dot_product(y,s)<1d10) then
                hess = hess - matmul(vec2mat(matmul(hess,s)),transpose(vec2mat(matmul(s,hess))))/&
                              dot_product(s,matmul(hess,s)) + matmul(vec2mat(y),transpose(vec2mat(y)))/dot_product(y,s)
            else
                ! hess = hess
            end if

            !!! Update the gradient
            grae0_rt = grad

            ! Update
            x0 = x

            if (iter==maxiter) then
                write(*,'(A)') 'BFGS: Did not converge before max iter reached'
            end if

        end do

        if (present(hess_out)) hess_out = hess

        if (printmod>0) write(*,'(A)') 'BFGS:__________________________________________________________'

    end subroutine sub_BFGS


    ! NOTE: THIS HAS NOT BEEN TESTED MUCH
    ! Gradient descent (GD)
    ! 
    ! Gradient descent method employing a line search
    ! 
    subroutine sub_gradDescent(x,f,scaling_in,toler_in,xtoler_in,maxiter_in,printmod_in)
        use mod_deriv
        use mod_matlab
        implicit none
        real(rt), dimension(1:), intent(inout) :: x
        interface 
            function f(x) 
                use mod_types
                implicit none
                real(rt), dimension(1:), intent(in) :: x
                real(rt) :: f
            end function f
        end interface
        real(rt), intent(in), optional :: scaling_in(1:size(x)),toler_in, xtoler_in
        integer, intent(in), optional :: maxiter_in,printmod_in
        !local
        integer :: maxiter,printmod,iter,n
        real(rt) :: toler, xtoler, f0, snorm, gnorm, lambda
        real(rt), allocatable :: x0(:),s(:), hess(:,:), grad(:), scaling(:)
        
        n = size(x)        
        allocate(x0(n),s(n),grad(n),hess(n,n),scaling(n))

        !Set defaults and overwrite if present
        maxiter = 1000; if (present(maxiter_in)) maxiter = maxiter_in
        xtoler = 1e-6_rt;  if (present(xtoler_in)) xtoler = xtoler_in
        toler = 1e-6_rt;   if (present(toler_in)) toler = toler_in
        printmod = 1;   if (present(printmod_in)) printmod = printmod_in
        scaling = 1e0_rt;  if (present(scaling_in)) scaling = scaling_in

        if (printmod>0) then
            write(*,'(A)') 'GS:________________________________________________________'            
            write(*,'(A)') 'GS:Beginning Gradient Descent Method       '          
            write(*,'(A)') 'GS:iter      f0       snorm      gnorm     '
        end if

        ! Do the gradient descent method
        do iter = 1,maxiter
         
            ! Compute the derivative information 
            call sub_numGradHess(x,f,scaling,grad,f0=f0)
            gnorm = sum(grad**2)

! print*,'x is ',x

            ! Compute the step via line search
            x0 = x
            grad = -grad
            call sub_lineSearch(x0,lambda,grad,f,f0,grad) ! x0 overwritten
            grad = -grad
            s = x-x0
            snorm = sum(s**2)

            ! Display output
            if (printmod>0 .and. mod(iter,printmod)==0) &
                write(*,'(A,i6,4e12.5)') 'GS:',iter,f0,snorm,gnorm

            ! Check for convergence
            if (snorm<xtoler*(1e0_rt+sum(x**2))) then
                if (gnorm<toler*(1e0_rt+abs(f0))) then
                    print*,'GS: Converged to critical point: gradient is nearly 0'
                else
                    print*,'GS: Converged to non optimal point: gradient is not nearly 0, but step is small'
                end if
                exit
            end if

            ! Update
            x = x0

            if (iter==maxiter) then
                print*,'GS: Did not converge before max iter reached'
            end if

        end do

        if (printmod>0) write(*,'(A)') 'GS:________________________________________________________'

    end subroutine sub_gradDescent


    ! Newton's method in N dimensions (NN)
    ! 
    ! Pure newton's method, calculating the hessian, without linesearch.
    ! 
    subroutine sub_newtonNdim(x,f,scaling_in,toler_in,xtoler_in,maxiter_in,printmod_in)
        use mod_matlab
        use mod_deriv
        implicit none
        real(rt), dimension(1:), intent(inout) :: x
        interface 
            function f(x) 
                use mod_types
                implicit none
                real(rt), dimension(1:), intent(in) :: x
                real(rt) :: f
            end function f
        end interface
        real(rt), intent(in), optional :: scaling_in(1:size(x)),toler_in, xtoler_in
        integer, intent(in), optional :: maxiter_in,printmod_in
        !local
        integer :: maxiter,printmod,iter,n
        real(rt) :: toler, xtoler, f0, snorm, gnorm
        real(rt), allocatable :: s(:), hess(:,:), grad(:), scaling(:)
        
        n = size(x)        
        allocate(s(n),grad(n),hess(n,n),scaling(n))

        !Set defaults and overwrite if present
        maxiter = 1000; if (present(maxiter_in)) maxiter = maxiter_in
        xtoler = 1e-6_rt;  if (present(xtoler_in)) xtoler = xtoler_in
        toler = 1e-6_rt;   if (present(toler_in)) toler = toler_in
        printmod = 1;   if (present(printmod_in)) printmod = printmod_in
        scaling = 1e0_rt;  if (present(scaling_in)) scaling = scaling_in

        if (printmod>0) then
            write(*,'(A)') 'NN:________________________________________________________'            
            write(*,'(A)') 'NN:Beginning Newton''s Method                     '          
            write(*,'(A)') 'NN:  iter          f0       snorm       gnorm     '
        end if

        ! Do Newton's method
        do iter = 1,maxiter
         
            ! Compute the derivative information 
            call sub_numGradHess(x,f,scaling,grad,hess,f0)
            gnorm = sum(grad**2)

            ! Compute the step
            s = -mldivide(hess,grad)
            snorm = sum(s**2)

            ! Display output
            if (printmod>0 .and. mod(iter,printmod)==0) &
                write(*,'(A,i6,4e12.5)') 'NN:',iter,f0,snorm,gnorm

            ! Check for convergence
            if (snorm<xtoler*(1e0_rt+sum(x**2))) then
                if (gnorm<toler*(1e0_rt+abs(f0))) then
                    if (sum(sum(hess,1))<0e0_rt) then
                        print*,'NN: Converged to min'
                    else
                        print*,'NN: Converged to max'
                    end if
                else
                    print*,'NN: Converged to non optimal point: gradient is not nearly 0, but step is small'
                end if
                exit
            end if

            ! Update
            x = x + s

            if (iter==maxiter) then
                print*,'NN: Did not converge before max iter reached'
            end if

        end do

        if (printmod>0) write(*,'(A)') 'NN:________________________________________________________'

    end subroutine sub_newtonNdim


    !Nelder-Mead algorithm
    ! 
    ! I should change that so it just treats anywhere outside of [a,b] as infinite.
    !
    ! This code essentially follows the algorithm in wikipedia
    ! However, I try a small modification where I accept the reflection point if it is better
    ! than the worst point, not the second worse point. This does not affect the volume of the 
    ! simplex, but gets around one pathology I encountered with a non-continuous function.
    ! 
    ! There are two routines here: 
    !   sub_simplex: full featured but slow. Ideal for flexibility, debugging, and calibration. 
    !   sub_simplex_core: not as many options and info, but faster (and so suitable for core routines)
    ! I recently added a new option: "randomize". If this is selected, the initial simplex is 
    ! drawn randomly with x_j = x + randn*e so that is e is the standard deviation of the shock
    !
    subroutine sub_simplex_core(x,f,e,xtoler,maxiter,printmod)
        use mod_sort
        implicit none
        real(rt), dimension(1:), intent(inout) :: x
        interface 
            function f(x) 
                use mod_types
                implicit none
                real(rt), dimension(1:), intent(in) :: x
                real(rt) :: f
            end function f
        end interface
        real(rt), dimension(1:size(x)), intent(in) :: e
        real(rt), intent(in) :: xtoler
        integer, intent(in) :: maxiter,printmod
        !local
        !controls
        real(rt), parameter :: alpha = 1.e0_rt, beta = 1.e0_rt, gamma = .5e0_rt, eta = .5e0_rt
        !tools
        real(rt), dimension(1:size(x)+1,1:size(x)) :: x_simplex
        real(rt), dimension(1:size(x)+1) :: f_simplex
        real(rt), dimension(1:size(x)) :: x_gravity,x_reflect,x_expand,x_contract
        real(rt) :: f_reflect,f_expand,f_contract
        integer, dimension(1:size(x)+1) :: ranking
        real(rt) :: dist
        integer :: n, i, iter
        logical :: doNotTryReflect

        n = size(x)        

        !Create initial simplex and evaluate at simplex corners
        x_simplex(1,:) = x 
        do i = 2,n+1
            x_simplex(i,:) = x
            x_simplex(i,i-1) = x(i-1) + e(i-1)            
        end do

        !Evaluate simplex at corners
        do i = 1,n+1
            f_simplex(i) = f(x_simplex(i,:))
        end do

        if (printmod>0) write(*,'(A)') 'NM:iter     min(f)        max(f)        simplex size'

        !Main routine
        doNotTryReflect = .false.
        do iter = 1,maxiter

            ! Reorder the simplex from smallest to largest (ascending sort)
            !!! Use insertion sort because almost surely this will be a very small list
            call isort(f_simplex,ranking) 
            x_simplex = x_simplex(ranking,:)
 
            ! Output
            if (printmod>0 .AND. mod(iter,printmod)==0) write(*,'(A,i6,A,e12.5,A,e12.5,A,e12.5)') &
                'NM:',iter,'  ', f_simplex(1),'  ', f_simplex(n+1),'  ', dist

            !Check whether length of simplex sides from best point are less than xtoler
            dist = 0.e0_rt
            do i = 2,n+1
                dist = max(dist,sqrt(sum((x_simplex(i,:)-x_simplex(1,:))**2)))                
            end do

            ! Check for convergence or end of loop
            if (dist<xtoler .or. iter==maxiter) then
                if (printmod>0) write(*,'(A,i6,A,e12.5,A,e12.5,A,e12.5)') &
                    'NM:',iter,'  ', f_simplex(1),'  ', f_simplex(n+1),'  ', dist
                ! Finished
                x = x_simplex(1,:)
                return
            end if

            !Try to replace the worst point, n+1, by another better point

            !Calc center of gravity of n other pts
            x_gravity = sum(x_simplex(1:n,:),1)/real(n,rt) 

            !Reflect xr = xg + alpha*(xg-x(n+1)) and calculate value
            x_reflect = x_gravity + alpha*(x_gravity-x_simplex(n+1,:))
            if (.not. doNotTryReflect) then
                f_reflect = f(x_reflect)
            else
                f_reflect = huge(0e0_rt)
                doNotTryReflect = .false.
            end if
            
            !Determine what to do with reflected point
            !if (f_simplex(1)<f_reflect .and. f_reflect<f_simplex(n)) then ! The old algorithm had this
            if (f_simplex(1)<f_reflect .and. f_reflect<f_simplex(n+1)) then 
                ! NOTE: next time, reflect will not satisfy this criterion---could optimize by avoiding it.
                ! If this is true, then next time reflect is tried, it will be the old point.
                ! We already know this is worse -- so don't waste an feval!
                if (f_reflect>=f_simplex(n)) then
                    doNotTryReflect = .true.
                end if 

                !Reflect: If f(x1)>f(xr)>f(xn) then replace and continue
                x_simplex(n+1,:) = x_reflect
                f_simplex(n+1) = f_reflect

            elseif (f_reflect>=f_simplex(n)) then

                ! If f_reflect is worst point, generate a new point by contraction
                x_contract = x_gravity + gamma*(x_simplex(n+1,:)-x_gravity)
                f_contract = f(x_contract)

                if (f_contract<f_simplex(n+1)) then

                    ! Contract: found a better point, so use it
                    x_simplex(n+1,:) = x_contract
                    f_simplex(n+1) = f_contract
                
                else
                
                    ! Contract all: no better point found, contract the entire simplex toward the best point
                    do i = 2,n+1                        
                        x_simplex(i,:) = x_simplex(1,:) + eta*(x_simplex(i,:)-x_simplex(1,:))
                        f_simplex(i) = f(x_simplex(i,:))
                    end do

                end if 

            elseif (f_reflect<=f_simplex(1)) then
                
                ! If f_reflect is best point, then try expanding in this direction for further improve
                x_expand = x_reflect + beta*(x_reflect-x_gravity)
                f_expand = f(x_expand)

                if (f_expand<f_reflect) then
                    ! Expand
                    x_simplex(n+1,:) = x_expand
                    f_simplex(n+1) = f_expand
                else
                    ! Reflect
                    x_simplex(n+1,:) = x_reflect
                    f_simplex(n+1) = f_reflect
                end if
            else 
                !Case not found
                STOP 'simplex: case not found'
            end if

        end do

    end subroutine sub_simplex_core
    subroutine sub_simplex(x,f,e_in,a_in,b_in,toler_in,xtoler_in,maxiter_in,printmod_in,randomize_in,contract_in,uid_in)
        use mod_util
        use mod_matlab
        implicit none
        real(rt), dimension(1:), intent(inout) :: x
        interface 
            function f(x) 
                use mod_types
                implicit none
                real(rt), dimension(1:), intent(in) :: x
                real(rt) :: f
            end function f
        end interface
        real(rt), dimension(:), intent(in), optional :: e_in,a_in,b_in
        real(rt), intent(in), optional :: toler_in, xtoler_in, contract_in
        integer, intent(in), optional :: maxiter_in,printmod_in,uid_in
        logical, intent(in), optional :: randomize_in
        !local
        !controls
        real(rt) :: toler, xtoler
        integer :: maxiter,printmod
        real(rt), dimension(1:size(x)) :: e
        real(rt), parameter :: alpha = 1.e0_rt, beta = 1.e0_rt
        real(rt) :: gamma, eta
        
        !tools
        real(rt), dimension(1:size(x)+1,1:size(x)) :: x_simplex
        real(rt), dimension(1:size(x)+1) :: f_simplex, tmp
        real(rt), dimension(1:size(x)) :: x_gravity,x_reflect,x_expand,x_contract
        real(rt) :: f_reflect,f_expand,f_contract
        integer, dimension(1:size(x)+1) :: ranking
        real(rt) :: dist
        integer :: n, i, iter, fevals, uid, flag
        character(len=15) :: action
        logical :: is_constrained, doNotTryReflect, randomize

        n = size(x)        

        ! Check the sizes conform
        if (present(e_in)) then 
            if (size(e_in)/=n) STOP 'ERROR: sub_simplex: e_in does not conform to x'
        end if
        if (present(a_in)) then 
            if (size(a_in)/=n) STOP 'ERROR: sub_simplex: a_in does not conform to x'
        end if
        if (present(b_in)) then 
            if (size(b_in)/=n) STOP 'ERROR: sub_simplex: b_in does not conform to x'
        end if

        !Set defaults
        maxiter = 1000
        e = max(.1e0_rt*abs(x),.0001e0_rt) !inital step size
        xtoler = 1e-12_rt
        toler = -huge(0.e0_rt) !stop iterations if f(x)<toler for some x (note that f may be negative)
        printmod = 1
        randomize = .false.
        gamma = .5e0_rt ! contraction param
        eta = .5e0_rt ! contraction param      
        uid = 6 ! stdout

        !Overwrite if present
        if (present(e_in)) e = e_in
        if (present(xtoler_in)) xtoler = xtoler_in
        if (present(toler_in)) toler = toler_in
        if (present(maxiter_in)) maxiter = maxiter_in
        if (present(printmod_in)) printmod = printmod_in
        if (present(randomize_in)) randomize = randomize_in
        if (present(uid_in)) uid = uid_in
        if (present(contract_in)) then
            gamma = (1e0_rt-contract_in)
            eta = (1e0_rt-contract_in)
            if (contract_in<=0e0_rt .or. contract_in>=1e0_rt) then
                STOP 'ERROR: sub_simplex: contract_in invalid'
            end if
        end if

        ! Check that x is inside a,b if present
        if (present(a_in) .and. present(b_in)) then
            call check_guess_in_range(flag,x,a_in,b_in)
            if (flag/=0) write(0,*) 'WARNING: x was not in range, so bounding and proceeding'
            x = min(max(a_in,x),b_in)
        end if

        if (printmod==1) then
            write(uid,'(A)') 'NM:'//repeat('_',120) 
            write(uid,'(A)') 'NM:Beginning Simplex Method (Nelder-Mead)'            
        end if

        is_constrained = present(a_in) .and. present(b_in)

        !Create initial simplex and evaluate at simplex corners
        x_simplex(1,:) = x 
        if (.not. randomize) then
            do i = 2,n+1
                x_simplex(i,:) = x
                x_simplex(i,i-1) = x(i-1) + e(i-1)            
            end do
        else
            do i = 2,n+1
                x_simplex(i,:) = x + randn(n)*e
            end do
        end if

        !Evaluate simplex at corners
        do i = 1,n+1
            if (.not. is_constrained) then
                f_simplex(i) = f(x_simplex(i,:))
            else
                if (any(x_simplex(i,:)<a_in .or. x_simplex(i,:)>b_in)) then
                    f_simplex(i) = huge(0e0_rt)
                else
                    f_simplex(i) = f(x_simplex(i,:))
                end if
            end if
        end do
        fevals = n+1
    
        if (printmod==1) then
            write(uid,'(A)') 'NM: Initial simplex: '
            do i = 1,n+1
                write(uid,'(A,i4,A,A)') 'NM: i ',i,' x ',printx(x_simplex(i,:))
            end do
            write(uid,'(A)') 'NM: Initial simplex values: '
            write(uid,'(A,A)') 'NM: ',printx(f_simplex)
        end if

        dist = -1e0_rt

        !Main routine
!         if (printmod>0) write(uid,'(A)') 'NM:iter     min(f)        max(f)        simplex size  fevals   action              bestx'
        action = ' Init'
        doNotTryReflect = .false.
        do iter = 1,maxiter

            !Reorder the simplex from smallest to largest (ascending sort)
            call sort2(tmp,ranking,f_simplex)
            f_simplex = f_simplex(ranking)
            x_simplex = x_simplex(ranking,:)
            
! print*,'f_simplex',f_simplex
! print*,'x_simplex'
! call disp(x_simplex)

            if (printmod>0 .AND. mod(iter,printmod)==0) then
                write(uid,'(A,i6,A,e12.5,A,e12.5,A,e12.5,2A)') &
                'NM:',iter,' best ', f_simplex(1),' worst ', f_simplex(n+1),' s.size ', dist, action, printx(x_simplex(1,:))
            end if

            !Check whether length of simplex sides from best point are less than xtoler
            dist = 0.e0_rt
            do i = 2,n+1
                dist = max(dist,sqrt(sum((x_simplex(i,:)-x_simplex(1,:))**2)))                
            end do
            if (dist<xtoler) then
                if (printmod>0) &
                    write(uid,'(A,i6,A,e12.5,A,e12.5,A,e12.5,2A)') &
                    'NM:',iter,' best ', minval(f_simplex),' worst ', maxval(f_simplex),&
                    ' s.size ', dist, action, printx(x_simplex(1,:))
                if (printmod==1) write(uid,'(A)') 'NM:Simplex finished: simplex size less than xtoler'
                exit
            end if
            if (f_simplex(1)<toler) then
                if (printmod>0) &
                    write(uid,'(A,i6,A,e12.5,A,e12.5,A,e12.5,2A)') &
                    'NM:',iter,' best ', minval(f_simplex),' worst ', maxval(f_simplex),&
                    ' s.size ', dist, action, printx(x_simplex(1,:))
                if (printmod==1) write(uid,'(A)') 'NM:Simplex finished: f is less than toler'
                exit
            end if            

            !Try to replace the worst pt, N+1, by another better point

            !Calc center of gravity of n other pts
            x_gravity = sum(x_simplex(1:n,:),1)/real(n,rt) 

            !Reflect xr = xg + alpha*(xg-x(n+1)) and calculate value
            x_reflect = x_gravity + alpha*(x_gravity-x_simplex(n+1,:))
! print*,'x_reflect', x_reflect            
            if (.not. doNotTryReflect) then
                if (.not. is_constrained) then
                    f_reflect = f(x_reflect)
                else
                    if (any(x_reflect<a_in .or. x_reflect>b_in)) then
                        f_reflect = huge(0e0_rt)
                    else
                        f_reflect = f(x_reflect)
                    end if
                end if
! print*,'f_reflect', f_reflect            
                fevals = fevals+1
            else
                f_reflect = huge(0e0_rt)
                doNotTryReflect = .false.
            end if
            
            !Determine what to do with reflected point
            !if (f_simplex(1)<f_reflect .and. f_reflect<f_simplex(n)) then ! The old algorithm had this
            if (f_simplex(1)<f_reflect .and. f_reflect<f_simplex(n+1)) then 
                ! NOTE: next time, reflect will not satisfy this criterion---could optimize by avoiding it.
                ! If this is true, then next time reflect is tried, it will be the old point.
                ! We already know this is worse -- so don't waste an feval!
                if (f_reflect>=f_simplex(n)) then
                    doNotTryReflect = .true.
                end if 
                action = ' Reflected'
                !If f(x1)>f(xr)>f(xn) then replace and continue
                x_simplex(n+1,:) = x_reflect
                f_simplex(n+1) = f_reflect

            elseif (f_reflect>=f_simplex(n)) then
                !If f_reflect is worst point, generate a new point by contraction
                x_contract = x_gravity + gamma*(x_simplex(n+1,:)-x_gravity)
                if (.NOT. is_constrained) then
                    f_contract = f(x_contract)
                else
                    if (any(x_contract<a_in .or. x_contract>b_in)) then
                        f_contract = huge(0e0_rt)
                    else
                        f_contract = f(x_contract)
                    end if
                end if
                fevals = fevals+1
                if (f_contract<f_simplex(n+1)) then
                    action = ' Contracted'
                    !If found a better point, then use it
                    x_simplex(n+1,:) = x_contract
                    f_simplex(n+1) = f_contract
                else
                    !If no better point found, contract the entire simplex toward the best point
                    action = ' Contracted All'
                    do i = 2,n+1                        
                        x_simplex(i,:) = x_simplex(1,:) + eta*(x_simplex(i,:)-x_simplex(1,:))
                        !f_simplex(i) = f(x_simplex(i,:))
                        if (.not. is_constrained) then
                            f_simplex(i) = f(x_simplex(i,:))
                        else
                            if (any(x_simplex(i,:)<a_in .or. x_simplex(i,:)>b_in)) then
                                f_simplex(i) = huge(0e0_rt)
                            else
                                f_simplex(i) = f(x_simplex(i,:))
                            end if
                        end if
                        fevals = fevals+1
                    end do
                end if 
            elseif (f_reflect<=f_simplex(1)) then
                
                !If f_reflect is best point, then try expanding in this direction for further improve
                x_expand = x_reflect + beta*(x_reflect-x_gravity)
                !f_expand = f(x_expand)
                if (.not. is_constrained) then
                    f_expand = f(x_expand)
                else
                    if (any(x_expand<a_in .or. x_expand>b_in)) then
                        f_expand = huge(0e0_rt)
                    else
                        f_expand = f(x_expand)
                    end if
                end if
                fevals = fevals+1
                if (f_expand<f_reflect) then
                    action = ' Expanded'
                    x_simplex(n+1,:) = x_expand
                    f_simplex(n+1) = f_expand
                else
                    action = ' Reflected'
                    x_simplex(n+1,:) = x_reflect
                    f_simplex(n+1) = f_reflect
                end if
            else 
                !Case not found
                print*,'f_reflect:',f_reflect
                print*,'f_simplex:',f_simplex
                STOP 'simplex: case not found'
            end if

        end do
 
        !Set out
        x = x_simplex(minloc(f_simplex,1),:)

        if (printmod==1) then
            write(uid,'(A)') 'NM: Final simplex: '
            do i = 1,n+1
                write(uid,'(A,i4,A,99(e12.5))') 'NM: i ',i,' x ',x_simplex(i,:)
            end do
            write(uid,'(A)') 'NM: Final simplex values: '
            write(uid,'(A,99(e12.5))') 'NM: ',f_simplex
            write(uid,'(A)') 'NM:Final minimum x: '
            write(uid,'(A,99(e12.5))') 'NM: ',x
            write(uid,'(A)') 'NM:Final minimum f: '
            write(uid,'(A,e12.5)') 'NM:',minval(f_simplex)
            write(uid,'(A)') 'NM:'//repeat('_',120)
        end if

    end subroutine sub_simplex

    ! OpenMP version of simplex
!     subroutine sub_parallelsimplex_OMP(x,f,e_in,toler_in,xtoler_in,maxiter_in,printmod_in,randomize_in,contract_in)
!         use mod_matlab
!         use mod_util
!         implicit none
!         real(rt), dimension(1:), intent(inout) :: x
!         interface 
!             function f(x) 
!                 implicit none
!                 real(rt), dimension(1:), intent(in) :: x
!                 real(rt) :: f
!             end function f
!         end interface
!         real(rt), dimension(:), intent(in), optional :: e_in
!         real(rt), intent(in), optional :: toler_in, xtoler_in, contract_in
!         integer, intent(in), optional :: maxiter_in,printmod_in
!         logical, intent(in), optional :: randomize_in
!         !local
!         !controls
!         real(rt) :: toler, xtoler
!         integer :: maxiter,printmod
!         real(rt), dimension(1:size(x)) :: e
!         real(rt), parameter :: alpha = 1.e0_rt, beta = 1.e0_rt
!         real(rt) :: gamma, eta
!         !tools
!         real(rt), dimension(:,:), allocatable :: x_simplex
!         real(rt), dimension(:), allocatable :: f_simplex, tmp
!         integer, dimension(:), allocatable :: ranking
!         logical, dimension(:), allocatable :: didImprove
!         real(rt), dimension(:), allocatable :: x_gravity,x_reflect,x_expand,x_contract
!         real(rt) :: f_reflect,f_expand,f_contract
!         real(rt) :: dist
!         integer :: n, i, iter, fevals, ip, np, ilo, ihi
!         logical :: randomize
!         
! 
!         n = size(x)        
! 
!         allocate(x_simplex(n+1,n))
!         allocate(f_simplex(n+1))
!         allocate(tmp(n+1))
!         allocate(ranking(n+1))
!         allocate(didImprove(n+1))
!         allocate(x_gravity(n))
!         allocate(x_reflect(n))
!         allocate(x_expand(n))
!         allocate(x_contract(n))
! 
!         ! Get number of threads
!         np = 0
!         !$OMP PARALLEL REDUCTION(+:np)
!         np = np + 1
!         !$OMP END PARALLEL 
!         np = min(n,np)
! 
!         !Set defaults
!         maxiter = 1000
!         e = max(.1e0_rt*abs(x),.0001e0_rt) !inital step size
!         xtoler = 1d-12
!         toler = -huge(0.e0_rt) !stop iterations if f(x)<toler for some x (note that f may be negative)
!         printmod = 1
!         randomize = .false.
!         gamma = .5e0_rt ! contraction param
!         eta = .5e0_rt ! contraction param      
! 
!         !Overwrite if present
!         if (present(e_in)) e = e_in
!         if (present(xtoler_in)) xtoler = xtoler_in
!         if (present(toler_in)) toler = toler_in
!         if (present(maxiter_in)) maxiter = maxiter_in
!         if (present(printmod_in)) printmod = printmod_in
!         if (present(randomize_in)) randomize = randomize_in
!         if (present(contract_in)) then
!             gamma = (1e0_rt-contract_in)
!             eta = (1e0_rt-contract_in)
!             if (contract_in<=0e0_rt .or. contract_in>=1e0_rt) then
!                 STOP 'ERROR: sub_simplex: contract_in invalid'
!             end if
!         end if
! 
!         if (randomize) STOP 'ERROR: sub_parallelsimplex: not sure randomization is okay across multiple processes'
! 
!         printmod = merge(printmod,-1,mympi_iproc()==0) 
! 
!         !Create initial simplex and evaluate at simplex corners
!         x_simplex(1,:) = x 
!         if (.not. randomize) then
!             do i = 2,n+1
!                 x_simplex(i,:) = x
!                 x_simplex(i,i-1) = x(i-1) + e(i-1)            
!             end do
!         else
!             do i = 2,n+1
!                 x_simplex(i,:) = x + randn(n)*e
!             end do
!         end if
! 
!         !Evaluate simplex at corners
!         !$OMP PARALLEL DO DEFAULT(SHARED)
!         do i = 1,n+1
!             f_simplex(i) = f(x_simplex(i,:))
!         end do
!         !$OMP PARALLEL DO
! 
!         fevals = n+1
!     
!         if (printmod>0 .AND. printmod==1) then
!             write(*,'(A)') 'PNM:Initial simplex: '
!             do i = 1,n+1
!                 write(*,'(A,i4,A,99(e12.5))') 'PNM: i ',i,' x ',x_simplex(i,:)
!             end do
!             write(*,'(A)') 'PNM: Initial simplex values: '
!             write(*,'(A,99(e12.5))') 'PNM: ',f_simplex
!         end if
! 
!         dist = -1e0_rt
! 
!         !Main routine
!         if (printmod>0) write(*,'(A)') 'PNM:iter     min(f)        max(f)        simplex size  '            
!         do iter = 1,maxiter
! 
!             !Reorder the simplex from smallest to largest (ascending sort)
!             call sort2(tmp,ranking,f_simplex)
!             f_simplex = f_simplex(ranking); x_simplex = x_simplex(ranking,:)
! 
!             if (printmod>0 .AND. mod(iter,printmod)==0) write(*,'(A,i6,A,e12.5,A,e12.5,A,e12.5,2A)') 'PNM:',iter,'  ', f_simplex(1),'  ', f_simplex(n+1),'  ', dist,' xbest = ',printx(x_simplex(1,:))
! 
!             !Check whether length of simplex sides from best point are less than xtoler
!             dist = 0.e0_rt
!             do i = 2,n+1
!                 dist = max(dist,sqrt(sum((x_simplex(i,:)-x_simplex(1,:))**2)))                
!             end do
!             if (dist<xtoler) then
!                 if (printmod>0) &
!                     write(*,'(A,i6,A,e12.5,A,e12.5,A,e12.5)') 'PNM:',iter,'  ', minval(f_simplex),'  ', maxval(f_simplex),'  ', dist
!                 if (printmod>0) &
!                     write(*,'(A)') 'PNM:Simplex finished: simplex size less than xtoler'
!                 exit
!             end if
!             if (f_simplex(1)<toler) then
!                 if (printmod>0) &
!                     write(*,'(A,i6,A,e12.5,A,e12.5,A,e12.5)') 'PNM:',iter,'  ', minval(f_simplex),'  ', maxval(f_simplex),'  ', dist
!                 if (printmod>0) &
!                     write(*,'(A)') 'PNM:Simplex finished: f is less than toler'
!                 exit
!             end if            
! 
!             !Calc center of gravity of n other pts
!             x_gravity = sum(x_simplex(1:n+1-np,:),1)/real(n+1-np,8) 
! 
!             !!!!! PARALLEL ALGORITHM !!!!!!
!             ! Try to replace the worst P points, N-P,N-P+1,...,N+1, with other points.
!             !$OMP PARALLEL DO
!             do ip = n-np,n+1
! 
!                 !Reflect xr = xg + alpha*(xg-x(n+1)) and calculate value
!                 x_reflect(:,ip) = x_gravity + alpha*(x_gravity-x_simplex(ip,:))
!                 f_reflect(ip) = f(x_reflect)
! 
!             end do
!             !$OMP END PARALLEL DO
! 
!             ! For each of the reflected points, see if they improved and which ones to keep
!             ! Sort the new points in with the simplex, and see how many of the old ones are no longer in the simplex.
!             do ip = n-np,n+1
!                 if (f_simplex()
!             end do
!             call sort2()
!             call sort2(tmp,ranking,f_simplex)
! 
!             do ip = n-np,n+1
!                 
!                 !Determine what to do with reflected point
!                 if (f_reflect(ip)<=f_simplex(1)) then
!                     
!                     !If f_reflect is best point, then try expanding in this direction for further improve
!                     x_expand = x_reflect(:,ip) + beta*(x_reflect(:,ip)-x_gravity); f_expand = f(x_expand)
! 
!                     if (f_expand<f_reflect(ip)) then
!                         x_simplex(ip,:) = x_expand;  f_simplex(ip) = f_expand
!                     else
!                         x_simplex(ip,:) = x_reflect(:,ip); f_simplex(ip) = f_reflect(ip)
!                     end if
! 
!                     didImprove(ip) = .true.
! 
!                 elseif (f_simplex(1)<f_reflect(ip) .and. f_reflect(ip)<f_simplex(n+1-np)) then 
! 
!                     x_simplex(ip,:) = x_reflect; f_simplex(ip) = f_reflect(ip)
!                     didImprove(ip) = .true.
! 
!                 elseif (f_reflect(ip)>=f_simplex(n+1-np)) then
! 
!                     !If f_reflect is worst point, generate a new point by contraction
!                     if (f_reflect(ip)>=f_simplex(ip)) then
!                         x_reflect = x_simplex(ip,:); f_reflect(ip) = f_simplex(ip)
!                     end if
! 
!                     x_contract = x_gravity + gamma*(x_reflect-x_gravity)
!                     f_contract = f(x_contract)
! 
!                     if (f_contract<f_reflect(ip)) then
!                         !If found a better point, then use it
!                         x_simplex(ip,:) = x_contract; f_simplex(ip) = f_contract
!                         didImprove(ip) = .true.
!                     else
!                         ! These are not actually the reflection points, but the original point
!                         x_simplex(ip,:) = x_reflect; f_simplex(ip) = f_reflect(ip)
!                         didImprove(ip) = .false.
!                     end if 
!                 else 
!                     !Case not found
!                     print*,'f_reflect(ip):',f_reflect(ip)
!                     print*,'f_simplex:',f_simplex
!                     STOP 'simplex: case not found'
!                 end if
!             end do
! 
!             ! Now every process has updated a simplex point. Use all gather to collect them
!             ilo = n+1-np+1
!             ihi = n+1
!             call mympi_allgatherv(f_simplex(ilo:ihi))
!             x_simplex_transpose = transpose(x_simplex)
!             call mympi_allgatherv(x_simplex_transpose(:,ilo:ihi))
!             x_simplex = transpose(x_simplex_transpose)
!             call mympi_allgatherv(didImprove(ilo:ihi))
! 
!             if (.not. any(didImprove(ilo:ihi))) then
! 
!                 ! Contract all
!                 do i = 2,n+1
!                     x_simplex(i,:) = x_simplex(1,:) + eta*(x_simplex(i,:)-x_simplex(1,:))
!                 end do
! 
!                 ! Evaluate at the new simplex 
!                 do i = 1+mympi_a(n),1+mympi_b(n)
!                     f_simplex(i) = f(x_simplex(i,:))
!                 end do
!                 call mympi_allgatherv(f_simplex(2:n+1))
! 
!             end if
! 
!         end do
!  
!         !Set out
!         x = x_simplex(minloc(f_simplex,1),:)
! 
!         if (printmod>0 .AND. printmod==1) then
!             write(*,'(A)') 'PNM: Final simplex: '
!             do i = 1,n+1
!                 write(*,'(A,i4,A,99(e12.5))') 'PNM: i ',i,' x ',x_simplex(i,:)
!             end do
!             write(*,'(A)') 'PNM: Final simplex values: '
!             write(*,'(A,99(e12.5))') 'PNM: ',f_simplex
!             write(*,'(A)') 'PNM:Final minimum x: '
!             write(*,'(A,99(e12.5))') 'PNM: ',x
!             write(*,'(A)') 'PNM:Final minimum f: '
!             write(*,'(A,e12.5)') 'PNM:',minval(f_simplex)
!         end if
! 
!     end subroutine sub_parallelsimplex_OMP

    ! NOTE: currently this tries to replace the worst P points. However, this does not scale well to a large number of processors
    ! Instead, what could be done is a line search. When a reflection is called for, try reflecting at P different lengths.
    ! When a contraction is called for, try contracting at P different lengths. When a contract all is called for, do a round-robin 
    ! distributing N/P but checking multiple lengths.
    ! 
    subroutine sub_parallelsimplex(x,f,e_in,toler_in,xtoler_in,maxiter_in,printmod_in,randomize_in,contract_in)
        use mod_mympi
        use mod_matlab
        use mod_util
        implicit none
        real(rt), dimension(1:), intent(inout) :: x
        interface 
            function f(x) 
                use mod_types
                implicit none
                real(rt), dimension(1:), intent(in) :: x
                real(rt) :: f
            end function f
        end interface
        real(rt), dimension(1:size(x)), intent(in), optional :: e_in
        real(rt), intent(in), optional :: toler_in, xtoler_in, contract_in
        integer, intent(in), optional :: maxiter_in,printmod_in
        logical, intent(in), optional :: randomize_in
        !local
        !controls
        real(rt) :: toler, xtoler
        integer :: maxiter,printmod
        real(rt), dimension(1:size(x)) :: e
        real(rt), parameter :: alpha = 1.e0_rt, beta = 1.e0_rt
        real(rt) :: gamma, eta
        !tools
        real(rt), dimension(1:size(x)+1,1:size(x)) :: x_simplex
        real(rt), dimension(1:size(x)+1) :: f_simplex, tmp
        real(rt), dimension(1:size(x)) :: x_gravity,x_reflect,x_expand,x_contract
        real(rt) :: f_reflect,f_expand,f_contract
        integer, dimension(1:size(x)+1) :: ranking
        real(rt) :: dist
        integer :: n, i, iter, fevals, ip, np, ilo, ihi
        logical :: randomize
        logical, dimension(1:size(x)+1) :: didImprove
        real(rt), dimension(1:size(x),1:size(x)+1) :: x_simplex_transpose

        n = size(x)        

        ! Get number of nodes
        np = min(n,mympi_nproc())

        !Set defaults
        maxiter = 1000
        e = max(.1e0_rt*abs(x),.0001e0_rt) !inital step size
        xtoler = 1e-10_rt
        toler = -huge(0.e0_rt) !stop iterations if f(x)<toler for some x (note that f may be negative)
        printmod = 1
        randomize = .false.
        gamma = .5e0_rt ! contraction param
        eta = .5e0_rt ! contraction param      

        !Overwrite if present
        if (present(e_in)) e = e_in
        if (present(xtoler_in)) xtoler = xtoler_in
        if (present(toler_in)) toler = toler_in
        if (present(maxiter_in)) maxiter = maxiter_in
        if (present(printmod_in)) printmod = printmod_in
        if (present(randomize_in)) randomize = randomize_in
        if (present(contract_in)) then
            gamma = (1e0_rt-contract_in)
            eta = (1e0_rt-contract_in)
            if (contract_in<=0e0_rt .or. contract_in>=1e0_rt) then
                STOP 'ERROR: sub_simplex: contract_in invalid'
            end if
        end if

        if (randomize) STOP 'ERROR: sub_parallelsimplex: not sure randomization is okay across multiple processes'

        printmod = merge(printmod,-1,mympi_iproc()==0) 

        !Create initial simplex and evaluate at simplex corners
        x_simplex(1,:) = x 
        if (.not. randomize) then
            do i = 2,n+1
                x_simplex(i,:) = x
                x_simplex(i,i-1) = x(i-1) + e(i-1)            
            end do
        else
            do i = 2,n+1
                x_simplex(i,:) = x + randn(n)*e
            end do
        end if

        !Evaluate simplex at corners
        do i = mympi_a(n+1),mympi_b(n+1)
            f_simplex(i) = f(x_simplex(i,:))
        end do
        call mympi_allgatherv(f_simplex)

        fevals = n+1
    
        if (printmod>0 .AND. printmod==1) then
            write(*,'(A)') 'PNM:Initial simplex: '
            do i = 1,n+1
                write(*,'(A,i4,A,99(e12.5))') 'PNM: i ',i,' x ',x_simplex(i,:)
            end do
            write(*,'(A)') 'PNM: Initial simplex values: '
            write(*,'(A,99(e12.5))') 'PNM: ',f_simplex
        end if

        dist = -1e0_rt

        !Main routine
        if (printmod>0) write(*,'(A)') 'PNM:iter     min(f)        max(f)        simplex size  '            
        do iter = 1,maxiter

            !Reorder the simplex from smallest to largest (ascending sort)
            call sort2(tmp,ranking,f_simplex)
            f_simplex = f_simplex(ranking); x_simplex = x_simplex(ranking,:)

            if (printmod>0 .AND. mod(iter,printmod)==0) write(*,'(A,i6,A,e12.5,A,e12.5,A,e12.5,2A)') &
                'PNM:',iter,'  ', f_simplex(1),'  ', f_simplex(n+1),'  ', dist,' xbest = ',printx(x_simplex(1,:))

            !Check whether length of simplex sides from best point are less than xtoler
            dist = 0.e0_rt
            do i = 2,n+1
                dist = max(dist,sqrt(sum((x_simplex(i,:)-x_simplex(1,:))**2)))                
            end do
            if (dist<xtoler) then
                if (printmod>0) &
                    write(*,'(A,i6,A,e12.5,A,e12.5,A,e12.5)') 'PNM:',iter,'  ', minval(f_simplex),'  ', maxval(f_simplex),'  ', dist
                if (printmod>0) &
                    write(*,'(A)') 'PNM:Simplex finished: simplex size less than xtoler'
                exit
            end if
            if (f_simplex(1)<toler) then
                if (printmod>0) &
                    write(*,'(A,i6,A,e12.5,A,e12.5,A,e12.5)') 'PNM:',iter,'  ', minval(f_simplex),'  ', maxval(f_simplex),'  ', dist
                if (printmod>0) &
                    write(*,'(A)') 'PNM:Simplex finished: f is less than toler'
                exit
            end if            

            !Calc center of gravity of n other pts
            x_gravity = sum(x_simplex(1:n+1-np,:),1)/real(n+1-np,rt) 

            !!!!! PARALLEL ALGORITHM !!!!!!
            ! Try to replace the worst P points, N-P,N-P+1,...,N+1, with other points.
            do ip = n+1-np+ mympi_a(np),n+1-np+ mympi_b(np)

                !Reflect xr = xg + alpha*(xg-x(n+1)) and calculate value
                x_reflect = x_gravity + alpha*(x_gravity-x_simplex(ip,:))
                f_reflect = f(x_reflect)
                
                !Determine what to do with reflected point
                if (f_reflect<=f_simplex(1)) then
                    
                    !If f_reflect is best point, then try expanding in this direction for further improve
                    x_expand = x_reflect + beta*(x_reflect-x_gravity); f_expand = f(x_expand)

                    if (f_expand<f_reflect) then
                        x_simplex(ip,:) = x_expand;  f_simplex(ip) = f_expand
                    else
                        x_simplex(ip,:) = x_reflect; f_simplex(ip) = f_reflect
                    end if

                    didImprove(ip) = .true.

                elseif (f_simplex(1)<f_reflect .and. f_reflect<f_simplex(n+1-np)) then 

                    x_simplex(ip,:) = x_reflect; f_simplex(ip) = f_reflect
                    didImprove(ip) = .true.

                elseif (f_reflect>=f_simplex(n+1-np)) then

                    !If f_reflect is worst point, generate a new point by contraction
                    if (f_reflect>=f_simplex(ip)) then
                        x_reflect = x_simplex(ip,:); f_reflect = f_simplex(ip)
                    end if

                    x_contract = x_gravity + gamma*(x_reflect-x_gravity)
                    f_contract = f(x_contract)

                    if (f_contract<f_reflect) then
                        !If found a better point, then use it
                        x_simplex(ip,:) = x_contract; f_simplex(ip) = f_contract
                        didImprove(ip) = .true.
                    else
                        ! These are not actually the reflection points, but the original point
                        x_simplex(ip,:) = x_reflect; f_simplex(ip) = f_reflect  
                        didImprove(ip) = .false.
                    end if 
                else 
                    !Case not found
                    print*,'f_reflect:',f_reflect
                    print*,'f_simplex:',f_simplex
                    STOP 'simplex: case not found'
                end if

            end do

            ! Now every process has updated a simplex point. Use all gather to collect them
            ilo = n+1-np+1
            ihi = n+1
            call mympi_allgatherv(f_simplex(ilo:ihi))
            x_simplex_transpose = transpose(x_simplex)
            call mympi_allgatherv(x_simplex_transpose(:,ilo:ihi))
            x_simplex = transpose(x_simplex_transpose)
            call mympi_allgatherv(didImprove(ilo:ihi))

            if (.not. any(didImprove(ilo:ihi))) then

                ! Contract all
                do i = 2,n+1
                    x_simplex(i,:) = x_simplex(1,:) + eta*(x_simplex(i,:)-x_simplex(1,:))
                end do

                ! Evaluate at the new simplex 
                do i = 1+mympi_a(n),1+mympi_b(n)
                    f_simplex(i) = f(x_simplex(i,:))
                end do
                call mympi_allgatherv(f_simplex(2:n+1))

            end if

        end do
 
        !Set out
        x = x_simplex(minloc(f_simplex,1),:)

        if (printmod>0 .AND. printmod==1) then
            write(*,'(A)') 'PNM: Final simplex: '
            do i = 1,n+1
                write(*,'(A,i4,A,99(e12.5))') 'PNM: i ',i,' x ',x_simplex(i,:)
            end do
            write(*,'(A)') 'PNM: Final simplex values: '
            write(*,'(A,99(e12.5))') 'PNM: ',f_simplex
            write(*,'(A)') 'PNM:Final minimum x: '
            write(*,'(A,99(e12.5))') 'PNM: ',x
            write(*,'(A)') 'PNM:Final minimum f: '
            write(*,'(A,e12.5)') 'PNM:',minval(f_simplex)
        end if

    end subroutine sub_parallelsimplex

    !Shrinking hypercube
    ! Starting with an initial guess and a hypercube centered on that guess, 
    ! this draws a random point from within the hypercube.
    ! If the value is better (less) than the current value, the hypercube edges are doubled
    ! If the value is worse (greater), the hypercube edges are decreased by a small amount
    ! 
    ! I originally got this idea from the web. It is discussed in http://linneus20.ethz.ch:8080/4_4_2.html
    ! which looks like it was copied from somewhere. If you know where the original article is from, please
    ! let me know and I'll credit appropriately.
    subroutine sub_shrinkingcube(x,f,e_in,maxiter_in,printmod_in,xtoler_in,shrinkFactor_in)
        implicit none
        real(rt), dimension(1:), intent(inout) :: x
        interface 
            function f(x) 
                use mod_types
                implicit none
                real(rt), dimension(1:), intent(in) :: x
                real(rt) :: f
            end function f
        end interface
        real(rt), dimension(1:size(x)), intent(in), optional :: e_in
        integer, intent(in), optional :: maxiter_in,printmod_in
        real(rt), intent(in), optional :: xtoler_in, shrinkFactor_in
        !local
        integer :: maxiter, printmod, iter
        real(rt), dimension(1:size(x)) :: e, bestx, draw
        real(rt) :: xtoler, fx, bestf, shrinkFactor
        integer :: seedSize,seed(100)

        ! Get the current state of the seed and save it to not interfere with a main program
        call random_seed(size=seedSize)
        if (seedSize>size(seed)) STOP 'sub_shrinkingcube: must increase seed size'
        call random_seed(get=seed(1:seedSize))        

        !Set default values
        e = .02e0_rt*max(abs(x),1e-5_rt) !the size of neighborhood about each point which is a 
                                   !hypersphere stretched to edge length e in each dimension
        maxiter = 1000
        printmod = 1
        xtoler = 1e-8_rt
        shrinkFactor = .99e0_rt !Default suggested on website

        !Overwrite if applicable
        if (present(maxiter_in)) maxiter = maxiter_in
        if (present(printmod_in)) printmod = printmod_in
        if (present(e_in)) e = e_in
        if (present(xtoler_in)) xtoler = xtoler_in
        if (present(shrinkFactor_in)) shrinkFactor = shrinkFactor_in

        if (printmod>0) write(*,'(A)') 'SH: ________________________________________________________'            
        if (printmod>0) write(*,'(A)') 'SH: Beginning Shrinking Hypercube'            

        ! Initialize
        fx = f(x)
        bestf = fx
        bestx = x        

        ! 
        do iter = 1,maxiter

            if (printmod>0) write(*,'(A,i10,A,d12.5,A,d12.5,A,d12.5)') &
                'SH: iter',iter, ' f',fx,' bestf',bestf, ' cubeVol', product(e)

            ! If a cube edge is less than xtoler stop
            if (any(e<xtoler)) exit
        
            ! Take random draw from within cube
            call random_number(draw)
            draw = draw*2.e0_rt-1.e0_rt !now uniform in [-.5,.5]
            x = bestx + e*draw

            ! Evalute current point
            fx = f(x)

            ! If best save it and expand, otherwise shrink it
            if (fx<bestf) then
                bestf = fx
                bestx = x
                e = e*2.e0_rt
            else
                e = e*shrinkFactor
            end if

        end do

        x = bestx

        if (printmod>0 .AND. printmod==1) write(*,'(A)') 'SH: Final minimum x: '
        if (printmod>0 .AND. printmod==1) write(*,*) 'SH:', x !call disp(x)            
        if (printmod>0 .AND. printmod==1) write(*,'(A)') 'SH: Final minimum f: '
        if (printmod>0 .AND. printmod==1) write(*,'(A,d12.5)') 'SH:',bestf
        if (printmod>0) write(*,'(A)') 'SH:________________________________________________________'

    end subroutine sub_shrinkingcube

    ! Simulated Annealing (SA)
    ! 
    ! Algorithm description: 
    ! Begins at x and draws random normal innovations to generate x'. x' is accepted whenever f(x')<f(x)
    ! Even when f(x')>f(x), it still accepted according to the probability that U < exp((fx-f(x'))/T).
    ! The temperature T "cools" slowly trending to 0. The innovations are drawn with a decreasing variance.
    ! 
    ! The temperature begins at T0 and cools with T(k) = T0/(1+log(k)).
    ! The variance begins at Var(0)=e**2 and decreases with Var(k) = Var(0)/(1+log(k)).
    !
    ! Global convergence is guaranteed for T0 "large enough". Unfortunately, this is not really helpful.
    ! Part of the program output is an acceptance rate. One may try to tune T0 and e so that this is not much
    ! more than 30% but also not 0%. Also, the "currf" and "bestf" should not be too far apart
    ! 
    ! Optional input:
    ! e: initial standard deviation of the shocks to x.
    ! T0: initial temperature.
    ! a: lower bound on parameter space
    ! b: upper bound on parameter space
    ! maxiter: self-explanatory
    ! printmod: self-explanatory
    ! 
    subroutine sub_simanneal(x,f,e_in,a_in,b_in,T0_in,maxiter_in,printmod_in,aggressive_in)
        use mod_matlab
        implicit none
        real(rt), dimension(1:), intent(inout) :: x
        interface 
            function f(x) 
                use mod_types
                implicit none
                real(rt), dimension(1:), intent(in) :: x
                real(rt) :: f
            end function f
        end interface
        real(rt), dimension(1:size(x)), intent(in), optional :: e_in,a_in,b_in
        real(rt), intent(in), optional :: T0_in
        integer, intent(in), optional :: maxiter_in,printmod_in
        logical, intent(in), optional :: aggressive_in
        !local
        real(rt), parameter :: fracEps = .02e0_rt
        integer :: maxiter, printmod
        real(rt) :: T,T0
        real(rt), dimension(1:size(x)) :: a, b, e, y, xbest
        real(rt) :: fx,fy,fxbest
        integer :: seedSize,seed(100)
        integer :: n, iter, naccept, nreject
        real(rt) :: randnDraw(size(x)), randuDraw
        logical :: aggressive

        n = size(x)

        ! Check seedSize is large enough
        call random_seed(size=seedSize)
        if (seedSize>size(seed)) STOP 'sub_simanneal: must increase seed size'

        !Set defaults
        maxiter = 10000
        printmod = 1
        a = -huge(0.e0_rt) !the lowerbound on the state space
        b = huge(0.e0_rt)  !the upperbound on the state space
        e = .01e0_rt*max(abs(x),1e-4_rt) !the size of neighborhood about each point which is a hypersphere stretched to diameter e in each dimension
        T0 = 1e0_rt !temperature scaling  (setting T0 sufficiently ensures a.s. convergence)
        aggressive = .false.

        !Overwrite if present
        if (present(maxiter_in)) maxiter = maxiter_in
        if (present(printmod_in)) printmod = printmod_in
        if (present(a_in)) a = a_in
        if (present(b_in)) b = b_in
        if (present(e_in)) e = e_in
        if (present(T0_in)) T0 = T0_in
        if (present(aggressive_in)) aggressive = aggressive_in

        ! Check inputs
        if (any(x<a) .or. any(x>b)) STOP 'sub_simanneal: x was not initialized to be between a and b!'

        ! Evaluate f at x
        fx = f(x)

        ! Initialize best values
        fxbest = fx
        xbest = x

        naccept = 0
        nreject = 0

        if (printmod>0) then
            write(*,'(A)') 'SA: ________________________________________________________'            
            write(*,'(A)') 'SA: Beginning Simulated Annealing Method '            
        end if
        do iter = 1,maxiter

!             ! Save the seed state, draw a random normal (size n) and random uniform 
!             !!! Get the current state of the seed and save it to not interfere with a main program
!             call random_seed(get=seed(1:seedSize))
            !!! Draw the variables
            randnDraw = randn(n)
            call random_number(randuDraw)
!             !!! Return the state of the seed to what it was before sim annealing was called
!             call random_seed(put=seed(1:seedSize))

            ! Set the current temperature
            T = T0/(1e0_rt+log(real(iter,rt))) ! T0 to 0
! T = T0/(1e0_rt+real(iter-1,8)) ! T0 to 0

            ! Draw a newpoint within the neighborhood of current x
            !!! I saw here http://stats.stackexchange.com/questions/27923/should-the-proposal-distribution-in-simulated-annealing-depend-on-the-temperatur
            !!! that can draw from this variance decreasing in T and will still get convergence (for large enough T0)
            if (.not. aggressive) then
                y = x + e*randnDraw*sqrt(T/T0) ! Initial innovations are drawn from N(0,e**2). Subsquent, N(0,e**2/(1+log(k))**2)
            else
                y = x + e*randnDraw*T/T0 ! Initial innovations are drawn from N(0,e**2). Subsquent, N(0,e**2/(1+log(k))**2)
            end if


            ! Bound y so it remains in the parameter space
            y = min(max(y,a),b)

            ! Evaluate f at y 
            fy = f(y)

            ! Determine accept reject
            if (fy<fx) then 
                ! In this variant we always accept better moves
                x = y
                fx = fy
                naccept = naccept + 1
            else
                if (randuDraw<exp((fx-fy)/T)) then
                    ! Move from x to y
                    x = y
                    fx = fy
                    naccept = naccept + 1
                else
                    ! Keep current x
                    nreject = nreject + 1
                end if
            end if

            ! Try
            if (printmod>0 .and. mod(iter,printmod)==0) then
                write(*,'(A,i6,A,d12.5,A,d12.5,A,d12.5,A,d12.5,A,d12.5)') &
                    'SA: iter ',iter,' ftrial',fy,' fcurr',fx,' fbest',fxbest,' Temp', T,&
                    ' Accept%', 100.e0_rt*real(naccept,8)/real(naccept+nreject,8)
            end if

            ! Save best f
            if (fx<fxbest) then 
                fxbest = fx
                xbest = x
            end if

        end do

        ! Return the best function value
        if (printmod>0) then
            if (fxbest/=fx) then
                write(*,'(A,d12.5,A)') 'SA: best value ever encountered, ',fxbest,', being returned, it is /= fx'
            end if
        end if
        x = xbest
        fx = fxbest

    end subroutine sub_simanneal

    ! Parallel Simulated Annealing (PSA)
    ! 
    ! All the argumetns are the same as the regular simulated annealing. 
    ! The algorithm is the same except that at each iteration, P proposals are drawn at the same time.
    ! The best of the proposals is taken. 
    ! 
    subroutine sub_parallelsimanneal(x,f,e_in,a_in,b_in,T0_in,maxiter_in,printmod_in,aggressive_in)
        use mod_mympi
        use mod_matlab
        implicit none
        real(rt), dimension(1:), intent(inout) :: x
        interface 
            function f(x) 
                use mod_types
                implicit none
                real(rt), dimension(1:), intent(in) :: x
                real(rt) :: f
            end function f
        end interface
        real(rt), dimension(1:size(x)), intent(in), optional :: e_in,a_in,b_in
        real(rt), intent(in), optional :: T0_in
        integer, intent(in), optional :: maxiter_in,printmod_in
        logical, intent(in), optional :: aggressive_in
        !local
        real(rt), parameter :: fracEps = .02e0_rt
        integer :: maxiter, printmod, n, iter, naccept, nreject, np, ip, itmp(1), seedSize, seed(100)
        real(rt) :: fx,fy,fxbest,T,T0,randuDraw
        real(rt), dimension(1:size(x)) :: a, b, e, y, xbest
        real(rt), allocatable :: z(:,:),fz(:)
        logical :: aggressive

        n = size(x)
        np = mympi_nproc()

        allocate(z(n,np),fz(np))

        ! Check seedSize is large enough
        call random_seed(size=seedSize)
        if (seedSize>size(seed)) STOP 'sub_simanneal: must increase seed size'

        !Set defaults
        maxiter = 10000
        printmod = 1
        a = -huge(0.e0_rt) !the lowerbound on the state space
        b = huge(0.e0_rt)  !the upperbound on the state space
        e = .01e0_rt*max(abs(x),1e-4_rt) !the size of neighborhood about each point which is a hypersphere stretched to diameter e in each dimension
        T0 = 1e0_rt !temperature scaling  (setting T0 sufficiently ensures a.s. convergence)
        aggressive = .false.

        !Overwrite if present
        if (present(maxiter_in)) maxiter = maxiter_in
        if (present(printmod_in)) printmod = printmod_in
        if (present(a_in)) a = a_in
        if (present(b_in)) b = b_in
        if (present(e_in)) e = e_in
        if (present(T0_in)) T0 = T0_in
        if (present(aggressive_in)) aggressive = aggressive_in

        printmod = merge(printmod,-1,mympi_iproc()==0)

        ! Check inputs
        if (any(x<a) .or. any(x>b)) STOP 'sub_simanneal: x was not initialized to be between a and b!'

        ! Initialize best values
        fxbest = huge(0e0_rt)
        xbest = 0e0_rt

        naccept = 0
        nreject = 0

        if (printmod>0) then
            write(*,'(A)') 'PSA: ________________________________________________________'            
            write(*,'(A)') 'PSA: Beginning Parallel Simulated Annealing Method '            
        end if
        do iter = 1,maxiter

            ! Set the current temperature
            T = T0/(1e0_rt+log(real(iter,rt))) ! T0 to 0

            !!! Draw the variables
            ! Draw a newpoint within the neighborhood of current x
            !!! I saw here http://stats.stackexchange.com/questions/27923/should-the-proposal-distribution-in-simulated-annealing-depend-on-the-temperatur
            !!! that can draw from this variance decreasing in T and will still get convergence (for large enough T0)
            do ip = 1,np

                if (iter==1 .and. ip==1) then
                    z(:,ip) = x
                else
                    do while (.true.)

                        if (.not. aggressive) then
                            z(:,ip) = x + e*randn(n)*sqrt(T/T0) ! Initial innovations are drawn from N(0,e**2). Subsquent, N(0,e**2/(1+log(k))**2)
                        else
                            z(:,ip) = x + e*randn(n)*T/T0 ! Initial innovations are drawn from N(0,e**2). Subsquent, N(0,e**2/(1+log(k))**2)
                        end if

                        ! Use rejection sampling to get the new point
                        if (all(z(:,ip)>=a) .and. all(z(:,ip)<=b)) exit

                    end do
                end if

            end do

            ! Evaluate f in parallel and share
            do ip = mympi_a(np),mympi_b(np)
                fz(ip) = f(z(:,ip))
            end do

            call mympi_allgatherv(fz)

            ! Take the best
            fy = minval(fz)
            itmp = minloc(fz)
            y = z(:,itmp(1))

            ! Determine accept reject
            if (iter==1) then
                x = y
                fx = fy
            elseif (fy<fx) then 
                ! In this variant we always accept better moves
                x = y
                fx = fy
                naccept = naccept + 1
            else
                call random_number(randuDraw)
                if (randuDraw<exp((fx-fy)/T)) then
                    ! Move from x to y
                    x = y
                    fx = fy
                    naccept = naccept + 1
                else
                    ! Keep current x
                    nreject = nreject + 1
                end if
            end if

            ! Try
            if (printmod>0 .and. mod(iter,printmod)==0) then
                write(*,'(A,i6,A,d12.5,A,d12.5,A,d12.5,A,d12.5,A,d12.5)') &
                    'PSA: iter ',iter,' ftrial',fy,' fcurr',fx,' fbest',fxbest,' Temp', T,&
                    ' Accept%', 100.e0_rt*real(naccept,8)/real(naccept+nreject,8)
            end if

            ! Save best f
            if (fx<fxbest) then 
                fxbest = fx
                xbest = x
            end if

        end do

        ! Return the best function value
        if (printmod>0) then
            if (fxbest/=fx) then
                write(*,'(A,d12.5,A)') 'PSA: best value ever encountered, ',fxbest,', being returned, it is /= fx'
            end if
        end if
        x = xbest
        fx = fxbest

    end subroutine sub_parallelsimanneal

    ! Grid search (GS)
    ! 
    ! Performs a brute force grid search. The tensor product grid is formed from the variables "grids" and "n".
    ! The grid in the ith dimension is grids(1:n(i),i). 
    subroutine sub_gridsearch(x,f,n,grids,printmod_in)
        use mod_matlab
        implicit none
        real(rt), dimension(1:), intent(inout) :: x
        interface 
            function f(x) 
                use mod_types
                implicit none
                real(rt), dimension(1:), intent(in) :: x
                real(rt) :: f
            end function f
        end interface
        real(rt), dimension(:,:), intent(in) :: grids
        integer, dimension(:), intent(in) :: n
        integer, intent(in), optional :: printmod_in
        !local
        real(rt), dimension(size(x)) :: xbest
        real(rt) :: fxbest,fx
        integer :: printmod, d, j, i
        integer, dimension(size(x)) :: ivec

        ! Get dimension of the problem
        d = size(x)

        ! Check 
        if (d/=size(n)) STOP 'ERROR: sub_gridsearch: n has wrong size'
        if (d/=size(grids,2)) STOP 'ERROR: sub_gridsearch: grids has wrong size 2'
        if (maxval(n)>size(grids,1)) STOP 'ERROR: sub_gridsearch: grids has wrong size 1'

        ! Set defaults
        printmod = 1

        ! Overwrite if present
        if (present(printmod_in)) printmod = printmod_in

        ! Check inputs
!         if (any(b<a)) STOP 'sub_gridsearch: ERROR: b<a' 
        if (any(n<=0)) STOP 'sub_gridsearch: ERROR: n<=0'

        if (printmod>0) then
            write(*,'(A)') 'GS: ________________________________________________________'            
            write(*,'(A)') 'GS: Beginning Grid Search'            
        end if

        ! Do the grid search
        fxbest = huge(0e0_rt)
        ivec(:) = 1
        do i = 1,product(n)

            ! Get x
            do j = 1,d
                x(j) = grids(ivec(j),j)
            end do

            ! Evaluate 
            fx = f(x)

            ! Store the best
            if (fx<fxbest) then
                xbest = x
                fxbest = fx
            end if

            ! Try
            if (printmod>0 .and. mod(i,printmod)==0) then
                write(*,'(A,i6,A,d12.5,A,d12.5)') 'GS: iter ',i,' f',fx,' fbest',fxbest
            end if

            ! Increment ivec
            do j = 1,d
                if (mod(ivec(j),n(j))==0) then
                    ivec(j) = 1 
                else
                    ivec(j) = ivec(j)+1
                    exit
                end if
            end do
        end do
    
        ! Return the best
        x = xbest

        ! Return the best function value
        if (printmod>0) then
            write(*,'(A)') 'GS: finished.'
        end if

    end subroutine sub_gridsearch


!     ! Grid search (GS)
!     ! 
!     ! Performs a basic grid search over the domain [a,b] with n giving the points in each dimension
!     !
!     subroutine sub_gridsearch(x,f,a,b,n,printmod_in)
!         use mod_matlab
!         implicit none
!         real(rt), dimension(1:), intent(inout) :: x
!         interface 
!             function f(x) 
!                 implicit none
!                 real(rt), dimension(1:), intent(in) :: x
!                 real(rt) :: f
!             end function f
!         end interface
!         real(rt), dimension(1:size(x)), intent(in) :: a,b
!         integer, dimension(1:size(x)), intent(in) :: n
!         integer, intent(in), optional :: printmod_in
!         !local
!         real(rt), dimension(size(x)) :: xbest
!         real(rt) :: fxbest,fx
!         integer :: printmod, d, j, i
!         integer, dimension(size(x)) :: ivec
! 
!         ! Get dimension of the problem
!         d = size(x)
! 
!         ! Set defaults
!         printmod = 1
! 
!         ! Overwrite if present
!         if (present(printmod_in)) printmod = printmod_in
! 
!         ! Check inputs
!         if (any(b<a)) STOP 'sub_gridsearch: ERROR: b<a' 
!         if (any(n<=0)) STOP 'sub_gridsearch: ERROR: n<=0'
! 
!         if (printmod>0) then
!             write(*,'(A)') 'GS: ________________________________________________________'            
!             write(*,'(A)') 'GS: Beginning Grid Search'            
!         end if
! 
!         ! Do the grid search
!         fxbest = huge(0e0_rt)
!         ivec(:) = 1
!         do i = 1,product(n)
! 
! !             print*,'ivec',ivec
! 
!             ! Get x
!             do j = 1,d
!                 x(j) = a(j) + real(ivec(j)-1,8)/real(n(j)-1,8)*(b(j)-a(j))
!             end do
! 
!             ! Evaluate 
!             fx = f(x)
! 
!             ! Store the best
!             if (fx<fxbest) then
!                 xbest = x
!                 fxbest = fx
!             end if
! 
!             ! Try
!             if (printmod>0 .and. mod(i,printmod)==0) then
!                 write(*,'(A,i6,A,d12.5,A,d12.5)') 'GS: iter ',i,' f',fx,' fbest',fxbest
!             end if
! 
!             ! Increment ivec
!             do j = 1,d
!                 if (mod(ivec(j),n(j))==0) then
!                     ivec(j) = 1 
!                 else
!                     ivec(j) = ivec(j)+1
!                     exit
!                 end if
!             end do
!         end do
!     
!         ! Return the best
!         x = xbest
! 
!         ! Return the best function value
!         if (printmod>0) then
!             write(*,'(A)') 'GS: finished.'
!         end if
! 
!     end subroutine sub_gridsearch

    ! Genetic algorithm --- MPI version
    ! 
    ! attempts to solve max f(x) s.t. x in {0,1}^n 
    ! 
    ! It generates N random uniform points.
    ! It takes the worst point and tries to get rid of it by reflecting it
    ! (like in Nelder-Mead) through a simplex of points. This simplex is chosen 
    ! by randomly selecting n+1 points (whose function values are known) with 
    ! replacement, giving a total of n+1 points. 
    !
    ! The centroid of the simplex is then used for reflecting the worst point. 
    ! If there reflected point improves on the worst point (and is in a,b), it is
    ! accepted. O/w, the process repeats.
    ! 
    ! This is modified for parallel purposes by attempting to simultaneously get 
    ! rid of several of the worst points simultaneously. The parallel design here,
    ! in contrast to CRS2LM, is more advantageous when the functions time to completion
    ! is not uniform across x values. The algorithm itself is more global.
    !
    ! The optional argument sleeptime is useful when there are only a few ranks. 
    ! It specifies how long the master should sleep if no free workers are found. This
    ! keeps processing cycles free for those ranks actually doing work. 
    subroutine sub_ga_MPI(xstar,f,opts)
        use mod_util
        use mod_mympi
        use mod_matlab
        implicit none
        integer, intent(inout) :: xstar(:)
        type(minNdim_opts), intent(in), optional :: opts
        interface 
            function f(x) 
                use mod_types
                implicit none
                integer, dimension(:), intent(in) :: x
                real(rt) :: f
            end function
        end interface
        ! local other
        integer, allocatable, target :: X(:,:),slavexlist(:,:) ! NOTE: these have target attribute and will be pointed at
        real(rt), allocatable, target :: y(:),slaveylist(:) ! NOTE: these have target attribute and will be pointed at
        logical, allocatable :: finishedtask(:),slaveeverassignedlist(:)
        integer, allocatable :: fevaldsn(:),requestlist(:) ! ,lasti(:)
        real(rt) :: tmp
        integer :: nx, N, ix, iter, i, who, itmp(1), prop, ib, iw
        type(minNdim_opts) :: o

        if (present(opts)) o = opts

        call disp_opts(o)

        nx = size(xstar)

        if (mympi_nproc()<=1) stop 'ERROR: sub_ga_MPI: called with only one rank'

        ! N is the pool of genes size
        N = mympi_nproc()-1 ! ensure at least N-1 ranks, o/w they would be doing nothing
        if (o%poolsize/=-1) N = max(N,o%poolsize) ! if specified, ensure at least that poolsize is used
        N = max(N,100) ! ensure at least 100

        ! If not master, go to the "slave" routine and wait for a x to evaluate
        if (mympi_iproc()/=0) then
            call slave()
            return
        end if

        ! Generate N random points uniformly from [a,b]
        allocate(finishedtask(mympi_nproc()-1))
        allocate(X(nx,N))
        allocate(y(N))
        allocate(fevaldsn(mympi_nproc()-1))
        allocate(requestlist(mympi_nproc()-1))
        allocate(slavexlist(nx,mympi_nproc()-1))
        allocate(slaveylist(mympi_nproc()-1))
        allocate(slaveeverassignedlist(mympi_nproc()-1))

        ! For random results, initialize based on the clock.
        if (o%reproducible==0) then 
            call init_random_seed()
        elseif (o%reproducible==1) then
            ! should draw from some particular stream each time. 
            ! for now though, just use the default seed (the issue 
            ! is this will only be reproducible if f never changes
            ! the number or order of function calls.
        else
            stop 'ERROR: sub_ga_MPI: opts%reproducible should be 0 or 1'
        end if 

        ! Generate initial pool of points
        X(:,1) = xstar
        do i = 2,N
            do ix = 1,nx
                ! call random_number(tmp)
                ! X(ix,i) = merge(1,0,tmp>.5d0)
                X(ix,i) = mod(myirand(),2)
            end do
        end do

        fevaldsn = 0

        finishedtask = .true.

        ! Evaluate at each of the N points in parallel
        write(*,'(A,i4,A)') 'GA: Solving for initial pool of ',N,' points'
        do i = 1,N

            ! Wait for a free worker
            if (o%dbglvl>0) print*,'MASTER: WAITING FOR FREE WORKER'
            who = wait_for_free_worker()

            ! Assign the point
            if (o%dbglvl>0) print*,'MASTER: X(:,i)',X(:,i)
            call assign_xy_to_worker(X(:,i),y(i),who)

            if (o%printmod>0) write(*,'(A,i4,A,i4,A,i4)') 'GA: image ', who, ' solving for point ',i,' of ',N

            ! For statistical purposes, collect who was assigned that feval
            fevaldsn(who) = fevaldsn(who) + 1

        end do

        if (o%dbglvl>1) print*,'waiting for everyone to finish'

        ! Wait for everyone to finish
        do while (any(.not. finishedtask))

            if (o%dbglvl>1) print*,'finishedtask',finishedtask
            if (o%dbglvl>1) print*,'requestlist',requestlist
            do who = 1,mympi_nproc()-1
                if (.not. finishedtask(who)) then
                    finishedtask(who) = mympi_test(requestlist(who)) ! requestlist is not valid after mympi_test returns true
                end if
            end do

        end do

        finishedtask = .true.
        slaveeverassignedlist = .false.

        if (o%dbglvl>1) print*,'done initial point distribution, proceeding to main loop'

        do iter = 1,o%maxiter

            ! Wait for a slave. If there were previously assigned a task, then record its value.
            who = wait_for_free_worker()

            ! If they had a previously assigned point, use its value
            if (slaveeverassignedlist(who)) then

                ! CRS assumption:

                ! Replace the worst point if found a better one
                ! itmp = minloc(y) 
                ! if (slaveylist(who)>y(itmp(1))) then
                !     y(itmp(1)) = slaveylist(who)
                !     x(:,itmp(1)) = slavexlist(:,who)
                ! end if

                ! GA assumption: replace a random parent. If elitist, ensure best is not replaced
                itmp = maxloc(y) 
                prop = itmp(1)
                do while (prop==itmp(1)) ! ensures best is never replaced (elitist)
                    call random_number(tmp)
                    tmp = tmp*N + .5d0 ! U[.5,N+.5]
                    prop = nint(tmp) 
                    if (prop<1 .or. prop>N) prop = itmp(1)
                end do 
                y(prop) = slaveylist(who)
                x(:,prop) = slavexlist(:,who)

            end if

            ! After this point, "who" will be assigned an x
            slaveeverassignedlist(who) = .true.

            ! Sort the current points (descending)
            ! tmpy = -y
            ! call sort2(y,tmpI,tmpy) 
            ! y = -y
            ! X = X(:,tmpI) ! sort X

            ! Get the best and the worst points
            itmp = maxloc(y); ib = itmp(1) ! best
            itmp = minloc(y); iw = itmp(1) ! worst

            ! Display progress
            if (iter==1 .or. (o%printmod>0 .and. mod(iter,o%printmod)==0)) then
                write(*,'(A,i8,2(A,e14.6))') 'GA: iter',iter,'  best ',y(ib),'  worst ',y(iw)
            end if

            ! Check stopping criteria
            if (abs(y(ib)-y(iw))<o%yabstoler .or. iter==o%maxiter) then

                if (abs(y(ib)-y(iw))<o%yabstoler) then
                    write(*,'(A)') 'GA: convergence criteria satisfied'
                else
                    write(*,'(A)') 'GA: max iterations reached'
                end if

                ! Shutdown and distribute xstar
                xstar = X(:,ib)
                call shutdown_slaves_sharebest()

                ! Wait for everyone to finish 
                call mympi_barrier()

                return

            end if

            ! Generate a new gene by selecting two parents
            if (y(iw)<0._rt) stop 'ERROR: GA: function values should always be positive'
            call gen_gene(slavexlist(:,who),X,y)

            ! Request that they solve for x and store it in slavey
            call assign_xy_to_worker(slavexlist(:,who),slaveylist(who),who)

            fevaldsn(who) = fevaldsn(who) + 1

        end do

    contains

        subroutine gen_gene(new,X,y)
            use mod_discreteAR1
            implicit none
            integer, intent(out) :: new(:) ! new gene
            integer, intent(in) :: X(:,:) ! pool (size new x size y)
            real(rt), intent(in) :: y(:) ! fitnesses
            ! local
            real(rt), allocatable :: prob(:)
            real(rt) :: tmp
            integer :: p1,p2,c1,c2,i,j

            allocate(prob(size(y)))

            ! select two parents according to their fitness
            prob = y - minval(y) 
            prob = prob/sum(prob) ! always drops the worst one

            p1 = drawIid(prob,0)

            prob(p1) = 0d0
            prob = prob/sum(prob)

            p2 = drawIid(prob,0)

            ! select double crossover
            ! call random_number(tmp)
            ! tmp = tmp*size(X,1)
            ! tmp = tmp+.5d0
            ! c1 = nint(tmp)
            ! c1 = min(max(1,c1),size(X,1))

            c1 = 1+mod(myirand(),size(X,1))

            ! call random_number(tmp)
            ! tmp = tmp*size(X,1)
            ! tmp = tmp+.5d0
            ! c2 = nint(tmp)
            ! c2 = min(max(1,c2),size(X,1))

            c2 = 1+mod(myirand(),size(X,1))

            if (c1>c2) then
                i = c1
                c1 = c2
                c2 = i
            end if

            if (c2<c1) stop 'Failed to generate crossover points'

            ! do double crossover
            new(1:c1) = X(1:c1,p1)
            new(c1+1:c2) = X(c1+1:c2,p2)
            new(c2+1:size(new)) = X(c2+1:size(new),p1)

            if (size(new)<10000) then
                ! do bit mutation
                do i = 1,size(new)
                    call random_number(tmp)
                    if (tmp<=o%mutationprob) new(i) = mod(new(i)+1,2)
                end do
            else
                ! need to replace o%mutationprob percent bits randomly.
                ! for large sizes, a more efficient way to do this is to choose o%mutationprob*size(new) indices which
                ! should be accurate roughly via a law of large numbers
                do i = 1,max(1,int(size(new)*o%mutationprob)) ! okay if bigger than size(new)
                    j = 1+mod(myirand(),size(new))
                    new(j) = mod(new(j)+1,2)
                end do
            end if

        end subroutine 

        integer function myirand()
            implicit none
            real(8) :: tmp
            integer, parameter :: bigi = huge(0)-1
            call random_number(tmp)
            myirand = int(bigi*tmp)
        end function 

        ! This routine sends the value of x to a slave "who", who will then store the value in y. 
        subroutine assign_xy_to_worker(x,y,who)
            use iso_c_binding, only: c_loc,c_ptr,c_f_pointer ! GETS THE ACTUALY ADDRESS OF Y WHICH WILL PERSIST AFTER CALL TO ASSIGN ENDS
            implicit none
            integer, intent(in), target :: x(:)
            real(rt), intent(out), target :: y
            integer, intent(in) :: who ! who you want assigned
            ! local
            type(c_ptr) :: yptr_c ! pointer in C
            real(rt), pointer :: yptr_f ! pointer in Fortran 
            integer :: flag,ierr

            if (o%dbglvl>1) print*,'MASTER: assign_xy: x',x(1:nx)

            ! Assign it
            flag = 1
            finishedtask(who) = .false.

            if (o%dbglvl>0) print*,'MASTER: SENDING FLAG: ',flag

            call mympi_send(flag,who) ! blocking send (tell them to proceed to execute)

            if (o%dbglvl>1) print*,'MASTER: FINISHED SENDING FLAG'
            if (o%dbglvl>1) print*,'MASTER: who',who
            if (o%dbglvl>1) print*,'MASTER: SENDING X: ',x(1:nx)

            call mympi_send(x,who) ! assign an x value (wait for it to complete)

            if (o%dbglvl>0) print*,'MASTER: IRECVING Y'

            ! In order to be able to do this assignment where the memory location associated with y will be stored even after 
            ! the subroutine is exited, the safest/most reliable way I have found is to send MPI the actual memory address.
            ! This can be accomplished by getting the memory address of y using the F2003 intrinsic "c_loc" and assigning that
            ! to a C pointer. Then, using c_f_pointer, copy that to a Fortran pointer. Then pass that Fortran poitner to MPI. 
            !
            ! This may be a hack (it seems like completely legal Fortran though), but it simplifies bookkeeping substantially.
            ! I.e., it becomes "set it and forget it."

            yptr_c = c_loc(y) ! get the actual address of y using "c_loc" and have yptr_c point to that
            call c_f_pointer(yptr_c,yptr_f) ! get a fortran pointer to that same address
            call MPI_Irecv(yptr_f, 1, mympi_rt, who, who, MPI_COMM_WORLD, requestlist(who), ierr)

        end subroutine 

        ! a routine to shut down all slaves
        subroutine shutdown_slaves_sharebest() 
            implicit none
            ! local
            integer :: i
            logical :: ft

            ! Cycle through and shutdown
            do i = 1,mympi_nproc()-1

                ! Wait for rank to finish if not finished task
                if (.not. finishedtask(i)) then

                    ! Infinite loop, wait for them to finish 
                    do 

                        ! Check whether i has finished
                        ft = mympi_test(requestlist(i))
                        if (ft) then
                            finishedtask(i) = .true. ! show that they are now free
                            exit
                        end if

                        ! If sleeptime, then sleep
                        if (o%sleeptime>=1) then

                            call sleep(o%sleeptime)

                        else

                            ! Do some busy work so as to not continually check the status (which may cause MPI problems)
                            call short_sleep()

                        end if

                    end do

                end if
                    
                ! Shut down, sending flag = -1 and xstar
                call mympi_send(-1,i) 
                call mympi_send(xstar,i) 

            end do


        end subroutine


        ! a routine returning the index of a free slave
        ! if no worker is available, who is -1
        function wait_for_free_worker() result(who)
            implicit none
            integer :: who
            ! local
            integer :: i
            logical :: ft

            do 

                ! Infinite loop
                do i = 1,mympi_nproc()-1

                    if (finishedtask(i)) then

                        who = i
                        return  

                    else

                        ! Check whether i has finished
                        if (o%dbglvl>1) print*,'MASTER: TESTING MESSAGE'
                        if (o%dbglvl>2) print*,'MASTER: finishedtask',finishedtask
                        if (o%dbglvl>2) print*,'MASTER: REQUESTLIST',REQUESTLIST
                        if (o%dbglvl>2) print*,'MASTER: i: ',i

                        ft = mympi_test(requestlist(i))
                        if (o%dbglvl>1) print*,'MASTER: FINISHED TEST'
                        if (ft) then
                            finishedtask(i) = .true. ! show that they are now free
                            who = i
                            if (o%dbglvl>0) print*,'MASTER: RECEIVED MESSAGE'
                            return
                        end if

                    end if

                end do

                ! If sleeptime, then sleep
                if (o%sleeptime>=1) then

                    call sleep(o%sleeptime)

                else

                    ! Do some busy work so as to not thrash the system
                    call short_sleep()

                end if

            end do

        end function

        ! Do something according to the value of flag(this_image())[1]
        !  1 := evaluate slavex and store in slavey
        ! -1 := exit the slave routine
        subroutine slave()
            implicit none
            integer :: flag
            real(rt) :: slavey
            integer, allocatable :: slavex(:)

            allocate(slavex(nx))

            do 

                ! Wait for a flag 
                if (o%dbglvl>0) print*,'SLAVE: WAITING TO RECEIVE FLAG'
                call mympi_recv(flag,0)
                if (o%dbglvl>1) print*,'SLAVE: flag is ',flag

! call sleep(1)

                ! Wait for an x
                if (o%dbglvl>0) print*,'SLAVE: WAITING TO RECEIVE SLAVEX'
                call mympi_recv(slavex,0)
                if (o%dbglvl>1) print*,'SLAVE: slavex is ',slavex

                ! If flag==1, execute f(x) and store on master.
                if (flag==1) then

                    ! Compute the result
                    slavey = f(slavex)

                    ! Send the result to master tagged by mympi_iproc()
                    if (o%dbglvl>0) print*,'SLAVE: SENDING SLAVEY'
                    call mympi_send(slavey,0,mympi_iproc()) 
                    if (o%dbglvl>0) print*,'SLAVE: FINISHED SENDING SLAVEY'

                end if

                if (flag==-1) then
                    xstar = slavex

                    ! Wait for everyone to finish
                    call mympi_barrier()
                    exit
                end if

            end do

        end subroutine


    end subroutine

    subroutine short_sleep()
        use mod_mympi
        implicit none
        real(8) :: tic,toc,tmp ! leave as double precision
        integer :: uid

        tic = MPI_WTIME()
        toc = tic
        tmp = 0d0
        do while (toc-tic<1d-6)
            toc = MPI_WTIME()
            tmp = tmp + 1d0
        end do

        ! Trick the compiler into not optimizing away the busy work (hopefully)
        uid = 152
        open(uid,file='/lkandsfkasn/makecompilernotoptimize.dat',status='unknown',err=999) ! file should not exist (but could), so this will jump to line 999
        write(uid,*) tmp
        close(uid)
999     tmp = 0d0

    end subroutine 


end module mod_minNdim

module mod_minNdim_test
    use mod_minNdim
    implicit none
contains

    function mysumf(x) result(y)
        use mod_types
        implicit none
        integer, intent(in) :: x(:)
        real(rt) :: y
        y = sum(real(x,rt)) / real(size(x),rt)
    end function 

    subroutine test_GA_MPI()
        implicit none
        integer, allocatable :: x(:)

        allocate(x(10001))
        x = 0
        call sub_ga_MPI(x,mysumf,optimset('maxiter=1000000'))

    end subroutine 

    function testfn_noisyLM(x,nf) result(y)
        use mod_matlab
        implicit none
        real(rt), intent(in) :: x(:)
        integer, intent(in) :: nf
        real(rt) :: y(nf)

        ! y = sin(x) - real(colon(1,size(x)),8)*(1e0_rt-cos(x)) + sum(cos(x)) - size(x) ! + floor(mod(exp(x)-1+.2,1)) + exp(.3*x)-1

!         y = log(abs(x(1)*x(2) + sin(x(1)) - exp(cos(x(2)))))

        ! x = exp(x)
        y = x - sin(x)**2



    end function 
!     subroutine test_parnoisyLM
!         use mod_mympi
!         implicit none
!         real(rt) :: x(2)
! 
!         x = [-5e0_rt,-5e0_rt]
!         call mympi_barrier()
!         call sub_parnoisyLM(x,testfn_noisyLM,size(x),1d-3 + 0e0_rt*x,ndraws=1000,opts=optimset('maxiter',300e0_rt))
!         write(*,*) 'x: ',x
!         write(*,*) 'fx: ',testfn_noisyLM(x,size(x))
! 
!     end subroutine 
!     subroutine test_noisyLM
!         use mod_mympi
!         implicit none
!         real(rt) :: x(2)
!         if (mympi_iproc()==0) then
!             x = [-5e0_rt,-5e0_rt]
!             call sub_noisyLM(x,testfn_noisyLM,size(x),1d-3 + 0e0_rt*x,opts=optimset('maxiter',300e0_rt))
!             write(*,*) 'x: ',x
!             write(*,*) 'fx: ',testfn_noisyLM(x,size(x))
!         end if
!     end subroutine 

    subroutine test_opt_along_dim
        implicit none
        real(rt) :: x(3)
        x = (/.3e0_rt, .1e0_rt, .5e0_rt /)

        call sub_opt_along_dim(x,rosenbrock,-1._rt,1._rt,2)

    end subroutine 

    function lm_foo(x,m) 
        use mod_types
        implicit none
        real(rt), dimension(:), intent(in) :: x
        integer, intent(in) :: m
        real(rt) :: lm_foo(m)
        ! local
        integer :: i

        ! If m==size(x), then solution for min SSRS problem is 1,2,3,...
        do i = 1,m
            lm_foo(i) = abs(x(min(i,size(x)))-i)**2
        end do
        print*,'x',x

    end function

    subroutine test_levenbergmarquardt()
        implicit none
        integer, parameter :: m = 3, n = 3
        real(rt) :: xstar(n)
        xstar = 10._rt
        call sub_levenbergmarquardt(xstar,lm_foo,m,optimset('yreltoler=1d-6,xreltoler=1d-6,fcnerr=1d12,printmod=1')) ! fcnerr=1d16 step size is 100%, 1d14 by 10%, 1d12 by 1%
        if (.not. all(abs(xstar-[1._rt,2._rt,3._rt])<1e-6_rt)) stop 'test_levenbergmarquardt: failed'
        print*,'test_levenbergmarquardt: passed'
    end subroutine

    subroutine test_parmin
        implicit none
        real(rt) :: x(3)
        x = (/.3e0_rt, .1e0_rt, .5e0_rt /)
!         call sub_parmin_MPI(x,rosenbrock,opt=0,e=abs(x)*.5e0_rt)
        call sub_parmin_OMP(x,rosenbrock,x*0e0_rt,x*10e0_rt,[.1e0_rt,.1e0_rt,.1e0_rt],optimset('printmod=1'))
        stop 'checking test_parmin'
        write(*,*) 'x: ',x
        write(*,*) 'fx: ',rosenbrock(x)
    end subroutine 
!     subroutine test_subplex
!         implicit none
!         real(rt) :: x(3)
!         x = (/.3e0_rt, .1e0_rt, .5e0_rt /)
!         call sub_subplex(x,rosenbrock)
!         write(*,*) 'x: ',x
!         write(*,*) 'fx: ',rosenbrock(x)
!     end subroutine 
    subroutine test_simplex
        implicit none
        real(rt) :: x(3)
        x = (/.3e0_rt, .1e0_rt, .5e0_rt /)
        call sub_simplex(x,rosenbrock,toler_in=1e-6_rt,maxiter_in=10000,printmod_in=1,randomize_in=.false.)
        ! call sub_simplex(x,rosenbrock,optimset('toler=1d-6,maxiter=10000,printmod=15'))
        write(*,*) 'x: ',x
        write(*,*) 'fx: ',rosenbrock(x)
    end subroutine 
    subroutine test_parallelsimplex
        use mod_mympi
        implicit none
        real(rt) :: x(3)
        x = (/.3e0_rt, .1e0_rt, .5e0_rt /)
        call sub_parallelsimplex(x,rosenbrock,toler_in=1e-6_rt,maxiter_in=10000,printmod_in=1)
        write(*,*) 'x: ',x
        write(*,*) 'fx: ',rosenbrock(x)
    end subroutine 


    subroutine test_parallelsimanneal
        implicit none
        real(rt) :: x(3),a(3),b(3)
        x = (/.3e0_rt,.1e0_rt,.5e0_rt/)
        a = 0.e0_rt
        b = (/1.2e0_rt,3e0_rt,99e0_rt/)
        call random_seed
        call sub_parallelsimanneal(x,rosenbrock,a_in=a,b_in=b,e_in=1e0_rt*(/.01e0_rt,.01e0_rt,.01e0_rt/),&
                                   printmod_in=1000,maxiter_in=10000,T0_in=.001e0_rt)
        write(*,*) 'x: ',x
        write(*,*) 'fx: ',rosenbrock(x)
    end subroutine 

    subroutine test_simanneal
        implicit none
        real(rt) :: x(3),a(3),b(3)
        x = (/.3e0_rt,.1e0_rt,.5e0_rt/)
        a = 0.e0_rt
        b = (/1.2e0_rt,3e0_rt,99e0_rt/)
        call random_seed
        call sub_simanneal(x,rosenbrock,a_in=a,b_in=b,e_in=1e0_rt*(/.01e0_rt,.01e0_rt,.01e0_rt/),printmod_in=1000,maxiter_in=10000,T0_in=.001e0_rt)
        write(*,*) 'x: ',x
        write(*,*) 'fx: ',rosenbrock(x)
    end subroutine 

    subroutine test_parCRS_MPI
        use mod_mympi
        implicit none
        real(rt) :: x(3),a(3),b(3),e(3)

        a = 0.e0_rt
        b = [1.2e0_rt,3e0_rt,99e0_rt]

        x = [.3e0_rt,.1e0_rt,.5e0_rt]
        call sub_parCRS_MPI(x,rosenbrock,a,b,opts=optimset('printmod=5,sleeptime=0,yabstoler=1d-10,maxiter=125'))
        if (mympi_iproc()==0) then
            write(*,*) 'x: ',x
            write(*,*) 'fx: ',rosenbrock(x)
        end if

        x = [.3e0_rt,.1e0_rt,.5e0_rt]
        call sub_parCRS_MPI(x,rosenbrock,a,b,opts=optimset('printmod=100,sleeptime=0,yabstoler=1d-10,maxiter=10000'))
        if (mympi_iproc()==0) then
            write(*,*) 'x: ',x
            write(*,*) 'fx: ',rosenbrock(x)
        end if

        x = [.3e0_rt,.1e0_rt,.5e0_rt]
        e = [.01_rt,.01_rt,.01_rt]
        call sub_parCRS_MPI(x,rosenbrock,a,b,e_in=e,opts=optimset('printmod=100,sleeptime=0,yabstoler=1d-10,maxiter=10000,poolsize=2'))
        if (mympi_iproc()==0) then
            write(*,*) 'x: ',x
            write(*,*) 'fx: ',rosenbrock(x)
        end if

    end subroutine 

!     subroutine test_CRS
!         implicit none
!         real(rt) :: x(3),a(3),b(3)
!         x = [.3e0_rt,.1e0_rt,.5e0_rt]
!         a = 0.e0_rt
!         b = [1.2e0_rt,3e0_rt,99e0_rt]
!         call sub_CRS(x,rosenbrock,a,b,optimset('xabstoler=1d-12,printmod=2500,yabstoler=1d-12,madeuparg=200,maxiter=12345'))
!         write(*,*) 'x: ',x
!         write(*,*) 'fx: ',rosenbrock(x)
!     end subroutine 


    subroutine test_gridsearch
        use mod_matlab, only: linspace
        implicit none
        real(rt) :: x(3),grids(40,3)
        integer :: n(3)

        n = (/19,29,05/)
        grids(1:n(1),1) = linspace(0e0_rt,2e0_rt,n(1))
        grids(1:n(2),2) = linspace(0e0_rt,2e0_rt,n(2))
        grids(1:n(3),3) = linspace(0e0_rt,2e0_rt,n(3))
        call sub_gridsearch(x,rosenbrock,n,grids,printmod_in=1000)

        write(*,*) 'x: ',x
        write(*,*) 'fx: ',rosenbrock(x)

    end subroutine

    subroutine test_shrinkingcube
        implicit none
        real(rt) :: x(3)
        x = (/.3e0_rt, .1e0_rt, .5e0_rt /)
        call sub_shrinkingcube(x,rosenbrock)
        write(*,*) 'x: ',x
        write(*,*) 'fx: ',rosenbrock(x)        
    end subroutine 

    subroutine test_parrandom
        implicit none
        real(rt) :: x(3)
        x = (/.3e0_rt, .1e0_rt, .5e0_rt /)
        call sub_parrandom(x,rosenbrock,printmod_in=100)
        write(*,*) 'x: ',x
        write(*,*) 'fx: ',rosenbrock(x)        
    end subroutine 

    subroutine test_newtonNdim
        implicit none
        real(rt) :: x(3)
        x = (/.3e0_rt, .1e0_rt, .5e0_rt /)
        call sub_newtonNdim(x,rosenbrock)
        write(*,*) 'x: ',x
        write(*,*) 'fx: ',rosenbrock(x)        
    end subroutine

!     subroutine test_compass
!         implicit none
!         real(rt) :: x(3)
!         x = (/.3e0_rt, .1e0_rt, .5e0_rt /)
!         call sub_compass(x,rosenbrock,shrink_in=.5e0_rt,printmod_in=50)
!         write(*,*) 'x: ',x
!         write(*,*) 'fx: ',rosenbrock(x)        
!     end subroutine
! 
!     subroutine test_sequentialmin
!         implicit none
!         real(rt) :: x(3)
!         x = (/.3e0_rt, .1e0_rt, .5e0_rt /)
!         call sub_sequentialmin(x,rosenbrock,shrink_in=.5e0_rt,printmod_in=1)
!         write(*,*) 'x: ',x
!         write(*,*) 'fx: ',rosenbrock(x)        
!     end subroutine 

    subroutine test_gradDescent
        implicit none
        real(rt) :: x(3)
        x = (/.3e0_rt, .1e0_rt, .5e0_rt /)
        call sub_gradDescent(x,rosenbrock,maxiter_in=10000,xtoler_in=1e-10_rt,toler_in=1e-10_rt,printmod_in=1000)
        write(*,*) 'x: ',x
        write(*,*) 'fx: ',rosenbrock(x)        
    end subroutine

    subroutine test_stochGradDescent
        implicit none
        real(rt) :: x(3)
        x = (/.3e0_rt, .1e0_rt, .5e0_rt /)
        call sub_stochGradDescent(x,rosenbrock,maxiter_in=10000,xtoler_in=1e-10_rt,printmod_in=1000,lambda_in=.1e0_rt)
        write(*,*) 'x: ',x
        write(*,*) 'fx: ',rosenbrock(x)        
    end subroutine

    subroutine test_BFGS
        implicit none
        real(rt) :: x(3)
        x = (/.3e0_rt, .1e0_rt, .5e0_rt /)
        call sub_BFGS(x,rosenbrock)
        write(*,*) 'x: ',x
        write(*,*) 'fx: ',rosenbrock(x)        
    end subroutine

!     subroutine test_smolyaksearch
!         implicit none
!         real(rt) :: x(3)
!         x = (/.3e0_rt, .1e0_rt, .5e0_rt /)
!         call sub_smolyaksearch(x,rosenbrock,e_in=.1e0_rt*(/10e0_rt,10e0_rt,10e0_rt/),mu_in=2,maxiter_in=1000,xtoler_in=1d-6)
!         write(*,*) 'x: ',x
!         write(*,*) 'fx: ',rosenbrock(x)        
!     end subroutine

    subroutine test_numGradHess
        use mod_deriv
        use mod_matlab
        implicit none
        real(rt) :: x(3), grad(3), hess(3,3), gradT(3), hessT(3,3), err

        hessT = reshape((/2e0_rt,3e0_rt,0e0_rt,3e0_rt,2e0_rt,3e0_rt,0e0_rt,3e0_rt,2e0_rt/),(/3,3/)) ! Not a function of x

        x = 0e0_rt
        gradT = 0e0_rt
        call sub_numGradHess(x,quadratic,(/1e0_rt,1e0_rt,1e0_rt/),grad,hess)
        err = maxval(abs(gradT-grad))
        err = max(err,maxval(abs(hessT-hess)))
!         print*,'Gradient',grad
!         print*,'Hess'
!         call disp(hess)

        x = 1e0_rt
        gradT = (/5e0_rt,8e0_rt,5e0_rt/)
        call sub_numGradHess(x,quadratic,(/1e0_rt,1e0_rt,1e0_rt/),grad,hess)
        err = max(err,maxval(abs(gradT-grad)))
        err = max(err,maxval(abs(hessT-hess)))
!         print*,'Gradient',grad
!         print*,'Hess'
!         call disp(hess)

        if (err>1d-07) then
            print*,'Error was: ',err
            STOP 'test_numGradHess: failed'
        else
            print*,'test_numGradHess: passed'
        end if

    end subroutine

    function quadratic(x) result(f)
        implicit none
        real(rt), dimension(1:), intent(in) :: x
        real(rt) :: f
        !local
        real(rt) :: tmp
        integer :: i

        tmp = sum(x**2)
        do i = 1,size(x)-1
            tmp = tmp + 3e0_rt*x(i)*x(i+1)
        end do
        f = tmp

    end function

    function rosenbrock(x) result(f)
        use mod_util
        implicit none
        real(rt), dimension(1:), intent(in) :: x
        real(rt) :: f
        !local
        integer :: i,n
        real(rt) :: f_tmp, tmp

        call random_number(tmp)
        if (tmp<.0001_rt) then
            tmp = -1._rt
            f = sqrt(tmp)
            return
        end if

        n = size(x)
        f_tmp = 0.e0_rt
        do i = 1,n-1
             f_tmp = f_tmp + 100.e0_rt*(x(i+1)-x(i)**2)**2 + (1.e0_rt-x(i))**2
        end do
        f = f_tmp

!         call sleep(1)

    end function
end module mod_minNdim_test

!     private :: minN_f_template
! 
!     abstract interface
!         function minN_f_template_w_cache(x,f,cache) result(y)
!            use mod_minNdim_w_cache ! only way I could figure out how to declare minN_f_template and minN_f_cache here
!            implicit none
!            real(rt), intent(in) :: x(1:) 
!            procedure(minN_f_template) :: f
!            type(minN_f_cache), intent(inout) :: cache
!            real(rt) :: y
!        end function
!     end interface
! 
! module mod_minNdim_w_cache
! 
!     type minN_f_cache 
!         real(rt), allocatable :: x(:,:)
!         real(rt), allocatable :: y(:)
!         integer :: n,maxn
! !     contains
! !         procedure, pass :: init => init_cache
! !         procedure, pass :: put => put_cache
! !         procedure, pass :: check => check_cache
! !         procedure, pass :: free => empty_cache
!     end type
! 
!     abstract interface
!         function minN_f_template(x) result(y)
!             implicit none
!             real(rt), intent(in) :: x(1:)
!             real(rt) :: y
!         end function 
!     end interface
! 
! contains
! 
!     ! The function g will first search the cache and if not found, will evaluate f
!     function f_without_cache(x,f,cache) result(g)
!         implicit none
!         procedure(minN_f_template) :: f
!         type(minN_f_cache), intent(inout) :: cache ! not used
!         real(rt), intent(in) :: x(1:)
!         real(rt) :: g
! 
!         g = f(x)
! 
!     end function
! 
!     function f_with_cache(x,f,cache) result(g)
!         implicit none
!         procedure(minN_f_template) :: f
!         type(minN_f_cache), intent(inout) :: cache
!         real(rt), intent(in) :: x(1:)
!         real(rt) :: g
!         ! local
!         real(rt) :: h
!         integer :: flag
! 
!         call check_cache(cache,x,h,flag)
! 
!         if (flag == 1) then
!             ! Cached value found
!             g = h
!             return
!         else
!             ! Cached value not found
!             h = f(x)
!             call put_cache(cache,x,h)
!             g = h
!             return
!         end if
! 
!     end function
! 
!     subroutine init_cache(cache,nx,maxn)
!         implicit none
!         type(minN_f_cache), intent(out) :: cache
!         integer, intent(in) :: nx, maxn
! 
!         cache%maxn = maxn
!         cache%n = 0
!         allocate(cache%x(nx,maxn))
!         allocate(cache%y(maxn))
! 
!     end subroutine
!      
!     subroutine free_cache(cache)
!         implicit none
!         type(minN_f_cache), intent(out) :: cache
!         if (allocated(cache%x)) deallocate(cache%x)
!         if (allocated(cache%y)) deallocate(cache%y)
!     end subroutine
! 
!     subroutine put_cache(cache,x,fx)
!         implicit none
!         real(rt), intent(in) :: x(1:),fx
!         type(minN_f_cache), intent(inout) :: cache
! 
!         if (cache%n>=cache%maxn) then
!             print*,'WARNING: cache size exceeded, not caching more values' 
!         else
!             cache%n= cache%n + 1
!             cache%x(:,cache%n) = x
!             cache%y(cache%n) = fx
!         end if
! 
!     end subroutine 
! 
!     subroutine check_cache(cache,x,fx,flag)
!         implicit none
!         type(minN_f_cache), intent(in) :: cache
!         real(rt), intent(in) :: x(1:)
!         real(rt), intent(inout) :: fx ! Overwritten if found in cache
!         integer, intent(out) :: flag
!         ! local
!         integer :: i
! 
!         flag = 0
!         ! Search the cache
!         do i = 1,cache%n
!             if (all(x==cache%x(:,i))) then
!                 flag = 1
!                 fx = cache%y(i)
!                 return
!             end if
!         end do
!         
!     end subroutine 
!     
! end module 
! Written by Grey Gordon based on http://www.tek-tips.com/viewthread.cfm?qid=1627306
real(8) function d1mach(i)
    implicit none
    integer, intent(in) :: i

    select case (i)
        case (1); d1mach = tiny(0d0)
        case (2); d1mach = huge(0d0)
        case (3); d1mach = epsilon(0d0)/2d0
        case (4); d1mach = epsilon(0d0) 
        case (5); d1mach = log10(2d0)
    end select

end function
module mod_stata
    use mod_types
    implicit none


    integer, parameter :: STATA_MAX_VARS = 10000
    real(rt), parameter :: STATA_INVALID = huge(0._rt)/1e1_rt
    real(rt), parameter :: STATA_MISSING = huge(0._rt)/1e2_rt 

    integer, parameter :: STATA_NAME_LEN = 16
    ! Formats for displaying output to terminal (all fixed width with STATA_NAME_LEN)
    character(len=*), parameter :: STATA_F_FMT = '(f16.5)'
    character(len=*), parameter :: STATA_E_FMT = '(e16.7e3)'
    character(len=*), parameter :: STATA_S_FMT = '(A16)'
    character(len=*), parameter :: STATA_I_FMT = '(I16)'
    ! Formats for exporting data (no data loss, no extra spaces for lower memory)
    character(len=*), parameter :: STATA_Eneg_XFMT = '(ES24.16E3)' ! 
    character(len=*), parameter :: STATA_Epos_XFMT = '(ES23.16E3)' ! save a digit
    character(len=*), parameter :: STATA_I_XFMT = '(I10)'

    type stata_wkspace
        real(rt), allocatable :: d(:,:) ! data set (dims not necessarily nobs x ns)
        real(rt), allocatable :: w(:) ! weights (dims not necessarily nobs)

        integer :: ns = 0 ! number of series
        integer :: nobs = 0 ! number of obs

        character(STATA_NAME_LEN) :: snames(STATA_MAX_VARS) ! data series names

        ! Like in actual stata, we can include a return list
        real(rt)              :: e_r2 ! r2
        real(rt), allocatable :: e_b(:) ! regression coefficients
        real(rt), allocatable :: e_V(:,:) ! variance-covariance of regression coefficients

    contains 
        procedure, pass :: free => stata_free
        procedure, pass :: gen => stata_gen_assumedshape
        procedure, pass :: setweight => stata_setweight_assumedsize
        procedure, pass :: varlist => stata_varlist
        procedure, pass :: v => stata_v
        procedure, pass :: i => stata_i
        procedure, pass :: summarize => stata_summarize
        procedure, pass :: summarize_var => stata_summarize_var
        procedure, pass :: stdev => stata_stdev
        procedure, pass :: median => stata_median
        procedure, pass :: pctile => stata_pctile
        procedure, pass :: corr => stata_corr
        procedure, pass :: pwcorr => stata_pwcorr
        procedure, pass :: regress => stata_regress
        procedure, pass :: list => stata_list
        procedure, pass :: hist => stata_hist
        procedure, pass :: scatter => stata_scatter
        procedure, pass :: export => stata_export
    end type

    interface stata_gen
        module procedure stata_gen_assumedshape, stata_gen_assumedsize
    end interface
    interface stata_setweight
        module procedure stata_setweight_assumedshape, stata_setweight_assumedsize
    end interface


    private :: corr_core_stable,corr_core_fast,isvalid

contains

!     subroutine stata_set_obs()
!         implicit none 
! 
!     end subroutine 
! 
!     subroutine stata_set_data_pt(idata,ivar)
!         implicit none
!         integer, intent(in) :: idata
! 
!     end subroutine

    elemental function stata_log(x) result(y)
        implicit none
        real(rt), intent(in) :: x
        real(rt) :: y

        if (x>0._rt .and. x/=STATA_MISSING .and. x/=STATA_INVALID) then
            y = log(x)
        elseif (x==STATA_MISSING) then
            y = STATA_MISSING
        else
            y = STATA_INVALID
        end if

    end function 

    elemental function isvalid(x) result(tf)
        implicit none
        real(rt), intent(in) :: x
        logical :: tf
        tf = x/=STATA_INVALID .and. x/=STATA_MISSING
    end function

    subroutine stata_export(this,filename)
        implicit none
        class(stata_wkspace), intent(in) :: this
        character(len=*), intent(in), optional :: filename
        ! local
        integer :: col,row,uid

        uid = 6
        if (present(filename)) open(newunit=uid,file=filename,action='write',status='replace',err=999)

        ! Header row
        do col = 1,this%ns
            write(uid,'(A)',advance='no') trim(this%snames(col))
            ! write(uid,conv_i_to_s(len(trim(this%snames(col))))//')',advance='no') adjustr(this%snames(col))
            write(uid,'(A1)',advance='no') ','
        end do
        ! Add weight
        write(uid,'(A)') 'weight'

        ! The rest
        do row = 1,this%nobs
            ! Check whether it can be stored as an integer
            call disp_stata_value(uid,this%d(row,1),ishighprec=.true.,dotryint=.true.)
            do col = 2,this%ns
                write(uid,'(A1)',advance='no') ','
                call disp_stata_value(uid,this%d(row,col),ishighprec=.true.,dotryint=.true.)
            end do
            write(uid,'(A1)',advance='no') ','
            call disp_stata_value(uid,this%w(row),ishighprec=.true.,dotryint=.true.)
            write(uid,*)
        end do

        if (present(filename)) close(uid)

        return

999     write(0,*) 'ERROR: could not export ',filename,', encountered error while opening file.'

    end subroutine
    
    ! Lists the data series
    subroutine stata_list(this)
        implicit none
        class(stata_wkspace), intent(inout) :: this
        ! local
        integer :: row,col

        do col = 1,this%ns
            write(*,'(A)',advance='no') adjustr(this%snames(col))
        end do
        write(*,*)

        do row = 1,this%nobs
            do col = 1,this%ns
                call disp_stata_value(6,this%d(row,col))
            end do
            write(*,*)
        end do

    end subroutine

    subroutine stata_scatter(this,name1,name2,filename)
        use mod_plot
        implicit none
        class(stata_wkspace), intent(in) :: this
        character(len=*), intent(in) :: name1,name2,filename
        ! local
        logical, allocatable :: incl(:)
        integer :: uid,i1,i2

        call clear()

        uid = initplot(filename)
        call add_gnuplot_cmd(uid,'set style data points')

        i1 = this%i(name1)
        i2 = this%i(name2)

        allocate(incl(this%nobs))
        incl =     this%d(:,i1)/=STATA_INVALID .and. this%d(:,i1)/=STATA_MISSING &
             .and. this%d(:,i2)/=STATA_INVALID .and. this%d(:,i2)/=STATA_MISSING

        call triplet(pack(this%d(:,i1),mask=incl),&
                     pack(this%d(:,i2),mask=incl),'')

        call xlabel(uid,trim(this%snames(i1)))
        call ylabel(uid,trim(this%snames(i2)))
        call title(uid,'Scatter plot of '//trim(this%snames(i1))//' vs '//trim(this%snames(i2)))
        call drawnow(uid)
        if (.not. disable_plotting) write(*,*) 'Plot written to '//filename
        call clear()

    end subroutine

    ! Adds the data series "x" to the workspace labeling it using "name"
    subroutine stata_gen_assumedsize(this,x,name,nx)
        implicit none
        class(stata_wkspace), intent(inout) :: this
        real(rt), intent(in) :: x(*)
        character(len=*), intent(in) :: name
        integer, intent(in) :: nx
        call stata_gen_assumedshape(this,x(1:nx),name)
    end subroutine
    subroutine stata_gen_assumedshape(this,x,name)
        implicit none
        class(stata_wkspace), intent(inout) :: this
        real(rt), intent(in) :: x(:)
        character(len=*), intent(in) :: name
        ! local
        real(rt), allocatable :: dold(:,:)

        if (.not. allocated(this%d)) then

            this%nobs = size(x)
            this%ns = 1 

            ! Add the series 
            allocate(this%d(this%nobs,1))
            this%d(:,1) = x
            this%snames(1) = name

            if (.not. allocated(this%w)) then
                allocate(this%w(this%nobs))
                this%w = 1._rt/this%nobs
            end if

        else

            ! Check that the size agrees
            if (this%nobs/=size(x)) then
                write(0,*) 'ERROR: stata: gen: while generating ', name,' found size ', size(x), ' was inconsistent with number of observations ',this%nobs
                stop 
            end if

            ! See whether there is room
            if (this%ns==size(this%d,2)) then

                ! No: need to expand the data set

                allocate(dold(this%nobs,this%ns))

                dold = this%d

                deallocate(this%d)

                allocate(this%d(this%nobs,this%ns*2)) ! double the possible number of series so need not keep doing this

                this%d(:,1:this%ns) = dold

            end if

            ! Now for sure there is room
            this%ns = this%ns + 1
            this%d(:,this%ns) = x
            this%snames(this%ns) = adjustl(name)

        end if

    end subroutine

    ! Sets weights to be used in summarizing, etc
    subroutine stata_setweight_assumedsize(this,w,nw)
        implicit none
        class(stata_wkspace), intent(inout) :: this
        real(rt), intent(in) :: w(*)
        integer, intent(in) :: nw
        call stata_setweight_assumedshape(this,w(1:nw))
    end subroutine
    subroutine stata_setweight_assumedshape(this,w)
        implicit none
        class(stata_wkspace), intent(inout) :: this
        real(rt), intent(in) :: w(:)

        if (.not. allocated(this%d)) then
            write(0,*) 'WARNING: use gen to add a series before calling setweight. Weight has not been set.'
            return
        end if 

        if (this%nobs/=size(w)) then
            write(0,*) 'WARNING: weight size ',size(w),' does not agree with nobs ',this%nobs,'. Weight has not been set.'   
            return
        end if

        if (.not. allocated(this%w)) allocate(this%w(this%nobs)) 

        this%w = w

    end subroutine

    ! Produces a copy of the variable "name"
    function stata_v(this,name) result(x)
        implicit none
        class(stata_wkspace), intent(inout) :: this
        character(len=*), intent(in) :: name
        real(rt) :: x(this%nobs) 
        ! local
        integer :: i

        i = stata_i(this,name)

        if (i>0 .and. i<=this%ns) then
            x = this%d(:,i)
        else
            x = STATA_INVALID
        end if

    end function 

    ! Returns the data series index of the variable "name"
    function stata_i(this,name) result(i)
        implicit none
        class(stata_wkspace), intent(in) :: this
        character(len=*), intent(in) :: name 
        integer :: i
        ! local
        character(len=STATA_NAME_LEN) :: namecopy
        integer :: is

        namecopy = ''
        namecopy = name

        do is = 1,this%ns
            if (namecopy==this%snames(is)) then
                i = is
                return
            end if
        end do

        write(0,*) 'WARNING: stata: variable ',namecopy,' not found.'
        write(0,*) 'Existings names are as follows:'
        do is = 1,this%ns
            write(0,*) this%snames(is)
        end do

        i = -1
        return

    end function

    ! varlist: returns a logical vector corresponding to whether the series are selected or not
    ! if names is just length 1 and equal to "*", then all series are returned.
    function stata_varlist(this,namelist) result(tf)
        use mod_util
        implicit none
        class(stata_wkspace), intent(in) :: this
        character(*), intent(in) :: namelist
        logical, dimension(this%ns) :: tf
        ! local
        character(STATA_NAME_LEN), allocatable :: names(:)
        integer :: i, nvars, ind

        nvars = count_delim_arg(namelist,',') 

        allocate(names(nvars))
        do i = 1,nvars
            names(i) = adjustl(get_delim_arg(namelist,',',i))
        end do

        if (size(names)==1 .and. trim(adjustl(names(1)))=="*") then
            tf = .true.
        else
            tf = .false.
            do i = 1,size(names)
                ind = this%i(names(i))
                if (ind>=1 .and. ind<=this%ns) then
                    tf(ind) = .true.
                else
                    write(0,*) 'ERROR: stata_varlist: could not find requested variable '//names(i)//' in workspace'
                    stop
                end if
            end do
        end if

    end function

    subroutine stata_hist(this,name,filename)
        use mod_matlab, only: linspace,plot,bsearch
        implicit none
        class(stata_wkspace), intent(in) :: this
        character(len=*), intent(in) :: name
        character(len=*), intent(in), optional :: filename ! where to save the plot (if desired)
        ! local
        real(rt), allocatable :: cnt(:), edge(:), table(:,:)
        integer :: ivar,iobs,nbins,ibin

        ! Get the index of the variable
        ivar = this%i(name)

        ! Summarize to the max,min
        allocate(table(this%ns,5)) 
        call this%summarize(table)

        ! Compute the mass
        nbins = 100
        allocate(cnt(nbins))
        allocate(edge(nbins+1))

        edge = linspace(table(ivar,4),table(ivar,5),nbins+1)

        cnt = 0d0
        do iobs = 1,this%nobs

            if (this%d(iobs,ivar)==STATA_INVALID .or. this%d(iobs,ivar)==STATA_MISSING) cycle

            call bsearch(ibin,this%d(iobs,ivar),edge)

            if (allocated(this%w)) then
                cnt(ibin) = cnt(ibin) + this%w(iobs)
            else
                cnt(ibin) = cnt(ibin) + 1d0
            end if

        end do

        ! Plot this hist
        if (.not. present(filename)) then
            call plot(edge(1:nbins),cnt,'Histogram of '//name)
        else
            call plot(edge(1:nbins),cnt,'Histogram of '//name,fname_=filename)
        end if

        
    end subroutine 

    function stata_stdev(this,varname) result(stdev)
        implicit none
        class(stata_wkspace), intent(in) :: this
        character(*), intent(in) :: varname
        real(rt) :: stdev
        ! local
        real(rt) :: std

        call stata_summarize_var(this,varname,std=stdev)

    end function 

    ! Summarizes the data giving nobs, mean, stdev, min, and max
    subroutine stata_summarize_var(this,varname,mean,std,min,max,mass)
        implicit none
        class(stata_wkspace), intent(in) :: this
        character(*), intent(in) :: varname
        real(rt), intent(out), optional :: mean,std,min,max,mass
        ! local
        real(rt) :: stats(5),w
        integer :: is,iobs

        stats(1) = 0._rt ! nobs
        stats(2) = 0._rt ! mean
        stats(3) = 0._rt ! std
        stats(4) = huge(0._rt) ! min
        stats(5) = -huge(0._rt) ! max

        is = this%i(varname)
        
        if (is<1 .or. is>this%ns) return

        do iobs = 1,this%nobs

            ! Check if missing or invalid
            if (this%d(iobs,is)==STATA_INVALID .or. this%d(iobs,is)==STATA_MISSING) cycle

            if (allocated(this%w)) then
                w = this%w(iobs)
                if (w==0._rt) cycle
            else
                w = 1._rt
            end if

            ! If here, valid entry
            stats(1) = stats(1) + w
            stats(2) = stats(2) + this%d(iobs,is)*w
            stats(3) = stats(3) + this%d(iobs,is)**2*w
            if (stats(4)<this%d(iobs,is)) stats(4) = this%d(iobs,is) ! min
            if (stats(5)>this%d(iobs,is)) stats(5) = this%d(iobs,is) ! max
            
        end do
        
        ! Now summarize
        if (stats(1)==0._rt) then
            stats(2:5) = STATA_INVALID
        else
            stats(2) = stats(2)/stats(1) ! mean
            stats(3) = stats(3)/stats(1) - stats(2)**2 ! var = Ex^2-(Ex)^2
            if (stats(3)>=-epsilon(0._rt)) stats(3) = sqrt(abs(stats(3))) 
        end if

        if (present(mass)) mass = stats(1)
        if (present(mean)) mean = stats(2)
        if (present(std)) std = stats(3)
        if (present(min)) min = stats(4)
        if (present(max)) max = stats(5)

    end subroutine


    ! Summarizes the data giving nobs, mean, stdev, min, and max
    ! If "table" is present, will place the statistics in order into it.
    ! If "filename" is present, will write the resulting table to file.
    ! If neither is true, the result is written to screen.
    ! Note that each row corresponds to a data series, so table should be nseries x 4
    subroutine stata_summarize(this,table,filename)
        implicit none
        class(stata_wkspace), intent(in) :: this
        real(rt), intent(out), optional :: table(:,:)
        character(len=*), intent(in), optional :: filename
        ! local
        real(rt), allocatable :: tbl(:,:)
        real(rt) :: w
        integer :: is,iobs,col,row,uid

        allocate(tbl(this%ns,5)) 
        tbl(:,1) = 0._rt ! nobs
        tbl(:,2) = 0._rt ! mean
        tbl(:,3) = 0._rt ! std
        tbl(:,4) = huge(0._rt) ! min
        tbl(:,5) = -huge(0._rt) ! max

        do is = 1,this%ns
            do iobs = 1,this%nobs

                ! Check if missing or invalid
                if (this%d(iobs,is)==STATA_INVALID .or. this%d(iobs,is)==STATA_MISSING) then
                    cycle
                end if

                if (allocated(this%w)) then
                    w = this%w(iobs)
                    if (w==0._rt) cycle
                else
                    w = 1._rt
                end if

                ! If here, valid entry
                tbl(is,1) = tbl(is,1) + 1._rt*w
                tbl(is,2) = tbl(is,2) + this%d(iobs,is)*w
                tbl(is,3) = tbl(is,3) + this%d(iobs,is)**2*w
                tbl(is,4) = min(tbl(is,4),this%d(iobs,is))
                tbl(is,5) = max(tbl(is,5),this%d(iobs,is))
                
            end do

            if (tbl(is,1)==0._rt) then
                tbl(is,2:5) = STATA_INVALID
            else
                tbl(is,2) = tbl(is,2)/tbl(is,1) ! mean
                tbl(is,3) = tbl(is,3)/tbl(is,1) - tbl(is,2)**2 ! var = Ex^2-(Ex)^2
                if (tbl(is,3)>=-epsilon(0._rt)) tbl(is,3) = sqrt(abs(tbl(is,3)))
            end if

        end do

        ! Display, save, or output?
        if (present(table)) then
            ! Check dims are correct
            if (size(table,1)/=this%ns .or. size(table,2)/=5) then
                write(0,*) 'WARNING: table is incorrectly sized: ',shape(table),' Should be ',this%ns,5
                table = -1
                return
            end if

            table = tbl
            return

        end if

        uid = 6
        if (present(filename)) open(newunit=uid,file=filename,action='write')

        write(uid,STATA_S_FMT,advance='no') 'variable'; write(uid,'(A1)',advance='no') ','
        write(uid,STATA_S_FMT,advance='no') 'nobs'    ; write(uid,'(A1)',advance='no') ','
        write(uid,STATA_S_FMT,advance='no') 'mean'    ; write(uid,'(A1)',advance='no') ','
        write(uid,STATA_S_FMT,advance='no') 'stdev'   ; write(uid,'(A1)',advance='no') ','
        write(uid,STATA_S_FMT,advance='no') 'min'     ; write(uid,'(A1)',advance='no') ','
        write(uid,STATA_S_FMT,advance='no') 'max'
        write(uid,*)
        do row = 1,this%ns
            write(uid,STATA_S_FMT,advance='no') adjustr(this%snames(row))
            do col = 1,5
                write(uid,'(A1)',advance='no') ','
                call disp_stata_value(uid,tbl(row,col),col==1 .and. .not. allocated(this%w))
            end do
            write(uid,*)
        end do

        if (present(filename)) close(uid)

    end subroutine


    ! Computes the median of a given varlist
    ! Uses an interpolation based formula, interpolating the cdf to get a value 
    function stata_median(this,varname,method) result(med) 
        implicit none
        class(stata_wkspace), intent(inout) :: this
        character(len=*), intent(in) :: varname
        integer, intent(in), optional :: method
        real(rt) :: med
        ! local
        real(rt) :: median(1)

        median = this%pctile(varname,[50d0])
        med = median(1)

    end function 

    ! pctile : computes the percentiles (should be in 0,100) and returns the values corresponding to them
    function stata_pctile(this,varname,pct) result(c)
        use mod_util
        use mod_matlab
        implicit none
        class(stata_wkspace), intent(inout) :: this
        character(*), intent(in) :: varname
        real(rt), intent(in) :: pct(:)
        real(rt) :: c(size(pct))
        ! local
        real(rt), allocatable :: x(:),cdf(:),tmp(:)
        integer, allocatable :: I(:)
        logical, allocatable :: mask(:)
        real(rt) :: mlo
        integer :: n,is,ilo

        is = this%i(varname)

        call assert(is>=1 .and. is<=this%ns,'stata_median: could not find variable '//varname)

!         methodslct = 1
!         if (present(method)) methodslct = method
!         call assert(methodslct>=1 .and. methodslct<=2,'stata_median: selected method must be 1 or 2 but was ',methodslct

        allocate(mask(this%nobs))

        mask = this%d(:,is)/=STATA_MISSING .and. & 
               this%d(:,is)/=STATA_INVALID .and. & 
               this%w>0._rt

        n = count(mask)

        allocate(x(n),cdf(n),I(n),tmp(n))

        x = pack(this%d(:,is),mask)

        if (maxval(this%w,mask)==minval(this%w,mask)) then

            ! If the observations are uniformly weighted, just sort the list and check the middle
            call sort2(tmp,I,x)
            x = tmp

            c = x(min(size(x),max(1,nint(size(x)*pct/100d0))))
            ! if (mod(size(x),2)==0) then
            !     med = .5d0*sum(x(size(x)/2:size(x)/2+1))
            ! else
            !     med = x(size(x)/2+1)
            ! end if

        else

            ! If the observations are not uniformly weighted, need to interpolate cdf
            cdf = pack(this%w,mask)

            ! Now, need to sort
            call sort2(tmp,I,x)
            x = x(I)
            cdf = cdf(I)
            call assert(all(tmp==x),'stata_median: error in sorting')

            ! Now, cumsum to get actual cdf
            tmp = cdf/sum(cdf)
            cdf = cumsum(tmp)

            ! Now, find median (should allow user to select desired formula, but here I am forcing an interpolation formula)
            c = interp1(cdf,x,pct/100d0)
            ! call noexinterp1qlocweight(ilo,mlo,.5d0,cdf)
            ! med = x(ilo)*mlo + x(ilo+1)*(1d0-mlo)

        end if
        
    end function 

    ! Regresses
    ! 
    subroutine stata_regress(this,depvar,indepvarlist,nocons,nodisp)
        use mod_matlab
        implicit none
        class(stata_wkspace), intent(inout) :: this
        character(*), intent(in) :: depvar
        logical, intent(in) :: indepvarlist(:)
        logical, intent(in), optional :: nocons,nodisp
        ! local
        integer :: idep,i,j
        real(rt), allocatable :: y(:),X(:,:),XPX(:,:),XPy(:),w(:)
        real(rt), allocatable :: b(:),b_stderr(:),varcov(:,:),res(:),II(:,:),tab(:,:)
        real(rt) :: s2,RSS,YSS,Rsq
!         real(rt), allocatable :: t(:,:),w(:)
!         logical, allocatable :: mask(:)
!         logical :: totable, toval, toscreen
        character(STATA_NAME_LEN), allocatable :: row(:),col(:)
        integer :: nindep
        logical :: hascons,dodisp

        hascons = .true.
        if (present(nocons)) hascons = .not. nocons

        nindep = count(indepvarlist) + merge(1,0,hascons)

        ! Error chcek
        idep = this%i(depvar)
        if (idep<1 .or. idep>this%ns) then
            write(0,*) 'ERROR: stata_regress: depvar ',depvar,' not found'
            stop
        end if
        if (size(indepvarlist)/=this%ns) stop 'ERROR: stata_regress: indepvarlist has the wrong size' 

!         toscreen = .not. present(table) .and. .not. present(val)
!         totable = present(table)
!         toval = present(val)

        allocate(y(this%nobs))
        allocate(X(this%nobs,nindep))
        allocate(w(this%nobs))
        allocate(XPX(nindep,nindep))
        allocate(XPy(nindep))
        allocate(b(nindep))
        allocate(res(this%nobs))
        allocate(b_stderr(nindep))
        allocate(varcov(nindep,nindep))
        allocate(II(nindep,nindep))
        allocate(tab(nindep,2))
        allocate(row(size(tab,1)))
        allocate(col(size(tab,2)))

        w = this%w
        where (this%d(:,idep)==STATA_INVALID .or. this%d(:,idep)==STATA_MISSING) w = 0._rt
        do i = 1,this%ns
            if (indepvarlist(i)) where(this%d(:,i)==STATA_INVALID .or. this%d(:,i)==STATA_MISSING) w = 0._rt
        end do

        ! Compute conditional
        w = w/sum(w)
!         where (w>0._rt) w = 1._rt

        y = sqrt(w)*this%d(:,idep)
        if (hascons) then
            j = 1
            X(:,1) = sqrt(w)*1._rt
            row(1) = '_cons'
        else
            j = 0
        end if
        do i = 1,this%ns
            if (indepvarlist(i)) then
                j = j+1
                X(:,j) = sqrt(w)*this%d(:,i)
                row(j) = this%snames(i)
            end if
        end do

        ! Compute X'X
        call mygemm(XPX,X,X,this%nobs,nindep,this%nobs,nindep,'T','N')

        ! Compute X'y
        call mygemv(XPy,X,y,this%nobs,nindep,'T')

        ! Compute b = (X'X)^-1*(X'y)
        b = mldivide(XPX,XPy)
        where (abs(b)<10._rt*epsilon(0._rt)) b = 0._rt

        ! Compute the residuals
        res = y-matmul(X,b)

        ! Variance estimate
        call stats_core_stable(res,this%w,var=RSS)
        ! RSS = dot_product(res,res)
        s2 = RSS/(count(w>0._rt)-nindep) 

        ! Variance-covariance is V = s2*(XPX)^{-1} => V * XPX = s2 * I => XPX'*V' = s2*I
        II = eye(nindep)
        II = s2*II
        varcov = transpose(mldivide(transpose(XPX),transpose(II)))
        where (abs(varcov)<10._rt*epsilon(0._rt)) varcov = 0._rt
        b_stderr = sqrt(diag(varcov))

        ! Compute R-squared
        call stats_core_stable(y,this%w,var=YSS)
        ! YSS = sum(this%w*(y-sum(y,mask=this%w>0._rt)/sum(this%w))**2)/sum(this%w)
        Rsq = 1._rt-RSS/YSS

!         call disp('y',y)
!         call disp('X',X)
!         call disp('w',w)
!         call disp('RSS',RSS)
!         call disp('YSS',YSS)
!         call disp('Rsq',Rsq)
!         call disp('s2',s2)
!         call disp('XPX',XPX)
!         call disp('XPy',XPy)
!         call disp('b',b)
!         call disp('varcov',varcov)
!         call disp('b_stderr',b_stderr)

        ! Output values
        this%e_r2 = Rsq
        if (allocated(this%e_b)) deallocate(this%e_b); allocate(this%e_b(nindep)); this%e_b = b
        if (allocated(this%e_V)) deallocate(this%e_V); allocate(this%e_V(nindep,nindep)); this%e_V = varcov

        ! Table
        dodisp = .true.
        if (present(nodisp)) dodisp = .not. nodisp

        if (dodisp) then
            tab(:,1) = b
            tab(:,2) = b_stderr

            col(1) = 'Coef.'
            col(2) = 'Std. err.'
            
            write(6,STATA_S_FMT,advance='no') ' '
            write(6,'(A1)',advance='no') ','
            write(6,STATA_S_FMT,advance='no') 'R-squared'
            write(6,'(A1)',advance='no') ','
            call disp_stata_value(6,Rsq)
            write(6,*) 
            write(6,STATA_S_FMT,advance='no') ' '
            write(6,'(A1)',advance='no') ','
            write(6,STATA_S_FMT,advance='no') 'Number of obs'
            write(6,'(A1)',advance='no') ','
            call disp_stata_value(6,real(count(w>0._rt),rt))
            write(6,*) 
            write(6,*) 
            call stata_table(6,tab,row,col)
        end if

    end subroutine

    
    subroutine stata_table(uid,tab,row,col)
        implicit none
        integer, intent(in) :: uid
        real(rt), intent(in) :: tab(:,:)
        character(STATA_NAME_LEN), intent(in), optional :: row(:),col(:)
        integer :: i,j

        if (present(col)) then
            if (present(row)) call addentry()
            do i = 1,size(col)
                call addentry(s=adjustr(col(i)),suppress=i==size(col))
            end do
            call newline()
        end if

        do i = 1,size(tab,1) 

            if (present(row)) call addentry(s=adjustr(row(i)))
            do j = 1,size(tab,2)
                call addentry(r=tab(i,j),suppress=j==size(tab,2))
            end do
            call newline()

        end do

    contains
        subroutine addentry(r,s,suppress)
            implicit none
            real(rt), intent(in), optional :: r
            character(*), intent(in), optional :: s
            logical, intent(in), optional :: suppress ! whether trailing comma should be suppressed
            if (.not. present(r) .and. .not. present(s)) then
                write(uid,STATA_S_FMT,advance='no') ' '
            elseif (present(r)) then
                call disp_stata_value(uid,r)
            elseif (present(s)) then
                write(uid,STATA_S_FMT,advance='no') s
            end if

            if (present(suppress)) then
                if (suppress) return
            end if

            write(uid,'(A1)',advance='no') ','

        end subroutine 
        subroutine newline()
            implicit none
            write(uid,*) 
        end subroutine
    end subroutine 


    ! Correlates variable 1 and variable 2
    function stata_pwcorr(this,v1,v2) result(c)
        implicit none
        class(stata_wkspace), intent(inout) :: this
        character(*), intent(in) :: v1,v2
        real(rt) :: c 
        ! local
        integer :: i1,i2

        i1 = this%i(v1)
        i2 = this%i(v2)

        c = STATA_INVALID 

        if (i1<1 .or. i1>this%ns) then
            write(0,*) 'WARNING: pwcorr: invalid variable name '//v1
            return
        end if
        if (i2<1 .or. i2>this%ns) then
            write(0,*) 'WARNING: pwcorr: invalid variable name '//v2
            return
        end if
        
        c = corr_core_stable(this%d(:,i1),this%d(:,i2),this%w,this%nobs)

    end function 



    ! Correlates the varlists varlist1 and varlist2, both integer vectors
    ! 
    ! The output is either to screen, to table, or to val.
    ! If to val, then varlist1 and varlist2 must be singletons
    ! 
    subroutine stata_corr(this,varlist1,varlist2,table,val)
        implicit none
        class(stata_wkspace), intent(inout) :: this
        logical, intent(in) :: varlist1(:),varlist2(:)
        real(rt), intent(out), optional :: table(:,:),val
        ! local
        integer, allocatable :: v1(:),v2(:)
        real(rt), allocatable :: t(:,:),w(:)
        logical, allocatable :: mask(:)
        logical :: totable, toval, toscreen
        integer :: nrows,ncols,irow,icol,i,uid,r,c

        ! Error chcek
        if (size(varlist1)/=this%ns) stop 'ERROR: stata_corr: varlist1 has the wrong size' 
        if (size(varlist2)/=this%ns) stop 'ERROR: stata_corr: varlist1 has the wrong size' 

        toscreen = .not. present(table) .and. .not. present(val)
        totable = present(table)
        toval = present(val)

        nrows = count(varlist1)
        ncols = count(varlist2)

        allocate(v1(nrows))
        allocate(v2(ncols))

        irow = 0
        icol = 0
        do i = 1,this%ns
            if (varlist1(i)) then
                irow = irow + 1
                v1(irow) = i
            end if
            if (varlist2(i)) then
                icol = icol + 1
                v2(icol) = i
            end if
        end do

        allocate(mask(this%nobs))
        allocate(w(this%nobs))
        allocate(t(nrows,ncols))

        do irow = 1,nrows
            do icol = 1,ncols

                r = v1(irow)
                c = v2(icol)
                
                mask = this%d(:,r)/=STATA_MISSING .and. this%d(:,r)/=STATA_INVALID .and. &
                       this%d(:,c)/=STATA_MISSING .and. this%d(:,c)/=STATA_INVALID

                where (mask)
                    w = this%w
                elsewhere
                    w = 0._rt
                endwhere
                if (sum(w)>0._rt) w = w / sum(w)

                if (any(mask)) then
                    t(irow,icol) = corr_core_stable(this%d(:,r),this%d(:,c),w,this%nobs)
                    if (t(irow,icol)/=STATA_INVALID .and. t(irow,icol)/=STATA_MISSING) then
                        if (abs(t(irow,icol))>1.0_rt+sqrt(abs(1.0_rt))) write(0,*) 'WARNING: mod_stata: corr: computed correlation found to be invalid: ',t(irow,icol)
                    end if
                else
                    t(irow,icol) = STATA_MISSING
                end if

            end do
        end do

        ! Now have all the correlations, output
        if (toscreen) then
            uid = 6
            ! Header row
            write(uid,STATA_S_FMT,advance='no') 'variable' ! Top left 
            do icol = 1,ncols
                write(uid,'(A1)',advance='no') ','
                c = v2(icol)
                write(uid,STATA_S_FMT,advance='no') adjustr(this%snames(c)) ! varlist2 names
            end do
            write(uid,*)

            ! The meat
            do irow = 1,nrows
                r = v1(irow)
                write(uid,STATA_S_FMT,advance='no') adjustr(this%snames(r)) ! varlist1 names
                do icol = 1,ncols
                    write(uid,'(A1)',advance='no') ','
                    call disp_stata_value(uid,t(irow,icol))
                end do
                write(uid,*)
            end do
        end if
        if (totable) then
            if (any(shape(t)/=shape(table))) stop 'ERROR: stata_corr: table has the wrong dimensions'
            table = t
        end if
        if (toval) then
            if (any(shape(t)/=[1,1])) stop 'ERROR: stata_corr: val present but more than one pair was requested'
            val = t(1,1)
        end if


    end subroutine

    ! corr_core
    ! This 
    function corr_core_fast(x,y,n,mask) result(c)
        implicit none
        real(rt), intent(in) :: x(*),y(*)
        integer, intent(in) :: n
        logical, intent(in), optional :: mask(*)
        real(rt) :: c
        ! local
        real(rt) :: Ex,Ey,Ex2,Ey2,Exy,sdx,sdy
        integer :: i,m

        Ex = 0e0_rt
        Ex2 = 0e0_rt
        Ey = 0e0_rt
        Ey2 = 0e0_rt
        Exy = 0e0_rt
        m = 0
        if (present(mask)) then
            do i = 1,n
                if (mask(i)) then
                    Ex = Ex + x(i)
                    Ey = Ey + y(i)
                    Ex2 = Ex2 + x(i)**2
                    Ey2 = Ey2 + y(i)**2
                    Exy = Exy + x(i)*y(i)
                    m = m + 1
                end if
            end do
        else
            m = n
            do i = 1,n
                Ex = Ex + x(i)
                Ey = Ey + y(i)
                Ex2 = Ex2 + x(i)**2
                Ey2 = Ey2 + y(i)**2
                Exy = Exy + x(i)*y(i)
            end do
        end if

        sdx = sqrt(max(0._rt,m*Ex2 - Ex**2)) ! in case numerical error induces = -epsilon, bound at 0
        sdy = sqrt(max(0._rt,m*Ey2 - Ey**2))

        if (sdx>0e0_rt .and. sdy>0e0_rt) then
            c = (m*Exy-Ex*Ey)/sdx/sdy
        else
            c = STATA_INVALID
        end if

    end function 

    ! Numerically stable version
    subroutine stats_core_stable(x,weight,mean,std,var,min,max,mass)
        implicit none
        real(rt), intent(in) :: x(:),weight(:)
        real(rt), intent(out), optional :: mean,std,var,min,max,mass
        ! local
        real(rt) :: sumw,Ex,stdx,tmp,minx,maxx,varx
        integer :: i,n

        n = size(x)

        sumw = 0._rt
        Ex = 0._rt
        varx = 0._rt
        minx = huge(0._rt)
        maxx = -huge(0._rt)

        ! Compute the mean and sum
        do i = 1,n
            if (weight(i)>0._rt .and. x(i)/=STATA_INVALID .and. x(i)/=STATA_MISSING) then
                sumw = sumw + weight(i)
                Ex = Ex + x(i)*weight(i)
            end if
        end do

        if (sumw<=0._rt) then
            Ex = STATA_INVALID
            varx = STATA_INVALID
            stdx = STATA_INVALID 
            minx = STATA_INVALID
            maxx = STATA_INVALID
        else
            Ex = Ex/sumw

            ! Now compute the stdevs
            do i = 1,n
                if (weight(i)>0._rt .and. x(i)/=STATA_INVALID .and. x(i)/=STATA_MISSING) then
                    varx = varx + (x(i)-Ex)**2*weight(i)
                end if
            end do
            varx = varx/sumw
            stdx = merge(sqrt(varx),0e0_rt,stdx>=0e0_rt) 

        end if

        ! Output requested stats
        if (present(mean)) mean = Ex
        if (present(std)) std = stdx
        if (present(var)) var = varx
        if (present(mass)) mass = sumw
        if (present(min)) min = minx
        if (present(max)) max = maxx

    end subroutine

    ! This is a numerically stable version
    function corr_core_stable(x,y,weight,n) result(c)
        implicit none
        real(rt), intent(in) :: x(*),y(*),weight(*)
        integer, intent(in) :: n
        real(rt) :: c
        ! local
        real(rt) :: Ex,Ey,stdx,stdy,tmp,sumw
        integer :: i,m,nobs
        logical :: tf

        sumw = 0e0_rt
        Ex = 0e0_rt
        Ey = 0e0_rt
        stdx = 0e0_rt
        stdy = 0e0_rt

        ! Compute how much mass there is total
        do i = 1,n
            if (isvalid(x(i)) .and. isvalid(y(i))) sumw = sumw + weight(i)
        end do
         
        if (sumw<=0._rt) then
            c = STATA_INVALID
            return
        end if

        ! Now compute the means
        do i = 1,n
            if (isvalid(x(i)) .and. isvalid(y(i)) .and. weight(i)>0._rt) then
                Ex = Ex + x(i)*weight(i)
                Ey = Ey + y(i)*weight(i)
            end if
        end do
        Ex = Ex/sumw
        Ey = Ey/sumw

        ! Now compute the stdevs
        do i = 1,n
            if (isvalid(x(i)) .and. isvalid(y(i)) .and. weight(i)>0._rt) then
                stdx = stdx + (x(i)-Ex)**2*weight(i)
                stdy = stdy + (y(i)-Ey)**2*weight(i)
            end if
        end do
        stdx = stdx/sumw
        stdy = stdy/sumw
        stdx = merge(sqrt(stdx),0e0_rt,stdx>=0e0_rt) 
        stdy = merge(sqrt(stdy),0e0_rt,stdy>=0e0_rt) 

        ! Check for valid standard deviations
        if (stdx<=epsilon(0e0_rt) .or. stdy==epsilon(0e0_rt)) then
            c = STATA_INVALID
            return
        end if

        ! If still here, compute the correlation
        tmp = 0d0
        do i = 1,n
            if (isvalid(x(i)) .and. isvalid(y(i)) .and. weight(i)>0._rt) then
                tmp = tmp + (x(i)-Ex)/stdx*(y(i)-Ey)/stdy*weight(i)
            end if
        end do
        tmp = tmp/sumw
        c = tmp

    end function 




    ! Frees any memory in the stata workspace
    subroutine stata_free(this)
        implicit none
        class(stata_wkspace), intent(inout) :: this

    end subroutine


    !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    !                 AUXILIARY ROUTINES             !
    ! These are not meant to be called by the user.  !
    !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

    subroutine test_stata()
        real(rt) :: x(8),y(8),table(3,3),tmp
        integer :: i
        type(stata_wkspace) :: w

        x = [2.0_rt,3.0_rt,4.0_rt,5.0_rt,10.0_rt,-1.0_rt,STATA_MISSING,STATA_INVALID]
        where (x /= STATA_MISSING .and. x/=STATA_INVALID) y = x**2 
        y(7) = STATA_MISSING
        y(8) = STATA_INVALID

        call w%gen(x,'x')
        call w%gen(stata_log(x),'ln_x')
        call w%gen(y,'y')
        call w%list()
        write(*,*) 

        i = w%i('x'); if (i/=1) stop 'ERROR: test_stata: failed test 1'
        i = w%i('ln_x'); if (i/=2) stop 'ERROR: test_stata: failed test 2'

        ! i = w%i('asdf'); if (i/=-1) stop 'ERROR: test_stata: failed test 3'

        print*
        print*,'SUMMARIZE'
        print*
        call w%summarize()

        print*
        print*,'MEDIAN'
        print*
        tmp = w%median('x')
        print*,'Median of x: ',tmp
        print*,'Median of ln_x: ',w%median('ln_x')
        print*,'10,25,50,75,90 pctiles of x', w%pctile('x',[10d0,25d0,50d0,75d0,90d0])

        print*
        print*,'CORRELATION'
        print*
        tmp = w%pwcorr('x','ln_x')
        print*,'pwcorr correlation of x,ln_x: ',tmp

        call w%corr(w%varlist('x,ln_x'),w%varlist('*'))
        call w%corr(w%varlist('x'),w%varlist('y'))
        call w%corr(w%varlist('y,x'),w%varlist('ln_x'))


        call w%corr(w%varlist('*'),w%varlist('*'))
        call w%corr(w%varlist('*'),w%varlist('*'),table)

        print*,'table',table
        print*,'Correlation between x and ln_x : ',table(w%i('x'),w%i('ln_x'))

        ! call w%scatter('x','y','/tmp/scatter.eps')

        call w%regress('ln_x',w%varlist('x,y')) 

        ! call w%export('/tmp/statatest.csv')
        ! write(*,*) 'Stata wkspace exported to /tmp/statatest.csv'


    end subroutine

    subroutine disp_stata_value(uid,v,isint,ishighprec,dotryint)
        use mod_util
        implicit none
        integer, intent(in) :: uid
        real(rt), intent(in) :: v
        logical, optional, intent(in) :: isint
        logical, optional, intent(in) :: ishighprec 
        logical, optional, intent(in) :: dotryint ! whether to try to write as in integer
        ! local
        logical :: isint_local,ishighprec_local,dotryint_local

        isint_local = .false.
        ishighprec_local = .false.
        dotryint_local = .false.
        if (present(isint)) isint_local = isint
        if (present(ishighprec)) ishighprec_local = ishighprec
        if (present(dotryint)) dotryint_local = dotryint

        ! If dotryint is specified, then compare the value with an integer to see whether we can save on space
        if (dotryint_local) then
            if (v<huge(1)) then
                if (abs(nint(v)-v)<=epsilon(0d0)*abs(v)) then
                    isint_local = .true.
                end if
            end if
        end if

        if (ishighprec_local) then 
            if (v==STATA_MISSING .or. v==STATA_INVALID) then
                write(uid,'(A1)',advance='no') '.'
            else
                if (.not. isint_local) then
                    if (v<0d0) then
                        write(uid,STATA_Eneg_XFMT,advance='no') v
                    else
                        write(uid,STATA_Epos_XFMT,advance='no') v
                    end if
                else
                    ! write(uid,'(A)',advance='no') trim(conv_i_to_s(nint(v)))
                    select case (nint(v))
                        case (0:9);             write(uid,'(I1)',advance='no') nint(v)
                        case (10:99);           write(uid,'(I2)',advance='no') nint(v)
                        case (100:999);         write(uid,'(I3)',advance='no') nint(v)
                        case (1000:9999);       write(uid,'(I4)',advance='no') nint(v)
                        case (10000:99999);     write(uid,'(I5)',advance='no') nint(v)
                        case (100000:999999);   write(uid,'(I6)',advance='no') nint(v)
                        case (-9:-1);           write(uid,'(I2)',advance='no') nint(v)
                        case (-99:-10);         write(uid,'(I3)',advance='no') nint(v)
                        case (-999:-100);       write(uid,'(I4)',advance='no') nint(v)
                        case (-9999:-1000);     write(uid,'(I5)',advance='no') nint(v)
                        case (-99999:-10000);   write(uid,'(I6)',advance='no') nint(v)
                        case (-999999:-100000); write(uid,'(I7)',advance='no') nint(v)
                        case default; write(uid,'(I10)',advance='no') nint(v)
                    end select
                end if
            end if
        else
            if (v==STATA_MISSING) then
                write(uid,STATA_S_FMT,advance='no') ' .'
            elseif (v==STATA_INVALID) then
                write(uid,STATA_S_FMT,advance='no') ' invalid'
            else
                if (.not. isint_local) then
                    if ((abs(v)>1e9_rt .or. abs(v)<1e-4_rt) .and. abs(v)/=0._rt) then
                        write(uid,STATA_E_FMT,advance='no') v
                    else
                        write(uid,STATA_F_FMT,advance='no') v
                    end if
                else
                    write(uid,STATA_I_FMT,advance='no') nint(v)
                end if
            end if
        end if

    end subroutine

    ! Increases the number of series
!     subroutine stata_expand(this,newnobs,newns)
!         implicit none
!         class(stata_wkspace), intent(inout) :: this
!         integer, intent(in) :: newnobs,newns
!         ! local
!         real(rt), allocatable :: dold
! 
!         if (allocated(this%d)) then
! 
!             ! Check that the new number of observations is equal to current
!             if (newnobs /= this%nobs) then
!                 stop 'ERROR: stata: new number of observations does not agree with current'
!             end if
! 
!             ! Need to create a copy of d
!             allocate(dold(size(d,1),size(d,2)))
!             dold = this%d
!             
!             ! Expand 
!             deallocate(this%d)
!             allocate(this%d(
! 
!         end if
! 
! 
! 
!         if (this%ns==0) then
!             allocate(this%d(
!         else
! 
!         end if
! 
!     end subroutine 

    ! Applies function f on an element by element basis to the variable name 
!     subroutine bsxfun(this,f,name)
!         implicit none
!         class(stata_wkspace), intent(inout) :: this
!         real(rt) function f(x)
!             use mod_types
!             implicit none
!             real(rt), intent(in) :: x
!         end function 
!         character(len=*), intent(in) :: name
! 
!     end subroutine 
    



end module
! Grey Gordon, 2016
! 
! A master/slave helper routine using MPI. The slaves should call "slave(fcn,nx)" and then use the type-bound procedures in type_gang and type_slave 
! to request the slaves compute fcn(y,x) and access the value when done. 
! 
! See "test_mpislave" for a usage example; sorry for the lack of documentation.
!
module mod_mpislave
    use mod_types
    use mod_mympi
    implicit none

!     integer, parameter :: dbglevel = 1

    type type_slave 
        real(rt), pointer :: x(:) => null()
        real(rt), pointer :: y(:) => null()
        integer :: ny ! size of y (inferred automatically)
        integer :: rank
        integer :: requestid
        logical :: busy = .false.
        logical :: ever_assigned = .false.
    contains
        procedure, pass :: assign_x
        procedure, pass :: check_busy
        procedure, pass :: shutdown => shutdown_slave
    end type 

    type type_gang
        type(type_slave), dimension(:), allocatable :: slaves
        integer :: sleeptime = 0
    contains
        procedure, pass :: init => init_gang
        procedure, pass :: shutdown => shutdown_gang
        procedure, pass :: wait_for_free
    end type

contains

    subroutine test_mpislave()
        use mod_util
        implicit none
        type(type_gang) :: gang
        integer :: i,who
        integer, parameter :: nx=2,sleeptime=0
        logical, parameter :: dbg = .false.
        real(rt) :: x(nx),piest
        type(type_seed) :: randstate

        if (mympi_nproc()==1) then
            call mympi_stop('ERROR: test_mpislave: at least 2 ranks are required')
        end if

        call randstate%randomize()

        if (mympi_iproc()/=0) then 
            call slave(fcn,nx)
            return
        end if

        if (dbg) write(*,'(A)',advance='no') 'initializing gang ...'
        call gang%init(nx,sleeptime)
        if (dbg) write(*,'(A)') ' done.'

        piest = 3._rt

        do i = 2,100000

            if (dbg) write(*,'(A)',advance='no') 'waiting for free ...'
            who = gang%wait_for_free()
            if (dbg) write(*,'(A)') ' done.'
            
            if (gang%slaves(who)%ever_assigned) then
                if (dbg) write(*,'(A)',advance='no') 'getting y value ...'
                piest = gang%slaves(who)%y(1)/i + piest*real(i-1,rt)/i
                if (dbg) write(*,'(A)') ' done.'
            end if

            if (dbg) write(*,'(A)',advance='no') 'assigning x...'
            call random_number(x)
            call gang%slaves(who)%assign_x(x)
            if (dbg) write(*,'(A)') ' done.'

        end do

        if (dbg) write(*,'(A)',advance='no') 'shutting down ...'
        call gang%shutdown()
        if (dbg) write(*,'(A)') ' done.'

        if (abs(piest-3.14d0)<.1d0) then
            print*,'test_mpislave: passed'
        else
            stop 'test_mpislave: failed, pi estimate is not close to pi'
        end if

    contains

        subroutine fcn(y,x)
            real(rt), intent(out), allocatable :: y(:)
            real(rt), intent(in) :: x(:)

            ! Determine whether x1,x2 lies insides a circle inscribed in a square, or outside of it. 

            ! Area ratio is pi over 4, so to get estimate of pi, multiply by 4
            allocate(y(1))
            y(1) = merge(4._rt,0._rt,norm2(x)<1._rt) 

        end subroutine

    end subroutine

    ! Do something according to the value of flag(this_image())[1]
    !  1 := evaluate slavex and store in slavey
    ! -1 := exit the slave routine
    subroutine slave(fcn,nx)
        use mod_mympi
        implicit none
        integer, intent(in) :: nx
        interface 
            subroutine fcn(y,x) 
                use mod_types
                implicit none
                real(rt), dimension(:), intent(out), allocatable :: y ! dimension inferred after fcn is called
                real(rt), dimension(:), intent(in) :: x
            end subroutine
        end interface
        ! local
        integer :: flag
        real(rt), allocatable :: slavex(:),slavey(:)

        allocate(slavex(nx))

        do 

            ! Wait for a flag 
            call mympi_recv(flag,0)

            ! If flag==-1, then shutdown
            if (flag==-1) return

            ! If still here, wait for an x
            call mympi_recv(slavex,0)

            ! If flag==1, execute f(x) and store on master.
            if (flag==1) then

                ! Compute the result
                call fcn(slavey,slavex) 

                ! Note that here, slavey should be allocated. If not, abort! 
                if (.not. allocated(slavey)) call mympi_stop('ERROR: mod_mpislave: slave: fcn must allocate slavey, but was not allocated.')

                ! Send the dimension to master tagged by mympi_iproc()
                call mympi_send(size(slavey),0,tag=mympi_iproc()) 

                ! Send the result to master tagged by mympi_nproc()+mympi_iproc()
                ! call mympi_send(slavey,0,tag=mympi_nproc()+mympi_iproc()) 
                call mympi_send(slavey,0) !,tag=mympi_nproc()+mympi_iproc()) 

            end if

        end do

    end subroutine

    ! a routine to shut down all slaves
    subroutine shutdown_slave(this,sleeptime) 
        use mod_util
        implicit none
        class(type_slave), intent(inout) :: this
        integer, intent(in) :: sleeptime
        ! local
        integer :: j
        character(1), volatile :: abc ! volatile prevents optimization (this is used for a wait loop)

        ! Wait for rank to finish if not finished task
        do while (this%busy) 

            ! Check if still busy (if not, then this%busy will be set to .false.
            call this%check_busy()

            ! If still busy, then sleep or do something else for a little while to not thrash the system
            if (this%busy) then

                ! If sleeptime, then sleep
                if (sleeptime>=1) then

                    call sleep(sleeptime)

                else

                    do j = 1,1000000
                        abc = 'a'
                    end do

                end if


            end if

        end do
            
        ! Once not busy, then send the shut down signal of flag==-1
        call mympi_send(-1,this%rank) 

    end subroutine

    subroutine check_busy(this)
        implicit none
        class(type_slave), intent(inout) :: this
        ! local
        logical :: tf

        if (this%busy) then

            ! Check whether i has finished
            tf = mympi_test(this%requestid)

            if (tf) then


                ! Finished sending ny, so need to allocate y and receive it
                if (associated(this%y)) then
                    if (this%ny/=size(this%y)) then
                        deallocate(this%y)
                        allocate(this%y(this%ny))
                        ! Now y has the correct shape
                    end if
                else
                    allocate(this%y(this%ny))
                end if

                ! Here, this%y is allocated and is size this%ny. Receive it.
                call mympi_recv(this%y,this%rank)

                this%busy = .false.

            end if

        end if

    end subroutine

    ! This routine sends the value of x to a slave "who", who will then store the value in y. 
    subroutine assign_x(this,x) !,y)
        implicit none
        class(type_slave), intent(inout) :: this
        real(rt), intent(in), target :: x(:)
        integer :: ierr

        this%busy = .true.
        this%ever_assigned = .true.

        this%x = x

        call mympi_send(1,this%rank) ! blocking send (tell them to proceed to execute)
        call mympi_send(x,this%rank) ! assign an x value (wait for it to complete)

        ! Request that the value of ny be sent and stored in this%ny (once finished, will indicate completed task)
        call MPI_Irecv(this%ny, 1, MPI_INTEGER, this%rank, this%rank, MPI_COMM_WORLD, this%requestid, ierr)

        ! Once the slave has finished evaluating f(x), it's value will be stored in this%y 

    end subroutine 

    subroutine init_gang(this,nx,sleeptime)
        implicit none
        class(type_gang), intent(inout) :: this
        integer, intent(in) :: nx ! size of x
        integer, intent(in), optional :: sleeptime
        ! local
        integer :: i,nranks

        if (allocated(this%slaves)) deallocate(this%slaves)

        if (present(sleeptime)) this%sleeptime = sleeptime

        nranks = mympi_nproc()

        allocate(this%slaves(nranks-1))
        do i = 1,nranks-1
            if (associated(this%slaves(i)%x)) deallocate(this%slaves(i)%x)
            allocate(this%slaves(i)%x(nx))
            this%slaves(i)%rank = i
        end do

    end subroutine

    ! a routine to shut down all slaves
    subroutine shutdown_gang(this) 
        implicit none
        class(type_gang), intent(inout) :: this
        ! local
        integer :: i

        do i = 1,size(this%slaves)
            call this%slaves(i)%shutdown(this%sleeptime)
        end do
        
    end subroutine

    ! a routine returning the index of a free slave
    function wait_for_free(this) result(who)
        use mod_util
        implicit none
        class(type_gang), intent(inout) :: this
        integer :: who
        ! local
        real(8) :: tic, toc, tmp
        integer :: i

        do 

            do i = 1,size(this%slaves)

                ! Test whether a particular slave is busy
                call this%slaves(i)%check_busy()
                if (.not. this%slaves(i)%busy) then
                    who = this%slaves(i)%rank
                    return
                end if

            end do


            ! If sleeptime, then sleep
            if (this%sleeptime>=1) then

                call sleep(this%sleeptime)

            else

                ! Do some busy work so as to not continually check the status (which may cause MPI problems)
                tic = MPI_WTIME()
                toc = tic
                tmp = 0d0
                do while (toc-tic<1d-6)
                    toc = MPI_WTIME()
                    tmp = tmp + i 
                end do
                
                ! Trick the compiler into not optimizing away the busy work (hopefully)
                open(newunit=i,file='/lkandsfkasn/makecompilernotoptimize.dat',status='unknown',err=999) ! file should not exist (but could), so this will jump to line 999
                write(i,*) tmp
                close(i)
999             tmp = 0d0

            end if

        end do

    end function



end module
! Module containing sub_crs, the Fortran routine used by 
! "A Practical Approach to Testing Calibration Strategies" (Yongquan Cao and Grey Gordon)
! See sub_crs for the main documentation of the routine. 
! 
! Grey Gordon, 2016. SDG.
module mod_crs
    use mod_types
    use mod_util
    use mod_mympi
    use mod_mpislave
    implicit none

    ! Options for minNdim 
    type type_crs_opts 
        ! Commonly used ones
        integer :: printmod = 1, maxiter = 10000
        real(rt) :: xabstoler = 1e-6_rt !, xreltoler = 1e-6_rt
        real(rt) :: yabstoler = 1e-10_rt, yreltoler = 1e-10_rt
        ! Others (maybe only a few algorithms use them)
        ! For these, -1 means there is no default specified
        integer :: algorithm = 1
        integer :: poolsize = -1
        integer :: sleeptime = -1 
        integer :: displaybestx = 1 ! 1==means true (the default)
    end type 

    private
    public :: sub_crs, test_crs

contains

    ! Controlled random search --- MPI version
    ! 
    ! CRS attempts to solve min f(x) s.t. x in [a,b] (where all are in Rn)
    ! or simply evaluates f at an arbitrary number of random x draws from [a,b]
    ! 
    ! When minimizing, it first generates a random pool of points (the number 
    ! is controlled by the poolsize option in opts). Then, it tries to improve
    ! this pool by taking the worst point and trying to get rid of it. This is
    ! done by reflecting it through a simplex of points (like in Nelder-Mead).
    ! 
    ! This simplex is chosen by randomly selecting n+1 points (whose function 
    ! values are known) with replacement. The centroid of the simplex is then
    ! used for reflecting the worst point. 
    !
    ! If the reflected point improves on the worst point (and is in a,b), it is
    ! accepted. O/w, the process repeats.
    ! 
    ! This is modified for parallel purposes by attempting to simultaneously get 
    ! rid of several of the worst points simultaneously. 
    !
    ! The optional argument sleeptime is useful when there are only a few ranks. 
    ! It specifies how long the master should sleep if no free workers are found. This
    ! keeps processing cycles free for those ranks actually doing work. If using 
    ! sleeptime>0, the number of ranks should typically be one more than the physical 
    ! number of cores.
    !
    ! If e_in is present, then random draws will not be taken uniformly from [a,b], but rather
    ! from [xstar-e,xstar+e] (component-wise). Combined with a small poolsize (which can be set 
    ! to 0 which will then be overridden), this makes the routine effectively like a parallel 
    ! version of Nelder-Mead.
    !
    ! algorithm
    ! 0           : no optimization : just computes the pool of points and stops. This is used for just testing calibration strategies.  
    ! 1 (default) : crs1 algorithm : purely global with random simplex chosen irrespective of best point and random simplex.
    ! 2           : crs2 algorithm : when random trial point fails, local mutation is used to potentially move towards the best point. 
    ! 
    ! Note: was formerly called sub_parCRS_MPI_v2
    !
    ! INPUT/OUTPUT
    !   xstar : initial guess on input (may not be used depending on algorithm) and final guess on exit
    ! 
    ! INPUT
    !   f       : subroutine f(y,x) with x input and y a vector of outputs. The minimization is done wrt y(1) which will typically be 
    !             the SSRS. However, additional output may be returned which will be saved in a csv file if desired (see "fname" input).
    !   a,b     : lower and upper bound on x space
    !   e_in    : optional, see description above
    !   fname   : optional, gives the path and name where a csv file should be saved (if desired)
    !   xlabels : optional, comma separated list of labels for the x vector (defaults to x1,x2,x3,...)
    !   ylabel  : optional, as above, but for y vector
    !   targets : optional, includes an extra row in the csv file where the x values are listed as . (Stata for missing) and the y values
    !             are listed as targets(1),targets(2),... . The Stata code for "A Practical Approach to Testing Calibration Strategies" 
    !             interprets this row as giving the target moments.
    !   opts    : optional, comma separated list of options. Possible options are xabstoler,yabstoler,yreltoler,printmod,maxiter,poolsize,
    !             sleeptime,displaybestx,algorithm. For instance, opts="printmod=5,poolsize=1000,algorithm=2" will display info every 5th 
    !             time f is computed, use a poolsize of 1000, and use algorithm 2 (described above). 
    ! 
    ! REFERENCE: P. Kaelo and M. M. Ali, "Some variants of the controlled random search algorithm for global optimization," J. Optim. Theory Appl. 130 (2), 253-264 (2006). 
    ! 
    subroutine sub_crs(xstar,f,a,b,e_in,fname,xlabels,ylabels,targets,opts)
        implicit none
        real(rt), intent(inout) :: xstar(:)
        real(rt), intent(in) :: a(:),b(:)
        real(rt), intent(in), optional :: e_in(:), targets(:)
        character(*), intent(in), optional :: fname,xlabels,ylabels,opts
        interface 
            subroutine f(y,x) 
                use mod_types
                implicit none
                real(rt), dimension(:), intent(out), allocatable :: y
                real(rt), dimension(:), intent(in) :: x
            end subroutine
        end interface
        ! local other
        type(type_seed) :: randstate
        real(rt), allocatable :: xtmp(:)
        real(rt), allocatable :: X(:,:),Y(:,:) 
        real(rt), allocatable :: R(:),g(:),smx(:,:)
        real(rt) :: pool_dist
        integer :: nx, N, M, iter, ism, who, itmp(1), flag, effPoolSize, ibest, iworst, smx_iter, uid
        type(type_crs_opts) :: o
        type(type_gang) :: gang

        if (present(opts)) o = optimset(opts)

        if (o%algorithm<0 .or. o%algorithm>2) then
            write(0,*) 'sub_crs: selected algorithm not valid, use 0, 1 or 2; selecting algorithm==1'
            o%algorithm = 1
        end if

        if (mympi_nproc()<=1) then
            write(0,*) 'WARNING: sub_crs: called with only one rank, aborting'
            return
        end if

        nx = size(xstar)
        if (size(a)/=nx) stop 'ERROR: sub_crs: size of a and x do not agree'
        if (size(b)/=nx) stop 'ERROR: sub_crs: size of b and x do not agree'
        if (any(a>=b)) stop 'ERROR: sub_crs: a>=b at least once'

        ! Check that input x is in a,b for all i
        call check_guess_in_range(flag,xstar,a,b)
        if (flag/=0) then
            write(0,*) 'WARNING: sub_crs: x is not in a,b or b<a for some dimension. Will bound and continue'
            xstar = min(b,max(a,xstar))
        end if 

        ! M is the number to simultaneously try to get rid of
        M = mympi_nproc()-1

        ! N is number of trial points (the pool of points)
        ! N = max(2*M, ((100+nx)/M+1)*M, nx+2+M)
        N = max(((100+nx)/M+1)*M, nx+2+M)
        ! N = ((100+nx)/M+1)*M ! roughly evenly divide 100+nx points across the M ranks
        if (o%poolsize/=-1) N = max(nx+2+M,o%poolsize)

        if (N-M<=nx+1) call mympi_stop('ERROR: sub_crs: too few N points set relative to M')

        ! If not master, go to the "slave" routine and wait for a x to evaluate
        if (mympi_iproc()/=0) then
            call slave(f,nx)
            call mympi_bcast(xstar,0) ! receive the best point 
            call mympi_barrier()
            return
        end if

        ! If still here, then you are the master. Create the gang of workers.
        call gang%init(nx,o%sleeptime)

        ! Generate N random points uniformly from [a,b]
        allocate(xtmp(nx))
        allocate(X(nx,N))
        allocate(R(nx))
        allocate(g(nx))
        allocate(smx(nx,nx+1))                    

        ! Randomize for the master
        call randstate%randomize()

        ! Until have reached poolsize, expand
        effPoolSize = 0
        do iter = 1,o%maxiter

            ! Get the best point and the worst point
            if (effPoolSize>1) then
                itmp = minloc(Y(1,1:effPoolSize))
                ibest = itmp(1)
                itmp = maxloc(Y(1,1:effPoolSize))
                iworst = itmp(1)

                ! Report 
                pool_dist = norm2(maxval(X,2)-minval(X,2))
                if (o%printmod>0 .and. mod(iter,o%printmod)==0) then
                    write(*,'(A,i8,3(A,ES14.6),2A)') 'sub_crs: iter ',iter,' pool dist ',pool_dist,' worst ',Y(1,iworst),' best ',Y(1,ibest),'  xbest = ',printx(X(:,ibest))
                    if (effPoolSize==2 .or. mod(iter,100*o%printmod)==0) then
                        call report()
                    end if
                end if

            end if

            ! Check stopping criteria
            if (effPoolSize==N) then
                if (o%algorithm==0) then
                    write(*,'(A)') 'sub_crs: select algorithm was 0 and pool size met, so exiting.'
                    exit
                end if
                if (pool_dist<o%xabstoler) then
                    write(*,'(A,ES14.6,A)') 'sub_crs: xabstoler of ',o%xabstoler,' met.'
                    exit
                end if
                if (abs(Y(1,iworst)-Y(1,ibest))<o%yreltoler*.5*abs(Y(1,iworst)+Y(1,ibest))) then
                    write(*,'(A,ES14.6,A)') 'sub_crs: yreltoler of ',o%yreltoler,' met.'
                    exit
                end if
                if (abs(Y(1,iworst)-Y(1,ibest))<o%yabstoler) then
                    write(*,'(A,ES14.6,A)') 'sub_crs: yabstoler of ',o%yabstoler,' met.'
                    exit
                end if
            end if

            ! Generate a point
            if (effPoolSize<N) then

                ! If pool size is still not large enough, generate a random point.
                ! If local algorithm and the first iteration, include xstar in the pool.
                ! Otherwise, take a draw from uniform distibution over [a,b].
                if (iter==1 .and. o%algorithm==2) then
                    xtmp = xstar
                else
                    call random_number(xtmp)
                    if (.not. present(e_in)) then
                        xtmp = xtmp*a + (1._rt-xtmp)*b
                    else
                        xtmp = xtmp*max(a,xstar-e_in) + (1._rt-xtmp)*min(b,xstar+e_in)
                    end if
                end if

            else

                do smx_iter = 1,100

                    ! If here, at the refinement stage. So take the simplex of points and pick a better one
                    select case (o%algorithm)
                        case (1)
                            ! Choose a truly random simplex
                            do ism = 1,nx+1
                                smx(:,ism) = x(:,randi(1,N)) 
                            end do
                        case (2)
                            do ism = 1,nx+1
                                ! CRS with local mutation where simplex always includes best point
                                if (ism==1) then
                                    smx(:,ism) = x(:,ibest) 
                                else
                                    do 
                                        itmp(1) = randi(1,N) 
                                        if (ibest/=itmp(1) .and. iworst/=itmp(1)) exit
                                    end do
                                    smx(:,ism) = X(:,itmp(1))
                                end if
                            end do
                    end select

                    ! Compute the centroid
                    g = sum(smx,2)/(nx+1)

                    ! Reflect the worst point
                    R = 2._rt*g - X(:,iworst) ! pure reflection

                    ! If reflection point is interior, use it. 
                    ! If this is not satisfied before the max iterations is reached, then R will be used, but bounded
                    if (all(R>=a) .or. any(R<=b)) exit

                end do

                ! If goes through boundary, then bound at boundary
                xtmp = min(max(R,a),b)

            end if

            ! Here, xtmp is the point that should be assigned

            ! Wait for a free worker
            who = gang%wait_for_free()

            ! If ever previously assigned, use that value
            if (gang%slaves(who)%ever_assigned) then

                if (isnan(gang%slaves(who)%y(1))) then
                    write(0,*) 'WARNING: sub_crs: function returned a NaN value for x :',printx(gang%slaves(who)%x),' replacing with huge(0d0)'
                    gang%slaves(who)%y(1) = huge(0d0)
                end if

                if (effPoolSize<N) then

                    ! Add to pool
                    effPoolSize = effPoolSize + 1
                    X(:,effPoolSize) = gang%slaves(who)%x
                    if (.not. allocated(Y)) then

                        ! Now, the size of Y can be inferred, so allocate it.
                        allocate(Y(size(gang%slaves(who)%y),size(X,2)))

                        ! If fname is present, then export the x and y values to file.
                        if (present(fname)) then
                            call opencsv(uid,'replace')
                            call writeheader(uid)
                            close(uid) ! close the file, and reopen it later
                        end if

                    end if
                    Y(:,effPoolSize) = gang%slaves(who)%y

                    if (effPoolSize==N) write(*,'(A)') 'sub_crs: achieved full pool, moving to refinement stage'

                else

                    ! Replace the worst point if found a better one
                    if (gang%slaves(who)%y(1)<Y(1,iworst)) then

                        X(:,iworst) = gang%slaves(who)%x
                        Y(:,iworst) = gang%slaves(who)%y

                    else

                        if (o%algorithm==2) then 
                            ! If CRS2 is the algorithm, then in the case of a failure, move randomly close to the best point
                            ! The new point is a sort of random reflection of the reflection through the best point 
                            call random_number(xtmp) 
                            xtmp = (1._rt+xtmp)*X(:,ibest) - xtmp*gang%slaves(who)%x 
                            xtmp = min(max(xtmp,a),b)
                        end if

                    end if

                end if

                ! Save the value to file
                call appendtocsv(gang%slaves(who)%x,gang%slaves(who)%y)

            end if

            ! Assign them a value
            call gang%slaves(who)%assign_x(xtmp)

        end do

        ! Do a final report
        call report()
        if (iter==o%maxiter+1) &
        write(*,'(A,i8)')     'sub_crs: failed to converge in allotted iterations'
        if (iter<=o%maxiter) &
        write(*,'(A,i8)')     'sub_crs: converged successfully'

        ! Output the best
        xstar = X(:,ibest)

        call gang%shutdown()
        call mympi_bcast(xstar,0) ! receive the best point 
        call mympi_barrier()

    contains
    
        subroutine writeheader(uid)
            implicit none
            integer, intent(in) :: uid
            integer :: ix, iy

            ! Header
            if (present(xlabels)) then
                write(uid,'(A)',advance='no') xlabels ! should be comma separated list
                write(uid,'(A)',advance='no') ','
            else
                do ix = 1,nx
                    write(uid,'(A)',advance='no') 'x'//trim(conv_i_to_s(ix))//','
                end do
            end if

            if (present(ylabels)) then
                write(uid,'(A)',advance='no') ylabels
            else
                do iy = 1,size(Y,1)
                    write(uid,'(A)',advance='no') 'stat'//trim(conv_i_to_s(iy))
                    if (iy<size(Y,1)) write(uid,'(A)',advance='no') ','
                end do
            end if
            write(uid,*) 

            ! Row listing targets (if applicable)
            if (present(targets)) then

                if (size(targets)/=size(Y,1)) then
                    write(0,*) 'WARNING: size(targets) is incorrect, does not equal dimension of function''s output variable. Targets will not be written.'
                    return
                end if

                do ix = 1,nx
                    write(uid,'(A2)',advance='no') '.,' ! missing
                end do
                do iy = 1,size(Y,1)
                    write(uid,'(ES24.16E3)',advance='no') targets(iy)
                    if (iy<size(Y,1)) write(uid,'(A)',advance='no') ','
                end do
                write(uid,*) 

            end if

        end subroutine

        ! Computes the mean and variance-covariance matrix of some data x(j,i) where i gives the observation and j gives the variable
        subroutine sub_varcov(mean,varcov,d)
            implicit none
            real(8), intent(in) :: d(:,:)
            real(8), intent(out) :: mean(:),varcov(:,:)
            ! local 
            integer :: i,j,ni,nj,j2

            nj = size(d,1)
            ni = size(d,2)

            if (size(mean)/=nj) stop 'ERROR: sub_varcov: mean has the wrong size'
            if (any(shape(varcov)/=[nj,nj])) stop 'ERROR: sub_varcov: varcov has the wrong size'

            mean = sum(d,2)/ni

            varcov = 0d0
            do i = 1,size(d,2)
                do j = 1,nj
                    do j2 = j,nj
                        varcov(j,j2) = varcov(j,j2) + (d(j,i)-mean(j))*(d(j2,i)-mean(j2))/ni
                    end do
                end do
            end do
            
            ! Now the variance-covariance upper triangular is set
            ! Do the transpose to fill in the lower triangular
            do j = 1,nj
                do j2 = 1,j-1
                    varcov(j,j2) = varcov(j2,j)
                end do
            end do

        end subroutine

        subroutine report()
            implicit none
            real(8), allocatable :: mu(:),sigma(:,:)
            integer :: i

            allocate(mu(size(a)))
            allocate(sigma(size(a),size(a)))

            call sub_varcov(mu,sigma,X(:,1:effPoolSize))


            write(*,'(2A)')   'sub_crs:  ',repeat('*',120)
            write(*,'(A,i3)') 'sub_crs:            algorithm = ',o%algorithm
            if (present(fname)) &
            write(*,'(2A)')   'sub_crs:     x y history file = ',fname
            if (.not. present(fname)) &
            write(*,'(2A)')   'sub_crs:     x y history file = ','(no history file created) '
            write(*,'(A,i8)') 'sub_crs:                 iter = ',iter 
            write(*,'(A,i8)') 'sub_crs:    current pool size = ',effPoolSize
            write(*,'(A,i8)') 'sub_crs:    desired pool size = ',N
            write(*,'(2A)')   'sub_crs:                    a = ',printx(a)
            write(*,'(2A)')   'sub_crs:                    b = ',printx(b)
            if (present(e_in)) &
            write(*,'(2A)')   'sub_crs:                 e_in = ',printx(e_in)
            write(*,'(2A)')   'sub_crs:   element-wise max y = ',printx(maxval(Y(:,1:effPoolSize),2))
            write(*,'(2A)')   'sub_crs:   element-wise min y = ',printx(minval(Y(:,1:effPoolSize),2))
            write(*,'(2A)')   'sub_crs:   element-wise max x = ',printx(maxval(X(:,1:effPoolSize),2))
            write(*,'(2A)')   'sub_crs:   element-wise min x = ',printx(minval(X(:,1:effPoolSize),2))
            write(*,'(2A)')   'sub_crs:         x pool range = ',printx(maxval(X(:,1:effPoolSize),2)-minval(X(:,1:effPoolSize),2))
            write(*,'(2A)')   'sub_crs:  x range % of space  = ',printx((maxval(X(:,1:effPoolSize),2)-minval(X(:,1:effPoolSize),2))/max(b-a,epsilon(0._rt)))
            write(*,'(2A)')   'sub_crs:          x pool mean = ',printx(mu)
            do i = 1,size(a)
            write(*,'(A,I3,2A)')'sub_crs: x pool varcov(',i,',:) = ',printx(sigma(i,:))
            end do
            write(*,'(2A)')   'sub_crs:                ybest = ',printx(Y(:,ibest))
            write(*,'(2A)')   'sub_crs:                xbest = ',printx(X(:,ibest))
            write(*,'(2A)')   'sub_crs:  ',repeat('*',120)
        end subroutine 

        ! Opens fname for writing, returning the unit id corresponding to it. Catches errors in opening.
        subroutine opencsv(uid,status)
            integer, intent(out) :: uid
            character(len=*), intent(in) :: status
            do 
                ! open(newunit=uid,file=fname,action='write',status=status,position='append',err=1234)
                open(newunit=uid,file=fname,action='write',position='append',err=1234)
                exit ! If no error, escape the loop
1234            write (0,*) 'ERROR: sub_crs: File name '//trim(fname)//' could not be created'
                call mympi_abort()
            end do
        end subroutine 

        ! Appends the x and stats values to the csv file
        subroutine appendtocsv(x,y)
            implicit none
            real(8), intent(in) :: x(:),y(:)
            integer :: i

            if (present(fname)) then
                call opencsv(uid,'old')
                do i = 1,nx
                    write(uid,'(ES24.16E3)',advance='no') x(i)
                    write(uid,'(A)',advance='no') ','
                end do
                do i = 1,size(Y,1)
                    write(uid,'(ES24.16E3)',advance='no') y(i)
                    if (i<size(Y,1)) write(uid,'(A)',advance='no') ','
                end do
                write(uid,*) ! new line
                close(uid)
            end if

        end subroutine


    end subroutine


    ! Generate a random integer in a,...,b
    function randi(a,b) result(i)
        implicit none
        integer, intent(in) :: a,b
        integer :: i
        real(rt) :: tmp

        ! Generate a U[a,b+1-epsilon]
        ! Then floor to get a in 
        call random_number(tmp)
        tmp = tmp*(b+1-a) + a ! U[a,b+1]
        tmp = min(tmp,b+.99999999e0_rt)
        i = floor(tmp)

        if (floor(tmp)<a .or. floor(tmp)>b) STOP 'ERROR: randi: failed to generate i in a,...,b'

    end function

    subroutine rosenbrock(y,x)
        implicit none
        real(rt), dimension(:), intent(in) :: x
        real(rt), dimension(:), intent(out), allocatable :: y
        !local
        integer :: i,n
        real(rt) :: f_tmp, tmp

        allocate(y(1))

        call random_number(tmp)
        if (tmp<.0001_rt) then
            tmp = -1._rt
            y = sqrt(tmp)
            return
        end if

        n = size(x)
        f_tmp = 0.e0_rt
        do i = 1,n-1
             f_tmp = f_tmp + 100.e0_rt*(x(i+1)-x(i)**2)**2 + (1.e0_rt-x(i))**2
        end do
        y = f_tmp

!         call sleep(1)

    end subroutine

    subroutine check_guess_in_range(flagout,x,a,b)
        implicit none
        real(rt), intent(in) :: x(:),a(:),b(:)
        integer, intent(out) :: flagout
        ! local
        integer :: i,n,flag

        n = size(x)
        if (size(a)/=n) stop 'ERROR: check_guess_in_range: a and x do not have same length'
        if (size(b)/=n) stop 'ERROR: check_guess_in_range: b and x do not have same length'

        flag = 0
        do i = 1,n 
        if (x(i)<a(i)) then; write(0,*) 'WARNING: check_guess_in_range: x<a for element',i; flag = merge(-1,-999,flag==0); end if
        if (x(i)>b(i)) then; write(0,*) 'WARNING: check_guess_in_range: x>b for element',i; flag = merge(-2,-999,flag==0); end if
        if (b(i)<a(i)) then; write(0,*) 'WARNING: check_guess_in_range: b<a for element',i; flag = merge(-3,-999,flag==0); end if
        end do
        flagout = flag

    end subroutine 


    subroutine test_crs
        implicit none
        real(rt) :: x(3),a(3),b(3),e(3)

        a = 0.2_rt
        b = [1.2e0_rt,3e0_rt,99e0_rt]
        e = [.01_rt,.01_rt,.01_rt]

        print*
        x = [.3e0_rt,.1e0_rt,.5e0_rt]
        call sub_crs(x,rosenbrock,a,b,opts='printmod=100,sleeptime=0,yabstoler=1d-10,maxiter=100000,algorithm=1')

        print*
        x = [.3e0_rt,.1e0_rt,.5e0_rt]
        call sub_crs(x,rosenbrock,a,b,opts='printmod=1000,sleeptime=0,yabstoler=1d-10,maxiter=100000,algorithm=2')

        print*
        x = [.3e0_rt,.1e0_rt,.5e0_rt]
        call sub_crs(x,rosenbrock,a,b,e_in=e,opts='printmod=1000,sleeptime=0,yabstoler=1d-10,maxiter=100000,poolsize=15')

        print*
        call sub_crs(x,rosenbrock,a,b,fname='/tmp/statmat.csv',xlabels='param1,param2,param3',ylabels='SSRS',opts='printmod=1000,sleeptime=0,yabstoler=1d-10,maxiter=100000,poolsize=2')

    end subroutine 

    ! Here, options are specified via a string such as "xtoler=1d6,printmod=10000"
    function optimset(str) result(opts)
        implicit none
        character(len=*), intent(in) :: str
        type(type_crs_opts) :: opts
        ! local
        character(len=len(str)) :: s,arg,val
        type(type_crs_opts) :: def
        integer :: nargs,i

        nargs = count_delim_arg(str,",")

        do i = 1,nargs

            ! Get the substring
            s = ''
            s = trim(get_delim_arg(str,",",i))

            ! Process to get what is left and right of the equal sign
            arg = ''
            val = ''
            arg = get_delim_arg(s,"=",1)
            val = get_delim_arg(s,"=",2)

            if     (trim(arg)=="xabstoler")     then; read(val,'(d30.16)',err=999) opts%xabstoler 
!             elseif (trim(arg)=="xreltoler")     then; read(val,'(d30.16)',err=999) opts%xreltoler 
            elseif (trim(arg)=="yabstoler")     then; read(val,'(d30.16)',err=999) opts%yabstoler 
            elseif (trim(arg)=="yreltoler")     then; read(val,'(d30.16)',err=999) opts%yreltoler 
            elseif (trim(arg)=="printmod")      then; read(val,'(i25)',   err=999) opts%printmod
            elseif (trim(arg)=="maxiter")       then; read(val,'(i25)',   err=999) opts%maxiter
            elseif (trim(arg)=="poolsize")      then; read(val,'(i25)',   err=999) opts%poolsize
            elseif (trim(arg)=="sleeptime")     then; read(val,'(i25)',   err=999) opts%sleeptime
            elseif (trim(arg)=="displaybestx")  then; read(val,'(i25)',   err=999) opts%displaybestx
            elseif (trim(arg)=="algorithm")     then; read(val,'(i25)',   err=999) opts%algorithm
            else
                write(0,*) 'WARNING: optimset: did not find option corresponding to argument ',&
                           trim(arg),'  with value ',trim(val),' and overall input ',trim(str)
            end if

        end do

        return

999     write(0,*) 'ERROR: optimset: error encountered processing ',trim(str),', with arg ',trim(arg),' and val ',trim(val),' falling back to default arguments'
        opts = def
    

    end function



end module
