! Grey Gordon, 2014-2017. Soli Deo Gloria.
! 
! This has several key routines 
!
! sub_bmono_disc: solves a univariate problem with discrete states and choices
! sub_bmono_cont: solves a univariate problem with discrete states and a continous choice space
! sub_bmono_2s_disc: solves a problem with two state variables and one choice all discrete
! 
module mod_binarymono
    implicit none
contains

    ! Solves V(i) = max_{i' in 1,...n'} f(i,i') with policy function g(i)
    ! 
    ! monotone | monotonicity speedup
    !        0 | brute force (do not assume monotone)
    !        1 | simple monotonicity 
    !        2 | binary monotonicity
    ! 
    !  concave | concavity speedup 
    !        0 | brute force (do not assume concave)
    !        1 | simple concavity
    !        2 | binary concavity
    !
    subroutine sub_bmono_disc(V,g,f,np,monotone,concave,evalcnt)
        implicit none
        interface
            function f(i,ip) 
                implicit none
                integer, intent(in) :: i,ip 
                real(8) :: f
            end function 
        end interface
        real(8), intent(out) :: V(:)
        integer, intent(out) :: g(:)
        integer, intent(in) :: np, monotone, concave
        integer, intent(out), optional :: evalcnt
        ! local
        integer :: n,evalcnt_
        
        n = size(V)

        evalcnt_ = 0

        select case (monotone)
            case (0)
                call brutemono()
            case (1)
                call simplemono()
            case (2)
                ! call binarymono_rec() ! recursive version
                call binarymono_nonrec() ! non-recursive version
            case default
                stop 'ERROR: sub_bmono_disc: option <monotone> must be 0, 1, or 2'
        end select

        if (present(evalcnt)) evalcnt = evalcnt_

    contains

        subroutine brutemono()
            implicit none
            ! local
            integer :: i

            do i = 1,n
                call maxstep(i,1,np)
            end do

        end subroutine 
        subroutine simplemono()
            implicit none
            ! local
            integer :: i

            call maxstep(1,1,np)
            do i = 2,n
                call maxstep(i,g(i-1),np)
            end do

        end subroutine 

        subroutine binarymono_rec()
            implicit none
            call maxstep(1,1,np)
            call maxstep(n,g(1),np)
            call binarymono_rec_work(1,n)

        end subroutine 
        recursive subroutine binarymono_rec_work(ilo,ihi)
            implicit none
            integer, intent(in) :: ilo,ihi
            ! local
            integer :: m

            m = (ilo+ihi)/2
            call maxstep(m,g(ilo),g(ihi))

            if (m>ilo+1) call binarymono_rec_work(ilo,m)
            if (ihi>m+1) call binarymono_rec_work(m,ihi)

        end subroutine 

        subroutine binarymono_nonrec()
            implicit none
            integer :: k !,ind
            integer, allocatable :: l(:),u(:)

            allocate(l(1+ceiling(log(real(n-1,8))/log(2d0))))
            allocate(u(1+ceiling(log(real(n-1,8))/log(2d0))))

            call maxstep(1,1,np)
            call maxstep(n,g(1),np)

            l(1) = 1
            u(1) = n

            k = 1
            
            ! Main loop (step 1 completed)
            outer: do

                ! Expand list (step 2)
                do 
                
                    ! Check stopping criterion for list expansion
                    if (u(k)==l(k)+1) exit

                    k = k + 1

                    l(k) = l(k-1)
                    u(k) = (l(k-1)+u(k-1))/2

                    ! Compute g for u(k) searching from g(l(k-1)) to g(u(k-1)) 
                    call maxstep(u(k),g(l(k-1)),g(u(k-1)))

                end do

                ! Check whether should move to a higher interval (step 3) or stop the algorithm
                do 

                    if (k==1) exit outer

                    if (u(k)/=u(k-1)) exit

                    k = k - 1

                end do

                ! Move to the right subinterval (step 4)
                l(k) = u(k)
                u(k) = u(k-1)

            end do outer


        end subroutine

        ! Solves for g(i) restricting the search space to i' \in \{a,...,b\}
        subroutine maxstep(i,a,b)
            implicit none
            integer, intent(in) :: i,a,b

            select case (concave)
                case (0)
                    call bruteconc(i,a,b)
                case (1)
                    call simpleconc(i,a,b)
                case (2)
                    call binaryconc(i,a,b)
                case default
                    stop 'ERROR: sub_bmono_disc: option <concave> must be either 0, 1, or 2'
            end select

        end subroutine

        subroutine binaryconc(i,a_,b_)
            implicit none
            ! local
            integer, intent(in) :: i,a_,b_
            integer :: a,b,m !,n
            real(8) :: fa,fb,fm,fmp
            logical :: faknown,fbknown

            a = a_
            b = b_

            ! Handle degenerate case
            if (a==b) then
                g(i) = a
                V(i) = f(i,a)
                evalcnt_ = evalcnt_ + 1
                return
            end if

            faknown = .false.
            fbknown = .false.

            do 

                ! "m" is used both as "n", the cardinality of {a,...,b} and as the midpoint of a,b
                m = b-a+1

                if (m==2) then

                    ! Compute values if not known
                    if (.not. faknown) then
                        fa = f(i,a)
                        evalcnt_ = evalcnt_ + 1
                    end if
                    if (.not. fbknown) then
                        fb = f(i,b)
                        evalcnt_ = evalcnt_ + 1
                    end if


                    ! Compute best of a,b
                    if (fa>=fb) then
                        g(i) = a
                        V(i) = fa
                    else
                        g(i) = b
                        V(i) = fb
                    end if

                    return
                    
                elseif (m==3) then

                    m = (a+b)/2
                    fm = f(i,m) 
                    evalcnt_ = evalcnt_ + 1

                    if (.not. (faknown .or. fbknown)) then
                        fa = f(i,a)
                        evalcnt_ = evalcnt_ + 1
                        faknown = .true.
                    end if

                    if (faknown) then
                        if (fa>fm) then
                            V(i) = fa
                            g(i) = a
                            return
                        else
                            a = m
                            fa = fm
                            faknown = .true.
                        end if
                    else ! fbknown --- must be true here
                        if (fb>fm) then
                            V(i) = fb
                            g(i) = b
                            return
                        else
                            b = m
                            fb = fm
                            fbknown = .true.
                        end if
                    end if

                else

                    m = (a+b)/2
                    fm = f(i,m)
                    fmp = f(i,m+1)
                    evalcnt_ = evalcnt_ + 2

                    if (fm>=fmp) then
                        b = m
                        fb = fm
                        fbknown = .true.
                    else
                        a = m+1
                        fa = fmp
                        faknown = .true.
                    end if

                end if

            end do

        end subroutine

        subroutine bruteconc(i,a,b)
            implicit none
            ! local
            integer, intent(in) :: i,a,b
            integer :: j 
            real(8) :: tmp

            V(i) = f(i,a)
            evalcnt_ = evalcnt_ + 1
            g(i) = a
            do j = a+1,b
                tmp = f(i,j)
                evalcnt_ = evalcnt_ + 1
                if (tmp>V(i)) then
                    V(i) = tmp
                    g(i) = j
                end if
            end do

        end subroutine

        subroutine simpleconc(i,a,b)
            implicit none
            ! local
            integer, intent(in) :: i,a,b
            integer :: j 
            real(8) :: tmp

            V(i) = f(i,a)
            evalcnt_ = evalcnt_ + 1
            g(i) = a
            do j = a+1,b
                tmp = f(i,j)
                evalcnt_ = evalcnt_ + 1
                if (tmp>=V(i)) then
                    V(i) = tmp
                    g(i) = j
                else
                    return
                end if
            end do

        end subroutine

    end subroutine


    ! Solves V(i,j) = max_{i' in 1,...n'} f(i,j,i') with policy function g(i,j) monotone in both arguments.
    ! 
    ! monotone | monotonicity speedup
    !        0 | brute force (do not assume monotone)
    !        2 | binary monotonicity
    ! 
    !  concave | concavity speedup 
    !        0 | brute force (do not assume concave)
    !        1 | simple concavity
    !        2 | binary concavity
    !
    ! transpose | if true, will transpose the algorithm, exploiting monotonicity in i before j. Note there is no point in doing that for brute force.
    subroutine sub_bmono_2s_disc(V,g,f,np,monotone,concave,dotranspose)
        implicit none
        interface
            function f(i,j,ip) 
                implicit none
                integer, intent(in) :: i,j,ip 
                real(8) :: f
            end function 
        end interface
        real(8), intent(out) :: V(:,:)
        integer, intent(out) :: g(:,:)
        integer, intent(in) :: np,monotone,concave
        logical, intent(in), optional :: dotranspose
        ! local
        real(8), allocatable :: Vtran(:,:)
        integer, allocatable :: gtran(:,:)
        logical :: tran

        tran = .false.
        if (present(dotranspose)) tran = dotranspose

        if (dotranspose) then
            allocate(Vtran(size(V,2),size(V,1)))
            allocate(gtran(size(g,2),size(g,1)))
            call solvemax_2state_work(Vtran,gtran,ftran,np,monotone,concave)
            V = transpose(Vtran)
            g = transpose(gtran)
        else
            ! Don't do any transpose, just call solvemax_2state_work
            call solvemax_2state_work(V,g,f,np,monotone,concave)
        end if

    contains
    
        ! Transposed function
        function ftran(j,i,ip) 
            implicit none
            integer, intent(in) :: i,j,ip 
            real(8) :: ftran
            ftran = f(i,j,ip)
        end function 

    end subroutine
    subroutine solvemax_2state_work(V,g,f,np,monotone,concave)
        implicit none
        interface
            function f(i,j,ip) 
                implicit none
                integer, intent(in) :: i,j,ip 
                real(8) :: f
            end function 
        end interface
        real(8), intent(out) :: V(:,:)
        integer, intent(out) :: g(:,:)
        integer, intent(in) :: np,monotone,concave
        ! local
        integer :: n1,n2,JSLICE,glo,ghi,ilo,ihi
        
        n1 = size(V,1)
        n2 = size(V,2)

        if (monotone==1) stop 'ERROR: solvemax_2state: simple monotonicity is not implemented for the 2-stage problem'

        if (monotone==0) then
            do JSLICE = 1,n2
                call sub_bmono_disc(V(:,JSLICE),g(:,JSLICE),fslice,np,monotone,concave)
            end do
            return
        end if

        ! If still here, using binary monotonicity

        ! Step 1 : solve for g(.,1) using binary monotonicity
        JSLICE = 1
        call sub_bmono_disc(V(:,1),g(:,1),fslice,np,monotone,concave)

        ! Step 2 : solve for g(.,jhi)
        associate (l=>g(:,1))

            ! Compute g(.,n2)
            ! Step 2(a)
            call maxstep(1,n2,l(1),np)
            ilo = 1
            glo = g(1,n2)
            ! Step 2(b)
            call maxstep(n1,n2,max(g(1,n2),l(n1)),np)
            ihi = n1
            ghi = g(n1,n2)
            ! Step 2(c)
            call rec_binmono_fixedj(ilo,ihi,glo,ghi,n2,1,-1)

        end associate

! stop 'here'

        ! Step 3 is also executed by the below routine
        call binmono_fixedj(1,n2) 



    contains

        real(8) function fslice(i,ip) 
            implicit none
            integer, intent(in) :: i,ip
            fslice = f(i,JSLICE,ip) 
        end function 

        ! Solve for g(.,.) recursively taking g(.,jlo) and g(.,jhi) as known
        recursive subroutine binmono_fixedj(jlo,jhi)
            implicit none
            integer, intent(in) :: jlo,jhi
            ! local
            integer :: m2,glo,ilo,ghi,ihi

            ! Check stopping criterion
            if (jhi==jlo+1) return

            ! If still here, compute midpoint and g(.,m2) unless jhi==-1 meaning it has not been computed yet

            associate (l=>g(:,jlo),u=>g(:,jhi))

                ! Step 3(b) (already checked stopping criterion)
                m2 = (jlo+jhi)/2

                ! Step 3(b)(i) : solve for g(1,m2)
                call maxstep(1,m2,l(1),u(1))
                glo = g(1,m2)
                ilo = 1

                ! Step 3(b)(ii) : solve for g(n1,m2)
                call maxstep(n1,m2,max(glo,l(n1)),u(n1))
                ghi = g(n1,m2)
                ihi = n1

            end associate

            ! Step 3(b)(iii) : solve for g(.,m2) taking g(1,m2) and g(n1,m2) as known
            call rec_binmono_fixedj(ilo,ihi,glo,ghi,m2,jlo,jhi)

            ! Now, at step 3(c), recurse twice 
            call binmono_fixedj(jlo,m2)
            call binmono_fixedj(m2,jhi)

        end subroutine 

        ! Solve for g(i,j) for all i in ilo,...,ihi assuming g(.,jlo) and g(.,jhi) are known.
        recursive subroutine rec_binmono_fixedj(ilo,ihi,glo,ghi,j,jlo,jhi)
            implicit none
            integer, intent(in) :: ilo,ihi,glo,ghi,j,jlo,jhi
            ! local
            integer :: m1

! print*,'ilo,ihi,j,jlo,jhi',ilo,ihi,j,jlo,jhi
            ! Solve for g(i,j) for all i in ilo,ihi
!             if (ihi==ilo+1) return


            ! Solve for the maximium at the midpoint
            m1 = (ilo + ihi)/2 ! automatically rounds down
            if (jhi/=-1) then
                call maxstep(m1,j,max(glo,g(m1,jlo)),min(ghi,g(m1,jhi))) ! l(.) = g(.,jlo) and u(.) = g(.,jho)
            else
                call maxstep(m1,j,max(glo,g(m1,jlo)),ghi) ! here u(.) = n' so ghi is always less that it
            end if

            ! Recurse (check stopping criterion here to prevent unnecessary stack increases)
            if (m1/=ilo+1) call rec_binmono_fixedj(ilo,m1,glo,g(m1,j),j,jlo,jhi)
            if (ihi/=m1+1) call rec_binmono_fixedj(m1,ihi,g(m1,j),ghi,j,jlo,jhi)

        end subroutine 

        subroutine maxstep(i,j,a,b)
            implicit none
            integer, intent(in) :: i,j,a,b

            if (concave==0) then
                call bruteconc(i,j,a,b)
            elseif (concave==1) then
                call simpleconc(i,j,a,b)
            elseif (concave==2) then
                call binaryconc(i,j,a,b)
            end if

        end subroutine 

        subroutine bruteconc(i,j,a,b)
            implicit none
            ! local
            integer, intent(in) :: i,j,a,b
            integer :: ip
            real(8) :: tmp

            V(i,j) = f(i,j,a)
            g(i,j) = a
            do ip = a+1,b
                tmp = f(i,j,ip)
                if (tmp>V(i,j)) then
                    V(i,j) = tmp
                    g(i,j) = ip
                end if
            end do

        end subroutine

        subroutine simpleconc(i,j,a,b)
            implicit none
            ! local
            integer, intent(in) :: i,j,a,b
            integer :: ip 
            real(8) :: tmp

            V(i,j) = f(i,j,a)
            g(i,j) = a
            do ip = a+1,b
                tmp = f(i,j,ip)
                if (tmp>=V(i,j)) then
                    V(i,j) = tmp
                    g(i,j) = ip
                else
                    return
                end if
            end do

        end subroutine

        subroutine binaryconc(i,j,a_,b_)
            implicit none
            ! local
            integer, intent(in) :: i,j,a_,b_
            integer :: a,b,m !,n
            real(8) :: fa,fb,fm,fmp
            logical :: faknown,fbknown

            a = a_
            b = b_

            ! Handle degenerate case
            if (a==b) then
                g(i,j) = a
                V(i,j) = f(i,j,a)
                return
            end if

            faknown = .false.
            fbknown = .false.

            do 

                ! "m" is used both as "n", the cardinality of {a,...,b} and as the midpoint of a,b
                m = b-a+1

                if (m==2) then

                    ! Compute values if not known
                    if (.not. faknown) fa = f(i,j,a)
                    if (.not. fbknown) fb = f(i,j,b)

                    ! Compute best of a,b
                    if (fa>=fb) then
                        g(i,j) = a
                        V(i,j) = fa
                    else
                        g(i,j) = b
                        V(i,j) = fb
                    end if

                    return
                    
                elseif (m==3) then

                    m = (a+b)/2
                    fm = f(i,j,m) 

                    if (.not. (faknown .or. fbknown)) then
                        fa = f(i,j,a)
                        faknown = .true.
                    end if

                    if (faknown) then
                        if (fa>fm) then
                            V(i,j) = fa
                            g(i,j) = a
                            return
                        else
                            a = m
                            fa = fm
                            faknown = .true.
                        end if
                    else ! fbknown --- must be true here
                        if (fb>fm) then
                            V(i,j) = fb
                            g(i,j) = b
                            return
                        else
                            b = m
                            fb = fm
                            fbknown = .true.
                        end if
                    end if

                else

                    m = (a+b)/2
                    fm = f(i,j,m)
                    fmp = f(i,j,m+1)

                    if (fm>=fmp) then
                        b = m
                        fb = fm
                        fbknown = .true.
                    else
                        a = m+1
                        fa = fmp
                        faknown = .true.
                    end if

                end if

            end do

        end subroutine



        
    end subroutine 

    ! Solves V(b) = max_{b'} u(z(b) - x(b')) + W(b') using interpolation of W and x
    ! Note: there need not be any relation between b and b'. 
    ! 
    ! monotone | monotonicity speedup
    !        0 | brute force (do not assume monotone)
    !        1 | simple monotonicity 
    !        2 | binary monotonicity
    ! 
    !  concave | concavity speedup 
    !        0 | brute force (do not assume concave)
    !        1 | simple concavity
    !        2 | binary concavity
    ! 
    ! Inputs
    ! B       : (nB) exogenous grid on which the value and policy function will be recovered. 
    ! z       : (nB) gives z(b) for each b in B (often, this may just == B). 
    ! Bp      : (nBp) interpolation knots for choice space, also giving the boundaries of the search space 
    ! W       : (nBp) continuation utility at exogenous Bp grid
    ! x       : (nBp) expenditure function (little w) at exogenous Bp grid
    ! monotone : (scalar) see above
    ! concave : (scalar) see above
    ! iscontinuous : (scalar) if .true., then continuous optimization is done. if .false., then only grid search.
    ! crra    : (scalar) gives utility function u(c) = c**(1-crra)/(1-crra) (note that special cases of crra=2 and 1 are coded for speed)
    ! [dowarn] : (scalar, optional) if present and false, warnings will be displayed to screen when appropriate. 
    ! [domaxexact] : (scalar, optional) if present and false, an approximate maximization routine will be used, using global grid search followed by local C0 search
    ! [evalcnt] : (scalar, optional) Gives the number of times the utility function is evaluated
    !
    ! Outputs
    ! V       : (nB) value function at exogenous grid B
    ! c       : (nB) consumption policy function at exogenous grid B, c = z - x
    ! g       : (nB) "asset" policy function at exogenous grid B (continuous choice, not necessarily in Bp)
    ! 
    ! Note: the only reason both B and z are present is because whenever interpolation is required, interpolation
    ! is done with respect to b rather than z.
    subroutine sub_bmono_cont(V,c,g,B,z,Bp,W,x,monotone,concave,iscontinuous,crra,dowarn,domaxexact,evalcnt)
        implicit none
        real(8), intent(out) :: V(:),c(:),g(:)
        real(8), intent(in) :: B(:),z(:),Bp(:),W(:),x(:),crra
        integer, intent(in) :: monotone,concave
        logical, intent(in) :: iscontinuous
        logical, intent(in), optional :: dowarn,domaxexact
        integer, intent(out), optional :: evalcnt
        ! local
        real(8), allocatable :: slope_xinv(:),duinv_slope_Wx(:)
        integer, allocatable :: g_ilo(:) !, gi(:)
        logical, allocatable :: g_ongrid(:)
        integer :: nB,nBp
        real(8), parameter :: death = -sqrt(huge(0d0))
        real(8), parameter :: DUINV_INVALID = sqrt(huge(0d0))*.123456789d0
        integer :: evalcnt_
        logical :: printwarn,domaxexact_


        nB = size(B)
        nBp = size(Bp)
        
        ! Check inputs
        if (monotone<0 .or. monotone>2) stop 'ERROR: sub_bmono_cont: monotone must be in {0,1,2}'
        if (concave<0 .or. concave>2) stop 'ERROR: sub_bmono_cont: concave must be in {0,1,2}'
        if (size(z)/=nB) stop 'ERROR: sub_bmono_cont: B and z don''t conform'
        if (size(W)/=nBp) stop 'ERROR: sub_bmono_cont: Bp and W don''t conform'
        if (size(x)/=nBp) stop 'ERROR: sub_bmono_cont: Bp and W don''t conform'
        if (size(V)/=nB) stop 'ERROR: sub_bmono_cont: B and V don''t conform'
        if (size(c)/=nB) stop 'ERROR: sub_bmono_cont: B and c don''t conform'
        if (size(g)/=nB) stop 'ERROR: sub_bmono_cont: B and g don''t conform'

        printwarn   = .true.; if (present(dowarn)) printwarn = dowarn
        domaxexact_ = .true.; if (present(domaxexact)) domaxexact_ = domaxexact
        evalcnt_ = 0

        allocate(g_ilo(nB))
        allocate(g_ongrid(nB))

        if (concave>0) write(0,*) 'WARNING: sub_bmono_cont: concavity is not exploited in this version of the code.'

        ! Precompute some functions that only depend on b' 
        if (iscontinuous) call precompute()

        select case (monotone)
            case (0) 
                call nomono()
            case (1)
                call simple()
            case (2)
                call maximize(1,-1,-1) ! solve at left end point
                call maximize(nB,1,-1) ! solve at right end point
                call binary(1,nB) ! main recursion
        end select

        ! If assuming monotonicity, check that the implied policy and value functions are monotone, warn if not
        if (present(evalcnt)) evalcnt = evalcnt_

    contains

        logical function isascend(x,strict)
            implicit none
            real(8), intent(in) :: x(:)
            logical, intent(in) :: strict
            integer :: i
            isascend = .true.
            do i = 2,size(x)
                if (strict) then
                    if (x(i)<=x(i-1)) then
                        isascend=.false.
                        return
                    end if
                else
                    if (x(i)<x(i-1)) then
                        isascend=.false.
                        return
                    end if
                end if
            end do
        end function

        subroutine precompute()
            implicit none
            integer :: iBp
            real(8) :: slope_W,slope_x

            allocate(slope_xinv(nBp),duinv_slope_Wx(nBp))

            do iBp = 1,nBp-1

                ! Compute slopes
                slope_x = (x(iBp+1)-x(iBp))/(Bp(iBp+1)-Bp(iBp))
                slope_W = (W(iBp+1)-W(iBp))/(Bp(iBp+1)-Bp(iBp))

                if (slope_W<=0d0 .or. slope_x<=0d0) then
                    duinv_slope_Wx(iBp) = DUINV_INVALID
                else
                    slope_xinv(iBp) = 1d0/slope_x
                    duinv_slope_Wx(iBp) = duinv_(slope_W*slope_xinv(iBp)) 
                end if

            end do

        end subroutine

        subroutine nomono()
            implicit none
            ! local
            integer :: i

            do i = 1,nB
                call maximize(i,-1,-1)
            end do

        end subroutine 

        subroutine simple()
            implicit none
            ! local
            integer :: i

            call maximize(1,-1,-1)
            do i = 2,nB
                call maximize(i,i-1,-1)
            end do

        end subroutine 

        ! 
        recursive subroutine binary(l,u)
            implicit none
            integer, intent(in) :: l,u
            ! local
            integer :: m

            ! Solve the maximization step at the midpoint
            m = (l+u)/2
            call maximize(m,l,u)

            ! If still here, divide and conquer
            if (m>l+1) call binary(l,m)
            if (u>m+1) call binary(m,u)

        end subroutine 
        
        subroutine maximize(i,ilb,iub)
            implicit none
            integer, intent(in) :: i,ilb,iub

            if (iscontinuous) then
                call maximize_cont(i,ilb,iub,domaxexact_)
            else
                call maximize_discrete(i,ilb,iub)
            end if

        end subroutine

        ! For index i, searches for optimal b' in {lb...ub}
        ! Saves it to g(i) along with V(i), c(i), and g_ilo(i) (which provides the exact index giving g)
        subroutine maximize_discrete(i,ilb,iub)
            implicit none
            integer, intent(in) :: i,ilb,iub
            ! local
            real(8) :: Vval,cval,xval,Wval,mass,bpval
            integer :: ip,iplb,ipub

            ! Locate index for g(ilb) and g(iub)
            if (ilb==-1) then 
                iplb = 1
            else
                iplb = g_ilo(ilb) 
            end if
            if (iub==-1) then
                ipub = nBp
            else
                ipub = g_ilo(iub)
            end if

            ! Set default policy
            V(i) = -huge(0d0)
            g(i) = Bp(iplb)
            c(i) = 0d0
            g_ilo(i) = iplb

            ! Do grid search on discrete grid
            do ip = iplb,ipub
                cval = z(i) - x(ip)
                Vval = u_(cval) + W(ip)
                if (Vval>V(i)) then
                    V(i) = Vval
                    g(i) = Bp(ip)
                    c(i) = cval
                    g_ilo(i) = ip
                end if
            end do

        end subroutine

        ! For index i, searches for optimal b' in [lb,ub]
        ! Saves it to g(i) along with V(i), c(i), and g_ilo(i) (which provides the closest index bounding g)
        subroutine maximize_cont(i,ilb,iub,doglobal)
            implicit none
            integer, intent(in) :: i,ilb,iub
            logical, intent(in) :: doglobal
            ! local
            real(8) :: Vval,cval,xval,Wval,mass,bpval
            integer :: ip,iplb_cont,iplb_grid,ipub_cont,ipub_grid

            ! Setup the search space
            !!! Lower bound
            if (ilb==-1) then 
                iplb_grid = 1
                iplb_cont = 1
            else
                if (g_ongrid(ilb)) then
                    ! Here, want to start exactly at the grid point
                    iplb_grid = g_ilo(ilb) 
                    iplb_cont = g_ilo(ilb) 
                else
                    ! If it's not on the grid, then no need to check g_ilo(ilb) 
                    iplb_grid = g_ilo(ilb)+1
                    iplb_cont = g_ilo(ilb) ! but sitll need to check the interval
                end if
            end if
            !!! Upper bound
            if (iub==-1) then
                ipub_grid = nBp
                ipub_cont = nBp
            else
                if (g_ongrid(iub)) then
                    ! Here, need to check that grid point and, for the continuous, check one below it.
                    ipub_grid = g_ilo(iub)
                    ipub_cont = g_ilo(iub)
                else
                    ! If it's not on the grid, then no need to check g_ilo(iub)
                    ipub_grid = g_ilo(iub)+1-1 
                    ipub_cont = g_ilo(iub)+1 ! but still need to check the interval
                end if
            end if

            ! Note that it is possible for ipub_grid<iplb_grid.
            ! In that case, one must check the interval (g_ilo(ilb),g_ilo(ilb)+1) when doing the continuous search.
            ! Consequently, when setting the default policy, set g_ilo(i) = g_ilo(ilb) to ensure this interval is checked
            V(i) = -huge(0d0)
            g_ilo(i) = g_ilo(ilb)

            ! Do grid search on discrete grid (it is possible no grid search is performed at this step)
            do ip = iplb_grid,ipub_grid
                cval = z(i) - x(ip)
                Vval = u_(cval) + W(ip)
                if (Vval>V(i)) then
                    V(i) = Vval
                    g(i) = Bp(ip)
                    c(i) = cval
                    g_ilo(i) = ip
                    g_ongrid(i) = .true.
                end if
            end do


            ! If doing global proceed. But if doing locally about the grid point, then overwrite iplb_cont and ipub_cont
            if (.not. doglobal) then
                iplb_cont = max(1,g_ilo(i)-1)
                ipub_cont = min(nBp,g_ilo(i)+1)
            end if

            ! Also, check values in each interval the closed form solution in each interval for an exact recovery
            do ip = iplb_cont,ipub_cont-1

                evalcnt_ = evalcnt_ + 1 ! this will be subtracted off below if u_ is called to avoid double counting

                ! Two lines after this section should be equivalent to the following commented code:  
!                   slope_x = (x(ip+1)-x(ip))/(Bp(ip+1)-Bp(ip))
!                   slope_W = (W(ip+1)-W(ip))/(Bp(ip+1)-Bp(ip))
!                   if (slope_W<=0d0 .or. slope_x<=0d0) cycle 
!                   bpval = Bp(ip) + (-x(ip)+ z(i) - duinv_(slope_W/slope_x) ) / slope_x

                ! Compute slopes : Note that if W's slope is 0 or x's slope is 0, this will be invalid
                if (duinv_slope_Wx(ip) == DUINV_INVALID) cycle

                ! Formula for b' making the euler equation hold when W and x are linearly interpolated
                bpval = Bp(ip) + (-x(ip)+ z(i) - duinv_slope_Wx(ip)) * slope_xinv(ip)

                if (bpval>Bp(ip) .and. bpval<Bp(ip+1)) then

                    ! interior solution case, might be optimal
                    mass = (Bp(ip+1)-bpval)/(Bp(ip+1)-Bp(ip))
                    xval = mass*x(ip) + (1d0-mass)*x(ip+1)
                    Wval = mass*W(ip) + (1d0-mass)*W(ip+1)

                    cval = z(i) - xval
                    Vval = u_(cval) + Wval

                    evalcnt_ = evalcnt_ - 1 ! to avoid double counting (u_ increments evalcnt_)

                    ! compare with current best
                    if (Vval>V(i)) then

                        V(i) = Vval
                        g(i) = bpval
                        c(i) = cval
                        g_ilo(i) = ip ! gives lower bound
                        g_ongrid(i) = .false.

                    end if

                end if
            end do

        end subroutine

        function duinv_(du)
            implicit none
            real(8), intent(in) :: du
            real(8) :: duinv_
            
            ! For crra, u' = c**-crra so u'-1 = du**(-1/crra)
            if (crra==2d0) then
                duinv_ = 1d0/sqrt(du)
            elseif (crra==1d0) then
                duinv_ = 1d0/du
            else
                duinv_ = du**(-1d0/crra)
            end if

        end function 

        function u_(c)
            implicit none
            real(8), intent(in) :: c
            real(8) :: u_

            evalcnt_ = evalcnt_ + 1

            if (c<=0d0) then
                u_ = death
                return
            end if
            
            ! For crra, u = c**(1-crra)/(1-crra) (except for log)
            if (crra==2d0) then
                u_ = -1d0/c
            elseif (crra==1d0) then
                u_ = log(c)
            else
                u_ = c**(1d0-crra)/(1d0-crra)
            end if

        end function 

    end subroutine 
! 
end module
