! Grey Gordon, 2016
! 
! A master/slave helper routine using MPI. The slaves should call "slave(fcn,nx)" and then use the type-bound procedures in type_gang and type_slave 
! to request the slaves compute fcn(y,x) and access the value when done. 
! 
! See "test_mpislave" for a usage example; sorry for the lack of documentation.
!
module mod_mpislave
    use mod_types
    use mod_mympi
    implicit none

!     integer, parameter :: dbglevel = 1

    type type_slave 
        real(rt), pointer :: x(:) => null()
        real(rt), pointer :: y(:) => null()
        integer :: ny ! size of y (inferred automatically)
        integer :: rank
        integer :: requestid
        logical :: busy = .false.
        logical :: ever_assigned = .false.
    contains
        procedure, pass :: assign_x
        procedure, pass :: check_busy
        procedure, pass :: shutdown => shutdown_slave
    end type 

    type type_gang
        type(type_slave), dimension(:), allocatable :: slaves
        integer :: sleeptime = 0
    contains
        procedure, pass :: init => init_gang
        procedure, pass :: shutdown => shutdown_gang
        procedure, pass :: wait_for_free
    end type

contains

    subroutine test_mpislave()
        use mod_util
        implicit none
        type(type_gang) :: gang
        integer :: i,who
        integer, parameter :: nx=2,sleeptime=0
        logical, parameter :: dbg = .false.
        real(rt) :: x(nx),piest
        type(type_seed) :: randstate

        if (mympi_nproc()==1) then
            call mympi_stop('ERROR: test_mpislave: at least 2 ranks are required')
        end if

        call randstate%randomize()

        if (mympi_iproc()/=0) then 
            call slave(fcn,nx)
            return
        end if

        if (dbg) write(*,'(A)',advance='no') 'initializing gang ...'
        call gang%init(nx,sleeptime)
        if (dbg) write(*,'(A)') ' done.'

        piest = 3._rt

        do i = 2,100000

            if (dbg) write(*,'(A)',advance='no') 'waiting for free ...'
            who = gang%wait_for_free()
            if (dbg) write(*,'(A)') ' done.'
            
            if (gang%slaves(who)%ever_assigned) then
                if (dbg) write(*,'(A)',advance='no') 'getting y value ...'
                piest = gang%slaves(who)%y(1)/i + piest*real(i-1,rt)/i
                if (dbg) write(*,'(A)') ' done.'
            end if

            if (dbg) write(*,'(A)',advance='no') 'assigning x...'
            call random_number(x)
            call gang%slaves(who)%assign_x(x)
            if (dbg) write(*,'(A)') ' done.'

        end do

        if (dbg) write(*,'(A)',advance='no') 'shutting down ...'
        call gang%shutdown()
        if (dbg) write(*,'(A)') ' done.'

        if (abs(piest-3.14d0)<.1d0) then
            print*,'test_mpislave: passed'
        else
            stop 'test_mpislave: failed, pi estimate is not close to pi'
        end if

    contains

        subroutine fcn(y,x)
            real(rt), intent(out), allocatable :: y(:)
            real(rt), intent(in) :: x(:)

            ! Determine whether x1,x2 lies insides a circle inscribed in a square, or outside of it. 

            ! Area ratio is pi over 4, so to get estimate of pi, multiply by 4
            allocate(y(1))
            y(1) = merge(4._rt,0._rt,norm2(x)<1._rt) 

        end subroutine

    end subroutine

    ! Do something according to the value of flag(this_image())[1]
    !  1 := evaluate slavex and store in slavey
    ! -1 := exit the slave routine
    subroutine slave(fcn,nx)
        use mod_mympi
        implicit none
        integer, intent(in) :: nx
        interface 
            subroutine fcn(y,x) 
                use mod_types
                implicit none
                real(rt), dimension(:), intent(out), allocatable :: y ! dimension inferred after fcn is called
                real(rt), dimension(:), intent(in) :: x
            end subroutine
        end interface
        ! local
        integer :: flag
        real(rt), allocatable :: slavex(:),slavey(:)

        allocate(slavex(nx))

        do 

            ! Wait for a flag 
            call mympi_recv(flag,0)

            ! If flag==-1, then shutdown
            if (flag==-1) return

            ! If still here, wait for an x
            call mympi_recv(slavex,0)

            ! If flag==1, execute f(x) and store on master.
            if (flag==1) then

                ! Compute the result
                call fcn(slavey,slavex) 

                ! Note that here, slavey should be allocated. If not, abort! 
                if (.not. allocated(slavey)) call mympi_stop('ERROR: mod_mpislave: slave: fcn must allocate slavey, but was not allocated.')

                ! Send the dimension to master tagged by mympi_iproc()
                call mympi_send(size(slavey),0,tag=mympi_iproc()) 

                ! Send the result to master tagged by mympi_nproc()+mympi_iproc()
                ! call mympi_send(slavey,0,tag=mympi_nproc()+mympi_iproc()) 
                call mympi_send(slavey,0) !,tag=mympi_nproc()+mympi_iproc()) 

            end if

        end do

    end subroutine

    ! a routine to shut down all slaves
    subroutine shutdown_slave(this,sleeptime) 
        use mod_util
        implicit none
        class(type_slave), intent(inout) :: this
        integer, intent(in) :: sleeptime
        ! local
        integer :: j
        character(1), volatile :: abc ! volatile prevents optimization (this is used for a wait loop)

        ! Wait for rank to finish if not finished task
        do while (this%busy) 

            ! Check if still busy (if not, then this%busy will be set to .false.
            call this%check_busy()

            ! If still busy, then sleep or do something else for a little while to not thrash the system
            if (this%busy) then

                ! If sleeptime, then sleep
                if (sleeptime>=1) then

                    call sleep(sleeptime)

                else

                    do j = 1,1000000
                        abc = 'a'
                    end do

                end if


            end if

        end do
            
        ! Once not busy, then send the shut down signal of flag==-1
        call mympi_send(-1,this%rank) 

    end subroutine

    subroutine check_busy(this)
        implicit none
        class(type_slave), intent(inout) :: this
        ! local
        logical :: tf

        if (this%busy) then

            ! Check whether i has finished
            tf = mympi_test(this%requestid)

            if (tf) then


                ! Finished sending ny, so need to allocate y and receive it
                if (associated(this%y)) then
                    if (this%ny/=size(this%y)) then
                        deallocate(this%y)
                        allocate(this%y(this%ny))
                        ! Now y has the correct shape
                    end if
                else
                    allocate(this%y(this%ny))
                end if

                ! Here, this%y is allocated and is size this%ny. Receive it.
                call mympi_recv(this%y,this%rank)

                this%busy = .false.

            end if

        end if

    end subroutine

    ! This routine sends the value of x to a slave "who", who will then store the value in y. 
    subroutine assign_x(this,x) !,y)
        implicit none
        class(type_slave), intent(inout) :: this
        real(rt), intent(in), target :: x(:)
        integer :: ierr

        this%busy = .true.
        this%ever_assigned = .true.

        this%x = x

        call mympi_send(1,this%rank) ! blocking send (tell them to proceed to execute)
        call mympi_send(x,this%rank) ! assign an x value (wait for it to complete)

        ! Request that the value of ny be sent and stored in this%ny (once finished, will indicate completed task)
        call MPI_Irecv(this%ny, 1, MPI_INTEGER, this%rank, this%rank, MPI_COMM_WORLD, this%requestid, ierr)

        ! Once the slave has finished evaluating f(x), it's value will be stored in this%y 

    end subroutine 

    subroutine init_gang(this,nx,sleeptime)
        implicit none
        class(type_gang), intent(inout) :: this
        integer, intent(in) :: nx ! size of x
        integer, intent(in), optional :: sleeptime
        ! local
        integer :: i,nranks

        if (allocated(this%slaves)) deallocate(this%slaves)

        if (present(sleeptime)) this%sleeptime = sleeptime

        nranks = mympi_nproc()

        allocate(this%slaves(nranks-1))
        do i = 1,nranks-1
            if (associated(this%slaves(i)%x)) deallocate(this%slaves(i)%x)
            allocate(this%slaves(i)%x(nx))
            this%slaves(i)%rank = i
        end do

    end subroutine

    ! a routine to shut down all slaves
    subroutine shutdown_gang(this) 
        implicit none
        class(type_gang), intent(inout) :: this
        ! local
        integer :: i

        do i = 1,size(this%slaves)
            call this%slaves(i)%shutdown(this%sleeptime)
        end do
        
    end subroutine

    ! a routine returning the index of a free slave
    function wait_for_free(this) result(who)
        use mod_util
        implicit none
        class(type_gang), intent(inout) :: this
        integer :: who
        ! local
        real(8) :: tic, toc, tmp
        integer :: i

        do 

            do i = 1,size(this%slaves)

                ! Test whether a particular slave is busy
                call this%slaves(i)%check_busy()
                if (.not. this%slaves(i)%busy) then
                    who = this%slaves(i)%rank
                    return
                end if

            end do


            ! If sleeptime, then sleep
            if (this%sleeptime>=1) then

                call sleep(this%sleeptime)

            else

                ! Do some busy work so as to not continually check the status (which may cause MPI problems)
                tic = MPI_WTIME()
                toc = tic
                tmp = 0d0
                do while (toc-tic<1d-6)
                    toc = MPI_WTIME()
                    tmp = tmp + i 
                end do
                
                ! Trick the compiler into not optimizing away the busy work (hopefully)
                open(newunit=i,file='/lkandsfkasn/makecompilernotoptimize.dat',status='unknown',err=999) ! file should not exist (but could), so this will jump to line 999
                write(i,*) tmp
                close(i)
999             tmp = 0d0

            end if

        end do

    end function



end module
