#include <iostream>
#include <cstdio>
#include <new>
#include <cstdlib>
#include <cmath>
#include <cfloat>
#include <cublas_v2.h>
#include "culibgrey.h"
#include <ctime> 

void STOP(std::string msg, std::string file, int line){
    std::cerr<<"ERROR: "<<file<<"("<<line<<"): "<<msg<<std::endl;
    exit(EXIT_FAILURE);
}

double testfunctions_squaredmin2(double x){return x*x - 2.;};

int testbisect(){

    // compute the sqrt of 2 via f(x) = 0 where f x*x = 2 
    optsettings<double> o;
    o.pmod=5;
    double x = bisect(testfunctions_squaredmin2,0.,3.33333,o);
    std::cout<<"sqrt of 2 is "<<x<<std::endl;
    return 0;

}

__global__ void cubellman(float *V, float *c, int *api, float *W, float *gA, float *gE, int nAp, int nA, int nE){

    float vtmp,ctmp,vbest,cbest,ytmp;
    int iE,iA,iAp,i,iApbest;

    i = threadIdx.x + blockDim.x * blockIdx.x;
    // i = iA + iE*nA  so  iE = i/nA and iA = i - iE*nA 
    iE = i/nA;
    iA = i - iE*nA;

    if (iE>=nE || iA<0 || iA>=nA) return;

    ytmp = gA[iA] + gE[iE];
    vbest = -1e20f;
    cbest = -1.f;
    iApbest = -1;
    for (iAp=0; iAp<nAp; iAp++){

        // Program in the budget constraint here
        ctmp = ytmp - .96f*gA[iAp];

        // Compute the value 
        vtmp = -1.f/max(ctmp,FLT_EPSILON) + W[iAp + iE*nAp];

        if (vtmp>vbest) {
            vbest = vtmp;
            cbest = ctmp;
            iApbest = iAp;
        }

    }
    V[i] = vbest;
    c[i] = cbest;
    api[i] = iApbest;

    return;

}

// here, blockIdx gives the (a,e) pair responsible for. threadIdx gives the (ap) values responsible for.
// specificaly, blockIdx.x gives the a and blockIdx.y gives the e
__global__ void cubellman2(float *V, float *c, int *api, float *W, float *gA, float *gE, int nAp, int nA, int nE){

    float vbest;
    int iE,iA,iAp,iApbest;

    //__shared__ float sV[128]; 
    extern __shared__ float sV[], sC[]; // the size for this is specified by the third argument in the <<<...>>> notation
    extern __shared__ float sapi[]; 

    iA = blockIdx.x;
    iE = blockIdx.y;
    iAp = threadIdx.x; 

    // compute util from your choice and store in shared memory
    sC[iAp] = gA[iA] + gE[iE] - .96f*gA[iAp];
    sV[iAp] = -1.f/max(sC[iAp],FLT_EPSILON) + W[iAp+iE*nAp];
    
    // compute maximum across threads
    __syncthreads(); // wait for everyone to finish

    // store the best iAp values 
//    if (iA==0 & iE==0) {
//printf("I am thread (%d) in block (%d, %d)\n",threadIdx.x, blockIdx.x, blockIdx.y);
 //   }
        
//    // STUPID EASY WAY: have thread 0 compute max over entire shared memory
//    if (iAp==0) {
//        vbest = -1e10f;
//        iApbest = 0;
//        for (unsigned int j = 0; j<nAp; j++) {
//            if (sV[j]>vbest) {
//                vbest = sV[j];
//                iApbest = j;
//            }
//        }
//        
//        V[iA + iE*nA] = vbest;
//        api[iA + iE*nA] = iApbest;
//        //c[iA + iE*nA] = gA[iA] + gE[iE] - .96f*gA[iApbest];
//        c[iA + iE*nA] = sC[iApbest];
//
//    }

    // FASTER WAY: have threads compute max in parallel. 
    // Specifically, recursively compute the maximum over shared memory by comparing having each thread compare only 
    // two elements at each step
    for (unsigned int s = blockDim.x/2; s>0; s>>=1) { 
        if (iAp<s) {
            if (sV[iAp]<sV[iAp+s]) {
                sV[iAp] = sV[iAp+s];
                sC[iAp] = sC[iAp+s];
                sapi[iAp] = sapi[iAp+s];
            }
        }
        
        __syncthreads();
    }
    // Write result back to global memory
    if (iAp==0) {
        V[iA + iE*nA] = sV[0];
        api[iA + iE*nA] = sC[0];
        c[iA + iE*nA] = sapi[0];
    }



    return;

}

// An attempt to optimize cubellman
// Assumes that nA==nAp
__global__ void cubellman3(float *V, float *c, int *api, float *W, float *gA, float *gE, int nAp, int nA, int nE){

    float vtmp,ctmp,vbest,cbest,ytmp;
    int iE,iA,iAp,iApbest;

    extern __shared__ float shmem[];  
    float *sW;
    float *sgA;
    sW = &shmem[0];
    sgA = &shmem[nA];
    //__shared__ float sW[64], sgA[64]; // 
    
    // Assume nA = nAp and nE is number of blocks and nA is the number of threads per block
    iA = threadIdx.x; // 
    iE = blockIdx.x; // Assume the iE values always separate across blocks (the first dim may not be). 

    // Load the continuation utility and asset grid into local memory
    sW[iA] = W[iA + iE*nA];
    sgA[iA] = gA[iA];
    __syncthreads();

    // Now compute the maximum
    ytmp = sgA[iA] + gE[iE];
    vbest = -1e10f;
    cbest = -1.f;
    iApbest = -1;
    for (iAp=0; iAp<nAp; iAp++){

        // Program in the budget constraint here
        ctmp = ytmp - .96f*sgA[iAp];

        // Compute the value 
        vtmp = -1.f/max(ctmp,FLT_EPSILON) + sW[iAp];

        if (vtmp>vbest) {
            vbest = vtmp;
            cbest = ctmp;
            iApbest = iAp;
        }

    }
    V[iA + iE*nA] = vbest;
    c[iA + iE*nA] = cbest;
    api[iA + iE*nA] = iApbest;


}




// computes y = ax + y
int axpy(float a, const devicearray<float> &x,devicearray<float> &y, cublasHandle_t handle){

    cublasStatus_t stat;
    stat = cublasSaxpy(handle,x.size,&a,x.data,1,y.data,1); if (stat != CUBLAS_STATUS_SUCCESS) std::cerr<<"ERROR DETECTED IN axpy WHEN CALLING cublasSaxpy"<<std::endl;
    return 0;

}

// routine to copy a to b. both must be allocated and have the same size.
int copy(const devicearray<float> &a, devicearray<float> &b, cublasHandle_t handle){
    cublasStatus_t stat;
    if (a.size==b.size) {
        stat = cublasScopy(handle,a.size,a.data,1,b.data,1); if (stat != CUBLAS_STATUS_SUCCESS) std::cerr<<"ERROR DETECTED IN copy WHEN CALLING cublasScopy"<<std::endl;
        return 0;
    } else {
        std::cerr<<"copy: cannot copy arrays, they do not have the same size"<<std::endl;
        return 1;
    }
};

// computes c = a-b
int minus(devicearray<float> &c,const devicearray<float> &a,const devicearray<float> &b, cublasHandle_t handle){

    // copy a to c
    copy(a,c,handle);   

    // call c = (-1)*b + a which now is c = (-1)*b + c
    axpy(-1.f,b,c,handle);

    return 0;
}

__global__ void global_max(float *x, float c){ x[blockIdx.x] = fmaxf(x[blockIdx.x],c);}
__host__ void max(devicearray<float> &x, float c){global_max<<<x.size,1>>>(x.data,c);}
__global__ void global_min(float *x, float c){ x[blockIdx.x] = fminf(x[blockIdx.x],c);}
__host__ void min(devicearray<float> &x, float c){global_min<<<x.size,1>>>(x.data,c);}

// Compute the maximum value of an array and its index
// To do this via blas calls, I use the mathematical relation 
// If maxval(|x|)>0, x has a positive element. In this case, max x = maxval(|x|). 
// If maxval(|x|)=0, x has no positive elements. In this case, max x = -minval(|x|).

// If minval(|-x|)>0, x has a negative element. 
// If minval(|-x|)=0, x has no negative elements. 
__host__ int maxloc(devicearray<float> &x, cublasHandle_t handle){
    cublasStatus_t stat;
    int imax;
    
    // find maxloc(|x|)
    stat = cublasIsamax(handle, x.size, x.data, 1, &imax); if (stat != CUBLAS_STATUS_SUCCESS) std::cerr<<"ERROR DETECTED IN MAXLOC WHEN CALLING cublasIsamax"<<std::endl;

    imax = imax - 1; // adjust to zeros based indexing


    if (x[imax]>=FLT_EPSILON){

        // x has a positive element, so max is just maxval(|max(x,0)), the index of which is imax
        return imax;

    } else {
        
        // x has no positive elements, so find smallest |x| value
        stat = cublasIsamin(handle, x.size, x.data, 1, &imax); if (stat != CUBLAS_STATUS_SUCCESS) std::cerr<<"ERROR DETECTED IN MAXLOC WHEN CALLING cublasIsamin"<<std::endl;
        imax = imax - 1; // adjust to zeros based indexing

        // negate it to get the max
        return imax;
    }
}
__host__ float maxval(devicearray<float> &x, cublasHandle_t handle){
    int imax = maxloc(x,handle);
    return x[imax];
}

//routine for computing the supnorm
__host__ float supnorm(devicearray<float> &a, devicearray<float> &b, devicearray<float> &work, cublasHandle_t handle){
    float x;
    int i;
//    cublasStatus_t stat;

    // work = a - b
    minus(work,a,b,handle);

    // work = abs(a-b);
    abs(work);

    // compute maxval(abs(a-b))
    i = maxloc(work,handle);
    x = work[i]; // thunk
    //maxval(x,i,work,handle); 

    if (x<0.f) {
        std::cout<<"x negative "<<x<<" i "<<i<<" a[i] "<<a[i]<<" b[i] "<<b[i]<<" work[i] "<<work[i]<<std::endl;
    }

    return x;
    
//    // 
//    stat = cublasIsamax(cublasHandle_t handle, int n, const float *x, int incx, int *result);
}

extern "C" void sgemm_( char * ta, char * tb, int * m, int *n, int *k, float *alpha, float *a, int *lda, float *b, int *ldb, float *beta, float *c, int *ldc);
extern "C" void dgemm_( char * ta, char * tb, int * m, int *n, int *k, double *alpha, double *a, int *lda, double *b, int *ldb, double *beta, double *c, int *ldc);

void mtimes(array<double>& c,array<double>& a,array<double>& b, char ta='N', char tb='N', double alpha=1., double beta=0.){
    int na1 = a.n[0], na2 = a.n[1], nb1 = b.n[0], nb2 = b.n[1];
    int m,n,k;

    m = (ta=='N' ? na1 : na2);
    n = (tb=='N' ? nb2 : nb1);
    k = (ta=='N' ? na2 : na1);

    dgemm_(&ta,&tb,&m,&n,&k,&alpha,a.data,&na1,b.data,&nb1,&beta,c.data,&na1);

}

void mtimes(double *c, double *a, double *b, int na1, int na2, int nb1, int nb2, char ta='N', char tb='N', double alpha=1., double beta=0.){
    int m,n,k;

    m = (ta=='N' ? na1 : na2);
    n = (tb=='N' ? nb2 : nb1);
    k = (ta=='N' ? na2 : na1);

    dgemm_(&ta,&tb,&m,&n,&k,&alpha,a,&na1,b,&nb1,&beta,c,&na1);

}

void mtimes(float *c, float *a, float *b, int na1, int na2, int nb1, int nb2, char ta='N', char tb='N', float alpha=1.f, float beta=0.f){
    int m,n,k;

    m = (ta=='N' ? na1 : na2);
    n = (tb=='N' ? nb2 : nb1);
    k = (ta=='N' ? na2 : na1);

    sgemm_(&ta,&tb,&m,&n,&k,&alpha,a,&na1,b,&nb1,&beta,c,&na1);
//    std::cout<<"not doing anything in mtimes"<<std::endl;

}

int mtimes(devicearray<float> &c, devicearray<float> &a, devicearray<float> &b, char tA, char tB, float alpha, float beta, cublasHandle_t handle){
    cublasStatus_t stat;

    cublasOperation_t transa = (tA=='N' ? CUBLAS_OP_N : CUBLAS_OP_T);
    cublasOperation_t transb = (tB=='N' ? CUBLAS_OP_N : CUBLAS_OP_T);
    
    int m = (tA=='N' ? a.n[0] : a.n[1]);
    int n = (tB=='N' ? b.n[1] : b.n[0]);
    int k = (tA=='N' ? a.n[1] : a.n[0]);

    // check that dimensions conform
    stat = cublasSgemm(handle,transa,transb,m,n,k,&alpha,a.data,a.n[0],b.data,b.n[0],&beta,c.data,c.n[0]); if (stat != CUBLAS_STATUS_SUCCESS) std::cerr<<"ERROR DETECTED IN MAXLOC WHEN CALLING cublasSgemm"<<std::endl;

    return 0;
}

void testcubellman(){

    int nA = 8*1024, nE = 64, nAp = nA; 
    //int nA = 64, nE = 8, nAp = nA; 
    float sup, lastsup = 1.f;
    const float beta = .96f;
    time_t tstart,tend;

    devicearray<float> V(nA,nE), V0(nA,nE), c(nA,nE), W(nA,nE), Fee(nE,nE), workAE(nA,nE);
    devicearray<float> gA(nA), gE(nE);
    devicearray<int> api(nA,nE);

    cublasHandle_t handle;
    cublasStatus_t stat;

    // start the clock
    time(&tstart);

    // initialize cublas
    stat = cublasCreate(&handle); if (stat != CUBLAS_STATUS_SUCCESS) std::cerr<<"ERROR DETECTED IN cublasCreate"<<std::endl;

    // set up grids and shocks
    linspace(gA,0.f,10.f);
    tauchen(gE,Fee,3.f,.95f,.17f);

    // check them
    disp("gE",gE);
    disp("gA",gA);
    disp("Fee",Fee);

    // zero out some things
    zeros(V);
    zeros(V0);
    zeros(W);

    for (int iter=0; iter<2000; iter++){

        // compute C0 utility
        mtimes(W,V,Fee,'N','T',beta,0.,handle);

        // save previous value function
        copy(V,V0,handle);

        // solve the bellman equation
        // Method 1: distribute the iA and iE across threads and have each thread maximize over all the iAp values.
        int nThreadsPerBlock = min(1024,nA*nE);
        int nBlocks = 1 + (nE*nA-1)/nThreadsPerBlock; // ensures that if nA and nE are not divisible by 1024, cover everything
        cubellman<<<nBlocks,nThreadsPerBlock>>>(V.data,c.data,api.data,W.data,gA.data,gE.data,nAp,nA,nE);

        // Method 2: 
        //dim3 blockDim(nAp);
        //dim3 gridDim(nA,nE);
        //cubellman2<<<gridDim,blockDim,nAp*sizeof(int)+2*nAp*sizeof(float)>>>(V.data,c.data,api.data,W.data,gA.data,gE.data,nAp,nA,nE);
        //cubellman2<<<gridDim,blockDim>>>(V.data,c.data,api.data,W.data,gA.data,gE.data,nAp,nA,nE);

        // Method 3:
        //dim3 blockDim(nA); // assumes nA = nAp
        //dim3 gridDim(nE);
        //cubellman3<<<gridDim,blockDim,2*nA*sizeof(float)>>>(V.data,c.data,api.data,W.data,gA.data,gE.data,nAp,nA,nE);
        //disp("V",V);
        //STOP("checking first iter",__FILE__,__LINE__);

        //std::cout<<" maxval V after bellman: "<<maxval(V,handle)<<std::endl;

        // compute the supnorm
        sup = supnorm(V,V0,workAE,handle);

        // check for convergence
        copy(V,workAE,handle);
        abs(workAE);
        float z = maxval(workAE,handle); // just to check a particular value for making sure everything works

        if (iter%50==0 || sup<1e-6f*z) std::cout<<"iter: "<<iter<<" sup: "<<sup<<" modulus: "<<sup/lastsup<<" termination: "<<1e-6f*z<<" maxval V: "<<maxval(V,handle)<<std::endl;
        if (sup<1e-6f*z) break;

        lastsup = sup;

    }

    std::cout<<"max of V: "<<maxval(V,handle)<<std::endl;

    cublasDestroy(handle);

    time(&tend);
    std::cout<<"elapsed time: "<<difftime(tend,tstart)/60.0<<" minutes"<<std::endl;

    return;
}




