#include <iostream>
#include <iomanip>
#include <fstream>
#include <cstdio>
#include <cstdlib>
#include <cmath>
#include <cublas_v2.h>
#include <cuda_runtime.h>

#define CLIBGREYMAXRANK 6

//template <class T>
//__host__ save(std::string name, array<T> x){
//}

void STOP(std::string msg, std::string file, int line);

#define gpuErrchk(ans) { gpuAssert((ans), __FILE__, __LINE__); }
inline void gpuAssert(cudaError_t code, const char* file, int line, bool abort=true){
    if (code != cudaSuccess) {
        //fprintf(stderr,"GPUassert: %s %s %d\n", cudaGetErrorString(code), file, line);
        std::cerr<<"GPUassert: "<<cudaGetErrorString(code)<<file<<line<<std::endl;
        if (abort) exit(code);
    }
}

template <class T>
class array{
public:
    T *data;
    int n[CLIBGREYMAXRANK]; // dimensions
    int rank; // number of dimensions
    int size; // number of elements
    bool isAllocated; // whether or not allocated
    
    // constructor (this calls init but separate in case want to free or reallocate later)
    array(int n1, int n2, int n3, int n4, int n5, int n6){isAllocated=false;init(n1,n2,n3,n4,n5,n6);isAllocated=true;};
    array(int n1, int n2, int n3, int n4, int n5){isAllocated=false;init(n1,n2,n3,n4,n5);isAllocated=true;};
    array(int n1, int n2, int n3, int n4){isAllocated=false;init(n1,n2,n3,n4);isAllocated=true;};
    array(int n1, int n2, int n3){isAllocated=false;init(n1,n2,n3);isAllocated=true;};
    array(int n1, int n2){isAllocated=false;init(n1,n2);isAllocated=true;};
    array(int n1){isAllocated=false;init(n1);isAllocated=true;};
    array(){isAllocated=false; size=-1; rank=-1;}; // uninitialized

    // destructor 
    ~array(){if (isAllocated) free();};

    // initialize
    void init(int n1, int n2, int n3, int n4, int n5, int n6){
        if (!isAllocated) {
            for (int i=0; i<CLIBGREYMAXRANK; i++) {
                n[i] = 1;
            }
            rank = 6;
            n[0] = n1; n[1] = n2; n[2] = n3; n[3] = n4; n[4] = n5; n[5] = n6;
            size = n1*n2*n3*n4*n5*n6;
            data = new T [size];
            isAllocated = true;
        } else {
            std::cerr<<"Error: "<<__FILE__<<" line "<<__LINE__<<" array already initialized"<<std::endl;
//            std::exit(EXIT_FAILURE);
        }
    }
    void init(int n1, int n2, int n3, int n4, int n5){init(n1,n2,n3,n4,n5,1); rank = 5;}
    void init(int n1, int n2, int n3, int n4){init(n1,n2,n3,n4,1,1);rank = 4;}
    void init(int n1, int n2, int n3){init(n1,n2,n3,1,1,1);rank = 3;}
    void init(int n1, int n2){init(n1,n2,1,1,1,1);rank = 2;}
    void init(int n1){init(n1,1,1,1,1,1);rank = 1;}

    // free
    void free(){
        if (isAllocated) {delete[] (data);};
        isAllocated = false;
    }

    // overload brackets for easy access to data
    T& operator[](int i){return data[i];}; //setter

    // overload parentheses for easy access to multidimensional arrays
    T& operator()(int i, int j){return data[i + j*n[0]];};
    T& operator()(int i, int j, int k){return data[i + j*n[0] + k*n[0]*n[1]];};

    // overloaded values
    array<T>& operator=(const array<T> &source); // this = source
    array<T>& operator=(const T x); // this = x
    array<T> operator-(const array<T> &rhs); // return = this - rhs
    array<T> operator+(const array<T> &rhs); // return = this + rhs

    // display some properties
    void stats(){
        std::cout<<"allocated: "<<isAllocated<<std::endl;
        std::cout<<"rank:      "<<rank<<std::endl;
        std::cout<<"shape:     ";
        for (int i=0; i<rank; i++) {std::cout<<n[i]<<" ";};
        std::cout<<std::endl;
        std::cout<<"size:      "<<size<<std::endl;
    };

    // fill the array with any argument of the same type
    //array<T>& fill(T val){for (int i=0;i<size;i++) data[i] = val;}
    void fill(T val){for (int i=0;i<size;i++) data[i] = val;}

    // reshape
    void reshape(int n1, int n2, int n3, int n4, int n5, int n6);
    void reshape(int n1, int n2, int n3, int n4, int n5){reshape(n1,n2,n3,n4,n5,1);rank=5;};
    void reshape(int n1, int n2, int n3, int n4){reshape(n1,n2,n3,n4,1,1);rank=4;};
    void reshape(int n1, int n2, int n3){reshape(n1,n2,n3,1,1,1);rank=3;};
    void reshape(int n1, int n2){reshape(n1,n2,1,1,1,1);rank=2;};
    void reshape(int n1){reshape(n1,1,1,1,1,1);rank=1;};

};

template <class T>
class devicearray{
public:
    T *data;
    int n[10]; // dimensions
    int rank; // number of dimensions
    int size; // number of elements
    bool isAllocated; // whether or not allocated
    
    // constructor (this calls init but separate in case want to free or reallocate later)
    devicearray(int n1, int n2, int n3){isAllocated=0;init(n1,n2,n3);};
    devicearray(int n1, int n2){isAllocated=0;init(n1,n2);};
    devicearray(int n1){isAllocated=0;init(n1);};
//    devicearray(){}; // could define as a scalar, but this will likely lead to incorrect usage when declaration is accidentally left out

    // destructor 
    ~devicearray(){if (isAllocated) final();};

    // initialize
    void init(int n1, int n2, int n3, int n4, int n5){
        if (!isAllocated) {
            for (int i=0; i<10; i++) {
                n[i] = 1;
            }
            rank = 5;
            n[0] = n1;
            n[1] = n2;
            n[2] = n3;
            n[3] = n4;
            n[4] = n5;
            size = n1*n2*n3*n4*n5;
            gpuErrchk(cudaMalloc((void **) &data, sizeof(T) * size));
            isAllocated = 1;
        } else {
            std::cerr<<"Error: "<<__FILE__<<" line "<<__LINE__<<" array already initialized"<<std::endl;
            std::exit(EXIT_FAILURE);
        }
    }
    void init(int n1, int n2, int n3, int n4){init(n1,n2,n3,n4,1);rank = 4;}
    void init(int n1, int n2, int n3){init(n1,n2,n3,1,1);rank = 3;}
    void init(int n1, int n2){init(n1,n2,1,1,1);rank = 2;}
    void init(int n1){init(n1,1,1,1,1);rank = 1;}

    // finalize
    void final(){
        cudaFree(data);
        isAllocated = 0;
    }

    // overload brackets for each access to device memory
    T operator[](int i){
        T x;
        cudaMemcpy(&x,&data[i],sizeof(T) * 1,cudaMemcpyDeviceToHost);
        return x;
    };

};

template <class T>
void array<T>::reshape(int n1, int n2, int n3, int n4, int n5, int n6){
    if (n1*n2*n3*n4*n5*n6 == size){
        n[0] = n1;
        n[1] = n2;
        n[2] = n3;
        n[3] = n4;
        n[4] = n5;
        n[5] = n6;
        rank = 6;
    } else {
        STOP("cannot reshape array, number of elements is not the same.",__FILE__,__LINE__);
    }
}

// function to check whether dims are the same
template <class T>
bool sameshape(const array<T> &a,const array<T> &b){
    if (!a.isAllocated || !b.isAllocated){
        return false; 
    } 
    if (a.rank!=b.rank) {
        return false;
    }
    for (int i = 0; i<a.rank; i++){
        if (a.n[i]!=b.n[i]) {return false;}
    }
    return true;
};

template <class T>
int init_from_source(array<T> &lhs,const array<T> &rhs){
    // now target is free and can reinit
    if (rhs.rank>6) STOP("init_from_source: Have not programmed this many dims yet",__FILE__,__LINE__);
    lhs.init(rhs.n[0],rhs.n[1],rhs.n[2],rhs.n[3],rhs.n[4],rhs.n[5]);
    lhs.rank = rhs.rank;
    return 0;
}


// assignment array operator (a = b is like a.=(b). this refers to a and is well-defined.)
template <class T>
array<T>& array<T>::operator=(const array<T> &source){

    if (sameshape(*this,source)){
        // in this case, no allocation or reallocation is necessary, just copy all elements over
    } else if (!isAllocated) {
        // now target is free and can reinit
        init_from_source(*this,source);
    } else {
        // here, they do not conform and the lhs is allocated. I do not allow this case.
        STOP("LHS is allocated but does have same size as RHS.",__FILE__,__LINE__);
    }

    // copy over the data
    for (int i = 0; i<source.size; i++){data[i] = source.data[i];}

    return *this;
}

template <class T>
void minus(array<T> &c, const array<T>&a, const array<T>&b){

    if (!c.isAllocated) init_from_source(c,a);           

    if (!sameshape(c,a) || !sameshape(a,b)) STOP("Minus: Dimensions do not conform. Please allocate LHS first.",__FILE__,__LINE__);

    for (int i = 0; i<a.size; i++){c.data[i] = a.data[i] - b.data[i];}
    
}

template <class T>
void plus(array<T> &c, const array<T>&a, const array<T>&b){

    if (!c.isAllocated) init_from_source(c,a);           

    if (!sameshape(c,a) || !sameshape(a,b)) STOP("Minus: Dimensions do not conform. Please allocate LHS first.",__FILE__,__LINE__);

    for (int i = 0; i<a.size; i++){c.data[i] = a.data[i] + b.data[i];}
    
}

template <class T>
void abs(array<T> &x){

    // perform the minus operation
    for (int i = 0; i<x.size; i++){x.data[i] = fabs(x.data[i]);};
    
}

// minus array operator, works as lhs = this - rhs
template <class T>
array<T> array<T>::operator-(const array<T> &rhs){
    array<T> lhs;
    init_from_source(lhs,rhs);
    minus(lhs,*this,rhs);
    return lhs;
}

// plus array operator, works as lhs = this + rhs
template <class T>
array<T> array<T>::operator+(const array<T> &rhs){
    array<T> lhs;
    init_from_source(lhs,rhs);
    plus(lhs,*this,rhs);
    return lhs;
}



template <class T>
__host__ __device__ T max(T x, T y){return ( x > y ? x : y );}

template <class T>
__host__ __device__ T min(T x, T y){return ( x < y ? x : y );}

template <class T>
T maxval(array<T> &x){
    int i;
    T sup;
    
    sup = fabs(x.data[0]);
    for (i=1;i<x.size;i++){
        sup = max<T>(sup,fabs(x.data[i]));
    }

    return sup;
}


template <class T>
__host__ __device__ T supnorm(T *x, T *y, int n){
    int i;
    T sup;
    
    sup = fabs(y[0]-x[0]);
    for (i=1;i<n;i++){
        sup = max<T>(sup,fabs(y[i]-x[i]));
    }

    return sup;
}

template <class T>
__host__ void disp(T* x,int n1=1, int n2=1){
    int i,j;
    for (i=0; i<n1; i++) {
        for (j=0; j<n2; j++) {
            // can use std::fixed to print zeros everywhere like matlab
            std::cout<<std::setw(15)<<std::setprecision(6)<<x[i+j*n1];
        }
        std::cout<<std::endl;
    }
}

template <class T>
__host__ void disp(std::string str, devicearray<T> &x){

     std::cout<<str<<":"<<std::endl;

     if (x.rank>2) 
         std::cout<<" only displaying values for (:,:,1)"<<std::endl;
 
     for (int i=0; i<x.n[0]; i++) {
         for (int j=0; j<x.n[1]; j++) {
             std::cout<<std::setw(15)<<std::setprecision(6)<<x[i + j*x.n[0]]; // brackets call cudaMemcopy
         }
         std::cout<<std::endl;
     }
 
}


template <class T> __host__ void disp(std::string s){std::cout<<s<<std::endl;}
template <class T> __host__ void disp(std::string str, T* x, int n1=1, int n2=1){std::cout<<str<<":"<<std::endl; disp(x,n1,n2);}
template <class T> __host__ void disp(std::string str, array<T> &x){std::cout<<str<<":"<<std::endl; disp(x);}
template <class T> __host__ void disp(array<T> &x){disp(x.data,x.n[0],x.n[1]);}

template <class T>
__host__ __device__ T normcdf(T x,T mean,T std){
    T one = 1, two = 2;
    return one/two*(one+erf((x-mean)/(std*sqrt(two))));
}

template <class T>
__host__ __device__ void linspace(T* x,T a,T b, int n){
    T m;
    int i;
    const T one = 1;

    x[0] = a;
    for (i=1; i<n-1; i++){
        m = T(i) / (n-1); // use a double, but case x into T
        x[i] = a*(one-m) + b*m;
    }
    x[n-1] = b;

    return;

}

template <class T>
__host__ __device__ void linspace(array<T> &x, T a, T b){ linspace(x.data,a,b,x.size); }

template <class T>
__host__ __device__ void tauchen(T *z, T *Fzz, T cover, T ar1, T cdnalstd, int n){

    T unstd,cdnalmean,mid,mid2;
    const T one=1,two=2,half=one/two;
    int i,j;

    // unconditional stdev
    unstd = cdnalstd/sqrt(one - ar1*ar1);

    // set grid
    linspace(z,-cover*unstd,cover*unstd,n);

    // construct transition matrix
    for (i=0; i<n; i++){
        // get conditional mean
        cdnalmean = ar1*z[i];
    
        // at left boundary, use weight F(m(0))
        j = 0;
        mid = half*z[j] + half*z[j+1];
        Fzz[i + j*n] = normcdf(mid,cdnalmean,cdnalstd);

        for (j=1; j<n-1; j++){
            // use weight F(m(j+1)) - F(m(j))
            mid = half*z[j-1] + half*z[j];
            mid2 = half*z[j] + half*z[j+1];
            Fzz[i + j*n] = normcdf(mid2,cdnalmean,cdnalstd)-normcdf(mid,cdnalmean,cdnalstd);
        }

        // at right boundary, use weight F(m(n-1))
        j = n-1;
        mid = half*z[j-1] + half*z[j];
        Fzz[i + j*n] = one - normcdf(mid,cdnalmean,cdnalstd);

    }

    // convert grid out of logs
    for (i=0; i<n; i++){
        z[i] = exp(z[i]);
    }

}

template <class T>
__host__ __device__ void tauchen(array<T> &z, array<T> &Fzz, T cover, T ar1, T cdnalstd){tauchen(z.data,Fzz.data,cover,ar1,cdnalstd,z.size);}


// bisection method
template <class T>
class optsettings {
public:
    T xtoler;
    int maxiter, pmod;

    // constructor
    optsettings(){xtoler = 1e-6f; maxiter= 10000; pmod = 1;};

    // destructor
    ~optsettings(){};

//    optsettings optimset(std::string s, void * x);
};

template <class T>
T bisect(T (*f)(T x),T a,T b, const optsettings<T> &o){
    T a_ = a, b_ = b, fa, fb;
    T half = (T) 1 / (T) 2;
    int iter = 1;
    fa = f(a);
    fb = f(b);
    return bisectcore(f,a_,b_,o,fa,fb,iter,half);
}

template <class T>
T bisectcore(T (*f)(T x),T& a,T& b, const optsettings<T> &o, T& fa, T& fb, int& iter, const T& half){
    
    // display info
    if (iter%o.pmod==0) std::cout<<"bisect: iter "<<iter<<" a "<<a<<" b "<<b<<" fa "<<fa<<" fb "<<fb<<std::endl;

    // check that the function is bracketed
    if ((fa>0 && fb > 0) || (fa<0 && fb<0)) STOP("Root is not bracketed",__FILE__,__LINE__);
    
    // check termination criterion
    if (fabs(a-b)<o.xtoler || iter>o.maxiter) return (fabs(fa)<fabs(fb) ? a : b );

    // construct midpoint and evalaute
    T m = half*a + half*b;
    T fm = f(m); 
    
    // call bisect again
    if (fm*fa>0){
        return bisectcore(f,m,b,o,fm,fb,++iter,half);
    } else {
        return bisectcore(f,a,m,o,fa,fm,++iter,half);
    }

}


template <class T>
T bisect(T (*f)(T x),T a,T b, const optsettings<T> &o);

template <class T>
T bisectcore(T (*f)(T x),T& a,T& b, const optsettings<T> &o, T& fa, T& fb, int& iter, const T& half);

double testfunctions_squaredmin2(double x);

int testbisect();



template <class T>
__host__ void disp(std::string str, devicearray<T> &x);

int mtimes(devicearray<float> &c, devicearray<float> &a, devicearray<float> &b, char tA, char tB, float alpha, float beta, cublasHandle_t handle);
void mtimes(array<double>& c,array<double>& a,array<double>& b, char ta, char tb, double alpha, double beta);
void mtimes(double *c, double *a, double *b, int na1, int na2, int nb1, int nb2, char ta, char tb, double aplha, double beta);
void mtimes(double *c, double *a, double *b, int na1, int na2, int nb1, int nb2, double aplha, double beta);
void mtimes(float *c, float *a, float *b, int na1, int na2, int nb1, int nb2, char ta, char tb, double aplha, double beta);
void mtimes(float *c, float *a, float *b, int na1, int na2, int nb1, int nb2, double aplha, double beta);

void testbellman();
void testcubellman();

template <class T>
__global__ void global_zeros(T *x, int n){
    int i = threadIdx.x + blockDim.x * blockIdx.x;
    if (i<n)
        x[i] = 0;
}

template <class T>
void zeros(devicearray<T> &x){
    int nThreadsPerBlock = 1024;
    int nBlocks = 1+(x.size-1)/1024;
    global_zeros<<<nBlocks,nThreadsPerBlock>>>(x.data,x.size);
}

// inplace absolute value
template <class T>
__global__ void global_abs(T *x,int n){
    int i = threadIdx.x + blockDim.x * blockIdx.x;
    if (i<n)
        x[i] = abs(x[i]);
}

template <class T>
__host__ void abs(devicearray<T> &x){
    int nThreadsPerBlock = 1024;
    int nBlocks = 1+(x.size-1)/1024;
    global_abs<<<nBlocks,nThreadsPerBlock>>>(x.data,x.size);
}

// inplace max (in Fortran sense). Sets x = max(x,c) element by element.
// This is useful for maxval, enabling it to be solved by cublas.
__global__ void global_max(float *x, float c);
__host__ void max(devicearray<float> &x, float c);



// Use binary search to find the lowerbound on a bracketing interval for 
// Given sorted ascending array, find lowerbound on bracketing interval. Note that
// ilo will always be in 1,...,size(x)-1.
// 
// This is supposed to work much like the C++ standard library where the interface comes from
template <class T>
__host__ __device__ int bsearch(T xhat,T *x,int n){
    int ilo,ihi,mid;
        
    ilo = 0;
    ihi = n-1;
    mid = (ihi+ilo)/2; // mid will have ilo<=mid<ihi as long as ihi>ilo. If ihi=ilo, then mid=ilo, so the 
                       // while loop is not entered and ilo=1 is returned 
    
    while (mid>ilo){ // mid=ilo means ihi=ilo+1.

        // Here, we must have ihi>ilo and ilo<=mid<ihi

        // Bisect. 
        // If xhat<x(1), then xhat>=x(mid) will never occur and ilo will never be modified.
        // If xhat>=x(nx), then xhat>=x(mid) will always occur (note mid<ihi). Consequently, 
        //      ilo = mid occurs until mid==ilo, which only occures if ihi==ilo+1.
        if (xhat>=x[mid]) 
            ilo = mid;
        else 
            ihi = mid;

        mid = (ihi+ilo)/2;

    }

    return ilo;

}

template <class T>
__host__ __device__ void interp1qlocweight(int &ilo, T &mlo, T xi, T *x, int n){
    ilo = bsearch(xi,x,n); 
    mlo = (x[ilo+1]-xi)/(x[ilo+1]-x[ilo]);
    return;
}
template <class T>
__host__ __device__ T interp1q(T *x,T *y,T xi, int nx){
    int ilo;
    T mlo, one = 1;

    // Get interpolation location and weights
    interp1qlocweight(ilo,mlo,xi,x,nx);

    // Perform the interpolation
    return y[ilo]*mlo + y[ilo+1]*(one - mlo);

}


template <class T>
T rand01(){
    return (T)rand() / (T)RAND_MAX;
}

template <class T>
void rand01(T *x, int n){
    int i;
    for (i=0; i<n; i++)
        x[i] = rand01<T>();
}


template <class T>
__host__ __device__ void eye(T *x, int n){
    T zero = 0, one = 1;
    int i;
    for (i=0;i<n*n;i++)
        x[i] = zero;

    for (i=0;i<n;i++)
        x[i + i*n] = one;
}

template <class T>
__global__ void global_linspace(T *x, T a, T b, int n){
    if (blockIdx.x==0) 
        linspace(x,a,b,n);

}

template <class T>
void linspace(devicearray<T> &x, T a, T b){
    global_linspace<<<1,1>>>(x.data,a,b,x.size);
}

template <class T>
__global__ void global_tauchen(T *z, T *Fzz, T cover, T ar1, T cdnalstd, int n){
    if (blockIdx.x==0) 
        tauchen(z,Fzz,cover,ar1,cdnalstd,n);
}

template <class T>
void tauchen(devicearray<T> &z, devicearray<T> &Fzz, T cover, T ar1, T cdnalstd){
    global_tauchen<<<1,1>>>(z.data,Fzz.data,cover,ar1,cdnalstd,z.size);
}
