! subroutine sub_powell(x,f,toler_in,xtoler_in,maxiter_in,printmod_in,eps_in,maxfev_in,factor_in,fjac_in,info_out,uid_in,derivstep_in)
    use omp_lib
    implicit none   
    procedure(rootN_f_template) :: f
    procedure(rootN_fjac_template), optional :: fjac_in
    real(8), dimension(1:), intent(inout) :: x    
    !optional parameters
    real(8), intent(in), optional :: toler_in, xtoler_in, eps_in, factor_in, derivstep_in(:)
    integer, intent(in), optional :: maxiter_in, printmod_in, maxfev_in, uid_in
    integer, intent(out), optional :: info_out
    !local 
    real(8) :: toler, xtoler, eps, factor
    integer :: maxiter, printmod, maxfev, mode, info, uid
    ! more local
    procedure(rootN_f_template), pointer :: pow_f_ptr => NULL()
    procedure(rootN_fjac_template), pointer :: pow_fjac_ptr => NULL()
    integer :: pow_f_eval_count, pow_f_printmod
    real(8) :: pow_f_toler
    real(8) :: pow_bestx(1000) ! max size

    ! Error check
    if (omp_get_num_threads()/=1) STOP 'ERROR: sub_powell might not be thread safe, '//&
                                    'depends on handling of threadprivate procedure pointers'
    if (size(pow_bestx)<size(x)) STOP 'ERROR: sub_powell: size of pow_bestx must be increased'
    
    ! Point the module variable pow_f_ptr to the function f
    pow_f_ptr => f 
    if (present(fjac_in)) pow_fjac_ptr => fjac_in
    if (present(derivstep_in)) stop 'abandoned this b/c internal procedure' !pow_fjac_ptr => finiteDiffJacob

    !set defaults for optional parameters
    toler = 1d-12   !function termination criterion |f|<toler
    eps = 1d-12    !precision of f function.  if <=0, then use machine precision ??or is this the tolerence??
    xtoler = 1d-12   !x termination criterion. it is relative the values of x
    maxiter = 200  !max # of iterations in powell
    maxfev = 200*size(x) !max # of function evaluations
    printmod = 1   !print modulus.  if -1, don't print anything.    
    factor = 100d0 !Used in determining initial step bound.  Should be in (.1d0,100d0).  100d0 recommended.
    mode = 1       !mode 1 <=> var scaled internally, mode 2 <=> scale specified by input diag
    uid = 6        ! Write to stdout by default

    !override default values if argument present
    if (present(toler_in)) toler = toler_in
    if (present(xtoler_in)) xtoler = xtoler_in
    if (present(eps_in)) eps = eps_in
    if (present(maxfev_in)) maxfev = maxfev_in
    if (present(maxiter_in)) maxiter = maxiter_in
    if (present(printmod_in)) printmod = printmod_in
    if (present(factor_in)) factor = factor_in
    if (present(uid_in)) uid = uid_in
       
    !check for inappropriate parameters
    if (toler<=0d0 .OR. xtoler<=0d0) STOP 'ERROR: POW: inappropriate parameters'

    pow_f_eval_count = 0
    pow_f_printmod = printmod
    pow_f_toler = toler
    if (.not. present(fjac_in) .and. .not. present(derivstep_in)) then
        ! Finite difference jacobian, call hybrd
        call hybrd_driver(x,xtoler,maxfev,eps,mode,factor,0) ! the 0 is a print control
    else
        ! Analytic jacobian, call hybrj
        call hybrj_driver(x,xtoler,maxfev,mode,factor,1) ! the 1 is a print control
    end if

    ! Output info if requested
    if (present(info_out)) info_out = info

contains 
    function finiteDiffJacob(x) result(Jac)
        implicit none
        real(8), intent(in) :: x(:)
        real(8) :: Jac(size(x),size(x))
        ! local
        real(8) :: xph(size(x)),fx(size(x)),fxph(size(x))
        integer :: j,n

        n = size(x)

        ! Evaluate f at x (this is wasteful but I don't have an expedient way to not do it currently)
        fx = f(x)

        ! Compute the Jacobian
        do j=1,n

            xph = x
            xph(j) = x(j) + derivstep_in(j)
            
            fxph = f(xph)

            Jac(:,j) = (fxph - fx)/derivstep_in(j)

        end do

    end function 

    ! To call this routine, the user should have already pointed pow_f_ptr
    ! to the correct function.
    subroutine hybrd_driver(x,xtol,maxfev,epsfcn,mode,factor,nprint)
        implicit none
        real(8), intent(inout) :: x(1:)
        integer, intent(in) :: maxfev,mode,nprint
        real(8), intent(in) :: xtol,factor,epsfcn
        ! local
        integer :: n,ldfjac,nfev,lr,ml,mu
        real(8), dimension(:), allocatable :: fvec,r,qtf,wa1,wa2,wa3,wa4,diag
        real(8), dimension(:,:), allocatable :: fjac

        n = size(x)
        ldfjac = n
        lr = (n*(n+1))/2

        ! For non-banded jacobian, set ml and mu to at least n-1
        ml = n-1
        mu = n-1
        
        allocate(fvec(n),fjac(ldfjac,n),r(lr),diag(n),&
                 qtf(n),wa1(n),wa2(n),wa3(n),wa4(n))

        if (pow_f_printmod>0) then 
            write(uid,*) 'POW:________________________________________________________________'
            write(uid,*) 'POW: Beginning Powell''s Hybrid Method                              '
            write(uid,*) 'POW:   iter       normx       normf'
        end if

        call hybrd(hybrd_private,n,x,fvec,xtol,maxfev,ml,mu,epsfcn,diag,mode, &
                   factor,nprint,info,nfev,fjac,ldfjac,r,lr, &
                   qtf,wa1,wa2,wa3,wa4)

        !output results    
        if (pow_f_printmod>0) then
            select case (info)
                case (0)
                    write(uid,*) 'POW: Improper input parameters.'
                case (1)                
                    write(uid,*) 'POW: Successfully converged: relative error between two iterates is < xtol'
                case (2)
                    write(uid,*) 'POW: Failed to converge: max fevals reached '
                case (3)
                    write(uid,*) 'POW: Failed to converge: xtol is too small, no further improvement possible '
                case (4)
                    write(uid,*) 'POW: Failed to converge: iteration is not making good progress'
                    write(uid,*) 'POW: as measured by the improvement from the last five jacobian evals.'
                case (5)
                    write(uid,*) 'POW: Failed to converge: iteration is not making good progress'
                    write(uid,*) 'POW: as measured by the improvement from the last ten iterations.'
                case (-1)
                    ! This is a user-defined (me) case
                    x = pow_bestx(1:n)
                    write(uid,*) 'POW: Successfully converged: max|f|<toler. Note: last output may not '
                    write(uid,*) 'POW: correspond to final x.'
            end select

        elseif (pow_f_printmod==0 .and. abs(info) /= 1) then        

            write(uid,*) 'POW: Error in powell.  Set printmod>0 for more info.  Exit flag:', info

        endif

    end subroutine 
    ! Instead of passing an internal function, could make this external with a function pointer 
    ! as a common object. The function pointer would have to be a thread private object in order to be safe.
    subroutine hybrd_private(n,x,fvec,iflag)
        integer :: n, iflag
        real(8) :: x(n),fvec(n)

        fvec = pow_f_ptr(x)
        pow_f_eval_count = pow_f_eval_count  + 1

        if (pow_f_printmod>0 .and. mod(pow_f_eval_count,pow_f_printmod)==0) then
            write(uid,'(A,i6,2d12.5)') ' POW: ',pow_f_eval_count,sqrt(sum(x**2)),sqrt(sum(fvec**2))
        end if

        ! Also check toler termination criterion (not done in hybr*)
        if (maxval(abs(fvec))<pow_f_toler) then
            pow_bestx(1:n) = x ! save best x
            iflag = -1 ! tells hybr* to exit
        end if

    end subroutine hybrd_private


    ! To call this routine, the user should have already pointed pow_f_ptr
    ! and pow_fjac_ptr to the correct functions.
    subroutine hybrj_driver(x,xtol,maxfev,mode,factor,nprint)
        implicit none
        real(8), intent(inout) :: x(1:)
        integer, intent(in) :: maxfev,mode,nprint
        real(8), intent(in) :: xtol,factor
        ! local
        integer :: n,ldfjac,nfev,njev,lr
        real(8), dimension(:), allocatable :: fvec,r,qtf,wa1,wa2,wa3,wa4,diag
        real(8), dimension(:,:), allocatable :: fjac

        n = size(x)
        ldfjac = n
        lr = (n*(n+1))/2
        
        allocate(fvec(n),fjac(ldfjac,n),r(lr),diag(n),&
                 qtf(n),wa1(n),wa2(n),wa3(n),wa4(n))

        if (pow_f_printmod>0) then 
            write(uid,*) 'POWJ:________________________________________________________________'
            write(uid,*) 'POWJ: Beginning Powell''s Hybrid Method                              '
            write(uid,*) 'POWJ:   iter       normx       normf'
        end if

        ! Call hybrj
        call hybrj(hybrj_private,n,x,fvec,fjac,ldfjac,xtol,maxfev,diag,&
                   mode,factor,nprint,info,nfev,njev,r,lr,qtf,&
                   wa1,wa2,wa3,wa4)

        ! Report
        if (pow_f_printmod>0) then
            select case (info)
                case (0)
                    write(uid,*) 'POWJ: improper input parameters.'
                case (1)
                    write(uid,*) 'POWJ: relative error between two consecutive iterates'//&
                           'is at most xtol.'
                case (2)
                    write(uid,*) 'POWJ: number of calls to fcn with iflag = 1 has'//&
                           'reached maxfev.'
                case (3)
                    write(uid,*) 'POWJ: xtol is too small. no further improvement in'//&
                           'the approximate solution x is possible.'
                case (4)
                    write(uid,*) 'POWJ: iteration is not making good progress, as'//&
                           'measured by the improvement from the last'//&
                           'five jacobian evaluations.'
                case (5)
                    write(uid,*) 'POWJ: iteration is not making good progress, as'//&
                           'measured by the improvement from the last'//&
                           'ten iterations.'
                case (-1)
                    ! This is a user-defined (me) case
                    x = pow_bestx(1:n)
                    write(uid,*)  'POWJ: Successfully converged: max|f|<toler. Note: last output may not '
                    write(uid,*)  'POWJ: correspond to final x.'
                case default
                    STOP 'POWJ: info case not found'
            end select
        elseif (pow_f_printmod==0 .and. abs(info) /= 1) then        
            write(uid,*) 'POWJ: Error in powell.  Set printmod>0 for more info.  Exit flag:', info
        end if

    end subroutine 

    ! A subroutine conforming to the interface of hybrj
    subroutine hybrj_private(n,x,fvec,fjac,ldfjac,iflag)
        use mod_matlab, only: disp
        implicit none
        integer :: n,ldfjac,iflag
        real(8) :: x(n),fvec(n),fjac(ldfjac,n)

        ! if iflag = 1 calculate the functions at x and
        ! return this vector in fvec. do not alter fjac.
        ! if iflag = 2 calculate the jacobian at x and
        ! return this matrix in fjac. do not alter fvec.

        if (iflag==1) then
            fvec = pow_f_ptr(x)
            pow_f_eval_count = pow_f_eval_count + 1

            ! Also check toler termination criterion (not done in hybr*)
            if (maxval(abs(fvec))<pow_f_toler) then
                pow_bestx(1:n) = x
                iflag = -1 ! tells hybr* to exit
            end if
        elseif (iflag==2) then
            fjac = pow_fjac_ptr(x)
            pow_f_eval_count = pow_f_eval_count + 1 ! not right to add 1, but easy with printmod
        elseif (iflag==0) then
            if (pow_f_printmod>0 .and. mod(pow_f_eval_count,pow_f_printmod)==0) then
                write(uid,'(A,i6,2d12.5)') ' POWJ: ',pow_f_eval_count,sqrt(sum(x**2)),sqrt(sum(fvec**2))
            end if
        end if

    end subroutine 


    
end subroutine  
