! Grey Gordon, 2014. Soli Deo Gloria.
module mod_binarymono
    use mod_types
    implicit none
contains

    ! Solves V(i) = max_{i' in 1,...n'} f(i,i') with policy function g(i)
    ! 
    ! monotone | monotonicity speedup
    !        0 | brute force (do not assume monotone)
    !        1 | simple monotonicity 
    !        2 | binary monotonicity
    ! 
    !  concave | concavity speedup 
    !        0 | brute force (do not assume concave)
    !        1 | simple concavity
    !        2 | binary concavity
    !
    subroutine solvemax(V,g,f,np,monotone,concave)
        implicit none
        interface
            function f(i,ip) 
                use mod_types
                implicit none
                integer, intent(in) :: i,ip 
                real(rt) :: f
            end function 
        end interface
        real(rt), intent(out) :: V(:)
        integer, intent(out) :: g(:)
        integer, intent(in) :: np, monotone, concave
        ! local
        integer :: n
        
        n = size(V)
        
        select case (monotone)
            case (0)
                call brutemono()
            case (1)
                call simplemono()
            case (2)
                call binarymono()
        end select

    contains

        subroutine brutemono()
            implicit none
            ! local
            integer :: i

            do i = 1,n
                call maxstep(i,1,np)
            end do

        end subroutine 
        subroutine simplemono()
            implicit none
            ! local
            integer :: i

            call maxstep(1,1,np)
            do i = 2,n
                call maxstep(i,g(i-1),np)
            end do

        end subroutine 

        subroutine binarymono()
            implicit none
            integer :: k !,ind
            integer, allocatable :: l(:),u(:)

            allocate(l(1+ceiling(log(real(n-1,8))/log(2d0))))
            allocate(u(1+ceiling(log(real(n-1,8))/log(2d0))))

            call maxstep(1,1,np)
            call maxstep(n,g(1),np)

            l(1) = 1
            u(1) = n

            k = 1
            
            ! Main loop (step 1 completed)
            outer: do

                ! Expand list (step 2)
                do 
                
                    ! Check stopping criterion for list expansion
                    if (u(k)==l(k)+1) exit

                    k = k + 1

                    l(k) = l(k-1)
                    u(k) = (l(k-1)+u(k-1))/2

                    ! Compute g for u(k) searching from g(l(k-1)) to g(u(k-1)) 
                    call maxstep(u(k),g(l(k-1)),g(u(k-1)))

                end do

                ! Check whether should move to a higher interval (step 3) or stop the algorithm
                do 

                    if (k==1) exit outer

                    if (u(k)/=u(k-1)) exit

                    k = k - 1

                end do

                ! Move to the right subinterval (step 4)
                l(k) = u(k)
                u(k) = u(k-1)

            end do outer


        end subroutine

        subroutine maxstep(i,a,b)
            implicit none
            integer, intent(in) :: i,a,b

            if (concave==0) then
                call bruteconc(i,a,b)
            elseif (concave==1) then
                call simpleconc(i,a,b)
            elseif (concave==2) then
                call binaryconc(i,a,b)
            end if

        end subroutine

        subroutine binaryconc(i,a_,b_)
            implicit none
            ! local
            integer, intent(in) :: i,a_,b_
            integer :: a,b,m !,n
            real(rt) :: fa,fb,fm,fmp
            logical :: faknown,fbknown

            a = a_
            b = b_

            ! Handle degenerate case
            if (a==b) then
                g(i) = a
                V(i) = f(i,a)
                return
            end if

            faknown = .false.
            fbknown = .false.

            do 

                ! "m" is used both as "n", the cardinality of {a,...,b} and as the midpoint of a,b
                m = b-a+1

                if (m==2) then

                    ! Compute values if not known
                    if (.not. faknown) fa = f(i,a)
                    if (.not. fbknown) fb = f(i,b)

                    ! Compute best of a,b
                    if (fa>=fb) then
                        g(i) = a
                        V(i) = fa
                    else
                        g(i) = b
                        V(i) = fb
                    end if

                    return
                    
                elseif (m==3) then

                    m = (a+b)/2
                    fm = f(i,m) 

                    if (.not. (faknown .or. fbknown)) then
                        fa = f(i,a)
                        faknown = .true.
                    end if

                    if (faknown) then
                        if (fa>fm) then
                            V(i) = fa
                            g(i) = a
                            return
                        else
                            a = m
                            fa = fm
                            faknown = .true.
                        end if
                    else ! fbknown --- must be true here
                        if (fb>fm) then
                            V(i) = fb
                            g(i) = b
                            return
                        else
                            b = m
                            fb = fm
                            fbknown = .true.
                        end if
                    end if

                else

                    m = (a+b)/2
                    fm = f(i,m)
                    fmp = f(i,m+1)

                    if (fm>=fmp) then
                        b = m
                        fb = fm
                        fbknown = .true.
                    else
                        a = m+1
                        fa = fmp
                        faknown = .true.
                    end if

                end if

            end do

        end subroutine

        subroutine bruteconc(i,a,b)
            implicit none
            ! local
            integer, intent(in) :: i,a,b
            integer :: j 
            real(rt) :: tmp

            V(i) = f(i,a)
            g(i) = a
            do j = a+1,b
                tmp = f(i,j)
                if (tmp>V(i)) then
                    V(i) = tmp
                    g(i) = j
                end if
            end do

        end subroutine

        subroutine simpleconc(i,a,b)
            implicit none
            ! local
            integer, intent(in) :: i,a,b
            integer :: j 
            real(rt) :: tmp

            V(i) = f(i,a)
            g(i) = a
            do j = a+1,b
                tmp = f(i,j)
                if (tmp>=V(i)) then
                    V(i) = tmp
                    g(i) = j
                else
                    return
                end if
            end do

        end subroutine

    end subroutine


end module
