! Module containing sub_crs, the Fortran routine used by 
! "A Practical Approach to Testing Calibration Strategies" (Yongquan Cao and Grey Gordon)
! See sub_crs for the main documentation of the routine. 
! 
! Grey Gordon, 2016. SDG.
module mod_crs
    use mod_types
    use mod_util
    use mod_mympi
    use mod_mpislave
    implicit none

    ! Options for minNdim 
    type type_crs_opts 
        ! Commonly used ones
        integer :: printmod = 1, maxiter = 10000
        real(rt) :: xabstoler = 1e-6_rt !, xreltoler = 1e-6_rt
        real(rt) :: yabstoler = 1e-10_rt, yreltoler = 1e-10_rt
        ! Others (maybe only a few algorithms use them)
        ! For these, -1 means there is no default specified
        integer :: algorithm = 1
        integer :: poolsize = -1
        integer :: sleeptime = -1 
        integer :: displaybestx = 1 ! 1==means true (the default)
    end type 

    private
    public :: sub_crs, test_crs

contains

    ! Controlled random search --- MPI version
    ! 
    ! CRS attempts to solve min f(x) s.t. x in [a,b] (where all are in Rn)
    ! or simply evaluates f at an arbitrary number of random x draws from [a,b]
    ! 
    ! When minimizing, it first generates a random pool of points (the number 
    ! is controlled by the poolsize option in opts). Then, it tries to improve
    ! this pool by taking the worst point and trying to get rid of it. This is
    ! done by reflecting it through a simplex of points (like in Nelder-Mead).
    ! 
    ! This simplex is chosen by randomly selecting n+1 points (whose function 
    ! values are known) with replacement. The centroid of the simplex is then
    ! used for reflecting the worst point. 
    !
    ! If the reflected point improves on the worst point (and is in a,b), it is
    ! accepted. O/w, the process repeats.
    ! 
    ! This is modified for parallel purposes by attempting to simultaneously get 
    ! rid of several of the worst points simultaneously. 
    !
    ! The optional argument sleeptime is useful when there are only a few ranks. 
    ! It specifies how long the master should sleep if no free workers are found. This
    ! keeps processing cycles free for those ranks actually doing work. If using 
    ! sleeptime>0, the number of ranks should typically be one more than the physical 
    ! number of cores.
    !
    ! If e_in is present, then random draws will not be taken uniformly from [a,b], but rather
    ! from [xstar-e,xstar+e] (component-wise). Combined with a small poolsize (which can be set 
    ! to 0 which will then be overridden), this makes the routine effectively like a parallel 
    ! version of Nelder-Mead.
    !
    ! algorithm
    ! 0           : no optimization : just computes the pool of points and stops. This is used for just testing calibration strategies.  
    ! 1 (default) : crs1 algorithm : purely global with random simplex chosen irrespective of best point and random simplex.
    ! 2           : crs2 algorithm : when random trial point fails, local mutation is used to potentially move towards the best point. 
    ! 
    ! Note: was formerly called sub_parCRS_MPI_v2
    !
    ! INPUT/OUTPUT
    !   xstar : initial guess on input (may not be used depending on algorithm) and final guess on exit
    ! 
    ! INPUT
    !   f       : subroutine f(y,x) with x input and y a vector of outputs. The minimization is done wrt y(1) which will typically be 
    !             the SSRS. However, additional output may be returned which will be saved in a csv file if desired (see "fname" input).
    !   a,b     : lower and upper bound on x space
    !   e_in    : optional, see description above
    !   fname   : optional, gives the path and name where a csv file should be saved (if desired)
    !   xlabels : optional, comma separated list of labels for the x vector (defaults to x1,x2,x3,...)
    !   ylabel  : optional, as above, but for y vector
    !   targets : optional, includes an extra row in the csv file where the x values are listed as . (Stata for missing) and the y values
    !             are listed as targets(1),targets(2),... . The Stata code for "A Practical Approach to Testing Calibration Strategies" 
    !             interprets this row as giving the target moments.
    !   opts    : optional, comma separated list of options. Possible options are xabstoler,yabstoler,yreltoler,printmod,maxiter,poolsize,
    !             sleeptime,displaybestx,algorithm. For instance, opts="printmod=5,poolsize=1000,algorithm=2" will display info every 5th 
    !             time f is computed, use a poolsize of 1000, and use algorithm 2 (described above). 
    ! 
    ! REFERENCE: P. Kaelo and M. M. Ali, "Some variants of the controlled random search algorithm for global optimization," J. Optim. Theory Appl. 130 (2), 253-264 (2006). 
    ! 
    subroutine sub_crs(xstar,f,a,b,e_in,fname,xlabels,ylabels,targets,opts)
        implicit none
        real(rt), intent(inout) :: xstar(:)
        real(rt), intent(in) :: a(:),b(:)
        real(rt), intent(in), optional :: e_in(:), targets(:)
        character(*), intent(in), optional :: fname,xlabels,ylabels,opts
        interface 
            subroutine f(y,x) 
                use mod_types
                implicit none
                real(rt), dimension(:), intent(out), allocatable :: y
                real(rt), dimension(:), intent(in) :: x
            end subroutine
        end interface
        ! local other
        type(type_seed) :: randstate
        real(rt), allocatable :: xtmp(:)
        real(rt), allocatable :: X(:,:),Y(:,:) 
        real(rt), allocatable :: R(:),g(:),smx(:,:)
        real(rt) :: pool_dist
        integer :: nx, N, M, iter, ism, who, itmp(1), flag, effPoolSize, ibest, iworst, smx_iter, uid
        type(type_crs_opts) :: o
        type(type_gang) :: gang

        if (present(opts)) o = optimset(opts)

        if (o%algorithm<0 .or. o%algorithm>2) then
            write(0,*) 'sub_crs: selected algorithm not valid, use 0, 1 or 2; selecting algorithm==1'
            o%algorithm = 1
        end if

        if (mympi_nproc()<=1) then
            write(0,*) 'WARNING: sub_crs: called with only one rank, aborting'
            return
        end if

        nx = size(xstar)
        if (size(a)/=nx) stop 'ERROR: sub_crs: size of a and x do not agree'
        if (size(b)/=nx) stop 'ERROR: sub_crs: size of b and x do not agree'
        if (any(a>=b)) stop 'ERROR: sub_crs: a>=b at least once'

        ! Check that input x is in a,b for all i
        call check_guess_in_range(flag,xstar,a,b)
        if (flag/=0) then
            write(0,*) 'WARNING: sub_crs: x is not in a,b or b<a for some dimension. Will bound and continue'
            xstar = min(b,max(a,xstar))
        end if 

        ! M is the number to simultaneously try to get rid of
        M = mympi_nproc()-1

        ! N is number of trial points (the pool of points)
        ! N = max(2*M, ((100+nx)/M+1)*M, nx+2+M)
        N = max(((100+nx)/M+1)*M, nx+2+M)
        ! N = ((100+nx)/M+1)*M ! roughly evenly divide 100+nx points across the M ranks
        if (o%poolsize/=-1) N = max(nx+2+M,o%poolsize)

        if (N-M<=nx+1) call mympi_stop('ERROR: sub_crs: too few N points set relative to M')

        ! If not master, go to the "slave" routine and wait for a x to evaluate
        if (mympi_iproc()/=0) then
            call slave(f,nx)
            call mympi_bcast(xstar,0) ! receive the best point 
            call mympi_barrier()
            return
        end if

        ! If still here, then you are the master. Create the gang of workers.
        call gang%init(nx,o%sleeptime)

        ! Generate N random points uniformly from [a,b]
        allocate(xtmp(nx))
        allocate(X(nx,N))
        allocate(R(nx))
        allocate(g(nx))
        allocate(smx(nx,nx+1))                    

        ! Randomize for the master
        call randstate%randomize()

        ! Until have reached poolsize, expand
        effPoolSize = 0
        do iter = 1,o%maxiter

            ! Get the best point and the worst point
            if (effPoolSize>1) then
                itmp = minloc(Y(1,1:effPoolSize))
                ibest = itmp(1)
                itmp = maxloc(Y(1,1:effPoolSize))
                iworst = itmp(1)

                ! Report 
                pool_dist = norm2(maxval(X,2)-minval(X,2))
                if (o%printmod>0 .and. mod(iter,o%printmod)==0) then
                    write(*,'(A,i8,3(A,ES14.6),2A)') 'sub_crs: iter ',iter,' pool dist ',pool_dist,' worst ',Y(1,iworst),' best ',Y(1,ibest),'  xbest = ',printx(X(:,ibest))
                    if (effPoolSize==2 .or. mod(iter,100*o%printmod)==0) then
                        call report()
                    end if
                end if

            end if

            ! Check stopping criteria
            if (effPoolSize==N) then
                if (o%algorithm==0) then
                    write(*,'(A)') 'sub_crs: select algorithm was 0 and pool size met, so exiting.'
                    exit
                end if
                if (pool_dist<o%xabstoler) then
                    write(*,'(A,ES14.6,A)') 'sub_crs: xabstoler of ',o%xabstoler,' met.'
                    exit
                end if
                if (abs(Y(1,iworst)-Y(1,ibest))<o%yreltoler*.5*abs(Y(1,iworst)+Y(1,ibest))) then
                    write(*,'(A,ES14.6,A)') 'sub_crs: yreltoler of ',o%yreltoler,' met.'
                    exit
                end if
                if (abs(Y(1,iworst)-Y(1,ibest))<o%yabstoler) then
                    write(*,'(A,ES14.6,A)') 'sub_crs: yabstoler of ',o%yabstoler,' met.'
                    exit
                end if
            end if

            ! Generate a point
            if (effPoolSize<N) then

                ! If pool size is still not large enough, generate a random point.
                ! If local algorithm and the first iteration, include xstar in the pool.
                ! Otherwise, take a draw from uniform distibution over [a,b].
                if (iter==1 .and. o%algorithm==2) then
                    xtmp = xstar
                else
                    call random_number(xtmp)
                    if (.not. present(e_in)) then
                        xtmp = xtmp*a + (1._rt-xtmp)*b
                    else
                        xtmp = xtmp*max(a,xstar-e_in) + (1._rt-xtmp)*min(b,xstar+e_in)
                    end if
                end if

            else

                do smx_iter = 1,100

                    ! If here, at the refinement stage. So take the simplex of points and pick a better one
                    select case (o%algorithm)
                        case (1)
                            ! Choose a truly random simplex
                            do ism = 1,nx+1
                                smx(:,ism) = x(:,randi(1,N)) 
                            end do
                        case (2)
                            do ism = 1,nx+1
                                ! CRS with local mutation where simplex always includes best point
                                if (ism==1) then
                                    smx(:,ism) = x(:,ibest) 
                                else
                                    do 
                                        itmp(1) = randi(1,N) 
                                        if (ibest/=itmp(1) .and. iworst/=itmp(1)) exit
                                    end do
                                    smx(:,ism) = X(:,itmp(1))
                                end if
                            end do
                    end select

                    ! Compute the centroid
                    g = sum(smx,2)/(nx+1)

                    ! Reflect the worst point
                    R = 2._rt*g - X(:,iworst) ! pure reflection

                    ! If reflection point is interior, use it. 
                    ! If this is not satisfied before the max iterations is reached, then R will be used, but bounded
                    if (all(R>=a) .or. any(R<=b)) exit

                end do

                ! If goes through boundary, then bound at boundary
                xtmp = min(max(R,a),b)

            end if

            ! Here, xtmp is the point that should be assigned

            ! Wait for a free worker
            who = gang%wait_for_free()

            ! If ever previously assigned, use that value
            if (gang%slaves(who)%ever_assigned) then

                if (isnan(gang%slaves(who)%y(1))) then
                    write(0,*) 'WARNING: sub_crs: function returned a NaN value for x :',printx(gang%slaves(who)%x),' replacing with huge(0d0)'
                    gang%slaves(who)%y(1) = huge(0d0)
                end if

                if (effPoolSize<N) then

                    ! Add to pool
                    effPoolSize = effPoolSize + 1
                    X(:,effPoolSize) = gang%slaves(who)%x
                    if (.not. allocated(Y)) then

                        ! Now, the size of Y can be inferred, so allocate it.
                        allocate(Y(size(gang%slaves(who)%y),size(X,2)))

                        ! If fname is present, then export the x and y values to file.
                        if (present(fname)) then
                            call opencsv(uid,'replace')
                            call writeheader(uid)
                            close(uid) ! close the file, and reopen it later
                        end if

                    end if
                    Y(:,effPoolSize) = gang%slaves(who)%y

                    if (effPoolSize==N) write(*,'(A)') 'sub_crs: achieved full pool, moving to refinement stage'

                else

                    ! Replace the worst point if found a better one
                    if (gang%slaves(who)%y(1)<Y(1,iworst)) then

                        X(:,iworst) = gang%slaves(who)%x
                        Y(:,iworst) = gang%slaves(who)%y

                    else

                        if (o%algorithm==2) then 
                            ! If CRS2 is the algorithm, then in the case of a failure, move randomly close to the best point
                            ! The new point is a sort of random reflection of the reflection through the best point 
                            call random_number(xtmp) 
                            xtmp = (1._rt+xtmp)*X(:,ibest) - xtmp*gang%slaves(who)%x 
                            xtmp = min(max(xtmp,a),b)
                        end if

                    end if

                end if

                ! Save the value to file
                call appendtocsv(gang%slaves(who)%x,gang%slaves(who)%y)

            end if

            ! Assign them a value
            call gang%slaves(who)%assign_x(xtmp)

        end do

        ! Do a final report
        call report()
        if (iter==o%maxiter+1) &
        write(*,'(A,i8)')     'sub_crs: failed to converge in allotted iterations'
        if (iter<=o%maxiter) &
        write(*,'(A,i8)')     'sub_crs: converged successfully'

        ! Output the best
        xstar = X(:,ibest)

        call gang%shutdown()
        call mympi_bcast(xstar,0) ! receive the best point 
        call mympi_barrier()

    contains
    
        subroutine writeheader(uid)
            implicit none
            integer, intent(in) :: uid
            integer :: ix, iy

            ! Header
            if (present(xlabels)) then
                write(uid,'(A)',advance='no') xlabels ! should be comma separated list
                write(uid,'(A)',advance='no') ','
            else
                do ix = 1,nx
                    write(uid,'(A)',advance='no') 'x'//trim(conv_i_to_s(ix))//','
                end do
            end if

            if (present(ylabels)) then
                write(uid,'(A)',advance='no') ylabels
            else
                do iy = 1,size(Y,1)
                    write(uid,'(A)',advance='no') 'stat'//trim(conv_i_to_s(iy))
                    if (iy<size(Y,1)) write(uid,'(A)',advance='no') ','
                end do
            end if
            write(uid,*) 

            ! Row listing targets (if applicable)
            if (present(targets)) then

                if (size(targets)/=size(Y,1)) then
                    write(0,*) 'WARNING: size(targets) is incorrect, does not equal dimension of function''s output variable. Targets will not be written.'
                    return
                end if

                do ix = 1,nx
                    write(uid,'(A2)',advance='no') '.,' ! missing
                end do
                do iy = 1,size(Y,1)
                    write(uid,'(ES24.16E3)',advance='no') targets(iy)
                    if (iy<size(Y,1)) write(uid,'(A)',advance='no') ','
                end do
                write(uid,*) 

            end if

        end subroutine

        ! Computes the mean and variance-covariance matrix of some data x(j,i) where i gives the observation and j gives the variable
        subroutine sub_varcov(mean,varcov,d)
            implicit none
            real(8), intent(in) :: d(:,:)
            real(8), intent(out) :: mean(:),varcov(:,:)
            ! local 
            integer :: i,j,ni,nj,j2

            nj = size(d,1)
            ni = size(d,2)

            if (size(mean)/=nj) stop 'ERROR: sub_varcov: mean has the wrong size'
            if (any(shape(varcov)/=[nj,nj])) stop 'ERROR: sub_varcov: varcov has the wrong size'

            mean = sum(d,2)/ni

            varcov = 0d0
            do i = 1,size(d,2)
                do j = 1,nj
                    do j2 = j,nj
                        varcov(j,j2) = varcov(j,j2) + (d(j,i)-mean(j))*(d(j2,i)-mean(j2))/ni
                    end do
                end do
            end do
            
            ! Now the variance-covariance upper triangular is set
            ! Do the transpose to fill in the lower triangular
            do j = 1,nj
                do j2 = 1,j-1
                    varcov(j,j2) = varcov(j2,j)
                end do
            end do

        end subroutine

        subroutine report()
            implicit none
            real(8), allocatable :: mu(:),sigma(:,:)
            integer :: i

            allocate(mu(size(a)))
            allocate(sigma(size(a),size(a)))

            call sub_varcov(mu,sigma,X(:,1:effPoolSize))


            write(*,'(2A)')   'sub_crs:  ',repeat('*',120)
            write(*,'(A,i3)') 'sub_crs:            algorithm = ',o%algorithm
            if (present(fname)) &
            write(*,'(2A)')   'sub_crs:     x y history file = ',fname
            if (.not. present(fname)) &
            write(*,'(2A)')   'sub_crs:     x y history file = ','(no history file created) '
            write(*,'(A,i8)') 'sub_crs:                 iter = ',iter 
            write(*,'(A,i8)') 'sub_crs:    current pool size = ',effPoolSize
            write(*,'(A,i8)') 'sub_crs:    desired pool size = ',N
            write(*,'(2A)')   'sub_crs:                    a = ',printx(a)
            write(*,'(2A)')   'sub_crs:                    b = ',printx(b)
            if (present(e_in)) &
            write(*,'(2A)')   'sub_crs:                 e_in = ',printx(e_in)
            write(*,'(2A)')   'sub_crs:   element-wise max y = ',printx(maxval(Y(:,1:effPoolSize),2))
            write(*,'(2A)')   'sub_crs:   element-wise min y = ',printx(minval(Y(:,1:effPoolSize),2))
            write(*,'(2A)')   'sub_crs:   element-wise max x = ',printx(maxval(X(:,1:effPoolSize),2))
            write(*,'(2A)')   'sub_crs:   element-wise min x = ',printx(minval(X(:,1:effPoolSize),2))
            write(*,'(2A)')   'sub_crs:         x pool range = ',printx(maxval(X(:,1:effPoolSize),2)-minval(X(:,1:effPoolSize),2))
            write(*,'(2A)')   'sub_crs:  x range % of space  = ',printx((maxval(X(:,1:effPoolSize),2)-minval(X(:,1:effPoolSize),2))/max(b-a,epsilon(0._rt)))
            write(*,'(2A)')   'sub_crs:          x pool mean = ',printx(mu)
            do i = 1,size(a)
            write(*,'(A,I3,2A)')'sub_crs: x pool varcov(',i,',:) = ',printx(sigma(i,:))
            end do
            write(*,'(2A)')   'sub_crs:                ybest = ',printx(Y(:,ibest))
            write(*,'(2A)')   'sub_crs:                xbest = ',printx(X(:,ibest))
            write(*,'(2A)')   'sub_crs:  ',repeat('*',120)
        end subroutine 

        ! Opens fname for writing, returning the unit id corresponding to it. Catches errors in opening.
        subroutine opencsv(uid,status)
            integer, intent(out) :: uid
            character(len=*), intent(in) :: status
            do 
                ! open(newunit=uid,file=fname,action='write',status=status,position='append',err=1234)
                open(newunit=uid,file=fname,action='write',position='append',err=1234)
                exit ! If no error, escape the loop
1234            write (0,*) 'ERROR: sub_crs: File name '//trim(fname)//' could not be created'
                call mympi_abort()
            end do
        end subroutine 

        ! Appends the x and stats values to the csv file
        subroutine appendtocsv(x,y)
            implicit none
            real(8), intent(in) :: x(:),y(:)
            integer :: i

            if (present(fname)) then
                call opencsv(uid,'old')
                do i = 1,nx
                    write(uid,'(ES24.16E3)',advance='no') x(i)
                    write(uid,'(A)',advance='no') ','
                end do
                do i = 1,size(Y,1)
                    write(uid,'(ES24.16E3)',advance='no') y(i)
                    if (i<size(Y,1)) write(uid,'(A)',advance='no') ','
                end do
                write(uid,*) ! new line
                close(uid)
            end if

        end subroutine


    end subroutine


    ! Generate a random integer in a,...,b
    function randi(a,b) result(i)
        implicit none
        integer, intent(in) :: a,b
        integer :: i
        real(rt) :: tmp

        ! Generate a U[a,b+1-epsilon]
        ! Then floor to get a in 
        call random_number(tmp)
        tmp = tmp*(b+1-a) + a ! U[a,b+1]
        tmp = min(tmp,b+.99999999e0_rt)
        i = floor(tmp)

        if (floor(tmp)<a .or. floor(tmp)>b) STOP 'ERROR: randi: failed to generate i in a,...,b'

    end function

    subroutine rosenbrock(y,x)
        implicit none
        real(rt), dimension(:), intent(in) :: x
        real(rt), dimension(:), intent(out), allocatable :: y
        !local
        integer :: i,n
        real(rt) :: f_tmp, tmp

        allocate(y(1))

        call random_number(tmp)
        if (tmp<.0001_rt) then
            tmp = -1._rt
            y = sqrt(tmp)
            return
        end if

        n = size(x)
        f_tmp = 0.e0_rt
        do i = 1,n-1
             f_tmp = f_tmp + 100.e0_rt*(x(i+1)-x(i)**2)**2 + (1.e0_rt-x(i))**2
        end do
        y = f_tmp

!         call sleep(1)

    end subroutine

    subroutine check_guess_in_range(flagout,x,a,b)
        implicit none
        real(rt), intent(in) :: x(:),a(:),b(:)
        integer, intent(out) :: flagout
        ! local
        integer :: i,n,flag

        n = size(x)
        if (size(a)/=n) stop 'ERROR: check_guess_in_range: a and x do not have same length'
        if (size(b)/=n) stop 'ERROR: check_guess_in_range: b and x do not have same length'

        flag = 0
        do i = 1,n 
        if (x(i)<a(i)) then; write(0,*) 'WARNING: check_guess_in_range: x<a for element',i; flag = merge(-1,-999,flag==0); end if
        if (x(i)>b(i)) then; write(0,*) 'WARNING: check_guess_in_range: x>b for element',i; flag = merge(-2,-999,flag==0); end if
        if (b(i)<a(i)) then; write(0,*) 'WARNING: check_guess_in_range: b<a for element',i; flag = merge(-3,-999,flag==0); end if
        end do
        flagout = flag

    end subroutine 


    subroutine test_crs
        implicit none
        real(rt) :: x(3),a(3),b(3),e(3)

        a = 0.2_rt
        b = [1.2e0_rt,3e0_rt,99e0_rt]
        e = [.01_rt,.01_rt,.01_rt]

        print*
        x = [.3e0_rt,.1e0_rt,.5e0_rt]
        call sub_crs(x,rosenbrock,a,b,opts='printmod=100,sleeptime=0,yabstoler=1d-10,maxiter=100000,algorithm=1')

        print*
        x = [.3e0_rt,.1e0_rt,.5e0_rt]
        call sub_crs(x,rosenbrock,a,b,opts='printmod=1000,sleeptime=0,yabstoler=1d-10,maxiter=100000,algorithm=2')

        print*
        x = [.3e0_rt,.1e0_rt,.5e0_rt]
        call sub_crs(x,rosenbrock,a,b,e_in=e,opts='printmod=1000,sleeptime=0,yabstoler=1d-10,maxiter=100000,poolsize=15')

        print*
        call sub_crs(x,rosenbrock,a,b,fname='/tmp/statmat.csv',xlabels='param1,param2,param3',ylabels='SSRS',opts='printmod=1000,sleeptime=0,yabstoler=1d-10,maxiter=100000,poolsize=2')

    end subroutine 

    ! Here, options are specified via a string such as "xtoler=1d6,printmod=10000"
    function optimset(str) result(opts)
        implicit none
        character(len=*), intent(in) :: str
        type(type_crs_opts) :: opts
        ! local
        character(len=len(str)) :: s,arg,val
        type(type_crs_opts) :: def
        integer :: nargs,i

        nargs = count_delim_arg(str,",")

        do i = 1,nargs

            ! Get the substring
            s = ''
            s = trim(get_delim_arg(str,",",i))

            ! Process to get what is left and right of the equal sign
            arg = ''
            val = ''
            arg = get_delim_arg(s,"=",1)
            val = get_delim_arg(s,"=",2)

            if     (trim(arg)=="xabstoler")     then; read(val,'(d30.16)',err=999) opts%xabstoler 
!             elseif (trim(arg)=="xreltoler")     then; read(val,'(d30.16)',err=999) opts%xreltoler 
            elseif (trim(arg)=="yabstoler")     then; read(val,'(d30.16)',err=999) opts%yabstoler 
            elseif (trim(arg)=="yreltoler")     then; read(val,'(d30.16)',err=999) opts%yreltoler 
            elseif (trim(arg)=="printmod")      then; read(val,'(i25)',   err=999) opts%printmod
            elseif (trim(arg)=="maxiter")       then; read(val,'(i25)',   err=999) opts%maxiter
            elseif (trim(arg)=="poolsize")      then; read(val,'(i25)',   err=999) opts%poolsize
            elseif (trim(arg)=="sleeptime")     then; read(val,'(i25)',   err=999) opts%sleeptime
            elseif (trim(arg)=="displaybestx")  then; read(val,'(i25)',   err=999) opts%displaybestx
            elseif (trim(arg)=="algorithm")     then; read(val,'(i25)',   err=999) opts%algorithm
            else
                write(0,*) 'WARNING: optimset: did not find option corresponding to argument ',&
                           trim(arg),'  with value ',trim(val),' and overall input ',trim(str)
            end if

        end do

        return

999     write(0,*) 'ERROR: optimset: error encountered processing ',trim(str),', with arg ',trim(arg),' and val ',trim(val),' falling back to default arguments'
        opts = def
    

    end function



end module
