module mod_gini
    use mod_types
    implicit none
    integer, parameter :: nLorenzPts = 10001

contains 

! Computes an estimate of the gini coefficient via a monte carlo sampling of size k
function fn_gini_approx(k,dsn,grid,n) result(gini)
    use mod_matlab
    implicit none
    real(rt) :: gini
    integer, intent(in) :: k,n
    real(rt), intent(in) :: dsn(*),grid(*) ! size n
    ! local
    real(rt), dimension(:), allocatable :: cdsn,rg
    integer, dimension(:), allocatable :: redo
    real(rt) :: x,SN,SD
    integer :: i,j

    ! Sort the grid
    allocate(cdsn(0:n))
    allocate(rg(k),redo(k))

    cdsn(0) = 0._rt
    do i = 1,n
        cdsn(i) = cdsn(i-1) + dsn(i)
    end do
!     cdsn(0) = 0._rt
!     cdsn(1:n) = cumsum(dsn(1:n))
    if (abs(cdsn(n)-1d0)>sqrt(epsilon(0._rt))) then
        write(0,*) 'ERROR: fn_gini_approx: dsn doesn''t sum to one but to ',cdsn(n),' trying to recover'
    end if
    cdsn(n) = 1._rt

    ! Random sample
    do i = 1,k

        call random_number(x)
        call bsearch(j,x,cdsn)
        rg(i) = grid(j)
        
    end do

    ! Sort rg
    rg = sort(rg)

    ! Compute the gini
    SN = 0._rt
    SD = 0._rt
    do i = 1,k
        SN = SN + i*rg(i)
        SD = SD + rg(i)
    end do

    gini = (2*SN)/(k*SD) - (k+1)/k

end function


! Returns the requested lorenz points, as well as the gini
subroutine sub_lorenzpts(gini,lorenzpts,dsn,grid,n)
    use mod_sort, only: qsort
    use mod_matlab
    implicit none
    real(rt), intent(inout) :: gini, lorenzpts(:)
    real(rt), intent(in) :: dsn(*),grid(*)
    integer, intent(in) :: n
    ! local
    real(rt), dimension(:), allocatable :: Lorenz,Lorenz_grid,sdsn,sgrid
    integer, dimension(:), allocatable :: redo
    integer :: i

    ! Sort the grid
    allocate(sgrid(n),sdsn(n),redo(n))

    sgrid = grid(1:n)
    call qsort(sgrid,redo)
    sdsn = dsn(redo)

    allocate(Lorenz(nLorenzPts),Lorenz_grid(nLorenzPts))

    ! Get the Lorenz curve
    call sub_gini(gini,Lorenz,Lorenz_grid,sdsn,sgrid)

    ! Interpolate 
    if (size(lorenzpts)>0) then
        lorenzpts = interp1(Lorenz_grid,Lorenz,lorenzpts)
    end if

end subroutine 

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! y_in is the value and fy_in is the associated 
! density function.  e.g., a grid on wealth and 
! the invariant distribution of wealth
function fn_gini(y_in,fy_in) result(G)
    use mod_matlab
    implicit none
    real(rt), dimension(:), intent(in) :: y_in,fy_in
    real(rt), dimension(1:size(y_in)):: y,fy
    integer, dimension(1:size(y_in)) :: srt_ind
    real(rt), dimension(1:size(y_in)+1) :: S
    real(rt), dimension(1:size(y_in)) :: Siminus1_Si
    real(rt) :: G
    integer :: n, i
    
    n=size(y)
    y=y_in(1:n)
    fy=fy_in(1:n)
    
    do i=1,n
        srt_ind(i) = i
    end do
    
    call sort2(y,srt_ind,y)
    fy=fy(srt_ind)
    
    S(1)=0.e0_rt
    call sub_cumsum(S(2:n+1),y*fy)
    
    Siminus1_Si=S(2:n+1)+S(1:n)
    
    G=1-dot_product(fy,Siminus1_Si)/S(n+1)

end function fn_gini

subroutine sub_gini(gini,Lorenz,Lorenz_grid,dsn,grid)
    use mod_matlab, only: interp1, linspace, cumsum, issorted, unique
    implicit none
    real(rt), intent(out) :: gini
    real(rt), dimension(nLorenzPts), intent(out)  :: Lorenz,Lorenz_grid
    real(rt), dimension(1:), intent(in) :: dsn,grid
    !local
    real(rt), dimension(:), allocatable :: pct_of_stuff, pct_of_pop, pctpop, pctstuff
    integer, dimension(:), allocatable :: undo, redo

    allocate(pct_of_stuff(size(dsn)+2),pct_of_pop(size(dsn)+2),undo(size(dsn)+2))

    !Check inputs
    if (size(dsn)/=size(grid)) STOP 'sub_gini: dsn and grid size don''t agree'
    if (.not. issorted(grid,'ascend')) STOP 'sub_gini: grid is not sorted'
    
    Lorenz_grid = linspace(0.e0_rt,1.e0_rt,nLorenzPts)
    
    pct_of_stuff(1) = 0.e0_rt
    pct_of_stuff(2:size(dsn)+1) = cumsum(grid*dsn)/dot_product(grid,dsn)
    pct_of_stuff(size(dsn)+2) = 1.e0_rt
    
    pct_of_pop(1) = 0.e0_rt
    pct_of_pop(2:size(dsn)+1) = cumsum(dsn)/sum(dsn)
    pct_of_pop(size(dsn)+2) = 1.e0_rt


!     call unique(pctpop,redo,undo,pct_of_pop)
!     allocate(pctstuff(size(pctpop)))
!     pctstuff = pct_of_stuff(redo)
! 
!     Lorenz = interp1(pctpop,pctstuff,Lorenz_grid)
!     gini = sum(Lorenz_grid-Lorenz)/(sum(Lorenz_grid-Lorenz)+sum(Lorenz))  

    Lorenz = interp1(pct_of_pop,pct_of_stuff,Lorenz_grid)
    gini = sum(Lorenz_grid-Lorenz)/(sum(Lorenz_grid-Lorenz)+sum(Lorenz))  

end subroutine sub_gini


end module mod_gini
