! Based on https://en.wikipedia.org/wiki/Heapsort
module mod_heapsort
    implicit none

    interface sub_heapsort
        module procedure sub_heapsort_1d_rank_inplace,&
                         sub_heapsort_1d_norank_inplace,&
                         sub_heapsort_1d_rank_outplace,&
                         sub_heapsort_1d_norank_outplace
    end interface
contains

    subroutine sub_heapsort_1d_rank_inplace(a,rank)
        implicit none
        real(8), intent(inout) :: a(:)
        integer, intent(inout) :: rank(:)
        call sub_heapsort_work(a,rank,size(a))
    end subroutine 
    subroutine sub_heapsort_1d_norank_inplace(a)
        implicit none
        real(8), intent(inout) :: a(:)
        integer, allocatable :: rank(:)
        allocate(rank(size(a)))
        call sub_heapsort_work(a,rank,size(a))
    end subroutine 

    subroutine sub_heapsort_1d_rank_outplace(a,rank,a0)
        implicit none
        real(8), intent(in) :: a0(:)
        real(8), intent(out) :: a(:)
        integer, intent(inout) :: rank(:)
        a = a0
        call sub_heapsort_work(a,rank,size(a))
    end subroutine 
    subroutine sub_heapsort_1d_norank_outplace(a,a0)
        implicit none
        real(8), intent(in) :: a0(:)
        real(8), intent(out) :: a(:)
        integer, allocatable :: rank(:)
        allocate(rank(size(a0)))
        a = a0
        call sub_heapsort_work(a,rank,size(a))
    end subroutine 



    subroutine sub_heapsort_work(a,rank,n)
        implicit none
        integer, intent(in) :: n ! number of elements of a
        real(8), intent(inout) :: a(0:*) ! 0 to n-1
        integer, intent(inout) :: rank(0:*) ! 0 to n-1, but will return ones-based indices
        ! implicit none
        integer :: finish,i

        do i = 0,n-1 
            rank(i) = i+1 
        end do

        ! Build the heap
        call heapify(a,rank,n)

        ! Now, because of the heap structure, a(0) is the largest element. 
        ! It should be at the end, so the final array will be sorted ascending. 
        ! So do that swap. Then from 0 to one before the end, reheapify and repeat the process.
        finish = n - 1
        do while (finish>0) 
              
            ! Get the largest element from the heap and put it and the end
            call doswap(a(0),a(finish))
            call doswapi(rank(0),rank(finish))

            finish = finish - 1

            ! Because of the swap, have to repair the heap for the remaining elements, so do that
            call siftdown(a,rank,0,finish)

        end do

    end subroutine 

    ! Use zero-based indexing to be consistent with algorithm 
    subroutine heapify(a,rank,n)
        implicit none
        integer, intent(in) :: n ! number of elements of a
        real(8), intent(inout) :: a(0:*) ! 0 to n-1
        integer, intent(inout) :: rank(0:*) ! 0 to n-1
        integer :: start

        start = iparent(n-1)
        do while (start>=0)

            ! Sift down
            call siftdown(a,rank,start,n-1)

            ! Go to the next parent 
            start = start - 1

        end do

    end subroutine 

    ! Repairs the heap assuming the children are heaped
    subroutine siftdown(a,rank,start,finish)
        implicit none
        real(8), intent(inout) :: a(0:*) 
        integer, intent(inout) :: rank(0:*) 
        integer, intent(in) :: start, finish
        integer :: root,child,swap

        root = start

        ! While the root has at least one child
        do while (ilchild(root) <= finish)

            ! Take the left child of the root
            child = ilchild(root)

            ! Keep track of child to potentially swap with
            swap = root

            ! Among the left child, right child, and parent, need the biggest one to be the new parent

            ! If left child is bigger than parent, switch them so parent is bigger
            if (a(swap)<a(child)) swap = child

            ! If right child is bigger than the bigger of left child and parent, then want it to be the parent
            if (child+1<=finish) then
                if (a(swap)<a(child+1)) swap = child+1
            end if

            ! If swap==root, then the parent should be the new biggest parent. So done because all the children
            ! are assumed to be heaped.
            if (swap==root) return
            
            ! If still here, need to swap so new parent is biggest
            call doswap(a(swap),a(root))
            call doswapi(rank(swap),rank(root))

            ! B/c did a swap, need to ensure that heap order is preserved for children. 
            ! So, set root to whichever child was used in the swap and continue
            root = swap

        end do

    end subroutine

    ! test whether heapified
    subroutine test_heapified(a,n)
        implicit none
        integer, intent(in) :: n ! number of elements of a
        real(8), intent(in) :: a(0:*) ! 0 to n-1
        integer :: i
        do i = 0,(n-1)/2
            print*,'i',i
            if (a(ilchild(i))>a(i)) stop 'Not heapified'
            if (ilchild(i)+1<n-1) then
                if (a(ilchild(i)+1)>a(i)) stop 'Not heapified'
            end if
        end do

    end subroutine 

    subroutine doswap(a,b)
        implicit none
        real(8), intent(inout) :: a,b
        real(8) :: tmp
        tmp = a
        a = b
        b = tmp
    end subroutine 

    subroutine doswapi(a,b)
        implicit none
        integer, intent(inout) :: a,b
        integer :: tmp
        tmp = a
        a = b
        b = tmp
    end subroutine 

    integer function iparent(i)
        implicit none
        integer :: i
        iparent = (i-1)/2 ! does floor automatically
    end function 
    integer function ilchild(i)
        implicit none
        integer :: i
        ilchild = 2*i+1
    end function 

end module

module mod_heapsort_test
    use mod_heapsort
    implicit none
contains
    subroutine test_heapsort
        implicit none
        real(8) :: x(9),x0(size(x))
        integer :: rank(size(x)),i,iter

        do iter = 1,10000

            if (iter==1) then
                x0 = [6,5,3,1,8,7,2,4,-30]
            else
                call random_number(x0)
                x0 = (x0-.5d0)*10d0
                if (mod(iter,5)==0) then
                    x0 = real(nint(x0),8)
                end if
            end if

            ! Test in place
            x = x0
            call sub_heapsort(x,rank)
            if (maxval(abs(x0(rank)-x))>10d0*epsilon(0d0)) stop 'sub_heapsort: failed test 1'
            do i = 1,size(x)-1
                if (x(i)>x(i+1)) stop 'sub_heapsort: failed test 2'
            end do

            ! Test out of place
            call sub_heapsort(x,rank,x0)
            if (maxval(abs(x0(rank)-x))>10d0*epsilon(0d0)) stop 'sub_heapsort: failed test 3'
            do i = 1,size(x)-1
                if (x(i)>x(i+1)) stop 'sub_heapsort: failed test 4'
            end do

        end do

        
        write(*,'(A)') 'test_heapsort: passed'
    end subroutine

end module 
