module mod_ident
    use mod_types
    implicit none

    private :: test_function

contains

    !
    subroutine test_function(x,stats) 
        use mod_types
        implicit none
        real(rt), dimension(:), intent(in) :: x
        real(rt), dimension(:), intent(out) :: stats
        ! local
        real(rt) :: err

        if (size(x)/=3) stop 'ERROR: test_function: size(x) should be 3'
        if (size(stats)/=4) stop 'ERROR: test_function: size(stats) should be 4'

        stats(1) = dot_product(x,x)
        stats(2) = log(abs(x(2)))
        stats(3) = log(abs(x(1)+x(2)))
        stats(4) = sum(x) + 1d0

        call random_number(err)
        err = 1d0 +(2d0*err-1d0)/1000d0
        stats = stats * err ! .1% error

    end subroutine

    subroutine test_identranddraws_MPI()
        use mod_matlab
        use mod_mympi
        implicit none
        integer, parameter :: nx = 3, nstats = 4, ndraws = 1000
        real(rt) :: a(nx),b(nx),targets(nstats)
        real(rt), allocatable :: statmat(:,:)

        targets = [25d0,.5d0,1d0,2d0]

        a = -5d0
        b = 5d0

        allocate(statmat(ndraws,nx+nstats))

        call sub_identranddraws_MPI(test_function,nstats,a,b,ndraws,'/tmp/statmat.csv',statmat,&
                                    xlabels='xval1,xval2,xval3',&
                                    statlabels='dotprod,logabsx2,logabsx1px2,sumxp1',&
                                    targets=targets)
        if (mympi_iproc()==0) then  
            call disp('statmat',statmat)
            write(*,*) 'File written to /tmp/statmat.csv'
        end if
        
    end subroutine 

    ! Draws a desired number of of U[a,b] draws and evaluates f at them.
    ! Given x, f(x,s) should fill s with the desired statistics (of which there should be nstats).
    ! If fname is present, will write a csv file "fname" containing the draws and function values
    ! If statmat is present, will output the x and f(x) values 
    subroutine sub_identranddraws_MPI(f,nstats,a,b,ndraws,fname,statmat,xlabels,statlabels,targets)
        use mod_minNdim
        use mod_util
        use mod_mympi
        use mod_matlab
        implicit none
        real(rt), intent(in) :: a(:),b(:)
        integer, intent(in) :: nstats,ndraws
        character(len=*), intent(in), optional :: fname
        real(rt), intent(out), optional :: statmat(:,:)
        character(*), intent(in), optional :: xlabels ! if present, the csv file will have a header row giving xlabels instead of the generic x1,x2,...
        character(*), intent(in), optional :: statlabels ! if present, the csv file will have a header row giving statlabels instead of the generic stat1,stat2,...
        real(rt), intent(in), optional :: targets(:) ! if present, the first non-header row of the csv file will have "." for the x values and targets(:) for the f(x,s) values. 
        interface
            subroutine f(x,stats) 
                use mod_types
                implicit none
                real(rt), dimension(:), intent(in) :: x
                real(rt), dimension(:), intent(out) :: stats
           end subroutine
        end interface
        ! local other
        real(rt), allocatable, target :: X(:,:),y(:,:) ! NOTE: these have target attribute and will be pointed at
        logical, allocatable :: finishedtask(:)
        integer, allocatable :: requestlist(:) ! ,lasti(:)
        integer, allocatable :: previouslyassignedind(:) 
        real(rt), allocatable :: xstar(:) ! just a dummy variable. including so as not to have to modify code (code was from PARCRS)
        real(rt) :: tmp
        integer :: nx, N, ix, iter, i, ism, who, itmp(1), flag, uid
        integer, parameter :: debuglevel = 0
        type(minNdim_opts) :: o

        nx = size(a)
        if (size(b)/=nx) stop 'ERROR: sub_identranddraws_MPI: size of b and a do not agree'
        if (any(a>=b)) stop 'ERROR: sub_identranddraws_MPI: a>=b at least once'

        if (present(targets)) then
            if (size(targets)/=nstats) stop 'ERROR: sub_identranddraws_MPI: size(targets) does not equal nstats'
        end if

        if (present(xlabels)) then
            if (nx/=count_delim_arg(xlabels,',')) stop 'ERROR: sub_identranddraws_MPI: xlabels has an incorrect number of entries'
        end if
        if (present(statlabels)) then
            if (nstats/=count_delim_arg(statlabels,',')) stop 'ERROR: sub_identranddraws_MPI: statlabels has an incorrect number of entries'
        end if


        if (.not. present(fname) .and. .not. present(statmat)) then
            stop 'ERROR: sub_identranddraws_MPI: at least one of fname or statmat should be present, o/w what are you doing?'
        end if

        if (mympi_nproc()<=1) stop 'ERROR: sub_identranddraws_MPI: must call with at least two ranks'

        if (present(fname)) then
            if (mympi_iproc()==0) then
            
                ! Open up the file and write the header and target
                call opencsv(uid,'replace')

                ! Header
                do ix = 1,nx
                    if (present(xlabels)) then
                        write(uid,'(A)',advance='no') trim(adjustl(get_delim_arg(xlabels,',',ix)))
                    else
                        write(uid,'(A)',advance='no') 'x'//trim(conv_i_to_s(ix))
                    end if
                    write(uid,'(A)',advance='no') ','
                end do
                do ix = 1,nstats
                    if (present(statlabels)) then
                        write(uid,'(A)',advance='no') trim(adjustl(get_delim_arg(statlabels,',',ix)))
                    else
                        write(uid,'(A)',advance='no') 'stat'//trim(conv_i_to_s(ix))
                    end if
                    if (ix<nstats) write(uid,'(A)',advance='no') ','
                end do
                write(uid,*) 

                ! Row listing targets (if applicable)
                if (present(targets)) then
                    do ix = 1,nx
                        write(uid,'(A2)',advance='no') '.,' ! missing
                    end do
                    do ix = 1,nstats
                        write(uid,'(ES24.16E3)',advance='no') targets(ix)
                        if (ix<nstats) write(uid,'(A)',advance='no') ','
                    end do
                end if
                write(uid,*) 

                close(uid) ! close the file, and reopen it later

            end if
        end if

        if (present(statmat)) then
            if (ndraws/=size(statmat,1)) then
                print*,'ndraws',ndraws
                print*,'nx',nx
                print*,'nstats',nstats
                print*,'ERROR: sub_identranddraws_MPI: smat should have shape [ndraws,nx+nstats]'
            end if
            if (nx+nstats/=size(statmat,2)) then
                print*,'ndraws',ndraws
                print*,'nx',nx
                print*,'nstats',nstats
                stop 'ERROR: sub_identranddraws_MPI: smat should have shape [ndraws,nx+nstats]'
            end if
        end if

        N = ndraws

        ! If not master, go to the "slave" routine and wait for a x to evaluate
        if (mympi_iproc()/=0) then
            call slave()
            if (present(statmat)) call mympi_bcast(statmat,0) ! once returned, collect statmat
            call mympi_barrier() ! wait again (not sure if necessary, but be safe)
            return
        end if

        ! Generate N random points uniformly from [a,b]
        if (debuglevel>=3) print*,'allocating'
        allocate(finishedtask(mympi_nproc()-1))
        allocate(X(nx,N))
        allocate(y(nstats,N))
        allocate(requestlist(mympi_nproc()-1))
        allocate(previouslyassignedind(mympi_nproc()-1))
        allocate(xstar(nX)) 

        ! For random results, initialize based on the clock.
        if (debuglevel>=3) print*,'generating random seed'
        call init_random_seed()

        ! Generate initial pool of points
        if (debuglevel>=3) print*,'generating initial pool of points'
        do i = 1,N
            do ix = 1,nx
                call random_number(tmp)
                X(ix,i) = tmp*a(ix) + (1d0-tmp)*b(ix)
            end do
        end do


        previouslyassignedind = -1
        finishedtask = .true. ! whether busy or not

        ! Evaluate at each of the N points in parallel
        write(*,'(A,i4,A)') 'sub_identranddraws_MPI: Solving for initial pool of ',N,' points'
        do i = 1,N

            ! Wait for a free worker
            if (debuglevel>0) print*,'MASTER: WAITING FOR FREE WORKER'
            who = wait_for_free_worker()

            ! Assign the point
            if (debuglevel>0) print*,'MASTER: X(:,i)',X(:,i)
            previouslyassignedind(who) = i
            call assign_xy_to_worker(X(:,i),y(:,i),who)

            if (o%printmod>0) then
                write(*,'(A,i4,A,i4,A,i4,A,A)') 'sub_identranddraws_MPI: image ', who, ' solving for point ',i,' of ',N,' x = ',printx(X(:,i))
            end if

        end do

        if (debuglevel>1) print*,'waiting for everyone to finish'

        ! Wait for everyone to finish
        do while (any(.not. finishedtask))

            if (debuglevel>1) print*,'finishedtask',finishedtask
            if (debuglevel>1) print*,'requestlist',requestlist
            do who = 1,mympi_nproc()-1
                if (.not. finishedtask(who)) then
                    finishedtask(who) = mympi_test(requestlist(who)) ! requestlist is not valid after mympi_test returns true

                    ! Write the result to file
                    if (finishedtask(who)) call appendtocsv(X(:,previouslyassignedind(who)),Y(:,previouslyassignedind(who)))
                end if
            end do

        end do

        finishedtask = .true.

        ! Shutdown and distribute xbest
        call shutdown_slaves_sharebest()

        ! Wait for slaves to finish shutting down
        call mympi_barrier()
        
        ! Output to statmat
        if (present(statmat)) then
            statmat(:,1:nx) = transpose(X)
            statmat(:,nx+1:nx+nstats) = transpose(y) ! output to statmat
            call mympi_bcast(statmat,0)
        end if

        ! Output to file (note: only the master executes this code)
!         if (present(fname)) then
! 
! 
!             ! Main data
!             do i = 1,ndraws
!                 do ix = 1,nx
!                     write(uid,'(ES24.16E3)',advance='no') X(ix,i)
!                     write(uid,'(A)',advance='no') ','
!                 end do
!                 do ix = 1,nstats
!                     write(uid,'(ES24.16E3)',advance='no') Y(ix,i)
!                     if (ix<nstats) write(uid,'(A)',advance='no') ','
!                 end do
!                 write(uid,*) ! new line
!             end do
!             close(uid)
!         end if

        ! Wait for everyone to receive smat and for writing to finish
        call mympi_barrier()

    contains

        ! Opens fname for writing, returning the unit id corresponding to it. Catches errors in opening.
        subroutine opencsv(uid,status)
            integer, intent(out) :: uid
            character(len=*), intent(in) :: status
            do 
                open(newunit=uid,file=fname,action='write',status=status,position='append',err=1234)
                exit ! If no error, escape the loop
1234                write (0,*) 'ERROR: sub_identranddraws_MPI: File name '//trim(fname)//' could not be created'
                call mympi_abort()
            end do
        end subroutine 

        ! This routine sends the value of x to a slave "who", who will then store the value in y. 
        subroutine assign_xy_to_worker(x,y,who)
            use iso_c_binding, only: c_loc,c_ptr,c_f_pointer ! GETS THE ACTUALY ADDRESS OF Y WHICH WILL PERSIST AFTER CALL TO ASSIGN ENDS
            implicit none
            real(rt), intent(in), target :: x(:)
            real(rt), intent(out), target :: y(:)
            integer, intent(in) :: who ! who you want assigned
            ! local
            type(c_ptr) :: yptr_c ! pointer in C
            real(rt), pointer :: yptr_f(:) ! pointer in Fortran 
            integer :: flag,ierr

            if (debuglevel>1) print*,'MASTER: assign_xy: x',x(1:nx)

            ! Assign it
            flag = 1
            finishedtask(who) = .false.

            if (debuglevel>0) print*,'MASTER: SENDING FLAG: ',flag

            call mympi_send(flag,who) ! blocking send (tell them to proceed to execute)

            if (debuglevel>1) print*,'MASTER: FINISHED SENDING FLAG'
            if (debuglevel>1) print*,'MASTER: who',who
            if (debuglevel>1) print*,'MASTER: SENDING X: ',x(1:nx)

            call mympi_send(x,who) ! assign an x value (wait for it to complete)

            if (debuglevel>0) print*,'MASTER: IRECVING Y'

            ! In order to be able to do this assignment where the memory location associated with y will be stored even after 
            ! the subroutine is exited, the safest/most reliable way I have found is to send MPI the actual memory address.
            ! This can be accomplished by getting the memory address of y using the F2003 intrinsic "c_loc" and assigning that
            ! to a C pointer. Then, using c_f_pointer, copy that to a Fortran pointer. Then pass that Fortran poitner to MPI. 
            !
            ! This may be a hack (it seems like completely legal Fortran though), but it simplifies bookkeeping substantially.
            ! I.e., it becomes "set it and forget it."

            yptr_c = c_loc(y) ! get the actual address of y using "c_loc" and have yptr_c point to that
            call c_f_pointer(yptr_c,yptr_f,shape(y)) ! get a fortran pointer to that same address
            call MPI_Irecv(yptr_f, size(y), MPI_DOUBLE_PRECISION, who, who, MPI_COMM_WORLD, requestlist(who), ierr)
            if (ierr/=MPI_SUCCESS) stop 'MASTER: error occurred in calling MPI_Irecv'

        end subroutine 

        ! a routine to shut down all slaves
        subroutine shutdown_slaves_sharebest() 
            implicit none
            ! local
            integer :: i
            logical :: ft
            character(1), volatile :: abc ! volatile prevents optimization (this is used for a wait loop)
            integer :: j

            ! Cycle through and shutdown
            do i = 1,mympi_nproc()-1

                ! Wait for rank to finish if not finished task
                if (.not. finishedtask(i)) then

                    ! Infinite loop, wait for them to finish 
                    do 

                        ! Check whether i has finished
                        ft = mympi_test(requestlist(i))
                        if (ft) then
                            finishedtask(i) = .true. ! show that they are now free
                            exit
                        end if

                        ! If sleeptime, then sleep
                        if (o%sleeptime>=1) then

                            call sleep(o%sleeptime)

                        else

                            ! Do some busy work so as to not thrash the system
                            do j = 1,1000000
                                abc = 'a'
                            end do

                        end if

                    end do

                end if
                    
                ! Shut down, sending flag = -1 and xstar
                call mympi_send(-1,i) 
                xstar = -1234d0
                call mympi_send(xstar,i) 

            end do


        end subroutine

        ! Appends the x and stats values to the csv file
        subroutine appendtocsv(x,stats)
            implicit none
            real(8), intent(in) :: x(:),stats(:)
            integer :: i

            if (present(fname)) then
                call opencsv(uid,'old')
                do i = 1,size(x)
                    write(uid,'(ES24.16E3)',advance='no') x(i)
                    write(uid,'(A)',advance='no') ','
                end do
                do i = 1,size(stats)
                    write(uid,'(ES24.16E3)',advance='no') stats(i)
                    if (i<size(stats)) write(uid,'(A)',advance='no') ','
                end do
                write(uid,*) ! new line
                close(uid)
            end if

        end subroutine


        ! a routine returning the index of a free slave
        ! if no worker is available, who is -1
        function wait_for_free_worker() result(who)
            implicit none
            integer :: who
            ! local
            integer :: i,ix
            logical :: ft
            character(1), volatile :: abc ! volatile prevents optimization (this is used for a wait loop)
            integer :: j

            do 

                ! Infinite loop
                do i = 1,mympi_nproc()-1

                    if (finishedtask(i)) then

                        ! This branch only gets executed on the first pass when a rank has not yet been assigned a value
                        if (previouslyassignedind(i)/=-1) stop 'ERROR: sub_identranddraws_MPI: wait_for_free_worker: thought this case was impossible'

                        who = i
                        return  

                    else

                        ! This is where a value was assigned. Need to test to see whether it is finished or not.

                        ! Check whether i has finished
                        if (debuglevel>0) print*,'MASTER: TESTING MESSAGE'
                        if (debuglevel>1) print*,'MASTER: finishedtask',finishedtask
                        if (debuglevel>1) print*,'MASTER: REQUESTLIST',REQUESTLIST
                        if (debuglevel>1) print*,'MASTER: i: ',i

                        ft = mympi_test(requestlist(i))
                        if (debuglevel>0) print*,'MASTER: FINISHED TEST'
                        if (ft) then
                            finishedtask(i) = .true. ! show that they are now free

                            ! If they were previously assigned value, then write it to file if applicable
                            if (previouslyassignedind(i)/=-1) call appendtocsv(X(:,previouslyassignedind(i)),Y(:,previouslyassignedind(i)))

                            who = i
                            if (debuglevel>0) print*,'MASTER: RECEIVED MESSAGE'
                            return
                        end if

                    end if

                end do

                ! If sleeptime, then sleep
                if (o%sleeptime>=1) then

                    call sleep(o%sleeptime)

                else

                    ! Do some busy work so as to not thrash the system
    !                     call tic
                    do j = 1,1000000
                        abc = 'a'
                    end do
    !                     call toc

                end if

            end do

        end function

        ! Do something according to the value of flag(this_image())[1]
        !  1 := evaluate slavex and store in slavey
        ! -1 := exit the slave routine
        subroutine slave()
            implicit none
            integer :: flag
            real(rt), allocatable :: slavex(:),slavey(:)

            allocate(slavex(nx))
            allocate(slavey(nstats))

            do 

                ! Wait for a flag 
                if (debuglevel>0) print*,'SLAVE: WAITING TO RECEIVE FLAG'
                call mympi_recv(flag,0)
                if (debuglevel>1) print*,'SLAVE: flag is ',flag

    ! call sleep(1)

                ! Wait for an x
                if (debuglevel>0) print*,'SLAVE: WAITING TO RECEIVE SLAVEX'
                call mympi_recv(slavex,0)
                if (debuglevel>1) print*,'SLAVE: slavex is ',slavex

                ! If flag==1, execute f(x) and store on master.
                if (flag==1) then

                    ! Compute the result
                    call f(slavex,slavey)

                    ! Send the result to master tagged by mympi_iproc()
                    if (debuglevel>0) print*,'SLAVE: SENDING SLAVEY'
                    call mympi_send(slavey,0,mympi_iproc()) 
                    if (debuglevel>0) print*,'SLAVE: FINISHED SENDING SLAVEY'

                end if

                if (flag==-1) then

                    ! xstar = slavex

                    ! Wait for everyone to finish
                    call mympi_barrier()
                    exit
                end if

            end do

        end subroutine


    end subroutine

end module
