! Grey Gordon

! ! For the stubs library (which will compile without mpi), include the appropriate file.
! include 'include_mympi.f90'
! ! include 'include_mympi_stubs.f90'

! A module that attempts to simplify the process of using mpi w/o a significant
! computational cost (which of course would defeat the point!)
!
! Most routines, if available, are listed as myXXX where XXX is an mpi command.
! E.g., mympi_allgatherv calls MPI_Allgatherv.
! There are some exceptions, such as mympi_stop which calls MPI_Abort after
! displaying an error message.
module mod_mympi
    use mod_types
    implicit none

    include 'mpif.h'

    logical, private :: mympi_initialized = .false.
    integer, private :: nproc,iproc


    interface mympi_allgatherv
        module procedure mympi_allgatherv_vec, mympi_allgatherv_mat,  mympi_allgatherv_3d,&
                         mympi_allgatherv_4d, mympi_allgatherv_5d, mympi_allgatherv_6d,&
                         mympi_allgatherv_ivec,mympi_allgatherv_imat, mympi_allgatherv_i3d, mympi_allgatherv_i4d, &
                         mympi_allgatherv_l1d, mympi_allgatherv_l2d,mympi_allgatherv_l4d, mympi_allgatherv_l5d
    end interface
    interface mympi_allgatherv_r4
        module procedure mympi_allgatherv_r4vec,mympi_allgatherv_r4mat,mympi_allgatherv_r43d,mympi_allgatherv_r44d
    end interface

    interface mympi_allreduce
        module procedure mympi_allreduce_1d,mympi_allreduce_2d,mympi_allreduce_3d, &
                         mympi_allreduce_i0d,mympi_allreduce_i1d,mympi_allreduce_i2d
    end interface

    interface mympi_bcast
        module procedure mympi_bcast_vec, mympi_bcast_mat, mympi_bcast_3d, mympi_bcast_4d, mympi_bcast_5d, mympi_bcast_6d, &
                         mympi_bcast_ivec, mympi_bcast_imat, mympi_bcast_i3d, mympi_bcast_i4d, mympi_bcast_i5d, mympi_bcast_i6d, &
                         mympi_bcast_logical_vec, mympi_bcast_logical_mat, mympi_bcast_logical_3d, mympi_bcast_logical_4d, &
                         mympi_bcast_logical_5d, mympi_bcast_logical_6d
    end interface

    interface mympi_ibcast
        module procedure mympi_ibcast_vec, &
                         mympi_ibcast_ivec
    end interface

    interface mympi_irecv
        module procedure mympi_irecv_1d, mympi_irecv_0d, mympi_irecv_i1d
    end interface

    interface mympi_recv
        module procedure mympi_recv_2d, mympi_recv_1d, mympi_recv_0d, & 
                         mympi_recv_i2d, mympi_recv_i1d, mympi_recv_i0d
    end interface

    interface mympi_send
        module procedure mympi_send_0d, mympi_send_1d, mympi_send_2d, &
                         mympi_send_i0d, mympi_send_i1d, mympi_send_i2d
    end interface

    private :: mympi_bcast_core
    private :: mympi_allgatherv_core
    private :: mympi_allgatherv_core_logical
    private :: mympi_allgatherv_icore
    private :: checkinit
    private :: cumsum
    
contains

    ! Performs a reduction operation and shares the result with all nodes
    ! [ MPI_MAX] maximum
    ! [ MPI_MIN] minimum
    ! [ MPI_SUM] sum
    ! [ MPI_PROD] product
    ! [ MPI_LAND] logical and
    ! [ MPI_BAND] bit-wise and
    ! [ MPI_LOR] logical or
    ! [ MPI_BOR] bit-wise or
    ! [ MPI_LXOR] logical xor
    ! [ MPI_BXOR] bit-wise xor
    ! [ MPI_MAXLOC] max value and location
    ! [ MPI_MINLOC] min value and location 
    subroutine mympi_allreduce_3d(B,A,operation)
        implicit none
        real(rt), intent(in) :: A(:,:,:)
        real(rt), intent(out) :: B(:,:,:)
        integer, intent(in) :: operation

        if (size(A)/=size(B)) STOP 'ERROR: mympi_allreduce: input and output do not have the same size'

        call mympi_allreduce_core_d(B,A,size(A),operation)

    end subroutine 
    subroutine mympi_allreduce_2d(B,A,operation)
        implicit none
        real(rt), intent(in) :: A(:,:)
        real(rt), intent(out) :: B(:,:)
        integer, intent(in) :: operation

        if (size(A)/=size(B)) STOP 'ERROR: mympi_allreduce: input and output do not have the same size'

        call mympi_allreduce_core_d(B,A,size(A),operation)

    end subroutine 
    subroutine mympi_allreduce_1d(B,A,operation)
        implicit none
        real(rt), intent(in) :: A(:)
        real(rt), intent(out) :: B(:)
        integer, intent(in) :: operation

        if (size(A)/=size(B)) STOP 'ERROR: mympi_allreduce: input and output do not have the same size'

        call mympi_allreduce_core_d(B,A,size(A),operation)

    end subroutine 
    subroutine mympi_allreduce_core_d(B,A,nelts,operation)
        implicit none
        real(rt), intent(in) :: A(*)
        real(rt), intent(out) :: B(*)
        integer, intent(in) :: nelts,operation
        ! local
        integer :: ierr

        ! Call the mpi routine
        call MPI_Allreduce(A(1),& ! first element assigned to iproc
                           B(1),& ! where to put the reduction  
                           nelts,& ! total numbber of elements assigned to iproc --- may be zero
                           mympi_rt,& ! data type of send elements
                           operation,& ! type of operation
                           MPI_COMM_WORLD,&
                           ierr) 

    end subroutine
    
    subroutine mympi_allreduce_i2d(B,A,operation)
        implicit none
        integer, intent(in) :: A(:,:)
        integer, intent(out) :: B(:,:)
        integer, intent(in) :: operation

        if (size(A)/=size(B)) STOP 'ERROR: mympi_allreduce: input and output do not have the same size'

        call mympi_allreduce_core_i(B,A,size(A),operation)

    end subroutine 
    subroutine mympi_allreduce_i1d(B,A,operation)
        implicit none
        integer, intent(in) :: A(:)
        integer, intent(out) :: B(:)
        integer, intent(in) :: operation

        if (size(A)/=size(B)) STOP 'ERROR: mympi_allreduce: input and output do not have the same size'

        call mympi_allreduce_core_i(B,A,size(A),operation)

    end subroutine 
    subroutine mympi_allreduce_i0d(B,A,operation)
        implicit none
        integer, intent(in) :: A
        integer, intent(out) :: B
        integer, intent(in) :: operation
        integer :: ierr

        call MPI_Allreduce(A,& ! first element assigned to iproc
                           B,& ! where to put the reduction  
                           1,& ! total number of elements assigned to iproc --- may be zero
                           MPI_INTEGER,& ! data type of send elements
                           operation,& ! type of operation
                           MPI_COMM_WORLD,&
                           ierr) 

    end subroutine 
    subroutine mympi_allreduce_core_i(B,A,nelts,operation)
        implicit none
        integer, intent(in) :: A(*)
        integer, intent(out) :: B(*)
        integer, intent(in) :: nelts,operation
        ! local
        integer :: ierr

        ! Call the mpi routine
        call MPI_Allreduce(A(1),& ! first element assigned to iproc
                           B(1),& ! where to put the reduction  
                           nelts,& ! total numbber of elements assigned to iproc --- may be zero
                           MPI_INTEGER,& ! data type of send elements
                           operation,& ! type of operation
                           MPI_COMM_WORLD,&
                           ierr) 

    end subroutine

function mympi_getprocessorname() result(nomen)
    implicit none
    character(len=MPI_MAX_PROCESSOR_NAME) :: nomen
    ! local
    integer :: ierr,lenout
    character(len=MPI_MAX_PROCESSOR_NAME) :: tmp

    
    call MPI_GET_PROCESSOR_NAME(tmp,lenout,ierr)

    nomen = ''
    nomen(1:lenout) = tmp(1:lenout) 

end function 

function get_ierr_msg(ierr) result(msg)
    implicit none
    integer, intent(in) :: ierr
    character(len=50) :: msg

    select case (ierr)
        case (MPI_SUCCESS)
            msg='Successful return code'
        case (MPI_ERR_BUFFER)
            msg='Invalid buffer pointer'
        case (MPI_ERR_COUNT)
            msg='Invalid count argument'
        case (MPI_ERR_TYPE)
            msg='Invalid datatype argument'
        case (MPI_ERR_TAG)
            msg='Invalid tag argument'
        case (MPI_ERR_COMM)
            msg='Invalid communicator'
        case (MPI_ERR_RANK)
            msg='Invalid rank'
        case (MPI_ERR_ROOT)
            msg='Invalid root'
        case (MPI_ERR_GROUP)
            msg='Null group passed to function'
        case (MPI_ERR_OP)
            msg='Invalid operation'
        case (MPI_ERR_TOPOLOGY)
            msg='Invalid topology'
        case (MPI_ERR_DIMS)
            msg='Illegal dimension argument'
        case (MPI_ERR_ARG)
            msg='Invalid argument'
        case (MPI_ERR_UNKNOWN)
            msg='Unknown error'
        case (MPI_ERR_TRUNCATE)
            msg='message truncated on receive'
        case (MPI_ERR_OTHER)
            msg='Other error; use Error_string'
        case (MPI_ERR_INTERN)
            msg='internal error code'
        case (MPI_ERR_IN_STATUS)
            msg='Look in status for error value'
        case (MPI_ERR_PENDING)
            msg='Pending request'
        case (MPI_ERR_REQUEST)
            msg='illegal mpi_request handle'
        case (MPI_ERR_LASTCODE)
            msg='Last error code -- always at end'
    end select

end function

subroutine checkinit()
    implicit none
    if (.not. mympi_initialized) STOP 'ERROR: mpi has not been initialized'
end subroutine 

integer function mympi_iproc()
    implicit none
    call checkinit()
    mympi_iproc = iproc
end function 
integer function mympi_nproc()
    implicit none
    call checkinit()
    mympi_nproc = nproc
end function 

subroutine mympi_barrier()
    implicit none
    integer :: ierr

    call checkinit()
    call MPI_Barrier(MPI_COMM_WORLD,ierr)

end subroutine 

! Aborts all mpi routines and also displays a message
subroutine mympi_abort()
    implicit none
    call mympi_stop('')
end subroutine 
subroutine mympi_stop(msg)
    implicit none
    character(len=*), intent(in) :: msg
    ! local
    integer :: errorcode
    integer :: ierr

    call checkinit()

    ! Display the error message
    write(0,*) msg

    ! Pass error code to environment
    errorcode = 1 
    
    ! Terminate all processes
    call MPI_Abort(MPI_COMM_WORLD,errorcode,ierr)

end subroutine 

! Non blocking test for a message
! function mympi_iprobe(source,tag) result(tf)
!     implicit none
!     integer, intent(in) :: source
!     integer, intent(in), optional :: tag 
!     logical :: tf
!     ! local
!     integer :: ierr, status(MPI_STATUS_SIZE)
!     logical :: flag
! 
!     if (present(tag)) then
!         call MPI_Iprobe(source, tag, MPI_COMM_WORLD, flag, status, ierr)
!     else
!         call MPI_Iprobe(source, MPI_ANY_TAG, MPI_COMM_WORLD, flag, status, ierr)
!     end if
! 
!     tf = flag
! 
! end function 

! Divides n elts as evenly as possible into ranges [a,b]
! If b<a, then b is set to -1 and a to 0. This is convenient as b-a+1 
! then returns the number of elements for each node regardless of whether 
! the range is 0 or not. Also, since b<a, "do i = a,b" is not executed. 
! NOTE: if used to assign trailing dimensions of an array, 
! as it defines contiguous array section this is very useful
subroutine mympi_assign_range(a,b,n)
    implicit none
    integer, intent(in) :: n
    integer, intent(out) :: a,b
    
    call checkinit()
    call mympi_assign_range_core(a,b,iproc,n)

end subroutine 
! Helper functions
integer function mympi_a(n)
    implicit none
    integer, intent(in) :: n
    ! local
    integer :: b
    call checkinit()
    call mympi_assign_range(mympi_a,b,n)
end function mympi_a
integer function mympi_b(n)
    implicit none
    integer, intent(in) :: n
    ! local
    integer :: a
    call checkinit()
    call mympi_assign_range(a,mympi_b,n)
end function mympi_b
! Gives the combined list
subroutine mympi_assign_range_list(a,b,n)
    implicit none
    integer, intent(in) :: n
    integer, intent(out) :: a(nproc),b(nproc)
    ! local
    integer :: ip

    call checkinit()

    do ip = 0,nproc-1
        call mympi_assign_range_core(a(ip+1),b(ip+1),ip,n)
    end do
end subroutine 

! The core routine for dividing elements
! ip = 0,...,nproc-1
subroutine mympi_assign_range_core(a,b,ip,n)
    implicit none
    integer, intent(in) :: ip,n
    integer, intent(out) :: a,b
    ! local
    integer :: rem,jproc,times

    call checkinit()
    if (n<nproc) then
        ! If here, can assign one element to each node. 
        ! Hence, ip gets the ip+1 element if ip+1<=n
        if (ip<n) then
            a = ip+1
            b = ip+1
        else
            a = 0
            b = -1
        end if
    else
        ! If here, need to assign at least n/nproc to each element and the remainder
        ! across the first elements

        rem = mod(n,nproc)
        times = n/nproc ! >=1

        if (n/=times*nproc + rem) call mympi_stop('ERROR: mympi_assign_range: arithmetic not as expected')

        b = 0
        do jproc = 0,ip
            if (jproc<rem) then
                ! Gets times+1 elements
                a = 1 + b
                b = times + 1 + b
            else
                ! Gets times elements
                a = 1 + b
                b = times + b
            end if
        end do

    end if

end subroutine


subroutine mympi_bcast_logical_6d(A,root)
    implicit none
    logical, intent(inout) :: A(:,:,:,:,:,:)
    integer, intent(in) :: root
    call mympi_bcast_logical_core(A,size(A),root)
end subroutine
subroutine mympi_bcast_logical_5d(A,root)
    implicit none
    logical, intent(inout) :: A(:,:,:,:,:)
    integer, intent(in) :: root
    call mympi_bcast_logical_core(A,size(A),root)
end subroutine
subroutine mympi_bcast_logical_4d(A,root)
    implicit none
    logical, intent(inout) :: A(:,:,:,:)
    integer, intent(in) :: root
    call mympi_bcast_logical_core(A,size(A),root)
end subroutine
subroutine mympi_bcast_logical_3d(A,root)
    implicit none
    logical, intent(inout) :: A(:,:,:)
    integer, intent(in) :: root
    call mympi_bcast_logical_core(A,size(A),root)
end subroutine
subroutine mympi_bcast_logical_mat(A,root)
    implicit none
    logical, intent(inout) :: A(:,:)
    integer, intent(in) :: root
    call mympi_bcast_logical_core(A,size(A),root)
end subroutine
subroutine mympi_bcast_logical_vec(A,root)
    implicit none
    logical, intent(inout) :: A(:)
    integer, intent(in) :: root
    call mympi_bcast_logical_core(A,size(A),root)
end subroutine

subroutine mympi_bcast_logical_core(A,n,root)
    implicit none
    integer, intent(in) :: n, root
    logical, intent(inout) :: A(*)
    ! local
    integer :: ierr

    call checkinit()

    ! Call the mpi routine
    call MPI_Bcast(A(1),n,MPI_LOGICAL,root,MPI_COMM_WORLD,ierr)

end subroutine

subroutine mympi_ibcast_ivec(request,A,root)
    implicit none
    integer, intent(out) :: request
    integer, intent(inout) :: A(:)
    integer, intent(in) :: root

    call mympi_ibcast_icore(request,A,size(A),root)

end subroutine
subroutine mympi_ibcast_icore(request,A,n,root)
    implicit none
    integer, intent(out) :: request
    integer, intent(inout) :: A(*)
    integer, intent(in) :: n,root
    ! local
    integer :: ierr

    call checkinit()

    ! Call the mpi routine
    call MPI_Ibcast(A(1:n),n,MPI_INTEGER,root,MPI_COMM_WORLD,request,ierr)

    if (ierr/=MPI_SUCCESS) then
        call mympi_stop('error calling mympi_ibcast_icore')
    end if

end subroutine

subroutine mympi_ibcast_vec(request,A,root)
    implicit none
    integer, intent(out) :: request
    real(rt), intent(inout) :: A(:)
    integer, intent(in) :: root
    call mympi_ibcast_core(request,A,size(A),root)
end subroutine
subroutine mympi_ibcast_core(request,A,n,root)
    implicit none
    integer, intent(out) :: request
    integer, intent(in) :: n,root
    real(rt), intent(inout) :: A(*)
    ! local
    integer :: ierr

    call checkinit()

    ! Call the mpi routine
    call MPI_Ibcast(A(1),n,mympi_rt,root,MPI_COMM_WORLD,request,ierr)

end subroutine

subroutine mympi_bcast_6d(A,root)
    implicit none
    real(rt), intent(inout) :: A(:,:,:,:,:,:)
    integer, intent(in) :: root
    call mympi_bcast_core(A,size(A),root)
end subroutine
subroutine mympi_bcast_5d(A,root)
    implicit none
    real(rt), intent(inout) :: A(:,:,:,:,:)
    integer, intent(in) :: root
    call mympi_bcast_core(A,size(A),root)
end subroutine
subroutine mympi_bcast_4d(A,root)
    implicit none
    real(rt), intent(inout) :: A(:,:,:,:)
    integer, intent(in) :: root
    call mympi_bcast_core(A,size(A),root)
end subroutine
subroutine mympi_bcast_3d(A,root)
    implicit none
    real(rt), intent(inout) :: A(:,:,:)
    integer, intent(in) :: root
    call mympi_bcast_core(A,size(A),root)
end subroutine
subroutine mympi_bcast_mat(A,root)
    implicit none
    real(rt), intent(inout) :: A(:,:)
    integer, intent(in) :: root
    call mympi_bcast_core(A,size(A),root)
end subroutine
subroutine mympi_bcast_vec(A,root)
    implicit none
    real(rt), intent(inout) :: A(:)
    integer, intent(in) :: root
    call mympi_bcast_core(A,size(A),root)
end subroutine

subroutine mympi_bcast_core(A,n,root)
    implicit none
    integer, intent(in) :: n,root
    real(rt), intent(inout) :: A(*)
    ! local
    integer :: ierr

    call checkinit()

    ! Call the mpi routine
    call MPI_Bcast(A(1),n,mympi_rt,root,MPI_COMM_WORLD,ierr)

end subroutine

subroutine mympi_bcast_i6d(A,root)
    implicit none
    integer, intent(inout) :: A(:,:,:,:,:,:)
    integer, intent(in) :: root
    call mympi_bcast_icore(A,size(A),root)
end subroutine
subroutine mympi_bcast_i5d(A,root)
    implicit none
    integer, intent(inout) :: A(:,:,:,:,:)
    integer, intent(in) :: root
    call mympi_bcast_icore(A,size(A),root)
end subroutine
subroutine mympi_bcast_i4d(A,root)
    implicit none
    integer, intent(inout) :: A(:,:,:,:)
    integer, intent(in) :: root
    call mympi_bcast_icore(A,size(A),root)
end subroutine
subroutine mympi_bcast_i3d(A,root)
    implicit none
    integer, intent(inout) :: A(:,:,:)
    integer, intent(in) :: root
    call mympi_bcast_icore(A,size(A),root)
end subroutine
subroutine mympi_bcast_imat(A,root)
    implicit none
    integer, intent(inout) :: A(:,:)
    integer, intent(in) :: root
    call mympi_bcast_icore(A,size(A),root)
end subroutine
subroutine mympi_bcast_ivec(A,root)
    implicit none
    integer, intent(inout) :: A(:)
    integer, intent(in) :: root
    call mympi_bcast_icore(A,size(A),root)
end subroutine

subroutine mympi_bcast_icore(A,n,root)
    implicit none
    integer, intent(in) :: n,root
    integer, intent(inout) :: A(*)
    ! local
    integer :: ierr

    call checkinit()

    ! Call the mpi routine
    call MPI_Bcast(A(1),n,MPI_INTEGER,root,MPI_COMM_WORLD,ierr)

end subroutine

function mympi_test(request) result(isComplete)
    implicit none
    integer, intent(in) :: request
    logical :: isComplete
    ! local
    logical :: flag
    integer :: ierr ! , status(MPI_STATUS_SIZE)

    call MPI_Test(request,flag,MPI_STATUS_IGNORE,ierr)

    isComplete = flag

end function

subroutine mympi_irecv_1d(request,A,source,tag)
    implicit none
    integer, intent(out) :: request
    real(rt), intent(out) :: A(:)
    integer, intent(in) :: source
    integer, intent(in), optional :: tag

    if (present(tag)) then
        call mympi_irecv_core(request,A,size(A),source,tag)
    else
        ! call mympi_irecv_core(request,A,size(A),source,MPI_ANY_TAG)
        call mympi_irecv_core(request,A,size(A),source,0)
    end if

end subroutine
subroutine mympi_irecv_0d(request,A,source,tag)
    implicit none
    integer, intent(out) :: request
    real(rt), intent(out) :: A
    integer, intent(in) :: source
    integer, intent(in), optional :: tag
    ! local
    real(rt) :: tmp(1)

    if (present(tag)) then
        call mympi_irecv_core(request,tmp,1,source,tag)
    else
        ! call mympi_irecv_core(request,tmp,1,source,MPI_ANY_TAG)
        call mympi_irecv_core(request,tmp,1,source,0)
    end if

    A = tmp(1)

end subroutine


subroutine mympi_irecv_core(request,A,n,source,tag)
    implicit none
    integer, intent(out) :: request
    real(rt), intent(out) :: A(*)
    integer, intent(in) :: n,source,tag
    ! local
    integer :: ierr

    call MPI_Irecv(A(1), n, mympi_rt, source, tag, &
                  MPI_COMM_WORLD, request, ierr)

    if (ierr/=MPI_SUCCESS) then
        print*,'ERROR: mympi_irecv: ',trim(get_ierr_msg(ierr))
        call MPI_Abort(MPI_COMM_WORLD,1,ierr)
    end if

end subroutine 

subroutine mympi_irecv_i1d(request,A,source,tag)
    implicit none
    integer, intent(out) :: request
    integer, intent(out) :: A(:)
    integer, intent(in) :: source
    integer, intent(in), optional :: tag

    if (present(tag)) then
        call mympi_irecv_icore(request,A,size(A),source,tag)
    else
        ! call mympi_irecv_icore(request,A,size(A),source,MPI_ANY_TAG)
        call mympi_irecv_icore(request,A,size(A),source,0)
    end if

end subroutine

subroutine mympi_irecv_icore(request,A,n,source,tag)
    implicit none
    integer, intent(out) :: request
    integer, intent(out) :: A(*)
    integer, intent(in) :: n,source,tag
    ! local
    integer :: ierr

    call MPI_Irecv(A(1), n, MPI_INTEGER, source, tag, &
                  MPI_COMM_WORLD, request, ierr)

    if (ierr/=MPI_SUCCESS) then
        print*,'ERROR: mympi_irecv: ',trim(get_ierr_msg(ierr))
        call MPI_Abort(MPI_COMM_WORLD,1,ierr)
    end if

end subroutine 

subroutine mympi_waitall(array_of_requests)
    implicit none
    integer, intent(in) :: array_of_requests(:)
    ! local
    integer :: ierr

    call MPI_Waitall(size(array_of_requests),array_of_requests(1),MPI_STATUSES_IGNORE,ierr)

    if (ierr/=MPI_SUCCESS) then
        print*,'ERROR: mympi_waitall: ',trim(get_ierr_msg(ierr))
        call MPI_Abort(MPI_COMM_WORLD,1,ierr)
    end if
end subroutine

subroutine mympi_wait(rqst)
    implicit none
    integer, intent(in) :: rqst
    ! local
    integer :: ierr

    call MPI_Wait(rqst,MPI_STATUS_IGNORE,ierr)

    if (ierr/=MPI_SUCCESS) then
        print*,'ERROR: mympi_wait: ',trim(get_ierr_msg(ierr))
        call MPI_Abort(MPI_COMM_WORLD,1,ierr)
    end if
end subroutine



subroutine mympi_recv_2d(A,source,tag)
    implicit none
    real(rt), intent(out) :: A(:,:)
    integer, intent(in) :: source
    integer, intent(in), optional :: tag

    if (present(tag)) then
        call mympi_recv_core(A,size(A),source,tag)
    else
        ! print*,'about to call recv_core'
        ! call mympi_recv_core(A,size(A),source,MPI_ANY_TAG)
        call mympi_recv_core(A,size(A),source,0)
    end if

end subroutine
subroutine mympi_recv_1d(A,source,tag)
    implicit none
    real(rt), intent(out) :: A(:)
    integer, intent(in) :: source
    integer, intent(in), optional :: tag

    if (present(tag)) then
        call mympi_recv_core(A,size(A),source,tag)
    else
        ! print*,'about to call recv_core'
        ! call mympi_recv_core(A,size(A),source,MPI_ANY_TAG)
        call mympi_recv_core(A,size(A),source,0)
    end if

end subroutine
subroutine mympi_recv_0d(A,source,tag)
    implicit none
    real(rt), intent(out) :: A
    integer, intent(in) :: source
    integer, intent(in), optional :: tag
    ! local
    real(rt) :: tmp(1)

    if (present(tag)) then
        call mympi_recv_core(tmp,1,source,tag)
    else
        ! call mympi_recv_core(tmp,1,source,MPI_ANY_TAG)
        call mympi_recv_core(tmp,1,source,0)
    end if

    A = tmp(1)

end subroutine


subroutine mympi_recv_core(A,n,source,tag)
    implicit none
    real(rt), intent(out) :: A(*)
    integer, intent(in) :: n,source,tag
    ! local
    integer :: ierr !, status(MPI_STATUS_SIZE)

    ! print*,'about to call recv'
    ! print*,'source: ',source
    ! print*,'tag: ',tag,'MPI_ANY_TAG',MPI_ANY_TAG
    ! print*,'MPI_STATUS_IGNORE: ',MPI_STATUS_IGNORE
    ! print*,'n',n
    ! print*,'A(1)',A(1)
    ! print*,'A(2)',A(2)
    ! print*,'A(3)',A(3)
    ! print*,'MPI_COMM_WORLD',MPI_COMM_WORLD
    call MPI_Recv(A(1), n, mympi_rt, source, tag, &
                  MPI_COMM_WORLD, MPI_STATUS_IGNORE, ierr)

    if (ierr/=MPI_SUCCESS) then
        print*,'ERROR: mympi_recv: ',trim(get_ierr_msg(ierr))
        call MPI_Abort(MPI_COMM_WORLD,1,ierr)
    end if

end subroutine 

subroutine mympi_recv_i0d(A,source,tag)
    implicit none
    integer, intent(out) :: A
    integer, intent(in) :: source
    integer, intent(in), optional :: tag
    ! local
    integer :: tmp(1)

    if (present(tag)) then
        call mympi_recv_icore(tmp,1,source,tag)
    else
        ! call mympi_recv_icore(tmp,1,source,MPI_ANY_TAG)
        call mympi_recv_icore(tmp,1,source,0)
    end if

    A = tmp(1)

end subroutine
subroutine mympi_recv_i2d(A,source,tag)
    implicit none
    integer, intent(out) :: A(:,:)
    integer, intent(in) :: source
    integer, intent(in), optional :: tag

    if (present(tag)) then
        call mympi_recv_icore(A,size(A),source,tag)
    else
        ! call mympi_recv_icore(A,size(A),source,MPI_ANY_TAG)
        call mympi_recv_icore(A,size(A),source,0)
    end if

end subroutine
subroutine mympi_recv_i1d(A,source,tag)
    implicit none
    integer, intent(out) :: A(:)
    integer, intent(in) :: source
    integer, intent(in), optional :: tag

    if (present(tag)) then
        call mympi_recv_icore(A,size(A),source,tag)
    else
        ! call mympi_recv_icore(A,size(A),source,MPI_ANY_TAG)
        call mympi_recv_icore(A,size(A),source,0)
    end if

end subroutine

subroutine mympi_recv_icore(A,n,source,tag)
    implicit none
    integer, intent(out) :: A(*)
    integer, intent(in) :: n,source,tag
    ! local
    integer :: ierr !, status(MPI_STATUS_SIZE)

    call MPI_Recv(A(1), n, MPI_INTEGER, source, tag, &
                  MPI_COMM_WORLD, MPI_STATUS_IGNORE, ierr)

    if (ierr/=MPI_SUCCESS) then
        print*,'ERROR: mympi_recv: ',trim(get_ierr_msg(ierr))
        call MPI_Abort(MPI_COMM_WORLD,1,ierr)
    end if

end subroutine 


subroutine mympi_send_2d(A,dest,tag)
    implicit none
    real(rt), intent(in) :: A(:,:)
    integer, intent(in) :: dest
    integer, intent(in), optional :: tag

    if (present(tag)) then
        call mympi_send_core(A,size(A),dest,tag)
    else
        ! NOTE: MPI_ANY_TAG may only be used in "RECV"s
        ! print*,'about to call mympi_send_core'
        call mympi_send_core(A,size(A),dest,0)
    end if

end subroutine
subroutine mympi_send_1d(A,dest,tag)
    implicit none
    real(rt), intent(in) :: A(:)
    integer, intent(in) :: dest
    integer, intent(in), optional :: tag

    if (present(tag)) then
        call mympi_send_core(A,size(A),dest,tag)
    else
        ! NOTE: MPI_ANY_TAG may only be used in "RECV"s
        ! print*,'about to call mympi_send_core'
        call mympi_send_core(A,size(A),dest,0)
    end if

end subroutine
subroutine mympi_send_0d(A,dest,tag)
    implicit none
    real(rt), intent(in) :: A
    integer, intent(in) :: dest
    integer, intent(in), optional :: tag
    ! local
    real(rt) :: tmp(1)

    tmp(1) = A

    if (present(tag)) then
        call mympi_send_core(tmp,1,dest,tag)
    else
        ! NOTE: MPI_ANY_TAG may only be used in "RECV"s
        call mympi_send_core(tmp,1,dest,0)
    end if

end subroutine

subroutine mympi_send_core(A,n,dest,tag)
    implicit none
    real(rt), intent(in) :: A(*)
    integer, intent(in) :: n,dest,tag
    ! local
    integer :: ierr

    ! print*,'about to send'
    call MPI_send(A(1), n, mympi_rt, dest, tag, &
                   MPI_COMM_WORLD, ierr)

    if (ierr/=MPI_SUCCESS) then
        print*,'ERROR: mympi_send: ',trim(get_ierr_msg(ierr))
        call MPI_Abort(MPI_COMM_WORLD,1,ierr)
    end if

end subroutine 

subroutine mympi_send_i2d(A,dest,tag)
    implicit none
    integer, intent(in) :: A(:,:)
    integer, intent(in) :: dest
    integer, intent(in), optional :: tag

    if (present(tag)) then
        call mympi_send_icore(A,size(A),dest,tag)
    else
        ! NOTE: MPI_ANY_TAG may only be used in "RECV"s
        call mympi_send_icore(A,size(A),dest,0)
    end if

end subroutine
subroutine mympi_send_i1d(A,dest,tag)
    implicit none
    integer, intent(in) :: A(:)
    integer, intent(in) :: dest
    integer, intent(in), optional :: tag

    if (present(tag)) then
        call mympi_send_icore(A,size(A),dest,tag)
    else
        ! NOTE: MPI_ANY_TAG may only be used in "RECV"s
        call mympi_send_icore(A,size(A),dest,0)
    end if

end subroutine
subroutine mympi_send_i0d(A,dest,tag)
    implicit none
    integer, intent(in) :: A
    integer, intent(in) :: dest
    integer, intent(in), optional :: tag
    ! local
    integer :: tmp(1)

    tmp(1) = A

    if (present(tag)) then
        call mympi_send_icore(tmp,1,dest,tag)
    else
        ! NOTE: MPI_ANY_TAG may only be used in "RECV"s
        call mympi_send_icore(tmp,1,dest,0)
    end if

end subroutine


subroutine mympi_send_icore(A,n,dest,tag)
    implicit none
    integer, intent(in) :: A(*)
    integer, intent(in) :: n,dest,tag
    ! local
    integer :: ierr

    call MPI_send(A(1), n, MPI_INTEGER, dest, tag, &
                   MPI_COMM_WORLD, ierr)

    if (ierr/=MPI_SUCCESS) then
        print*,'ERROR: mympi_send: ',trim(get_ierr_msg(ierr))
        call MPI_Abort(MPI_COMM_WORLD,1,ierr)
    end if

end subroutine 



subroutine mympi_allgatherv_l5d(A)
    implicit none
    logical, intent(inout) :: A(:,:,:,:,:)
    call mympi_allgatherv_core_logical(A,size(A,1)*size(A,2)*size(A,3)*size(A,4),size(A,5))
end subroutine 
subroutine mympi_allgatherv_l4d(A)
    implicit none
    logical, intent(inout) :: A(:,:,:,:)
    call mympi_allgatherv_core_logical(A,size(A,1)*size(A,2)*size(A,3),size(A,4))
end subroutine 
subroutine mympi_allgatherv_l2d(A)
    implicit none
    logical, intent(inout) :: A(:,:)
    call mympi_allgatherv_core_logical(A,size(A,1),size(A,2))
end subroutine 
subroutine mympi_allgatherv_l1d(A)
    implicit none
    logical, intent(inout) :: A(:)
    call mympi_allgatherv_core_logical(A,1,size(A,1))
end subroutine 

! This gathers data from all tasks and delivers the combined data to all tasks.
! NOTE: it assumes the work was distributed across the trailing dimension according
!       to mympi_assign_range_list.
subroutine mympi_allgatherv_6d(A)
    implicit none
    real(rt), intent(inout) :: A(:,:,:,:,:,:)
    call mympi_allgatherv_core(A,size(A,1)*size(A,2)*size(A,3)*size(A,4)*size(A,5),size(A,6))
end subroutine 
subroutine mympi_allgatherv_5d(A)
    implicit none
    real(rt), intent(inout) :: A(:,:,:,:,:)
    call mympi_allgatherv_core(A,size(A,1)*size(A,2)*size(A,3)*size(A,4),size(A,5))
end subroutine 
subroutine mympi_allgatherv_4d(A)
    implicit none
    real(rt), intent(inout) :: A(:,:,:,:)
    call mympi_allgatherv_core(A,size(A,1)*size(A,2)*size(A,3),size(A,4))
end subroutine 
subroutine mympi_allgatherv_3d(A)
    implicit none
    real(rt), intent(inout) :: A(:,:,:)
    call mympi_allgatherv_core(A,size(A,1)*size(A,2),size(A,3))
end subroutine 
subroutine mympi_allgatherv_mat(A)
    implicit none
    real(rt), intent(inout) :: A(:,:)
    call mympi_allgatherv_core(A,size(A,1),size(A,2))
end subroutine 
subroutine mympi_allgatherv_vec(A)
    implicit none
    real(rt), intent(inout) :: A(:)
    call mympi_allgatherv_core(A,1,size(A,1))
end subroutine 
subroutine mympi_allgatherv_core(A,nrow,ncol)
    implicit none
    integer, intent(in) :: nrow,ncol
    real(rt), intent(inout) :: A(nrow,*)
    ! local
    real(rt), allocatable :: B(:,:)
    integer :: lo(nproc),hi(nproc),nelts(nproc),offset(nproc)
    integer :: ierr,lo_iproc,icol

    call checkinit()

    ! Get how the list was distributed
    call mympi_assign_range_list(lo,hi,ncol)

    ! Number of elements (columns)
    nelts = hi - lo + 1

    ! Offset
    offset(1) = 0
    offset(2:nproc) = cumsum(nrow*nelts(1:nproc-1))

    ! Check whether number of elements to be sent is > 0 
    if (nelts(iproc+1)>0) then
        ! If so, then first element of A to be sent occurs at lo_iproc
        lo_iproc = lo(iproc+1)
    else
        ! Otherwise, need to specify one knowing that no values are sent 
        lo_iproc = 1
    end if

    allocate(B(nrow,ncol))

    ! Call the mpi routine
    call MPI_Allgatherv(A(1,lo_iproc),& ! first element assigned to iproc
                        nrow*nelts(iproc+1),& ! total number of elements assigned to iproc --- may be zero
                        mympi_rt,& ! data type of send elements
                        B(1,1),& ! first element of receive 
                        nrow*nelts,& ! number of elements to be received by each process
                        offset,& ! Entry iproc specifies the displacement (relative to recvbuf) at which to place data from iproc
                        mympi_rt,& ! data type of receive elements
                        MPI_COMM_WORLD,&
                        ierr) 

    ! Now all has been received in B, so copy out to A (this is a restriction of MPI)
    do icol = 1,ncol
        A(1:nrow,icol) = B(:,icol)
    end do
    deallocate(B)

end subroutine

! Same as mympi_allgatherv_core but with logical
subroutine mympi_allgatherv_core_logical(A,nrow,ncol)
    implicit none
    integer, intent(in) :: nrow,ncol
    logical, intent(inout) :: A(nrow,*)
    ! local
    logical, allocatable :: B(:,:)
    integer :: lo(nproc),hi(nproc),nelts(nproc),offset(nproc)
    integer :: ierr,lo_iproc,icol

    call checkinit()


    ! Get how the list was distributed
    call mympi_assign_range_list(lo,hi,ncol)

    ! Number of elements (columns)
    nelts = hi - lo + 1

    ! Offset
    offset(1) = 0
    offset(2:nproc) = cumsum(nrow*nelts(1:nproc-1))

    ! Check whether number of elements to be sent is > 0 
    if (nelts(iproc+1)>0) then
        ! If so, then first element of A to be sent occurs at lo_iproc
        lo_iproc = lo(iproc+1)
    else
        ! Otherwise, need to specify one knowing that no values are sent 
        lo_iproc = 1
    end if

    allocate(B(nrow,ncol))

    ! Call the mpi routine
    call MPI_Allgatherv(A(1,lo_iproc),& ! first element assigned to iproc
                        nrow*nelts(iproc+1),& ! total number of elements assigned to iproc --- may be zero
                        MPI_LOGICAL,& ! data type of send elements
                        B(1,1),& ! first element of receive 
                        nrow*nelts,& ! number of elements to be received by each process
                        offset,& ! Entry iproc specifies the displacement (relative to recvbuf) at which to place data from iproc
                        MPI_LOGICAL,& ! data type of receive elements
                        MPI_COMM_WORLD,&
                        ierr) 

    ! Now all has been received in B, so copy out to A (this is a restriction of MPI)
    do icol = 1,ncol
        A(1:nrow,icol) = B(:,icol)
    end do
    deallocate(B)


end subroutine

! Single precision version of the above
subroutine mympi_allgatherv_r44d(A)
    implicit none
    real(4), intent(inout) :: A(:,:,:,:)
    call mympi_allgatherv_r4core(A,size(A,1)*size(A,2)*size(A,3),size(A,4))
end subroutine 
subroutine mympi_allgatherv_r43d(A)
    implicit none
    real(4), intent(inout) :: A(:,:,:)
    call mympi_allgatherv_r4core(A,size(A,1)*size(A,2),size(A,3))
end subroutine 
subroutine mympi_allgatherv_r4mat(A)
    implicit none
    real(4), intent(inout) :: A(:,:)
    call mympi_allgatherv_r4core(A,size(A,1),size(A,2))
end subroutine 
subroutine mympi_allgatherv_r4vec(A)
    implicit none
    real(4), intent(inout) :: A(:)
    call mympi_allgatherv_r4core(A,1,size(A))
end subroutine 
subroutine mympi_allgatherv_r4core(A,nrow,ncol)
    implicit none
    integer, intent(in) :: nrow,ncol
    real(4), intent(inout) :: A(nrow,*)
    ! local
    real(4), allocatable :: B(:,:)
    integer :: lo(nproc),hi(nproc),nelts(nproc),offset(nproc)
    integer :: ierr,lo_iproc,icol

    call checkinit()


    ! Get how the list was distributed
    call mympi_assign_range_list(lo,hi,ncol)

    ! Number of elements (columns)
    nelts = hi - lo + 1

    ! Offset
    offset(1) = 0
    offset(2:nproc) = cumsum(nrow*nelts(1:nproc-1))

    ! Check whether number of elements to be sent is > 0 
    if (nelts(iproc+1)>0) then
        ! If so, then first element of A to be sent occurs at lo_iproc
        lo_iproc = lo(iproc+1)
    else
        ! Otherwise, need to specify one knowing that no values are sent 
        lo_iproc = 1
    end if

    allocate(B(nrow,ncol))

    ! Call the mpi routine
    call MPI_Allgatherv(A(1,lo_iproc),& ! first element assigned to iproc
                        nrow*nelts(iproc+1),& ! total number of elements assigned to iproc --- may be zero
                        MPI_REAL4,& ! data type of send elements
                        B(1,1),& ! first element of receive 
                        nrow*nelts,& ! number of elements to be received by each process
                        offset,& ! Entry iproc specifies the displacement (relative to recvbuf) at which to place data from iproc
                        MPI_REAL4,& ! data type of receive elements
                        MPI_COMM_WORLD,&
                        ierr) 

    ! Now all has been received in B, so copy out to A (this is a restriction of MPI)
    do icol = 1,ncol
        A(1:nrow,icol) = B(:,icol)
    end do
    deallocate(B)


end subroutine


! Integer versions of the above
subroutine mympi_allgatherv_i4d(A)
    implicit none
    integer, intent(inout) :: A(:,:,:,:)
    call mympi_allgatherv_icore(A,size(A,1)*size(A,2)*size(A,3),size(A,4))
end subroutine 
subroutine mympi_allgatherv_i3d(A)
    implicit none
    integer, intent(inout) :: A(:,:,:)
    call mympi_allgatherv_icore(A,size(A,1)*size(A,2),size(A,3))
end subroutine 
subroutine mympi_allgatherv_imat(A)
    implicit none
    integer, intent(inout) :: A(:,:)
    call mympi_allgatherv_icore(A,size(A,1),size(A,2))
end subroutine 
subroutine mympi_allgatherv_ivec(A)
    implicit none
    integer, intent(inout) :: A(:)
    call mympi_allgatherv_icore(A,1,size(A))
end subroutine 
subroutine mympi_allgatherv_icore(A,nrow,ncol)
    implicit none
    integer, intent(in) :: nrow,ncol
    integer, intent(inout) :: A(nrow,*)
    ! local
    integer, allocatable :: B(:,:)
    integer :: lo(nproc),hi(nproc),nelts(nproc),offset(nproc)
    integer :: ierr,lo_iproc,icol

    call checkinit()


    ! Get how the list was distributed
    call mympi_assign_range_list(lo,hi,ncol)

    ! Number of elements (columns)
    nelts = hi - lo + 1

    ! Offset
    offset(1) = 0
    offset(2:nproc) = cumsum(nrow*nelts(1:nproc-1))

    ! Check whether number of elements to be sent is > 0 
    if (nelts(iproc+1)>0) then
        ! If so, then first element of A to be sent occurs at lo_iproc
        lo_iproc = lo(iproc+1)
    else
        ! Otherwise, need to specify one knowing that no values are sent 
        lo_iproc = 1
    end if

    allocate(B(nrow,ncol))

    ! Call the mpi routine
    call MPI_Allgatherv(A(1,lo_iproc),& ! first element assigned to iproc
                        nrow*nelts(iproc+1),& ! total number of elements assigned to iproc --- may be zero
                        MPI_INTEGER,& ! data type of send elements
                        B(1,1),& ! first element of receive 
                        nrow*nelts,& ! number of elements to be received by each process
                        offset,& ! Entry iproc specifies the displacement (relative to recvbuf) at which to place data from iproc
                        MPI_INTEGER,& ! data type of receive elements
                        MPI_COMM_WORLD,&
                        ierr) 

    ! Now all has been received in B, so copy out to A (this is a restriction of MPI)
    do icol = 1,ncol
        A(1:nrow,icol) = B(:,icol)
    end do
    deallocate(B)



end subroutine

subroutine mympi_init()
    implicit none
    integer :: ierr,flag

    ! Test whether intialized already, and, if not, initialize
    call MPI_initialized(flag,ierr)
    if (flag==0) then 
        call MPI_init(ierr)

        ! Get the number of processors
        call MPI_Comm_size(MPI_COMM_WORLD,nproc,ierr)

        ! Get the processor id
        call MPI_Comm_rank(MPI_COMM_WORLD,iproc,ierr)

        ! Initialize mpi
        mympi_initialized = .true.

    end if

end subroutine
subroutine mympi_finalize()
    implicit none
    integer :: ierr

    call checkinit()
    call MPI_Finalize(ierr)

end subroutine 

subroutine mympi_hello_world()
    implicit none
    character(len=MPI_MAX_PROCESSOR_NAME) :: procname
    integer :: lenprocname,ierr

    call checkinit()

    ! Get the processor name
    call MPI_Get_processor_name(procname,lenprocname,ierr)

    ! Display
    write(*,'(3A,i0,A,i0,A)') 'Hello from ',trim(procname),'. I am process ',iproc,' of ',nproc,'.'

end subroutine 

function cumsum(x_in) result(x_out)
    implicit none
    integer, dimension(1:), intent(in) :: x_in
    integer, dimension(1:size(x_in)) :: x_out
    integer :: i,accum

    accum = 0
    do i = 1,size(x_in)
        accum = accum + x_in(i)
        x_out(i) = accum
    end do

end function 


end module mod_mympi

module mod_mympi_test
    use mod_mympi
    implicit none
contains
    
    ! mympi_init and mympi_finalize should be called before and after this routine
    subroutine test_mympi
        implicit none
        call test_mympi_core(1,7)
        call test_mympi_core(2,7)
        call test_mympi_core(3,7)
        call test_mympi_core(4,7)
        call test_mympi_core(5,7)
        call test_mympi_core(11,7)
        call test_mympi_core(24,7)
        call test_mympi_core(31,7)
        call test_mympi_core(1017,7)
        if (mympi_iproc()==0) print*,'test_mympi: passed'
    end subroutine
    subroutine test_mympi_core(ncol,ndim1)
        implicit none
        integer, intent(in) :: ncol, ndim1 
        real(rt), allocatable :: x_d_1d(:),x_d_2d(:,:),x_d_3d(:,:,:)
        real(4), allocatable :: x_f_1d(:),x_f_2d(:,:)
        integer, allocatable :: x_i_1d(:),x_i_2d(:,:)
        logical, allocatable :: x_l_1d(:),x_l_2d(:,:)
        real(rt), allocatable :: y_d_1d(:),y_d_2d(:,:),y_d_3d(:,:,:)
        real(4), allocatable :: y_f_1d(:),y_f_2d(:,:)
        integer, allocatable :: y_i_1d(:),y_i_2d(:,:)
        logical, allocatable :: y_l_1d(:),y_l_2d(:,:)
        integer, allocatable :: lo(:),hi(:),nelts(:),rqstlst(:)
        integer :: a,b,itmp,nproc,icol,cnt,itmp2
        real(rt) :: tmp

        nproc = mympi_nproc()
        allocate(lo(nproc),hi(nproc),nelts(nproc))
        allocate(rqstlst(ncol))

        ! Make space
        allocate(x_d_1d(ncol),x_d_2d(ndim1,ncol),x_d_3d(ndim1,ncol,3)) 
        allocate(x_f_1d(ncol),x_f_2d(ndim1,ncol)) 
        allocate(x_i_1d(ncol),x_i_2d(ndim1,ncol)) 
        allocate(x_l_1d(ncol),x_l_2d(ndim1,ncol)) 
        allocate(y_d_1d(ncol),y_d_2d(ndim1,ncol),y_d_3d(ndim1,ncol,3)) 
        allocate(y_f_1d(ncol),y_f_2d(ndim1,ncol)) 
        allocate(y_i_1d(ncol),y_i_2d(ndim1,ncol)) 
        allocate(y_l_1d(ncol),y_l_2d(ndim1,ncol)) 
        
        ! Hello world
!         call mympi_hello_world()

        ! Check list assignments
        call mympi_assign_range_list(lo,hi,ncol)

        ! Check list assignments
        nelts = hi - lo + 1
        if (any(nelts<0)) call mympi_stop('ERROR: List assignment has negative number of elements')
        if (maxval(nelts)>minval(nelts)+1) call mympi_stop('ERROR: List assignment was uneven')

        call mympi_assign_range(a,b,ncol)
        if (mympi_a(ncol)/=a) call mympi_stop('ERROR: in mympi_a')
        if (mympi_b(ncol)/=b) call mympi_stop('ERROR: in mympi_b')

        !!! Check that ever element was assigned to exactly one rank
        do icol = 1,ncol
            cnt = 0
            do itmp = 1,nproc
                if (icol>=lo(itmp) .and. icol<=hi(itmp)) cnt = cnt + 1
            end do
            if (cnt/=1) call mympi_stop('ERROR: failed to assign complete range or assigned multiply')
        end do

        ! Test barrier
        call mympi_barrier() ! prettier i/o

        ! Test mympi_allgatherv
        x_d_1d = -1e0_rt; x_d_2d = -1e0_rt
        x_f_1d = -1e0; x_f_2d = -1e0
        x_i_1d = -1  ; x_i_2d = -1
        x_l_1d = -1>0; x_l_2d = -1>0

        !!! Have each rank write its rank to x
        do icol = mympi_a(ncol),mympi_b(ncol)
            x_d_1d(  icol) = real(mympi_iproc(),rt)
            x_d_2d(:,icol) = real(mympi_iproc(),rt)
            x_f_1d(  icol) = real(mympi_iproc(),4)
            x_f_2d(:,icol) = real(mympi_iproc(),4)
            x_i_1d(  icol) = mympi_iproc()  
            x_i_2d(:,icol) = mympi_iproc()
            x_l_1d(  icol) = .true. 
            x_l_2d(:,icol) = .true.
        end do

        !!! Gather
        call mympi_allgatherv(x_d_1d)
        call mympi_allgatherv(x_d_2d)
        call mympi_allgatherv_r4(x_f_1d)
        call mympi_allgatherv_r4(x_f_2d)
        call mympi_allgatherv(x_i_1d)
        call mympi_allgatherv(x_i_2d)
        call mympi_allgatherv(x_l_1d)
        call mympi_allgatherv(x_l_2d)

        !!! Test that it is as it should be
        do icol = 1,ncol
            ! Search for who this column was assigned to
            do itmp = 1,nproc
                if (icol>=lo(itmp) .and. icol<=hi(itmp)) exit
            end do
            if (itmp<=nproc) then ! found an assignment
                itmp = itmp - 1
!                 print*,'icol',icol,'x_d_1d(icol)',x_d_1d(icol),'itmp',itmp
                if (   (x_d_1d(  icol) /= real(itmp,rt))) call mympi_stop('ERROR: failed test mympi_allgatherv d_1d')
                if (any(x_d_2d(:,icol) /= real(itmp,rt))) call mympi_stop('ERROR: failed test mympi_allgatherv d_2d')
                if (   (x_f_1d(  icol) /= real(itmp,4))) call mympi_stop('ERROR: failed test mympi_allgatherv f_1d')
                if (any(x_f_2d(:,icol) /= real(itmp,4))) call mympi_stop('ERROR: failed test mympi_allgatherv f_2d')
                if (   (x_i_1d(  icol) /= itmp))         call mympi_stop('ERROR: failed test mympi_allgatherv i_1d')
                if (any(x_i_2d(:,icol) /= itmp))         call mympi_stop('ERROR: failed test mympi_allgatherv i_2d')
                if (   (x_l_1d(  icol) .neqv. .true. ))      call mympi_stop('ERROR: failed test mympi_allgatherv l_1d')
                if (any(x_l_2d(:,icol) .neqv. .true.))       call mympi_stop('ERROR: failed test mympi_allgatherv l_2d')
            end if
        end do

        !!! Wait and display pass
        call mympi_barrier()

        ! Test reduction

        !!! Have each rank write its rank to x
        x_d_1d = real(mympi_iproc(),rt)
        x_d_2d = real(mympi_iproc(),rt)
        x_d_3d = real(mympi_iproc(),rt)

        !!! Reduce
        call random_number(y_d_1d)
        call random_number(y_d_2d)
        call random_number(y_d_3d)

        call mympi_allreduce(y_d_1d,x_d_1d,MPI_SUM)
        call mympi_allreduce(y_d_2d,x_d_2d,MPI_SUM)
        call mympi_allreduce(y_d_3d,x_d_3d,MPI_SUM)

        
        cnt = 0
        do itmp2 = 0,mympi_nproc()-1
            cnt = cnt + itmp2 + 2

        end do

        itmp = -1000000000
        call mympi_allreduce(itmp,2+mympi_iproc(),MPI_SUM)
        if (itmp/=cnt) call mympi_stop('ERROR: failed test mympi_allreduce i_0d')

        call mympi_barrier()
        stop 'checking'

        !!! Test that it is as it should be
        tmp = 0e0_rt
        do itmp = 0,nproc-1
            tmp = tmp + real(itmp,rt)
        end do
        x_d_1d = tmp
        x_d_2d = tmp
        x_d_3d = tmp

        if (any(y_d_1d/=x_d_1d)) call mympi_stop('ERROR: failed test mympi_allreduce d_1d')
        if (any(y_d_2d/=x_d_2d)) call mympi_stop('ERROR: failed test mympi_allreduce d_2d')
        if (any(y_d_3d/=x_d_3d)) call mympi_stop('ERROR: failed test mympi_allreduce d_3d')

        ! Check mympi_ibcast
        x_i_2d = 0
        if (mympi_iproc()==0) then
            do icol = 1,ncol
                x_i_2d(:,icol) = icol
            end do
        end if
        do icol = 1,ncol
            call mympi_ibcast(rqstlst(icol),x_i_2d(:,icol),0)
        end do
        call mympi_waitall(rqstlst) 
        do icol = 1,ncol
            if (any(x_i_2d(:,icol)/=icol)) call mympi_stop('ERROR: failed test mympi_ibcast')
        end do


    end subroutine


end module mod_mympi_test

