! Grey Gordon, 2014 
module mod_mynlopt
    implicit none
    include 'nlopt.f'

contains

    subroutine sub_mynlopt(x,f,algorithm,a_in,b_in,e_in,printmod_in,xtolerrel_in,xtolerabs_in,ftolerrel_in,ftolerabs_in,maxiter_in,grad_in)
        implicit none
        real(8), intent(inout) :: x(:)
        integer, intent(in) :: algorithm ! call using one of the constants named in nlopt.f which is included in this module
        real(8), intent(in), optional :: a_in(:), b_in(:), e_in(:)
        integer, intent(in), optional :: printmod_in,maxiter_in
        real(8), intent(in), optional :: xtolerrel_in, xtolerabs_in, ftolerrel_in, ftolerabs_in
        interface
            function f(x)
                implicit none
                real(8), intent(in) :: x(:)
                real(8) :: f
            end function 
        end interface
        interface
            ! Grad shoudl return both 
            subroutine grad(fx,gradx,x)
                implicit none
                real(8), intent(in) :: x(:)
                real(8), intent(out) :: fx,gradx(:)
            end subroutine
        end interface
        procedure(grad), optional :: grad_in
        ! local
        real(8) :: f_data_local(1),fx,xtolerabs,xtolerrel,ftolerabs,ftolerrel,fgoodenough
        integer :: n,ires,uid,iter,printmod,maxiter
        integer(8) :: ptr
        character(len=10) :: nstr

        iter = 0
        printmod = 1; if (present(printmod_in)) printmod = printmod_in
        fgoodenough = -huge(0d0)
        ftolerabs = 1d-12; if (present(ftolerabs_in)) ftolerabs = ftolerabs_in ! -huge(0d0)
        ftolerrel = 1d-12; if (present(ftolerrel_in)) ftolerrel = ftolerrel_in ! -huge(0d0)  ! -1d0 ! any negative value means no stopping
        xtolerabs = 1d-12; if (present(xtolerabs_in)) xtolerabs = xtolerabs_in
        xtolerrel = 1d-12; if (present(xtolerrel_in)) xtolerrel = xtolerrel_in
        maxiter = 100000; if (present(maxiter_in)) maxiter = maxiter_in
        uid = 6
        
        n = size(x)

        write(nstr,'(i10)') n
        nstr = adjustl(nstr)

        ! If the chosen algorithm is known, display it
        if (printmod==1) then
            select case (algorithm)
                case (NLOPT_GN_DIRECT_L)
                    write(uid,'(A)') 'NL: algorithm DIRECT-L selected (dividing rectangles with local search)'
                case (NLOPT_GN_CRS2_LM)
                    write(uid,'(A)') 'NL: algorithm CRS2-LM selected (controlled random search with local mutation)'
                case (NLOPT_LN_COBYLA)
                    write(uid,'(A)') 'NL: algorithm COBYLA selected (constrained optimization by linear approximation)'
                case (NLOPT_LN_PRAXIS) 
                    write(uid,'(A)') 'NL: algorithm PRAXIS (principle axis method)'
                case (NLOPT_LN_NELDERMEAD)
                    write(uid,'(A)') 'NL: algorithm Nelder Mead'
                case (NLOPT_LD_LBFGS)
                    write(uid,'(A)') 'NL: algorithm Low Storage BFGS'
            end select
        end if

        ! Initialize nlopt
        ptr = 0
        call nlo_create(ptr,algorithm,n)

        ! Set the stopping criteria 
        call nlo_set_stopval(ires,ptr,fgoodenough) ! stops if function value less than fgoodenough
        call nlo_set_ftol_abs(ires,ptr,ftolerabs) ! stops if function value less than ftoler found
        call nlo_set_ftol_rel(ires,ptr,ftolerrel) ! stops if relative change in function value is less than this
        call nlo_set_xtol_abs1(ires,ptr,xtolerabs) ! absolute stopping (checks absolute change in x)
        call nlo_set_xtol_rel(ires,ptr,xtolerrel) ! relative stopping (checks percentage change in x)
        call nlo_set_maxeval(ires,ptr,maxiter) ! maximum number of evaluations

        ! Set bounds (if applicable)
        if (present(a_in)) call nlo_set_lower_bounds(ires,ptr,a_in)
        if (present(b_in)) call nlo_set_upper_bounds(ires,ptr,b_in)

        ! Set initial step size (if applicable)
        if (present(e_in)) call nlo_set_initial_step(ires,ptr,e_in)

        ! Specify the function to be minimized
        call nlo_set_min_objective(ires,ptr,nloptfcn,f_data_local)

        ! Do the optimization
        call nlo_optimize(ires,ptr,x,fx)
        if (printmod>0) write(*,'(A,f18.12,A,'//trim(nstr)//'f18.12)',advance='no') 'NL: completed:  fx ',fx,' x ',x

        ! Report success or failure
        if (printmod>0) then
            select case (ires)
                case (NLOPT_STOPVAL_REACHED);  write(uid,'(A)') ' : success, f<=fgoodenough'
                case (NLOPT_FTOL_REACHED);     write(uid,'(A)') ' : success, ftolerabs or ftolerrel achieved'
                case (NLOPT_XTOL_REACHED);     write(uid,'(A)') ' : success, xtolerabs or xtolerrel reached'
                case (NLOPT_SUCCESS);          write(uid,'(A)') ' : success, generic'
                case (NLOPT_MAXEVAL_REACHED);  write(uid,'(A)') ' : failure, maxeval reached'
                case (NLOPT_MAXTIME_REACHED);  write(uid,'(A)') ' : failure, maxtime reached'
                case (NLOPT_INVALID_ARGS);     write(uid,'(A)') ' : failure, invalid arguments'
                case (NLOPT_OUT_OF_MEMORY);    write(uid,'(A)') ' : failure, out of memory'
                case (NLOPT_ROUNDOFF_LIMITED); write(uid,'(A)') ' : failure, roundoff errors limit further progress'
                case (NLOPT_FORCED_STOP);      write(uid,'(A)') ' : failure, forced stop'
                case (NLOPT_FAILURE);          write(uid,'(A)') ' : failure, generic'
                case default;                  write(uid,'(A)') ' : termination code not found'
            end select
        end if

        ! Free memory
        call nlo_destroy(ptr)

    contains
        
        subroutine nloptfcn(fx, n, x, grad, need_gradient, f_data)
            implicit none
            integer, intent(in) :: need_gradient
            integer, intent(in) :: n
            real(8) :: x(n), grad(n)
            real(8), intent(out) :: fx
            real(8), intent(inout) :: f_data(*)
            ! local
            integer :: i
            real(8) :: step,tmp(n)

            if (need_gradient/=0) then
!                 STOP 'ERROR: nloptfcn: gradient requested, but have not programmed this'

                if (present(grad_in)) then

                    ! Use analytic gradient if provided
                    call grad_in(fx,grad,x)
                    
                else
                    ! Use numerical derivatves if not availanle
                    fx = f(x)
                    
                    do i = 1,n
                        step = max(abs(x(i))*1d-6,1d-6)
                        tmp = x
                        tmp(i) = x(i) + step
                        grad(i) = (f(tmp) - fx)/step
                    end do

                end if

!                 print*,'Gradient is ',grad
                
            else

                fx = f(x)

            endif

            iter = iter + 1

            if (printmod>0 .and. mod(iter,printmod)==0) then
                if (need_gradient==0) then
                    write(*,'(A,i6,A,ES14.6E3,A,'//trim(nstr)//'f18.12)') 'NL: iter ',iter,' fx ',fx,' x ',x
                else
                    write(*,'(A,i6,A,ES14.6E3,2(A,'//trim(nstr)//'f18.12))') 'NL: iter ',iter,' fx ',fx,' x ',x,' grad ',grad
                end if
            end if

        end subroutine

    end subroutine

end module

module mod_mynlopt_test
    use mod_mynlopt
    implicit none

contains

    function convexprob(x) result(f)
        implicit none
        real(8), dimension(1:), intent(in) :: x
        real(8) :: f
        !local
        integer :: i,n
        real(8) :: f_tmp
        n = size(x)
        f_tmp = 0.d0
        f_tmp = f_tmp + log(10d0-sum(x(1:n)))
        do i = 1,n
             f_tmp = f_tmp + log(x(i))
        end do
        f = -f_tmp

        

    end function


    function rosenbrock(x) result(f)
        implicit none
        real(8), dimension(1:), intent(in) :: x
        real(8) :: f
        !local
        integer :: i,n
        real(8) :: f_tmp
        n = size(x)
        f_tmp = 0.d0
        do i = 1,n-1
             f_tmp = f_tmp + 100.d0*(x(i+1)-x(i)**2)**2 + (1.d0-x(i))**2
        end do
        f = f_tmp
    end function

    subroutine test_mynlopt()
        implicit none
        real(8) :: x(3), a(3), b(3)

        x = [.1d0,.2d0,.3d0]
        a = 0d0
        b = 100d0

        call sub_mynlopt(x,rosenbrock,NLOPT_LD_LBFGS,a_in=a,b_in=b,printmod_in=1)
        call sub_mynlopt(x,convexprob,NLOPT_LD_LBFGS,a_in=a,b_in=b,printmod_in=1)
        call sub_mynlopt(x,rosenbrock,NLOPT_LN_COBYLA,a_in=a,b_in=b,printmod_in=250)
        call sub_mynlopt(x,rosenbrock,NLOPT_LN_NELDERMEAD,a_in=a,b_in=b,printmod_in=250)
        call sub_mynlopt(x,rosenbrock,NLOPT_LN_PRAXIS,a_in=a,b_in=b,printmod_in=250)
        call sub_mynlopt(x,rosenbrock,NLOPT_GN_DIRECT_L,a_in=a,b_in=b,printmod_in=250)
        call sub_mynlopt(x,rosenbrock,NLOPT_GN_CRS2_LM,a_in=a,b_in=b,printmod_in=250)
        
        

    end subroutine

end module 
