! http://stackoverflow.com/questions/6931846/sleep-in-fortran
! module mod_sleep
!     implicit none
! 
!     private :: callcsleep
! 
!     interface
!         function callcsleep(seconds)  bind ( C, name="sleep" )
!             use, intrinsic :: iso_c_binding, only: c_int
!             import
!             implicit none
!             integer (c_int) :: callcsleep
!             integer (c_int), value :: seconds
!         end function 
!     end interface
! 
! contains
! 
!     subroutine sleep(sec)
!         use, intrinsic :: iso_c_binding, only: c_int
!         implicit none
!         integer, intent(in) :: sec
!         integer (c_int) :: sec_c
!         integer (c_int) :: err
! 
!         sec_c = int(sec,c_int)
! 
!         err = callcsleep(sec_c)
! 
!     end subroutine
! 
! end module 

module mod_parminNdim
    implicit none

!     interface sub_parCRS2LM
!         module procedure :: sub_parCRS2LM_v2
!     end interface

contains


    ! Controlled random search
    ! CRS attempts to solve min f(x) s.t. x in [a,b] (where all are in Rn)
    ! It generates N random uniform points.
    ! It takes the worst point and tries to get rid of it by reflecting it
    ! (like in Nelder-Mead) through a simplex of points. This simplex is chosen 
    ! by randomly selecting n+1 points (whose function values are known) with 
    ! replacement, giving a total of n+1 points. 
    !
    ! The centroid of the simplex is then used for reflecting the worst point. 
    ! If there reflected point improves on the worst point (and is in a,b), it is
    ! accepted. O/w, the process repeats.
    ! 
    ! This is modified for parallel purposes by attempting to simultaneously get 
    ! rid of several of the worst points simultaneously. The parallel design here,
    ! in contrast to CRS2LM, is more advantageous when the functions time to completion
    ! is not uniform across x values. The algorithm itself is more global.
    !
    ! The optional argument sleeptime is useful when there are only a few images. 
    ! It specifies how long the master should sleep if no free workers are found. This
    ! keeps processing cycles free for those images actually doing work. 
    subroutine sub_parCRS(xstar,f,a,b,sleeptime)
        use mod_mympi
        use mod_matlab
        use mod_util, only: sleep
        implicit none
        real(8), intent(inout) :: xstar(:)
        real(8), intent(in) :: a(:),b(:)
        integer, intent(in), optional :: sleeptime
        interface 
            function f(x) 
                implicit none
                real(8), dimension(:), intent(in) :: x
                real(8) :: f
            end function
        end interface
        ! local coarrays
        integer, allocatable :: flag[:]
        real(8), allocatable :: slavex(:)[:]
        real(8), allocatable :: slavey(:)[:]
        logical, allocatable :: finishedtask(:)[:]
        ! local other
        real(8), allocatable :: X(:,:),R(:,:),g(:),smx(:,:)
        real(8), allocatable :: y(:),yR(:),tmpy(:)
        integer, allocatable :: tmpI(:),lasti(:), fevaldsn(:)
        real(8) :: tmp
        integer :: nx, N, M, ix, iter, i, ism, fevals, inner, who, j
        
        if (num_images()<=1) error stop 'ERROR: sub_parCRS: must have at least two images!' 

        ! For random results, initialize based on the clock
        call init_random_seed(this_image())

        nx = size(xstar)
        if (size(a)/=nx) error stop 'ERROR: sub_parCRS2LM: size of a and x do not agree'
        if (size(b)/=nx) error stop 'ERROR: sub_parCRS2LM: size of b and x do not agree'
        if (any(a>=b)) error stop 'ERROR: sub_parCRS2LM: a>=b at least once'

        ! Check that input x is in a,b for all i
        if (any(xstar<a) .or. any(xstar>b)) error stop 'ERROR: sub_parCRS2LM: x is not in a,b for all dimensions'

        ! M is the number to simultaneously try to get rid of
        M = num_images()

        ! N is number of trial points (the pool of points)
        N = max(5*M, ((100+nx)/num_images()+1)*num_images(), nx+2+M)

        if (N-M<=nx+1) then
            error stop 'ERROR: sub_parCRS2LM: too few N points set relative to M'
        end if

        allocate(flag[*])
        allocate(slavex(nx)[*])
        allocate(slavey(num_images())[*])
        allocate(finishedtask(num_images())[*])


        flag = 0

        ! If not master, go to the "slave" routine and wait for a x to evaluate
        if (this_image()/=1) then
            sync images(this_image())
            call slave()
            return
        end if

        ! Generate N random points uniformly from [a,b]
        allocate(X(nx,N))
        allocate(y(N))
        allocate(R(nx,M))
        allocate(yR(M))
        allocate(g(nx))
        allocate(smx(nx,nx+1))                    
        allocate(tmpy(N),tmpI(N))
        allocate(lasti(num_images()))
        allocate(fevaldsn(num_images()))

        inner = 1

        ! For random results, initialize based on the clock
        call init_random_seed()

        ! Generate initial pool of points
        X(:,1) = xstar
        do i = 2,N
            do ix = 1,nx
                call random_number(tmp)
                X(ix,i) = tmp*a(ix) + (1d0-tmp)*b(ix)
            end do
        end do

        fevaldsn = 0

        lasti = -1
        finishedtask = .true.

        sync images(*)

        ! Evaluate at each of the N points in parallel
        write(*,'(A,i4,A)') 'PARCRS: Generating initial pool of ',N,' points'
        do i = 1,N
            ! Get a slave. If there were previously assigned a task, then record its value
            who = -1
            do while (who==-1) 
                who = findworker()
            end do

            write(*,'(A,i4,A,i4,A,i4,A)',advance='no') 'PARCRS: image ', who, ' solving for point ',i,' of ',N,' x = '
            call printx(X(:,i))
            write(*,*) 

            ! Get previously computed value
            if (lasti(who)>=1) then 
                ! Ensure that slave has same view as master 
                sync images(who) ! sync values with worker and allow them to prepare for new x values 
                y(lasti(who)) = slavey(who)
                lasti(who) = -1
            end if

            ! Request that they solve for x and store it in slavey
            slavex(:)[who] = X(:,i) ! point to be executed
            flag[who] = 1 ! execute
            lasti(who) = i ! save location to store
            finishedtask(who) = .false.
            sync images(who) ! release x to worker who will then begin executing

            ! For statistical purposes, collect who was assigned that feval
            fevaldsn(who) = fevaldsn(who) + 1

        end do

        print*,'waiting for everyone to finish'

        ! Wait for everyone to finished
        do while (any(lasti>=1))

            ! See if they have finished but their value has not yet been stored
            do who = 2,num_images()
                if (lasti(who)>=1 .and. isworkerfree(who)) then
                    ! Now, found one that finished, store its value
                    sync images(who)
                    y(lasti(who)) = slavey(who)
                    lasti(who) = -1
                end if
            end do

        end do
        fevals = N

        lasti = -1
        finishedtask = .true.

        print*,'done proceeding to main loop'

        do iter = 1,10000

!             print*,'finishedtask',finishedtask,'lasti',lasti

!             ! From any free workers (i.e., ones who have completed any previous task),
!             ! get previously computed values (if there are any). This is not executed in iter==1
!             ! If there are no free workers, then wait for them here.
!             do who = 2,num_images()
! 
!                 if (isworkerfree(who) .and. lasti(who)>=1) then
! 
!                     if (iter==1) error stop 'this should not evaluate to true'
! 
!                     ! Ensure that slave has same view as master 
!                     sync images(who)
! 
!                     yR(lasti(who)) = slavey(who)
! 
!                     print*,'yR',yR(lasti(who)),'y(N)',y(N)
! 
!                     ! Add it to the pool if better than worst point
!                     if (yR(lasti(who))<y(N)) then
! 
!                         y(N) = yR(lasti(who))
!                         X(:,N) = R(:,lasti(who))
!                         lasti(who) = -1
! 
!                         ! Resort so that 
!                         call sort2(tmpy,tmpI,y) 
!                         X = X(:,tmpI) ! sort X
!                         y = tmpy ! sort y
! 
!                     end if
! 
!                     lasti(who) = -1
!                 end if
! 
!             end do

            ! Sort the current points (ascending)
            call sort2(tmpy,tmpI,y) 
            X = X(:,tmpI) ! sort X
            y = tmpy ! sort y

            ! Display progress
            write(*,'(2(A,i8),2(A,e14.6),A)',advance='no') 'PARCRS: iter',iter,' fevals ',fevals,'  best ',y(1),'  worst ',y(N),'  xbest = '; call printx(X(:,1)); write(*,*) 
 
            ! In case exit either through maxiter being reached or satisfying exit criteria, save xstar
            xstar = X(:,1)

            ! Check stopping criteria
            if (abs(y(1)-y(N))<1d-5 .or. iter==10000) then

                if (abs(y(1)-y(N))<1d-5) then
                    write(*,'(A)') 'PARCRS: convergence criteria satisfied'
                else
                    write(*,'(A)') 'PARCRS: max iterations reached'
                end if

                do i = 1,num_images()

                    ! Put the best
                    slavex(:)[i] = xstar

                    ! Send stop message to workers
                    flag[i] = -1

                end do

                print*, 'fevaldsn was ',fevaldsn

                sync images(*)

                return

            end if

            ! Wait for a slave. If there were previously assigned a task, then record its value.
!             print*,'waiting for a free worker'
            who = -1
            do while (who==-1) 
                who = findworker()

                if (present(sleeptime)) then

                    ! tmp = mpi_wtime()
                    ! do while (mpi_wtime() - tmp < sleeptime) 

                    ! end do

                    call sleep(sleeptime) 
                    ! call execute_command_line ("sleep "//trim(num2str(sleeptime,'(i4)')), wait=.true.)
                end if
            end do

            ! For every free worker, have them execute a function
!             print*,'looping through search for free workers'
            do j = 2,num_images()

                if (isworkerfree(j)) then
                    who = j
!                     print*,'worker ',j,' is free'
                else
                    cycle
                end if

                ! If he previously computed a task, then save it
!                 print*,'collecting previous point and sorting'
                if (lasti(who)>=1) then

                    if (iter==1) error stop 'this should not evaluate to true'

                    ! Ensure that slave has same view as master 
                    sync images(who)

                    yR(lasti(who)) = slavey(who)

!                     print*,'yR',yR(lasti(who)),'y(N)',y(N)

                    ! Add it to the pool if better than worst point
                    if (yR(lasti(who))<y(N)) then

                        y(N) = yR(lasti(who))
                        X(:,N) = R(:,lasti(who))
                        lasti(who) = -1

                        ! Resort so that 
                        call sort2(tmpy,tmpI,y) 
                        X = X(:,tmpI) ! sort X
                        y = tmpy ! sort y

                    end if

                    lasti(who) = -1

                end if

                ! Form a random simplex excluding current worst point 
!                 print*,'generating random point'
                do ism = 1,nx+1
                    
                    ! Pure CRS: choose simplex completely randomly
                    ! smx(:,ism) = x(:,randi(1,N-1))

                    ! CRS with local mutation where simplex always includes best point
                    if (ism==1) then
                        smx(:,ism) = x(:,1) 
                    else
                        smx(:,ism) = x(:,randi(1,N-1))
                    end if

                    ! Genetic algorithm: choose according to fitness
                    ! tmpy = 1d0/N - min(y(1:N),1d5)/sum(min(y(1:N),1d5)) ! let this be the probabiliy of *not* being chosen
    !                 tmpy = y(1:N) - y(1) ! >= 0 
    !                 tmpy = tmpy / sum(tmpy) ! let this be the probability of *not* being chosen
    !                 print*,'pdf',tmpy
    !                 print*,'1-pdf',1d0-tmpy
    !                 tmpy = cumsum(tmpy) ! let thsis be the pdf
    !                 print*,'sum(tmpy)',sum(tmpy)
    !                 call random_number(tmp)
    !                 call bsearch(randitmp,tmp,tmpy) ! bsearch will never let this be N
    !                 smx(:,ism) = x(:,randitmp)

                end do

                ! Compute the centroid
                g = sum(smx,2)/(nx+1)

                ! Reflect the worst point
                R(:,who) = 2d0*g - X(:,N) ! pure reflection

                ! If goes through boundary, then bound at boundary
                R(:,who) = max(R(:,who),a)
                R(:,who) = min(R(:,who),b)

                ! To make more local, can 50% of the time combine with local search which is convex combination with best point
                call random_number(tmp)
                if (tmp<=.5d0) then
                    do ix = 1,nx
                        call random_number(tmp)
                        R(ix,who) = tmp*x(ix,1) + (1d0-tmp)*R(ix,who) ! convex combination, so in [a,b]
                    end do
                end if
                
                ! Request that they solve for x and store it in slavey
!                 print*,'requesting new point'
                slavex(:)[who] = R(:,who) ! point to be executed
                flag[who] = 1 ! execute
                lasti(who) = who ! store that they are working on R(:,who)
!                 print*,'master: releasing to worker'
                finishedtask(who) = .false.
                sync images(who) ! release x to worker who will then begin executing
!                 print*,'master: finished releasing'

                fevals = fevals + 1
                fevaldsn(who) = fevaldsn(who) + 1

!                 print*,'finishedtask',finishedtask

            end do

!             do while 
! 
!                 ! Get previously computed value
!                 if (lasti(who)>=1) then 
!                     ! Ensure that slave has same view as master 
!                     sync images(who)
!                     y(lasti(who)) = slavey(who)
!                     lasti(who) = -1
!                 end if
! 
! 
!             ! Generate random simplices and trial points
!             do i = 1,M
! 
!                 ! Potentially infinite loop
!                 do inner = 1,1000000
! 
!                     ! Get some random simplex draws using the best point x(:,1) 
!                     ! along with N more randomly chosen points
!                     smx(:,1) = x(:,1)
!                     do ism = 2,nx+1
! 
!                         ! Pick an integer between 1 to N-M and use that x as a simplex point
!                         smx(:,ism) = x(:,randi(1,N-M))
! 
!                     end do
! 
!                     ! Compute the centroid
!                     g = sum(smx,2)/(nx+1)
! 
!                     ! Reflect the ith worst point
!                     R(:,i,1) = 2d0*g - X(:,N-i+1)
! 
!                     ! Check that R(:,i) is in [a,b], if so, exit the loop. O/w, go to the top and form a new simplex.
!                     if (all(R(:,i,1)>=a) .and. all(R(:,i,1)<=b)) exit 
! 
!                     ! Note: the above reflectio nmay fail even with a large number of iterations if the point is close to the boundary. 
!                     ! May need to adjust this to bound the reflection point at a,b (elementwise) to prevent failures.
!                     if (inner>1000) then
!                         ! If here after one thousand tries, likely at a border case. So just bound element wise and proceed
!                         R(:,i,1) = max(R(:,i,1),a)
!                         R(:,i,1) = min(R(:,i,1),b)
!                         exit
!                     end if
!                 
!                 end do
! 
!                 ! Generate another trial point based on a random combination between best point and the reflected point 
!                 do ix = 1,nx
!                     call random_number(tmp)
!                     R(ix,i,2) = tmp*x(ix,1) + (1d0-tmp)*R(ix,i,1) ! convex combination, so in [a,b]
!                 end do
! 
!             end do
! 
!             ! Now evaluate the functions
!             do j = 1,2
!                 do i = 1,M
! 
!                     
!                     who = 1 + mod(i,num_images()) ! Who is responsible for this i?
! 
!                     slavex(:)[who] = R(:,i,j) ! point to be executed
!                     flag[who] = 1 ! execute
!                     sync images(who) ! release x to worker who will then begin executing
! 
!                     if (who==1) then
!                         slavey(1) = f(slavex) ! release x to master
!                         sync images(*) ! collect y from everyone
!                         yR(i-num_images()+1:i-1,j) = slavey(2:num_images()) ! the slaves have stored the value in slavey(them)[1]
!                         yR(i,j) = slavey(1)
!                     end if
! 
!                 end do
!             end do
!             fevals = fevals + 2*M
! 
!             ! Check whether the trial point is accepted or not
!             do i = 1,M
! 
!                 ! Is the reflected point better than the ith worst point? If so, 
!                 ! then accept it unless the 2nd point is better than the reflected point and 
!                 ! better than the best point.
!                 ! Although this discard potentially useful info from the second point, prevents
!                 ! shrinking too quickly. Most of the time, the reflected point is not better anyway.
!                 if (opt==0) then
!                     ! DEFAULT
!                     if (yR(i,2)<y(1) .and. yR(i,2)<yR(i,1)) then
! 
!                         ! Accept the second point
!                         y(N-i+1) = yR(i,2)
!                         x(:,N-i+1) = R(:,i,2)
! 
!                     ! The default except in exceptional cases
!                     elseif (yR(i,1)<y(N-i+1)) then
! 
!                         ! Accept it
!                         y(N-i+1) = yR(i,1)
!                         x(:,N-i+1) = R(:,i,1)
!                     
!                     ! Otherwise, is the combination of the reflected and best better? If so, then accept it.
!                     elseif (yR(i,2)<y(N-i+1)) then
! 
!                         ! Accept it
!                         y(N-i+1) = yR(i,2)
!                         x(:,N-i+1) = R(:,i,2)
! 
!                     ! O/w, do nothing.
!                     end if
!                 else
!                     ! MORE LOCAL SEARCH
!                     if (min(yR(i,1),yR(i,2))<y(N-i+1)) then
!                         ! Replace with best of yR(i,1) and yR(i,2). Otherwise, do nothing.
!                         if (yR(i,1)<yR(i,2)) then
!                             y(N-i+1) = yR(i,1)
!                             x(:,N-i+1) = R(:,i,1)
!                         else
!                             y(N-i+1) = yR(i,2)
!                             x(:,N-i+1) = R(:,i,2)
!                         end if
!                     end if
!                 end if
! 
!             end do

        end do

    contains

        ! a routine returning the index of a free slave
        ! if no worker is available, who is -1
        function findworker() result(who)
            implicit none
            integer :: who
            ! local
            integer :: i
            logical :: ft

            do i = 2,num_images()
!                 call atomic_ref(ft,finishedtask(i))
                ft = finishedtask(i)
                if (ft) then
                    who = i
                    return
                end if
            end do

            ! if here, no one was finished task.
            who = -1

        end function
        function isworkerfree(who) result(tfout)
            implicit none
            integer, intent(in) :: who
            logical :: tfout
            ! local
            logical :: tf
!             call atomic_ref(tf,finishedtask(who))
            tf = finishedtask(who)
            tfout = tf
        end function

        ! Do something according to the value of flag(this_image())[1]
        !  1 := evaluate slavex and store in slavey
        ! -1 := exit the slave routine
        subroutine slave()
            implicit none
            ! local
            integer :: me

            me = this_image()

            do 

                ! Logic here is wait at sync images(1) for data to be placed from master.
                ! When it is, check the flag that was sent to me. The flag can only say
                ! two things, either evaluate f or stop.
                ! In the first case, evaluate f at slavex and store in slavey. Then call 
                ! sync images(1) again to tell the master you are finished. 
                ! In the second case, you want to terminate. So, slavex contains the best
                ! point, just return.

                ! Create a barrier, wait for sync images(me) to be called
                sync images(1) ! INPUT: should have same view of slavex and flag

                ! Now sync images(me) has been called, so have same view of slavex and flag.

                ! If flag==1, execute f(x) and store on master.
                if (flag==1) then
!                     call atomic_define(finishedtask(me)[1],.false.)
!                     print*,'worker ',me,' begin working'
                    slavey(me)[1] = f(slavex)
!                     print*,'worker ',me,' finished working'
                    finishedtask(me)[1] = .true.
!                     call atomic_define(finishedtask(me)[1],.true.)

                    sync images(1) ! OUTPUT: have computed value
                end if

                if (flag==-1) then
                    xstar = slavex
!                     sync images(1) ! OUTPUT: have computed value
                    exit
                end if

            end do

        end subroutine


        subroutine printx(x)
            implicit none
            real(8), intent(in) :: x(:)
            write(*,'(A)',advance='no') '['
            do ix = 1,nx-1
                write(*,'(d16.8,A)',advance='no') x(ix),', '
            end do
            write(*,'(d16.8)',advance='no') x(nx)
            write(*,'(A)',advance='no') ']'
        end subroutine

        ! Generate a random integer in a,...,b
        function randi(a,b) result(i)
            implicit none
            integer, intent(in) :: a,b
            integer :: i
            real(8) :: tmp

            ! Generate a U[a,b+1-epsilon]
            ! Then floor to get a in 
            call random_number(tmp)
            tmp = tmp*(b+1-a) + a ! U[a,b+1]
            tmp = min(tmp,b+.99999999d0)
            i = floor(tmp)

            if (floor(tmp)<a .or. floor(tmp)>b) STOP 'ERROR: randi: failed to generate i in a,...,b'

        end function
    end subroutine

!     ! Controlled random search with local mutation 
!     ! CRSLM attempts to solve min f(x) s.t. x in [a,b] (where all are in Rn)
!     ! It generates N random uniform points.
!     ! It takes the worst point and tries to get rid of it by reflecting it
!     ! (like in Nelder-Mead) through a simplex of points. This simplex is chosen 
!     ! by taking the best point found so far, along with n randomly points (whose 
!     ! function values are known) with replacement, giving a total of n+1 points. 
!     ! The centroid of the simplex is then used for reflecting the worst point. 
!     ! If there reflected point improves on the worst point (and is in a,b), it is
!     ! accepted. If not, then this reflected point, say xR, by taking a random, 
!     ! convex combination of xR and the reflected point, say yR. If yR fails to 
!     ! on the worst point, then a new simplex is chosen randomly and the process 
!     ! repeats. If yR does improve, then it is accepted.
!     ! 
!     ! This is modified for parallel purposes by attempting to simultaneously get 
!     ! rid of the worst M points. For each of the M points, it attempts the 
!     ! procedure above. Currently though, it doesn't do the yR, only xR.
!     ! 
!     ! option controls how local vs global the search is
!     ! option = 0 or not present: more global search
!     ! option = 1: always use the best of the reflection and local search (more local search)
!     subroutine sub_parCRS2LM_v2(xstar,f,a,b,option)
!         use mod_mympi
!         use mod_matlab
!         implicit none
!         real(8), intent(inout) :: xstar(:)
!         real(8), intent(in) :: a(:),b(:)
!         integer, intent(in), optional :: option
!         interface 
!             function f(x) 
!                 implicit none
!                 real(8), dimension(:), intent(in) :: x
!                 real(8) :: f
!             end function
!         end interface
!         ! local coarrays
!         integer, allocatable :: flag[:]
!         real(8), allocatable :: slavey(:)[:],slavex(:)[:]
!         ! local other
!         real(8), allocatable :: X(:,:),R(:,:,:),g(:),smx(:,:)
!         real(8), allocatable :: y(:),yR(:,:),tmpy(:)
!         integer, allocatable :: tmpI(:)
!         real(8) :: tmp
!         integer :: nx, N, M, ix, iter, i, ism, fevals, inner, who, j, opt
! 
!         if (present(option)) then
!             opt = option
!         else
!             opt = 0
!         end if
! 
!         nx = size(xstar)
!         if (size(a)/=nx) error stop 'ERROR: sub_parCRS2LM: size of a and x do not agree'
!         if (size(b)/=nx) error stop 'ERROR: sub_parCRS2LM: size of b and x do not agree'
!         if (any(a>=b)) error stop 'ERROR: sub_parCRS2LM: a>=b at least once'
! 
!         ! Check that input x is in a,b for all i
!         if (any(xstar<a) .or. any(xstar>b)) error stop 'ERROR: sub_parCRS2LM: x is not in a,b for all dimensions'
! 
!         ! M is the number to simultaneously try to get rid of
!         M = num_images()
! 
!         ! N is number of trial points (the pool of points)
!         N = max(5*M, ((100+nx)/num_images()+1)*num_images(), nx+2+M)
! 
!         if (N-M<=nx+1) then
!             error stop 'ERROR: sub_parCRS2LM: ooo few N points set relative to M'
!         end if
! 
! 
!         allocate(flag[*])
!         allocate(slavex(nx)[*])
!         allocate(slavey(num_images())[*])
! 
!         flag = 0
! 
!         ! If not master, go to the "slave" routine and wait for a x to evaluate
!         if (this_image()/=1) then
! 
!             call slave()
! 
!             return
! 
!         end if
! 
!         ! Generate N random points uniformly from [a,b]
!         allocate(X(nx,N))
!         allocate(y(N))
!         allocate(R(nx,M,2))
!         allocate(yR(M,2))
!         allocate(g(nx))
!         allocate(smx(nx,nx+1))                    
!         allocate(tmpy(N),tmpI(N))
! 
!         inner = 1
! 
!         ! For random results, initialize based on the clock
!         call init_random_seed()
! 
!         ! Generate initial pool of points
!         X(:,1) = xstar
!         do i = 2,N
!             do ix = 1,nx
!                 call random_number(tmp)
!                 X(ix,i) = tmp*a(ix) + (1d0-tmp)*b(ix)
!             end do
!         end do
! 
!         ! Evaluate at each of the N points in parallel
!         write(*,'(A,i4,A)') 'PARCRS2LM: Generating initial pool of ',N,' points'
!         do i = 1,N
!             write(*,'(A,i4,A,i4,A)',advance='no') 'PARCRS2LM: Solving for point ',i,' of ',N,' x = '
!             call printx(X(:,i))
!             write(*,*) 
! 
!             ! Who is responsible for this i?
!             who = 1 + mod(i,num_images())
! 
!             slavex(:)[who] = X(:,i) ! point to be executed
!             flag[who] = 1 ! execute
!             sync images(who) ! release x to worker who will then begin executing
! 
!             if (who==1) then ! 
!                 slavey(1) = f(slavex) ! release x to master
!                 sync images(*) ! collect y from everyone
!                 y(i-num_images()+1:i-1) = slavey(2:num_images()) ! the slaves have stored the value in slavey(them)[1]
!                 y(i) = slavey(1)
!             end if
! 
!         end do
!         fevals = N
! 
!         
!         do iter = 1,10000
! 
!             ! Sort the points (ascending)
!             call sort2(tmpy,tmpI,y) 
!             X = X(:,tmpI) ! sort X
!             y = tmpy ! sort y
! 
! !             write(*,'(2(A,i8),2(A,e14.6),A)',advance='no') 'PARCRS2LM: iter',iter,' fevals ',fevals,'  best ',y(1),'  worst ',y(N),'  xbest = '
!             call printx(X(:,1))
!             write(*,*) 
!  
!             ! In case exit either through maxiter being reached or satisfying exit criteria, save xstar
!             xstar = X(:,1)
! 
! 
!             ! Check stopping criteria
!             if (abs(y(1)-y(N))<1d-5 .or. iter==10000) then
!                 if (abs(y(1)-y(N))<1d-5) then
!                     write(*,'(A)') 'PARCRS2LM: convergence criteria satisfied'
!                 else
!                     write(*,'(A)') 'PARCRS2LM: max iterations reached'
!                 end if
! 
!                 do i = 1,num_images()
! 
!                     ! Put the best
!                     slavex(:)[i] = xstar
! 
!                     ! Send stop message to workers
!                     flag[i] = -1
! 
!                 end do
! 
!                 sync images(*)
! 
!                 return
! 
!             end if
! 
!             !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!             ! Debugging: assert X(:,i) is in [a,b] for all i
!             do i = 1,N
!                 if (any(X(:,i)<a) .or. any(X(:,i)>b)) then
!                     write(0,*) 'ERROR: X(:,i) is not in [a,b] for all i'
!                     write(0,*) 'Debug info:'
!                     call disp('X',X(:,:),0)
!                     call disp('a',a,0)
!                     call disp('b',b,0)
!                     error stop  'ERROR: X(:,i) is not in [a,b] for all i'
!                 end if
!             end do
!             !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!             
! 
! 
!             ! Generate random simplices and trial points
!             do i = 1,M
! 
!                 ! Potentially infinite loop
!                 do inner = 1,1000000
! 
!                     ! Get some random simplex draws using the best point x(:,1) 
!                     ! along with N more randomly chosen points
!                     smx(:,1) = x(:,1)
!                     do ism = 2,nx+1
! 
!                         ! Pick an integer between 1 to N-M and use that x as a simplex point
!                         smx(:,ism) = x(:,randi(1,N-M))
! 
!                     end do
! 
!                     ! Compute the centroid
!                     g = sum(smx,2)/(nx+1)
! 
!                     ! Reflect the ith worst point
!                     R(:,i,1) = 2d0*g - X(:,N-i+1)
! 
!                     ! Check that R(:,i) is in [a,b], if so, exit the loop. O/w, go to the top and form a new simplex.
!                     if (all(R(:,i,1)>=a) .and. all(R(:,i,1)<=b)) exit 
! 
!                     ! Note: the above reflectio nmay fail even with a large number of iterations if the point is close to the boundary. 
!                     ! May need to adjust this to bound the reflection point at a,b (elementwise) to prevent failures.
!                     if (inner>1000) then
!                         ! If here after one thousand tries, likely at a border case. So just bound element wise and proceed
!                         R(:,i,1) = max(R(:,i,1),a)
!                         R(:,i,1) = min(R(:,i,1),b)
!                         exit
!                     end if
!                 
!                 end do
! 
!                 ! Generate another trial point based on a random combination between best point and the reflected point 
!                 do ix = 1,nx
!                     call random_number(tmp)
!                     R(ix,i,2) = tmp*x(ix,1) + (1d0-tmp)*R(ix,i,1) ! convex combination, so in [a,b]
!                 end do
! 
!             end do
! 
!             ! Now evaluate the functions
!             do j = 1,2
!                 do i = 1,M
! 
!                     
!                     who = 1 + mod(i,num_images()) ! Who is responsible for this i?
! 
!                     slavex(:)[who] = R(:,i,j) ! point to be executed
!                     flag[who] = 1 ! execute
!                     sync images(who) ! release x to worker who will then begin executing
! 
!                     if (who==1) then
!                         slavey(1) = f(slavex) ! release x to master
!                         sync images(*) ! collect y from everyone
!                         yR(i-num_images()+1:i-1,j) = slavey(2:num_images()) ! the slaves have stored the value in slavey(them)[1]
!                         yR(i,j) = slavey(1)
!                     end if
! 
!                 end do
!             end do
!             fevals = fevals + 2*M
! 
!             ! Check whether the trial point is accepted or not
!             do i = 1,M
! 
!                 ! Is the reflected point better than the ith worst point? If so, 
!                 ! then accept it unless the 2nd point is better than the reflected point and 
!                 ! better than the best point.
!                 ! Although this discard potentially useful info from the second point, prevents
!                 ! shrinking too quickly. Most of the time, the reflected point is not better anyway.
!                 if (opt==0) then
!                     ! DEFAULT
!                     if (yR(i,2)<y(1) .and. yR(i,2)<yR(i,1)) then
! 
!                         ! Accept the second point
!                         y(N-i+1) = yR(i,2)
!                         x(:,N-i+1) = R(:,i,2)
! 
!                     ! The default except in exceptional cases
!                     elseif (yR(i,1)<y(N-i+1)) then
! 
!                         ! Accept it
!                         y(N-i+1) = yR(i,1)
!                         x(:,N-i+1) = R(:,i,1)
!                     
!                     ! Otherwise, is the combination of the reflected and best better? If so, then accept it.
!                     elseif (yR(i,2)<y(N-i+1)) then
! 
!                         ! Accept it
!                         y(N-i+1) = yR(i,2)
!                         x(:,N-i+1) = R(:,i,2)
! 
!                     ! O/w, do nothing.
!                     end if
!                 else
!                     ! MORE LOCAL SEARCH
!                     if (min(yR(i,1),yR(i,2))<y(N-i+1)) then
!                         ! Replace with best of yR(i,1) and yR(i,2). Otherwise, do nothing.
!                         if (yR(i,1)<yR(i,2)) then
!                             y(N-i+1) = yR(i,1)
!                             x(:,N-i+1) = R(:,i,1)
!                         else
!                             y(N-i+1) = yR(i,2)
!                             x(:,N-i+1) = R(:,i,2)
!                         end if
!                     end if
!                 end if
! 
!             end do
! 
!         end do
! 
!     contains
! 
!         ! Do something according to the value of flag(this_image())[1]
!         !  1 := evaluate slavex and store in slavey
!         ! -1 := exit the slave routine
!         subroutine slave()
!             implicit none
!             ! local
!             integer :: me
! 
!             me = this_image()
! 
!             do 
! 
!                 ! Logic here is wait at sync images(1) for data to be placed from master.
!                 ! When it is, check the flag that was sent to me. The flag can only say
!                 ! two things, either evaluate f or stop.
!                 ! In the first case, evaluate f at slavex and store in slavey. Then call 
!                 ! sync images(1) again to tell the master you are finished. 
!                 ! In the second case, you want to terminate. So, slavex contains the best
!                 ! point, just return.
! 
!                 ! Create a barrier, wait for sync images(me) to be called
!                 sync images(1)
! 
!                 ! Now sync images(me) has been called, check my data
!                 if (flag==1) then
!                     slavey(me)[1] = f(slavex)
!                     sync images(1)
!                 end if
! 
!                 if (flag==-1) then
!                     xstar = slavex
!                     exit
!                 end if
! 
!             end do
! 
!         end subroutine
! 
! 
!         subroutine printx(x)
!             implicit none
!             real(8), intent(in) :: x(:)
!             write(*,'(A)',advance='no') '['
!             do ix = 1,nx-1
!                 write(*,'(d16.8,A)',advance='no') x(ix),', '
!             end do
!             write(*,'(d16.8)',advance='no') x(nx)
!             write(*,'(A)',advance='no') ']'
!         end subroutine
! 
!         ! Generate a random integer in a,...,b
!         function randi(a,b) result(i)
!             implicit none
!             integer, intent(in) :: a,b
!             integer :: i
!             real(8) :: tmp
! 
!             ! Generate a U[a,b+1-epsilon]
!             ! Then floor to get a in 
!             call random_number(tmp)
!             tmp = tmp*(b+1-a) + a ! U[a,b+1]
!             tmp = min(tmp,b+.99999999d0)
!             i = floor(tmp)
! 
!             if (floor(tmp)<a .or. floor(tmp)>b) STOP 'ERROR: randi: failed to generate i in a,...,b'
! 
!         end function
!     end subroutine
! 
! 
! 
!     ! Controlled random search with local mutation 
!     ! CRSLM attempts to solve min f(x) s.t. x in [a,b] (where all are in Rn)
!     ! It generates N random uniform points.
!     ! It takes the worst point and tries to get rid of it by reflecting it
!     ! (like in Nelder-Mead) through a simplex of points. This simplex is chosen 
!     ! by taking the best point found so far, along with n randomly points (whose 
!     ! function values are known) with replacement, giving a total of n+1 points. 
!     ! The centroid of the simplex is then used for reflecting the worst point. 
!     ! If there reflected point improves on the worst point (and is in a,b), it is
!     ! accepted. If not, then this reflected point, say xR, by taking a random, 
!     ! convex combination of xR and the reflected point, say yR. If yR fails to 
!     ! on the worst point, then a new simplex is chosen randomly and the process 
!     ! repeats. If yR does improve, then it is accepted.
!     ! 
!     ! This is modified for parallel purposes by attempting to simultaneously get 
!     ! rid of the worst M points. For each of the M points, it attempts the 
!     ! procedure above. Currently though, it doesn't do the yR, only xR.
!     ! 
!     subroutine sub_parCRS2LM_v1(xstar,f,a,b,uid)
!         use mod_mympi
!         use mod_matlab
!         implicit none
!         real(8), intent(inout) :: xstar(:)
!         real(8), intent(in) :: a(:),b(:)
!         integer, intent(in) :: uid
!         interface 
!             function f(x) 
!                 implicit none
!                 real(8), dimension(:), intent(in) :: x
!                 real(8) :: f
!             end function
!         end interface
!         ! local
!         real(8), allocatable :: X(:,:)[:],R(:,:,:)[:],g(:),smx(:,:)
!         real(8), allocatable :: y(:)[:],yR(:,:)[:],tmpy(:)
!         integer, allocatable :: tmpI(:)
!         real(8) :: tmp
!         integer :: nx, N, M, ix, iter, i, ism, fevals, inner
!         logical, allocatable :: errfindingR[:]
! 
!         nx = size(xstar)
!         if (size(a)/=nx) error stop 'ERROR: sub_parCRS2LM: size of a and x do not agree'
!         if (size(b)/=nx) error stop 'ERROR: sub_parCRS2LM: size of b and x do not agree'
!         if (any(a>=b)) error stop 'ERROR: sub_parCRS2LM: a>=b at least once'
! 
! 
!         ! Check that input x is in a,b for all i
!         if (any(xstar<a) .or. any(xstar>b)) error stop 'ERROR: sub_parCRS2LM: x is not in a,b for all dimensions'
! 
!         ! M is the number to simultaneously try to get rid of
!         M = num_images()
! 
!         ! N is number of trial points (the pool of points)
!         N = max(5*M, ((100+nx)/num_images()+1)*num_images(), nx+2+M)
! 
!         if (N-M<=nx+1) then
!             error stop 'ERROR: sub_parCRS2LM: ooo few N points set relative to M'
!         end if
! 
! 
!         ! Generate N random points uniformly from [a,b]
!         allocate(X(nx,N)[*])
!         allocate(y(N)[*])
!         allocate(R(nx,M,2)[*])
!         allocate(yR(M,2)[*])
!         allocate(g(nx))
!         allocate(smx(nx,nx+1))                    
!         allocate(errfindingR[*])
! 
!         errfindingR = .false.
!         inner = 1
! 
!         ! For random results, initialize based on the clock
!         call init_random_seed()
! 
!         ! Generate initial pool of points
!         X(:,1) = xstar
!         do i = 2,N
!             do ix = 1,nx
!                 call random_number(tmp)
!                 X(ix,i) = tmp*a(ix) + (1d0-tmp)*b(ix)
!             end do
!         end do
! 
!         sync all
! 
!         ! Everyone get their point from the master
!         X(:,:) = X(:,:)[1]
! 
!         sync all
!         ! Assert that everyone has the same image
!         if (any(X(:,:)/=X(:,:)[1])) then
!             ! To try to get to "critical" section (gfortran has a bug)
!             call sleep(this_image()*2)
!             write(0,*) 'ERROR: X and y do not agree across images'
!             write(0,*) 'Debug info:'
!             write(0,*) 'this_image()',this_image(),'num_images()',num_images()
!             call disp('X from this image',X(:,:),0)
!             call disp('X from root      ',X(:,:)[1],0)
!         end if
!         sync all 
! 
! 
!         ! Evaluate at each of the N points in parallel
!         if (this_image()==1) write(*,'(A,i4,A)') 'PARCRS2LM: Generating initial pool of ',N,' points'
!         do i = 1,N
!             if (this_image()==1+mod(i,num_images())) then
! 
!                 write(uid,'(A,i4,A,i4,A)',advance='no') 'PARCRS2LM: Solving for point ',i,' of ',N,' x = '
!                 call printx(X(:,i))
!                 write(uid,*) 
! 
!                 ! Then you are responsible
!                 y(i) = f(X(:,i))
! 
!             end if
!         end do
!         fevals = N
! 
!         ! Collect the values
!         sync all
!         do i = 1,N
!             y(i) = y(i)[1+mod(i,num_images())]
!         end do
!         
!         ! Need another sync because y is modified below by the sorting routine. So ensure transfer is finished. 
!         sync all 
! 
!         do iter = 1,10000
! 
!             sync all
! 
!             ! Sort the points (ascending)
!             allocate(tmpy(N),tmpI(N))
!             call sort2(tmpy,tmpI,y) 
!             X = X(:,tmpI) ! sort X
!             y = tmpy ! sort y
!             deallocate(tmpy,tmpI)
! 
!             if (this_image()==1) then
!                 write(*,'(2(A,i8),2(A,e14.6),A)',advance='no') 'PARCRS2LM: iter',iter,' fevals ',fevals,'  best ',y(1),'  worst ',y(N),'  xbest = '
!                 call printx(X(:,1))
!                 write(*,*) 
!             end if
!  
!             ! In case exit either through maxiter being reached or satisfying exit criteria, save xstar
!             xstar = X(:,1)
! 
!             !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!             ! Debugging: check that X(:,:) and y(:) are the same across all images
!             !!! Stop, make sure everyone has same view of data
!             sync all
!             if (any(X(:,:)/=X(:,:)[1]) .or. any(y(:)/=y(:)[1])) then
!                 ! To try to get to "critical" section (gfortran has a bug)
!                 call sleep(this_image()*2)
!                 write(0,*) 'ERROR: X and y do not agree across images'
!                 write(0,*) 'Debug info:'
!                 write(0,*) 'this_image()',this_image(),'num_images()',num_images()
!                 call disp('X from this image',X(:,:),0)
!                 call disp('X from root      ',X(:,:)[1],0)
!                 call disp('y from this image',y(:),0)
!                 call disp('y from root      ',y(:)[1],0)
!                 error stop 'PARCRS2LM: ERROR: X and y do not agree across images'
!             end if
!             !!! Stop again, make sure that no data is modified before there is a chance for everyone to finish.
!             sync all 
!             !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! 
!             ! Check stopping criteria
!             if (abs(y(1)-y(N))<1d-5) then
!                 if (this_image()==1) then
!                     write(*,'(A)') 'PARCRS2LM: convergence criteria satisfied'
!                 end if
!                 sync all
!                 return
!             end if
! 
!             !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!             ! Debugging: assert X(:,i) is in [a,b] for all i
!             do i = 1,N
!                 if (any(X(:,i)<a) .or. any(X(:,i)>b)) then
!                     ! To try to get to "critical" section (gfortran has a bug)
!                     call sleep(this_image()*2)
!                     write(0,*) 'ERROR: X(:,i) is not in [a,b] for all i'
!                     write(0,*) 'Debug info:'
!                     write(0,*) 'this_image()',this_image(),'num_images()',num_images()
!                     call disp('X from this image',X(:,:),0)
!                     call disp('X(:,i) from this image',X(:,i),0)
!                     call disp('a from this image',a,0)
!                     call disp('b from this image',b,0)
!                     error stop  'ERROR: X(:,i) is not in [a,b] for all i'
!                 end if
!             end do
!             !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!             
! 
! 
!             if (this_image()==1) then
! 
!                 ! Generate random simplices and trial points
!                 do i = 1,M
! 
!                     ! Potentially infinite loop
!                     do inner = 1,1000000
! 
!                         ! Get some random simplex draws using the best point x(:,1) 
!                         ! along with N more randomly chosen points
!                         smx(:,1) = x(:,1)
!                         do ism = 2,nx+1
! 
!                             ! Pick an integer between 1 to N-M and use that x as a simplex point
!                             smx(:,ism) = x(:,randi(1,N-M))
! 
!                         end do
! 
!                         ! Compute the centroid
!                         g = sum(smx,2)/(nx+1)
! !     print*,'g',g
! 
!                         ! Reflect the ith worst point
!                         R(:,i,1) = 2d0*g - X(:,N-i+1)
! 
!                         ! Check that R(:,i) is in [a,b], if so, exit the loop. O/w, go to the top and form a new simplex.
!                         if (all(R(:,i,1)>=a) .and. all(R(:,i,1)<=b)) exit 
! 
!                         ! Note: the above reflectio nmay fail even with a large number of iterations if the point is close to the boundary. 
!                         ! May need to adjust this to bound the reflection point at a,b (elementwise) to prevent failures.
!                         if (inner>1000) then
!                             ! If here after one thousand tries, likely at a border case. So just bound element wise and proceed
!                             R(:,i,1) = max(R(:,i,1),a)
!                             R(:,i,1) = min(R(:,i,1),b)
!                             exit
!                         end if
!                     
!                     end do
! 
!                     ! Generate another trial point based on a random combination between best point and the reflected point 
!                     do ix = 1,nx
!                         call random_number(tmp)
!                         R(ix,i,2) = tmp*x(ix,1) + (1d0-tmp)*R(ix,i,1) ! convex combination, so in [a,b]
!                     end do
! 
!                 end do
! 
!                 if (.not. errfindingR) errfindingR = inner>=1000000
! 
!             end if
! 
!             sync all
!             if (errfindingR[1]) then
!                 ! To try to get to "critical" section (gfortran has a bug)
!                 call sleep(this_image()*2)
!                 write(0,*) 'PARCRS2LM: ERROR: failed to generate reflection point in [a,b]. Returning best x found so far'
!                 write(0,*) 'Debug info:'
!                 write(0,*) 'this_image()',this_image(),'num_images()',num_images()
!                 call disp('X from this image',X(:,:),0)
!                 call disp('X from root      ',X(:,:)[1],0)
!                 call disp('a from this image',a(:),0)
!                 call disp('b from this image',b(:),0)
!                 error stop 'PARCRS2LM: ERROR: failed to generate reflection point in [a,b]. Returning best x found so far'
!             end if
!             sync all
! 
!             ! Distribute the reflection points to each node
!             if (this_image()==1) then
!                 do i = 1,num_images()
!                     R(:,:,:)[i] = R(:,:,:)
!                 end do
!             end if
! 
!             sync all
! 
!             ! Now every image has the same view of R
! 
!             ! Now evaluate the functions
!             do i = 1,M
!                 if (this_image()==1+mod(i,num_images())) then
!                     yR(i,1) = f(R(:,i,1))
!                     yR(i,2) = f(R(:,i,2))
! !                     print*,'this_image()',this_image(),'yR(i,1)',yR(i,1),'yR(i,2)',yR(i,2)
!                 end if
!             end do
!             fevals = fevals + 2*M
! 
!             sync all
! 
!             ! Collect the function values
!             do i = 1,M
!                 yR(i,:) = yR(i,:)[1+mod(i,num_images())]
!             end do
! 
!             sync all
! 
!             ! Now every image has the same yR points
! 
!             ! Check whether the trial point is accepted or not
!             do i = 1,M
! 
!                 ! Is the reflected point better than the ith worst point? If so, then accept it.
!                 ! Although this discard potentially useful info from the second point, prevents
!                 ! shrinking too quickly. Most of the time, the reflected point is not better anyway.
!                 if (yR(i,1)<y(N-i+1)) then
! 
!                     ! Accept it
!                     y(N-i+1) = yR(i,1)
!                     x(:,N-i+1) = R(:,i,1)
!                 
!                 ! Otherwise, is the combination of the reflected and best better? If so, then accept it.
!                 elseif (yR(i,2)<y(N-i+1)) then
! 
!                     ! Accept it
!                     y(N-i+1) = yR(i,2)
!                     x(:,N-i+1) = R(:,i,2)
! 
!                 ! O/w, do nothing.
!                 end if
! 
!             end do
! 
!         end do
! 
!     contains
! 
!         subroutine printx(x)
!             implicit none
!             real(8), intent(in) :: x(:)
!             write(uid,'(A)',advance='no') '['
!             do ix = 1,nx-1
!                 write(uid,'(f16.6,A)',advance='no') x(ix),','
!             end do
!             write(uid,'(f16.6)',advance='no') x(nx)
!             write(uid,'(A)',advance='no') ']'
!         end subroutine
! 
!         ! Generate a random integer in a,...,b
!         function randi(a,b) result(i)
!             implicit none
!             integer, intent(in) :: a,b
!             integer :: i
!             real(8) :: tmp
! 
!             ! Generate a U[a,b+1-epsilon]
!             ! Then floor to get a in 
!             call random_number(tmp)
!             tmp = tmp*(b+1-a) + a ! U[a,b+1]
!             tmp = min(tmp,b+.99999999d0)
!             i = floor(tmp)
! 
!             if (floor(tmp)<a .or. floor(tmp)>b) STOP 'ERROR: randi: failed to generate i in a,...,b'
! 
!         end function
!     end subroutine
! 
!     ! Unique global min for size(x)==3 at (1,1,1) 
!     ! Two minima for size(x)>=4 & size(x)<=7, global at ones and local at -1,1,...,1 
    function rosenbrock(x) result(yout)
        implicit none
        real(8), intent(in) :: x(:)
        real(8) :: yout
        real(8) :: y
        integer :: i

        y = 0d0
        do i = 1,size(x)-1
            y = y + (1d0-x(i))**2 + 100d0*(x(i+1)-x(i)**2)**2
        end do
        yout = y

        
!         call random_number(y)
!         call sleep(merge(1,0,y>.99d0))
!         call random_number(y)
!         y = y * 5d0
!         print*,'waiting ',floor(y)
!         call sleep(floor(y))
!         do i = 1,100000
!             call random_number(y) ! burn some time
!         end do
    end function 
! 
!     subroutine test_sub_parCRS2LM()
!         implicit none
!         integer, parameter :: n = 3
!         real(8) :: x(n),a(n),b(n)
! 
!         a = -2d0
!         b = 3d0
!         x = 0d0
! 
!         ! Global min at (1,1,1)
!         call sub_parCRS2LM(x,rosenbrock,a,b)
! 
!         print*,'x is ',x
!         print*,'global min is 1,...,1'
! 
!         a = 1.04d0
!         b = 3d0
!         x = 2.2d0
! 
!         ! Global min at (1,1,1)
!         call sub_parCRS2LM(x,rosenbrock,a,b,option=1)
! 
!     end subroutine

    subroutine test_sub_parCRS()
        implicit none
        integer, parameter :: n = 3
        real(8) :: x(n),a(n),b(n)

        a = -2d0
        b = 3d0
        x = 0d0

        ! Global min at (1,1,1)
        call sub_parCRS(x,rosenbrock,a,b)

        print*,'x is ',x
        print*,'global min is 1,...,1'

        a = 1.04d0
        b = 3d0
        x = 2.2d0

        ! Global min at (1,1,1)
        call sub_parCRS(x,rosenbrock,a,b,sleeptime=0)

    end subroutine


end module
