! A module for plotting from Fortran using gnuplot

! subroutine clear()
! subroutine scratchdir(s)
! subroutine checkinit()
! subroutine legend(s1,s2,s3,s4,s5,s6,s7,s8,s9,s10,s11,s12)
! subroutine axis(x)
! subroutine xlabel(s)
! subroutine ylabel(s)
! subroutine title(s)

! subroutine triplet(x,y,s)
! * Inputs
! --- x,y are the data
! --- s is a string giving the style. One of "lines", "linespoints", "points", "dots"

! subroutine saveplot(fname)
! subroutine drawnow()
! subroutine test_plot()
module mod_plot
    use mod_types
    implicit none

    logical :: disable_plotting = .false.

    character(len=1000) :: dir = '/tmp/'

    ! Defaults
    character(len=7) :: default_colors(1000) 
    integer :: n_default_colors = 1

    ! Triplet data type
    type xys
        real(rt), allocatable :: x(:), y(:)
        integer :: nx,ncol
        character(len=4) :: s ! style
    end type

    ! Plot type
    type type_plotopts
        type(xys) :: p(10) ! stores the plot data
        integer :: npairs = 0
        logical :: haslegend = .false.
        character(len=1000), dimension(12) :: legendlist ! hardcoded max legend limit is 10
        integer :: nlegendlist = 0 
        character(len=1000) :: legendlocation = ''
        integer :: persist 
        real(rt) :: xub,xlb,yub,ylb
        logical :: hasxub=.false.,hasxlb=.false.,hasyub=.false.,hasylb=.false.
        logical :: dosaveplot = .false.
        character(len=1000) :: savefname= ''
    end type

    type(type_plotopts), dimension(100), target :: plot_list ! list of all plots
    type(type_plotopts), pointer :: cf => null() ! current figure
    
    logical :: isinit = .false.

    interface triplet
        module procedure triplet_vec, triplet_mat
    end interface

    private 
    public :: xlabel,ylabel,title,legend,triplet,drawnow,axis,scratchdir,clear,test_plot,initplot,add_gnuplot_cmd,disable_plotting,gnuplot

    interface title_renamed
        module procedure title
    end interface 

    interface gnuplot
        module procedure gnuplot_mat, gnuplot_vec
    end interface

contains

    subroutine gnuplot_vec(x,y,title,saveas)
        implicit none
        real(8), intent(in) :: x(:)
        real(8), intent(in) :: y(:)
        character(*), intent(in), optional :: title,saveas

        write(0,*) 'WARNING: passing optional arguments, this may segfault'
        call gnuplot_mat(x,reshape(y,[size(y),1]),title,saveas)

    end subroutine 

    subroutine gnuplot_mat(x,y,title,saveas)
        implicit none
        real(8), intent(in) :: x(:)
        real(8), intent(in) :: y(:,:)
        character(*), intent(in), optional :: title,saveas
        ! local
        integer :: uid

        call clear()

        if (present(saveas)) then
            uid = initplot(saveas)
        else
            uid = initplot('/tmp/gnuplot_core.eps')
        end if

        if (present(title)) then
            call title_renamed(uid,title)
        end if

        call triplet(x,y,' ')

        call drawnow(uid)
        

    end subroutine 

    ! Creates "handle" (really just a unit number) for a particular plot
    ! * Inputs
    ! --- fname : file name for eps file
    ! 
    function initplot(fname) result(uid)
        implicit none
        character(len=*), intent(in) :: fname
        integer :: uid
        ! local
        integer :: uidtmp

        call checkinit()

        open(newunit=uidtmp,file=trim(dir)//"mod_plot_tmp.cfg",status='replace',err=999)
        uid = uidtmp

        ! Write the default style file 

        write(uid,'(3A)') 'set output "',trim(fname),'"'
        write(uid,'(A)') 'set terminal postscript eps colour enhanced "Helvetica" 14' 
        write(uid,'(A)') 'set linetype 1 lc rgb "#4169E1" linewidth 4 ' ! royal blue
        write(uid,'(A)') 'set linetype 2 lc rgb "#FF7F50" linewidth 4 dashtype "-." # coral' ! coral
        write(uid,'(A)') 'set linetype 3 lc rgb "#3CB371" linewidth 4 dashtype "._. ._."' ! medium sea-green
        write(uid,'(A)') 'set linetype 4 lc rgb "#DC143C" linewidth 4 ' ! crimson
        write(uid,'(A)') 'set linetype cycle 4'
        write(uid,'(A)') 'set border linewidth 1'
        write(uid,'(A)') 'set key off' ! by default, no legend, but if legend is called, turned on
        write(uid,'(A)') 'set style data linespoints' ! set line as default style 

        return

999     write(0,*) 'WARNING: could not open ',trim(dir)//'mod_plot_tmp.cfg, plot will not be created'
        uid = 0

    end function 

    subroutine add_gnuplot_cmd(uid,s)
        implicit none
        integer, intent(in) :: uid
        character(len=*), intent(in) :: s
        write(uid,'(A)') s
    end subroutine

    ! legend
    ! Creates a legend with names and location specified 
    !
    ! * Inputs
    ! --- names : comma separated list of data names
    ! --- <location> : gnuplot style location, e.g. "inside bottom right"
    subroutine legend(uid,names,location)
        use mod_util
        implicit none
        integer, intent(in) :: uid
        character(len=*), intent(in) :: names
        character(len=*), intent(in), optional :: location
        ! local
        integer :: n, i

        if (present(location)) then
            write(uid,'(A)') 'set key '//location
        else
            write(uid,'(A)') 'set key on'
        end if

        cf%haslegend = .true.

        cf%nlegendlist = count_delim_arg(names,',')

        do i = 1,cf%nlegendlist 
            cf%legendlist(i) = get_delim_arg(names,',',i)
        end do

    end subroutine

    subroutine xlabel(uid,s)
        implicit none
        integer, intent(in) :: uid
        character(len=*), intent(in) :: s
        write(uid,'(A)') 'set xlabel "'//s//'"'
    end subroutine
    subroutine ylabel(uid,s)
        implicit none
        integer, intent(in) :: uid
        character(len=*), intent(in) :: s
        write(uid,'(A)') 'set ylabel "'//s//'"'
    end subroutine
    subroutine title(uid,s)
        implicit none
        integer, intent(in) :: uid
        character(len=*), intent(in) :: s
        write(uid,'(A)') 'set title "'//s//'"'
    end subroutine
    
    ! frees the specified figure our, by default, current figure
    subroutine clear()
        implicit none
        ! local
        integer :: i
        type(type_plotopts) :: defaulttype

        call checkinit()
        
        ! I don't think it is necessarily, but to be careful, manually free x and y data 
        do i = 1,cf%npairs
            if (allocated(cf%p(i)%x)) deallocate(cf%p(i)%x)
            if (allocated(cf%p(i)%y)) deallocate(cf%p(i)%y)
!             cf%p(i)%nx = -1
!             cf%p(i)%ncol = -1
!             cf%p(i)%s = ''

        end do
!         cf%npairs = 0

        ! Use this to reset to whatever the default specification is
        cf = defaulttype

    end subroutine

    subroutine write_default_style(uid)
        implicit none
        integer, intent(in) :: uid

        write(uid,'(A)') '# Temporary gnuplot configuration file created by grey''s plot routine'
        write(uid,'(A)') 'set terminal postscript eps colour enhanced "Helvetica" 16' 
        write(uid,'(A)') 'set linetype 1 lc rgb "#4169E1" linewidth 4 ' ! royal blue
        write(uid,'(A)') 'set linetype 2 lc rgb "#FF7F50" linewidth 4 dashtype "-." # coral' ! coral
        write(uid,'(A)') 'set linetype 3 lc rgb "#DC143C" linewidth 4 ' ! crimson
        write(uid,'(A)') 'set linetype 4 lc rgb "#3CB371" linewidth 4 dashtype "._. ._."' ! medium sea-green
        write(uid,'(A)') 'set linetype cycle 4'
        write(uid,'(A)') 'set border linewidth 1'

    end subroutine

    
    subroutine scratchdir(s)
        implicit none
        character(len=*), intent(in) :: s
        dir = s
    end subroutine 

    subroutine checkinit()
        implicit none
        if (.not. isinit) then
            cf => plot_list(1)
            isinit = .true.
        end if
    end subroutine

!     subroutine legend(s1,s2,s3,s4,s5,s6,s7,s8,s9,s10,s11,s12)
!         implicit none
!         character(len=*), intent(in), optional :: s1,s2,s3,s4,s5,s6,s7,s8,s9,s10,s11,s12
!         ! local
!         integer :: narg
! 
!         cf%haslegend = .true.
! 
!         if (present(s1))  then; cf%legendlist(1) = s1; narg = 1; end if
!         if (present(s2))  then; cf%legendlist(2) = s2; narg = 2; end if
!         if (present(s3))  then; cf%legendlist(3) = s3; narg = 3; end if
!         if (present(s4))  then; cf%legendlist(4) = s4; narg = 4; end if
!         if (present(s5))  then; cf%legendlist(5) = s5; narg = 5; end if
!         if (present(s6))  then; cf%legendlist(6) = s6; narg = 6; end if
!         if (present(s7))  then; cf%legendlist(7) = s7; narg = 7; end if
!         if (present(s8))  then; cf%legendlist(8) = s8; narg = 8; end if
!         if (present(s9))  then; cf%legendlist(9) = s9; narg = 9; end if
!         if (present(s10)) then; cf%legendlist(10) = s10; narg = 10; end if
!         if (present(s11)) then; cf%legendlist(11) = s11; narg = 11; end if
!         if (present(s12)) then; cf%legendlist(12) = s12; narg = 12; end if
! 
!         cf%nlegendlist = narg
! 
!         if (trim(cf%legendlist(narg-1))=='location') then
!             cf%legendlocation = trim(cf%legendlist(narg))
!             cf%nlegendlist = narg - 2
!         end if
!         
!     end subroutine

    subroutine axis(x)
        implicit none
        real(rt), intent(in) :: x(4)
        call checkinit()
        
        if (x(1)/=-huge(0e0_rt)) then; cf%xlb = x(1); cf%hasxlb = .true.; end if
        if (x(2)/=huge(0e0_rt)) then; cf%xub = x(2); cf%hasxub = .true.; end if
        if (x(3)/=-huge(0e0_rt)) then; cf%ylb = x(3); cf%hasylb = .true.; end if
        if (x(4)/=huge(0e0_rt)) then; cf%yub = x(4); cf%hasyub = .true.; end if

    end subroutine
    subroutine triplet_vec(x,y,s)
        implicit none
        real(rt), intent(in) :: x(:),y(:)
        character(len=*), intent(in) :: s
        call triplet_core(x,y,s,1)
    end subroutine
    subroutine triplet_mat(x,y,s)
        implicit none
        real(rt), intent(in) :: x(:),y(:,:)
        character(len=*), intent(in) :: s
        call triplet_core(x,y,s,size(y,2))
    end subroutine
    subroutine triplet_core(x,y,s,ncol)
        implicit none
        real(rt), intent(in) :: x(:),y(*)
        integer, intent(in) :: ncol
        character(len=*), intent(in) :: s

        call checkinit()

        cf%npairs = cf%npairs + 1

        associate (p => cf%p(cf%npairs)) 

            p%s = s

            p%nx = size(x)
            allocate(p%x(p%nx))
            p%x = x

            p%ncol = ncol

            allocate(p%y(p%nx*p%ncol))
            p%y(1:p%nx*p%ncol) = y(1:p%nx*p%ncol)

        end associate

    end subroutine

!     subroutine saveplot(fname)
!         implicit none
!         character(len=*), intent(in) :: fname
! 
!         cf%dosaveplot = .true.
!         cf%savefname = fname
!         call drawnow()
!         cf%dosaveplot = .false.
!         cf%savefname = ''
! 
!     end subroutine 

    subroutine drawnow(uid)
        implicit none
        integer, intent(in) :: uid
        integer :: i, unitno !,j,ni,nj
        character(len=3), parameter :: newline = ', \'
        character(len=3) :: tmpchar
        character(len=100) :: configfile, datafile 
        integer :: ierr,ierr2
        integer :: irow,nrow,icol,xaxiscol,ilegend
        character(len=*), parameter :: doublefmt = '1PE18.9E3'
        character(len=100) :: persistcmd
        ! FOR PGI, need below line
!         include 'lib3f.h' 

        if (disable_plotting) then
            close(uid)
            return
        end if

!         !$OMP SINGLE

        ! As an extra precaution, I try to make sure the threads write to unique files
        datafile = trim(dir)//'mod_plot_tmp'//'.dat'
        configfile = trim(dir)//'mod_plot_tmp'//'.cfg'

        ! Write raw data to a temporary file

        ! To handle uneven data like triplet(#2,#2) triplet(#3,#3), need to figure out the max length, (#3 in this example) and write all the rows at one time
        ! Data:
        ! 0 1.2 0 1.3
        ! 1 2   .5 1.6
        !       2 2
        ! Then, when displaying, select 1st col and 2nd col as first set of data and 3rd col and 4th col as second set of data
        nrow = -1
        do i = 1,cf%npairs
            nrow = max(nrow,cf%p(i)%nx)
        end do

        open(newunit=unitno,file=trim(datafile),status='unknown',err=999)
        do irow = 1,nrow
            do i = 1,cf%npairs
                if (irow<=cf%p(i)%nx) then
                    
                    write(unitno,'('//doublefmt//',A)',advance='no') cf%p(i)%x(irow),' '
                    do icol = 1,cf%p(i)%ncol
                        write(unitno,'('//doublefmt//',A)',advance='no') cf%p(i)%y(irow + (icol-1)*cf%p(i)%nx),' '
                    end do
                else 
                    ! write(unitno,'(A)',advance='no') repeat('? ',1+cf%p(i)%ncol)
                    write(unitno,'('//doublefmt//',A)',advance='no') cf%p(i)%x(cf%p(i)%nx),' '
                    do icol = 1,cf%p(i)%ncol
                        write(unitno,'(A19)',advance='no') '? '
                    end do
                end if
            end do
            write(unitno,*)
        end do
        close(unitno)

        ! Write the configuration file
        


        ! Write config to a temporary file
        unitno = uid
        ! open(newunit=unitno,file=trim(configfile),status='unknown')

        ! write(unitno,*) 'set term x11'
        ! write(unitno,*) 'set term epslatex'


        ! if (cf%xlabel/='') write(unitno,*) 'set xlabel "'//trim(cf%xlabel)//'"'
        ! if (cf%ylabel/='') write(unitno,*) 'set ylabel "'//trim(cf%ylabel)//'"'
        ! if (cf%title/='') write(unitno,*) 'set title "'//trim(cf%title)//'"'


        write(unitno,*) 'set datafile missing "?" ' ! set line as default style 

        ! Legend entries
        if (.not. cf%haslegend) write(unitno,*) 'set key off'
        if (cf%haslegend) then
                if (trim(cf%legendlocation)=='northwest') then; write(unitno,*) 'set key on top left'; 
            elseif (trim(cf%legendlocation)=='northeast') then; write(unitno,*) 'set key on top right'; 
            else
                ! do nothing
            end if

        end if

        ! Autoscaling is a bit tricky, so determine appropriate rnages

        ! Set xrange
        if (cf%hasylb .or. cf%hasyub .or. cf%hasxub .or. cf%hasxlb) then
            write(unitno,'(A)',advance='no') 'set xr ['
            if (cf%hasxlb) then
                write(unitno,'('//doublefmt//')',advance='no') cf%xlb
            else
                cf%xlb = huge(0e0_rt)
                do i = 1,cf%npairs 
                    cf%xlb = min(cf%xlb,minval(cf%p(i)%x))
                end do
                write(unitno,'('//doublefmt//')',advance='no') cf%xlb
            end if
            write(unitno,'(A)',advance='no') ':'
            if (cf%hasxub) then
                write(unitno,'('//doublefmt//')',advance='no') cf%xub
            else
                cf%xub = -huge(0e0_rt)
                do i = 1,cf%npairs 
                    cf%xub = max(cf%xub,maxval(cf%p(i)%x))
                end do
                write(unitno,'('//doublefmt//')',advance='no') cf%xub
            end if
            write(unitno,'(A)',advance='yes') ']'
        end if
        ! Set yrange
        if (cf%hasylb .or. cf%hasyub .or. cf%hasxub .or. cf%hasxlb) then
            write(unitno,'(A)',advance='no') 'set yr ['
            if (cf%hasylb) then
                write(unitno,'('//doublefmt//')',advance='no') cf%ylb
            else
                cf%ylb = huge(0e0_rt)
                do i = 1,cf%npairs 
                    cf%ylb = min(cf%ylb,minval(cf%p(i)%x))
                end do
                write(unitno,'('//doublefmt//')',advance='no') cf%ylb
            end if
            write(unitno,'(A)',advance='no') ':'
            if (cf%hasyub) then
                write(unitno,'('//doublefmt//')',advance='no') cf%yub
            else
                cf%yub = -huge(0e0_rt)
                do i = 1,cf%npairs 
                    cf%yub = max(cf%yub,maxval(cf%p(i)%y))
                end do
                write(unitno,'('//doublefmt//')',advance='no') cf%yub
            end if
            write(unitno,'(A)',advance='yes') ']'
        end if


    !     ! Check inputs
    !     ni = size(x)
    !     nj = size(y,2)
    !     if (size(y,1)/=ni) STOP 'sub_plot_x_y2: x and y dims don''t agree'
    ! 
    !     ! Set defaults and replace them if optional arg present
    !     do j = 1,nJ
    !         write(jAsString,'(i0)') j
    !         dls(j) = 'data '//jAsString
    !     end do
    !     ms = 'line'
    !     cls = ''
    ! 
    !     ! Write data to a temporary file
    !     open(unitno,file=trim(datafile),status='unknown')
    !     write(unitno,*) '# Temporary data file created by grey''s plot routine' 
    !     do i = 1,ni
    !         write(unitno,'(1PE24.15E3)',advance='no') x(i)
    !         do j = 1,nj
    !             write(unitno,'(1PE24.15E3)',advance='no') y(i,j)
    !         end do
    !         write(unitno,*) 
    !     end do
    !     close(unitno)

        write(unitno,'(A)',advance='no') 'plot' ! this is the first part of the command

        xaxiscol = 1
        ilegend = 1
! print*,'cf%npairs',cf%npairs
        do i = 1,cf%npairs

! print*,'cf%p(i)%ncol',cf%p(i)%ncol
            do icol = 1,cf%p(i)%ncol
                
                ! Determine whether new line is needed
                if (i<cf%npairs .or. icol<cf%p(i)%ncol) then
                    tmpchar = newline
                else
                    tmpchar = '   '
                end if

                ! Output the command. NOTE: must use advance no here to prevent ifort from breaking the line
                if (cf%haslegend .and. ilegend<=cf%nlegendlist) then
                    write(unitno,'(A)',advance='no') ' "'//trim(datafile)//'" using '//trim(i_to_s(xaxiscol))//':' &
                            //trim(i_to_s(xaxiscol+icol))//' title "'//trim(cf%legendlist(ilegend))&
                            //'"'//trim(tmpchar) 
                elseif (cf%haslegend .and. ilegend>cf%nlegendlist) then
                    write(unitno,'(A)',advance='no') ' "'//trim(datafile)//'" using '//trim(i_to_s(xaxiscol))//':' &
                            //trim(i_to_s(xaxiscol+icol))//' title "'//' '//'" '//trim(tmpchar) ! Title in latex is more like legend in matlab
                else 
                    write(unitno,'(A)',advance='no') ' "'//trim(datafile)//'" using '//trim(i_to_s(xaxiscol))//':' &
                            //trim(i_to_s(xaxiscol+icol))//trim(tmpchar) ! Title in latex is more like legend in matlab
                end if
                ! Make a new line
                write(unitno,*)

                ilegend = ilegend + 1
            end do

            xaxiscol = xaxiscol + 1 + cf%p(i)%ncol

        end do

!         if (cf%dosaveplot) then 
! !             write(unitno,*) 'set terminal postscript
! !             write(unitno,*) 'set terminal postscript eps size 3.5,2.62 enhanced color font ''Helvetica,20'' linewidth 2 '
! !             write(unitno,*) 'set terminal postscript eps enhanced color font ''Helvetica,20'' linewidth 2 '
!             write(unitno,*) 'set terminal postscript eps enhanced color font ''Helvetica,14'' linewidth 2 '
!             write(unitno,*) 'set output "'//trim(cf%savefname)//'.eps"'
!             write(unitno,*) 'replot'
! 
!             ! write(unitno,*) 'set terminal epslatex size 3.5,2.62 color colortext'
!             ! write(unitno,*) 'set output "'//trim(cf%savefname)//'.tex"'
!             ! write(unitno,*) 'replot'
!         end if

        close(unitno)

!         if (cf%dosaveplot) then
!             persistcmd = ''
!         else
!             persistcmd = '-persist'
!         end if

        ! Now use a nonstandard fortran routine (appearing in ifort and gfortran) to call gnuplot
        ! NOTE: if gnuplot is not available or an error is received, just report it and the program
        ! will continue.
        ! ierr = system('gnuplot -persist '//trim(configfile)) ! for PGI
        call execute_command_line('gnuplot -persist '//trim(configfile)//' 2> /dev/null',exitstat=ierr,cmdstat=ierr2)
!         call execute_command_line('gnuplot '//trim(persistcmd)//' '//trim(configfile)),exitstat=ierr)
!         if (ierr/=0) print*,'sub_plot_x_y2: WARNING: error in calling gnuplot'

!         !$OMP END SINGLE
        
        return

999     write(0,*) 'WARNING: could not open ',trim(datafile),', plot will not be created'

    end subroutine drawnow

    ! Converts an integer into a string
    function i_to_s(i) result(s)
        implicit none
        integer, intent(in) :: i 
        character(len=100) :: s
        ! local
        character(len=100) :: slocal

        write(slocal,'(i100)') i
        s = adjustl(slocal)

    end function 

    subroutine test_plot()
        implicit none
        integer :: i

        i = initplot('/tmp/deleteme.eps') 
        call title(i,'test plot')
        call xlabel(i,'x axis')
        call ylabel(i,'y axis')
        call legend(i,'first triplet,second triplet,third triplet','inside bottom right')

        call triplet([0e0_rt,1e0_rt],reshape([.1e0_rt,2e0_rt,.1e0_rt,1e0_rt],[2,2]),'r-')
        call triplet([0e0_rt,.5e0_rt,1e0_rt],[.1e0_rt,2.4e0_rt,2e0_rt],'b-.')
    !     call axis([.3e0_rt,1.1e0_rt,-huge(0e0_rt),huge(0e0_rt)])

        call drawnow(i)

        print*,'Saved plot to /tmp/deleteme.eps'
!         call saveplot('/tmp/savedplot')

    end subroutine

end module

! 
! program testplot
!     use mod_plot
!     implicit none
! 
! 
! end program
! 
!     ! Returns rgb value 
!     ! Hex Code #RRGGBB
!     ! Values taken from http://www.rapidtables.com/web/color/RGB_Color.htm
!     function rgb(color) 
!         implicit none
!         character(7) :: rgb
!         character(30), intent(in) :: color
!         character(30,2), dimension(139) :: colors
!         integer :: i
! 
!         i = 1
!         colors(i,1) = "maroon      ";                       colors(i,2) = "#800000"; i = i + 1
!         colors(i,1) = "dark red        ";                   colors(i,2) = "#8B0000"; i = i + 1
!         colors(i,1) = "brown       ";                       colors(i,2) = "#A52A2A"; i = i + 1
!         colors(i,1) = "firebrick       ";                   colors(i,2) = "#B22222"; i = i + 1
!         colors(i,1) = "crimson     ";                       colors(i,2) = "#DC143C"; i = i + 1
!         colors(i,1) = "red     ";                           colors(i,2) = "#FF0000"; i = i + 1
!         colors(i,1) = "tomato      ";                       colors(i,2) = "#FF6347"; i = i + 1
!         colors(i,1) = "coral   ";                           colors(i,2) = "#FF7F50"; i = i + 1
!         colors(i,1) = "indian red  ";                       colors(i,2) = "#CD5C5C"; i = i + 1
!         colors(i,1) = "light coral     ";                   colors(i,2) = "#F08080"; i = i + 1
!         colors(i,1) = "dark salmon         ";               colors(i,2) = "#E9967A"; i = i + 1
!         colors(i,1) = "salmon  ";                           colors(i,2) = "#FA8072"; i = i + 1
!         colors(i,1) = "light salmon        ";               colors(i,2) = "#FFA07A"; i = i + 1
!         colors(i,1) = "orange red      ";                   colors(i,2) = "#FF4500"; i = i + 1
!         colors(i,1) = "dark orange         ";               colors(i,2) = "#FF8C00"; i = i + 1
!         colors(i,1) = "orange  ";                           colors(i,2) = "#FFA500"; i = i + 1
!         colors(i,1) = "gold        ";                       colors(i,2) = "#FFD700"; i = i + 1
!         colors(i,1) = "dark golden rod         ";           colors(i,2) = "#B8860B"; i = i + 1
!         colors(i,1) = "golden rod  ";                       colors(i,2) = "#DAA520"; i = i + 1
!         colors(i,1) = "pale golden rod         ";           colors(i,2) = "#EEE8AA"; i = i + 1
!         colors(i,1) = "dark khaki  ";                       colors(i,2) = "#BDB76B"; i = i + 1
!         colors(i,1) = "khaki   ";                           colors(i,2) = "#F0E68C"; i = i + 1
!         colors(i,1) = "olive       ";                       colors(i,2) = "#808000"; i = i + 1
!         colors(i,1) = "yellow  ";                           colors(i,2) = "#FFFF00"; i = i + 1
!         colors(i,1) = "yellow green        ";               colors(i,2) = "#9ACD32"; i = i + 1
!         colors(i,1) = "dark olive green        ";           colors(i,2) = "#556B2F"; i = i + 1
!         colors(i,1) = "olive drab  ";                       colors(i,2) = "#6B8E23"; i = i + 1
!         colors(i,1) = "lawn green      ";                   colors(i,2) = "#7CFC00"; i = i + 1
!         colors(i,1) = "chart reuse         ";               colors(i,2) = "#7FFF00"; i = i + 1
!         colors(i,1) = "green yellow    ";                   colors(i,2) = "#ADFF2F"; i = i + 1
!         colors(i,1) = "dark green  ";                       colors(i,2) = "#006400"; i = i + 1
!         colors(i,1) = "green   ";                           colors(i,2) = "#008000"; i = i + 1
!         colors(i,1) = "forest green        ";               colors(i,2) = "#228B22"; i = i + 1
!         colors(i,1) = "lime    ";                           colors(i,2) = "#00FF00"; i = i + 1
!         colors(i,1) = "lime green  ";                       colors(i,2) = "#32CD32"; i = i + 1
!         colors(i,1) = "light green     ";                   colors(i,2) = "#90EE90"; i = i + 1
!         colors(i,1) = "pale green  ";                       colors(i,2) = "#98FB98"; i = i + 1
!         colors(i,1) = "dark sea green  ";                   colors(i,2) = "#8FBC8F"; i = i + 1
!         colors(i,1) = "medium spring green         ";       colors(i,2) = "#00FA9A"; i = i + 1
!         colors(i,1) = "spring green    ";                   colors(i,2) = "#00FF7F"; i = i + 1
!         colors(i,1) = "sea green   ";                       colors(i,2) = "#2E8B57"; i = i + 1
!         colors(i,1) = "medium aqua marine      ";           colors(i,2) = "#66CDAA"; i = i + 1
!         colors(i,1) = "medium sea green    ";               colors(i,2) = "#3CB371"; i = i + 1
!         colors(i,1) = "light sea green         ";           colors(i,2) = "#20B2AA"; i = i + 1
!         colors(i,1) = "dark slate gray     ";               colors(i,2) = "#2F4F4F"; i = i + 1
!         colors(i,1) = "teal    ";                           colors(i,2) = "#008080"; i = i + 1
!         colors(i,1) = "dark cyan   ";                       colors(i,2) = "#008B8B"; i = i + 1
!         colors(i,1) = "aqua    ";                           colors(i,2) = "#00FFFF"; i = i + 1
!         colors(i,1) = "cyan        ";                       colors(i,2) = "#00FFFF"; i = i + 1
!         colors(i,1) = "light cyan      ";                   colors(i,2) = "#E0FFFF"; i = i + 1
!         colors(i,1) = "dark turquoise      ";               colors(i,2) = "#00CED1"; i = i + 1
!         colors(i,1) = "turquoise       ";                   colors(i,2) = "#40E0D0"; i = i + 1
!         colors(i,1) = "medium turquoise    ";               colors(i,2) = "#48D1CC"; i = i + 1
!         colors(i,1) = "pale turquoise  ";                   colors(i,2) = "#AFEEEE"; i = i + 1
!         colors(i,1) = "aqua marine         ";               colors(i,2) = "#7FFFD4"; i = i + 1
!         colors(i,1) = "powder blue     ";                   colors(i,2) = "#B0E0E6"; i = i + 1
!         colors(i,1) = "cadet blue  ";                       colors(i,2) = "#5F9EA0"; i = i + 1
!         colors(i,1) = "steel blue      ";                   colors(i,2) = "#4682B4"; i = i + 1
!         colors(i,1) = "corn flower blue    ";               colors(i,2) = "#6495ED"; i = i + 1
!         colors(i,1) = "deep sky blue   ";                   colors(i,2) = "#00BFFF"; i = i + 1
!         colors(i,1) = "dodger blue         ";               colors(i,2) = "#1E90FF"; i = i + 1
!         colors(i,1) = "light blue      ";                   colors(i,2) = "#ADD8E6"; i = i + 1
!         colors(i,1) = "sky blue    ";                       colors(i,2) = "#87CEEB"; i = i + 1
!         colors(i,1) = "light sky blue  ";                   colors(i,2) = "#87CEFA"; i = i + 1
!         colors(i,1) = "midnight blue       ";               colors(i,2) = "#191970"; i = i + 1
!         colors(i,1) = "navy    ";                           colors(i,2) = "#000080"; i = i + 1
!         colors(i,1) = "dark blue   ";                       colors(i,2) = "#00008B"; i = i + 1
!         colors(i,1) = "medium blue     ";                   colors(i,2) = "#0000CD"; i = i + 1
!         colors(i,1) = "blue        ";                       colors(i,2) = "#0000FF"; i = i + 1
!         colors(i,1) = "royal blue      ";                   colors(i,2) = "#4169E1"; i = i + 1
!         colors(i,1) = "blue violet         ";               colors(i,2) = "#8A2BE2"; i = i + 1
!         colors(i,1) = "indigo  ";                           colors(i,2) = "#4B0082"; i = i + 1
!         colors(i,1) = "dark slate blue     ";               colors(i,2) = "#483D8B"; i = i + 1
!         colors(i,1) = "slate blue      ";                   colors(i,2) = "#6A5ACD"; i = i + 1
!         colors(i,1) = "medium slate blue   ";               colors(i,2) = "#7B68EE"; i = i + 1
!         colors(i,1) = "medium purple   ";                   colors(i,2) = "#9370DB"; i = i + 1
!         colors(i,1) = "dark magenta        ";               colors(i,2) = "#8B008B"; i = i + 1
!         colors(i,1) = "dark violet     ";                   colors(i,2) = "#9400D3"; i = i + 1
!         colors(i,1) = "dark orchid         ";               colors(i,2) = "#9932CC"; i = i + 1
!         colors(i,1) = "medium orchid   ";                   colors(i,2) = "#BA55D3"; i = i + 1
!         colors(i,1) = "purple      ";                       colors(i,2) = "#800080"; i = i + 1
!         colors(i,1) = "thistle         ";                   colors(i,2) = "#D8BFD8"; i = i + 1
!         colors(i,1) = "plum        ";                       colors(i,2) = "#DDA0DD"; i = i + 1
!         colors(i,1) = "violet  ";                           colors(i,2) = "#EE82EE"; i = i + 1
!         colors(i,1) = "magenta             ";               colors(i,2) = "#FF00FF"; i = i + 1
!         colors(i,1) = "orchid  ";                           colors(i,2) = "#DA70D6"; i = i + 1
!         colors(i,1) = "medium violet red   ";               colors(i,2) = "#C71585"; i = i + 1
!         colors(i,1) = "pale violet red         ";           colors(i,2) = "#DB7093"; i = i + 1
!         colors(i,1) = "deep pink   ";                       colors(i,2) = "#FF1493"; i = i + 1
!         colors(i,1) = "hot pink        ";                   colors(i,2) = "#FF69B4"; i = i + 1
!         colors(i,1) = "light pink  ";                       colors(i,2) = "#FFB6C1"; i = i + 1
!         colors(i,1) = "pink    ";                           colors(i,2) = "#FFC0CB"; i = i + 1
!         colors(i,1) = "antique white       ";               colors(i,2) = "#FAEBD7"; i = i + 1
!         colors(i,1) = "beige   ";                           colors(i,2) = "#F5F5DC"; i = i + 1
!         colors(i,1) = "bisque      ";                       colors(i,2) = "#FFE4C4"; i = i + 1
!         colors(i,1) = "blanched almond         ";           colors(i,2) = "#FFEBCD"; i = i + 1
!         colors(i,1) = "wheat       ";                       colors(i,2) = "#F5DEB3"; i = i + 1
!         colors(i,1) = "corn silk       ";                   colors(i,2) = "#FFF8DC"; i = i + 1
!         colors(i,1) = "lemon chiffon       ";               colors(i,2) = "#FFFACD"; i = i + 1
!         colors(i,1) = "light golden rod yellow         ";   colors(i,2) = "#FAFAD2"; i = i + 1
!         colors(i,1) = "light yellow        ";               colors(i,2) = "#FFFFE0"; i = i + 1
!         colors(i,1) = "saddle brown    ";                   colors(i,2) = "#8B4513"; i = i + 1
!         colors(i,1) = "sienna      ";                       colors(i,2) = "#A0522D"; i = i + 1
!         colors(i,1) = "chocolate       ";                   colors(i,2) = "#D2691E"; i = i + 1
!         colors(i,1) = "peru        ";                       colors(i,2) = "#CD853F"; i = i + 1
!         colors(i,1) = "sandy brown     ";                   colors(i,2) = "#F4A460"; i = i + 1
!         colors(i,1) = "burly wood  ";                       colors(i,2) = "#DEB887"; i = i + 1
!         colors(i,1) = "tan     ";                           colors(i,2) = "#D2B48C"; i = i + 1
!         colors(i,1) = "rosy brown  ";                       colors(i,2) = "#BC8F8F"; i = i + 1
!         colors(i,1) = "moccasin        ";                   colors(i,2) = "#FFE4B5"; i = i + 1
!         colors(i,1) = "navajo white        ";               colors(i,2) = "#FFDEAD"; i = i + 1
!         colors(i,1) = "peach puff      ";                   colors(i,2) = "#FFDAB9"; i = i + 1
!         colors(i,1) = "misty rose  ";                       colors(i,2) = "#FFE4E1"; i = i + 1
!         colors(i,1) = "lavender blush  ";                   colors(i,2) = "#FFF0F5"; i = i + 1
!         colors(i,1) = "linen       ";                       colors(i,2) = "#FAF0E6"; i = i + 1
!         colors(i,1) = "old lace        ";                   colors(i,2) = "#FDF5E6"; i = i + 1
!         colors(i,1) = "papaya whip         ";               colors(i,2) = "#FFEFD5"; i = i + 1
!         colors(i,1) = "sea shell       ";                   colors(i,2) = "#FFF5EE"; i = i + 1
!         colors(i,1) = "mint cream  ";                       colors(i,2) = "#F5FFFA"; i = i + 1
!         colors(i,1) = "slate gray      ";                   colors(i,2) = "#708090"; i = i + 1
!         colors(i,1) = "light slate gray    ";               colors(i,2) = "#778899"; i = i + 1
!         colors(i,1) = "light steel blue        ";           colors(i,2) = "#B0C4DE"; i = i + 1
!         colors(i,1) = "lavender    ";                       colors(i,2) = "#E6E6FA"; i = i + 1
!         colors(i,1) = "floral white    ";                   colors(i,2) = "#FFFAF0"; i = i + 1
!         colors(i,1) = "alice blue  ";                       colors(i,2) = "#F0F8FF"; i = i + 1
!         colors(i,1) = "ghost white     ";                   colors(i,2) = "#F8F8FF"; i = i + 1
!         colors(i,1) = "honeydew    ";                       colors(i,2) = "#F0FFF0"; i = i + 1
!         colors(i,1) = "ivory   ";                           colors(i,2) = "#FFFFF0"; i = i + 1
!         colors(i,1) = "azure       ";                       colors(i,2) = "#F0FFFF"; i = i + 1
!         colors(i,1) = "snow    ";                           colors(i,2) = "#FFFAFA"; i = i + 1
!         colors(i,1) = "black       ";                       colors(i,2) = "#000000"; i = i + 1
!         colors(i,1) = "dim gray ";                          colors(i,2) = "#696969"; i = i + 1
!         colors(i,1) = "gray                ";               colors(i,2) = "#808080"; i = i + 1
!         colors(i,1) = "dark gray               ";           colors(i,2) = "#A9A9A9"; i = i + 1
!         colors(i,1) = "silver      ";                       colors(i,2) = "#C0C0C0"; i = i + 1
!         colors(i,1) = "light gray                      ";   colors(i,2) = "#D3D3D3"; i = i + 1
!         colors(i,1) = "gainsboro   ";                       colors(i,2) = "#DCDCDC"; i = i + 1
!         colors(i,1) = "white smoke     ";                   colors(i,2) = "#F5F5F5"; i = i + 1
!         colors(i,1) = "white       ";                       colors(i,2) = "#FFFFFF"; i = i + 1
! 
!         ! Default to black
!         rgb = "#000000"
!     
!         ! Search for colors
!         do i = 1,size(colors)
!             if (colors(i,1)==color) then
!                 rgb = trim(colors(i,2))
!                 return
!             else 
!                 write(0,*) 'WARNING: rgb: color not found'
!             end if
!         end do
! 
!     end function

! A triplet specification consists of three things: 
! color (can be any name listed inside rgb)
! linestyle (one one lines, linespoints, lin

! b     blue          .     point              -     solid
! g     green         o     circle             :     dotted
! r     red           x     x-mark             -.    dashdot 
! c     cyan          +     plus               --    dashed   
! m     magenta       *     star             (none)  no line
! y     yellow        s     square
! k     black         d     diamond
! w     white         v     triangle (down)
!                     ^     triangle (up)
!                     <     triangle (left)
!                     >     triangle (right)
!                     p     pentagram
!                     h     hexagram
! 
!     
! 
!     function tripletspec(color,mrk,lne) result(cmdout)
!         implicit none
!         character(len=*), optional, intent(in) :: col
!         character(len=*), optional, intent(in) :: mrk,lne
! 
!         character(len=50) :: cmd,cmdout
!         character(len=1) :: c,marker
!         character(len=2) :: style
!         integer :: i
! 
! 
!         marker = ''
!         do i = 1,4
!             select case (t%s(i:i))
!                 case ('b','g','r','c','m','y','k','w')
!                     c = t%s(i:i)
!                 case ('-',':') 
!                     if (t%s(i+1:i+1)=='.') then
!                         style = '-.'
!                     else
!                         style(1:1) = t%s(i:i)
!                         style(2:2) = ' '
!                     end if
!                 case default 
!             end select
!         end do
! 
! 
!         cmd = ''
!         select case (style)
!             case ("- ")
!                 cmd = trim(cmd)//" with lines "
!             case (": ")
!                 cmd = trim(cmd)//" with dots "
!             case ("-.")
!                 cmd = trim(cmd)//" with linespoints "
!             case ("  ")
!                 cmd = trim(cmd)//" with points "
!             case default
!         end select
!         select case (c)
!             case ('b')
!                 cmd = trim(cmd)//' lt rgb "blue" '
!             case ('g')
!                 cmd = trim(cmd)//' lt rgb "green" '
!             case ('r')
!                 cmd = trim(cmd)//' lt rgb "red" '
!             case ('c')
!                 cmd = trim(cmd)//' lt rgb "cyan" '
!             case ('m')
!                 cmd = trim(cmd)//' lt rgb "magenta" '
!             case ('y')
!                 cmd = trim(cmd)//' lt rgb "yellow" '
!             case ('k')
!                 cmd = trim(cmd)//' lt rgb "black" '
!             case ('w')
!                 cmd = trim(cmd)//' lt rgb "white" '
!             case default
!         end select
!         select case (marker)
!             ! Currently not implemented
!             case default
!         end select
!             
! !         print*,'cmd:',cmd
!         cmdout = trim(cmd)
! 
!     end function 
