module mod_quad
    use mod_types
    implicit none
    private :: en_r2_11_1, en_r2_11_1_size
contains


! Integrates 
! \int^b_a f(x) dx 
! using Gauss-Legendre quadrature adapted from [-1,1] with weighting function 1
subroutine legendrequad(x,w,a,b)
    use mod_matlab
    implicit none
    real(rt), intent(out) :: x(:),w(:)
    real(rt), intent(in) :: a,b
    real(rt) :: tmp
    integer :: i,order,uid
    logical :: found
    
    found = .false.

    open(newunit=uid,file='/tmp/legendrequadnodes.dat',status='old',action='read',err=999)
    inquire(uid,recl=i)
    if (i==size(x)) then
        read(uid,*) x
        found = .true.
    else
        found = .false.
    end if
    close(uid)

    ! If still here, check  they are increasing, and they are distinct
    if (any(x(1:size(x)-1)>=x(2:size(x)))) found = .false.

    ! Check that they are actually roots
    if (found) then
        !$OMP PARALLEL DO DEFAULT(SHARED) PRIVATE(i,tmp) SCHEDULE(STATIC,50) 
        do i = 1,order
            call legendre_eval(x(i),order,P=tmp)
            if (abs(tmp)>sqrt(epsilon(0._rt))) found = .false.
        end do
        !$OMP END PARALLEL DO 
    end if

    ! Get roots of legendre polynomials if couldn't read from file
999 if (.not. found) call legendre_roots(x)

    order = size(x)

    ! Get the weights
    !$OMP PARALLEL DO DEFAULT(SHARED) PRIVATE(i) SCHEDULE(STATIC,50) 
    do i = 1,order
        call legendre_eval(x(i),order,dP=w(i))
        ! Now w(i) stores P'_{n+1}. Use equation 12 here http://mathworld.wolfram.com/Legendre-GaussQuadrature.html (7/20/16)
        w(i) = 2d0/((1d0-x(i)**2)*w(i)**2)
    end do
    !$OMP END PARALLEL DO 

    ! Try to save to file 
    if (.not. found) then
        open(newunit=uid,file='/tmp/legendrequadnodes.dat',status='replace',action='write',err=998)
        write(uid,*,err=998) x
        close(uid)
    end if

    ! Now adapt to [a,b] via a linear change of variables
998 x = a+.5d0*(b-a)*(x+1d0)
    w = .5d0*(b-a)*w

end subroutine

! Computes the n roots of H_n, the Legendre polynomials
! P_0 = 1
! P_1 = x
!
! with the recurrence relations http://mathworld.wolfram.com/LegendrePolynomial.html (retrieved 7/20/16)
! P_n = (2*(n-1) + 1) * P_{n-1} - (n-1) * P_{n-2})/n
! P'_n = (-n * x * P_n + n * P_{n-1}) / (1d0-x**2)
! 
subroutine legendre_roots(x)
    implicit none
    real(rt), intent(inout) :: x(:)
    ! local
    integer :: order,i,n
    real(rt), allocatable :: xlast(:)

    n = size(x)

    allocate(xlast(n))

    do order = 1,n

        if (order==1) then
            x(1:order) = 0d0 
        else

            !$OMP PARALLEL DO DEFAULT(SHARED) PRIVATE(i) SCHEDULE(STATIC,20)
            do i = 1,order/2
                ! Look for a hermite root between the two previous ones
                if (i==1) then
                    call legendre_bracket(x(i),order,-1d0,xlast(1))
                elseif (i==order) then
                    call legendre_bracket(x(i),order,xlast(i-1),1d0)
                else
                    call legendre_bracket(x(i),order,xlast(i-1),xlast(i))
                end if
            end do
            !$OMP END PARALLEL DO

            if (mod(order,2)==1) then
                ! Odd number
                x((order+1)/2) = 0d0
                do i = (order+1)/2+1,order
                    x(i) = -x(order-i+1)
                end do
            else
                ! Even number
                do i = order/2+1,order
                    x(i) = -x(order-i+1)
                end do
            end if

        end if

        xlast = x

    end do

!     print*,'x',x

end subroutine

! Uses bracketing method to find root of P_n in a,b
subroutine legendre_bracket(x,n,a,b)
    use mod_root1dim
    implicit none
    real(rt), intent(inout) :: x
    real(rt), intent(in) :: a,b
    integer, intent(in) :: n
    ! local
    real(rt) :: s,tmp,P,dP,tmpP
    integer :: iter

    if (n==0) then
        write(0,*) 'WARNING: legendre_bracket: there is no root for P_0'
        return
    end if

    if (n==1) then
        x = 0d0
    else

        ! Try Newton
        x = .5d0*(a+b)
        do iter = 1,100
            call legendre_eval(x,n,P,dP)
            s = P/dP
            x = x - s
            if (abs(s)<1e2_rt*epsilon(0._rt)) exit
        end do

        if (abs(s)<1e2_rt*epsilon(0._rt) .and. x>a .and. x<b) then
!             print*,'Newton succeeded!'
            return
        end if

!         print*,'Newton failed :('

        ! If still here, Newton didn't work, so use Brent
        call sub_brent(x,res,a,b,printmod_in=-1)

        ! Try to polish off with Newton
        tmp = x
        call legendre_eval(x,n,tmpP)

        do iter = 1,10
            call legendre_eval(x,n,P,dP)
            s = P/dP
            x = x - s
        end do

        call legendre_eval(x,n,P)

        if (x<a .or. x>b .or. abs(tmpP)<abs(P)) then
            ! Newton failed to improve, so revert
            x = tmp
        end if

    end if

contains 

    real(rt) function res(x)
        implicit none
        real(rt), intent(in) :: x
        real(rt) :: P

        call legendre_eval(x,n,P)
        res = P
    end function 

end subroutine 

! Gives the P_n and P'_n at a given x
subroutine legendre_eval(x,n,P,dP)
    implicit none
    real(rt), intent(in) :: x
    integer, intent(in) :: n
    real(rt), intent(out), optional :: P,dP
    ! local
    real(8) :: Hzero,Hone,Htwo
    integer :: i

    ! H_0 = 1
    ! H_1 = x

    if (n==0) then
        if (present(P)) P = 1d0
        if (present(dP)) dP = 0d0
        return
    elseif (n==1) then
        if (present(P)) P = x
        if (present(dP)) dP = 1d0
        return
    end if

    Htwo = 1d0
    Hone = x
    do i = 2,n

        ! P_n = (2*(n-1) + 1) * x * P_{n-1} - (n-1) * P_{n-2})/n
        ! P_n = (2*n- 1) * x * P_{n-1} - (n-1) * P_{n-2})/n
        Hzero = ( (2*i-1)*x*Hone - (i-1)*Htwo)/i

        if (i<n) then
            Htwo = Hone
            Hone = Hzero
        end if
    end do

    ! Output the results
    if (present(P)) P = Hzero

    ! P'_n = (-n * x * P_n + n * P_{n-1}) / (1d0-x**2)
    if (present(dP)) then
        if (abs(x)/=1d0) then
            dP = (-n * x * Hzero + n*Hone )/(1._rt-x**2)
        else
            dP = (-n * x * Hzero + n*Hone )*1e10_rt
        end if
    end if


end subroutine 

! Computes the n roots of H_n, the hermite polynomials
! H_0 = 1
! H_1 = 2*x
! H_2 = 4*x**2 - 2
!
! with the recurrence relations http://mathworld.wolfram.com/HermitePolynomial.html (retrieved 7/20/16)
! H_n = 2 * x * H_{n-1} - 2 * (n-1) * H_{n-2}
! H'_n = 2 * n * H_{n-1}
! 
! 
! For Newton's method, need x' = x - H_n/H'_n. This is 
! x' = x - (x/n - (n-1)/n * H_{n-2}/H_{n-1})
! So, only need Hermite values rather than the derivatives
subroutine hermite_roots(x)
    implicit none
    real(rt), intent(inout) :: x(:)
    ! local
    integer :: order,i,n
    real(rt), allocatable :: xlast(:)

    n = size(x)

    allocate(xlast(n))

    do order = 1,n

        if (order==1) then
            x(1:order) = 0d0 
        elseif (order==2) then
            x(1:order) = .5d0*[-sqrt(2d0),sqrt(2d0)]
        elseif (order==3) then
            x(1:order) = .5d0*[-sqrt(6d0),0d0,sqrt(6d0)]
        else
            do i = 1,order
                ! Look for a hermite root between the two previous ones
                if (i==1) then
                    ! x(i) = (1d0+1d-3)*xlast(1) ! will be negative
                    ! call hermite_newton(x(i),order)
                    call hermite_bracket(x(i),order,-1d3,xlast(1))
                elseif (i==order) then
                    ! x(i) = (1d0+1d-3)*xlast(i-1)
                    ! call hermite_newton(x(i),order)
                    call hermite_bracket(x(i),order,xlast(i-1),1d3)
                else
                    ! x(i) = .5d0*(xlast(i-1)+xlast(i))
                    ! call hermite_newton(x(i),order)
                    call hermite_bracket(x(i),order,xlast(i-1),xlast(i))
                end if
            end do

        end if

        xlast = x

    end do

!     print*,'x',x

end subroutine

! Uses bracketing method to find root of H_n in a,b
subroutine hermite_bracket(x,n,a,b)
    use mod_root1dim
    implicit none
    real(rt), intent(inout) :: x
    real(rt), intent(in) :: a,b
    integer, intent(in) :: n
    ! local
    real(rt) :: s,H,Hp
    integer :: iter

    if (n==0) then
        write(0,*) 'WARNING: hermite_newton: there is no root for H_0'
        return
    end if

    if (n==1) then
        x = 0d0
    else
        call sub_brent(x,res,a,b,printmod_in=-1)
    end if

contains 

    real(rt) function res(x)
        implicit none
        real(rt), intent(in) :: x
        real(rt) :: H

        call hermite_eval(x,n,H)
        res = H
    end function 

end subroutine 


! Uses Newton's method to find a root of the nth polynomial starting from x
subroutine hermite_newton(x,n)
    implicit none
    real(rt), intent(inout) :: x
    integer, intent(in) :: n
    ! local
    real(rt) :: s,H,Hp
    integer :: iter

    if (n==0) then
        write(0,*) 'WARNING: hermite_newton: there is no root for H_0'
        return
    end if

    if (n==1) then
        x = 0d0
    else
        do iter = 1,1000

            call hermite_eval(x,n,H,Hp)
            print*,'n',n,'x,H,Hp',x,H,Hp
            s = H/Hp
            if (abs(s)<1d-10) return
            x = x - s
        end do
        write(0,*) 'WARNING: hermite_newton: failed to find root for H_n, step size is',s
    end if

end subroutine 

! Gives the H_n and H'_n at a given x
subroutine hermite_eval(x,n,H,Hp)
    implicit none
    real(rt), intent(in) :: x
    integer, intent(in) :: n
    real(rt), intent(out), optional :: H,Hp
    ! local
    real(8) :: Hzero,Hone,Htwo
    integer :: i

    ! H_0 = 1
    ! H_1 = 2*x
    ! H_2 = 4*x**2 - 2

    if (n==0) then
        if (present(H)) H = 1d0
        if (present(Hp)) Hp = 0d0
        return
    elseif (n==1) then
        if (present(H)) H = 2d0*x
        if (present(Hp)) Hp = 2d0
        return
    elseif (n==2) then
        if (present(H)) H = 4d0*x**2 - 2d0
        if (present(Hp)) Hp = 8d0*x
        return
    end if

    Htwo = 2d0*x
    Hone = 4d0*x**2 - 2d0
    do i = 3,n
        ! H_n = 2 * x * H_{n-1} - 2 * (n-1) * H_{n-2}
        Hzero = 2d0 * x * Hone - 2d0 * (i-1) * Htwo

        if (i<n) then
            Htwo = Hone
            Hone = Hzero
        end if
    end do

    ! Output the results
    if (present(H)) H = Hzero
    if (present(Hp)) Hp = 2d0 * n * Hone


end subroutine 

! Integrates 
! \int^b_a f(x) dx 
! using Chebyshev quadrature adapted from [-1,1] with weighting function 1/sqrt(1-x**2).
subroutine chebyquad(x,w,a,b)
    use mod_matlab
    implicit none
    real(rt), intent(out) :: x(:),w(:)
    real(rt), intent(in) :: a,b
    ! local
    real(rt), allocatable :: xsort(:)
    integer, allocatable :: Isort(:)
    real(rt) :: tmp
    integer :: n,i

    n = size(x)

    if (size(w)/=n) STOP 'ERROR: chebyquad: x,w don''t conform'

    do i = 1,n
        tmp = (2e0_rt*real(i,rt)-1e0_rt)/2e0_rt/real(n,rt)*pi
        tmp = cos(tmp)
        x(i) = a + (b-a)*(tmp+1e0_rt)/2e0_rt
        w(i) = (b-a)/2e0_rt*pi/real(n,rt)*sqrt(1e0_rt-tmp**2)
    end do

    ! Sort to make sure they are ascending
    allocate(xsort(n),Isort(n))
    call sort2(xsort,Isort,x)
    x = xsort
    w = w(Isort)

!     if (abs(sum(w)-(b-a))>1d-6*abs(b-a)) then
!         print*,'w',w
!         print*,'x',x
!         stop 'checking chebyquad'
!         print*,'b',b
!         print*,'a',a
!         print*,'b-a',b-a
!         print*,'sum(w)',sum(w)
!         STOP 'ERROR: chebyquad: weights do not sum to b-a'
!     end if


end subroutine

! gauss-legendre here has
!  \int^1_{-1} f(x) dx 
! but I use a change of variables to handle
!  \int^b_a f(x) dx
! n-point rules exactly integrate a 2n-1 degree polynomial wrt a weighting function
! n is inferred from the size of x and w
subroutine legendre(x,w,a,b)
    implicit none
    real(rt), intent(out) :: x(:),w(:)
    real(rt), intent(in) :: a,b
    ! local

    integer :: n

    n = size(x)

    if (size(w)/=n) STOP 'ERROR: x,w don''t conform'

    ! Get the nodes and weights for [-1,1] 
    select case (n)
        case (1)
            x = 0e0_rt
            w = 2e0_rt
        case (2)
            x = [-sqrt(1e0_rt/3e0_rt),sqrt(1e0_rt/3e0_rt)]
            w = 1e0_rt
        case (3)
            x(1) = -sqrt(3e0_rt/5e0_rt)
            x(2) = 0e0_rt
            x(3) = sqrt(3e0_rt/5e0_rt)
            w(1) = 5e0_rt/9e0_rt
            w(2) = 8e0_rt/9e0_rt
            w(3) = 5e0_rt/9e0_rt
        case (5) 

            x(1) = -0.90618e0_rt
            x(2) = -0.538469e0_rt
            x(3) = 0e0_rt 
            x(4) = -x(2)
            x(5) = -x(1)

            w(1) = 0.236927e0_rt
            w(2) = 0.478629e0_rt
            w(3) = 0.568889e0_rt
            w(4) = w(2)
            w(5) = w(1)

        case (11)

            ! http://keisan.casio.com/exec/system/1329114617
            x(1)  = -0.978228658146056992804e0_rt;  w(1)  = 0.0556685671161736664828e0_rt
            x(2)  = -0.8870625997680952990752e0_rt; w(2)  = 0.1255803694649046246347e0_rt
            x(3)  = -0.730152005574049324093e0_rt;  w(3)  = 0.1862902109277342514261e0_rt
            x(4)  = -0.519096129206811815926e0_rt;  w(4)  = 0.2331937645919904799185e0_rt
            x(5)  = -0.2695431559523449723315e0_rt; w(5)  = 0.262804544510246662181e0_rt
            x(6)  = 0e0_rt                        ; w(6)  = 0.2729250867779006307145e0_rt
            x(7)  = 0.2695431559523449723315e0_rt;  w(7)  = 0.262804544510246662181e0_rt
            x(8)  = 0.5190961292068118159257e0_rt;  w(8)  = 0.233193764591990479919e0_rt
            x(9)  = 0.7301520055740493240934e0_rt;  w(9)  = 0.186290210927734251425e0_rt
            x(10) = 0.887062599768095299075e0_rt;   w(10) = 0.125580369464904624635e0_rt
            x(11) = 0.9782286581460569928039e0_rt;  w(11) = 0.055668567116173666483e0_rt

        case (21)

            x(1 ) = -0.99375217062038950026e0_rt; w(1 ) = 0.016017228257774333324e0_rt
            x(2 ) = -0.967226838566306294317e0_rt; w(2 ) = 0.0369537897708524938e0_rt
            x(3 ) = -0.92009933415040082879e0_rt; w(3 ) = 0.0571344254268572082836e0_rt
            x(4 ) = -0.853363364583317283647e0_rt; w(4 ) = 0.076100113628379302017e0_rt
            x(5 ) = -0.768439963475677908616e0_rt; w(5 ) = 0.0934444234560338615533e0_rt
            x(6 ) = -0.667138804197412319306e0_rt; w(6 ) = 0.108797299167148377663e0_rt
            x(7 ) = -0.551618835887219807059e0_rt; w(7 ) = 0.1218314160537285341954e0_rt
            x(8 ) = -0.4243421202074387835737e0_rt; w(8 ) = 0.1322689386333374617811e0_rt
            x(9 ) = -0.288021316802401096601e0_rt; w(9 ) = 0.1398873947910731547221e0_rt
            x(10) = -0.145561854160895090937e0_rt; w(10) = 0.1445244039899700590638e0_rt
            x(11) = 0e0_rt;                        w(11) = 0.146081133649690427192e0_rt
            x(12) = 0.145561854160895090937e0_rt; w(12) = 0.144524403989970059064e0_rt
            x(13) = 0.2880213168024010966008e0_rt; w(13) = 0.139887394791073154722e0_rt
            x(14) = 0.4243421202074387835737e0_rt; w(14) = 0.132268938633337461781e0_rt
            x(15) = 0.551618835887219807059e0_rt; w(15) = 0.1218314160537285341956e0_rt
            x(16) = 0.667138804197412319306e0_rt; w(16) = 0.108797299167148377663e0_rt
            x(17) = 0.768439963475677908616e0_rt; w(17) = 0.0934444234560338615534e0_rt
            x(18) = 0.853363364583317283647e0_rt; w(18) = 0.0761001136283793020169e0_rt
            x(19) = 0.92009933415040082879e0_rt; w(19) = 0.057134425426857208284e0_rt
            x(20) = 0.967226838566306294317e0_rt; w(20) = 0.0369537897708524938e0_rt
            x(21) = 0.99375217062038950026e0_rt; w(21) = 0.016017228257774333324e0_rt

        case default
            STOP 'ERROR: legendre: formula for requested n not programmed yet'
    end select

    ! Adapt to [a,b]
    w = (b-a)/2e0_rt*w
    x = (b-a)/2e0_rt*x + (a+b)/2e0_rt

    ! Note: the weights should definitely sum to b-a because even with 1 quadrature point, \int 1 di can be integrated exactly.
    if (abs(sum(w)-(b-a))>4.0_rt*epsilon(0.0_rt)*abs(b-a)) then
        print*,'w',w
        print*,'x',x
        print*,'b',b
        print*,'a',a
        print*,'b-a',b-a
        print*,'sum(w)',sum(w)
        STOP 'ERROR: legendre: weights do not sum to b-a'
    end if
end subroutine


! Trapezoidal rule for a predetermined x grid
! This specifies the quadrature weights for the trapezoidal rule
! at the nodes x. Note: if y = f(x) is piecewise linear, then
! this integrates the function exactly.
function trapzquad_fixedx(x) result(w)
    implicit none
    real(rt), intent(in) :: x(1:)
    real(rt) :: w(size(x))
    ! local
    integer :: n,i

    n = size(x)

    w(1) = .5e0_rt*h(1)
    w(n) = .5e0_rt*h(n-1)
    do i = 2,n-1
        w(i) = .5e0_rt*(h(i-1)+h(i))
    end do

contains
    real(rt) function h(i)
        integer, intent(in) :: i
        h = x(i+1)-x(i)
    end function
end function 

! Simpson's rule 
! NOTE: n must be odd
subroutine sub_simpsonquad(x,w,a,b)
    use mod_matlab
    implicit none
    real(rt), intent(out) :: x(:),w(:)
    real(rt), intent(in) :: a,b
    ! local
    real(rt) :: h
    integer :: i,n

    n = size(x)

    if (n/=size(w)) STOP 'ERROR: simpson: size of x and w do not agree'
    if (mod(n,2)==0) STOP 'ERROR: simpson: n must be odd'
    if (n<3) STOP 'ERROR: simpson: n must be >= 3'

    x = linspace(a,b,n)

    ! Step size from x(i) to x(i+1)
    h = (b-a)/real(n-1,rt) 

    ! End point weights
    w(1) = 1e0_rt
    w(n) = 1e0_rt

    ! Even interior values
    do i = 2,n-1,2
        w(i) = 4e0_rt
    end do

    ! Odd interior values
    do i = 3,n-2,2
        w(i) = 2e0_rt
    end do

    ! Scale by h/3
    w = w*h/3e0_rt

end subroutine sub_simpsonquad

! Integrate the piecewise linear function f(x) over domain [a,b] wrt the Lebesgue measure
! That is, \int^b_a f(x) dx. It is assumed the knots x are strictly increasing. 
! If a or b is outside the domain of x, linear extrapolation is used
function intPiecewiseLinear(xin,yin,a,b) result(qout)
    use mod_matlab
    implicit none
    real(rt), intent(in) :: a,b,xin(1:),yin(1:)
    real(rt) :: qout
    ! local
    real(rt), allocatable :: x(:),y(:)
    real(rt) :: q,m,xlo,xhi
    integer :: nx,ialo,ibhi,i

    nx = size(xin) 
    if (nx/=size(yin)) STOP 'x and y must have the same size'
    if (nx<2) STOP 'x must have at least length 2'
    if (b<a) STOP 'b must be >= a'

    allocate(x(nx),y(nx))

    x = xin
    y = yin

    ! Since linearly extrapolating, can just extend the range of x to a and b if a<x(1) or b>x(nx)
    if (a<x(1)) then
        ! Replace (x(1),y(1)) with (a,f(a))
        m = (y(2)-y(1))/(x(2)-x(1))       
        y(1) = y(1) + m*(a-x(1))
        x(1) = a
    end if
    if (b>x(nx)) then
        ! Replace (x(nx),y(nx)) with (b,f(b))
        m = (y(nx)-y(nx-1))/(x(nx)-x(nx-1))       
        y(nx) = y(nx-1) + m*(b-x(nx-1))
        x(nx) = b
    end if
    
    ! Now, everything is interior with x(1)<=a<=b<=x(n)
    call bsearch(ialo,a,x)
    call bsearch(ibhi,b,x)
    if (b>x(ibhi)) ibhi = ibhi + 1 ! Note: bsearch always returns lower bracket. We want higher.

    !!! Check that conditions hold
    if (a<x(1)) STOP 'ERROR: intPiecewiseLinear: should have x(1)<=a'
    if (b<a) STOP 'ERROR: intPiecewiseLinear: should have a<=b'
    if (x(nx)<b) STOP 'ERROR: intPiecewiseLinear: should have b<=x(nx)'
    if (.not. (a>=x(ialo) .and. a<=x(ialo+1))) STOP 'ERROR: intPiecewiseLinear: error in getting ialo'
    if (.not. (b>=x(ibhi-1) .and. b<=x(ibhi))) STOP 'ERROR: intPiecewiseLinear: error in getting ibhi'

    ! Integrate looping through the intervals
    q = 0e0_rt
    do i = ialo,ibhi-1
        ! In range [max(a,x(i)),min(b,x(i+1))], integrate
        xlo = max(a,x(i))
        xhi = min(b,x(i+1))

        m = (y(i+1)-y(i))/(x(i+1)-x(i))

        q = q + y(i)*(xhi-xlo) + m*(xhi-x(i))**2/2e0_rt - m*(xlo-x(i))**2/2e0_rt

    end do

    qout = q

end function

! Provides quadrature nodes (x) and weights (w) that can be used to 
! integrate a function f of dim d with respect to a multivariate normal
! distribution N(mean,varcov) using monomial rules taken from Stroud.
! There are several methods available
! 
! Method name      Description
! 'deg3B'          A degree 3 rule using 2*d spread out equally weighted points 
! ! 'deg5'           A degree 5 rule using 2d^2+1 pts at various distances
subroutine myquad(x,w,d,mean,varcov,method)
    use mod_matlab, only: chol,det,pi,strcmp
    implicit none
    real(rt), allocatable, intent(out) :: x(:,:),w(:)
    integer, intent(in) :: d
    real(rt), intent(in) :: mean(d),varcov(d,d)
    character(len=*), intent(in) :: method
    !local
    real(rt) :: Omega(d,d), absdetOmega

    ! Construct the cholesky decomposition of varcov
    Omega = transpose(chol(varcov))
    absdetOmega = abs(det(Omega))

    ! Get the weights for standard form
    if (strcmp(method,'deg3B')) then
        call int_deg3ruleB(x,w,d)
!     elseif (strcmp(method,'deg5')) then
!         call int_deg5rule(x,w,d)
    else
        STOP 'myquad: Method not found.  Remember it''s case sensitive.'
    end if

    ! Convert to non-standard form
    x = sqrt(2.e0_rt)*x*transpose(Omega) + transpose(spread(mean,2,size(x,1)))
    w = w*(2.e0_rt**(real(d,rt)/2.e0_rt))*absdetOmega*1.e0_rt/sqrt(abs(det(varcov)))*(2.e0_rt*pi)**(-real(d,rt)/2.e0_rt)

end subroutine myquad

subroutine int_deg3ruleB(x,w,d)
    use mod_matlab, only: pi, diag
    implicit none
    real(rt), intent(out), allocatable :: x(:,:),w(:)
    integer, intent(in) :: d
    !local
    real(rt) :: r,V

    allocate(x(2*d,d),w(2*d))

    r = sqrt(real(d,rt)/2.e0_rt);    
    x(1:d,:) = diag(spread(r,1,d))
    x(d+1:2*d,:) = diag(spread(-r,1,d))

    V = pi**(real(d,rt)/2.e0_rt)
!     E = V/(2*d)*sum(f);
    w = V/(2.e0_rt*real(d,rt))
end subroutine int_deg3ruleB


subroutine test_quad_deg5rule
    use mod_matlab, only: randn, mean
    implicit none
    integer, parameter :: dint = 3
    real(rt) :: x(2*dint**2+1,dint) !
    real(rt) :: w(2*dint**2+1) !
    
    call sub_quad_deg5rule(x,w,dint)

    ! integrate sum(2*(.9*x-3)**5+1,2) using formula
    write(*,*) 'E according to quad rule', dot_product(sum(2.e0_rt*(.9e0_rt*x-3.e0_rt)**5+1.e0_rt,dim=2),w)

    write(*,*) 'E according to MC       ', mean(sum(2.e0_rt*(.9e0_rt*randn(100000,dint)-3.e0_rt)**5+1.e0_rt,dim=2))

end subroutine test_quad_deg5rule

! Integrate \int_{R^d} f(x) (2\pi)^{-d/2} e^{-(1/2) \sum^d_{i=1} x^2_i dx}
! Which is the integral of f wrt a multivariate unit normal distribution. 
!
! Based on integrating \int_{R^d} f(x) e^{-\sum^d_{i=1} x^2_i dx} but uses
! a change of variables to get it into a more useful form.
! Uses a degree 5 rule as in Judd p.275.
! 
! x are the quadrature points  2*d**2+1 by d
! w are the quadrature weights  2*d**2+1
! d is the dimension
!
subroutine sub_quad_deg5rule(x,w,dint)
    use mod_matlab, only: pi, diag
    implicit none
    integer, intent(in) :: dint
    real(rt), intent(inout) :: x(1:,1:) !(2*dint**2+1,dint)
    real(rt), intent(inout) :: w(1:) !(2*dint**2+1)
    !local
    integer :: N, ind, i, j
    real(rt) :: d, V, A, B, C !, E
    real(rt) :: r,s, tmpPP(dint),tmpPM(dint),tmpMP(dint),tmpMM(dint)

    ! Check for correct sizes
    if (size(x,1)/=2*dint**2+1) STOP 'sub_quad_deg5rule: x has wrong size1'
    if (size(x,2)/=dint) STOP 'sub_quad_deg5rule: x has wrong size2'
    if (size(w)/=2*dint**2+1) STOP 'sub_quad_deg5rule: w has wrong size'

    ! Get quadrature points
    N = 2*dint**2+1 !Number of points used in formula
    d = real(dint,rt)
    r = sqrt(1.e0_rt+.5e0_rt*d)
    s = sqrt(.5e0_rt+.25e0_rt*d)
    
    !x = 0            
    x(1,:) = 0.e0_rt
    ! x= r*ei
    x(2:2+dint-1,:) = diag(spread(r,1,dint))
    ! x= -r*ei
    x(2+dint:2+dint+dint-1,:) = diag(spread(-r,1,dint))
    ! x = +-s*ei +-s*ej
    ind = 2+dint+dint
    do i = 1,dint-1
        do j = i+1,dint
            tmpPP = 0.e0_rt
            tmpPP(i) = s
            tmpPP(j) = s
            tmpPM = 0.e0_rt
            tmpPM(i) = s
            tmpPM(j) = -s
            tmpMP = 0.e0_rt
            tmpMP(i) = -s
            tmpMP(j) = s
            tmpMM = 0.e0_rt
            tmpMM(i) = -s
            tmpMM(j) = -s

            x(ind,:) = tmpPP
            x(ind+1,:) = tmpPM
            x(ind+2,:) = tmpMP
            x(ind+3,:) = tmpMM
            ind = ind+4
        end do
    end do

    ! Change of variables to integrate f wrt unit normal
    x = sqrt(2.e0_rt)*x

    ! Get quadrature weights
    N = 2*dint**2+1 !Number of points used in formula
    d = real(dint,rt)

    V = pi**(d/2.e0_rt)
    A = 2.e0_rt/(d + 2.e0_rt)*V
    B = (4.e0_rt-d)/(2.e0_rt*(d+2.e0_rt)**2)*V
    C = V/(d+2.e0_rt)**2

    w(1) = A
    w(2:2+dint-1) = B
    w(2+dint:2+dint+dint-1) = B
    w(2+dint+dint:2*dint**2+1) = C

    ! Use change of variables formula
    w = 1.e0_rt/(sqrt(pi)**dint)*w  

end subroutine sub_quad_deg5rule

!Integrate \int_{R^d} f(x) e^{-\sum^d_{i=1} x^2_i dx}
!using a degree 5 rule as in Judd p.275.
!Proceed in two steps.  In the first step, you passed values of x for which
!you must evaluate f.  Then passing the values back, the integration is calculated.
!
function int_deg5rule_step1(dint) result(x)
    use mod_matlab
    implicit none
    integer, intent(in) :: dint
    real(rt) :: x(2*dint**2+1,dint)
    !local
    integer :: N, ind, i, j
    real(rt) :: d,r,s, tmpPP(dint),tmpPM(dint),tmpMP(dint),tmpMM(dint)
    !Values of x that need to be evaluated    
    !Variable definitions using monomial approach due to Stroud (1971) and found in Judd
    !integer, parameter :: dint = 4 !underscore J is for Judd

    N = 2*dint**2+1 !Number of points used in formula
    d = real(dint,rt)
    r = sqrt(1.e0_rt+.5e0_rt*d)
    s = sqrt(.5e0_rt+.25e0_rt*d)
    
    !x = 0            
    x(1,:) = 0.e0_rt
    ! x= r*ei
    x(2:2+dint-1,:) = diag(spread(r,1,dint))
    ! x= -r*ei
    x(2+dint:2+dint+dint-1,:) = diag(spread(-r,1,dint))

    ! x = +-s*ei +-s*ej
    ind = 2+dint+dint
    do i = 1,dint-1
        do j = i+1,dint
            tmpPP = 0.e0_rt
            tmpPP(i) = s
            tmpPP(j) = s
            tmpPM = 0.e0_rt
            tmpPM(i) = s
            tmpPM(j) = -s
            tmpMP = 0.e0_rt
            tmpMP(i) = -s
            tmpMP(j) = s
            tmpMM = 0.e0_rt
            tmpMM(i) = -s
            tmpMM(j) = -s

            x(ind,:) = tmpPP
            x(ind+1,:) = tmpPM
            x(ind+2,:) = tmpMP
            x(ind+3,:) = tmpMM
            ind = ind+4
        end do
    end do

end function

! See help for int_deg5rule_step1
function int_deg5rule_step2(dint,fx) result(E)
    use mod_matlab
    implicit none
    integer, intent(in) :: dint
    real(rt), dimension(1:), intent(in) :: fx
    real(rt) :: E
    !local
    integer :: N
    real(rt) :: d, V, A, B, C

    N = 2*dint**2+1 !Number of points used in formula
    d = real(dint,rt)

    V = pi**(d/2e0_rt)
    A = 2e0_rt/(d + 2e0_rt)*V
    B = (4e0_rt-d)/(2e0_rt*(d+2e0_rt)**2)*V
    C = V/(d+2e0_rt)**2

    E = (A*fx(1)+ B*sum(fx(2:2+dint-1)+fx(2+dint:2+dint+dint-1)) + C*sum(fx(2+dint+dint:size(fx,1))))

end function


! Degree 11 rule due to Stroud
function deg11rule_pts(d) result(x)
    implicit none
    integer, intent(in) :: d
    real(rt), allocatable :: x(:,:)
    !local
    real(rt), allocatable :: w(:)
    real(rt), allocatable :: xlocal(:,:)
    integer :: o
    if (d<3 .or. d>5) STOP 'deg11rule can only handle a dimension of 3,4, or 5'

!     N    O
!    __  ___
!     3  151
!     4  417
!     5  983
    if (d==3) o = 151
    if (d==4) o = 417
    if (d==5) o = 983

    allocate(xlocal(o,d),x(o,d),w(o))

    call en_r2_11_1(d,1,o,xlocal,w)
    
    ! Modify the points so they can be thought of as innovations to a normal distribution
    x = sqrt(2.e0_rt)*xlocal

end function deg11rule_pts


! THE TWO ROUTINES ARE ADAPTED TO FORTRAN FROM JON BURKARDT'S WEBSITE:
! http://people.sc.fsu.edu/~jburkardt/m_src/stroud/en_r2_11_1.m
subroutine en_r2_11_1 ( n, option, o, x, w )
!*****************************************************************************80
!
!! EN_R2_11_1 implements the Stroud rule 11.1 for region EN_R2.
!
!  Discussion:
!
!    The rule has order 
!
!      O = ( 4 * N^5 - 20 * N^4 + 140 * N^3 - 130 * N^2 + 96 * N + 15 ) / 15.
!
!    The rule has precision P = 11.
!
!    EN_R2 is the entire N-dimensional space with weight function
!
!      w(x) = product ( 1 <= i <= n ) ( exp ( - x(i)^2 ) 
!
!    There are two versions of each rule, chosen by setting the
!    OPTION variable to 1 or 2.
!
!    The rule as tabulated by Stenger is available for N = 2 through 20.
!    This function accepts N = 3 through 5.
!
!     N    O
!    __  ___
!     3  151
!     4  417
!     5  983
!
!  Licensing:
!
!    This code is distributed under the GNU LGPL license.
!
!  Modified:
!
!    23 January 2010
!
!  Author:
!
!    John Burkardt
!
!  Reference:
!
!    Arthur Stroud,
!    Approximate Calculation of Multiple Integrals,
!    Prentice Hall, 1971,
!    ISBN: 0130438936,
!    LC: QA311.S85.
!
!  Parameters:
!
!    Input, integer ( kind = 4 ) N, the spatial dimension.
!    3 <= N <= 5.
!
!    Input, integer ( kind = 4 ) OPTION, chooses rule option 1 or 2.
!
!    Input, integer ( kind = 4 ) O, the order.
!
!    Output, real ( kind = 8 ) X(N,O), the abscissas.
!
!    Output, real ( kind = 8 ) W(O), the weights.
!
  implicit none

  integer ( kind = 4 ) n
  integer ( kind = 4 ) o

  real    ( kind = rt ) b0
  real    ( kind = rt ) b1
  real    ( kind = rt ) b2
  real    ( kind = rt ) b3
  real    ( kind = rt ) b4
  real    ( kind = rt ) b5
  real    ( kind = rt ) b6
  real    ( kind = rt ) b7
  real    ( kind = rt ) b8
  real    ( kind = rt ) b9
  real    ( kind = rt ) b10
  real    ( kind = rt ) b11
  real    ( kind = rt ) b12
  real    ( kind = rt ) b13
  real    ( kind = rt ) b14
  real    ( kind = rt ) b15
  integer ( kind = 4 ) i
  integer ( kind = 4 ) i1
  integer ( kind = 4 ) i2
  integer ( kind = 4 ) i3
  integer ( kind = 4 ) i4
  integer ( kind = 4 ) i5
  integer ( kind = 4 ) j
  integer ( kind = 4 ) k
  integer ( kind = 4 ) l
  integer ( kind = 4 ) m
!   logical              more
  integer ( kind = 4 ) option
  real    ( kind = rt ), parameter :: pi = 3.141592653589793e+00_rt
  real    ( kind = rt ) u
  real    ( kind = rt ) v
  real    ( kind = rt ) volume
  real    ( kind = rt ) w2
  real    ( kind = rt ) w(o)
  real    ( kind = rt ) x(n,o)

  if ( n < 3 .or. 5 < n ) then
    write ( *, '(a)' ) ' '
    write ( *, '(a)' ) 'EN_R2_11_1 - Fatal error!'
    write ( *, '(a)' ) '  3 <= N <= 5 required.'
    stop
  end if

  if ( option < 1 .or. 2 < option ) then
    write ( *, '(a)' ) ' '
    write ( *, '(a)' ) 'EN_R2_11_1 - Fatal error!'
    write ( *, '(a)' ) '  1 <= OPTION <= 2 required.'
    stop
  end if

  volume = sqrt ( pi ** n )

  if ( n == 3 .and. option == 1 ) then
    u =     0.235060497367449e+01_rt
    v =     0.436077411927617e+00_rt
    w2 =    0.133584907401370e+01_rt
    b0 =  - 0.881591029957858e+01_rt
    b1 =  - 0.751996143360650e-01_rt
    b2 =    0.621743189471515e+01_rt
    b3 =    0.241426451456494e+00_rt
    b4 =  - 0.120709739276065e-02_rt
    b5 =  - 0.427751221210138e+01_rt
    b6 =    0.550169924840163e-01_rt
    b7 =    0.237084999634707e-01_rt
    b8 =  - 0.169791992887741e-02_rt
    b9 =  - 0.252266276123350e-04_rt
    b10 =   0.326777873717691e+01_rt
    b11 =   0.968469949206802e-02_rt
    b12 =   0.789754514877422e-03_rt
    b13 =   0.000000000000000e+00_rt
    b14 =   0.000000000000000e+00_rt
    b15 =   0.000000000000000e+00_rt
  else if ( n == 3 .and. option == 2 ) then
    u =     0.235060497367449e+01_rt
    v =     0.133584907401370e+01_rt
    w2 =    0.436077411927617e+00_rt
    b0 =  - 0.141214037032900e+02_rt
    b1 =  - 0.803730274707282e-01_rt
    b2 =    0.235546545595906e+00_rt
    b3 =    0.888123191556611e+01_rt
    b4 =    0.142467131155533e-03_rt
    b5 =    0.582993124006494e-01_rt
    b6 =  - 0.561099173155661e+01_rt
    b7 =  - 0.204028691521686e-02_rt
    b8 =    0.252880089932256e-01_rt
    b9 =  - 0.814378678627283e-04_rt
    b10 =   0.804353953375146e-02_rt
    b11 =   0.393451849690453e+01_rt
    b12 =   0.171183493169724e-03_rt
    b13 =   0.000000000000000e+00_rt
    b14 =   0.000000000000000e+00_rt
    b15 =   0.000000000000000e+00_rt
  else if ( n == 4 .and. option == 1 ) then
    u =     0.235060497367449e+01_rt
    v =     0.436077411927617e+00_rt
    w2 =    0.133584907401370e+01_rt
    b0 =    0.241502736147339e+03_rt
    b1 =  - 0.196095938531478e+00_rt
    b2 =  - 0.128675737999280e+03_rt
    b3 =    0.307568784278696e+00_rt
    b4 =  - 0.480908422319460e-02_rt
    b5 =    0.698087019367085e+02_rt
    b6 =    0.631837143743771e-01_rt
    b7 =    0.392226151971179e-01_rt
    b8 =  - 0.300948471646799e-02_rt
    b9 =  - 0.650235306755170e-04_rt
    b10 = - 0.386951974646715e+02_rt
    b11 =   0.171656829095787e-01_rt
    b12 =   0.139980343116450e-02_rt
    b13 =   0.101552487093372e-04_rt
    b14 =   0.222435922356439e+02_rt
    b15 =   0.000000000000000e+00_rt
  else if ( n == 4 .and. option == 2 ) then
    u =     0.235060497367449e+01_rt
    v =     0.133584907401370e+01_rt
    w2 =    0.436077411927617e+00_rt
    b0 =  - 0.151944464736584e+03_rt
    b1 =  - 0.223498438689039e+00_rt
    b2 =    0.243574919068010e+00_rt
    b3 =    0.634373877008693e+02_rt
    b4 =  - 0.782065187814018e-04_rt
    b5 =    0.911833754536616e-01_rt
    b6 =  - 0.238927288245914e+02_rt
    b7 =  - 0.422314408318853e-02_rt
    b8 =    0.448218289217760e-01_rt
    b9 =  - 0.138053374667391e-03_rt
    b10 =   0.607473265800655e-02_rt
    b11 =   0.697375246129742e+01_rt
    b12 =   0.303414841680135e-03_rt
    b13 = - 0.314574391771792e-05_rt
    b14 =   0.409103498175100e-02_rt
    b15 =   0.000000000000000e+00_rt
  else if ( n == 5 .and. option == 1 ) then
    u =     0.235060497367449e+01_rt
    v =     0.436077411927617e+00_rt
    w2 =    0.133584907401370e+01_rt
    b0 =    0.255885269311763e+04_rt
    b1 =  - 0.439598677491526e+00_rt
    b2 =  - 0.106541406144610e+04_rt
    b3 =    0.453540909054264e+00_rt
    b4 =  - 0.132100905623778e-01_rt
    b5 =    0.418606568954203e+03_rt
    b6 =    0.511394563043680e-01_rt
    b7 =    0.645581013845604e-01_rt
    b8 =  - 0.533417277494500e-02_rt
    b9 =  - 0.137981626254496e-03_rt
    b10 = - 0.147436933189884e+03_rt
    b11 =   0.304253807765057e-01_rt
    b12 =   0.248108698207828e-02_rt
    b13 =   0.113652094546015e-04_rt
    b14 =   0.394257407160391e+02_rt
    b15 =   0.331725011358320e-05_rt
  else if ( n == 5 .and. option == 2 ) then
    u =     0.235060497367449e+01_rt
    v =     0.133584907401370e+01_rt
    w2 =    0.436077411927617e+00_rt
    b0 =  - 0.761305347548192e+03_rt
    b1 =  - 0.536360805019297e+00_rt
    b2 =    0.110669832078736e+00_rt
    b3 =    0.246421088923968e+03_rt
    b4 =  - 0.773649327968607e-03_rt
    b5 =    0.169088641205970e+00_rt
    b6 =  - 0.670700680243651e+02_rt
    b7 =  - 0.856090560229205e-02_rt
    b8 =    0.794446232770302e-01_rt
    b9 =  - 0.220272863263544e-03_rt
    b10 = - 0.373515812228225e-02_rt
    b11 =   0.123606544052884e+02_rt
    b12 =   0.537788804557843e-03_rt
    b13 = - 0.122101861480881e-04_rt
    b14 =   0.725117070759373e-02_rt
    b15 =   0.331725011358320e-05_rt
  end if

  x(1:n,1:o) = 0.0D+00

  k = 0
!
!  1 point.
!
  k = k + 1
! x(1:n,k) = 0.0D+00
  w(k) = b0
!
!  2 * N points.
!
  do i = 1, n
    k = k + 1
    x(i,k) = - u
    w(k) = b1
    k = k + 1
    x(i,k) = + u
    w(k) = b1
  end do
!
!  2 * N points.
!
  do i = 1, n
    k = k + 1
    x(i,k) = - v
    w(k) = b2
    k = k + 1
    x(i,k) = + v
    w(k) = b2
  end do
!
!  2 * N points.
!
  do i = 1, n
    k = k + 1
    x(i,k) = - w2
    w(k) = b3
    k = k + 1
    x(i,k) = + w2
    w(k) = b3
  end do
!
!  4 * ( N * ( N - 1 ) / 2 ) points.
!
  do i = 1, n - 1
    do j = i + 1, n
      k = k + 1
      x(i,k) = - u
      x(j,k) = - u
      w(k) = b4
      k = k + 1
      x(i,k) = - u
      x(j,k) = + u
      w(k) = b4
      k = k + 1
      x(i,k) = + u
      x(j,k) = - u
      w(k) = b4
      k = k + 1
      x(i,k) = + u
      x(j,k) = + u
      w(k) = b4
    end do
  end do
!
!  4 * ( N * ( N - 1 ) / 2 ) points.
!
  do i = 1, n - 1
    do j = i + 1, n
      k = k + 1
      x(i,k) = - v
      x(j,k) = - v
      w(k) = b5
      k = k + 1
      x(i,k) = - v
      x(j,k) = + v
      w(k) = b5
      k = k + 1
      x(i,k) = + v
      x(j,k) = - v
      w(k) = b5
      k = k + 1
      x(i,k) = + v
      x(j,k) = + v
      w(k) = b5
    end do
  end do
!
!  4 * ( N * ( N - 1 ) / 2 ) points.
!
  do i = 1, n - 1
    do j = i + 1, n
      k = k + 1
      x(i,k) = - w2
      x(j,k) = - w2
      w(k) = b6
      k = k + 1
      x(i,k) = - w2
      x(j,k) = + w2
      w(k) = b6
      k = k + 1
      x(i,k) = + w2
      x(j,k) = - w2
      w(k) = b6
      k = k + 1
      x(i,k) = + w2
      x(j,k) = + w2
      w(k) = b6
    end do
  end do
!
!  4 * ( N * ( N - 1 ) ) points.
!
  do i = 1, n - 1
    do j = i + 1, n
      k = k + 1
      x(i,k) = - u
      x(j,k) = - v
      w(k) = b7
      k = k + 1
      x(i,k) = - u
      x(j,k) = + v
      w(k) = b7
      k = k + 1
      x(i,k) = + u
      x(j,k) = - v
      w(k) = b7
      k = k + 1
      x(i,k) = + u
      x(j,k) = + v
      w(k) = b7
      k = k + 1
      x(i,k) = - v
      x(j,k) = - u
      w(k) = b7
      k = k + 1
      x(i,k) = - v
      x(j,k) = + u
      w(k) = b7
      k = k + 1
      x(i,k) = + v
      x(j,k) = - u
      w(k) = b7
      k = k + 1
      x(i,k) = + v
      x(j,k) = + u
      w(k) = b7
    end do
  end do
!
!  4 * ( N * ( N - 1 ) ) points.
!
  do i = 1, n - 1
    do j = i + 1, n
      k = k + 1
      x(i,k) = - u
      x(j,k) = - w2
      w(k) = b8
      k = k + 1
      x(i,k) = - u
      x(j,k) = + w2
      w(k) = b8
      k = k + 1
      x(i,k) = + u
      x(j,k) = - w2
      w(k) = b8
      k = k + 1
      x(i,k) = + u
      x(j,k) = + w2
      w(k) = b8
      k = k + 1
      x(i,k) = - w2
      x(j,k) = - u
      w(k) = b8
      k = k + 1
      x(i,k) = - w2
      x(j,k) = + u
      w(k) = b8
      k = k + 1
      x(i,k) = + w2
      x(j,k) = - u
      w(k) = b8
      k = k + 1
      x(i,k) = + w2
      x(j,k) = + u
      w(k) = b8
    end do
  end do
!
!  8 * ( N * ( N - 1 ) * ( N - 2 ) / 6 ) points.
!
  do i = 1, n - 2
    do j = i + 1, n - 1
      do l = j + 1, n
        k = k + 1
        x(i,k) = - u
        x(j,k) = - u
        x(l,k) = - u
        w(k) = b9
        k = k + 1
        x(i,k) = - u
        x(j,k) = - u
        x(l,k) = + u
        w(k) = b9
        k = k + 1
        x(i,k) = - u
        x(j,k) = + u
        x(l,k) = - u
        w(k) = b9
        k = k + 1
        x(i,k) = - u
        x(j,k) = + u
        x(l,k) = + u
        w(k) = b9
        k = k + 1
        x(i,k) = + u
        x(j,k) = - u
        x(l,k) = - u
        w(k) = b9
        k = k + 1
        x(i,k) = + u
        x(j,k) = - u
        x(l,k) = + u
        w(k) = b9
        k = k + 1
        x(i,k) = + u
        x(j,k) = + u
        x(l,k) = - u
        w(k) = b9
        k = k + 1
        x(i,k) = + u
        x(j,k) = + u
        x(l,k) = + u
        w(k) = b9
      end do
    end do
  end do
!
!  8 * ( N * ( N - 1 ) * ( N - 2 ) / 6 ) points.
!
  do i = 1, n - 2
    do j = i + 1, n - 1
      do l = j + 1, n
        k = k + 1
        x(i,k) = - v
        x(j,k) = - v
        x(l,k) = - v
        w(k) = b10
        k = k + 1
        x(i,k) = - v
        x(j,k) = - v
        x(l,k) = + v
        w(k) = b10
        k = k + 1
        x(i,k) = - v
        x(j,k) = + v
        x(l,k) = - v
        w(k) = b10
        k = k + 1
        x(i,k) = - v
        x(j,k) = + v
        x(l,k) = + v
        w(k) = b10
        k = k + 1
        x(i,k) = + v
        x(j,k) = - v
        x(l,k) = - v
        w(k) = b10
        k = k + 1
        x(i,k) = + v
        x(j,k) = - v
        x(l,k) = + v
        w(k) = b10
        k = k + 1
        x(i,k) = + v
        x(j,k) = + v
        x(l,k) = - v
        w(k) = b10
        k = k + 1
        x(i,k) = + v
        x(j,k) = + v
        x(l,k) = + v
        w(k) = b10
      end do
    end do
  end do
!
!  8 * ( N * ( N - 1 ) * ( N - 2 ) / 6 ) points.
!
  do i = 1, n - 2
    do j = i + 1, n - 1
      do l = j + 1, n
        k = k + 1
        x(i,k) = - w2
        x(j,k) = - w2
        x(l,k) = - w2
        w(k) = b11
        k = k + 1
        x(i,k) = - w2
        x(j,k) = - w2
        x(l,k) = + w2
        w(k) = b11
        k = k + 1
        x(i,k) = - w2
        x(j,k) = + w2
        x(l,k) = - w2
        w(k) = b11
        k = k + 1
        x(i,k) = - w2
        x(j,k) = + w2
        x(l,k) = + w2
        w(k) = b11
        k = k + 1
        x(i,k) = + w2
        x(j,k) = - w2
        x(l,k) = - w2
        w(k) = b11
        k = k + 1
        x(i,k) = + w2
        x(j,k) = - w2
        x(l,k) = + w2
        w(k) = b11
        k = k + 1
        x(i,k) = + w2
        x(j,k) = + w2
        x(l,k) = - w2
        w(k) = b11
        k = k + 1
        x(i,k) = + w2
        x(j,k) = + w2
        x(l,k) = + w2
        w(k) = b11
      end do
    end do
  end do
!
!  8 * ( N * ( N - 1 ) * ( N - 2 ) / 2 ) points.
!
  do i = 1, n - 2
    do j = i + 1, n - 1
      do l = j + 1, n
        k = k + 1
        x(i,k) = - u
        x(j,k) = - u
        x(l,k) = - v
        w(k) = b12
        k = k + 1
        x(i,k) = - u
        x(j,k) = - u
        x(l,k) = + v
        w(k) = b12
        k = k + 1
        x(i,k) = - u
        x(j,k) = + u
        x(l,k) = - v
        w(k) = b12
        k = k + 1
        x(i,k) = - u
        x(j,k) = + u
        x(l,k) = + v
        w(k) = b12
        k = k + 1
        x(i,k) = + u
        x(j,k) = - u
        x(l,k) = - v
        w(k) = b12
        k = k + 1
        x(i,k) = + u
        x(j,k) = - u
        x(l,k) = + v
        w(k) = b12
        k = k + 1
        x(i,k) = + u
        x(j,k) = + u
        x(l,k) = - v
        w(k) = b12
        k = k + 1
        x(i,k) = + u
        x(j,k) = + u
        x(l,k) = + v
        w(k) = b12
        k = k + 1
        x(i,k) = - u
        x(j,k) = - v
        x(l,k) = - u
        w(k) = b12
        k = k + 1
        x(i,k) = - u
        x(j,k) = - v
        x(l,k) = + u
        w(k) = b12
        k = k + 1
        x(i,k) = - u
        x(j,k) = + v
        x(l,k) = - u
        w(k) = b12
        k = k + 1
        x(i,k) = - u
        x(j,k) = + v
        x(l,k) = + u
        w(k) = b12
        k = k + 1
        x(i,k) = + u
        x(j,k) = - v
        x(l,k) = - u
        w(k) = b12
        k = k + 1
        x(i,k) = + u
        x(j,k) = - v
        x(l,k) = + u
        w(k) = b12
        k = k + 1
        x(i,k) = + u
        x(j,k) = + v
        x(l,k) = - u
        w(k) = b12
        k = k + 1
        x(i,k) = + u
        x(j,k) = + v
        x(l,k) = + u
        w(k) = b12
        k = k + 1
        x(i,k) = - v
        x(j,k) = - u
        x(l,k) = - u
        w(k) = b12
        k = k + 1
        x(i,k) = - v
        x(j,k) = - u
        x(l,k) = + u
        w(k) = b12
        k = k + 1
        x(i,k) = - v
        x(j,k) = + u
        x(l,k) = - u
        w(k) = b12
        k = k + 1
        x(i,k) = - v
        x(j,k) = + u
        x(l,k) = + u
        w(k) = b12
        k = k + 1
        x(i,k) = + v
        x(j,k) = - u
        x(l,k) = - u
        w(k) = b12
        k = k + 1
        x(i,k) = + v
        x(j,k) = - u
        x(l,k) = + u
        w(k) = b12
        k = k + 1
        x(i,k) = + v
        x(j,k) = + u
        x(l,k) = - u
        w(k) = b12
        k = k + 1
        x(i,k) = + v
        x(j,k) = + u
        x(l,k) = + u
        w(k) = b12
      end do
    end do
  end do
!
!  16 * ( N * ( N - 1 ) * ( N - 2 ) * ( N - 3 ) / 24 ) points.
!
  do i = 1, n - 3
    do j = i + 1, n - 2
      do l = j + 1, n - 1
        do m = l + 1, n
          k = k + 1
          x(i,k) = - u
          x(j,k) = - u
          x(l,k) = - u
          x(m,k) = - u
          w(k) = b13
          k = k + 1
          x(i,k) = - u
          x(j,k) = - u
          x(l,k) = - u
          x(m,k) = + u
          w(k) = b13
          k = k + 1
          x(i,k) = - u
          x(j,k) = - u
          x(l,k) = + u
          x(m,k) = - u
          w(k) = b13
          k = k + 1
          x(i,k) = - u
          x(j,k) = - u
          x(l,k) = + u
          x(m,k) = + u
          w(k) = b13
          k = k + 1
          x(i,k) = - u
          x(j,k) = + u
          x(l,k) = - u
          x(m,k) = - u
          w(k) = b13
          k = k + 1
          x(i,k) = - u
          x(j,k) = + u
          x(l,k) = - u
          x(m,k) = + u
          w(k) = b13
          k = k + 1
          x(i,k) = - u
          x(j,k) = + u
          x(l,k) = + u
          x(m,k) = - u
          w(k) = b13
          k = k + 1
          x(i,k) = - u
          x(j,k) = + u
          x(l,k) = + u
          x(m,k) = + u
          w(k) = b13
          k = k + 1
          x(i,k) = + u
          x(j,k) = - u
          x(l,k) = - u
          x(m,k) = - u
          w(k) = b13
          k = k + 1
          x(i,k) = + u
          x(j,k) = - u
          x(l,k) = - u
          x(m,k) = + u
          w(k) = b13
          k = k + 1
          x(i,k) = + u
          x(j,k) = - u
          x(l,k) = + u
          x(m,k) = - u
          w(k) = b13
          k = k + 1
          x(i,k) = + u
          x(j,k) = - u
          x(l,k) = + u
          x(m,k) = + u
          w(k) = b13
          k = k + 1
          x(i,k) = + u
          x(j,k) = + u
          x(l,k) = - u
          x(m,k) = - u
          w(k) = b13
          k = k + 1
          x(i,k) = + u
          x(j,k) = + u
          x(l,k) = - u
          x(m,k) = + u
          w(k) = b13
          k = k + 1
          x(i,k) = + u
          x(j,k) = + u
          x(l,k) = + u
          x(m,k) = - u
          w(k) = b13
          k = k + 1
          x(i,k) = + u
          x(j,k) = + u
          x(l,k) = + u
          x(m,k) = + u
          w(k) = b13
        end do
      end do
    end do
  end do
!
!  16 * ( N * ( N - 1 ) * ( N - 2 ) * ( N - 3 ) / 24 ) points.
!
  do i = 1, n - 3
    do j = i + 1, n - 2
      do l = j + 1, n - 1
        do m = l + 1, n
          k = k + 1
          x(i,k) = - v
          x(j,k) = - v
          x(l,k) = - v
          x(m,k) = - v
          w(k) = b14
          k = k + 1
          x(i,k) = - v
          x(j,k) = - v
          x(l,k) = - v
          x(m,k) = + v
          w(k) = b14
          k = k + 1
          x(i,k) = - v
          x(j,k) = - v
          x(l,k) = + v
          x(m,k) = - v
          w(k) = b14
          k = k + 1
          x(i,k) = - v
          x(j,k) = - v
          x(l,k) = + v
          x(m,k) = + v
          w(k) = b14
          k = k + 1
          x(i,k) = - v
          x(j,k) = + v
          x(l,k) = - v
          x(m,k) = - v
          w(k) = b14
          k = k + 1
          x(i,k) = - v
          x(j,k) = + v
          x(l,k) = - v
          x(m,k) = + v
          w(k) = b14
          k = k + 1
          x(i,k) = - v
          x(j,k) = + v
          x(l,k) = + v
          x(m,k) = - v
          w(k) = b14
          k = k + 1
          x(i,k) = - v
          x(j,k) = + v
          x(l,k) = + v
          x(m,k) = + v
          w(k) = b14
          k = k + 1
          x(i,k) = + v
          x(j,k) = - v
          x(l,k) = - v
          x(m,k) = - v
          w(k) = b14
          k = k + 1
          x(i,k) = + v
          x(j,k) = - v
          x(l,k) = - v
          x(m,k) = + v
          w(k) = b14
          k = k + 1
          x(i,k) = + v
          x(j,k) = - v
          x(l,k) = + v
          x(m,k) = - v
          w(k) = b14
          k = k + 1
          x(i,k) = + v
          x(j,k) = - v
          x(l,k) = + v
          x(m,k) = + v
          w(k) = b14
          k = k + 1
          x(i,k) = + v
          x(j,k) = + v
          x(l,k) = - v
          x(m,k) = - v
          w(k) = b14
          k = k + 1
          x(i,k) = + v
          x(j,k) = + v
          x(l,k) = - v
          x(m,k) = + v
          w(k) = b14
          k = k + 1
          x(i,k) = + v
          x(j,k) = + v
          x(l,k) = + v
          x(m,k) = - v
          w(k) = b14
          k = k + 1
          x(i,k) = + v
          x(j,k) = + v
          x(l,k) = + v
          x(m,k) = + v
          w(k) = b14
        end do
      end do
    end do
  end do
!
!  All quintuples UUUUU with 32 sign combinations.
!
  do i1 = 1, n - 4
    do i2 = i1 + 1, n - 3
      do i3 = i2 + 1, n - 2
        do i4 = i3 + 1, n - 1
          do i5 = i4 + 1, n
            k = k + 1
            x(i1,k) = - u
            x(i2,k) = - u
            x(i3,k) = - u
            x(i4,k) = - u
            x(i5,k) = - u
            w(k) = b15
            k = k + 1
            x(i1,k) = - u
            x(i2,k) = - u
            x(i3,k) = - u
            x(i4,k) = - u
            x(i5,k) = + u
            w(k) = b15
            k = k + 1
            x(i1,k) = - u
            x(i2,k) = - u
            x(i3,k) = - u
            x(i4,k) = + u
            x(i5,k) = - u
            w(k) = b15
            k = k + 1
            x(i1,k) = - u
            x(i2,k) = - u
            x(i3,k) = - u
            x(i4,k) = + u
            x(i5,k) = + u
            w(k) = b15
            k = k + 1
            x(i1,k) = - u
            x(i2,k) = - u
            x(i3,k) = + u
            x(i4,k) = - u
            x(i5,k) = - u
            w(k) = b15
            k = k + 1
            x(i1,k) = - u
            x(i2,k) = - u
            x(i3,k) = + u
            x(i4,k) = - u
            x(i5,k) = + u
            w(k) = b15
            k = k + 1
            x(i1,k) = - u
            x(i2,k) = - u
            x(i3,k) = + u
            x(i4,k) = + u
            x(i5,k) = - u
            w(k) = b15
            k = k + 1
            x(i1,k) = - u
            x(i2,k) = - u
            x(i3,k) = + u
            x(i4,k) = + u
            x(i5,k) = + u
            w(k) = b15
            k = k + 1
            x(i1,k) = - u
            x(i2,k) = + u
            x(i3,k) = - u
            x(i4,k) = - u
            x(i5,k) = - u
            w(k) = b15
            k = k + 1
            x(i1,k) = - u
            x(i2,k) = + u
            x(i3,k) = - u
            x(i4,k) = - u
            x(i5,k) = + u
            w(k) = b15
            k = k + 1
            x(i1,k) = - u
            x(i2,k) = + u
            x(i3,k) = - u
            x(i4,k) = + u
            x(i5,k) = - u
            w(k) = b15
            k = k + 1
            x(i1,k) = - u
            x(i2,k) = + u
            x(i3,k) = - u
            x(i4,k) = + u
            x(i5,k) = + u
            w(k) = b15
            k = k + 1
            x(i1,k) = - u
            x(i2,k) = + u
            x(i3,k) = + u
            x(i4,k) = - u
            x(i5,k) = - u
            w(k) = b15
            k = k + 1
            x(i1,k) = - u
            x(i2,k) = + u
            x(i3,k) = + u
            x(i4,k) = - u
            x(i5,k) = + u
            w(k) = b15
            k = k + 1
            x(i1,k) = - u
            x(i2,k) = + u
            x(i3,k) = + u
            x(i4,k) = + u
            x(i5,k) = - u
            w(k) = b15
            k = k + 1
            x(i1,k) = - u
            x(i2,k) = + u
            x(i3,k) = + u
            x(i4,k) = + u
            x(i5,k) = + u
            w(k) = b15
            k = k + 1
            x(i1,k) = + u
            x(i2,k) = - u
            x(i3,k) = - u
            x(i4,k) = - u
            x(i5,k) = - u
            w(k) = b15
            k = k + 1
            x(i1,k) = + u
            x(i2,k) = - u
            x(i3,k) = - u
            x(i4,k) = - u
            x(i5,k) = + u
            w(k) = b15
            k = k + 1
            x(i1,k) = + u
            x(i2,k) = - u
            x(i3,k) = - u
            x(i4,k) = + u
            x(i5,k) = - u
            w(k) = b15
            k = k + 1
            x(i1,k) = + u
            x(i2,k) = - u
            x(i3,k) = - u
            x(i4,k) = + u
            x(i5,k) = + u
            w(k) = b15
            k = k + 1
            x(i1,k) = + u
            x(i2,k) = - u
            x(i3,k) = + u
            x(i4,k) = - u
            x(i5,k) = - u
            w(k) = b15
            k = k + 1
            x(i1,k) = + u
            x(i2,k) = - u
            x(i3,k) = + u
            x(i4,k) = - u
            x(i5,k) = + u
            w(k) = b15
            k = k + 1
            x(i1,k) = + u
            x(i2,k) = - u
            x(i3,k) = + u
            x(i4,k) = + u
            x(i5,k) = - u
            w(k) = b15
            k = k + 1
            x(i1,k) = + u
            x(i2,k) = - u
            x(i3,k) = + u
            x(i4,k) = + u
            x(i5,k) = + u
            w(k) = b15
            k = k + 1
            x(i1,k) = + u
            x(i2,k) = + u
            x(i3,k) = - u
            x(i4,k) = - u
            x(i5,k) = - u
            w(k) = b15
            k = k + 1
            x(i1,k) = + u
            x(i2,k) = + u
            x(i3,k) = - u
            x(i4,k) = - u
            x(i5,k) = + u
            w(k) = b15
            k = k + 1
            x(i1,k) = + u
            x(i2,k) = + u
            x(i3,k) = - u
            x(i4,k) = + u
            x(i5,k) = - u
            w(k) = b15
            k = k + 1
            x(i1,k) = + u
            x(i2,k) = + u
            x(i3,k) = - u
            x(i4,k) = + u
            x(i5,k) = + u
            w(k) = b15
            k = k + 1
            x(i1,k) = + u
            x(i2,k) = + u
            x(i3,k) = + u
            x(i4,k) = - u
            x(i5,k) = - u
            w(k) = b15
            k = k + 1
            x(i1,k) = + u
            x(i2,k) = + u
            x(i3,k) = + u
            x(i4,k) = - u
            x(i5,k) = + u
            w(k) = b15
            k = k + 1
            x(i1,k) = + u
            x(i2,k) = + u
            x(i3,k) = + u
            x(i4,k) = + u
            x(i5,k) = - u
            w(k) = b15
            k = k + 1
            x(i1,k) = + u
            x(i2,k) = + u
            x(i3,k) = + u
            x(i4,k) = + u
            x(i5,k) = + u
            w(k) = b15
          end do
        end do
      end do
    end do
  end do

  return
end subroutine

subroutine en_r2_11_1_size ( n, option, o )

!*****************************************************************************80
!
!! EN_R2_11_1_SIZE sizes the Stroud rule 11.1 for region EN_R2.
!
!  Discussion:
!
!    The rule has order 
!
!      O = ( 4 * N^5 - 20 * N^4 + 140 * N^3 - 130 * N^2 + 96 * N + 15 ) / 15.
!
!    The rule has precision P = 11.
!
!    EN_R2 is the entire N-dimensional space with weight function
!
!      w(x) = product ( 1 <= i <= n ) ( exp ( - x(i)^2 ) 
!
!    There are two versions of each rule, chosen by setting the
!    OPTION variable to 1 or 2.
!
!    The rule as tabulated by Stenger is available for N = 2 through 20.
!    This function accepts N = 3 through 5.
!
!     N    O
!    __  ___
!     3  151
!     4  417
!     5  983
!
!  Licensing:
!
!    This code is distributed under the GNU LGPL license.
!
!  Modified:
!
!    27 January 2010
!
!  Author:
!
!    John Burkardt
!
!  Reference:
!
!    Arthur Stroud,
!    Approximate Calculation of Multiple Integrals,
!    Prentice Hall, 1971,
!    ISBN: 0130438936,
!    LC: QA311.S85.
!
!  Parameters:
!
!    Input, integer ( kind = 4 ) N, the spatial dimension.
!    3 <= N <= 5.
!
!    Input, integer ( kind = 4 ) OPTION, chooses rule option 1 or 2.
!
!    Output, integer ( kind = 4 ) O, the order.
!
  implicit none

  integer ( kind = 4 ) n
  integer ( kind = 4 ) o
  integer ( kind = 4 ) option

  if ( n < 3 .or. 5 < n ) then
    write ( *, '(a)' ) ' '
    write ( *, '(a)' ) 'EN_R2_11_1_SIZE - Fatal error!'
    write ( *, '(a)' ) '  3 <= N <= 5 required.'
    stop
  end if

  if ( option < 1 .or. 2 < option ) then
    write ( *, '(a)' ) ' '
    write ( *, '(a)' ) 'EN_R2_11_1_SIZE - Fatal error!'
    write ( *, '(a)' ) '  1 <= OPTION <= 2 required.'
    stop
  end if

  o = ( 4 * n ** 5 - 20 * n ** 4 + 140 * n ** 3 - 130 * n ** 2 &
    + 96 * n + 15 ) / 15


  return
end subroutine

end module mod_quad


module mod_quad_test
    use mod_quad
contains

    subroutine test_gausslegendre()
        use mod_matlab
        implicit none
        real(rt) :: x(5),xtrue(5),w(5),wtrue(5),tmp
        real(rt), allocatable :: roots(:),P(:)
        integer :: i,order
        integer, parameter :: maxorder = 1001

        allocate(roots(maxorder),P(maxorder))
        do order = 1,maxorder

            ! Do randomly because this is costly
            if (randu()>.01d0) cycle 

            call legendre_roots(roots(1:order))
            ! Check that P is zero at the roots
            do i = 1,order
                call legendre_eval(roots(i),order,P(i))
            end do

            print*,'order : ',order,' max abs P ',maxval(abs(P(1:order))),'min distance between roots',minval(roots(2:order)-roots(1:order-1))
            if (maxval(abs(P(1:order)))>sqrt(epsilon(0._rt))) stop 'test_gausslegendre: failed test 0'
        end do


        ! x = [-1d0,-.5d0,0d0,.5d0,1d0]
        ! do i = 1,size(x)
        !     call legendre_eval(x(i),3,xtrue(i))
        ! end do
        ! call plot(x,xtrue,'4th order Legendre Polynomial')
        ! stop

        call legendrequad(x,w,-1d0,1d0)

        print*,'x',x
        print*,'w',w

        xtrue(1:3) = [-.90618d0,-.538469d0,0d0]
        xtrue(4) = -xtrue(2)
        xtrue(5) = -xtrue(1)

        wtrue(1:3) = [0.236927d0,0.478629d0,0.568889d0]
        wtrue(4) = wtrue(2)
        wtrue(5) = wtrue(1)

        if (maxval(abs(x-xtrue))>1d-4) stop 'test_gausslegendre: failed test 1'
        if (maxval(abs(w-wtrue))>1d-4) stop 'test_gausslegendre: failed test 2'

        ! Integrate a 4th degree polynomial from -5 to 10, so should be exact : int 5*x**4 dx = x**5 | 10,-5 = 10**5 + 5**5 
        call legendrequad(x,w,-5d0,10d0)
        x = 5d0*x**4
        tmp = 10**5+5**5
        ! print*,'dot_product(x,w)',dot_product(x,w),'true',10**5+5**5
        if (abs(dot_product(x,w)-tmp)>1e-6*tmp) stop 'test_gausslegendre: failed test 3'

        print*,'test_gausslegendre: passed'

    end subroutine 

    subroutine test_gausshermite()
        implicit none
        real(rt) :: x(5),xtrue(5)

        call hermite_roots(x)

        xtrue(1:3) = [-2.02018d0,-.958572d0,0d0]
        xtrue(4) = -xtrue(2)
        xtrue(5) = -xtrue(1)

        if (maxval(abs(x-xtrue))>1d-4) stop 'test_gausshemite: failed'

    end subroutine 

    subroutine test_sub_simpsonquad()
        implicit none
        real(rt) :: x(11),w(11),y(11)
        real(rt) :: lo,hi, Eapprox,E
        
        lo = 2e0_rt
        hi = 5e0_rt
        call sub_simpsonquad(x,w,lo,hi)

        ! Integrate a quadratic function (should be exact)
        ! y = \int^h_l a + bx + cx**2 
        !   = ax + bx**2/2 + c x**3/3 |^h_l
        y = -.5e0_rt - 5e0_rt*x + 3e0_rt*x**2
        Eapprox = dot_product(w,y)
        E = -.5e0_rt*hi - 5e0_rt*hi**2/2e0_rt + 3e0_rt*hi**3/3e0_rt
        E = E - (-.5e0_rt*lo - 5e0_rt*lo**2/2e0_rt + 3e0_rt*lo**3/3e0_rt)

        if (abs(E-Eapprox)>abs(E)*sqrt(epsilon(0.0_rt))) STOP 'ERROR: test_sub_simpsonquad: not exact'

        print*,'test_sub_simpsonquad: passed'

    end subroutine

    subroutine test_intPiecewiseLinear()
        implicit none
        real(rt) :: x(4),y(4),a,b,q,qtrue,q2
        integer :: i

        x = [-3e0_rt,0e0_rt,10e0_rt,20e0_rt]
        y = -sqrt(2e0_rt)*x + 1e0_rt  ! Linear function so should integrate exactly

        a = x(1)
        b = x(size(x))
        qtrue = (b-a) - sqrt(2e0_rt)/2e0_rt*(b**2-a**2)
        q2 = dot_product(y,trapzquad_fixedx(x))
        if (abs(qtrue-q2)>abs(qtrue)*sqrt(epsilon(0._rt))) then
            print*,'qtrue',qtrue
            print*,'q2',q2
            print*,'sum(trapzquad_fixedx(x))',sum(trapzquad_fixedx(x))
            print*,'|qtrue-q2|=',abs(qtrue-q2)
            print*,'tolerance=',abs(qtrue)*sqrt(epsilon(0._rt))
            STOP 'qtrue and q2 are far apart'
        end if

        do i = 1,10000

            call random_number(a)
            call random_number(b)

            ! Get a in range [-25e0_rt,25e0_rt]
            a = a*50e0_rt - 25e0_rt
            ! Get b-a in range [0e0_rt,30e0_rt]
            b = a + b*30e0_rt

            ! Integrate
            q = intPiecewiseLinear(x,y,a,b)
            qtrue = (b-a) - sqrt(2e0_rt)/2e0_rt*(b**2-a**2)

            if (abs(qtrue-q)>abs(qtrue)*sqrt(epsilon(0.0_rt))) then
                print*,'i',i,'a',a,'b',b
                print*,'qtrue',qtrue
                print*,'q',q
                print*,'|qtrue-q|=',abs(qtrue-q)
                print*,'tolerance=',abs(qtrue)*sqrt(epsilon(0._rt))
                STOP 'qtrue and q are far apart'
            end if


        end do
        print*,'test_intPiecewiseLinear: passed'

    end subroutine
end module mod_quad_test 
