module mod_quicksort
    implicit none

    interface sub_quicksort
        module procedure sub_quicksort_1d_rank_inplace,&
                         sub_quicksort_1d_norank_inplace,&
                         sub_quicksort_1d_rank_outplace,&
                         sub_quicksort_1d_norank_outplace
    end interface
contains

    subroutine sub_quicksort_1d_rank_inplace(a,rank)
        real(8), intent(inout) :: a(:)
        integer, intent(inout) :: rank(:)
        integer :: i
        do i = 1,size(a)
            rank(i) = i
        end do
        call quicksort_work(a,rank,1,size(a))
    end subroutine
    subroutine sub_quicksort_1d_rank_outplace(a,rank,a0)
        real(8), intent(out) :: a(:)
        real(8), intent(in) :: a0(:)
        integer, intent(out) :: rank(:)
        integer :: i
        do i = 1,size(a0)
            rank(i) = i
        end do
        a = a0
        call quicksort_work(a,rank,1,size(a))
    end subroutine

    subroutine sub_quicksort_1d_norank_inplace(a)
        real(8), intent(inout) :: a(:)
        integer, allocatable :: rank(:)
        allocate(rank(size(a)))
        call quicksort_work(a,rank,1,size(a))
    end subroutine
    subroutine sub_quicksort_1d_norank_outplace(a,a0)
        real(8), intent(out) :: a(:)
        real(8), intent(in) :: a0(:)
        integer, allocatable :: rank(:)
        allocate(rank(size(a)))
        a = a0
        call quicksort_work(a,rank,1,size(a))
    end subroutine


    ! https://en.wikipedia.org/wiki/Quickselect
    ! Returns the k-th smallest element of list within left..right inclusive
    ! (i.e. left <= k <= right).
    function quickselect(a,k) result(kval)
        implicit none
        real(8) :: kval
        real(8), intent(in) :: a(:)
        integer, intent(in) :: k
        ! local
        real(8), allocatable :: acopy(:)
        real(8) :: tmp

        allocate(acopy(size(a)))
        acopy = a

        call quickselect_work(tmp,acopy,1,size(a),k)

        kval = tmp

    end function 

    ! The search space within the array is changing for each round - but the list
    !  is still the same size. Thus, k does not need to be updated with each round.
    recursive subroutine quickselect_work(aval,a,lo,hi,k)
        implicit none
        real(8), intent(out) :: aval
        real(8), intent(inout) ::  a(*)
        integer, intent(in) :: lo,hi,k
        integer :: pivotind

        ! Check stopping condition
        if (lo==hi) then
            aval = a(lo)
            return
        end if

        ! For selecting median (seems to give better performance)
        pivotind = lo+(hi-lo)/2
        if (A(pivotind)>=max(A(lo),A(hi))) then
            ! pivot currently is max, want median
            pivotind = merge(lo,hi,A(lo)>A(hi))
        elseif (A(pivotind)<=min(A(lo),A(hi))) then
            ! pivot currently is min, want median
            pivotind = merge(lo,hi,A(lo)<A(hi))
        end if
        
        pivotind = partition(lo,hi,pivotind)

        if (k==pivotind) then
            aval = a(pivotind)
            return
        elseif (k<pivotind) then
            call quickselect_work(aval,a,lo,pivotind-1,k)
        else
            call quickselect_work(aval,a,pivotind+1,hi,k)
        end if

!         ! Check stopping condition
!         if (lo==hi) then
!             aval = a(lo)
!             return
!         end if
! 
!         ! For selecting median (seems to give better performance)
!         pivot = A(lo+(hi-lo)/2)
!         if (pivot>=max(A(lo),A(hi))) then
!             ! pivot currently is max, want median
!             pivot = merge(A(lo),A(hi),A(lo)>A(hi))
!         elseif (pivot<=min(A(lo),A(hi))) then
!             ! pivot currently is min, want median
!             pivot = merge(A(lo),A(hi),A(lo)<A(hi))
!         end if
!         
!         i = lo
!         j = hi
! 
!         ! Need to put all elements less than pivot to the left, all greater to the right
!         do
! 
!             ! Moving from left, find the first element >= pivot
!             do while (a(i)<pivot)
!                 i = i + 1
!             end do
! 
!             ! Moving from right, find the first element <= pivot
!             do while (pivot<a(j))
!                 j = j-1
!             end do
! 
!             ! If i>=j, it means every element a(i)<pivot<a(j) for all i < pivot index < j, which is the goal.
!             if (i >= j) exit
! 
!             ! If here, have found two elements out of position, i.e., a(i)>=pivot and a(j)<=pivot. Swap so that a(i) <= pivot <= a(j)
!             call doswap(a(i),a(j))
! 
!             ! Now that have exchanged i and j, they are in correct order from lo to i and j to hi. Continue starting at i+1 and j-1
!             i = i+1
!             j = j-1
! 
!         end do
! 
!         if (j==k) then
!             aval = a(j)
!             return
!         elseif (k<j) then
!             call quickselect_work(aval,a,lo,j-1,k)
!         else
!             call quickselect_work(aval,a,j+1,hi,k)
!         end if

    contains

        integer function partition(lo,hi,pivot) result(p)
            implicit none
            ! local
            integer, intent(in) :: lo,hi,pivot
            real(8) :: pivotval
            integer :: storeind,i

            pivotval = a(pivot)
            call doswap(a(pivot),a(hi))
            storeind = lo
            do i = lo,hi-1
                if (a(i)<pivotval) then
                    call doswap(a(i),a(storeind))
                    storeind = storeind + 1
                end if
            end do
            call doswap(a(storeind),a(hi))

            p = storeind

        end function 

    end subroutine 

    ! http://rosettacode.org/wiki/Sorting_algorithms/Quicksort
    recursive subroutine quicksort_work(a,rank,lo,hi)
        implicit none
        real(8), intent(inout) ::  a(*)
        integer, intent(inout) :: rank(*)
        integer, intent(in) :: lo,hi
        integer :: i,j
        real(8) :: pivot

        ! If very few elements, use insertion sort
        if (hi-lo<=20) then
            call insert()
            return
        end if
        
        ! If here, need to sort 

        ! For selecting median (seems to give better performance)
        pivot = A(lo+(hi-lo)/2)
        if (pivot>=max(A(lo),A(hi))) then
            ! pivot currently is max, want median
            pivot = merge(A(lo),A(hi),A(lo)>A(hi))
        elseif (pivot<=min(A(lo),A(hi))) then
            ! pivot currently is min, want median
            pivot = merge(A(lo),A(hi),A(lo)<A(hi))
        end if
        
        i = lo
        j = hi

        ! Need to put all elements less than pivot to the left, all greater to the right
        do

            ! Moving from left, find the first element >= pivot
            do while (a(i)<pivot)
                i = i + 1
            end do

            ! Moving from right, find the first element <= pivot
            do while (pivot<a(j))
                j = j-1
            end do

            ! If i>=j, it means every element a(i)<pivot<a(j) for all i < pivot index < j, which is the goal.
            if (i >= j) exit

            ! If here, have found two elements out of position, i.e., a(i)>=pivot and a(j)<=pivot. Swap so that a(i) <= pivot <= a(j)
            call doswap(A(i),A(j))
            call doswapi(rank(i),rank(j))

            ! Now that have exchanged i and j, they are in correct order from lo to i and j to hi. Continue starting at i+1 and j-1
            i = i+1
            j = j-1

        end do

        if (j>lo)   call quicksort_work(a,rank,lo,j)
        if (hi>j+1) call quicksort_work(a,rank,j+1,hi)

    contains

        subroutine insert()
            implicit none
            integer :: i,j
            real(8) :: x
            integer :: ranktmp

            ! adapted from wiki
            do i = lo+1,hi
                x = a(i)
                ranktmp = rank(i)
                j = i-1
                do 
                    if (j<lo) exit
                    if (a(j)<=x) exit
                    a(j+1) = a(j)
                    rank(j+1) = rank(j)
                    j = j-1
                end do
                a(j+1) = x
                rank(j+1) = ranktmp
            end do

        end subroutine 

    end subroutine 

    subroutine doswap(a,b)
        implicit none
        real(8), intent(inout) :: a,b
        real(8) :: tmp
        tmp = a
        a = b
        b = tmp
    end subroutine 

    subroutine doswapi(a,b)
        implicit none
        integer, intent(inout) :: a,b
        integer :: tmp
        tmp = a
        a = b
        b = tmp
    end subroutine 
end module

module mod_quicksort_test
    use mod_quicksort
    implicit none
contains
    subroutine test_quicksort
!         use mod_matlab
        implicit none
        real(8), allocatable :: x(:),x0(:),y(:)
        integer, allocatable :: rank(:)
        integer :: i,iter
        real(8) :: tmp


!         allocate(x(101))
!         do i = 1,size(x)
!             x(i) = i
!         end do
! 
!         do i = 1,size(x)
!             print*,quickselect(x,i),x(i)
!         end do
!         deallocate(x)
! 
!         return
        
        

        do iter = 1,100

            call random_number(tmp)
            allocate(x(1+nint(tmp*100))) ! pick a length in 1 to 101
            allocate(x0(size(x)))
            allocate(y(size(x)))
            allocate(rank(size(x)))

            if (iter==1 .and. size(x)==9) then
                x0(1:9) = [6,5,3,1,8,7,2,4,-30]
            else
                call random_number(x0)
                x0 = (x0-.5d0)*10d0
                if (mod(iter,5)==0) then
                    x0 = real(nint(x0),8)
                end if
            end if

            ! Test in place
            x = x0
            call sub_quicksort(x,rank)
            if (maxval(abs(x0(rank)-x))>10d0*epsilon(0d0)) stop 'sub_quicksort: failed test 1'

            do i = 1,size(x)-1
                if (x(i)>x(i+1)) then
                    print*,'x',x
                    stop 'sub_quicksort: failed test 2'
                end if
            end do

!             call sort2(y,rank,x0)
!             if (any(y/=x)) stop 'sub_quicksort: failed test 5'

            call sub_quicksort(x,rank,x0)
            if (maxval(abs(x0(rank)-x))>10d0*epsilon(0d0)) stop 'sub_quicksort: failed test 3'

            do i = 1,size(x)-1
                if (x(i)>x(i+1)) stop 'sub_quicksort: failed test 4'
            end do

!             call sort2(y,rank,x0)
!             if (any(y/=x)) stop 'sub_quicksort: failed test 6'

            ! Test quickselect too
            do i = 1,size(x)
                tmp = quickselect(x0,i)
                ! print*,'n',size(x),'quickselect',tmp,'true',x(i),'error',tmp-x(i)
                tmp = tmp-x(i) ! error
                if (abs(tmp)>epsilon(0d0)) stop 'sub_quickselect: failed test 1'
            end do

            deallocate(x,x0,rank,y)

        end do
        
        write(*,'(A)') 'test_quicksort: passed'
        write(*,'(A)') 'test_quickselect: passed'
    end subroutine
end module 
