module mod_sparse2
    implicit none
   
    type sparse_mat
        integer :: nnz !number of nonzero elts
        real(8), dimension(:), allocatable :: val
        integer, dimension(:), allocatable :: row,col
    end type sparse_mat    

    interface operator (+)
        module procedure plus
    end interface

    interface operator (-)
        module procedure minus
    end interface

    interface operator (*)
        module procedure times
    end interface

    interface transpose
        module procedure transpose_sparse
    end interface    

contains

    subroutine allocate_sparse(spmat,nnz)
        implicit none
        type(sparse_mat), intent(inout) :: spmat
        integer, intent(in) :: nnz
        
        if (allocated(spmat%val) .OR. allocated(spmat%row) .OR. allocated(spmat%col)) STOP 'spmat already allocated!'
        spmat%nnz = nnz
        allocate(spmat%val(1:nnz),spmat%row(1:nnz),spmat%col(1:nnz))
        spmat%val = 0.d0
        spmat%row = 1
        spmat%col = 1
        
    end subroutine allocate_sparse

    subroutine deallocate_sparse(spmat)
        implicit none
        type(sparse_mat), intent(inout) :: spmat
        
        if (.NOT.(allocated(spmat%val) .OR. allocated(spmat%row) .OR. allocated(spmat%col))) STOP 'spmat not allocated!'
        spmat%nnz = -1
        deallocate(spmat%val,spmat%row,spmat%col)
        
    end subroutine deallocate_sparse

    function speye(nnz) result(eye)
        implicit none
        integer, intent(in) :: nnz
        type(sparse_mat) :: eye
        !local
        integer :: i
        
        call allocate_sparse(eye,nnz)        
        
        eye%val = 0.d0
        do i = 1,nnz
            eye%val(i) = 1.d0
            eye%row(i) = i
            eye%col(i) = i
        end do
        
    end function speye

    function minus(A,B) result(C)
        implicit none
        type(sparse_mat), intent(in) :: A,B
        type(sparse_mat) :: C
        !local
        type(sparse_mat) :: negB
        call allocate_sparse(negB,B%nnz)
        negB%val = -B%val
        negB%row = B%row
        negB%col = B%col
        C = A+negB
    end function

    function plus(A,B) result(D) 
        use mod_matlab, only: sort2
        implicit none
        type(sparse_mat), intent(in) :: A,B
        type(sparse_mat) :: D
        !local
        type(sparse_mat) :: C
        integer :: i,j,k,bothAB
        integer :: nrows, nnz
        integer, dimension(:), allocatable :: ind, ranking
  
        call allocate_sparse(C,A%nnz + B%nnz)
        C%val(1:A%nnz) = A%val
        C%row(1:A%nnz) = A%row
        C%col(1:A%nnz) = A%col
        C%val(A%nnz+1:B%nnz) = B%val
        C%row(A%nnz+1:B%nnz) = B%row
        C%col(A%nnz+1:B%nnz) = B%col

        !Now figure out which values (col,row) pairs appear twice and add them.
        !Assign all combinations a unique number.
        allocate(ind(C%nnz), ranking(C%nnz))
        nrows = maxval(C%row)
        ind = C%row + nrows*(C%col-1)

        call sort2(ind,ranking,ind) !sorts ind and gives ranking
        !Recursively add elts if they appear more than once
        nnz = C%nnz
        do i = size(ind)-1,1,-1
            if (ind(i)==ind(i+1)) then
                C%val(i) = C%val(i) + C%val(i+1)
                C%val(i+1) = 0.d0
                nnz = nnz-1
            end if
        end do
        deallocate(ind,ranking)
              
        call allocate_sparse(D,nnz)
        D%val = C%val(1:nnz)
        D%row = C%row(1:nnz)
        D%col = C%col(1:nnz)

!         !find out how many nonzero values after addition     
!         bothAB = 0
!         do i=1,A%nnz
!             if (any(A%row(i)==B%row .AND. A%col(i)==B%col)) bothAB = bothAB+1
!         end do
!         call allocate_sparse(C, A%nnz + B%nnz - bothAB)       
! 
!         !perform addition
!         k = 1
!         !just A
!         do i=1,A%nnz
!             if (any(A%row(i)==B%row .AND. A%col(i)==B%col)) cycle
!             C%val(k) = A%val(i)
!             C%row(k) = A%row(i)
!             C%col(k) = A%col(i)
!             k = k + 1
!         end do
!         !just B
!         do j=1,B%nnz                
!             if (any(B%row(j)==A%row .AND. B%col(j)==A%col)) cycle
!             C%val(k) = B%val(j)
!             C%row(k) = B%row(j)
!             C%col(k) = B%col(j)                   
!             k = k + 1
!         end do
!         !both A and B
!         do i = 1,A%nnz
!             do j = 1,B%nnz
!                 if (A%row(i)==B%row(j) .AND. A%col(i)==B%col(j)) then
!                         C%val(k) = A%val(i) + B%val(j)
!                         C%row(k) = A%row(i)
!                         C%col(k) = A%col(i)
!                         k = k + 1
!                 end if    
!             end do
!         end do        
          
    end function plus
    
    !C = A*b where b is a scalar
    function times(A,b) result(C)
        implicit none
        type(sparse_mat), intent(in) :: A
        real(8), intent(in) :: b
        type(sparse_mat) :: C
        
        call allocate_sparse(C,A%nnz)
        C%val= b*A%val
        C%row = A%row
        C%col = A%col
        
    end function times 
    
    !c=A*b for b and c a vector and A square
    function spmtimes(A,b) result(c)
        implicit none
        type(sparse_mat), intent(in) :: A
        real(8), dimension(1:), intent(in) :: b
        real(8), dimension(1:size(b)) :: c
        !local
        integer :: i,j,k
!         real(8), dimension(1:size(b)) :: c_tmp
!         !THE BELOW WORKS BUT IS INEFFICIENT WITH MEMORY
!         !c = A*b
!         !c_j  = sum_i A_ij b_j
!         c_tmp = 0.d0
!         do k = 1,A%nnz
!             if (A%val(k)==0.d0) cycle
!             i = A%row(k)
!             j = A%col(k)
!             c_tmp(i) = c_tmp(i) + A%val(k)*b(j)
!         end do
!         c = c_tmp

        !THE MORE EFFICIENT MEMORY VERSION
        call mkl_dcoogemv('N', size(b), A%val, A%row, A%col, A%nnz, b, c)
       
    end function spmtimes
    
    subroutine write_sparse(A)
        implicit none
        type(sparse_mat) :: A
        integer :: i
        do i = 1,A%nnz
            write(*,*) A%val(i),A%row(i),A%col(i)
        end do
    end subroutine write_sparse
    
    function transpose_sparse(A) result(B)
        implicit none
        type(sparse_mat), intent(in) :: A
        type(sparse_mat) :: B
        
        call allocate_sparse(B,A%nnz)
        B%val = A%val
        B%row = A%col
        B%col = A%row
        
    end function transpose_sparse  
    
    !SPARSE Create sparse matrix.
    !    S = SPARSE(X) converts a sparse or full matrix to sparse form by
    !    squeezing out any zero elements.    
    function sparse(A) result(B)
        implicit none
        real(8), dimension(1:,1:), intent(in) :: A
        type(sparse_mat) :: B
        !local
        integer :: i,j,k
        
        call allocate_sparse(B,count(A/=0.d0))
        
        k=1
        do i = 1,size(A,1)
            do j = 1,size(A,2)
                if (A(i,j)/=0.d0) then
                    B%val(k) = A(i,j)
                    B%row(k) = i
                    B%col(k) = j
                    k=k+1
                end if
            end do
        end do        
        
    end function sparse
    
    ! Take a sparse mat A and convert to full B with dimensions m,n
    function full(A,m,n) result(B)
        implicit none
        type(sparse_mat), intent(in) :: A
        integer, intent(in) :: m,n
        real(8), dimension(1:m,1:n) :: B
        !local
        integer :: k
        
        if (maxval(A%row)>m .OR. maxval(A%col)>n) STOP 'full: m and n not compatible with A and B'
        
        B = 0.d0
        do k = 1,A%nnz
            B(A%row(k),A%col(k)) = A%val(k)
        end do
        
    end function full
    
end module mod_sparse2

