module mod_stat
    use mod_types
    implicit none

    real(rt), parameter :: INVAL_STAT = -sqrt(huge(0._rt))
    integer, parameter :: LENTTL_STAT = 16
    type mcstats_type
        real(rt) :: mean=INVAL_STAT
        real(rt) :: var=INVAL_STAT
        real(rt) :: std=INVAL_STAT
        real(rt) :: min=INVAL_STAT
        real(rt) :: median=INVAL_STAT
        real(rt) :: max=INVAL_STAT
        real(rt) :: gini=INVAL_STAT
        real(rt) :: pop=INVAL_STAT
        character(len=LENTTL_STAT) :: rvname=' '
    end type
contains

    ! Computes the mean and variance-covariance matrix of some data x(j,i) where i gives the observation and j gives the variable
    subroutine sub_varcov(mean,varcov,d,obsstride)
        implicit none
        real(8), intent(in) :: d(:,:)
        real(8), intent(out) :: mean(:),varcov(:,:)
        integer, intent(in), optional :: obsstride ! if present, observations will be 1,1+obsstride,1+2*obsstride,etc
        ! local 
        integer :: i,j,ni,nj,j2
        integer :: istr

        nj = size(d,1)

        istr = 1
        if (present(obsstride)) istr = obsstride

        if (size(mean)/=nj) stop 'ERROR: sub_varcov: mean has the wrong size'
        if (any(shape(varcov)/=[nj,nj])) stop 'ERROR: sub_varcov: varcov has the wrong size'

        mean = 0d0
        ni = 0
        do i = 1,size(d,2),istr
            mean = mean + d(:,i)
            ni = ni + 1
        end do
        mean = mean/ni

        varcov = 0d0
        do i = 1,size(d,2),istr
            do j = 1,nj
                do j2 = j,nj
                    varcov(j,j2) = varcov(j,j2) + (d(j,i)-mean(j))*(d(j2,i)-mean(j2))/ni
                end do
            end do
        end do
        
        ! Now the variance-covariance upper triangular is set
        ! Do the transpose to fill in the lower triangular
        do j = 1,nj
            do j2 = 1,j-1
                varcov(j,j2) = varcov(j2,j)
            end do
        end do

    end subroutine

function statslct(statname,rvname,stats) result(s)
    implicit none
    real(rt) :: s
    type(mcstats_type), intent(in) :: stats(:)
    character(len=*), intent(in) :: statname
    character(len=*), intent(in) :: rvname
    ! 
    integer :: irv,nrvs,flag

    nrvs = size(stats)

    s = INVAL_STAT

    do irv = 1,nrvs
        if (len(rvname)<=LENTTL_STAT) then
            if (rvname==stats(irv)%rvname) then
                exit
            end if
        else 
            if (rvname(1:LENTTL_STAT)==stats(irv)%rvname) then
                exit
            end if
        end if
    end do
    
    if (irv==nrvs+1) then
        write(0,*) 'WARNING: statslct: did not find random variable ',trim(rvname)
        return 
    end if

    flag = -1
    if (statname=='mean') then; s = stats(irv)%mean; flag=0; end if
    if (statname=='std') then; s = stats(irv)%std; flag=0;  end if
    if (statname=='var') then; s = stats(irv)%var; flag=0;  end if
    if (statname=='min') then; s = stats(irv)%min; flag=0;  end if
    if (statname=='max') then; s = stats(irv)%max; flag=0;  end if
    if (statname=='median') then; s = stats(irv)%median; flag=0; end if
    if (statname=='gini') then; s = stats(irv)%gini; flag=0; end if
    if (statname=='pop') then; s = stats(irv)%pop; flag=0;  end if
    
    if (flag==-1) then
        write(0,*) 'WARNING: statslct: did not find statistic by name of',statname
        return 
    end if

end function 

subroutine disp_mcstats(stats)
    implicit none
    type(mcstats_type), intent(in) :: stats(:)
    character(len=*), parameter :: sfmt = '(A16)'
    character(len=*), parameter :: rfmt = '(f16.6)'
    character(len=16) :: ttl
    integer :: irvs,nrvs,row
    real(rt), allocatable :: s(:)

    nrvs = size(stats)

    allocate(s(nrvs))

    write(*,sfmt,advance='no') ' '
    do irvs = 1,nrvs
        write(*,sfmt,advance='no') trim(adjustl(stats(irvs)%rvname))
    end do
    write(*,*) ' '

    do row = 1,8
        select case (row)
            case (1); s = stats%mean;   ttl = 'mean'
            case (2); s = stats%std;    ttl = 'std'
            case (3); s = stats%var;    ttl = 'var'
            case (4); s = stats%min;    ttl = 'min'
            case (5); s = stats%max;    ttl = 'max'
            case (6); s = stats%median; ttl = 'median'
            case (7); s = stats%gini;   ttl = 'gini'
            case (8); s = stats%pop;    ttl = 'pop'
        end select
        write(*,sfmt,advance='no') ttl
        do irvs = 1,nrvs
            write(*,rfmt,advance='no') s(irvs)
        end do
        write(*,*) ' '
    end do

end subroutine 

! Computes stats from an invariant distribution
subroutine stats_from_invar(stats,dsn,rvs,ndsn,nrvs,title)
    use mod_matlab, only: sort2
    implicit none
    integer, intent(in) :: ndsn,nrvs
    type(mcstats_type), intent(out) :: stats(:) ! stats(nrvs)
    real(rt), intent(in) :: dsn(*), rvs(ndsn,*) ! dsn(ndsn), rvs(ndsn,nrvs)
    character(len=*), intent(in), optional :: title(:) ! title(nrvs)
    ! local
    real(rt), allocatable :: y(:),p(:)
    integer, allocatable :: order(:)
    real(rt) :: x,giniSl,giniSu,Ey,Ey2,mass,sumdsn,minv,maxv,mass_so_far,giniG
    integer :: i,irv,strd

    if (nrvs/=size(stats)) then
        write(0,*) 'ERROR: mcstats_from_invar: stats has the wrong size ',size(stats),' relative to nrvs ',nrvs
        stop 
    end if
    if (present(title)) then
        if (nrvs/=size(title)) then
            write(0,*) 'ERROR: mcstats_from_invar: stats has the wrong size ',size(stats),' relative to nrvs ',nrvs
            stop 
        end if
        do irv = 1,nrvs
            stats(irv)%rvname = title(irv)
        end do
    end if

    allocate(y(ndsn))
    allocate(p(ndsn))
    allocate(order(ndsn))

    ! Get the sum of the dsn
    sumdsn = sum(dsn(1:ndsn))

    ! If no mass, return (conditional dsn not well-defined)
    if (sumdsn==0d0) return

    ! Now have random sample, mcinds
    do irv = 1,nrvs

        ! For each random variable, compute the stats
        call sort2(y,order,rvs(:,irv))
        p = dsn(order)

        ! Compute the stats
        Ey = 0._rt
        Ey2 = 0._rt
        giniSu = 0._rt
        giniG = 0._rt
        minv = huge(0._rt)
        maxv = -huge(0._rt)
        mass_so_far = 0._rt

        i = 1
        do while (i<=ndsn)

            ! If y values are not strictly increasing, then make them
            ! If they are are strictly increasing, then order will just be 1,...,n
            ! If they are not, then it will jump
            mass = p(i) 
            strd = 1
            do while (i+strd<=ndsn .and. y(i)==y(min(ndsn,i+strd)))
                mass = mass + p(i+strd)
                strd = strd + 1
            end do

            ! Get mass from conditional distribution
            mass = mass/sumdsn

            Ey = Ey + y(i)*mass
            Ey2 = Ey2 + y(i)**2*mass

            if (mass>0._rt) then
                minv = min(minv,y(i))
                maxv = max(maxv,y(i))
            end if

            mass_so_far = mass_so_far + mass
            if (mass_so_far>=.5_rt) then

                ! this is median
                stats(irv)%median = y(i)

                ! ensure condition not satisfied again
                mass_so_far = -huge(0._rt)

            end if

            giniSl = giniSu
            giniSu = giniSl + y(i)*mass
            giniG = giniG + mass*(giniSu+giniSl)

            i = i + strd

        end do

        stats(irv)%mean = Ey
        stats(irv)%var = max(0._rt,Ey2 - stats(irv)%mean**2) ! in case = -epsilon
        stats(irv)%std = sqrt(stats(irv)%var)
        if (giniSu/=0._rt) then
            stats(irv)%gini = 1._rt - giniG/giniSu
        else
            ! print*,'y(1:ndsn)',y(1:ndsn)
            ! print*,'p(1:ndsn)',p(1:ndsn)
            print*,'giniG',giniG
            print*,'giniSU',giniSU
            print*,'invalid'
            stats(irv)%gini = INVAL_STAT
        end if  
        stats(irv)%min = minv
        stats(irv)%max = maxv
        stats(irv)%pop = sumdsn

    end do

end subroutine

! Computes stats from an invariant distribution by drawing randomly. This is probably not useful. 
subroutine mcstats_from_invar(stats,nsmpl,dsn,rvs,ndsn,nrvs)
    use m_refsor
    use mod_matlab
    implicit none
    integer, intent(in) :: nsmpl,ndsn,nrvs
    type(mcstats_type), intent(out) :: stats(:) ! stats(nrvs)
    real(rt), intent(in) :: dsn(*), rvs(ndsn,*) ! dsn(ndsn), rvs(ndsn,nrvs)
    ! local
    real(rt), dimension(:), allocatable :: cdsn,y
    integer, dimension(:), allocatable :: mcinds
    real(rt) :: x,giniSN,giniSD,Ey,Ey2
    integer :: i,irv

    if (nrvs/=size(stats)) then
        write(0,*) 'ERROR: mcstats_from_invar: stats has the wrong size ',size(stats),' relative to nrvs ',nrvs
        stop 
    end if

    ! Construct a random sample from the invariant distribution
    allocate(cdsn(0:ndsn))
    allocate(mcinds(nsmpl)) 

    ! Cumulatively sum the distribution
    cdsn(0) = 0._rt
    do i = 1,ndsn
        cdsn(i) = cdsn(i-1) + dsn(i)
    end do
!     if (abs(cdsn(n)-1d0)>sqrt(epsilon(0._rt))) then
!         write(0,*) 'ERROR: fn_gini_approx: dsn doesn''t sum to one but to ',cdsn(n),' trying to recover'
!     end if
!     cdsn(n) = 1._rt
    
    if (cdsn(ndsn)<10._rt*epsilon(0._rt)) then
        write(0,*) 'Distribtion summed to almost zero, returning invalid stats'
        return
    end if

    ! Random sample
    do i = 1,nsmpl

        call random_number(x)
        x = x * cdsn(ndsn) ! if dsn doesn't sum to one, then compute a draw conditional on the positive mass
        call bsearch(mcinds(i),x,cdsn)
        
    end do

    deallocate(cdsn)
    allocate(y(nsmpl))

    ! Now have random sample, mcinds
    do irv = 1,nrvs

        ! For each random variable, compute the stats
        y = rvs(mcinds,irv)

        ! Sort the rv
        call refsor(y) ! sorts y using quicksort

        ! Compute the stats
        Ey = 0._rt
        Ey2 = 0._rt
        giniSN = 0._rt
        giniSD = 0._rt
        do i = 1,nsmpl

            Ey = Ey + y(i)
            Ey2 = Ey2 + y(i)**2

            giniSN = giniSN + i*y(i)
            giniSD = giniSD + y(i) ! note this is just Ey

        end do

        stats(irv)%mean = Ey/nsmpl
        stats(irv)%var = max(0._rt,Ey2/nsmpl - stats(irv)%mean**2) ! in case = -epsilon
        stats(irv)%std = sqrt(stats(irv)%var)
        stats(irv)%gini = (2*giniSN)/(nsmpl*giniSD) - (nsmpl+1)/real(nsmpl,rt)
        stats(irv)%min = y(1) 
        stats(irv)%median = y((nsmpl+1)/2) 
        stats(irv)%max = y(nsmpl) 
        stats(irv)%pop = nsmpl

    end do

end subroutine


real(rt) function median(dsn,grid,n) result(m)
    use mod_matlab
    implicit none
    real(rt), intent(in) :: dsn(*), grid(*)
    integer, intent(in) :: n
    ! local
    real(rt), allocatable :: sgrid(:),sdsn(:)
    integer, allocatable :: redo(:) 
    integer :: i 

    allocate(sdsn(n),sgrid(n),redo(n))

    ! sort 
    call sort2(sgrid,redo,grid(1:n))
    sdsn = dsn(redo)

    ! cumsum the dsn
    sdsn = cumsum(sdsn)

    ! find the median via interpolation
    if (sdsn(1)>=.5d0) then
        m = sgrid(1)
    elseif (sdsn(n)<=.5d0) then
        m = sgrid(n)
    else
        m = interp1(sdsn,sgrid,.5d0)
    end if

end function

function R2(obs,pred)
    use mod_matlab, only: mean
    implicit none
    real(rt) :: R2
    real(rt), dimension(1:), intent(in) :: obs, pred
    !local
    real(rt) :: obs_mean

    !Check dims
    if (size(obs)/=size(pred)) then
        write(*,*) 'R2: dims don''t agree.  Obs', size(obs), ' Pred',size(pred)
        STOP 'R2: stopping.'
    end if

    obs_mean = mean(obs)
    r2 = 1 - sum((obs-pred)**2)/sum((obs-obs_mean)**2)

end function R2

end module mod_stat
