module mod_stata
    use mod_types
    implicit none


    integer, parameter :: STATA_MAX_VARS = 10000
    real(rt), parameter :: STATA_INVALID = huge(0._rt)/1e1_rt
    real(rt), parameter :: STATA_MISSING = huge(0._rt)/1e2_rt 

    integer, parameter :: STATA_NAME_LEN = 16
    ! Formats for displaying output to terminal (all fixed width with STATA_NAME_LEN)
    character(len=*), parameter :: STATA_F_FMT = '(f16.5)'
    character(len=*), parameter :: STATA_E_FMT = '(e16.7e3)'
    character(len=*), parameter :: STATA_S_FMT = '(A16)'
    character(len=*), parameter :: STATA_I_FMT = '(I16)'
    ! Formats for exporting data (no data loss, no extra spaces for lower memory)
    character(len=*), parameter :: STATA_Eneg_XFMT = '(ES24.16E3)' ! 
    character(len=*), parameter :: STATA_Epos_XFMT = '(ES23.16E3)' ! save a digit
    character(len=*), parameter :: STATA_I_XFMT = '(I10)'

    type stata_wkspace
        real(rt), allocatable :: d(:,:) ! data set (dims not necessarily nobs x ns)
        real(rt), allocatable :: w(:) ! weights (dims not necessarily nobs)

        integer :: ns = 0 ! number of series
        integer :: nobs = 0 ! number of obs

        character(STATA_NAME_LEN) :: snames(STATA_MAX_VARS) ! data series names

        ! Like in actual stata, we can include a return list
        real(rt)              :: e_r2 ! r2
        real(rt), allocatable :: e_b(:) ! regression coefficients
        real(rt), allocatable :: e_V(:,:) ! variance-covariance of regression coefficients

    contains 
        procedure, pass :: free => stata_free
        procedure, pass :: gen => stata_gen_assumedshape
        procedure, pass :: setweight => stata_setweight_assumedsize
        procedure, pass :: varlist => stata_varlist
        procedure, pass :: v => stata_v
        procedure, pass :: i => stata_i
        procedure, pass :: summarize => stata_summarize
        procedure, pass :: summarize_var => stata_summarize_var
        procedure, pass :: stdev => stata_stdev
        procedure, pass :: median => stata_median
        procedure, pass :: pctile => stata_pctile
        procedure, pass :: corr => stata_corr
        procedure, pass :: pwcorr => stata_pwcorr
        procedure, pass :: regress => stata_regress
        procedure, pass :: list => stata_list
        procedure, pass :: hist => stata_hist
        procedure, pass :: scatter => stata_scatter
        procedure, pass :: export => stata_export
    end type

    interface stata_gen
        module procedure stata_gen_assumedshape, stata_gen_assumedsize
    end interface
    interface stata_setweight
        module procedure stata_setweight_assumedshape, stata_setweight_assumedsize
    end interface


    private :: corr_core_stable,corr_core_fast,isvalid

contains

!     subroutine stata_set_obs()
!         implicit none 
! 
!     end subroutine 
! 
!     subroutine stata_set_data_pt(idata,ivar)
!         implicit none
!         integer, intent(in) :: idata
! 
!     end subroutine

    elemental function stata_log(x) result(y)
        implicit none
        real(rt), intent(in) :: x
        real(rt) :: y

        if (x>0._rt .and. x/=STATA_MISSING .and. x/=STATA_INVALID) then
            y = log(x)
        elseif (x==STATA_MISSING) then
            y = STATA_MISSING
        else
            y = STATA_INVALID
        end if

    end function 

    elemental function isvalid(x) result(tf)
        implicit none
        real(rt), intent(in) :: x
        logical :: tf
        tf = x/=STATA_INVALID .and. x/=STATA_MISSING
    end function

    subroutine stata_export(this,filename)
        implicit none
        class(stata_wkspace), intent(in) :: this
        character(len=*), intent(in), optional :: filename
        ! local
        integer :: col,row,uid

        uid = 6
        if (present(filename)) open(newunit=uid,file=filename,action='write',status='replace',err=999)

        ! Header row
        do col = 1,this%ns
            write(uid,'(A)',advance='no') trim(this%snames(col))
            ! write(uid,conv_i_to_s(len(trim(this%snames(col))))//')',advance='no') adjustr(this%snames(col))
            write(uid,'(A1)',advance='no') ','
        end do
        ! Add weight
        write(uid,'(A)') 'weight'

        ! The rest
        do row = 1,this%nobs
            ! Check whether it can be stored as an integer
            call disp_stata_value(uid,this%d(row,1),ishighprec=.true.,dotryint=.true.)
            do col = 2,this%ns
                write(uid,'(A1)',advance='no') ','
                call disp_stata_value(uid,this%d(row,col),ishighprec=.true.,dotryint=.true.)
            end do
            write(uid,'(A1)',advance='no') ','
            call disp_stata_value(uid,this%w(row),ishighprec=.true.,dotryint=.true.)
            write(uid,*)
        end do

        if (present(filename)) close(uid)

        return

999     write(0,*) 'ERROR: could not export ',filename,', encountered error while opening file.'

    end subroutine
    
    ! Lists the data series
    subroutine stata_list(this)
        implicit none
        class(stata_wkspace), intent(inout) :: this
        ! local
        integer :: row,col

        do col = 1,this%ns
            write(*,'(A)',advance='no') adjustr(this%snames(col))
        end do
        write(*,*)

        do row = 1,this%nobs
            do col = 1,this%ns
                call disp_stata_value(6,this%d(row,col))
            end do
            write(*,*)
        end do

    end subroutine

    subroutine stata_scatter(this,name1,name2,filename)
        use mod_plot
        implicit none
        class(stata_wkspace), intent(in) :: this
        character(len=*), intent(in) :: name1,name2,filename
        ! local
        logical, allocatable :: incl(:)
        integer :: uid,i1,i2

        call clear()

        uid = initplot(filename)
        call add_gnuplot_cmd(uid,'set style data points')

        i1 = this%i(name1)
        i2 = this%i(name2)

        allocate(incl(this%nobs))
        incl =     this%d(:,i1)/=STATA_INVALID .and. this%d(:,i1)/=STATA_MISSING &
             .and. this%d(:,i2)/=STATA_INVALID .and. this%d(:,i2)/=STATA_MISSING

        call triplet(pack(this%d(:,i1),mask=incl),&
                     pack(this%d(:,i2),mask=incl),'')

        call xlabel(uid,trim(this%snames(i1)))
        call ylabel(uid,trim(this%snames(i2)))
        call title(uid,'Scatter plot of '//trim(this%snames(i1))//' vs '//trim(this%snames(i2)))
        call drawnow(uid)
        if (.not. disable_plotting) write(*,*) 'Plot written to '//filename
        call clear()

    end subroutine

    ! Adds the data series "x" to the workspace labeling it using "name"
    subroutine stata_gen_assumedsize(this,x,name,nx)
        implicit none
        class(stata_wkspace), intent(inout) :: this
        real(rt), intent(in) :: x(*)
        character(len=*), intent(in) :: name
        integer, intent(in) :: nx
        call stata_gen_assumedshape(this,x(1:nx),name)
    end subroutine
    subroutine stata_gen_assumedshape(this,x,name)
        implicit none
        class(stata_wkspace), intent(inout) :: this
        real(rt), intent(in) :: x(:)
        character(len=*), intent(in) :: name
        ! local
        real(rt), allocatable :: dold(:,:)

        if (.not. allocated(this%d)) then

            this%nobs = size(x)
            this%ns = 1 

            ! Add the series 
            allocate(this%d(this%nobs,1))
            this%d(:,1) = x
            this%snames(1) = name

            if (.not. allocated(this%w)) then
                allocate(this%w(this%nobs))
                this%w = 1._rt/this%nobs
            end if

        else

            ! Check that the size agrees
            if (this%nobs/=size(x)) then
                write(0,*) 'ERROR: stata: gen: while generating ', name,' found size ', size(x), ' was inconsistent with number of observations ',this%nobs
                stop 
            end if

            ! See whether there is room
            if (this%ns==size(this%d,2)) then

                ! No: need to expand the data set

                allocate(dold(this%nobs,this%ns))

                dold = this%d

                deallocate(this%d)

                allocate(this%d(this%nobs,this%ns*2)) ! double the possible number of series so need not keep doing this

                this%d(:,1:this%ns) = dold

            end if

            ! Now for sure there is room
            this%ns = this%ns + 1
            this%d(:,this%ns) = x
            this%snames(this%ns) = adjustl(name)

        end if

    end subroutine

    ! Sets weights to be used in summarizing, etc
    subroutine stata_setweight_assumedsize(this,w,nw)
        implicit none
        class(stata_wkspace), intent(inout) :: this
        real(rt), intent(in) :: w(*)
        integer, intent(in) :: nw
        call stata_setweight_assumedshape(this,w(1:nw))
    end subroutine
    subroutine stata_setweight_assumedshape(this,w)
        implicit none
        class(stata_wkspace), intent(inout) :: this
        real(rt), intent(in) :: w(:)

        if (.not. allocated(this%d)) then
            write(0,*) 'WARNING: use gen to add a series before calling setweight. Weight has not been set.'
            return
        end if 

        if (this%nobs/=size(w)) then
            write(0,*) 'WARNING: weight size ',size(w),' does not agree with nobs ',this%nobs,'. Weight has not been set.'   
            return
        end if

        if (.not. allocated(this%w)) allocate(this%w(this%nobs)) 

        this%w = w

    end subroutine

    ! Produces a copy of the variable "name"
    function stata_v(this,name) result(x)
        implicit none
        class(stata_wkspace), intent(inout) :: this
        character(len=*), intent(in) :: name
        real(rt) :: x(this%nobs) 
        ! local
        integer :: i

        i = stata_i(this,name)

        if (i>0 .and. i<=this%ns) then
            x = this%d(:,i)
        else
            x = STATA_INVALID
        end if

    end function 

    ! Returns the data series index of the variable "name"
    function stata_i(this,name) result(i)
        implicit none
        class(stata_wkspace), intent(in) :: this
        character(len=*), intent(in) :: name 
        integer :: i
        ! local
        character(len=STATA_NAME_LEN) :: namecopy
        integer :: is

        namecopy = ''
        namecopy = name

        do is = 1,this%ns
            if (namecopy==this%snames(is)) then
                i = is
                return
            end if
        end do

        write(0,*) 'WARNING: stata: variable ',namecopy,' not found.'
        write(0,*) 'Existings names are as follows:'
        do is = 1,this%ns
            write(0,*) this%snames(is)
        end do

        i = -1
        return

    end function

    ! varlist: returns a logical vector corresponding to whether the series are selected or not
    ! if names is just length 1 and equal to "*", then all series are returned.
    function stata_varlist(this,namelist) result(tf)
        use mod_util
        implicit none
        class(stata_wkspace), intent(in) :: this
        character(*), intent(in) :: namelist
        logical, dimension(this%ns) :: tf
        ! local
        character(STATA_NAME_LEN), allocatable :: names(:)
        integer :: i, nvars, ind

        nvars = count_delim_arg(namelist,',') 

        allocate(names(nvars))
        do i = 1,nvars
            names(i) = adjustl(get_delim_arg(namelist,',',i))
        end do

        if (size(names)==1 .and. trim(adjustl(names(1)))=="*") then
            tf = .true.
        else
            tf = .false.
            do i = 1,size(names)
                ind = this%i(names(i))
                if (ind>=1 .and. ind<=this%ns) then
                    tf(ind) = .true.
                else
                    write(0,*) 'ERROR: stata_varlist: could not find requested variable '//names(i)//' in workspace'
                    stop
                end if
            end do
        end if

    end function

    subroutine stata_hist(this,name,filename)
        use mod_matlab, only: linspace,plot,bsearch
        implicit none
        class(stata_wkspace), intent(in) :: this
        character(len=*), intent(in) :: name
        character(len=*), intent(in), optional :: filename ! where to save the plot (if desired)
        ! local
        real(rt), allocatable :: cnt(:), edge(:), table(:,:)
        integer :: ivar,iobs,nbins,ibin

        ! Get the index of the variable
        ivar = this%i(name)

        ! Summarize to the max,min
        allocate(table(this%ns,5)) 
        call this%summarize(table)

        ! Compute the mass
        nbins = 100
        allocate(cnt(nbins))
        allocate(edge(nbins+1))

        edge = linspace(table(ivar,4),table(ivar,5),nbins+1)

        cnt = 0d0
        do iobs = 1,this%nobs

            if (this%d(iobs,ivar)==STATA_INVALID .or. this%d(iobs,ivar)==STATA_MISSING) cycle

            call bsearch(ibin,this%d(iobs,ivar),edge)

            if (allocated(this%w)) then
                cnt(ibin) = cnt(ibin) + this%w(iobs)
            else
                cnt(ibin) = cnt(ibin) + 1d0
            end if

        end do

        ! Plot this hist
        if (.not. present(filename)) then
            call plot(edge(1:nbins),cnt,'Histogram of '//name)
        else
            call plot(edge(1:nbins),cnt,'Histogram of '//name,fname_=filename)
        end if

        
    end subroutine 

    function stata_stdev(this,varname) result(stdev)
        implicit none
        class(stata_wkspace), intent(in) :: this
        character(*), intent(in) :: varname
        real(rt) :: stdev
        ! local
        real(rt) :: std

        call stata_summarize_var(this,varname,std=stdev)

    end function 

    ! Summarizes the data giving nobs, mean, stdev, min, and max
    subroutine stata_summarize_var(this,varname,mean,std,min,max,mass)
        implicit none
        class(stata_wkspace), intent(in) :: this
        character(*), intent(in) :: varname
        real(rt), intent(out), optional :: mean,std,min,max,mass
        ! local
        real(rt) :: stats(5),w
        integer :: is,iobs

        stats(1) = 0._rt ! nobs
        stats(2) = 0._rt ! mean
        stats(3) = 0._rt ! std
        stats(4) = huge(0._rt) ! min
        stats(5) = -huge(0._rt) ! max

        is = this%i(varname)
        
        if (is<1 .or. is>this%ns) return

        do iobs = 1,this%nobs

            ! Check if missing or invalid
            if (this%d(iobs,is)==STATA_INVALID .or. this%d(iobs,is)==STATA_MISSING) cycle

            if (allocated(this%w)) then
                w = this%w(iobs)
                if (w==0._rt) cycle
            else
                w = 1._rt
            end if

            ! If here, valid entry
            stats(1) = stats(1) + w
            stats(2) = stats(2) + this%d(iobs,is)*w
            stats(3) = stats(3) + this%d(iobs,is)**2*w
            if (stats(4)<this%d(iobs,is)) stats(4) = this%d(iobs,is) ! min
            if (stats(5)>this%d(iobs,is)) stats(5) = this%d(iobs,is) ! max
            
        end do
        
        ! Now summarize
        if (stats(1)==0._rt) then
            stats(2:5) = STATA_INVALID
        else
            stats(2) = stats(2)/stats(1) ! mean
            stats(3) = stats(3)/stats(1) - stats(2)**2 ! var = Ex^2-(Ex)^2
            if (stats(3)>=-epsilon(0._rt)) stats(3) = sqrt(abs(stats(3))) 
        end if

        if (present(mass)) mass = stats(1)
        if (present(mean)) mean = stats(2)
        if (present(std)) std = stats(3)
        if (present(min)) min = stats(4)
        if (present(max)) max = stats(5)

    end subroutine


    ! Summarizes the data giving nobs, mean, stdev, min, and max
    ! If "table" is present, will place the statistics in order into it.
    ! If "filename" is present, will write the resulting table to file.
    ! If neither is true, the result is written to screen.
    ! Note that each row corresponds to a data series, so table should be nseries x 4
    subroutine stata_summarize(this,table,filename)
        implicit none
        class(stata_wkspace), intent(in) :: this
        real(rt), intent(out), optional :: table(:,:)
        character(len=*), intent(in), optional :: filename
        ! local
        real(rt), allocatable :: tbl(:,:)
        real(rt) :: w
        integer :: is,iobs,col,row,uid

        allocate(tbl(this%ns,5)) 
        tbl(:,1) = 0._rt ! nobs
        tbl(:,2) = 0._rt ! mean
        tbl(:,3) = 0._rt ! std
        tbl(:,4) = huge(0._rt) ! min
        tbl(:,5) = -huge(0._rt) ! max

        do is = 1,this%ns
            do iobs = 1,this%nobs

                ! Check if missing or invalid
                if (this%d(iobs,is)==STATA_INVALID .or. this%d(iobs,is)==STATA_MISSING) then
                    cycle
                end if

                if (allocated(this%w)) then
                    w = this%w(iobs)
                    if (w==0._rt) cycle
                else
                    w = 1._rt
                end if

                ! If here, valid entry
                tbl(is,1) = tbl(is,1) + 1._rt*w
                tbl(is,2) = tbl(is,2) + this%d(iobs,is)*w
                tbl(is,3) = tbl(is,3) + this%d(iobs,is)**2*w
                tbl(is,4) = min(tbl(is,4),this%d(iobs,is))
                tbl(is,5) = max(tbl(is,5),this%d(iobs,is))
                
            end do

            if (tbl(is,1)==0._rt) then
                tbl(is,2:5) = STATA_INVALID
            else
                tbl(is,2) = tbl(is,2)/tbl(is,1) ! mean
                tbl(is,3) = tbl(is,3)/tbl(is,1) - tbl(is,2)**2 ! var = Ex^2-(Ex)^2
                if (tbl(is,3)>=-epsilon(0._rt)) tbl(is,3) = sqrt(abs(tbl(is,3)))
            end if

        end do

        ! Display, save, or output?
        if (present(table)) then
            ! Check dims are correct
            if (size(table,1)/=this%ns .or. size(table,2)/=5) then
                write(0,*) 'WARNING: table is incorrectly sized: ',shape(table),' Should be ',this%ns,5
                table = -1
                return
            end if

            table = tbl
            return

        end if

        uid = 6
        if (present(filename)) open(newunit=uid,file=filename,action='write')

        write(uid,STATA_S_FMT,advance='no') 'variable'; write(uid,'(A1)',advance='no') ','
        write(uid,STATA_S_FMT,advance='no') 'nobs'    ; write(uid,'(A1)',advance='no') ','
        write(uid,STATA_S_FMT,advance='no') 'mean'    ; write(uid,'(A1)',advance='no') ','
        write(uid,STATA_S_FMT,advance='no') 'stdev'   ; write(uid,'(A1)',advance='no') ','
        write(uid,STATA_S_FMT,advance='no') 'min'     ; write(uid,'(A1)',advance='no') ','
        write(uid,STATA_S_FMT,advance='no') 'max'
        write(uid,*)
        do row = 1,this%ns
            write(uid,STATA_S_FMT,advance='no') adjustr(this%snames(row))
            do col = 1,5
                write(uid,'(A1)',advance='no') ','
                call disp_stata_value(uid,tbl(row,col),col==1 .and. .not. allocated(this%w))
            end do
            write(uid,*)
        end do

        if (present(filename)) close(uid)

    end subroutine


    ! Computes the median of a given varlist
    ! Uses an interpolation based formula, interpolating the cdf to get a value 
    function stata_median(this,varname,method) result(med) 
        implicit none
        class(stata_wkspace), intent(inout) :: this
        character(len=*), intent(in) :: varname
        integer, intent(in), optional :: method
        real(rt) :: med
        ! local
        real(rt) :: median(1)

        median = this%pctile(varname,[50d0])
        med = median(1)

    end function 

    ! pctile : computes the percentiles (should be in 0,100) and returns the values corresponding to them
    function stata_pctile(this,varname,pct) result(c)
        use mod_util
        use mod_matlab
        implicit none
        class(stata_wkspace), intent(inout) :: this
        character(*), intent(in) :: varname
        real(rt), intent(in) :: pct(:)
        real(rt) :: c(size(pct))
        ! local
        real(rt), allocatable :: x(:),cdf(:),tmp(:)
        integer, allocatable :: I(:)
        logical, allocatable :: mask(:)
        real(rt) :: mlo
        integer :: n,is,ilo

        is = this%i(varname)

        call assert(is>=1 .and. is<=this%ns,'stata_median: could not find variable '//varname)

!         methodslct = 1
!         if (present(method)) methodslct = method
!         call assert(methodslct>=1 .and. methodslct<=2,'stata_median: selected method must be 1 or 2 but was ',methodslct

        allocate(mask(this%nobs))

        mask = this%d(:,is)/=STATA_MISSING .and. & 
               this%d(:,is)/=STATA_INVALID .and. & 
               this%w>0._rt

        n = count(mask)

        allocate(x(n),cdf(n),I(n),tmp(n))

        x = pack(this%d(:,is),mask)

        if (maxval(this%w,mask)==minval(this%w,mask)) then

            ! If the observations are uniformly weighted, just sort the list and check the middle
            call sort2(tmp,I,x)
            x = tmp

            c = x(min(size(x),max(1,nint(size(x)*pct/100d0))))
            ! if (mod(size(x),2)==0) then
            !     med = .5d0*sum(x(size(x)/2:size(x)/2+1))
            ! else
            !     med = x(size(x)/2+1)
            ! end if

        else

            ! If the observations are not uniformly weighted, need to interpolate cdf
            cdf = pack(this%w,mask)

            ! Now, need to sort
            call sort2(tmp,I,x)
            x = x(I)
            cdf = cdf(I)
            call assert(all(tmp==x),'stata_median: error in sorting')

            ! Now, cumsum to get actual cdf
            tmp = cdf/sum(cdf)
            cdf = cumsum(tmp)

            ! Now, find median (should allow user to select desired formula, but here I am forcing an interpolation formula)
            c = interp1(cdf,x,pct/100d0)
            ! call noexinterp1qlocweight(ilo,mlo,.5d0,cdf)
            ! med = x(ilo)*mlo + x(ilo+1)*(1d0-mlo)

        end if
        
    end function 

    ! Regresses
    ! 
    subroutine stata_regress(this,depvar,indepvarlist,nocons,nodisp)
        use mod_matlab
        implicit none
        class(stata_wkspace), intent(inout) :: this
        character(*), intent(in) :: depvar
        logical, intent(in) :: indepvarlist(:)
        logical, intent(in), optional :: nocons,nodisp
        ! local
        integer :: idep,i,j
        real(rt), allocatable :: y(:),X(:,:),XPX(:,:),XPy(:),w(:)
        real(rt), allocatable :: b(:),b_stderr(:),varcov(:,:),res(:),II(:,:),tab(:,:)
        real(rt) :: s2,RSS,YSS,Rsq
!         real(rt), allocatable :: t(:,:),w(:)
!         logical, allocatable :: mask(:)
!         logical :: totable, toval, toscreen
        character(STATA_NAME_LEN), allocatable :: row(:),col(:)
        integer :: nindep
        logical :: hascons,dodisp

        hascons = .true.
        if (present(nocons)) hascons = .not. nocons

        nindep = count(indepvarlist) + merge(1,0,hascons)

        ! Error chcek
        idep = this%i(depvar)
        if (idep<1 .or. idep>this%ns) then
            write(0,*) 'ERROR: stata_regress: depvar ',depvar,' not found'
            stop
        end if
        if (size(indepvarlist)/=this%ns) stop 'ERROR: stata_regress: indepvarlist has the wrong size' 

!         toscreen = .not. present(table) .and. .not. present(val)
!         totable = present(table)
!         toval = present(val)

        allocate(y(this%nobs))
        allocate(X(this%nobs,nindep))
        allocate(w(this%nobs))
        allocate(XPX(nindep,nindep))
        allocate(XPy(nindep))
        allocate(b(nindep))
        allocate(res(this%nobs))
        allocate(b_stderr(nindep))
        allocate(varcov(nindep,nindep))
        allocate(II(nindep,nindep))
        allocate(tab(nindep,2))
        allocate(row(size(tab,1)))
        allocate(col(size(tab,2)))

        w = this%w
        where (this%d(:,idep)==STATA_INVALID .or. this%d(:,idep)==STATA_MISSING) w = 0._rt
        do i = 1,this%ns
            if (indepvarlist(i)) where(this%d(:,i)==STATA_INVALID .or. this%d(:,i)==STATA_MISSING) w = 0._rt
        end do

        ! Compute conditional
        w = w/sum(w)
!         where (w>0._rt) w = 1._rt

        y = sqrt(w)*this%d(:,idep)
        if (hascons) then
            j = 1
            X(:,1) = sqrt(w)*1._rt
            row(1) = '_cons'
        else
            j = 0
        end if
        do i = 1,this%ns
            if (indepvarlist(i)) then
                j = j+1
                X(:,j) = sqrt(w)*this%d(:,i)
                row(j) = this%snames(i)
            end if
        end do

        ! Compute X'X
        call mygemm(XPX,X,X,this%nobs,nindep,this%nobs,nindep,'T','N')

        ! Compute X'y
        call mygemv(XPy,X,y,this%nobs,nindep,'T')

        ! Compute b = (X'X)^-1*(X'y)
        b = mldivide(XPX,XPy)
        where (abs(b)<10._rt*epsilon(0._rt)) b = 0._rt

        ! Compute the residuals
        res = y-matmul(X,b)

        ! Variance estimate
        call stats_core_stable(res,this%w,var=RSS)
        ! RSS = dot_product(res,res)
        s2 = RSS/(count(w>0._rt)-nindep) 

        ! Variance-covariance is V = s2*(XPX)^{-1} => V * XPX = s2 * I => XPX'*V' = s2*I
        II = eye(nindep)
        II = s2*II
        varcov = transpose(mldivide(transpose(XPX),transpose(II)))
        where (abs(varcov)<10._rt*epsilon(0._rt)) varcov = 0._rt
        b_stderr = sqrt(diag(varcov))

        ! Compute R-squared
        call stats_core_stable(y,this%w,var=YSS)
        ! YSS = sum(this%w*(y-sum(y,mask=this%w>0._rt)/sum(this%w))**2)/sum(this%w)
        Rsq = 1._rt-RSS/YSS

!         call disp('y',y)
!         call disp('X',X)
!         call disp('w',w)
!         call disp('RSS',RSS)
!         call disp('YSS',YSS)
!         call disp('Rsq',Rsq)
!         call disp('s2',s2)
!         call disp('XPX',XPX)
!         call disp('XPy',XPy)
!         call disp('b',b)
!         call disp('varcov',varcov)
!         call disp('b_stderr',b_stderr)

        ! Output values
        this%e_r2 = Rsq
        if (allocated(this%e_b)) deallocate(this%e_b); allocate(this%e_b(nindep)); this%e_b = b
        if (allocated(this%e_V)) deallocate(this%e_V); allocate(this%e_V(nindep,nindep)); this%e_V = varcov

        ! Table
        dodisp = .true.
        if (present(nodisp)) dodisp = .not. nodisp

        if (dodisp) then
            tab(:,1) = b
            tab(:,2) = b_stderr

            col(1) = 'Coef.'
            col(2) = 'Std. err.'
            
            write(6,STATA_S_FMT,advance='no') ' '
            write(6,'(A1)',advance='no') ','
            write(6,STATA_S_FMT,advance='no') 'R-squared'
            write(6,'(A1)',advance='no') ','
            call disp_stata_value(6,Rsq)
            write(6,*) 
            write(6,STATA_S_FMT,advance='no') ' '
            write(6,'(A1)',advance='no') ','
            write(6,STATA_S_FMT,advance='no') 'Number of obs'
            write(6,'(A1)',advance='no') ','
            call disp_stata_value(6,real(count(w>0._rt),rt))
            write(6,*) 
            write(6,*) 
            call stata_table(6,tab,row,col)
        end if

    end subroutine

    
    subroutine stata_table(uid,tab,row,col)
        implicit none
        integer, intent(in) :: uid
        real(rt), intent(in) :: tab(:,:)
        character(STATA_NAME_LEN), intent(in), optional :: row(:),col(:)
        integer :: i,j

        if (present(col)) then
            if (present(row)) call addentry()
            do i = 1,size(col)
                call addentry(s=adjustr(col(i)),suppress=i==size(col))
            end do
            call newline()
        end if

        do i = 1,size(tab,1) 

            if (present(row)) call addentry(s=adjustr(row(i)))
            do j = 1,size(tab,2)
                call addentry(r=tab(i,j),suppress=j==size(tab,2))
            end do
            call newline()

        end do

    contains
        subroutine addentry(r,s,suppress)
            implicit none
            real(rt), intent(in), optional :: r
            character(*), intent(in), optional :: s
            logical, intent(in), optional :: suppress ! whether trailing comma should be suppressed
            if (.not. present(r) .and. .not. present(s)) then
                write(uid,STATA_S_FMT,advance='no') ' '
            elseif (present(r)) then
                call disp_stata_value(uid,r)
            elseif (present(s)) then
                write(uid,STATA_S_FMT,advance='no') s
            end if

            if (present(suppress)) then
                if (suppress) return
            end if

            write(uid,'(A1)',advance='no') ','

        end subroutine 
        subroutine newline()
            implicit none
            write(uid,*) 
        end subroutine
    end subroutine 


    ! Correlates variable 1 and variable 2
    function stata_pwcorr(this,v1,v2) result(c)
        implicit none
        class(stata_wkspace), intent(inout) :: this
        character(*), intent(in) :: v1,v2
        real(rt) :: c 
        ! local
        integer :: i1,i2

        i1 = this%i(v1)
        i2 = this%i(v2)

        c = STATA_INVALID 

        if (i1<1 .or. i1>this%ns) then
            write(0,*) 'WARNING: pwcorr: invalid variable name '//v1
            return
        end if
        if (i2<1 .or. i2>this%ns) then
            write(0,*) 'WARNING: pwcorr: invalid variable name '//v2
            return
        end if
        
        c = corr_core_stable(this%d(:,i1),this%d(:,i2),this%w,this%nobs)

    end function 



    ! Correlates the varlists varlist1 and varlist2, both integer vectors
    ! 
    ! The output is either to screen, to table, or to val.
    ! If to val, then varlist1 and varlist2 must be singletons
    ! 
    subroutine stata_corr(this,varlist1,varlist2,table,val)
        implicit none
        class(stata_wkspace), intent(inout) :: this
        logical, intent(in) :: varlist1(:),varlist2(:)
        real(rt), intent(out), optional :: table(:,:),val
        ! local
        integer, allocatable :: v1(:),v2(:)
        real(rt), allocatable :: t(:,:),w(:)
        logical, allocatable :: mask(:)
        logical :: totable, toval, toscreen
        integer :: nrows,ncols,irow,icol,i,uid,r,c

        ! Error chcek
        if (size(varlist1)/=this%ns) stop 'ERROR: stata_corr: varlist1 has the wrong size' 
        if (size(varlist2)/=this%ns) stop 'ERROR: stata_corr: varlist1 has the wrong size' 

        toscreen = .not. present(table) .and. .not. present(val)
        totable = present(table)
        toval = present(val)

        nrows = count(varlist1)
        ncols = count(varlist2)

        allocate(v1(nrows))
        allocate(v2(ncols))

        irow = 0
        icol = 0
        do i = 1,this%ns
            if (varlist1(i)) then
                irow = irow + 1
                v1(irow) = i
            end if
            if (varlist2(i)) then
                icol = icol + 1
                v2(icol) = i
            end if
        end do

        allocate(mask(this%nobs))
        allocate(w(this%nobs))
        allocate(t(nrows,ncols))

        do irow = 1,nrows
            do icol = 1,ncols

                r = v1(irow)
                c = v2(icol)
                
                mask = this%d(:,r)/=STATA_MISSING .and. this%d(:,r)/=STATA_INVALID .and. &
                       this%d(:,c)/=STATA_MISSING .and. this%d(:,c)/=STATA_INVALID

                where (mask)
                    w = this%w
                elsewhere
                    w = 0._rt
                endwhere
                if (sum(w)>0._rt) w = w / sum(w)

                if (any(mask)) then
                    t(irow,icol) = corr_core_stable(this%d(:,r),this%d(:,c),w,this%nobs)
                    if (t(irow,icol)/=STATA_INVALID .and. t(irow,icol)/=STATA_MISSING) then
                        if (abs(t(irow,icol))>1.0_rt+sqrt(abs(1.0_rt))) write(0,*) 'WARNING: mod_stata: corr: computed correlation found to be invalid: ',t(irow,icol)
                    end if
                else
                    t(irow,icol) = STATA_MISSING
                end if

            end do
        end do

        ! Now have all the correlations, output
        if (toscreen) then
            uid = 6
            ! Header row
            write(uid,STATA_S_FMT,advance='no') 'variable' ! Top left 
            do icol = 1,ncols
                write(uid,'(A1)',advance='no') ','
                c = v2(icol)
                write(uid,STATA_S_FMT,advance='no') adjustr(this%snames(c)) ! varlist2 names
            end do
            write(uid,*)

            ! The meat
            do irow = 1,nrows
                r = v1(irow)
                write(uid,STATA_S_FMT,advance='no') adjustr(this%snames(r)) ! varlist1 names
                do icol = 1,ncols
                    write(uid,'(A1)',advance='no') ','
                    call disp_stata_value(uid,t(irow,icol))
                end do
                write(uid,*)
            end do
        end if
        if (totable) then
            if (any(shape(t)/=shape(table))) stop 'ERROR: stata_corr: table has the wrong dimensions'
            table = t
        end if
        if (toval) then
            if (any(shape(t)/=[1,1])) stop 'ERROR: stata_corr: val present but more than one pair was requested'
            val = t(1,1)
        end if


    end subroutine

    ! corr_core
    ! This 
    function corr_core_fast(x,y,n,mask) result(c)
        implicit none
        real(rt), intent(in) :: x(*),y(*)
        integer, intent(in) :: n
        logical, intent(in), optional :: mask(*)
        real(rt) :: c
        ! local
        real(rt) :: Ex,Ey,Ex2,Ey2,Exy,sdx,sdy
        integer :: i,m

        Ex = 0e0_rt
        Ex2 = 0e0_rt
        Ey = 0e0_rt
        Ey2 = 0e0_rt
        Exy = 0e0_rt
        m = 0
        if (present(mask)) then
            do i = 1,n
                if (mask(i)) then
                    Ex = Ex + x(i)
                    Ey = Ey + y(i)
                    Ex2 = Ex2 + x(i)**2
                    Ey2 = Ey2 + y(i)**2
                    Exy = Exy + x(i)*y(i)
                    m = m + 1
                end if
            end do
        else
            m = n
            do i = 1,n
                Ex = Ex + x(i)
                Ey = Ey + y(i)
                Ex2 = Ex2 + x(i)**2
                Ey2 = Ey2 + y(i)**2
                Exy = Exy + x(i)*y(i)
            end do
        end if

        sdx = sqrt(max(0._rt,m*Ex2 - Ex**2)) ! in case numerical error induces = -epsilon, bound at 0
        sdy = sqrt(max(0._rt,m*Ey2 - Ey**2))

        if (sdx>0e0_rt .and. sdy>0e0_rt) then
            c = (m*Exy-Ex*Ey)/sdx/sdy
        else
            c = STATA_INVALID
        end if

    end function 

    ! Numerically stable version
    subroutine stats_core_stable(x,weight,mean,std,var,min,max,mass)
        implicit none
        real(rt), intent(in) :: x(:),weight(:)
        real(rt), intent(out), optional :: mean,std,var,min,max,mass
        ! local
        real(rt) :: sumw,Ex,stdx,tmp,minx,maxx,varx
        integer :: i,n

        n = size(x)

        sumw = 0._rt
        Ex = 0._rt
        varx = 0._rt
        minx = huge(0._rt)
        maxx = -huge(0._rt)

        ! Compute the mean and sum
        do i = 1,n
            if (weight(i)>0._rt .and. x(i)/=STATA_INVALID .and. x(i)/=STATA_MISSING) then
                sumw = sumw + weight(i)
                Ex = Ex + x(i)*weight(i)
            end if
        end do

        if (sumw<=0._rt) then
            Ex = STATA_INVALID
            varx = STATA_INVALID
            stdx = STATA_INVALID 
            minx = STATA_INVALID
            maxx = STATA_INVALID
        else
            Ex = Ex/sumw

            ! Now compute the stdevs
            do i = 1,n
                if (weight(i)>0._rt .and. x(i)/=STATA_INVALID .and. x(i)/=STATA_MISSING) then
                    varx = varx + (x(i)-Ex)**2*weight(i)
                end if
            end do
            varx = varx/sumw
            stdx = merge(sqrt(varx),0e0_rt,stdx>=0e0_rt) 

        end if

        ! Output requested stats
        if (present(mean)) mean = Ex
        if (present(std)) std = stdx
        if (present(var)) var = varx
        if (present(mass)) mass = sumw
        if (present(min)) min = minx
        if (present(max)) max = maxx

    end subroutine

    ! This is a numerically stable version
    function corr_core_stable(x,y,weight,n) result(c)
        implicit none
        real(rt), intent(in) :: x(*),y(*),weight(*)
        integer, intent(in) :: n
        real(rt) :: c
        ! local
        real(rt) :: Ex,Ey,stdx,stdy,tmp,sumw
        integer :: i,m,nobs
        logical :: tf

        sumw = 0e0_rt
        Ex = 0e0_rt
        Ey = 0e0_rt
        stdx = 0e0_rt
        stdy = 0e0_rt

        ! Compute how much mass there is total
        do i = 1,n
            if (isvalid(x(i)) .and. isvalid(y(i))) sumw = sumw + weight(i)
        end do
         
        if (sumw<=0._rt) then
            c = STATA_INVALID
            return
        end if

        ! Now compute the means
        do i = 1,n
            if (isvalid(x(i)) .and. isvalid(y(i)) .and. weight(i)>0._rt) then
                Ex = Ex + x(i)*weight(i)
                Ey = Ey + y(i)*weight(i)
            end if
        end do
        Ex = Ex/sumw
        Ey = Ey/sumw

        ! Now compute the stdevs
        do i = 1,n
            if (isvalid(x(i)) .and. isvalid(y(i)) .and. weight(i)>0._rt) then
                stdx = stdx + (x(i)-Ex)**2*weight(i)
                stdy = stdy + (y(i)-Ey)**2*weight(i)
            end if
        end do
        stdx = stdx/sumw
        stdy = stdy/sumw
        stdx = merge(sqrt(stdx),0e0_rt,stdx>=0e0_rt) 
        stdy = merge(sqrt(stdy),0e0_rt,stdy>=0e0_rt) 

        ! Check for valid standard deviations
        if (stdx<=epsilon(0e0_rt) .or. stdy==epsilon(0e0_rt)) then
            c = STATA_INVALID
            return
        end if

        ! If still here, compute the correlation
        tmp = 0d0
        do i = 1,n
            if (isvalid(x(i)) .and. isvalid(y(i)) .and. weight(i)>0._rt) then
                tmp = tmp + (x(i)-Ex)/stdx*(y(i)-Ey)/stdy*weight(i)
            end if
        end do
        tmp = tmp/sumw
        c = tmp

    end function 




    ! Frees any memory in the stata workspace
    subroutine stata_free(this)
        implicit none
        class(stata_wkspace), intent(inout) :: this

    end subroutine


    !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    !                 AUXILIARY ROUTINES             !
    ! These are not meant to be called by the user.  !
    !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

    subroutine test_stata()
        real(rt) :: x(8),y(8),table(3,3),tmp
        integer :: i
        type(stata_wkspace) :: w

        x = [2.0_rt,3.0_rt,4.0_rt,5.0_rt,10.0_rt,-1.0_rt,STATA_MISSING,STATA_INVALID]
        where (x /= STATA_MISSING .and. x/=STATA_INVALID) y = x**2 
        y(7) = STATA_MISSING
        y(8) = STATA_INVALID

        call w%gen(x,'x')
        call w%gen(stata_log(x),'ln_x')
        call w%gen(y,'y')
        call w%list()
        write(*,*) 

        i = w%i('x'); if (i/=1) stop 'ERROR: test_stata: failed test 1'
        i = w%i('ln_x'); if (i/=2) stop 'ERROR: test_stata: failed test 2'

        ! i = w%i('asdf'); if (i/=-1) stop 'ERROR: test_stata: failed test 3'

        print*
        print*,'SUMMARIZE'
        print*
        call w%summarize()

        print*
        print*,'MEDIAN'
        print*
        tmp = w%median('x')
        print*,'Median of x: ',tmp
        print*,'Median of ln_x: ',w%median('ln_x')
        print*,'10,25,50,75,90 pctiles of x', w%pctile('x',[10d0,25d0,50d0,75d0,90d0])

        print*
        print*,'CORRELATION'
        print*
        tmp = w%pwcorr('x','ln_x')
        print*,'pwcorr correlation of x,ln_x: ',tmp

        call w%corr(w%varlist('x,ln_x'),w%varlist('*'))
        call w%corr(w%varlist('x'),w%varlist('y'))
        call w%corr(w%varlist('y,x'),w%varlist('ln_x'))


        call w%corr(w%varlist('*'),w%varlist('*'))
        call w%corr(w%varlist('*'),w%varlist('*'),table)

        print*,'table',table
        print*,'Correlation between x and ln_x : ',table(w%i('x'),w%i('ln_x'))

        ! call w%scatter('x','y','/tmp/scatter.eps')

        call w%regress('ln_x',w%varlist('x,y')) 

        ! call w%export('/tmp/statatest.csv')
        ! write(*,*) 'Stata wkspace exported to /tmp/statatest.csv'


    end subroutine

    subroutine disp_stata_value(uid,v,isint,ishighprec,dotryint)
        use mod_util
        implicit none
        integer, intent(in) :: uid
        real(rt), intent(in) :: v
        logical, optional, intent(in) :: isint
        logical, optional, intent(in) :: ishighprec 
        logical, optional, intent(in) :: dotryint ! whether to try to write as in integer
        ! local
        logical :: isint_local,ishighprec_local,dotryint_local

        isint_local = .false.
        ishighprec_local = .false.
        dotryint_local = .false.
        if (present(isint)) isint_local = isint
        if (present(ishighprec)) ishighprec_local = ishighprec
        if (present(dotryint)) dotryint_local = dotryint

        ! If dotryint is specified, then compare the value with an integer to see whether we can save on space
        if (dotryint_local) then
            if (v<huge(1)) then
                if (abs(nint(v)-v)<=epsilon(0d0)*abs(v)) then
                    isint_local = .true.
                end if
            end if
        end if

        if (ishighprec_local) then 
            if (v==STATA_MISSING .or. v==STATA_INVALID) then
                write(uid,'(A1)',advance='no') '.'
            else
                if (.not. isint_local) then
                    if (v<0d0) then
                        write(uid,STATA_Eneg_XFMT,advance='no') v
                    else
                        write(uid,STATA_Epos_XFMT,advance='no') v
                    end if
                else
                    ! write(uid,'(A)',advance='no') trim(conv_i_to_s(nint(v)))
                    select case (nint(v))
                        case (0:9);             write(uid,'(I1)',advance='no') nint(v)
                        case (10:99);           write(uid,'(I2)',advance='no') nint(v)
                        case (100:999);         write(uid,'(I3)',advance='no') nint(v)
                        case (1000:9999);       write(uid,'(I4)',advance='no') nint(v)
                        case (10000:99999);     write(uid,'(I5)',advance='no') nint(v)
                        case (100000:999999);   write(uid,'(I6)',advance='no') nint(v)
                        case (-9:-1);           write(uid,'(I2)',advance='no') nint(v)
                        case (-99:-10);         write(uid,'(I3)',advance='no') nint(v)
                        case (-999:-100);       write(uid,'(I4)',advance='no') nint(v)
                        case (-9999:-1000);     write(uid,'(I5)',advance='no') nint(v)
                        case (-99999:-10000);   write(uid,'(I6)',advance='no') nint(v)
                        case (-999999:-100000); write(uid,'(I7)',advance='no') nint(v)
                        case default; write(uid,'(I10)',advance='no') nint(v)
                    end select
                end if
            end if
        else
            if (v==STATA_MISSING) then
                write(uid,STATA_S_FMT,advance='no') ' .'
            elseif (v==STATA_INVALID) then
                write(uid,STATA_S_FMT,advance='no') ' invalid'
            else
                if (.not. isint_local) then
                    if ((abs(v)>1e9_rt .or. abs(v)<1e-4_rt) .and. abs(v)/=0._rt) then
                        write(uid,STATA_E_FMT,advance='no') v
                    else
                        write(uid,STATA_F_FMT,advance='no') v
                    end if
                else
                    write(uid,STATA_I_FMT,advance='no') nint(v)
                end if
            end if
        end if

    end subroutine

    ! Increases the number of series
!     subroutine stata_expand(this,newnobs,newns)
!         implicit none
!         class(stata_wkspace), intent(inout) :: this
!         integer, intent(in) :: newnobs,newns
!         ! local
!         real(rt), allocatable :: dold
! 
!         if (allocated(this%d)) then
! 
!             ! Check that the new number of observations is equal to current
!             if (newnobs /= this%nobs) then
!                 stop 'ERROR: stata: new number of observations does not agree with current'
!             end if
! 
!             ! Need to create a copy of d
!             allocate(dold(size(d,1),size(d,2)))
!             dold = this%d
!             
!             ! Expand 
!             deallocate(this%d)
!             allocate(this%d(
! 
!         end if
! 
! 
! 
!         if (this%ns==0) then
!             allocate(this%d(
!         else
! 
!         end if
! 
!     end subroutine 

    ! Applies function f on an element by element basis to the variable name 
!     subroutine bsxfun(this,f,name)
!         implicit none
!         class(stata_wkspace), intent(inout) :: this
!         real(rt) function f(x)
!             use mod_types
!             implicit none
!             real(rt), intent(in) :: x
!         end function 
!         character(len=*), intent(in) :: name
! 
!     end subroutine 
    



end module
