! Fortran interface to Merseene Twister C code available at 
! http://www.math.sci.hiroshima-u.ac.jp/~m-mat/MT/VERSIONS/C-LANG/ver991029.html
! Retrieved 1/7/2017
! Slighlty modified C code with Fortran 2003 interfaces
! Grey Gordon
module mod_twister
    implicit none

    interface 
        subroutine sgenrand(seed) bind(C,name='sgenrand')
            use iso_c_binding, only: C_LONG
            implicit none
            integer(C_LONG) :: seed
        end subroutine
        subroutine resetsgenrand() bind(C,name='resetsgenrand')
            implicit none
        end subroutine
        function genrand() bind(C,name='genrand')
            use iso_c_binding, only: C_DOUBLE
            implicit none
            real(C_DOUBLE) :: genrand
        end function
    end interface

    interface mt_random_number
        module procedure :: mtrn0,mtrn1,mtrn2,mtrn3,mtrn4,mtrn5
    end interface

    private :: w,mtrn0,mtrn1,mtrn2,mtrn3,mtrn4,mtrn5

contains

    subroutine mt_random_seed(seed)
        use iso_c_binding, only: C_LONG
        implicit none
        integer, intent(in), optional :: seed
        !$OMP CRITICAL
        if (present(seed)) then
            if (seed<=0) then
                write(0,*) 'WARNING: mt_random_seed: invalid seed, set to > 0, using default seed'
                call resetsgenrand() ! reset to their default
            else
                call sgenrand(int(seed,C_LONG))
            end if
        else
            call resetsgenrand() ! reset to their default
        end if
        !$OMP END CRITICAL
    end subroutine 

    subroutine mtrn0(x)
        implicit none
        real(8), intent(out) :: x
        !$OMP CRITICAL
        x = genrand()
        !$OMP END CRITICAL
    end subroutine 

    subroutine mtrn1(x); real(8) :: x(:);         call w(x,size(x)); end subroutine 
    subroutine mtrn2(x); real(8) :: x(:,:);       call w(x,size(x)); end subroutine 
    subroutine mtrn3(x); real(8) :: x(:,:,:);     call w(x,size(x)); end subroutine 
    subroutine mtrn4(x); real(8) :: x(:,:,:,:);   call w(x,size(x)); end subroutine 
    subroutine mtrn5(x); real(8) :: x(:,:,:,:,:); call w(x,size(x)); end subroutine 

    subroutine w(x,nx)
        implicit none
        real(8), intent(out) :: x(*)
        integer, intent(in) :: nx
        integer :: i
        !$OMP CRITICAL
        do i = 1,nx
            x(i) = genrand()
        end do
        !$OMP END CRITICAL
    end subroutine 

end module 

module mod_twister_test
    use mod_twister
    implicit none
contains

    subroutine test_twister()
        implicit none
        real(8), allocatable :: x(:,:)
        real :: ta,tb
        integer :: i

        allocate(x(1000,3))
        
        ! This should be the same as what they provide in their out file
        call mt_random_number(x(:,1))
        if (maxval(abs(x(:,1)-testdata()))>1d-7) stop 'test_twister: failed test 0'

        ! Change the seed and verify it is different
        call mt_random_seed(30001)
        call mt_random_number(x(:,2))
        if (any(x(:,2)==x(:,1))) stop 'test_twister: failed test 1'

        ! write(*,'(5(f10.8," "))') x(:,2)

        ! Revert the seed to default
        call mt_random_seed()
        call mt_random_number(x(:,3))
        if (.not. all(x(:,3)==x(:,1))) stop 'test_twister: failed test 2'

        ! Speed comparison
        call cpu_time(ta)
        do i = 1,100000000
            call mt_random_number(x(1,1))
        end do
        call cpu_time(tb)

        print*,'Mersenne-Twister time:',tb-ta

        call cpu_time(ta)
        do i = 1,100000000
            call random_number(x(1,1))
        end do
        call cpu_time(tb)
        print*,'Default time         :',tb-ta

        print*,'test_twister : passed'

    end subroutine


    function testdata() result(x)
        implicit none
        real :: x(1000)
        integer :: i

        i = 1; 

        x(i:i+4) = [0.66757648,0.36908387,0.72483069,0.68775863,0.57364695]; i = i+5
        x(i:i+4) = [0.81077819,0.27108403,0.83777019,0.13736374,0.95745406]; i = i+5
        x(i:i+4) = [0.17860795,0.95299339,0.30790312,0.78163234,0.76835241]; i = i+5
        x(i:i+4) = [0.84970713,0.85602622,0.34343780,0.60581002,0.79646820]; i = i+5
        x(i:i+4) = [0.58030504,0.00665827,0.44261972,0.28159319,0.83360769]; i = i+5
        x(i:i+4) = [0.57935190,0.01859701,0.77031648,0.67844436,0.97001810]; i = i+5
        x(i:i+4) = [0.44956184,0.77149456,0.60622814,0.65834100,0.45450425]; i = i+5
        x(i:i+4) = [0.32715062,0.45621146,0.51734903,0.95696471,0.23879988]; i = i+5
        x(i:i+4) = [0.85454026,0.91521138,0.60093783,0.31445925,0.90979591]; i = i+5
        x(i:i+4) = [0.93474556,0.21777190,0.40560878,0.74110343,0.87901243]; i = i+5
        x(i:i+4) = [0.62175155,0.14704158,0.77954247,0.66942131,0.26849801]; i = i+5
        x(i:i+4) = [0.24094272,0.98148164,0.78444039,0.56903055,0.11400358]; i = i+5
        x(i:i+4) = [0.34932469,0.87404238,0.16636981,0.29357386,0.68747168]; i = i+5
        x(i:i+4) = [0.77700075,0.12931921,0.01741686,0.26816707,0.32850746]; i = i+5
        x(i:i+4) = [0.26081908,0.51879284,0.94471072,0.72524766,0.59233503]; i = i+5
        x(i:i+4) = [0.82239323,0.53261521,0.47330885,0.34851109,0.77604562]; i = i+5
        x(i:i+4) = [0.42312313,0.05839693,0.27006566,0.15956065,0.82463138]; i = i+5
        x(i:i+4) = [0.10756423,0.73272163,0.78488413,0.17240949,0.38178793]; i = i+5
        x(i:i+4) = [0.20741402,0.42296261,0.17240298,0.04327979,0.99182907]; i = i+5
        x(i:i+4) = [0.92160498,0.99312528,0.21458565,0.96731334,0.18748237]; i = i+5
        x(i:i+4) = [0.70184654,0.42031099,0.13070361,0.62999246,0.36046608]; i = i+5
        x(i:i+4) = [0.58470953,0.24153172,0.80754590,0.52701693,0.02341943]; i = i+5
        x(i:i+4) = [0.31606620,0.84152189,0.03988034,0.53137092,0.82027225]; i = i+5
        x(i:i+4) = [0.54687890,0.02497169,0.99563849,0.80733361,0.05267620]; i = i+5
        x(i:i+4) = [0.26443902,0.27307048,0.37345968,0.55682407,0.73387997]; i = i+5
        x(i:i+4) = [0.25687306,0.73453675,0.25915843,0.82343475,0.68741401]; i = i+5
        x(i:i+4) = [0.28851792,0.59806914,0.66306494,0.76804766,0.87679883]; i = i+5
        x(i:i+4) = [0.42346814,0.41643898,0.29711001,0.22104934,0.12570099]; i = i+5
        x(i:i+4) = [0.58565603,0.76329179,0.51112485,0.73125562,0.52288791]; i = i+5
        x(i:i+4) = [0.92501953,0.21589842,0.11901197,0.64867112,0.90554553]; i = i+5
        x(i:i+4) = [0.33995496,0.81055839,0.61600598,0.59548759,0.70043684]; i = i+5
        x(i:i+4) = [0.54439747,0.86514897,0.54195165,0.38894885,0.09278668]; i = i+5
        x(i:i+4) = [0.91089129,0.43573767,0.19075151,0.37337378,0.52223266]; i = i+5
        x(i:i+4) = [0.74667570,0.69716858,0.81010343,0.56540978,0.69970076]; i = i+5
        x(i:i+4) = [0.13894581,0.99567377,0.14665349,0.38001195,0.49245637]; i = i+5
        x(i:i+4) = [0.18654659,0.89124594,0.45647667,0.41794316,0.69371423]; i = i+5
        x(i:i+4) = [0.02033509,0.73324065,0.14992572,0.81887529,0.48147565]; i = i+5
        x(i:i+4) = [0.65209065,0.04573058,0.40765170,0.34219265,0.67796685]; i = i+5
        x(i:i+4) = [0.46309207,0.65200856,0.65320173,0.51036734,0.85563082]; i = i+5
        x(i:i+4) = [0.62438484,0.98426651,0.63535972,0.77525738,0.44391929]; i = i+5
        x(i:i+4) = [0.73162364,0.26874948,0.96943570,0.41880865,0.93454721]; i = i+5
        x(i:i+4) = [0.47240659,0.51596869,0.43055283,0.95628885,0.58562796]; i = i+5
        x(i:i+4) = [0.82009466,0.46187088,0.47145557,0.82367849,0.91507510]; i = i+5
        x(i:i+4) = [0.01677295,0.41464613,0.82151208,0.05860020,0.27681455]; i = i+5
        x(i:i+4) = [0.45335699,0.08624149,0.88801548,0.03782344,0.88876984]; i = i+5
        x(i:i+4) = [0.70426919,0.02663374,0.74290453,0.63726167,0.23433229]; i = i+5
        x(i:i+4) = [0.84906979,0.06704500,0.26773391,0.11348281,0.49147618]; i = i+5
        x(i:i+4) = [0.97085464,0.63053349,0.38698528,0.11862192,0.25593954]; i = i+5
        x(i:i+4) = [0.68441813,0.95854416,0.52087730,0.58967510,0.44819302]; i = i+5
        x(i:i+4) = [0.52970914,0.46601702,0.01998145,0.27799436,0.66620801]; i = i+5
        x(i:i+4) = [0.23612172,0.23207917,0.96955989,0.25392320,0.49604321]; i = i+5
        x(i:i+4) = [0.17908208,0.23175815,0.21657196,0.01244777,0.22504003]; i = i+5
        x(i:i+4) = [0.44390366,0.80582624,0.90367141,0.52995965,0.02774875]; i = i+5
        x(i:i+4) = [0.41099963,0.52810009,0.25262355,0.62300157,0.10668494]; i = i+5
        x(i:i+4) = [0.33876737,0.23698669,0.61686954,0.69836698,0.09049847]; i = i+5
        x(i:i+4) = [0.79518412,0.70399305,0.35529014,0.57099395,0.47830871]; i = i+5
        x(i:i+4) = [0.84556855,0.52367918,0.49086827,0.59759319,0.14306320]; i = i+5
        x(i:i+4) = [0.39515950,0.44396248,0.45752610,0.85800966,0.42118261]; i = i+5
        x(i:i+4) = [0.77738501,0.87649970,0.90901211,0.15236771,0.78444909]; i = i+5
        x(i:i+4) = [0.04021164,0.41665851,0.02307698,0.04591255,0.30446493]; i = i+5
        x(i:i+4) = [0.76262273,0.73626249,0.50273361,0.56620328,0.11160146]; i = i+5
        x(i:i+4) = [0.44833702,0.61185087,0.18440434,0.11005170,0.97270536]; i = i+5
        x(i:i+4) = [0.60519276,0.79958488,0.44343314,0.43668535,0.64782387]; i = i+5
        x(i:i+4) = [0.43183192,0.07800938,0.18224609,0.77993019,0.48207379]; i = i+5
        x(i:i+4) = [0.33636839,0.96241462,0.70340082,0.66551682,0.98945885]; i = i+5
        x(i:i+4) = [0.72538204,0.65792949,0.75068180,0.96658137,0.99798036]; i = i+5
        x(i:i+4) = [0.29253233,0.35340504,0.17358288,0.64194354,0.93444911]; i = i+5
        x(i:i+4) = [0.56427628,0.50952936,0.89161637,0.36309527,0.17361471]; i = i+5
        x(i:i+4) = [0.80093093,0.03751695,0.86256897,0.48730053,0.86793819]; i = i+5
        x(i:i+4) = [0.25844010,0.83032378,0.30121456,0.44455501,0.56349591]; i = i+5
        x(i:i+4) = [0.53661566,0.84637289,0.82700509,0.99852184,0.15577177]; i = i+5
        x(i:i+4) = [0.87104550,0.51220783,0.50480298,0.39526474,0.24933757]; i = i+5
        x(i:i+4) = [0.64975530,0.21643175,0.88900918,0.36142797,0.46198154]; i = i+5
        x(i:i+4) = [0.46849597,0.77152779,0.74209345,0.87159297,0.80466664]; i = i+5
        x(i:i+4) = [0.86160078,0.15681999,0.85708844,0.14900673,0.21142697]; i = i+5
        x(i:i+4) = [0.06643849,0.81045159,0.63471125,0.79935807,0.97321938]; i = i+5
        x(i:i+4) = [0.84261641,0.67640103,0.25754596,0.49006639,0.80017935]; i = i+5
        x(i:i+4) = [0.58569104,0.55223658,0.23739079,0.84469850,0.20389842]; i = i+5
        x(i:i+4) = [0.08635077,0.68473020,0.15959708,0.17418036,0.88410115]; i = i+5
        x(i:i+4) = [0.24477070,0.53151348,0.11113551,0.94362181,0.79413150]; i = i+5
        x(i:i+4) = [0.63054362,0.91038647,0.18661656,0.73696949,0.72152386]; i = i+5
        x(i:i+4) = [0.04716607,0.71173065,0.73892523,0.39655731,0.33339456]; i = i+5
        x(i:i+4) = [0.36044696,0.95434741,0.93717326,0.39039334,0.74162154]; i = i+5
        x(i:i+4) = [0.32070520,0.15882017,0.05341297,0.12486067,0.30334311]; i = i+5
        x(i:i+4) = [0.51749927,0.79193143,0.07974914,0.81300209,0.20497391]; i = i+5
        x(i:i+4) = [0.53283376,0.59608995,0.21064801,0.67586340,0.52143474]; i = i+5
        x(i:i+4) = [0.32149937,0.35529888,0.18606804,0.89970702,0.17882435]; i = i+5
        x(i:i+4) = [0.59965903,0.86156567,0.32245210,0.41552510,0.48249092]; i = i+5
        x(i:i+4) = [0.52925015,0.01700478,0.44096038,0.12912483,0.23340165]; i = i+5
        x(i:i+4) = [0.05701807,0.44580366,0.72250673,0.95469388,0.00351167]; i = i+5
        x(i:i+4) = [0.96364870,0.06669990,0.00654904,0.99784653,0.75747954]; i = i+5
        x(i:i+4) = [0.19758164,0.17506852,0.67815743,0.99306636,0.76540895]; i = i+5
        x(i:i+4) = [0.71536673,0.34126924,0.26058940,0.39734235,0.38130604]; i = i+5
        x(i:i+4) = [0.30299881,0.23163295,0.73859811,0.64891197,0.94007861]; i = i+5
        x(i:i+4) = [0.46607060,0.71442945,0.44556854,0.03838125,0.78853184]; i = i+5
        x(i:i+4) = [0.54961601,0.12548533,0.26134595,0.50379613,0.88960521]; i = i+5
        x(i:i+4) = [0.11016112,0.09385041,0.18176689,0.44619499,0.49763107]; i = i+5
        x(i:i+4) = [0.69223586,0.36344271,0.08919716,0.65746979,0.02227179]; i = i+5
        x(i:i+4) = [0.25632302,0.47842123,0.33361583,0.00918456,0.90007729]; i = i+5
        x(i:i+4) = [0.52232804,0.07607001,0.87652747,0.89765526,0.89049552]; i = i+5
        x(i:i+4) = [0.40978921,0.90573503,0.00891906,0.75643755,0.23228934]; i = i+5
        x(i:i+4) = [0.30315809,0.26413128,0.87511469,0.89793836,0.93036177]; i = i+5
        x(i:i+4) = [0.46517147,0.54517078,0.96369087,0.98238836,0.67069820]; i = i+5
        x(i:i+4) = [0.36178296,0.17227293,0.62125561,0.28366943,0.40203458]; i = i+5
        x(i:i+4) = [0.91942652,0.29481153,0.70148921,0.52166847,0.08780925]; i = i+5
        x(i:i+4) = [0.21124205,0.59623447,0.04913357,0.43425462,0.39126768]; i = i+5
        x(i:i+4) = [0.70001842,0.27845033,0.12548004,0.18192453,0.37806754]; i = i+5
        x(i:i+4) = [0.49811947,0.16183850,0.29506583,0.15007203,0.62336950]; i = i+5
        x(i:i+4) = [0.25128931,0.59257764,0.12502416,0.79392712,0.07641970]; i = i+5
        x(i:i+4) = [0.89130398,0.69203443,0.61444017,0.04846100,0.77313389]; i = i+5
        x(i:i+4) = [0.78321358,0.50332935,0.70495746,0.65620157,0.42759664]; i = i+5
        x(i:i+4) = [0.62866350,0.86406004,0.63036488,0.52445634,0.12621910]; i = i+5
        x(i:i+4) = [0.06772138,0.01525184,0.97705369,0.08335805,0.21231837]; i = i+5
        x(i:i+4) = [0.65240669,0.74245828,0.14353654,0.97234090,0.81323412]; i = i+5
        x(i:i+4) = [0.18525218,0.07981979,0.89667167,0.57686567,0.22681910]; i = i+5
        x(i:i+4) = [0.48199554,0.70981122,0.81376251,0.15136839,0.00345939]; i = i+5
        x(i:i+4) = [0.35192942,0.30182722,0.05603578,0.97045692,0.50327985]; i = i+5
        x(i:i+4) = [0.03468732,0.22421088,0.30447416,0.77415169,0.30796931]; i = i+5
        x(i:i+4) = [0.40374548,0.55635645,0.08908196,0.10071154,0.07256218]; i = i+5
        x(i:i+4) = [0.49450104,0.96809050,0.73443314,0.69341799,0.30055386]; i = i+5
        x(i:i+4) = [0.46748246,0.75922816,0.61480063,0.79258162,0.37544248]; i = i+5
        x(i:i+4) = [0.71217758,0.99829254,0.68052207,0.70617895,0.81760155]; i = i+5
        x(i:i+4) = [0.23095584,0.14993299,0.56734341,0.83274618,0.55343761]; i = i+5
        x(i:i+4) = [0.28454225,0.85574309,0.46394642,0.64788899,0.27953923]; i = i+5
        x(i:i+4) = [0.03715739,0.15826703,0.73619860,0.13723131,0.24323187]; i = i+5
        x(i:i+4) = [0.45757685,0.39901923,0.28298861,0.42217752,0.70939637]; i = i+5
        x(i:i+4) = [0.32808287,0.75695066,0.82242039,0.60908726,0.32272598]; i = i+5
        x(i:i+4) = [0.04062714,0.92638830,0.65684054,0.49064401,0.80589136]; i = i+5
        x(i:i+4) = [0.11439019,0.49646593,0.97376522,0.47284227,0.21563978]; i = i+5
        x(i:i+4) = [0.42910902,0.00113945,0.25409239,0.60046464,0.79847317]; i = i+5
        x(i:i+4) = [0.80675877,0.94228278,0.80807688,0.11385570,0.37008334]; i = i+5
        x(i:i+4) = [0.57919961,0.01165382,0.97437180,0.11137551,0.95058400]; i = i+5
        x(i:i+4) = [0.69462399,0.92857962,0.68617807,0.63748421,0.15153586]; i = i+5
        x(i:i+4) = [0.04820435,0.10926267,0.62481920,0.70259732,0.03795410]; i = i+5
        x(i:i+4) = [0.39425341,0.90896522,0.70401651,0.69061450,0.30984713]; i = i+5
        x(i:i+4) = [0.16645158,0.11348071,0.74370218,0.03075622,0.80838600]; i = i+5
        x(i:i+4) = [0.77608956,0.22453614,0.43140626,0.79120270,0.31831407]; i = i+5
        x(i:i+4) = [0.70509658,0.45985545,0.32841860,0.21663841,0.41848252]; i = i+5
        x(i:i+4) = [0.23863590,0.98169491,0.68144766,0.64749301,0.68689040]; i = i+5
        x(i:i+4) = [0.05927982,0.57301120,0.72202073,0.94679581,0.52993362]; i = i+5
        x(i:i+4) = [0.46699535,0.16361573,0.08204070,0.93877394,0.90075423]; i = i+5
        x(i:i+4) = [0.71303221,0.46212466,0.26128921,0.42757976,0.70993807]; i = i+5
        x(i:i+4) = [0.81611594,0.32974144,0.70115395,0.84260721,0.61336456]; i = i+5
        x(i:i+4) = [0.12152844,0.29310132,0.84912503,0.11238740,0.13425191]; i = i+5
        x(i:i+4) = [0.23880796,0.37711937,0.68262252,0.55499206,0.97053602]; i = i+5
        x(i:i+4) = [0.90799536,0.63449957,0.38267489,0.95773604,0.39469768]; i = i+5
        x(i:i+4) = [0.21639232,0.96168053,0.76327704,0.07885930,0.61037994]; i = i+5
        x(i:i+4) = [0.17701924,0.42548931,0.68777647,0.65250575,0.22168134]; i = i+5
        x(i:i+4) = [0.36423438,0.35157707,0.58608833,0.67037882,0.32057486]; i = i+5
        x(i:i+4) = [0.12697130,0.06413693,0.41133618,0.19592106,0.53063194]; i = i+5
        x(i:i+4) = [0.03743389,0.39245431,0.02260016,0.01639602,0.62810921]; i = i+5
        x(i:i+4) = [0.30548560,0.06471907,0.58493305,0.82888639,0.86326006]; i = i+5
        x(i:i+4) = [0.41957691,0.51555913,0.85680825,0.65185340,0.48061191]; i = i+5
        x(i:i+4) = [0.58642359,0.24937033,0.76580400,0.65943286,0.14594271]; i = i+5
        x(i:i+4) = [0.65702487,0.97914652,0.72003879,0.30915253,0.49157231]; i = i+5
        x(i:i+4) = [0.78161023,0.49047193,0.22831025,0.55284752,0.92503172]; i = i+5
        x(i:i+4) = [0.42263712,0.68315185,0.57438940,0.61632286,0.44636741]; i = i+5
        x(i:i+4) = [0.17643947,0.65202509,0.80607796,0.43422435,0.86417596]; i = i+5
        x(i:i+4) = [0.75541319,0.25209786,0.45490765,0.83182100,0.42972104]; i = i+5
        x(i:i+4) = [0.39431439,0.44497822,0.00317034,0.35558258,0.21981674]; i = i+5
        x(i:i+4) = [0.86635254,0.70108236,0.85212617,0.81285140,0.28096475]; i = i+5
        x(i:i+4) = [0.57373917,0.06030610,0.08187271,0.08103981,0.05521743]; i = i+5
        x(i:i+4) = [0.60366858,0.80840766,0.22519752,0.45208357,0.77613377]; i = i+5
        x(i:i+4) = [0.98809198,0.66074977,0.50609124,0.49925814,0.95129356]; i = i+5
        x(i:i+4) = [0.63774387,0.40093002,0.23960409,0.89959119,0.73796666]; i = i+5
        x(i:i+4) = [0.40425617,0.25315515,0.87111423,0.69927431,0.94043851]; i = i+5
        x(i:i+4) = [0.22482295,0.34799002,0.62883121,0.43375657,0.47967471]; i = i+5
        x(i:i+4) = [0.61386312,0.53480418,0.49800254,0.19342872,0.56108443]; i = i+5
        x(i:i+4) = [0.22027592,0.15743168,0.54430525,0.45811257,0.31760671]; i = i+5
        x(i:i+4) = [0.26099883,0.35044470,0.73967791,0.88207901,0.51721941]; i = i+5
        x(i:i+4) = [0.67004141,0.43027511,0.24685628,0.69409327,0.09190582]; i = i+5
        x(i:i+4) = [0.18060550,0.57402781,0.05961282,0.15650883,0.93932405]; i = i+5
        x(i:i+4) = [0.15185844,0.05230872,0.65950081,0.60320652,0.37636188]; i = i+5
        x(i:i+4) = [0.11059014,0.40273726,0.24052930,0.85677431,0.45301339]; i = i+5
        x(i:i+4) = [0.96441333,0.80133448,0.84363009,0.67261392,0.89616395]; i = i+5
        x(i:i+4) = [0.64598140,0.17080951,0.36307550,0.14222415,0.43349758]; i = i+5
        x(i:i+4) = [0.06753002,0.45846346,0.68672443,0.96548375,0.39795624]; i = i+5
        x(i:i+4) = [0.72285575,0.56221110,0.86196534,0.44750392,0.69995902]; i = i+5
        x(i:i+4) = [0.59544870,0.45478250,0.80917089,0.80112927,0.43769621]; i = i+5
        x(i:i+4) = [0.95902850,0.12208225,0.52342082,0.37452434,0.45324745]; i = i+5
        x(i:i+4) = [0.37212588,0.80644936,0.10630277,0.35234753,0.44668821]; i = i+5
        x(i:i+4) = [0.69975833,0.28361668,0.60866604,0.34436061,0.58327394]; i = i+5
        x(i:i+4) = [0.72971772,0.48689157,0.51989691,0.40516765,0.51277047]; i = i+5
        x(i:i+4) = [0.49979911,0.49448107,0.79164538,0.73163248,0.26263339]; i = i+5
        x(i:i+4) = [0.85113333,0.86145969,0.20285442,0.35671599,0.20565903]; i = i+5
        x(i:i+4) = [0.47554786,0.26583844,0.76954785,0.52301379,0.34483407]; i = i+5
        x(i:i+4) = [0.57444831,0.24517665,0.26264016,0.13488078,0.95328813]; i = i+5
        x(i:i+4) = [0.77127258,0.61438517,0.98361796,0.22709089,0.64742944]; i = i+5
        x(i:i+4) = [0.75649177,0.40462684,0.13090446,0.72346026,0.37970215]; i = i+5
        x(i:i+4) = [0.77391791,0.62590549,0.39044919,0.32394191,0.26753559]; i = i+5
        x(i:i+4) = [0.03737445,0.43973621,0.28036720,0.60531698,0.95381788]; i = i+5
        x(i:i+4) = [0.50138943,0.52025153,0.52482530,0.65166013,0.44344669]; i = i+5
        x(i:i+4) = [0.71764913,0.14407532,0.73642009,0.17391811,0.96678625]; i = i+5
        x(i:i+4) = [0.84782413,0.08101573,0.19726966,0.80169785,0.98081749]; i = i+5
        x(i:i+4) = [0.49861194,0.55174662,0.94342959,0.56679545,0.52948777]; i = i+5
        x(i:i+4) = [0.91718010,0.08518465,0.41046841,0.87667656,0.99765508]; i = i+5
        x(i:i+4) = [0.27429135,0.27119230,0.53659872,0.60580929,0.15388982]; i = i+5
        x(i:i+4) = [0.07549456,0.28821052,0.32386421,0.72722712,0.75192038]; i = i+5
        x(i:i+4) = [0.15067533,0.51887623,0.81139097,0.89741567,0.86488425]; i = i+5
        x(i:i+4) = [0.44807063,0.06424586,0.75766097,0.40567560,0.23996701]; i = i+5

    end function 


end module
