! A module for maximizing the Bellman equation one way or another
!
! 1) one-dimensional maximization routine 
! 2) two-dimensional maximization routine
module mod_valueopt
    use mod_types
    use mod_spline
    implicit none

    real(rt), parameter, private :: deathv = -sqrt(huge(0._rt)), deathc = -1e0_rt, deathap = -1e0_rt
    integer, parameter, private :: deathapi = -1 ! must be <= 0

    private :: sub_bellman_continuous_core

    interface sub_bellman_discrete_choose2
        module procedure sub_bellman_discrete_choose2_1d
    end interface sub_bellman_discrete_choose2

    interface sub_bellman_interp_choose2
        module procedure sub_bellman_interp_choose2_1d
    end interface sub_bellman_interp_choose2

    interface sub_bellman_discrete
        module procedure sub_bellman_discrete_1d, &
                         sub_bellman_discrete_2d, &
                         sub_bellman_discrete_3d, &
                         sub_bellman_discrete_4d
    end interface sub_bellman_discrete

    interface sub_bellman_continuous
        module procedure sub_bellman_continuous_2d, &
                         sub_bellman_continuous_3d, &
                         sub_bellman_continuous_4d
    end interface sub_bellman_continuous

    interface sub_bellman_interp
        module procedure sub_bellman_interp_2d, &
                         sub_bellman_interp_3d, &
                         sub_bellman_interp_4d
    end interface sub_bellman_interp


    interface checkShape
        module procedure checkShapeReal,checkShapeInt,&
                         checkShapeReal2,checkShapeInt2,&
                         checkShapeReal3,checkShapeInt3,&
                         checkShapeReal4,checkShapeInt4
    end interface


    abstract interface
        function PUcustom_template(c) result(PU)
            use mod_types
            implicit none
            real(rt), intent(in) :: c
            real(rt) :: PU
        end function
    end interface

    procedure(PUcustom_template), pointer, private :: PUcustom 

    ! Don't point to null, weird language issue: http://software.intel.com/en-us/forums/topic/329011
    type(spline), private, save :: negV_pp
    real(rt), dimension(:), pointer, private :: negV_x, negV_W, negV_gAp
    real(rt), private :: negV_y, negV_crra
    logical, private :: negV_isSchumaker

    !$OMP THREADPRIVATE(negV_pp,negV_x,negV_W,negV_gAp,negV_y,negV_crra,negV_isSchumaker)

contains

    ! Solves V(i) = max_{1<=i'<=nip} u(i,i') + W(i') for 1<=i<=ni where u and W are functions taking 
    ! integer arguments. The resulting policy function is called g(i). ni is inferred from the size of V, but
    ! nip must be passed as an argument. 
    !
    ! The function u should return u(i,i') as well as a flag saying whether the choice is feasible (0) or not.
    ! The function W should return W(i') as well as a flag saying whether the choice is feasible or not.
    ! Both functions should take assumed-size arrays realargs and intargs. These arrays are never accessed by
    ! bellman and may be size zero. The point of these arrays is to pass additional arguments to u and W in a 
    ! convenient way. Note that realargs and intargs may be any rank or size when bellman is called.
    !
    ! The routine uses speedups, based on what the user selects via opt.
    ! 
    ! Monotonicity:
    ! opt(1) | meaning
    ! 2      | binary monotonicity
    !
    ! Concavity:
    ! opt(2) | meaning
    ! 0      | none
    ! 
    subroutine bellman(V,g,u,W,nip,opt,realargs,intargs)
        implicit none
        real(rt), intent(out) :: V(:)
        integer, intent(out) :: g(:)
        integer, intent(in) :: nip
        integer, intent(in) :: opt(:)
        real(rt), intent(inout) :: realargs(*)
        integer, intent(inout) :: intargs(*)
        interface
            subroutine u(pu,i,ip,flag,realargs,intargs)
                use mod_types
                implicit none
                real(rt), intent(out) :: pu
                integer, intent(in) :: i,ip
                integer, intent(out) :: flag
                real(rt), intent(inout) :: realargs(*)
                integer, intent(inout) :: intargs(*)
            end subroutine
            subroutine W(cv,ip,flag,realargs,intargs)
                use mod_types
                implicit none
                real(rt), intent(out) :: cv
                integer, intent(in) :: ip
                integer, intent(out) :: flag
                real(rt), intent(inout) :: realargs(*)
                integer, intent(inout) :: intargs(*)
            end subroutine
        end interface
        ! local
        integer :: ni

        if (size(opt)<2) STOP 'ERROR: bellman: opt must be at least length 2'
        if (any(opt(1:2)/=[2,0])) STOP 'ERROR: bellman: requested optimization not programmed' 

        ni = size(V)

        ! Compute the optimum at the endpoints
        call brute(1,1,nip) ! computes V(1),g(1)
        call brute(ni,g(1),nip) ! computes V(ni),g(ni) (assumes mono since lowerbound is g(1)

!         print*,1,V(1),g(1)
!         print*,ni,V(ni),g(ni)

        ! Uses binary monotonicity
        call binarymono(1,ni)

    contains

        ! Computes V(i) and g(i) for all i between ia and ib assuming g(ia) and g(ib) are known
        recursive subroutine binarymono(ia,ib)
            implicit none
            integer, intent(in) :: ia,ib
            ! local
            real(rt) :: putmp,cvtmp
            integer :: i,ip,im,flag


            if (ib<=ia+1) then
                ! Since g(ia) and g(ib) are assumed known, desired values are known
                return
            elseif (g(ia)==g(ib)) then

                ! Now know the entire policy, so fill out V and g
                ip = g(ia)

                call W(cvtmp,ip,flag,realargs,intargs)
                do i = ia+1,ib-1

                    call u(putmp,i,ip,flag,realargs,intargs)
                    V(i) = putmp + cvtmp
                    g(i) = ip

                end do
                
            else

                ! Evaluate the midpoint
                im = (ia+ib)/2
                call brute(im,g(ia),g(ib)) ! computes V(im),g(im)
!                 print*,im,V(im),g(im)

                ! Divide and conquer
                call binarymono(ia,im)
                call binarymono(im,ib)

            end if

        end subroutine

        ! Computes V(i),g(i) by doing a brute force search from ipa to ipb
        subroutine brute(i,ipa,ipb)
            implicit none
            integer, intent(in) :: i,ipa,ipb
            ! local
            real(rt) :: Vtmp,putmp,cvtmp
            integer :: ip,flag

            

            V(i) = -huge(0e0_rt)
            g(i) = 1
            do ip = ipa,ipb
                call u(putmp,i,ip,flag,realargs,intargs)
                call W(cvtmp,ip,flag,realargs,intargs)
                Vtmp = putmp + cvtmp
                if (Vtmp>V(i)) then
                    V(i) = Vtmp
                    g(i) = ip
                end if
            end do

        end subroutine

    end subroutine

    subroutine setPUcustom(PU)
        implicit none
        procedure(PUcustom_template) :: PU

        PUcustom => PU

    end subroutine

    ! Check that a function W(a',e) defined as the linear interpolant of 
    ! W at knots gAp is concave for each e. To check that it is globally 
    ! concave may computationally intensive because for each interval [a_i,a_j]
    ! all intermediate points k s.t. i<k<j must be checked. Instead, this routine
    ! just checks a random point and asks whether it lies above all the
    ! convex combinations. This can still be slow, but is much quicker than 
    ! checking all the combinations.
    function randCheckLinInterpConcave(W,gAP,lenAp,lenE) result(isconcave)
        use mod_matlab
        implicit none
        logical :: isconcave
        integer, intent(in) :: lenAp,lenE
        real(rt), intent(in) :: W(lenAp,*)
        real(rt), intent(in) :: gAP(lenAp)
        ! local
        real(rt) :: theta, aprand, Wrand
        integer :: iE,i,j, ilo

        ! Get the random value
        call random_number(aprand)
        aprand = gAP(1) + aprand*(gAP(lenAP)-gAP(1))

        call bsearch(ilo,aprand,gAP)

        ! Check it
        do iE = 1,lenE

            Wrand = interp1(gAP,W(:,iE),aprand)

            do i = 1,ilo
                do j = ilo+1,lenAp
                    ! The convex combination of X(i) and X(j) must lie below X(k)
                    theta = (aprand-gAp(i))/(gAp(j)-gAp(i))
                    if (Wrand<(1e0_rt-theta)*W(i,iE) + theta*W(j,iE) - 2e0_rt*abs(Wrand)*epsilon(0e0_rt)) then
                        ! not concave
                        print*,'randCheckLinInterpConcave: W is not concave'
                        print*,'aprand,Wrand: ',aprand,Wrand
                        print*,'gAP(i),W(i): ',gAP(i),W(i,iE)
                        print*,'gAP(j),W(j): ',gAP(j),W(j,iE)
                        print*,'Convex combination: ',(1e0_rt-theta)*W(i,iE) + theta*W(j,iE)
                        isconcave = .false.                            
                        return
                    end if
                end do
            end do
        end do
        isconcave = .true.

    end function 

    ! Check that a function W(a',e) defined as the linear interpolant of 
    ! W at knots gAp is concave for each e. To check that it is globally 
    ! concave may computationally intensive because for each interval [a_i,a_j]
    ! all intermediate points k s.t. i<k<j must be checked
    function checkLinInterpConcave(W,gAP,lenAp,lenE) result(isconcave)
        implicit none
        logical :: isconcave
        integer, intent(in) :: lenAp,lenE
        real(rt), intent(in) :: W(lenAp,*)
        real(rt), intent(in) :: gAP(lenAp)
        ! local
        real(rt) :: theta
        integer :: iE,i,j,k

        do iE = 1,lenE
            do i = 1,lenAp
                do j = lenAp,i+1,-1
                    do k = i+1,j-1
                        ! The convex combination of X(i) and X(j) must lie below X(k)
                        theta = (gAp(k)-gAp(i))/(gAp(j)-gAp(i))
                        if (W(k,iE)<(1e0_rt-theta)*W(i,iE) + theta*W(j,iE)) then
                            ! not concave
                            isconcave = .false.                            
                            return
                        end if
                    end do
                end do
            end do
        end do
        isconcave = .true.

    end function 

!     ! Both sub_bellman_discrete and sub_bellman_continuous solve
!     ! V(a,e) = max_{a'} u(y(a,e) - x(a',e)) + W(a',e)
!     ! Here, y(a,e) is income and x(a',e) could be q(a',e)a' as in CCNR
!     ! or just a' as in a RBC model. Options may be specified to control
!     ! various speedups and optimization options. 
!     ! 
!     ! Note: code is written to handle any problem with two or more dimensions
!     ! in the sense that any dimensions, other than the first, are treated as 
!     ! a single dimension. 
!     !
!     ! The code is threaded.
!     ! 
!     subroutine sub_policyiter_continuous(V,c,ap,W,y,x,gAP,lenA,lenE,lenAp,crra,opt,xopt)
!         use mod_matlab, only: bsearch, disp, interp1q
!         implicit none
!         integer, intent(in) :: lenA,lenE,lenAp
!         real(rt), dimension(lenA,*), intent(out) :: V
!         real(rt), dimension(lenA,*), intent(in) :: c,ap
!         real(rt), dimension(lenA,*), intent(in) :: y
!         real(rt), dimension(lenAp,*), intent(in) :: W,x
!         real(rt), dimension(lenAp), intent(in) :: gAP ! grid for AP
!         real(rt), intent(in) :: crra
!         integer, intent(in) :: opt(1:)
!         real(rt), intent(in) :: xopt(1:)
!         ! local
!         real(rt) :: tmp,step,work(3)
!         integer :: iA,iE,iwork(1)
! 
!         step = real(lenAp-1,8)/(gAp(lenAp)-gAp(1))
! 
!         STOP ' I need to update policiter_continuous if I want to use it '
! 
!         !$OMP PARALLEL DO DEFAULT(SHARED) PRIVATE(iA,iE,tmp,work,iwork)
!         do iE = 1,lenE
!             do iA = 1,lenA
!                 call getH(V(iA,iE),tmp,ap(iA,iE),iA,iE,W,y,x,gAP,step,crra,lenA,lenAp,work,iwork,opt)
!             end do
!         end do
!         !$OMP END PARALLEL DO
! 
!     end subroutine sub_policyiter_continuous

    ! Continuous case: linear interpolation

    ! Model options, governed by opt(1)
    ! 0         ERROR         
    ! 1         x(a',e) is increasing in a'
    !
    ! Speedup options, governed by opt(2)
    ! 0         ERROR
    ! 1         ERROR
    ! 2         Monotonicity and concavity assumption
    ! 
    ! Speedup options, governed by opt(3): NOTE: this is not currently used
    ! 0         Arbitrarily spaced gAP grid
    ! 1         Linearly spaced gAP grid
    ! 
    ! Interpolation options, governed by opt(4)
    ! 0         Linear interpolation
    ! 1         Schumaker shape-preserving spline interpolation
    ! 
    ! Tolerance, governed by xopt(1)
    ! xopt(1)   Termination critera, stop when bracket length<xopt(1)
    ! 
    subroutine sub_bellman_continuous_4d(V,c,ap,W,y,x,gAP,crra,opt,xopt)
        real(rt), dimension(:,:,:,:), intent(out) :: V,c
        real(rt), dimension(:,:,:,:), intent(out) :: ap
        real(rt), dimension(:,:,:,:), intent(in) :: y,W,x
        real(rt), dimension(:), intent(in) :: gAP ! grid for AP
        real(rt), intent(in) :: crra
        integer, intent(in) :: opt(1:)
        real(rt), intent(in) :: xopt(1:)
        ! local
        integer :: lenA,lenE1,lenE2,lenE3,lenAp

        lenA = size(V,1)
        lenE1 = size(V,2)
        lenE2 = size(V,3)
        lenE3 = size(V,4)
        lenAp = size(W,1)

        call checkShape(V,lenA,lenE1,lenE2,lenE3)
        call checkShape(c,lenA,lenE1,lenE2,lenE3)
        call checkShape(y,lenA,lenE1,lenE2,lenE3)
        call checkShape(x,lenAp,lenE1,lenE2,lenE3)
        call checkShape(W,lenAp,lenE1,lenE2,lenE3)
        call checkShape(ap,lenA,lenE1,lenE2,lenE3)
        
        call sub_bellman_continuous_core(V,c,ap,W,y,x,gAP,lenA,lenE1*lenE2*lenE3,lenAp,crra,opt,xopt)

    end subroutine 

    subroutine sub_bellman_continuous_3d(V,c,ap,W,y,x,gAP,crra,opt,xopt)
        real(rt), dimension(:,:,:), intent(out) :: V,c
        real(rt), dimension(:,:,:), intent(out) :: ap
        real(rt), dimension(:,:,:), intent(in) :: y,W,x
        real(rt), dimension(:), intent(in) :: gAP ! grid for AP
        real(rt), intent(in) :: crra
        integer, intent(in) :: opt(1:)
        real(rt), intent(in) :: xopt(1:)
        ! local
        integer :: lenA,lenE1,lenE2,lenAp

        lenA = size(V,1)
        lenE1 = size(V,2)
        lenE2 = size(V,3)
        lenAp = size(W,1)

        call checkShape(V,lenA,lenE1,lenE2)
        call checkShape(c,lenA,lenE1,lenE2)
        call checkShape(y,lenA,lenE1,lenE2)
        call checkShape(x,lenAp,lenE1,lenE2)
        call checkShape(W,lenAp,lenE1,lenE2)
        call checkShape(ap,lenA,lenE1,lenE2)
        
        call sub_bellman_continuous_core(V,c,ap,W,y,x,gAP,lenA,lenE1*lenE2,lenAp,crra,opt,xopt)

    end subroutine 

    subroutine sub_bellman_continuous_2d(V,c,ap,W,y,x,gAP,crra,opt,xopt)
        real(rt), dimension(:,:), intent(out) :: V,c
        real(rt), dimension(:,:), intent(out) :: ap
        real(rt), dimension(:,:), intent(in) :: y,W,x
        real(rt), dimension(:), intent(in) :: gAP ! grid for AP
        real(rt), intent(in) :: crra
        integer, intent(in) :: opt(1:)
        real(rt), intent(in) :: xopt(1:)
        ! local
        integer :: lenA,lenE,lenAp

        lenA = size(V,1)
        lenE = size(V,2)
        lenAp = size(W,1)

        call checkShape(V,lenA,lenE)
        call checkShape(c,lenA,lenE)
        call checkShape(y,lenA,lenE)
        call checkShape(x,lenAp,lenE)
        call checkShape(W,lenAp,lenE)
        call checkShape(ap,lenA,lenE)
        
        call sub_bellman_continuous_core(V,c,ap,W,y,x,gAP,lenA,lenE,lenAp,crra,opt,xopt)

    end subroutine 

    subroutine sub_bellman_continuous_core(V,c,ap,W,y,x,gAP,lenA,lenE,lenAp,crra,opt,xopt)
        use mod_matlab, only: bsearch, disp, interp1q
        use mod_interp, only: isearch
        use mod_spline
        use mod_min1dim
        implicit none
        integer, intent(in) :: lenA,lenE,lenAp
        real(rt), dimension(lenA,*), intent(out) :: V,c
        real(rt), dimension(lenA,*), intent(out) :: ap
        real(rt), dimension(lenA,*), intent(in) :: y
        real(rt), dimension(lenAp,*), intent(in), target :: W,x
        real(rt), dimension(lenAp), intent(in), target :: gAP ! grid for AP
        real(rt), intent(in) :: crra
        integer, intent(in) :: opt(1:)
        real(rt), intent(in) :: xopt(1:)
        ! local 
        integer, allocatable :: api(:,:)
        integer :: iA,iE,iAp_lb,iAp_ub,iAp
        real(rt) :: apmin,apmax
        logical :: isSchumaker


        if (size(xopt)<1) STOP 'sub_bellman_continuous: xopt must have length at least 1'
        if (xopt(1)<=2e0_rt*epsilon(0e0_rt)) STOP 'sub_bellman_continuous: xopt is set to small, should be >=eps'
        if (size(opt)<4) STOP 'sub_bellman_continuous: opt must have length at least 4'
        if (opt(2)/=2) STOP 'sub_bellman_continuous: concavity must be assumed'
        if (opt(1)/=1) STOP 'sub_bellman_continuous: x must be increasing in a'

        isSchumaker = opt(4)==1

        allocate(api(lenA,lenE))

        ! First, find the best value on the grid. Then, since objective function is concave,
        ! the optimum is bracketed in an interval + or - 1 index from the optimum. Since 
        ! concavity is assumed and it is much cheaper to check values on the grid than off
        ! the grid (since no interpolation is involved), this is extremely quick.
        call sub_bellman_discrete_core(V,c,api,W,y,x,lenA,lenE,lenAp,crra,[opt(1:2),1])

        ! api now has the optimum. 


!!! I THOUGHT THE OMP CODE SHOULD WORK, BUT SUPNORM IS SLIGHTLY DIFFERENT. AVOID UNTIL 
!!! HAVE TIME TO PROPERLY TEST IT. 
        !$OMP PARALLEL DO DEFAULT(SHARED) PRIVATE(iA,iE,iAp_lb,iAp_ub,iAp,apmin,apmax)
        do iE = 1,lenE

            ! Compute a spline of the continuation value if doing schumaker
            if (isSchumaker) then
                call schumaker(negV_pp,gAp,W(:,iE))
            end if

            ! Let negV know what isSchumaker is
            negV_isSchumaker = isSchumaker

            do iA = 1,lenA

                ! Note: need to determine feasible region, o/w the flat region created
                ! by deathv and result in multiple local minima.

                ! First, need to find a bracket that is feasible at both end points
                ! Since x is increasing in a', check that iAp=1 is feasible. O/w, nothing is.
                if (y(iA,iE)<x(1,iE)) then
                    ! Consumption is always negative
                    V(iA,iE) = deathv
                    c(iA,iE) = deathc
                    ap(iA,iE) = deathap
                    print*,'WARNING: no feasible choice found'
                    cycle
                else
                    apmin = gAP(1)
                end if

                ! Now, have to find a' such that c(a') = y-x(a') = 0 
                ! Begin by iterating from the top down and stopping if c>0. Since x is increasing
                ! in a', then c will be positive from gAP(1) to a'
                !!! Note: need to find iAp such that c(iAp)>=0e0_rt and c(iAp+1)<=0e0_rt if one exists. Equivalently,
                !!! find iAp such that y>=x(iAp) and y<=x(iAp+1). The routine isearch will find such an iAp, 
                !!! First though, check whether y>x(nAp) in which case all iAp are feasible.
                if (y(iA,iE)>x(lenAp,iE)) then
                    iAp = lenAp
                ! elseif (y(iA,iE)<x(1,iE)) then ! NOTE: this case has already been checked above
                else
                    call isearch(iAp,y(iA,iE),x(:,iE))
                end if
                
                if (iAp==lenAp) then
                    ! Entire choice set is feasible
                    apmax = gAP(lenAp)
                else
                    ! c > 0 at iAp but is <= at iAp+1
                    ! Now, need to find exact end point [aphi,gAp(iAp+1)]
                    ! y(iA,iE) = xval AND
                    ! xval = x(iAp,iE) + (x(iAp+1,iE)-x(iAp,iE))/(gAp(iAp+1)-gAp(iAp))*(ap-gAp(iAp))
                    ! <=> (y(iA,iE)-x(iAp,iE)) = (x(iAp+1,iE)-x(iAp,iE))/(gAp(iAp+1)-gAp(iAp))*(ap-gAp(iAp))
                    ! <=> (y(iA,iE)-x(iAp,iE))/(x(iAp+1,iE)-x(iAp,iE))*(gAp(iAp+1)-gAp(iAp)) = (ap-gAp(iAp))
                    ! <=> gAp(iAp) + (y(iA,iE)-x(iAp,iE))/(x(iAp+1,iE)-x(iAp,iE))*(gAp(iAp+1)-gAp(iAp)) = ap
                    apmax = gAp(iAp) + (y(iA,iE)-x(iAp,iE))/(x(iAp+1,iE)-x(iAp,iE))*(gAp(iAp+1)-gAp(iAp))
                end if

                ! Find the optimum using brent's method within the interval [lb,ub]
                iAp_lb = max(api(iA,iE)-1,1)
                iAp_ub = min(api(iA,iE)+1,lenAp)


                negV_crra = crra ! just a scalar, copy
                negV_y = y(iA,iE) ! "
                negV_gAp => gAp(iAp_lb:iAp_ub) ! variable length vectors, avoid copy
                negV_W => W(iAp_lb:iAp_ub,iE) ! "
                negV_x => x(iAp_lb:iAp_ub,iE) ! "

!                 print*,'negV_y',negV_y
!                 print*,'negV_W',negV_W
!                 print*,'negV_x',negV_x
!                 print*,'negV_gAp',negV_gAp
! 
!                 call negV_core(V(iA,iE),c(iA,iE),.5e0_rt)
!                 print*,'V',V(iA,iE)
!                 print*,'c',c(iA,iE)

                ! Use the minimization routine by brent. NOTE: only searching feasible region [apmin,apmax]
                apmin = max(gAp(iAp_lb),apmin)
                apmax = min(gAp(iAp_ub),apmax)
                call sub_brent_min(ap(iA,iE),negV,apmin,apmax,printmod_in=-1,xtoler_in=1e-8_rt)

                ! Evaluate last time at the optimum to get value and consumption policy
                call negV_core(V(iA,iE),c(iA,iE),ap(iA,iE))

            end do
        end do
        !$OMP END PARALLEL DO

!         STOP 'checking sub_brent_min'

    end subroutine sub_bellman_continuous_core

    ! Compute the negative of u(c)+W for ap 
    function negV(ap)
        implicit none
        real(rt), intent(in) :: ap
        real(rt) :: negV
        ! local
        real(rt) :: c,V

        call negV_core(V,c,ap)
        negV = -V

    end function 
    subroutine negV_core(V,c,ap) 
        use mod_interp, only: bsearch
        use mod_spline
        implicit none           
        real(rt), intent(in) :: ap
        real(rt), intent(out) :: V,c
        ! local
        integer :: ilo 
        real(rt) :: whi,xval,Wval
            
        ! Find the location of ap in the grid
        call bsearch(ilo,ap,negV_gAp)

        ! Interpolate to get x(a',e) and W(a',e)
        whi = (ap-negV_gAP(ilo))/(negV_gAP(ilo+1)-negV_gAP(ilo))
        xval = negV_x(ilo) + whi*(negV_x(ilo+1)-negV_x(ilo))

        if (negV_isSchumaker) then
            ! Evaluate the spline
            call ppval(Wval,ap,negV_pp)
        else
            Wval = negV_W(ilo) + whi*(negV_W(ilo+1)-negV_W(ilo))
        end if
! 
! print*,'ilo',ilo
! print*,'xval',xval
! print*,'Wval',Wval

        ! Evaluate consumption and utility
        c = negV_y - xval
! print*,'c',c
        if (c>0e0_rt) then
            if (negV_crra==2e0_rt) then
                V = -1e0_rt/c + Wval
            elseif (negV_crra==1e0_rt) then
                V = log(c) + Wval
            else
                V = c**(1e0_rt-negV_crra)/(1e0_rt-negV_crra) + Wval
            end if
        else
            V = deathv
            c = deathc
        end if
! print*,'V',V

    end subroutine  

!     subroutine getH(H,c,ap,iA,iE,W,y,x,gAP,step,crra,lenA,lenAp,work,iwork,opt) 
!         use mod_interp, only: bsearch,isearch_with_step
!         implicit none           
!         real(rt), intent(out) :: H,c
!         real(rt), intent(in) :: ap,step,crra
!         integer, intent(in) :: iA,iE
!         integer, intent(in) :: lenA,lenAp
!         real(rt), dimension(lenA,*), intent(in) :: y
!         real(rt), dimension(lenAp,*), intent(in) :: W,x
!         real(rt), dimension(lenAp), intent(in) :: gAP 
!         real(rt), intent(out) :: work(3)
!         integer, intent(out) :: iwork(1)
!         integer, intent(in) :: opt(1:)
! 
! ! integer :: itmp
!         associate (whi=>work(1), xval=>work(2), Wval=>work(3), ilo=>iwork(1))
! 
!         ! Get position of ap in gAP
!         if (opt(3)==0) then
!             ! Arbitrarily spaced grid
!             call bsearch(ilo,ap,gAP)
!         else
!             ! Linearly spaced grid
!             ! if gAP is linearly spaced, will find it in one iteration, 
!             ! but is robust to non-linear spacing.
!             call isearch_with_step(ilo,ap,gAP,step) 
!         end if
! 
!         ! Interpolate to get x(a',e) and W(a',e)
!         whi = (ap-gAP(ilo))/(gAP(ilo+1)-gAP(ilo))
!         xval = x(ilo,iE) + whi*(x(ilo+1,iE)-x(ilo,iE))
!         Wval = W(ilo,iE) + whi*(W(ilo+1,iE)-W(ilo,iE))
! 
!         ! Evaluate consumption and utility
!         c = y(iA,iE) - xval
!         if (c>0e0_rt) then
!             if (crra==2e0_rt) then
!                 H = -1e0_rt/c + Wval
!             elseif (crra==1e0_rt) then
!                 H = log(c) + Wval
!             else
!                 H = c**(1e0_rt-crra)/(1e0_rt-crra) + Wval
!             end if
!         else
!             H = deathv
!             c = deathc
!         end if
! 
!         end associate
! 
!     end subroutine  


    ! Grid search with interpolation
    ! 
    ! Identical to sub_bellman_discrete except that it first interpolates W and x
    ! from gAP onto gAPcheck and optimizes over the gAPfine grid. 
    ! Interpolation can be done via either linear interpolation or Schumaker shape-preserving
    ! splines. 
    ! 
    ! Interpolation options, governed by opt(1)
    ! 0         Linear interpolation
    ! XXX       Schumaker spline interpolation for W, linear for X (not implemented yet)
    ! XXX       Schumaker for both W and X (not implemented yet)
    !    
    ! Model options, governed by opt(2)
    ! 0         x is general form 
    ! 1         x is increasing in a' for each e (speedup available)
    !
    ! Speedup options, governed by opt(3)
    ! 0         Brute force grid search
    ! 1         Monotonicity assumption a'(a,e) monotone in a
    ! 2         Monotonicity and concavity assumption
    !
    ! Parallel options, governed by opt(4)
    ! 0         No parallelization
    ! 1         Parallelized over the e values using openmp
    ! 
    ! NOTE: gAP is assumed to be increasing so fast interpolation can be done
    ! NOTE: the concavity assumption is exploited using binary search (Heer and Maussner)
    ! NOTE: if concavity is assumed, then x must be increasing in a'
    subroutine sub_bellman_interp_4d(V,c,ap,W,y,x,gAP,gAPfine,crra,opt)
        implicit none
        real(rt), dimension(:,:,:,:), intent(out) :: V,c,ap
        real(rt), dimension(:,:,:,:), intent(in) :: y,W,x
        real(rt), dimension(:), intent(in) :: gAP,gAPfine
        real(rt), intent(in) :: crra
        integer, intent(in) :: opt(:)
        ! local
        integer :: lenA,lenE1,lenE2,lenE3,lenAp,lenApfine

        lenA = size(V,1)
        lenE1 = size(V,2)
        lenE2 = size(V,3)
        lenE3 = size(V,4)
        lenAp = size(W,1)
        lenApfine = size(gAPfine)

        call checkShape(V,lenA,lenE1,lenE2,lenE3)
        call checkShape(c,lenA,lenE1,lenE2,lenE3)
        call checkShape(y,lenA,lenE1,lenE2,lenE3)
        call checkShape(x,lenAp,lenE1,lenE2,lenE3)
        call checkShape(W,lenAp,lenE1,lenE2,lenE3)
        call checkShape(ap,lenA,lenE1,lenE2,lenE3)
        call checkShape(gAP,lenAp)
        call checkShape(opt,4)
        
        call sub_bellman_interp_core(V,c,ap,W,y,x,gAP,gAPfine,lenA,lenE1*lenE2*lenE3,lenAp,crra,opt)

    end subroutine 
    subroutine sub_bellman_interp_3d(V,c,ap,W,y,x,gAP,gAPfine,crra,opt)
        implicit none
        real(rt), dimension(:,:,:), intent(out) :: V,c,ap
        real(rt), dimension(:,:,:), intent(in) :: y,W,x
        real(rt), dimension(:), intent(in) :: gAP,gAPfine
        real(rt), intent(in) :: crra
        integer, intent(in) :: opt(:)
        ! local
        integer :: lenA,lenE1,lenE2,lenAp,lenApfine

        lenA = size(V,1)
        lenE1 = size(V,2)
        lenE2 = size(V,3)
        lenAp = size(W,1)
        lenApfine = size(gAPfine)

        call checkShape(V,lenA,lenE1,lenE2)
        call checkShape(c,lenA,lenE1,lenE2)
        call checkShape(y,lenA,lenE1,lenE2)
        call checkShape(x,lenAp,lenE1,lenE2)
        call checkShape(W,lenAp,lenE1,lenE2)
        call checkShape(ap,lenA,lenE1,lenE2)
        call checkShape(gAP,lenAp)
        call checkShape(opt,4)
        
        call sub_bellman_interp_core(V,c,ap,W,y,x,gAP,gAPfine,lenA,lenE1*lenE2,lenAp,crra,opt)

    end subroutine 
    subroutine sub_bellman_interp_2d(V,c,ap,W,y,x,gAP,gAPfine,crra,opt)
        implicit none
        real(rt), dimension(:,:), intent(out) :: V,c,ap
        real(rt), dimension(:,:), intent(in) :: y,W,x
        real(rt), dimension(:), intent(in) :: gAP,gAPfine
        real(rt), intent(in) :: crra
        integer, intent(in) :: opt(:)
        ! local
        integer :: lenA,lenE1,lenAp,lenApfine

        lenA = size(V,1)
        lenE1 = size(V,2)
        lenAp = size(W,1)
        lenApfine = size(gAPfine)

        call checkShape(V,lenA,lenE1)
        call checkShape(c,lenA,lenE1)
        call checkShape(y,lenA,lenE1)
        call checkShape(x,lenAp,lenE1)
        call checkShape(W,lenAp,lenE1)
        call checkShape(ap,lenA,lenE1)
        call checkShape(gAP,lenAp)
        call checkShape(opt,4)
        
        call sub_bellman_interp_core(V,c,ap,W,y,x,gAP,gAPfine,lenA,lenE1,lenAp,crra,opt)

    end subroutine 
    subroutine sub_bellman_interp_1d(V,c,ap,W,y,x,gAP,gAPfine,crra,opt)
        implicit none
        real(rt), dimension(:), intent(out) :: V,c,ap
        real(rt), dimension(:), intent(in) :: y,W,x
        real(rt), dimension(:), intent(in) :: gAP,gAPfine
        real(rt), intent(in) :: crra
        integer, intent(in) :: opt(:)
        ! local
        integer :: lenA,lenAp,lenApfine

        lenA = size(V,1)
        lenAp = size(W,1)
        lenApfine = size(gAPfine)

        call checkShape(V,lenA)
        call checkShape(c,lenA)
        call checkShape(y,lenA)
        call checkShape(x,lenAp)
        call checkShape(W,lenAp)
        call checkShape(ap,lenA)
        call checkShape(gAP,lenAp)
        call checkShape(opt,4)
        
        call sub_bellman_interp_core(V,c,ap,W,y,x,gAP,gAPfine,lenA,1,lenAp,crra,opt)

    end subroutine 

    subroutine sub_bellman_interp_core(V,c,ap,W,y,x,gAP,gAPfine,lenA,lenE,lenAp,crra,opt)
        use mod_matlab, only: bsearch, disp, interp1q
        use mod_interp, only: isearch
        use mod_spline
        use mod_min1dim
        implicit none
        integer, intent(in) :: lenA,lenE,lenAp
        real(rt), dimension(lenA,*), intent(out) :: V,c
        real(rt), dimension(lenA,*), intent(out) :: ap
        real(rt), dimension(lenA,*), intent(in) :: y
        real(rt), dimension(lenAp,*), intent(in) :: W,x
        real(rt), dimension(lenAp), intent(in) :: gAP ! grid for AP
        real(rt), dimension(:), intent(in) :: gAPfine ! interpolation grid
        real(rt), intent(in) :: crra
        integer, intent(in) :: opt(:)
        ! local 
        real(rt), allocatable :: Wfine(:,:), xfine(:,:)
        integer, allocatable :: api(:,:)
        integer :: lenAPfine,iE

        if (size(opt)<4) STOP 'sub_bellman_interp_core: opt must have length at least 5'

        lenAPfine = size(gAPfine)

        allocate(api(lenA,lenE))
        allocate(Wfine(lenAPfine,lenE))
        allocate(xfine(lenAPfine,lenE))

        Wfine = interp1q(gAP,W(:,1:lenE),gAPfine)
        xfine = interp1q(gAP,x(:,1:lenE),gAPfine)

        ! Call bellman discrete to do the actual optimization
        call sub_bellman_discrete_core(V,c,api,Wfine,y,xfine,lenA,lenE,lenAPfine,crra,opt(2:4))

        ! Recover the a' values
        do iE = 1,lenE
            where (c(:,iE)>0e0_rt) 
                ap(:,iE) = gAPfine(api(:,iE))
            elsewhere
                ap(:,iE) = deathap
            end where
        end do

    end subroutine sub_bellman_interp_core

    ! Discrete case: no interpolation
    !    
    ! Model options, governed by opt(1)
    ! 0         x is general form 
    ! 1         x is increasing in a' for each e (speedup available)
    !
    ! Speedup options, governed by opt(2)
    ! 0         Brute force grid search
    ! 1         Monotonicity assumption a'(a,e) monotone in a
    ! 2         Monotonicity and concavity assumption
    ! 
    ! Parallel options, governed by opt(3)
    ! 0         No parallelization
    ! 1         Parallelized over the e values using openmp
    !
    ! NOTE: the concavity assumption is exploited using binary search (Heer and Maussner)
    ! NOTE: if concavity is assumed, then x must be increasing in a'
    subroutine sub_bellman_discrete_4d(V,c,api,W,y,x,crra,opt)
        implicit none
        real(rt), dimension(:,:,:,:), intent(out) :: V,c
        integer, dimension(:,:,:,:), intent(out) :: api
        real(rt), dimension(:,:,:,:), intent(in) :: y,W,x
        real(rt), intent(in) :: crra
        integer, intent(in) :: opt(:)
        ! local
        integer :: lenA,lenE1,lenE2,lenE3,lenAp

        lenA = size(V,1)
        lenE1 = size(V,2)
        lenE2 = size(V,3)
        lenE3 = size(V,4)
        lenAp = size(W,1)

        call checkShape(V,lenA,lenE1,lenE2,lenE3)
        call checkShape(c,lenA,lenE1,lenE2,lenE3)
        call checkShape(y,lenA,lenE1,lenE2,lenE3)
        call checkShape(x,lenAp,lenE1,lenE2,lenE3)
        call checkShape(W,lenAp,lenE1,lenE2,lenE3)
        call checkShape(api,lenA,lenE1,lenE2,lenE3)
        
        call sub_bellman_discrete_core(V,c,api,W,y,x,lenA,lenE1*lenE2*lenE3,lenAp,crra,opt)

    end subroutine 
    subroutine sub_bellman_discrete_3d(V,c,api,W,y,x,crra,opt)
        implicit none
        real(rt), dimension(:,:,:), intent(out) :: V,c
        integer, dimension(:,:,:), intent(out) :: api
        real(rt), dimension(:,:,:), intent(in) :: y,W,x
        real(rt), intent(in) :: crra
        integer, intent(in) :: opt(:)
        ! local
        integer :: lenA,lenE1,lenE2,lenAp

        lenA = size(V,1)
        lenE1 = size(V,2)
        lenE2 = size(V,3)
        lenAp = size(W,1)

        call checkShape(V,lenA,lenE1,lenE2)
        call checkShape(c,lenA,lenE1,lenE2)
        call checkShape(y,lenA,lenE1,lenE2)
        call checkShape(x,lenAp,lenE1,lenE2)
        call checkShape(W,lenAp,lenE1,lenE2)
        call checkShape(api,lenA,lenE1,lenE2)
        
        call sub_bellman_discrete_core(V,c,api,W,y,x,lenA,lenE1*lenE2,lenAp,crra,opt)

    end subroutine 
    subroutine sub_bellman_discrete_2d(V,c,api,W,y,x,crra,opt)
        implicit none
        real(rt), dimension(:,:), intent(out) :: V,c
        integer, dimension(:,:), intent(out) :: api
        real(rt), dimension(:,:), intent(in) :: y,W,x
        real(rt), intent(in) :: crra
        integer, intent(in) :: opt(1:)
        ! local
        integer :: lenA,lenE,lenAp

        lenA = size(V,1)
        lenE = size(V,2)
        lenAp = size(W,1)

        call checkShape(V,lenA,lenE)
        call checkShape(c,lenA,lenE)
        call checkShape(y,lenA,lenE)
        call checkShape(x,lenAp,lenE)
        call checkShape(W,lenAp,lenE)
        call checkShape(api,lenA,lenE)

        call sub_bellman_discrete_core(V,c,api,W,y,x,lenA,lenE,lenAp,crra,opt)

    end subroutine 
    subroutine sub_bellman_discrete_1d(V,c,api,W,y,x,crra,opt)
        implicit none
        real(rt), dimension(:), intent(out) :: V,c
        integer, dimension(:), intent(out) :: api
        real(rt), dimension(:), intent(in) :: y,W,x
        real(rt), intent(in) :: crra
        integer, intent(in) :: opt(1:)
        ! local
        integer :: lenA,lenAp

        lenA = size(V,1)
        lenAp = size(W,1)

        call checkShape(V,lenA)
        call checkShape(c,lenA)
        call checkShape(y,lenA)
        call checkShape(x,lenAp)
        call checkShape(W,lenAp)
        call checkShape(api,lenA)

        call sub_bellman_discrete_core(V,c,api,W,y,x,lenA,1,lenAp,crra,opt)

    end subroutine 

!         subroutine checkShape(x,i,j)
!             implicit none
!             class(*), intent(in) :: x(:,:)
!             integer, intent(in) :: i,j
!             select type (a=>x)
!                 type is (real)
!                     if (any(shape(a)/=[i+1,j])) STOP 'ERROR: sub_bellman_discrete: incompatible dimensions'
!                 type is (integer)
!                     if (any(shape(a)/=[i+1,j])) STOP 'ERROR: sub_bellman_discrete: incompatible dimensions'
!             end select
!         end subroutine
    subroutine checkShapeReal4(x,i,j,k,l)
        implicit none
        real(rt), intent(in) :: x(:,:,:,:)
        integer, intent(in) :: i,j,k,l
        if (any(shape(x)/=[i,j,k,l])) STOP 'ERROR: bellman: incompatible dimensions'
    end subroutine
    subroutine checkShapeInt4(x,i,j,k,l)
        implicit none
        integer, intent(in) :: x(:,:,:,:)
        integer, intent(in) :: i,j,k,l
        if (any(shape(x)/=[i,j,k,l])) STOP 'ERROR: bellman: incompatible dimensions'
    end subroutine

    subroutine checkShapeReal3(x,i,j,k)
        implicit none
        real(rt), intent(in) :: x(:,:,:)
        integer, intent(in) :: i,j,k
        if (any(shape(x)/=[i,j,k])) STOP 'ERROR: bellman: incompatible dimensions'
    end subroutine
    subroutine checkShapeInt3(x,i,j,k)
        implicit none
        integer, intent(in) :: x(:,:,:)
        integer, intent(in) :: i,j,k
        if (any(shape(x)/=[i,j,k])) STOP 'ERROR: bellman: incompatible dimensions'
    end subroutine

    subroutine checkShapeReal2(x,i,j)
        implicit none
        real(rt), intent(in) :: x(:,:)
        integer, intent(in) :: i,j
        if (any(shape(x)/=[i,j])) STOP 'ERROR: bellman: incompatible dimensions'
    end subroutine
    subroutine checkShapeInt2(x,i,j)
        implicit none
        integer, intent(in) :: x(:,:)
        integer, intent(in) :: i,j
        if (any(shape(x)/=[i,j])) STOP 'ERROR: bellman: incompatible dimensions'
    end subroutine

    subroutine checkShapeReal(x,i)
        implicit none
        real(rt), intent(in) :: x(:)
        integer, intent(in) :: i
        if (any(shape(x)/=[i])) STOP 'ERROR: bellman: incompatible dimensions'
    end subroutine
    subroutine checkShapeInt(x,i)
        implicit none
        integer, intent(in) :: x(:)
        integer, intent(in) :: i
        if (any(shape(x)/=[i])) STOP 'ERROR: bellman: incompatible dimensions'
    end subroutine

    subroutine sub_bellman_discrete_core(V,c,api,W,y,x,lenA,lenE,lenAp,crra,opt)
        implicit none
        integer, intent(in) :: lenA,lenE,lenAp
        real(rt), dimension(lenA,*), intent(out) :: V,c
        integer, dimension(lenA,*), intent(out) :: api
        real(rt), dimension(lenA,*), intent(in) :: y
        real(rt), dimension(lenAp,*), intent(in) :: W,x
        real(rt), intent(in) :: crra
        integer, intent(in) :: opt(1:)
        ! local
        logical :: isConcave, isMono, xIsIncr, doOpenMP
        integer :: iE,iA,lastiap,lb,ub

        ! Check inputs
        if (size(opt)<3) then
            STOP 'Opt length must be at least three'
        end if

        ! Unpack the option values
        xIsIncr = opt(1)==1 ! whether x(a',e) is increasing in a'
        isMono = opt(2)>=1 ! whether a'(a,e) is mono in a
        isConcave = opt(2)==2 ! whether u+W is concave in a' for all e

        doOpenMP = opt(3)==1 ! use openmp

        if (isConcave .and. .not. xIsIncr) STOP 'ERROR: sub_bellman_discrete: x must be increasing if concave'

        !$OMP PARALLEL DO DEFAULT(SHARED) PRIVATE(iE,iA,lastiap,lb,ub) IF(doOpenMP)
        do iE = 1,lenE

            lastiap = 1

            do iA = 1,lenA

                ! Set defaults
                V(iA,iE) = deathv
                api(iA,iE) = deathapi
                c(iA,iE) = deathc

                if (.not. isConcave) then 

                    ! If not concave, then perform a grid search
                    call gridSearch(lastiap,lenAp,iA,iE)

                else

                    ! If concave and monotone, perform a binary search
                    lb = lastiap
                    ub = lenAp
                    do while (ub>lb)
                        call updateLbUb(lb,ub,iA,iE)
                    end do

                    ! Update the value and policy functions
                    if (lb/=ub) STOP 'ERROR: sub_bellman_discrete: logic did not work' 
                    call gridSearch(lb,ub,iA,iE) ! only checks one value

                end if

                ! If monotonicity assumption, then restrict search space 
                if (isMono) then
                    lastiap = max(api(iA,iE),1)
                end if

            end do
        end do
        !$OMP END PARALLEL DO

    contains

        ! Given a bound on the search space, lb and ub, replaces either lb or ub 
        ! with a midpoint such that [lb,ub] is cut roughly in half. It does this assuming
        ! concavity and checking the midpoint(s).
        subroutine updateLbUb(lb,ub,iA,iE)
            implicit none
            integer, intent(inout) :: lb,ub
            integer, intent(in) :: iA,iE
            ! local
            integer :: mid
            real(rt) :: clo,chi,vlo,vhi

            mid = floor(.5e0_rt*(lb+ub))

            ! Evaluate utility at midpoints
            call evalH(vlo,clo,mid,iA,iE)
            call evalH(vhi,chi,mid+1,iA,iE)

            if (clo>0e0_rt .and. chi>0e0_rt) then
                ! If both are feasible choices (typical case) then compare utility
                if (vlo>vhi) then
                    ub = mid
                else
                    lb = mid+1
                end if
            else
                ! If either is negative, then since x is increasing in iAp, know the optimum must be in 1,mid
                ub = mid
            end if
            
        end subroutine 

        ! Given bounds on search space, lb and ub, finds the index of a policy 
        ! resulting in V higher than the current V(iA,iE). If no such policy
        ! exists, V(iA,iE),c(iA,iE),api(iA,iE) are unchanged.
        subroutine gridSearch(lb,ub,iA,iE)
            implicit none
            integer, intent(in) :: lb,ub,iA,iE
            ! local
            integer :: iAp
            real(rt) :: cval,vval

            ! Search for the optimal policy
            do iAp = lb,ub

                ! Evaluate utility at a specific iAp
                call evalH(vval,cval,iAp,iA,iE)

                ! Check whether consumption is negative
                if (cval>0e0_rt) then
                    if (vval>V(iA,iE)) then
                        V(iA,iE) = vval
                        c(iA,iE) = cval
                        api(iA,iE) = iAp
                    end if
                elseif (xIsIncr) then
                    ! Since x is increasing in a', any subsequent values will result in lower c
                    exit
                end if

            end do

        end subroutine gridSearch

        ! Compute consumption and utility conditional on iA,iE for a given iAp. 
        ! If cval <= 0e0_rt, vval is not assigned a value.
        subroutine evalH(vval,cval,iAp,iA,iE)
            implicit none
            integer, intent(in) :: iAp,iA,iE
            real(rt), intent(out) :: cval,vval


            cval = y(iA,iE) - x(iAp,iE)  

            if (cval>0e0_rt) then

                ! Compute utility 
                if (crra==2e0_rt) then
                    vval = -1e0_rt/cval + W(iAp,iE)
                elseif (crra==1e0_rt) then
                    vval = log(cval) + W(iAp,iE)
                elseif (crra>0e0_rt) then
                    vval = cval**(1e0_rt-crra)/(1e0_rt-crra) + W(iAp,iE)
                else
                    vval = PUcustom(cval) + W(iAp,iE)
                end if

!                 ! Compute utility 
!                 if (crra==2e0_rt) then
!                     vval = -1e0_rt/(cval+1d-6) + W(iAp,iE)
!                 else
!                     STOP 'adding constant to utility'
!                 end if
            end if

        end subroutine

    end subroutine sub_bellman_discrete_core


    subroutine sub_bellman_discrete_choose2_1d(V,c,api,bpi,W,y,x,crra,opt)
        implicit none
        real(rt), dimension(:), intent(out) :: V,c
        integer, dimension(:), intent(out) :: api,bpi
        real(rt), dimension(:), intent(in) :: y
        real(rt), dimension(:,:), intent(in) :: W,x
        real(rt), intent(in) :: crra
        integer, intent(in) :: opt(1:)
        ! local
        integer :: lenA,lenAp,lenBp

        lenA = size(V,1)
        lenAp = size(W,1)
        lenBp = size(W,2)

        call checkShape(V,lenA)
        call checkShape(c,lenA)
        call checkShape(y,lenA)
        call checkShape(x,lenAp,lenBp)
        call checkShape(W,lenAp,lenBp)
        call checkShape(api,lenA)
        call checkShape(bpi,lenA)

        call sub_bellman_discrete_choose2_core(V,c,api,bpi,W,y,x,lenA,1,lenAp,lenBp,crra,opt)

    end subroutine 

    ! 2-dimensional optimization
    ! 
    ! Solves 
    !   V(a,e) = max_{a',b'} u(y(a,e)-x(a',b',e)) + W(a',b',e) for 
    ! a' and b' on grids (whose values need not be known).
    ! 
    ! Model options, governed by opt(1)
    ! 0         x is general form 
    ! 1         x is increasing in a' for each e (speedup available)
    ! 2 [same as 0] x is increasing in b' for each e
    ! 3 [same as 1] x is increasing in a' and b' for each e
    !
    ! Speedup options, governed by opt(2)
    ! 0         Brute force grid search
    ! 1         Monotonicity assumption a'(a,e) monotone in a
    ! 2         Monotonicity assumption b'(a,e) monotone in a
    ! 3         Monotonicity assumption a'(a,e) and b'(a,e) monotone in a
    ! XXX       Concavity Not programmed yet
    ! 
    subroutine sub_bellman_discrete_choose2_core(V,c,api,bpi,W,y,x,lenA,lenE,lenAp,lenBp,crra,opt)
        implicit none
        integer, intent(in) :: lenA,lenE,lenAp,lenBp
        real(rt), dimension(lenA,*), intent(out) :: V,c
        integer, dimension(lenA,*), intent(out) :: api,bpi
        real(rt), dimension(lenA,*), intent(in) :: y
        real(rt), dimension(lenAp,lenBp,*), intent(in) :: W,x
        real(rt), intent(in) :: crra
        integer, intent(in) :: opt(1:)
        ! local
        logical :: isMonoA, isMonoB, xIsIncrA, xIsIncrB ! isConcave
        integer :: iE,iA,lastiap,lastibp !,lb,ub

        ! Check inputs
        if (size(opt)<2) then
            STOP 'Opt length must be at least two'
        end if

        ! Unpack the option values
        xIsIncrA = opt(1)==1 .or. opt(1)==3 ! whether x(a',b',e) is increasing in a'
        xIsIncrB = opt(1)==2 .or. opt(1)==3 ! whether x(a',b',e) is increasing in b'
        isMonoA = opt(2)==1 .or. opt(2)==3 ! whether a'(a,e) is mono in a
        isMonoB = opt(2)==2 .or. opt(2)==3 ! whether b'(a,e) is mono in a

!         !$OMP PARALLEL DO DEFAULT(SHARED) PRIVATE(iE,iA,lastiap,lastibp)
    !!! CRAY compiler threw an error here saying xisincra is used before defined with openmp
        do iE = 1,lenE

            lastiap = 1
            lastibp = 1

            do iA = 1,lenA

                ! Set defaults
                V(iA,iE) = deathv
                api(iA,iE) = deathapi
                bpi(iA,iE) = deathapi
                c(iA,iE) = deathc

                ! If not concave, then perform a grid search
                call gridSearch(lastiap,lenAp,lastibp,lenBp,iA,iE)

                ! If monotonicity assumption, then restrict search space 
                if (isMonoA) lastiap = max(api(iA,iE),1)
                if (isMonoB) lastibp = max(bpi(iA,iE),1)

            end do
        end do
!         !$OMP END PARALLEL DO

    contains

        ! Given bounds on search space, lb and ub, finds the index of a policy 
        ! resulting in V higher than the current V(iA,iE). If no such policy
        ! exists, V(iA,iE),c(iA,iE),api(iA,iE),bpi(ia,iE) are unchanged.
        subroutine gridSearch(a_lb,a_ub,b_lb,b_ub,iA,iE)
            implicit none
            integer, intent(in) :: iA,iE
            integer, intent(in) :: a_lb,a_ub,b_lb,b_ub
            real(rt) :: cval,vval
            integer :: iAp,iBp

            ! Search for the optimal policy
            do iBp = b_lb,b_ub
                do iAp = a_lb,a_ub

                    ! Evaluate utility at a specific iAp
                    call evalH(vval,cval,iAp,iBp,iA,iE)

                    ! Check whether consumption is negative
                    if (cval>0e0_rt) then
                        if (vval>V(iA,iE)) then
                            V(iA,iE) = vval
                            c(iA,iE) = cval
                            api(iA,iE) = iAp
                            bpi(iA,iE) = iBp
                        end if
                    elseif (xIsIncrA) then
                        ! Since x is increasing in a', any subsequent values will result in lower c
                        exit
                    end if

                end do
            end do

        end subroutine gridSearch

        ! Compute consumption and utility conditional on iA,iE for a given iAp,iBp. 
        ! If cval <= 0e0_rt, vval is not assigned a value.
        subroutine evalH(vval,cval,iAp,iBp,iA,iE)
            implicit none
            integer, intent(in) :: iAp,iBp,iA,iE
            real(rt), intent(out) :: cval,vval

            cval = y(iA,iE) - x(iAp,iBp,iE)  

            if (cval>0e0_rt) then

                ! Compute utility 
                if (crra==2e0_rt) then
                    vval = -1e0_rt/cval + W(iAp,iBp,iE)
                elseif (crra==1e0_rt) then
                    vval = log(cval) + W(iAp,iBp,iE)
                elseif (crra>0e0_rt) then
                    vval = cval**(1e0_rt-crra)/(1e0_rt-crra) + W(iAp,iBp,iE)
                else
                    vval = PUcustom(cval) + W(iAp,iBp,iE)
                end if

            else

                vval = deathv

            end if

        end subroutine

    end subroutine sub_bellman_discrete_choose2_core

    subroutine sub_bellman_interp_choose2_1d(V,c,ap,bp,W,y,x,gAP,gAPfine,gBP,gBPfine,crra,opt)
        implicit none
        real(rt), dimension(:), intent(out) :: V,c,ap,bp
        real(rt), dimension(:), intent(in) :: y,gAP,gAPfine,gBP,gBPfine
        real(rt), dimension(:,:), intent(in) :: W,x
        real(rt), intent(in) :: crra
        integer, intent(in) :: opt(1:)
        ! local
        integer :: lenA,lenAp,lenBp

        lenA = size(V,1)
        lenAp = size(W,1)
        lenBp = size(W,2)

        call checkShape(V,lenA)
        call checkShape(c,lenA)
        call checkShape(y,lenA)
        call checkShape(x,lenAp,lenBp)
        call checkShape(W,lenAp,lenBp)
        call checkShape(ap,lenA)
        call checkShape(bp,lenA)
        call checkShape(gAP,lenAp)
        call checkShape(gBP,lenBp)

        call sub_bellman_interp_choose2_core(V,c,ap,bp,W,y,x,gAP,gAPfine,gBP,gBPfine,lenA,1,lenAp,lenBp,crra,opt)

    end subroutine 

    subroutine sub_bellman_interp_choose2_core(V,c,ap,bp,W,y,x,gAP,gAPfine,gBP,gBPfine,lenA,lenE,lenAp,lenBp,crra,opt)
        use mod_matlab, only: bsearch, disp, interp2q
        use mod_interp, only: isearch
        use mod_spline
        use mod_min1dim
        implicit none
        integer, intent(in) :: lenA,lenE,lenAp,lenBp
        real(rt), dimension(lenA,*), intent(out) :: V,c
        real(rt), dimension(lenA,*), intent(out) :: ap,bp
        real(rt), dimension(lenA,*), intent(in) :: y
        real(rt), dimension(lenAp,lenBp,*), intent(in) :: W,x
        real(rt), dimension(lenAp), intent(in) :: gAP
        real(rt), dimension(lenBp), intent(in) :: gBP 
        real(rt), dimension(:), intent(in) :: gAPfine, gBPfine ! interpolation grids
        real(rt), intent(in) :: crra
        integer, intent(in) :: opt(1:)
        ! local 
        real(rt), allocatable :: Wfine(:,:,:), xfine(:,:,:)
        integer, allocatable :: api(:,:),bpi(:,:)
        integer :: lenAPfine,lenBPfine,iE

        if (size(opt)<3) STOP 'sub_bellman_interp_core: opt must have length >= 3'

        lenAPfine = size(gAPfine)
        lenBPfine = size(gBPfine)

        allocate(api(lenA,lenE))
        allocate(bpi(lenA,lenE))
        allocate(Wfine(lenAPfine,lenBPfine,lenE))
        allocate(xfine(lenAPfine,lenBPfine,lenE))

        ! NOTE: it would probably be a lot faster to not use interp2 but
        ! rather manually do the interpolation and not call discrete_choose2. 
        ! For now, I don't pursue this for time's sake.
        Wfine = interp2q(gAP,gBP,W(:,:,1:lenE),gAPfine,gBPfine)
        xfine = interp2q(gAP,gBP,x(:,:,1:lenE),gAPfine,gBPfine)

        ! Call bellman discrete to do the actual optimization
        call sub_bellman_discrete_choose2_core(V,c,api,bpi,Wfine,y,xfine,lenA,lenE,lenAPfine,lenBPfine,crra,opt(2:3))

        ! Recover the a' values
        do iE = 1,lenE
            where (c(:,iE)>0e0_rt) 
                ap(:,iE) = gAPfine(api(:,iE))
                bp(:,iE) = gBPfine(bpi(:,iE))
            elsewhere
                ap(:,iE) = deathap
                bp(:,iE) = deathap
            end where
        end do

    end subroutine sub_bellman_interp_choose2_core

end module mod_valueopt


module mod_valueopt_test
    use mod_valueopt
    implicit none
    integer, parameter :: nK = 20, nZ = 3, maxiter = 3000 ! 1500
    ! integer, parameter :: nK = 2048, nZ = 64, maxiter = 3000 ! 1500
    real(rt), parameter :: alpha = .36e0_rt, beta = .99e0_rt, crra = 2e0_rt, deprec = .025e0_rt
    real(rt) :: Fzz(nZ,nZ),Z(nZ),K(nK),sup,lastsup,time(0:12),xopt(1)
    integer :: opt(5),iter,iZ,option
    real(rt), dimension(:,:), allocatable :: V,c,W,y,x,V0,kp,suphistory
    integer, dimension(:,:), allocatable :: kpi
contains

    subroutine u(pu,iK,iKp,flag,realargs,intargs)
        implicit none
        real(rt), intent(out) :: pu
        integer, intent(in) :: iK,iKp
        integer, intent(out) :: flag
        real(rt), intent(inout) :: realargs(*)
        integer, intent(inout) :: intargs(*)
        ! local
        real(rt) :: ctmp
        integer :: iZ

        iZ = intargs(1)

        ! Compute period utility
        ! ctmp = Z(iZ)*K(iK)**alpha + (1e0_rt-deprec)*K(iK) - K(iKp)
        ctmp = Y(iK,iZ) - K(iKp)
        if (ctmp>0e0_rt) then
            pu = -1e0_rt/ctmp
            flag = 0
        else
            pu = -sqrt(huge(0._rt))
            flag = 1
        end if

    end subroutine
    subroutine contutil(cv,iKp,flag,realargs,intargs)
        implicit none
        real(rt), intent(out) :: cv
        integer, intent(in) :: iKp
        integer, intent(out) :: flag
        real(rt), intent(inout) :: realargs(*)
        integer, intent(inout) :: intargs(*)
        ! local
        integer :: iZ

        iZ = intargs(1)

        ! Compute period utility
        cv = W(iKp,iZ)

        flag = 0

    end subroutine



    ! test_valueopt 
    ! Computes the solution to 
    ! V(k,z) = max u(c) + beta*E(V(k',z'))
    ! with c = -k' + z*k**alpha + (1e0_rt-deprec)*k
    subroutine test_valueopt
        use mod_discreteAR1
        use mod_matlab
        implicit none
        real(rt) :: realargs(1)
        integer :: intargs(1)

        allocate(V(nK,nZ))
        allocate(c(nK,nZ))
        allocate(W(nK,nZ))
        allocate(y(nK,nZ))
        allocate(x(nK,nZ))
        allocate(V0(nK,nZ))
        allocate(kp(nK,nZ))
        allocate(kpi(nK,nZ))
        allocate(suphistory(maxiter,0:12))

        ! Discretize the AR1
        call sub_rouwenhorst(Z,Fzz,.95e0_rt,.007e0_rt,nZ)
        Z = exp(Z)

!         call disp('Z',Z)
!         call disp('Fzz',Fzz)

        ! Get capital grid
        K = linspace(.7e0_rt,1.3e0_rt,nK)*((1e0_rt/beta+deprec-1e0_rt)/alpha)**(1e0_rt/(alpha-1e0_rt))

!         call disp('K',K)

        ! Compute y and x
        do iZ = 1,nZ
            y(:,iZ) = Z(iZ)*K**alpha + (1e0_rt-deprec)*K
            x(:,iZ) = K ! expenditure
        end do

        do option = 0,12

            ! Set opt according to which cycle we are doing
            select case (option)
                case (0)
                    print*,'new bellman: binary monotonicity'
               case (1)
                    print*,'Checking brute force w/ OpenMP'
                    opt(1:3) = [0,0,1]
                case (2)
                    print*,'Checking monotonicity w/ OpenMP'
                    opt(1:3) = [0,1,1]
                case (3)
                    print*,'Checking brute force w increasing x w/ OpenMP'
                    opt(1:3) = [1,0,1]
                case (4)
                    print*,'Checking monotonicity w increasing x w/ OpenMP'
                    opt(1:3) = [1,1,1]
                case (5)
                    print*,'Checking concavity w increasing x w/ OpenMP'
                    opt(1:3) = [1,2,1]
                case (6)
                    print*,'Checking brute force w/o OpenMP'
                    opt(1:3) = [0,0,0]
                case (7)
                    print*,'Checking monotonicity w/o OpenMP'
                    opt(1:3) = [0,1,0]
                case (8)
                    print*,'Checking brute force w increasing x w/o OpenMP'
                    opt(1:3) = [1,0,0]
                case (9)
                    print*,'Checking monotonicity w increasing x w/o OpenMP'
                    opt(1:3) = [1,1,0]
                case (10)
                    print*,'Checking concavity w increasing x w/o OpenMP'
                    opt(1:3) = [1,2,0]
                case (11)
                    print*,'Continuous: checking lininterp'
                    opt(1:4) = [1,2,1,0]
                    xopt(1) = sqrt(epsilon(0e0_rt))
                case (12)
                    print*,'Continuous: checking Schumaker'
                    opt(1:4) = [1,2,1,1]
                    xopt(1) = sqrt(epsilon(0e0_rt))
            end select

            suphistory(:,option) = -1e0_rt

            call tic()

            ! VFI 
            do iZ = 1,nZ
                V0(:,iZ) = 0e0_rt !log(K) ! - log(K(nK))! NOTE: guess must be strictly concave! 
            end do
            ! V0 = -1d-20
            do iter = 1,10000

                W = beta*matmul(V0,transpose(Fzz))

                if (option==0) then
                    !$OMP PARALLEL DO PRIVATE(iZ,intargs) DEFAULT(SHARED)
                    do iZ = 1,nZ
                        intargs(1) = iZ                        
                        call bellman(V(:,iZ),kpi(:,iZ),u,contutil,nK,[2,0],realargs,intargs)
                    end do
                    !$OMP END PARALLEL DO
                    ! call plot(V)
                    ! call plot(real(kpi,8))
                    ! if (iter==3) STOP 'check'
                elseif (option<=10) then
                    call sub_bellman_discrete(V,c,kpi,W,y,x,crra,opt)
                else
                    call sub_bellman_continuous(V,c,kp,W,y,x,K,crra,opt,xopt)
                end if

                ! Compute the supnorm
                if (iter>1) lastsup = sup

                sup = maxval(abs(V-V0)) 

                if (iter>1) then
                    if (sup/lastsup>beta+1e-3_rt*beta) then
                        print*,'option',option
                        print*,'iter',iter
                        print*,'sup',sup
                        print*,'lastsup',lastsup
                        print*,'sup/lastsup',sup/lastsup
                        print*,'beta',beta
                        print*,'ERROR: supnorm did not fall by beta'
                    end if
                end if
                suphistory(iter,option) = sup

                if (mod(iter,500)==0) print*,'iter',iter,'supnorm',sup

                if (sup<1e-3_rt*maxval(abs(V))) exit

                ! Update V
                V0 = V

            end do

            ! Check that the supnorm history agrees 
            if (option>0 .and. option<=10) then
                ! supnorm history should be completely the same in the discrete case
                if (any(suphistory(:,option)/=suphistory(:,option-1))) then 
                    STOP 'ERROR: supnorm history does not agree'
                end if
            end if

! call plot(K,V,'V')
! call plot(K,real(kpi,8),'kpi')
! 
            call toc(time(option))

!         STOP 'only checking valueopt option 0'

        end do

!         call plot(K,c,'c')
!         call plot(suphistory,'supnorm')

!         call disp('supnorm',suphistory)

        print*,'max diff in supnorm history from continuous and discrete'
        print*,maxval(abs(suphistory(:,1)-suphistory(:,11))),maxval(abs(suphistory(:,1)-suphistory(:,12)))
        print*,'max diff in supnorm history between two continuous methods'
        print*,maxval(abs(suphistory(:,11)-suphistory(:,12)))

        call disp('option times',time)

        print*,'test_valueopt: passed'
!         STOP 'only checking valueopt'

    end subroutine
end module 
