% Uses the homotopy to try to find an x s.t. f(x) = 0.
% The particular homotopy used is 
% H(x,t) = f(x) - (1-t)*f(start) where start is a user supplied (and
% ideally would be the true solution). 
% 
% Test code is at the bottom and may be called by passing no arguments.
% 
function [x] = homotopy(f_in,start,opt)

    if nargin==0
        test_homotopy();
        return
    end
    
    assert(nargin>=2,'Must provide f and an initial x value');
    
    sz_in = size(start);
    vec = @(y) y(:);
    f = @(x) vec(f_in(reshape(x,sz_in)));
    start = start(:);
    
    nx = prod(sz_in);
    ny = nx + 1;
    
    o.newton = 1; % Newton type homotopy if 1, otherwise fixed point
    o.greedy = 1;
    o.errtol = .001; % will try fsolve correction when L2 norm of H is bigger than errtol. To not have a correction, set to Inf
    o.maxiter = 10000;
    o.pmod = 1;    
    o.step = 1d-3;
    
    if nargin>=3 
        options = fieldnames(o);
        for i = 1:length(options)
            if isfield(opt,options{i})
                o.(options{i}) = opt.(options{i});
            end
        end
    end
    
    if o.newton
        %  Newton homotopy -- deforms f(x)-f(x0), which has a root at x0, into f(x)
        fstart = f(start);
        Hxt = @(x,t) f(x) - (1-t)*fstart; % This is the Newton homotopy / root-finding one
        dH_t = @(x,t) fstart;
        dH_x = @(x,t) numderiv(f,x);
        H = @(y) Hxt(y(1:nx),y(nx+1));    
        dH = @(y) [dH_x(y(1:nx),y(ny)) dH_t(y(1:nx),y(ny))];
    else
        % Fixed-point homotopy like in Judd -- deforms x-x0 into f(x)
        % Does not seem to work very well outside of his simple example
        % In fact, there may be something wrong with it. 
        Hxt = @(x,t) t*f(x) + (1-t)*(x-start);
        H = @(y) Hxt(y(1:nx),y(nx+1));    
        dH = @(y) numderiv(H,y);
    end
    
    y0 = [start(:);0];
    
    if exist('ode45')==2
        
        odeopt = odeset('Stats','on','OutputFcn',@myhomotopeplot,'Events',@checkstop);
        
        [~,yout] = ode45(@(s,y) dyds(y,dH),[0,Inf],y0,odeopt);
%         [~,yout] = ode23s(@(s,y) dyds(y,dH),[0,Inf],y0,odeopt);
%         [~,yout] = ode23t(@(s,y) dyds(y,dH),[0,Inf],y0,odeopt);
%         [~,yout] = ode15s(@(s,y) dyds(y,dH),[0,Inf],y0,odeopt);
        tout = yout(:,ny);
        plot(tout,yout(:,1:nx)); xlabel('t'); ylabel('x');
        x = yout(end,1:nx)';
        
%         [tout,xout] = ode45(@(t,x) dxdt([x;t],dH),[0,1],y0(1:nx),odeopt);
%         plot(tout,xout); xlabel('t'); ylabel('x')

    else
        
        % How much do we want to step along the path
        ds = o.step;

        % Begin path following        
        hist = NaN(o.maxiter,ny);
        for iter = 1:o.maxiter

            Hy = H(y0); % zero if exactly on the path

            if (mod(iter,o.pmod)==0 || y0(ny)>1)
                fprintf(1,'iter %6i t %f x norm %f H norm %f ',iter,y0(ny),norm(y0(1:nx),2),norm(Hy,2));        
                if (iter>1) 
                    fprintf(1,'lambda %3i',lambda);
                end
            end

            % If the error is bad, try to get the homotopy back on track
            if (norm(Hy,2)>o.errtol) 
                y0(1:nx) = fsolve(@(x) Hxt(x,y0(ny)),y0(1:nx),optimset('display','none'));
                Hy = H(y0); 
                if (mod(iter,o.pmod)==0 || y0(ny)>1)
                    fprintf(1,' fixed, new norm %f',norm(Hy,2));
                end
            end
            if (mod(iter,o.pmod)==0 || y0(ny)>1)
                fprintf(1,' \n');
            end

            if y0(ny)>1
                break
            end

            % Note: here I am using a very crude ODE solver like in Judd's
            % example. One might be able to use Runge-Kutta and it would likely 
            % be more efficient.
            dy = dyds(y0,dH);
%             dx = dxdt(y0,dH);
%             dt = 1d-3;
            
            % Increment have Euclidean norm of roughly ds
            ynext = y0 + ds*dy/norm(dy,2); 
%             ynext = y0 + [dx * dt; dt];
%             lambda = 1;

            % Try to be greedy by checking different path lengths
%             if o.greedy
%                 for magn = 1:5
%                     ytrial = y0 + 10^magn * ds*dy/norm(dy,2); 
%                     % When close to t=1, don't want to overshoot
%                     if (ytrial(ny)>1) 
%                         break
%                     end
%                     % Check whether the trial point has acceptable error 
%                     Hy = H(ytrial);
%                     if (norm(Hy,2)<o.errtol) 
%                         ynext = ytrial;
%                         lambda = 10^magn;
%                     else
%                         break
%                     end
%                 end
%             end
            y0 = ynext;

            hist(iter,:) = y0;

        end
        iter = min(iter,o.maxiter);

        x = y0(1:nx);

        plot(hist(:,ny),hist(:,1:nx),'-')
        xlabel('t')
        for i = 1:nx
            lab{i} = sprintf('x(%i)',i);
        end
        legend(lab{:})
        title('Path of variables over deformation')


    end
    
    x = reshape(x,sz_in);
    
end

function status = myhomotopeplot(~,y,flag)

    persistent ITER HIST NY LAB
    
    status = 0;

    if strcmp(flag,'init')
        ITER = 0;
        HIST = [];
        NY = length(y);
        for i = 1:NY-1
            LAB{i} = sprintf('x(%i)',i);
        end
        
        return
    end
    
    ITER = ITER + 1;

    if isempty(flag)
        
        fprintf(1,'iter %i ',ITER);
        fprintf(1,'t %f x',y(NY));
        fprintf(1,' %f',y(1:NY-1));
        fprintf(1,'\n');
        
        HIST = [HIST;y(:)'];
        if mod(ITER,1)==0
            plot(HIST(:,NY),HIST(:,1:NY-1),'-o')
            xlabel('t')
            ylabel('x')
            axis([-Inf 1 -Inf Inf]);
%             legend(LAB{:})
            
            drawnow
        end
%         hold on
%         plot(y(end),y(1:end-1),'o'); xlabel('t'); ylabel('x');
%         hold off
    end
    
    
end

function [position,isterminal,direction] = checkstop(s,y) 
    t = y(end); 
    position = t - 1; % want this to be 0
    isterminal = 1;
    direction = 0;
end
        
function [dx] = dxdt(y0,dH) 
    dy = dyds(y0,dH);
    ny = length(y0);
    nx = ny - 1;
    dx = dy(1:nx)/dy(ny);
end

function [dy] = dyds(y0,dH) 

    ny = length(y0);
    nx = ny - 1;

    % Get how the homotopy changes 
    dH0 = dH(y0);

    % Get how y changes implicitly as function of path variable s
    dy = NaN(ny,1);
    DD = NaN(nx,nx);
    for i = 1:ny
        DD(:,1:i-1) = dH0(:,1:i-1);
        DD(:,i:nx) = dH0(:,i+1:ny);
        dy(i) = (-1)^(i) * det(DD);
    end

end


function [fp] = numderiv(f,x)

    step = max(1d-6 * abs(x),1d-8);
    
    fx = f(x);

    fp = NaN(length(fx),length(x));
    
    for i = 1:length(x)
        e = zeros(length(x),1);
        e(i) = step(i);
        fp(:,i) = (f(x+e)-fx)/step(i);
    end

end

function [] = test_homotopy()

    hold off 
    close all
    
    opt.pmod = 200;
    opt.step = .001;

    % Judd's example
    fprintf(1,'Example #1 \n');
    
    f = @(x) 2*x - 4 + sin(2*pi*x);    
    xguess = 0;
    xstar = homotopy(f,xguess,opt);
    
    % Economic example
    fprintf(1,'\nExample #2 \n');
    crra = 25;
    w = 1.3;
    eta = 10;
    frisch = 3; 
    lump = .1;

    intra = @(c,l) c^(-crra) * w - eta * l.^(1/frisch);
    budg = @(c,l) - c + w * l  + .5 * w * max(0,l-.8).^2 + lump;
        
    f = @(x) [intra(x(1),x(2));budg(x(1),x(2))];

    xguess = [1;1];
    xsoln_fsolve = fsolve(f,xguess,optimset('display','none'));

    xsoln_homo = homotopy(f,xguess,opt);
    
    xsoln_fsolve
    xsoln_homo
    
    % 

    return

end
