% A = import_fortran_csv(file)
% 
% Returns an array A matching the description and data in the csv file. The
% csv file must be created by the corresponding fortran code. 
% 
% Note: Currently, this code will fail if a non-number appears in the data part, such
% as "infinity" or "NaN".
% 
% Grey Gordon, 2014 
function [A] = import_fortran_csv(file,varargin)

    maxMB = realmax;
    if (nargin>=2) 
        maxMB = varargin{1};
    end

    % Replace home directory if user passed a tilde
    inds = find(file(:)=='~');
    if (~isempty(inds))
        % Found this trick at http://www.mathworks.com/matlabcentral/fileexchange/15885-get-user-home-directory/content/etuserdir.m
        homedir = getenv('HOME');
        file = [file(1:inds-1),homedir,file(inds+1:end)];
    end

    % If file doesn't end in .csv, then add .csv
    if (~strcmp(file(max(1,end-3):end),'.csv')) 
        file = [file '.csv'];
    end

    rank = [];
    dims = [];
    
    % Check that this is a csv file written by fortran
    [fid msg] = fopen(file,'r');
    if (fid==-1)
        disp(['Failed to open ' file ', will return []. Error message was ' msg]);
        A = [];
        return
    end
    tmp = fgetl(fid);
    tmp = tmp(1:min(14,length(tmp)));
    fclose(fid);
    if ~strcmp(tmp,'Variable name:') % presumably only true if written from fortran
        A = []; 
        disp(['File ',file,' was a .csv file but not created by the fortran code, returning [].'])
        return
    end 

    % Get the array shape
    rank = dlmread(file,',',[2 1 2 1]); % rank is 3rd row 2nd column 
    if (rank==0) 
        dims = [1 1];
    else
        dims = dlmread(file,',',[3 1 3 rank]); % read in dims in the 4th row and columns 2+
        if (length(dims)==1) 
            dims = [dims 1];
        end
    end

    if (prod(dims)*8/1024/1024>maxMB) 
        A = -123456;
    else

        % Read in the raw data (which begins at row 10 and column 1 and goes to
        % the end of the file
%         A = csvread(file,9,0); 

%         A = textread(file,'','delimiter',',','headerlines',9); % much faster than csvread
%         B = A;


        fid = fopen(file,'r');
        for i = 1:9
            fgetl(fid); % skip to 10th line
        end

        d2 = [prod(dims)/dims(2) dims(2)];

        % Construct the format for importing 
        fmt = '%g';
        for i = 2:d2(2)
            fmt = [fmt ',%g'];
        end 
        fmt = [fmt,'\n'];

        A = fscanf(fid,fmt,[d2(2),d2(1)]); % reverse dim order here since need to transpose
        A = A'; % transpose
        fclose(fid);

        if numel(A)~=prod(dims)

            warning(['Failed to correctly import ' file ' There may be invalid values in the .csv. Trying a more robust file-reading method. If no further warning occurs, then file was imported.'])
            A = dlmread(file,',',9,0,'emptyvalue',NaN); 

            if numel(A)~=prod(dims)
                error(['Failed to correctly import ' file '. There are ' sprintf('%i',numel(A)) ' elts in imported and supposed to be ' sprintf('%i',prod(dims)) ' according to the header']);
            end 

        end


%         err = ~min(B(:)==A(:));
%         if (err)
%             error(['Failed to correctly import ',file])
%         end 



        % Reshape to correct dimensions
        %%% NO! A is organized with dims 2 as columns and other dims as
        %%% rows.
        if (length(dims)==1)
            % do nothing, already a vector
        elseif (length(dims)==2)
            % just convert to matrix
            A = reshape(A,dims);
        elseif (length(dims)>2)
            
            
            ni = dims(1);
            nj = dims(2);
            nk = prod(dims(3:end));
            
            A = reshape(A,[ni*nk,nj]); % reshape to what it looks like in csv file
            
            B = zeros(ni,nj,nk);
            for ik = 1:nk
                B(:,:,ik) = A(1+(ik-1)*ni:ik*ni,:);
            end
            A = reshape(B,dims); 
            
        end

    end 
   
%     catch me
%         
%         disp(['Error in importing file ',file]);
%         if ~isempty(rank)
%             disp('Imported rank:')
%             disp(rank)
%         end
%         if ~isempty(dims)
%             disp('Imported dims:')
%             disp(dims)
%         end
%         rethrow(me);
%         
%     end

 
    
end
