% [varargout] = import_from_fortran(filedir,varargin)	
% 
% Imports the files in filedir, probably will only be useful for me. But
% there is a data file *.txt containing a vector and a header file *_h.txt 
% containing the dimensions of the original array. This loads the vectors
% and reshapes them. The output is saved to 'results_octave.mat' in the
% present directory unless varargout{1} is present. In this case, a
% structure is returned with the fields as all the loaded files. Note, to
% access a field one can use 'eval' to postprocess if the desire is to load 
% into memory. There may be more efficient ways. If saved to file, the
% structure is removed and just the individual variables are saved.
% Additionally, the user may specify files to be loaded as strings
% e.g. import_from_fortran('~/fruit/','bananas','oranges') will search in
% the directory '~/fruit' for the files bananas and oranges.
%
% Copyright (C) 2010 Grey Gordon
% 
% This program is free software; you can redistribute it and/or modify
% it under the terms of the GNU General Public License as published by
% the Free Software Foundation; either version 2 of the License, or
% (at your option) any later version.
% 
% This program is distributed in the hope that it will be useful,
% but WITHOUT ANY WARRANTY; without even the implied warranty of
% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
% GNU General Public License for more details.
% 
% You should have received a copy of the GNU General Public License
% along with this program; if not, write to the Free Software
% Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

% Author: Grey Gordon
% Created: 2010-07-20
% Modified: 2013-03-27
%
function [varargout] = import_from_fortran(filedir,varargin)	

    if (nargout==0) 
        doSaveResults = 1;
    elseif (nargout==1) 
        doSaveResults = 0;
    else
        error('There can be at most one output argument.');
    end
    if (nargin==1)
        findAll = 1;
    elseif (nargin>1)
        findAll = 0;
    end 
    
    firstFile = 1;

    % Check for absolute path. If not an absolute path, convert to it.
    if ~strcmp(filedir(1),'/')
        if (strcmp(filedir(1:2),'./'))
            newfiledir = [pwd,'/',filedir(3:end)];
        elseif (strcmp(filedir(1:3),'../'))
            tmp3 = pwd;
            cd ..
            tmp4 = pwd;
            cd(tmp3)
            newfiledir = [tmp4,'/',filedir(4:end)];
            clear tmp3 tmp4
        elseif (strcmp(filedir(1:2),'~/'))
            tmp3 = pwd;
            cd ~
            tmp4 = pwd;
            cd(tmp3)
            newfiledir = [tmp4,'/',filedir(3:end)];
            clear tmp3 tmp4
        end
    else
        newfiledir = filedir;
    end    
    filedir = newfiledir;
    if ~strcmp(filedir(end),'/')
        filedir = [filedir,'/'];
    end
    
    if (findAll)
        fprintf(1,'Processing %s \n\n',filedir);
    end
   
    % Now that the file directory is set, find the requested files
    if (~findAll)
        
        % Will look only for objects specified in varargin
        objList = varargin;
        
    else
        
        % Will look for all objects, so build the objList
        objList = cell(0);
        
        % Get the file list
        filelist = dir(filedir);    
        tmp = cell(size(filelist,1),1);
        for ind=1:size(filelist,1)
            tmp{ind,1} = filelist(ind).name;
        end
        filelist = tmp;
        clear tmp
        
        % Process the list and search for files ending in '_h'
        fprintf(1,'containing %i files \n',size(filelist,1));
        for i = 1:size(filelist,1)
            file = filelist{i};
            fprintf(1,'Processing file %s ... is txt? ',file);
            %Does file end in .txt?
            if (numel(file)>4)
                fprintf(1,'Y');
            else
                fprintf(1,'N');
            end
            fprintf(1,' ... is header? ');
            %If .txt, is this file a header, _h.txt ?
            if (numel(file)>6)
                ishead = strcmp(file(end-5:end),'_h.txt');
                if (ishead)
                    basefile = [file(1:end-6),'.txt'];
                    %Search for corresponding file
                    ismatched = 0;
                    for j = 1:size(filelist,1)
                        ismatched = ismatched+...
                            strcmp(basefile,filelist{j});
                    end
                    if ismatched==0
                        ishead=0;
                    end
                end
            else
                ishead = 0;
            end
            if (ishead)
                fprintf(1,'Y');
            else
                fprintf(1,'N');
            end
%             %If .txt is a header (_h.txt), is it a sparse file?
%             fprintf(1,' ... is header for sparse? ');
%             if (numel(file)>16)  %Length of .sparseVAL_h.txt
%                 issparse = strcmp(file(end-15:end),'.sparseVAL_h.txt') | ...
%                            strcmp(file(end-15:end),'.sparseCOL_h.txt') | ...
%                            strcmp(file(end-15:end),'.sparseROW_h.txt') | ...
%                            strcmp(file(end-15:end),'.sparseNNZ_h.txt') | ...
%                            strcmp(file(end-15:end),'.sparseNNC_h.txt') | ...
%                            strcmp(file(end-15:end),'.sparseNNR_h.txt');
%             else 
%                 issparse = 0;
%             end
%             if (issparse)
%                 fprintf(1,'Y');
%             else
%                 fprintf(1,'N');
%             end
            
            if (ishead)
                % Save the file stem
                objList{end+1} = file(1:end-6);
            end                
    		fprintf(1,'\n');
        
        end        
    end
    
    % Process all the valid objects
    for i = 1:numel(objList)
        
        % If isn't sparse
%         if (~issparse)
            stem = objList{i};
            mainfile = [filedir,stem,'.txt'];
            headfile = [filedir,stem,'_h.txt'];

            % Check whether the header and main file exist
            if (~exist(mainfile,'file')) 
                error('File %s does not exist \n',mainfile);
            elseif (~exist(headfile,'file'))
                error('File %s does not exist \n',headfile);
            end
            
            % Load the header file to get the dimensions
            load('-ascii',headfile);
            hloc = load('-ascii',headfile);
            
            % Check whether the mex file exists
            if (3==exist('myimportfile')) 
                % Import new way using the mex file
                tosave.(stem) = myimportfile(mainfile,prod(hloc));
            else
                % Import old way                                    
                tmpfid = fopen(mainfile);
                tosave.(stem) = fscanf(tmpfid,'%e',prod(hloc));
%                 maxblocksize = .5*8*1028^3; %.5GB file
%                 if (prod(hloc)<=maxblocksize)
%                     tosave.(stem) = fscanf(tmpfid,'%e',prod(hloc));
%                 else                   
%                     tosave.(stem) = [];
%                     nelts = prod(hloc);
%                     nleft = prod(hloc);
%                     tosave.(stem) = NaN(nleft,1);
%                     block = 1;
%                     while nleft>0
%                         lb = 1 + maxblocksize*(block-1);
%                         ub = min(nelts,maxblocksize*block);
%                         nread = min(maxblocksize,nleft);
%                         tosave.(stem)(lb:ub) = fscanf(tmpfid,'%e',nread);
%                         nleft = nleft - min(maxblocksize,nleft);
%                         block = block + 1;
%                     end 
%                     
%                     if (any(bugcheck~=tosave.(stem))) 
%                         error('importing in blocks did not work')
%                     end
%                     
%                 end
                fclose(tmpfid);
            end                

            stem = regexprep(stem,'%','_');

            %Reshape main according to header if not a vector
            if ~(isscalar(hloc))
                % Check that number of elements is the same
                if (numel(tosave.(stem))~=prod(hloc))
                    error(sprintf('cant reshape %s because number of elements is inconsistent',stem));
                end
                tosave.(stem) = reshape(tosave.(stem),hloc');
            end

    
            % Saves the fields of the structure rather than their name
            if (doSaveResults) 
                if (firstFile) 
                    save('results_octave.mat','-struct','tosave',stem) 
                    firstFile = 0;
                else
                    save('results_octave.mat','-struct','tosave',stem,'-append') 
                end
                tosave = rmfield(tosave,stem);
            end
            
            
%         else
%         % If is sparse, then process if is the .sparseVAL_h.txt file
%             if (strcmp(file(end-15:end),'.sparseVAL_h.txt'))
%                 stem = file(1:end-16);
%                 tmpVAL = load('-ascii',[filedir,stem,'.sparseVAL.txt']);
%                 tmpROW = load('-ascii',[filedir,stem,'.sparseROW.txt']);
%                 tmpCOL = load('-ascii',[filedir,stem,'.sparseCOL.txt']);
%                 tmpNNZ = load('-ascii',[filedir,stem,'.sparseNNZ.txt']);
%                 tmpNNR = load('-ascii',[filedir,stem,'.sparseNNR.txt']);
%                 tmpNNC = load('-ascii',[filedir,stem,'.sparseNNC.txt']);
% 
%                 tosave.(stem) = sparse(tmpROW,tmpCOL,tmpVAL,tmpNNR,tmpNNC,tmpNNZ);
% 
%                 clear tmpVAL tmpROW tmpCOL tmpNNR tmpNNC tmpNNZ
%             end
%         end
    end

    if (doSaveResults)
        fprintf(1,'Files written to results_octave.mat\n');
    else
        varargout{1} = tosave;
    end
    
%             % Saves the fields of the structure rather than their name
%             if (doSaveResults) 
%                 save('results_octave.mat','-struct','tosave') 
%                 fprintf(1,'Files written to results_octave.mat\n');
%             else
%                 varargout{1} = tosave;
%             end
%     fprintf(1,'Successfully imported %i files. \n',numel(fieldnames(tosave)));

end
