classdef param_classdef
    %PARAM_CLASSDEF Contains all commonly used parameters of the model
    %
    %   Detailed explanation goes here
    
    % Fixed value (true parameters in Fortran sense)
    properties (Constant = true)

        % Determine whether using Aiyagari or bankruptcy.
        is_Aiyagari = false(1)
        is_MeansTested = false(1)
        is_Commit = false(1)
        %Note on the debt limit: w*emin*1/(1-rho*qbar)
        %The steady state natural limit is ~1.2148 = 1.164*.0502*1/(1-.9519)
        %The sequential limit obviously varies, but worst observed so far would be ~ 1.1786 = 1.1387*.0502*1/(1-.9515)
        %I set the limit tighter than both of these to ensure that if there's forecast error, that no one gets negative consumption
        Aiyagari_debt_limit_SS = -1.1d0
        Aiyagari_debt_limit = param_classdef.Aiyagari_debt_limit_SS*ones(2,2) %reshape((/.2d0,.0d0,.2d0,0.d0/),(/2,2/))
        e_mean = 1.d0 %used for means testing

        %%%%%%%%%%%%%%%%%%%%%%%
        % Lengths
        %
        len_z = 2 % 2 = #{z=g,z=b}
        len_h = 2 % 2 = #{h=0,h=1}
        len_N = 1 % # of N points.  If F_e'|e,z,z' = F_e'|e, only need one
        len_A1 = 1 %2
        len_A2 = 3 %3
        len_A3 = 6
        len_m = param_classdef.len_z*param_classdef.len_N*...
                param_classdef.len_A1*param_classdef.len_A2*...
                param_classdef.len_A3
        len_T = 150 % # of periods to simulate
        len_t0 = 50  % # of period to throwout (only use len_t0+1...T)
        len_kpp = 1250 %750 %1250 % # of positive grid points used when calculating distributions (len_k = len_kpp+len_Amm+1)

        %%%%%%%%%%%%%%%%%%%%%%%%
        % Estimated parameters
        %
        discount = .917354d0*.975d0           %CCNR have effective discount rate of .8902
        chi        = .0189d0                    %Percent cost of labor income lost
        trans_11 = 0.01979519441340765740d0 %.02d0
        trans_12 = 1.d0-param_classdef.trans_11
        trans_13 = 0.d0
        trans_21 = 0.00081778249954317654d0 %.001d0
        trans_23 = 1.d0 - 0.00081778249954317654d0 - 0.78068033809607384477d0; %.219d0
        trans_22 = 1.d0-param_classdef.trans_21-param_classdef.trans_23;
        trans_33 = 0.96397436291636229289d0;
        trans_32 = 1.d0-param_classdef.trans_33;
        trans_31 = 0.d0;

        phi_SS = 0.46972182750383228989d0; %.47d0
        phi = param_classdef.phi_SS*ones(param_classdef.len_z,...
                                         param_classdef.len_z);
        
        %%%%%%%%%%%%%%%%%%%%%%%%
        % Multigrid Parameters
        % 
        mg_steps = 1 %2
        mg_n_1 = 6*ones(param_classdef.mg_steps,1)  %1 %2 %1 %(/2,4/)
        mg_n_2 = 18*ones(param_classdef.mg_steps,1)   %4 %2 %4 %(/3,8/)
        mg_n_3 = 40*ones(param_classdef.mg_steps,1)   %24 %3  %24 %(/4,24/)
        mg_len_e = param_classdef.mg_n_1+param_classdef.mg_n_2+param_classdef.mg_n_3
        mg_xi1_start = 1*ones(param_classdef.mg_steps,1)  %Gives starting index for xi==1
        mg_xi1_end = param_classdef.mg_n_1
        mg_xi2_start = param_classdef.mg_n_1 + 1
        mg_xi2_end = param_classdef.mg_n_1 + param_classdef.mg_n_2
        mg_xi3_start = param_classdef.mg_n_1 + param_classdef.mg_n_2 + 1
        mg_xi3_end = param_classdef.mg_len_e

        %%%%%%%%%%%%%%%%%%%%%%%%
        % Bankruptcy Parameters
        %
        death = -1d150  %Value of negative consumption
        sigma = 2.d0            %Coefficient of consumption risk aversion
        alpha_k = .36d0              %Capital share
        delta = .10d0                %Depreciation rate
        lambda = .9d0          %Prob bad stays on record %1/(1-lamba) is average # periods spent in bankruptcy

        %%%%%%%%%%%%%%%%%%%%%%%%
        % Shock parameters
        %
        rho = .975d0                %Conditional probability of survival
        emin_3_SS = 1d0/19.93674729116d0
        emax_3_SS = 30.46468670327320893421d0/19.93674729116d0
        emin_2_SS = 39.24604493098055257860d0/19.93674729116d0
        emax_2_SS = 116.34074857650554690736d0/19.93674729116d0
        emin_1_SS = 14335.69235206637677038088d0/19.93674729116d0
        emax_1_SS = 21263.86711017461493611336d0/19.93674729116d0
        emax_3 = [1.266619d0,param_classdef.emax_3_SS;param_classdef.emax_3_SS,1.634849d0]
        emax_2 = [5.151411d0,param_classdef.emax_2_SS;param_classdef.emax_2_SS,6.466954d0]
        emax_1 = [1005.090923d0,param_classdef.emax_1_SS;param_classdef.emax_1_SS,1123.313219d0]
        emin_3 = [.172967d0,param_classdef.emin_3_SS;param_classdef.emin_3_SS,.000001d0]
        emin_2 = [2.289856d0,param_classdef.emin_2_SS;param_classdef.emin_2_SS,1.671917d0]
        emin_1 = [747.935167d0,param_classdef.emin_1_SS;param_classdef.emin_1_SS,692.403574d0]

        g = 1.02d0 %value of z in good state %cf KMS calibration
        b = 0.98d0 %value of z in bad state
        g_ind = 1 % index of z grid associated with the good agg state
        b_ind = 2 % index of z grid associated with the bad agg state
        trans_gg = .5d0 %prob of trans from g to g %avg business cycle duration ~ 2yrs
        trans_bb = .5d0 %prob of trans from b to b
        grid_z = [param_classdef.g;param_classdef.b]
        trans_z = [param_classdef.trans_gg,1.d0-param_classdef.trans_gg;...
                   1.d0-param_classdef.trans_bb,param_classdef.trans_bb]

        %%%%%%%%%%%%%%%%%%%%%%%%
        % Targets
        % 
        target_r = 0.1169d0              %Target MPK
        target_qbar = param_classdef.rho/(1.d0 + param_classdef.target_r - param_classdef.delta)
        target_KN = (param_classdef.target_r/param_classdef.alpha_k).^(1.d0/(param_classdef.alpha_k-1.d0)) %Target capital labor ratio
        target_KY = 1.d0/(param_classdef.target_r/param_classdef.alpha_k)   %Target capital output ratio
        target_w = (1.d0-param_classdef.alpha_k)*param_classdef.target_KN.^param_classdef.alpha_k %Target MPL
        target_pop_in_debt = 0.036d0     %Target measuare of hhs in ddbt
        target_debt_to_output = 0.0036d0 %Target debt/output ratio
        target_pct_hhs_filing = 0.0029d0 %Target % of hhs filing for bankruptcy
        target_pct_in_bnk = param_classdef.target_pct_hhs_filing/(1.d0-param_classdef.rho*param_classdef.lambda) %Fraction of households who will be in default
        target_income_gini = .61d0
        target_wealth_gini = .80d0


        %%%%%%%%%%%%%%%%%%%%%%%%
        % Grids 
        %
        max_a = 35000.d0*param_classdef.target_w
        min_a = -7.d0*param_classdef.target_w
        
    end
    
    % Changeable value "parameters" (not parameters in the Fortran sense)
    properties

        %%%%%%%%%%%%%%%%%%%%%%%%
        % Shock values (can change)
        %
        n_1 = param_classdef.mg_n_1(param_classdef.mg_steps) %10 %2 %3 %100 %3 %  10 %2 %10 %1 %10 %20 %3 % 1 %4 %12 
        n_2 = param_classdef.mg_n_2(param_classdef.mg_steps) %30 %2 %16 %100 %16 % 20 %2 %20 %4 %40 %44 %11 %3 %8 %44 
        n_3 = param_classdef.mg_n_3(param_classdef.mg_steps) %70 %2 %40 %100 %40 %26 % 40 %2 %40 %8 %80 %64 %16 %4 %20 %64 
        xi1_beg = param_classdef.mg_xi1_start(param_classdef.mg_steps)
        xi1_end = param_classdef.mg_xi1_end(param_classdef.mg_steps)
        xi2_beg = param_classdef.mg_xi2_start(param_classdef.mg_steps)
        xi2_end = param_classdef.mg_xi2_end(param_classdef.mg_steps)
        xi3_beg = param_classdef.mg_xi3_start(param_classdef.mg_steps)
        xi3_end = param_classdef.mg_xi3_end(param_classdef.mg_steps)
        len_e = param_classdef.mg_len_e(param_classdef.mg_steps) %# of idiosyncratic shocks, should be mod 8 = 0


        %%%%%%%%%%%%%%%%%%%%%%%%
        % Grids 
        %
        grid_N = zeros(param_classdef.len_N,1)
        grid_A1 = zeros(param_classdef.len_A1,1)
        grid_A2 = zeros(param_classdef.len_A2,1)
        grid_A3 = zeros(param_classdef.len_A3,1)


        
        %%%%%%%%%%%%%%%%%%%%%%%
        % Loop controls, tolerances
        max_iter_VFI = 100           %Maximum number of iterations to perform during VFI
        print_mod = 10
        toler = 1.d-6
        toler_VFI = 1.d-6
        %toler_newton = 1.d-6

    end
    
    methods
        %function out = set_indivmom_grids(in)        
        function out = param_classdef()        
            %out = in;
            %if ~in.is_Aiyagari 
            if ~param_classdef.is_Aiyagari 
                out.grid_A1 = [ 1.d0 ]*.0041d0;
                out.grid_A2 = [ .75d0; 1.d0; 1.25d0 ]*2d0;
                out.grid_A3 = [ .75d0; .85d0; .95d0; 1.05d0; 1.15d0; 1.25d0]*4.54d0;
                out.grid_N = 1.d0;
            else 
                out.grid_A1 = [ 1.d0 ]*.0041d0;
                out.grid_A2 = [ .5d0; .75d0; 1.d0 ]*2d0;
                out.grid_A3 = [ .55d0; .65d0; .75d0; .85d0; .95d0; 1.05d0]*4.54d0;
                out.grid_N = 1.d0;
            end
        end
    end
    
end

