% [knots abc varargout] = schumakerSplineCoeff(x,y,varargin)
% 
% Constructs a Schumaker (1983) shape preserving spline approximation of
% the data (x,y). If y' values are supplied in varargin{1}, these are used
% as well. The output is knots (which include but are not neccessarily
% limited to x) and coeffs to use in schumakerSpline[Eval|Deriv]. 
% Optionally, returns the "pp" form if a third output argument is
% requested.
%
% Grey Gordon, 2012
% SDG
function [knots abc varargout] = schumakerSplineCoeff(x,y,varargin)

    % Ensure column vectors
    if (~iscolumn(x))
        error('x must be a column vector')
    elseif (~iscolumn(y))
        error('y must be a column vector')
    end
    
    n = length(x);
    
    if (length(y)~=n) 
        error('x and y not same length')
    end
    
    % Compute forward differences
    dx = diff(x);
    dy = diff(y);
    
    % Compute slopes
    m = dy./dx;
    
    % Get the first derivatives
    if (isempty(varargin))
        
        % Compute approximations
        L = sqrt(dx.^2 + dy.^2);
        Lm = L.*m;
        
        yp = NaN(n,1);
        yp(2:n-1) = (m(1:n-2).*m(2:n-1)>0).*(Lm(1:n-2)+Lm(2:n-1))./(L(1:n-2)+L(2:n-1));
        yp(1) = (3*m(1) - yp(2))/2;
        yp(n) = (3*m(n-1) - yp(n-1))/2;
        
    else
        
        % Get actual from user
        yp = varargin{1};
        if (~iscolumn(yp))
            error('yp must be a column vector')
        end
        if (length(yp)~=n)
            error('yp does not have same length as x')
        end
        
    end
    
    % Vectorized computation
    dyp = diff(yp);
    yplo_minus_m = yp(1:n-1)-m;
    yphi_minus_m = yp(2:n)-m;
    abs_yplo_minus_m = abs(yplo_minus_m);
    abs_yphi_minus_m = abs(yphi_minus_m);
    same_sign = (yphi_minus_m>=0 & yplo_minus_m>=0) | (yphi_minus_m<=0 & yplo_minus_m<=0);
    
    % Case where lemma applies and a quadratic spline works perfectly 
    notSpecialCase = ~(abs((yp(1:n-1)+yp(2:n))/2 - m)<eps(0));

    % Allocate the correct number of break points and coeffs
    nb = 2*(n-1) - sum(~notSpecialCase);
    A = NaN(nb,1);
    B = NaN(nb,1);
    C = NaN(nb,1);
    knots = NaN(nb+1,1);

    % Compute the knots and coefficients
    if (all(notSpecialCase)) 

        % If none are special case, we can vectorize the construction and
        % speedup things considerably.
        case1 = find(same_sign(1:n-1));
        case2 = find(~same_sign(1:n-1) & abs_yplo_minus_m(1:n-1)>abs_yphi_minus_m(1:n-1));
        case3 = find(~same_sign(1:n-1) & abs_yplo_minus_m(1:n-1)<=abs_yphi_minus_m(1:n-1));
        
        xsi = NaN(n-1,1);
        xsibar = NaN(n-1,1);
        % Typical case where quadratic won't fit the data and need another knot.
        % Construct the knot so that monotonicity and concavity preserved
        xsi(case1) = (x(case1)+x(case1+1))/2;
        xsibar(case2) = x(case2) + 2*dx(case2).*yphi_minus_m(case2)./dyp(case2);
        xsi(case2) = (x(case2)+xsibar(case2))/2;
        xsibar(case3) = x(case3+1) + 2*dx(case3).*yplo_minus_m(case3)./dyp(case3);
        xsi(case3) = (x(case3+1) + xsibar(case3))/2;
        % Given the knot, compute the uniquely determined coefficients
        alpha = xsi-x(1:n-1);
        beta = x(2:n)-xsi;
        ypbar = 2*m(1:n-1) - (alpha.*yp(1:n-1) + beta.*yp(2:n))./dx(1:n-1);
        
        I1 = 1 + (0:2:nb-2);
        I2 = 2 + (0:2:nb-2);
        
        A(I1) = y(1:n-1);
        B(I1) = yp(1:n-1);
        C(I1) = (ypbar - yp(1:n-1))./(2*alpha);
        knots(I1) = x(1:n-1);
        A(I2) = A(I1) + alpha.*B(I1) + alpha.^2.*C(I1); 
        B(I2) = ypbar;
        C(I2) = (yp(2:n)-ypbar)./(2*beta);
        knots(I2) = xsi;
        
    else

        % Loop over intervals and compute coeffs for each
        j = 1;
        for i = 1:n-1

            % Compute the coefficients and knots
            if (notSpecialCase(i))
                % Typical case where quadratic won't fit the data and need another knot.
                % Construct the knot so that monotonicity and concavity preserved
                if (same_sign(i)) 
                    xsi = (x(i)+x(i+1))/2;
                elseif (abs_yplo_minus_m(i)>abs_yphi_minus_m(i))
                    xsibar = x(i) + 2*dx(i)*yphi_minus_m(i)/dyp(i);
                    xsi = (x(i) + xsibar)/2;
                else
                    xsibar = x(i+1) + 2*dx(i)*yplo_minus_m(i)/dyp(i);
                    xsi = (x(i+1) + xsibar)/2;
                end
                % Given the knot, compute the uniquely determined coefficients
                alpha = xsi-x(i);
                beta = x(i+1)-xsi;
                ypbar = 2*m(i) - (alpha*yp(i) + beta*yp(i+1))/dx(i);
                A(j) = y(i);
                B(j) = yp(i);
                C(j) = (ypbar - yp(i))/(2*alpha);
                knots(j) = x(i);
                A(j+1) = A(j) + alpha*B(j) + alpha^2*C(j); 
                B(j+1) = ypbar;
                C(j+1) = (yp(i+1)-ypbar)/(2*beta);
                knots(j+1) = xsi;
                j = j + 2;
            else 
                % Special case where quadratic fits the data
                A(j) = y(i);
                B(j) = yp(i);
                C(j) = .5*dyp(i)/dx(i); 
                knots(j) = x(i);
                j = j + 1;            
            end

        end
    
    end

    knots(nb+1) = x(n); % Closing the knot grid
    
    % Put the coefficients together
    abc = [A,B,C];
    
    % Form the "pp" form for matlab. NOTE: they use a different convention
    if (nargout==3)
        pp = mkpp(knots,[C,B,A]);
        varargout{1} = pp;
    end
    
end