! Grey Gordon and Qiu Shi, 2016
! Solves the Arellano model with production as described in the paper using discrete everything.
module mod_data
    use mod_matlab
    use mod_discreteAR1
    use mod_util
    implicit none

    type(timer) :: time_max, time_quad, time_tot, time_sup
    real(8) :: maxTime,avgcnt
    logical :: useStdOut
    integer :: cashopt
    integer, parameter :: itercut = 200 ! since possibly does not converge, only get stats for first itercut iterations

    integer, protected :: nK, nB, nZ
    integer, protected :: iBeq0

    ! Parameters
    real(8), parameter :: crra = 2d0
    real(8), parameter :: beta = .95282d0
    real(8), parameter :: deprec = .025d0, alpha= .36d0
    real(8), parameter :: rstar = .017d0 ! world-risk-free rate
    real(8) :: theta = .282d0
    real(8) :: kappa = .05d0 ! output loss conditional on being in default
    real(8) :: def_std = .025d0, def_ar1 = .945d0
    real(8), parameter :: kSS = ((1d0/beta - 1d0 + deprec)/alpha)**(1d0/(alpha-1d0))
    real(8), parameter :: ySS = kSS**alpha
    real(8), parameter :: cSS = kSS**alpha - deprec*kSS
    real(8), parameter :: vSS = cSS**(1d0-crra)/(1d0-crra)/(1d0-beta)

    ! Grids and shocks
    real(8), allocatable :: gB(:),gK(:),gZ(:),Fzpcz(:,:)
    ! Conditional on repayment (nd at end stands for "no default")
    real(8), allocatable :: q(:,:,:),Vnd(:,:,:),cnd(:,:,:),CVnd(:,:,:)
    integer, allocatable :: bpind(:,:,:),kpind(:,:,:)
    ! Conditional on default (d at end stands for "default")
    real(8), allocatable :: Vd(:,:),cd(:,:),CVd(:,:)
    integer, allocatable :: kpid(:,:)
    ! Best of both
    real(8), allocatable :: V(:,:,:),d(:,:,:)

contains

    subroutine allocatedata(nBin,nKin,nZin)
        implicit none
        integer, intent(in) :: nBin, nKin, nZin

        nB = nBin
        nK = nKin
        nZ = nZin

        iBeq0 = 7*nB/10

        if (allocated(gB))       deallocate(gB);        allocate(gB(nB))
        if (allocated(gK))       deallocate(gK);        allocate(gK(nK))
        if (allocated(gZ))       deallocate(gZ);        allocate(gZ(nZ))
        if (allocated(Fzpcz))    deallocate(Fzpcz);     allocate(Fzpcz(nZ,nZ))
        if (allocated(V))        deallocate(V);         allocate(V(nB,nK,nZ))
        if (allocated(q))        deallocate(q);         allocate(q(nB,nK,nZ))
        if (allocated(d))        deallocate(d);         allocate(d(nB,nK,nZ))

        if (allocated(Vnd))      deallocate(Vnd);       allocate(Vnd(nB,nK,nZ))
        if (allocated(cnd))      deallocate(cnd);       allocate(cnd(nB,nK,nZ))
        if (allocated(bpind))    deallocate(bpind);     allocate(bpind(nB,nK,nZ))
        if (allocated(kpind))    deallocate(kpind);     allocate(kpind(nB,nK,nZ))
        if (allocated(CVnd))     deallocate(CVnd);      allocate(CVnd(nB,nK,nZ))

        if (allocated(Vd))       deallocate(Vd);        allocate(Vd(nK,nZ))
        if (allocated(cd))       deallocate(cd);        allocate(cd(nK,nZ))
        if (allocated(kpid))     deallocate(kpid);      allocate(kpid(nK,nZ))
        if (allocated(CVd))      deallocate(CVd);       allocate(CVd(nK,nZ))

    end subroutine

end module 

module mod_routines
    implicit none
contains

    subroutine vfiDEF(iter,overtime,opt)
        use mod_util
        use mod_matlab
        use mod_binmono_disc_sort
        use mod_data
        implicit none
        integer, intent(in) :: opt(2)
        logical, intent(out) :: overtime
        integer, intent(out) :: iter ! vfi iteration
        ! local
        real(8), allocatable :: work(:,:,:),cash(:,:,:),expend(:,:,:),cashd(:,:),gCASH(:),workcash(:,:),work2(:,:),mlo(:,:,:)
        integer, allocatable :: iworkcash(:,:),ilo(:,:,:)
        real(8) :: sup, time, onecnt
        integer :: iK,iZ,iB,tmpcnt,nCASH

        ! Start the overall timer
        call time_tot%start()

        nCASH = nK

        allocate(work(nB,nK,nZ))
        allocate(work2(nB,4))
        allocate(cash(nB,nK,nZ))
        allocate(expend(nB,nK,nZ))
        allocate(cashd(nK,nZ))
        allocate(gCASH(nCASH))
        allocate(workcash(nCASH,4))
        allocate(iworkcash(nCASH,2))
        allocate(ilo(nB,nK,nZ))
        allocate(mlo(nB,nK,nZ))
        
        overtime = .false.

        ! Make initial guess
        Vnd = vSS
        Vd = vSS
        V = vSS
        d = 0d0
        avgcnt = 0d0

        ! Precompute "cash" from the state (b,k) --- does not change over time
        do iZ = 1,nZ
            do iK = 1,nK
                cash(:,iK,iZ) = gB + gZ(iZ) * gK(iK)**alpha + (1d0-deprec)*gK(iK)
                cashd(iK,iZ) = (1d0-kappa) * gZ(iZ) * gK(iK)**alpha + (1d0-deprec)*gK(iK)
            end do
        end do
        gCASH(1) = minval(cash)
        gCASH(nCASH) = maxval(cash)
        gCASH = linspace(gCASH(1),gCASH(nCASH),nCASH)

        ! Precompute how to do linear interpolation from f(cash) to f(b,k)
        do iZ = 1,nZ
            do iK = 1,nK
                do iB = 1,nB
                    call noexinterp1qlocweight(ilo(iB,iK,iZ),mlo(iB,iK,iZ),cash(iB,iK,iZ),gCASH)
                end do
            end do
        end do


        do iter = 1,500
        
            work = V

            ! Compute bond prices and continuation utility
            if (iter<=itercut) call time_quad%start()

            call mygemm(CVnd,V,Fzpcz,nB*nK,nZ,nZ,nZ,alpha=beta)
            call mygemm(q,d,Fzpcz,nB*nK,nZ,nZ,nZ,alpha=1d0/(1d0+rstar)) ! qbar * E d
            q = 1d0/(1d0+rstar) - q ! qbar - qbar * E d = qbar * E(1-d)
            CVd = beta*matmul(theta*V(iBeq0,:,:) + (1d0-theta)*Vd,Fzpcz)

            if (iter<=itercut) call time_quad%stop()

            onecnt = 0d0
            do iZ = 1,nZ

                ! Check if have exceeded time limit
                call toc(time)
                if (time>maxTime) then
                    overtime = .true.
                    return ! do not modify cnt or time_out
                end if

                !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
                ! Solve the Vnd problem
                !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

                ! Compute the expenditures
                do iK = 1,nK
                    expend(:,iK,iZ) = q(:,iK,iZ)*gB + gK(iK) ! q(b',k',z) b' + k'
                end do
                
                ! Solve

                if (cashopt==0) then
                    ! Do the normal optimization (don't use cash at hand)
                    if (iter<=itercut) call time_max%start()

                    call sub_binmono_disc_sort(Vnd(:,:,iZ),cnd(:,:,iZ),kpind(:,:,iZ),cash(:,:,iZ),expend(:,:,iZ),CVnd(:,:,iZ),crra,nB*nK,nB*nK,mono=opt(1),cnt=tmpcnt)
                    ! Note that the kpind here is NOT the capital index. Rather, it gives which element is optimal in the combined tensor grid.
                    bpind(:,:,iZ) = 1+mod(kpind(:,:,iZ)-1,nB)
                    kpind(:,:,iZ) = 1+(kpind(:,:,iZ)-1)/nB

                    ! not assuming ascending is much slower
                    if (iter<=itercut) call time_max%stop()

                else

                    ! Solve maximization
                    if (iter<=itercut) call time_max%start()
                    if (cashopt==1) then
                        call sub_binmono_disc_sort(workcash(:,1),workcash(:,2),iworkcash(:,1),gCASH,expend(:,:,iZ),CVnd(:,:,iZ),crra,nCASH,nB*nK,mono=opt(1),cnt=tmpcnt,fsorted=.true.)
                        ! Recover the asset policy
                        ! workcash(:,3) = gB(1+mod(iworkcash(:,1)-1,nB))
                        ! workcash(:,4) = gK(1+(iworkcash(:,1)-1)/nB)
                    else
                        call sub_binmono_disc_ascend(workcash(:,1),workcash(:,2),iworkcash(:,1),iworkcash(:,2),gCASH,expend(:,:,iZ),CVnd(:,:,iZ),crra,nCASH,[nB,nK],mono=opt(1),cnt=tmpcnt,fsorted=.true.)
                        ! Recover the asset policy
                        ! workcash(:,3) = gB(iworkcash(:,1))
                        ! workcash(:,4) = gK(iworkcash(:,2))
                    end if
                    if (iter<=itercut) call time_max%stop()

                    ! Recover on the (b,k) grid
                    do iK = 1,nK

                        ! Do the interpolation

                        ! Interpolation just V, precompute
                        do iB = 1,nB
                            Vnd(iB,iK,iZ) = workcash(ilo(iB,iK,iZ)+1,1) + mlo(iB,iK,iZ)*(workcash(ilo(iB,iK,iZ),1)-workcash(ilo(iB,iK,iZ)+1,1))
                        end do

                        ! Interpolation just V, no precompute
                        ! Vnd(:,iK,iZ) = interp1q(gCASH,workcash(:,1),cash(:,iK,iZ))

                        ! Interpolation of everything, no precompute
                        ! work2(:,1:4) = interp1q(gCASH,workcash(:,1:4),cash(:,iK,iZ))
                        ! Vnd(:,iK,iZ) = work2(:,1)
                        ! cnd(:,iK,iZ) = work2(:,2)
                        ! ! Do binary search to simulate the cost of interpolating using the policies
                        ! do iB = 1,nB
                        !     call bsearch(bpind(iB,iK,iZ),work2(iB,3),gB)
                        !     call bsearch(kpind(iB,iK,iZ),work2(iB,4),gK)
                        ! end do
                    end do

                end if

                if (iter<=itercut) onecnt = onecnt + real(tmpcnt,8) / (nB*nK*nZ) ! evaluations per state


                !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
                ! Solve the Vd problem
                !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
                if (iter<=itercut) call time_max%start()
                call sub_binmono_disc(Vd(:,iZ),cd(:,iZ),kpid(:,iZ),cashd(:,iZ),gK,CVd(:,iZ),crra,nK,nK,mono=opt(1)) ! no need to sort here, satisfies Lemma 1 Prop 1
                if (iter<=itercut) call time_max%stop()

                !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
                ! Solve the V = max(Vnd,Vd) problem
                !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
                do iK = 1,nK
                    do iB = 1,nB
                        if (Vd(iK,iZ)>Vnd(iB,iK,iZ)) then
                            ! Default
                            d(iB,iK,iZ) = 1d0
                            V(iB,iK,iZ) = Vd(iK,iZ)
                        else
                            ! Repay
                            d(iB,iK,iZ) = 0d0
                            V(iB,iK,iZ) = Vnd(iB,iK,iZ)
                        end if
                    end do
                end do

            end do

            ! Update the average count
            if (iter<=itercut) avgcnt = (onecnt + avgcnt*(iter-1))/iter

            ! Comptue the supnorm
            if (iter<=itercut) call time_sup%start()
            work = work - V
            sup = maval(work,size(work)) ! sup = maxval(abs(work))

            if (mod(iter,1)==0 .and. useStdOut) write(*,*) 'iter: ',iter,'supnorm V',sup
            
            if (sup<1d-7*maval(V,size(V))) exit ! if (sup<1d-7*maxval(abs(V))) exit

            if (iter<=itercut) call time_sup%stop()
            if (iter==itercut) call time_tot%stop() ! stop the overall timer

        end do

    contains

        ! Additional argument says whether to interpolate along first (the default) or second dimension 
!         subroutine sub_interp(yi,x,y,xi,nx,ny2,nxi)
!             implicit none
!             integer, intent(in) :: nx,ny2,nxi
!             real(8), dimension(*), intent(out) :: yi(nxi,*)
!             real(8), dimension(*), intent(in) :: x(*),y(nx,*),xi(*)
!             ! local 
!             real(8) :: mlo
!             integer :: i,j,ilo
! 
!             do i = 1,nxi
!                 call bsearch(ilo,xi(i),x(1:nx))
!                 mlo = (x(ilo+1)-xi(i))/(x(ilo+1)-x(ilo))
!                 do j = 1,ny2
!                     yi(ilo,j) = y(ilo,j)*mlo + y(ilo+1,j)*(1d0-mlo)
!                 end do
!             end do
! 
!         end subroutine

        real(8) function maval(x,nx)
            implicit none
            real(8), dimension(*), intent(in) :: x
            integer, intent(in) :: nx
            ! local
            integer :: i
!             external idamax

            interface 
                integer function idamax(n,dx,incx)
                    integer :: incx,n
                    real(8) :: dx(*)
                end function 
            end interface

            ! BLAS to get the index of the largest value in absolute terms
            i = idamax(nx,x,1)
            
            ! Return the level
            maval = abs(x(i))

        end function 

    end subroutine 

!     subroutine invarDEF()
!         use mod_matlab
!         use mod_data
!         implicit none
!         real(8), allocatable :: dsn2(:,:,:)
!         real(8) :: sup,mlo
!         integer :: iZ,iK,iter,ilo
! 
!         allocate(dsn2(nK,nZ,2))
! 
!         dsn = 0d0
!         dsn(nK,(nZ+1)/2,1) = 1d0  
! 
!         do iter = 1,merge(10000,1000,theta>0d0)
! 
!             dsn2 = 0d0
! 
!             ! h = 0
!             do iZ = 1,nZ
!                 do iK = 1,nK
!                     if (d(iK,iZ)==1d0) then
!                         dsn2(iKeq0,iZ,2) = dsn2(iKeq0,iZ,2) + dsn(iK,iZ,1)
!                     else
!                         ! dsn2(kpi(iK,iZ),iZ,1) = dsn2(kpi(iK,iZ),iZ,1) + dsn(iK,iZ,1)
!                         call noexinterp1qlocweight(ilo,mlo,kp(iK,iZ),gK)
!                         dsn2(ilo:ilo+1,iZ,1) = dsn2(ilo:ilo+1,iZ,1) + [mlo,1d0-mlo]*dsn(iK,iZ,1)
!                     end if
!                 end do
!             end do
! 
!             ! h = 1
!             do iZ = 1,nZ
!                 dsn2(iKeq0,iZ,1) = dsn2(iKeq0,iZ,1) + theta*dsn(iKeq0,iZ,2)
!                 dsn2(iKeq0,iZ,2) = dsn2(iKeq0,iZ,2) + (1d0-theta)*dsn(iKeq0,iZ,2)
!             end do
! 
!             work = dsn(:,:,1)
!             work2 = dsn(:,:,2)
! 
!             ! dsn(:,:,1) = matmul(dsn2(:,:,1),transpose(Fzpcz))
!             ! dsn(iKeq0,:,2) = matmul(dsn2(iKeq0,:,2),transpose(Fzpcz))
!             call mygemm(dsn(:,:,1),dsn2(:,:,1),Fzpcz,nK,nZ,nZ,nZ,tb='T')
!             dsn(iKeq0,:,2) = matmul(Fzpcz,dsn2(iKeq0,:,2))  ! b*A' = (A*b')'
! 
!             sup = max(maxval(abs(dsn(:,:,1)-work)),maxval(abs(dsn(:,:,2)-work2)))
! 
!             if (mod(iter,1000)==0) print*,'DSN: iter ',iter,' sup ',sup, 'sum ',sum(dsn)
! 
!             if (sup<1d-10) exit
! 
!         end do
! 
! 
!     end subroutine 


end module mod_routines

program main
    use mod_util
    use mod_data
    use mod_routines
    implicit none
    integer :: vfi_iter,opt
    real(8) :: minutes!,stats(7)
    logical :: overtime
!     real(8), allocatable :: extrastats(:)
!     character(50) :: arg
!     character(50*9) :: arglist

    ! If using Mac, single threaded BLAS via the Accelerate framework can be done by setting the environment variable VECLIB_MAXIMUM_THREADS to 1'
    ! call execute_command_line('export VECLIB_MAXIMUM_THREADS=1')

    maxTime = 60d0 ! this is in terms of minutes

    if (command_argument_count()/=6) stop 'Please provide 6 command line arguments'

    ! Allocate space
    call allocatedata(getcmdarg_i(1),getcmdarg_i(2),getcmdarg_i(3))

    opt = getcmdarg_i(4)
    useStdOut = getcmdarg_i(5)==1
    cashopt = getcmdarg_i(6)

    if (useStdOut) then
        write(*,'(A,I12)') '               nB: ',nB
        write(*,'(A,I12)') '               nK: ',nK
        write(*,'(A,I12)') '               nZ: ',nZ
        write(*,'(A,I12)') '              opt: ',opt
        write(*,'(A,L12)') '        useStdOut: ',useStdOut
        write(*,'(A,I12)') '          cashopt: ',cashopt
    end if

    ! Note: Arellano baseline has nZ = 21 and nK = 200
    ! Scale the debt by steady state output
    gB(1:iBeq0) = ySS * linspace(-.35d0,0d0,iBeq0)
    gB(iBeq0:nB) = ySS * linspace(0d0,.15d0,nB-iBeq0+1)
    gK = linspace(.5d0*kSS,1.5d0*kSS,nK)

    ! +/- 3stdev corresponds to Arellano's AER grid (but she uses Tauchen and Hussey not Tauchen)
    call sub_tauchen(gZ,Fzpcz,def_ar1,def_std,nZ,3d0)
    gZ = exp(gZ)
    Fzpcz = transpose(Fzpcz)

    ! Start the timer
    call tic()

    select case (opt)
        ! Monotonicity is the first argument, concavity is the second
        case (1); call vfiDEF(vfi_iter,overtime,[0,0])
        case (4); call vfiDEF(vfi_iter,overtime,[1,0])
        case (7); call vfiDEF(vfi_iter,overtime,[2,0])
        case default; STOP 'ERROR: invalid opt'
    end select


    if (useStdOut) then
        write(*,*) 'Average count    :',avgcnt
        write(*,*) 'Minutes          :',minutes
        write(*,*) 'VFI iter cnt     :',vfi_iter
        write(*,*) 'Time integrating :',time_quad%elapsed()
        write(*,*) 'Time sup         :',time_sup%elapsed()
        write(*,*) 'Time maximizing  :',time_max%elapsed()
        write(*,*) 'Time total       :',time_tot%elapsed()

        call gnuplot(gB,q(:,nK/2,:),title='q for fixed k',saveas='pics/qfk.eps')
        call gnuplot(gB,Vnd(:,nK/2,:),title='Vnd for fixed k',saveas='pics/Vndfk.eps')
        call gnuplot(gB,cnd(:,nK/2,:),title='cnd for fixed k',saveas='pics/cndfk.eps')
        call gnuplot(gB,real(bpind(:,nK/2,:),8),title='bpind for fixed k',saveas='pics/bpindfk.eps')
        call gnuplot(gB,real(kpind(:,nK/2,:),8),title='kpind for fixed k',saveas='pics/kpindfk.eps')
        call gnuplot(gK,Vd,title='Vd',saveas='pics/Vd.eps')
        call gnuplot(gK,cd,title='cd',saveas='pics/cd.eps')

    end if


    write(*,'(4(A,A1),2(I6,A1),3(E25.17E3,A1),I2)') &
               trim(adjustl(getcmdarg_s(1))),',',&
               trim(adjustl(getcmdarg_s(2))),',',&
               trim(adjustl(getcmdarg_s(3))),',',&
               trim(adjustl(getcmdarg_s(4))),',',&
               vfi_iter,',',&
               merge(1,0,overtime),',',&
               time_max%elapsed(),',',&
               time_tot%elapsed(),',',&
               avgcnt,',',&
               cashopt

end program
