! Grey Gordon and Shi Qiu, 2016
module mod_binmono_disc_sort
    implicit none
contains
    ! Solves problems of the form 
    ! 
    ! V(i) = max_{j} u(f(i) - g(j)) + W(j)
    ! 
    ! for any f,g,W and u(c) = c**(1-crra)/(1-crra). Note that f,g,W may be arrays of any dimension. The optimal policy will be in terms of indices.
    !
    subroutine sub_binmono_disc_sort(V,c,j,f,g,W,crra,nstate,nchoice,mono,cnt,fsorted,Wsorted)
        use mod_matlab
        use mod_quicksort
        implicit none
        real(8), dimension(*), intent(out) :: V,c 
        integer, dimension(*), intent(out) :: j
        real(8), dimension(*), intent(in) :: f,g,W
        real(8), intent(in) :: crra 
        integer, intent(in) :: nstate,nchoice ! total number
        integer, intent(in) :: mono ! 0 for brute force, 1 for simple mono, 2 for binary mono
        integer, intent(out), optional :: cnt ! gives evaluation count (this is for the paper's analysis, in general, ignore).
        logical, intent(in), optional :: fsorted, Wsorted ! whether f and W, respectively, are sorted
        ! local
        real(8), dimension(:), allocatable :: Wsort(:),gsort(:),fsort(:),Vsort(:),csort(:)
        integer, dimension(:), allocatable :: IsortW(:),Isortf(:),Iunsortf(:),jsort(:)
        integer :: i,evalcnt
        logical :: dosortW, dosortf

        dosortW = .true.
        dosortf = .true.
        if (present(Wsorted)) dosortW = .not. Wsorted
        if (present(fsorted)) dosortf = .not. fsorted

        ! If using brute force, there is no need to sort. 
        if (mono==0) then
            call sub_binmono_disc(V,c,j,f,g,W,crra,nstate,nchoice,mono,evalcnt)
            if (present(cnt)) cnt = evalcnt
            return
        end if

        ! Allocate
        allocate(Wsort(nchoice))
        allocate(gsort(nchoice))
        allocate(IsortW(nchoice))
        allocate(fsort(nstate))
        allocate(Vsort(nstate))
        allocate(csort(nstate))
        allocate(jsort(nstate))
        allocate(Isortf(nstate))
        allocate(Iunsortf(nstate))

        ! Sort ascending and get unsorting method
        if (dosortW) then
            call sub_quicksort(Wsort,IsortW,W(1:nchoice))
            ! call sort2(Wsort,IsortW,W(1:nchoice))
            gsort = g(IsortW)
        else
            Wsort = W(1:nchoice)
            gsort = g(1:nchoice)
        end if
        if (dosortf) then
            call sub_quicksort(fsort,Isortf,f(1:nstate))
            ! call sort2(fsort,Isortf,f(1:nstate))
            do i = 1,nstate
                Iunsortf(Isortf(i)) = i
            end do
        else
            fsort = f(1:nstate)
        end if

        ! Pass to the usual binary mono
        call sub_binmono_disc(Vsort,csort,jsort,fsort,gsort,Wsort,crra,nstate,nchoice,mono,evalcnt)
        if (present(cnt)) cnt = evalcnt

        ! Need to unsort 
        ! Currently, policies/value are evaluated at the sorted f values as states. Undo this.
        if (dosortf) then
            j(1:nstate) = jsort(Iunsortf)
            c(1:nstate) = csort(Iunsortf)
            V(1:nstate) = Vsort(Iunsortf)
            ! For policy index, the index was the optimal index on the Wsort grid. Undo this.
            if (dosortW) then
                j(1:nstate) = IsortW(j(1:nstate))
            else
                ! Do nothing : j stores the optimum
            end if
        else
            j(1:nstate) = jsort
            c(1:nstate) = csort
            V(1:nstate) = Vsort
            ! For policy index, the index was the optimal index on the Wsort grid. Undo this.
            if (dosortW) then
                j(1:nstate) = IsortW(j(1:nstate))
            else
                ! Do nothing, j stores the optimum
            end if
        end if
        
    end subroutine

    ! Assumes sorted increasing f and W
    subroutine sub_binmono_disc(V,c,j,f,g,W,crra,nstate,nchoice,mono,cnt)
        use mod_matlab
        implicit none
        real(8), dimension(*), intent(out) :: V,c 
        integer, dimension(*), intent(out) :: j
        real(8), dimension(*), intent(in) :: f,g,W
        real(8), intent(in) :: crra 
        integer, intent(in) :: nstate,nchoice
        integer, intent(in) :: mono
        integer, intent(out), optional :: cnt
        ! local
        integer :: i,ni,nip,evalcnt
        real(8), parameter :: death = -sqrt(huge(0d0))

        evalcnt = 0

        ni = nstate
        nip = nchoice

        select case (mono)
            case (0)

                ! Brute force
                do i = 1,ni
                    call solvemax(i,1,nip)
                end do

            case (1)

                ! Simple monotonicity 
                call solvemax(1,1,nip)
                do i = 2,ni
                    call solvemax(i,j(i-1),nip)
                end do

            case (2)

                ! Binary monotonicity
                call solvemax(1,1,nip)
                call solvemax(ni,j(1),nip)
                call binary(1,ni)

        end select

        if (present(cnt)) cnt = evalcnt
        
    contains
        
        ! Binary monotonicity : solves for optimal policy between i=l and i=u assuming l and u policies are known
        recursive subroutine binary(l,u)
            implicit none
            integer, intent(in) :: l,u
            integer :: m

            ! Solve at midpoint
            m = (l+u)/2
            call solvemax(m,j(l),j(u))

            ! Divide and conquer
            if (m>l+1) call binary(l,m)
            if (u>m+1) call binary(m,u)

        end subroutine 
        
        ! Brute force maximizer searches from lb to ub
        subroutine solvemax(i,lb,ub)
            implicit none
            integer, intent(in) :: i,lb,ub
            ! local
            integer :: ip
            real(8) :: vtmp,ctmp

            evalcnt = evalcnt + ub - lb + 1

            do ip = lb,ub
                ctmp = f(i) - g(ip)
                vtmp = u(ctmp) + W(ip)
                if (ip==lb .or. vtmp>=V(i)) then
                    j(i) = ip
                    c(i) = ctmp
                    V(i) = vtmp
                end if
            end do

        end subroutine

        real(8) function u(c)
            implicit none
            real(8), intent(in) :: c

            if (c<=0d0) then
                u = death
                return
            end if
            if (crra==2d0) then
                u = -1d0/c
            elseif (crra==1d0) then
                u = log(c)
            else
                u = c**(1d0-crra)/(1d0-crra)
            end if

        end function 
        
    end subroutine

    ! Solves problems of the form 
    ! 
    ! V(i) = max_{j1,j2} u(f(i) - g(j1,j2)) + W(j1,j2)
    ! 
    ! for any g, with f increasing, and W increasing along at least one of its 2 dimensions with u(c) = c**(1-crra)/(1-crra). 
    ! 
    ! The user specifies whether W is increasing along its first dim (default) or second. If both, select one. 
    ! 
    ! The goal here is to avoid potentially costly sorting of W. If this is low cost, sub_binmono_disc_sort may be used instead.
    !
    subroutine sub_binmono_disc_ascend(V,c,j1,j2,f,g,W,crra,nstate,nchoice,mono,cnt,fsorted)
        use mod_matlab
        use mod_quicksort
        implicit none
        real(8), dimension(*), intent(out) :: V,c 
        integer, dimension(*), intent(out) :: j1,j2
        real(8), dimension(*), intent(in) :: f
        real(8), intent(in) :: crra 
        integer, intent(in) :: nstate,nchoice(:) ! total number
        real(8), dimension(nchoice(1),*), intent(in) :: W,g ! should have shape == nchoice
        integer, intent(in) :: mono ! 0 for brute force, 1 for simple mono, 2 for binary mono
        integer, intent(out), optional :: cnt ! gives evaluation count (this is for the paper's analysis, in general, ignore).
        logical, optional, intent(in) :: fsorted
        ! local
        real(8), allocatable :: Vtmp(:),ctmp(:)
        integer, allocatable :: jtmp(:)
        integer :: dim2,evalcnt,tmpcnt,istate
        real(8), dimension(:), allocatable :: fsort(:),Vsort(:),csort(:)
        integer, dimension(:), allocatable :: Isortf(:),Iunsortf(:),j1sort(:),j2sort(:)
        logical :: dosortf

        ! If using brute force, there is no need to sort. 
        if (mono==0) then
            allocate(jtmp(nstate))
            call sub_binmono_disc(V,c,jtmp,f,g,W,crra,nstate,product(nchoice),mono,evalcnt)
            j1(1:nstate) = 1+mod(jtmp-1,nchoice(1))
            j2(1:nstate) = 1+(jtmp-1)/nchoice(1)
            if (present(cnt)) cnt = evalcnt
            return
        end if


        dosortf = .true.
        if (present(fsorted)) dosortf = .not. fsorted

        ! In case need to sort f, create arrays to hold the values etc. on the sorted grid
        allocate(fsort(nstate))
        allocate(Vsort(nstate))
        allocate(csort(nstate))
        allocate(j1sort(nstate))
        allocate(j2sort(nstate))
        allocate(Isortf(nstate))
        allocate(Iunsortf(nstate))
        if (dosortf) then
            call sub_quicksort(fsort,Isortf,f(1:nstate))
            ! call sort2(fsort,Isortf,f(1:nstate))
            do istate = 1,nstate
                Iunsortf(Isortf(istate)) = istate
            end do
        else
            fsort = f(1:nstate)
        end if

        ! Pass to the usual binary mono
        allocate(Vtmp(nstate))
        allocate(ctmp(nstate))
        allocate(jtmp(nstate))
        evalcnt = 0
        do dim2 = 1,nchoice(2)
            call sub_binmono_disc(Vtmp,ctmp,jtmp,fsort,g(:,dim2),W(:,dim2),crra,nstate,nchoice(1),mono,tmpcnt) ! solve for the monotone policy
            evalcnt = evalcnt + tmpcnt
            if (dim2==1) then
                Vsort = Vtmp
                csort = ctmp
                j1sort = jtmp
                j2sort = dim2
            else
                do istate = 1,nstate
                    if (Vtmp(istate)>Vsort(istate)) then
                        Vsort(istate) = Vtmp(istate)
                        csort(istate) = ctmp(istate)
                        j1sort(istate) = jtmp(istate)
                        j2sort(istate) = dim2
                    end if
                end do
            end if
        end do

        ! Unsort the states
        if (dosortf) then
            V(1:nstate) = Vsort(Iunsortf)
            c(1:nstate) = csort(Iunsortf)
            j1(1:nstate) = j1sort(Iunsortf)
            j2(1:nstate) = j2sort(Iunsortf)
        else
            V(1:nstate) = Vsort
            c(1:nstate) = csort
            j1(1:nstate) = j1sort
            j2(1:nstate) = j2sort
        end if

        if (present(cnt)) cnt = evalcnt
        
    end subroutine

end module
