module mod_sort
    implicit none
    integer, parameter, private :: qsort_nthresh = 100 ! When a list is <= this threshold, insertion sort is called. MUST be >= 3

    ! I found using ifort that a high number like 60 or 100 was close to optimal.
    ! I usually have nearly sorted data, it's better to err on the high side.
    ! I also found middle pivot was better slightly, but will keep median as it is more robust.
    
    interface msort
        module procedure msort_i,msort_r
    end interface msort
    interface qsort
        module procedure qsort_i,qsort_r
    end interface qsort
    interface isort
        module procedure isort_i,isort_r
    end interface isort

    ! qsort is based on algorithms I found in several places online, most citing textbooks
    ! isort is based on wiki I believe
    ! msort I think I wrote completely on my own (which is why it's so bad!)

contains

    ! tests whether x is sorted according to optional argument mode ("ascend" or "descend").
    ! By default, checks for ascended sort.
    function issorted(x,mode) result(sorted)
        implicit none
        logical :: sorted
        real(8), dimension(1:) :: x
        character(len=*), intent(in) :: mode    !either 'ascend' or 'descend'
        !local
        integer :: i

        if (size(x)==1) then
            sorted = .TRUE. 
            return
        end if

        if (mode=='ascend') then
            do i = 2,size(x)
                if (x(i)<x(i-1)) then
                    sorted = .FALSE.
                    return
                end if
            end do
        elseif (mode=='descend') then
            do i = 2,size(x)
                if (x(i)>x(i-1)) then
                    sorted = .FALSE.
                    return
                end if
            end do
        else
            write(*,*) 'issorted: error: sorting mode not found'
            STOP 'issorted: error: sorting mode not found'
        end if

        sorted = .TRUE.

    end function issorted

    ! Insertion sort
    pure subroutine isort_i(X,I)
        implicit none
        integer, dimension(1:), intent(inout) :: X
        integer, dimension(1:size(X)), intent(out) :: I
        !local
        real(8), dimension(1:size(X)) :: Xreal

        Xreal = real(X,8) ! Convert to double prec
        call isort_r(Xreal,I) ! Sort the double prec
        X = nint(Xreal) ! Nearest integer to Xreal

    end subroutine isort_i 
    pure subroutine isort_r(X,I) 
        implicit none
        real(8), dimension(1:), intent(inout) :: X ! Input/Output array
        integer, dimension(1:size(X)), intent(out) :: I ! I s.t. X(I) is sorted
        !local
        integer :: j

        do j = 1,size(X)
            I(j) = j
        end do

        call isort_r_core(X,I)

    end subroutine isort_r
    pure subroutine isort_r_core(X,I) 
        implicit none
        real(8), dimension(1:), intent(inout) :: X ! Input/Output array
        integer, dimension(1:), intent(inout) :: I ! I s.t. X(I) is sorted
        !local
        real(8) :: val
        integer :: k,j,n,ival
        
        n = size(x)

        ! Loop over items to be sorted 
        do k = 2,n
            ! 1 to k-1 is sorted. Now place X(i) in the sorted list, shifting
            ! the sorted elements to the right as necessary.
            val = X(k)
            ival = I(k)
            ! Going backwards from j=k-1 to 1, if X(k)<X(j), then we know X(k) is to the 
            ! left of X(j) somewhere. Consequently, X(j) should be shifted to the right.
            ! If not, then X(k)>=X(j), and so X(k) should be put in the j+1th position
            do j = k-1,1,-1
                if (val<X(j)) then
                    X(j+1) = X(j) ! Shift forward
                    I(j+1) = I(j) ! Ditto
                else
                    ! Exit here
                    ! When we exit, j retains it's last value
                    exit
                end if
            end do
            ! The j value here is either what it was when "exit" was executed or 0 if the loop completed
            ! Consequently, we always want to replace the j+1 th value.
            X(j+1) = val ! Put
            I(j+1) = ival
        end do

    end subroutine isort_r_core

    ! Quick sort
    ! Sorts X in place and also returns I such that X=X(I) with RHS X input.
    subroutine qsort_i(X,I)
        implicit none
        integer, dimension(1:), intent(inout) :: X
        integer, dimension(1:size(X)), intent(out) :: I
        !local
        real(8), dimension(1:size(X)) :: Xreal
print*,'WARNING: I AM NOT SURE WHETHER THIS VERSION OF QUICKSORT IS STABLE!'

        Xreal = real(X,8) ! Convert to double prec
        call qsort_r(Xreal,I) ! Sort the double prec
        X = nint(Xreal) ! Nearest integer to Xreal
        
    end subroutine qsort_i
    subroutine qsort_r(X,I)
        implicit none
        real(8), dimension(1:), intent(inout) :: X ! Input/Output array
        integer, dimension(1:size(X)), intent(out) :: I ! I s.t. X(I) is sorted
        ! local
        integer :: l,n ! # of elts
print*,'WARNING: I AM NOT SURE WHETHER THIS VERSION OF QUICKSORT IS STABLE!'
        
        n = size(X)
        
        ! Test whether should go directly to insertion sort
        if (n<=qsort_nthresh) then
            call isort_r(X,I)
            return
        end if 

        ! Set the index array
        do l = 1,n
            I(l) = l
        end do

        ! Call the recursion 
        call qsort_r_core(X,I,1,n)

        ! I have seen it suggested that one should not do the many isorts in the core, 
        ! but one large one at the end since isort works well on almost sorted data.
        ! I do not pursue that here.
        ! call isort_core(I,X)

    end subroutine qsort_r

    ! The core routine for qsort
    pure recursive subroutine qsort_r_core(X,I,left,right)
        implicit none
        real(8), dimension(1:), intent(inout) :: X ! Input/Output array
        integer, dimension(1:), intent(inout) :: I ! I s.t. X(I) is sorted
        integer, intent(in) :: left,right
        !local
        integer :: p,l

        ! If size is small, use isort
        if (right-left+1<=qsort_nthresh) then
            call isort_r_core(X(left:right),I(left:right))
            return
        end if

        ! Middle
!         p = (left + right)/2

        ! Median of three
        l = (left + right)/2 ! l used for temporary storage
        if (X(l)>=X(left)) then
            ! l >= left
            if (X(l)<=X(right)) then
                ! l <= right
                p = l
            else
                ! l>=left and l>right
                if (X(left)<=X(right)) then
                    ! left<=right
                    p = right
                else
                    p = left
                end if
            end if
        else
            ! l < left
            if (X(l)>=X(right)) then
                ! l >= right
                p = l
            else
                ! l <= right
                if (X(left)<=X(right)) then
                    ! left<=right and l<left
                    p = left
                else
                    ! left > right and l<=right
                    p = right
                end if
            end if
        end if

        ! Partition the list
        call qsort_r_partition(l,X,I,left,right,p)

        ! Now, recursively sort the sublists.
        if (left<l-1) call qsort_r_core(X,I,left,l-1)
        if (l<right) call qsort_r_core(X,I,l,right)
        
    end subroutine qsort_r_core

    ! Given the array X and bounds left,right, and a pivot p, 
    ! rearranges X(left:right) so that X(left:l)<=X(p), X(l+1:right)>=X(p) 
    ! where l is integer output. 
    pure subroutine qsort_r_partition(l,X,I,left,right,p)
        implicit none
        real(8), dimension(1:), intent(inout) :: X
        integer, dimension(1:), intent(inout) :: I
        integer, intent(in) :: left,right,p
        integer, intent(out) :: l
        ! local
        real(8) :: pval, tmp
        integer :: r, tmpi

        ! Get the pivot value
        pval = X(p)

        ! 
        l = left
        r = right
        do while (l<=r) 

            ! Find the first l greater or equal to the current l such that X(l)>=pval is found
            do while(X(l)<pval) ! Exits when X(l)>=pval. Since pval lies in left,right, this is not an infinite loop
                l = l+1
            end do
            ! Find the first r less than or equal to the current r such that X(r)<=pval is found
            do while (X(r)>pval) ! Exits when X(r)<=pval
                r = r-1
            end do

            ! Here, X(l)>=pval and X(r)<=pval and l<r, so swap these elements to guarantee
            if (l<=r) then
                tmp = X(l)
                X(l) = X(r)
                X(r) = tmp
                tmpi = I(l)
                I(l) = I(r)
                I(r) = tmpi
                ! At this point, left:l satisfies X(left:l)<=pval and X(r:right)>=pval. 

                ! Increment l and r
                l = l+1
                r = r-1
            end if

        end do
        
        ! Here, since l>r (we exited the loop) and X(left:l-1)<=pval and X(l:right)>=pval, we 
        ! have partitioned X 
        
    end subroutine qsort_r_partition 



    ! Performs mergesort for real numbers
    subroutine msort_r(A,I)
        use mod_orderpack
        implicit none
        real(8), dimension(1:), intent(inout) :: A !list to be sorted
        integer, dimension(1:size(A)), intent(out) :: I !index of sorting s.t. to sort B according to A use B=B(I)

        ! Call the ORDERPACK 2.0 routine for a stable mergesort
        call mrgrnk(A,I)
        A = A(I)

    end subroutine msort_r

!     !Returns sorted vector A and indices I sorted according to A
!     ! I am embarassed by this routine. I should go back and fix it to use
!     ! pointers rather than straight up copying a million times with automatic
!     ! local arrays.
!     recursive subroutine msort_recursion_r(A,I)
!         implicit none
!         real(8), dimension(1:), intent(inout) :: A
!         integer, dimension(1:), intent(inout) :: I
!         !local
!         real(8), dimension(1:size(A)) :: Acopy
!         integer, dimension(1:size(I)) :: Icopy
!         integer :: na,j,a1,a2
!         na = size(a)
! 
!         !If na is 0 or 1, the list is sorted, so return
!         if (na<=1) then        
!             return
!         end if
! 
!         !If na>1 then divide the list in half
!         call msort_recursion_r(A(1:na/2),I(1:na/2))
!         call msort_recursion_r(A(na/2+1:na),I(na/2+1:na))
! 
!         !Merge these two sorted lists
!         a1 = 1
!         a2 = na/2+1
!         do j=1,na      
!             if (a1<=na/2 .AND. a2<=na) then
!                 if (A(a1)<=A(a2)) then
!                     Acopy(j) = A(a1)
!                     Icopy(j) = I(a1)
!                     a1 = a1+1
!                 else
!                     Acopy(j) = A(a2)
!                     Icopy(j) = I(a2)
!                     a2 = a2+1
!                 end if
!                 !if (j==na) exit
!             elseif (a1>na/2) then !first list has been exhausted, fill remaining elements with 2nd list
!                 Acopy(j:na) = A(a2:na)
!                 Icopy(j:na) = I(a2:na)
!                 exit
!             elseif (a2>na) then !second list has been exhausted, fill remaining elements with 1st list
!                 Acopy(j:na) = A(a1:na/2)
!                 Icopy(j:na) = I(a1:na/2)
!                 exit
!             end if
!         end do
! 
!         A = Acopy
!         I = Icopy
! 
!     end subroutine msort_recursion_r

    subroutine msort_i(A,I)
        use mod_orderpack
        implicit none
        integer, dimension(1:), intent(inout) :: A !list to be sorted
        integer, dimension(1:size(A)), intent(out) :: I !index of sorting s.t. to sort B according to A use B=B(I)

        ! Call the ORDERPACK 2.0 routine for a stable mergesort
        call mrgrnk(A,I)
        A = A(I)

    end subroutine msort_i

!     !Returns sorted vector A and indices I sorted according to A
!     recursive subroutine msort_recursion_i(A,I)
!         implicit none
!         integer, dimension(1:), intent(inout) :: A
!         integer, dimension(1:), intent(inout) :: I
!         !local
!         integer, dimension(1:size(A)) :: Acopy
!         integer, dimension(1:size(I)) :: Icopy
!         integer :: na,j,a1,a2
!         na = size(a)
! 
!         !If na is 0 or 1, the list is sorted, so return
!         if (na<=1) then        
!             return
!         end if
! 
!         !If na>1 then divide the list in half
!         call msort_recursion_i(A(1:na/2),I(1:na/2))
!         call msort_recursion_i(A(na/2+1:na),I(na/2+1:na))
! 
!         !Merge these two sorted lists
!         a1 = 1
!         a2 = na/2+1
!         do j=1,na      
!             if (a1<=na/2 .AND. a2<=na) then
!                 if (A(a1)<=A(a2)) then
!                     Acopy(j) = A(a1)
!                     Icopy(j) = I(a1)
!                     a1 = a1+1
!                 else
!                     Acopy(j) = A(a2)
!                     Icopy(j) = I(a2)
!                     a2 = a2+1
!                 end if
!                 !if (j==na) exit
!             elseif (a1>na/2) then !first list has been exhausted, fill remaining elements with 2nd list
!                 Acopy(j:na) = A(a2:na)
!                 Icopy(j:na) = I(a2:na)
!                 exit
!             elseif (a2>na) then !second list has been exhausted, fill remaining elements with 1st list
!                 Acopy(j:na) = A(a1:na/2)
!                 Icopy(j:na) = I(a1:na/2)
!                 exit
!             end if
!         end do
! 
!         A = Acopy
!         I = Icopy
! 
!     end subroutine msort_recursion_i

    subroutine test_sortmethod(sortMethod,n)
        implicit none
        interface 
            subroutine sortMethod(X,I)
                implicit none
                real(8), dimension(1:), intent(inout) :: X
                integer, dimension(1:size(X)), intent(out) :: I
            end subroutine sortMethod
        end interface
        integer, intent(in) :: n
        real(8), allocatable :: X(:),X0(:)
        integer, allocatable :: I(:) ! Output sorted ints
        logical :: passed
        real :: ta, tb
        ! Test data

        allocate(X(n),X0(n),I(n))

        call cpu_time(ta)

        ! Random input
        call random_number(X)
        X0 = X; call sortMethod(X,I); passed = issorted(X,'ascend') .and. all(X0(I)==X)
        if (.not. passed) STOP 'failed random data' 

        ! Sorted data
        X0 = X; call sortMethod(X,I); passed = issorted(X,'ascend') .and. all(X0(I)==X)
        if (.not. passed) STOP 'failed sorted data' 

        ! Descendingly sorted data
        X = -X;
        X0 = X; call sortMethod(X,I); passed = issorted(X,'ascend') .and. all(X0(I)==X)
        if (.not. passed) STOP 'failed descendingly sorted data' 

        ! Random, but not distinct
        ! gcc4.9.0 has a internal compiler error here, so I disabled this test
!         call random_number(X)
!         X = floor(X*1000d0) 
!         X0 = X; call sortMethod(X,I); passed = issorted(X,'ascend') .and. all(X0(I)==X)
!         if (.not. passed) STOP 'failed random but not distinct pts' 

        call cpu_time(tb)

        print*,'sorting time',tb-ta

        deallocate(X,X0,I)

    end subroutine test_sortmethod
    subroutine test_sort
        implicit none
        call test_sortmethod(isort_r,10); print*,'test_sort: isort_r passed'
        call test_sortmethod(qsort_r,1000); print*,'test_sort: qsort_r passed'
        call test_sortmethod(msort_r,1000); print*,'test_sort: msort_r passed'
    end subroutine test_sort

end module mod_sort
